/* -*- C++ -*- */

/* -*- C++ -*- */

/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Crystal animation and 3D modeling application
 *
 * Authors:   Tommi Ilmonen 2001-2004
 * Copyright: Tommi Ilmonen and/or Tekes 2001-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <crystal_blow_recorder.h>

#include <crystal_action_list.h>
#include <crystal_audio_action.h>
#include <crystal_dom_util.h>
#include <crystal_engine.h>
#include <crystal_force_editor.h>
#include <crystal_gl_data.h>
#include <crystal_engine.h>
#include <crystal_location_action.h>
#include <crystal_remove_action.h>
#include <crystal_tools.h>
#include <crystal_value_action.h>

#include <vee_error.h>
#include <vee_strs.h>

#include <Fluid/Fistiness.hpp>
#include <Fluid/TrackerSensorSelector.hpp>

namespace Crystal { 

  static bool debugBlow = true;

  CreateBlowAction::~CreateBlowAction()
  {}
  
  void CreateBlowAction::act(ActionList *al)
  {
    if(debugBlow)
      VEE_trace("CreateBlowAction::act # %d %d", 
		m_id, al->getScene()->totalForces());

    VEE_Blow *blow = new VEE_Blow();

    blow->setAgeLimit(1e+10);
    blow->setLocation(m_location);
    blow->setRange(5.0);
    blow->setMagnitude(5.0);

    al->setObject(m_id, blow);
    al->getScene()->addNewForce(blow);
    
    if(debugBlow)
      VEE_trace("CreateBlowAction::act # DONE %d %d", 
		m_id, al->getScene()->totalForces());
  }

  QDomElement CreateBlowAction::writeDom(QDomDocument &doc)
  {
    QDomElement elem = doc.createElement("CreateBlowAction");

    writeDomTimeStamp(elem, doc);

    addTextChild(elem, doc, "location", m_location);
    addTextChild(elem, doc, "id", m_id);

    return elem;
  }

  bool CreateBlowAction::readDom(QDomElement &elem) 
  { 
    if(elem.tagName() != "CreateBlowAction")
      return false;

    readDomTimeStamp(elem);

    m_location = getVector3(elem, "location");
    m_id       = getInt(elem, "id", -1);

    return true;
  }


  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  BlowRecorder::BlowRecorder(bool motion)
    : m_motion(motion),
      m_blowId(-1),
      m_trackerCount(0),
      m_range(0),
      m_magnitude(0)
  {
    m_location.clear();
    m_direction.clear();
  }

  BlowRecorder::~BlowRecorder()
  {}

  void BlowRecorder::prepare(ActionList *al)
  {
    Engine *engine = al->engine();
    m_trackerCount = (long) 
      engine->weakHandLoc()->locations().getSampleCount();
  }

  void BlowRecorder::beginRecording(ActionList *al, 
				    const TimeStamp &now)
  {
    const char * fname = "BlowRecorder::beginRecording";

    VEE_trace("%s", fname);

    Engine *engine = al->engine();
    m_trackerCount = (long) engine->tracker()->outputBuffer().getSampleCount();

    VEE_trace("%s # %d", fname, 1);  

    IONodeBuffer<Vector3> * handLoc = 
      & engine->weakHandLoc()->locations();
    
    Vector3 location = handLoc->getIndex(m_trackerCount);
  
    m_blowId = al->allocateObjectId();

    if(debugBlow)
      VEE_trace("%s # blowId = %d", fname, m_blowId);  

    al->appendAction(new CreateBlowAction(m_blowId, location, now));
    
    m_audioSourceId = al->allocateAudioId();
    AudioActionNoise * noteon = 
      new AudioActionNoise(m_audioSourceId, 0.5, 200, 0.5, now);
  
    al->appendAction(noteon);
    m_noteOnActionId = al->addMappedAction(noteon);
  
    m_audioLocator = 
      new AudioActionSourceLocation2(m_audioSourceId,  
				     m_blowId, now); 
    // Make sure the action is not wiped out accidentally.
    TimeStamp tmpts;
    tmpts.setSeconds(1.0e+8);
    m_audioLocator->setEndTimeStamp(tmpts);

    al->appendAction(m_audioLocator);

    m_rateControl = 
      new AudioActionControl(m_noteOnActionId, 
			     m_blowId, 
			     "magnitude",
			     "pitch",
			     now,
			     new MJ_MapperMultiplyLimit(200.0, 50.0, 10000.0));
			
    m_rateControl->setEndTimeStamp(tmpts);

    al->appendAction(m_rateControl);  		   

    m_mode = RECORD;

    if(debugBlow)
      VEE_trace("BlowRecorder::beginRecording # EXIT");    
  }

  void BlowRecorder::endRecording(ActionList *al, 
				  const TimeStamp &now)
  {
    VEE_trace("BlowRecorder::endRecording");

    TimeStamp ts = now;

    if(!m_motion) {
      TimeStamp tmp;
      tmp.setSeconds(1.0e+5);
      ts += tmp;
    }

    al->appendAction(new RemoveAction(m_blowId, ts));
    al->appendAction(new AudioActionNoteoff(m_noteOnActionId, ts));

    m_audioLocator->setEndTimeStamp(now - TimeStamp(1));
    m_rateControl->setEndTimeStamp(now - TimeStamp(1));

    m_mode = MUTE;
  }

  static std::string locStr("location");
  static std::string directionStr("direction");
  static std::string rangeStr("range");
  static std::string magnitudeStr("magnitude");

  void BlowRecorder::record(ActionList *al, 
			    const TimeStamp &now)
  {  
    const char * fname = "BlowRecorder::record";

    if(debugBlow)
      VEE_trace(fname);

    Engine *engine = al->engine();

    IONodeBuffer<Vector3> * handLoc = 
      & engine->weakHandLoc()->locations();

    IONodeBuffer<Vector3> * handLoc2 = 
      & engine->strongHandLoc()->locations();

    ValueGraphs * graphs = & al->getGraph(m_blowId);
  
    for(;m_trackerCount < (long) handLoc->getSampleCount(); m_trackerCount++) {
    
      m_location  = handLoc->getIndex(m_trackerCount);
      Vector3 location2 = handLoc2->getIndex(m_trackerCount);

      m_direction = m_location - location2;
      m_range = m_direction.length() * 5;

      if(m_range < 1e-5) {
	m_direction[0] = 0.01;
	m_range = 0.01;
      }

      if(m_mode == RECORD) {

	TimeStamp sampleTime = handLoc->getTimeStamp(m_trackerCount);
	TimeStamp ts = al->realTimeToRecordTime(sampleTime);
	double dts = ts.secondsD();
	
	graphs->appendVector3(locStr, dts, location2);
	graphs->appendVector3(directionStr, dts, m_direction);
	
	graphs->appendFloat(rangeStr, dts, m_range);
	// graphs->appendFloat(magnitudeStr, dts, range * 0.7);
      }
    }

    Fistiness  *fistiness = engine->fistiness();

    IONodeBuffer<float> * values = & fistiness->buffer();

    float magnitude = engine->tools()->forceEditor()->current()->magnitude();
    float sensitivity = 
      engine->tools()->forceEditor()->current()->sensitivity();

    float fist = 1.0 - values->getNewest();
    m_magnitude = fist * sensitivity + magnitude;

    if(m_mode == RECORD) {
      // RRR ("+ 2.0")
      graphs->appendFloat(magnitudeStr, now.secondsD(), m_magnitude);
    }
    
    if(debugBlow)
      VEE_trace("%s # EXIT", fname);
  }

  void BlowRecorder::render(Engine *e, GlData *gld)
  {
    if(m_mode == RECORD)
      return;

    VEE_Blow b;

    b.setLocation(m_location);
    b.setDirection(m_direction);
    b.setRange(m_range);
    b.setMagnitude(m_magnitude);
    b.setAge(e->timing().currentTimeToRecordTime().secondsF());

    gld->m_veeRenderer.initRenderState();
    gld->m_veeRenderer.setForceFrontColor(& VEE_Rgba::Green);
    b.render(& gld->m_veeRenderer);
    gld->m_veeRenderer.setForceFrontColor(0);
    gld->m_veeRenderer.closeRenderState();
  }

}
