/* -*- C++ -*- */

/* -*- C++ -*- */

/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Crystal animation and 3D modeling application
 *
 * Authors:   Tommi Ilmonen 2001-2004
 * Copyright: Tommi Ilmonen and/or Tekes 2001-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <crystal_component_editor.h>

#include <crystal_action_list.h>
#include <crystal_dom_util.h>
#include <crystal_engine.h>
#include <crystal_gl_data.h>
#include <crystal_tools.h>
#include <crystal_visualize.h>

#include <vee_error.h>
#include <vee_math.h>

#include <Fluid/MessageFloat.hpp>

#include <GL/gl.h>


namespace Crystal { 

  ComponentEditor::ComponentData::ComponentData() 
    : m_plane(0),
      m_list(0), 
      m_listId(0), 
      m_enabled(true), 
      m_hidden(false), 
      m_visualize(false),
      m_host(0)
  {}

  ComponentEditor::ComponentData::~ComponentData()
  {}

  bool ComponentEditor::ComponentData::msgProcess
  (const char *id, Fluid::Message *)
  {
    bool ok = false;
    
    if(strcmp(id, "graphs") == 0) {

      VEE_trace("Toggle graphs %d", m_listId);

      toggleGraphs();
      ok = true;
    }

    return ok;
  }

  void ComponentEditor::ComponentData::toggle(ComponentEditor *e)
  {
    assert(m_list && m_button.ptr());

    m_enabled = !m_enabled;

    m_list->setMode(m_enabled ? 
		    ActionList::PLAY :  ActionList::MUTE);

    ColorGroup cg;

    if(!m_enabled)
      cg = e->disableColors();

    m_button.ptr()->setColors(cg);
  }

  void ComponentEditor::ComponentData::setEnabled
  (ComponentEditor *e, bool enable)
  {
    assert(m_list && m_button.ptr());

    if(m_hidden) return;

    m_enabled = enable;

    m_list->setMode(m_enabled ? 
		    ActionList::PLAY :  ActionList::MUTE);

    ColorGroup cg;

    if(!m_enabled)
      cg = e->disableColors();

    m_button.ptr()->setColors(cg);
  }

  void ComponentEditor::ComponentData::setHidden()
  {
    assert(m_list && m_button.ptr());

    m_enabled = false;
    m_hidden = true;
    m_list->setMode(ActionList::MUTE);
    m_button.ptr()->setState(Widget::HIDDEN);
  }

  void ComponentEditor::ComponentData::startEdit(ComponentEditor *e)
  {
    UiHandler * uih = dynamic_cast<UiHandler *> (m_list->visualization());
    if(!uih) return;
    
    m_visualize = true;

    e->host()->setUiHandler(uih);
    uih->start();
  }

  void ComponentEditor::ComponentData::build(ComponentEditor *e)
  {
    Menu * txb = new Menu(e->host());
    txb->setSounds(e->host()->buttonSounds());
  
    txb->setLabel(m_label);
  
    char buf[64];

    sprintf(buf, "visualize_%d", m_listId);
    txb->addItem("Show",  buf, e);

    sprintf(buf, "move_%d", m_listId);
    txb->addItem("Move",  buf, e);

    sprintf(buf, "edit_%d", m_listId);
    txb->addItem("Edit",  buf, e);

    m_graphId = txb->addItem("Graphs",  "graphs", this);

    sprintf(buf, "enable_%d", m_listId);
    txb->addItem("Enable",  buf, e);

    sprintf(buf, "disable_%d", m_listId);
    txb->addItem("Disable", buf, e);

    sprintf(buf, "delete_%d", m_listId);
    txb->addItem("Delete",  buf, e);

    m_button = txb;

    m_list = e->host()->engine()->actionLists()->getList(m_listId);

    m_button.ptr()->setChecked(0, m_visualize);

    if(!m_list)
      VEE_fatal("ComponentEditor::ComponentData::build # null list %d ",
		m_listId);

    e->addWidget(txb);
  }

  void ComponentEditor::ComponentData::toggleGraphs()
  {
    if(m_graphs.ptr()) {
      m_button.ptr()->setChecked(m_graphId, false);
      m_graphs.clear();
      return;
    }

    m_button.ptr()->setChecked(m_graphId, true);

    m_graphs = new GraphEditor;

    for(ActionList::graph_iterator it = m_list->graphs().begin();
	it != m_list->graphs().end(); it++) {
      
      ValueGraphs * vg = & (*it).second;
      
      for(ValueGraphs::iterator vit = vg->graphs().begin();
	  vit != vg->graphs().end(); vit++)
	m_graphs.ptr()->addGraph((*vit).first.c_str(), * (*vit).second);
    }
    
    m_host->addWidget(m_graphs.ptr());
    
    Vector3 low  = m_button.ptr()->boundingBox().low();
    Vector3 high = m_button.ptr()->boundingBox().high();

    low.y   = high.y;
    high.y += 2.0;

    m_graphs.ptr()->setTimePerUnit(m_host->secsPerUnit());
    m_graphs.ptr()->setVertices(low.vector2(), high.vector2());
  }

  QDomElement ComponentEditor::ComponentData::writeDom(QDomDocument &doc)
  {
    QDomElement elem = doc.createElement("ComponentData");
  
    addTextChild(elem, doc, "list_id", m_listId);
    addTextChild(elem, doc, "enabled", (int) m_enabled);
    addTextChild(elem, doc, "hidden",  (int) m_hidden);
    addTextChild(elem, doc, "visualize",  (int) m_visualize);
    addTextChild(elem, doc, "label",   m_button.ptr()->label().c_str());

    return elem;  
  }

  bool ComponentEditor::ComponentData::readDom(QDomElement &elem)
  {
    if(elem.tagName() != "ComponentData")
      return false;

    VEE_trace("ComponentEditor::ComponentData::readDom");
 
    m_list = 0;
  
    m_listId    = getInt(elem, "list_id", -1);
    m_enabled   = getInt(elem, "enabled", 0);
    m_hidden    = getInt(elem, "hidden", 0);
    m_visualize = getInt(elem, "visualize", 0);
    m_label     = getString(elem, "label").ascii();
 
    return true;
  }


  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  void ComponentEditor::Track::build(ComponentEditor *e)
  {
    for(uint i=0; i < m_components.size(); i++)
      m_components[i].ptr()->build(e);

    Menu * menu = new Menu(e->host());
    menu->setState(Widget::INACTIVE);
    menu->setSounds(e->host()->buttonSounds());
    menu->setLabel(m_label);

    m_menu = menu;

    e->fixRowMenu(menu, m_id, m_label);

    e->addWidget(menu);
  }

  int ComponentEditor::Track::findComponent(int id)
  { 
    for(uint i=0; i < m_components.size(); i++)
      if(m_components[i].ptr()->m_listId == id)
	return i;

    return -1;
  }

  bool ComponentEditor::Track::spaceIsOccupied
  (float left, float right, int plane, int until)
  {
    for(uint i=0; i < m_components.size() && (int) i < until; i++) {

      ComponentData * cd = m_components[i].ptr();
      
      float cleft  = cd->m_button.ptr()->low().x;
      float cright = cd->m_button.ptr()->high().x;
      int   cplane = cd->m_plane;

      if(plane == cplane &&
	 ((left  < cright && left  > cleft) ||
	  (right < cright && right > cleft) ||
	  (left < cleft && right > cright)))
	return true;
    }

    return false;
  }

  QDomElement ComponentEditor::Track::writeDom(QDomDocument &doc)
  {
    QDomElement elem = doc.createElement("Track");

    if(m_components.size()) {

      QDomElement items = doc.createElement("components");
  
      QString tmp;

      for(uint i=0; i < m_components.size(); i++) {
	tmp.setNum(m_components[i].ptr()->m_listId);
	addTextChild(items, doc, "component", tmp);
      }
      
      elem.appendChild(items);
    }

    addTextChild(elem, doc, "id", m_id);
    addTextChild(elem, doc, "label", m_label.c_str());
  
    return elem;  
  }

  bool ComponentEditor::Track::readDom(QDomElement &elem, 
				       ComponentEditor *editor)
  {
    if(elem.tagName() != "Track")
      return false;

    VEE_trace("ComponentEditor::Track::readDom");
 
    m_components.clear();

    VEE_trace("ComponentEditor::Track::readDom # cleared");
 
    QDomElement components = elem.namedItem("components").toElement();

    if(!components.isNull()) {
      QDomNode n = components.firstChild();
      while ( !n.isNull() ) {
	if ( n.isElement() ) {
	  QDomElement e = n.toElement();
	  bool ok = true;
	  int id = e.firstChild().toText().data().toInt(&ok);

	  if(ok)
	    m_components.push_back(editor->getComponent(id));
	
	}
	n = n.nextSibling();
      }
    }

    m_id     = getInt(elem, "id", -1);
    m_label  = getString(elem, "label").ascii();
 
    VEE_trace("ComponentEditor::Track::readDom # EXIT % u",
	      m_components.size());

    return true;
  }

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  ComponentEditor::ComponentEditor(Tools *t)
    : m_host(t),
      m_recordTrack(0),
      m_newRecords(false),
      m_topLine(2.5),
      m_trackHeight(0.5),
      m_left(-2.0),
      m_right(0.0),
      m_secsPerUnit(1.0),
      m_moveComponent(0)
  {
    const char * fname = "ComponentEditor::ComponentEditor";

    m_pointerLocation.clear();
    m_widgets.setGrabGroup(t->grabs());

    VEE_Rgba inactive(0.7, 0.0, 0.0, 1.0); // Medium red
    VEE_Rgba active  (1.0, 0.6, 0.0, 1.0); // Red / orange
    VEE_Rgba grab    (1.0, 1.0, 0.5, 1.0); // Bright orange

    m_disableColors.setColor(ColorGroup::INACTIVE, inactive);
    m_disableColors.setColor(ColorGroup::ACTIVE,   active);
    m_disableColors.setColor(ColorGroup::GRAB,     grab);

    m_moveColors.constantColor(VEE_Rgba::Green);

    Slider * slider = 
      new Slider(m_host, 0.0, 0.0, 10.0);
  
    slider->setSounds(m_host->buttonSounds());
    slider->setTexture("");

    slider->msgAddSink("", "offset", this);
  
    float sliderH = 0.2;
    slider->setLocation(VEE_Vector3(m_left - sliderH * 0.5f, m_topLine, 0.0f));
    slider->setSize(10 + sliderH, sliderH);
    slider->setOrientation(Slider::HORIZONTAL);
    slider->setDisplayValue(false);
    slider->setBorder(0.0);
  
    m_offset = slider;

    m_widgets.addWidget(slider);

    int tracks = 8;
  
    m_tracks.resize(tracks);

    char buf[64];

    for(int i=0; i < tracks; i++) {
      Track & t = m_tracks[i].ref();
      t.m_id = i;
      sprintf(buf, "Track %d", i);
      t.m_label = buf;
      t.build(this);
    }

    ConfigReader::Chunk audioChunk = m_host->engine()->config()->get("Audio");
    
    std::string path = audioChunk.get("path").getString();
    std::string file = audioChunk.get("file").getString();

    if(path.size() && file.size())
      m_audioFile = path + file;

    VEE_trace("%s # audio track = \"%s\"", fname, m_audioFile.c_str());
    
    if(m_audioFile.size())
      m_waveForm.setFile(m_audioFile.c_str(), 1000);
  }

  ComponentEditor::~ComponentEditor()
  {}
  
  bool ComponentEditor::msgProcess(const char *id, Fluid::Message *m)
  {
    const char * fname = "ComponentEditor::msgProcess";

    VEE_trace("%s # \"%s\"", fname, id);

    bool ok = false;

    char * endPtr = 0;

    if(strncmp(id, "enable_", 7) == 0) {

      int i = (int) strtol(id + 7, & endPtr, 0);
      VEE_trace("%s # Enabling button %d", fname, i);

      m_components[i].ptr()->setEnabled(this, true);

      ok = true;
    }
    if(strncmp(id, "move_", 5) == 0) {

      int i = (int) atoi(id + 5);

      VEE_trace("%s # Moving component %d", fname, i);

      ComponentData * cd = getComponent(i).ptr();

      cd->m_button.ptr()->setColors(m_moveColors);
      m_moveComponent = cd;
      m_moveOrigin = m_pointerLocation;    
      m_originalTime = cd->m_list->timeOffset();

      ok = true;
    }
    if(strncmp(id, "edit_", 5) == 0) {

      int i = (int) atoi(id + 5);

      VEE_trace("%s # Edit component %d", fname, i);

      getComponent(i).ptr()->startEdit(this);

    }
    else if(strncmp(id, "disable_", 8) == 0) {

      int i = (int) strtol(id + 8, & endPtr, 0);
      VEE_trace("%s # Disabling button %d", fname, i);

      m_components[i].ptr()->setEnabled(this, false);
    
      ok = true;
    }
    else if(strncmp(id, "delete_", 7) == 0) {

      int i = (int) strtol(id + 7, & endPtr, 0);

      VEE_trace("%s # Deleting list %d", fname, i);
      m_components[i].ptr()->setHidden();
    
      ok = true;
    }
    else if(strncmp(id, "enabletrack_", 11) == 0) {

      int i = (int) strtol(id + 11, & endPtr, 0);

      VEE_trace("%s # Enabled track %d ", fname, i);
      m_tracks[i].ref().setEnabled(this, true);

      ok = true;
    }
    else if(strncmp(id, "disabletrack_", 12) == 0) {

      int i = (int) strtol(id + 12, & endPtr, 0);

      VEE_trace("%s # Disabled track %d", fname, i);
      m_tracks[i].ref().setEnabled(this, false);
    
      ok = true;
    }
    else if(strncmp(id, "visualize_", 10) == 0) {

      int i = (int) strtol(id + 10, & endPtr, 0);

      m_components[i].ptr()->toggleVisual();

      // cd.m_button.ptr()->setChecked(0, cd.m_visualize);

      VEE_trace("%s # toggled visualization for %d", fname, i);

      ok = true;
    }
    else if(strcmp(id, "offset") == 0) {

      Fluid::MessageFloat * mf = dynamic_cast<Fluid::MessageFloat *> (m);

      if(mf) {
	float r = mf->value();
	m_host->engine()->setOffset(r);
	ok = true;

	VEE_trace("%s # offset = %f", fname, r);

      }
    }
  
    return ok;
  }

  void ComponentEditor::render(GlData *gld)
  {
    glDisable(GL_BLEND);

    m_widgets.renderGl(gld);

    // Render the track separators:
    glDisable(GL_TEXTURE_2D);

    glColor4fv(VEE_Rgba::White.data());

    glLineWidth(2);

    glBegin(GL_LINES);

    uint i, ntracks = m_tracks.size();

    for(i=0; i <= ntracks; i++) {

      float y = m_topLine - i * m_trackHeight;

      glVertex3f(m_left,  y, 0.0);
      glVertex3f(m_right, y, 0.0);
    }

    glEnd();

    // Render the lines that connect floating elements to their places:

    glBegin(GL_TRIANGLES);
    glColor4f(1.0, 0.5, 0.0, 1.0);

    for(i=0; i < m_components.size(); i++) {

      ComponentData * cd = m_components[i].ptr();

      if(cd->m_hidden)
	continue;

      Menu * menu = cd->m_button.ptr();
      Vector3 loc;
      loc.x = menu->low().x;
      loc.y = menu->high().y;
      loc.z = menu->location().z;

      if(loc.z != 0.0) {
	glVertex3fv(loc.data());
	glVertex3f(loc.x, menu->low().y, loc.z);
	glVertex3f(loc.x, loc.y, 0.0);
      }
    }
    
    glEnd();

    // Render the audio waveform:

    glColor3f(0.5, 0.5, 0.0);

    if(m_waveFormData.size()) {
      glPushMatrix();
      
      glColor4fv(VEE_Rgba::Red.data());
      glTranslatef(m_left, m_topLine + 0.2, -0.005);
      
      QAWL_GlRenderer::renderCollector(&m_waveFormData);
      
      glPopMatrix();
    }

    // Render the vertical time bar
    glEnable(GL_BLEND);
    float bottomLine = m_topLine - ntracks * m_trackHeight;

    float barX = m_left + m_host->engine()->timing().m_time / m_secsPerUnit;

    glColor4f(1.0, 0.4, 0.0, 0.7);

    glLineWidth(4);

    float w = 0.03;

    glBegin(GL_QUADS);
    glVertex3f(barX - w , bottomLine, 0.03);
    glVertex3f(barX + w , bottomLine, 0.03);
    glVertex3f(barX + w , m_topLine, 0.03);
    glVertex3f(barX - w , m_topLine, 0.03);
    glEnd();

    // VEE_trace("ComponentEditor::render # %f", barX);

    // Render the timing information
    glColor4fv(VEE_Rgba::White.data());
    glDisable(GL_BLEND);

    char buf[64];
    float x = m_left;
    float xmax = x + 10.0;

    float interval = ceil(m_secsPerUnit);

    float tmpx1, tmpx2, tmpy1, tmpy2;
  
    for(; x < xmax; i++) {

      float time = interval * i;
      int seconds = (int) time;
      int fractions = (int) ((time - seconds) * 100.0);
      sprintf(buf, "%.2d:%.2d", seconds, fractions);

      gld->m_font.getBBox(buf, 0.15, 0.0, &tmpx1, &tmpx2, &tmpy1, &tmpy2);

      float textWidth = tmpx2 - tmpx1;
    
      gld->m_fntRenderer.start3f(x - textWidth * 0.5, m_topLine + 0.1, 0.0);
      gld->m_fntRenderer.puts(buf);
    
      glDisable(GL_TEXTURE_2D);
      glBegin(GL_LINES);
      glVertex3f(x, m_topLine, 0.0);
      glVertex3f(x, m_topLine + 0.1, 0.0);
      glEnd();

      x += interval / m_secsPerUnit;
    }

  
  }

  void ComponentEditor::postRender(GlData *gld)
  {
    // Now render the visualizations if necessary:

    for(component_iterator it = m_components.begin(); 
	it != m_components.end(); it++) {

      ComponentData * cd = (*it).second.ptr();
      if(cd->m_visualize && cd->m_list && !cd->m_hidden)
	cd->m_list->visualize(gld);
    }

  }

  void ComponentEditor::update()
  {
    m_widgets.update();
  }

  void ComponentEditor::handleEvent(const InEvent *e)
  { 
    if(m_moveComponent) {

      VEE_Vector3 newLoc = e->location();

      // Move the component to proper x-location. 
      float xmove = newLoc.x - m_moveOrigin.x;
      m_moveOrigin = newLoc;

      float moveSeconds = xmove * m_secsPerUnit;
      Fluid::TimeStamp offset;
      offset.setSeconds(moveSeconds);

      m_moveComponent->m_list->moveTimeOffset(offset);

      VEE_trace("ComponentEditor::handleEvent # %lf",
		m_moveComponent->m_list->beginTime());

      // Move the component to proper track:
      int track = (int) ((m_topLine - newLoc.y) / m_trackHeight);
      int listId = m_moveComponent->m_listId;
      track = VEE_clamp(track, 0, (int) m_tracks.size() - 1);
    
      for(int i=0; i < (int) m_tracks.size(); i++) {

	Track * t = & m_tracks[i].ref();
	int index = t->findComponent(listId);

	if(index >= 0) {
	  if(i != track) {
	    m_tracks[track].ref().m_components.push_back(getComponent(listId));
	    t->removeComponent(index);
	    VEE_trace("ComponentEditor::handleEvent # Moved component from %d to %d",
		      i, track);
	  }
	  i = m_tracks.size();
	}
      }

      redesignLayout();
    
      if(e->buttonState() && !e->prevButtonState()) {
	m_moveComponent->toggle(this); // Reset the colors
	m_moveComponent->toggle(this);
	m_moveComponent = 0;
	m_pointerLocation = newLoc;
      }
    }
    else {
      m_pointerLocation = e->location();
    
      m_widgets.handleEvent(e); 
    }  

    if(m_waveFormData.size() == 0 && m_audioFile.size() > 0)
      recalculateWaveform();
  }

  void ComponentEditor::addList(ActionList *al)
  {
    int typeIndex = ++m_typeCounters[m_recordName].m_value;

    ComponentData *cdp = new ComponentData;
    cdp->m_host = this;
    VEE_RefPtr<ComponentData> cd(cdp);

    int listId = al->myId();
    cdp->m_listId = listId;

    char buf[64];

    /* Menu * txb = new Menu(m_host);
       txb->setSounds(m_host->buttonSounds()); */

    sprintf(buf, "%s %d", m_recordName.c_str(), typeIndex);

    cdp->m_label = buf;
    cdp->build(this);

    // txb->setLabel(buf);

    VEE_trace("ComponentEditor::addList # Added component %d", typeIndex);

    m_components[listId] = cd;
    m_tracks[m_recordTrack].ref().m_components.push_back(cd);

    m_newRecords = true;

    redesignLayout();
  }

  void ComponentEditor::startRecording(const std::string &name)
  {
    if(m_newRecords)
      m_recordTrack = (m_recordTrack + 1) % m_tracks.size();
    m_recordName = name;

    m_newRecords = false;
  }

  void ComponentEditor::redesignLayout()
  {
    m_trackHeight = 0.3;
  
    float margin = m_trackHeight * 0.1;
    m_right = m_left;
    int group = 0;

    for(iterator it = m_tracks.begin(); it != m_tracks.end(); it++) {

      Track * t = & (*it).ref();

      float top = m_topLine - group * m_trackHeight;
      float bottom = top - m_trackHeight;
    
      float buttonW = 1.4;

      Fluid::Vector2 menulow(m_left - buttonW - 0.1, bottom);
      Fluid::Vector2 menuhigh(m_left - 0.1, top);
      t->m_menu.ptr()->setVertices(menulow, menuhigh);
  
      for(uint i=0; i < t->m_components.size(); i++) {
      
	ActionList *al = t->m_components[i].ptr()->m_list;
            
	double beginTime = al->beginTime();
	double endTime = al->endTime();
      
	if(endTime < beginTime + 0.5)
	  endTime = beginTime + 0.5;
      
	float left = beginTime / m_secsPerUnit + m_left;
	float right = endTime / m_secsPerUnit + m_left;
      
	Fluid::Vector2 low(left, bottom + margin);
	Fluid::Vector2 high(right, top - margin);

	t->m_components[i].ptr()->m_button.ptr()->setVertices(low, high);
      
	if(m_right < right)
	  m_right = right;

	int plane = 0;

	t->m_components[i].ptr()->m_button.ptr()->
	  setLocation(Vector3(0.0, 0.0, 0.0));
	
	while(t->spaceIsOccupied(left, right, plane, i)) {
	  plane++;	  

	  t->m_components[i].ptr()->m_button.ptr()->
	    setLocation(Vector3(0.0, 0.0, plane * 0.4));
	}
	t->m_components[i].ptr()->m_plane = plane;
      }

      group++;
    }

    if(m_right < m_left + 10.0)
      m_right = m_left + 10.0;
  }

  void ComponentEditor::clear()
  {
    /* for(iterator it = m_tracks.begin(); it != m_tracks.end(); it++) {
       (*it).ref().clear();
       }*/
  
    m_tracks.clear();
    m_components.clear();
  }

  void ComponentEditor::fixRowMenu
  (Menu *menu, int id, const std::string &label)
  {
    char buf[64];

    menu->clear();

    sprintf(buf, "enabletrack_%d", id);
    menu->addItem("Enable",  buf, this);
    sprintf(buf, "disabletrack_%d", id);
    menu->addItem("Disable", buf, this);

    menu->setLabel(label);
  }

  QDomElement ComponentEditor::writeDom(QDomDocument &doc)
  {
    QDomElement elem = doc.createElement("Components");
  
    char buf[64];
    uint i=0;

    if(m_components.size()) {
      QDomElement components = doc.createElement("components");
      
      for(component_iterator it = m_components.begin(); 
	  it != m_components.end(); it++) {
	components.appendChild((*it).second.ptr()->writeDom(doc));
      }
      
      elem.appendChild(components);
    }

    for(iterator it = m_tracks.begin(); it != m_tracks.end(); it++) {
      sprintf(buf, "track_%u", i);
    
      QDomElement tmp = doc.createElement(buf);
    
      tmp.appendChild((*it).ref().writeDom(doc));
      elem.appendChild(tmp);
      i++;
    }

  
    return elem;
  }

  bool ComponentEditor::readDom(QDomElement &elem)
  {
    if(elem.tagName() != "Components") {
      VEE_error("ComponentEditor::readDom # bad tag (%s)", 
		elem.tagName().ascii());
      return false;
    }
  
    clear();

    // Get components:
  
    QDomElement components = elem.namedItem("components").toElement();

    if(!components.isNull()) {

      QDomNode n = components.firstChild();
      while ( !n.isNull() ) {
	if ( n.isElement() ) {
	  QDomElement e = n.toElement();
	  ComponentData *cdp = new ComponentData;
	  cdp->m_host = this;
	  VEE_RefPtr<ComponentData> cd(cdp);

	  if(cdp->readDom(e))
	    m_components[cdp->m_listId] = cd;
	}
	n = n.nextSibling();
      }
    }

    VEE_trace("ComponentEditor::readDom # Got %u components", 
	      m_components.size());

    // Get the tracks:
 
    char buf[64];
    bool loop = true;

    for(uint i=0; loop; i++) {
      sprintf(buf, "track_%u", i);
    
      QDomElement e = elem.namedItem(buf).toElement();

      if(e.isNull())
	loop = false;
      else {
	VEE_ReferenceObject<Track> track;
    
	QDomElement e2 = e.firstChild().toElement();
	if((*track).readDom(e2, this)) {
	  (*track).build(this);
	  m_tracks.push_back(track);
	
	}
      }
    }
  
    VEE_trace("ComponentEditor::readDom # Got %u tracks", 
	      m_tracks.size());

    redesignLayout();

    return true;
  }

  void ComponentEditor::recalculateWaveform()
  {
    VEE_trace("ComponentEditor::recalculateWaveform");

    float sr = 44100;
    float secs = 10.0;
    
    QAWL_GlRenderer r;

    r.render(& m_waveForm, 0, (ulong) (sr * secs),
	     0, 
	     secs / m_secsPerUnit, 0.01, 0.3, 0.3, 
	     0, & m_waveFormData);

    VEE_trace("ComponentEditor::recalculateWaveform # EXIT");
  }
  
}

