/* -*- C++ -*- */

/* -*- C++ -*- */

/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Crystal animation and 3D modeling application
 *
 * Authors:   Tommi Ilmonen 2001-2004
 * Copyright: Tommi Ilmonen and/or Tekes 2001-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <crystal_density_editor.h>

#include <crystal_dom_util.h>
#include <crystal_engine.h>
#include <crystal_gl_data.h>
#include <crystal_model.h>
#include <crystal_tools.h>

#include <Fluid/MessageFloat.hpp>
#include <Fluid/MessageString.hpp>
#include <Fluid/Mouse.hpp>
#include <Fluid/TrackerSensorSelector.hpp>

#include <Solar/MtlParser.hpp>
#include <Solar/GraphicTriangles.hpp>

#include <vee_algorithm.h>
#include <vee_color.h>
#include <vee_error.h>
#include <vee_mesh_conversion.h>

namespace Crystal { 

  
  DensityEditor::Element::Element()
    : m_hsva(VEE_Hsva::Yellow),
      m_texture("stripes"),
      m_dirty(false),
      m_renderCount(0),
      m_uvScale(1.0)
  {
    m_gridCenter.clear(); 
    m_material.setFlags(Solar::GlMaterial::LIGHTING);
    m_material.setColor(m_hsva.rgba());
  }

  DensityEditor::Element::~Element()
  {}

  void DensityEditor::Element::createModelFromGrid(Engine *e)
  {
    m_dirty = false;

    m_model.clear();

    m_grid.triangulate();

    if(!m_grid.triangleCount())
      return;

    Solar::GraphicTriangles * gti = new Solar::GraphicTriangles();
    gti->setMaterial(m_materialName);
    gti->setGarbageBin(e->deletedGraphics());
    
    VEE_MeshConversion::append(gti, &m_grid, m_uvScale);
    
    m_model.addGraphic(gti);
    m_model.centralize();
  }


  QDomElement DensityEditor::Element::writeDom(QDomDocument &doc)
  {
    QDomElement elem = doc.createElement("MetaballMesh");

    addTextChild(elem, doc, "texture",  m_texture.c_str());
    addTextChild(elem, doc, "material", m_materialName.c_str());
    addTextChild(elem, doc, "color",    toStr(m_hsva));
    addTextChild(elem, doc, "voxelSize", m_grid.voxelSize());
    addTextChild(elem, doc, "uvScale", m_uvScale);
  
    VEE_MetaBalls * balls = m_grid.metaBalls();

    if(balls->nofBalls()) {

      QDomElement items = doc.createElement("MetaBalls");
      
      for(uint i=0; i < balls->nofBalls(); i++) {
	const VEE_MetaBalls::Ball & b = balls->getBall(i);
	
	QDomElement mb = doc.createElement("MetaBall");
	addTextChild(mb, doc, "center", b.m_center);
	addTextChild(mb, doc, "radius", b.m_radius);
	
	items.appendChild(mb); 
      }
      
      elem.appendChild(items);
    }

    return elem;
  }

  bool DensityEditor::Element::readDom(QDomElement &elem, Engine *engine)
  {
    if(elem.tagName() != "MetaballMesh")
      return false;

    QString texture = getString(elem, "texture").ascii();
    if(!texture.isEmpty())
      m_texture = texture.ascii();

    m_materialName = getString(elem, "material").ascii();
    m_hsva         = toHsva(getString(elem, "color"));
    m_material.setColor(m_hsva.rgba());
    m_uvScale = getFloat(elem, "uvScale", 1.0);
    
    float voxelSize = getFloat(elem, "voxelSize", 0.3);

    m_grid.clear();
    m_grid.setVoxelSize(voxelSize);
    
    QDomElement items = elem.namedItem("MetaBalls").toElement();

    if(!items.isNull()) {
      QDomNode n = items.firstChild();
      while ( !n.isNull() ) {
	if ( n.isElement() ) {
	  QDomElement e = n.toElement();
	  VEE_MetaBalls::Ball b;
	  b.m_center = getVector3(e, "center");
	  b.m_radius = getFloat(e, "radius", 1.0);

	  m_grid.addBall(b.m_center, b.m_radius);
	}
	n = n.nextSibling();
      }
    }

    createModelFromGrid(engine);

    m_renderCount = 0;
    
    return true;
  }
      

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  DensityEditor::DensityEditor(Tools *host)
    : Transformation(host),
      m_host(host),
      m_widgets(host->grabs()),
      m_selectedEditor(0),
      m_current(0)
  {
    const char * fname = "DensityEditor::DensityEditor";

    char buf[32];
      
    m_recorder = new DensityRecorder(host);
    
    const uint fields = 6;
    
    m_fields.resize(fields);

    for(uint i=0; i < m_fields.size(); i++) {
      Element * e = new Element;
      m_fields[i] = e;
      e->m_grid.setVoxelSize(0.3);
      sprintf(buf, "volume_%u", i);
      e->m_materialName = buf;
    }

    Vector3 texLow, texSpan;
    
    m_low.make(-2.0, 1.0, 0.0);
    m_span.make(0.7);
    m_move.make(0.9, 0.0, 0.0);
    
    texLow.make(-2.5, 2.0, 1.0);
    texSpan.make(0.5, 0.3, 0.0);

    Vector2 uvLoc, uvSize;
    uvLoc.make(-3.0, 0.0);
    uvSize.make(0.5, 2.0);

    rebuildBoxes();

    // Now build texture selectors:

    std::string mtlFile = m_host->engine()->config()->get("OpenGL").
      get("materials").getString("vee_materials.mtl");

    ConfigReader::Config conf;

    if(mtlFile.size())
      Solar::MtlParser::load(&conf, mtlFile.c_str());
			     
    const uint cols = 1;

    uint i = 0;

    std::set<std::string> loadedTextures;

    bool showTexNames = 
      m_host->engine()->config()->get("DensityEditor").
      get("show_texnames").getString() == "Yes";

    for(ConfigReader::Config::iterator it = conf.begin();
	it != conf.end(); it++) {

      std::string tex2 = (*it).second.get("map_Ka").getString();

      // Simplify the texture name (if it is a file)
      const char * str = tex2.c_str();
      const char * slash = 0;

      int l = strlen(str);

      for(int j = l - 1; j > 0; j--) {
	if(str[j] == '/') {
	  slash = str + j;
	  j = 0;
	}
      }

      std::string tex = slash ? slash + 1 : str;

      if(!tex2.size() || loadedTextures.find(tex) != loadedTextures.end())
	continue;

      loadedTextures.insert(tex);

      uint row = i / cols;
      uint col = i % cols;

      Vector3 loc = texLow;
      loc.x += col * texSpan.x;
      loc.y -= row * texSpan.y;

      TextureButton * txb = new TextureButton(this);

      txb->setLocation(loc);
      txb->setSize(texSpan.x, texSpan.y);
      
      /* VEE_trace("%s # txb loc = %f %f %f [%u %u %u]", 
	 fname, loc.x, loc.y, loc.z, col, row, i); */

      if(showTexNames)
	txb->setLabel(tex);

      txb->setTexture((*it).first.c_str());

      txb->msgAddSink("press", std::string("texture_") + (*it).first, this);

      m_widgets.addWidget(txb);
      i++;
    }

    VEE_trace("%s # %u textures", fname, i);


    VEE_BoundingBox colorCubeBox;

    colorCubeBox.low().make(-2.5, 0.0, 0.0);
    colorCubeBox.high() = colorCubeBox.low() + Vector3(1.0, 1.0, 1.0);

    ColorCube * cc = new ColorCube(this);
    cc->msgAddSink("color", "color", this);
    cc->setBoundingBox(colorCubeBox);
    m_widgets.addWidget(cc);

    m_uvScale = new Slider(this, 1.0, 0.02, 4.0);    
    m_uvScale->setArea(uvLoc, uvSize);
    m_uvScale->msgAddSink("value", "uvScale", this);
    m_widgets.addWidget(m_uvScale);

    msgProcess("sphere_3", 0);
    msgProcess("sphere_4", 0);
    msgProcess("sphere_5", 0);

    select(0);
  }

  DensityEditor::~DensityEditor()
  {}

  bool DensityEditor::msgProcess(const char *id, Fluid::Message *m)
  {
    const char * fname = "DensityEditor::msgProcess";

    bool ok = false;

    if(strncmp(id, "volume_", 7) == 0) {
      int index = atoi(id + 7);

      VEE_trace("%s # Selected volume %d", fname, index);
    
      select(index);

      ok = true;
    }
    if(strncmp(id, "edit_", 5) == 0) {
      int index = atoi(id + 5);

      VEE_trace("%s # Edit volume %d", fname, index);
      select(index);

      m_host->setUiHandler(m_recorder);

      ok = true;
    }
    if(strncmp(id, "clear_", 6) == 0) {
      int index = atoi(id + 6);

      VEE_trace("%s # Cleared volume %d", fname, index);
      select(index);

      Element * dd = m_fields[index].ptr();

      dd->m_grid.clear();
      dd->createModelFromGrid(m_host->engine());

      ok = true;
    }
    if(strncmp(id, "sphere_", 7) == 0) {
      int index = atoi(id + 7);

      VEE_trace("%s # Sphere volume %d", fname, index);
      select(index);
        
      Element * dd = m_fields[index].ptr();
      // dd->m_metaBalls.clear();
      dd->m_grid.clear();
      dd->m_grid.addBall(VEE_Vector3ZERO, 1.0);
      dd->m_grid.triangulate();
      dd->createModelFromGrid(m_host->engine());

      ok = true;
    }
    else if(strncmp(id, "animate_", 8) == 0) {
      int index = atoi(id + 8);
      animate(index);

      ok = true;
    }
    else if(strncmp(id, "texture_", 8) == 0) {
      return selectTexture(id + 8);
    }
    else if(strcmp(id, "color") == 0) {

      if(!m_current)
	return false;

      float h = 0.5, s = 0.5, v = 0.5;
      Fluid::MessageString *ms = dynamic_cast<Fluid::MessageString *> (m);
      
      if(!ms)
	return false;

      int ns = sscanf(ms->string().c_str(), "%f %f %f", &h, &s, &v);
      if(ns != 3)
	return false;

      m_current->m_hsva.set(h, s, v, 1.0, 0.0);
      m_current->m_material.setColor(m_current->m_hsva.rgba());
      m_current->m_renderCount = 0;

      return true;
    }
    else if(strcmp(id, "uvScale") == 0) {
      
      Fluid::MessageFloat * mf = dynamic_cast<Fluid::MessageFloat *> (m);

      if(mf) {
	m_current->m_uvScale = mf->value();
	ok = true;
      }
    }


    return ok;
  }

  void DensityEditor::render(GlData *gld)
  {
    m_widgets.renderGl(gld);
    renderVolumes(gld, &m_cubes);
  }

  void DensityEditor::update()
  {
    for(uint i=0; i < m_fields.size(); i++) {
      m_fields[i].ptr()->m_renderCount++;
    }
  }

  void DensityEditor::startRecording()
  {}
  

  bool DensityEditor::select(int index)
  {
    if((uint) index >= m_fields.size())
      return false;

    m_selectedEditor = index;
    m_current = m_fields[index].ptr();
    m_recorder->setData(m_current);

    for(uint i=0; i < m_fields.size(); i++)
      m_cubes[i]->setColors(m_host->usualColor());

    m_cubes[m_selectedEditor]->setColors(m_host->selectedColor());

    m_uvScale->setValue(m_current->m_uvScale);

    return true;
  }

  Model * DensityEditor::createModel(uint i)
  {
    if(i >= m_fields.size())
      return 0;

    Model * m = new Model;
    Element * e = m_fields[i].ptr();

    m->object().setModel( & e->m_model);

    VEE_trace("DensityEditor::createModel # %d (%u tris)", 
	      i, e->m_grid.triangleCount());

    return m;
  }

  void DensityEditor::createVolumeBoxes
  (const VEE_Vector3 &first,
   const VEE_Vector3 &span,
   const VEE_Vector3 &move,
   uint fields,
   Fluid::MessageSink *sink,
   Widgets *widgets,
   Transformation * trans,
   Tools * tools,
   std::vector<CubeWidget*> *store,
   std::vector<Menu*> * mstore)
  {
    char buf[64];

    Fluid::Vector2 low2, high2;
  
    VEE_Vector3 low = first;

    for(uint i=0; i < fields; i++) {

      // Add the cube:
      CubeWidget * cube = new CubeWidget(trans);

      cube->setSounds(tools->buttonSounds());
      cube->setBorder(0.05);

      VEE_BoundingBox bbox;

      bbox.low()  = low + move * (float) i;
      bbox.high() = bbox.low() + span;
    
      cube->setBoundingBox(bbox);
      sprintf(buf, "volume_%u", i);
      cube->msgAddSink("press", buf, sink);

      cube->setColors(tools->usualColor());

      widgets->addWidget(cube);

      store->push_back(cube);

      if(mstore) {
	// Add a menu:

	Menu * menu = new Menu(trans);

	menu->setSounds(tools->buttonSounds());
	sprintf(buf, "Vol %u", i + 1);
	menu->setLabel(buf);
      
	sprintf(buf, "edit_%u", i);
	menu->addItem("Edit", buf, sink);
      
	sprintf(buf, "clear_%u", i);
	menu->addItem("Clear", buf, sink);
      
	sprintf(buf, "sphere_%u", i);
	menu->addItem("Sphere", buf, sink);
      
	sprintf(buf, "animate_%u", i);
	menu->addItem("Animate", buf, sink);
      
	low2.make(bbox.low()[0], bbox.high()[1] + 0.1);
	high2.make(bbox.high()[0], bbox.high()[1] + 0.7);
      
	menu->setVertices(low2, high2);
	menu->setMenuBelow(false);
      
	widgets->addWidget(menu);

	mstore->push_back(menu);
      }
    }  
  }

  void DensityEditor::renderVolumes
  (GlData *gld, 
   std::vector<CubeWidget*> * cubes)
  {
    uint i;

    VEE_Vector3 zero;
    zero.clear();

    if(cubes->size() != m_fields.size())
      VEE_fatal("DensityEditor::renderVolumes # Sizes do not match");

    /* for(i=0; i < m_fields.size(); i++) {

      Element * dd = m_fields[i].ptr();

      const VEE_BoundingBox & bbox = (*cubes)[i]->boundingBox();
      VEE_BoundingBox bbox2;
      dd->m_grid.getBoundingBox(&bbox2);

      VEE_Vector3 center  = bbox.center();
      VEE_Vector3 center2 = bbox2.center();
      float span = bbox.span().minimum() / bbox2.span().maximum();

      glPushMatrix();

      glTranslatef(center.x, center.y, center.z);
      glScalef(span, span, span);

      gld->m_veeRenderer.drawVolumeGrid2(-center2, &dd->m_grid);

      glPopMatrix();
      }*/

    // Now rebuild the necessary materials:    
    
    for(i=0; i < m_fields.size(); i++) {

      Element * e = m_fields[i].ptr();

      if(e->m_renderCount < 2) {

	Solar::GlMaterial mat = e->m_material;
	
	VEE_GlTexture2D * tex = gld->m_textures.getTexture(e->m_texture);
	
	if(tex)
	  mat.setTextureId(tex->id());
	else 
	  VEE_error("DensityEditor::renderVolumes # no texture \"%s\"",
		    e->m_texture.c_str());
	
	gld->m_materials.addMaterial(e->m_materialName, mat);
      }
    }

    // Now render the models as Solar objects beneath the boxes:

    glPushAttrib(~0);

    gld->m_solarRenderer.prepareToRender();

    for(i=0; i < m_fields.size(); i++) {

      Element * dd = m_fields[i].ptr();

      const VEE_BoundingBox & bbox = (*cubes)[i]->boundingBox();
      const VEE_BoundingBox & bbox2= dd->m_model.boundingBox();

      VEE_Vector3 center  = bbox.center();
      VEE_Vector3 center2 = bbox2.center();
      float span = bbox.span().minimum() / bbox2.span().maximum();

      glPushMatrix();

      glTranslatef(center.x, center.y, center.z);
      glScalef(span, span, span);
      glTranslatef(-center2.x, -center2.y, -center2.z);

      dd->m_model.render(& gld->m_solarRenderer);

      glPopMatrix();
    }
    glPopAttrib();
  }

  VEE_Volume * DensityEditor::cloneVolume(int i)
  {
    if((uint) i >= m_fields.size())
      VEE_fatal("DensityEditor::cloneVolume # Argument too large");

    return m_fields[i].ptr()->m_grid.clone();
  }

  QDomElement DensityEditor::writeDom(QDomDocument &doc)
  {
    QDomElement elem = doc.createElement("Solids");

    for(uint i=0; i < m_fields.size(); i++)
      elem.appendChild(m_fields[i].ptr()->writeDom(doc));

    return elem;
  }

  bool DensityEditor::readDom(QDomElement &elem)
  {
    VEE_trace("DensityEditor::readDom # %s", elem.tagName().ascii());

    if(elem.tagName() != "Solids")
      return false;

    m_fields.clear();

    QDomNode n = elem.firstChild();

    while ( !n.isNull() ) {
      if ( n.isElement() ) {
	QDomElement e = n.toElement();

	Element * el = new Element;

	bool ok = el->readDom(e, m_host->engine());

	VEE_trace("DensityEditor::readDom # New element %d %p", (int) ok, el);

	if(!ok)
	  delete el;
	else
	  addElement(el);

      }
      n = n.nextSibling();
    }

    rebuildBoxes();

    select(0);

    return true;
  }

  bool DensityEditor::selectTexture(const char * t)
  {
    if(!m_current)
      return false;

    VEE_trace("DensityEditor::selectTexture # %p %s", m_current, t);

    m_current->m_texture = t;
    m_current->m_renderCount = 0;

    return true;
  }

  uint DensityEditor::expandOne()
  {
    uint r = m_fields.size();

    m_fields.resize(r + 1);

    m_fields[r] = new Element;

    Element * elem = m_fields[r].ptr();

    char buf[32];
    
    sprintf(buf, "material_%u", r);

    elem->m_materialName = buf;

    return r;
  }

  void DensityEditor::animate(int index)
  {
    VEE_trace("DensityEditor::animate # %d", index);

    select(index);

    m_host->engine()->setNextRecorder(Engine::RECORDER_MESH, "MeshRecorder");

    m_host->engine()->skipMode(Engine::MODE_PLAY);
  }

  void DensityEditor::addElement(Element *e)
  {
    uint n = m_fields.size();

    m_fields.resize(n + 1);

    m_fields[n] = e;
  }
  
  void DensityEditor::rebuildBoxes()
  {
    VEE_Algorithm::deleteAll(m_cubes.begin(), m_cubes.end());
    VEE_Algorithm::deleteAll(m_menus.begin(), m_menus.end());

    m_cubes.clear();
    m_menus.clear();

    createVolumeBoxes(m_low, m_span, m_move, m_fields.size(), this, 
		      &m_widgets, this, m_host, & m_cubes, &m_menus);
  }

  

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  DensityRecorder::DensityRecorder(Tools *host)
    : m_host(host),
      m_data(0)
  {
    m_prevLoc.clear();
  }

  DensityRecorder::~DensityRecorder()
  {}

  void DensityRecorder::handleInput()
  {
    Fluid::TrackerSensorSelector2 *left  = m_host->engine()->strongHandLoc();
    Fluid::TrackerSensorSelector2 *right = m_host->engine()->weakHandLoc();
    Fluid::Mouse                  *mouse = m_host->engine()->mouse();

    bool gotNew = false;

    VEE_Time tmpTime;
    tmpTime.getTimeOfDay();
    
    // Discard the first samples (mouse has not been released yet).
    bool discard = m_startTime.secsTo(tmpTime) < 1.0;

    for(; m_samples < (long) left->locations().getSampleCount(); m_samples++) {

      VEE_Vector3 leftLoc = left->locations().getIndex(m_samples);
      VEE_Vector3 rightLoc = right->locations().getIndex(m_samples);

      VEE_Vector3 center = (leftLoc + rightLoc) * 0.5;

      float radius = (leftLoc - center).length();
      float d = (center - m_prevCenter).length();

      Fluid::TimeStamp t = left->locations().getTimeStamp(m_samples);
      long mouseIndex = mouse->outputBuffer().timeStampToIndex(t);

      Fluid::MouseSample *ms = & mouse->outputBuffer().getIndex(mouseIndex);
    
      /* VEE_trace("DensityRecorder::handleInput # %d %d %f %f", 
	 (int) ms->buttonPressed(0), m_metaBalls.nofBalls(), d, radius); */

      if((d > radius * 0.2 || m_first) && ms->buttonPressed(0) && !discard) {
	m_data->m_grid.addBall(center, radius);
      
	m_prevCenter = center;
	m_first = false;
	gotNew = true;
      }

      m_prevLoc = center;
      m_prevRad = radius;
    }

    if(gotNew) {

      m_data->m_grid.triangulate();
    
      VEE_trace("DensityRecorder::handleInput # %d triangles, %3.1f kilobytes",
		(int) m_data->m_grid.triangleCount(), 
		(float) m_data->m_grid.byteCount() / 1024.0f);
      
      VEE_BoundingBox bbox;

      m_data->m_grid.getBoundingBox(&bbox);
      m_data->m_gridCenter = bbox.center();
      m_data->m_dirty = true;
    }
  }

  void DensityRecorder::update()
  {}

  void DensityRecorder::render(GlData *gld)
  {
    if(!m_data) return;

    glColor4fv(VEE_Rgba::White.data());

    glPushMatrix();
    glTranslatef(m_prevLoc.x, m_prevLoc.y, m_prevLoc.z);
    glRotatef(90.0, 1.0, 0.0, 0.0);

    gluQuadricDrawStyle(gld->m_quadric, GLU_FILL);
    gluQuadricTexture(gld->m_quadric, GL_TRUE);
    
    const VEE_GlMaterial & mat = gld->m_veeRenderer.materials().getMaterial
      (VEE_GlRenderer::CHECK);
    
    if(mat.hasTexture()) {
      glEnable(GL_TEXTURE_2D);
      glBindTexture(GL_TEXTURE_2D, mat.textureId());
    }
    else
      glDisable(GL_TEXTURE_2D);

    gluSphere(gld->m_quadric, m_prevRad * 0.7, 10, 10);
  
    glPopMatrix();

    if(!m_data->m_grid.ballCount())
      return;

    VEE_Vector3 zero;
    zero.clear();

    gld->m_veeRenderer.drawVolumeGrid2(zero, &m_data->m_grid, false);
  }

  void DensityRecorder::start()
  {
    Fluid::TrackerSensorSelector2 *left = m_host->engine()->strongHandLoc();

    m_samples = (long) left->locations().getSampleCount();

    m_first = true;
    m_prevCenter.clear();
    m_prevLoc.clear();
    m_prevRad = 0.0;
    m_startTime.getTimeOfDay();
  }

  void DensityRecorder::stop()
  {
    if(m_data->m_dirty)
      m_data->createModelFromGrid(m_host->engine());
  }

}
