/* -*- C++ -*- */

/* -*- C++ -*- */

/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Crystal animation and 3D modeling application
 *
 * Authors:   Tommi Ilmonen 2001-2004
 * Copyright: Tommi Ilmonen and/or Tekes 2001-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <crystal_emitter_editor.h>

#include <crystal_density_editor.h>
#include <crystal_dom_util.h>
#include <crystal_engine.h>
#include <crystal_gl_data.h>
#include <crystal_particle_editor.h>
#include <crystal_tools.h>

#include <vee_error.h>
#include <vee_gl_renderer.h>
#include <vee_math.h>
#include <vee_operator_random.h>

#include <Fluid/MessageFloat.hpp>
#include <Fluid/MessageString.hpp>

namespace Crystal { 

  static const char * parammap[EmitterEditor::SLIDERS] = {
    "rate",
    "speed",
    "spread"
  };

  static const char * emissions[] = {
    "running",
    "static",
    0
  };

  static const char * emissionLabels[] = {
    "Emitter",
    "Emitter (static)",
    0
  };

  EmitterEditor::EmitterData::EmitterData()
    : m_particle(0),
      m_volume(0)
  { 
    m_params[RATE] = 40.0;
    m_params[SPEED] = 0.3;
    m_params[SPREAD] = 0.2;

    VEE_OperatorRandomSphereVector * speed = new VEE_OperatorRandomSphereVector;
    speed->setId("addvelocity");
    m_emitter.appendOperator2(speed, "speed");

    VEE_OperatorRandomSphereVector * spread = new VEE_OperatorRandomSphereVector;
    spread->setId("addlocation");
    m_emitter.appendOperator2(spread, "spread");
  }

  void EmitterEditor::EmitterData::init(EmitterEditor *host)
  {
    ParticleEditor *pe = host->host()->particleEditor();
    VEE_Object * particle = pe->createParticle(m_particle);

    VEE_trace("EmitterEditor::EmitterData::init # particle %d has texture %d",
	      (int) m_particle, (int) particle->texture());

    m_emitter.setType(particle);
    m_emitter.setValueFloat("rate", m_params[RATE] * 0.5);
    m_emitter.setValueFloat("childops/speed/radius", m_params[SPEED] * 0.5);
    m_emitter.setValueFloat("childops/spread/radius", m_params[SPREAD] * 0.5);
  
    m_volumeOp.setId("addlocation");
    m_volumeOp.setVolume(host->host()->densityEditor()->cloneVolume(m_volume));

    m_emitter.removeOperator("volume");
    m_emitter.appendOperator2(m_volumeOp.clone(), "volume");
  }

  QDomElement EmitterEditor::EmitterData::writeDom(QDomDocument &doc)
  {
    QDomElement elem = doc.createElement("EmitterData");

    uint i;

    for(i=0; i < SLIDERS; i++)
      addTextChild(elem, doc, parammap[i], m_params[i]);

    addTextChild(elem, doc, "particle", (int) m_particle);
    addTextChild(elem, doc, "volume",   m_volume);
  
    return elem;
  }

  bool EmitterEditor::EmitterData::readDom(QDomElement &elem)
  {
    if(elem.tagName() != "EmitterData")
      return false;

    for(uint i=0; i < SLIDERS; i++)
      m_params[i] = getFloat(elem, parammap[i], m_params[i]);
  
    m_particle = getInt(elem, "particle", m_particle);
    m_volume   = getInt(elem, "volume",   m_volume);

    return true;
  }

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  EmitterEditor::EmitterEditor(Tools * host)
    : Transformation(host),
      m_host(host),
      m_currentEmitter(0),
      m_selectedEmitter(0),
      m_widgets(host->grabs())
  {
    char buf[64];
    Fluid::Vector2 v1, v2, v3;
  
    m_emitterBox.set(-2.0, 0.7, 0.0, 
		     0.0, 2.0, 0.0);

    m_particleBox.set(0.5, 0.7, 0.0,
		      2.5, 2.8, 0.0);

    Fluid::Vector2 b1, b2, b3; // Emission type selectors

    b1.make(-2.0, 2.2);
    b2.make(1.0, 0.0);
    b3.make(0.9, 0.4);

    Fluid::Vector3 w1, w2; // Slider locations
    w1.make(-2.3, 1.0, 2.7);
    w2.make(0.0, 0.7, 0.0);

  
    Fluid::Vector3 clow, cspan, cmove; // Volume locations

    clow.make(-2.0, 0.0, 0.0);
    cspan.make(0.6, 0.6, 0.6);
    cmove.make(0.7, 0.0, 0.0);

    uint emitterRows = 4;
    uint emitterCols = 3;
    uint i, j;
  
    Fluid::Vector3 buttonSize = m_emitterBox.span();
    buttonSize[0] /= (float) emitterCols;
    buttonSize[1] /= (float) emitterRows;

    v2.make(buttonSize[0], buttonSize[1]); 
    v3.make(m_emitterBox.low()[0], m_emitterBox.low()[1]);

    m_emitters.resize(emitterRows, emitterCols);
  
    for(i = 0; i < emitterRows; i++) { 
      for(j = 0; j < emitterCols; j++) { 
	v1.make(v2[0] * j, v2[1] * i);
	v1 += v3;

	uint index = i * emitterCols + j;

	TextureButton * txb = new TextureButton(m_host);
      
	txb->setArea(v1, v2);
	txb->setSounds(m_host->buttonSounds());

	sprintf(buf, "%u", index + 1);
	txb->setLabel(buf);

	m_widgets.addWidget(txb);
	sprintf(buf, "emitter_%u", index);

	txb->msgAddSink("press", buf, this);
	m_emitters.get(i, j).m_button = txb;
	m_emitters.get(i, j).init(this);

	if(index == m_selectedEmitter) {
	  txb->setColors(m_host->selectedColor());
	  m_currentEmitter = & m_emitters.get(i, j);
	}
      }
    }

    m_particles = m_host->particleEditor()->particles();

    Transformation nullTrans;

    ParticleEditor::createParticleSelectors
      (&m_widgets, nullTrans, 
       m_host, m_host, this, & m_particleBox, &m_particles);

    DensityEditor::createVolumeBoxes
      (clow, cspan, cmove, m_host->densityEditor()->volumes(), 
       this, &m_widgets, this, m_host, & m_cubes);
  
    // Misc sliders:
  
    Fluid::Matrix3 m1, m2;
    m1.rotateX(-0.3);
    m2.rotateY(VEE_HALF_PI);

    m1 = m2 * m1;

    float parammin[SLIDERS] = {
      10.0,
      0.0,
      0.0
    };

    float parammax[SLIDERS] = {
      100.0,
      20.0,
      2.0
    };

    for(i = 0; i < SLIDERS; i++) { 

      Slider * slider = 
	new Slider(m_host, m_currentEmitter->m_params[i],
		   parammin[i], parammax[i]);

      slider->setSounds(m_host->buttonSounds());

      slider->msgAddSink("", parammap[i], this);
    
      slider->setLocation(w1);
      slider->setRotation(m1);
      slider->setSize(2.5, 0.3);
      slider->setLabel(parammap[i]);
      slider->setOrientation(Slider::HORIZONTAL);
      slider->setDisplayValue(false);
    
      m_widgets.addWidget(slider);
      m_sliders[i] = slider;

      w1 += w2;
    }

    for(i=0; emissions[i]; i++) {
      TextureButton * txb = new TextureButton(m_host);
    
      txb->setArea(b1, b3);
      txb->setSounds(m_host->buttonSounds());
      txb->setLabel(emissions[i]);
    
      sprintf(buf, "emission-%u", i);
      txb->msgAddSink("press", buf, this);
    
      m_emissions.push_back(txb);

      b1 += b2;

      m_widgets.addWidget(txb);
    }

    selectEmitter(0);
  }

  EmitterEditor::~EmitterEditor()
  {}

  bool EmitterEditor::msgProcess(const char *id, Fluid::Message *m)
  {
    VEE_trace("EmitterEditor::msgProcess # %s %p", id, m);
  
    bool ok = false;

    if(strncmp(id, "emitter_", 8) == 0) {
      uint index = atoi(id + 8);
      selectEmitter(index);
      ok = true;
    }
    else if(strncmp(id, "particle_", 9) == 0) {

      uint index = atoi(id + 9);
      selectParticle(index);
      ok = true;
    }
    else if(strcmp(id, "rate") == 0) {

      Fluid::MessageFloat * mf = dynamic_cast<Fluid::MessageFloat *> (m);

      if(mf) {
	float r = mf->value();
	m_currentEmitter->m_params[RATE] = r;
	ok = true;
      }
    }
    else if(strcmp(id, "speed") == 0) {

      Fluid::MessageFloat * mf = dynamic_cast<Fluid::MessageFloat *> (m);

      if(mf) {
	float s = mf->value();
	m_currentEmitter->m_params[SPEED] = s;
	ok = true;
      }
    }
    else if(strcmp(id, "spread") == 0) {

      Fluid::MessageFloat * mf = dynamic_cast<Fluid::MessageFloat *> (m);

      if(mf) {
	float s = mf->value();
	m_currentEmitter->m_params[SPREAD] = s;
	ok = true;
      }
    }
    else if(strncmp(id, "emission-", 9) == 0) {

      int emissionType = atoi(id + 9);

      ok = selectEmissionType(emissionType);
    }
    else if(strncmp(id, "volume_", 7) == 0) {

      int volume = atoi(id + 7);

      selectVolume(volume);

      ok = true;
    }

    return ok;
  }

  void EmitterEditor::render(GlData *glData)
  {
    glDisable(GL_BLEND);
  
    m_widgets.renderGl(glData);

    glEnable(GL_BLEND);

    ParticleEditor::renderParticleSelectors(&m_particles, glData);

    m_host->densityEditor()->renderVolumes(glData, &m_cubes);
  }

  void EmitterEditor::update()
  {
    m_widgets.update();
  }

  void EmitterEditor::selectParticle(int index)
  {
    VEE_trace("EmitterEditor::selectParticle # %d", index);  
    m_currentEmitter->m_particle = index;

    for(uint i=0; i < m_particles.elements(); i++) {
      (m_particles.begin() + i)->m_button->setColors(m_host->usualColor());
    }
    (m_particles.begin() + index)->m_button->setColors(m_host->selectedColor());
  }

  void EmitterEditor::selectEmitter(int index)
  {
    VEE_trace("EmitterEditor::selectEmitter # %d", index);

    EmitterData * ed = m_emitters.begin() + m_selectedEmitter;
  
    m_currentEmitter->m_button->setColors(m_host->usualColor());
  
    m_selectedEmitter = index;
    ed = m_emitters.begin() + index;
    m_currentEmitter = ed;
    selectVolume(ed->m_volume);
    ed->init(this);

    ed->m_button->setColors(m_host->selectedColor());
    selectParticle(ed->m_particle);

    updateSliders();
  }

  uint EmitterEditor::emitterParticle(uint emitter)
  {
    if(emitter >= m_emitters.elements())
      VEE_fatal("EmitterEditor::emitterParticle # out of range");

    return (m_emitters.begin() + emitter)->m_particle;
  }

  VEE_Emitter * EmitterEditor::createEmitter(uint i)
  {
    uint row = i / m_emitters.cols();
    uint col = i % m_emitters.cols();

    if(row >= m_emitters.rows() || col >= m_emitters.cols()) {
      VEE_error("EmitterEditor::createEmitter # %u exceeds range", i);
    }
    EmitterData * ed = & m_emitters.get(row, col);
    ed->init(this);
  
    return (VEE_Emitter *) ed->m_emitter.clone();
  }

  void EmitterEditor::getEmitterParameters
  (uint i, std::vector<float> * v)
  {
    EmitterData * ed = m_emitters.begin() + i;

    v->resize(SLIDERS);
    for(uint i=0; i < SLIDERS; i++) {
      (*v)[i] = ed->m_params[i];
    }
  }

  void EmitterEditor::updateSliders()
  {
    for(uint i=0; i < SLIDERS; i++)
      m_sliders[i]->setValue(m_currentEmitter->m_params[i]);
  }

  bool EmitterEditor::selectEmissionType(uint index)
  {
    const char * fname = "EmitterEditor::selectEmissionType";

    VEE_trace("%s # %u", fname, index);

    if(index >= m_emissions.size())
      return false;

    ColorGroup cg = m_host->usualColor();
  
    for(uint i=0; i < m_emissions.size(); i++)
      m_emissions[i]->setColors(cg);

    m_emissions[index]->setColors(m_host->selectedColor());
  
    m_host->engine()->setNextRecorder((Engine::Recorder) index,
				      emissionLabels[index]);
  
    return true;
  }

  void EmitterEditor::selectVolume(int vol)
  {
    m_currentEmitter->m_volume = vol;

    for(uint i=0; i < m_cubes.size(); i++)
      m_cubes[i]->setColors(m_host->usualColor());

    m_cubes[vol]->setColors(m_host->selectedColor());
  }


  QDomElement EmitterEditor::writeDom(QDomDocument &doc)
  {
    QDomElement elem = doc.createElement("Emitters");

    char buf[64];

    for(uint i=0; i < m_emitters.elements(); i++) {
      sprintf(buf, "emitter_%u", i);
    
      QDomElement tmp = doc.createElement(buf);
    
      tmp.appendChild((m_emitters.begin() + i)->writeDom(doc));
      elem.appendChild(tmp);
    }

    return elem;
  }

  bool EmitterEditor::readDom(QDomElement &elem)
  {
    if(elem.tagName() != "Emitters") {
      VEE_error("EmitterEditor::readDom # bad tag (%s)", 
		elem.tagName().ascii());
      return false;
    }
  
    QDomNode n = elem.firstChild();
    while ( !n.isNull() ) {
      if ( n.isElement() ) {
	QDomElement e = n.toElement();
	const char * tag = e.tagName().ascii();

	if(strncmp(tag, "emitter_", 8) == 0) {
	  int index = atoi(tag + 8);
	  if(index < (int) m_emitters.elements()) {
	    QDomElement e2 = e.firstChild().toElement();
	  
	    (m_emitters.begin() + index)->readDom(e2);
	  }
	}
      }
      n = n.nextSibling();
    }
  
    selectEmitter(0);
  
    return true;
  }

}

