/* -*- C++ -*- */

/* -*- C++ -*- */

/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Crystal animation and 3D modeling application
 *
 * Authors:   Tommi Ilmonen 2001-2004
 * Copyright: Tommi Ilmonen and/or Tekes 2001-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <crystal_emitter_recorder.h>

#include <crystal_audio_action.h>
#include <crystal_action_creator.h>
#include <crystal_action_list.h>
#include <crystal_dom_util.h>
#include <crystal_emitter_editor.h>
#include <crystal_engine.h>
#include <crystal_remove_action.h>
#include <crystal_tools.h>
#include <crystal_visualize_location.h>
#include <crystal_visualize_particles.h>

#include <vee_error.h>

#include <vee_scene.h>
#include <vee_strs.h>

#include <Fluid/Fistiness.hpp>
#include <Fluid/TrackerSensorSelector.hpp>

#include <ConfigReader.hpp>

namespace Crystal { 

  static bool debugEmitter = true;

  void CreateEmitterAction::act(ActionList *al)
  {
    if(debugEmitter)
      VEE_trace("CreateEmitterAction::act # %d %d %d", 
		m_id, m_emitterType, al->getScene()->totalForces());
  
    EmitterEditor * editor = al->engine()->tools()->emitterEditor();
  
    VEE_Emitter * emitter = editor->createEmitter(m_emitterType);
  
    if(!emitter) 
      VEE_fatal("CreateEmitterAction::act # NULL emitter");
  
    emitter->setLocation(m_location);
  
    al->setObject(m_id, emitter);
    al->getScene()->addObject(emitter);
  
    if(debugEmitter)
      VEE_trace("CreateEmitterAction::act # DONE %d %d", 
		m_id, al->getScene()->totalForces());
  }

  QDomElement CreateEmitterAction::writeDom(QDomDocument &doc)
  {
    QDomElement elem = doc.createElement("CreateEmitterAction");

    writeDomTimeStamp(elem, doc);
    
    addTextChild(elem, doc, "location", m_location);
    addTextChild(elem, doc, "id", m_id);
    addTextChild(elem, doc, "emitter", m_emitterType);
  
    return elem;
  }

  bool CreateEmitterAction::readDom(QDomElement &elem)
  {
    if(elem.tagName() != "CreateEmitterAction")
      return false;

    readDomTimeStamp(elem);

    m_location = getVector3(elem, "location");
    m_id = getInt(elem, "id", -2);
    m_emitterType = getInt(elem, "emitter", -2);

    return true;
  }


  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  static std::string locStr("location");
  static std::string rotStr("childops/volume/volume/rotation");
  static std::string rateStr("rate");
  static std::string speedStr("childops/speed/radius");
  static std::string spreadStr("childops/spread/radius");

  EmitterRecorder::EmitterRecorder()
    : m_emitterId(-1),
      m_audioSourceId(-1),
      m_noteOnActionId(-1),
      m_trackerCount(0),
      m_gloveCount(0),
      m_audioLocator(0)
  {}

  EmitterRecorder::~EmitterRecorder()
  {}

  void EmitterRecorder::prepare(ActionList *)
  {
    // Engine *engine = al->engine();
  }

  void EmitterRecorder::beginRecording(ActionList *al, 
				       const Fluid::TimeStamp &now)
  {
    const char * fname = "EmitterRecorder::beginRecording";

    if(debugEmitter)
      VEE_trace("%s # %p %lf", fname, this, now.secondsD());
  
    Engine *engine = al->engine();
    m_trackerCount = (long) engine->tracker()->outputBuffer().getSampleCount();
    m_gloveCount = (long) engine->glove()->outputBuffer().getSampleCount();

    Fluid::TrackerSensorSelector2 * handLoc = engine->weakHandLoc();

    Fluid::Vector3 location = handLoc->locations().getNewest();
    Matrix3 rotation;
  
    m_emitterId = al->allocateObjectId();

    VisualizeLocation *vl = new VisualizeLocation();
    vl->set(m_emitterId, locStr);
    al->setVisualize(vl);
  
    uint se = engine->tools()->emitterEditor()->selectedEmitter();
  
    al->appendAction(new CreateEmitterAction
		     (m_emitterId, 
		      se,
		      location, 
		      now));
    
    

    engine->tools()->emitterEditor()->getEmitterParameters(se, & m_params);

    AudioActionGranular * noteon = 
      new AudioActionGranular(now);
  
    m_audioSourceId = al->allocateAudioId();
    noteon->set(m_audioSourceId, "grain.wav", 1.5);
  
    al->appendAction(noteon);
    m_noteOnActionId = al->addMappedAction(noteon);
  
    ValueGraphs * graphs = & al->getGraph(m_emitterId);

    { 
      Fluid::IONodeBuffer<Fluid::Matrix3> * handRot = 
	& engine->weakHandLoc()->rotations();
      
      Fluid::IONodeBuffer<Fluid::Vector3> * handLoc = 
	& engine->weakHandLoc()->locations();
    
      location = handLoc->getIndex(m_trackerCount - 1);
      rotation = handRot->getIndex(m_trackerCount - 1);

      Fluid::TimeStamp sampleTime = handLoc->getTimeStamp(m_trackerCount - 1);
      Fluid::TimeStamp ts = al->realTimeToRecordTime(sampleTime);
    
      double dts = ts.secondsD();
      graphs->appendVector3(locStr, dts, location);
      graphs->appendMatrix3(rotStr, dts, rotation);
    }
  
    { 
      Fluid::Fistiness  *fistiness = engine->fistiness();
  
      Fluid::IONodeBuffer<float> * values = & fistiness->buffer();
    
      float fisty = 1.0 - values->getIndex(m_gloveCount);

      float rate = fisty * m_params[EmitterEditor::RATE];
    
      Fluid::TimeStamp sampleTime = 
	values->getTimeStamp(m_gloveCount);
    
      Fluid::TimeStamp ts = al->realTimeToRecordTime(sampleTime);
    
      double dts = ts.secondsD();
    
      graphs->appendFloat(rateStr, dts, rate);
    }

    m_audioLocator = 
      new AudioActionSourceLocation2(m_audioSourceId,  
				     m_emitterId, now); 
    // Make sure the action is not wiped out accidentally.
    TimeStamp tmpts;
    tmpts.setSeconds(1.0e+8);
    m_audioLocator->setEndTimeStamp(tmpts);

    al->appendAction(m_audioLocator);    

    m_rateControl = 
      new AudioActionControl(m_noteOnActionId, 
			     m_emitterId, 
			     rateStr,
			     rateStr,
			     now,
			     new MJ_MapperMultiplyLimit(5.0, 10.0, 5000.0));
			
    m_rateControl->setEndTimeStamp(tmpts);

    al->appendAction(m_rateControl);  		   

    m_mode = RECORD;

    if(debugEmitter)
      VEE_trace("EmitterRecorder::beginRecording # EXIT");
  }

  void EmitterRecorder::endRecording(ActionList *al, 
				     const Fluid::TimeStamp &now)
  {
    VEE_trace("EmitterRecorder::endRecording # %lf", now.secondsD());
  
    al->appendAction(new RemoveAction(m_emitterId, now));
    al->appendAction(new AudioActionNoteoff(m_noteOnActionId, now));

    m_audioLocator->setEndTimeStamp(now - Fluid::TimeStamp(1));
    m_rateControl->setEndTimeStamp(now - Fluid::TimeStamp(1));

    m_audioLocator = 0;

    m_mode = MUTE;
  }

  void EmitterRecorder::record(ActionList *al, 
			       const Fluid::TimeStamp &now)
  {  
    const char * fname = "EmitterRecorder::record";

    if(m_mode != RECORD)
      return;

    /* if(debugEmitter)
       VEE_trace("%s # %p", fname, this); */

    Engine *engine = al->engine();

    Fluid::IONodeBuffer<Fluid::Vector3> * handLoc = 
      & engine->weakHandLoc()->locations();

    VEE_Vector3 location;

    ValueGraphs * graphs = & al->getGraph(m_emitterId);

    // Get location:

    for(; m_trackerCount < (long)handLoc->getSampleCount(); m_trackerCount++) {
	    
      location = handLoc->getIndex(m_trackerCount);

      Fluid::TimeStamp sampleTime = handLoc->getTimeStamp(m_trackerCount);
      Fluid::TimeStamp ts = al->realTimeToRecordTime(sampleTime);
    
      double dts = ts.secondsD();
      graphs->appendVector3(locStr, dts, location);
    }

    // Get params from glove

    Fluid::Fistiness  *fistiness = engine->fistiness();

    Fluid::IONodeBuffer<float> * values = & fistiness->buffer();

    for(; m_gloveCount < (long) values->getSampleCount(); m_gloveCount++) {

      float fist = 1.0 - values->getIndex(m_gloveCount);

      float rate = fist * m_params[EmitterEditor::RATE];
      float speed = fist * m_params[EmitterEditor::SPEED];
      float spread = fist * m_params[EmitterEditor::SPREAD];
    
      Fluid::TimeStamp sampleTime = 
	values->getTimeStamp(m_gloveCount);

      Fluid::TimeStamp ts = al->realTimeToRecordTime(sampleTime);
    
      double dts = ts.secondsD();

      graphs->appendFloat(rateStr, dts, rate);
      graphs->appendFloat(speedStr, dts, speed);
      graphs->appendFloat(spreadStr, dts, spread);

      double tdiff = sampleTime  > now ?
	(sampleTime - now).secondsD() : - (now - sampleTime).secondsD(); 

      VEE_trace("%s # rate = %f, time = %.3lf, tdiff = %.3lf", 
		fname, rate, dts, tdiff);
    }

    // VEE_trace("%s # EXIT", fname);
  }

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  EmitterRecorder2::EmitterRecorder2()
    : m_emitter(0),
      m_trackerCount(0)
  {
    m_prevTime.getTimeOfDay();
  }

  EmitterRecorder2::~EmitterRecorder2()
  {}

  void EmitterRecorder2::prepare(ActionList *)
  {}

  void EmitterRecorder2::beginRecording(ActionList *al, 
					const Fluid::TimeStamp &)
  {
    const char * fname = "EmitterRecorder2::beginRecording";

    if(debugEmitter)
      VEE_trace("%s # %p", fname, this);
  
    Engine *engine = al->engine();
    EmitterEditor *ee = engine->tools()->emitterEditor();
    m_trackerCount = engine->trackerSamples();
    Fluid::TrackerSensorSelector2 * handLoc = engine->weakHandLoc();

    Fluid::Vector3 location = handLoc->locations().getNewest();
    m_emitter = ee->createSelectedEmitter();
    m_particleId = ee->emitterParticle(ee->selectedEmitter());
    m_emitter->setLocation(location);
    uint se = ee->selectedEmitter();
    ee->getEmitterParameters(se, & m_params);

    al->getScene()->addObject(m_emitter);

    m_emitter->setAccumulate(true);
    m_emitter->setRate(0.0);

    VisualizeParticles * vp = new VisualizeParticles();
    al->setVisualize(vp);

    m_mode = RECORD;

    // m_prevTime.getTimeOfDay();
  }

  void EmitterRecorder2::endRecording(ActionList *al, 
				      const Fluid::TimeStamp &now)
  {
    AudioActionSample * noteon = 
      new AudioActionSample(now);
  
    noteon->set(-1, al->engine()->config()->get("Sounds").
		get("static_particles").getString().c_str(), 0.2, m_emitter->location());

    al->appendAction(noteon);

    if(m_emitter->sceneBox()) {
      m_emitter->sceneBox()->removeChildObject(m_emitter);
      m_emitter->setSceneBox(0);
    }
    delete m_emitter;

    m_mode = MUTE;
  }

  void EmitterRecorder2::record(ActionList *al, 
				const Fluid::TimeStamp &now)
  {
    const char * fname = "EmitterRecorder2::record";

    if(m_mode != RECORD)
      return;

    VEE_trace("%s # Enter", fname);

    Fluid::IONodeBuffer<Fluid::Vector3> * handLoc = 
      & al->engine()->weakHandLoc()->locations();

    // Get location:

    float sampleDur = 1.0f / handLoc->getSampleRate();

    { 
      Engine * engine = al->engine();

      // Set some parameters:
      Fluid::Fistiness  *fistiness = engine->fistiness();
    
      Fluid::IONodeBuffer<float> * values = & fistiness->buffer();
    
      float fist = 1.0 - values->getNewest();
    
      float rate = fist * m_params[EmitterEditor::RATE];
      float speed = fist * m_params[EmitterEditor::SPEED];
      float spread = fist * m_params[EmitterEditor::SPREAD];
    
      m_emitter->setValueFloat(rateStr.c_str(), rate);
      m_emitter->setValueFloat(speedStr.c_str(), speed);
      m_emitter->setValueFloat(spreadStr.c_str(), spread);
    }

    for(; m_trackerCount < (long)handLoc->getSampleCount(); m_trackerCount++) {

      m_emitter->setLocation(handLoc->getIndex(m_trackerCount));

      m_emitter->update(sampleDur);

      std::vector<VEE_Object *> * particles = & m_emitter->accumulated();
    
      /* This may look strange. First we create the particles with the
	 creator. Then we clone the objects to the list creator
	 objects. Then we remove the objects from scene since they
	 should not be aged etc, but they are private property of the
	 list creator.  :-| */

      for(uint i=0; i < particles->size(); i++) {

	VEE_Object * o = (*particles)[i];
	ActionCreator *ac = new ActionCreator
	  (o->location(), o->velocity(), m_particleId, now);
	al->appendAction(ac);

	if(o->sceneBox()) {
	  o->sceneBox()->removeChildObject(o);
	  o->setSceneBox(0);
	}
      }

      particles->clear();
    }

    VEE_trace("%s # Exit", fname);
  }

}
