/* -*- C++ -*- */

/* -*- C++ -*- */

/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Crystal animation and 3D modeling application
 *
 * Authors:   Tommi Ilmonen 2001-2004
 * Copyright: Tommi Ilmonen and/or Tekes 2001-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <crystal_engine.h>

#include <crystal_ac.h>
#include <crystal_action_list.h>
#include <crystal_blow_recorder.h>
#include <crystal_component_editor.h>
#include <crystal_gee_mesh.h>
#include <crystal_gee_config.h>
#include <crystal_emitter_editor.h>
#include <crystal_emitter_recorder.h>
#include <crystal_explosion_recorder.h>
#include <crystal_gl_data.h>
#include <crystal_log.h>
#include <crystal_mesh_recorder.h>
#include <crystal_scene_objs.h>
#include <crystal_vortex_recorder.h>
#include <crystal_tools.h>

#include <Fluid/Fistiness.hpp>
#include <Fluid/Glove.hpp>
#include <Fluid/GloveNavigator.hpp>
#include <Fluid/GloveRecognizer.hpp>
#include <Fluid/InputDevices.hpp>
#include <Fluid/Math.hpp>
#include <Fluid/MessageInt.hpp>
#include <Fluid/MessageMouse.hpp>
#include <Fluid/MessageString.hpp>
#include <Fluid/MotionNavigator.hpp>
#include <Fluid/MessageMouse.hpp>
#include <Fluid/MouseButtons.hpp>
#include <Fluid/MouseNavigator.hpp>
#include <Fluid/NapraToMouse.hpp>
#include <Fluid/SensorContainment.hpp>
#include <Fluid/SensorHitDetector.hpp>
#include <Fluid/Tracker.hpp>
#include <Fluid/TrackerSensorSelector.hpp>

#include <Solar/Model.hpp>
#include <Solar/RotatingObject.hpp>

#include <gee_debug.h>
#include <gee_engine.h>

#include <vee_blow.h>
#include <vee_color_table.h>
#include <vee_damping.h>
#include <vee_error.h>
#include <vee_gl_renderer.h>
#include <vee_gl_texture.h>
#include <vee_gravity.h>
#include <vee_grid_force.h>
#include <vee_magic_rain.h>
#include <vee_scene.h>
#include <vee_volume_ellipsoid.h>

#include <ac_attenuation.h>
#include <ac_control.h>
#include <ac_vr_control.h>

#include <ConfigReader.hpp>
 
#include <qfile.h> 

#include <strings.h>
#include <qtextstream.h> 

namespace Crystal { 

  static std::string pyramidName("pyramid");

  static DI_MutexAuto traceMutex;

  AC_VrControl * audio = 0;

  using namespace Fluid;

  static void traceLock(bool lock)
  {
    if(lock) traceMutex.lock();
    else traceMutex.unlock();
  }

  bool Engine::m_debug = false;
  Engine * Engine::m_engine = 0;
  static bool m_superDebug = false;

  Engine::Engine(VEE_SceneBox *scene,
		 const std::string & configFile )
    : m_done(false),
      m_renderSolars(false),
      m_updates(0),
      m_scene(scene),
      m_gravity(new VEE_Gravity(Fluid::Vector3(0.0, -1.0, 0.0))),
      m_isRealTime(true),
      m_showForces(true),
      m_actionLists(new ActionListManager(this)),
      m_tracker(0),
      m_handSensor1(-1),
      m_handSensor2(-1),
      m_headSensor(-1),
      m_glove(0),
      m_fistiness(0),
      m_clapDetector(0),
      m_navigator(0),
      m_navigatorLoc(0),
      m_navigatorRot(0),
      m_navigatorScale(0),
      m_strongHandLoc(0),
      m_weakHandLoc(0),
      m_headLoc(0),
      m_acControl(new AC_Control),
      m_acVrControl(new AC_VrControl(m_acControl)),
      m_particleVolume(0.3),
      m_soundTrackVolume(0.3),
      m_centralModel(0),
      m_centralObject(0),
      m_mode(MODE_TOOLS),
      m_nextRecorder(RECORDER_EMITTER),
      m_recorder(RECORDER_EMITTER),
      m_tools(0),
      m_forceToolsVisible(false),
      m_configFile(configFile),
      m_config(new ConfigReader::Config),
      m_geeEngine(new GEE_Engine),
      m_geeConfig(new GeeConfig(this)),
      m_rebuildGee(true),
      m_showGroups(true),
      m_curtainTop(0.0),
      m_curtainBottom(0.0),
      m_curtainColor(0.3)
  {
    if(!m_engine)
      m_engine = this;

    m_debug = GEE_Debug::debug("Crystal::Engine");
    m_superDebug = GEE_Debug::debug("Crystal::Engine::super");
    GEE_World::loadTextureDefinitions("vee-gl.config");

    VEE_setTraceStreamLock(&traceLock);
    // ConfigReader::setLockFunction(&traceLock);

    ConfigReader::readConfig(m_config, m_configFile.c_str());

    m_config->dump(std::cout);

    // We always reserve channel 0 for the "system notifications"
    m_audioAllocator.expand(10);
    allocateAudioSource();
    // Allocate two tracks to play the soundtrack from constant locations:
    m_audioTrackIds[0] = allocateAudioSource();
    m_audioTrackIds[1] = allocateAudioSource();
    m_audioNoteIds[1] = m_audioNoteIds[0] = -1;
    m_audioTrack = m_config->get("Audio").get("file").getString();

    m_mouseButtons[0] = m_mouseButtons[1] = 0;
    m_handPointer[0]  = m_handPointer[1]  = 0;
    m_handRender[0]   = m_handRender[1]   = true;

    m_rotation.identity();
    m_location.clear();
    m_locationOld.clear();
    m_scaleOld = 1.0f;
    m_scale = 1.0f;
    m_cursorSize = 1.0f;

    // Initialize audio system:

    VEE_trace("Engine::Engine");

    ConfigReader::Chunk audioChunk = m_config->get("Audio");
  
    std::string soundSocketAddress;// = 
    // audioChunk.get("address").getString("a2:10030");

    bool withAudio = soundSocketAddress.size() > 2;

    VEE_trace("Engine::Engine # Audio: %s", soundSocketAddress.c_str());

    if(withAudio) {
      withAudio = m_acControl->init(soundSocketAddress.c_str());
      if(!withAudio)
	VEE_trace("Engine::Engine # "
		  "Could not init sound controller !!");
    }
    
    if(!withAudio) {
      delete m_acControl;
      m_acControl = 0;
      
      delete m_acVrControl;
      m_acVrControl = 0;
    }
    else {
      AC_AttenuationPower * att = new AC_AttenuationPower(0.7);
      att->setMinMaxDist(4.0, 20.0);
      m_acVrControl->setDefaultAttenuation(att);
      // Locate the sound track.
      m_acControl->changeSourceOrientation(m_audioTrackIds[0],
					   0.1, -45.0, 0.0);
      m_acControl->changeSourceOrientation(m_audioTrackIds[1],
					   0.1,  45.0, 0.0);
    }

    VEE_trace("Engine::Engine # 2");

    audio = m_acVrControl;

    // Load scene if it seems to be available
    m_config->dump(std::cout);
    
    for(int i=0; i < 2; i++) {
      GeeRecorders * rs = new GeeRecorders;
      rs->init(this, m_geeEngine);
      rs->setHand((Hand) i);
      m_recorders[i] = rs;

      m_currentGeeRecorder[i] = 0;
    }

    for(int scene = 0; scene >= 0; scene++) {

      char buf[64];

      sprintf(buf, "Scene-%d", scene);
    
      ConfigReader::Chunk modelChunk = m_config->get(buf);
    
      if(modelChunk.size())
	loadScene(buf, & modelChunk);
      else {
	VEE_trace("Engine::Engine # No Scene chunk %d", scene);
	scene = -2;
      }
    }

    if(m_scenes.size())
      useScene(0);

    m_geeEngine->setGarbageBin(&m_veeGarbage);

    VEE_trace("Engine::Engine # EXIT");
  }

  Engine::~Engine()
  {
    delete m_acVrControl;
    delete m_acControl;
    delete m_centralObject;
    delete m_tools;
    delete m_geeEngine;

    if(m_engine == this)
      m_engine = 0;
  
    audio = 0;
  }
  
  static std::string mouseStr("mouse");
  static std::string trackerStr("tracker");
  static std::string strongHandStr("left-hand");
  static std::string weakHandStr("right-hand");
  static std::string headStr("head");
  static std::string gloveStr("left-glove");

  void Engine::init()
  {
    const char * fname = "Engine::init";

    if(m_debug) VEE_trace(fname);

    Log::init();
    Log::log("INIT");

    int i;

    ConfigReader::Config conf = *m_config;
  
    /* Add gravity to the scene. */
    m_scene->addNewForce(m_gravity);
    m_gravity->setVisible(false);

    VEE_trace("%s # Getting tracker %s", fname, trackerStr.c_str());
    m_tracker = Fluid::InputDevices::getTracker(trackerStr);
    
    VEE_trace("%s # Got tracker %p", fname, this);

    m_glove = Fluid::InputDevices::getGlove(gloveStr);

    m_trackerSamples = 0;
    // m_glove->start();

    if(!m_tracker)// || !m_glove)
      VEE_fatal("Engine::init # NULL tracker \"%s\"", trackerStr.c_str());

    std::string indevsPath("../inputdevices/");

    // Napra to mouse conversion:
    NapraToMouse * n2m[2];
    
    n2m[0] = new NapraToMouse(&m_processors, 
			      "mousedata",
			      indevsPath + "napra/output",
			      0);

    n2m[1] = new NapraToMouse(&m_processors, 
			      "mousedata2",
			      indevsPath + "napra/output",
			      1);

    for(i = 0; i < 2; i++) {
      if(!n2m[i]->start()) {
	VEE_trace("%s # Failed to create NapraToMouse", fname);
	delete n2m[i];
	n2m[i] = 0;
      }
    }


    VEE_trace("%s # Preparing to load mouse", fname);

    m_mouse = Fluid::InputDevices::getMouse(mouseStr);

    std::string buttonsName[2] = { 
      std::string("mousebuttons"), 
      std::string("mousebuttons2") 
    };
    
    {
      // Mouse button detector.
      std::string mouseNode[2];

      if(!n2m[0]) 
	mouseNode[0] = indevsPath + mouseStr + "/output";
      else {
	mouseNode[0] = "mousedata/output";
	mouseNode[1] = "mousedata2/output";
      }

      for(i = 0; i < 2; i++) {

	VEE_trace("%s # Preparing to load mouse button %d", fname, i);

	if(!mouseNode[i].size())
	  continue;

	MouseButton * mb = 
	  new Fluid::MouseButton(&m_processors, buttonsName[i].c_str(), 
				 mouseNode[i]);

	mb->start();

	if(!mb->ok()) {
	  VEE_error("%s # Failed to create mouse detector \"%s\"", 
		    fname, mouseNode[i].c_str());
	  delete mb;
	  mb = 0;
	}
	else {
	  // mouseButtons->addEventListener(this);
	  if(i == 0)
	    mb->msgAddSink("", "mouse", this);
	  else if(i == 1)
	    mb->msgAddSink("", "mouse-2", this);
	  m_processors.addNode(mb);
	}
	m_mouseButtons[i] = mb;
      }
    }
    
    std::string navigatorType = 
      m_config->get("Navigator").get("type").getString();

    if(navigatorType == "mouse" || navigatorType.size() == 0) {
      // Create mouse navigator.
      MouseNavigator * navi = new MouseNavigator(&m_processors, "navigator",
						 strongHandStr  + "/location",
						 weakHandStr + "/location",
						 strongHandStr  + "/rotation",
						 buttonsName[0],
						 2);
      m_processors.addNode(navi); 
      
      m_navigatorLoc = & navi->location();
      m_navigatorRot = & navi->rotation();
      m_navigatorScale = & navi->scale();
      m_navigator = navi;
    }
    else if(navigatorType == "motion") {

      MotionNavigator * navi = 
	new MotionNavigator(&m_processors, "navigator",
			    headStr + "/location", 
			    Vector3(0, 0, 0),
			    Math::PI * 1.25,
			    Vector3(2, 1, 1.5),
			    m_config->get("Navigator").get("style").getString());
      m_processors.addNode(navi); 

      m_navigatorLoc = & navi->location();
      m_navigatorRot = & navi->rotation();
      m_navigator = navi;
    }
    else {
      VEE_fatal("%s # Unsupportd navigator \"%s\"", 
		fname, navigatorType.c_str());
    }

    // Glove-related stuff.

    VEE_trace("%s # Preparing to load glove", fname);
  
    std::string gloveOutputStr(indevsPath + gloveStr + "/output");
    std::string gloveGesturesStr("glove-gestures");
  
    if(m_glove) {
      VEE_trace("Engine::init building fistiness detector");

      Fluid::FistinessRequest fr("fistiness", gloveOutputStr, 2.0);

      Fluid::ProcessorNode * pnode = m_processors.requestNode(&fr);

      if(pnode) {
	m_fistiness = (Fluid::Fistiness *) pnode;
      }
      else
	VEE_error("%s # Could not create fistiness object", fname);

      // m_fistiness = new Fluid::Fistiness("glove", 2.0);
      // m_processors.addNode(m_fistiness);
    }
    else 
      VEE_error("%s # No GLOVE \"%s\"", fname, gloveStr.c_str());

    m_headSensor = m_tracker->findSensor("head");
    m_handSensor1 = m_tracker->findSensor(strongHandStr);
    m_handSensor2 = m_tracker->findSensor(weakHandStr);
    VEE_trace("Engine::init # Sensor ids = %d, %d, %d", 
	      m_headSensor, m_handSensor1, m_handSensor2);

    if(m_headSensor < 0 || 
       m_handSensor1 < 0 || 
       m_handSensor2 < 0) {
      VEE_error("%s # Tracker config is broken", fname);    
    }

    if(m_glove) {
      // Glove Gestures:
      
      GloveRecognizerRequest request(gloveGesturesStr,
				     "glove-gestures.config", 
				     gloveOutputStr,
				     false);
      ProcessorNode * node = m_processors.requestNode(&request);
    
      GloveRecognizer * gr = dynamic_cast<GloveRecognizer *> (node);
      
      VEE_trace("%s # GloveRecognizer = %p", fname, gr);
    }


    {
      // Sensor selectors:
      Fluid::TrackerSensorSelector *tss1 = new Fluid::TrackerSensorSelector
	(&m_processors, strongHandStr, trackerStr, strongHandStr);
      (void) tss1;

      Fluid::TrackerSensorSelector *tss2 = new Fluid::TrackerSensorSelector
	(&m_processors, weakHandStr, trackerStr, weakHandStr);
      (void) tss2;
    
      Fluid::TrackerSensorSelector *tss3 = new Fluid::TrackerSensorSelector
	(&m_processors, headStr, trackerStr, headStr);
      (void) tss3;
    
      // Hit detector:
      Fluid::SensorHitDetectorRequest shdr(strongHandStr + "/location", 
					   weakHandStr + "/location",
					   0.5, 
					   3.0);
      Fluid::ProcessorNode * pnode = m_processors.requestNode(&shdr);
      m_clapDetector = dynamic_cast<Fluid::SensorHitDetector *> (pnode);

      if(!m_clapDetector) {
	VEE_error("%s # Failed to create clap detector", fname);
      }

      // Transformed tracker sensor selectors:

      if(m_navigator) {
	Vector3 handOffset(0.0, 0.35, 0.0);
	
	std::string naviScale = "navigator/scale";
	
	m_strongHandLoc = new TrackerSensorSelector2
	  (&m_processors, strongHandStr + "_trans", trackerStr, strongHandStr,
	   "navigator/rotation", "navigator/location", naviScale, handOffset);

	m_weakHandLoc = new TrackerSensorSelector2
	  (&m_processors, weakHandStr + "_trans", trackerStr, weakHandStr,
	   "navigator/rotation", "navigator/location", naviScale, handOffset);

	m_headLoc = new TrackerSensorSelector2
	  (&m_processors, headStr + "_trans", trackerStr, headStr,
	   "navigator/rotation", "navigator/location", naviScale, handOffset);

	VEE_trace("%s # selectors are: %p %p %p", fname, m_strongHandLoc, m_weakHandLoc, m_headLoc);
      }

      /* else 
	 m_clapDetector->msgAddSink(this); */
    }


    m_tools = new Tools(this);
    if(m_mouseButtons[0])
      m_mouseButtons[0]->msgAddSink("", "mouse", m_tools);
    m_tools->emitterEditor()->selectEmissionType(0);
    m_tools->start();

    VEE_trace("%s # Preparing to load Solar objects", fname);

    initSolarObjects();
    // initToolBoxes();

    m_timing.m_now.getTimeOfDay();
    m_timing.m_zeroTime = m_timing.m_now;

    VEE_trace("%s # Preparing to start Fluid", fname);

    Fluid::InputDevices::start(m_isRealTime ? 0 : &m_timing.m_now);
    if(!m_processors.start())
      VEE_fatal("%s # m_processors.start() failed", fname);

    m_changeTime.getTimeOfDay();
    m_lastUpdate = m_changeTime;

    m_geeConfig->init();

    centralizeTools();
  }

  void Engine::update(float dt)
  {
    const char * fname = "Engine::update";

    if(m_debug) VEE_trace(fname);

    m_veeGarbage.clear();
    m_deletedGraphics.clear();
    m_geeEngine->collector().clear();

    // Clear any "almost" selected geometries:
    m_geeEngine->world().deNearlySelectAll();

    if(m_isRealTime)
      m_timing.m_now.getTimeOfDay();
    else {
      Fluid::TimeStamp tmp;
      tmp.setSeconds(dt);
      m_timing.m_now += tmp;
    }

    if(dt <= 0.0) {
      VEE_Time now;
      now.getTimeOfDay();
      dt = m_lastUpdate.secsTo(now);
      m_lastUpdate = now;
    }

    if(m_mode == MODE_PLAY && 
       m_timing.m_playSpeed2 > 0.0)
      m_scene->update(40000, 20000, 10, dt * m_timing.m_playSpeed2);
    
    if(m_navigatorLoc)
      m_location = m_navigatorLoc->getNewest();

    if(m_navigatorRot)
      m_rotation = m_navigatorRot->getNewest();

    if(m_navigatorScale)
      m_scale = m_navigatorScale->getNewest();
  
    m_tools->setScale(0.7f / m_scale);

    Solar::Object **so = &* m_objects.begin();
    Solar::Object **sentinel = &* m_objects.end();

    while(so != sentinel) {
      (*so)->update(dt);
      so++;
    }

    if(m_debug) VEE_trace("%s # 1 ", fname);
  
    m_trackerSamples = (int) m_tracker->getLatestIndex();

    m_updates++;

    Vector3 right  = strongHandLoc()->locations().getNewest();
    Vector3 left = weakHandLoc()->locations().getNewest();

    if(m_handPointer[0]) {
      m_handPointer[0]->setLocation(right);
      m_handPointer[0]->setRotation(strongHandLoc()->rotations().getNewest());
      m_handPointer[0]->setScale(m_cursorSize / m_scale);
    }
    if(m_handPointer[1]) {
      m_handPointer[1]->setLocation(left);
      m_handPointer[1]->setRotation(weakHandLoc()->rotations().getNewest());
      m_handPointer[1]->setScale(m_cursorSize / m_scale);
    }

    if(m_debug)
      VEE_trace("%s # right = %f %f %f, left = %f %f %f", fname,
		right.x,  right.y,  right.z,
		left.x, left.y, left.z);

    Fluid::InputDevices::update(m_isRealTime ? 0 : & m_timing.m_now);
    
    /* This method may cause recursive calls to this object. One must be
       careful, the values of some variables may change. */
    m_processors.update();
 
    VEE_Time t;
    t.getTimeOfDay();


    float fistiness = 0.25;

    if(m_fistiness) {
      fistiness = m_fistiness->buffer().getNewest();
      // VEE_trace("%s # Fistiness = %f", fname, fistiness);
    }

    if(m_debug) VEE_trace("%s # 2 ", fname);

    if(m_mode != MODE_PLAY || m_forceToolsVisible) {
      if(m_debug) VEE_trace("%s # 2.5 ", fname);
      m_tools->handleInput();
      m_tools->update();
    }

    if(m_debug) VEE_trace("%s # 3 ", fname);

    float relRate = (0.5 - fistiness);

    if(relRate < 0.05) relRate = 0.05;

    TimeStamp recordTime = m_timing.currentTimeToRecordTime();
  
    // This MUST be before "m_geeEngine->update"
    for(int i = 0; i < 2; i++) {
      if(m_currentGeeRecorder[i])
	m_currentGeeRecorder[i]->update(recordTime);
    }

    if(m_mode == MODE_PLAY) {
      // Add "1" to make sure we get all elements at that time.
      m_actionLists->updateTo(recordTime + TimeStamp(1));
      m_geeEngine->update(m_timing.currentTimeToRecordTime());
      m_timing.updateTime();
    }
	
    // VEE_trace("%s # recTime = %lf", fname, recordTime.secondsD());

    if(m_acVrControl) {
      AC_Vector3 headPos(m_headLoc->locations().getNewest().data());
      AC_Matrix3 headRot(m_headLoc->rotations().getNewest().data());
      AC_Matrix3 worldRot(m_rotation.data());

      worldRot.transpose();

      m_acVrControl->setTransformations(headPos, headRot, worldRot, 0.15);
    }
    
    m_geeEngine->collector().setCameraLocation
      (m_headLoc->locations().getNewest());
    
    if(m_mode != MODE_TOOLS) {

      m_geeEngine->collectObjects(false);
      
      /*
      if(m_recorder == RECORDER_GEE && !currentActiveGeeRecorder()) {
	
	Vector3 diff = right - left;
	
	Vector3 test1 = diff;
	test1.y -= 1.0f;

	Vector3 test2 = diff;
	test2.y += 1.0f;
	
	int quickEditHand = -1;
	
	if(test1.length() < 1.0f)
	  quickEditHand = RIGHT_HAND;

	if(test2.length() < 1.0f)
	  quickEditHand = LEFT_HAND;
	
	if(quickEditHand >= 0)
	  m_tools->useQuickEditor(quickEditHand);
      }
      */
    }

    if(m_debug) VEE_trace("%s # EXIT", fname);
  }

  void Engine::render(GlData *gld,
		      VEE_GlRenderer *veeRenderer, 
		      // const Fluid::Vector3 &camPos,
		      // VEE_GlTexture2D *texture,
		      Solar::Renderer *solarRenderer,
		      bool moveCamera)
  {
    VEE_Time now, now2;
    now.getTimeOfDay();

    const char * fname = "Engine::render";

    static DI_MutexAuto superLock;

    int tid = (int) DI_Thread::myThreadId();

    if(m_superDebug)
      superLock.lock();

    if(m_debug) {
      VEE_trace("%s # %d (%s)", fname, tid, gluErrorString(glGetError()));
    }

    if(!solarRenderer || !veeRenderer)
      Fatal("%s # Missing a renderer", fname);

    if(m_debug) 
      VEE_trace("%s # %d Freelists", fname, tid);

    // gld->m_displayLists.freeLists(& m_deletedGraphics);
    gld->m_geeLists.freeLists(& m_veeGarbage);
    gld->m_geeHighlightLists.freeLists(& m_veeGarbage);
    gld->m_geeRenderer.selectLights(m_geeConfig->lightSet());
    gld->m_geeRenderer.setShowroupSructure(m_showGroups);

    now2.getTimeOfDay();
    gld->m_timings[0] += now.usecsTo(now2);
    now = now2;
    
    const VEE_Rgba &bg = m_geeEngine->world().backgroundColor();

    glClearColor(bg.red(), bg.green(), bg.blue(), bg.alpha());
    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    now2.getTimeOfDay();
    gld->m_timings[1] += now.usecsTo(now2);
    now = now2;


    glPushMatrix();

    if(moveCamera) { 
      // World transform
    
      glPushMatrix();

      Fluid::Vector3 vz;
      Fluid::Matrix3 mi;
      vz.clear();
      mi.identity();

      Fluid::Matrix3 rot = m_rotation;

      rot.transpose();
    
      glScalef(m_scale, m_scale, m_scale);
      GlData::applyGlTransform(rot, vz);
      GlData::applyGlTransform(mi, -m_location);
      // VEE_trace("%s # scale = %f", fname, m_scale);
    }

    if(m_debug) 
      VEE_trace("%s # %d Solids (%s)", fname, tid, 
		gluErrorString(glGetError()));

    // Render solids:
    solarRenderer->prepareToRender();
    gld->lightsOn();

    now2.getTimeOfDay();
    gld->m_timings[2] += now.usecsTo(now2);
    now = now2;

    glDisable(GL_COLOR_MATERIAL);
    for(int hand = 0; hand < 2; hand++)
      if(m_handRender[hand])
	m_handPointer[hand]->render(solarRenderer);

    if(m_renderSolars && false) { /// false -> RRR
      
      for(Solar::Object **obj = &* m_objects.begin(); 
	  obj != &* m_objects.end();
	  obj++) {
	(*obj)->render(solarRenderer);
      }
      
      for(Solar::Object **obj = &* m_aniObjects.begin(); 
	  obj != &* m_aniObjects.end();
	  obj++) {
	(*obj)->render(solarRenderer);
      }
    }

    // Render extra visualizations:
    if(false) /// RRR
      m_actionLists->render(gld);
    
    if(m_debug) 
      VEE_trace("%s # %d to GEE (%s)", fname, tid, 
		gluErrorString(glGetError()));

    now2.getTimeOfDay();
    gld->m_timings[3] += now.usecsTo(now2);
    now = now2;

    for(int i = 0; i < 2; i++)
      if(m_currentGeeRecorder[i])
	m_currentGeeRecorder[i]->render(gld);

    /* Render the GEE stuff. */
    GEE_GlRenderer * geeGl = & gld->m_geeRenderer;
    geeGl->initGlState();
    geeGl->renderFrontToBack(& m_geeEngine->collector().solids());
    geeGl->renderBackToFront(& m_geeEngine->collector().transparencies());
    geeGl->closeGlState();

    if(m_debug) 
      VEE_trace("%s # %d to VEE (%s)", fname, tid, 
		gluErrorString(glGetError()));

    now2.getTimeOfDay();
    gld->m_timings[4] += now.usecsTo(now2);
    now = now2;

    // Render the particles:
    if(moveCamera) {
      Fluid::Vector3 camPos = m_headLoc ? m_headLoc->locations().getNewest() : 
	m_tracker->getLocation(m_headSensor);
    
      veeRenderer->camPos() = camPos;
    }

    veeRenderer->setTexId();
    veeRenderer->prepareGlMode();
    if(m_mode != MODE_TOOLS) {
      veeRenderer->setForceMode
	(m_showForces ? VEE_GlRenderer::RenderTriangles :
	 VEE_GlRenderer::RenderNothing);
      veeRenderer->drawBackToFront(m_scene);
    }
    veeRenderer->endGlMode();

    if(m_mode != MODE_PLAY || m_forceToolsVisible) {
      m_tools->render(gld);
      solarRenderer->prepareToRender();
    }

    if(moveCamera)
      glPopMatrix();

    glPopMatrix();

    if(m_mode == MODE_PLAY && (m_curtainTop != 0.0 || m_curtainBottom != 0.0)){

      glMatrixMode (GL_PROJECTION);
      
      glPushMatrix();

      glLoadIdentity ();
      gluOrtho2D(0.0, 1.0, 0.0, 1.0);
      
      glMatrixMode (GL_MODELVIEW);
      glPushMatrix();
      glLoadIdentity ();
      
      glDisable(GL_LIGHTING);
      float color = m_curtainColor;
      glColor3f(color, color, color);
      
      glPushAttrib(~0);

      glDisable(GL_CULL_FACE);
      glDepthFunc(GL_ALWAYS);
      glBegin(GL_QUADS);

      glVertex2f(0.0, 0.0);
      glVertex2f(1.0, 0.0);
      glVertex2f(1.0, m_curtainBottom);
      glVertex2f(0.0, m_curtainBottom);

      float top = 1.0 - m_curtainTop;
      
      glVertex2f(0.0, top);
      glVertex2f(1.0, top);
      glVertex2f(1.0, 1.0);
      glVertex2f(0.0, 1.0);
      
      glEnd();

      glPopAttrib();

      glPopMatrix();
      
      glMatrixMode (GL_PROJECTION);
      glPopMatrix();
      
      glMatrixMode (GL_MODELVIEW);      
    }

    if(m_debug) 
      VEE_trace("%s # %d EXIT (%s)", fname, tid, gluErrorString(glGetError()));

    now2.getTimeOfDay();
    gld->m_timings[5] += now.usecsTo(now2);

    if(m_superDebug)
      superLock.unlock();
  }

  bool Engine::msgProcess(const char *id, Fluid::Message *m)
  {
    const char * fname = "Engine::msgProcess";

    assert(id && m);

    const MessageMouse *mm = 
      dynamic_cast<const Fluid::MessageMouse *> (m);
    const MessageString *ms = 
      dynamic_cast<const Fluid::MessageString *> (m);

    if(strcmp(id, "load") == 0) {
      if(!ms)
	return false;
      return readDomFile(ms->string().c_str());
    }
    else if(strcmp(id, "save") == 0) {
      if(!ms)
	return false;
      return writeDomFile(ms->string().c_str());
    }
    
    /* const Fluid::MessageInt *mi = 
       dynamic_cast<const Fluid::MessageInt *> (m); */

    VEE_trace("%s # \"%s\" %s", fname, id, typeid(*m).name());

    if(mm) {
      if(strcmp(id, "mouse") == 0) {
	if(m_mode == MODE_PLAY) {
	  if(mm->pressed(0)) {
	    VEE_trace("%s # Start recording", fname);
	    m_actionLists->continueRecording(m_timing.m_zeroTime, 
					     m_timing.realTimeToRecordTime
					     (m_timing.m_now));
	  }
	  else if(mm->released(0)) {
	    VEE_trace("%s # Pause recording", fname);
	    m_actionLists->pauseRecording(m_timing.realTimeToRecordTime
					  (m_timing.m_now));
	  }
	}
	
	if(mm->pressed(1) && !m_tools->grabs()->isGrabbed()) {
	  skipMode((m_mode == MODE_TOOLS) ? MODE_PLAY : MODE_TOOLS);
	}
	
	return true;
      }
      else if(strcmp(id, "mouse-2") == 0) {

	if(m_mode == MODE_PLAY) {
	  
	  if(mm->pressed(1))
	    m_tools->useQuickEditor(RIGHT_HAND);
	  else if(mm->pressed(2))
	    m_tools->useQuickEditor(LEFT_HAND);
	}
	else if(m_mode == MODE_QUICK_EDIT) {
	  if(mm->pressed(1) || mm->pressed(2))
	    skipMode(MODE_PLAY);
	}
      }
    }

    return false;
  }

  int Engine::allocateAudioSource() 
  { 
    int id = m_audioAllocator.allocateChannel(); 

    VEE_trace("Engine::allocateAudioSource # Allocated %d", id);

    return id;
  }

  void Engine::deallocateAudioSource(int id)
  { 
    VEE_trace("Engine::deallocateAudioSource # Deallocating %d", id);
    m_audioAllocator.deallocateChannel(id); 
  }

  bool Engine::loadScene(const char *name, ConfigReader::Chunk * config)
  {
    SceneObjs * so = new SceneObjs;

    so->loadScene(name, config);
  
    if(!so->model()) {
      delete so;
      VEE_error("Engine::loadScene # no model");
      return false;
    }
  
    m_models[name] = so->model();
    so->model()->setGarbageBin(& m_deletedGraphics);

    m_scenes.push_back(so);

    return true;
  }

  void Engine::setGravity(double g)
  {
    m_gravity->setValueVector3("gravity", Fluid::Vector3(0.0, -1.0, 0.0) * g);
  }


  void Engine::skipMode(Mode newMode, bool forceToolsVisible)
  {
    const char * fname = "Engine::skipMode";

    static const char * modeNames[MODE_SIZEOF] = {
      "TOOLS",
      "PLAY"
    };

    Log::log("SKIP_MODE %s %s", 
	     modeNames[newMode], m_nextRecorderName[0].c_str());
  
    VEE_trace("%s # %d %d", fname, (int) newMode, (int) forceToolsVisible);
    
    m_forceToolsVisible = forceToolsVisible;

    m_showGroups = m_geeConfig->getValueInt("showgroups", 1);

    float speed  = m_geeConfig->getValueFloat("timescale", 1.0);
    float offset = m_geeConfig->getValueFloat("start", 0.0);

    VEE_trace("%s # %f %f", fname, speed, offset);

    setPlaySpeed(speed);
    setOffset(offset);

    {
      m_curtainTop    = m_geeConfig->getValueFloat("topcurtain", 0.0);
      m_curtainBottom = m_geeConfig->getValueFloat("bottomcurtain", 0.0);
      m_curtainColor  = m_geeConfig->getValueFloat("curtaincolor", 0.3);
    }

    if(m_acControl) {
      m_acControl->allNotesOff(true);
    }
    
    m_aniObjects.clear();

    m_audioAllocator.clearAllocations();
    allocateAudioSource();
    // Allocate two tracks to play the soundtrack from EVE-constant locations:
    m_audioTrackIds[0] = allocateAudioSource();
    m_audioTrackIds[1] = allocateAudioSource();

    for(int i = 0; i < 2; i++) {
      if(m_currentGeeRecorder[i]) {
	m_currentGeeRecorder[i]->
	  deactivate(m_timing.currentTimeToRecordTime());
	m_currentGeeRecorder[i] = 0;
      }
    }

    if(m_mode == MODE_TOOLS && !m_forceToolsVisible) {
      if(m_tools->uiHandler()) {

	m_tools->setUiHandler(0);

	if((m_location - m_locationOld).length() > 1.0 ||
	   m_scale != m_scaleOld)
	  centralizeTools();

	return;
      }
      
      m_tools->stop();
    }
    m_actionLists->rewind(m_timing.currentTimeToRecordTime());

    m_mode = newMode;

    m_timing.m_zeroTime.getTimeOfDay();
    m_timing.m_now = m_timing.m_zeroTime;

    m_actionLists->setTiming(m_timing);

    if(m_mode != MODE_PLAY) {
      m_recorder = RECORDER_SIZEOF;
      m_tools->start();
    
      // m_geeEngine->toBegin();

      // Now move the tool package to the center of the CAVE
      if((m_location - m_locationOld).length() > 1.0 ||
	 m_scale != m_scaleOld)
	centralizeTools();
    }
    else if(m_mode == MODE_PLAY) {

      TimeStamp time = m_timing.currentTimeToRecordTime();
      
      if(m_rebuildGee) {
	m_geeEngine->rewind(time);
	m_rebuildGee = false;
      }
      else
	m_geeEngine->update(time);

      bool notrunning = (m_nextRecorder == RECORDER_EMITTER_STATIC) ||
	(m_nextRecorder == RECORDER_BLOW_STATIC) ||
	(m_nextRecorder == RECORDER_SWIRL_STATIC);

      if(!notrunning) {
	// Get rid of the old particles:
	m_scene->update(40000, 20000, 10, 100.0);
	m_scene->update(40000, 20000, 10, 100.0);
	m_scene->update(40000, 20000, 10, 100.0);
      }

      m_timing.setRunning(!notrunning);
      m_actionLists->setTiming(m_timing);

      if(!m_forceToolsVisible) {
	if(m_nextRecorder == RECORDER_EMITTER) {
	  m_actionLists->setRecorder(new EmitterRecorder);
	}
	else if(m_nextRecorder == RECORDER_EMITTER_STATIC) {
	  m_actionLists->setRecorder(new EmitterRecorder2);
	}
	else if(m_nextRecorder == RECORDER_BLOW) {
	  m_actionLists->setRecorder(new BlowRecorder);
	}
	else if(m_nextRecorder == RECORDER_BLOW_STATIC) {
	  m_actionLists->setRecorder(new BlowRecorder(false));
	}
	else if(m_nextRecorder == RECORDER_SWIRL) {
	  m_actionLists->setRecorder(new VortexRecorder);
	}
	else if(m_nextRecorder == RECORDER_SWIRL_STATIC) {
	  m_actionLists->setRecorder(new VortexRecorder(false));
	}
	else if(m_nextRecorder == RECORDER_MESH) {
	  m_actionLists->setRecorder(new MeshRecorder());
	}
	else if(m_nextRecorder == RECORDER_EXPLOSION) {
	  m_actionLists->setRecorder(new ExplosionRecorder());
	}
	else if(m_nextRecorder == RECORDER_GEE) {
	  m_actionLists->setRecorder(0);
	  
	  for(int i = 0; i < 2; i++) {

	    if(!m_mouseButtons[i])
	      continue;
	    
	    GeePathToX * rec = 
	      m_recorders[i]->getRecorder(m_nextRecorderName[i]);

	    m_currentGeeRecorder[i] = rec;

	    if(!rec)
	      VEE_error("%s # No GEE Recorder %d \"%s\"", 
			fname, i, m_nextRecorderName[i].c_str());
	    else {
	      rec->setIgnoreNapra(false);
	      rec->activate(m_timing.currentTimeToRecordTime());
	    }
	  }
	}
	else {
	  VEE_error("%s # Bad recorder type %d", fname, (int) m_nextRecorder);
	}
	
	m_recorder = m_nextRecorder;
	tools()->componentEditor()->startRecording(m_nextRecorderName[0]);
      }

      if(!notrunning && m_audioTrack.size() && m_acControl) {

	m_acControl->beginContiguous();

	char buf[128];

	float skip = m_timing.m_offset.secondsD();
	float vol  = m_soundTrackVolume * 0.2;
	float pitch = m_timing.m_playSpeed2;
	
	sprintf(buf, "%s-left", m_audioTrack.c_str());
	m_audioNoteIds[0] = m_acControl->playSamplePitched
	  (m_audioTrackIds[0], buf, vol, pitch, 
	   -45.0, 0.0, 0.0, 1.0, false, skip);

	sprintf(buf, "%s-right", m_audioTrack.c_str());
	m_audioNoteIds[1] = m_acControl->playSamplePitched
	  (m_audioTrackIds[1], buf, vol, pitch,
	   45.0, 0.0, 0.0, 1.0, false, skip);

	m_acControl->endContiguous();
      }	
    }  

    if(newMode != MODE_PLAY) {
      setRenderHand(0, true);
      setRenderHand(1, true);
    }

    VEE_trace("%s # EXIT", fname);
  }

  bool Engine::writeDomFile(const QString &str)
  {
    QDomDocument doc("AnimaLand");
    QDomElement elem = writeDom(doc);

    doc.appendChild(elem);
  
    QCString s = doc.toCString();

    QFile out(str);
    bool ok = out.open(IO_WriteOnly);
    QTextStream(&out) << s;

    m_geeEngine->write((str + ".gee").ascii());

    return ok;
  }

  QDomElement Engine::writeDom(QDomDocument &doc)
  {
    QDomElement elem = doc.createElement("Engine");
  
    elem.appendChild(m_actionLists->writeDom(doc)); 
    elem.appendChild(m_tools->writeDom(doc));

    return elem;
  }

  bool Engine::readDomFile(const QString &str)
  {
    VEE_trace("Engine::readDomFile # \"%s\"", str.ascii());

    m_geeEngine->read((str + ".gee").ascii());

    QDomDocument doc;

    QFile f(str);
  
    if ( !f.open( IO_ReadOnly ) ) {
      VEE_error("Engine::readDomFile # Open failed");
      return false;
    }
    QString error;
    int errline = -1;

    if ( !doc.setContent( &f, & error, &errline ) ) {
      VEE_error("Engine::readDomFile # failed on line %d (%s)",
		errline, error.ascii());
      f.close();
      return false;
    }

    QDomElement elem = doc.documentElement();

    bool ok = readDom(elem);

    MessageInt mi(0);
    msgSend("newfile", &mi);

    return ok;
  }

  bool Engine::readDom(QDomElement &e)
  {

    VEE_trace("Engine::readDom");

    QDomElement lists = e.namedItem("ActionListManager").toElement();

    bool ok = m_actionLists->readDom(lists);

    if(!ok)
      VEE_error("Engine::readDom # Action lists could not be read");

    QDomElement tools = e.namedItem("Tools").toElement();

    bool ok2 = m_tools->readDom(tools);

    if(!ok2)
      VEE_error("Engine::readDom # Tools could not be read");

    return ok && ok2;
  }

  void Engine::useScene(uint index)
  {
    VEE_trace("Engine::useScene # %u", index);

    for(uint i=0; i < m_scenes.size(); i++) {

      SceneObjs *so = m_scenes[i];

      std::vector<Solar::Object *>::iterator it;

      while((it = 
	     std::find(m_objects.begin(), m_objects.end(), so->object())) !=
	    m_objects.end()) {
    
	VEE_trace("Engine::useScene # removing %u", i);

	// Clear the previous scene away:
	m_objects.erase(it);

	if(so->grid()) {
	  so->grid()->unlinkBoxes();
	  if(so->grid()->sceneBox()) {
	    so->grid()->sceneBox()->forgetForce(so->grid());
	    so->grid()->sceneBox()->removeChildObject(so->grid());
	    so->grid()->setSceneBox(0);
	  }
	}

      }
    }

    assert(index < m_scenes.size());

    SceneObjs *so = m_scenes[index];

    m_objects.push_back(so->object());

    if(so->grid())
      m_scene->addNewObject(so->grid()/*->clone()*/);
  }

  void Engine::origin(const Vector3 & loc)
  {
    MouseNavigator * navi = dynamic_cast<MouseNavigator *> (m_navigator);

    if(!navi)
      return;

    navi->moveTo(loc, 0.0);
  }

  void Engine::setNextRecorder(Recorder r, 
			       const std::string &rec1,
			       const std::string &rec2)
  { 
    m_nextRecorder = r; 

    /* Only take the next recorder name if it is a valid tool name: */
    if(m_recorders[0]->getRecorder(rec1))
      m_nextRecorderName[0] = rec1;
    if(m_recorders[1]->getRecorder(rec2))
      m_nextRecorderName[1] = rec2;
  }

  void Engine::setDone(bool done) 
  { 
    Log::log("EXIT");
    Log::quit();

    MessageInt mi(done);
    msgSend("exit", &mi);

    m_done = done;
  }

  void Engine::addAniMesh(Solar::Object *mesh)
  {
    m_aniObjects.push_back(mesh);
  }

  void Engine::removeAniMesh(Solar::Object *mesh)
  {
    std::vector<Solar::Object *>::iterator it =
      std::find(m_aniObjects.begin(), m_aniObjects.end(), mesh);

    if(it != m_aniObjects.end())
      m_aniObjects.erase(it);
  }


  MessageSource * Engine::buttonMessages(int hand)
  {
    return m_mouseButtons[hand];
  }
  
  GeeRecorder * Engine::currentActiveGeeRecorder()
  {
    for(int i = 0; i < 2; i++)
      if(m_currentGeeRecorder[i])
	if(m_currentGeeRecorder[i]->recording())
	  return m_currentGeeRecorder[i];

    return 0;
  }

  /** This method initializes all Solar-related stuff. Solar is used to
      render the solid objects (background, 3d-cursor, pyramid) to
      the scene. */

  void Engine::initSolarObjects()
  {
    // const char * fname = "Engine::initSolarObjects";
  
    Fluid::Matrix3 rot;

    rot.rotateX(Fluid::Math::PI * 0.5);

    for(uint i=0; i < 2; i++) {
      char buf[16];

      sprintf(buf, "Cursor%d", i+1);

      Solar::Model * cursor = new Solar::Model;
      cursor->setGarbageBin(& m_deletedGraphics);

      cursor->createPyramid(0.1, buf);
      cursor->transform(rot);

      Solar::Object * obj = new Solar::Object(cursor);

      // m_objects.push_back(obj);

      m_handPointer[i] = obj;
    }
  }

  void Engine::centralizeTools()
  {
    VEE_Vector3 loc(0.0, -1.0, 0.0);
    VEE_Matrix3 rot;
    rot.identity();
    m_tools->setLocation(m_rotation * (loc / m_scale) + m_location);
    m_tools->setRotation(m_rotation * rot);

    m_locationOld = m_location;
    m_scaleOld = m_scale;
  }

}
