/* -*- C++ -*- */

/* -*- C++ -*- */

/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Crystal animation and 3D modeling application
 *
 * Authors:   Tommi Ilmonen 2001-2004
 * Copyright: Tommi Ilmonen and/or Tekes 2001-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <crystal_gee_config.h>

#include <crystal_engine.h>

#include <gee_engine.h>

#include <vee_error.h>
#include <vee_math.h>
#include <vee_values.h>
#include <vee_value_str.h>

#include <Fluid/InputDevices.hpp>
#include <Fluid/MessageFloat.hpp>
#include <Fluid/MessageInt.hpp>
#include <Fluid/MessageString.hpp>

namespace Crystal { 

  GeeConfig::GeeConfig(Engine *e)
    : m_engine(e),
      m_lightSet(0)
  {}

  GeeConfig::~GeeConfig()
  {}

  void GeeConfig::init()
  {
    bool withNapra = Fluid::InputDevices::getNapra("napra") != 0;

    // Generate a few default profiles:
    GEE_Profile prof;
    GEE_ProfileVertex2D vertex;
    vertex.m_normal.make(0, 1);

    Vector3 span(1, 1, 0);
    Vector2 normal(0, 1);

    prof.generateBar(1.0, 0.5);
    if(withNapra) prof.normalize(span);
    m_profiles["ibar"] = prof;

    if(withNapra) {

      // A stripe between thumb and index finger

      prof.clear();
      prof.addVertex(GEE_ProfileVertex2D(Vector2(-0.5, -0.5), normal));
      prof.addVertex(GEE_ProfileVertex2D(Vector2( 0.0,  0.5), normal, false));
    }
    else
      prof.generateBar(1.0, 0.0);

    m_profiles["stripe"] = prof;

    /* An x^2 arc that opens downwards: */
    prof.clear();
    for(int i = -3; i <= 3; i++) {
      float tmp = i / 3.0f;
      prof.addVertex(GEE_ProfileVertex2D(Vector2(tmp * 0.5, 
						 - tmp * tmp + 0.5), normal));
    }

    prof.getVertex(prof.size() - 1).m_continues = false;
    prof.normalize(span);
    m_profiles["arc"] = prof;

    prof.generateCircle(1.0, 3);
    if(withNapra) prof.normalize(span);
    m_profiles["tube3"] = prof;

    prof.getVertex(1).m_continues = false;
    m_profiles["cut3"] = prof;

    prof.generateCircle(1.0, 5);
    if(withNapra) prof.normalize(span);
    m_profiles["tube5"] = prof;

    prof.generateCircle(1.0, 10);
    if(withNapra) prof.normalize(span);
    m_profiles["tube10"] = prof;

    prof.generateCross(1.0);
    if(withNapra) prof.normalize(span);
    m_profiles["cross"] = prof;

    selectProfile("Mesh=stripe");
    selectProfile("MultiBrush=stripe");
  }


  bool GeeConfig::msgProcess(const char *id, Fluid::Message *m)
  {
    const char * fname = "GeeConfig::msgProcess";

    MessageFloat  * mf = dynamic_cast<MessageFloat *> (m);
    MessageInt    * mi = dynamic_cast<MessageInt *> (m);
    MessageString * ms = dynamic_cast<MessageString *> (m);

    int ival = 0;
    float fval = 0;
    std::string sval;
    
    if(mf) {
      fval = mf->value();
      ival = (int) fval;
      setValue(id, VEE_ValueFloat(fval));
    }
    else if(mi) {
      ival = mi->value();
      fval = ival;
      setValue(id, VEE_ValueInt(ival));
    }
    else if(ms) {
      sval = ms->string();
      setValue(id, VEE_ValueString(sval));
    }

    bool ok = false;

    if(strcmp(id, "hsv") == 0) {

      VEE_error("%s # Old fashioned hsv -command");

      /* Fluid::MessageString *ms = dynamic_cast<Fluid::MessageString *> (m);

      if(ms) {
	float h = 0.5, s = 0.5, v = 0.5;
	int ns = sscanf(ms->string().c_str(), "%f %f %f", &h, &s, &v);
	if(ns != 3)
	  return false;

	VEE_trace("%s # Got color %f %f %f", fname, h, s, v);
	VEE_Hsva hsva(h, s, v, 1.0, 1.0);
	VEE_Rgba rgba(hsva.rgba());
	rgba.setAlpha(m_color.alpha());
	rgba.setTheta(m_color.theta());
	updateColor(rgba);
	ok = true;
	}*/
    }
    else if(strcmp(id, "alpha") == 0) {
      VEE_error("%s # Old fashioned alpha -command");
    }
    else if(strcmp(id, "burn") == 0) {
      VEE_error("%s # Old fashioned burn -command");
    }
    else if(strcmp(id, "texture") == 0) {
      std::string str = GEE_World::textureName((int) fval);
      VEE_trace("%s # Texture \"%s\"", fname, str.c_str());
      setValue("texture", VEE_ValueString(str));
      ok = true;
    }
    else if(strncmp(id, "profile-", 8) == 0) {
      selectProfile(id + 8, m);
      ok = true;
    }
    else if(strncmp(id, "align-", 6) == 0) {
      int align = atoi(id + 6);
      setValue("align", VEE_ValueInt(align));
      ok = true;
    }
    else if(strcmp(id, "cursorsize") == 0) {

      m_engine->setCursorSize(VEE_clamp(fval * 15.0f, 0.1f, 3.0f));
      ok = true;
    }
    else if(strncmp(id, "lights-", 7) == 0) {
      m_lightSet = atoi(id + 7);
      VEE_trace("%s # Light set %d", fname, m_lightSet);
      ok = true;
    }
    else if(strncmp(id, "texture=", 8) == 0) {
      const char * texture = id + 8;
      setValueString("texturename", texture);
      VEE_trace("%s # Chose texture %s", fname, texture);
      ok = true;
    }

    return ok;
  }

  void GeeConfig::setValue(const std::string &name, const VEE_Value &v)
  { 
    m_values[name] = v.clone(); 
  }
  
  void GeeConfig::setValueInt(const std::string &name, int v)
  {
    m_values[name] = new VEE_ValueInt(v);
  }

  void GeeConfig::setValueFloat(const std::string &name, float v)
  {
    m_values[name] = new VEE_ValueFloat(v);
  }
  
  void GeeConfig::setValueString(const std::string &name, 
				 const std::string &val)
  {
    m_values[name] = new VEE_ValueString(val);
  }

  VEE_Value * GeeConfig::getValue(const std::string &name) 
  {
    iterator it = m_values.find(name);

    if(it == m_values.end()) {
      // VEE_error("GeeConfig::getValue # value \"%s\" undefined", name.c_str());
      return 0;
    }
    
    return (*it).second.ptr();
  }
  
  int GeeConfig::getValueInt(const std::string &name, int def)
  {
    VEE_Value * v = getValue(name);
    if(!v)
      return def;
    
    int tmp = def;
    if(v->getInt(&tmp))
      def = tmp;

    return def;
  }

  float GeeConfig::getValueFloat(const std::string &name, float def)
  {
    VEE_Value * v = getValue(name);
    if(!v)
      return def;
    
    v->getFloat(&def);

    return def;
  }

  std::string GeeConfig::getValueString(const std::string &name,
					const std::string &def)
  {
    VEE_Value * v = getValue(name);
    if(!v)
      return def;
    
    VEE_ValueString * str = dynamic_cast<VEE_ValueString *> (v);

    return str ? str->string() : def; 
  }

  VEE_Rgba GeeConfig::getValueRgba(const std::string &name, 
				   const VEE_Rgba & def)
  {
    VEE_Value * v = getValue(name);
    if(!v)
      return def;
    
    VEE_ValueRgba * rgba = dynamic_cast<VEE_ValueRgba *> (v);

    return rgba ? rgba->getValue() : def;    
  }

  VEE_Rgba GeeConfig::constructRgba(const std::string &name)
  {
    std::string hsv = getValueString(name + "/hsv");
    
    float h = 0.5, s = 0.5, v = 0.5, a, b;

    int ns = sscanf(hsv.c_str(), "%f %f %f", &h, &s, &v);
    if(ns != 3)
      VEE_error("GeeConfig::constructRgba # Missing hsv components from %s",
		name.c_str());
    
    a = getValueFloat(name + "/alpha", 1.0);
    b = getValueFloat(name + "/burn", 0.0);
    
    // VEE_trace("%s # Got color %f %f %f", fname, h, s, v);

    VEE_Hsva hsva(h, s, v, a, b);
    
    return hsva.rgba();
  }

  GEE_Material GeeConfig::material(const std::string &name)
  {
    VEE_Rgba color = constructRgba(name);

    GEE_Material mat;
    
    mat.setColor(Vector4(color.data()));
    mat.enableBlend(color.alpha() < 1.0f);
    mat.setBurn(color.theta());

    return mat;
  }

  const GEE_Profile & GeeConfig::profile(const std::string &tool)
  {
    if(m_profile.find(tool) == m_profile.end()) {
      VEE_error("GeeConfig::profile # Returning an empty profile");
      return m_empty;
    }

    return m_profile[tool];
  }

  /** Selects a profile for a tool. Either with format-string
      (Mesh=cut3 etc) or via MessageString. */

  void GeeConfig::selectProfile(const char * name, Message * m)
  {
    // Strip the name of the tool type into buf
    
    uint i;
    char buf[64];
    
    for(i = 0; i < strlen(name) && name[i] != '='; i++)
      buf[i] = name[i];

    buf[i] = '\0';

    if(name[i] == '=') {
      
      name += i + 1;
    }
    else {
      MessageString * ms = dynamic_cast<MessageString *> (m);
      if(ms)
	name = ms->string().c_str();
    }

    if(m_profiles.find(name) == m_profiles.end()) {
      VEE_error("GeeConfig::selectProfile # No profile \"%s\"", name);
      return;
    }

    VEE_error("GeeConfig::selectProfile # selected \"%s\" for \"%s\"", 
	      name, buf);

    m_profile[buf] = m_profiles[name];
  }
  
}
