/* -*- C++ -*- */

/* -*- C++ -*- */

/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Crystal animation and 3D modeling application
 *
 * Authors:   Tommi Ilmonen 2001-2004
 * Copyright: Tommi Ilmonen and/or Tekes 2001-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <crystal_gee_editor.h>

#include <crystal_engine.h>
#include <crystal_gee_config.h>
#include <crystal_tools.h>

#include <gee_engine.h>
#include <gee_macros.h>

#include <vee_error.h>
#include <vee_math.h>
#include <vee_particle.h>
#include <vee_values.h>
#include <vee_value_str.h>
#include <vee_vector4.h>

#include <Fluid/MessageFloat.hpp>
#include <Fluid/MessageString.hpp>

namespace Crystal { 

  class GeeEditorSliderParam
  {
  public:
    GeeEditorSliderParam(const char * name,
			 float def, float min, float max, 
			 MJ_Mapper * mapper = 0)
      : m_name(name),
	m_min(min),
	m_max(max),
	m_default(def),
	m_mapper(mapper)
    {}

    const char * m_name;
    float m_min, m_max, m_default;
    MJ_Mapper * m_mapper;
  };

  static const GeeEditorSliderParam GeeEditorSliderParams[] = {
    GeeEditorSliderParam("radius",  0.2, 0.0, 3.0),
    GeeEditorSliderParam("size",    0.2, 0.005, 3.0, new MJ_MapperLog),
    GeeEditorSliderParam("texture", 10, 0, 63),
    GeeEditorSliderParam("alpha",   1.0, 0.0, 1.0),
    GeeEditorSliderParam("burn",    0.0, 0.0, 1.0),
    GeeEditorSliderParam("timescale", 0.0, 0.0, 1.0),
    GeeEditorSliderParam("threshold", 0.2, 0.0, 3.0),
    GeeEditorSliderParam("pressure", 1.0, 0.0, 3.0),
    GeeEditorSliderParam(0, 0, 0, 0) // Terminator
  };

  enum {
    GEE_TOOL_LINES,
    GEE_TOOL_MESH,
    GEE_TOOL_PARTICLES,
    GEE_TOOL_MOVE,
    GEE_TOOL_MOVE_GEO,
    GEE_TOOL_MOVE_ROT_GEO,
    GEE_TOOL_SCALE_GEO,
    GEE_TOOL_SCALE_AXIS_GEO,
    GEE_TOOL_RETEXTURE,
    GEE_TOOL_COLOR,
    GEE_TOOL_COLOR_GEO,
    GEE_TOOL_ERASE,
    GEE_TOOL_ERASE_GEO,
    GEE_TOOL_GROUP,
    GEE_TOOL_SIZEOF    
  };

  static const char * toolIds[GEE_TOOL_SIZEOF] = {
    "Lines",
    "Mesh",
    "Particles",
    "Magnet",
    "Move",
    "MoveRot",
    "Scale",
    "ScaleAxis",
    "Retexture",
    "Color",
    "ColorGeo",
    "Erase",
    "EraseGeo",
    "Group"
  };
  
  GeeEditor::GeeEditor(Tools *host)
    : m_host(host),
      m_widgets(host->grabs()),
      m_color(VEE_Rgba::Blue),
      m_po(new GEE_ParticleObject)
  {
    GEE_ParticleData pd;
    m_po->addParticle(pd);
    m_pd = & (* m_po->begin());
    m_pd->m_particle = new VEE_Particle();
    m_pdId = GEE_IdContainer::allocateId();
    m_po->setId(m_pdId);
    m_pdd  = m_po;
    
    VEE_BoundingBox colorCubeBox;

    int i;

    colorCubeBox.set(-2.0, 0.5, 0.0, 
		     0.0, 2.0, 2.0);

    Vector2 b1, b2, b3, b21, b22; // Some vectors for selectors

    b1.make(0.0, 0.0);
    b2.make(0.0, 0.22);
    b3.make(1.2, 0.2);

    b21.make(b3.x + 0.1, 0.0);
    b22.make(b21.x + b3.x + 0.1, 0.0);

    char buf[32], buf2[32];

    for(i=0; i < GEE_TOOL_SIZEOF; i++) {

      TextureButton * txb = new TextureButton(m_host);
      
      txb->setArea(b1, b3);
      txb->setLabel(toolIds[i]);      

      sprintf(buf, "tool-%u", i);

      txb->msgAddSink("press", buf, this);
      
      m_widgets.addWidget(txb);

      b1 += b2;
    }

    /* NULL-terminated list of tool names. */
    static const char * unredo[] = { 
      "undo", 
      "redo", 
      "bgcolor", 
      "clear", 
      "scene", 
      0 
    };
    
    const char * label;
    for(i = 0; (label = unredo[i]); i++) {
      TextureButton * txb = new TextureButton(m_host);
      
      txb->setArea(b21, b3);
      txb->setSounds(m_host->buttonSounds());
      txb->setLabel(label);

      txb->msgAddSink("press", label, this);
      
      m_widgets.addWidget(txb);

      b21 += b2;      
    }
    
    /* NULL-terminated list of alignment types */
    static const char * alignTypes[] = { 
      "Motion", 
      "Sensor", 
      0 
    };
    
    Menu * aligns = new Menu(m_host);

    aligns->setArea(b21, b3);
    aligns->setSounds(m_host->buttonSounds());
    aligns->setLabel("Align");
    m_widgets.addWidget(aligns);
    
    for(i = 0; (label = alignTypes[i]); i++) {
      sprintf(buf,  "align-%d", i);
      aligns->addItem(label, buf, this);
    }
    b21 += b2;

    /* NULL-terminated list of profiles names */
    static const char * profiles[] = { 
      "Tube", 
      "I-Bar", 
      "|-Stripe",
      0 
    };

    for(i = 0; (label = profiles[i]); i++) {
      TextureButton * txb = new TextureButton(m_host);
      
      txb->setArea(b22, b3);
      txb->setSounds(m_host->buttonSounds());
      txb->setLabel(label);

      sprintf(buf, "profile-%d", i);

      txb->msgAddSink("press", buf, this);
      
      m_widgets.addWidget(txb);

      b22 += b2;      
    }

    // Layer menu

    Menu * layers = new Menu(m_host);

    layers->setArea(b21, b3);
    layers->setSounds(m_host->buttonSounds());
    layers->setLabel("Layers");
    m_widgets.addWidget(layers);

    for(i = -1; i < 10; i++) {
      sprintf(buf,  "layer_%d", i);
      sprintf(buf2, "Layer %d", i);
      layers->addItem(buf2, buf, this);
    }

    // Color cube

    ColorCube * cc = new ColorCube(m_host);
    cc->msgAddSink("color", "color", this);
    m_widgets.addWidget(cc);
    // m_colorCube = cc;
    cc->setBoundingBox(colorCubeBox);

    // Sliders


    Fluid::Vector3 w1, w2; // Slider locations
    w1.make(0.3, 0.4, 1.8);
    w2.make(0.0, 0.6, 0.0);

    Fluid::Matrix3 m1, m2;
    m1.rotateX(-0.3 * VEE_HALF_PI);
    m2.rotateY(-VEE_HALF_PI);
    m1 = m2 * m1;
    

    for(i=0; GeeEditorSliderParams[i].m_name != 0; i++) {

      const GeeEditorSliderParam & gesp = GeeEditorSliderParams[i];
      
      
      Slider *slider = new Slider(m_host, 
				  gesp.m_default,
				  gesp.m_min,
				  gesp.m_max);

      slider->setSounds(m_host->buttonSounds());
      m_widgets.addWidget(slider);
    
      slider->msgAddSink("value", gesp.m_name, this);
  
      slider->setLocation(w1);
      slider->setRotation(m1);
      slider->setSize(2.0, 0.3);
      slider->setLabel(gesp.m_name);
      slider->setOrientation(Slider::HORIZONTAL);
      slider->setDisplayValue(false);
      slider->setMapper(gesp.m_mapper ? gesp.m_mapper->clone() : 0);

      slider->setValue(gesp.m_default, true);

      w1 += w2;

    }
  }
  
  GeeEditor::~GeeEditor()
  {}

  bool GeeEditor::msgProcess(const char *id, Fluid::Message *m)
  {
    GEE_FNAME = "GeeEditor::msgProcess";

    bool ok = false;

    VEE_trace("%s # %s", fname, id);

    MessageFloat * mf = dynamic_cast<Fluid::MessageFloat *> (m);
    float fval = mf ? mf->value() : -1.0;
    GeeConfig * geeConfig = m_host->engine()->geeConfig();

    if(strcmp(id, "undo") == 0) {
      m_host->engine()->geeEngine()->undo();
      ok = true;
    }
    else if(strcmp(id, "redo") == 0) {
      m_host->engine()->geeEngine()->redo();
      ok = true;
    }
    else if(strcmp(id, "bgcolor") == 0) {
      m_host->engine()->geeEngine()->world().setBackgroundColor(m_color);
      ok = true;
    }
    else if(strcmp(id, "clear") == 0) {
      m_host->engine()->geeEngine()->actions().undo(1000000);
      ok = true;
    }
    else if(strcmp(id, "scene") == 0) {
      m_host->engine()->toggleRenderSolars();
      ok = true;
    }
    else if(strncmp(id, "layer_", 6) == 0) {
      int layer = atoi(id + 6);
      geeConfig->setValue("layer", VEE_ValueInt(layer));
      ok = true;
    }
    else if(strcmp(id, "size") == 0) {
      geeConfig->setValue("size", VEE_ValueFloat(fval));
      m_pd->setSize(fval);
      ok = true;
    }
    else if(strcmp(id, "radius") == 0) {
      geeConfig->setValue("radius", VEE_ValueFloat(fval));
      ok = true;
    }
    else if(strcmp(id, "pressure") == 0) {
      geeConfig->setValue("pressure", VEE_ValueFloat(fval));
      ok = true;
    }
    else if(strcmp(id, "texture") == 0) {
      std::string str = GEE_World::textureName((int) fval);
      VEE_trace("%s # Texture \"%s\"", fname, str.c_str());
      geeConfig->setValue("texture", VEE_ValueString(str));
      m_pd->setTexture(str, (int) fval);
      ok = true;
    }
    else if(strcmp(id, "alpha") == 0) {
      m_color.setAlpha(fval);
      updateColor(m_color);
      ok = true;
    }
    else if(strcmp(id, "burn") == 0) {
      m_color.setTheta(fval);
      updateColor(m_color);      
      ok = true;
    }
    else if(strcmp(id, "timescale") == 0) {
      geeConfig->setValue("timescale", VEE_ValueFloat(fval));
      m_host->engine()->setPlaySpeed(fval);
      ok = true;
    }
    else if(strcmp(id, "threshold") == 0) {
      geeConfig->setValue("threshold", VEE_ValueFloat(fval));
      ok = true;
    }
    else if(strncmp(id, "align-", 6) == 0) {
      int alignMode = atoi(id + 6);
      geeConfig->setValue("align", VEE_ValueInt(alignMode));
    }
    else if(strncmp(id, "tool-", 5) == 0) {
      int toolId = atoi(id + 5);
      
      Engine * e = m_host->engine();
      const char * toolName = toolIds[toolId];
      e->setNextRecorder(Engine::RECORDER_GEE, toolName);
      VEE_trace("%s # Selected tool \"%s\"", fname, toolName);
      ok = true;
    }
    else if(strncmp(id, "profile-", 8) == 0) {
      
      GEE_Profile p;
      int profId = atoi(id + 8);
      
      if(profId == 0)
	p.generateCircle(1.0, 5);
      else if(profId == 1)
	p.generateBar(1.0, 0.5);
      else if(profId == 2)
	p.generateBar(1.0, 0.0);
	
      geeConfig->setProfile(p);
    }
    else if(strcmp(id, "color") == 0) {
      Fluid::MessageString *ms = dynamic_cast<Fluid::MessageString *> (m);

      if(ms) {
	float h = 0.5, s = 0.5, v = 0.5;
	int ns = sscanf(ms->string().c_str(), "%f %f %f", &h, &s, &v);
	if(ns != 3)
	  return false;

	VEE_trace("%s # Got color %f %f %f", fname, h, s, v);
	VEE_Hsva hsva(h, s, v, 1.0, 1.0);
	VEE_Rgba rgba(hsva.rgba());
	rgba.setAlpha(m_color.alpha());
	rgba.setTheta(m_color.theta());

	updateColor(rgba);

	ok = true;
      }
    }
    else {
      VEE_error("%s # Unrecognized message id \"\"", fname, id);
      ok = false;
    }

    return ok;
  }

  void GeeEditor::render(GlData *gld)
  {
    m_widgets.renderGl(gld);
  }

  void GeeEditor::update()
  {
    m_widgets.update();
    m_pd->setLocation(m_host->worldLocation(VEE_Vector3(2.6, 1.5, 0.5)));
    m_host->engine()->geeEngine()->world().addGeometry2(m_pdd);
    /* VEE_trace("GeeEditor::update # %d %p", (int) m_po->particleCount(), 
       m_pd->m_particle.ptr()); */
  }

  QDomElement GeeEditor::writeDom(QDomDocument &doc)
  {
    QDomElement elem = doc.createElement("GeeEditor");

    return elem;
  }

  bool GeeEditor::readDom(QDomElement &)
  {
    return false;
  }

  void GeeEditor::updateColor(const VEE_Rgba & rgba)
  {    
    m_color = rgba;

    VEE_Vector4 color;
    color.copy(rgba.data());

    bool blend = rgba.alpha() < 1.0f;
    
    GeeConfig * geeConfig = m_host->engine()->geeConfig();
    geeConfig->material().setColor(color);
    geeConfig->material().enableBlend(blend);
    geeConfig->setValue("color", VEE_ValueRgba(rgba));
    
    m_pd->setColor(m_color);
  }

}
