/* -*- C++ -*- */

/* -*- C++ -*- */

/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Crystal animation and 3D modeling application
 *
 * Authors:   Tommi Ilmonen 2001-2004
 * Copyright: Tommi Ilmonen and/or Tekes 2001-2004
 *
 * firstname.lastname@hut.fi
 *
 */


#include <crystal_general_editor.h>

#include <crystal_dom_util.h>
#include <crystal_engine.h>
#include <crystal_scene_objs.h>
#include <crystal_tools.h>

#include <vee_error.h>
#include <vee_math.h>

#include <Fluid/MessageFloat.hpp>
#include <Fluid/MessageInt.hpp>

#include <GL/gl.h>

namespace Crystal { 

  static const char * parammap[GeneralEditor::SLIDERS] = {
    "speed",
    "gravity",
    "bell_sounds",
    "particle_sounds",
    "music_volume"
  };

  static const float paramdefault[GeneralEditor::SLIDERS] = {
    1.0,
    1.0,
    0.15,
    0.3,
    0.3
  };


  static const float parammin[GeneralEditor::SLIDERS] = {
    0.0,
    0.0,
    0.0,
    0.0,
    0.0
  };

  static const float parammax[GeneralEditor::SLIDERS] = {
    3.0,
    20.0,
    1.0,
    1.0,
    1.0
  };

  GeneralEditor::GeneralEditor(Tools *host)
    : Transformation(host),
      m_host(host)
  {
    char buf[64];
    uint i;

    m_widgets.setGrabGroup(host->grabs());

    // Location of the force visibility box.
    Fluid::Vector2 bloc(-1.0, 2.5), bsize(2.0, 0.4); 

    // Location of the load & save buttons
    Fluid::Vector2 floc(-2.0, 2.5), fsize(0.85, 0.4); 

    Fluid::Vector3 s1, s2, s3; // Scene selection locations
    s1.make(2.3, 1.5, 0.0);
    s2.make(0.8);
    s3.make(0.0, -1.0, 0.0);

    Fluid::Vector3 w1, w2; // Slider locations
    w1.make(-1.0, 2.0, 0.0);
    w2.make(0.0, -0.6, 0.0);

    Fluid::Matrix3 m1;
    m1.rotateX(-0.3);

    for(i = 0; i < SLIDERS; i++) { 

      Slider * slider = 
	new Slider(m_host, paramdefault[i], parammin[i], parammax[i]);

      slider->setSounds(m_host->buttonSounds());
      slider->msgAddSink("", parammap[i], this);
    
      slider->setLocation(w1);
      slider->setRotation(m1);
      slider->setSize(3.0, 0.3);
      slider->setOrientation(Slider::HORIZONTAL);
      slider->setLabel(parammap[i]);

      m_widgets.addWidget(slider);
      m_sliders[i] = slider;
      slider->setValue(paramdefault[i], true);

      w1 += w2;
    }

    CheckButton * cb = new CheckButton(this);

    cb->setSounds(m_host->buttonSounds());
    cb->setLabel("Show forces");
    cb->setArea(bloc, bsize);
    cb->msgAddSink("checked", "forcesvisible", this);
    cb->setChecked(true);
    m_forceToggle = cb;

    m_widgets.addWidget(cb);

    // Save & load:

    uint slots = 20;
    char bufload[64], bufsave[64], buf2[64];

    Menu * save = new Menu(m_host);
  
    save->setArea(floc, fsize);
    save->setSounds(m_host->buttonSounds());
    save->setLabel("Save");

    m_widgets.addWidget(save);

    Menu * load = new Menu(m_host);
  
    floc.y -= 0.6;

    load->setArea(floc, fsize);
    load->setSounds(m_host->buttonSounds());
    load->setLabel("Load");
    m_widgets.addWidget(load);

    for(i=0; i < slots; i++) {
      sprintf(bufsave, "save_%d", (int) i);
      sprintf(bufload, "load_%d", (int) i);
      sprintf(buf2, "Slot %d", (int) i);
      save->addItem(buf2, bufsave, this);
      load->addItem(buf2, bufload, this);
    }


    Menu * exit = new Menu(m_host);
  
    floc.y -= 0.6;

    exit->setArea(floc, fsize);
    exit->setSounds(m_host->buttonSounds());
    exit->setLabel("Exit");
    exit->addItem("Exit", "exit", this);
    m_widgets.addWidget(exit);

    TextureButton * origin = new TextureButton(m_host);
  
    floc.y -= 0.6;

    origin->setArea(floc, fsize);
    origin->setSounds(m_host->buttonSounds());
    origin->setLabel("Origin");
    origin->msgAddSink("press", "origin", this);
    m_widgets.addWidget(origin);

    TextureButton * far = new TextureButton(m_host);
  
    floc.y -= 0.6;

    far->setArea(floc, fsize);
    far->setSounds(m_host->buttonSounds());
    far->setLabel("Kauas");
    far->msgAddSink("press", "far", this);
    m_widgets.addWidget(far);

    // Scene selection:

    std::vector<SceneObjs *> * scenes = & m_host->engine()->scenes();

    for(uint s=0; s < scenes->size() && scenes->size() > 1; s++) {

      CubeWidget * cw = new CubeWidget(this);

      cw->setBoundingBox(VEE_BoundingBox(s1, s1 + s2));
      cw->setModel((*scenes)[s]->name());

      sprintf(buf, "scene_%u", s);
      cw->msgAddSink("press", buf, this);
      cw->setBorder(0.05);

      m_widgets.addWidget(cw);
    
      s1 += s3;

      m_scenes.push_back(cw);
    }

    selectScene(0);
  }

  GeneralEditor::~GeneralEditor()
  {}

  bool GeneralEditor::msgProcess(const char *id, Fluid::Message *m)
  {
    bool ok = false;

    if(strcmp(id, parammap[SPEED]) == 0) {
    
      Fluid::MessageFloat * mf = dynamic_cast<Fluid::MessageFloat *> (m);

      if(mf) {
	float s = VEE_clamp(mf->value(), parammin[SPEED], parammax[SPEED]);
	m_host->engine()->setPlaySpeed(s);
	ok = true;
	m_values[SPEED] = s;
      }
    }
    else if(strcmp(id, parammap[GRAVITY]) == 0) {
    
      Fluid::MessageFloat * mf = dynamic_cast<Fluid::MessageFloat *> (m);

      if(mf) {
	float s = VEE_clamp(mf->value(), parammin[GRAVITY], parammax[GRAVITY]);
	m_host->engine()->setGravity(s);
	ok = true;
	m_values[GRAVITY] = s;
      }
    }
    else if(strcmp(id, parammap[BELLS_SOUND_LEVEL]) == 0) {
    
      Fluid::MessageFloat * mf = dynamic_cast<Fluid::MessageFloat *> (m);

      if(mf) {
	float s = VEE_clamp(mf->value(), parammin[BELLS_SOUND_LEVEL], 
			    parammax[BELLS_SOUND_LEVEL]);
	(*m_host->buttonSounds()).setMasterVolume(s);
	ok = true;
	m_values[BELLS_SOUND_LEVEL] = s;
      }
    }
    else if(strcmp(id, parammap[PARTICLE_SOUND_LEVEL]) == 0) {
    
      Fluid::MessageFloat * mf = dynamic_cast<Fluid::MessageFloat *> (m);

      if(mf) {
	float s = VEE_clamp(mf->value(), parammin[PARTICLE_SOUND_LEVEL], 
			    parammax[PARTICLE_SOUND_LEVEL]);
	(*m_host->engine()).setParticleVolume(s);
	ok = true;
	m_values[PARTICLE_SOUND_LEVEL] = s;
      }
    }
    else if(strcmp(id, parammap[SOUNDTRACK_SOUND_LEVEL]) == 0) {
    
      Fluid::MessageFloat * mf = dynamic_cast<Fluid::MessageFloat *> (m);

      if(mf) {
	float s = VEE_clamp(mf->value(), parammin[SOUNDTRACK_SOUND_LEVEL], 
			    parammax[SOUNDTRACK_SOUND_LEVEL]);
	(*m_host->engine()).setSoundTrackVolume(s);
	ok = true;
	m_values[PARTICLE_SOUND_LEVEL] = s;
      }
    }
    else if(strcmp(id, "forcesvisible") == 0) {

      Fluid::MessageInt * mi = dynamic_cast<Fluid::MessageInt *> (m);
    
      if(mi) {
	bool forcesVisible = mi->value();
	m_host->engine()->setShowForces(forcesVisible);
	ok = true;
      }
    }
    else if(strncmp(id, "save_", 5) == 0) {
      QString f("dump_");
      f += (id + 5);
      ok = m_host->engine()->writeDomFile(f);
    }
    else if(strncmp(id, "load_", 5) == 0) {
      QString f("dump_");
      f += (id + 5);
      ok = m_host->engine()->readDomFile(f);
    }
    else if(strcmp(id, "exit") == 0) {
      m_host->engine()->setDone(true);
      ok = true;
    }
    else if(strcmp(id, "origin") == 0) {
      m_host->engine()->origin();
      ok = true;
    }
    else if(strcmp(id, "far") == 0) {
      m_host->engine()->origin(Vector3(24.915775, 192.718719, 9.249522));
      ok = true;
    }
    else if(strncmp(id, "scene_", 6) == 0) {
    
      int index = atoi(id + 6);

      selectScene(index);

      ok = true;
    }

    return ok;
  }

  void GeneralEditor::render(GlData *glData)
  {
    glDisable(GL_BLEND);
    m_widgets.renderGl(glData);
  }

  void GeneralEditor::update()
  {
    m_widgets.update();
  }

  void GeneralEditor::selectScene(int index)
  {
    VEE_trace("GeneralEditor::selectScene # %d %d %p", 
	      index, (int) m_scenes.size(), m_host);
    
    if((uint) index >= m_scenes.size())
      return;
    
    for(uint i=0; i < m_scenes.size(); i++) {
      m_scenes[i]->setColors(m_host->usualColor());
    }
    m_scenes[index]->setColors(m_host->selectedColor());
    
    m_host->engine()->useScene(index);
  }
  
  QDomElement GeneralEditor::writeDom(QDomDocument &doc)
  {
    QDomElement elem = doc.createElement("General");
  
    addTextChild(elem, doc, "show_forces", 
			 (int) m_host->engine()->showForces());

    for(uint i=0; i < SLIDERS; i++) {
      addTextChild(elem, doc, parammap[i], m_values[i]);
    }

    return elem;
  }

  bool GeneralEditor::readDom(QDomElement &elem)
  {
    if(elem.tagName() != "General")
      return false;

    for(uint i=0; i < SLIDERS; i++)
      m_sliders[i]->setValue
	(getFloat(elem, parammap[i], m_values[i]), true);

    bool showForces = getInt(elem, "show_forces", 1);

    m_forceToggle->setChecked(showForces);
    m_host->engine()->setShowForces(showForces);

    return true;
  }

}
