/* -*- C++ -*- */

/* -*- C++ -*- */

/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Crystal animation and 3D modeling application
 *
 * Authors:   Tommi Ilmonen 2001-2004
 * Copyright: Tommi Ilmonen and/or Tekes 2001-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <crystal_gl_data.h>

#include <crystal_engine.h>

#include <vee_error.h>
#include <vee_image.h>
#include <vee_test_qt.h>

#include <Fluid/ConfigReader.hpp>

#include <di_mutex.h>

#include <qdir.h> 

namespace Crystal { 

  /** Initialiaze light values with reasonable defaults. */
  
  GlData::Light::Light()
    : m_location(1.0, 1.0, 1.0, 0.0), // Last zero -> directional light
      m_ambient(0.3, 0.3, 0.3, 1.0),
      m_diffuse(1.0, 1.0, 1.0, 1.0),
      m_specular(0.4, 0.4, 0.4, 1.0)
  {}

  /** Apply the this these light parameters to specified OpenGL
      light. */
  void GlData::Light::apply(GLenum id)
  {
    glLightfv (id, GL_AMBIENT, m_ambient.data());
    glLightfv (id, GL_DIFFUSE, m_diffuse.data());
    glLightfv (id, GL_SPECULAR, m_specular.data());
    glLightfv (id, GL_POSITION, m_location.data());
  }

  /** Get all light parameters from a config chunk. If the values
     cannot be read, then old values are implicitly kept. */
  void GlData::Light::set(ConfigReader::Chunk *values)
  {
    values->get("Ka").getFloats(m_ambient.data(), 4);
    values->get("Kd").getFloats(m_diffuse.data(), 4);
    values->get("Ks").getFloats(m_specular.data(), 4);
    values->get("Location").getFloats(m_location.data(), 4);
  }

  /////////////////////////////////////////////////////////////////////////////

  static DI_MutexAuto dataMutex;

  std::vector<GlData *> GlData::m_instances;

  GlData::GlData()
    : m_quadric(0), m_models(0) 
  {
    // m_solarRenderer.setDisplayLists(&m_displayLists);
    
    m_geeRenderer.setMaterials( & m_geeMaterials);
    m_geeRenderer.setDisplayLists( & m_geeLists);
    m_geeRenderer.setHighlightDisplayLists( & m_geeHighlightLists);
    m_geeRenderer.setVeeRenderer( & m_veeRenderer);
    m_geeRenderer.setTextures( & m_textures);

    DI_Guard g(&dataMutex);

    m_instances.push_back(this);
  }

  GlData::~GlData()
  {
    DI_Guard g(&dataMutex);

    std::vector<GlData *>::iterator it = 
      std::find(m_instances.begin(), m_instances.end(), this);
    if(it != m_instances.end())
      m_instances.erase(it);
  }

  /** Initialize textures for this context. */

  void GlData::contextInit(Engine * engine, 
			   ConfigReader::Config * config)
  {
    // static DI_MutexAuto mutex;
    // mutex.lock();

    m_models = & engine->models();

    ConfigReader::Config tmp;
    ConfigReader::Chunk glChunk;

    // If there is no proper chunk we use a temporary (empty) chunk -> default
    if(!config)
      config = &tmp;

    glChunk = config->get("OpenGL");

    const char * fname = "GlData::contextInit";

    uint i;

    VEE_trace("%s # OpenGL chunk:", fname);
    glChunk.dump(std::cout);
    std::cout << std::flush;

    {
      // Need to lock since the m_veeRenderer.initializeGL() uses configreader

      int particlemode = glChunk.get("particlemode").getInt
	(VEE_GlRenderer::RenderQuads);
      int usetextures  = glChunk.get("usetextures").getInt(1);
      int boxesVisible = glChunk.get("boxesvisible").getInt(0);
      int subtexture   = glChunk.get("subtexture").getInt(11);

      float particleSize = glChunk.get("size").getDouble(1.0);
      float lengthScale  = glChunk.get("lengthscale").getDouble(0.5);
      float alphamul     = glChunk.get("alphamul").getDouble(1.0);
    
      m_veeRenderer.setExternalTextures(&m_blobTex, &m_polyTex);
      m_veeRenderer.initializeGL();
      m_veeRenderer.setBlend(true);
      m_veeRenderer.setParticleMode((VEE_GlRenderer::Mode) particlemode);
      m_veeRenderer.setForceMode(VEE_GlRenderer::RenderPoints);
      m_veeRenderer.setSubTexture(subtexture);
      m_veeRenderer.setParticleSize(particleSize);

      m_veeRenderer.setAlphaMul(alphamul);
      m_veeRenderer.setLengthScale(lengthScale);
      m_veeRenderer.setUseTextures(usetextures);
      m_veeRenderer.setDrawBoxEdges(boxesVisible);

      m_veeRenderer.defaultRenderClass().setAlign(VEE_ALIGN_BILLBOARD);

    }

    // mutex.unlock();

    m_materials.addUsualColors();
    m_solarRenderer.setMaterials(&m_materials);

    // Load fonts:

    std::string fontPath = glChunk.get("fontpath").getString
      ("/p/projects/tilmonen/local/txf-fonts");
    std::string fontFile = glChunk.get("font").getString("Helvetica.txf");
    fontFile = fontPath + "/" + fontFile;
    i = m_font.load(fontFile.c_str());

    if(i == FNT_FALSE) {
      VEE_error("%s # Failed to load font %s", fname, fontFile.c_str());
    }

    m_fntRenderer.setFont(&m_font);

    // Load Solar materials for the scene:

    std::string mtlFile = 
      glChunk.get("materials").getString("vee_materials.mtl");

    VEE_TestQt::loadMaterials(mtlFile, &m_textures, & m_materials);

    // More and models:

    const uint faces = 5;

    const char * faceStrs[faces] = {
      "Face0",
      "Face1",
      "Face2",
      "Face3",
      "Floor"
    };

    {
      // Add a few materials for the 3D cursors.

      std::string pointerStrs[2] = { "Cursor1", "Cursor2" };
      Solar::Rgba colors[2] = 
	{ Solar::Rgba (0.8, 0.0, 0.0), Solar::Rgba (0.8, 0.0, 0.0) };

      for(uint i=0; i < 2; i++) {
	for(uint j=0; j < faces; j++) {
	  Solar::GlMaterial mat(70.0, colors[i]);
	  mat.setFlags(mat.flags() | Solar::GlMaterial::WITH_MESH);
	  m_materials.addMaterial(pointerStrs[i] + faceStrs[j],
				  mat);
	}
      }
    }

    uint checkSize = 16, checkInterval = 4;

    VEE_Image bitmap;
  
    bitmap.resize(checkSize, checkSize, 1);
  
    unsigned char gray1 = 0xFF;
    unsigned char gray2 = 0x0;
  
    bitmap.checkTexture(checkInterval, checkInterval, & gray1, & gray2 );

    VEE_GlTexture2D checkTex;
    checkTex.genTexture(&bitmap, true);
  
    m_textures.addTexture("check", checkTex);

    bitmap.load("images/button-bg.png");
    
    VEE_GlTexture2D buttonTex;
    buttonTex.genTexture(&bitmap, true);
  
    m_textures.addTexture("button_bg", buttonTex);
  
    m_materials.addMaterial
      ("particles", 
       Solar::GlMaterial(50, Solar::Rgba::White, m_veeRenderer.getTexture()));

    // Load default textures:

    QString texdir = glChunk.get("texturedir").getString().c_str();

    VEE_Qt::loadTextureDir( & m_textures, texdir);

    // Lights:

    char buf[32];

    for(i = 0; i < 8; i++) {

      sprintf(buf, "Light-%d", i);

      ConfigReader::Chunk c = config->get(buf);
      if(!c.isEmpty()) {
	Light l;
	l.set(&c);
	m_lights.push_back(l);
      }
    }

    m_geeRenderer.loadLights("lights-%d");

    m_quadric = gluNewQuadric();
  }


  void GlData::lightsOn()
  {
    if(!m_lights.size()) {
      glDisable(GL_LIGHTING);
      return;
    }

    glEnable(GL_LIGHTING);

    for(int i=0; i < 8; i++) {

      GLenum id = GL_LIGHT0 + i;

      if(i < (int) m_lights.size()) {	
	m_lights[i].apply(id);
	glEnable(id);
      }
      else
	glDisable(id);
    }

    // glLightModelfv (GL_LIGHT_MODEL_AMBIENT, ambientLight.data());

    glLightModeli (GL_LIGHT_MODEL_LOCAL_VIEWER, 1);
  }

  void GlData::applyGlTransform(const Fluid::Matrix3 &rot,
				const Fluid::Vector3 &loc)
  {
    float oglMatrix[16];

    oglMatrix[0] = rot[0][0];
    oglMatrix[1] = rot[1][0];
    oglMatrix[2] = rot[2][0];
    oglMatrix[3] = 0.0f;

    oglMatrix[4] = rot[0][1];
    oglMatrix[5] = rot[1][1];
    oglMatrix[6] = rot[2][1];
    oglMatrix[7] = 0.0f;

    oglMatrix[8] = rot[0][2];
    oglMatrix[9] = rot[1][2];
    oglMatrix[10] = rot[2][2];
    oglMatrix[11] = 0.0f;

    oglMatrix[12] = loc[0];
    oglMatrix[13] = loc[1];
    oglMatrix[14] = loc[2];
    oglMatrix[15] = 1.0f;

    glMultMatrixf(oglMatrix);
  }

}

