/* -*- C++ -*- */

/* -*- C++ -*- */

/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Crystal animation and 3D modeling application
 *
 * Authors:   Tommi Ilmonen 2001-2004
 * Copyright: Tommi Ilmonen and/or Tekes 2001-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <crystal_graph_editor.h>

#include <crystal_gl_data.h>
#include <crystal_value_graph.h>

#include <vee_error.h>
#include <vee_math.h>
#include <vee_ptr.h>

#include <GL/gl.h>

namespace Crystal { 

  GraphEditor::Graph::~Graph()
  {}

  void GraphEditor::Graph::renderGl(GlData *)
  {
    uint n = m_size, stride = m_stride;

    float * time  = m_times;
    float * value = m_values;

    glColor4fv(m_color.data());

    glBegin(GL_QUAD_STRIP);

    float width = 0.02;

    float min = m_min;

    float scale = renderScale();

    for(uint i=0; i < n; i++) {

      float y = (*value - min) * scale;

      glVertex2f(*time, y + width);
      glVertex2f(*time, y - width);

      time  = VEE_Ptr::offset(time, stride);
      value = VEE_Ptr::offset(value, stride);
    }
    
    glEnd();
  }

  void GraphEditor::Graph::init()
  {
    if(!m_size) return;

    m_min = m_max = *m_values;

    for(uint i=0; i < m_size; i++) {
      float v =  * VEE_Ptr::offset(m_values, i * m_stride);
      if(m_min > v) m_min = v;
      if(m_max < v) m_max = v;
    }

    VEE_trace("GraphEditor::Graph::init # \"%s\" %f to %f %p", 
	      m_label.c_str(), m_min, m_max, m_values);
  }

  void GraphEditor::Graph::setPoint(uint i, float v)
  {
    if(i >= m_size)
      VEE_fatal("GraphEditor::Graph::setPoint # %u >= %u",
		i, m_size);

    * VEE_Ptr::offset(m_values, m_stride * i) = v;
  }

  int GraphEditor::Graph::findIndex(float time)
  {
    if(!m_size)
      return -1;

    int index = 0;
    float diff = VEE_abs(m_times[0] - time);

    for(uint i=1; i < m_size; i++) {
      float d = VEE_abs(* VEE_Ptr::offset(m_times, i * m_stride) - time);
      if(diff > d) {
	diff = d;
	index = i;
      }
    }

    return index;
  }

  /////////////////////////////////////////////////////////////////////////////

  static const float saturation = 0.5;
  static const float invSaturation = 1.0 - saturation;

  static const uint colorCount = 7;

  static VEE_Rgba colors[colorCount] = {
    VEE_Rgba::Gray * invSaturation + VEE_Rgba::Red     * saturation,
    VEE_Rgba::Gray * invSaturation + VEE_Rgba::Green   * saturation,
    VEE_Rgba::Gray * invSaturation + VEE_Rgba::Blue    * saturation,

    VEE_Rgba::Gray * invSaturation + VEE_Rgba::Cyan    * saturation,
    VEE_Rgba::Gray * invSaturation + VEE_Rgba::Magenta * saturation,
    VEE_Rgba::Gray * invSaturation + VEE_Rgba::Yellow  * saturation,

    VEE_Rgba::Gray
  };

  GraphEditor::GraphEditor()
    : m_selected(-1),
      m_prevIndex(-1),
      m_prevValue(0.0),
      m_timePerUnit(1.0)
  {
    m_fontSize = 0.1;
  }

  GraphEditor::~GraphEditor()
  {}

  void GraphEditor::handleEvent(const InEvent *e)
  {
    TextureButton::handleEvent(e);

    Vector3 loc = applyMyTransform(e->location());

    if(loc.x < boundingBox().low().x - 0.05) {

      for(uint i=0; i < m_graphs.size(); i++) {

	if(m_graphs[i].inside(Vector2(loc.x, loc.y))) {

	  if(e->buttonState() & m_buttonMask) {
	    transmitSound(ColorGroup::ACTIVE, loc);
	    m_selected = i;
	  }
	  else
	    m_state = GRAB;

	}
      }

      m_prevIndex = -1;
      return;
    }

    if(m_state != GRAB) {
      m_prevIndex = -1;
      return;
    }

    Graph * g = & m_graphs[m_selected];

    float time = (loc.x - low().x) * m_timePerUnit;
    
    int index = g->findIndex(time);

    if(index < 0)
      return;

    float h = high().y - low().y;
    float y = (loc.y - low().y) / h;
    // float scale = g->renderScale();
    float value = y * (g->m_max - g->m_min) + g->m_min;

    if(g->m_limitMin && value < g->m_min)
      value = g->m_min;
    if(g->m_limitMax && value > g->m_max)
      value = g->m_max;

    g->setPoint(index, value);

    if(index != m_prevIndex && m_prevIndex >= 0) {
      // Interpolate the values between this and the previous hit:

      int step = m_prevIndex < index ? 1 : -1;
      float span = m_prevIndex - index;

      for(int i = m_prevIndex; i != index; i+= step) {

	float diff = i - index;
	float w = diff / span;

	g->setPoint(i, m_prevValue * w + (1.0f - w) * value);
	
      }
    }

    VEE_trace("GraphEditor::handleEvent # %s [%d] %f", 
	      g->m_label.c_str(), index, value);

    m_prevValue = value;
    m_prevIndex = index;
  }

  void GraphEditor::renderGl(GlData *gld, bool transform)
  {
    pushLocalMatrix(transform);

    TextureButton::renderGl(gld, false);
    
    const float step = 0.01;
    float offset = step;
    float yspan = boundingBox().high().y - low().y;

    glDisable(GL_BLEND);
    glDisable(GL_TEXTURE_2D);

    glPushMatrix();
    glTranslatef(low().x, low().y, 0.0);
    glScalef(1.0 / m_timePerUnit, yspan, 1.0);

    /* Render the graphs to different levels to make them look
       better (no flashing). */
    int i=0;
    for(iterator it = m_graphs.begin(); it != m_graphs.end(); it++) {

      if(i != m_selected) {
	
	glPushMatrix();
	glTranslatef(0.0, 0.0, offset);
	
	(*it).renderGl(gld);
	
	glPopMatrix();
      
	offset += step;
      }
      i++;
    }

    // Finally render the selected graph to the top:
    if(m_selected >= 0) {
      glPushMatrix();
      glTranslatef(0.0, 0.0, offset);
      
      m_graphs[m_selected].renderGl(gld);
      
      glPopMatrix();
    }

    glPopMatrix();

    Vector2 tes = determineTextEntrySize(gld);
    tes *= 1.2;

    Vector2 bottomLeft, topRight(low().x, high().y);

    i = 0;

    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    for(iterator it = m_graphs.begin(); it != m_graphs.end(); it++) {

      bottomLeft = topRight - tes;

      // Mark the selected graph
      if(i == m_selected)
	bottomLeft.x -= 0.2;

      (*it).m_low  = bottomLeft;
      (*it).m_high = topRight;

      glColor4fv((*it).m_color.data());

      glDisable(GL_BLEND);
      glDisable(GL_TEXTURE_2D);
      glBegin(GL_QUADS);
      
      glVertex2f(bottomLeft.x, bottomLeft.y);
      glVertex2f(topRight.x, bottomLeft.y);
      glVertex2f(topRight.x, topRight.y);
      glVertex2f(bottomLeft.x, topRight.y);

      glEnd();

      glEnable(GL_BLEND);
      glEnable(GL_TEXTURE_2D);
      glColor4fv(VEE_Rgba::Black.data());

      gld->m_fntRenderer.start3f(bottomLeft.x + 0.1,
				 ((*it).m_high.y + (*it).m_low.y) * 0.5 - 
				 m_fontSize * 0.5,
				 step);

      gld->m_fntRenderer.puts((*it).m_label.c_str());

      topRight.y -= tes.y;
      i++;
    }

    popLocalMatrix(transform);
  }

  void GraphEditor::update()
  {}

  void GraphEditor::addGraph(const char * label, ValueGraphBase *graph)
  {
    bool loop = true;

    for(uint i=0; loop; i++) {

      Graph g;
      float *low = 0, * high = 0;
      const char * id = 0;
      loop = graph->getParameters(i, & g.m_times, & g.m_values, 
				  & g.m_stride, & g.m_size, & id, 
				  & low, & high);

      if(loop) {
	m_selected = m_graphs.size();

	g.m_label = label;
	if(id && strlen(id))
	  g.m_label += std::string(".") + id;

	g.init();
	g.m_color = colors[m_graphs.size() % colorCount];

	if(low) {
	  g.m_min = *low;
	  g.m_limitMin = true;
	}
	if(high) {
	  g.m_max = *high;
	  g.m_limitMax = true;
	}

	m_graphs.push_back(g);
      }
      
    }
  }

  Vector2 GraphEditor::determineTextEntrySize(GlData *gld)
  {
    float w = 0.0, h = 0.0;
    float x1, x2, y1, y2;

    gld->m_fntRenderer.setPointSize(m_fontSize);

    for(uint i=0; i < m_graphs.size(); i++) {
      gld->m_font.getBBox(m_graphs[i].m_label.c_str(), 
			  m_fontSize, 0.0,
			  &x1, &x2, &y1, &y2);
      float wt = x2 - x1;
      float ht = y2 - y1;

      if(w < wt)
	w = wt;
      if(h < ht)
	h = ht;
    }

    return Vector2(w, h);
  }

}
