/* -*- C++ -*- */

/* -*- C++ -*- */

/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Crystal animation and 3D modeling application
 *
 * Authors:   Tommi Ilmonen 2001-2004
 * Copyright: Tommi Ilmonen and/or Tekes 2001-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <crystal_helma_editor.h>

#include <crystal_dom_util.h>
#include <crystal_engine.h>
#include <crystal_gee_config.h>
#include <crystal_gee_mesh.h>
#include <crystal_gl_data.h>
#include <crystal_tools.h>

#include <gee_containment.h>
#include <gee_engine.h>
#include <gee_path.h>
#include <gee_particle_object.h>
#include <gee_sphere.h>

#include <vee_bounding_square.h>
#include <vee_error.h>
#include <vee_math.h>
#include <vee_values.h>

#include <Fluid/Message.hpp>
#include <Fluid/MessageInt.hpp>
#include <Fluid/MessageString.hpp>
#include <Fluid/TrackerSensorSelector.hpp>

#include <mj_array2dtmpl.C>

#include <qfile.h>
#include <qtextstream.h> 

#include <GL/gl.h>

namespace Crystal { 

  QDomElement WidgetGroup::Connection::writeDom(QDomDocument &doc)
  {
    QDomElement e = doc.createElement("connection");

    e.setAttribute("sourcename", m_sourceName.c_str());
    e.setAttribute("targetname", m_targetName.c_str());
    e.setAttribute("targetid",   m_targetId.c_str());
    
    return e;
  }

  bool WidgetGroup::Connection::readDom(QDomElement &e)
  {
    if(e.tagName() != "connection")
      return false;

    m_sourceName = e.attribute("sourcename").ascii();
    m_targetName = e.attribute("targetname").ascii();
    m_targetId   = e.attribute("targetid").ascii();

    return true;
  }

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  QDomElement WidgetGroup::Object::writeConnections(QDomElement &e, 
						    QDomDocument &doc)
  {
    for(std::list<Connection>::iterator it = m_connections.begin();
	it != m_connections.end(); it++)
      e.appendChild((*it).writeDom(doc));
    
    return e;
  }
  
  bool WidgetGroup::Object::readConnections(QDomElement &e)
  {
    // VEE_trace("WidgetGroup::Object::readConnections");

    QDomNode n = e.namedItem("connection");
    
    while(!n.isNull()) {
      QDomElement ne = n.toElement();
      
      if(ne.tagName() == "connection") {
	Connection r;
	r.readDom(ne);
	m_connections.push_back(r);
      }
      
      n = n.nextSibling();
    }
    
    return true;
  }
  
  QDomElement WidgetGroup::Object::writeSize(QDomElement &e, 
					     QDomDocument &doc)
  {
    addTextChild(e, doc, "rows", m_rows);
    addTextChild(e, doc, "cols", m_cols);

    return e;
  }

  bool WidgetGroup::Object::readSize(QDomElement &e)
  {
    m_rows = getInt(e, "rows", m_rows);
    m_cols = getInt(e, "cols", m_cols);

    return true;
  }

  void WidgetGroup::Object::applyConnections(HelmaEditor *e, HelmaTab *tab)
  {
    Engine * engine = e->host()->engine();
    GeeConfig * config = engine->geeConfig();
    Demo * demo = tab->demo();

    for(iterator it = m_connections.begin(); it != m_connections.end(); it++) {
      Connection & c = (*it);
      MessageSink * target = 0;

      if(c.m_targetId == "config")
	target = config;
      if(c.m_targetId == "demo")
	target = demo;
      if(c.m_targetId == "editor")
	target = e;
      if(c.m_targetId == "tab")
	target = tab;
      if(c.m_targetId == "engine")
	target = engine;
      
      Widget * w = m_widget.ptr();

      if(target && w)
	w->msgAddSink(c.m_sourceName, c.m_targetName, target);
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  WidgetGroup::WidgetGroup(Transformation * host, HelmaTab * tab)
    : Transformation(host),
      m_widgets(tab ? tab->host()->host()->grabs() : 0),
      m_size(1.5, 1.5, 1.5),
      m_border(0.05),
      m_widgetSpace(0.03),
      m_move(STATIONARY),
      m_tab(tab),
      m_trackHand(-1)
  {}

  WidgetGroup::~WidgetGroup()
  {}

  QDomElement WidgetGroup::writeDom(QDomDocument &doc)
  {
    QDomElement e = doc.createElement("group");
    
    addTextChild(e, doc, "lowerleft", m_location);
    addTextChild(e, doc, "size",      m_size);
    addTextChild(e, doc, "rotation",  m_rotation);
    addTextChild(e, doc, "rows",(int) m_grid.rows());
    addTextChild(e, doc, "cols",(int) m_grid.cols());
    if(m_id.size())
      addTextChild(e, doc, "id", m_id.c_str());

    QDomElement widgets = doc.createElement("widgets");

    for(iterator it = m_objects.begin(); it != m_objects.end(); it++) {
      QDomElement we = (*it).m_widget.ptr()->writeDom(doc);
      (*it).writeConnections(we, doc);
      (*it).writeSize(we, doc);
      widgets.appendChild(we);
    }

    e.appendChild(widgets);

    return e;
  }

  bool WidgetGroup::readDom(QDomElement &e)
  {
    const char * fname = "WidgetGroup::readDom";

    m_location = getVector3(e, "lowerleft", m_location);
    m_size     = getVector3(e, "size",      m_size);
    m_rotation = getMatrix3(e, "rotation",  m_rotation);
    int rows   = getInt(e, "rows", m_grid.rows());
    int cols   = getInt(e, "cols", m_grid.cols());
    m_id       = getString(e, "id", "").ascii();
    
    m_grid.resize(rows, cols);
    
    QDomElement widgets = e.namedItem("widgets").toElement();

    QDomNode n = widgets.firstChild();
    
    while ( !n.isNull() ) {
      if ( n.isElement() ) {
	QDomElement ne = n.toElement();

	QString tag = ne.tagName();

	Widget * w = Widget::create(tag.ascii());

	VEE_trace("%s # %s -> %p", fname, tag.ascii(), w);

	if(!w) continue;

	w->setHostTransformation(this);

	if(!w->readDom(ne)) {
	  VEE_error("%s # Read failed for %s", fname, w->type());
	  delete w;
	}
	else {
	  Object o;
	  o.m_widget = w;
	  o.readConnections(ne);
	  o.readSize(ne);
	  o.applyConnections(m_tab->host(), m_tab);
	  addObject(o);
	}
      }
      n = n.nextSibling();
    }

    calculateLayout();

    VEE_trace("%s # EXIT", fname);

    return true;
  }

  void WidgetGroup::calculateLayout(bool signal)
  {
    const char * fname = "WidgetGroup::calculateLayout";

    VEE_trace("%s # %d widgets to %d x %d grid", 
	      fname, (int) m_objects.size(), 
	      (int) m_grid.rows(), (int) m_grid.cols());

    m_grid.setAll(false);

    Vector3 cellSize = m_size;
    cellSize.descale(Vector3(m_grid.cols(), m_grid.rows(), 1));

    for(iterator it = m_objects.begin(); it != m_objects.end(); it++) {

      Object & o = (*it);
      int rows = o.m_rows;
      int cols = o.m_cols;

      Widget * w = o.m_widget.ptr();
      if(!w) 
	VEE_fatal("%s # NULL widget", fname);

      // Now seek an area in the grid for the widget:

      int row, col;
      bool found = false;

      // Print the bloody grid to the screen:
      /* printf("\nGRID: \n");
      for(uint ii = 0; ii < m_grid.rows(); ii++) {
	for(uint jj = 0; jj < m_grid.cols(); jj++)
	  printf("%d  ", (int) m_grid.get(ii, jj));
	printf("\n");
      }

      fflush(0); */

      /* Iterate over each possible top-left cell. The layout begins
	 from top-left cell and proceeds one column at a time.  */
      for(row = 0; row + rows <= (int) m_grid.rows() && !found; row++) {
	for(col = 0; col + cols <= (int) m_grid.cols() && !found; col++) {
	  bool occupied = false;

	  // Check if the next cell is a possible target:
	  for(int colcheck = col; colcheck < col + o.m_cols; colcheck++)
	    for(int rowcheck = row; rowcheck < row + o.m_rows; rowcheck++)
	      if(m_grid.get(rowcheck, colcheck))
		occupied = true;
	  
	  if(!occupied)
	    found = true;
	}
      }
      
      if(!found) {
	VEE_error("%s # Failed to locate %s (needing %d %d)", 
		  fname, typeid(*w).name(), rows, cols);
	continue;
      }

      // Get rid of the extra "++" ofthe previous for-loops:
      col--;
      row--;

      VEE_trace("%s # Positioning %s to %d, %d", 
		fname, typeid(*w).name(), row, col);
      
      Vector3 vlow = 
	Vector3(col * cellSize.x, m_size.y - (row + 1) * cellSize.y, 0);
      Vector3 vspan = Vector3(cols * cellSize.x, rows * cellSize.y, 0);
      vspan -= (m_widgetSpace * 2.0f);
      if(w->useExtraZ())
	vspan.z = vspan.x;
      else
	vspan.z = VEE_min(vspan.x, vspan.y);
      
      w->setLocation(vlow + m_widgetSpace);
      w->boundingBox().set(VEE_Vector3ZERO, vspan);

      // Make sure the widget is completely in front of the frame:
      /* Matrix4 r;
	 r.identity();
	 r.setRotation(w->rotation()); */
      VEE_BoundingBox b = w->rotation() * w->boundingBox();
      Vector3 l = w->location();
      l.z -= b.low().z;
      w->setLocation(l);

      // Now we mark the newly occupied cells:

      int endcol = col + o.m_cols;
      int endrow = row + o.m_rows;

      for(; col < endcol; col++)
	for(int srow = row; srow < endrow; srow++)
	  m_grid.get(srow, col) = true;	  
    }
    
    if(signal) {
      MessageInt mi(0);
      msgSend("layoutchange", &mi);
    }
  }
  
  void WidgetGroup::addObject(Object &o)
  { 
    m_objects.push_back(o);
    m_widgets.addWidget(o.m_widget.ptr()); 
  }
  
  void WidgetGroup::activate()
  {
    for(iterator it = m_objects.begin(); it != m_objects.end(); it++)
      (*it).m_widget.ptr()->transmitValue();
  }

  void WidgetGroup::renderGl(GlData * gld) 
  { 
    glPushMatrix();

    GlData::applyGlTransform(m_rotation, m_location);
    glScalef(m_scale, m_scale, m_scale);

    m_widgets.renderGl(gld); 
    
    // Draw the frame surrounding the group:

    glDisable(GL_TEXTURE_2D);
    
    float normalColor[4] = {1, 1, 1, 1};
    float moveColor[4]   = {1, 0.5, 0, 1};

    glColor4fv(m_move != STATIONARY ? moveColor : normalColor);

    Vector2 borderv;
    borderv.make(m_border * 2.0);

    Widget::frameGl(- borderv, 
		    m_size.vector2() + borderv, 
		    m_border);

    glPopMatrix();
  }

  
  void WidgetGroup::beforeEvents()
  {
    m_widgets.beforeEvents();
  }

  void WidgetGroup::handleEvent(const InEvent *e) 
  { 
    checkMove(e);
    
    InEvent e2(*e);
  
    e2.setLocation(applyMyTransform(e->location()));
    e2.setRotation(applyMyTransform(e->rotation()));

    if(m_move == STATIONARY)
      m_widgets.handleEvent(&e2);
  }

  void WidgetGroup::calculatePreferredSize(uint widgets, 
					   float widgetAspectRatio,
					   uint * rows, uint * cols)
  {
    float error = 1.0e+10;

    *rows = widgets;
    *cols = 1;

    for(uint c = 1; c <= widgets; c++) {

      float r = ceil(widgets / (double) c);
      float w = m_size.x / (float) c;
      float h = m_size.y / (float) r;
      float thisaspect = w / h;

      float thiserror = VEE_abs(thisaspect - widgetAspectRatio);
      
      if(thiserror < error) {
	*rows = (uint) r;
	*cols = c;
	error = thiserror;
      }
    }
  }

  void WidgetGroup::calculateBoundingBox(VEE_BoundingBox *bbox)
  {
    m_widgets.calculateBoundingBox(bbox);
  }

  void WidgetGroup::checkMove(const InEvent *e)
  {
    if(m_widgets.grabs()->someOneElsesGrab(this))
      return;
    
    if((m_trackHand >= 0) && (m_trackHand != e->hand()))
      return;

    InEvent e2(*e);
  
    e2.setLocation(applyMyTransform(e->location()));
    e2.setRotation(applyMyTransform(e->rotation()));

    if(e2.buttonState() & DEFAULT_MASK) {
      if(m_move != STATIONARY) {
	doMove((m_move == MOVE) ? e : &e2);
	return;
      }
    }
    else {
      m_move = STATIONARY;
      m_trackHand = -1;
      if(m_widgets.grabs()->myGrab(this))
	if(!m_widgets.grabs()->ungrab(this))
	  VEE_error("WidgetGroup::checkMove # ungrab failed");
      return;
    }
    
    // The inner square where widgets are:
    VEE_BoundingSquare inner(Vector2(0, 0), m_size.vector2());

    if(inner.contains(e2.location().vector2()))
      return;

    VEE_BoundingSquare outer(inner);

    Vector2 borderv;
    borderv.make(m_border * 2.0);

    outer.low()  -= borderv;
    outer.high() += borderv;

    if(!outer.contains(e2.location().vector2()))
      return;
    
    if(VEE_abs(e2.location().z) > 0.5)
      return;

    if(e2.location().x > inner.high().x)
      m_move = STRETCH_X;
    else if(e2.location().y < inner.low().y)
      m_move = STRETCH_Y;
    else
      m_move = MOVE;

    m_prevMove = (m_move == MOVE) ? e->location() : e2.location();
    m_widgets.grabs()->grab(this);
    m_trackHand = e->hand();
  }

  void WidgetGroup::doMove(const InEvent *e)
  {
    if(m_move == MOVE) {
      m_location += (e->location() - m_prevMove);
      m_prevMove = e->location();
      MessageInt mi(0);
      msgSend("move", &mi);
    }
    else if(m_move == STRETCH_X) {
      float diff = e->location().x - m_prevMove.x;
      m_size.x += diff;
      calculateLayout();
      m_prevMove = e->location();
    }
    else {
      float diff = e->location().y - m_prevMove.y;
      m_size.y -= diff;
      m_location.y += diff;
      calculateLayout();
      m_prevMove = e->location();
      m_prevMove.y -= diff;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  Demo::Demo(Engine *e)
    : m_recorders(new GeeRecorders),
      m_engine(new GEE_Engine),
      m_cengine(e)
  {
    m_recorders->init(e, m_engine);
    m_config = e->geeConfig();

    m_engine->setGarbageBin(e->veeGarbage());

    m_coolDemos.insert("Lines");
    m_coolDemos.insert("Mesh");
    m_coolDemos.insert("MultiBrush");
    m_coolDemos.insert("Particles");
    // m_coolDemos.insert("Check");
  }

  Demo::~Demo()
  {}

  bool Demo::msgProcess(const char *id, Fluid::Message *)
  {
    // VEE_trace("Demo::msgProcess # %s", id);

    bool ok = false;

    if(strcmp(id, "update") == 0) {
      updateDemo();
      ok = true;
    }
    else if(strncmp(id, "tool-", 5) == 0) {
      m_recorderName = id + 5;
      updateDemo();
      ok = true;
    }
    else if(strcmp(id, "retexture") == 0) {
      retexture();
      ok = true;
    }
    else if(strcmp(id, "recolor") == 0) {
      recolor();
      ok = true;
    }

    return ok;
  }
  
  void Demo::updateDemo()
  {
    VEE_trace("Demo::updateDemo");

    // Clean up previous crap:
    m_engine->actions().stop();
    m_engine->world().clear();
    m_engine->actions().clear();
    // m_engine->actions().undo(2);
    // m_engine->actions().undo(m_engine->actions().actionCount());

    if(m_coolDemos.find(m_recorderName) != m_coolDemos.end()) {

      // Now rebuild the demo:
      GeePathToX * rec = m_recorders->getRecorder(m_recorderName);
      
      if(!rec)
	return;

      if(m_recorderName != "MultiBrush")
	rec->setIgnoreNapra(true);
      
      rec->activate(TimeStamp(0));
      rec->start(TimeStamp(0));
      
      rec->pathA1()->create3DSpiral(3, 30);
      rec->pathA2()->create3DSpiral(3, 30);

      m_engine->rewind(0);
      m_engine->update(TimeStamp().setSeconds(20.0));

      rec->deactivate(TimeStamp().setSeconds(25.0));

      m_engine->actions().stop();

      m_coolDemo = true;
    }
    else {
      GEE_Sphere * sphere = new GEE_Sphere();

      float radius = m_cengine->geeConfig()->getValueFloat("radius", 1.0);
      VEE_Vector4 color(m_cengine->geeEngine()->world().
			backgroundColor().data());

      for(int c = 0; c < 3; c++)
	color[c] = color[c] > 0.5 ? 0.0 : 1.0;

      GEE_Material mat;
      mat.setColor(color);

      sphere->setRadius(radius);
      sphere->setMaterial(mat);

      m_engine->world().addGeometry(sphere);

      m_coolDemo = false;
    }

    VEE_trace("Demo::updateDemo # EXIT");
  }

  void Demo::retexture()
  {
    std::string texture = m_config->getValueString("texture", "noise-1");
    int textureId = GEE_World::textureId(texture);
   
    VEE_trace("Demo::retexture # %s %d", texture.c_str(), textureId);
 
    for(GEE_World::iterator it = m_engine->world().begin();
	it != m_engine->world().end(); it++) {

      GEE_Geometry * g = GEE_World::getGeometry(it).ptr();
      GEE_ParticleObject * po = dynamic_cast<GEE_ParticleObject *> (g);
      if(po) {
	for(GEE_ParticleObject::iterator it2 = po->begin();
	    it2 != po->end(); it2++) {
	  GEE_ParticleData & pd = GEE_ParticleObject::getParticleData(it2);
	  pd.setTexture(texture, textureId);
	}
      }
    }
  }

  void Demo::recolor()
  {
    const GEE_Material &m = m_config->material(m_recorderName);

    for(GEE_World::iterator it = m_engine->world().begin();
	it != m_engine->world().end(); it++) {

      GEE_Geometry * g = GEE_World::getGeometry(it).ptr();
      g->setMaterial(m);
    }
  }

  void Demo::render(GEE_Renderer *r)
  {
    m_engine->world().render(r);
  }

  void Demo::setGeometryTransformation(const Matrix4 &m)
  {
    VEE_Matrix4d md(m.data());
    for(GEE_World::iterator it = m_engine->world().begin();
	it != m_engine->world().end(); it++) {
      
      GEE_Geometry * g = GEE_World::getGeometry(it).ptr();
      g->setTransformation(md);
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  HelmaTab::HelmaTab(Transformation *t, HelmaEditor *he)
    : m_transform(t),
      m_host(he),
      m_demo(he->host()->engine())
  {}

  HelmaTab::~HelmaTab()
  {}

  bool HelmaTab::msgProcess(const char *, Fluid::Message *)
  {
    return false;
  }

  QDomElement HelmaTab::writeDom(QDomDocument &doc)
  {
    QDomElement e = doc.createElement("tab");

    QDomAttr aname = doc.createAttribute("name");
    aname.setValue(m_name.c_str());
    e.attributes().setNamedItem(aname);

    QDomAttr atool = doc.createAttribute("tool");
    atool.setValue(m_tool.c_str());
    e.attributes().setNamedItem(atool);

    for(iterator it = m_groups.begin(); it != m_groups.end(); it++) {
      e.appendChild((*it).ptr()->writeDom(doc));
    }

    return e;
  }

  bool HelmaTab::readDom(QDomElement &elem)
  {
    if(elem.tagName() != "tab")
      return false;

    m_name = elem.attributes().namedItem("name").nodeValue().ascii();
    m_tool = elem.attributes().namedItem("tool").nodeValue().ascii();
    
    QDomNode n = elem.firstChild();
    
    while ( !n.isNull() ) {
      if ( n.isElement() ) {
	QDomElement e = n.toElement();
	if(e.tagName() == "group") {
	  WidgetGroup * wg = new WidgetGroup(m_transform, this);
	  wg->readDom(e);
	  m_groups.push_back(wg);
	}
      }
      n = n.nextSibling();
    }

    activate();

    m_demo.setTool(m_tool);

    return true;
  }

  void HelmaTab::activate()
  {
    for(iterator it = m_groups.begin(); it != m_groups.end(); it++) 
      (*it).ptr()->activate();
  }

  void HelmaTab::update()
  {
    for(iterator it = m_groups.begin(); it != m_groups.end(); it++) 
      (*it).ptr()->update();
  }

  void HelmaTab::renderGl(GlData * gld)
  {
    for(iterator it = m_groups.begin(); it != m_groups.end(); it++) 
      (*it).ptr()->renderGl(gld);
  }

  void HelmaTab::beforeEvents()
  {
    for(iterator it = m_groups.begin(); it != m_groups.end(); it++) 
      (*it).ptr()->beforeEvents();
  }

  void HelmaTab::handleEvent(const InEvent *e)
  {
    for(iterator it = m_groups.begin(); it != m_groups.end(); it++) 
      (*it).ptr()->handleEvent(e);
  }

  WidgetGroup * HelmaTab::widgetGroup(const std::string &id, bool alwaysGetOne)
  {
    WidgetGroup * any = 0;

    for(iterator it = m_groups.begin(); it != m_groups.end(); it++) {
      WidgetGroup * wg = (*it).ptr();
      if(wg->id() == id)
	return wg;
      else if(!any)
	any = wg;
    }

    return alwaysGetOne ? any : 0;
  }

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  Gallery::Gallery(Transformation *t, HelmaEditor *h)
    : HelmaTab(t, h),
      m_group(new WidgetGroup(t, this)),
      m_delete(-1)
  {
    VEE_RefPtr<WidgetGroup> tmp(m_group);
    addGroup(tmp);

    VEE_trace("Gallery::Gallery");

    m_name = "Gallery";

    m_group->msgAddSink("layoutchange", "layoutchange", this);
    m_group->msgAddSink("move", "layoutmove", this);
    
    h->host()->msgAddSink("move", "layoutmove", this);
    h->host()->engine()->msgAddSink("newfile", "layoutchange", this);

    rebuild();
  }

  Gallery::~Gallery()
  {}
  
  void Gallery::activate()
  {
    HelmaTab::activate();

    recalculateTransforms();
  }

  /** Pushes all the models into the global collector renderer. */

  void Gallery::update()
  {
    GEE_CollectorRenderer * r = 
      & m_host->host()->engine()->geeEngine()->collector();
    
    for(uint i = 0; i < m_items.size(); i++) {
      Item & item = m_items[i];
      r->pushTransform(item.m_transform);
      item.m_model.ptr()->render(r);
      r->popTransform();
    }    
  }

  bool Gallery::msgProcess(const char * id, Message *m)
  {
    bool ok = false;

    if(strcmp(id, "snapshot") == 0) {
      takeSnapshot(false);
      ok = true;
    }
    else if(strcmp(id, "selectedshot") == 0) {
      takeSnapshot(true);
      ok = true;
    }
    else if(strcmp(id, "layoutmove") == 0) {
      recalculateTransforms();
      ok = true;
    }
    else if(strcmp(id, "layoutchange") == 0) {
      rebuild();
      ok = true;
    }
    else if(strncmp(id, "erase_", 6) == 0) {
      m_delete = atoi(id + 6);
      ok = true;
    }
    else if(strncmp(id, "insert_", 7) == 0) {
      int index = atoi(id + 7);
      insert(index);
      ok = true;
    }
    else if(strcmp(id, "merge") == 0) {
      MessageString * ms = dynamic_cast<MessageString *> (m);
      if(!ms)
	VEE_error("Gallery::msgProcess # message does not contain a string");
      else
	merge(ms->string().c_str());
    }

    return ok;
  }

  QDomElement Gallery::writeDom(QDomDocument &doc)
  {
    QDomElement e = HelmaTab::writeDom(doc);
    
    e.toElement().setAttribute("special", "Gallery");
    
    return e;
  }

  bool Gallery::readDom(QDomElement &elem)
  {
    VEE_trace("Gallery::readDom");

    m_name = elem.attributes().namedItem("name").nodeValue().ascii();
    m_tool = elem.attributes().namedItem("tool").nodeValue().ascii();

    QDomNode n = elem.firstChild();
    
    while ( !n.isNull() ) {
      if ( n.isElement() ) {
	QDomElement e = n.toElement();
	if(e.tagName() == "group" && 
	   e.namedItem("id").toElement().text() == "tools") {
	  WidgetGroup * wg = new WidgetGroup(m_transform, this);
	  wg->readDom(e);
	  m_groups.push_back(wg);
	  VEE_trace("Gallery::readDom # Got a group with %u widgets",
		    wg->widgets()->size());
	}
      }
      n = n.nextSibling();
    }

    rebuild();

    activate();

    return true;
  }

  void Gallery::takeSnapshot(bool onlySelected)
  {
    VEE_trace("Gallery::takeSnapshot");

    GEE_Engine * ge = m_host->host()->engine()->geeEngine();

    GEE_Gallery * gal = & ge->gallery();
    
    GEE_id_t id = gal->takeSnapshot( & ge->world(), -1, onlySelected);

    GEE_World * w = gal->item(id);

    if(w) {
      w->setAllGenerations(0);
    }
   
    rebuild();
  }

  void Gallery::handleEvent(const InEvent *e)
  {
    HelmaTab::handleEvent(e);

    if(m_delete >= 0) {
      erase(m_delete);
      rebuild();
      m_delete = -1;
    }
  }

  /** Rebuilds the widgets etc.*/

  void Gallery::rebuild()
  {
    VEE_trace("Gallery::rebuild");

    char buf[64];

    GEE_Gallery * g = & m_host->host()->engine()->geeEngine()->gallery();

    uint n = g->size();

    m_items.clear();
    m_items.resize(n);
    uint i = 0;

    uint rows, cols;

    m_group->clear();
    m_group->calculatePreferredSize(n, 1.0, &rows, & cols);
    m_group->resizeGrid(rows, cols);

    // Then gallery items:
    for(GEE_Gallery::iterator it = g->begin(); it != g->end(); it++) {

      Item & item = m_items[i];
            
      item.m_model = GEE_Gallery::getWorld(it);

      CubeWidget * cw = new CubeWidget(m_group);
      item.m_cube = cw;

      WidgetGroup::Object gobj;
      gobj.m_widget = item.m_cube;

      m_group->addObject(gobj);

      Menu * menu = new Menu(m_group);

      menu->setLabel("Model");
      
      sprintf(buf, "erase_%u", i);
      menu->addItem("Erase", buf, this);

      sprintf(buf, "insert_%u", i);
      menu->addItem("Insert", buf, this);

      menu->setMenuBelow(false);

      item.m_menu = menu;

      m_group->widgets()->addWidget(menu);
      
      i++;
    }

    m_group->calculateLayout(false);
    recalculateTransforms();
    
    /* The transformations get calculated in the previous call. */
  }

  /** Sets the model transformations so that they fit inside the
      cubes. */

  void Gallery::recalculateTransforms()
  {
    const char * fname = "Gallery::recalculateTransforms";

    VEE_trace(fname);

    // Move the worlds to their final positions
    for(uint i = 0; i < m_items.size(); i++) {

      Item & item = m_items[i];

      Widget * w = item.m_cube.ptr();

      Matrix4 m, m2;
      m.identity();
      m2 = m;

      VEE_BoundingBox bw, bc; // World and Cube centers

      item.m_model.ptr()->getBoundingBox(&bw);
      bc = w->boundingBox();
      
      /*
      VEE_trace("Gallery::recalculateTransforms # wbox = %f %f %f - %f %f %f",
		bw.low().x,  bw.low().y,  bw.low().z, 
		bw.high().x, bw.high().y, bw.high().z);*/

      // Move model to the origin:
      m.setColumn3(3, -bw.center());

      Vector3 vscale = bc.span();
      vscale.descale(bw.span() / w->totalScale());
      
      float spanMin = vscale.minimum();
      
      // float scale = (spanMin > 1.0e-6) ? (1.0f / spanMin) : 1.0;
      float scale = spanMin;

      // Scale the model to fit inside the cube:
      m2[0][0] = m2[1][1] = m2[2][2] = scale;

      m = m2 * m;

      // Rotate the model with the rotation of the cube:
      /* m2.setRotation(w->worldRotation());
	 m = m2 * m; */
      
      // Move the object to the cube center:
      m2.identity();
      m2.setColumn3(3, w->worldLocation(bc.center()));

      item.m_transform = m2 * m;

      // Now set the location of the corresponding menu

      Menu * menu = item.m_menu.ptr();

      menu->setTransformation(*w);
      Vector3 l = menu->location();
      l.z += bc.high().z + 0.001;
      menu->setLocation(l);

      Vector2 area = bc.span().vector2();

      menu->setArea(Vector2(0, 0), Vector2(area.x * 0.75, area.y * 0.2));

      VEE_trace("%s # s = %f", fname, scale);
      std::cout << "spans = " << bc.span() << " # "  << bw.span() << "\n";
    }
  }

  void Gallery::merge(const char * file)
  {
    VEE_trace("Gallery::merge # %s", file);
    
    std::string tmp(file);

    m_host->host()->engine()->geeEngine()->
      mergeGalleriesFrom((tmp + ".gee").c_str());

    rebuild();
  }

  void Gallery::erase(int index)
  {
    GEE_Gallery * g = & m_host->host()->engine()->geeEngine()->gallery();

    GEE_Gallery::iterator it = g->getIndex(index);

    if(it == g->end()) {
      VEE_error("Gallery::erase # Index %d out of range",
		index);
      return;
    }

    g->erase(it);
  }
   
  void Gallery::insert(int index)
  {
    VEE_trace("Gallery::insert # %d", index);

    Engine * e = m_host->host()->engine();

    GEE_Gallery * g = & e->geeEngine()->gallery();

    GEE_Gallery::iterator it = g->getIndex(index);

    if(it == g->end()) {
      VEE_error("Gallery::insert # Index %u out of range",
		index);
      return;
    }
    
    e->geeConfig()->setValueInt("insert_source_id", 
				GEE_Gallery::getWorld(it).ptr()->id());

    e->skipMode(Engine::MODE_PLAY);
  }

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  HelmaEditor::HelmaEditor(Tools *host, const char * uiFile)
    : m_host(host),
      m_currentTab(0),
      m_tabArea(host, this),
      m_tabButtons(new WidgetGroup(host, &m_tabArea)),
      m_widgets(host->grabs())
  {
    char buf[64], buf2[64];

    m_tabButtons.ptr()->setLocation(Vector3(-2, 1, 0));
    m_tabArea.addGroup(m_tabButtons);

    m_host->engine()->setPlaySpeed(0.0f);
    
    GeeConfig * gc = m_host->engine()->geeConfig();

    gc->setValue("timescale", VEE_ValueFloat(0.0));

    const char * fname = "HelmaEditor::HelmaEditor";

    VEE_trace(fname);

    // Load UI spec from an XML file:
    QFile f(uiFile);
  
    if ( !f.open( IO_ReadOnly ) ) {
      VEE_error("%s # Open failed", fname);
      return;
    }
    QString error;
    int errline = -1;
    
    if(!m_uiSetup.setContent( &f, & error, &errline))
      VEE_error("%s # failed on line %d (%s)", fname, errline, error.ascii());
    
    VEE_Vector2 b1, b2, b3;
    b1.make(-1.0, 2.6);
    b2.make(0.8, 0.2);
    b3.make(b2.x + 0.2, 0.0);

    static const char * unredo[] = { 
      "undo", 
      "redo", 
      "clear",
      0
    };

    const char * label;
    for(int i = 0; (label = unredo[i]); i++) {
      TextureButton * txb = new TextureButton(m_host);
      
      txb->setArea(b1, b2);
      txb->setSounds(m_host->buttonSounds());
      txb->setLabel(label);

      txb->msgAddSink("press", label, this);
      
      m_widgets.addWidget(txb);

      b1 += b3;
    }
    
    Menu * lights = new Menu(m_host);

    lights->setLabel("Lights");
    lights->setArea(b1, b2);
  
    m_widgets.addWidget(lights);

    // Test the existense of lighting files by opening them:
    for(int l=0; l < 20; l++) {
      sprintf(buf, "lights-%d", l);
      FILE * f = fopen(buf, "r");
      if(f) {
	sprintf(buf2, "Light set %d", l + 1);
	lights->addItem(buf2, buf, gc);
	fclose(f);
      }
      else
	break;
    }

    buildTabs();

    host->engine()->msgAddSink("exit", "save", this);

    saveTabs("save-first.xui");

    if(m_tabs.size()) {
      const char * tool = (*m_tabList.begin()).ptr()->name().c_str();
      selectTool(tool, RIGHT_HAND);
      selectTool(tool, LEFT_HAND);
    }
    VEE_trace("%s # EXIT", fname);
  }

  HelmaEditor::~HelmaEditor()
  {}

  bool HelmaEditor::msgProcess(const char *id, Fluid::Message *m)
  {
    VEE_trace("HelmaEditor::msgProcess # %s", id);

    bool ok = false;

    Engine * e = m_host->engine();

    if(strncmp(id, "tool-", 5) == 0) {
      MessageInt * mi = dynamic_cast<MessageInt *> (m);
      
      ok = selectTool(id + 5, mi ? mi->value() : RIGHT_HAND);
    }
    else if(strcmp(id, "save") == 0)
      ok = saveTabs("save-exit.xui");
    else if(strcmp(id, "undo") == 0) {
      e->geeEngine()->undo(false);
      e->setRebuildGee(true);
      ok = true;
    }
    else if(strcmp(id, "redo") == 0) {
      e->geeEngine()->redo(false);
      e->setRebuildGee(true);
      ok = true;
    }
    else if(strcmp(id, "clear") == 0) {
      e->geeEngine()->actions().undo(1000000);
      e->setRebuildGee(true);
      ok = true;
    }
    else if(strcmp(id, "bgcolor") == 0) {
      Fluid::MessageString *ms = dynamic_cast<Fluid::MessageString *> (m);

      if(ms) {
	float h = 0.5, s = 0.5, v = 0.5;
	int ns = sscanf(ms->string().c_str(), "%f %f %f", &h, &s, &v);
	if(ns != 3)
	  return false;

	VEE_Hsva hsva(h, s, v, 1.0, 1.0);
	VEE_Rgba rgba(hsva.rgba());

	e->geeEngine()->world().setBackgroundColor(rgba);
	ok = true;
      }
    }
    else if(strcmp(id, "clearselection") == 0) {
      e->geeEngine()->world().selectAll(0);
      ok = true;
    }
    else if(strcmp(id, "selectall") == 0) {
      e->geeEngine()->world().selectAll(GEE_SELECT_BIT);
      ok = true;
    }
    
    return ok;
  }

  void HelmaEditor::start()
  {
    if(m_currentTab)
      m_currentTab->activate();
  }

  void HelmaEditor::render(GlData *gld)
  {
    m_widgets.renderGl(gld);

    if(m_currentTab)
      m_currentTab->renderGl(gld);
    
    m_tabArea.renderGl(gld);
  }

  void HelmaEditor::update()
  {
    m_widgets.update();

    if(m_currentTab)
      m_currentTab->update();
    
    m_tabArea.update();
    
    if(m_currentTab) {

      VEE_Vector3 loc = m_host->worldLocation(VEE_Vector3(0, 1, -2.5));
      VEE_Matrix4 m;
      m.identity();
      m.setColumn3(3, loc);

      Demo * demo = m_currentTab->demo();
      demo->setGeometryTransformation(m);
      demo->render(&m_host->engine()->geeEngine()->collector());
    }
  }

  void HelmaEditor::beforeEvents()
  {
    if(m_currentTab)
      m_currentTab->beforeEvents();

    m_tabArea.beforeEvents();

    m_widgets.beforeEvents();
  }

  void HelmaEditor::handleEvent(const InEvent *e)
  {
    if(m_currentTab)
      m_currentTab->handleEvent(e);

    m_tabArea.handleEvent(e);

    m_widgets.handleEvent(e);
  }

  QDomElement HelmaEditor::writeDom(QDomDocument &doc)
  {
    QDomElement e = doc.createElement("HelmaEditor");

    return e;
  }

  bool HelmaEditor::readDom(QDomElement &)
  {
    return false;
  }

  bool HelmaEditor::selectTool(const char * toolName, int hand)
  {
    m_currentTabName = toolName;
    iterator it = m_tabs.find(m_currentTabName);

    if(it == m_tabs.end()) {
      VEE_error("HelmaEditor::selectTool # No tab %s", toolName);
      return false;
    }

    m_host->engine()->setCursorSize(1.0);

    m_currentTab = (*it).second.ptr();
    m_currentTab->activate();

    const std::string & rec = m_currentTab->tool();

    m_currentTools[hand] = rec;
    
    m_host->engine()->setNextRecorder(Engine::RECORDER_GEE, 
				      m_currentTools[0], m_currentTools[1]);

    // m_demo.setTool(m_currentTab->tool());

    VEE_trace("HelmaEditor::selectTool # Selected tool %s %s %d", 
	      toolName, rec.c_str(), hand);

    return true;
  }

  HelmaTab * HelmaEditor::tab(const std::string &tool)
  {
    for(iterator2 it = m_tabList.begin(); it != m_tabList.end(); it++) {

      HelmaTab * t = (*it).ptr();

      if(t->tool() == tool)
	return t;
    }

    return 0;
  }

  bool HelmaEditor::buildTabs()
  {
    const char * fname = "HelmaEditor::buildTabs";

    m_tabs.clear();
    m_tabList.clear();

    WidgetGroup * buttons = m_tabButtons.ptr();

    QDomNode n = m_uiSetup.documentElement().firstChild();

    uint tabcount = 0;

    while ( !n.isNull() ) {

      if(n.toElement().tagName() == "tab")
	tabcount++;

      n = n.nextSibling();
    }    

    buttons->resizeGrid((tabcount + 1) / 2, 2);
    
    // This XML-node specifies the UI package

    char buf[64];

    n = m_uiSetup.documentElement().firstChild();

    while ( !n.isNull() ) {
      if ( n.isElement() ) {
	QDomElement e = n.toElement();

	QString tag = e.tagName();
	
	VEE_trace("%s # got tag \"%s\"", fname, tag.ascii());

	if(tag == "tab") {

	  HelmaTab * ht;
	  
	  QString special = e.attribute("special");
	  
	  if(special == "Gallery") 
	    ht = new Gallery(m_host, this);
	  else
	    ht = new HelmaTab(m_host, this);

	  ht->readDom(e);
	  std::string name = ht->name();
	  VEE_RefPtr<HelmaTab> htr = ht;
	  m_tabs[name] = htr;
	  m_tabList.push_back(htr);
	  
	  VEE_trace("%s # added tab \"%s\"", fname, ht->name().c_str());

	  TextureButton * txb = new TextureButton(m_tabButtons.ptr());
	  txb->setLabel(name);
	  
	  sprintf(buf, "tool-%s", name.c_str());
	  txb->msgAddSink("press", buf, this);

	  WidgetGroup::Object o;
	  o.m_widget = txb;
	  m_tabButtons.ptr()->addObject(o);
	}
	else if(tag == "buttons") {
	  buttons->setLocation(getVector3(e, "lowerleft",buttons->location()));
	  buttons->setSize(getVector3(e, "size", buttons->size()));
	}
	else
	  VEE_error("%s # Unrecognized tag \"%s\"", fname, tag.ascii());

      }
      n = n.nextSibling();
    }

    buttons->calculateLayout();
    
    return true;    
  }

  bool HelmaEditor::saveTabs(const char * filename)
  {
    QDomDocument doc("HelmaUI");
 
    QDomElement base = doc.createElement("ui");

    WidgetGroup * buttons = m_tabButtons.ptr();

    QDomElement be = doc.createElement("buttons");

    addTextChild(be, doc, "lowerleft", buttons->location());
    addTextChild(be, doc, "size", buttons->size());

    base.appendChild(be);
   
    for(iterator2 it = m_tabList.begin(); it != m_tabList.end(); it++) {
      QDomElement e = (*it).ptr()->writeDom(doc);
      base.appendChild(e);
    }

    doc.appendChild(base);

    QCString s = doc.toCString();

    QFile out(filename);
    bool ok = out.open(IO_WriteOnly);
    QTextStream(&out) << s;
    
    return ok;
  }

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  
  HelmaQuickEdit::HelmaQuickEdit(Tools * host)
    : m_host(host),
      m_widgets(0)
  {}
  
  HelmaQuickEdit::~HelmaQuickEdit()
  {}

  void HelmaQuickEdit::start()
  {
    m_host->setRenderingTransforms(false);
  }

  void HelmaQuickEdit::stop()
  {
    if(!m_widgets)
      return;

    m_widgets->setLocation(m_origLoc);
    m_widgets->setRotation(m_origRot);
    m_widgets->setHostTransformation(m_origTransform);

    m_widgets = 0;
  }

  void HelmaQuickEdit::render(GlData *gld)
  {
    if(m_widgets)
      m_widgets->renderGl(gld);
  }

  void HelmaQuickEdit::update()
  {
    if(m_widgets)
      m_widgets->update();
  }

  void HelmaQuickEdit::beforeEvents()
  {
    if(m_widgets)
      m_widgets->beforeEvents();
  }

  void HelmaQuickEdit::handleEvent(const InEvent *e)
  {
    // VEE_trace("HelmaQuickEdit::handleEvent # %p %p", m_widgets, e);

    if(!m_widgets)
      return;

    InEvent e2(*e);

    /* The event is in tools' coordinate system, thus revert to the
       world coordinates: */
    Vector3 loc = m_host->unapplyMyTransform(e->location());
    
    Vector3 loc2 = m_widgets->applyMyTransform(loc);
    VEE_BoundingBox bb;
    m_widgets->calculateBoundingBox(&bb);
    
    if(bb.axisDistance(loc2) > 0.7f && !m_host->grabs()->isGrabbed()) {
      VEE_trace("HelmaQuickEdit::handleEvent # stopping");
      stop();
      m_host->engine()->skipMode(Engine::MODE_PLAY);
      return;
    }

    e2.setLocation(loc);
    e2.setRotation(m_host->unapplyMyTransform(e->rotation()));

    m_widgets->handleEvent(&e2);
  }


  QDomElement HelmaQuickEdit::writeDom(QDomDocument &)
  {
    return QDomElement();
  }

  bool HelmaQuickEdit::readDom(QDomElement &)
  {
    return false;
  }

  void HelmaQuickEdit::init(HelmaEditor * editor,
			    int hand,
			    const std::string & tool,
			    const std::string & group)
  {
    const char * fname = "HelmaQuickEdit::init";

    VEE_trace("%s # %d %s %s", fname, hand, tool.c_str(), group.c_str());

    stop();
    
    HelmaTab * tab = editor->tab(tool);

    if(!tab) {
      VEE_error("%s # No tab with tool %s", fname, tool.c_str());
      return;
    }

    WidgetGroup * wg = tab->widgetGroup(group);

    if(!wg) {
      VEE_error("%s # No widget group %s", fname, group.c_str());
      return;
    }
    
    m_widgets = wg;
    m_origTransform = wg->hostTransformation();
    m_origLoc = wg->location();
    m_origRot = wg->rotation();

    TrackerSensorSelector2 * sensor = m_host->engine()->handLoc(hand);

    Vector3 loc = sensor->locations().getNewest();

    VEE_BoundingBox bb;
    m_widgets->calculateBoundingBox(&bb);
    Vector3 center = bb.center();
    loc -= center;
    loc.y -= center.y * 0.5f;

    wg->setLocation(loc);
    wg->setRotation(editor->host()->engine()->rotation());
    wg->setHostTransformation(0);
  }

}

template class MJ_Array2D<bool>;
