/* -*- C++ -*- */

/* -*- C++ -*- */

/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Crystal animation and 3D modeling application
 *
 * Authors:   Tommi Ilmonen 2001-2004
 * Copyright: Tommi Ilmonen and/or Tekes 2001-2004
 *
 * firstname.lastname@hut.fi
 *
 */


#include <crystal_mesh_recorder.h>

#include <crystal_action_list.h>
#include <crystal_density_editor.h>
#include <crystal_dom_util.h>
#include <crystal_gl_data.h>
#include <crystal_engine.h>
#include <crystal_model.h>
#include <crystal_tools.h>

#include <Fluid/TrackerSensorSelector.hpp>

#include <vee_error.h>

namespace Crystal { 
  
  CreateMeshAction::~CreateMeshAction()
  {}
  
  void CreateMeshAction::act(ActionList *al)
  {
    Model * m = 
      al->engine()->tools()->densityEditor()->createModel(m_meshType);

    m->object().setLocation(m_location);
    m->object().setRotation(m_rotation);

    al->setModel(m_id, m);
    al->engine()->addAniMesh(& m->object());

    VEE_trace("CreateMeshAction::act # %p ", m);
  }

  QDomElement CreateMeshAction::writeDom(QDomDocument &doc)
  {
    QDomElement elem = doc.createElement("CreateMeshAction");

    writeDomTimeStamp(elem, doc);

    addTextChild(elem, doc, "location", m_location);
    addTextChild(elem, doc, "rotation", m_rotation);
    addTextChild(elem, doc, "id", m_id);
    addTextChild(elem, doc, "mesh", m_meshType);

    return elem;
  }

  bool CreateMeshAction::readDom(QDomElement &elem)
  {
    if(elem.tagName() != "CreateMeshAction")
      return false;

    readDomTimeStamp(elem);

    m_location = getVector3(elem, "location");
    m_rotation = getMatrix3(elem, "rotation");

    m_id       = getInt(elem, "id", -1);
    m_meshType = getInt(elem, "mesh", -1);

    return true;
  }

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////


  DeleteMeshAction::~DeleteMeshAction()
  {}
  
  void DeleteMeshAction::act(ActionList *al)
  {
    Model * m = al->getModel(m_id);
    if(m) {
      al->engine()->removeAniMesh(& m->object());
      delete m;
      al->setModel(m_id, 0);
    }

    VEE_trace("DeleteMeshAction::act# %p", m);
  }

  QDomElement DeleteMeshAction::writeDom(QDomDocument &doc)
  {
    QDomElement elem = doc.createElement("DeleteMeshAction");

    writeDomTimeStamp(elem, doc);

    addTextChild(elem, doc, "id", m_id);

    return elem;
  }

  bool DeleteMeshAction::readDom(QDomElement &elem)
  {
    if(elem.tagName() != "DeleteMeshAction")
      return false;

    readDomTimeStamp(elem);

    m_id = getInt(elem, "id", -1);

    return true;
  }

  
  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////


  MeshRecorder::MeshRecorder()
    : m_meshId(-1),
      m_trackerCount(-1)
  {
    m_location.clear();
    m_rotation.identity();
  }

  MeshRecorder::~MeshRecorder()
  {}

  
  void MeshRecorder::prepare(ActionList *) {}
  
  void MeshRecorder::beginRecording(ActionList *al, 
				    const TimeStamp & now)
  {
    const char * fname = "MeshRecorder::beginRecording";

    VEE_trace("%s", fname);

    Engine *engine = al->engine();
    
    IONodeBuffer<Vector3> * handLoc = 
      & engine->weakHandLoc()->locations();

    IONodeBuffer<Matrix3> * handRot = 
      & engine->weakHandLoc()->rotations();
    
    m_trackerCount = (long) handLoc->getSampleCount();

    Vector3 location = handLoc->getIndex(m_trackerCount);
    Matrix3 rotation = handRot->getIndex(m_trackerCount);
  
    m_meshId = al->allocateMeshId();

    int meshType = engine->tools()->densityEditor()->currentMesh();
    
    al->appendAction
      (new CreateMeshAction(m_meshId, meshType, location, rotation, now));

    m_mode = RECORD;

    VEE_trace("%s # %d %d", fname, m_meshId, meshType);
  }
  
  void MeshRecorder::endRecording(ActionList *al, 
				  const TimeStamp &now)
  {
    VEE_trace("MeshRecorder::endRecording");

    al->appendAction(new DeleteMeshAction(m_meshId, now));
    
    m_meshId = -1;
    m_mode = MUTE;
  }

  static std::string locStr("location");
  static std::string rotStr("rotation");

  void MeshRecorder::record(ActionList *al, const TimeStamp &)
  {
    VEE_trace("MeshRecorder::record");

    Engine *engine = al->engine();

    Fluid::IONodeBuffer<Fluid::Vector3> * handLoc = 
      & engine->weakHandLoc()->locations();

    IONodeBuffer<Matrix3> * handRot = 
      & engine->weakHandLoc()->rotations();
    
    ValueGraphs * graphs = & al->getGraph(m_meshId);

    graphs->setTargetType(ValueGraphs::MODEL);
  
    for(;m_trackerCount < (long) handLoc->getSampleCount(); m_trackerCount++) {
    
      m_location = handLoc->getIndex(m_trackerCount);
      m_rotation = handRot->getIndex(m_trackerCount);

      if(m_mode == RECORD) {

	Fluid::TimeStamp sampleTime = handLoc->getTimeStamp(m_trackerCount);
	Fluid::TimeStamp ts = al->realTimeToRecordTime(sampleTime);
	double dts = ts.secondsD();
	
	graphs->appendVector3(locStr, dts, m_location);
	graphs->appendMatrix3(rotStr, dts, m_rotation);
      }
    }

  }

  static std::string _green("green");

  void MeshRecorder::render(Engine *engine, GlData *gld)
  {
    if(m_mode == RECORD)
      return;
        
    gld->m_materials.forceMaterial(_green);

    glPushAttrib(~0);

    gld->m_solarRenderer.prepareToRender();
    gld->m_solarRenderer.pushTransform(m_rotation, m_location, 1.0);
    engine->tools()->densityEditor()->currentElement()->m_model.render
      (& gld->m_solarRenderer);
    gld->m_solarRenderer.popTransform();
    
    glPopAttrib();

    gld->m_materials.unforceMaterial();

    // VEE_trace("MeshRecorder::render # DONE");
  }

}
