/* -*- C++ -*- */

/* -*- C++ -*- */

/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Crystal animation and 3D modeling application
 *
 * Authors:   Tommi Ilmonen 2001-2004
 * Copyright: Tommi Ilmonen and/or Tekes 2001-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <crystal_particle_editor.h>

#include <crystal_dom_util.h>
#include <crystal_gl_data.h>
#include <crystal_graph_editor.h>
#include <crystal_tools.h>

#include <vee_error.h>
#include <vee_gl_renderer.h>
#include <vee_math.h>

#include <Fluid/MessageFloat.hpp>
#include <Fluid/MessageString.hpp>

namespace Crystal { 

  static const char * parammap[ParticleEditor::SLIDERS] = {
    "agelimit",
    "mass",
    "area"
  };

  void ParticleEditor::ParticleData::init()
  {
    int colors = 5;

    m_colorsHsva.resize(colors);
  
    m_colorsHsva[0] = VEE_Hsva::Yellow;
    m_colorsHsva[1] = VEE_Hsva::Red;
    m_colorsHsva[2] = VEE_Hsva::Blue;
    m_colorsHsva[3] = VEE_Hsva::White;
    m_colorsHsva[4] = VEE_Hsva::Yellow;
    m_colorsHsva[4].setValue(0.3);
    m_colorsHsva[4].setAlpha(0.3);

    const float alphas[5] = { 0.5, 1.0, 1.0, 1.0, 0.1 };
    const float thetas[5] = { 0.5, 0.5, 0.0, 0.0, 0.0 };
    const float sizes[5]  = { 0.2, 0.2, 0.2, 0.2, 0.1 };


    m_graphs["alpha"].copy(alphas, 5);
    m_graphs["additive"].copy(thetas, 5);
    m_graphs["size"].copy(sizes, 5);
    
    m_graphs["alpha"].setLimits(0.0, 1.0, true, true);
    m_graphs["additive"].setLimits(0.0, 1.0, true, true);
    m_graphs["size"].setLimits(0.1, 1.0, true, true);

    transmitColors();

    (*m_particle).setLooks(m_colors.clone());
    (*m_particle).setAgeLimit(5.0);
    (*m_particle).setSize(0.3);
    (*m_particle).setArea(0.7);
    (*m_particle).setMass(0.7);  
    (*m_particle).setTexture(42);  

    m_sizeCurve.setId("size");
    m_sizeCurve.setScaling(VEE_TS_AGE_MATCH);

    (*m_particle).appendSelfOperator(m_sizeCurve.clone());
  }

  void ParticleEditor::ParticleData::transmitColors()
  {
    VEE_RgbaTable *ctable = & m_colors.colorTable();

    ctable->clear();

    for(uint i = 0; i < m_colorsHsva.size(); i++) {
      m_colorsHsva[i].setAlpha(m_graphs["alpha"].m_graph.getPoint(i).m_value);
      m_colorsHsva[i].setTheta(m_graphs["additive"].m_graph.getPoint(i).m_value);
      ctable->appendColor(m_colorsHsva[i].rgba());
    }  
  }

  void ParticleEditor::ParticleData::update()
  {
    transmitColors();

    m_sizeCurve.clearCurve();

    VEE_LinearInterpolation<float> * sizes = & m_graphs["size"].m_graph;

    for(uint i=0; i < sizes->size(); i++) {
      float v = sizes->getPoint(i).m_value;
      v *= v;
      v *= 5.0; // Make it non-linear
      m_sizeCurve.appendPoint(v, (float) i);
    }
  }

  QDomElement ParticleEditor::ParticleData::writeDom(QDomDocument &doc)
  {
    QDomElement elem = doc.createElement("ParticleData");

    QDomElement colors = doc.createElement("colors");

    uint i;

    for(i=0; i < m_colorsHsva.size(); i++) {
      addTextChild(colors, doc, "color", toStr(m_colorsHsva[i]));
    }

    elem.appendChild(colors);

    addTextChild(elem, doc, "agelimit", (*m_particle).ageLimit());
    addTextChild(elem, doc, "size", (*m_particle).size());
    addTextChild(elem, doc, "mass", (*m_particle).mass());
    addTextChild(elem, doc, "area", (*m_particle).area());
    addTextChild(elem, doc, "texture", (*m_particle).texture());

    return elem;
  }

  bool ParticleEditor::ParticleData::readDom(QDomElement &elem)
  {
    VEE_trace("ParticleEditor::ParticleData::readDom");

    if(elem.tagName() != "ParticleData")
      return false;

    QDomNode colors = elem.namedItem("colors");

    if(colors.isElement()) {
      QDomElement ce = colors.toElement();

      QDomNode n = ce.firstChild();
    
      m_colorsHsva.resize(0);

      while(!n.isNull()) {
      
	if ( n.isElement() ) {
	  QDomElement e = n.toElement();
	  if(e.tagName() != "color")
	    VEE_error("ParticleEditor::ParticleData::readDom");
	  else {
	    m_colorsHsva.push_back(toHsva(e.firstChild().toText().data()));
	  }
	}

	n = n.nextSibling();
      }
    
    }

    VEE_Particle & p = (*m_particle);
  
    p.setAgeLimit(getFloat(elem, "agelimit", 1.0f));
    p.setSize(getFloat(elem, "size", 1.0f));
    p.setMass(getFloat(elem, "mass", 1.0f));
    p.setTexture(getInt(elem, "texture", 0));

    transmitColors();

    VEE_trace("ParticleEditor::ParticleData::readDom # EXIT ok");

    return true;
  }

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  ParticleEditor::ParticleEditor(Tools *tools)
    : Transformation(tools),
      m_host(tools),
      m_widgets(tools->grabs()),
      m_selectedTexture(3),
      m_selectedParticle(0),
      m_colorPoint(0),
      m_currentParticle(0)
  {
    uint i, j;
    Vector2 v1, v2, v3;
    char buf[64];

    VEE_BoundingBox colorCubeBox;

    colorCubeBox.set(-2.0, 0.5, 0.0, 
		     0.0, 2.0, 2.0);

    // Texture selection:

    m_colorBox.set(-2.0, 2.5, 0.0, 
		   0.0, 3.0, 0.0);

    m_texBox.set(0.5, 0.5, 0.0, 
		 2.5, 2.5, 0.0);

    m_particleBox.set(0.0, 0.5, 0.0,
		      2.0, 3.5, 0.0);

    Transformation particleTransform;
    particleTransform.setLocation(Vector3(2.7, 0.0, 0.5));
    Matrix3 rot;
    rot.rotateY(-0.7 * VEE_HALF_PI);
    particleTransform.setRotation(rot);
    

    Fluid::Vector3 w1, w2; // Slider locations
    w1.make(0.3, 0.4, 1.8);
    w2.make(0.0, 0.0, -0.6);

    uint division = 8;

    Fluid::Vector3 texButtonSize = m_texBox.high() - m_texBox.low();
    texButtonSize /= (float) division;
  
    v2.make(texButtonSize[0], texButtonSize[1]); 
    v3.make(m_texBox.low()[0], m_texBox.low()[1]);
  
    m_textureButtons.resize(division, division);
    m_textureButtons.setAll(0);

    for(i=0; i < division; i++) {
      for(j=0; j < division; j++) {

	v1.make(v2[0] * j, v2[1] * (division - i - 1));
	v1 += v3;

	TextureButton * txb = new TextureButton(m_host);
      
	txb->setArea(v1, v2);
	txb->setSounds(m_host->buttonSounds());
	txb->setFontSize(0.13);

	// sprintf(buf, "%u", i * division +  j + 1);
	// txb->setLabel(buf);
	txb->setTexture("");
      
	m_widgets.addWidget(txb);
	sprintf(buf, "texture_%u_%u", i, j);

	txb->msgAddSink("press", buf, this);
	m_textureButtons.get(i, j) = txb;

	txb->setBorder(0.01);
      }
    }

    // Color cube

    ColorCube * cc = new ColorCube(m_host);
    cc->msgAddSink("color", "color", this);
    m_widgets.addWidget(cc);
    m_colorCube = cc;
    cc->setBoundingBox(colorCubeBox);

    // Available particles:

    uint particleRows = 5;
    uint particleCols = 3;
    // uint totalParticles = particleRows * particleCols;

    m_particles.resize(particleRows, particleCols);

    m_currentParticle = m_particles.begin() + m_selectedParticle;

    createParticleSelectors(& m_widgets, particleTransform, m_host, m_host,
			    this, & m_particleBox, &m_particles);
 
    uint totalTextures = division * division;

    for(i=0; i < m_particles.elements(); i++) {
      (*( (*(m_particles.begin() + i)).m_particle)).
	setTexture(i % totalTextures);
    }

    selectTexture((*m_currentParticle->m_particle).texture());

    // The horizontal panel that has the color control points:
  
    float dist = (m_colorBox.high()[0] - m_colorBox.low()[0]) / 
      m_currentParticle->m_colorsHsva.size();
  
    v2.make(dist);

    for(i=0; i < m_currentParticle->m_colorsHsva.size(); i++) {

      v1[0] = m_colorBox.low()[0];
      v1[1] = m_colorBox.low()[1];
    
      v1[0] += i * dist;
      v1[1] -= dist;

      TextureButton * tb = new TextureButton(m_host);
      tb->setArea(v1, v2);
      tb->setSounds(m_host->buttonSounds());
      tb->setTexture("");

      sprintf(buf, "color_point_%u", i);

      tb->msgAddSink("press", buf, this);
      m_widgets.addWidget(tb);
      m_colorPointButtons.push_back(tb);
    }

    m_colorPointButtons[m_colorPoint]->setColors(m_host->selectedColor());

    // Misc sliders:
  
    Fluid::Matrix3 m1, m2;
    m1.rotateX(-0.8 * VEE_HALF_PI);
    m2.rotateY(VEE_HALF_PI);

    // m1 = m2 * m1;

    float parammin[SLIDERS] = {
      0.05,
      -1.0,
      0.0
    };

    float parammax[SLIDERS] = {
      20.0,
      1.0,
      1.0
    };

    MJ_Mapper * mappers[SLIDERS] = {
      new MJ_MapperLog,
      0,
      0
    };

    for(i=0; i < SLIDERS; i++) {
    
      float v = (*m_currentParticle->m_particle).getVariableFloat(parammap[i]);

      Slider *slider = new Slider(m_host, 
				  v, parammin[i], parammax[i]);

      slider->setSounds(m_host->buttonSounds());
      m_widgets.addWidget(slider);
    
      slider->msgAddSink("value", parammap[i], this);
  
      slider->setLocation(w1);
      slider->setRotation(m1);
      slider->setSize(2.0, 0.3);
      slider->setLabel(parammap[i]);
      slider->setOrientation(Slider::HORIZONTAL);
      slider->setDisplayValue(false);
      slider->setMapper(mappers[i]);

      m_sliders[i] = slider;

      w1 += w2;

    }

    // The graphs about size etc:
    m_graphs = new GraphEditor();
    for(ParticleData::giterator it = m_currentParticle->m_graphs.begin();
	it != m_currentParticle->m_graphs.end(); it++) {
      m_graphs->addGraph((*it).first.c_str(), & (*it).second);
    }
    v1.make(2.0, 1.0);
    v2.make(2.0, 1.0);
    m_graphs->setArea(v1, v2);
    m_graphs->setRotation(m2);
    m_graphs->setLocation(Vector3(-2.0, 1.2, 4.1));
    m_graphs->setTimePerUnit(2.0);

    m_widgets.addWidget(m_graphs);

    msgProcess("color_point_0", 0);
  }

  ParticleEditor::~ParticleEditor()
  {}

  bool ParticleEditor::msgProcess(const char *id, Fluid::Message *m)
  {
    const char * fname = "ParticleEditor::msgProcess";

    VEE_trace("%s # %s %p", fname, id, m);

    bool ok = false;

    if(strncmp(id, "texture_", 8) == 0) {
      uint row = 0, col = 0;
      id += 8;
      sscanf(id, "%u_%u", & row, & col);
      uint texture = row * m_textureButtons.cols() + col;
      selectTexture(texture);

      VEE_trace("%s # Selected texture %u", fname, texture);
      ok = true;
    }
    else if(strncmp(id, "color_point_", 12) == 0) {
      int index = 0;
      id += 12;
      sscanf(id, "%d", & index);
    
      ColorGroup cg;

      m_colorPointButtons[m_colorPoint]->setColors(cg);

      m_colorPoint = index;

      m_colorPointButtons[m_colorPoint]->setColors(m_host->selectedColor());

      updateSliders();

      VEE_trace("%s # Selected color point %d", fname, index);
      m_colorCube->setColor(m_currentParticle->m_colorsHsva[m_colorPoint]);
      ok = true;
    }
    else if(strncmp(id, "particle_", 9) == 0) {
      int index = 0;
      id += 9;
      sscanf(id, "%d", & index);

      selectParticle(index);
    
      VEE_trace("%s # Selected particle %d", fname, index);
      ok = true;
    }
    else if(strcmp(id, "color") == 0) {
      float h = 0.5, s = 0.5, v = 0.5;
      Fluid::MessageString *ms = dynamic_cast<Fluid::MessageString *> (m);

      if(ms) {
	int ns = sscanf(ms->string().c_str(), "%f %f %f", &h, &s, &v);
	if(ns != 3)
	  return false;

	VEE_trace("%s # Got color %f %f %f", fname, h, s, v);
      
	VEE_Hsva & hsva = m_currentParticle->m_colorsHsva[m_colorPoint];
	hsva.set(h, s, v, hsva.alpha());

	m_currentParticle->transmitColors();
	ok = true;
      }

      // VEE_trace("%s # Selected color point %d", fname, index);
    }
    else if(strcmp(id, "alpha") == 0) {
      Fluid::MessageFloat * mf = dynamic_cast<Fluid::MessageFloat *> (m);

      if(mf) {
	float a = mf->value();
	m_currentParticle->m_colorsHsva[m_colorPoint].setAlpha(a);
	m_currentParticle->transmitColors();
	ok = true;
	m_colorCube->setAlpha(a);
      }
    }
    else if(strcmp(id, "additive") == 0) {
      Fluid::MessageFloat * mf = dynamic_cast<Fluid::MessageFloat *> (m);

      if(mf) {
	float add = mf->value();
	m_currentParticle->m_colorsHsva[m_colorPoint].setTheta(add);
	m_currentParticle->transmitColors();
	ok = true;
	m_colorCube->setTheta(add);
      }
    }
    else if(strcmp(id, "size") == 0 ||
	    strcmp(id, "agelimit") == 0 ||
	    strcmp(id, "mass") == 0 ||
	    strcmp(id, "area") == 0) {
      Fluid::MessageFloat * mf = dynamic_cast<Fluid::MessageFloat *> (m);

      if(mf) {
	float a = mf->value();
	(*m_currentParticle->m_particle).setValueFloat(id, a);
	ok = true;
      }
    }

    if(!ok) {
      VEE_error("%s # Unhandled message \"%s\"", fname, id);
    }

    return ok;
  }

  void ParticleEditor::render(GlData *glData)
  {
    // Start with the background:
  
    uint i, subTextureDivision = glData->m_veeRenderer.subTextureDivision();

    // Render a rectangle that holds the texture:
  
    VEE_Vector3 v1, v2;

    // glDepthMask(GL_TRUE);

    // Render the available particles:

    glDepthMask(GL_FALSE);
  
    renderParticleSelectors(&m_particles, glData);

    // Render the color examples to show what the user is going to get:

    glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);

    glBegin(GL_QUADS);

    float dist = (m_colorBox.high()[0] - m_colorBox.low()[0]) / 
      m_currentParticle->m_colorsHsva.size();

    float texStep = 1.0f / subTextureDivision;

    int ptex = (*m_currentParticle->m_particle).texture();

    int trow = ptex / subTextureDivision;
    int tcol = ptex % subTextureDivision;

    Fluid::Vector2 texSize;
    texSize.make(1);
    texSize /= subTextureDivision;

    Fluid::Vector2 texLow(texSize[0] * tcol, texSize[1] * trow);

    for(i=0; i < m_currentParticle->m_colorsHsva.size();i++) {

      v1 = m_colorBox.low();
    
      v1[0] += i * dist;
      v1[1] -= dist;
      v2.make(v1.x + dist, v1.y + dist, 0.0);
    
      VEE_Rgba color = glData->m_veeRenderer.transformColor
	(m_currentParticle->m_colors.colorTable().getColor(i));

      glColor4fv(color.data());

      glTexCoord2f(texLow.x, texLow.y);
      glVertex3f(v1[0], v1[1], 0.0);

      glTexCoord2f(texLow.x + texStep, texLow.y);
      glVertex3f(v2[0], v1[1], 0.0);

      glTexCoord2f(texLow.x + texStep, texLow.y + texStep);
      glVertex3f(v2[0], v2[1], 0.0);

      glTexCoord2f(texLow.x, texLow.y + texStep);
      glVertex3f(v1[0], v2[1], 0.0);
    }

    glEnd();

    // Render the color gradient to show how the colors will change:
  
    glDisable(GL_TEXTURE_2D);
  
    glBegin(GL_QUAD_STRIP);
  
    for(i=0; i < m_currentParticle->m_colorsHsva.size();i++) {
    
      v1 = m_colorBox.low();
    
      v1[0] += i * dist + dist * 0.5;
      v2.make(v1.x + dist, v1.y + dist, 0.0);
    
      VEE_Rgba color = glData->m_veeRenderer.transformColor
	(m_currentParticle->m_colors.colorTable().getColor(i));

      glColor4fv(color.data());

      glVertex3f(v1[0], v1[1], 0.0);
      glVertex3f(v1[0], v2[1], 0.0);
    }

    glEnd();

    glDepthMask(GL_TRUE);
    glDisable(GL_BLEND);

    m_widgets.renderGl(glData);

    glEnable(GL_BLEND);

    // Render the particle textures:
    
    // The background panel:
    glDisable(GL_TEXTURE_2D);

    glColor4f(0.0, 0.0, 0.0, 0.7);
    glBegin(GL_QUADS);

    // Lower left
    v1 = m_texBox.low();
    v1[2] -= 0.02;
    glVertex3fv(v1.data());

    // Lower right
    v1[0] = m_texBox.high()[0];
    v1[2] -= 0.02;
    glVertex3fv(v1.data());

    // Upper Right
    v1 = m_texBox.high();
    v1[2] -= 0.02;
    glVertex3fv(v1.data());

    // Upper Left
    v1[0] = m_texBox.low()[0];
    v1[2] -= 0.02;
    glVertex3fv(v1.data());
    
    glEnd();

    // The texture quad:
    glBlendFunc(GL_SRC_ALPHA, GL_ONE);

    GLuint tex = glData->m_veeRenderer.getTexture();

    if(tex == (GLuint) -1) {
      VEE_error("ParticleEditor::render # No texture");
      return;
    }
  
    glBindTexture(GL_TEXTURE_2D, tex);
    glEnable(GL_TEXTURE_2D);

    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, VEE_Rgba::White.data());
    glColor4fv(VEE_Rgba::White.data());
  
    glBegin(GL_QUADS);

    v1 = m_texBox.low();
    v1[2] -= 0.01;
    glTexCoord2f(0.0, 1.0);
    glVertex3fv(v1.data());

    v1[0] = m_texBox.high()[0];
    v1[2] -= 0.01;
    glTexCoord2f(1.0, 1.0);
    glVertex3fv(v1.data());

    v1 = m_texBox.high();
    v1[2] -= 0.01;
    glTexCoord2f(1.0, 0.0);
    glVertex3fv(v1.data());

    v1[0] = m_texBox.low()[0];
    v1[2] -= 0.01;
    glTexCoord2f(0.0, 0.0);
    glVertex3fv(v1.data());
  
    glEnd();

    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    glDisable(GL_TEXTURE_2D);
    glColor4fv(VEE_Rgba::White.data());
  }


  void ParticleEditor::update()
  {
    m_widgets.update();
    if(m_graphs->state() == Widget::GRAB)
      m_currentParticle->update();
  }

  void ParticleEditor::selectTexture(int t)
  {
    uint col = m_selectedTexture % m_textureButtons.cols();
    uint row = m_selectedTexture / m_textureButtons.cols();

    ColorGroup cg;

    m_textureButtons.get(row, col)->setColors(cg);

    m_selectedTexture = t;

    col = m_selectedTexture % m_textureButtons.cols();
    row = m_selectedTexture / m_textureButtons.cols();
  
    m_textureButtons.get(row, col)->setColors(m_host->selectedColor());

    (*m_currentParticle->m_particle).setTexture(t);
  }

  void ParticleEditor::selectParticle(int index)
  {
    ColorGroup cg;

    m_currentParticle->m_button->setColors(cg);

    m_selectedParticle = index;
    m_currentParticle  = m_particles.begin() + index;

    m_currentParticle->m_button->setColors(m_host->selectedColor());

    selectTexture((*m_currentParticle->m_particle).texture());
    updateSliders();

    m_colorCube->setColor(m_currentParticle->m_colorsHsva[m_colorPoint]);
  }

  VEE_Object * ParticleEditor::createParticle(uint i)
  {
    uint row = i / m_particles.cols();
    uint col = i % m_particles.cols();
    if(row >= m_particles.rows() || col >= m_particles.cols()) {
      VEE_fatal("ParticleEditor::createParticle # %u exceeds range", i);
    }
  
    m_particles.get(row, col).update();

    return (*m_particles.get(row, col).m_particle).clone();
  }

  void ParticleEditor::createParticleSelectors
  (Widgets *widgets,
   const Transformation &selftrans,
   Transformation *transHost,
   Tools *tools,
   Fluid::MessageSink *host,
   VEE_BoundingBox *bb,
   MJ_Array2D<ParticleData> *particles)
  {
    uint i, j;
    Fluid::Vector2 v1, v2, v3;
    char buf[64];
    Fluid::Vector3 texButtonSize;

    VEE_Vector3 pbspan = bb->span();

    pbspan[0] /= particles->cols();
    pbspan[1] /= particles->rows();

    float s = VEE_min(pbspan[0], pbspan[1]);
    texButtonSize.make(s);

    v2.make(texButtonSize[0], texButtonSize[1]); 
    v3.make(bb->low()[0], bb->low()[1]);
  
    for(i=0; i < particles->rows(); i++) {
      for(j=0; j < particles->cols(); j++) {
	uint index = i * particles->cols() + j;

	v1.make(v2[0] * j, v2[1] * i);
	v1 += v3;

	TextureButton * txb = new TextureButton(transHost);

	txb->setSounds(tools->buttonSounds());
	txb->setTransformation(selftrans);

	txb->setArea(v1, v2);
	sprintf(buf, "%u", i * particles->cols() +  j + 1);
	txb->setLabel(buf);
      
	sprintf(buf, "particle_%u", index);
      
	txb->msgAddSink("press", buf, host);

	widgets->addWidget(txb);
      
	particles->get(i,j).m_button = txb;
	particles->get(i,j).init();
      
      }
    }
  }

  void ParticleEditor::renderParticleSelectors
  (MJ_Array2D<ParticleData> *particles, GlData *glData)
  {  
    uint particleRows = particles->rows();
    uint particleCols = particles->cols();
  
    VEE_GlRenderer * veeGl = & glData->m_veeRenderer;

    glBindTexture(GL_TEXTURE_2D, veeGl->getTexture());
    glEnable(GL_TEXTURE_2D);

    glPushMatrix();

    const Transformation * trans = particles->get(0, 0).m_button;

    GlData::applyGlTransform(trans->rotation(), trans->location());

    glBegin(GL_QUADS);

    int division = veeGl->subTextureDivision();
  
    Fluid::Vector2 texSize;
    texSize.make(1);
    texSize /= division;

    for(uint i=0; i < particleRows; i++) {
      for(uint j=0; j < particleCols; j++) {
      
	ParticleData * pd = & particles->get(i, j);

	TextureButton * txb = pd->m_button;

	Fluid::Vector2 high = txb->boundingBox().high().vector2();
	Fluid::Vector2 low  = txb->boundingBox().low().vector2();

	high -= txb->border();
	low  += txb->border();

	int tex = (*pd->m_particle).texture();
	int trow = tex / division;
	int tcol = tex % division;
      
	Fluid::Vector2 texLow(texSize[0] * tcol, texSize[1] * trow);
	Fluid::Vector2 texHigh(texLow + texSize);
	Fluid::Vector2 texCrop = (texHigh - texLow) * 0.2;
	texLow  += texCrop;
	texHigh -= texCrop;

	const VEE_RgbaTable * colors = & pd->m_colors.colorTable();


	Fluid::Vector2 tsize = high - low;
	tsize /= (float) colors->nofColors();

	// Fluid::Vector2 tsize((high[0] + low[0]) * 0.5, low[1]);
	float minDim = VEE_min(tsize[0], tsize[1]);
	tsize.make(minDim);

	for(uint k=0; k < colors->nofColors(); k++) {
	
	  const VEE_Rgba & color = colors->getColor(k);
      
	  glColor4fv(color.data());

	  Fluid::Vector2 high2 = high;
	  high2[1] -= k * tsize[1];
	  Fluid::Vector2 low2 = high2 - tsize;
	  float offset = 0.01;

	  glTexCoord2f(texLow.x, texLow.y);
	  glVertex3f(low2.x, low2.y, offset);
	
	  glTexCoord2f(texHigh.x, texLow.y);
	  glVertex3f(high2.x, low2.y, offset);
	
	  glTexCoord2f(texHigh.x, texHigh.y);
	  glVertex3f(high2.x, high2.y, offset);
	
	  glTexCoord2f(texLow.x, texHigh.y);
	  glVertex3f(low2.x, high2.y, offset);
	}
      }
    }
  
    glEnd();

    glPopMatrix();
  }

  QDomElement ParticleEditor::writeDom(QDomDocument &doc)
  {
    QDomElement elem = doc.createElement("Particles");

    char buf[64];

    for(uint i=0; i < m_particles.elements(); i++) {
      sprintf(buf, "particle_%u", i);

      QDomElement tmp = doc.createElement(buf);
    
      tmp.appendChild((m_particles.begin() + i)->writeDom(doc));
      elem.appendChild(tmp);
    }

    return elem;
  }

  bool ParticleEditor::readDom(QDomElement &elem)
  {
    if(elem.tagName() != "Particles") {
      VEE_error("ParticleEditor::readDom # bad tag (%s)", 
		elem.tagName().ascii());
      return false;
    }
  
    QDomNode n = elem.firstChild();
    while ( !n.isNull() ) {
      if ( n.isElement() ) {
	QDomElement e = n.toElement();
	const char * tag = e.tagName().ascii();

	if(strncmp(tag, "particle_", 9) == 0) {
	  int index = atoi(tag + 9);
	  if(index < (int) m_particles.elements()) {
	    QDomElement e2 = e.firstChild().toElement();
	  
	    (m_particles.begin() + index)->readDom(e2);
	  }
	}
      }
      n = n.nextSibling();
    }
  
    selectParticle(0);

    return true;
  }

  void ParticleEditor::updateSliders()
  {
    for(uint i=0; i < SLIDERS; i++) {
    
      float v = (*m_currentParticle->m_particle).getVariableFloat(parammap[i]);
	
      m_sliders[i]->setValue(v);

      VEE_trace("ParticleEditor::updateSliders # %s is %f",
		parammap[i], v);
    }

    m_graphs->clear();
    
    for(ParticleData::giterator it = m_currentParticle->m_graphs.begin();
	it != m_currentParticle->m_graphs.end(); it++) {
      m_graphs->addGraph((*it).first.c_str(), & (*it).second);
    }

  }
}
