/* -*- C++ -*- */

/* -*- C++ -*- */

/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Crystal animation and 3D modeling application
 *
 * Authors:   Tommi Ilmonen 2001-2004
 * Copyright: Tommi Ilmonen and/or Tekes 2001-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#ifdef __OS_LINUX__
// Fix the broken NVidia headers:
#define GL_GLEXT_PROTOTYPES 1
#define GLAPI 
#include <GL/gl.h>
#include <GL/glext.h>
#endif // __OS_LINUX__


#include <crystal_tools.h>

#include <crystal_component_editor.h>
#include <crystal_density_editor.h>
#include <crystal_emitter_editor.h>
#include <crystal_engine.h>
#include <crystal_force_editor.h>
#include <crystal_general_editor.h>
#include <crystal_gee_editor.h>
#include <crystal_gl_data.h>
#include <crystal_helma_editor.h>
#include <crystal_particle_editor.h>

#include <Fluid/InputDevices.hpp>
#include <Fluid/MessageInt.hpp>
#include <Fluid/MessageMouse.hpp>
#include <Fluid/Mouse.hpp>
#include <Fluid/Processors.hpp>
#include <Fluid/TrackerSensorSelector.hpp>

#include <vee_error.h>
#include <vee_math.h>

#include <ConfigReader.hpp>

#include <GL/gl.h>

namespace Crystal { 

  static const char * editorNames[Tools::EDITORS] = {
    "General",
    "Helma",
    "HelmaQuick",
    "Particles",
    "Emitters",
    "Timeline",
    "Volume",
    "Force"
  };

  Tools::Tools(Engine * e)
    : m_engine(e),
      m_uiRecorder(0),
      m_trackerCount(0),
      // m_activeEditor(DENSITY_EDITOR),
      m_activeEditor(HELMA_EDITOR2),
      m_oldEditor(HELMA_EDITOR2),
      m_runningRecord(false),
      m_bgImage("check-low-contrast"),
      m_moveTools(false),
      m_renderingTransforms(true),
      m_handMask(0x3)
  {
    char buf[64];

    m_mouse[0] = m_mouse[1] = 0;
    m_mouseCount[0] = m_mouseCount[1] = 0;

    m_widgets.setGrabGroup(grabs());

    m_rotation.rotateY(-VEE_HALF_PI * 0.8);
    m_location.make(0.0, 1.0, 1.0);

    m_cursorLoc.clear();

    m_selectedColor.constantColor(VEE_Rgba::Red);

    std::string uiFile = 
      e->config()->get("UI").get("file").getString("helma.xui");
  
    // The order *is* important:
    m_densityEditor   = new DensityEditor(this);
    m_particleEditor  = new ParticleEditor(this);
    m_emitterEditor   = new EmitterEditor(this);
    m_generalEditor   = new GeneralEditor(this);
    m_componentEditor = new ComponentEditor(this);
    m_forceEditor     = new ForceEditor(this);
    // m_geeEditor       = new GeeEditor(this);
    m_helmaEditor     = new HelmaEditor(this, uiFile.c_str());
    m_helmaQuickEdit  = new HelmaQuickEdit(this);
    m_scale = 0.2;

    m_editors["Solids"]     = m_densityEditor;
    m_editors["Particles"]  = m_particleEditor;
    m_editors["Emitters"]   = m_emitterEditor;
    m_editors["General"]    = m_generalEditor;
    m_editors["Components"] = m_componentEditor;
    m_editors["Forces"]     = m_forceEditor;
    m_editors["Helma"]      = m_helmaEditor;
    m_editors["HelmaQuick"] = m_helmaQuickEdit;

    m_editors2[PARTICLE_EDITOR]  = m_particleEditor;
    m_editors2[EMITTER_EDITOR]   = m_emitterEditor;
    m_editors2[GENERAL_EDITOR]   = m_generalEditor;
    m_editors2[COMPONENT_EDITOR] = m_componentEditor;
    m_editors2[DENSITY_EDITOR]   = m_densityEditor;
    m_editors2[FORCE_EDITOR]     = m_forceEditor;
    m_editors2[HELMA_EDITOR2]    = m_helmaEditor;
    m_editors2[HELMA_QUICK_EDIT] = m_helmaQuickEdit;
    
    m_currentLoc[0].clear();
    m_currentLoc[1].clear();
    m_currentRot[0].identity();
    m_currentRot[1].identity();

    Fluid::Vector2 v1, v2, v3, v4;
    v1.make(-1.0, 3.0);
    v2.make( 4.0, 0.4);
  
    v3.make(-2.0, 3.0);
    v4.make(0.9, 0.4);


    TextureButton * txb = new TextureButton(this);
    txb->setArea(v1, v2);
    txb->setLabel("Tools");
    txb->msgAddSink("press", "startmove", this);
    txb->msgAddSink("release", "endmove", this);
    txb->setFontSize(0.3);
    txb->setButtonMask(~0); // receive all button types.

    m_headLine = txb;

    m_widgets.addWidget(txb);


    uint i;

    Menu * menu = new Menu(this);

    menu->setSounds(buttonSounds());
    menu->setLabel("Menu");
    menu->setArea(v3, v4);
  
    m_widgets.addWidget(menu);

    menu->addItem("Play", "play", this);
      
    for(i = 0; i < EDITORS; i++) {

      if(i == HELMA_QUICK_EDIT) // Do not include the quick editor
	continue;
      
      sprintf(buf, "tool-%u", i);
      menu->addItem(editorNames[i], buf, this);
    }

    selectEditor(m_activeEditor);  

    ConfigReader::Chunk soundDefs = m_engine->config()->get("Sounds");

    SoundGroup *sg = & * m_buttonSounds;

    sg->setSound(ColorGroup::ACTIVE, 
		 soundDefs.get("button_active").getString("active.wav"));
    sg->setSound(ColorGroup::GRAB, 
		 soundDefs.get("button_grab").getString("grab.wav"));
    sg->setMasterVolume(0.1);

    
  }

  Tools::~Tools()
  {
    delete m_emitterEditor;
    delete m_particleEditor;
    delete m_generalEditor;
    delete m_densityEditor;
    delete m_componentEditor;
    delete m_forceEditor;
  }

  bool Tools::msgProcess(const char *id, Fluid::Message *m)
  {
    VEE_trace("Tools::msgProcess # \"%s\"", id);

    bool ok = false;

    if(strcmp(id, "startmove") == 0) {

      MessageInt * mi = dynamic_cast<MessageInt *> (m);

      int hand = mi ? mi->value() : RIGHT_HAND;
      m_dragHand = hand;
      
      m_moveTools = true;
      m_dragBeginLoc = m_currentLoc[hand];
      m_dragBeginRot = m_currentRot[hand];

      m_dragBeginLocRel = ((m_currentLoc[hand] - m_location) /*/ m_scale */) * m_rotation;
      //m_dragBeginLocRel = m_rotation * ((m_currentLoc - m_location) / m_scale);
      m_dragBeginRot2 = m_rotation;
      ok = true;

      VEE_trace("Tools::msgProcess # startMove %.3f %.3f %.3f",
		m_dragBeginLocRel.x, m_dragBeginLocRel.y, m_dragBeginLocRel.z);
    }
    else if(strcmp(id, "endmove") == 0) {
      m_moveTools = false;    
      ok = true;
    }
    else if(strncmp(id, "tool-", 5) == 0) {
      int tool = atoi(id + 5);

      selectEditor((Tool) tool);

      ok = true;
    }
    else if(strcmp(id, "play") == 0) {
      engine()->skipMode(Engine::MODE_PLAY, true);

      ok = true;
    }

    return ok;
  }
  
  void Tools::handleInput()
  {
    const char * fname = "Tools::handleInput";

    if(m_uiRecorder) {
      m_uiRecorder->handleInput();
      return;
    }

    m_editors2[m_activeEditor]->beforeEvents();
    m_widgets.beforeEvents();

    for(int hand = 0; hand < 2; hand++) {

      IONodeBuffer<MouseSample> * mouse = m_mouse[hand];

      if(!mouse || !((0x1 << hand) & m_handMask))
	continue;

      Fluid::IONodeBuffer<Fluid::Vector3> * handLoc = 
	& m_engine->handLoc(hand)->locations();
      
      Fluid::IONodeBuffer<Fluid::Matrix3> * handRot = 
	& m_engine->handLoc(hand)->rotations();
      
      bool anyEvent = false;
    
      // bool gotEvent = false;
      
      /* VEE_trace("%s # hand = %d, mc = %ld, sc = %ld", fname, hand,
	 m_mouseCount[hand], (long) mouse->getSampleCount()); */

      for(; m_mouseCount[hand] <= (long) mouse->getSampleCount(); 
	  m_mouseCount[hand]++) {
	
	// gotEvent = true;
	anyEvent = true;
	
	long count = m_mouseCount[hand];
	  
	Fluid::TimeStamp t = mouse->getTimeStamp(count);
	
	long trackerSample = handLoc->timeStampToIndex(t);
	
	if(trackerSample <= 0) {
	  VEE_trace("%s # Discarded tracker sample with index %ld", 
		    fname, trackerSample);
	  continue;
	}

	m_currentRot[hand] = handRot->getIndex(trackerSample);
	m_currentLoc[hand] = handLoc->getIndex(trackerSample);
	
	// Fluid::Vector3 loc = ((m_currentLoc - m_location) / m_scale) * m_rotation;
	Fluid::Matrix3 rot = applyMyTransform(m_currentRot[hand]);
	Fluid::Vector3 loc = applyMyTransform(m_currentLoc[hand]);
	
	Fluid::MouseSample ms = mouse->getIndex(count);
	
	InEvent e;
	
	e.set(loc, rot, ms.state, m_pastState[hand], hand, t);
	
	// VEE_trace("%s # Mouse buttons = %x", fname, (int) ms.state);
		
	m_editors2[m_activeEditor]->handleEvent(&e);
	
	m_widgets.handleEvent(&e);
	
	m_pastState[hand] = ms.state;
      }

      
      if(!anyEvent) {
	
	// No new events, generate one to satisfy GUI needs.
	
	m_currentRot[hand] = handRot->getNewest();
	m_currentLoc[hand] = handLoc->getNewest();
	
	// Fluid::Vector3 loc = ((m_currentLoc - m_location) / m_scale) * m_rotation;
	Fluid::Vector3 loc = applyMyTransform(m_currentLoc[hand]);
	Fluid::Matrix3 rot = applyMyTransform(m_currentRot[hand]);
	
	Fluid::TimeStamp t = handRot->getTimeStamp(m_mouseCount[hand]);
	
	InEvent e;
	
	e.set(loc, rot, 
	      m_pastState[hand], m_pastState[hand], 
	      hand, t);
	
	m_editors2[m_activeEditor]->handleEvent(&e);
	
	m_widgets.handleEvent(&e);
      }
    }
  }

  void Tools::update()
  {
    if(!m_engine) {
      VEE_fatal("Tools::update # Need an engine and a mouse");
    }

    Fluid::IONodeBuffer<Fluid::Vector3> * handLoc = 
      & m_engine->strongHandLoc()->locations();
  
    Fluid::Vector3 loc = handLoc->getNewest();
    if(!handLoc->getSampleCount()) {
      loc.clear();
    }
    loc.length();
    m_location.length();
    Fluid::Vector3 locdiff = loc - m_location;
    m_cursorLoc = (locdiff / m_scale) * m_rotation;

    if(m_activeEditor == DENSITY_EDITOR)
      m_densityEditor->update();
    else if(m_activeEditor == PARTICLE_EDITOR)
      m_particleEditor->update();
    else if(m_activeEditor == EMITTER_EDITOR)
      m_emitterEditor->update();
    else if(m_activeEditor == FORCE_EDITOR)
      m_forceEditor->update();
    else if(m_activeEditor == HELMA_EDITOR2)
      m_helmaEditor->update();
    else if(m_activeEditor == HELMA_QUICK_EDIT)
      m_helmaQuickEdit->update();
  
    if(m_moveTools) {

      Fluid::Matrix3 m1, m2, m3;

      m2.rotateY(VEE_HALF_PI);
      m3.rotateZ(VEE_HALF_PI);

      m2 = m3 * m2;

      Fluid::Vector3 v1 = m_dragBeginRot * m2 * Fluid::Vector3(0.0, 0.0, 1.0);
      float angle1 = -atan2f(v1.x, v1.z);

      Fluid::Vector3 v2 = m_currentRot[m_dragHand] * 
	m2 * Fluid::Vector3(0.0, 0.0, 1.0);
      float angle2 = -atan2f(v2.x, v2.z);
      m1.rotateY(angle1 - angle2);
      m_rotation = m1 * m_dragBeginRot2;

      /* Fluid::Vector3 dragMove = m_currentLoc - m_dragBeginLoc;
	 m_location = m_dragBeginLoc + dragMove; */

      m_location = m_currentLoc[m_dragHand] - m_rotation * m_dragBeginLocRel;

      // VEE_trace("Tools::update # %.3f  %.3f", angle1, angle2);

      MessageInt mi(0);
      msgSend("move", & mi);
    }
  }
  
  void Tools::render(GlData *glData)
  {
    // Set the OpenGL parameters to what we need:
    // glDisable(GL_DEPTH_TEST);
    glDisable(GL_LIGHTING);

    if(!m_renderingTransforms) {
      m_editors2[m_activeEditor]->render(glData);
      return;
    }

    glPushMatrix();

    Fluid::Vector3 vz;
    Fluid::Matrix3 mi;
    vz.clear();
    mi.identity();
  
    GlData::applyGlTransform(mi, m_location);
    GlData::applyGlTransform(m_rotation, vz);
    glScalef(m_scale, m_scale, m_scale);

    // Draw cross on top of the cursor:
    /*
    glLineWidth(1);
    glBegin(GL_LINES);
    glColor4fv(VEE_Rgba::White.data());

    float l = 0.5;

    glVertex3f(m_cursorLoc[0] - l, m_cursorLoc[1], m_cursorLoc[2]);
    glVertex3f(m_cursorLoc[0] + l, m_cursorLoc[1], m_cursorLoc[2]);

    glVertex3f(m_cursorLoc[0], m_cursorLoc[1] - l, m_cursorLoc[2]);
    glVertex3f(m_cursorLoc[0], m_cursorLoc[1] + l, m_cursorLoc[2]);

    glVertex3f(m_cursorLoc[0], m_cursorLoc[1], m_cursorLoc[2] - l);
    glVertex3f(m_cursorLoc[0], m_cursorLoc[1], m_cursorLoc[2] + l);

    glEnd();
    */

    glBlendColor(0, 1, 0, 1);
    glBlendEquation(GL_FUNC_ADD);

    glDepthMask(GL_TRUE);

    if(!m_uiRecorder) {
      // Render the active editor:
      m_editors2[m_activeEditor]->render(glData);

      m_widgets.renderGl(glData);
    }

    glPopMatrix();

    // Density recorder prefers to draw into world coordinates:
    if(m_uiRecorder)
      m_uiRecorder->render(glData);
  
    m_componentEditor->postRender(glData);
    // Debug transformations:
    /* 
       MJ_Array2D<ParticleEditor::ParticleData> * particles = 
       & m_particleEditor->particles();

       glDisable(GL_TEXTURE_2D);

       glBegin(GL_POINTS);

       glColor4f(1, 1, 1, 1);

       for(uint y = 0; y < particles->rows(); y++) {
       for(uint x = 0; x < particles->cols(); x++) {
       Fluid::Vector2 center2 = particles->get(y, x).m_button->center();
       VEE_Vector3 center3(center2.x, center2.y, 0.0f);
       VEE_Vector3 inworld = worldLocation(center3);
       glVertex3fv(inworld.data());
       }
       }

       glEnd();
    */
  }

  void Tools::start()
  {
    m_trackerCount = (long) m_engine->tracker()->outputBuffer().getSampleCount();
    
    IONode * source = m_engine->processors().findIONode("mousedata/output");
    m_mouse[0] = dynamic_cast<IONodeBuffer<MouseSample> *> 
      (source);

    source = m_engine->processors().findIONode("mousedata2/output");
    m_mouse[1] = dynamic_cast<IONodeBuffer<MouseSample> *> 
      (source);
    
    if(!m_mouse[0]) {
      m_mouse[0] = & m_engine->mouse()->outputBuffer();
      VEE_trace("Tools::start # Using direct mouse (no napra)");
    }
    else
      VEE_trace("Tools::start # Got wrapped mouse (napra) %p %p",
		m_mouse[0], m_mouse[1]);
    
    m_mouseCount[0] = (long) m_mouse[0]->getSampleCount();
    if(m_mouse[1])
      m_mouseCount[1] = (long) m_mouse[1]->getSampleCount();

    if(m_activeEditor == HELMA_QUICK_EDIT)
      selectEditor(m_oldEditor);
  }

  void Tools::stop()
  {
    m_editors2[m_activeEditor]->stop();  
  }

  void Tools::nextEditor()
  {
    if(m_activeEditor == DENSITY_EDITOR) {
      if(m_uiRecorder) m_uiRecorder = 0;
      else {
	m_uiRecorder = m_densityEditor->recorder();
	m_uiRecorder->start();
      }
    }

    // selectEditor((Editor) (((int) m_activeEditor + 1) % EDITORS));
    // VEE_trace("Tools::nextEditor # %d", (int) m_activeEditor);
  }

  void Tools::toggleDensityRecord()
  {
    if(m_uiRecorder) {
      m_uiRecorder->stop();
      m_uiRecorder = 0;
    }
    else {
      m_uiRecorder = m_densityEditor->recorder();
      m_uiRecorder->start();
    }
  }

  void Tools::useQuickEditor(int hand)
  {
    VEE_trace("Tools::useQuickEditor # hand %d", hand);

    m_engine->skipMode(Engine::MODE_QUICK_EDIT);

    selectEditor(HELMA_QUICK_EDIT);
    m_handMask = 0x1 << hand;

    m_helmaQuickEdit->init(m_helmaEditor, hand, 
			   m_engine->nextRecorderName(hand));
  }

  void Tools::selectEditor(Tool e)
  {
    VEE_trace("Tools::selectEditor # %d", (int) e);
 
    if((int) e >= EDITORS || !m_headLine)
      VEE_fatal("Tools::selectEditor");

    if(e == HELMA_QUICK_EDIT && m_activeEditor != HELMA_QUICK_EDIT)
      m_oldEditor = m_activeEditor;

    m_handMask = 0x3;

    m_editors2[m_activeEditor]->stop();
    m_activeEditor = e;
    m_renderingTransforms = true;
    m_editors2[m_activeEditor]->start();

    char buf[64];

    sprintf(buf, "Tools / %s", editorNames[e]);
    m_headLine->setLabel(buf);  
  }

  void Tools::setUiHandler(UiHandler *r) 
  { 
    if(m_uiRecorder) m_uiRecorder->stop();

    m_uiRecorder = r; 

    if(r) r->start();
  }

  QDomElement Tools::writeDom(QDomDocument &doc)
  {
    QDomElement elem = doc.createElement("Tools");

    for(iterator it = m_editors.begin(); it != m_editors.end(); it++) {
      /* QDomElement tmp = doc.createElement((*it).first.c_str());
	 tmp.appendChild((*it).second->writeDom(doc));
	 elem.appendChild(tmp); */
      elem.appendChild((*it).second->writeDom(doc));
    }
    return elem;
  }

  bool Tools::readDom(QDomElement &elem)
  {
    VEE_trace("Tools::readDom # tag = \"%s\"", elem.tagName().ascii());

    if(elem.tagName() != "Tools") {
      return false;
    }
    QDomNode n = elem.firstChild();

    bool ok = true;
  
    while ( !n.isNull() ) {
      if ( n.isElement() ) {
	QDomElement e = n.toElement();

	QString tag = e.tagName();

	iterator it = m_editors.find(tag.ascii());

	VEE_trace("Tools::readDom # \"%s\"", tag.ascii());

	if(it == m_editors.end())
	  ok = false;
	else 
	  ok = ok && (*it).second->readDom(e);
      }
      n = n.nextSibling();
    }
    return ok;
  }

}

