/* -*- C++ -*- */

/* -*- C++ -*- */

/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Crystal animation and 3D modeling application
 *
 * Authors:   Tommi Ilmonen 2001-2004
 * Copyright: Tommi Ilmonen and/or Tekes 2001-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <crystal_vortex_recorder.h>

#include <crystal_action_list.h>
#include <crystal_engine.h>
#include <crystal_dom_util.h>
#include <crystal_force_editor.h>
#include <crystal_gl_data.h>
#include <crystal_location_action.h>
#include <crystal_remove_action.h>
#include <crystal_tools.h>
#include <crystal_value_action.h>

#include <vee_vortex.h>
#include <vee_error.h>
#include <vee_strs.h>
#include <vee_vortex.h>

#include <Fluid/Fistiness.hpp>
#include <Fluid/TrackerSensorSelector.hpp>

namespace Crystal { 

  static bool debugVortex = true;

  CreateVortexAction::~CreateVortexAction() {}
  
  void CreateVortexAction::act(ActionList *al)
  {
    if(debugVortex)
      VEE_trace("CreateVortexAction::act # %d %d", 
		m_id, al->getScene()->totalForces());

    VEE_Vortex *vortex = new VEE_Vortex;
    vortex->setLocation(m_location);
    vortex->setAgeLimit(1e+10);

    al->setObject(m_id, vortex);
    al->getScene()->addNewForce(vortex);
    
    if(debugVortex)
      VEE_trace("CreateVortexAction::act # DONE %d %d", 
		m_id, al->getScene()->totalForces());
  }

  QDomElement CreateVortexAction::writeDom(QDomDocument &doc)
  {
    QDomElement elem = doc.createElement("CreateVortexAction");

    writeDomTimeStamp(elem, doc);

    addTextChild(elem, doc, "location", m_location);
    addTextChild(elem, doc, "id", m_id);

    return elem;
  }

  bool CreateVortexAction::readDom(QDomElement &elem) 
  { 
    if(elem.tagName() != "CreateVortexAction")
      return false;

    readDomTimeStamp(elem);

    m_location = getVector3(elem, "location");
    m_id       = getInt(elem, "id", -1);

    return true; 
  }

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  VortexRecorder::VortexRecorder(bool motion)
    : m_motion(motion),
      m_vortexId(-1),
      m_trackerCount(0),
      m_gloveCount(0),
      m_range(0),
      m_magnitude(0)
  {
    m_location.clear();
    m_axes.clear();
  }

  VortexRecorder::~VortexRecorder()
  {}

  void VortexRecorder::prepare(ActionList *al)
  {
    Engine *engine = al->engine();
    m_trackerCount = (long) 
      engine->weakHandLoc()->locations().getSampleCount();
    m_gloveCount = (long) engine->glove()->outputBuffer().getSampleCount();
  }

  void VortexRecorder::beginRecording(ActionList *al, 
				     const Fluid::TimeStamp &now)
  {
    const char * fname = "VortexRecorder::beginRecording";

    if(debugVortex)
      VEE_trace("%s # %p", fname);

    Engine *engine = al->engine();
    m_trackerCount = (long) engine->weakHandLoc()->locations().getSampleCount();
    m_gloveCount = (long) engine->glove()->outputBuffer().getSampleCount();

    if(debugVortex)
      VEE_trace("%s # %d", fname, 1);  
  
    Fluid::IONodeBuffer<Fluid::Vector3> * handLoc = 
      & engine->weakHandLoc()->locations();
    
    Fluid::Vector3 location = handLoc->getIndex(m_trackerCount);
  
    m_vortexId = al->allocateObjectId();

    if(debugVortex)
      VEE_trace("%s # vortexId = %d", fname, m_vortexId);  

    al->appendAction(new CreateVortexAction(m_vortexId, location, 
					   now));

    m_mode = RECORD;

    if(debugVortex)
      VEE_trace("VortexRecorder::beginRecording # EXIT");
  }

  void VortexRecorder::endRecording(ActionList *al, 
				   const Fluid::TimeStamp &now)
  {
    VEE_trace("VortexRecorder::endRecording");

    Fluid::TimeStamp ts = now;

    if(!m_motion) {
      Fluid::TimeStamp tmp;
      tmp.setSeconds(1.0e+5);
      ts += tmp;
    }

    al->appendAction(new RemoveAction(m_vortexId, ts));

    m_mode = MUTE;
  }

  static std::string locStr("location");
  static std::string rangeStr("range");
  static std::string heightStr("height");
  static std::string axesStr("axes");
  static std::string magnitudeStr("magnitude");

  void VortexRecorder::record(ActionList *al, 
			     const Fluid::TimeStamp &)
  {  
    const char * fname = "VortexRecorder::record";
  
    if(debugVortex)
      VEE_trace(fname);

    Engine *engine = al->engine();

    Fluid::IONodeBuffer<Fluid::Vector3> * handLoc = 
      & engine->weakHandLoc()->locations();

    Fluid::IONodeBuffer<Fluid::Vector3> * handLoc2 = 
      & engine->strongHandLoc()->locations();

    /* Fluid::IONodeBuffer<Fluid::Matrix3> * handRot = 
       & engine->weakHandLoc()->rotations(); */

    ValueGraphs * graphs = & al->getGraph(m_vortexId);
  
    for(;m_trackerCount < (long)handLoc->getSampleCount(); m_trackerCount++) {
    
      m_location  = handLoc->getIndex(m_trackerCount);
      Fluid::Vector3 location2 = handLoc2->getIndex(m_trackerCount);
      m_axes = m_location - location2;

      m_range = (m_location - location2).length();

      if(m_mode == RECORD) {

	Fluid::TimeStamp sampleTime = handLoc->getTimeStamp(m_trackerCount);
	Fluid::TimeStamp ts = al->realTimeToRecordTime(sampleTime);
	
	double dts = ts.secondsD();

	graphs->appendVector3(locStr, dts, m_location);
	graphs->appendVector3(axesStr, dts, m_axes);
	
	graphs->appendFloat(rangeStr, dts, m_range);
	graphs->appendFloat(heightStr, dts, m_range);
	
	if(debugVortex)
	  VEE_trace("%s # time = %lf, range = %f", fname, dts, m_range);
      }

    }
  
    Fluid::Fistiness  *fistiness = engine->fistiness();

    Fluid::IONodeBuffer<float> * values = & fistiness->buffer();

    float magnitude = engine->tools()->forceEditor()->current()->magnitude();
    float sensitivity = engine->tools()->forceEditor()->current()->sensitivity();

    for(; m_gloveCount < (long) values->getSampleCount(); m_gloveCount++) {

      float fist = 1.0 - values->getIndex(m_gloveCount);
    
      if(m_mode == RECORD) {

	Fluid::TimeStamp sampleTime = 
	  values->getTimeStamp(m_gloveCount);
	
	Fluid::TimeStamp ts = al->realTimeToRecordTime(sampleTime);
	
	double dts = ts.secondsD();
	
	m_magnitude = fist * sensitivity + magnitude;

	if(m_magnitude < 0.0) m_magnitude = 0.0;
	
	graphs->appendFloat(magnitudeStr, dts, m_magnitude);
      }
    }
    
    if(debugVortex)
      VEE_trace("%s # EXIT", fname);
  }

  void VortexRecorder::render(Engine *e, GlData *gld)
  {
    if(m_mode == RECORD)
      return;

    VEE_Vortex v;

    v.setLocation(m_location);
    v.setAxes(m_axes);
    v.setRange(m_range);
    v.setHeight(m_range);
    v.setMagnitude(m_magnitude);
    v.setAge(e->timing().currentTimeToRecordTime().secondsF());

    gld->m_veeRenderer.initRenderState();
    gld->m_veeRenderer.setForceFrontColor(& VEE_Rgba::Green);
    v.render(& gld->m_veeRenderer);
    gld->m_veeRenderer.setForceFrontColor(0);
  }

}
