/* -*- C++ -*- */

/* -*- C++ -*- */

/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Crystal animation and 3D modeling application
 *
 * Authors:   Tommi Ilmonen 2001-2004
 * Copyright: Tommi Ilmonen and/or Tekes 2001-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <crystal_widget.h>

#include <crystal_ac.h>
#include <crystal_dom_util.h>
#include <crystal_engine.h>
#include <crystal_gl_data.h>

#include <vee_error.h>
#include <vee_math.h>

#include <Fluid/MessageFloat.hpp>
#include <Fluid/MessageInt.hpp>
#include <Fluid/MessageString.hpp>

#include <Solar/Model.hpp>

#include <ac_vr_control.h>

#include <GL/gl.h>

#include <qdir.h>

namespace Crystal { 

  ColorGroup::ColorGroup()
  {
    m_colors[INACTIVE].set(0.7, 0.7, 0.7);
    m_colors[ACTIVE].set(1.0, 0.7, 0.3);
    m_colors[GRAB].set(1.0, 0.5, 0.0);
  }

  ColorGroup::~ColorGroup()
  {}

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  SoundGroup::SoundGroup()
    : m_masterVolume(0.15)
  {}

  SoundGroup::~SoundGroup()
  {}

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////


  Widget::Widget(Transformation * host)
    : Transformation(host),
      m_state(INACTIVE),
      m_fontSize(0.15),
      m_buttonMask(DEFAULT_MASK),
      m_trackHand(-1),
      m_widgetManager(0)
  {
    m_bbox.low().make(0.0);
    m_bbox.high().make(1.0);
  }

  Widget::~Widget()
  {
    if(m_widgetManager)
      m_widgetManager->forget(this);
  }

  void Widget::beforeEvents()
  {
    if(m_state == ACTIVE)
      m_state = INACTIVE;
  }

  void Widget::handleEvent(const InEvent *)
  {}

  void Widget::renderGl(GlData *, bool)
  {}

  void Widget::update()
  {}

  QDomElement Widget::writeDom(QDomDocument &doc)
  {
    QDomElement e = doc.createElement(type());

    addTextChild(e, doc, "rotation", m_rotation);
    addTextChild(e, doc, "location", m_location);
    addTextChild(e, doc, "scale",    m_scale);
    addTextChild(e, doc, "low",      m_bbox.low());
    addTextChild(e, doc, "high",     m_bbox.high());

    return e;
  }

  bool Widget::readDom(QDomElement &e)
  {
    if(e.tagName() != type()) {
      VEE_error("Widget::readDom # %s != %s", e.tagName().ascii(), type());
      return false;
    }
    
    m_rotation = getMatrix3(e, "rotation", m_rotation);
    m_location = getVector3(e, "location", m_location);
    m_scale    = getFloat(e, "scale", m_scale);
    m_bbox.low()  = getVector3(e, "low",  m_bbox.low());
    m_bbox.high() = getVector3(e, "high", m_bbox.high());

    
    QDomNode n = e.namedItem("rotate");
    
    while(!n.isNull()) {
      QDomElement ne = n.toElement();
      
      if(ne.tagName() == "rotate") {
	Vector3 axis = toVector3(ne.attribute("axis", "1 0 0"));
	float degrees = ne.attribute("degrees", "0").toFloat();
	Matrix3 r;
	r.rotateAroundAxis(axis, degrees / 180.0f * VEE_PI);
	m_rotation = r * m_rotation;
      }
      
      n = n.nextSibling();
    }
    
    
    return true;
  }

  void Widget::transmitValue()
  {}

  /** Returns true if this widget can take advantage of extra depth. */
  bool Widget::useExtraZ()
  {
    return false;
  }

  void Widget::renderFrameGl(Vector2 low,
			     Vector2 high,
			     float b, float offset)
  {
    glDisable(GL_TEXTURE_2D);

    ColorGroup::State s = colorState(m_state);
  
    const float * color = m_colors.color(s).data();
  
    float shine = 0.5;

    glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, & shine);
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, color);
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
    glColor4fv(color);
  
    frameGl(low, high, b, offset);
  }

  void Widget::frameGl(Vector2 low,
		       Vector2 high,
		       float b, float offset)
  {
    glBegin(GL_TRIANGLE_STRIP);
  
    glNormal3f(0.0f, 0.0f, 1.0f);

    glVertex3f(low[0]    , low[1], offset); // 0
    glVertex3f(low[0] + b, low[1] + b, offset);
    glVertex3f(high[0]   , low[1], offset); // 2
    glVertex3f(high[0]- b, low[1] + b, offset);
    glVertex3f(high[0]   , high[1], offset); // 4
    glVertex3f(high[0]- b, high[1] - b, offset);
    glVertex3f(low[0]    , high[1], offset); // 6
    glVertex3f(low[0] + b, high[1] - b, offset);
    glVertex3f(low[0]    , low[1], offset); // 8
    glVertex3f(low[0] + b, low[1] + b, offset);

    glEnd();
  }


  void Widget::transmitSound(ColorGroup::State s,
			     const Vector3 &l) const
  {
    if(!(*m_sounds).sound(s).size() || !audio)
      return;

    Vector3 wl = worldLocation(l);

    AC_Vector3 tmp(wl.data());

    VEE_trace("Widget::transmitSound # %d %s  [%f %f %f]",
	      (int) s, (*m_sounds).sound(s).c_str(), wl.x, wl.y, wl.z);

    float volume = (*m_sounds).masterVolume();

    audio->playSample(0, (*m_sounds).sound(s).c_str(), volume, 
		      tmp, false);
  }

  void Widget::quadAroundText(fntRenderer * render,
			      const char * text, 
			      float size,
			      float border)
  {
    fntFont * font = render->getFont();

    float x1, x2, y1, y2, z;

    font->getBBox(text, size, 0.0,
		  &x1, &x2, &y1, &y2);
    float w = x2 - x1;
    float h = y2 - y1;

    render->getCursor(& x1, & y1, & z);
    
    glBegin(GL_QUADS);

    glNormal3f(0, 0, 1);

    z -= 0.005;

    float expand = h * border;
    w += expand * 2.0f;
    x1 -= expand;

    h += expand * 2.0f;
    y1 -= expand;

    glVertex3f(x1,     y1,     z);
    glVertex3f(x1 + w, y1,     z);
    glVertex3f(x1 + w, y1 + h, z);
    glVertex3f(x1,     y1 + h, z);
    
    glEnd();
  }

  Widget * Widget::create(const char * type)
  {
    if(strcmp(type, "CheckButton") == 0)
      return new CheckButton();
    if(strcmp(type, "ColorCube") == 0)
      return new ColorCube();
    if(strcmp(type, "ColorCylinder") == 0)
      return new ColorCylinder();
    if(strcmp(type, "FileMenu") == 0)
      return new FileMenu();
    if(strcmp(type, "Menu") == 0)
      return new Menu();
    if(strcmp(type, "Slider") == 0)
      return new Slider();
    if(strcmp(type, "TextureButton") == 0)
      return new TextureButton();

    return 0;
  }

  static const char * locationNames[] = {
    "center",
    "left",
    "right",
    "above",
    "below",
    0
  };

  
  const char * Widget::locationName(Location l)
  {
    return locationNames[l];
  }

  Widget::Location Widget::locationId(const char *name)
  {
    for(int i=0; locationNames[i]; i++)
      if(strcmp(locationNames[i], name) == 0)
	return (Location) i;
    
    return CENTER;
  }


  void Widget::pushLocalMatrix(bool push)
  {
    if(!push)
      return;

    glPushMatrix();

    GlData::applyGlTransform(m_rotation, m_location);
  }

  void Widget::popLocalMatrix(bool pop)
  {
    if(!pop)
      return;

    glPopMatrix();
  }

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  ColorGroup::State Widget::colorState(State state)
  {
    ColorGroup::State s = ColorGroup::INACTIVE;

    if(state == ACTIVE)
      s = ColorGroup::ACTIVE;
    if(state == GRAB)
      s = ColorGroup::GRAB;
  
    return s;
  }

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////


  TextureButton::TextureButton(Transformation * host)
    : Widget(host),
      m_texture("button_bg"),
      m_labelLoc(Widget::CENTER),
      m_border(0.015),
      m_activeDistance(-1.0)
  {
    m_uv[0].make(0.0, 1.0);
    m_uv[1].make(1.0, 0.0);
    m_bbox.low().make(0, 0, 0);
    m_bbox.high().make(1, 1, 0);
  }

  TextureButton::~TextureButton()
  {}

  void TextureButton::handleEvent(const InEvent *e)
  {
    const char * fname = "TextureButton::handleEvent";

    if(m_state == HIDDEN)
      return;

    // VEE_trace(fname);

    if(m_state == DISABLED)
      return;

    if((m_trackHand >= 0) && (m_trackHand != e->hand()))
      return;

    Vector3 loc3 = applyMyTransform(e->location());

    Vector2 loc;
    loc[0] = loc3[0];
    loc[1] = loc3[1];

    bool inside;

    Vector3 span = m_bbox.span();

    float activeDistance = m_activeDistance > 0.0 ? 
      m_activeDistance : VEE_min(span.x, span.y) * 2.0f;

    // Test is it inside:
    if(loc[0] > m_bbox.low()[0] &&
       loc[1] > m_bbox.low()[1] &&
       loc[0] < m_bbox.high()[0] &&
       loc[1] < m_bbox.high()[1] &&
       VEE_abs(loc3[2]) < activeDistance) {

      if(m_state == INACTIVE) {
	m_state = ACTIVE;
	MessageInt mi(ACTIVE, e->time());
	msgSend("state", &mi);
      }
      inside = true;
    }
    else {
      inside = false;

      /* if(m_state != GRAB)
	 m_state = INACTIVE; */
    }

    // Test if we should grab:
    if(e->buttonState() & m_buttonMask) {
      if(m_state == ACTIVE) {
	m_state = GRAB;
	MessageInt mi(e->hand(), e->time());
	msgSend("press", &mi);
	VEE_trace("%s # Grab %p", fname, this);
	transmitSound(ColorGroup::ACTIVE, loc3);
	m_trackHand = e->hand();
	newState(m_state, e);
      }
    }
    // Test if the grab should be released:
    else if(m_state == GRAB) {
      m_state = inside ? ACTIVE : INACTIVE;
      MessageInt mi(e->hand(), e->time());
      msgSend("release", &mi);
      VEE_trace("%s # Release %p", fname, this);
      transmitSound(ColorGroup::INACTIVE, loc3);
      newState(m_state, e);
      m_trackHand = -1;
    }
  }

  void TextureButton::renderGl(GlData *gld, bool transform)
  {
    if(m_state == HIDDEN)
      return;
  
    pushLocalMatrix(transform);    

    renderFrameGl(low().vector2(), high().vector2(), m_border);

    VEE_GlTexture2D * tex = gld->m_textures.getTexture(m_texture);
  
    if(tex) {
  
      tex->use();

      VEE_Rgba color = m_colors.color(colorState(m_state));
      color.setAlpha(1.0);
    
      glEnable(GL_BLEND);
      glColor4fv(color.data());
      // glEnable(GL_TEXTURE_2D);
    
      /* VEE_trace("TextureButton::renderGl # %u %f %f %f %f", 
	 (uint) tex->id(), m_uv[0][0], m_uv[0][1], m_uv[1][0], m_uv[1][1]); */
    
      glBegin(GL_QUADS);
    
      glTexCoord2f(m_uv[0][0], m_uv[0][1]);
      glVertex2f(low()[0] + m_border, low()[1] + m_border);
    
      glTexCoord2f(m_uv[1][0], m_uv[0][1]);
      glVertex2f(high()[0] - m_border, low()[1] + m_border);

      glTexCoord2f(m_uv[1][0], m_uv[1][1]);
      glVertex2f(high()[0] - m_border, high()[1] - m_border);
    
      glTexCoord2f(m_uv[0][0], m_uv[1][1]);
      glVertex2f(low()[0] + m_border, high()[1] - m_border);
    
      glEnd();
    }
    else
      glDisable(GL_BLEND);
  
    if(m_label.size()) {

      gld->m_fntRenderer.setPointSize(m_fontSize);

      // Need a background quad for the label text:
      bool needBg = false;

      fntFont * font = gld->m_fntRenderer.getFont();
      
      if(m_labelLoc == Widget::ABOVE) {
	gld->m_fntRenderer.start3f(low()[0], 
				   high()[1],
				   0.01);
	needBg = true;
      }
      else if(m_labelLoc == Widget::BELOW) {
	gld->m_fntRenderer.start3f(low()[0], 
				   low()[1] - 0.2,
				   0.01);
	needBg = true;
      }
      else if(m_labelLoc == Widget::CENTER)
	gld->m_fntRenderer.start3f
	  (low()[0], 
	   (low()[1] + high()[1]) * 0.5 - m_fontSize * 0.5, 
	   0.01);
      else if(m_labelLoc == Widget::LEFT) {	
	float x1, x2, y1, y2;
	
	font->getBBox(m_label.c_str(), m_fontSize, 0.0,
		      &x1, &x2, &y1, &y2);
	float w = x2 - x1;

	gld->m_fntRenderer.start3f(low().x - w - m_border, low().y, 0.0);
	needBg = true;
      }

      if(needBg) {
	quadAroundText(&gld->m_fntRenderer, m_label.c_str(), m_fontSize);
      }

      glColor4fv(VEE_Rgba::White.data());
      gld->m_fntRenderer.puts(m_label.c_str());
    }

    popLocalMatrix(transform);
  }

  QDomElement TextureButton::writeDom(QDomDocument &doc)
  {
    QDomElement e = Widget::writeDom(doc);
    
    addTextChild(e, doc, "label",    m_label.c_str());
    addTextChild(e, doc, "labelloc", locationName(m_labelLoc));
    addTextChild(e, doc, "texture",  m_texture.c_str());
    addTextChild(e, doc, "border",   m_border);
    addTextChild(e, doc, "activedistance", m_activeDistance);

    return e;
  }

  bool TextureButton::readDom(QDomElement &e)
  {
    if(!Widget::readDom(e))
      return false;

    m_label = getString(e, "label", m_label.c_str()).ascii();
    m_labelLoc = locationId
      (getString(e, "labelloc", 
		 locationName(m_labelLoc)).ascii());
    m_texture  = getString(e, "texture", m_texture.c_str()).ascii();
    m_border = getFloat(e, "border", m_border);
    m_activeDistance = getFloat(e, "activedistance", m_activeDistance);

    return true;
  }

  const char * TextureButton::type()
  {
    return "TextureButton";
  }

  void TextureButton::autoUv(bool stretch)
  {
    float activeWidth  = high()[0] - low()[0] - 2 * m_border;
    float activeHeight = high()[1] - low()[1] - 2 * m_border;

    m_uv[0].clear();

    if(stretch)
      m_uv[1].make(activeWidth, activeHeight);

    if(activeWidth < activeHeight) {
      m_uv[1][0] = activeWidth / activeHeight;
      m_uv[1][1] = 1.0f;    
    }
    else {
      m_uv[1][0] = 1.0f;    
      m_uv[1][1] = activeHeight / activeWidth;
    }

    m_uv[1] *= VEE_max(activeWidth, activeHeight) / 
      VEE_min(activeWidth, activeHeight);
  }


  Vector2 TextureButton::textSize(GlData *gld)
  {
    if(!m_label.size())
      return Vector2(0.0, 0.0);

    float x1, x2, y1, y2;

    gld->m_font.getBBox(m_label.c_str(), m_fontSize, 0.0, &x1, &x2, &y1, &y2);

    return Vector2(x2 - x1, y2 - y1);
  }

  void TextureButton::newState(State, const InEvent *)
  {}


  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  CheckButton::CheckButton(Transformation * host)
    : TextureButton(host),
      m_checked(false)
  {}

  CheckButton::~CheckButton()
  {}


  void CheckButton::renderGl(GlData *gld, bool transform)
  {
    if(m_state == HIDDEN)
      return;

    pushLocalMatrix(transform);

    TextureButton::renderGl(gld, false);

    glColor4fv(VEE_Rgba::Red.data());

    float yspan = high()[1] - low()[1];
  
    float border = m_border * 0.5;

    Vector2 vlow(high()[0] - yspan + m_border, 
		       low()[1] + m_border), high;

    yspan -= 2.0 * m_border;

    high = vlow + yspan;

    renderFrameGl(vlow, high, border, 0.01);
  
    border *= 1.5f;
    vlow += border;
    high -= border;

    if(m_checked) {
      glColor4fv(VEE_Rgba::Red.data());
      drawCheck(vlow, high, 0.01);
    }

    popLocalMatrix(transform);
  }

  void CheckButton::update()
  {
    TextureButton::update();
  }

  QDomElement CheckButton::writeDom(QDomDocument &doc)
  {
    QDomElement e = TextureButton::writeDom(doc);
    
    addTextChild(e, doc, "checked", (int) m_checked);

    return e;
  }

  bool CheckButton::readDom(QDomElement &e)
  {
    if(!TextureButton::readDom(e))
      return false;

    m_checked = getInt(e, "checked", (int) m_checked);

    return true;
  }

  const char * CheckButton::type()
  {
    return "CheckButton";
  }

  void CheckButton::transmitValue()
  {
    MessageInt mi(m_checked);
    msgSend("checked", &mi);
  }

  void CheckButton::drawCheck(const Vector2 & low,
			      const Vector2 & high, float offset)
  {
    const int vertexCount = 6;

    static const Vector2 vertices[vertexCount] = {
      Vector2(0.0,  0.8),
      Vector2(0.0,  0.5),
      Vector2(0.44, 0.3),
      Vector2(0.44, 0.0),
      Vector2(0.75, 1.0),
      Vector2(1.0,  0.55)
    };

    glBegin(GL_TRIANGLE_STRIP);

    Vector2 span = high - low;
    Vector2 tmp, tmp2;
  
    tmp2 = vertices[0];
    tmp2.scale(span);
    tmp = low + tmp2;

    glVertex3f(tmp.x, tmp.y, offset);

    for(int i=1; i < vertexCount; i++) {
      tmp2 = vertices[i];
      tmp2.scale(span);
      tmp = low + tmp2;
    
      glVertex3f(tmp.x, tmp.y, offset);
    
    }

    glEnd();
  
  }

  void CheckButton::newState(State s, const InEvent *e)
  {
    if(s == GRAB) {
      m_checked = !m_checked;
      MessageInt mi(m_checked, e->time());

      msgSend("checked", &mi);
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  Slider::Slider(Transformation * host, 
		 float value, float minVal, float maxVal)
    : TextureButton(host),
      m_value(value),
      m_relative((value - minVal) / (maxVal - minVal)),
      m_min(minVal), 
      m_max(maxVal),
      m_displayValue(true),
      m_orientation(VERTICAL)
  {
    m_rotation.identity();
    m_location.clear();
    setValue(value);
    m_labelLoc = Widget::BELOW;
    m_cursorLoc.clear();
  }

  Slider::~Slider()
  {}

  void Slider::handleEvent(const InEvent *e)
  {
    if(m_state == HIDDEN)
      return;

    m_cursorLoc = applyMyTransform(e->location());

    TextureButton::handleEvent(e);

    if(m_state == GRAB) {

      if(m_trackHand != e->hand())
	return;

      float span, flow, y;

      if(m_orientation == VERTICAL) {
	span = high()[1] - low()[1];
	flow = low()[1];
	y = m_cursorLoc[1];    
	flow += span * 0.05;
	span *= 0.9;
      }
      else {
	span = high()[0] - low()[0];
	float draggerHeight = (high()[1] - low()[1]) * 0.45;
	float draggerWidth = VEE_min(0.05f * span, draggerHeight);
	float relSpan = 1.0 - 2.0 * draggerWidth / span;
	span *= relSpan;
	flow = low()[0] + draggerWidth;
	y = m_cursorLoc[0];
      }
    
      float relative = (y - flow) / span;
    
      m_relative = VEE_clamp(relative, 0.0f, 1.0f);
    
      MJ_Mapper * mapper = m_mapper.ptr();
      if(mapper) {
	
	double localMin = mapper->map(m_min);
	double localMax = mapper->map(m_max);

	m_value = mapper->unmap(m_relative * (localMax - localMin) + localMin);
      }
      else {
	m_value = (m_max - m_min) * m_relative + m_min;
      }

      MessageFloat mf(m_value, e->time());
      msgSend("value", &mf);

      /* VEE_trace("Slider::handleEvent # Value is %f (%f)", 
	 m_value, m_relative); */
    }
  }

  void Slider::renderGl(GlData *gld, bool transform)
  {
    if(m_state == HIDDEN)
      return;

    pushLocalMatrix(transform);

    TextureButton::renderGl(gld, false);

    float draggerHeight;
    float draggerWidth;
    float draggerY;
    float draggerX;

    if(m_orientation == VERTICAL) {
      float span = high()[1] - low()[1];

      draggerHeight = 0.05 * span;
      span *= 0.9;
      draggerWidth = (high()[0] - low()[0]) * 0.45;
      draggerY = m_relative * span + low()[1] + draggerHeight;
      draggerX = (high()[0] + low()[0]) * 0.5;
    }
    else {
      float span = high()[0] - low()[0];

      draggerHeight = (high()[1] - low()[1]) * 0.45;
      draggerWidth = VEE_min(0.05f * span, draggerHeight);
      float relSpan = 1.0 - 2.0 * draggerWidth / span;
      span *= relSpan;
      draggerX = m_relative * span + low()[0] + draggerWidth;
      draggerY = (high()[1] + low()[1]) * 0.5;
    }

    VEE_Rgba color = m_colors.color(colorState(m_state));
  
    glDisable(GL_TEXTURE_2D);

    glShadeModel(GL_FLAT);

    glBegin(GL_TRIANGLE_FAN);

    // Center:
    VEE_Rgba useColor = color.mulRgb(0.7);
    useColor.normalize();
    glColor4fv(useColor.data());
    glVertex3f(draggerX, draggerY, VEE_min(draggerHeight, draggerWidth));
  
    // Lower left:
    glVertex3f(draggerX - draggerWidth, draggerY - draggerHeight, 0.0);

    // Lower right:
    glVertex3f(draggerX + draggerWidth, draggerY - draggerHeight, 0.0);
  
    // Top right:
    useColor = color.mulRgb(0.8);
    useColor.normalize();
    glColor4fv(useColor.data());
    glVertex3f(draggerX + draggerWidth, draggerY + draggerHeight, 0.0);
  
    // Top left:
    useColor = color.mulRgb(1.3);
    useColor.normalize();
    glColor4fv(useColor.data());
    glVertex3f(draggerX - draggerWidth, draggerY + draggerHeight, 0.0);
  
    // Lower left (again):
    useColor = color.mulRgb(1.1);
    useColor.normalize();
    glColor4fv(useColor.data());
    glVertex3f(draggerX - draggerWidth, draggerY - draggerHeight, 0.0);
  
    glEnd();

    char buf[64];

    glColor4fv(VEE_Rgba::White.data());
    
    sprintf(buf, "%.3f", m_value);

    float span = high()[1] - low()[1];

    if(m_displayValue) {
    
      gld->m_fntRenderer.setPointSize(m_fontSize);
      gld->m_fntRenderer.start3f(low()[0], 
				 high()[1] + span * 0.2,
				 0.0);

      gld->m_fntRenderer.puts(buf);
    }

    if(m_state == GRAB) {
      // Draw it next to the slider mark (for easy visibility).
      gld->m_fntRenderer.start3f(draggerX + draggerWidth * 0.2, 
				 draggerY - span * 0.3,
				 draggerHeight);
      glColor4f(0.3, 0.3, 0.3, 1.0);
      quadAroundText(&gld->m_fntRenderer, buf, m_fontSize);
      glColor4fv(VEE_Rgba::White.data());
      gld->m_fntRenderer.puts(buf);    
    }

    popLocalMatrix(transform);  
  }

  void Slider::update()
  {
    if(m_displayValue) {
      // transmitSelectionSound(m_cursorLoc);
    }
    checkOrientation();
  }

  QDomElement Slider::writeDom(QDomDocument &doc)
  {
    QDomElement e = TextureButton::writeDom(doc);

    addTextChild(e, doc, "min", m_min);
    addTextChild(e, doc, "max", m_max);
    addTextChild(e, doc, "value", m_value);
    
    if(m_mapper.ptr()) {
      QDomElement mapdom = doc.createElement("mapper");
      mapdom.appendChild(m_mapper.ptr()->writeDom(doc));
      e.appendChild(mapdom);
    }

    return e;
  }

  bool Slider::readDom(QDomElement &e)
  {
    if(!TextureButton::readDom(e))
      return false;
    
    m_min   = getFloat(e, "min", m_min);
    m_max   = getFloat(e, "max", m_max);
    float value = getFloat(e, "value", m_value);
    
    QDomElement mapdom = e.namedItem("mapper").toElement();
    
    if(!mapdom.isNull()) {
      QDomElement mapdata = mapdom.firstChild().toElement();
      QString mapperName = mapdata.tagName();
      MJ_Mapper * m = MJ_Mapper::create(mapperName.ascii());
      if(m) {
	m->readDom(mapdata);
	m_mapper = m;
      }
    }

    setValue(value, false);

    return true;
  }

  const char * Slider::type()
  {
    return "Slider";
  }

  void Slider::transmitValue()
  {
    MessageFloat mf(m_value);
    msgSend("value", &mf);
  }

  void Slider::setValue(float v, bool transmit)
  {
    v = VEE_clamp(v, m_min, m_max);
    m_value = v;
    
    MJ_Mapper * mapper = m_mapper.ptr();
    if(mapper) {
      
      double localMin = mapper->map(m_min);
      double localMax = mapper->map(m_max);
      v = mapper->map(v);
      m_relative = (v - localMin) / (localMax - localMin);
    }
    else {
      m_relative = (v - m_min) / (m_max - m_min);
    }

    if(transmit)
      transmitValue();
  }

  void Slider::checkOrientation()
  {
    Vector3 span = m_bbox.span();
    m_orientation = span.x > span.y ? HORIZONTAL : VERTICAL;
  }

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  ColorCube::ColorCube(Transformation * host)
    : Widget(host),
      m_currentColor(0.4, 0.8, 0.7),
      m_needNewTextures(true),
      m_rebuildGlTextures(true)
  {
    m_bbox.set(0.0, 0.0, 0.0,
	       2.0, 2.0, 2.0);

    uint i, pixels = 128;

    for(i=0; i < WALLS; i++) {
      m_cubeWalls[i].resize(pixels, pixels, 3);
      m_cubeWalls[i].evenGray(0x7F);
    }

    generateColorTextures(m_currentColor);

    m_wallNames[BACK] = "colorcube/back";
    m_wallNames[LEFT] = "colorcube/left";
    m_wallNames[BOTTOM] = "colorcube/bottom";

    char buf[32];
    sprintf(buf, "-%p", this);

    for(i = 0; i < WALLS; i++)
      m_wallNames[i] += buf;
  }

  ColorCube::~ColorCube()
  {}

  void ColorCube::handleEvent(const InEvent *e)
  {  
    const char * fname = "ColorCube::handleEvent";

    //  VEE_trace(fname);
    if(m_state == HIDDEN)
      return;

    if(m_state == DISABLED)
      return;

    if((m_trackHand >= 0) && (m_trackHand != e->hand()))
      return;

    Vector3 loc3 = applyMyTransform(e->location());

    bool inside = m_bbox.contains(loc3);

    // Test if we should grab:  
    if((e->buttonState() & m_buttonMask) && inside) {
      if(m_state != GRAB) {
	m_state = GRAB;
	VEE_trace("%s # Grab %p", fname, this);
	MessageInt mi(m_state, e->time());
	msgSend("press", &mi);
	transmitSound(ColorGroup::ACTIVE, loc3);
	m_trackHand = e->hand();
      }
    }

    if(!e->buttonState() & m_buttonMask) {
      if(m_state == GRAB) {
	MessageInt mi(RELEASE, e->time());
	msgSend("release", &mi);
	VEE_trace("%s # Release %p", fname, this);
	transmitSound(ColorGroup::INACTIVE, loc3);
	m_trackHand = -1;
	m_state = inside ? ACTIVE : INACTIVE;
      }
    }

    if(m_state == GRAB) {

      Vector3 v = loc3 - m_bbox.low();
      v.descale(m_bbox.high() - m_bbox.low());
    
      m_currentColor.set(v.x, v.y, v.z, m_currentColor.alpha());

      m_currentColor.normalize();

      transmitValue();

      m_needNewTextures = true;
    }
  }

  void ColorCube::update()
  {
    if(m_needNewTextures) {
      generateColorTextures(m_currentColor);
      m_needNewTextures = false;
      m_rebuildGlTextures = true;
    }
    else {
      m_rebuildGlTextures = false;
    }
  }

  QDomElement ColorCube::writeDom(QDomDocument &doc)
  {
    QDomElement e = Widget::writeDom(doc);
    addTextChild(e, doc, "color", toStr(m_currentColor));
    return e;
  }
  
  bool ColorCube::readDom(QDomElement &e)
  {
    if(!Widget::readDom(e))
      return false;

    m_currentColor = toHsva(getString(e, "color", toStr(m_currentColor)));
    
    return true;
  }

  const char * ColorCube::type()
  {
    return "ColorCube";
  }

  void ColorCube::transmitValue()
  {
    char buf[128];
    
    sprintf(buf, "%f %f %f", m_currentColor.hue(), 
	    m_currentColor.saturation(), 
	    m_currentColor.value());
    
    MessageString ms(buf);
    msgSend("color", &ms);
  }

  void ColorCube::renderGl(GlData *gld, bool transform)
  {
    if(m_state == HIDDEN)
      return;

    pushLocalMatrix(transform);

    // Now lets render the color cube:

    VEE_Vector3 v1, v2, v3;

    glEnable(GL_TEXTURE_2D);
    glDepthMask(GL_TRUE);
    glEnable(GL_DEPTH_TEST);
    glDisable(GL_BLEND);

    glColor4fv(VEE_Rgba::White.data());

    v1 = m_bbox.low();
    v2 = m_bbox.high();

    // Back wall:
    buildGlTexture(gld, BACK);

    glBegin(GL_QUADS);
  
    glTexCoord2f(0.0, 0.0);
    glVertex3f(v1.x, v1.y, v1.z);

    glTexCoord2f(1.0, 0.0);
    glVertex3f(v2.x, v1.y, v1.z);

    glTexCoord2f(1.0, 1.0);
    glVertex3f(v2.x, v2.y, v1.z);

    glTexCoord2f(0.0, 1.0);
    glVertex3f(v1.x, v2.y, v1.z);

    glEnd();

    // Left wall:
    buildGlTexture(gld, LEFT);

    glBegin(GL_QUADS);
  
    glTexCoord2f(0.0, 0.0);
    glVertex3f(v1.x, v1.y, v1.z);

    glTexCoord2f(1.0, 0.0);
    glVertex3f(v1.x, v1.y, v2.z);

    glTexCoord2f(1.0, 1.0);
    glVertex3f(v1.x, v2.y, v2.z);

    glTexCoord2f(0.0, 1.0);
    glVertex3f(v1.x, v2.y, v1.z);

    glEnd();

    // Bottom wall:
    buildGlTexture(gld, BOTTOM);

    glBegin(GL_QUADS);
  
    glTexCoord2f(0.0, 1.0);
    glVertex3f(v1.x, v1.y, v2.z);

    glTexCoord2f(1.0, 1.0);
    glVertex3f(v2.x, v1.y, v2.z);

    glTexCoord2f(1.0, 0.0);
    glVertex3f(v2.x, v1.y, v1.z);

    glTexCoord2f(0.0,0.0);
    glVertex3f(v1.x, v1.y, v1.z);

    glEnd();

    // Render tthe ball that represents the location in the color cube

    // glEnable(GL_LIGHTING);
    glDisable(GL_TEXTURE_2D);
    // glDisable(GL_LIGHTING);

    glPushMatrix();
  
    v3 = m_bbox.high() - m_bbox.low();

    float radius = v3.x * 0.1;

    v3.scale(VEE_Vector3(m_currentColor.hue(), 
			 m_currentColor.saturation(), 
			 m_currentColor.value()));
    v2 = v3 + v1;
    glTranslatef(v2.x, v2.y, v2.z);


    // Now the axis marks
    glColor4fv(VEE_Rgba::Gray.data());

    float cylradius = radius * 0.2;

    GLUquadric * q = gld->m_quadric;

    gluQuadricDrawStyle(q, GLU_FILL);

    glPushMatrix(); // Axis transforms:
  
    glRotatef(180.0f,1.0f,0.0f,0.0f);
    gluCylinder(q, cylradius, cylradius, v3.z, 8, 1); //z axis
  
    glRotatef(-90.0f,1.0f,0.0f,0.0f);
    gluCylinder(q, cylradius, cylradius, v3.y, 8, 1); //y axis

    glRotatef(-90.0f,0.0f,1.0f,0.0f);
    gluCylinder(q, cylradius, cylradius, v3.x, 8, 1); //x axis

    glPopMatrix();  // End Axis transforms, back to the crossing of axes.

    VEE_Rgba rgba = m_currentColor.rgba();

    glColor4fv(rgba.data());

    glEnable(GL_BLEND);

    gluSphere(q, radius, 8, 8);

    glDisable(GL_BLEND);

    glPopMatrix();  // End color-cube
    
    popLocalMatrix(transform);
  }

  /* The HSV axis are: 

  Left-right: Hue
  Up-down:    Saturation
  Front-back: value
  */
  void ColorCube::generateColorTextures(const VEE_Hsva &hsva)
  {
    uint i, j, pixels = m_cubeWalls[0].w();

    float a = hsva.alpha();

    // Back wall, constant value
    float val = hsva.value();

    for(i=0; i < pixels; i++) {
      float h = (float) i / (float) (pixels - 1);

      for(j=0; j < pixels; j++) {
	float s = (float) j / (float) (pixels - 1);
      
	VEE_Hsva hsvPixel(h, s, val, a);
	VEE_Rgba rgba = hsvPixel.rgba();

	unsigned char * pixel = m_cubeWalls[BACK].pixel(i, j);
      
	pixel[0] = VEE_min((int) (rgba.red() * 255), 255);
	pixel[1] = VEE_min((int) (rgba.green() * 255), 255);
	pixel[2] = VEE_min((int) (rgba.blue() * 255), 255);
      }
    }
  
    // Bottom wall, constant saturation

    float sat = hsva.saturation();

    for(i=0; i < pixels; i++) {
      float v = (float) i / (float) (pixels - 1);

      for(j=0; j < pixels; j++) {
	float h = (float) j / (float) (pixels - 1);
      
	VEE_Hsva hsvPixel(h, sat, v, a);
	VEE_Rgba rgba = hsvPixel.rgba();

	unsigned char * pixel = m_cubeWalls[BOTTOM].pixel(j, i);
      
	pixel[0] = VEE_min((int) (rgba.red() * 255), 255);
	pixel[1] = VEE_min((int) (rgba.green() * 255), 255);
	pixel[2] = VEE_min((int) (rgba.blue() * 255), 255);
      }
    }

    // Left wall, constant hue

    float hue = hsva.hue();

    for(i=0; i < pixels; i++) {
      float s = (float) i / (float) (pixels - 1);

      for(j=0; j < pixels; j++) {
	float v = (float) j / (float) (pixels - 1);
      
	VEE_Hsva hsvPixel(hue, s, v, a);
	VEE_Rgba rgba = hsvPixel.rgba();

	unsigned char * pixel = m_cubeWalls[LEFT].pixel(j, i);
      
	pixel[0] = VEE_min((int) (rgba.red() * 256), 255);
	pixel[1] = VEE_min((int) (rgba.green() * 256), 255);
	pixel[2] = VEE_min((int) (rgba.blue() * 256), 255);
      }
    }

    // Now create a grid to the images to make things look 3D
    uint lineInterval = 32, lineWidth = 2;

    for(uint k=0; k < 3; k++) {
      for(i=0; i < pixels; i++) {
	for(j=0; j < pixels; j++) {

	  if(i % lineInterval < lineWidth ||
	     j % lineInterval < lineWidth) {
	    unsigned char * pixel = m_cubeWalls[k].pixel(i ,j);
	    pixel[0] = 0x7f;
	    pixel[1] = 0x7f;
	    pixel[2] = 0x7f;
	  }

	}
      }
    }
  }

  /** This method renews the current texture for given wall. As a side
      effect it leaves the current texture on. */

  void ColorCube::buildGlTexture(GlData * gld, uint i)
  {
    VEE_GlTexture2D * tex2d = gld->m_textures.getTexture(m_wallNames[i]);
  
    if(!tex2d) {
      VEE_GlTexture2D tmp;
      tmp.genTexture(& m_cubeWalls[i], true);
      gld->m_textures.addTexture(m_wallNames[i], tmp);
      VEE_trace("ParticleEditor::buildGlTexture # New texture (%u)",
		(uint) tmp.id());
    }
    else if(m_rebuildGlTextures) {
      tex2d->genTexture(& m_cubeWalls[i], true);
    }
    else {
      tex2d->use();
    }
  }


  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  ColorCylinder::ColorCylinder(Transformation * host)
    : Widget(host),
      m_currentColor(0.4, 0.8, 0.7),
      m_yellowExpand(0.4),
      m_needNewTextures(true),
      m_rebuildGlTextures(true),
      m_indicatorLights(false)
  {
    m_bbox.set(0.0, 0.0, 0.0,
	       2.0, 2.0, 2.0);

    uint i, pixels = 128;

    m_circle.resize(pixels, pixels, 4);
    m_bar.resize(pixels * 3);

    generateColorTextures(m_currentColor);

    m_texNames[CIRCLE] = "colorcylinder/circle";
    m_texNames[BAR] = "colorcylinder/bar";

    char buf[32];
    sprintf(buf, "-%p", this);

    for(i = 0; i < ELEMENTS; i++)
      m_texNames[i] += buf;
  }

  ColorCylinder::~ColorCylinder()
  {}

  static const float yellowNorm = 1.0f / 6.0f;

  inline float ColorCylinder::hueMap(float in)
  {
    if(in < yellowNorm)
      return in / yellowNorm * m_yellowExpand;
    
    return (1.0f - m_yellowExpand) * (in - yellowNorm) / (1.0f - yellowNorm) +
      m_yellowExpand;
  }
  
  inline float ColorCylinder::hueUnMap(float in)
  {
    if(in < m_yellowExpand)
      return (in / m_yellowExpand) * yellowNorm;

    return (1.0f - yellowNorm) * (in - m_yellowExpand) / 
      (1.0f - m_yellowExpand) + yellowNorm;
  }

  static const float valuePower = 0.7f;

  inline float ColorCylinder::valueMap(float in)
  {
    if(in < 0.0)
      return 0.0;

    return powf(in, valuePower);
  }

  inline float ColorCylinder::valueUnMap(float in)
  {
    if(in < 0.0)
      return 0.0;

    return powf(in, 1.0f / valuePower);
  }


  void ColorCylinder::handleEvent(const InEvent *e)
  {  
    const char * fname = "ColorCylinder::handleEvent";

    //  VEE_trace(fname);
    if(m_state == HIDDEN)
      return;

    if(m_state == DISABLED)
      return;

    if(m_trackHand >= 0 && m_trackHand != e->hand())
      return;

    Vector3 loc3 = applyMyTransform(e->location());

    bool inside = m_bbox.contains(loc3);

    // Test if we should grab:  
    if((e->buttonState() & m_buttonMask) && inside) {
      if(m_state != GRAB) {
	m_state = GRAB;
	VEE_trace("%s # Grab %p", fname, this);
	MessageInt mi(m_state, e->time());
	msgSend("press", &mi);
	transmitSound(ColorGroup::ACTIVE, loc3);
	m_trackHand = e->hand();
	Engine::engine()->setRenderHand(m_trackHand, false);
	
      }
    }

    if(!e->buttonState() & m_buttonMask) {
      if(m_state == GRAB) {
	MessageInt mi(RELEASE, e->time());
	msgSend("release", &mi);
	VEE_trace("%s # Release %p", fname, this);
	transmitSound(ColorGroup::INACTIVE, loc3);
	Engine::engine()->setRenderHand(m_trackHand, true);
	m_trackHand = -1;
	m_state = inside ? ACTIVE : INACTIVE;
      }
    }

    if(m_state == GRAB) {

      Vector3 v = loc3 - m_bbox.low();
      Vector3 span = m_bbox.span();

      if(span.x > span.z)
	span.x = span.z;
      else
	span.z = span.x;

      v.descale(span);

      v.x -= 0.5f;
      v.z -= 0.5f;

      float hue = (VEE_PI + atan2(v.x, v.z)) / VEE_TWO_PI;
      float sat = sqrtf(v.x * v.x + v.z * v.z) * 2.0f;

      m_currentColor.set(hueUnMap(hue), sat, 
			 valueUnMap(v.y), m_currentColor.alpha());

      m_currentColor.normalize();

      transmitValue();

      m_needNewTextures = true;
    }
  }

  void ColorCylinder::update()
  {
    if(m_needNewTextures) {
      generateColorTextures(m_currentColor);
      m_needNewTextures = false;
      m_rebuildGlTextures = true;
    }
    else {
      m_rebuildGlTextures = false;
    }
  }

  QDomElement ColorCylinder::writeDom(QDomDocument &doc)
  {
    QDomElement e = Widget::writeDom(doc);
    addTextChild(e, doc, "color", toStr(m_currentColor));
    addTextChild(e, doc, "lighting", (int) m_indicatorLights);
    return e;
  }
  
  bool ColorCylinder::readDom(QDomElement &e)
  {
    if(!Widget::readDom(e))
      return false;

    m_currentColor = toHsva(getString(e, "color", toStr(m_currentColor)));
    m_indicatorLights = getInt(e, "lighting", m_indicatorLights);
    
    return true;
  }

  const char * ColorCylinder::type()
  {
    return "ColorCylinder";
  }

  void ColorCylinder::transmitValue()
  {
    char buf[128];
    
    sprintf(buf, "%f %f %f", m_currentColor.hue(), 
	    m_currentColor.saturation(), 
	    m_currentColor.value());
    
    MessageString ms(buf);
    msgSend("color", &ms);
  }
  
  bool ColorCylinder::useExtraZ()
  {
    return true;
  }

  void ColorCylinder::renderGl(GlData *gld, bool transform)
  {
    if(m_state == HIDDEN)
      return;

    pushLocalMatrix(transform);

    // Now lets render the color plane:

    VEE_Vector3 v1, v2, v3;

    glEnable(GL_TEXTURE_2D);
    glDepthMask(GL_TRUE);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_ALPHA_TEST);
    glAlphaFunc(GL_GREATER, 0.5);
    glEnable(GL_BLEND);
    glDisable(GL_LIGHTING);

    glColor4fv(VEE_Rgba::White.data());

    v1 = m_bbox.low();
    v2 = m_bbox.high();
    v3 = v2 - v1;

    // Make it regular
    if(v3.x > v3.z)
      v2.x = v1.x + v3.z;
    else
      v2.z = v1.z + v3.x;

    // color plane:

    buildGlTexture(gld, CIRCLE);
    
    float v = m_currentColor.value();
    float h = v2.y * v + v1.y * (1.0f - v);

    glColor3f(1, 1, 1);

    glBegin(GL_QUADS);
  
    glTexCoord2f(0.0, 0.0);
    glVertex3f(v1.x, h, v1.z);

    glTexCoord2f(1.0, 0.0);
    glVertex3f(v2.x, h, v1.z);

    glTexCoord2f(1.0, 1.0);
    glVertex3f(v2.x, h, v2.z);

    glTexCoord2f(0.0, 1.0);
    glVertex3f(v1.x, h, v2.z);

    glEnd();

    glDisable(GL_ALPHA_TEST);
    glDisable(GL_TEXTURE_2D);

    // A little stick to show the selected color:

    float stickColor = (v > 0.5f) ? 0.0f : 1.0f;

    glColor3f(stickColor, stickColor, stickColor);
    
    glLineWidth(2);
    glBegin(GL_LINES);

    float angle = VEE_PI + hueMap(m_currentColor.hue()) * VEE_TWO_PI;
    float circleRadius = 0.5f * VEE_min(v3.x, v3.z);
    float radius = circleRadius * m_currentColor.saturation();
    float x = v1.x + circleRadius + sinf(angle) * radius;
    float z = v1.z + circleRadius + cosf(angle) * radius;

    glVertex3f(x, h - 0.1, z);
    glVertex3f(x, h + 0.1, z);

    glEnd();

    // Vertical bar:

    /*

    buildGlTexture(gld, BAR);
    


    glTranslatef(v1.x + radius + x, v1.y + radius + y, v1.z);
    glRotatef(90, 1, 0, 0);
    
    gluCylinder(q, 0.1, 0.1, v3.y, 6, 2);
    
    glDisable(GL_TEXTURE_1D);    */

    // Color Sphere

    glDisable(GL_TEXTURE_2D);

    glPushMatrix();

    VEE_Rgba rgba = m_currentColor.rgba();

    GLUquadric * q = gld->m_quadric;
    gluQuadricDrawStyle(q, GLU_FILL);

    if(m_indicatorLights) {
      glEnable(GL_LIGHTING);
      glEnable(GL_LIGHT0);
      glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, rgba.data());
      glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, rgba.data());
      glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, 1);
      gluQuadricNormals(q, GLU_SMOOTH);
    }

    glTranslatef(0.5 * (v1.x + v2.x), v2.y + circleRadius * 0.25, 0.0);
    glColor4fv(rgba.data());
    gluSphere(q, circleRadius * 0.2, 8, 8);

    if(m_indicatorLights)
      glDisable(GL_LIGHTING);

    glPopMatrix();


    popLocalMatrix(transform);
  }

  /* The HSV axis are: 

  Around the circle: Hue
  Edge-center:       Saturation
  Up-down:           Value
  */
  void ColorCylinder::generateColorTextures(const VEE_Hsva &hsva)
  {
    // First the plane texture:
    VEE_Hsva color(hsva);
    color.setValue(valueMap(color.value()));

    int x, w = m_circle.w();
    int y, h = m_circle.h();
    
    float scale = 2.0f / (float) w;

    for(x = 0; x < w; x++) {
      for(y = 0; y < h; y++) {

	float xd = x * scale - 1.0f;
	float yd = y * scale - 1.0f;

	float d = sqrtf(xd * xd + yd * yd);
	
	uchar * pixel = m_circle.pixel(x, y);

	if(d > 1.0f) {
	  * ((uint *) pixel) = 0;
	  continue;
	}

	float hue = (VEE_PI + atan2(xd, yd)) / VEE_TWO_PI;
	
	color.setHue(hueUnMap(hue));
	color.setSaturation(d);

	VEE_Rgba rgba = color.rgba();
	
	pixel[0] = VEE_min((int) (rgba.red() * 255), 255);
	pixel[1] = VEE_min((int) (rgba.green() * 255), 255);
	pixel[2] = VEE_min((int) (rgba.blue() * 255), 255);
	pixel[3] = 255;
      }
    }

    // Then comes the color bar texture:

    h = m_bar.size() / 3;

    color = hsva;

    scale = 1.0f / (float) h;

    for(y = 0; y < h; y++) {

      color.setValue(VEE_min((int) (y * scale), 255, 255));
      
      VEE_Rgba rgba = color.rgba();

      uchar * pixel = & m_bar[y * 3];	

      pixel[0] = VEE_min((int) (rgba.red() * 255), 255);
      pixel[1] = VEE_min((int) (rgba.green() * 255), 255);
      pixel[2] = VEE_min((int) (rgba.blue() * 255), 255);
    }
  }

  /** This method renews the current texture. As a side
      effect it leaves the current texture on. */

  void ColorCylinder::buildGlTexture(GlData * gld, uint i)
  {
    VEE_GlTexture2D * tex2d = gld->m_textures.buildTexture(m_texNames[i]);
  
    if(m_rebuildGlTextures) {
      if(i == CIRCLE)
	tex2d->genTexture(& m_circle, false);
      else
	tex2d->genTexture1D(&m_bar[0], m_bar.size() / 3, GL_RGB, true);
    }
    else if(i == CIRCLE)
      tex2d->use2D();
    else {
      glDisable(GL_TEXTURE_2D);
      tex2d->use1D();
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  CubeWidget::CubeWidget(Transformation * host)
    : Widget(host),
      m_border(0.1)
  {
  }

  CubeWidget::~CubeWidget()
  {}

  void CubeWidget::handleEvent(const InEvent *e)
  {
    const char * fname = "CubeWidget::handleEvent";

    // VEE_trace(fname);
    if(m_state == HIDDEN)
      return;

    if(m_state == DISABLED)
      return;

    bool inside;

    Vector3 loc3 = applyMyTransform(e->location());

    // Test is it inside:
    if(m_bbox.contains(loc3)) {
      if(m_state == INACTIVE) {
	m_state = ACTIVE;
	MessageInt mi(ACTIVE, e->time());
	msgSend("state", &mi);
      }
      inside = true;
    }
    else {
      inside = false;

      /* if(m_state != GRAB)
	 m_state = INACTIVE; */
    }

    // Test if we should grab:
    if(e->buttonState() & m_buttonMask) {
      if(m_state == ACTIVE) {
	m_state = GRAB;
	MessageInt mi(e->hand(), e->time());
	msgSend("press", &mi);
	VEE_trace("%s # Grab %p", fname, this);
	transmitSound(ColorGroup::ACTIVE, loc3);
      }
    }

    // Test if the grab should be released:
    else if(m_state == GRAB) {
      m_state = inside ? ACTIVE : INACTIVE;
      MessageInt mi(e->hand(), e->time());
      msgSend("release", &mi);
      VEE_trace("%s # Release %p", fname, this);
      transmitSound(ColorGroup::INACTIVE, loc3);
    }  
  }

  void CubeWidget::renderGl(GlData *gld, bool transform)
  {
    if(m_state == HIDDEN)
      return;

    pushLocalMatrix(transform);

    Vector2 low, high;

    Vector3 vlow = m_bbox.low();
    Vector3 span = m_bbox.span();

    low.make(0.0);
    high.make(span[0], span[1]);

    glPushMatrix();

    glEnable(GL_LIGHTING);

    // Front wall:
    glTranslatef(0.0, 0.0, span.z);
    glTranslatef(vlow[0], vlow[1], vlow[2]);
    renderFrameGl(low, high, m_border);
  
    // Back wall:
    glTranslatef(0.0, 0.0, -span.z);
    renderFrameGl(low, high, m_border);

    // Left Wall:
    glRotatef(-90, 0.0, 1.0, 0.0);
    high.make(span.z, span.y);
    renderFrameGl(low, high, m_border);

    // Right Wall:
    glTranslatef(0.0, 0.0, -span.x);
    renderFrameGl(low, high, m_border);

    // Bottom Wall:
    glRotatef(90, 1.0, 0.0, 0.0);
    high.make(span.z, span.x);
    renderFrameGl(low, high, m_border);

    // Top Wall:
    glTranslatef(0.0, 0.0, -span.y);
    renderFrameGl(low, high, m_border);

    glDisable(GL_LIGHTING);

    glPopMatrix();

    if(m_model.size()) {
      Solar::Model * model = gld->model(m_model);

      if(model) {
	const VEE_BoundingBox & bb = model->boundingBox();

	VEE_Vector3 modelCenter = bb.center();
	VEE_Vector3 modelSpan   = bb.span();
	VEE_Vector3 cubeCenter  = m_bbox.center();
	VEE_Vector3 cubeSpan    = m_bbox.span();
      
	cubeSpan.descale(modelSpan);

	float modelScale = cubeSpan.minimum() * 0.9;

	glPushMatrix();
	glPushAttrib(~0);

	glTranslatef(cubeCenter.x, cubeCenter.y, cubeCenter.z);
	glScalef(modelScale, modelScale, modelScale);
	glTranslatef(-modelCenter.x, -modelCenter.y, -modelCenter.z);
      
	gld->m_solarRenderer.prepareToRender();

	model->render(& gld->m_solarRenderer);

	glPopAttrib();
	glPopMatrix();
      }
    }

    popLocalMatrix(transform);
  }

  void CubeWidget::update()
  {}

  const char * CubeWidget::type()
  {
    return "CubeWidget";
  }

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  Menu::Item::Item()
    : m_check(NOCHECK)
  {}
  
  Menu::Item::~Item()
  {}

  void Menu::Item::select()
  {
    MessageInt mi(1);
    msgSend("select", &mi);
  }

  QDomElement Menu::Item::writeDom(QDomDocument &doc)
  {
    QDomElement e = doc.createElement("Item");

    addTextChild(e, doc, "label", m_label.c_str());
    addTextChild(e, doc, "check", (int) m_check);
    
    if(m_id.size())
      addTextChild(e, doc, "id", m_id.c_str());
    
    return e;
  }

  bool Menu::Item::readDom(QDomElement &e)
  {
    if(e.tagName() != "Item") {
      VEE_error("Menu::Item # %s != Item", e.tagName().ascii());
      return false;
    }
    
    m_label = getString(e, "label", "???").ascii();
    m_check = (Check) getInt(e, "check", NOCHECK);
    m_id = getString(e, "id").ascii();

    return m_label != "???";
  }

  ///////////////////////////////////////////
  ///////////////////////////////////////////

  Menu::Menu(Transformation * host)
    : TextureButton(host),
      m_onItem(-1),
      m_menuBelow(true),
      m_transmit(true),
      m_checkSelected(false)
  {
    m_entryGeometry.clear();
    m_cursorLoc.clear();
  }

  Menu::~Menu()
  {}

  void Menu::handleEvent(const InEvent *e)
  {
    if(m_state == HIDDEN)
      return;

    State oldState = m_state;

    TextureButton::handleEvent(e);

    if((m_trackHand >= 0) && (m_trackHand != e->hand()))
      return;
  
    Vector3 loc = applyMyTransform(e->location());
    m_cursorLoc = loc;

    if(m_state == GRAB) {
    
      m_onItem = -1;

      Vector2 corner = m_menuBelow ? low().vector2() :
	Vector2(low().x, high().y + 
		       m_entryGeometry[1] * m_items.size());

      for(int i=0; i < (int) m_items.size(); i++) {
	Vector2 leftBottom = corner;
	leftBottom[1] -= (i + 1) * m_entryGeometry[1];
	Vector2 rightTop = leftBottom + m_entryGeometry;

	if(loc.x > leftBottom.x && loc.x < rightTop.x &&
	   loc.y > leftBottom.y && loc.y < rightTop.y)
	  m_onItem = i;
      }
    
    }
    if(oldState == GRAB && m_state != GRAB && m_onItem >= 0) {
      /* Now it is selection time (!) */

      Item * item = m_items[m_onItem].ptr();
    
      item->select();


      updateChecks();

      m_onItem = -1;
      transmitSound(ColorGroup::GRAB, m_cursorLoc);
      if(item->id().size()) {
	MessageString ms(item->id());
	msgSend("select", &ms);
      }
      m_lastId = item->id();
    }
  }

  void Menu::renderGl(GlData *gld, bool transform)
  {
    if(m_state == HIDDEN)
      return;

    pushLocalMatrix(transform);

    TextureButton::renderGl(gld, false);

    if(m_state != GRAB) {
      popLocalMatrix(transform);
      return;
    }

    Vector2 textGeometry  = determineEntrySize(gld);
    Vector2 entryGeometry;
    entryGeometry.y  = textGeometry.y * 1.1;
    entryGeometry.x  = textGeometry.x;
    Vector2 textOffset = textGeometry * 0.05;
    if(textOffset.x < entryGeometry.y * 0.5)
      textOffset.x = entryGeometry.y * 0.5;

    entryGeometry.x += textOffset.x * 2.0f;

    m_entryGeometry = entryGeometry;

    Vector2 corner = m_menuBelow ? low().vector2() :
      Vector2(low().x, high().y + 
		     entryGeometry[1] * m_items.size());

    VEE_Rgba color, offset(0.15, 0.15, 0.15, 0.0, 0.0);

    glShadeModel(GL_SMOOTH);

    for(int i=0; i < (int) m_items.size(); i++) {

      Vector2 leftBottom = corner;
      leftBottom[1] -= (i + 1) * entryGeometry[1];
      Vector2 rightTop = leftBottom + entryGeometry;

      glDisable(GL_BLEND);
      glDisable(GL_TEXTURE_2D);
      if(i != m_onItem)
	color = m_colors.color(ColorGroup::INACTIVE);
      else
	color = m_colors.color(ColorGroup::ACTIVE);

      glBegin(GL_QUADS);
      glColor4fv(color.data());
      glVertex3f(leftBottom.x, leftBottom.y, 0.02);
      glColor4fv((color + offset).data());
      glVertex3f(rightTop.x, leftBottom.y, 0.02);
      glColor4fv(color.data());
      glVertex3f(rightTop.x, rightTop.y, 0.02);
      glColor4fv((color - offset).data());
      glVertex3f(leftBottom.x, rightTop.y, 0.02);
      glEnd();

      if(m_items[i].ptr()->isChecked()) {
	glColor4fv(VEE_Rgba::Red.data());

	float margin = textGeometry.y * 0.1;
	CheckButton::drawCheck
	  (Vector2(leftBottom.x + margin, leftBottom.y + margin), 
	   Vector2(leftBottom.x + textOffset.x - margin, 
			  rightTop.y - margin), 0.035);
	// VEE_trace("Menu::renderGl # Rendered a check %d", i);
      
      }

      glEnable(GL_BLEND);
      glColor4f(0.0, 0.0, 0.0, 1.0);
      glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

      leftBottom += textOffset;

      gld->m_fntRenderer.start3f(leftBottom.x, leftBottom.y, 0.03);
      gld->m_fntRenderer.setPointSize(m_fontSize);
      gld->m_fntRenderer.puts(m_items[i].ptr()->label().c_str());
    }

    popLocalMatrix(transform);
  }

  void Menu::update()
  {}

  QDomElement Menu::writeDom(QDomDocument &doc)
  {
    QDomElement e = TextureButton::writeDom(doc);
    
    if(m_lastId.size())
      addTextChild(e, doc, "last_id", m_lastId.c_str());

    addTextChild(e, doc, "checkselected", (int) m_checkSelected);

    QDomElement items = doc.createElement("items");

    for(uint i=0; i < m_items.size(); i++)
      items.appendChild(m_items[i].ptr()->writeDom(doc));
    
    e.appendChild(items);
    
    return e;
  }

  bool Menu::readDom(QDomElement &e)
  {
    if(!TextureButton::readDom(e))
      return false;

    m_lastId = getString(e, "last_id").ascii();    
    m_checkSelected = getInt(e, "checkselected", (int) m_checkSelected);

    QDomElement items = e.namedItem("items").toElement();
        
    QDomNode n = items.firstChild();
    
    while(!n.isNull()) {

      if ( n.isElement() ) {
	QDomElement ne = n.toElement();
	Item tmp;
	if(tmp.readDom(ne))
	  addItem(new Item(tmp));
      }

      n = n.nextSibling();
    }

    updateChecks();

    return true;
  }

  const char * Menu::type()
  {
    return "Menu";
  }
  
  void Menu::transmitValue()
  {
    if(m_lastId.size() && m_transmit) {
      MessageString ms(m_lastId);
      msgSend("select", &ms);
    }
  }

  int Menu::addItem(Item *item)
  {
    int r = m_items.size();
    m_items.push_back(item);
    return r;
  }

  int Menu::addItem
  (const char *label, const char *sinkId, MessageSink *sink)
  {
    Item * it = new Item();

    if(label)
      it->setLabel(label);
    it->msgAddSink("select", sinkId, sink);

    return addItem(it);
  }

  void Menu::clear()
  {
    m_items.clear();
    m_onItem = -1;

    msgClearSinks();
  }

  void Menu::updateChecks()
  {
    if(m_checkSelected)
      for(uint i = 0; i < m_items.size(); i++)
	m_items[i].ptr()->setChecked(i == (uint) m_onItem);
  }

  Vector2 Menu::determineEntrySize(GlData *gld)
  {
    float w = 0.0, h = 0.0;
    float x1, x2, y1, y2;

    for(uint i=0; i < m_items.size(); i++) {
      gld->m_font.getBBox(m_items[i].ptr()->label().c_str(), 
			  m_fontSize, 0.0,
			  &x1, &x2, &y1, &y2);
      float wt = x2 - x1;
      float ht = y2 - y1;

      if(w < wt)
	w = wt;
      if(h < ht)
	h = ht;
    }

    return Vector2(w, h);
  }

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  FileMenu::FileMenu(Transformation * host)
    : Menu(host),
      m_write(false)
  {
    m_transmit = false;
  }
  
  FileMenu::~FileMenu()
  {}

  QDomElement FileMenu::writeDom(QDomDocument &doc)
  {
    QDomElement e = Menu::writeDom(doc);

    addTextChild(e, doc, "write", (int) m_write);

    return e;
  }

  bool FileMenu::readDom(QDomElement &e)
  {
    if(!Menu::readDom(e))
      return false;

    m_write = getInt(e, "write", m_write);
    
    return true;
  }
  
  const char * FileMenu::type()
  {
    return "FileMenu";
  }


  void FileMenu::newState(State s, const InEvent *)
  {
    const char * fname = "FileMenu::newState";
    
    if(s != GRAB)
      return;

    m_items.clear();

    if(m_write) {

      QString name;
      QString name2;
      
      for(int i = 1; i < 1000; i++) {

	name.sprintf("slot-%d", i);
	name2.sprintf("slot-%d.gee", i);

	QFileInfo info(name2);

	if(!info.exists()) {
	  Item * item = new Item;

	  item->setLabel((QString("Empty: ") + name).ascii());
	  item->setId(name.ascii());

	  addItem(item);
	  
	  break;
	}
      }
    }

    // Fill the menu with all existing files:
    
    QDir dir(".");
    
    QFileInfoList filelist = * dir.entryInfoList("*.gee");

    for(uint i = 0; i < filelist.count(); i++) {

      const QFileInfo * info = filelist.at(i);

      if(m_write && !info->isWritable())
	continue;

      Item * item = new Item;
      
      std::string name = info->baseName().ascii();

      item->setLabel(name);
      item->setId(name);

      addItem(item);

      VEE_trace("%s # Added file %s (%s)", fname,
		item->label().c_str(), item->id().c_str());
    }
  }
  
  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  bool GrabGroup::grab(void *grabber) 
  { 
    if(m_grabbed) {
      VEE_error("GrabGroup::grab # Already grabbed");
      return false;
    }

    VEE_trace("GrabGroup::grab # %p", grabber);

    m_grabbed = grabber; 
    return true;
  }

  bool GrabGroup::ungrab(void *grabber) 
  { 
    if(grabber != m_grabbed) {
      VEE_error("GrabGroup::ungrab # %p != %p", grabber, m_grabbed);
      return false;
    }

    VEE_trace("GrabGroup::ungrab # %p", grabber);

    m_grabbed = 0; 
    return true;
  }

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  Widgets::Widgets(GrabGroup * group)
    : m_grabWidget(-1),
      m_grabs(group)
  {}

  Widgets::~Widgets()
  {
    clear();
  }

  void Widgets::clear(bool alsoDelete)
  {
    for(int i = ((int) m_widgets.size()) - 1; i >= 0 && alsoDelete; i--)
      delete m_widgets[i];

    m_widgets.clear();
  }

  void Widgets::beforeEvents()
  {
    for(uint i = 0; i < m_widgets.size(); i++)
      m_widgets[i]->beforeEvents();
  }

  void Widgets::handleEvent(const InEvent *e)
  {

    if(m_grabs) {
      if(m_grabs->someOneElsesGrab(this))
	return;
    }

    if(m_grabWidget >= 0) {

      if((uint) m_grabWidget < m_widgets.size()) {

	Widget * w = m_widgets[m_grabWidget];
	w->handleEvent(e);
	Widget::State s = w->state();

	if(s != Widget::GRAB)
	  m_grabWidget = -1;
      }
      else
	m_grabWidget = -1;

      if(m_grabs && (m_grabWidget < 0))
	if(!m_grabs->ungrab(this))
	  VEE_error("Widgets::handleEvent # ungrab failed");
    }
  
    for(uint i = 0; (i < m_widgets.size()) && (m_grabWidget < 0); i++) {

      Widget * w = m_widgets[i];

      if(w->state() != Widget::HIDDEN)
	w->handleEvent(e);

      if(w->state() == Widget::GRAB) {
	m_grabWidget = i;

	if(m_grabs)
	  m_grabs->grab(this);

	return;
      }
    }
    
    /* 
       VEE_trace("Widgets::handleEvent # gw = %d type = %s", 
       m_grabWidget, 
       m_grabWidget >= 0 ? 
       typeid(*m_widgets[m_grabWidget]).name() : "NONE");
    */
  }

  void Widgets::update()
  {
    for(uint i=0; i < m_widgets.size(); i++)
      m_widgets[i]->update();  
  }

  void Widgets::renderGl(GlData *gld)
  {
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    for(uint i=0; i < m_widgets.size(); i++)
      m_widgets[i]->renderGl(gld);
  }

  void Widgets::setStateAll(Widget::State s)
  {
    for(uint i=0; i < m_widgets.size(); i++)
      m_widgets[i]->setState(s);
  }

  void Widgets::forget(Widget *w)
  {
    iterator it = 
      std::find(m_widgets.begin(), m_widgets.end(), w);

    if(it != m_widgets.end())
      m_widgets.erase(it);
  }

  void Widgets::calculateBoundingBox(VEE_BoundingBox *bbox)
  {
    if(!m_widgets.size())
      bbox->clear();
    else {
      *bbox = m_widgets[0]->boundingBox();

      for(uint i = 1; i < m_widgets.size(); i++)
	bbox->expand(m_widgets[i]->boundingBox());
    }
  }

}
