/* -*- C++ -*- */

/* -*- C++ -*- */

/* -*- C++ -*- */

/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Crystal animation and 3D modeling application
 *
 * Authors:   Tommi Ilmonen 2001-2004
 * Copyright: Tommi Ilmonen and/or Tekes 2001-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#ifndef CRYSTAL_WIDGET_H
#define CRYSTAL_WIDGET_H

#include <crystal_transformation.h>

#include <Fluid/Matrix3.hpp>
#include <Fluid/MessageSource.hpp>
#include <Fluid/TimeStamp.hpp>
#include <Fluid/Vector2.hpp>
#include <Fluid/Vector3.hpp>

#include <vee_bounding_box.h>
#include <vee_clonable_pointer.h>
#include <vee_color.h>
#include <vee_image.h>
#include <vee_reference_object.h>

#include <mj_mapper.h>

#include <Solar/Garbage.hpp>

#include <vector>

class fntRenderer;

namespace Crystal { 

  enum {
    DEFAULT_MASK = 0x1 | 0x8 | 0x10
  };

  class GlData;
  class Widgets;

  class InEvent
  {
  public:
    InEvent() : m_buttonState(0), m_prevButtonState(0), m_hand(RIGHT_HAND)
    { m_location.clear(); m_rotation.identity(); }

    const Vector3 & location() const { return m_location; }
    void setLocation(const Vector3 &loc) { m_location = loc; }

    const Matrix3 & rotation() const { return m_rotation; }
    void setRotation(const Matrix3 &r) { m_rotation = r; }

    int           buttonState() const { return m_buttonState; }
    int           prevButtonState() const { return m_prevButtonState; }
    int           hand() const { return m_hand; }
    TimeStamp     time() const { return m_time; }

    void set(const Vector3 & location,
	     const Matrix3 & rotation,
	     int            buttonState,
	     int            prevButtonState,
	     int            hand,
	     const TimeStamp & time)
    { 
      m_location = location; 
      m_rotation = rotation; 
      m_buttonState = buttonState; 
      m_prevButtonState = prevButtonState; 
      m_hand = hand;
      m_time = time; 
    }
  
  protected:

    Vector3   m_location;
    Matrix3   m_rotation;
    int       m_buttonState;
    int       m_prevButtonState;
    int       m_hand;
    TimeStamp m_time;
  };

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  class ColorGroup
  {
  public:

    enum State {
      INACTIVE,
      ACTIVE,
      GRAB,
      STATE_SIZEOF
    };

    ColorGroup();
    ~ColorGroup();

    void constantColor( const VEE_Rgba &c) 
    { for(uint i=0; i < STATE_SIZEOF; i++) m_colors[i] = c; }
    void setColor(State s, const VEE_Rgba &c) { m_colors[s] = c; }
    const VEE_Rgba & color(State s) const { return m_colors[s]; }

  protected:
    VEE_Rgba m_colors[STATE_SIZEOF];
  };

  class SoundGroup
  {
  public:
    SoundGroup();
    ~SoundGroup();

    void setSound(ColorGroup::State s, const std::string &sound)
    { m_sounds[s] = sound; }
    const std::string & sound(ColorGroup::State s) const
    { return m_sounds[s]; }

    void setMasterVolume(float volume) { m_masterVolume = volume; }
    float masterVolume() const { return m_masterVolume; }
  protected:
    std::string m_sounds[ColorGroup::STATE_SIZEOF];
    float m_masterVolume;
  };

  typedef VEE_ReferenceObject<SoundGroup> SoundGroupRef;

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  class Widget : public MessageSource,
		 public Transformation,
		 public Solar::Garbage
  {
  public:
    enum State {
      /// Widget is hidden
      HIDDEN,
      /// Widget is not active
      INACTIVE,
      /// Widget is disabled
      DISABLED,
      /// Widget is active (cursor is close)
      ACTIVE,
      /// Widget grabs the cursor
      GRAB,
      /// Widget releases the cursor
      RELEASE,
      /// The number of states
      STATE_SIZEOF
    };

    enum Location {
      CENTER,
      LEFT,
      RIGHT,
      ABOVE,
      BELOW
    };

    Widget(Transformation * host = 0);
    virtual ~Widget();

    virtual void beforeEvents();
    virtual void handleEvent(const InEvent *);
    virtual void renderGl(GlData *, bool transform = true);
    virtual void update();
    virtual QDomElement writeDom(QDomDocument &doc);
    virtual bool readDom(QDomElement &elem);
    virtual const char * type() = 0;
    virtual void transmitValue();
    virtual bool useExtraZ();

    void renderFrameGl(Vector2 low,
		       Vector2 high,
		       float border, float offset = 0.0);
    static void frameGl(Vector2 low,
			Vector2 high,
			float border, float offset = 0.0);

    State state() const { return m_state; }
    void setState(State s) { m_state = s; }

    void setColors(const ColorGroup &c) { m_colors = c; }

    static ColorGroup::State colorState(State);

    void transmitSound(ColorGroup::State,
		       const Vector3 &) const;
  
    SoundGroupRef sounds() { return m_sounds; }
    void setSounds(const SoundGroupRef & s) { m_sounds = s; }

    float                 fontSize() const { return m_fontSize; }
    void                  setFontSize(float size) { m_fontSize = size; }

    void setButtonMask(int mask) { m_buttonMask = mask; }

    void setWidgetManager(Widgets *wm) { m_widgetManager = wm; }

    static void quadAroundText(fntRenderer *, 
			       const char * text, 
			       float size,
			       float border = 0.1);
    
    static Widget * create(const char * type);

    static const char * locationName(Location);
    static Location locationId(const char *name);

    void setBoundingBox(const VEE_BoundingBox &bb) { m_bbox = bb; }
    VEE_BoundingBox & boundingBox() { return m_bbox; }
    Vector3 & low()  { return m_bbox.low(); }
    Vector3 & high() { return m_bbox.high(); }
    void move(const VEE_Vector3 &translation) { m_bbox.move(translation); }

  protected:

    void pushLocalMatrix(bool push);
    void popLocalMatrix(bool pop);

    ColorGroup    m_colors;
    SoundGroupRef m_sounds;
    State                 m_state;
    float                 m_fontSize;
    int                   m_buttonMask;
    int                   m_trackHand;
    Widgets              *m_widgetManager;
    VEE_BoundingBox       m_bbox;
  };

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  class TextureButton : public Widget
  {
  public:

    TextureButton(Transformation * host = 0);
    virtual ~TextureButton();

    virtual void handleEvent(const InEvent *);
    virtual void renderGl(GlData *, bool transform = true);
    virtual QDomElement writeDom(QDomDocument &doc);
    virtual bool readDom(QDomElement &elem);
    virtual const char * type();

    void autoUv(bool stretch = true);
    void setUv(const Vector2 & low,
	       const Vector2 & high)
    { m_uv[0] = low; m_uv[1] = high; }
    void scaleUv(float s) { m_uv[0] *= s; m_uv[1] *= s; }

    void setArea(const Vector2 & low,
		 const Vector2 & size, bool autouv = true)
    { 
      Vector2 high = low + size;

      m_bbox.low().make(low.x, low.y, 0.0);
      m_bbox.high().make(high.x, high.y, 0.0);

      if(autouv) autoUv(true);
    }

    void setVertices(const Vector2 & low,
		     const Vector2 & high, bool autouv = true)
    { 
      m_bbox.low().make(low.x, low.y, 0.0);
      m_bbox.high().make(high.x, high.y, 0.0);

      if(autouv) autoUv(true);
    }

    void setSize(float w, float h, bool autouv = true)
    {
      m_bbox.high().x = m_bbox.low().x + w;
      m_bbox.high().y = m_bbox.low().y + h;

      if(autouv)
	autoUv(true);
    }

    void setTexture(const char * tex) { m_texture = tex; }

    void setLabel(const std::string & label) { m_label = label; }
    const std::string label() const { return m_label; }

    float border() const { return m_border; }
    void setBorder(float b) { m_border = b; }

    Vector2 textSize(GlData *);

  protected:

    virtual void newState(State s, const InEvent *e);

    std::string    m_texture;
    std::string    m_label;
    Widget::Location m_labelLoc;

    Vector2 m_uv[2];
    float          m_border;
    float          m_activeDistance;
  };

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  class CheckButton : public TextureButton
  {
  public:
    CheckButton(Transformation * host = 0);
    virtual ~CheckButton();

    virtual void renderGl(GlData *, bool transform = true);
    virtual void update();
    virtual QDomElement writeDom(QDomDocument &doc);
    virtual bool readDom(QDomElement &elem);
    virtual const char * type();

    virtual void transmitValue();

    static void drawCheck(const Vector2 & low,
			  const Vector2 & high,
			  float offset = 0.0);
  
    void setChecked(bool checked) { m_checked = checked; }

  protected:

    virtual void newState(State s, const InEvent *e);

    bool m_checked;
  };


  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  class Slider : public TextureButton
  {
  public:

    enum Orientation {
      VERTICAL,
      HORIZONTAL
    };

    Slider(Transformation * host = 0,
	   float value = 0.0, float minVal = 0.0, float maxVal = 1.0);
    virtual ~Slider();

    virtual void handleEvent(const InEvent *);
    virtual void renderGl(GlData *, bool transform = true);
    virtual void update();
    virtual QDomElement writeDom(QDomDocument &doc);
    virtual bool readDom(QDomElement &elem);
    virtual const char * type();
    virtual void transmitValue();

    void setDisplayValue(bool b) { m_displayValue = b; }
    void setValue(float v, bool transmit = false);
    float value() const { return m_value;}

    Orientation orientation() const { return m_orientation; }
    void setOrientation(Orientation o) { m_orientation = o; }

    void setMapper(MJ_Mapper *m) { m_mapper = m; }

  protected:

    void checkOrientation();

    VEE_Vector3 m_cursorLoc;
    float m_value, m_relative;
    float m_min, m_max;
    bool  m_displayValue;
    Orientation m_orientation;

    VEE_ClonablePointer<MJ_Mapper> m_mapper;
  };

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  class ColorCube : public Widget
  {
  public:

    enum Wall {
      BACK,
      LEFT,
      BOTTOM,
      WALLS
    };

    ColorCube(Transformation * host = 0);
    virtual ~ColorCube();

    virtual void handleEvent(const InEvent *);
    virtual void renderGl(GlData *, bool transform = true);
    virtual void update();
    virtual QDomElement writeDom(QDomDocument &doc);
    virtual bool readDom(QDomElement &elem);
    virtual const char * type();
    virtual void transmitValue();

    void generateColorTextures(const VEE_Hsva &);

    void setColor(const VEE_Hsva &color)
    {
      m_currentColor = color;
      m_needNewTextures = true;
    }

    void setAlpha(float a) 
    {
      m_currentColor.setAlpha(a);
    }
    void setTheta(float a) 
    {
      m_currentColor.setTheta(a);
    }

  protected:

    void buildGlTexture(GlData *, uint);

    VEE_Hsva        m_currentColor;

    std::string m_wallNames[ColorCube::WALLS];

    VEE_Image       m_cubeWalls[WALLS];
    bool            m_needNewTextures;
    bool            m_rebuildGlTextures;
  };

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  class ColorCylinder : public Widget
  {
  public:

    enum Element {
      CIRCLE,
      BAR,
      ELEMENTS
    };

    ColorCylinder(Transformation * host = 0);
    virtual ~ColorCylinder();

    virtual void handleEvent(const InEvent *);
    virtual void renderGl(GlData *, bool transform = true);
    virtual void update();
    virtual QDomElement writeDom(QDomDocument &doc);
    virtual bool readDom(QDomElement &elem);
    virtual const char * type();
    virtual void transmitValue();
    virtual bool useExtraZ();

    void generateColorTextures(const VEE_Hsva &);

    void setColor(const VEE_Hsva &color)
    {
      m_currentColor = color;
      m_needNewTextures = true;
    }

    void setAlpha(float a) 
    {
      m_currentColor.setAlpha(a);
    }

    void setTheta(float a) 
    {
      m_currentColor.setTheta(a);
    }

  protected:

    void buildGlTexture(GlData *, uint);

    inline float hueMap(float in);
    inline float hueUnMap(float in);
    inline float valueMap(float in);
    inline float valueUnMap(float in);

    VEE_Hsva        m_currentColor;

    std::string m_texNames[2];

    VEE_Image          m_circle;
    std::vector<uchar> m_bar;

    float           m_yellowExpand;

    bool            m_needNewTextures;
    bool            m_rebuildGlTextures;
    bool            m_indicatorLights;
  };

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////


  class CubeWidget : public Widget
  {
  public:

    CubeWidget(Transformation * host = 0);
    virtual ~CubeWidget();

    virtual void handleEvent(const InEvent *);
    virtual void renderGl(GlData *, bool transform = true);
    virtual void update();
    virtual const char * type();

    void setLabel(const std::string & label) { m_label = label; }

    void setBorder(float border) { m_border = border; }

    void setModel(const std::string & model) 
    { m_model = model; }

  protected:

    std::string     m_label;

    float           m_border;
    std::string     m_model;
  };

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  class Menu : public TextureButton
  {
  public:

    class Item : public MessageSource
    {
    public:
      enum Check {
	NOCHECK,
	UNCHECK,
	CHECK
      };

      Item();
      ~Item();

      void select();

      void setLabel(const std::string & label) { m_label = label; }
      const std::string & label() const { return m_label; }

      void setId(const std::string & id) { m_id = id; }
      const std::string & id() const { return m_id; }

      void setChecked(bool b)
      { m_check = b ? CHECK : UNCHECK; }

      bool isChecked() const { return m_check == CHECK; }

      QDomElement writeDom(QDomDocument &doc);
      bool readDom(QDomElement &elem);

    protected:

      std::string m_label;
      std::string m_id;
      Check       m_check;
    };
  

    Menu(Transformation * host = 0);
    virtual ~Menu();

    virtual void handleEvent(const InEvent *);
    virtual void renderGl(GlData *, bool transform = true);
    virtual void update();
    virtual QDomElement writeDom(QDomDocument &doc);
    virtual bool readDom(QDomElement &elem);
    virtual const char * type();
    virtual void transmitValue();

    int addItem(Item *);
    int addItem(const char *label, const char *sinkId, MessageSink *);

    void setMenuBelow(bool menuBelow)
    { m_menuBelow = menuBelow; }

    void clear();

    void setChecked(int i, bool b) { m_items[i].ptr()->setChecked(b); }

  public:

    void updateChecks();

    Vector2 determineEntrySize(GlData *);

    std::vector<VEE_RefPtr<Item> > m_items;
    int                 m_onItem;
    Vector2      m_entryGeometry;
    Vector3      m_cursorLoc;

    bool         m_menuBelow;
    std::string  m_lastId;

    bool         m_transmit;
    bool         m_checkSelected;
  };

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  class FileMenu : public Menu
  {
  public:
    FileMenu(Transformation * host = 0);
    virtual ~FileMenu();

    virtual QDomElement writeDom(QDomDocument &doc);
    virtual bool readDom(QDomElement &elem);
    virtual const char * type();

  protected:

    virtual void newState(State s, const InEvent *e);
    
    bool m_write;
    
  };

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  class GrabGroup
  {
  public:
    GrabGroup() : m_grabbed(0) {}
    ~GrabGroup() {}
  
    bool myGrab(void * caller) const { return m_grabbed == caller; }

    bool someOneElsesGrab(void * caller) const 
    { return (m_grabbed != caller) && m_grabbed; }

    bool isGrabbed() const { return m_grabbed != 0; }

    bool grab(void *grabber);
    bool ungrab(void *grabber);

  protected:
    void * m_grabbed;
  };

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  class Widgets 
  {
  public:
    typedef std::vector<Widget *> container;
    typedef container::iterator iterator;

    Widgets(GrabGroup * group = 0);
    ~Widgets();

    void addWidget(Widget *w) 
    { m_widgets.push_back(w); w->setWidgetManager(this); }

    void clear(bool alsoDelete = true);

    void beforeEvents();
    void handleEvent(const InEvent *);
    void update();
    void renderGl(GlData *);

    void setStateAll(Widget::State);
    void disableAll() { setStateAll(Widget::DISABLED); }
    void enableAll()  { setStateAll(Widget::INACTIVE); }

    void setGrabGroup(GrabGroup *g) { m_grabs = g; }

    void forget(Widget *);

    GrabGroup * grabs() { return m_grabs; }

    void calculateBoundingBox(VEE_BoundingBox *bbox);

    uint size() const { return m_widgets.size(); }
    
  protected:
    container m_widgets;
    int m_grabWidget;

    GrabGroup * m_grabs;
  };

}

#endif
