/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Crystal animation and 3D modeling application
 *
 * Authors:   Tommi Ilmonen 2001-2004
 * Copyright: Tommi Ilmonen and/or Tekes 2001-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <crystal_vrsmuggler.h>

#include <crystal_engine.h>
#include <crystal_gl_data.h>

#include <vee_error.h>

#include <Fluid/Config.hpp>

#include <iostream>
#include <unistd.h>

namespace Crystal { 
  
  static DI_MutexAuto _smuggler_mutex;

  VrSmuggler::VrSmuggler(Engine *engine)
    : m_sampleGl(0),
      m_engine(engine)
  {
    VEE_GlRenderer::createPrivateTextures(false);
    m_statistics = fopen("statistics.txt", "w");
  }

  VrSmuggler::~VrSmuggler()
  {
    if(m_statistics)
      fclose(m_statistics);
  }
  
  void VrSmuggler::init()
  {
    const char * fname = "VrSmuggler::init";

    VEE_trace(fname);

    m_engine->init();

    m_frameCount = 0;
    m_lastFrameCount = 0;
    m_lastFrameTime.getTimeOfDay();  

    
    m_updateUSCount = 0;
    m_renderUSCount = 0;
  }

  void VrSmuggler::preFrame()
  {
    m_frameCount++;

    VEE_Time now;
    now.getTimeOfDay();
    
    m_renderUSCount += m_lastFrameEnd.usecsTo(now);

    double sinceFrameUpdate = m_lastFrameTime.secsTo(now);

    if(sinceFrameUpdate >  5.0) {

      long maxRenderTime = 0;
      uint index = 0;

      for(uint i = 0; i < GlData::m_instances.size(); i++) {

	long t = GlData::m_instances[i]->m_timings[0];
	for(uint j = 1;j < 4; j++)
	  t += GlData::m_instances[i]->m_timings[j];

	if(maxRenderTime < t) {
	  maxRenderTime = t;
	  index = i;
	}
      }

      m_sampleGl = GlData::m_instances[index];

      double rate = (double) (m_frameCount - m_lastFrameCount) 
	/ sinceFrameUpdate;
      m_lastFrameTime = now;

      VEE_trace("VrSmuggler::preFrame # Frame rate is %f Hz", rate);
      m_lastFrameCount = m_frameCount;

      if(m_statistics && m_sampleGl) {
	GEE_GlRenderer * g = & m_sampleGl->m_geeRenderer;
	fprintf(m_statistics, "%u  %f %ld %ld %ld %ld %ld %ld %ld %ld %ld ", 
		m_frameCount, rate, 
		g->vertexCount(), g->triangleCount(), 
		g->particleCount(), g->geometryCount(),
		(long) m_sampleGl->m_geeLists.size(),
		g->renderDisplayListCount(), 
		g->generateDisplayListCount(),
		m_updateUSCount, m_renderUSCount);

	for(uint i = 0; i < 6; i++)
	  fprintf(m_statistics, " %.2f", (float) m_sampleGl->m_timings[i] * 0.000001f);

	fprintf(m_statistics, " %u %.2f ", index, (float) maxRenderTime * 0.000001f);
	
	fprintf(m_statistics, "\n");

	fflush(m_statistics);

	m_updateUSCount = 0;
	m_renderUSCount = 0;
	m_sampleGl->clearTimings();
      }
    }

    m_lastFrameEnd.getTimeOfDay();

    m_updateUSCount += now.usecsTo(m_lastFrameEnd);

    m_engine->update();

    if(m_engine->isDone())
      quit(true);
  }

  static int contextCount = 0;

  void VrSmuggler::contextInit(VRS_Window * context)
  {
    const char * fname = "VrSmuggler::contextInit";

    contextCount++;

    GlData *glData = new GlData;

    if(!m_sampleGl)
      m_sampleGl = glData;
  
    uint index = context->wall()->getIndex();

    _smuggler_mutex.lock();
    m_glData[index] = glData;
    _smuggler_mutex.unlock();

    VEE_trace("%s # %d %p", fname, contextCount, glData);

    glData->contextInit(m_engine, &m_config);

    glData->lightsOn();

    // glEnable(GL_BLEND);
    m_frameCount = 0;
    m_lastFrameCount = 0;
    m_lastFrameTime.getTimeOfDay();  
  
    glFlush();

    VEE_trace("%s # EXIT", fname);
  }

  void VrSmuggler::draw(VRS_Window * context, int)
  {
    container::iterator it = m_glData.find(context->wall()->getIndex());

    if(it == m_glData.end())
      VEE_fatal("VrSmuggler::draw # No GL data");

    GlData *glData = (*it).second;
    glData->m_geeRenderer.resetCounters();

    VEE_GlRenderer *veeGl = & glData->m_veeRenderer;
    veeGl->setBlend(true);

    glEnable(GL_BLEND);

    glDisable(GL_LIGHTING);

    m_engine->render(glData, veeGl, & glData->m_solarRenderer);
    
  }

  void VrSmuggler::loadConfig(const char * filename)
  {
    if(!filename) return;

    ConfigReader::readConfig(& m_config, filename);
  }
}
