/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Crystal animation and 3D modeling application
 *
 * Authors:   Tommi Ilmonen 2001-2004
 * Copyright: Tommi Ilmonen and/or Tekes 2001-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <gee_debug.h>

#include <vee_error.h>
#include <vee_force_plane.h>
#include <vee_gl_renderer.h>
#include <vee_gravitate.h>
#include <vee_magic_rain.h>
#include <vee_scene.h>
#include <vee_qt.h>

#include <di_data_buffer.h>

#include <crystal_engine.h>
#include <crystal_vrsmuggler.h>

#include <Fluid/Config.hpp>
#include <Fluid/InputDevices.hpp>

int main(int argc, char **argv)
{
  const char * fluidConfig = "fluid-config";
  const char * animaConfig = "anima-config";
  const char * vrsConfig   = "eve.vrjfluid.config";
  const char * preload = 0;

  int processedArgs = 1;

  while(processedArgs < argc && argv[processedArgs][0] == '-') {

    const char *arg = argv[processedArgs];

    if(strcmp(arg, "--fluid") == 0 && processedArgs + 1 < argc) {
      fluidConfig = argv[++processedArgs];
      processedArgs++;
      VEE_trace("New FLUID config \"%s\"", fluidConfig);
    }
    else if(strcmp(arg, "--vrs") == 0 && processedArgs + 1 < argc) {
      vrsConfig = argv[++processedArgs];
      processedArgs++;
      VEE_trace("New VRSmuggler config \"%s\"", vrsConfig);
    }
    else if(strcmp(arg, "--config") == 0 && processedArgs + 1 < argc) {
      animaConfig = argv[++processedArgs];
      processedArgs++;
      VEE_trace("New ANIMA config \"%s\"", animaConfig);
    }
    else if(strcmp(arg, "--load") == 0 && processedArgs + 1 < argc) {
      preload = argv[++processedArgs];
      processedArgs++;
      VEE_trace("New preload file \"%s\"", preload);
    }
    else if(strcmp(arg, "--debug") == 0) {
      if(processedArgs + 1 >= argc || argv[processedArgs + 1][0] == '-') {
	GEE_Debug::setDebugAll(1);
	processedArgs++;
      }
      else {
	for(processedArgs++; 
	    processedArgs  < argc && argv[processedArgs][0] != '-'; 
	    processedArgs++) {
	  GEE_Debug::setDebug(argv[processedArgs]);
	}
      }	
    }
    else {
      VEE_error("Unrecognized option \"%s\"", arg);
      processedArgs++;
    }
  }

  /* LOAD FLUID: */
    
  VEE_trace("Initializing FLUID");

  Fluid::DeviceConfig::init(fluidConfig);

  /* Load more configurations. */


  VEE_Qt::initImageIO();

  VEE_SceneBox *scene = new VEE_SceneBox(0, 
					 Fluid::Vector3(-50.0, -50.0, -50.0),
					 Fluid::Vector3(50.0, 50.0, 50.0));

  VEE_trace("Creating CRYSTAL engine");

  Crystal::Engine engine(scene, animaConfig);
  
#ifndef VEE_FAST
  VEE_trace("%d particles in box", (int) scene->totalChildren());
  VEE_trace("%d particles left", (int) VEE_Object::count());
  VEE_trace("%d scene boxes left", (int) VEE_SceneBox::count());
#endif

  /* Fluid::Glove *glove = Fluid::InputDevices::getGlove("hanska");

  if(!glove) 
  VEE_fatal("Glove is NULL"); */
  
  Crystal::VrSmuggler* application = 
    new Crystal::VrSmuggler(&engine);

  if(animaConfig) {
    VEE_trace("Reading AnimaLand config \"%s\"", animaConfig);
    application->loadConfig(animaConfig);
  }
  else {
    VEE_trace("No AnimaLand configuration specified");
  }
  
  Fluid::InputDevices::init();
  Fluid::InputDevices::start();

  VEE_trace("Fluid::InputDevices::init DONE");

  // Load VRS config file

  application->loadVrConfig(vrsConfig);

  if(preload)
    if(!engine.readDomFile(preload))
      VEE_error("There were problems loading %s", preload);

  application->init();
  
  application->exec(false);

  return 0;
}



