/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Crystal animation and 3D modeling application
 *
 * Authors:   Tommi Ilmonen 2001-2004
 * Copyright: Tommi Ilmonen and/or Tekes 2001-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <crystal_qmain.h>

#include <crystal_engine.h>
#include <crystal_qsettings.h>

#include <vee_error.h>

#include <di_sleep.h>

#include <qfileinfo.h>

#include <ConfigReader.hpp>

#ifdef VEE_WITH_RENDERMAN
#include <vee_rman_renderer.h>
#include <ri.h>
#endif

CRYSTAL_QMain::CRYSTAL_QMain(VEE_Qt * host,
			     Crystal::Engine *engine,
			     QWidget * parent,
			     const char * name)
  : VEE_QtHidden(host, engine->scene(), parent, name),
    m_engine(engine),
    m_settings(new CRYSTAL_QSettings),
    m_ribFrame(0),
    m_ribIndex(0),
    m_frameWidth(640),
    m_frameHeight(480),
    m_extraSleep(0.0)
#ifdef VEE_WITH_RENDERMAN
    , m_rmanRenderer(0)
#endif

{
  m_headTrack = false;
  
  m_controls = new QPopupMenu();
  
  m_host->menuBar()->insertItem( "&Controls", m_controls);

  m_playId   = m_controls->insertItem("&Play", this, SLOT(play()));
  m_recordId = m_controls->insertItem("&Record", this, SLOT(record()));
  m_stopId   = m_controls->insertItem("&Stop", this, SLOT(stop()));
  m_controls->insertItem("&Head tracking", this, SLOT(headTrack()));
  m_controls->insertItem("Extra sleep", this, SLOT(extraSleep()));

  m_controls->insertSeparator();
  m_controls->insertItem("S&ettings", this, SLOT(showSettings()));

  delete m_veeGl;
  m_veeGl = & m_glData.m_veeRenderer;
  m_veeGl->setSceneBox(m_engine->scene());
}

CRYSTAL_QMain::~CRYSTAL_QMain()
{
  stop();

  m_veeGl = 0;

  delete m_settings;
}

void CRYSTAL_QMain::initializeGL()
{
  VEE_QtHidden::initializeGL();
  
  ConfigReader::Config conf;  
  ConfigReader::readConfig(& conf, m_configFile.ascii());

  m_glData.contextInit(m_engine, & conf);
}

void CRYSTAL_QMain::reFresh()
{
  if(m_extraSleep > 0.0)
    DI_Sleep::sleepMs(m_extraSleep * 1000.0);

  // If recording -> not in real-time
#ifdef VEE_WITH_RENDERMAN
  m_engine->update(m_rmanRenderer ? 1.0 / m_fixedRate : 0.0);
#else
  m_engine->update(m_fixedRate > 1.0 ? 1.0 / m_fixedRate : 0.0);
#endif

  VEE_QtHidden::reFresh();

  renderRibFrame();
}

void CRYSTAL_QMain::play()
{
  m_controls->setItemEnabled(m_playId, false);
  m_controls->setItemEnabled(m_recordId, false);
  m_controls->setItemEnabled(m_stopId, true);

  m_calculusUpdate = false;
  m_host->toggleCalculusUpdate();

  m_fixedRate = 0.0;

  m_engine->setRealTime(true);
}

void CRYSTAL_QMain::record()
{
  finalizeRibSegment();

  m_settings->getRibFileName(&m_ribBase);
  m_settings->getFrameFileName(&m_frameBase);
  m_settings->getFramesPerFile(&m_ribFramesPerSegment);
  
  m_fixedRate = m_settings->getFrameRate();

  m_controls->setItemEnabled(m_playId, false);
  m_controls->setItemEnabled(m_recordId, false);
  m_controls->setItemEnabled(m_stopId, true);

  VEE_trace("CRYSTAL_QMain::record # %s -> %s, %f Hz, %d frames / file",
	    m_ribBase.ascii(), m_frameBase.ascii(), m_fixedRate, 
	    m_ribFramesPerSegment);

  // Not really needed, but lets do it anyway:
#ifdef VEE_WITH_RENDERMAN
  delete m_rmanRenderer;
  m_rmanRenderer = new VEE_RManRenderer(1.0 / m_fixedRate);
#endif

  m_ribFrame = 1;
  m_ribIndex = 1;

  beginRibSegment();

  m_calculusUpdate = true;

  m_overSample = 1;

  m_engine->setRealTime(false);
}

void CRYSTAL_QMain::stop()
{
  VEE_trace("CRYSTAL_QMain::stop # %d frames", m_ribFrame);

  m_controls->setItemEnabled(m_playId, true);
  m_controls->setItemEnabled(m_recordId, true);
  m_controls->setItemEnabled(m_stopId, false);

  finalizeRibSegment();
  m_ribFrame = 0;
  m_ribIndex = 0;

  m_calculusUpdate = true;
  m_host->toggleCalculusUpdate();

  m_overSample = 1;
  m_fixedRate = 0.0;

  m_engine->setRealTime(true);

#ifdef VEE_WITH_RENDERMAN
  delete m_rmanRenderer;
  m_rmanRenderer = 0;
#endif
}

void CRYSTAL_QMain::showSettings()
{
  m_settings->show();
  m_settings->raise();
}

void CRYSTAL_QMain::headTrack()
{
  m_headTrack = !m_headTrack;
}

void  CRYSTAL_QMain::extraSleep()
{
  m_extraSleep = 0.3;
}

void CRYSTAL_QMain::veeRender(bool clear)
{
  // Solar::GlRenderer solarGl;

  // m_veeGl->paintGL(m_showBoxes, true, true, m_lastFrameDuration);

  if(clear)
    m_veeGl->clearGlScreen();

  m_veeGl->moveGlCamera();
  m_veeGl->moveCamera(m_lastFrameDuration);

  m_glData.lightsOn();

  m_engine->render(&m_glData,
		   m_veeGl,
		   & m_glData.m_solarRenderer,
		   m_headTrack);
}

void CRYSTAL_QMain::renderRibFrame()
{
#ifdef VEE_WITH_RENDERMAN
  if(m_ribFrame < 1 || !m_rmanRenderer)
    return;

  // VEE_trace("CRYSTAL_QMain::renderRibFrame # %d", m_ribFrame);
  
  if(m_ribFrame % m_ribFramesPerSegment == 0) {
    nextRibSegment();
  }

  char buf[20];
  sprintf(buf, "-%.4d.tif", m_ribFrame);

  QString frameName(m_frameBase + buf);
  
  RiOrientation(RI_RH);
  float fov = 70.0;
  RiProjection(RI_PERSPECTIVE, RI_FOV, & fov, RI_NULL);

  VEE_RManRenderer::placeCamera(m_veeGl->camPos(), m_veeGl->camDir());

  RiFrameBegin(m_ribFrame % m_ribFrame + 1);
  RiDisplay((char *) frameName.ascii(), RI_FILE,RI_RGBA, RI_NULL);

  RiShutter(0.0, 1.0);
  RiWorldBegin();
  
  RiIdentity();
  RiSurface("particle", RI_NULL);
  
  /* if(m_fixedRate > 0.0) UNNECESSARY
     m_rmanRenderer->setFrameDur(1.0 / m_fixedRate); */

  m_rmanRenderer->paintRecursive(m_scene);
    
  RiWorldEnd();
  RiFrameEnd();  

  m_ribFrame++;
#endif
}

void CRYSTAL_QMain::nextRibSegment()
{
  finalizeRibSegment();
  beginRibSegment();
}

void CRYSTAL_QMain::beginRibSegment()
{
#ifdef VEE_WITH_RENDERMAN
  char buf[20];

  sprintf(buf, "-%.3d.rib", m_ribIndex);

  m_ribName = m_ribBase + buf;

  RiBegin((RtToken) m_ribName.ascii());
  RiFormat(m_frameWidth, m_frameHeight, 0.0);
  RiCropWindow(0.0, 1.0, 0.0, 1.0);
#endif
}

void CRYSTAL_QMain::finalizeRibSegment()
{
#ifdef VEE_WITH_RENDERMAN
  if(m_ribIndex < 1)
    return;

  RiEnd();

  fflush(0);

  QString gzName(m_ribName + ".gz");

  QFileInfo figz(gzName);
  
  if(figz.exists()) {
    // Remove the old gz-file if it did exist
    QFile tmp(gzName);
    tmp.remove();
  }
  
  QString command("gzip ");
  command += m_ribName;

  VEE_trace("CRYSTAL_QMain::finalizeRibSegment # Running \"%s\"",
	    command.ascii());

  system(command.ascii());
  m_ribIndex++;
#endif
}
