/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Crystal animation and 3D modeling application
 *
 * Authors:   Tommi Ilmonen 2001-2004
 * Copyright: Tommi Ilmonen and/or Tekes 2001-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <crystal_vrjuggler.h>

#include <crystal_engine.h>
#include <crystal_gl_data.h>

#include <vee_error.h>

#include <Fluid/Config.hpp>

#include <iostream>
#include <unistd.h>

namespace Crystal { 
  
  VrJuggler::VrJuggler(Engine *engine)
    : m_sampleGl(0),
      m_engine(engine)
  {
    VEE_GlRenderer::createPrivateTextures(false);
    m_statistics = fopen("statistics.txt", "w");
  }

  VrJuggler::~VrJuggler()
  {
    if(m_statistics)
      fclose(m_statistics);
  }
  
  void VrJuggler::init()
  {
    const char * fname = "VrJuggler::init";

    VEE_trace(fname);

    m_engine->init();

    m_frameCount = 0;
    m_lastFrameCount = 0;
    m_lastFrameTime.getTimeOfDay();  

    
    m_updateUSCount = 0;
    m_renderUSCount = 0;
  }

  void VrJuggler::apiInit()
  {
    VEE_trace("VrJuggler::apiInit");
  }

  void VrJuggler::preFrame()
  {
    m_frameCount++;

    VEE_Time now;
    now.getTimeOfDay();
    
    m_renderUSCount += m_lastFrameEnd.usecsTo(now);

    double sinceFrameUpdate = m_lastFrameTime.secsTo(now);

    if(sinceFrameUpdate >  5.0) {

      long maxRenderTime = 0;
      uint index = 0;

      for(uint i = 0; i < GlData::m_instances.size(); i++) {

	long t = GlData::m_instances[i]->m_timings[3];
	
	if(maxRenderTime < t) {
	  maxRenderTime = t;
	  index = i;
	}
      }

      double rate = (double) (m_frameCount - m_lastFrameCount) 
	/ sinceFrameUpdate;
      m_lastFrameTime = now;

      VEE_trace("VrJuggler::preFrame # Frame rate is %f Hz", rate);
      m_lastFrameCount = m_frameCount;

      if(m_statistics && m_sampleGl) {
	GEE_GlRenderer * g = & m_sampleGl->m_geeRenderer;
	fprintf(m_statistics, "%u  %f %ld %ld %ld %ld %ld %ld %ld %ld %ld ", 
		m_frameCount, rate, 
		g->vertexCount(), g->triangleCount(), 
		g->particleCount(), g->geometryCount(),
		(long) m_sampleGl->m_geeLists.size(),
		m_displayListCount, g->generateDisplayListCount(),
		m_updateUSCount, m_renderUSCount);

	for(uint i = 0; i < 4; i++)
	  fprintf(m_statistics, " %.2f", m_sampleGl->m_timings[i] * 0.000001f);

	fprintf(m_statistics, " %u %.2f ", index, maxRenderTime * 0.000001f);
	
	fprintf(m_statistics, "\n");

	fflush(m_statistics);

	m_updateUSCount = 0;
	m_renderUSCount = 0;
	m_sampleGl->clearTimings();
      }
    }

    m_lastFrameEnd.getTimeOfDay();

    m_updateUSCount += now.usecsTo(m_lastFrameEnd);

    m_engine->update();
  }

  void VrJuggler::intraFrame()
  {}

  void VrJuggler::postFrame()
  {}

  static int contextCount = 0;

  void VrJuggler::contextInit()
  {
    const char * fname = "VrJuggler::contextInit";

    contextCount++;

    // sleep(1); // Maybe this makes for less core dumps (!!??)

    GlData *glData = new GlData;

    if(!m_sampleGl)
      m_sampleGl = glData;
  
    (*m_glData) = glData;

    vjDEBUG(vjDBG_ALL, 0) << clrSetNORM(clrRED) 
			  << fname << "  "
			  << contextCount << "\n" << vjDEBUG_FLUSH;
    // printf("VrJuggler::contextInit %d\n"), count;

    // Fluid::ConfigChunk * glConfig = Fluid::Config::getConfigChunk("OpenGL");

    glData->contextInit(m_engine, &m_config);

    glData->lightsOn();

    // glEnable(GL_BLEND);
    m_frameCount = 0;
    m_lastFrameCount = 0;
    m_lastFrameTime.getTimeOfDay();  
  
    glFlush();

    vjDEBUG(vjDBG_ALL, 0) << clrSetNORM(clrGREEN) 
			  << "VrJuggler::contextInit EXIT\n"
			  << vjDEBUG_FLUSH;
  }

  void VrJuggler::draw()
  {
    GlData *glData = (*m_glData);
    glData->m_geeRenderer.resetCounters();

    VEE_GlRenderer *veeGl = & glData->m_veeRenderer;
    veeGl->setBlend(true);

    glEnable(GL_BLEND);

    glDisable(GL_LIGHTING);

    /* Draw a single line to show the positive X-axis. */
    /* glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
       glBegin(GL_LINES);

       glVertex3f(0.0, 0.0, 0.0);
       glVertex3f(10.0, 0.0, 0.0);

       glEnd(); */
  
    m_engine->render(glData, veeGl, & glData->m_solarRenderer);
    m_displayListCount = glData->m_geeRenderer.renderDisplayListCount();
  }

  void VrJuggler::loadConfig(const char * filename)
  {
    if(!filename) return;

    ConfigReader::readConfig(& m_config, filename);
  }
}
