/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Crystal animation and 3D modeling application
 *
 * Authors:   Tommi Ilmonen 2001-2004
 * Copyright: Tommi Ilmonen and/or Tekes 2001-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <gee_debug.h>

#include <vee_error.h>
#include <vee_force_plane.h>
#include <vee_gl_renderer.h>
#include <vee_gravitate.h>
#include <vee_magic_rain.h>
#include <vee_scene.h>
#include <vee_qt.h>

#include <di_data_buffer.h>

#include <crystal_engine.h>
#include <crystal_vrjuggler.h>

#include <Fluid/Config.hpp>
#include <Fluid/InputDevices.hpp>
#include <Fluid/VrjTrackerExport.hpp>

#include <Kernel/vjKernel.h>

int main(int argc, char **argv)
{  
  const char * fluidConfig = "fluid-config";
  const char * animaConfig = "anima-config";
  const char * vrjConfig = "eve.vrjfluid.config";
  const char * preload = 0;

  int processedArgs = 1;

  while(processedArgs < argc && argv[processedArgs][0] == '-') {

    const char *arg = argv[processedArgs];

    if(strcmp(arg, "--fluid") == 0 && processedArgs + 1 < argc) {
      fluidConfig = argv[++processedArgs];
      processedArgs++;
      VEE_trace("New FLUID config \"%s\"", fluidConfig);
    }
    else if(strcmp(arg, "--vjr") == 0 && processedArgs + 1 < argc) {
      vrjConfig = argv[++processedArgs];
      processedArgs++;
      VEE_trace("New VRJuggler config \"%s\"", vrjConfig);
    }
    else if(strcmp(arg, "--config") == 0 && processedArgs + 1 < argc) {
      animaConfig = argv[++processedArgs];
      processedArgs++;
      VEE_trace("New ANIMA config \"%s\"", animaConfig);
    }
    else if(strcmp(arg, "--load") == 0 && processedArgs + 1 < argc) {
      preload = argv[++processedArgs];
      processedArgs++;
      VEE_trace("New preload file \"%s\"", preload);
    }
    else if(strcmp(arg, "--debug") == 0) {
      if(processedArgs + 1 >= argc || argv[processedArgs + 1][0] == '-') {
	GEE_Debug::setDebugAll(1);
	processedArgs++;
      }
      else {
	for(processedArgs++; 
	    processedArgs  < argc && argv[processedArgs][0] != '-'; 
	    processedArgs++) {
	  GEE_Debug::setDebug(argv[processedArgs]);
	}
      }	
    }
    else {
      VEE_error("Unrecognized option \"%s\"", arg);
      processedArgs++;
    }
  }

  /* LOAD FLUID: */
    
  VEE_trace("Initializing FLUID");

  Fluid::DeviceConfig::init(fluidConfig);

  /* Load more configurations. */


  VEE_Qt::initImageIO();

  VEE_SceneBox *scene = new VEE_SceneBox(0, 
					 Fluid::Vector3(-50.0, -50.0, -50.0),
					 Fluid::Vector3(50.0, 50.0, 50.0));

  VEE_trace("Creating CRYSTAL engine");

  Crystal::Engine engine(scene, animaConfig);
  
#ifndef VEE_FAST
  VEE_trace("%d particles in box", (int) scene->totalChildren());
  VEE_trace("%d particles left", (int) VEE_Object::count());
  VEE_trace("%d scene boxes left", (int) VEE_SceneBox::count());
#endif

  /* Fluid::Glove *glove = Fluid::InputDevices::getGlove("hanska");

  if(!glove) 
  VEE_fatal("Glove is NULL"); */
  
  // Allocate the kernel object and the application object
  vjKernel* kernel = vjKernel::instance();
  Crystal::VrJuggler* application = 
    new Crystal::VrJuggler(&engine);

  if(animaConfig) {
    VEE_trace("Reading AnimaLand config \"%s\"", animaConfig);
    application->loadConfig(animaConfig);
  }
  else {
    VEE_trace("No AnimaLand configuration specified");
  }
  
  // IF not args passed to the program
  //    Display usage information and exit
  if ((argc - processedArgs) <= 0) {
    std::cout << "\n\n"
	      << "Usage: " << argv[0] 
	      << " vjconfigfile[0] vjconfigfile[1] ... vjconfigfile[n]" 
	      << std::endl;
    exit(1);
  }

  Fluid::InputDevices::init();

  VEE_trace("Fluid::InputDevices::init DONE");

  Fluid::VrjTrackerExport::configInit();

  // Load VRJ config file

  kernel->loadConfigFile(vrjConfig);

  if(preload)
    if(!engine.readDomFile(preload))
      VEE_error("There were problems loading %s", preload);
  
  // Start the kernel
  kernel->start();
  
  // Give the kernel an application to execute
  kernel->setApplication(application);
  
  // Run the app:
  while(!engine.isDone()) {
    usleep(250000);
  }

  return 0;
}



