/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, Janne Kontkanen 1999-2002.
 * Tommi.Ilmonen@hut.fi, Janne.Kontkanen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#include <Fluid/Config.hpp>

#include <Fluid/Debug.hpp>
#include <Fluid/Error.hpp>

#include <di_mutex.h>

#include <assert.h>

#include <fstream>

extern bool parse_configuration(const char *fname);

namespace Fluid {

  static DI_MutexAuto __ConfigMutex;

  DeviceConfig * DeviceConfig::instance = 0;

  /**  
   * Loads configuration file and initializes the Config singleton.
   * 
   * The system tries to load the configuration from the following
   * files (in this order):
   *
   * 1) Filename given in rcfile parameter
   * 2) 'fluid-config' in working directory
   * 3) 'fluid-config' in directory FLUID_ROOT (environment variable)
   *
   */
  
  void DeviceConfig::init(const std::string& rcfile)
  {
    DI_Guard guard(&__ConfigMutex);

    assert(instance == 0);

    if(instance == 0) {

      logStream.rdbuf(std::cout.rdbuf());
      errStream.rdbuf(std::cerr.rdbuf());
      
      instance = new DeviceConfig;
      using namespace std;

      Trace("DeviceConfig::init # Trying to get configuration from file \"%s\"",
	    rcfile.c_str());
      if(!addConfig(rcfile.c_str())) {
	char *rootdir = getenv("FLUID_ROOT");
	Trace("DeviceConfig::init # Trying to get configuration from file \"%s\"",
	      "./fluid-config");
	if(!addConfig("./fluid-config") && rootdir != NULL) {
	  std::string path = std::string(rootdir) + "/fluid-config";
	  if(!addConfig(path.c_str())) {
	    errStream << "FLUID ERROR # No configuration file found at all\n";
	    errStream << "FLUID ERROR # Tried: " << rcfile.c_str() << " ./fluid-config, $FLUID_ROOT/fluid-config\n"; 
	  }
	  else
	    logStream << "FLUID # Read configuration from: " << path << std::endl;
	}
	else {
	  logStream << "FLUID # Read configuration from: " << "./fluid-config" << std::endl;
	}
      }
      else {
	logStream << "FLUID # Read configuration from: " << rcfile << std::endl;
      }
    }

    dump(std::cout);

    logStream << "FLUID # Config init() finished" << std::endl;
  }
  
  bool DeviceConfig::addConfig(const std::string& rcfile)
  {
    DI_Guard guard(&__ConfigMutex);

    Config conf;

    if(!instance)
      Fatal("Config must be initialized bewfore calling DeviceConfig::addConfig");
    bool ok = readConfig(&conf, rcfile.c_str());
    
    if(ok)
      instance->m_config.merge(conf);

    return ok;
  }

  void DeviceConfig::cleanup()
  {
    DI_Guard guard(&__ConfigMutex);
    delete instance;
    instance = 0;
  }
  
  Chunk* DeviceConfig::getChunk(const std::string& chunkType)
  {
    assert(instance);

    DI_Guard guard(&__ConfigMutex);

    Config::iterator it = instance->m_config.find(chunkType);
    
    return (it != instance->m_config.end()) ? & (*it).second : 0;

    /* for(Config::iterator it = instance->m_config.begin();
	it != instance->m_config.end(); it++) {
      if((*it).second.get("name").getString() == chunkName)
	return & (*it).second;
    }

    return 0; */
  }


  

  void DeviceConfig::dump(std::ostream& os)
  {
    if(!instance)
      return;
    
    instance->m_config.dump(os);
  }

  Config * DeviceConfig::config()
  {
    if(!instance)
      return 0;
    
    return & instance->m_config;
  }


  long long    DeviceConfig::getStartupTimeStamp()
  {
    return instance->startupTimeStamp;
  }

  void         DeviceConfig::setStartupTimeStamp(long long t)
  {
    instance->startupTimeStamp = t;
  }
} // namespace
