/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, Janne Kontkanen 1999-2002.
 * Tommi.Ilmonen@hut.fi, Janne.Kontkanen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#ifndef FLUID_CONFIG_HPP
#define FLUID_CONFIG_HPP

#include <string>
#include <map>
#include <iostream>

#include <Fluid/ConfigReader.hpp>

namespace Fluid {

  /**
   * System wide configuration settings
   *
   * Config class is a collection of static service
   * functions that provide access to config singleton.
   *
   * Config consists of Chunks
   */

class DeviceConfig
{
public:
  /**  
   * Loads configuration file and initializes this object
   * 
   * Loads the configuration file .fluidrc, which contains
   * input device dependent configuration parameters for each
   * device. This function must be called before anything else
   * is done with the Configuration object - usually this is
   * done in the initialization phase of the system.
   */
  static void         init(const std::string& rcfile = "");

  /** Loads a new config file to the system. The system must be
      initialized. */

  static bool         addConfig(const std::string& rcfile);

  /**
   * Releases the configuration settings.
   * 
   * Deletes the Config-singelton instance and releases the 
   * memory for configuration settings. Usually this is done
   * in the cleanup-phase of the system.
   */
  static void         cleanup();

  /**
   * Returns the config-chunk matching to given name
   *
   * Returns the config chunk matching to given name. Config chunks
   * are usually created and added to Config when Config is initialized.
   *
   * Device drivers use this function to access the configuration 
   * information
   */
  static Chunk* getChunk(const std::string& chunkName);

  /**
   * Dumps all the configuration information on the given stream
   */
  static void         dump(std::ostream& os);

  static Config * config();
  

  static long long    getStartupTimeStamp();
  static void         setStartupTimeStamp(long long t);  
private:
  Config         m_config;
  static DeviceConfig *instance;
  long long startupTimeStamp;
};

} // namespace

#endif
