/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#ifndef FLUID_DATASOURCE_HPP
#define FLUID_DATASOURCE_HPP

#include "di_thread.h"
#include "mj_ringbuffer.h"

#include "Fluid/Error.hpp"
#include "Fluid/Mutex.hpp"
#include "Fluid/TimeStamp.hpp"
#include "Fluid/Debug.hpp"
#include "Fluid/DumpFile.hpp"

#include <string>
#include <typeinfo>

namespace Fluid
{

/**
 * Base class for all the data sources
 *
 * Data sources derive from this and device interface classes
 * use this as a data source. Polling data sources derive 
 * from this class directly, but stream data sources are
 * simpler to implement by deriving from StreamDataSource
 *
 */
template <class Sample>
class DataSource
{
public:
           DataSource(int bufferSize)
	     :buffer(bufferSize),readIdx(0),writeIdx(0),latestTimeStamp(0),dumpFile(0)
             {}
  virtual ~DataSource() { delete dumpFile; }
  /// Start the device driver thread
  virtual void    start() = 0;
  /// Stop the device driver thread
  virtual void    stop() = 0;
  /// Lock the buffer for reading/writing data
  void    lock() { mutex.lock(); }
  /// Release the buffer to allow the device to go on
  void    release() { mutex.unlock(); }
  /// Number of available samples
  int     getAvailable() { return writeIdx-readIdx; }
  /// Read one sample 
  void    read(Sample &sample) { sample = buffer[readIdx++]; }
  /// Return the time latest sample was read in
  long long getLatestTimeStamp() { return latestTimeStamp; }    
  /// Returns the sampling rate in hertz
  virtual float getSampleRate() const { return 0.0; }

  virtual std::string getDriverName() const { return "Unknown"; } 
  /// Number of samples the input head has written to intermediate buffer.
  long writtenCount() const { return (long) writeIdx; }

  ulong bufferSize() const { return buffer.size(); }

  void setDumpFile(const std::string& fname)
  {
    delete dumpFile;
    dumpFile = new DumpFile<Sample>(fname);
    Trace("Opened dump file \"%s\"", fname.c_str());
  }
protected:
  /// Write one sample - null operation if the buffer is full
  void    write(const Sample &sample) 
  { 
    lock();
    if(writeIdx-readIdx < buffer.size())
      {
	buffer[writeIdx] = sample; 
	writeIdx++;
      }
    else
      {
	//errStream << "Warning - buffer overflow\n";
      }
    latestTimeStamp = TimeStamp::getTimeOfDayStamp();
    if(dumpFile)
      {
	dumpFile->write(latestTimeStamp,sample);
      }
    release();
  }
private:
  MJ_RingBuffer<Sample> buffer;
  Mutex    mutex;
  unsigned int readIdx;
  unsigned int writeIdx;
  long long latestTimeStamp;
  DumpFile<Sample>  *dumpFile;
};

  /**
   * Base class for stream-like datasources
   *
   * Stream-like data sources are data sources
   * that can easily provide sample read 
   * operation that blocks until next sample
   * is available
   */

template <class Sample>
class StreamDataSource : public DataSource<Sample>, public DI_Thread
{
public:         
  StreamDataSource(int bufferSize)
    :DataSource<Sample>(bufferSize)
  {}
 
  void start()
  {
    DI_Thread::run();
  }
  
  void stop()
  {
    DI_Thread::kill();
  }

protected:
  virtual void get(Sample& s) = 0;
private:
  void childLoop()
  {
    logStream << "FLUID # childLoop for " << typeid(*this).name() 
	      << " StreamDataSource started\n";
    while(true)
      {
	Sample s;
	get(s);
	write(s);
      }
  }  
};

} // namespace

#endif
