/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, Janne Kontkanen 1999-2002.
 * Tommi.Ilmonen@hut.fi, Janne.Kontkanen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#ifndef FLUID_FILEDATASOURCE_HPP
#define FLUID_FILEDATASOURCE_HPP

#include <Fluid/DataSource.hpp>
#include <Fluid/Error.hpp>
#include <Fluid/InputDevices.hpp>

#include <di_data_buffer.h>
// #include <di_socket.h>
#include <di_sleep.h>

#include <string>
#include <typeinfo>

namespace Fluid
{

  template <class Sample>
  class FileDataSource : public StreamDataSource<Sample>
  {
  public:
    FileDataSource(const std::string& filename, int bufferSize,
		   float sr);
    ~FileDataSource();
    void get(Sample& s);
    virtual float getSampleRate() const { return sampleRate; }
    virtual std::string getDriverName() const { return "File"; }
  private:
    DI_DataBuffer2 buffer;
    std::string fname;
    FILE *fp;
    // TimeStamp timeshift;
    float sampleRate;
  
    long m_count;
  };

  template <class Sample>
  inline FileDataSource<Sample>::FileDataSource
  (const std::string& filename, int bufferSize, float sr)
    :StreamDataSource<Sample>(bufferSize),
    buffer(256),fname(filename)/* ,timeshift(0)*/, sampleRate(sr), m_count(0)
  {
    fp = fopen(filename.c_str(),"rb");

    Trace("%s::FileDataSource # %s %p", 
	  typeid(*this).name(), filename.c_str(), fp);
  }

  template <class Sample>
  inline FileDataSource<Sample>::~FileDataSource()
  {
    fclose(fp);
  }

  /**
   * Fetch one sample from file
   *
   * This function reads one sample from the file. Samples
   * are timestamped, so the function blocks for some time 
   * if the next sample in file is scheduled for the future.
   * 
   * @param s = destination for the read    
   */

  template <class Sample>
  inline void FileDataSource<Sample>::get(Sample& s)
  {
    // Trace("%s::get # %p", typeid(*this).name(), &s);

    if(fp) {

      if(buffer.read(fp) < 4) {
	// Trace("%s::get # Failed to read data", typeid(*this).name());
	s.clear();
	return;
      }
    
      buffer.rewind();
      
      TimeStamp tstamp = buffer.readInt64();

      // Trace("%s::get # 1", typeid(*this).name());

      if(InputDevices::timeShift() == 0) {
	// Trace("%s::get # 2", typeid(*this).name());
	//timeshift = tstamp - TimeStamp::getTimeOfDayStamp();
	InputDevices::setTimeShift(InputDevices::latestTimeStamp() - tstamp);
      }
      else if(m_count) {
	// Trace("%s::get # 3", typeid(*this).name());

	// Trace("%s::get # %d", typeid(*this).name(), m_count++);
	InputDevices::threadWaitUntil(tstamp + InputDevices::timeShift());
	  
	/*
	  long long stime = 
	  tstamp - (TimeStamp::getTimeOfDayStamp()+timeshift);
	  if(stime > 0)
	  {
	  // We will not sleep more than one second here.
	  float subsecs = TimeStamp(stime).subSeconds();
	  DI_Sleep::sleepUs(subsecs * 1000000.0);
	  }
	*/
      }
      
      m_count++;

      // Trace("%s::get # 4", typeid(*this).name());
      s.read(buffer);
      // buffer.moveIndex(0);
    }
    else {
      errStream << "FLUID ERROR # FileDataSource::get(): read error" 
		<< std::endl;
      s.clear();
    }

    // Trace("%s::get # EXIT", typeid(*this).name(), &s);
  }

}

#endif
