/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#include <Fluid/Glove.hpp>
#include <di_data_buffer.h>

#include <mj_ringbuffertmpl.h>

namespace Fluid
{

void GloveSample::clear()
{
  for(int i = 0; i < 5; i++)
    for(int j = 0; j < 3; j++)
      fingerJoints[i][j] = 0.0f;
  for(int i = 0; i < 4; i++)
    abductionAngles[i] = 0.0f;
}

/**
 * De-serializes the glove sample from data buffer. This
 * function is used by SocketDataSource template to
 * fetch samples from network sockets
 *
 * @param buffer = buffer from which to read
 * @note DI_XXX classes are part of Diva-library, check the
 *       documentation of Diva for details
 */

void GloveSample::read(DI_DataBuffer2 &buffer)
{
  for(int i = 0; i < 5; i++)
    for(int j = 0; j < 3; j++)
      fingerJoints[i][j] = buffer.readFloat32();
  for(int i = 0; i < 4; i++)
    abductionAngles[i] = buffer.readFloat32();
}

/**
 * Writes the glove sample into data buffer. This
 * function is used by SocketDataSource template to
 * write samples to network socket
 *
 * @param buffer = buffer in which to write
 * @note DI_XXX classes are part of Diva-library, check the
 *       documentation of Diva for details
 */
void  GloveSample::write(DI_DataBuffer2& buffer) const
{
  for(int i = 0; i < 5; i++)
    for(int j = 0; j < 3; j++)
      buffer.writeFloat32(fingerJoints[i][j]);
  for(int i = 0; i < 4; i++)
    buffer.writeFloat32(abductionAngles[i]);
}

/**
 * Returns the bytes consumed by single sample serialized
 * by write(DI_DataBuffer2& buffer)
 *
 */
int GloveSample::getMaxSerializedSize() const
{
  return (15+4)*4+4; // 4 for padding
}

  /**
   * Retrieves the current status of specific joint of specific finger.
   *
   *  @param finger = finger of interest     
   *  @param joint  = joint of interest
   *  @return normalized float between in range [0,1] (1.0 = maximum flexture)
   */

float Glove::getJointAngle(Finger finger, FingerJoint joint)
{
  return getJointAngle(finger,joint,getLatestIndex());
}

  /**
   * Retrieves the status of specific joint of specific finger at 
   * given sample.
   *
   *  @param finger = finger of interest     
   *  @param joint  = joint of interest
   *  @param n = sample of interest
   *  @return normalized float between in range [0,1] (1.0 = maximum flexture)
   */

float Glove::getJointAngle(Finger finger, FingerJoint joint, ulong n)
{
  return getSample(n).fingerJoints[finger][joint];
}

float Glove::getAbductionAngle(AbductionAngle s)
{
  return getAbductionAngle(s,getLatestIndex());
}

float Glove::getAbductionAngle(AbductionAngle s, ulong n)
{
  return getSample(n).abductionAngles[s];
}

std::string Glove::getClassName() const
{
  return "Glove";
}

} // namespace

template class MJ_RingBuffer<Fluid::GloveSample>;
