/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#ifndef FLUID_IO_NODE_BUFFER_HPP
#define FLUID_IO_NODE_BUFFER_HPP

#include <mj_ringbuffer.h>

#include <Fluid/IONode.hpp>
#include <Fluid/TimeStamp.hpp>

namespace Fluid {

  /// IO node that has one ring buffer.

template <class T>
class IONodeBuffer : public IONode, public MJ_RingBufferDelay<T>
{
public:
  IONodeBuffer() : m_sampleRate(1.0), m_latestTimeStamp(0) {}
  IONodeBuffer(unsigned size) 
    : MJ_RingBufferDelay<T>(size), m_sampleRate(1.0) {}
  virtual ~IONodeBuffer() {}

  void setSampleRate(float f) { m_sampleRate = f; }
  float getSampleRate() const { return m_sampleRate; }
    
  inline TimeStamp   getTimeStamp(long) const;
  TimeStamp   getLatestTimeStamp() const { return m_latestTimeStamp; }
  void        setLatestTimeStamp(const TimeStamp &t) { m_latestTimeStamp = t; }

  inline long timeStampToIndex(const TimeStamp &t) const;

  inline void resizeMemory(float seconds, unsigned extra = 0);

  template <class S>
  void copyParameters(const IONodeBuffer<S> &that)
  { 
    m_sampleRate = that.getSampleRate(); 
    m_latestTimeStamp = that.getLatestTimeStamp();
    resize(that.size());
    setSampleCount(that.getSampleCount());
  }

  template <class S>
  void copyState(const IONodeBuffer<S> &that)
  { 
    m_sampleRate = that.getSampleRate(); 
    m_latestTimeStamp = that.getLatestTimeStamp();
  }

protected:
  float m_sampleRate;

  TimeStamp m_latestTimeStamp;        /// Timestamp of latest sample
};

  
  template <class T>
  TimeStamp   IONodeBuffer<T>::getTimeStamp(long i) const
  {
    long d = (long) getSampleCount() - i;
    TimeStamp::stamp_t tslast = m_latestTimeStamp.stamp();
    
    TimeStamp::stamp_t delta = (TimeStamp::stamp_t) 
      (((float) ((uint) -1)) * (float) d / m_sampleRate);

  // if(delta > );

  return (tslast - delta);

  }

  template <class T>
  long IONodeBuffer<T>::timeStampToIndex(const TimeStamp &t) const
  {
    long diff = (long) (m_latestTimeStamp - t).stamp();
    if(diff < 0) diff = 0;
    return (long) getSampleCount() - 
      (long) (TimeStamp(diff).secondsF() * m_sampleRate);
  }

  template <class T>
  void IONodeBuffer<T>::resizeMemory(float seconds, unsigned extra)
  { 
    unsigned s = (unsigned) (m_sampleRate * seconds) + extra; 
    if(!s) s = 1;
    resize(s);
  }


}

#endif
