/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, Janne Kontkanen 1999-2002.
 * Tommi.Ilmonen@hut.fi, Janne.Kontkanen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */
#ifndef FLUID_INPUTDEVICE_HPP
#define FLUID_INPUTDEVICE_HPP

#include <Fluid/ConfigReader.hpp>
#include <Fluid/DataSource.hpp>
#include <Fluid/Debug.hpp>
#include <Fluid/InputDeviceBase.hpp>
#include <Fluid/IONodeBuffer.hpp>

#include <string>
#include <assert.h>
#include <unistd.h>

namespace Fluid {

  /**
   * Generic input device class
   *
   * Even though this class does not define pure virtual members,
   * it is not directly instantiable because of getClassName() abstract 
   * function of InputDeviceBase
   */

template <class Sample>
class InputDevice : public InputDeviceBase
{
public:
              InputDevice(Chunk* chunk, DataSource<Sample> *dataSource);

  virtual IONode *    ioGetOutput(const char *id)
  { return (strcmp(id, "output") == 0) ? & history : 0; }
  virtual std::string ioGetOutputName(unsigned i)
  { return i ? "" : "output"; }

  void        requireHistory(unsigned int n);
  unsigned    historySize() { return history.size(); }
  void        update();
  void        start();
  void        stop();
  bool        isRunning() const { return running; }
  float       getSampleRate() const { assert(dataSource); return dataSource->getSampleRate(); }
  ulong        getLatestIndex() const { return history.getSampleCount(); }
  TimeStamp   getLatestTimeStamp() const 
  { return history.getLatestTimeStamp(); }
  long        getHistorySize() const { return (long) history.size(); }
  const Sample&     getLatestSample() const 
  { return history[getLatestIndex()]; }
  Sample&     getSample(ulong n)  { return history.getIndex(n); }
  Sample&     getDelayedSample(int delay)  
  { return history.getNewest(delay); }
  virtual std::string getDriverName() const { assert(dataSource); return dataSource->getDriverName(); }
  virtual std::string getName() const { return name; }

  virtual void        setDumpFile(const std::string& fname);

  IONodeBuffer<Sample> & outputBuffer() { return history; }

private:
  DataSource<Sample> *dataSource;   /// The input DataSource
  std::string name;                 /// Name of the instance
  IONodeBuffer<Sample> history;     /// History buffer
  bool running;
};

template <class Sample>
inline InputDevice<Sample>::InputDevice(Chunk* chunk, DataSource<Sample> *dataSource)
  :dataSource(dataSource),
   name(chunk->get("name")),
   history(1),
   running(false)
{
} 


template <class Sample>
inline void InputDevice<Sample>::requireHistory(unsigned int n)
{
  if(history.size() < n) history.resize(n);
}

template <class Sample>
inline void InputDevice<Sample>::update()
{
  dataSource->lock();

  if(dataSource->getAvailable() > 0) 
    history.setLatestTimeStamp(dataSource->getLatestTimeStamp());

  while(dataSource->getAvailable())
    {
      history.advance();
      dataSource->read(history.getNewest());
    }
  dataSource->release();
}

template <class Sample>
inline void InputDevice<Sample>::start()
{
  assert(dataSource);

  if(running) return;

  history.resize(dataSource->bufferSize());

  history.setSampleCount(0);
  history.setLatestTimeStamp(0);

  dataSource->start();
  int c = 300;
  while(!dataSource->getAvailable() && c--)  
    {
      usleep(10000);
    }
  if(dataSource->getAvailable())
    update();
  else
    {
      errStream << "FLUID # WARNING: Device doesn't seem to give out any data\n"; 
      errStream << "FLUID # WARNING: ...waited 3 seconds in InputDevice::start()\n";
    }
  running = true;

  history.setSampleRate(dataSource->getSampleRate());
}

template <class Sample>
inline void InputDevice<Sample>::stop()
{
  assert(dataSource);
  if(!running) return;
  dataSource->stop();
  running = false;
}

template <class Sample>
inline void InputDevice<Sample>::setDumpFile(const std::string& fname)
{
  assert(dataSource);
  dataSource->setDumpFile(fname);
}


} // namespace Fluid

#endif
