/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, Janne Kontkanen 1999-2002.
 * Tommi.Ilmonen@hut.fi, Janne.Kontkanen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#include <Fluid/InputDevices.hpp>
#include <Fluid/Config.hpp>
#include <Fluid/DeviceFactory.hpp>
#include <Fluid/Debug.hpp>
#include <Fluid/Error.hpp>
#include <Fluid/Directory.hpp>
#include <Fluid/Napra.hpp>
#include <Fluid/TimeStamp.hpp>
#include <Fluid/Tracker.hpp>

#include <string>
#include <dlfcn.h>
#include <iostream>
#include <algorithm>

namespace Fluid {

InputDevices* InputDevices::instance = 0;
  

  InputDevices::InputDevices()
    : m_latestStamp(0),
      m_timeShift(0)
  {}

  /**
 * Find a glove of give name.
 * 
 * In addition to just calling InputDevices::getInputDevice(name),
 * this function checks that the instance that is found is of class
 * "Glove". If no glove of given name is found, NULL is returned.
 * 
 *  @param  name of the glove
 *  @return pointer to an instance of Glove or NULL 
 */
  
  Glove*   InputDevices::getGlove(const std::string& name)
  {
    InputDeviceBase *device = getInputDevice(name);
    if(device && device->getClassName() == "Glove")
      return (Glove*) device;
    else
      return NULL; 
  }

  /**
 * Find a mouse of give name.
 * 
 * In addition to just calling InputDevices::getInputDevice(name),
 * this function checks that the instance that is found is of class
 * "Mouse". If no mouse of given name is found, NULL is returned.
 * 
 *  @param  name of the mouse
 *  @return pointer to an instance of Mouse or NULL 
 */
 
  Mouse*   InputDevices::getMouse(const std::string& name)
  {
    InputDeviceBase *device = getInputDevice(name);
    if(device && device->getClassName() == "Mouse")
      return (Mouse*) device;
    else
      return NULL;
  }
  
  /**
 * Find a napra of give name.
 * 
 * In addition to just calling InputDevices::getInputDevice(name),
 * this function checks that the instance that is found is of class
 * "Napra". If no napra of given name is found, NULL is returned.
 * 
 *  @param  name of the napra
 *  @return pointer to an instance of Napra or NULL 
 */
 
  Napra*   InputDevices::getNapra(const std::string& name)
  {
    InputDeviceBase *device = getInputDevice(name);

    if(!device) return 0;
    return dynamic_cast<Napra*> (device);
  }
  
  /**
 * Find a tracker of give name.
 * 
 * In addition to just calling InputDevices::getInputDevice(name),
 * this function checks that the instance that is found is of class
 * "Tracker". If no tracker of given name is found, NULL is returned.
 * 
 *  @param  name of the tracker
 *  @return pointer to an instance of Tracker or NULL 
 */
  
  Tracker* InputDevices::getTracker(const std::string& name)
  {
    InputDeviceBase *device = getInputDevice(name);
    if(device)
      return dynamic_cast<Tracker*> (device);
    else
      return NULL;
  }

  /**
 * This function scans the device driver 
 * directory and loads all the shared libraries
 * that it can find from there. It also creates the
 * devices that correspond to current configuration
 * (see Config).
 *
 * This function must be called in order to use 
 * any input devices.
 *
 * If InputDevices-singleton is not created when this
 * function is called it will be created automatically.
 */

  void InputDevices::init()
  {
    ensure();
    char *rootdir = getenv("FLUID_ROOT");
    if(rootdir == 0)
      {
	errStream << 
	  "FLUID ERROR # FLUID_ROOT environment variable must\n"
	  "be defined to point to root directory\n"
	  "of fluid library\n";
	exit(0);
      }

    char *platform = getenv("FLUID_PLATFORM");
    if(platform == 0)
      {
	errStream << "FLUID ERROR # FLUID_PLATFORM environment variable must\n"
	  "be defined\n";
	exit(0);
      }

    std::string path = std::string(rootdir) + "/plugins/" + platform + "/";
    Directory directory(path);
    logStream << "FLUID # Loading driver plugins from " << path << "\n\n";
    for(int i = 0; i < directory.getFileCount(); i++) {
      std::string filename = directory.getFileName(i);
      if(filename.size() > 3 && filename.substr(filename.size()-3,filename.size()-1) == ".so") {
	logStream << "FLUID # Loading driver plugin: " << filename << '\n';
	void *h = dlopen((path + directory.getFileName(i)).c_str(),RTLD_NOW);
	if(!h)
	  {
	    errStream << "FLUID ERROR # Bad driver plugin. Failed to load " << filename << ", because: " << dlerror() << std::endl;
	  }
	else
	  {
	    typedef DeviceFactory* (*FPTR)();
	    FPTR fptr;
	    fptr = (FPTR) dlsym(h,"getFactory");
	    if(fptr == 0)
	      {
		errStream << "FLUID ERROR # Bad driver plugin: " << filename << " (getFactory function not found)\n";
	      }
	    else
	      {
		DeviceFactory *factory = (*fptr)();
		if(factory)
		  {
		    addDevices(factory->getDriverName(),factory);	      
		    logStream << "FLUID # Plugin " << filename  << " driver: "
			      << factory->getDriverName() 
			      << " successfully loaded\n\n";
		  }
	      
	      }
	  }
      }
    }
    logStream << "FLUID # Available Devices\n";
    logStream << "FLUID # *****************\n";
    for(std::map<std::string,InputDeviceBase*>::iterator i = instance->devices.begin(); i != instance->devices.end(); i++)
      {
	logStream << "FLUID # " << (*i).second->getName() << " ( Class: " << (*i).second->getClassName() << ", Driver: " << (*i).second->getDriverName() << " )\n";
      }
    logStream << std::endl;
    DeviceConfig::setStartupTimeStamp(TimeStamp:: getTimeOfDayStamp());
  }

  /**
 * Deallocates the InputDevices singleton object
 */

  void InputDevices::cleanup()
  {
    /* This should probably delete the devices as well */
    logStream << "FLUID # InputDevices::cleanup() called" << std::endl;
    assert(instance);
    delete instance;
  }

  /**
 * Queries for the input device of given name
 *
 * If there is a matching input device subscribed
 * into the system it is returned, otherwise NULL
 */

  InputDeviceBase* InputDevices::getInputDevice(const std::string& name)
  {
    if(!instance)
      {
	errStream << "FLUID ERROR # getInputDevice() called before InputDevices::init()" << std::endl;
	return NULL;
      }
    using namespace std;
    map<string,InputDeviceBase*>::iterator i = instance->devices.find(name);
    if(i != instance->devices.end()) return (*i).second;
    else return NULL;
  }


  /**
 * Adds new devices to collection
 *
 * This function is called by device drivers or applications that implement
 * custom data sources. If InputDevices-singleton is not created when this
 * function is called, it will be created automatically.
 *
 * @param driver = name of the driver (for example "5DTGlove","QTMouse"...)
 * @param factory = device factory capable of creating devices
 */

  void InputDevices::addDevices(const std::string& driver, DeviceFactory *factory)
  {
    using namespace std;
    ensure();

    Config * conf = DeviceConfig::config();

    for(Config::iterator it = conf->find(driver); it != conf->end(); it++) {

      if((*it).first != driver)
	return;
      
      Chunk *chunk = & Config::getType(it);

      InputDeviceBase *device = factory->create(chunk);

      if(device) {

	std::string deviceName = chunk->get("name");
	if(!deviceName.size())
	  Error("A device configuration with no name: %s", driver.c_str());
	
	std::string dumpFile = chunk->get("dumpFile");
	if(dumpFile.size()) {
	  device->setDumpFile(dumpFile);
	}
	
	instance->devices.insert(pair<string,InputDeviceBase*>
				 (deviceName,device));	  
      }
    }
  }

  /**
 * Calls update() for all the InputDevices that are up
 * and running (InputDevice::start() has been called)
 */

  void InputDevices::update(const TimeStamp *ts)
  {
    assert(instance);

    if(ts)
      instance->m_latestStamp = *ts;
    else 
      instance->m_latestStamp.getTimeOfDay();

    instance->m_condition.broadcast2(instance->m_mutex);

    for(std::map<std::string,InputDeviceBase*>::iterator i = instance->devices.begin(); i != instance->devices.end(); i++)
      {
	if(i->second->isRunning()) i->second->update();
      }
  }

  /**
 * Calls start() for all the InputDevices that are available, but
 * not yet started.
 *
 * NOTE: If you use this function, you should have only those devices
 * in your configuration file that you really wish to use (The other
 * option is to start&stop each input device you are using explicitly)
 */

  void InputDevices::start(const TimeStamp *ts)
  {
    assert(instance);

    if(ts)
      instance->m_latestStamp = *ts;
    else 
      instance->m_latestStamp.getTimeOfDay();

    for(std::map<std::string,InputDeviceBase*>::iterator i = 
	  instance->devices.begin(); i != instance->devices.end(); i++)
      {
	if(!i->second->isRunning()) i->second->start();
      } 
  }

  /**
 * Calls stop() for all the InputDevices that are available, but not 
 * yet started.
 * 
 * NOTE: If you use this function, you should have only those devices
 * in your configuration file that you really wish to use (The other
 * option is to start&stop each input device you are using explicitly)
 */

  void InputDevices::stop()
  {
    assert(instance);
    for(std::map<std::string,InputDeviceBase*>::iterator i = 
	  instance->devices.begin(); i != instance->devices.end(); i++)
      {
	if(i->second->isRunning()) i->second->stop();
      }  
  }

  /** Wait until the main thread has consumed the given time-stamp. This
    method should only be used by device threads that need to wait
    until they can deliver their data. In practice this means the
    file-reader data sources. */

  void InputDevices::threadWaitUntil(const TimeStamp &ts)
  {
    assert(instance);

    instance->m_mutex.lock();

    while(instance->m_latestStamp <= ts)
      instance->m_condition.wait(instance->m_mutex);

    instance->m_mutex.unlock();
  
    /* Trace("InputDevices::threadWaitUntil # %f", 
       (ts - instance->m_latestStamp).subSeconds());*/
  }


} // namespace

