/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#ifndef FLUID_KEYBOARD_HPP
#define FLUID_KEYBOARD_HPP

#include <Fluid/InputDevice.hpp>
#include <Fluid/TimeStamp.hpp>

namespace Fluid {

  /** Key codes related to normal keyboard usage. The enums are
      indexed in a way that should make it easy to add stuff in
      future. Keys that belong logically together are allocatated a
      contiguous index area. The areas are separated by a safe
      margin.  */

  enum Key {
    KEY_NONE = 0,

    // Letters:
    KEY_A = 1,
    KEY_B,
    KEY_C,
    KEY_D,
    KEY_E,
    KEY_F,
    KEY_G,
    KEY_H,
    KEY_I,
    KEY_J,
    KEY_K,
    KEY_L,
    KEY_M,
    KEY_N,
    KEY_O,
    KEY_P,
    KEY_Q,
    KEY_R,
    KEY_S,
    KEY_T,
    KEY_U,
    KEY_V,
    KEY_W,
    KEY_X,
    KEY_Y,
    KEY_Z,

    // Numbers:
    KEY_0 = 100,
    KEY_1,
    KEY_2,
    KEY_3,
    KEY_4,
    KEY_5,
    KEY_6,
    KEY_7,
    KEY_8,
    KEY_9,

    // Scapes:
    KEY_SPACE = 110,
    KEY_TAB,
    KEY_ENTER,

    // Cursor motion:
    KEY_LEFT = 120,
    KEY_RIGHT,
    KEY_UP,
    KEY_DOWN,
    KEY_HOME,
    KEY_END,
    KEY_PAGE_UP,
    KEY_PAGE_DOWN,
    KEY_INSERT,
    KEY_DELETE,
    KEY_BACKSPACE,
    KEY_ESCAPE,

    // Parentheses, braces and brackets:
    KEY_LEFT_PAREN = 150,
    KEY_RIGHT_PAREN,
    KEY_LEFT_BRACE,
    KEY_RIGHT_BRACE,
    KEY_LEFT_BRACKET,
    KEY_RIGHT_BRACKET,

    // Mathematical:
    KEY_LESS = 170,
    KEY_EQUAL,
    KEY_GREATER,
    KEY_PLUS,
    KEY_MINUS,
    
    // Special characters:
    KEY_ASTERISK = 200, // "*"
    KEY_SLASH,      // "/"
    KEY_BACKSLASH,  // "\"
    KEY_QUESTION,   // "?"
    KEY_EXCLAM,     // "!"
    KEY_COMMA,      // ","
    KEY_PERIOD,     // "."
    KEY_COLON,      // ":"
    KEY_SEMICOLON,  // ";"
    KEY_UNDERSCORE, // "_"
    KEY_QUOTE,      // "
    KEY_DOLLAR,     // $
    KEY_AT,         // @
    KEY_TILDE,      // ~
    KEY_PERCENT,    // %
    KEY_AMPERSAND,  // &
    KEY_APOSTROPHE, // '
    KEY_NUMBERSIGN, // #
    KEY_CIRCUM,     // ^
    KEY_BAR,        // |
    /// The last key, seldom useful.
    KEY_FINAL
  };

  enum Modifier 
  {
    MOD_CONTROL   = 0x01,
    MOD_SHIFT     = 0x02,
    MOD_CAPS_LOCK = 0x04,
    MOD_ALT       = 0x08,
    MOD_ALTGR     = 0x10
  };
  
  char KeyToChar(Key k, int modifiers = 0);

  class KeyboardSample
  {
  public:
    KeyboardSample() { clear(); }
    ~KeyboardSample() {}

    void       clear() { time = 0; key = KEY_NONE; modifiers = 0; }
    void       read(DI_DataBuffer2 &buffer);
    void       write(DI_DataBuffer2 &buffer) const;
    int        getMaxSerializedSize() const
    { return sizeof(TimeStamp) + 1; }
  

    TimeStamp time;

    Key key;
    int modifiers;

  };


  class Keyboard : public InputDevice<KeyboardSample>
  {
  public:

    Keyboard(Chunk* chunk, DataSource<KeyboardSample>* source) 
      : InputDevice<KeyboardSample>(chunk, source){}

    virtual ~Keyboard();

    virtual std::string   getClassName() const;

  };

  template <>
  inline TimeStamp IONodeBuffer<KeyboardSample>::getTimeStamp(long i) const
  {
    return getIndex(i).time;
  }

} // namespace Fluid

#endif
