/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, Janne Kontkanen 1999-2002.
 * Tommi.Ilmonen@hut.fi, Janne.Kontkanen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#ifndef FLUID_MATRIX2T_HPP
#define FLUID_MATRIX2T_HPP

#include <Fluid/Math.hpp>
#include <Fluid/Vector2.hpp>
#include <assert.h>

namespace Fluid {

template <class T>
class Matrix2T
{
public:
  Matrix2T() {}
  Matrix2T(T v11, T v12, T v21, T v22) { m[0].make(v11, v12); m[1].make(v21, v22); }
  Matrix2T(const Matrix2T &that) { m[0] = that.m[0]; m[1] = that.m[1]; }

  ~Matrix2T() {}

  Vector2T<T>&       row(int i)             { return m[i]; }
  const Vector2T<T>& row(int i) const       { return m[i]; }
  Vector2T<T>        column(int i) const    { return Vector2T<T>(m[0][i],m[1][i]); }
  Vector2T<T>&       operator[](int i)      { return row(i); }
  const Vector2T<T>& operator[](int i) const{ return row(i); }
  void               set(int r, int c, T v) { m[r][c] = v; }
  T                  get(int r, int c) const{ return m[r][c]; }
  T *                data() { return m[0].data(); }
  const T *           data() const { return m[0].data(); }

  void               transpose()            { swap(m[0][1],m[1][0]); }
  void               clear()                { m[0].clear(); m[1].clear(); } 
  void               identity()             { m[0].make(1.0, 0.0); m[1].make(0.0, 1.0);}
  void               rotate(T);

  static int         rows() { return 2; }
  static int         columns() { return 2; }
private:
  static void swap(T &a, T& b);
  Vector2T<T> m[2];
};

template <class T>
inline void Matrix2T<T>::swap(T &a, T& b)
{
  T t = a;
  a = b;
  b = t;
}

template <class T>
inline void Matrix2T<T>::rotate(T a)
{
  T ca = Math::Cos(a);
  T sa = Math::Sin(a);
  m[0].make(ca, -sa);
  m[1].make(sa, ca);
}

template <class T>
inline Matrix2T<T> operator*(const Matrix2T<T>& m1, const Matrix2T<T>& m2)
{
  Matrix2T<T> res;

  for(int i = 0; i < 2; i++) {
    Vector2T<T> t = m2.column(i);
    res[0][i] = dot(m1.row(0),t);
    res[1][i] = dot(m1.row(1),t);
  }

  return res;
}

template <class T>
inline Vector2T<T> operator*(const Matrix2T<T>& m1,const Vector2T<T>& m2)
{
  Vector2T<T> res;
  for(int i = 0; i < 2; i++)
    res[i] = dot(m1.row(i),m2);
  return res;
}

template <class T>
inline Vector2T<T> operator*(const Vector2T<T>& m2, const Matrix2T<T>& m1)
{
  Vector2T<T> res;
  for(int i = 0; i < 2; i++)
    res[i] = dot(m1.column(i),m2);
  return res;
}

typedef Matrix2T<float> Matrix2;

}

#endif

