/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, Janne Kontkanen 1999-2002.
 * Tommi.Ilmonen@hut.fi, Janne.Kontkanen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */
#ifndef FLUID_MOUSE_HPP
#define FLUID_MOUSE_HPP

#include "Fluid/InputDevice.hpp"
#include <string>
#include <iostream>
#include <assert.h>

class DI_DataBuffer2;

namespace Fluid {

  class MouseSample  
  {
  public:
    static const int BUTTON_NUMBER = 3;

    void       clear();
    void       read(DI_DataBuffer2 &buffer);
    void       write(DI_DataBuffer2 &buffer) const;
    int        getMaxSerializedSize() const;
    bool       buttonPressed(uint n) const { return state & (1 << n); }

    unsigned int state;
    int x,y, xmotion, ymotion;
  };


  /**
   * Mouse input device class.
   *
   * Actual mouse device driver classes are 
   * derived from this class, but this should be
   * transparent to application programmer
   */

class Mouse : public InputDevice<MouseSample>
{
public:

  /**
   * Mouse button enumeration
   */

  enum Button
    {
      BUTTON_LEFT,
      BUTTON_MIDDLE,
      BUTTON_RIGHT,
      BUTTONS
    };

  /**
   * Constructor - shouldn't be called by application programmer
   *
   * This function is called by system when devices are created
   * in the initialization phase of the system
   */
                        Mouse(Chunk* chunk, DataSource<MouseSample>* dataSource);

  /**
   * Returns the state of given button at given sample 
   *
   * Note that samples are numbered:
   *
   * 0...current
   *
   * So the first sample ever is zero and the current sample
   * index can be queried by getLatestIndex() function
   *
   * Note also that only samples in interval:
   *
   * [current-required history,current]
   *
   * Are available
   *
   * @param button = BUTTON_LEFT/BUTTON_MIDDLE/BUTTON_RIGHT
   * @param n = sample index
   * @returns boolean (true if button is down, false otherwise)
   */
  bool                  getButtonState(Button button, ulong n);

  /**
   * Returns the x coordinate of the mouse at given sample
   *
   * @param n = sample index
   * @see getButtonState() about sample numbering
   */
  int                   getX(ulong n);

  /**
   * Returns the y coordinate of the mouse at given sample
   *
   * @param n = sample index
   * @see getButtonState() about sample numbering
   */
  int                   getY(ulong n);

  /** 
   * Returns the state of given button at latest sample
   *
   * @param button = BUTTON_LEFT/BUTTON_MIDDLE/BUTTON_RIGHT
   * @returns true if button pressed false otherwise
   */
  bool                  getButtonState(Button button) 
  { 
    return getButtonState(button,getLatestIndex()); 
  }

  /// Returns the x coordinate of the mouse at latest sample
  int                   getX() { return getX(getLatestIndex()); }
  /// Returns the y coordinate of the mouse at latest sample
  int                   getY() { return getY(getLatestIndex()); }

  virtual std::string   getClassName() const;
};

} // namespace

#endif
