/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#include <Fluid/SerialPort.hpp>

#include <Fluid/Error.hpp>

#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <termios.h>

namespace Fluid
{

  SerialPort::SerialPort()
    : m_fd(-1)
  {}

  SerialPort::~SerialPort()
  {
    close();
  }
  
  bool SerialPort::open(const char * device, bool stopBit, bool parityBit,
			int baud, int bits, int waitBytes, int waitTimeUS)
  {
    close();

    const char * fname = "SerialPort::open";

    m_fd = ::open(device, O_RDWR);
    
    if(m_fd <= 0) {
      Error("%s # Failed to open \"%s\" (%s)", 
	    fname, device, strerror(errno));
      errno = 0;
      return false;
    }

    struct termio sttyArgs;
    ioctl(m_fd, TCGETA , &sttyArgs);    

    if(stopBit)
      sttyArgs.c_cflag &= ~CSTOPB;	/* One stop bit */
    else
      sttyArgs.c_cflag |= CSTOPB;

    if(parityBit)
      sttyArgs.c_cflag |= PARENB;
    else
      sttyArgs.c_cflag &= ~PARENB;

    if(bits == 5)
      sttyArgs.c_cflag |= CSIZE & CS5;
    else if(bits == 6)
      sttyArgs.c_cflag |= CSIZE & CS6;
    else if(bits == 7)
      sttyArgs.c_cflag |= CSIZE & CS7;
    else if(bits == 8)
      sttyArgs.c_cflag |= CSIZE & CS8;

    // Clear bauds:
    sttyArgs.c_cflag &= ~CBAUD;

    // Set bauds:
    if(baud == 1200)
      sttyArgs.c_cflag |=  B1200; 
    else if(baud == 2400)
      sttyArgs.c_cflag |=  B2400; 
    else if(baud == 4800)
      sttyArgs.c_cflag |=  B4800; 
    else if(baud == 9600)
      sttyArgs.c_cflag |=  B9600; 
    else if(baud == 19200)
      sttyArgs.c_cflag |=  B19200; 
    else if(baud == 38400)
      sttyArgs.c_cflag |=  B38400; 
    else if(baud == 57600)
      sttyArgs.c_cflag |=  B57600; 
    else if(baud == 115200)
      sttyArgs.c_cflag |=  B115200; 
#ifdef B230400
    else if(baud == 230400)
      sttyArgs.c_cflag |=  B230400; 
#endif

    sttyArgs.c_cflag |= CREAD;	/* Allow reading */
    sttyArgs.c_cflag |= CLOCAL;	/* No modem between us and device */
    
    sttyArgs.c_cc[VMIN] = waitBytes;
    sttyArgs.c_cc[VTIME] = waitTimeUS / 100000; /* Unit = 0.1 s*/
    
    sttyArgs.c_iflag = (IGNBRK | IGNPAR);  /* Ignore break & parity errs */
    sttyArgs.c_oflag = 0;                  /* Raw output, leave tabs alone */
    sttyArgs.c_lflag = 0;              /* Raw input (no KILL, etc.), no echo */

    if(ioctl(m_fd, TCSETA, &sttyArgs) < 0) {
      Error("%s # Failed to set TTY parameters (%s)", fname, strerror(errno));
      errno = 0;
      return false;
    }

    return true;
  }

  bool SerialPort::close()
  {
    if(m_fd < 0)
      return false;

    ::close(m_fd);

    m_fd = -1;

    return true;
  }
  
}
