/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, Janne Kontkanen 1999-2002.
 * Tommi.Ilmonen@hut.fi, Janne.Kontkanen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#ifndef FLUID_SOCKETDATASOURCE_HPP
#define FLUID_SOCKETDATASOURCE_HPP

#include <Fluid/DataSource.hpp>
#include <Fluid/Exec.hpp>

#include <di_socket.h>
#include <di_sleep.h>
#include <string>

namespace Fluid
{

  template <class Sample>
  class SocketDataSource : public StreamDataSource<Sample>
  {
  public:
    SocketDataSource(const std::string& address, 
		     const std::string& launcher,
		     int bufferSize,
		     float sampleRate);
    ~SocketDataSource();
    void                get(Sample& s);
    std::string         getAddress() const { return address; }
    virtual std::string getDriverName() const { return "Remote"; } 
    virtual float getSampleRate() const { return sampleRate; }
  private:
    void open();

    DI_Socket      socket;
    DI_DataBuffer2 buffer;
    std::string    address;
    std::string    m_launcher;
    bool           ok;
    int            failures;
    float          sampleRate;
  };


  template <class Sample>
  inline SocketDataSource<Sample>::SocketDataSource
  (const std::string& address,
   const std::string& launcher, 
   int bufferSize,
   float sampleRate)
    : StreamDataSource<Sample>(bufferSize),
      buffer(2048),
      address(address),
      m_launcher(launcher),
      ok(false), /* Kludge! */
      failures(0),
      sampleRate(sampleRate)
  {
  }

  /** Opens socket connection to the server. If the server is not
      responding, and launcher script is define, then an attemp is
      made to start the server.  */

  template <class Sample>
  inline void SocketDataSource<Sample>::open()
  {

    if(failures > 10)
      return;

    if(!ok && socket.open(address.c_str())) {
      Trace("Opened connection to: %s", address.c_str());
      ok = true;
    }
    else {
      Error("Failed to open connection to: %s", address.c_str());
      ok = false;
      
      failures++;
      
      if(m_launcher.size()) {
	if(failures == 1) {
	  Trace("Spawning server script [%s]", m_launcher.c_str());
	  ExecuteParallel(m_launcher.c_str());
	  DI_Sleep::sleepS(1);
	  open();
	}
	else 
	  DI_Sleep::sleepS(1);
      }
    }
  }

  template <class Sample>
  inline SocketDataSource<Sample>::~SocketDataSource()
  {
    socket.close();
  }

  template <class Sample>
  inline void SocketDataSource<Sample>::get(Sample &s)
  {
    if(!ok) open();

    if(DI_readFromSocket(&socket,&buffer) < 0) 
      {
	errStream << "FLUID ERROR # SocketDataSource::get(): read error" << std::endl;
	s.clear();
      }
    else
      {
	buffer.rewind();
	s.read(buffer);
      }
  }

} // namespace

#endif // SOCKETDATASOURCE_HPP
