/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#include <Fluid/Tracker.hpp>
#include <Fluid/ConfigReader.hpp>
#include <Fluid/Error.hpp>

#include <di_data_buffer.h>

#include <sstream>

#include <mj_ringbuffertmpl.h>

namespace Fluid
{

  Tracker::Tracker(Chunk* chunk, DataSource<Sample>* source)
    : InputDevice<Sample>(chunk,source),
    m_sensors(0)
  {
    for(int i = 0; i < 100; i++)
      {
	std::ostringstream os;
	os << "sensor" << i;
	std::string sensorName = (std::string) chunk->get(os.str());
	sensorMap[sensorName] = i;
      }
  }

  int Tracker::getSensorCount() const
  {
    if(!isRunning())
      {
	errStream << "FLUID # WARNING: Tracker device must be running to figure out the number of sensors\n";
	errStream << "FLUID # WARNING: Did you forget to call start()?\n";
	return 0;
      }
    else
      {
	int sensorCount = getLatestSample().getSensorCount();
	if(sensorCount < 1)
	  {
	    errStream << "FLUID # WARNING: Tracker doesn't seem to have any sensors (Tracker::getSensorCount())???\n";
	    return 0;
	  }
	else
	  {
	    return sensorCount;
	  }
      }
  }

  void Tracker::start()
  {
    InputDevice<TrackerSample>::start();

    if(isRunning()) {
      if(getLatestIndex() > 1) {

	m_sensors = getSample(1).getSensorCount();

	Trace("Tracker::start # resizing all sensors to %d", m_sensors);

	for(uint i=0; i < historySize(); i++) {
	  getSample(i).setSensorCount(m_sensors);
	}
      }
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  void TrackerSample::clear()
  {
    for(int i = 0; i < sensors; i++)
      {
	rotation[i].clear();
	location[i].clear();
      }
  }

  TrackerSample::TrackerSample(int sensors)
    :sensors(sensors)
  {
    if(sensors) {
      rotation = new Matrix3[sensors];
      location = new Vector3[sensors];
    }
    else {
      rotation = 0;
      location = 0;
    }
  }

  TrackerSample::TrackerSample(const TrackerSample& that)
    :sensors(that.sensors)
  {
    rotation = new Matrix3[sensors];
    location = new Vector3[sensors];
    for(int i = 0; i < sensors; i++) {
      rotation[i] = that.rotation[i];
      location[i] = that.location[i];
    }
  }

  TrackerSample& TrackerSample::operator=(const TrackerSample& that)
  {
    if(this != &that) {
      if(sensors != that.sensors)
	setSensorCount(that.sensors);
      for(int i = 0; i < sensors; i++) {
	rotation[i] = that.rotation[i];
	location[i] = that.location[i];
      }
    }
    return *this;
  }

  TrackerSample::~TrackerSample()
  {
    delete [] rotation;
    delete [] location;
  }

  void TrackerSample::setSensorCount(int sensors)
  {
    // Trace("TrackerSample::setSensorCount # %d", sensors);
  
    if(this->sensors != sensors) {
      delete [] rotation;
      delete [] location;
      this->sensors = sensors;
      
      if(sensors) {
	rotation = new Matrix3[sensors];
	location = new Vector3[sensors];
      }
      else {
	rotation = 0;
	location = 0;
      }
    }
  }

  void TrackerSample::read(DI_DataBuffer2& buffer)
  {

    int s = buffer.readInt32();

    // int ind = buffer.position();
    // Trace("TrackerSample::read # %d sensors, index %d", s, ind);

    setSensorCount(s);
    for(int s = 0; s < sensors; s++)
      {
	for(int i = 0; i < 9; i++) rotation[s][i/3][i%3] = buffer.readFloat32();
	for(int i = 0; i < 3; i++) location[s][i] = buffer.readFloat32();
      }

    // Trace("TrackerSample::read # EXIT");
  }

  void TrackerSample::write(DI_DataBuffer2& buffer) const
  {
    buffer.writeInt32(sensors);
    for(int s = 0;  s < sensors; s++)
      {
	for(int i = 0; i < 9; i++) buffer.writeFloat32(rotation[s][i/3][i%3]);
	for(int i = 0; i < 3; i++) buffer.writeFloat32(location[s][i]);
      }
  }

  int TrackerSample::getMaxSerializedSize() const
  {
    return 4+48*sensors+4; // four for padding
  }

  const Matrix3& TrackerSample::getRotation(int i) const
  {
    assert(i >= 0 && i < sensors);
    // assert(rotation);
    // static Matrix3 def(Vector3(0.0,0.0,0.0),Vector3(0.0,0.0,0.0),Vector3(0.0,0.0,0.0));
    // if(i >= 0 && i < sensors)
    return rotation[i];
    // else
    // return def;
  }

  const Vector3& TrackerSample::getLocation(int i) const 
  { 
    if((uint) i >= (uint) sensors)
      Fatal("TrackerSample::getLocation # ARRAY EXCEEDED # %u >= %u",
	    (uint) i, (uint) sensors);

    // assert(location);
    // static Vector3 def(0.0f,0.0f,0.0f);
    // if(i >= 0 && i < sensors)
    return location[i]; 
    // else 
    // return def;
  }


  void TrackerSample::setRotation(int i, const Matrix3& m)
  {
    rotation[i] = m;
  }

  void TrackerSample::setLocation(int i, const Vector3& v)
  {
    location[i] = v;
  }

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////
  
  int Tracker::findSensor(const std::string &name)
  {
    std::map<std::string, int>::iterator it = sensorMap.find(name);

    if(it == sensorMap.end())
      return -1;
    return (*it).second;
  }

}

template class MJ_RingBuffer<Fluid::TrackerSample>;
template class MJ_RingBuffer<Fluid::Vector3>;
template class MJ_RingBuffer<Fluid::Matrix3>;
template class MJ_RingBuffer<float>;

