/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, Janne Kontkanen 1999-2002.
 * Tommi.Ilmonen@hut.fi, Janne.Kontkanen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#ifndef FLUID_TRACKER_HPP
#define FLUID_TRACKER_HPP

#include <Fluid/Matrix3.hpp>
#include <Fluid/Vector3.hpp>
#include <Fluid/InputDevices.hpp>

#include <map>
#include <string>


class DI_DataBuffer2;


namespace Fluid
{

  /**
   * Datatype that represents a single tracker sample.
   *
   * A sample consists of rotation and location
   *
   * @see Tracker,Matrix3,Vector3
   */

  class TrackerSample
  {
  public:
    void       clear();
    void       read(DI_DataBuffer2 &buffer);
    void       write(DI_DataBuffer2 &buffer) const;
    int        getMaxSerializedSize() const;

    TrackerSample():sensors(0),rotation(0),location(0) {}
    TrackerSample(int sensors);
    TrackerSample(const TrackerSample& that);
    TrackerSample& operator=(const TrackerSample& that);
    ~TrackerSample();

    int            getSensorCount()   const { return sensors; }
    void           setSensorCount(int sensors);
    const Matrix3& getRotation(int i) const;
    const Vector3& getLocation(int i) const;
    void           setRotation(int i, const Matrix3& m);
    void           setLocation(int i, const Vector3& v);
  private:
    int sensors;
    Matrix3 *rotation;
    Vector3 *location;
  };
  
  /**
   * Tracker interface class
   *
   * @see TrackerSample
   */

  class Tracker : public InputDevice<TrackerSample>
  {
  public:
    typedef TrackerSample Sample;
    Tracker(Chunk* chunk, DataSource<Sample>* source);
    const Matrix3 &getRotation(int sensor, ulong n) { return getSample(n).getRotation(sensor); }
    const Vector3 &getLocation(int sensor, ulong n) { return getSample(n).getLocation(sensor); }
    const Matrix3 &getRotation(int sensor) { return getRotation(sensor,getLatestIndex()); }
    const Vector3 &getLocation(int sensor) { return getLocation(sensor,getLatestIndex()); }
    std::string getClassName() const { return "Tracker"; }

    int findSensor(const std::string &name);
    int getSensorCount()   const;

    virtual void start();

    void requireHistory(unsigned int n)
    {
      InputDevice<TrackerSample>::requireHistory(n);
      for(unsigned i=0; i < historySize(); i++) {
	getSample(i).setSensorCount(m_sensors);
      }
    }

  protected:
    /// Match from sensor names to their indices
    std::map<std::string, int> sensorMap;
    std::string name;

    unsigned m_sensors;
  };


} // namespace Fluid

#endif //TRACKER_HPP
