/* COPYRIGHT
 *
 * This file is part of the Geometry Engine - GEE
 *
 * Read the "GEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen 2003
 *
 * firstname.lastname@hut.fi
 *
 */

#include <BglProjections.hpp>

#include <bgl.h>

#include <GL/gl.h>

#include <fstream>

#include <strings.h>

BglProjections::BglProjections()
{}

BglProjections::~BglProjections()
{
  // m_cave.dump(std::cout);
}

/** Loads in a cave projection setup and corresponding BGL setup
    file. This method also calls bglInit with the address specified in
    the config file. 

    An example config file is distributed with the bglproj library.
*/

bool BglProjections::init(const char * caveProjFile,
			  const char * bglConfigFile)
{
  const char * fname = "BglProjections::init";

  if(!m_cave.init(caveProjFile))
    return false;

  m_cave.dump(std::cout);

  std::ifstream in;

  in.open(bglConfigFile);

  char buf[256];
  char key[256];
  char wall[256];
  char eye[256];
  
  while(!in.eof() && in.good() && !in.fail()) {
    bzero(buf, 256);
    in.getline(buf, 255); // Never erase trailing end-of-string

    if(buf[0] == '#')
      continue;

    int id = -1;
    int n = sscanf(buf, "%s %d %s %s", key, & id, wall, eye);

    if(strcmp(key, "Address") == 0) {
      int n = sscanf(buf, "%s %s", key, eye);
      if(n == 2) m_address = eye;
      continue;
    }
    
    if(n != 4 || strcmp(key, "Renderer") != 0)
      continue;

    int index = m_cave.getWallIndex(wall);

    if(index < 0) {
      fprintf(stderr, "%s # No wall \"%s\"\n", fname, wall);
      fflush(stderr);
      continue;
    }

    int ieye = -1;

    if(strcmp(eye, "left") == 0)
      ieye = CaveProj::EYE_LEFT;
    else if(strcmp(eye, "right") == 0)
      ieye = CaveProj::EYE_RIGHT;

    if(ieye < 0) {
      fprintf(stderr, "%s # Eye must be \"left\" or \"right\" (got %s)\n", 
	      fname, eye);
      fflush(stderr);
      continue;
    }

    printf("%s # Adding renderer %d with wall %d (%s) %s\n", 
	   fname, id, index, wall, 
	   (ieye == CaveProj::EYE_LEFT) ? "left" : "right");
    fflush(stdout);

    RendererInfo r;
    r.m_id = id;
    r.m_wallName = wall;
    r.m_wall = index;
    r.m_eye = ieye;
    
    m_renderers.push_back(r);
  }

  if(!m_renderers.size())
    return false;

  bglInit(m_address.c_str());
  bglCreateWindow(BGL_NOFRAME);

  /* We have read all config information successfully, now set up
     BGL: */

  bglDeSelectRenderer(-1);
  
  for(uint i = 0 ; i < m_renderers.size(); i++) {
    const RendererInfo & r = m_renderers[i];
    CaveProj::Wall & w = m_cave.getWall(r.m_wall);

    Fluid::Vector2 size = w.getSize();
    Fluid::Vector2 origin = w.getOrigin();

    bglSelectRenderer(r.m_id);
    bglMoveWindow((int) origin.x, (int) origin.y);
    bglResizeWindow((int) size.x, (int) size.y);
    glViewport(0, 0, (int) size.x, (int) size.y);
    bglDeSelectRenderer(r.m_id);

  }

  bglSelectRenderer(-1);

  return true;
}

/** Sets the projection matrices via BGL. After this method is called
    all renderer are listening and have their matrix mode set to
    GL_MODELVIEW. */

void BglProjections::setProjectionMatrices()
{
  bglSelectRenderer(-1);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  bglDeSelectRenderer(-1);
  
  for(uint i = 0 ; i < m_renderers.size(); i++) {
    const RendererInfo & r = m_renderers[i];

    bglSelectRenderer(r.m_id);

    Fluid::Matrix4 m = m_cave.getWall(r.m_wall).matrix(r.m_eye);

    // std::cout << "Renderer " << i << " Matrix = " << m << std::endl;
    
    m.transpose();
    glLoadMatrixf(m.data());

    bglDeSelectRenderer(r.m_id);
    
    // std::cout << i << " Matrix = " << m << std::endl;
  }

  bglSelectRenderer(-1);
  glMatrixMode(GL_MODELVIEW);

  glLoadIdentity();
}
