/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of DIVA base library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000-2001.
 * Jarmo.Hiipakka@hut.fi Tommi.Ilmonen@hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA base library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */


#ifndef DI_DATA_BUFFER_H
#define DI_DATA_BUFFER_H

#ifndef WIN32
#include <netinet/in.h>
#else
#include <winsock.h>
#endif

#include <di_endian.h>

#include <string>

#include <stdio.h>

/// The maximum length of an argument string.
const int DI_MAX_DATA_STRING = 512;

/** Data buffer base class for portable communication.

    This class knows how to read/write data from/to a character buffer
    in network byte order. The buffer contains the length of the data and
    the data itself. The length field is in the beginning of the buffer,
    and it is 4 bytes long. The write and read methods advance an internal
    pointer automatically; all you have to do to create a buffer full of
    data, is to construct the buffer and call the write functions in the
    desired order. Reading the buffer must be done in exactly the same
    order.

    You can send the buffer over, e.g., a socket connection, and the
    receiver will be able to read the data regardless of the byte orders
    of the different hosts.

    This class was designed to be independent of the rest of the 'libdibase'
    library (i.e., the basic library used by the DIVA team), so that it
    can be extracted from the library by copying the relevant files. For
    easy transportation over a socket connection, the 'di_socket.h' header
    file contains helper functions.

    @author Jarmo Hiipakka & Tommi Ilmonen, HUT, 2000-2001
    
*/

class DI_DataBuffer
{
public:

  /**@name Constructors */
  
  //@{
  
  /// Creates a buffer with desired size.

  DI_DataBuffer(/// Data size (in bytes), excluding the length field.
		unsigned int size = 1024);

  virtual ~DI_DataBuffer();

  //@}
  
  /** @name Write methods. */

  //@{

  /// Write a two-byte integer.
  
  void writeInt16(short input) { writeInt32((int)input); }

  /// Write a four-byte integer.
  
  void writeInt32(int input);

  /// Write an eight-byte integer.
  void writeInt64(llong input);

  /// Write a four-byte floating point number.
  
  void writeFloat32(float input);

  /// Write an eight-byte floating point number.
  
  void writeFloat64(double input);

  /// Write a NULL-terminated character string (max length DI_MAX_DATA_STRING).
  
  void writeString(const char * input);

  //@}

  /** @name Read methods. */

  //@{

  /// Read a two-byte integer.
  
  short readInt16() { return (short)readInt32(); }

  /// Read a four-byte integer.
  
  int readInt32();

  /// Read an eight-byte integer.

  llong readInt64();

  /// Read a four-byte floating point number.

  float readFloat32();

  /// Read an eight-byte floating point number.

  double readFloat64();

  /// Read a string.
  
  void readString(char *string);

  //@}

  /** @name Access methods. */

  //@{

  /// Move read/write head.

  inline bool moveIndex(unsigned int i)
  { if(i >= m_size) return false; m_index = i; return true; }

  /// Rewind the read/write head to the beginning.
  inline void rewind() { m_index = 0; if(m_buf) writeInt32(0); }

  /// Get the position of the read/write head

  inline unsigned index() const { return m_index; }

  /** Resize the buffer.

      The buffer must be resized to sufficient length before write can
      take place. The contents of the buffer will be unchanged up to
      the lesser of the new and old sizes.
  */

  bool resize(unsigned int size);

  /// Return a pointer to the internal data buffer.

  char * buffer() const { return m_buf; }

  /// Return the size of the data buffer (in bytes).

  unsigned int size() const { return m_size ? m_size - 4 : 0; }

  /// Return the number of bytes actually in use.

  unsigned int bytesUsed() const { return ntohl(*(unsigned int *)m_buf); }

  //@}

protected:

  /** Write bytes to the internal buffer with network byte-order
      conversion. */
  bool writeNetworkOrder(const void * data, unsigned int bytes);
  /** Read bytes from the internal buffer with network byte-order
      conversion. */
  bool readNetworkOrder(void * data, unsigned int bytes);

  static void failIn(const char *name);

  char         *m_buf;
  unsigned int  m_size;
  unsigned int  m_index;

private:
  /// Disabled constructor:
  DI_DataBuffer(const DI_DataBuffer &) {}
  /// Disabled copy operator:
  DI_DataBuffer & operator = (const DI_DataBuffer &) { return *this; }
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


/** Class to store data in little-endian encoding. This class works
    regardless of the architecture of the underlying hardware. 

    <B>This class is very similar to DI_DataBuffer with a few
    differences:</B> 
    <UL>

    <LI> The data is not padded in the memory (there are no empty
    spaces in the memory area.

    <LI> The data is stored in little-endian format

    <LI> The buffer resizes itself when necessary to accomodate new
    data.

    </UL>
*/
class DI_DataBuffer2
{
public:

  DI_DataBuffer2();
  DI_DataBuffer2(unsigned size);
  DI_DataBuffer2(const DI_DataBuffer2 &);
  virtual ~DI_DataBuffer2();

  bool isOverrun() const { return m_overrun; }
  bool isLittleEndian() const { return m_littleEndian; }

  void setLittleEndian(bool e) { m_littleEndian = e; }

  unsigned size() const { return m_size; }
  inline unsigned sizeLE() const;
  void resize(unsigned size);
  /// Expands the buffer to contain n bytes
  void expandTo(unsigned n) { if(n > m_size) resize(n); }

  /// Position of the read/write head.
  unsigned position() const { return m_position; }
  void setPosition(unsigned pos) 
  { m_position = pos; m_overrun = (m_position > m_size); }
  void movePosition(int move) 
  { m_position += (unsigned) move; m_overrun = (m_position > m_size); }
  inline unsigned positionLE() const;
  inline void setPositionLE(unsigned pos);
  void rewind() { m_position = 0; m_overrun = false;}

  unsigned char * buffer() { return m_buf; }
  const unsigned char * buffer() const { return m_buf; }
  
  /** @name Data access. */
  //@{
  void writeByte(char v);
  void writeBytes(const void *, unsigned);
  void writeInt16(short v) { write2(v); }
  void writeInt32(int v) { write4(v); }
  void writeInt64(llong v) { write8(v); }

  void writeFloat32(float v) { write4(* (int *) &v); }
  void writeFloat64(double v) { write8(* (llong *) &v); }

  void writeString(const char *);
  void writeString(const std::string *);

  char  readByte();
  void readBytes(void *, unsigned);
  short readInt16() { return read2(); }
  int   readInt32() { return read4(); }
  llong   readInt64() { return read8(); }

  float readFloat32() { int tmp = read4(); return * (float *) & tmp; }
  double readFloat64() { llong tmp = read8(); return * (double *) & tmp; }

  void readString(char *);
  void readString(std::string *);

  //@}

  /** @name The actual data access back-ends. */
  //@{

  void write2(unsigned short);
  void write4(unsigned int);
  void write8(ullong);

  short read2();
  int   read4();
  llong read8();
  //@}

  bool loadFile(const char *);
  
  int write(FILE *, bool flush = false);
  int read(FILE *);

  DI_DataBuffer2 & operator = (const DI_DataBuffer2 &);

protected:

  void readErr();

  unsigned char * m_buf;
  
  unsigned m_size;
  unsigned m_position;

  bool     m_overrun;
  bool     m_littleEndian;
};


/// Position of the read/write head in little-endian encoding.
inline unsigned DI_DataBuffer2::positionLE() const
{ 
#ifdef WORDS_BIGENDIAN
  return DI_byteSwap4(m_position);
#else
  return m_position;
#endif
}

/// Sets the read/write head position with little-endian argument.
inline void DI_DataBuffer2::setPositionLE(unsigned pos)
{
#ifdef WORDS_BIGENDIAN
  m_position = DI_byteSwap4(pos);
#else
  m_position = pos;
#endif
}

/// Position of the read/write head in little-endian encoding.
inline unsigned DI_DataBuffer2::sizeLE() const
{ 
#ifdef WORDS_BIGENDIAN
  return DI_byteSwap4(m_size);
#else
  return m_size;
#endif
}

int DI_match(const char * str, const char ** strs, uint nstrs);

#endif  // DI_DATA_BUFFER_H
