/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of DIVA base library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000-2001.
 * Jarmo.Hiipakka@hut.fi Tommi.Ilmonen@hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA base library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */

#include "di_mutex.h"
#include <errno.h>
#include <string.h>

#include <iostream>

// Trick for Linux
#ifndef PTHREAD_MUTEX_RECURSIVE
extern "C" {
extern int pthread_mutexattr_setkind_np __P ((pthread_mutexattr_t *__attr,
					   int __kind));
}
#define PTHREAD_MUTEX_RECURSIVE PTHREAD_MUTEX_RECURSIVE_NP
#define pthread_mutexattr_settype(x,y) pthread_mutexattr_setkind_np(x,y)
#endif

static bool DI_mutexDebug = false;

DI_Mutex::DI_Mutex()
  : m_active(false)
{}

DI_Mutex::~DI_Mutex()
{
  if(!m_active) return;

  int err = pthread_mutex_destroy(&m_mutex);
  if(err && DI_mutexDebug)
    std::cerr << "Error:DI_Mutex::~DI_Mutex " << strerror(err) << std::endl;
}

bool DI_Mutex::init(bool shared, bool prio_inherit, bool recursive)
{
  int ishared, iprio_inherit, err;
  const char ename[] = "Error:DI_Mutex::init"; // Error and name.

  if(shared && prio_inherit) {
    if(DI_mutexDebug) std::cerr << ename << " - bad argument" << std::endl;
    return false;
  }

  pthread_mutexattr_t mutex_attr;
  pthread_mutexattr_init(&mutex_attr);
  
  if(recursive) {
    err = pthread_mutexattr_settype(&mutex_attr, PTHREAD_MUTEX_RECURSIVE);
    if(err) {
      if(DI_mutexDebug) std::cerr << ename << strerror(err) << std::endl;
      return false;
    }
  }

#if defined(_SGI_MP_SOURCE) || _POSIX_THREAD_PROCESS_SHARED
  if(shared) ishared = PTHREAD_PROCESS_SHARED;
  else ishared = PTHREAD_PROCESS_PRIVATE;

  err = pthread_mutexattr_setpshared(&mutex_attr, ishared);
  if(err) {
    if(DI_mutexDebug) std::cerr << ename << strerror(err) << std::endl;
    return false;
  }
#else
  // Not supported (Linux).
  (void) ishared;
#endif

#if defined(_SGI_MP_SOURCE) || _POSIX_THREAD_PRIO_INHERIT
  if(prio_inherit) iprio_inherit = PTHREAD_PRIO_INHERIT;
  else iprio_inherit = PTHREAD_PRIO_NONE;
  err = pthread_mutexattr_setprotocol(&mutex_attr, iprio_inherit);
  if(err) {
    std::cerr << ename << strerror(err) << std::endl;
    return false;
  }
#else
  (void) iprio_inherit;
#endif

  err = pthread_mutex_init(&m_mutex, &mutex_attr);
  if(err && DI_mutexDebug) {
    std::cerr << ename << strerror(err) << std::endl;
  }
  else
    m_active = true;

  return ! err;
}

bool DI_Mutex::close()
{
  if(m_active) {
    m_active = false;
    return ! pthread_mutex_destroy(&m_mutex);
  }
  return false; // Nothing to close
}

bool DI_Mutex::lock()
{ 
  const char *fname = "DI_Mutex::lock ";

  int e = pthread_mutex_lock(&m_mutex); 
  if(e) std::cerr << fname << strerror(e);
  return !e;
}

bool DI_Mutex::lock(bool block)
{
  if(!block) return tryLock();
  else return lock();
}

bool DI_Mutex::tryLock() 
{ 
  const char *fname = "DI_Mutex::trylock ";

  int e = pthread_mutex_trylock(&m_mutex); 
  if(e) std::cerr << fname << strerror(e);
  return !e;
}

bool DI_Mutex::unlock()
{
  const char *fname = "DI_Mutex::unlock ";

  int e = pthread_mutex_unlock(&m_mutex); 
  if(e) std::cerr << fname << strerror(e);
  return !e;
}

void DI_Mutex::test()
{
  DI_Mutex m1;

  const char name[]= "DI_Mutex::test";
  
  std::cout << name << std::endl;

  if(!m1.init(false, true)) {
    std::cerr << name << " First init failed" << std::endl;
    
    if(!m1.init(false, false))
      std::cerr << name << " Second init failed" << std::endl;
  }

  std::cout << name << " trying recursive locking ... " << std::endl;
  m1.lock();
  std::cout << name << " locked once" << std::endl;
  m1.lock();
  std::cout << name << " locked twice" << std::endl;
  m1.unlock();
  m1.unlock();
  std::cout << name << " recursive locking worked " << std::endl;
  // Second test:
  std::cout << "\n" << name << " testing illegal: " << std::endl;
  DI_Mutex m2;
  m2.init(true, true);
}

DI_Mutex &DI_Mutex::operator = (const DI_Mutex &foo)
{
  (void) foo;
  return *this;
}
