/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of DIVA base library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000-2001.
 * Jarmo.Hiipakka@hut.fi Tommi.Ilmonen@hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA base library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */
//
// DIVA Library. Copyright (C) Jarmo Hiipakka, 1999-2001.
// DIVA Library. Copyright (C) Tommi Ilmonen,  1999-2001.
//
// $Log: di_socket.C,v $
// Revision 1.12  2004/11/08 08:12:54  tilmonen
// Fixes.
//
// Revision 1.11  2004/09/06 12:16:47  tilmonen
// Flush & nodelay.
//
// Revision 1.16  2004/08/23 06:55:52  tilmonen
// Cleaned.
//
// Revision 1.15  2004/08/23 07:02:03  tilmonen
// Fixed the include bug.
//
// Revision 1.14  2004/08/23 06:57:29  tilmonen
// NO_DELAY option in the socket code.
//
// Revision 1.13  2004/08/19 06:44:45  tilmonen
// New flush.
//
// Revision 1.12  2004/08/18 06:36:04  tilmonen
// Fixes.
//
// Revision 1.11  2004/06/09 12:51:36  tilmonen
// Added flushing and checks for hang-up.
//
// Revision 1.10  2002/08/12 14:00:07  tilmonen
// Bug fixes.
//
// Revision 1.9  2002/06/05 10:08:58  tilmonen
// Updates.
//
// Revision 1.8  2002/05/15 12:40:47  tilmonen
// Updates to DI_DataBuffer2.
//
// Revision 1.7  2002/02/22 11:28:53  tilmonen
// New (and possibly better) DI_DataBuffer2 class for binary
// data management.
//
// Revision 1.6  2002/01/18 12:18:29  tilmonen
// Cleanups.
//
// Revision 1.5  2001/10/08 12:25:53  tilmonen
// Updates.
//
// Revision 1.4  2001/07/27 08:23:17  tilmonen
// License text updates (=cosmetic changes).
//
// Revision 1.3  2001/07/26 08:49:23  tilmonen
// Updates is isPendingInput.
//
// Revision 1.2  2001/07/09 14:41:31  tilmonen
// Important bug fixes (endian issue in di_data_buffer).
//
// Revision 1.1.1.1  2001/06/04 09:04:23  tilmonen
// Mustajuuri 4.6.2001, initial CVS cersion
//

#include "di_socket.h"

#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/socket.h> 
#include <netinet/tcp.h> 
#include <unistd.h>
#include <arpa/inet.h>

#include <stropts.h>
#include <poll.h>

/* The two functions */

/* Take a service name, and a service type, and return a port number.  If the
   service name is not found, it tries it as a decimal number.  The number
   returned is byte ordered for the network. */
int atoport(char * service, char * proto)
{
  int port;
  long int lport;
  struct servent *serv;
  char *errpos;

  /* First try to read it from /etc/services */
  serv = getservbyname(service, proto);
  if (serv != NULL)
    port = serv->s_port;
  else { /* Not in services, maybe a number? */
    lport = strtol(service,&errpos,0);
    if ( (errpos[0] != 0) || (lport < 1) || (lport > 65535) )
      return -1; /* Invalid port address */
    port = htons(lport);
  }
  return port;
}

/* Converts ascii text to in_addr struct.  NULL is returned if the address
   can not be found. */
struct in_addr *atoaddr(char *address)

{
  struct hostent *host;
  static struct in_addr saddr;

  /* First try it as aaa.bbb.ccc.ddd. */
  saddr.s_addr = inet_addr(address);
  if ((int) saddr.s_addr != -1) {
    return &saddr;
  }
  host = gethostbyname(address);
  if (host != NULL) {
    return (struct in_addr *) *host->h_addr_list;
  }
  return NULL;
}

// --- Constructors -----------------------------------------------------------

DI_Socket::DI_Socket(void)
{
  init();
}

DI_Socket::~DI_Socket(void)
{
  close();
}

// --- Access -----------------------------------------------------------------

void DI_Socket::setBlocking(DI_BlockingMode blocking)
{
  m_blocking = blocking;

  int iFlags = fcntl(m_socketID,F_GETFL);

  if (getBlocking() == DI_BLOCK_NOT)
    iFlags |=  O_NONBLOCK;
  else
    iFlags &= ~O_NONBLOCK;

  if (fcntl(m_socketID,F_SETFL,iFlags) == (-1))
  {
    m_error = DI_ERR_SYSTEM;
    DI_error(m_error,"DI_Socket:Blocking:fcntl",strerror(errno));
  }
}

ulong DI_Socket::getBufferSize(DI_SocketStatus which) const
{
  ulong result = 0;

  /* For some reason, the last argument of, e.g., getsockopt() is different
     in linux and IRIX */
#ifdef __linux__
  int value;
  socklen_t size = sizeof(value);
#else
  int value,size = sizeof(value);
#endif  // __linux__

  switch(which)
  {
   case DI_SS_INPUT:
    if (!getsockopt(m_socketID,SOL_SOCKET,SO_RCVBUF,&value,&size))
      result = value;
    else
      DI_error(DI_ERR_SYSTEM,"DI_Socket:getBufferSize:getsockopt failed");
    break;
   case DI_SS_OUTPUT:
    if (!getsockopt(m_socketID,SOL_SOCKET,SO_SNDBUF,&value,&size))
      result = value;
    else
      DI_error(DI_ERR_SYSTEM,"DI_Socket:getBufferSize:getsockopt failed");
    break;
   default:
    DI_error(DI_ERR_ARGUMENT,"DI_Socket:getBufferSize");
    break;
  }

  return result;
}

void DI_Socket::setBufferSize(DI_SocketStatus which, ulong size)
{
  int value = (int)size;

  switch(which)
  {
   case DI_SS_INPUT:
    if (setsockopt(m_socketID,SOL_SOCKET,SO_RCVBUF,&value,sizeof(value)))
      DI_error(DI_ERR_SYSTEM,"DI_Socket:setBufferSize:setsockopt failed");
    break;
   case DI_SS_OUTPUT:
    if (setsockopt(m_socketID,SOL_SOCKET,SO_SNDBUF,&value,sizeof(value)))
      DI_error(DI_ERR_SYSTEM,"DI_Socket:setBufferSize:setsockopt failed");
    break;
   case DI_SS_INPUT_OUTPUT:
    setBufferSize(DI_SS_INPUT,size);
    setBufferSize(DI_SS_OUTPUT,size);
    break;
   default:
    DI_error(DI_ERR_ARGUMENT,"DI_Socket:setBufferSize");
    break;
  }
}

bool DI_Socket::setNoDelay(bool noDelay)
{
  int yes = noDelay;

  if(isStream())
    return false;

  // if (setsockopt(m_socketID, SOL_TCP, TCP_NODELAY, (char *)&yes,
  if (setsockopt(m_socketID, IPPROTO_TCP, TCP_NODELAY, (char *)&yes, 
		 sizeof(int))) {
    m_error = DI_ERR_SYSTEM;
    DI_error(m_error, "DI_Socket:setNoDelay:setsockopt(TCP_NODELAY)", 
	     strerror(errno));

    return false;
  }

  return true;
}

// --- Operations -------------------------------------------------------------

void DI_Socket::clear(void)
{
  m_error = DI_ERR_NONE;
}

bool DI_Socket::close(void)
{
  if (m_socketID >= 0)
  {
    ::close(m_socketID);
    m_socketID = -1;
  }

  m_blocking = DI_BLOCK_FULL;
  m_type  = DI_SS_UNINITIALIZED;
  m_error = DI_ERR_NONE;
  return true;
}

DI_Socket *DI_Socket::listen(void)
{
  if (m_type != DI_SS_SERVER)
  {
    m_error = DI_ERR_MATCH;
    DI_error(m_error,"DI_Socket:Listen:Invalid connection type");
    return (DI_Socket *)0;
  }

  DI_Socket *newSocket(NULL);
  DI_SocketAddress newAddress;

#ifdef __linux__
  socklen_t addressLength(sizeof(newAddress));
#else
  int addressLength(sizeof(newAddress));
#endif  // __linux__

  int newSocketID(accept(m_socketID,(sockaddr *)&newAddress,&addressLength));

  if (newSocketID < 0)
  {
    if (getBlocking() != DI_BLOCK_NOT || errno != EWOULDBLOCK)
    {
      DI_error(m_error,"DI_Socket:Listen:accept",strerror(errno));
    }
    else
    {
      errno = 0; // It's OK to fail if we do not block and no-one was waiting
    }
  }
  else
  {
    newSocket = new DI_Socket();

    newSocket->setStream(true);
    newSocket->m_type          = DI_SS_DATA;
    newSocket->m_addressLength = addressLength;
    newSocket->m_address       = newAddress;
    newSocket->m_socketID      = newSocketID;
  }

  return newSocket;
}

bool DI_Socket::open(const char * const address, bool server, bool stream)
{
  close();

  if (!address || !(*address))
  {
    m_error = DI_ERR_ARGUMENT;
    DI_error(m_error,"DI_Socket::open:Invalid address",address);
    return false;
  }

  setStream(stream); // Set stream/datagram mode
  m_type = server ? DI_SS_SERVER : DI_SS_DATA;

  // Resolve address

  memset(&m_address,0,sizeof(m_address)); // Absolutely necessary

  const char * separator = strchr(address, ':');
  std::string hostName, portName;

  if(separator) {
    portName = separator + 1;
    hostName = address;
    hostName.resize(separator - address);
  }
  else {
    portName = address;
    hostName = "localhost";
  }
  /* atoaddr and atoprot are from sockhelp.c */
  int port = atoport((char *) portName.c_str(), "tcp");
  in_addr * addr = atoaddr((char *) hostName.c_str()); 
  
  m_addressFamily                = AF_INET;
  m_address.in.sin_family      = (unsigned char)AF_INET;
  m_address.in.sin_port        = (unsigned short) port;
  m_address.in.sin_addr.s_addr = addr->s_addr;
  
  m_addressLength = sizeof(m_address.in);

  // Create socket
  
  int type = isStream() ? SOCK_STREAM : SOCK_DGRAM;
  m_socketID = socket(m_addressFamily,type,0);

  if (m_socketID < 0)
  {
    m_error = DI_ERR_SYSTEM;
    DI_error(m_error,"DI_Socket:Open:socket",strerror(errno));
    return false;
  }

  // Configure socket: Allow immediate reuse of addresses

  int yes = 1;
  
  if (setsockopt(m_socketID,SOL_SOCKET,SO_REUSEADDR,(char *)&yes,sizeof(int)))
  {
    m_error = DI_ERR_SYSTEM;
    DI_error(m_error,"DI_Socket:Open:setsockopt",strerror(errno));
    return false;
  }

  // Configure socket: Allow fast close-operations.

  struct linger linger_struct;

  linger_struct.l_onoff = 0;
  linger_struct.l_linger = 0;

  if (setsockopt(m_socketID,SOL_SOCKET,SO_LINGER,(char *)&linger_struct,
		 sizeof(linger_struct)))
  {
    m_error = DI_ERR_SYSTEM;
    DI_error(m_error,"DI_Socket:Open:setsockopt",strerror(errno));
    return false;
  }

  if (m_type == DI_SS_SERVER)
  {
    if (bind(m_socketID,(sockaddr *) & m_address, m_addressLength) != 0)
    {
      m_error = DI_ERR_SYSTEM;
      DI_error(m_error,"DI_Socket:Open:bind",strerror(errno));
      return false;
    }

    if (isStream())
    {
      if (::listen(m_socketID,1) != 0)
      {
	m_error = DI_ERR_SYSTEM;
	DI_error(m_error,"DI_Socket:Open:listen",strerror(errno));
	return false;
      }
    }
  }
  else
  {
    if (connect(m_socketID,(sockaddr *)&m_address,m_addressLength) != 0)
    {
      m_error = DI_ERR_SYSTEM;
      DI_error(m_error,"DI_Socket:Open:connect",address);
      return false;
    }
  }
  setBlocking(DI_BLOCK_FULL);

  return true;
}

int DI_Socket::read(void *buffer, ulong bytes)
{
  if (bytes == 0) return 0;

  int readBytes;

  if (m_type == DI_SS_UNINITIALIZED)
  {
    m_error = DI_ERR_UNINITIALIZED;
    DI_error(m_error,"DI_Socket:read:Connection not initialized");
    return -1;
  }
  
  if (isStream()) // Stream socket
  {
    if (m_type != DI_SS_DATA)
    {
      m_error = DI_ERR_MATCH;
      DI_error(m_error,"DI_Socket:read:Invalid connection type");
      return -1;
    }

    ulong total = 0;
    bool waitFlag(false);

    do
    {
      readBytes = ::read(m_socketID,&(((char *)buffer)[total]),bytes);

      if (readBytes > 0) // Something was read
      {
	total += readBytes;
	bytes -= readBytes;

	switch (getBlocking())
	{
	 case DI_BLOCK_FULL:
	  waitFlag = (bytes > 0);
	  break;
	 case DI_BLOCK_ONE:
	  waitFlag = false;
	  break;
	 case DI_BLOCK_NOT:
	  waitFlag = false;
	  break;
	 default:
	  DI_fatal(DI_ERR_INTERNAL,"DI_Socket:read:Invalid blocking mode");
	  break;
	}
      }
      else if (readBytes < 0) // Error condition
      {
	switch (getBlocking())
	{
	 case DI_BLOCK_FULL:
	  waitFlag = false;
	  readBytes = DI_SS_ERROR;
	  m_error = DI_ERR_INPUT;
	  break;
	 case DI_BLOCK_ONE:
	  waitFlag = false;
	  readBytes = DI_SS_ERROR;
	  m_error = DI_ERR_INPUT;
	  break;
	 case DI_BLOCK_NOT:
	  if (errno == EAGAIN) // Attempt to read from empty connection
	  {
	    waitFlag = false;
	    readBytes = 0;
	    errno = 0;
	  }
	  break;
	 default:
	  DI_fatal(DI_ERR_INTERNAL,"DI_Socket:read:Invalid blocking mode");
	  break;
	}
      }
      else // Nothing was read: => End of file
      {
	switch(getBlocking())
	{
	 case DI_BLOCK_FULL:
	  waitFlag = false;
	  if (total > 0)
	    readBytes = (int)total;
	  else
	    readBytes = DI_SS_EOF;
	  m_error = DI_ERR_EOF;
	  break;
	 case DI_BLOCK_ONE:
	  waitFlag = false;
	  readBytes = DI_SS_EOF;
	  m_error = DI_ERR_EOF;
	  break;
	 case DI_BLOCK_NOT:
	  waitFlag = false;
	  readBytes = DI_SS_EOF;
	  m_error = DI_ERR_EOF;
	  break;
	 default:
	  DI_fatal(DI_ERR_INTERNAL,"DI_Socket:Read:Invalid blocking mode");
	  break;
	}
      }
    }
    while (waitFlag);

    if (readBytes >= 0) readBytes = (int)total;
  }
  else // Datagram socket
  {
    readBytes = (int)recvfrom(m_socketID,buffer,(int)bytes,0,0,0);

    if (readBytes < 0)
    {
      switch(getBlocking())
      {
       case DI_BLOCK_NOT:
	break;
       case DI_BLOCK_FULL:
       case DI_BLOCK_ONE:
	readBytes = DI_SS_ERROR;
	m_error = DI_ERR_INPUT;
	DI_error(DI_ERR_SYSTEM,"DI_Socket:read:recv failed");
	break;
       default:
	DI_fatal(DI_ERR_INTERNAL,"DI_Socket:read:Invalid blocking mode");
	break;
      }
    }
  }

  return readBytes;
}

int DI_Socket::write(const void *buffer, ulong bytes)
{
  if (bytes == 0) return 0;

  int writtenBytes;

  if (m_type == DI_SS_UNINITIALIZED)
  {
    m_error = DI_ERR_UNINITIALIZED;
    DI_error(m_error,"DI_Socket:write:Connection not initialized");
    return -1;
  }
  
  if (isStream()) // Stream socket
  {
    if (m_type != DI_SS_DATA)
    {
      m_error = DI_ERR_MATCH;
      DI_error(m_error,"DI_Socket:write:Invalid connection type");
      return -1;
    }

    ulong total = 0;
    bool waitFlag(false);

    do
    {
      writtenBytes = ::write(m_socketID,&(((char *)buffer)[total]),bytes);

      if (writtenBytes > 0)
      {
	total += writtenBytes;
	bytes -= writtenBytes;

	switch(getBlocking())
	{
	 case DI_BLOCK_FULL:
	  waitFlag = (bytes > 0) ? true : false;
	  break;
	 case DI_BLOCK_ONE:
	  waitFlag = false;
	  break;
	 case DI_BLOCK_NOT:
	  waitFlag = false;
	  break;
	 default:
	  DI_fatal(DI_ERR_INTERNAL,"DI_Socket:write:Invalid blocking mode");
	  break;
	}
      }
      else if (writtenBytes < 0)
      {
	waitFlag = false;
      }
      else
      {
	switch(getBlocking())
	{
	 case DI_BLOCK_FULL:
	  waitFlag = true;
	  break;
	 case DI_BLOCK_ONE:
	  waitFlag = true;
	  break;
	 case DI_BLOCK_NOT:
	  waitFlag = false;
	  break;
	 default:
	  DI_fatal(DI_ERR_INTERNAL,"DI_Socket:write:Invalid blocking mode");
	  break;
	}
      }
    }
    while(waitFlag);
    
    if (writtenBytes < 0)
    {
      m_error = DI_ERR_SYSTEM;
      DI_error(m_error,"DI_Socket:write:write",strerror(errno));
    }
    else
    {
      writtenBytes = (int)total;
    }
  }

  else // Datagram socket
  {
    // writtenBytes = (int)send(m_socketID,buffer,(int)bytes,0);
    writtenBytes = (int)sendto(m_socketID, buffer, (int) bytes, 0,
			       (sockaddr *)&m_address, m_addressLength);
    
    if (writtenBytes < (int)bytes)
      DI_error(DI_ERR_SYSTEM,"DI_Socket:write:send failed");
  }
  
  return writtenBytes;
}

/** Flushes the written data to the target. */

void DI_Socket::flush() const
{
  if(m_socketID < 0)
    return;

  /* struct pollfd pfd;
     pfd.fd = m_socketID;
     pfd.events = POLLIN;
     poll(&pfd, 1, 0); */
  
  fsync(m_socketID);
}

/** Checks if the socket is hung up. */

bool DI_Socket::isHungUp() const
{
  if(m_socketID < 0)
    return false;

  struct pollfd pfd;
  pfd.fd = m_socketID;
  pfd.events = ~0;
  poll(&pfd, 1, 0);
  return pfd.revents & (POLLERR | POLLHUP | POLLNVAL) != 0;
}

// --- Predicates -------------------------------------------------------------

bool DI_Socket::isEOF(void) const
{
  //  DI_error(DI_ERR_UNIMPLEMENTED,"DI_Socket:IsEOF");
  return (m_error == DI_ERR_EOF);
}

bool DI_Socket::isOK(void) const
{
  return (m_error == DI_ERR_NONE); // Note '=='
}

bool DI_Socket::isPendingInput(unsigned waitMicroSeconds)
{
  if(m_socketID < 0)
    return false;

  struct pollfd pfd;
  pfd.fd = m_socketID;
  pfd.events = POLLIN;
  poll(&pfd, 1, waitMicroSeconds / 1000);
  return pfd.revents & POLLIN;
}

// --- Private members --------------------------------------------------------

void DI_Socket::init(void)
{
  m_blocking = DI_BLOCK_FULL;

  m_type     = DI_SS_UNINITIALIZED;
  m_socketID = -1;
  m_error    = DI_ERR_NONE;
}

// --- Helper functions -------------------------------------------------------

int DI_writeToSocket(DI_Socket *sock, const DI_DataBuffer *data)
{
  if (sock->isStream())
  {
    return sock->write(data->buffer(), data->bytesUsed()+4);
  }
  else  // Datagram socket
  {
    int written(0);
    written += sock->write(data->buffer(), 4);  // Write the data size
    written += sock->write(data->buffer()+4, (ulong)data->bytesUsed());
    return written;
  }
}

int DI_readFromSocket(DI_Socket *sock, DI_DataBuffer *data)
{
  int read(0);
  read += sock->read(data->buffer(), 4); // Read the data size
  // DI_trace("DI_readFromSocket", data->bytesUsed());
  read += sock->read(data->buffer()+4, (ulong)data->bytesUsed());
  return read;
}

int DI_writeToSocket(DI_Socket *sock, const DI_DataBuffer2 *data)
{
  int tmp = data->positionLE();
  tmp = sock->write(&tmp, 4);
  return tmp + sock->write(data->buffer(), data->position());
}

int DI_readFromSocket(DI_Socket *sock, DI_DataBuffer2 *data)
{
  int r(0), tmp;
  r += sock->read(&tmp, 4);

  if(sock->isEOF() || ! sock->isOK() || r != 4) {
    return -1;
  }

  data->setPositionLE(tmp);
  tmp = data->position();
  if(tmp > 100000)
    DI_trace2("DI_readFromSocket # Got %d bytes of data (!)", tmp);
  data->expandTo(tmp);
  // DI_trace("DI_readFromSocket", data->bytesUsed());
  r += sock->read(data->buffer(), tmp);
  return r;
}
