/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of DIVA base library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000-2001.
 * Jarmo.Hiipakka@hut.fi Tommi.Ilmonen@hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA base library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */

#include "di_thread.h"
#include "di_mutex.h"
#include "di_sleep.h"

#include <errno.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>

#include <iostream>


bool DI_Thread::m_threadDebug = false;
bool DI_Thread::m_threadWarnings = false;

DI_Thread::DI_Thread()
{}

DI_Thread::~DI_Thread()
{}

bool DI_Thread::run(bool prefer_system)
{
  if(m_threadDebug)
    std::cout << "DI_Thread::run " << prefer_system << " " << this << std::endl;

  if(prefer_system) {
    if(!runSystem()) {
      return runProcess();
    }
  }
  else if(!runProcess())
    return runSystem();

  return true;
}

bool DI_Thread::runSystem()
{
  if(m_threadDebug)
    std::cout << "DI_Thread::runSystem " << this << std::endl;

  pthread_attr_t thread_attr;

  pthread_attr_init(&thread_attr);
  int e = pthread_attr_setscope(&thread_attr, PTHREAD_SCOPE_SYSTEM);

  if(e) {
    if(m_threadDebug || m_threadWarnings) 
      std::cout << "failed - " << strerror(e) << std::endl;
    return false;
  }

  e = pthread_create(&m_thread, &thread_attr, entry, this);

  if(e && (m_threadDebug || m_threadWarnings)) 
    std::cout << "failed - " << strerror(e) << std::endl;

  return !e;
}

bool DI_Thread::runProcess()
{
  if(m_threadDebug)
    std::cout << "DI_Thread::runProcess " << this << std::endl;

  pthread_attr_t thread_attr;

  pthread_attr_init(&thread_attr);

  int e = pthread_attr_setscope(&thread_attr, PTHREAD_SCOPE_PROCESS);

  if(e) {
    if(m_threadDebug || m_threadWarnings) 
      std::cout << "failed - " << strerror(e) << std::endl;
    return false;
  }
  
  e = pthread_create(&m_thread, &thread_attr, entry, this);

  if(e && (m_threadDebug || m_threadWarnings)) 
    std::cout << "failed - " << strerror(e) << std::endl;

  return !e;
}

bool DI_Thread::waitEnd()
{
  if(m_threadDebug) {
    std::cout << "DI_Thread::waitEnd " << this << std::endl;
  }

  int e = pthread_join(m_thread, 0);
  
  if(e) {
    if(m_threadDebug || m_threadWarnings)
      std::cerr << "DI_Thread::waitEnd failed - " << strerror(e) << std::endl;
    return false;
  }
  return true;
}

void DI_Thread::kill()
{
  pthread_kill(m_thread, SIGKILL);
}

void DI_Thread::test()
{
  bool debug = m_threadDebug;
  m_threadDebug = true;  // Debugging texts on.

  const char name[] = "DI_Thread::test ";

  DI_Sleep sleep;

  std::cout << name << " pid: " << getpid() << " test 1" << std::endl;
  {
    std::cout << name << "creating thread that runs for 4 seconds" << std::endl;
    DI_ThreadTest t;
    std::cout << name << "starting thread..." << std::endl;
    t.run();
    sleep.sleepUs(100000);
    std::cout << name << "...thread running, sleeping 4 seconds" << std::endl;
    DI_Sleep::sleepS(4);
    std::cout << name << "ending thread" << std::endl;
    t.requestEnd();
    bool ok = t.waitEnd();
    std::cout << name << "success: " << ok << std::endl << std::endl;
  }

  std::cout << name << "test 2" << std::endl;
  {
    std::cout << name << "creating thread that runs for 2 seconds" << std::endl;
    DI_ThreadTest t;
    std::cout << name << "starting thread..." << std::endl;
    t.run();
    std::cout << name << "...thread running, sleeping 2 seconds" << std::endl;
    DI_Sleep::sleepS(2);
    std::cout << name << "deliberate deadlock now, USE CTRL+C to exit " << std::endl;
    /* "requestEnd" is not called. Therefore we deadlock. */ 
    bool ok = t.waitEnd();
    std::cout << name << "Failed to deadlock " << ok << std::endl;
  }
  m_threadDebug = debug; // Return the original value
}


void DI_Thread::threadExit()
{
  if(m_threadDebug) {
    std::cout << "DI_Thread::threadExit " << this << std::endl;
  }

  pthread_exit(0);
}

void DI_Thread::mainLoop()
{
  if(m_threadDebug) {
    std::cout << "DI_Thread::mainLoop " << this <<  " pid: " << getpid() << std::endl;
  }

  childLoop();
}

void *DI_Thread::entry(void *t)
{
  if(m_threadDebug)
    std::cout << "DI_Thread::entry " << t << std::endl;
  
  ((DI_Thread *) t)->mainLoop();

  return 0;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

bool DI_ThreadAvailable::m_systemLevel = false;
bool DI_ThreadAvailable::m_processLevel = false;
bool DI_ThreadAvailable::m_defined = false;

bool DI_ThreadAvailable::systemLevelWorks()
{
  if(!m_defined) check();
  return m_systemLevel;
}

bool DI_ThreadAvailable::processLevelWorks()
{
  if(!m_defined) check();
  return m_processLevel;
}

void DI_ThreadAvailable::check()
{
  {
    DI_ThreadTest t;
    t.requestEnd();
    m_systemLevel = t.runSystem();
    if(m_systemLevel)
      t.waitEnd();
  }
  {
    DI_ThreadTest t;
    t.requestEnd();
    m_processLevel = t.runProcess();
    if(m_processLevel)
      t.waitEnd();
  }
  m_defined = true;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

DI_ThreadTest::DI_ThreadTest()
  : m_mutex(new DI_MutexAuto),
    m_continue(true),
    m_counter(0)
{}

int DI_ThreadTest::count()
{
  m_mutex->lock();
  int i = m_counter;
  m_mutex->unlock();
  return i;
}

void DI_ThreadTest::requestEnd()
{
  m_mutex->lock();
  m_continue = false;
  m_mutex->unlock();
}

void DI_ThreadTest::childLoop()
{
  m_mutex->lock();
  DI_Sleep s;

  while(m_continue) {

    m_mutex->unlock();
    s.sleepUs(500000); // Half second sleep.
    m_mutex->lock();

    std::cout << "DI_ThreadTest::childLoop " << this 
	      << "  " << m_counter++ << std::endl;
  }
  m_mutex->unlock();
}
