/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of DIVA base library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000-2001.
 * Jarmo.Hiipakka@hut.fi Tommi.Ilmonen@hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA base library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */

#ifndef DI_THREAD_H
#define DI_THREAD_H

#define DI_PTHREAD

#include <pthread.h>

class DI_Mutex;

/** Thread class. Utilize this class by inheriting. Make use of this
    class by overriding method "childLoop".*/
class DI_Thread
{
public:
  typedef pthread_t id_t;

  static id_t myThreadId() { return pthread_self(); }

  /** Construct a thread structure. The thread is NOT activated by this
      method. */
  DI_Thread();
  
  /** */
  virtual ~DI_Thread();

  /** Starts the thread in either system-level or process-level
      scope. If the OS fails to deliver the desired thread type the
      other kind is silently used instead. This method uses the other
      two "run" methods */
  bool run(bool prefer_system = true);
  
  /// Runs a system-level thread.
  bool runSystem();
  /// Runs a process-level thread.
  bool runProcess();

  /** Waits until thread is finished. This method does nothing to
      kill the thread, it simply waits until the thread has run its
      course. */
  bool waitEnd();

  /** Kills the thread. A violent way to shut down a thread. You
      should only call this method in emergency situations. May result
      in application crash and other minor problems.*/
  void kill();

  /** Drive some self tests. */
  static void test();

protected:
  /// Exits the the calling thread.
  void threadExit();

  /// Calls childLoop.
  void mainLoop();

  /** The actual contents of the thread. You need to override this to
      add functionality to your software. */
  virtual void childLoop() = 0;

  static void *entry(void *);

  pthread_t       m_thread;

  static bool m_threadDebug;
  static bool m_threadWarnings;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** Status of available thread types. 
    
    This class can be used that certain thread types do work.

    Only static methods.
 */
class DI_ThreadAvailable
{
public:
  DI_ThreadAvailable() {}
  ~DI_ThreadAvailable() {}

  /// Returns true if system-level threads seem to work.
  static bool systemLevelWorks();
  /// Returns true if process-level threads seem to work.
  static bool processLevelWorks();
  /** Performs checks for different thread models. You do not need to
      call this method explicitly. If this method has not been called
      prior to "systemLevelWorks" etc it will be called to check the
      values. */
  static void check();
protected:
  static bool m_systemLevel, m_processLevel, m_defined;

};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/// A simple thread that does nothing important.
class DI_ThreadTest : public DI_Thread
{
public:
  DI_ThreadTest();
  ~DI_ThreadTest() {}

  int count();
  void requestEnd();
  
protected:
  void childLoop();

  DI_Mutex *m_mutex;

  bool m_continue;
  int  m_counter;
};

#endif
