/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of DIVA base library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000.
 * Jarmo.Hiipakka@hut.fi Tommi.Ilmonen@hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA base library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */

#ifndef DI_THREAD_DATA_H
#define DI_THREAD_DATA_H

#include <di_mutex.h>
#include <di_thread.h>

#include <map>

/** Thread-specific data holder. This class is used to hold
    thread-specific data. Any thread can place data to the holder. The
    holder has a list of data objects, with corresponding thread
    ids. Thus a caller only gets the data that is specific to that
    particular thread. 

    Keep in mind that one must be careful with how the data is
    organized. The data type must have a default constructor and copy
    operator. If you wish to use data objects that do not have these
    features, then your best bet is to store only pointers to the data
    here.

    This class is completely thread-safe. It uses an internal mutex to
    prevent simultaneous modifications from multiple threads.

    @author Tommi Ilmonen
*/

template <class T> class DI_ThreadData
{
public:

  DI_ThreadData() {}
  ~DI_ThreadData() {}

  /** Access operator. You can always call this method. If the data
      element did not exist before this call , then it is created
      on-demand. This method does mutex locking to guarantee that
      other threads do not corrupt the table while the operation is
      taking place. */
  inline T & operator * ()
  {
    m_mutex.lock();
    DI_Thread::id_t id = DI_Thread::myThreadId();
    T& ret =  m_table[id];
    m_mutex.unlock();
    return ret;
  }

  /// Const access operator.
  inline const T & operator * () const
  { return * ((DI_ThreadData *) this); }

  /// Superfast access operation with no safety locks.
  inline T & getFast()
  {return m_table[DI_Thread::myThreadId()]; }

  /** This method deletes the data element that is specific to this
      thread. Use with care. */
  void clearMyData()
  {
    m_mutex.lock();
    DI_Thread::id_t id = DI_Thread::myThreadId();
    typename std::map<DI_Thread::id_t, T>::iterator t = m_table.find(id);
    if(t != m_table.end()) m_table.erase(t);
    m_mutex.unlock();
  }

  /** Deletes all objects. Use with \b great care. */
  void clear()
  { m_mutex.lock(); m_table.clear(); m_mutex.unlock(); }

  /** Returns a pointer to the first object. If there are no objects,
      then NULL is returned. */
  T * first() 
  { 
    DI_Guard g(&m_mutex);
    if(!m_table.size()) 
      return 0; 
    return & (*m_table.begin()).second; 
  }

protected:
  DI_MutexAuto m_mutex;

  std::map<DI_Thread::id_t, T> m_table;
  
  /// Disabled
  DI_ThreadData & operator = (const DI_ThreadData &) { return * this; }
  /// Disabled
  DI_ThreadData(const DI_ThreadData &) {}
};

#endif
