/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#include <Fluid/InputDevices.hpp>
#include <Fluid/Glove.hpp>
#include <Fluid/Keyboard.hpp>
#include <Fluid/Mouse.hpp>
#include <Fluid/Napra.hpp>
#include <Fluid/Speech.hpp>
#include <Fluid/Tracker.hpp>
#include <Fluid/Config.hpp>

#include <di_data_buffer.h>

#include <stdio.h>
#include <unistd.h>
#include <iostream>
#include <dlfcn.h>

using namespace Fluid;

void testMouse(const std::string& mouseName)
{
  Mouse *mouse = InputDevices::getMouse(mouseName);

  if(mouse == NULL) 
    {
      std::cout << "Mouse \"" << mouseName << "\" not found" << std::endl;
      return;
    }

  mouse->requireHistory(100);

  int cnt = 200;
  mouse->start();
  ulong counter = 1;
  while(cnt--) {
    mouse->update();
    
    for(; counter <= mouse->getLatestIndex(); counter++) {
      
      MouseSample s = mouse->getSample(counter);
      printf("%ld: %d,%d %x\n", counter, s.x, s.y, s.state);
    }

    fflush(0);
    usleep(100000);
  }
  // mouse->stop();

  std::cout << "Mouse test done" << std::endl;
}

void testGlove(const std::string& gloveName)
{
  Glove *glove = (Glove*) InputDevices::getInputDevice(gloveName);
  if(glove == NULL)
    {
      std::cout << "Glove \"" << gloveName << "\" not found" << std::endl;
      return; 
    }
  
  glove->start();
  while(1)
    {
      glove->update();
      for(int f = Glove::FINGER_THUMB; f < Glove::FINGERS; f++)
	{
	  std::cout << glove->getJointAngle((Glove::Finger)f,
					    Glove::FINGERJOINT_SECOND) << " ";
	}
      std::cout << '\n';
      
      usleep(1000);
    }
  // glove->stop();
}

void testKeyboard(const std::string& keyboardName)
{
  Keyboard *keyboard = (Keyboard*) InputDevices::getInputDevice(keyboardName);
  if(keyboard == NULL)
    {
      std::cout << "Keyboard \"" << keyboardName << "\" not found" << std::endl;
      return; 
    }
  
  keyboard->start();

  ulong count = 1;

  while(1)
    {
      keyboard->update();

      while(count <= keyboard->outputBuffer().getSampleCount()) {
	KeyboardSample s = keyboard->outputBuffer().getIndex(count);
	Trace("Got %x %x -> '%c'", (int) s.key, s.modifiers, 
	      KeyToChar(s.key, s.modifiers));
	count++;
      }

      usleep(1000);
    }
  // keyboard->stop();
}

void testSpeech(const std::string& name)
{
  Speech *speech = dynamic_cast<Speech *> (InputDevices::getInputDevice(name));
  if(speech == NULL)
    {
      std::cout << "Speech \"" << name << "\" not found" << std::endl;
      return; 
    }
  
  speech->start();

  ulong count = 0;

  while(1) {
    speech->update();
    
    while(count <= speech->outputBuffer().getSampleCount()) {
      Phrase & s = speech->outputBuffer().getIndex(count);
      Trace("Got phrase \"%s\"", s.phrase.c_str());
      count++;
      }
    
    usleep(10000);
  }
  // keyboard->stop();
}

void testKeyboardAndMouse(const std::string& keyboardName, 
			  const std::string& mouseName)
{
  Trace("Testing Keyboard And Mouse \"%s\" \"%s\"", 
	keyboardName.c_str(), mouseName.c_str());

  Keyboard *keyboard = (Keyboard*) InputDevices::getInputDevice(keyboardName);

  if(keyboard == NULL) {
    std::cout << "Keyboard \"" << keyboardName << "\" not found" << std::endl;
    return; 
  }

  Mouse *mouse = InputDevices::getMouse(mouseName);

  if(mouse == NULL) {
    std::cout << "Mouse \"" << mouseName << "\" not found" << std::endl;
    return; 
  }
  
  keyboard->start();
  mouse->start();
  
  ulong count = 0;

  while(1) {
    keyboard->update();
    mouse->update();
    
    while(count < keyboard->outputBuffer().getSampleCount()) {
      KeyboardSample s = keyboard->outputBuffer().getIndex(count);
      Trace("Got %x %x -> '%c'", (int) s.key, s.modifiers, 
	    KeyToChar(s.key, s.modifiers));
      count++;
    }
    
    std::cout << "COORDS = " << mouse->getX() << ", " << mouse->getY() << '\n';
    
    usleep(1000);
  }
  // keyboard->stop();
}


void testTracker(const std::string& name)
{
  Tracker *tracker = (Tracker*) InputDevices::getInputDevice(name);
  if(tracker == NULL)
    {
      std::cout << "Tracker not found" << std::endl;
      return; 
    }
  
  int n = 600;
  tracker->requireHistory(100);
  tracker->start();
  int sensor = tracker->findSensor("paa");
  std::cout << "Sensor 'paa' is: " << sensor << std::endl; 
  while(n--) {
    tracker->update();
    Vector3 v = tracker->getLocation(sensor);
    Matrix3 m = tracker->getRotation(sensor);
    
    std::cout << " v: " << v[0] << ' ' << v[1] << ' ' << v[2] << "\n";
    std::cout << " m: "<< m[0][0] << ' ' << m[0][1] << ' ' << m[0][2] 
	      << "\n";
    usleep(200000);
  }
  // tracker->stop();
}

void testNapra(const std::string& name)
{
  Napra *napra = (Napra*) InputDevices::getInputDevice(name);
  if(napra == NULL) {
    std::cout << "Napra not found" << std::endl;
    return; 
  }
  
  int n = 600;
  napra->requireHistory(100);
  napra->start();

  while(n--) {
    napra->update();
    const NapraSample & ns = napra->getLatestSample();
    
    const Vector3 & v = ns.fingerTips[0][0];
    uint switches = ns.switches[0];
    
    std::cout << " v: " << v << " " << switches << std::endl;
    usleep(200000);
  }
  // napra->stop();
}

void makeTrackerFile();
void testTrackerFile();

int main(int argc, char **argv)
{
  int test = 0;

  const char * device = 0;

  if(argc > 1) {
    const char * testStr = argv[1]; 
    if(!strcmp(testStr, "tracker"))
      test = 0;
    else if(!strcmp(testStr, "glove"))
      test = 1;
    else if(!strcmp(testStr, "mouse"))
      test = 2;
    else if(!strcmp(testStr, "localmouse"))
      test = 3;
    else if(!strcmp(testStr, "filemouse"))
      test = 4;
    else if(!strcmp(testStr, "filetracker"))
      test = 5;
    else if(!strcmp(testStr, "fileglove"))
      test = 6;
    else if(!strcmp(testStr, "keyboard"))
      test = 7;
    else if(!strcmp(testStr, "keyboard-and-mouse"))
      test = 8;
    else if(!strcmp(testStr, "speech"))
      test = 9;
    else if(!strcmp(testStr, "napra"))
      test = 10;
    else {
      printf("Unrecognized option \"%s\"\n"
	     "Usage:  %s [tracker,glove,mouse,napra,localmouse,"
	     "filemouse,filetracker,speech]\n", testStr, argv[0]);
      fflush(0);
      return 0;
    }
  }

  if(argc > 2) {
    device = argv[2];
  }

  DeviceConfig::init();
  InputDevices::init();

  switch(test) {
  case 0:
    testTracker(device ? device : "trakkeri");
    break;
    
  case 1:
    testGlove(device ? device : "hanska");
    break;

  case 2:
    testMouse(device ? device : "hiiri");
    break;
  case 3:
    testMouse(device ? device : "lokaalihiiri");
    break;
  case 4:
    testMouse(device ? device : "tiedostohiiri");
  case 5:
    testTracker(device ? device : "tiedostotrakkeri");
  case 6:
    testGlove(device ? device : "tiedostohanska");
  case 7:
    testKeyboard(device ? device : "localkeyboard");
  case 8:
    testKeyboardAndMouse("localkeyboard2", "lokaalihiiri");
  case 9:
    testSpeech(device ? device : "speech");
  case 10:
    testNapra(device ? device : "speech");
  }


  std::cout << "Cleaning up input devices" << std::endl;

  InputDevices::cleanup();

  std::cout << "Cleaning up config" << std::endl;

  DeviceConfig::cleanup();

  sleep(1);

  std::cout << "Done test" << std::endl;

  return 0;
}

/* Some utility stuff for testing */

void testMouseFile()
{
  DI_DataBuffer2 buffer(256);
  FILE *fp = fopen("mousedata.dump","rb");

  while(buffer.read(fp) > 0)
    {
      buffer.rewind();
      buffer.readInt64();
      MouseSample m;
      m.read(buffer);
      std::cout << m.x << ',' << m.y << std::endl;
      
    }
  fclose(fp);
}
void testTrackerFile()
{
  DI_DataBuffer2 buffer(4096);
  FILE *fp = fopen("trackerdata.dump","rb");

  while(buffer.read(fp) > 0)
    {
      buffer.rewind();
      buffer.readInt64();
      TrackerSample m;
      m.read(buffer);
      std::cout << m.getLocation(0) << std::endl;
    }
  fclose(fp);
}
void makeTrackerFile()
{
  DI_DataBuffer2 buffer(10024);
  TrackerSample sample;
  sample.setSensorCount(8);

  FILE *fp = fopen("trackerdata.dump","wb");
  
  int cnt = 1000;
  while(cnt--)

    {
      buffer.rewind();
      buffer.writeInt64(TimeStamp::getTimeOfDayStamp());
      Matrix3 m;
      m.identity();
      sample.setRotation(0,m);
      sample.setLocation(0,Vector3(1.0,2.0,3.0));
      sample.write(buffer);
      buffer.write(fp, true);
      usleep(100);
    }

  fclose(fp);
}
