/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#include <Fluid/Config.hpp>
#include <Fluid/Debug.hpp>
#include <Fluid/Error.hpp>
#include <Fluid/GloveRecognizer.hpp>
#include <Fluid/InputDevices.hpp>
#include <Fluid/MessagePrinter.hpp>
#include <Fluid/MessageStringPrinter.hpp>
#include <Fluid/Processors.hpp>
#include <Fluid/SensorHitDetector.hpp>
#include <Fluid/TrackerSensorSelector.hpp>

#include <iostream>

#include <stdio.h>
#include <unistd.h>

using namespace Fluid;

static void gloveTest(const std::string & glovename)
{
  Glove *glove = InputDevices::getGlove(glovename);
  
  if(!glove) {
    std::cerr << "Glove not found. (Possible reason: No configuration chunk of name '" 
	      << glovename << "' in config file)" << std::endl;
    return;
  }
  else 
    std::cout << "Got glove" << std::endl;

  glove->start();

  std::string nodePath("../inputdevices/");
  nodePath += glovename;
  nodePath += "/output";

  GloveRecognizerRequest request("glove-gestures",
				 "glove-gestures.config", 
				 nodePath,
				 false);
  Processors processors;

  ProcessorNode * node = processors.requestNode(&request);
  
  GloveRecognizer *gr = dynamic_cast<GloveRecognizer *> (node);

  printf("Node  = %p, GloveRecognizer = %p\n", node, gr);
  fflush(0);
  
  if(gr) {

    std::cout << "nof gestures = " << gr->nofGestures() << std::endl;

    MessageStringPrinter *esp = new MessageStringPrinter;
    gr->msgAddSink("", "", esp);
    // gr->addEventStringListener(esp);
    // processors.addNode(esp);
  }

  int n = 100;

  processors.start();

  Trace("Running the system");

  while(n--) {

    // Trace("Loop %d", n);

    Fluid::InputDevices::update();

    processors.update();
    usleep(100000);

    if(n % 30 == 0) {
      printf(".");
      fflush(0);
    }
  }

  Trace("Stopping the processors");

  processors.stop();

  Trace("Stopping the glove");

  glove->stop();

  Trace("Stopped operation");
}

static void trackerTest(const std::string & trackername)
{
  Tracker *tracker = InputDevices::getTracker(trackername);
  
  if(!tracker) {
    printf("Tracker \"%s\" not found\n",trackername.c_str());
    fflush(0);
    return;
  }
  else 
    printf("Got tracker \"%s\" (%p)\n",trackername.c_str(), tracker);

  fflush(0);

  tracker->start();

  // std::string basePath("../inputdevices/");
  // basePath += trackername;

  Processors processors;
  
  TrackerSensorSelector *tss1 = 
    new TrackerSensorSelector(&processors, 
			      "left-hand", 
			      trackername, 
			      "left-hand");
  TrackerSensorSelector *tss2 = 
    new TrackerSensorSelector(&processors, 
			      "right-hand", 
			      trackername, 
			      "right-hand");
  
  SensorHitDetectorRequest request("left-hand/location", 
				   "right-hand/location",
				   10.0, 10000);

  ProcessorNode * node = processors.requestNode(&request);
  
  SensorHitDetector *shd = dynamic_cast<SensorHitDetector *> (node);

  Trace("Node  = %p, SHD = %p\n", node, shd);
  
  if(shd) {
    MessagePrinter *ep = new MessagePrinter;
    
    shd->msgAddSink(ep);
    processors.addNode(shd);
  }

  int n = 200;

  processors.start();

  if(shd)
    Trace("SHD is %s", shd->ok() ? "good" : "bad");

  Trace("tss1 is %s", tss1->ok() ? "good" : "bad");
  Trace("tss2 is %s", tss2->ok() ? "good" : "bad");

  while(n--) {
    tracker->update();
    processors.update();

    std::cout << tracker->getLocation(0) << std::endl;
    usleep(100000); // 0.1 seconds of sleep
  }

  processors.stop();

  tracker->stop();
}

int main(int argc, char **argv)
{
  int test = 0;

  std::string glovename = "hanska";
  std::string trackername = "trakkeri";
  std::string localglovename = "lokaalihanska";

  if(argc > 1) {
    const char * testStr = argv[1]; 
    if(!strcmp(testStr, "-tracker"))
       test = 0;
    else if(!strcmp(testStr, "-glove"))
       test = 1;
    else if(!strcmp(testStr, "-localglove"))
       test = 2;
    else {
      printf("Unrecognized option \"%s\"\n"
	     "Usage:  %s [-tracker,-glove,-localglove]\n", testStr, argv[0]);
      fflush(0);
      return 0;
    }
  }

  /* 
  logStream << "Reading the configuration" << std::endl;

  GloveRecognizer gr;

  gr.readConfiguration("glove-gestures.config");

  fflush(0);

  logStream << "Read the configuration" << std::endl;

  gr.write(&logStream);
  */

  DeviceConfig::init();
  InputDevices::init();

  switch(test) {
  case 0:
    trackerTest(trackername);
    break;
  case 1:
    gloveTest(glovename);
    break;
  case 2:
    gloveTest(localglovename);
    break;
  }

  Trace("Done test");

  InputDevices::cleanup();
  DeviceConfig::cleanup();

  return 0;
}
