/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#include <Fluid/DataSource.hpp>
#include <Fluid/Tracker.hpp>
#include <Fluid/Debug.hpp>
#include <Fluid/Config.hpp>
#include <Fluid/ConfigReader.hpp>
#include <Fluid/DeviceFactory.hpp>

#include <unistd.h>
#include <sstream>

#include "di_thread.h"
#include "di_socket.h"
#include "mostar/WL/Mostar.h"
 
using namespace Fluid;

class MoStarTrackerSource: public DataSource<TrackerSample>, public DI_Thread
{
public:
  MoStarTrackerSource(const std::string& server, double sampleRate, int bufferSize);
  void setRotation(const Matrix3& m) { calibrateRotation = m; }
  void setTranslation(const Vector3& t) { calibrateTranslation = t; }
  void setScale(const Vector3& s) { calibrateScale = s; }

  void start();
  void stop();
  void childLoop();
  virtual float getSampleRate() const { return sampleRate; }

  std::string getDriverName() const { return "MoStarTracker"; }
private:
  std::string server;
  double      sampleRate;
  wlMostar mostar;
  Matrix3 calibrateRotation;
  Vector3 calibrateTranslation;
  Vector3 calibrateScale;

  bool    m_continue;
};

MoStarTrackerSource::MoStarTrackerSource(const std::string& server, double sampleRate, int bufferSize)
  : DataSource<TrackerSample> (bufferSize),
  server(server),
  sampleRate(sampleRate),
  mostar(server.c_str())
{
}

/**
 * Connects the MoStar Terminal Server and spawns a thread to read the data from it.
 *
 * In case of failure this function prints error diagnostics to std::cout and 
 * returns without starting the device thread at all.
 * 
 */
void MoStarTrackerSource::start()
{
  m_continue = true;

  using namespace std;
  //
  // CONNECT
  //
  cerr << endl << "Attempting to connect to motionstar..." << endl;
  if (mostar.ConnectTCP() != 0)
    {
      cerr << "FLUID # ERROR - Error while connecting to: " << server << endl;
      cerr << "FLUID # ERROR - ... " << mostar.ErrorString() << endl;
      return;
    }
  cerr << "...Connected." << endl;

  //
  // WAKEUP
  //
  if (mostar.Wakeup() != 0)
    {
      cerr << "FLUIF # ERROR - Error During Wakeup(). " << mostar.ErrorString() << endl;
      mostar.Disconnect();
      return;
    }
  //
  // GET GENERAL STATUS
  //
  wlMostarSys status;
  if (mostar.GetSystemStatus(status) != 0)
    {
      cerr << "FLUID # ERROR: Error During GetSystemStatus(). "
           << mostar.ErrorString() << endl;
      mostar.Shutdown();
      mostar.Disconnect();
      return;
    }

  //
  // SET GENERAL STATUS
  //
  status.SetMeasurementRate((int)(sampleRate*1000)); 
  if (mostar.SetSystemStatus(status) != 0)
    {
      cerr << "FLUID # ERROR: Error During SetSystemStatus(). "
           << mostar.ErrorString() << endl;     
      mostar.Shutdown();
      mostar.Disconnect();
      return;
    }

  //
  // VERIFY GENERAL STATUS
  //
  if (mostar.GetSystemStatus(status) != 0)
    {
      cerr << "ERROR: During GetSystemStatus(). "
           << mostar.ErrorString() << endl;
      mostar.Shutdown();
      mostar.Disconnect();
      return;
    }
  status.Print(cerr);

  //
  // GET/SET DEVICE STATUS
  //
  wlMostarDev dev_status;
  for (int i = 1; i <= status.FlockCount(); i++)
    {
      if (mostar.GetDeviceStatus(dev_status, i) != 0)
        {
          cerr << "FLUID # ERROR: During GetDeviceStatus(). "
               << mostar.ErrorString() << endl;
          exit(1);
        }
      dev_status.SetHemisphere(wlMostar::FRONT_HEMISPHERE);
      //dev_status.SetDataFormat(wlMostar::FORMAT_POS_QUAT);
      dev_status.SetDataFormat(wlMostar::FORMAT_POS_MAT);
      // dev_status.SetDataFormat(wlMostar::FORMAT_POS_ANG);
      // dev_status.SetDataFormat(wlMostar::FORMAT_POS);
      dev_status.SetReportRate(1);
      if (mostar.SetDeviceStatus(dev_status, i) != 0)
        {
          cerr << "FLUID # ERROR: During SetDeviceStatus(). "
               << mostar.ErrorString() << endl;
	  return;
        }
    }

  //
  // VERIFY DEVICE STATUS
  //
  for (int i = 1; i <= status.FlockCount(); i++)
    {
      if (mostar.GetDeviceStatus(dev_status, i) != 0)
        {
          cerr << "FLUID # ERROR: During GetDeviceStatus(). "
               << mostar.ErrorString() << endl;
          return;
        }
      dev_status.Print(cerr);
    }

  //
  // RUN CONTINUOUS
  //
  if (mostar.RunContinuous() != 0)
    {
      cerr << "FLUID # ERROR: During RunContinuous(). "
           << mostar.ErrorString() << endl;
      mostar.Shutdown();
      mostar.Disconnect();
      return;
    }
  cerr << "FLUID # MoStarTracker Running" << endl;

  run();
}

/**
 * Kills the device thread and shuts down the connection to tracker terminal
 * server.
 *
 */

void MoStarTrackerSource::stop()
{
  using namespace std;
  m_continue = false;
  waitEnd();
  //
  // SHUTDOWN
  //
  cerr << "Stopping..." << endl;
  if (mostar.StopData() != 0)
    {
      cerr << "ERROR: During StopData(). " << mostar.ErrorString() << endl;
      return;
    }

  cerr << "Shutdown..." << endl;
  if (mostar.Shutdown() != 0)
    {
      cerr << "ERROR: During Shutdown(). " << mostar.ErrorString() << endl;
      return;
    }

  cerr << "Disconnecting..." << endl;
  if (mostar.Disconnect() != 0)
    {
      cerr << "ERROR: During Disconnect(). " << mostar.ErrorString() << endl;
      return;
    }

  cerr << "Disconnected." << endl;
}

/**
 * The main loop where the data is read from tracker.
 *
 * Orientation and location is read in in form of 4x4
 * matrix. This data is then calibrated by rotation,
 * translation and scale that are specified in configuration
 * file.
 *
 */

void MoStarTrackerSource::childLoop()
{
  while(m_continue)
    {

      wlMostarPack packet;

      if (mostar.GetDataPacket(packet) != 0)
        {
          cerr << "ERROR: During GetDataPacket(). "
               << mostar.ErrorString() << endl;
          mostar.Shutdown();
          mostar.Disconnect();
          return;
        }
      Matrix3 rotation1, rotation;
      Vector3 location;
      int n = packet.NumRecords();
      TrackerSample sample(n);
      for(int i = 0; i < n; i++)
	{

	  wlMostarRec record;
	  packet.GetRecord(i,record);
	  float m[4][4];
	  record.GetXform(m);

	  /*
	    MoStar coordinate system is like follows:
	    Looking cube from birds eye:

             x
	      
	    ***
            *Z*   y
            ***
             |
	     
	    Z is into the ground

	    (This is according to MoStar manual page 65.)
	  */
	  
	  rotation[0][0] = m[0][0];
	  rotation[0][1] = m[0][1];
	  rotation[0][2] = m[0][2];
	  rotation[1][0] = m[1][0];
	  rotation[1][1] = m[1][1];
	  rotation[1][2] = m[1][2];
	  rotation[2][0] = m[2][0];
	  rotation[2][1] = m[2][1];
	  rotation[2][2] = m[2][2];
	  
	  Matrix3 ct = calibrateRotation;
	  ct.transpose();
	  rotation = rotation * ct;
	  
	  location[0] = m[3][0];
	  location[1] = m[3][1];
	  location[2] = m[3][2]; 
	  location.scale(calibrateScale);
	  location = calibrateRotation * location;
	  location += calibrateTranslation;

	  /* Added one more transposition to make the matrix appear ok
	     to most users. -Tommi */
	  rotation.transpose();

	  sample.setRotation(i,rotation);
	  sample.setLocation(i,location);
	}
      
      write(sample);
    }
}

class MoStarFactory : public DeviceFactory
{
public:
  virtual std::string      getDriverName() const { return "MoStarTracker"; }
  MoStarTrackerSource*     createDataSource(Fluid::Chunk* chunk)
  {
    std::string server;
    double sampleRate;
    Vector3 scale,translation;
    Matrix3 rotation;
    int bufferSize;

    Variant v = chunk->get("sampleRate");
    if( v.getType() != Variant::INVALID)
      sampleRate = v;
    else 
      sampleRate = 105.8;
    
    v = chunk->get("server");
    if( v.getType() != Variant::INVALID)
      server = std::string(v);
    else
      server = "motion.tml.hut.fi";
    
    v = chunk->get("bufferSize");
    if( v.getType() != Variant::INVALID )
      bufferSize = v;
    else 
      bufferSize = 30;
    /* --- New --- */
    v = chunk->get("translation");
    if( v.getType() != Variant::INVALID )
      {
	sscanf(std::string(v).c_str(),"%f %f %f",&(translation[0]),&(translation[1]),&(translation[2]));
      }
    else
      translation.clear();
    v = chunk->get("scale");
    if( v.getType() != Variant::INVALID)
      {
	sscanf(std::string(v).c_str(),"%f %f %f",&(scale[0]),&(scale[1]),&(scale[2]));
	std::cout << "MostarScale = " << scale << std::endl;

      }
    else
      scale.make(1.0f,1.0f,1.0f);
    
    v = chunk->get("rotation");
    if( v.getType() != Variant::INVALID)
      {   
	sscanf(std::string(v).c_str(),"%f %f %f %f %f %f %f %f %f",
	       &(rotation[0][0]),&(rotation[0][1]),&(rotation[0][2]),
	       &(rotation[1][0]),&(rotation[1][1]),&(rotation[1][2]),
	       &(rotation[2][0]),&(rotation[2][1]),&(rotation[2][2]));

	std::cout << "MostarRotation = " << rotation << std::endl;
      }
    else
      {
	rotation.identity();
      }

    MoStarTrackerSource *source = new MoStarTrackerSource(server,sampleRate,bufferSize);
    source->setRotation(rotation);
    source->setTranslation(translation);
    source->setScale(scale);
    return source;
  }

  virtual InputDeviceBase* create(Chunk* chunk)
  {
    return new Tracker(chunk,createDataSource(chunk));
  }
};

MoStarFactory factory;

extern "C" DeviceFactory* getFactory()
{
  return &factory;
}

