/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, Janne Kontkanen 1999-2002.
 * Tommi.Ilmonen@hut.fi, Janne.Kontkanen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#include <Mostar.hpp>

#include <di_error.h>

const char * Mostar::HemisphereNames[Mostar::HEMISPHERE_COUNT] = {
  "front",
  "rear",
  "upper",
  "lower",
  "left",
  "right"  
};

Mostar::Packet::Packet(ushort index,
		       uchar type,
		       uchar xt)
  : m_index(index),
    m_time(0),
    m_milliSeconds(0),
    m_type(type),
    m_xtype(xt),
    m_protocol(1),
    m_errorCode(0),
    m_errorCodeExt(0),
    m_dataBytes(0)
{}
    
void Mostar::Packet::write(DI_DataBuffer2 &buf) const
{
  buf.writeInt16(m_index);
  buf.writeInt32(m_time);
  buf.writeInt16(m_milliSeconds);
  buf.writeByte(m_type);
  buf.writeByte(m_xtype);
  buf.writeByte(m_protocol);
  buf.writeByte(m_errorCode);
  buf.writeInt16(m_errorCodeExt);
  buf.writeInt16(m_dataBytes);
}

void Mostar::Packet::read(DI_DataBuffer2 &buf)
{
  bzero(this, sizeof(*this));

  m_index = buf.readInt16();
  m_time  = buf.readInt32();
  m_milliSeconds = buf.readInt16();
  m_type         = buf.readByte();
  m_xtype        = buf.readByte();
  m_protocol     = buf.readByte();
  m_errorCode    = buf.readByte();
  m_errorCodeExt = buf.readInt16();
  m_dataBytes    = buf.readInt16();
}

void Mostar::Packet::describe(char *buf) const
{
  sprintf(buf, "Packet %d\n"
	  " time = %d "
	  " msecs = %d\n"
	  " type = %d\n"
	  " xtype = %d\n"
	  " protocol = %d\n"
	  " error = %d, errorext = %d\n"
	  " databytes = %d\n", 
	  (int) m_index,
	  (int) m_time,
	  (int) m_milliSeconds,
	  (int) m_type,
	  (int) m_xtype,
	  (int) m_protocol,
	  (int) m_errorCode,
	  (int) m_errorCodeExt,
	  (int) m_dataBytes);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

void Mostar::Status::setSampleRate(double hz)
{
  int rate = (int) (hz * 1000.0);
  char buf[7];

  if(rate > 999999) rate = 999999;

  sprintf(buf, "%d", rate);
  memcpy(m_sampleRate, buf, 6);
}

void Mostar::Status::write(DI_DataBuffer2 &buf) const
{
  buf.writeByte(m_all);
  buf.writeByte(m_fbbError);
  buf.writeByte(m_flockCount);
  buf.writeByte(m_serverCount);

  unsigned char tmp = (m_xmitterAddr << 4) & 0xf0;
  buf.writeByte(tmp);
  
  buf.writeBytes(m_sampleRate, 6);
  buf.writeByte(m_chassisId);
  buf.writeByte(m_chassisCount);
  buf.writeByte(m_firstAddr);
  buf.writeInt16(0);
}

void Mostar::Status::read(DI_DataBuffer2 &buf)
{
  bzero(this, sizeof(*this));

  m_all         = buf.readByte();
  m_fbbError    = buf.readByte();
  m_flockCount  = buf.readByte();
  m_serverCount = buf.readByte();

  char tmp = buf.readByte();
  m_xmitterAddr = (tmp >> 4) & 0x0f;
    
  buf.readBytes(m_sampleRate, 6);

  m_chassisId    = buf.readByte();
  m_chassisCount = buf.readByte();
  m_firstAddr    = buf.readByte();
  buf.readBytes(m_fbbStatus, 120);
}

void Mostar::Status::describe(char *buf) const
{
  char tmp[7];
  memcpy(tmp, m_sampleRate, 6);
  tmp[6] = '\0';
    
  sprintf(buf, "Status:\n"
	  " all = %x\n"
	  " fbbError = %d\n"
	  " flockCount = %d\n"
	  " serverCount = %d\n"
	  " xmitterAddr = %d\n"
	  " sampleRate = %s\n"
	  " chassis = %d / %d\n"
	  " firstAddr = %d\n",
	  (int) m_all, 
	  (int) m_fbbError, 
	  (int) m_flockCount, 
	  (int) m_serverCount, 
	  (int) m_xmitterAddr, 
	  tmp,
	  (int) m_chassisId, (int) m_chassisCount, 
	  (int) m_firstAddr);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

void Mostar::BirdStatus::setDataFormat(Format f)
{ 
  uchar s = formatSize(f);
  uchar fmt = f;
  m_dataFormat = ((s << 4) & 0xf0) | (fmt &0x0f);
}

void Mostar::BirdStatus::write(DI_DataBuffer2 &buf) const
{ 
  buf.writeByte(m_status);
  buf.writeByte(m_id);
  buf.writeInt16(m_revision);
  buf.writeByte(m_error);
  buf.writeByte(m_setup);
  buf.writeByte(m_dataFormat);
  buf.writeByte(m_reportRate);
  buf.writeInt16(m_scaling);
  buf.writeByte(m_hemisphere);
  buf.writeByte(m_fbbAddr);
  buf.writeInt32(0); // Four bytes of zero

  buf.writeBytes(m_alphaMin, 14);
  buf.writeBytes(m_alphaMax, 14);
  buf.writeBytes(m_vm, 14);
  buf.writeBytes(m_xyzRefFrame, 6);
  buf.writeBytes(m_xyzAngAlign, 6);
}

void Mostar::BirdStatus::read(DI_DataBuffer2 &buf)
{
  bzero(this, sizeof(*this));

  m_status     = buf.readByte();
  m_id         = buf.readByte();
  m_revision   = buf.readInt16();
  m_error      = buf.readByte();
  m_setup      = buf.readByte();
  m_dataFormat = buf.readByte();
  m_reportRate = buf.readByte();
  m_scaling    = buf.readInt16();
  m_hemisphere = buf.readByte();
  m_fbbAddr    = buf.readByte();

  buf.readInt32(); // Jump four-bytes of zero

  buf.readBytes(m_alphaMin, 14);
  buf.readBytes(m_alphaMax, 14);
  buf.readBytes(m_vm, 14);
  buf.readBytes(m_xyzRefFrame, 6);
  buf.readBytes(m_xyzAngAlign, 6);  
}

void Mostar::BirdStatus::describe(char *buf) const
{
  std::string tmp;

  if(m_setup & FILTER_NARROW_AC)
    tmp = " NARROW_AC ";
  if(m_setup & FILTER_WIDE_AC)
    tmp += " WIDE_AC ";
  if(m_setup & FILTER_DC)
    tmp += " DC ";
  
  sprintf(buf, 
	  "Bird\n"
	  " status = %d\n"
	  " id = %d\n"
	  " revision = %d\n"
	  " error code = %d\n"
	  " setup = %d (%s)\n"
	  " data format = %d\n"
	  " report rate = %d\n"
	  " scaling = %d\n"
	  " hemisphere = %s\n"
	  " fbbAddr = %d\n",
	  (int) m_status, 
	  (int) m_id,
	  (int) m_revision, 
	  (int) m_error, 
	  (int) m_setup, tmp.c_str(),
	  (int) m_dataFormat, 
	  (int) m_reportRate, 
	  (int) m_scaling, 
	  HemisphereNames[(int) m_hemisphere], 
	  (int) m_fbbAddr);
}

int Mostar::BirdStatus::formatSize(Format f)
{
  switch(f) {
  case FORMAT_POS:
  case FORMAT_ANG:
    return 3;
  case FORMAT_MAT:
    return 9;
  case FORMAT_POS_ANG:
    return 6;
  case FORMAT_POS_MAT:
    return 12;
  case FORMAT_QUAT:
    return 4;
  case FORMAT_POS_QUAT:
    return 7;
  case FORMAT_NONE:
  default:
    return 0;
  }
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

void Mostar::Sensor::read(DI_DataBuffer2 &buf)
{
  uchar b0 = buf.readByte();
  uchar b1 = buf.readByte();

  m_fbbAddr  = b0 & 0x7f;
  m_btnFlag  = ((b0 & 0x80) == 0x80);
  m_dataBytes = 2 * (b1 &  0x0F);
  m_dataFmt  = (b1 &  0xF0) >> 4;

  buf.readBytes(m_data, m_dataBytes);
}

void Mostar::Sensor::getPosMatrix(float pos[3], float matrix[3][3], 
				  float posfactor) const
{
  if(m_dataFmt == FORMAT_POS_MAT) {
    getVector3(pos, m_data, posfactor);
    getMatrix3(matrix, m_data + 6);
  }
  else {
    DI_error2(DI_ERR_UNIMPLEMENTED, 
	      "Mostar::Sensor::getPosMatrix # No conversion from %d", 
	      (int) m_dataFmt);
  }
}

void Mostar::Sensor::getPos(float pos[3], float posfactor) const
{
  if(m_dataFmt == FORMAT_POS ||
     m_dataFmt == FORMAT_QUAT ||
     m_dataFmt == FORMAT_POS_MAT ||
     m_dataFmt == FORMAT_POS_ANG) {
    getVector3(pos, m_data, posfactor);
  }
  else {
    DI_error2(DI_ERR_UNIMPLEMENTED, 
	      "Mostar::Sensor::getPos # No conversion from %d", 
	      (int) m_dataFmt);
  }
}

void Mostar::Sensor::describe(char *buf) const
{
  sprintf(buf, "Sensor %d\n"
	  " btnFlag = %d\n"
	  " dataBytes = %d\n"
	  " dataFmt = %d\n"
	  " data = %f %f %f %f "
	  "%f %f %f %f "
	  "%f %f %f %f\n", 
	  (int) m_fbbAddr, 
	  (int) m_btnFlag, 
	  (int) m_dataBytes, 
	  (int) m_dataFmt,
	  getFloat(m_data), getFloat(m_data + 2), 
	  getFloat(m_data + 4), getFloat(m_data + 6),
	  getFloat(m_data + 8), getFloat(m_data + 10),
	  getFloat(m_data + 12), getFloat(m_data + 14),
	  getFloat(m_data + 16), getFloat(m_data + 18),
	  getFloat(m_data + 20), getFloat(m_data + 22));
}

void Mostar::Sensor::getVector3(float v[3], const uchar data[6], float factor)
{
  v[0] = factor * getFloat(data);
  v[1] = factor * getFloat(data + 2);
  v[2] = factor * getFloat(data + 4);
}

void Mostar::Sensor::getMatrix3(float m[3][3], const uchar data[18])
{
  for(uint i=0; i < 3; i++)
    for(uint j=0; j < 3; j++)
      m[j][i] = getFloat(data + 2 * (i * 3 + j));
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////



void Mostar::Sample::read(Packet &p, DI_DataBuffer2 &buf)
{
  // This may be buggy.

  m_index = p.index();
  m_time  = p.time();
  m_milliSeconds = p.milliSeconds();
  uint bytes = p.dataBytes();
  uint offset = buf.position();

  for(m_sensorCount = 0; buf.position() - offset < bytes; m_sensorCount++) {
    m_sensors[m_sensorCount].read(buf);
    
    if(m_sensors[m_sensorCount].btnFlag()) // Skip two bytes ????
      buf.readInt16();
    // buf.readInt16(); // Skip two (???)
  }

}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

Mostar::Mostar()
  : m_index(0),
    m_error(MERR_NONE)
{
  m_buffer.resize(2048);
  m_buffer.setLittleEndian(false);
}

Mostar::~Mostar()
{
  disconnect();
}

bool Mostar::connect(const char * address, bool useTcp)
{
  disconnect();

  m_error = MERR_NONE;

  m_address  = address;

  m_address += useTcp ? ":6000" : ":5000";
  
  bool ok = m_socket.open(m_address.c_str(), false, useTcp);

  if(ok && useTcp)
    m_socket.setNoDelay(true);

  return ok;
}

bool Mostar::connectAndStartSampling(const char * address, bool useTcp,
				     float sampleRate, Format dataFormat, 
				     bool wideACFilter,
				     bool narrowACFilter,
				     bool dcFilter,
				     bool verbose,
				     Hemisphere hemisphere)
{
  uint i;
  char buf[2048];

  const char * fname = "Mostar::connectAndStartSampling";

  if(!connect(address, useTcp)) {
    if(verbose) 
      DI_error2(DI_ERR_UNKNOWN, "%s # Failed to connect to %s - %s", 
		fname, address, errorStr());
    return false;
  }

  if(!wakeUp()) {
    if(verbose) 
      DI_error2(DI_ERR_UNKNOWN, "%s # Failed to wake up - %s", 
		fname, errorStr());
    return false;
  }

  Status status;

  if(!getStatus(status)) {
    if(verbose) 
      DI_error2(DI_ERR_UNKNOWN, "Failed to get status - %s", errorStr());
    return false;
  }

  if(verbose) DI_trace("%s # Setting general status", fname);
  status.setSampleRate(sampleRate);
  
  if(!setStatus(status)) {
    if(verbose) 
      DI_error2(DI_ERR_UNKNOWN, "%s # Failed to set status - %s", 
		fname, errorStr());
    return false;
  }
  
  for(i = 1; i <= status.flockCount(); i++) {
    BirdStatus b;
    
    if(verbose) DI_trace2("%s # Getting status for bird %d", fname, i);

    if(!getBirdStatus(b, i)) {
      if(verbose) 
	DI_error2(DI_ERR_UNKNOWN, 
		  "%s # Failed to get status for bird %d - %s", fname, 
		  i, errorStr());
      return false;
    }
    if(verbose) {
      b.describe(buf);
      puts(buf);
    }

    b.setHemisphere(hemisphere);
    b.setDataFormat(dataFormat);
    b.setReportRate(1);

    // Filter settings:
    b.setWideACFilter(wideACFilter);
    b.setNarrowACFilter(narrowACFilter);
    b.setDCFilter(dcFilter);

    if(verbose) DI_trace2("%s # Setting status for bird %d", fname, i);
    
    if(!setBirdStatus(b, i)) {
      if(verbose) 
	DI_error2(DI_ERR_UNKNOWN, "%s # Failed to set status for bird %d - %s", fname, 
		  i, errorStr());
      return false;
    }

    if(!getBirdStatus(b, i)) {
      if(verbose) 
	DI_error2(DI_ERR_UNKNOWN, 
		  "%s # Failed to re-get status for bird %d - %s", fname, 
		  i, errorStr());
      return false;
    }
    
    if(verbose) {
      b.describe(buf);
      puts(buf);
      DI_trace2("%s # Set status for bird %d", fname, i);
    }

  }

  if(!startSampling()) {
      if(verbose) 
	DI_error2(DI_ERR_UNKNOWN, "%s # Failed to start sampling - %s", fname, 
		  errorStr());
      return false;
  }
  
  if(verbose) 
    DI_trace2("%s # EXIT OK", fname);

  return true;
}

bool Mostar::disconnect()
{
  m_error = MERR_NONE;
  m_socket.close();

  return true;
}

bool Mostar::shutdown()
{
  m_error = MERR_NONE;

  Packet p(m_index++, COM_SHUTDOWN);

  send(p);

#if 0
  /* Ascension should reply. It does not. */
  receive(p);

  if(p.type() != RES_SHUTDOWN)
    m_error = MERR_BAD_RESPONSE;
#endif

  return true;
}

bool Mostar::wakeUp(ushort index)
{
  m_error = MERR_NONE;
  
  m_index = index;
  
  Packet p(index, COM_WAKEUP);

  if(!send(p))
    return false;

  if(!receive(p))
    return false;
  
  if(p.type() != RES_WAKEUP) {
    m_error = MERR_BAD_RESPONSE;
    return false;
  }

  return true;
}

bool Mostar::startSampling()
{
  m_error = MERR_NONE;
  
  Packet p(m_index++, COM_RUN_CONTINUOUS);

  if(!send(p))
    return false;

  if(!receive(p))
    return false;
  
  if(p.type() != RES_RUN_CONTINUOUS) {
    m_error = MERR_BAD_RESPONSE;
    return false;    
  }
  
  return true;
}

bool Mostar::getStatus(Status & s)
{
  m_error = MERR_NONE;
  
  Packet p(m_index++, COM_GET_STATUS);

  // puts("Mostar::getStatus: Sending");

  if(!send(p))
    return false;

  // puts("Mostar::getStatus: Receiving");

  if(!receive(p))
    return false;

  // puts("Got some response");
  
  if(p.type() != RES_GET_STATUS) { // GET_STATUS vs SEND_SETUP ???
    m_error = MERR_BAD_RESPONSE;
    return false;
  }
  
  s.read(m_buffer);

  return true;
}

bool Mostar::setStatus(const Status &s)
{
  m_error = MERR_NONE;
  
  Packet p(m_index++, COM_SEND_SETUP);
  p.setDataBytes(STATUS_SIZE);

  send(p);

  m_buffer.setPosition(0);
  s.write(m_buffer);
  m_socket.write(m_buffer.buffer(), STATUS_SIZE);

  receive(p);
  
  if(p.type() != RES_SEND_SETUP) {
    m_error = MERR_BAD_RESPONSE;
    return false;
  }
  
  return true;
}


bool Mostar::getBirdStatus(BirdStatus &b, int id)
{
  bzero(&b, sizeof(b));

  Packet p(m_index++, COM_GET_STATUS, id); // GET_STATUS AGAIN ???

  if(!send(p))
    return false;
  
  receive(p);
  
  if(p.type() != RES_GET_STATUS) {
    m_error = MERR_BAD_RESPONSE;
    return false;
  }
  
  b.read(m_buffer);

  return true;
}

bool Mostar::setBirdStatus(const BirdStatus &b, int id)
{
  m_error = MERR_NONE;
  
  Packet p(m_index++, COM_SEND_SETUP, id);
  p.setDataBytes(70);

  if(!send(p))
    return false;

  m_buffer.setPosition(0);
  b.write(m_buffer);
  m_socket.write(m_buffer.buffer(), m_buffer.position());

  DI_trace2("Mostar::setBirdStatus # sent %d bytes", m_buffer.position());
  
  if(!receive(p))
    return false;
  
  if(p.type() != RES_SEND_SETUP) {
    m_error = MERR_BAD_RESPONSE;
    return false;
  }

  return true;
}

bool Mostar::getSample(Sample &s)
{
  m_error = MERR_NONE;
  
  Packet p;

  // puts("GS1");

  if(!receive(p))
    return false;

  if(p.type() != COM_DATA_PACKET) { // COM and RES upside-down
    m_error = MERR_BAD_RESPONSE;
    return false;    
  }

  s.read(p, m_buffer);

  // puts("GS2");

  Packet p2(m_index++, RES_DATA_PACKET);

  if(!send(p2))
    return false;

  // puts("GS3");

  return true;
}

bool Mostar::send(const Packet &p)
{
  m_buffer.setPosition(0);
  p.write(m_buffer);
  
  if(m_socket.write(m_buffer.buffer(), 
		    PACKET_SIZE) != PACKET_SIZE) {
    m_error = MERR_BAD_SEND;
    return false;
  }

  return true;
}

bool Mostar::receive(Packet &p)
{
  // Clear the buffer to make sure old crap is not reused
  bzero(m_buffer.buffer(), PACKET_SIZE);
  bool ok = m_socket.read(m_buffer.buffer(), PACKET_SIZE) == PACKET_SIZE;
  
  if(!ok) {
    m_error = MERR_BAD_RESPONSE;
    return false;
  }

  m_buffer.setPosition(0);
  p.read(m_buffer);

  ushort dataBytes = p.dataBytes();

  if((uint) (dataBytes + PACKET_SIZE) > m_buffer.size())
    m_buffer.resize(dataBytes + PACKET_SIZE);

  if(dataBytes) {

    // DI_trace2("Mostar::receive # Reading %d extra bytes", (int) dataBytes);
    
    ok = m_socket.read(m_buffer.buffer() + PACKET_SIZE, 
		       dataBytes) == dataBytes;
  }

  if(!ok)
    m_error = MERR_BAD_RESPONSE;

  /* char buf[2048];
     p.describe(buf);
     puts(buf); */

  return ok;
}

bool Mostar::test(const char * address)
{
  // int i;
  char buf[2048];

  Mostar mostar;

  mostar.connectAndStartSampling(address, true, 68.6, FORMAT_POS_MAT, 
				 true, true, true, true);
  
  int samples = 300;
  
  Sample sample;

  for(int s = 0; s < samples; s++) {
    DI_trace("Getting a sample");
    if(!mostar.getSample(sample)) {
      DI_error(DI_ERR_UNKNOWN, "Failed to get sample - %s", 
	       mostar.errorStr());
      return false;
    }

    float loc[3], rot[3][3];
    
    const Sensor & sensor = sample.getSensor(0);
    
    sensor.getPosMatrix(loc, rot);
    printf("%d: %f  %f  %f\n-------------\n", s, loc[0], loc[1], loc[2]);
    fflush(0);


    for(uint h=0; h < sample.sensorCount(); h++) {
      sample.getSensor(h).describe(buf);
      puts(buf);
    }
  }

  DI_trace2("Mostar::test # exit ok");

  return true;
}

const char * Mostar::errorToStr(MostarError e)
{
  const char * str;

  if(e == MERR_NONE)
    str = "No error";
  else if(e == MERR_ONE_MISSING_PACK)
    str = "Missing packet";
  else if(e == MERR_FEW_MISSING_PACK)
    str = "Few missing packets";
  else if(e == MERR_MANY_MISSING_PACK)
    str = "Many missing packets";
  else if(e == MERR_REPEAT_SEQ_NUM)
    str = "Repeated sequence number";
  else if(e == MERR_UNEXPECTED_PACK)
    str = "Unexpected packet type";
  else if(e == MERR_BAD_PACK)
    str = "Bad packet";
  else if(e == MERR_BAD_STATUS)
    str = "Bad status";
  else if(e == MERR_BAD_SETUP)
    str = "Bad setup";
  else if(e == MERR_SYSTEM_NOT_READY)
    str = "System not ready";
  else if(e == MERR_BAD_RESPONSE)
    str = "Bad response";
  else if(e == MERR_BAD_SETUP)
    str = "Bad send";
  else 
    str = "Unknown error code";
  
  return str;
}
