/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, Janne Kontkanen 1999-2002.
 * Tommi.Ilmonen@hut.fi, Janne.Kontkanen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#ifndef FLUID_MOSTAR_HPP
#define FLUID_MOSTAR_HPP

#include <di_socket.h>

#include <netinet/in.h>
#include <strings.h>

#include <string>

class Mostar {
public:
  
  /// Available hemispheres
  enum Hemisphere {
    ///
    FRONT_HEMISPHERE = 0,
    ///
    REAR_HEMISPHERE = 1,
    ///
    UPPER_HEMISPHERE = 2,
    ///
    LOWER_HEMISPHERE = 3,
    ///
    LEFT_HEMISPHERE = 4,
    ///
    RIGHT_HEMISPHERE = 5,
    ///
    HEMISPHERE_COUNT = 6
  };

  static const char * HemisphereNames[HEMISPHERE_COUNT];

  enum FilterSettings {
    FILTER_WIDE_AC = 0x2,
    FILTER_NARROW_AC = 0x4,
    FILTER_DC = 0x1
  };

  /// Command/Response Codes (use in DataPacket::type field)
  enum Command	
  {
    ///
    COM_WAKEUP = 10,
    ///
    COM_SHUTDOWN = 11,
    ///
    COM_GET_STATUS = 101,
    ///
    COM_SEND_SETUP = 102,
    ///
    COM_SINGLE_SHOT = 103,	
    ///
    COM_RUN_CONTINUOUS = 104,	
    ///
    COM_STOP_DATA = 105,	
    ///
    COM_DATA_PACKET = 210
  };

  enum Response {
    ///
    RES_WAKEUP = 20,
    ///
    RES_SHUTDOWN = 21,
    ///
    RES_GET_STATUS = 201,
    ///
    RES_SEND_SETUP = 202,
    ///
    RES_SINGLE_SHOT = 203,
    ///
    RES_RUN_CONTINUOUS = 204,
    ///
    RES_STOP_DATA = 205,
    ///
    RES_DATA_PACKET = 211,
    ///
    RES_ILLEGAL = 40,
    ///
    RES_UNKNOWN = 50
  };


  ///
  enum
  {
    ///
    PACKET_SIZE = 16,
    ///
    STATUS_SIZE = 16,
    ///
    BIRD_SIZE = 16
  };

  ///
  enum MostarError
  {
    ///
    MERR_NONE =		0,			

    // Packet Header Errors //
    ///
    MERR_ONE_MISSING_PACK = 	1,
    ///
    MERR_FEW_MISSING_PACK =	2,
    ///
    MERR_MANY_MISSING_PACK =     3,
    ///
    MERR_REPEAT_SEQ_NUM =	4,
    ///
    MERR_UNEXPECTED_PACK =	6,
    ///
    MERR_BAD_PACK =		7,
    ///
    MERR_BAD_STATUS = 	        8,
    ///
    MERR_BAD_SETUP =		9,
    ///
    MERR_SYSTEM_NOT_READY =	100,
    ///
    MERR_BAD_RESPONSE =		300,
    ///
    MERR_BAD_SEND =		301
  };

  enum Format
  {
    // Data Format Codes (bits 0-3)
    ///
    FORMAT_NONE = 0,
    ///
    FORMAT_POS = 1,
    ///
    FORMAT_ANG = 2,
    ///
    FORMAT_MAT = 3,
    ///
    FORMAT_POS_ANG = 4,
    ///
    FORMAT_POS_MAT = 5,
    ///
    FORMAT_QUAT = 7,
    ///
    FORMAT_POS_QUAT = 8
  };

  class Packet {

  public:
    Packet(ushort index = 0,
	   uchar type   = RES_UNKNOWN,
	   uchar xt     = 0);

    ~Packet() {}

    ushort index()        const { return m_index; }
    uint time()           const { return m_time; }
    ushort milliSeconds() const { return m_milliSeconds; }

    uchar type()      const { return m_type; }
    uchar xtype()     const { return m_xtype; }
    uchar protocol()  const { return m_protocol; }
    uchar errorCode() const { return m_errorCode; }
    
    ushort dataBytes() const { return m_dataBytes; }
    void setDataBytes(ushort bytes) { m_dataBytes = bytes; }

    void write(DI_DataBuffer2 &) const;
    void read(DI_DataBuffer2 &);

    void describe(char *) const;

  private:

    ushort	m_index;
    uint	m_time;
    ushort	m_milliSeconds;
    uchar	m_type;
    uchar	m_xtype;
    uchar	m_protocol;
    uchar	m_errorCode;	
    ushort	m_errorCodeExt;
    ushort	m_dataBytes;
  };

  class Status {
  public:
    Status() { bzero(this, sizeof(*this)); }
    ~Status() {}

    uchar flockCount() const { return m_flockCount; }

    void setSampleRate(double hz);

    void write(DI_DataBuffer2 &) const;
    void read(DI_DataBuffer2 &);

    void describe(char *) const;

  private:
    
    uchar	m_all;
    uchar	m_fbbError;
    uchar	m_flockCount;
    uchar	m_serverCount;
    uchar	m_xmitterAddr;
    char	m_sampleRate[6];
    uchar	m_chassisId;
    uchar	m_chassisCount;
    uchar	m_firstAddr;
    
    uchar	m_fbbStatus[120];
  };

  class BirdStatus {
  public:
    BirdStatus() { bzero(this, sizeof(*this));}
    ~BirdStatus() {}
    
    void setHemisphere(uchar h) { m_hemisphere = h; }
    void setDataFormat(Format f);
    void setReportRate(uchar r) { m_reportRate = r; }

    void write(DI_DataBuffer2 &) const;
    void read(DI_DataBuffer2 &);

    void describe(char *) const;

    static int formatSize(Format f);

    void setNarrowACFilter(bool use)
    { 
      if(use) m_setup |= FILTER_NARROW_AC;
      else m_setup &= ~(FILTER_NARROW_AC);
    }

    void setWideACFilter(bool use)
    { 
      if(use) m_setup |= FILTER_WIDE_AC;
      else m_setup &= ~(FILTER_WIDE_AC);
    }

    void setDCFilter(bool use)
    { 
      if(use) m_setup |= FILTER_DC;
      else m_setup &= ~(FILTER_DC);
    }

  protected:

    uchar  m_status;
    uchar  m_id;
    ushort m_revision;
    uchar  m_error;
    uchar  m_setup;
    uchar  m_dataFormat;
    uchar  m_reportRate;
    ushort m_scaling;
    uchar  m_hemisphere;
    uchar  m_fbbAddr;
    uchar  m_alphaMin[14];
    uchar  m_alphaMax[14];
    uchar  m_vm[14];
    uchar  m_xyzRefFrame[6];
    uchar  m_xyzAngAlign[6];
  };

  class Sensor {
  public:
    Sensor() {}
    ~Sensor() {}

    void read(DI_DataBuffer2 &);

    void getPosMatrix(float pos[3], float matrix[3][3], 
		      float posfactor = 144.0f) const;

    void getPos(float pos[3], float posfactor = 144.0f) const;

    void describe(char *) const;

    bool btnFlag() const { return m_btnFlag; }

  private:

    static float getFloat(const uchar * data)
    {
      ushort tmp = (data[0] << 8) | data[1];
      return ((short) tmp) / 32768.0f;
    }
    static void getVector3(float v[3], const uchar data[6], float factor);
    static void getMatrix3(float m[3][3], const uchar data[18]);

    uchar m_fbbAddr;
    bool  m_btnFlag;
    uchar m_dataBytes;
    uchar m_dataFmt;
    
    uchar m_data[24];
  };

  class Sample {
  public:
    Sample() { bzero(this, sizeof(*this)); }
    ~Sample() {}

    void read(Packet &, DI_DataBuffer2 &);

    const Sensor & getSensor(uint i) { return m_sensors[i]; }

    uint sensorCount() { return m_sensorCount; }

  private:
    int  m_index;
    uint m_time;
    int  m_milliSeconds;
    uint m_sensorCount;

    Sensor m_sensors[120];
  };

  Mostar();
  ~Mostar();

  bool connect(const char * address, bool useTcp);
  bool connectAndStartSampling(const char * address, bool useTcp,
			       float sampleRate, Format dataFormat, 
			       bool wideACFilter = true,
			       bool narrowACFilter = false,
			       bool dcFilter = true,
			       bool verbose = true,
			       Hemisphere h = FRONT_HEMISPHERE);
  bool disconnect();
  bool shutdown();
  bool wakeUp(ushort index = 0);
  bool startSampling();
  
  bool getStatus(Status &);
  bool setStatus(const Status &);

  bool getBirdStatus(BirdStatus &, int id);
  bool setBirdStatus(const BirdStatus &, int id);

  bool getSample(Sample &);

  bool send(const Packet &);
  bool receive(Packet &);

  static bool test(const char * address = "motion.tml.hut.fi");

  MostarError error() const { return m_error; }
  const char * errorStr() const { return errorToStr(m_error); }

  static const char * errorToStr(MostarError);

protected:

  std::string    m_address;
  DI_Socket      m_socket;
  DI_DataBuffer2 m_buffer;

  ushort         m_index;

  MostarError    m_error;
};


#endif
