/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#include <Fluid/DataSource.hpp>
#include <Fluid/Tracker.hpp>
#include <Fluid/Debug.hpp>
#include <Fluid/Config.hpp>
#include <Fluid/ConfigReader.hpp>
#include <Fluid/DeviceFactory.hpp>

#include <unistd.h>
#include <sstream>

#include "di_thread.h"
#include "di_socket.h"
#include "Mostar.hpp"
 
using namespace Fluid;

class MoStarTrackerSource: public DataSource<TrackerSample>, public DI_Thread
{
public:
  MoStarTrackerSource(const std::string& server, double sampleRate, int bufferSize);
  void setRotation(const Matrix3& m) { calibrateRotation = m; }
  void setTranslation(const Vector3& t) { calibrateTranslation = t; }
  void setScale(const Vector3& s) { calibrateScale = s; }

  void start();
  void stop();
  void childLoop();
  virtual float getSampleRate() const { return sampleRate; }

  std::string getDriverName() const { return "MotionStarTracker"; }

  /// Use wide AC filter
  void setUseWideAC(bool use) { m_useWideAC = use; }
  void setUseNarrowAC(bool use) { m_useNarrowAC = use; }
  void setUseDC(bool use) { m_useDC = use; }
  void setHemisphere(Mostar::Hemisphere h) { m_hemisphere = h; }

private:
  std::string server;
  double      sampleRate;
  Mostar  mostar;
  Matrix3 calibrateRotation;
  Vector3 calibrateTranslation;
  Vector3 calibrateScale;

  bool    m_continue;

  // Filter stuff:
  bool    m_useWideAC;
  bool    m_useNarrowAC;
  bool    m_useDC;
  Mostar::Hemisphere m_hemisphere;
};

MoStarTrackerSource::MoStarTrackerSource
(const std::string& server, double sampleRate, int bufferSize)
  : DataSource<TrackerSample> (bufferSize),
     server(server),
    sampleRate(sampleRate),
    m_useWideAC(true),
    m_useNarrowAC(true),
    m_useDC(true),
    m_hemisphere(Mostar::FRONT_HEMISPHERE)
{
}

/**
 * Connects the MoStar Terminal Server and spawns a thread to read the data from it.
 *
 * In case of failure this function prints error diagnostics to std::cout and 
 * returns without starting the device thread at all.
 * 
 */
void MoStarTrackerSource::start()
{
  m_continue = true;

  using namespace std;
  //
  // CONNECT
  //
  cerr << endl << "Attempting to connect to motionstar \"" 
       << server << "\" ..." << endl;
  if (!mostar.connectAndStartSampling(server.c_str(), true, sampleRate, 
				      Mostar::FORMAT_POS_MAT,
				      m_useWideAC, m_useNarrowAC, m_useDC,
				      true,
				      m_hemisphere)) {
    cerr << "FLUID # ERROR - Error while connecting to: " << server << endl;
    cerr << "FLUID # ERROR - ... " << mostar.errorStr() << endl;
    return;
  }

  run();
}

/**
 * Kills the device thread and shuts down the connection to tracker terminal
 * server.
 *
 */

void MoStarTrackerSource::stop()
{
  using namespace std;
  m_continue = false;
  waitEnd();
  //
  // SHUTDOWN
  //
  /* cerr << "Stopping..." << endl;
  if (mostar.StopData() != 0)
    {
      cerr << "ERROR: During StopData(). " << mostar.errorStr() << endl;
      return;
    }

  cerr << "Shutdown..." << endl;
  if (mostar.Shutdown() != 0)
    {
      cerr << "ERROR: During Shutdown(). " << mostar.errorStr() << endl;
      return;
    }
  */

  cerr << "Disconnecting..." << endl;
  if (!mostar.disconnect())
    {
      cerr << "ERROR: During Disconnect(). " << mostar.errorStr() << endl;
      return;
    }

  cerr << "Disconnected." << endl;
}

/**
 * The main loop where the data is read from tracker.
 *
 * Orientation and location is read in in form of 4x4
 * matrix. This data is then calibrated by rotation,
 * translation and scale that are specified in configuration
 * file.
 *
 */

void MoStarTrackerSource::childLoop()
{
  Mostar::Sample msample;
    
  if (!mostar.getSample(msample) ) {
    Error("MoStarTrackerSource::childLoop # %s(). ", mostar.errorStr());
    mostar.disconnect();
    return;
  }

  int n = msample.sensorCount();
  TrackerSample sample(n);

  while(m_continue) {
    
    if (!mostar.getSample(msample) ) {
      Error("MoStarTrackerSource::childLoop # %s(). ", mostar.errorStr());
      mostar.disconnect();
      return;
    }

    Matrix3 rotation;
    Vector3 location;

    for(int i = 0; i < n; i++) {
	  
      const Mostar::Sensor & sensor = msample.getSensor(i);
      float v[3];
      float m[3][3];
      sensor.getPosMatrix(v, m);

      /*
	MoStar coordinate system is like follows:
	Looking cube from birds eye:

	x
	      
	***
	*Z*   y
	***
	|
	     
	Z is into the ground

	(This is according to MoStar manual page 65.)
      */
	  
      rotation[0][0] = m[0][0];
      rotation[0][1] = m[0][1];
      rotation[0][2] = m[0][2];
      rotation[1][0] = m[1][0];
      rotation[1][1] = m[1][1];
      rotation[1][2] = m[1][2];
      rotation[2][0] = m[2][0];
      rotation[2][1] = m[2][1];
      rotation[2][2] = m[2][2];
	  
      Matrix3 ct = calibrateRotation;
      ct.transpose();
      rotation = rotation * ct;
	  
      location[0] = v[0];
      location[1] = v[1];
      location[2] = v[2]; 
      location.scale(calibrateScale);
      location = calibrateRotation * location;
      location += calibrateTranslation;

      /* Added one more transposition to make the matrix appear ok
	 to most users. -Tommi */
      rotation.transpose();

      sample.setRotation(i, rotation);
      sample.setLocation(i, location);
      
      // DI_trace2("SENSOR %d: %f %f %f", i, location.x, location.y, location.z);
    

      /*
      char buf[2048];
      sensor.describe(buf);
      DI_trace2("SENSOR %d", i);
      puts(buf); */
    }
      
    write(sample);
  }

}

class MoStarFactory : public DeviceFactory
{
public:
  virtual std::string      getDriverName() const { return "MotionStarTracker"; }
  MoStarTrackerSource*     createDataSource(Fluid::Chunk* chunk)
  {
    std::string server;
    double sampleRate;
    Vector3 scale, translation;
    Matrix3 rotation;
    int bufferSize;

    scale.make(1.0);
    translation.make(0.0);
    rotation.identity();

    sampleRate = chunk->get("sampleRate").getDouble(105.8);
    
    server = chunk->get("server").getString("motion.tml.hut.fi");    
    
    bufferSize = chunk->get("bufferSize").getInt(50);
    /* --- New --- */
    std::string s = chunk->get("translation");
    if( s.size()) {
      sscanf(s.c_str(),"%f %f %f",
	     &(translation[0]),&(translation[1]),&(translation[2]));
    }
    else
      translation.clear();

    s = chunk->get("scale").getString();
    if( s.size()) {
      sscanf(s.c_str(),"%f %f %f",&(scale[0]),&(scale[1]),&(scale[2]));
      std::cout << "MostarScale = " << scale << std::endl;
      
    }
    else
      scale.make(1.0f,1.0f,1.0f);
    
    
    s = chunk->get("rotation").getString();
    if( s.size()) {
      sscanf(s.c_str(),"%f %f %f %f %f %f %f %f %f",
	     &(rotation[0][0]),&(rotation[0][1]),&(rotation[0][2]),
	     &(rotation[1][0]),&(rotation[1][1]),&(rotation[1][2]),
	     &(rotation[2][0]),&(rotation[2][1]),&(rotation[2][2]));
      
      std::cout << "MostarRotation = " << rotation << std::endl;
    }
    else
      rotation.identity();

    MoStarTrackerSource *source = new 
      MoStarTrackerSource(server,sampleRate,bufferSize);

    source->setRotation(rotation);
    source->setTranslation(translation);
    source->setScale(scale);


    source->setUseWideAC(chunk->get("wide_ac_filter").getInt(1));
    source->setUseNarrowAC(chunk->get("narrow_ac_filter").getInt(1));
    source->setUseDC(chunk->get("dc_filter").getInt(1));
    
    s = chunk->get("hemisphere").getString();
    if(s.size()) {
      int hemi = Mostar::FRONT_HEMISPHERE;
      
      for(int i = 0; i < Mostar::HEMISPHERE_COUNT; i++)
	if(strcmp(s.c_str(), Mostar::HemisphereNames[i]) == 0)
	  hemi = i;

      if(hemi >= 0)
	source->setHemisphere((Mostar::Hemisphere) hemi);
    }

    return source;
    }

  virtual InputDeviceBase* create(Chunk* chunk)
  {
    return new Tracker(chunk,createDataSource(chunk));
  }
};

MoStarFactory factory;

extern "C" DeviceFactory* getFactory()
{
  return &factory;
}

