/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, Janne Kontkanen 1999-2002.
 * Tommi.Ilmonen@hut.fi, Janne.Kontkanen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#ifndef FLUID_SIGNAL_CLEANER_HPP
#define FLUID_SIGNAL_CLEANER_HPP

#include <Fluid/Math.hpp>

#include <Iir1.hpp>

namespace Fluid {

  /** A filter that cleans noise pops from a signal. Developed to
        filter the flight-time data from the Napra.

      @author Tommi Ilmonen
  */

  template <class T>
  class SignalCleaner
  {
  public:
    
    /** Parameters for the cleaner. Used to provide the same set of
	parameters to multiple cleaners. */
    
    class Params
    {
    public:
      void init(T min, T max, T threshold, T sensitivity)
      {
	m_min = min;
	m_max = max;
	m_sensitivity = sensitivity;
	m_threshold = threshold;

	m_range = max - min;
      }
      
      T m_min, m_max;
      T m_sensitivity;
      T m_threshold;

      T m_range;
    };

    SignalCleaner() { }
    ~SignalCleaner() {}

    void init(const Params &p)
    {
      m_value = (p.m_min + p.m_max) * 0.5f;
      m_holdTime = 0;
      m_discarded = 0;
    }

    void put(T v, const Params & p)
    {
      if(v > p.m_max || v < p.m_min) {
	// Discard v completely
	m_discarded++;
      }
      else {
	T diff = Math::Abs(m_value - v);
	T badness = diff / p.m_range;

	if(badness < p.m_threshold * Math::Clamp(m_discarded, 1, 3)) {
	  m_holdTime = 0;
	  m_value = v;
	}
	else {
	  int time = (int) (badness * p.m_sensitivity);

	  if(time < m_holdTime)
	    m_holdTime = time;
	  else if(m_holdTime > 1)
	    m_holdTime--;
	  else if(m_holdTime == 1) {
	    m_value = v;
	    m_holdTime = 0;
	  }
	  else
	    m_holdTime = time;
	}
      }

      if(v == m_value) {
	if(m_discarded > 0)
	  m_discarded = 0;
	else
	  m_discarded--;
      }
      else {
	if(m_discarded < 0)
	  m_discarded = 1;
	else
	  m_discarded++;
      }
    }

    const T & get() const { return m_value; }

    int discarded() const { return m_discarded; }

  protected:

    T m_value;

    int m_holdTime;
    int m_discarded;
  };

  template <class T>
  class SignalCleaner2
  {
  public:
    SignalCleaner2() {}
    ~SignalCleaner2() {}

    SignalCleaner<T> & cleaner() { return m_cleaner; }
    Iir1<T>          & iir() { return m_iir; }
    Iir1<T>          & filterLevel() { return m_filterLevel; }

    void put(T v, const typename SignalCleaner<T>::Params & p)
    {
      m_cleaner.put(v, p);

      T level = m_cleaner.discarded() < -2 ? 0.0f : 1.0f;

      // Map between 0.4 and 0.93.
      level = level * (T) 0.43 + (T) 0.5;

      m_filterLevel.putRaiser(level, 0.7);
      m_filterLevel.put(level);
      
      m_iir.setA(m_filterLevel.get());
      m_iir.put(m_cleaner.get());
    }

    T get() const { return m_iir.get(); }

  protected:
    SignalCleaner<T> m_cleaner;
    Iir1<T>          m_iir;
    /// Filtering strength
    Iir1<T>          m_filterLevel;
  };

  template <class T>
  class SignalCleaner3
  {
  public:
    SignalCleaner3() {}
    ~SignalCleaner3() {}

    void put(int n, T v)
    {
      m_inputs[n] = v;
    }

    void calculate(const typename SignalCleaner<T>::Params & p)
    {
      int i;

      for(i = 0; i < 3; i++)
	m_cleaners[i].put(m_inputs[i], p);

      T res[3];
      T good[3];

      for(i = 0; i < 3; i++) {
	res[i]  = m_cleaners[i].get();
	good[i] = m_cleaners[i].filterLevel().get();
      }
      
      int fix = 0;
      
      for(i = 0; i < 3; i++) {
	if(good[i] > 0.6f)
	  fix = fix | (1 << i);
      }
      
      if(fix && (fix != 0x7)) {
	
	/* Move the signals to be switched by using the motion values
	   of the working sensors to calculate the offset. */

	int sources = 0x7 & ~fix;

	T sourceAverageOld = 0.0;
	T sourceAverageNew = 0.0;
	T sourceCount = 0.0;

	for(i = 0; i < 3; i++) {
	  if((sources >> i) == 1) {
	    sourceAverageOld += m_res[i];
	    sourceAverageNew +=   res[i];
	    sourceCount++;
	  }
	}
	
	sourceAverageOld /= sourceCount;
	sourceAverageNew /= sourceCount;

	T move = sourceAverageNew - sourceAverageOld;
	
	for(i = 0; i < 3; i++) {
	  if((fix >> i) == 1) {
	    // float w = Math::Clamp
	    res[i] = m_res[i] + move;
	  }
	}
	
      }
      
      for(i = 0; i < 3; i++)
	m_res[i] = res[i];
    }

    T get(int n) const { return m_res[n]; }

    SignalCleaner2<T> & cleaner(int n) { return m_cleaners[n]; }

protected:

    T m_inputs[3];
    T m_res[3];
    
    SignalCleaner2<T> m_cleaners[3];
  };

}


#endif
