/* Copyright
 *
 * This file is part of FLUID
 *
 * Authors: Karri Palovuori, Tommi Ilmonen, 2004.
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#include "math.h"
#include "helma.h"

#include <strings.h>

float helma_readtime(int index, helma_state *state);

void helma_init(helma_state *state)
{
  bzero(state, sizeof(*state));

  state->cleanerParams.init(0.001, 0.20, 0.03, 20);

  state->filtering = HELMA_FILTERED1;

  for(int i = 0; i < 2; i++) {
    for(int j = 0; j < 3; j++) {
      for(int k = 0; k < 3; k++) {
	for(int l = 0; l < 2; l++) {

	  Fluid::SignalCleaner2<float> & clean = 
	    state->cleaner[i][j][l].cleaner(k);

	  clean.cleaner().init(state->cleanerParams);
	  clean.iir().set(0.0, 0.9);
	  clean.filterLevel().set(0.5, 0.85);
	}
      }
    }
  }
}

int helma_parse(char newdata, helma_state *state)
{
  float distance[9];
  int index;
  int i, j;

  /* protection against transmission errors */
  if (state->messagecounter > 36) state->messagecounter = 0; 
  if (newdata & 0x80) state->messagecounter = 0;

  state->messagebuffer[state->messagecounter] = newdata;
  state->messagecounter++;

  if (state->messagecounter != 37) return(0); // not a complete record yet

  /* at this point we have a complete, new record */

  index = (state->messagebuffer[0] & 0x20) ? 1 : 0; /* 0 = master, 1 = slave */

  /* Negate the LSB. */
  state->switches[index] = (state->messagebuffer[0] & 0x1f) ^ 1;

  for(i = 0; i < 9; i++) {

    float offset = 12.5f + 21.8f;     /* calibration */
    // float offset = 0.0f;     /* calibration */
    
    float timeA = helma_readtime(i * 2, state) - offset;
    float timeB = helma_readtime(i * 2 + 1, state) - offset;

    timeA *= HELMA_SPEEDOFSOUND / 1000000.0f; // Now it is distance in meters
    timeB *= HELMA_SPEEDOFSOUND / 1000000.0f;

    // printf("B = %f\n", timeB);

    distance[i] = timeB;

    j = i + 9 * index;
    state->distances[j][0] = timeA;
    state->distances[j][1] = timeB;

    int finger = i % 3;
    int emitter = i / 3;

    Fluid::SignalCleaner3<float> & cA = 
      state->cleaner[index][finger][0];
    Fluid::SignalCleaner3<float> & cB = 
      state->cleaner[index][finger][1];

    cA.put(emitter, timeA);
    cB.put(emitter, timeB);

    float value;

    int dA = cA.cleaner(emitter).cleaner().discarded();
    int dB = cB.cleaner(emitter).cleaner().discarded();
    
    if(dB < -10)
      value = cB.get(emitter);
    else if(dA < 0 && dB < 0) {
      if((dB / 3) < dA)
	value = cB.get(emitter);
      else
	value = cA.get(emitter);
    }
    else if(dB > 0 && dA < -3)
      value = cA.get(emitter);
    /*else if(dB < 0 && dA > 0)
      value = cB.get(); */
    /* else if(dA < (dB / 2) && dA < 10)
       value = cA.get(); */
    else
      value = cB.get(emitter);
    
    state->cleaned[index][finger][emitter] = value;

    /*if ((timeA>0.0) && (timeB>0.0) && (fabs(timeA-timeB)<20.0))
      distance[i] = HELMA_SPEEDOFSOUND * timeB/1000000.0;
    else
    distance[i] = 0.0f;*/
  }
  
  for(i = 0; i < 3; i++) {
    state->cleaner[index][i][0].calculate(state->cleanerParams);
    state->cleaner[index][i][1].calculate(state->cleanerParams);
  }

  float bx = HELMA_BX;
  float cx = HELMA_CX;
  float cy = HELMA_CY;

  float bx2 = bx * bx;
  float cx2 = cx * cx;
  float cy2 = cy * cy;
  
  for(i = 0; i < 3; i++) {

    float a, b, c;
    
    if(state->filtering == HELMA_RAW) {
      a = distance[i];
      b = distance[i + 3];
      c = distance[i + 6];
    }
    else if(state->filtering == HELMA_FILTERED1) {
      a = state->cleaner[index][i][1].cleaner(0).get();
      b = state->cleaner[index][i][1].cleaner(1).get();
      c = state->cleaner[index][i][1].cleaner(2).get();
    }
    else if(state->filtering == HELMA_FILTERED2) {
      a = state->cleaner[index][i][1].get(0);
      b = state->cleaner[index][i][1].get(1);
      c = state->cleaner[index][i][1].get(2);
    }

    float a2 = a * a;
    float b2 = b * b;
    float c2 = c * c;
    
    float x = (a2 - b2 + bx2) / (2 * bx);
    float y = (a2 - c2 - 2 * x * cx + cx2 + cy2) / (2 * cy);
    float det = a2 - x * x - y * y;

    if (det > 0.0) { /* and it should be! */
      float z = sqrt(det);
      j = i + 3 * index;
      state->x[j] = x;
      state->y[j] = y;
      state->z[j] = z;
    }
    else if(state->verbose)
      printf("Discarded %d %d - %f (%f %f %f)\n", 
	     index, i, det, a, b, c);
  }
  return (index+1);
}

/* return time records in microseconds */
float helma_readtime(int index, helma_state * state)
{
  float time;
  time =  4 * (state->messagebuffer[index * 2 + 1] * 128 + 
	       state->messagebuffer[index * 2 + 2]);
  time /= 7.3728;
  return(time);
}
