/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#include <napragl.hpp>

#include <Fluid/Error.hpp>
#include <Fluid/SerialPort.hpp>

#include <di_sleep.h>

#include "helma.h"

#include <qapplication.h> 
#include <qlayout.h>
#include <qslider.h>
#include <qtimer.h>

#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <strings.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>


NapraGl::NapraGl(QWidget * parent, const char * dev)
  : QGLWidget(parent), 
    m_finger(-1),
    m_dev(dev),
    m_index(0),
    m_mode(RENDER_XYZ),
    m_run(true)
{
  bzero(m_buffers, sizeof(m_buffers));
  bzero(m_distances, sizeof(m_distances));
  QTimer * t = new QTimer(this);
  connect(t, SIGNAL(timeout()), this, SLOT(updateGL()));
  t->start(40);
  setFocusPolicy(QWidget::StrongFocus);
}

NapraGl::~NapraGl()
{}

void NapraGl::exec( int & argc, char ** argv, const char * dev)
{
  QApplication qa(argc, argv);

  QWidget * window = new QWidget();

  window->resize(600, 600);

  NapraGl * nlg = new NapraGl(window, dev);

  QSlider * slider = new QSlider(-1, 5, 0, -1, QSlider::Horizontal, window);
  connect(slider, SIGNAL(valueChanged(int)), nlg, SLOT(setGraph(int)));

  QSlider * slider2 = new QSlider(0, RENDER_SIZEOF - 1, 
				  0, 0, QSlider::Horizontal, window);
  connect(slider2, SIGNAL(valueChanged(int)), nlg, SLOT(setMode(int)));

  QVBoxLayout * layout = new QVBoxLayout(window);

  layout->addWidget(slider);
  layout->addWidget(slider2);
  layout->addWidget(nlg);

  qa.setMainWidget(window);
  window->show();
  nlg->run();
  qa.exec();
}

void NapraGl::setGraph(int n)
{
  m_finger = n;
  Fluid::Trace("NapraGl::setGraph # %d", n);  
}

void NapraGl::setMode(int n)
{
  static const char * modes[RENDER_SIZEOF] = {
    "XYZ",
    "Time A",
    "Time A Filtered",
    "Time A Filtered 2",
    "Time B",
    "Time B Filtered",
    "Time B Filtered 2",
    "Time cleaned"
  };

  m_mode = (RenderGraph) n;
  Fluid::Trace("NapraGl::setMode # %d - %s", n, modes[n]);
}

void NapraGl::keyPressEvent ( QKeyEvent * e )
{
  int k = e->key();

  if(k == Key_Space) {
    m_run = !m_run;
    Fluid::Trace("NapraGl::keyPressEvent # m_run = %d", (int) m_run);
  }
  else if(k >= Key_1 && k < (Key_1 + RENDER_SIZEOF)) {
    setMode(k - Key_1);
  }
  else if(k >= Key_Escape)
    qApp->quit();
}

void NapraGl::resizeGL (int w, int h)
{
  Fluid::Trace("NapraGl::resizeGL # %d x %d", w, h);
  QGLWidget::resizeGL(w, h);  
  glViewport (0, 0, w, h);
}
    
void NapraGl::paintGL()
{
  // Fluid::Trace("NapraGl::paintGL");

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  
  glMatrixMode (GL_PROJECTION);

  glLoadIdentity ();

  if(m_finger >= 0) {

    float ymax = -1.0 + 2.0 * RENDER_SIZEOF;

    gluOrtho2D(0.0, SAMPLES, -1.0, ymax);
    glMatrixMode (GL_MODELVIEW);
    
    glLoadIdentity ();

    squares(-1.0, ymax, RENDER_SIZEOF, false);
    seconds();

    for(int i=0; i < RENDER_SIZEOF; i++) {

      glPushMatrix();
      if(i)
	glTranslatef(0.0f, i * 2.0f - 0.5f, 0.0f);
      renderGraph(m_finger, (RenderGraph) i);
      glPopMatrix();

      /* gluOrtho2D(0.0, SAMPLES, -1.0, 1.0);
	 glMatrixMode (GL_MODELVIEW);
	 glLoadIdentity ();
	 renderGraph(m_finger, m_mode); 
      */
    }
  }
  else {


    gluOrtho2D(0.0, SAMPLES, -1.0, 11.0);
    glMatrixMode (GL_MODELVIEW);
    
    glLoadIdentity ();

    squares(-1.0, 11.0, 6, true);
    seconds();

    for(int i=0; i < 6; i++) {

      glPushMatrix();
      glTranslatef(0.0f, i * 2.0f, 0.0f);
      renderGraph(i, m_mode);
      glPopMatrix();
    }
  }  

  glColor3f(1, 1, 1);

  glBegin(GL_LINES);

  float x = m_index % SAMPLES;
  glVertex2f(x, -1.0);
  glVertex2f(x, 20.0);

  glEnd();

}

void NapraGl::childLoop()
{
  Fluid::SerialPort serial;
  helma_state teststate;
  char byte;
  
  int rec = open("record.dat", O_WRONLY | O_CREAT | O_TRUNC, 
		 S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);

  int source = -1;
  
  if(!serial.open(m_dev, true, false, 115200, 8, 1, 0)) {
    Fluid::Error("Could not open serial device %s", m_dev);
    serial.close();
  }
  
  if(!serial.isOpen()) {
    source = open(m_dev, O_RDONLY);
    if(source < 0) {
      Fluid::Error("Could not open file %s", m_dev);
      return;
    }
    else
      Fluid::Trace("Opened file, fd = %d", source);
  }

  helma_init(&teststate);

  Fluid::Trace("NapraGl::childLoop # Beginning Naeprae (%d)", rec);

  uint c = 0;
  uint bytes = 0;
  
  while(true) {

    /* bool first = true;
    
    byte = 0;
    while(!(byte & 0x80))
      serial.read(&byte, 1);
    */

    int res = 0;
    
    // Fluid::Trace("Got first byte!!");

    bool collect = true;

    while(!res && collect) {

      while(!m_run && collect)
	DI_Sleep::sleepMs(40);

      /* if(!first) {
	int n = serial.read(&byte, 1);
	
	
	// Fluid::Trace("Bytes = %d", n);
	
	if(n != 1) {
	  Fluid::Error("NapraGl::childLoop # Read failed (%d)", n);
	  return;
	}
      }
      else
      first = false;*/
      /*else
	serial.read(&byte, 1);*/

      /* if(byte & 0x80)
	first = true;

	else */

      if(serial.isOpen())
	serial.read(&byte, 1);
      else
	read(source, &byte, 1);
      
      bytes++;

      if(rec > 0)
	write(rec, &byte, 1);

      res = helma_parse(byte, &teststate);
    }

    if(!res) {
      continue;
    }

    m_index = c;
    
    if(res == 1) {
      c++;
    }

    if(source >= 0 && (c % 3) == 0)
      DI_Sleep::sleepMs(10);

    int hand = res - 1;
    uint id = (res - 1) * 3;
    uint ind = c % SAMPLES;
    
    // Iterate over fingers:
    for(int k = 0; k < 3; k++) {
      
      uint j = id + k;
      
      // Fluid::Trace("%u %d %u %u", bytes, res, j, ind);
      
      Fluid::Vector3 & l =  m_buffers[j][ind];

      l.x = teststate.x[j];
      l.y = teststate.y[j];
      l.z = teststate.z[j];

      Fluid::Vector3 & dA = m_distances[0][j][ind];
      Fluid::Vector3 & dB = m_distances[1][j][ind];
      Fluid::Vector3 & dFA = m_distances[2][j][ind];
      Fluid::Vector3 & dFB = m_distances[3][j][ind];
      Fluid::Vector3 & dFA2 = m_distances[4][j][ind];
      Fluid::Vector3 & dFB2 = m_distances[5][j][ind];
      Fluid::Vector3 & dClean = m_distances[6][j][ind];

      uint base = k + (res - 1) * 9;
      
      // Iterate over axes:
      for(uint l = 0; l < 3; l++) {
	uint ti = base + 3 * l;
	dA[l] = teststate.distances[ti][0];
	dB[l] = teststate.distances[ti][1];

	dFA[l]  = teststate.cleaner[hand][k][0].cleaner(l).get();
	dFB[l]  = teststate.cleaner[hand][k][1].cleaner(l).get();

	dFA2[l] = teststate.cleaner[hand][k][0].get(l);
	dFB2[l] = teststate.cleaner[hand][k][1].get(l);
      }
      dClean = teststate.cleaned[hand][k];
    }
    bytes = 0;
  }  
}


void NapraGl::renderGraph(int sensor, RenderGraph gr)
{
  static const Fluid::Vector3 colors[3] = {
    Fluid::Vector3(1, 0, 0),
    Fluid::Vector3(0, 1, 0),
    Fluid::Vector3(0.5, 0.5, 1)
  };

  glLineWidth(2);

  Fluid::Vector3 * data = 0;

  if(gr == RENDER_XYZ)
    data = m_buffers[sensor];
  else if(gr == RENDER_DIS_A)
    data = m_distances[0][sensor];
  else if(gr == RENDER_DIS_B)
    data = m_distances[1][sensor];
  else if(gr == RENDER_DIS_FA)
    data = m_distances[2][sensor];
  else if(gr == RENDER_DIS_FB)
    data = m_distances[3][sensor];
  else if(gr == RENDER_DIS_FA2)
    data = m_distances[4][sensor];
  else if(gr == RENDER_DIS_FB2)
    data = m_distances[5][sensor];
  else if(gr == RENDER_DIS_CLEAN)
    data = m_distances[6][sensor];

  for(int a = 0; a < 3; a++) {

    glColor3fv(colors[a].data());

    glDisable(GL_LIGHTING);

    glBegin(GL_LINE_STRIP);
    
    for(int i = 0; i < SAMPLES; i++)
      glVertex2f(i, data[i][a] * 100.0f / 15.0f);

    glEnd();    
  }
}

void NapraGl::squares(float ymin, float ymax, int n, bool isAll)
{
  glShadeModel(GL_FLAT);

  glBegin(GL_QUADS);

  float color[2] = { 0.0f, 0.2f };

  float step = (ymax - ymin) / n;

  float blueBoost = isAll ? 0.2f : 0.0f;

  for(int i=0; i < n; i++) {
    float c = color[i % 2];
    glColor3f(c, c, c + blueBoost);

    float low = ymin + i * step;
    float high = low + step;

    glVertex2f(0.0, low);
    glVertex2f(SAMPLES, low);
    glVertex2f(SAMPLES, high);
    glVertex2f(0.0, high);
  }
  
  glEnd();
}


void NapraGl::seconds()
{
  // Second marks:

  glEnable(GL_BLEND);

  glBegin(GL_LINES);

  glColor4f(0.4, 0.4, 0.4, 0.7);

  for(int i = 0; i < SAMPLES; i += FPS) {
    glVertex2f(i, -2.0f);
    glVertex2f(i, 15.0f);
  }

  glEnd();
}
