/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#include "KeyboardTrackerWidget.hpp"

#include <Fluid/Error.hpp>
#include <Fluid/Keyboard.hpp>

#include "QTKeyboard.hpp"

#include <qcursor.h>
#include <qtimer.h>
#include <qpainter.h>


void QtKeyToFluidKey(int qkey, int qstate, Fluid::Key & key, int & modifiers)
{
  static const int keymap[] = {
    Qt::Key_A, Fluid::KEY_A,
    Qt::Key_B, Fluid::KEY_B,
    Qt::Key_C, Fluid::KEY_C,
    Qt::Key_D, Fluid::KEY_D,
    Qt::Key_E, Fluid::KEY_E,
    Qt::Key_F, Fluid::KEY_F,
    Qt::Key_G, Fluid::KEY_G,
    Qt::Key_H, Fluid::KEY_H,
    Qt::Key_I, Fluid::KEY_I,
    Qt::Key_J, Fluid::KEY_J,
    Qt::Key_K, Fluid::KEY_K,
    Qt::Key_L, Fluid::KEY_L,
    Qt::Key_M, Fluid::KEY_M,
    Qt::Key_N, Fluid::KEY_N,
    Qt::Key_O, Fluid::KEY_O,
    Qt::Key_P, Fluid::KEY_P,
    Qt::Key_Q, Fluid::KEY_Q,
    Qt::Key_R, Fluid::KEY_R,
    Qt::Key_S, Fluid::KEY_S,
    Qt::Key_T, Fluid::KEY_T,
    Qt::Key_U, Fluid::KEY_U,
    Qt::Key_V, Fluid::KEY_V,
    Qt::Key_W, Fluid::KEY_W,
    Qt::Key_X, Fluid::KEY_X,
    Qt::Key_Y, Fluid::KEY_Y,
    Qt::Key_Z, Fluid::KEY_Z,

    Qt::Key_0, Fluid::KEY_0,
    Qt::Key_1, Fluid::KEY_1,
    Qt::Key_2, Fluid::KEY_2,
    Qt::Key_3, Fluid::KEY_3,
    Qt::Key_4, Fluid::KEY_4,
    Qt::Key_5, Fluid::KEY_5,
    Qt::Key_6, Fluid::KEY_6,
    Qt::Key_7, Fluid::KEY_7,
    Qt::Key_8, Fluid::KEY_8,
    Qt::Key_9, Fluid::KEY_9,

    
    Qt::Key_Space, Fluid::KEY_SPACE,
    Qt::Key_Tab,   Fluid::KEY_TAB,
    Qt::Key_Enter, Fluid::KEY_ENTER,

    Qt::Key_Left, Fluid::KEY_LEFT,
    Qt::Key_Right, Fluid::KEY_RIGHT,
    Qt::Key_Up, Fluid::KEY_UP,
    Qt::Key_Down, Fluid::KEY_DOWN,
    Qt::Key_Home, Fluid::KEY_HOME,
    Qt::Key_End, Fluid::KEY_END,
    Qt::Key_PageUp, Fluid::KEY_PAGE_UP,
    Qt::Key_PageDown, Fluid::KEY_PAGE_DOWN,
    Qt::Key_Insert, Fluid::KEY_INSERT,
    Qt::Key_Delete, Fluid::KEY_DELETE,
    Qt::Key_BackSpace, Fluid::KEY_BACKSPACE,
    Qt::Key_Escape,   Fluid::KEY_ESCAPE, 

    Qt::Key_ParenLeft,  Fluid::KEY_LEFT_PAREN,
    Qt::Key_ParenRight, Fluid::KEY_RIGHT_PAREN,
    Qt::Key_BraceLeft,  Fluid::KEY_LEFT_BRACE,
    Qt::Key_BraceRight, Fluid::KEY_RIGHT_BRACE,
    Qt::Key_BracketLeft,  Fluid::KEY_LEFT_BRACKET,
    Qt::Key_BracketRight, Fluid::KEY_RIGHT_BRACKET,

    Qt::Key_Less,    Fluid::KEY_LESS,
    Qt::Key_Equal,   Fluid::KEY_EQUAL,
    Qt::Key_Greater, Fluid::KEY_GREATER,
    Qt::Key_Plus,    Fluid::KEY_PLUS,
    Qt::Key_Minus,   Fluid::KEY_MINUS,

    Qt::Key_Asterisk, Fluid::KEY_ASTERISK,
    Qt::Key_Slash, Fluid::KEY_SLASH,
    Qt::Key_Backslash, Fluid::KEY_BACKSLASH,
    Qt::Key_Question, Fluid::KEY_QUESTION,
    Qt::Key_Exclam, Fluid::KEY_EXCLAM,
    Qt::Key_Comma, Fluid::KEY_COMMA,

    Qt::Key_Period, Fluid::KEY_PERIOD,
    Qt::Key_Colon, Fluid::KEY_COLON,
    Qt::Key_Semicolon, Fluid::KEY_SEMICOLON,
    Qt::Key_Underscore, Fluid::KEY_UNDERSCORE,
    Qt::Key_QuoteDbl, Fluid::KEY_QUOTE,
    Qt::Key_Dollar, Fluid::KEY_DOLLAR,
    Qt::Key_At, Fluid::KEY_AT,
    Qt::Key_AsciiTilde, Fluid::KEY_TILDE,
    Qt::Key_Percent, Fluid::KEY_PERCENT,
    Qt::Key_Ampersand, Fluid::KEY_AMPERSAND,
    Qt::Key_Apostrophe, Fluid::KEY_APOSTROPHE,
    Qt::Key_NumberSign, Fluid::KEY_NUMBERSIGN,
    Qt::Key_AsciiCircum, Fluid::KEY_CIRCUM,
    Qt::Key_Bar, Fluid::KEY_BAR,

    -1, -1
  };

  modifiers = 0;

  if(qstate & Qt::ShiftButton)
    modifiers |= MOD_SHIFT;
  if(qstate & Qt::ControlButton)
    modifiers |= MOD_CONTROL;
  if(qstate & Qt::AltButton)
    modifiers |= MOD_ALT;

  key = KEY_NONE;

  printf("qkey = %x   state = %x\n", qkey, qstate); fflush(0);

  for(int i=0; keymap[i] >= 0; i += 2) {
    if(keymap[i] == qkey) {
      key = (Fluid::Key) keymap[i + 1];
      return;
    }
  }
}


KeyboardTrackerWidget::KeyboardTrackerWidget
(QTKeyboardSource *driver, int sampleRate, bool grab)
  :QWidget(0,"MMTK QT Keyboard Driver"),
   sampleRate(sampleRate), 
   grab(grab),
   driver(driver)
{
  internalTimer = new QTimer( this ); // create internal timer
  connect( internalTimer, SIGNAL(timeout()), SLOT(timeout()) );
  grabbed = false;
  internalTimer->start( 1000/sampleRate ); // 50hz
}

KeyboardTrackerWidget::~KeyboardTrackerWidget()
{
  if(grabbed) releaseKeyboard();
  delete internalTimer;
}

/** Toggle grabbing behaviour. Space toggles grabbing, Esc disables
    grabbing. */

void KeyboardTrackerWidget::keyPressEvent ( QKeyEvent * e)
{
  Trace("KeyboardTrackerWidget::keyPressEvent # %x", e->key());

  if(e->key() == Key_Space) {
    if(!grab) {
      // grabMouse();
      grabKeyboard();
      grab = true;
    }
    else {
      // releaseMouse();
      releaseKeyboard();
      grab = false;
    }
  }
  else if(e->key() == Key_Escape) {
    // releaseMouse();
    releaseKeyboard();
    grab = false;
  }
  else {
    
    TimeStamp t;
    t.getTimeOfDay();

    KeyboardSample s;
    s.time = t;

    QtKeyToFluidKey(e->key(), e->state(), s.key, s.modifiers);
    
    // for(uint i=0; i < tmp.length(); i++) {
    // s.character = tmp[i];
    m_keys.push_back(s);
    // Trace("KeyboardTrackerWidget::keyPressEvent # added '%c'", s.character);
    // }
    
    while(m_keys.size() > 200)
      popKey();

    if(driver)
      driver->notify();
  }
}
void KeyboardTrackerWidget::timeout()
{
  if(grab && !grabbed) 
    {
      // grabMouse();
      grabKeyboard();
      grabbed = true;
    }
  
}


KeyboardSample KeyboardTrackerWidget::popKey()
{
  if(m_keys.size() == 0)
    return KeyboardSample();

  KeyboardSample s(m_keys[0]);

  m_keys.erase(m_keys.begin());

  return s;
}

