/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#include "MouseTrackerWidget.hpp"

#include <Fluid/Error.hpp>

#include "QTMouse.hpp"

#include <qcursor.h>
#include <qtimer.h>
#include <qpainter.h>

MouseTrackerWidget::MouseTrackerWidget
(QTMouseSource *driver, int sampleRate, bool grab)
  :QWidget(0,"MMTK QT Mouse Driver"),
   sampleRate(sampleRate), 
   grab(grab),
   driver(driver)
{
  buttonState = 0;

  internalTimer = new QTimer( this ); // create internal timer
  connect( internalTimer, SIGNAL(timeout()), SLOT(timeout()) );
  grabbed = false;
  internalTimer->start( 1000/sampleRate ); // 50hz
}

MouseTrackerWidget::~MouseTrackerWidget()
{
  if(grabbed) releaseMouse();
  delete internalTimer;
}

void MouseTrackerWidget::paintEvent(QPaintEvent* )
{
  QPainter p(this);
  QString sx,sy,s;
  sx.setNum(x);
  sy.setNum(y);
  s = QString("POSITION:  ") + sx + ',' + sy;
  p.drawText(0,20,"Driver version 0.1");
  p.drawText(0,34,s);

  s = QString("BUTTONS:   ");
  if(buttonState & 0x1)
    s+="left ";
  if(buttonState & 0x2)
    s+="middle ";
  if(buttonState & 0x4)
    s+="right ";
  
  p.drawText(0,48,s);
}

void MouseTrackerWidget::mousePressEvent(QMouseEvent *e)
{
  // Conversion from Qt to fluid order.
  uint tmp = e->stateAfter();
  uint rbutton = (tmp & 0x2) << 1; // Right
  uint mbutton = (tmp & 0x4) >> 1; // Middle
  
  buttonState = (tmp & 0xFFF9) | rbutton | mbutton;
}

void MouseTrackerWidget::mouseReleaseEvent(QMouseEvent *e)
{
  uint tmp = e->stateAfter();
  uint rbutton = (tmp & 0x2) << 1; // Right
  uint mbutton = (tmp & 0x4) >> 1; // Middle
  
  buttonState = (tmp & 0xFFF9) | rbutton | mbutton;

}

/** Toggle grabbing behaviour. Space toggles grabbing, Esc disables
    grabbing. */

void MouseTrackerWidget::keyPressEvent ( QKeyEvent * e)
{
  Trace("MouseTrackerWidget::keyPressEvent # %x", e->key());

  if(e->key() == Key_Space) {
    if(!grab) {
      grabMouse();
      grabKeyboard();
      grab = true;
    }
    else {
      releaseMouse();
      releaseKeyboard();
      grab = false;
      setCursor(QCursor(ArrowCursor));
    }
  }
  else if(e->key() == Key_Escape) {
    releaseMouse();
    releaseKeyboard();
    setCursor(QCursor(ArrowCursor));
    grab = false;
  }
}

int MouseTrackerWidget::getCurrentX()
{
  return x;
}

int MouseTrackerWidget::getCurrentY()
{
  return y;
}

int MouseTrackerWidget::getButtonState()
{
  return buttonState;
  /* switch(b)
    {
    case BUTTON_LEFT:
      return leftdown;
    case BUTTON_RIGHT:
      return rightdown;
    case BUTTON_MIDDLE:
      return middledown;
    }
    return false; */
}

void MouseTrackerWidget::timeout()
{
  
   x = QCursor::pos().x();
   y = QCursor::pos().y();
   driver->notify();
   static int counter = 0;
   if((counter++ % 5) == 0)
     {

       if(grab) 
	 {
	   grabMouse();
	   grabKeyboard();
	   grabbed = true;
	   setCursor(QCursor(BlankCursor));
	 }

       update();
     }
}

void MouseTrackerWidget::setSampleRate(int hz)
{
  internalTimer->changeInterval(1000/hz);
  sampleRate = hz;
}

