/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#include <SpeechViaVoice.hpp>

#include <Fluid/Config.hpp>
#include <Fluid/ConfigReader.hpp>
#include <Fluid/DeviceFactory.hpp>
#include <Fluid/Error.hpp>

using namespace Fluid;

SpeechViaVoice::SpeechViaVoice(const std::string & grammar, 
			       const std::string & lang, 
			       int bufferSize)
  : DataSource<Phrase>(bufferSize),
    m_grammarName(grammar),
    m_lang(lang),
    m_continue(false)
{
  const char * fname = "SpeechViaVoice::SpeechViaVoice";
  Trace("%s # grammar = \"%s\" lang = \"%s\"", fname,
	m_grammarName.c_str(), m_lang.c_str());
}

SpeechViaVoice::~SpeechViaVoice()
{}

void SpeechViaVoice::start()
{
  const char * fname = "SpeechViaVoice::start";

  Trace("%s # grammar = \"%s\" lang = \"%s\"", fname,
	m_grammarName.c_str(), m_lang.c_str());

  bool ok = open("FluidSpeechViaVoice", 0);
  
  if(!ok) {
    return;
  }

  std::string trGrammar = m_grammarName + "_tr.fsg";
  std::string langGrammar = m_grammarName + "_" + m_lang + ".fsg";
  
  Trace("%s # Preparing to load files \"%s\" and \"%s\"",
	fname, trGrammar.c_str(), langGrammar.c_str());

  char  currDir[512];
  
  if (getcwd(currDir, sizeof(currDir)-1) == NULL) {
    Error("%s # Could not get the current working directory", fname);
    return;
  }
  
  std::string pwd(currDir);

  if(trGrammar[0] != '/')
    trGrammar = pwd + "/" + trGrammar;
  
  if(langGrammar[0] != '/')
    langGrammar = pwd + "/" + langGrammar;

  
  if (!defineGrammar(m_grammarName.c_str(),langGrammar.c_str())) {
    Error("%s # Error defining grammar \"%s\"", fname, langGrammar.c_str());
    return;
  }

  if (!enableVocab(m_grammarName.c_str())) {
    Error("%s # Error enabling vocalulary \"%s\"", 
	  fname, m_grammarName.c_str());
    return ;
  }
    
  if (!loadTranslationGrammar(trGrammar.c_str())) {
    Error("%s # Error loading translation grammar \"%s\"", 
	  fname, trGrammar.c_str());
    return ;
  }

  requestFocus();
  micOn();
  recognizeNextWord();

  m_continue = true;

  sleep(1);

  Trace("%s # Spawning thread", fname);

  run();
}

void SpeechViaVoice::stop()
{
  m_continue = false;

  waitEnd();
  
  micOff();
  releaseFocus();
  close();
}

void SpeechViaVoice::childLoop()
{
  // const char * fname = "SpeechViaVoice::childLoop";

  while(m_continue) {
    dispatch();
  }
}

SpeechViaVoice * SpeechViaVoice::create(Fluid::Chunk * chunk)
{
  std::string grammarName;
  std::string lang;
  int bufferSize;
  
  Variant v;
  
  v = chunk->get("grammar");
  if( v.getType() != Variant::INVALID)
    grammarName = std::string(v);
  else
    grammarName = "cave_grammar";
  
  v = chunk->get("lang");
  if( v.getType() != Variant::INVALID)
    lang = std::string(v);
  else
    lang = "en";
    
  v = chunk->get("bufferSize");
  if( v.getType() != Variant::INVALID )
    bufferSize = v;
  else 
    bufferSize = 30;    

  SpeechViaVoice *source = 
    new SpeechViaVoice(grammarName, lang, bufferSize);

  return source;

}

void SpeechViaVoice::deliver(const char *str)
{
  Phrase s;
  s.time.getTimeOfDay();
  s.phrase = str;

  write(s);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

class SpeechViaVoiceFactory : public DeviceFactory
{
public:
  virtual std::string      getDriverName() const { return "SpeechViaVoice"; }
  
  virtual InputDeviceBase* create(Chunk* chunk)
  {
    return new Speech(chunk,SpeechViaVoice::create(chunk));
  }
};

SpeechViaVoiceFactory factory;

extern "C" DeviceFactory* getFactory()
{
  return &factory;
}
