/* -*- C++ -*- */

//
// DIVA Library. Copyright (C) Jarmo Hiipakka, 1999-2000.
//
// $Log: di_TTS_base.C,v $
// Revision 1.1  2002/10/16 08:45:54  tilmonen
// Wonderful Speech recognition system by
//  - IBM: ViaVoice.
//  - Jarmo Hiipakka: C++ wrapping
//  - Tommi Ilmonen: Fluid wrapping
//

#include "di_error.h"
#include "di_TTS_base.h"

// -- Constructors etc. -------------------------------------------------------

DI_TTSBase::DI_TTSBase() :
  m_audioBuffer(NULL),
  m_phonemeBuffer(NULL),
  m_handle(eciNew())
{
  if (m_handle == NULL_ECI_HAND)
    DI_fatal(DI_ERR_INTERNAL, "ViaVoiceTTS instance not succesfully created");
  eciRegisterCallback(m_handle, m_callback, this);
}

DI_TTSBase::~DI_TTSBase()
{
  eciDelete(m_handle);
}

bool DI_TTSBase::reset()
{
  bool res((bool)eciReset(m_handle));
  eciRegisterCallback(m_handle, m_callback, this);
  m_audioBuffer = NULL;
  m_phonemeBuffer = NULL;
  
  return res;
}

bool DI_TTSBase::speakText(const char *text, bool annotations)
{
  return (bool)eciSpeakText((ECIInputText)text, (ECIBoolean)annotations);
}

// -- Output control ----------------------------------------------------------

bool DI_TTSBase::setOutputBuffer(int size, short *buffer)
{
  bool res = (bool)eciSetOutputBuffer(m_handle, size, buffer);

  if (!res)
  {
    m_error("eciSetOutputBuffer failed");
    return false;
  }
  
  if ((size == 0) || (buffer == NULL))
    m_audioBuffer = NULL;
  else
    m_audioBuffer = buffer;

  return true;
}

bool DI_TTSBase::setOutputDevice(int deviceNum)
{
  if (!eciSetOutputDevice(m_handle, deviceNum))
  {
    m_error("eciSetOutputDevice failed");
    return false;
  }

  return true;
}

bool DI_TTSBase::setOutputFilename(const char *filename)
{
  if (!eciSetOutputFilename(m_handle, filename))
  {
    m_error("eciSetOutputFilename failed");
    return false;
  }

  return true;
}

// -- Dictionary maintenance --------------------------------------------------

bool DI_TTSBase::deleteDict(ECIDictHand dictionary)
{
  return (eciDeleteDict(m_handle, dictionary) == NULL_DICT_HAND);
}

bool DI_TTSBase::dictFindFirst(ECIDictHand dictHandle,
			       ECIDictVolume whichDictionary,
			       const char **key, const char **translation)
{
  if (eciDictFindFirst(m_handle,dictHandle,whichDictionary,key,translation)
      == DictNoError)
  {
    return true;
  }
  else
  {
    DI_error(DI_ERR_UNKNOWN,"DI_TTSBase::dictFindFirst",
	     "eciDictFindFirst failed");
    return false;
  }
}

bool DI_TTSBase::dictFindNext(ECIDictHand dictHandle,
			      ECIDictVolume whichDictionary,
			      const char **key, const char **translation)
{
  if (eciDictFindNext(m_handle,dictHandle,whichDictionary,key,translation)
      == DictNoError)
  {
    return true;
  }
  else
  {
    DI_error(DI_ERR_UNKNOWN,"DI_TTSBase::dictFindNext",
	     "eciDictFindNext failed");
    return false;
  }
}

const char *DI_TTSBase::dictLookup(ECIDictHand dictHandle,
				   ECIDictVolume whichDictionary,
				   const char *key)
{
  return eciDictLookup(m_handle, dictHandle, whichDictionary, key);
}

ECIDictHand DI_TTSBase::getDict()
{
  return eciGetDict(m_handle);
}

bool DI_TTSBase::loadDict(ECIDictHand dictHandle,
			  ECIDictVolume whichDictionary,
			  const char *filename)
{
  ECIDictError dictError = eciLoadDict(m_handle, dictHandle, whichDictionary,
				       filename);
  
  if (dictError == DictFileNotFound)
  {
    DI_error(DI_ERR_ARGUMENT,"DI_TTSBase::loadDict","File not found");
    return false;
  }
  else if (dictError != DictNoError)
  {
    DI_error(DI_ERR_ARGUMENT,"DI_TTSBase::loadDict","eciLoadDict failed");
    return false;
  }
  return true;
}

ECIDictHand DI_TTSBase::newDict()
{
  return eciNewDict(m_handle);
}

bool DI_TTSBase::saveDict(ECIDictHand dictHandle,
			  ECIDictVolume whichDictionary,
			  const char *filename)
{
  ECIDictError dictError = eciSaveDict(m_handle, dictHandle, whichDictionary,
				       filename);
  
  if (dictError == DictAccessError)
  {
    DI_error(DI_ERR_SYSTEM,"DI_TTSBase::saveDict","Access error");
    return false;
  }
  else if (dictError != DictNoError)
  {
    DI_error(DI_ERR_ARGUMENT,"DI_TTSBase::saveDict","eciSaveDict failed");
    return false;
  }
  return true;
}

bool DI_TTSBase::setDict(ECIDictHand dictHandle)
{
  return (eciSetDict(m_handle,dictHandle) == DictNoError);
}

bool DI_TTSBase::updateDict(ECIDictHand dictHandle,
			    ECIDictVolume whichDictionary,
			    const char *key, const char *translation)
{
  return (eciUpdateDict(m_handle,dictHandle,whichDictionary,key,translation)
	  == DictNoError);
}

// -- Synthesis control methods -----------------------------------------------

bool DI_TTSBase::addText(const char *text)
{
  if (!eciAddText(m_handle, (ECIInputText)text))
  {
    m_error("eciAddText failed");
    return false;
  }
  return true;
}

bool DI_TTSBase::clearInput()
{
  if (!eciClearInput(m_handle))
  {
    m_error("eciClearInput failed");
    return false;
  }
  return true;
}

bool DI_TTSBase::generatePhonemes(int size, char *buffer)
{
  if (!eciGeneratePhonemes(m_handle, size, buffer))
  {
    m_error("eciGeneratePhonemes failed");
    return false;
  }

  if ((size == 0) || (buffer == NULL))
    m_phonemeBuffer = NULL;
  else
    m_phonemeBuffer = buffer;

  return true;
}

int DI_TTSBase::getIndex()
{
  return (bool)eciGetIndex(m_handle);
}

bool DI_TTSBase::insertIndex(int indexNum)
{
  if (!eciInsertIndex(m_handle, indexNum))
  {
    m_error("eciInsertIndex failed");
    return false;
  }
  return true;
}

bool DI_TTSBase::pause(bool pauseOn)
{
  if (!eciPause(m_handle, pauseOn))
  {
    m_error("eciPause failed");
    return false;
  }
  return true;
}

bool DI_TTSBase::speaking()
{
  return (bool)eciSpeaking(m_handle);
}

bool DI_TTSBase::stop()
{
  if (!eciStop(m_handle))
  {
    m_error("eciStop failed");
    return false;
  }
  return true;
}

bool DI_TTSBase::synchronize()
{
  if (!eciSynchronize(m_handle))
  {
    m_error("eciSynchronize failed");
    return false;
  }
  return true;
}

bool DI_TTSBase::synthesize()
{
  if (!eciSynthesize(m_handle))
  {
    m_error("eciSynthesize failed");
    return false;
  }
  return true;
}

bool DI_TTSBase::synthesizeFile(const char *filename)
{
  if (!eciSynthesizeFile(m_handle, filename))
  {
    m_error("eciSynthesizeFile failed");
    return false;
  }
  return true;
}

// -- Private methods ---------------------------------------------------------

ECICallbackReturn DI_TTSBase::m_callback(ECIHand handle, ECIMessage message,
					 long lparam, void *data)
{
  (void)handle;
  DI_TTSBase *thisPtr = (DI_TTSBase *)data;

  switch (message)
  {
   case eciWaveformBuffer:
    return thisPtr->waveformBufferMsg( lparam );
   case eciPhonemeBuffer:
    return thisPtr->phonemeBufferMsg( lparam );
   case eciIndexReply:
    return thisPtr->indexReplyMsg( lparam );
   case eciPhonemeIndexReply:
    return thisPtr->phonemeIndexReplyMsg( (ECIMouthData *)lparam );
   default:
    DI_error(DI_ERR_INTERNAL, "Unanticipated message type");
    return eciDataProcessed;
  }
}
