/* -*- C++ -*- */

//
// DIVA Library. Copyright (C) Jarmo Hiipakka, 1999-2000.
//
// $Log: di_TTS_base.h,v $
// Revision 1.1  2002/10/16 08:45:55  tilmonen
// Wonderful Speech recognition system by
//  - IBM: ViaVoice.
//  - Jarmo Hiipakka: C++ wrapping
//  - Tommi Ilmonen: Fluid wrapping
//

#ifndef DI_TTS_BASE_H
#define DI_TTS_BASE_H

#include <eci.h>

const int DI_TTS_MAX_NAME_LENGTH(ECI_VOICE_NAME_LENGTH+1);

enum DI_TTSLang {
  DI_TTS_LANG_AMERICAN_ENGLISH = eciGeneralAmericanEnglish,
  DI_TTS_LANG_BRITISH_ENGLISH = eciBritishEnglish,
  DI_TTS_LANG_CASTILIAN_SPANISH = eciCastilianSpanish,
  DI_TTS_LANG_MEXICAN_SPANISH = eciMexicanSpanish,
  DI_TTS_LANG_STANDARD_FRENCH = eciStandardFrench,
  DI_TTS_LANG_CANADIAN_FRENCH = eciCanadianFrench,
  DI_TTS_LANG_STANDARD_GERMAN = eciStandardGerman,
  DI_TTS_LANG_STANDARD_ITALIAN = eciStandardItalian,
  DI_TTS_LANG_SIMPLIFIED_CHINESE = eciSimplifiedChinese,
  DI_TTS_LANG_BRAZILIAN_PORTUGUESE = eciBrazilianPortuguese,
};

/** A base class for text-to-speech synthesis classes.

    This is a base class for text-to-speech synthesis classes. Currently
    it is a wrapper class for IBM:s ViaVoiceTTS system. To build operational
    TTS instances you have to overwrite the virtual functions of this class.
*/

class DI_TTSBase
{
public:
  
  /**name Constructors etc. */

  //@{
  
  ///
  
  DI_TTSBase();

  ///
  
  virtual ~DI_TTSBase();

  /// Reset the object

  bool reset();
  
  /** Synthesize text to the default audio device.

      This method is used for simple text-to-speech purposes. Creates
      a new TTS instance, synthesizes the given NULL-terminated C-string
      'text' to the default audio device, giving no way (but the
      annotations) to control synthesis, and destroys the TTS instance
      it created. If 'annotations' is 'true', annotations within 'text'
      are processed, and the speech output is altered accordingly.

      This function returns only after all speech has been synthesized
      and played. It returns 'true' on success, 'false' otherwise.
      
  */

  static bool speakText(const char *text, bool annotations);
  
  //@}

  /**@name Synthesis state parameters */

  //@{

  /// Return true if abbreviations dictionaries are used.

  bool getAbbrDictionary() const
  { return (eciGetParam(m_handle,eciDictionary) == 0); }

  /// Set the usage of the abbreviations dictionaries.

  bool setAbbrDictionary(bool useDict)
  { return ( eciSetParam(m_handle, eciDictionary, useDict ? 0 : 1) != -1 ); }

  /** Return input type.
      0: plain text input (default)
      1: annotated text input */
  
  int getInputType() const { return eciGetParam(m_handle,eciInputType); }

  /// Set input type.

  bool setInputType(int type)
  { return ( eciSetParam(m_handle, eciInputType, type) != -1 ); }
  
  /// Return the language in use.

  DI_TTSLang getLanguageDialect() const
  { return (DI_TTSLang)eciGetParam(m_handle,eciLanguageDialect); }

  /// Set language.

  bool setLanguageDialect(DI_TTSLang lang)
  { return ( eciSetParam(m_handle, eciLanguageDialect, (int)lang) != -1 ); }

  /** Return the number mode for 4-digit numbers.
      0: "nonyears"
      1: "years" (default) */

  int getNumberMode() const { return eciGetParam(m_handle,eciNumberMode); }

  /// Set number mode.

  bool setNumberMode(int mode)
  { return ( eciSetParam(m_handle, eciNumberMode, mode) != -1 ); }

  /// True if we use real world units as opposed to ECI values (default).

  bool getRealWorldUnits() const 
  { return (eciGetParam(m_handle,eciRealWorldUnits) == 1); }

  /// Set the usage of the real world units.

  bool setRealWorldUnits(bool rwu)
  { return ( eciSetParam(m_handle, eciRealWorldUnits, rwu ? 1 : 0) != -1 ); }

  /// Return the current sample rate (8000 Hz, 11025 Hz, 22050 Hz).
  
  int getSampleRate() const { return eciGetParam(m_handle,eciSampleRate); }

  /// Set the sampling rate.

  bool setSampleRate(int rate)
  { return ( eciSetParam(m_handle, eciSampleRate, rate) != -1 ); }

  /** Get the synthesis mode.

      0: Sentence: The input buffer is synthesized and cleared at the end
      of sentence (default).

      1: Manual: Synthesis and input clearing is controlled by commands only
  */
  
  int getSynthMode() const ;

  /// Set the synthesis mode.

  bool setSynthMode(int mode);
  
  /** Return the current text mode.

      0: default; no special interpretation (default).
      1: AlphaSpell; letters and digits are spelled out, punctuation normal.
      2: AllSpell; all symbols are spelled out.
      3: IRCSpell; like AlphaSpell but using International Radio Code.
  */

  int getTextMode() const { return eciGetParam(m_handle,eciTextMode); }

  /// Set text mode.

  bool setTextMode(int mode)
  { return ( eciSetParam(m_handle, eciTextMode, mode) != -1 ); }

  /// Return true, if phoneme indices are sent.

  bool getWantPhonemeIndices() const
  { return (eciGetParam(m_handle,eciWantPhonemeIndices) == 1); }

  /// Set the usage of the phoneme indices.
  
  bool setWantPhonemeIndices(bool want)
  { return ( eciSetParam(m_handle,eciWantPhonemeIndices, want?1:0) != -1 ); }
  
  //@}

  /**@name Voice parameter control

     A "voice" is a set of voice control parameters. There are sixteen
     voices that can store the voice parameters. Five of these voices are
     preset voices and three more are reserved for future use.

     The preset voices are:

     1. An adult male

     2. An adult female

     3. A child

     4. - 6. Reserved for future use

     7. An elderly female

     8. An elderly male.

     User defined voices (9.-16.) are undefined by default. You can create
     a new voice, e.g., by copying any of the predefined voices to the range
     9-16 using the copyVoice() method, and the modifying the parameters.

     Voice 0 is the active voice, i.e., you can take a voice into use
     by copying it to the voice number 0.
     
  */

  //@{

  /** Makes a copy of a set of voice parameters */

  bool copyVoice(int voiceFrom, int voiceTo)
  { return (bool)eciCopyVoice(m_handle,voiceFrom,voiceTo); }

  /// Get the current breathiness setting. Range: 0-100.

  int getBreathiness(int voiceNum) const
  { return eciGetVoiceParam(m_handle, voiceNum, eciBreathiness); }

  /// Set breathiness.

  bool setBreathiness(int voiceNum, int value)
  { return (eciSetVoiceParam(m_handle,voiceNum,eciBreathiness,value) != -1); }
  
  /// Get the current gender setting. 0: male, 1: female.

  int getGender(int voiceNum) const
  { return eciGetVoiceParam(m_handle, voiceNum, eciGender); }
  
  /// Set gender.

  bool setGender(int voiceNum, int value)
  { return (eciSetVoiceParam(m_handle,voiceNum,eciGender,value) != -1); }
  
  /// Get the current head size setting. Range: 0-100.

  int getHeadSize(int voiceNum) const
  { return eciGetVoiceParam(m_handle, voiceNum, eciHeadSize); }
  
  /// Set head size.

  bool setHeadSize(int voiceNum, int value)
  { return (eciSetVoiceParam(m_handle,voiceNum,eciHeadSize,value) != -1); }
  
  /** Get the current pitch baseline setting.
      Range: 0-100 (ECI units); 40-422 (real world units = Hz). */

  int getPitchBaseline(int voiceNum) const
  { return eciGetVoiceParam(m_handle, voiceNum, eciPitchBaseline); }
  
  /// Set pitch baseline.

  bool setPitchBaseline(int voiceNum, int value)
  { return (eciSetVoiceParam(m_handle,voiceNum,eciPitchBaseline,value) != -1);}
  
  /// Get the current pitch fluctuation setting. Range: 0-100.

  int getPitchFluctuation(int voiceNum) const
  { return eciGetVoiceParam(m_handle, voiceNum, eciPitchFluctuation); }
  
  /// Set pitch fluctuation.

  bool setPitchFluctuation(int voiceNum, int value)
  { return ( eciSetVoiceParam(m_handle,voiceNum,eciPitchFluctuation,value)
	     != -1 ) ; }
  
  /// Get the current roughness setting. Range: 0-100.

  int getRoughness(int voiceNum) const
  { return eciGetVoiceParam(m_handle, voiceNum, eciRoughness); }
  
  /// Set roughness.

  bool setRoughness(int voiceNum, int value)
  { return (eciSetVoiceParam(m_handle,voiceNum,eciRoughness,value) != -1); }
  
  /** Get the current speed setting.
      Range: 0-250 (ECI units); 70-1297 (rwu = words per minute). */

  int getSpeed(int voiceNum) const
  { return eciGetVoiceParam(m_handle, voiceNum, eciSpeed); }
  
  /// Set speed.

  bool setSpeed(int voiceNum, int value)
  { return (eciSetVoiceParam(m_handle,voiceNum,eciSpeed,value) != -1); }
  
  /** Get the current volume setting.
      Range: 0-100 (ECI units); 1-65535 (real world units). */

  int getVolume(int voiceNum) const
  { return eciGetVoiceParam(m_handle, voiceNum, eciVolume); }
  
  /// Set volume.

  bool setVolume(int voiceNum, int value)
  { return (eciSetVoiceParam(m_handle,voiceNum,eciVolume,value) != -1); }

  /** Get the name of a voice.

      Using this method it is possible to get the name of any of the voice
      parameter sets. 'nameBuf' is a pointer to a pre-allocated buffer,
      where the resulting NULL-terminated C-string can be copied. The buffer
      must be at least DI_TTS_MAX_NAME_LENGTH bytes long.
  */

  bool getVoiceName(int voiceNum, char *nameBuf) const
  { return (bool)eciGetVoiceName(m_handle, voiceNum, nameBuf); }
  
  //@}

  /**@name Output control */

  //@{

  /** Enable the use of a sample output buffer.

      This method registers an audio output buffer to receive 16-bit signed
      PCM audio samples. Calling this method with a size of 0 or a NULL
      pointer reverts synthesis to the default destination.

      When a buffer full of audio data has been produced, waveformBufferMsg()
      method will be called. The buffer may be accessed from overwritten
      virtual functions using the protected variable 'm_audioBuffer'.
      
  */

  bool setOutputBuffer(/// Buffer size (in samples)
		       int size,
		       /// Pointer to a pre-allocated buffer of shorts
		       short *buffer);

  /// Set the audio output device. Cancel file and buffer output.

  bool setOutputDevice(int deviceNum);

  /** Set the filename to be used for audio output.

      This method sets the file to be used for speech synthesis audio
      output. Buffer and device outputs are disabled. If filename is
      NULL or the string is empty, synthesis reverts to the default
      destination.

      The extension of the file name determines the format of the audio
      samples written to the file:

      .wav:  PCM sound in Windows format.

      .au:   u-law with header.

      .rau:  raw u-law without header.

      If the file does not exist, it is created. If it already exists,
      output is appended to the existing contents.
      
  */

  bool setOutputFilename(const char *filename);
  
  //@}

  /**@name Dictionary maintenance */

  //@{

  /// Deletes a specified dictionary set.
  
  bool deleteDict(/// A handle to a dictionary (from getDict, newDict, ...).
		  ECIDictHand dictionary);

  /// Retrieve the first entry from a dictionary
  
  bool dictFindFirst(/// A handle to a dictionary
		     ECIDictHand dictHandle,
		     /// One of: eciMainDict, eciRootDict, eciAbbvDict.
		     ECIDictVolume whichDictionary,
		     /// The key of the first entry in the dictionary.
		     const char **key,
		     /// The translation of the specific key.
		     const char **translation);

  /// Retrieve the next dictionary entry.
  
  bool dictFindNext(/// A handle to a dictionary
		    ECIDictHand dictHandle,
		    /// One of: eciMainDict, eciRootDict, eciAbbvDict.
		    ECIDictVolume whichDictionary,
		    /// The key of the next entry in the dictionary.
		    const char **key,
		    /// The translation of the specific key.
		    const char **translation);

  /// Retrieve the translation for a key. Return NULL if the key was not found.
  
  const char *dictLookup(/// A handle to a dictionary
			 ECIDictHand dictHandle,
			 /// One of: eciMainDict, eciRootDict, eciAbbvDict.
			 ECIDictVolume whichDictionary,
			 /// The key to be retrieved.
			 const char *key);

  /// Return a handle to the currently active dictionary.
  
  ECIDictHand getDict();

  /// Load a dictionary from a file.
  
  bool loadDict(/// A handle to the dictionary
		ECIDictHand dictHandle,
		/// One of: eciMainDict, eciRootDict, eciAbbvDict.
		ECIDictVolume whichDictionary,
		/// The file name.
		const char *filename);

  /// Create a new dictionary set.
  
  ECIDictHand newDict();

  /// Write the contents of a dictionary to a file
  
  bool saveDict(/// A handle to the dictionary
		ECIDictHand dictHandle,
		/// One of: eciMainDict, eciRootDict, eciAbbvDict.
		ECIDictVolume whichDictionary,
		/// The file name.
		const char *filename);

  /// Activate a specific dictionary
  
  bool setDict(/// A handle to the dictionary
	       ECIDictHand dictHandle);

  /// Add a key-translation pair to a dictionary.
  
  bool updateDict(/// A handle to the dictionary
		  ECIDictHand dictHandle,
		  /// One of: eciMainDict, eciRootDict, eciAbbvDict.
		  ECIDictVolume whichDictionary,
		  /// The key to be added.
		  const char *key,
		  /// The translation to be added.
		  const char *translation);
  
  //@}
  
  /**@name Synthesis control methods */

  //@{

  /// Append new text to the input buffer
  
  bool addText(const char *text);

  /** Clear the input buffer.

      This method clears the input buffer. It does not affect any
      synthesis already in progress. The input buffer can be cleared
      only if the SynthMode parameter (see getSynthMode()) is set to
      manual.

  */
  
  bool clearInput();

  /** Convert text to phonemes.

      This method converts text to phonemes if SynthMode is set to manual.
      The phoneme buffer given using this method is accessible from the
      phonemeBufferMsg() method using the protected m_phonemeBuffer variable.
  */
  
  bool generatePhonemes(int size, char *buffer);

  /// Return the last index reached in an output buffer.
  
  int getIndex();

  /// Insert an index into the input buffer.
  
  bool insertIndex(int indexNum);

  /// Pauses or unpauses speech synthesis and playback.
  
  bool pause(/// Pause (true) or resume (false).
	     bool pauseOn);

  /** Indicates whether synthsis is in progress.

      This method can be used to poll for synthesis in progress. This
      method provides an alternative to blocking your thread's execution
      on the synchronize() method. This method gives the TTS instance
      a chance to check for messages and calling the message-related
      virtual functions.
  */
  
  bool speaking();

  /// Stop synthesis.
  
  bool stop();

  /** Wait in an efficient state until synthesis has finished.

      Waits for the synthesis to be finished. While waiting, calls the
      message-related virtual functions, if any messages have been produced
      by the engine.
  */
  
  bool synchronize();

  /// Starts synthesis of the text in the input buffer. Returns immediately.
  
  bool synthesize();

  /** Synthesize the contents of a file.

      Synthesizes the contents of a file. Returns immediately, i.e., does
      not wait for synthesis to complete. You should use speaking() or
      synchronize() to wait for the synthesis to complete before terminating.
  */

  bool synthesizeFile(const char *filename);
  
  //@}

  /**@name Message-related functions.

     These functions get called in a callback style manner during speech
     synthesis. All these functions should return 'eciDataProcessed' if
     they have processed the relevant data and don't wish to see the
     same message again. If the data was not processed and the function
     would like to see the same data again, it should return
     'eciDataNotProcessed'. The default implementations of these methods
     do nothing but return 'eciDataProcessed'.

  */

  //@{
  
  /** A buffer full (nSamples long) of waveform data received.

      This method is called when the synthesis engine has produced
      a block of audio output. This method will only be called if
      the setOutputBuffer() has been used succesfully to set the output
      buffer.

  */
  
  virtual ECICallbackReturn waveformBufferMsg(long nSamples)
  { (void)nSamples; return eciDataProcessed; }

  /// A buffer full (nBytes long) of phoneme data received.
  
  virtual ECICallbackReturn phonemeBufferMsg(long nBytes)
  { (void)nBytes; return eciDataProcessed; }

  /// Speech synthesis has reached an index (see the insertIndex() method).
  
  virtual ECICallbackReturn indexReplyMsg(long index)
  { (void)index; return eciDataProcessed; }

  /// Speech synthesis is generating the sound for a specific phoneme.
  
  virtual ECICallbackReturn phonemeIndexReplyMsg(ECIMouthData *mouthData)
  { (void)mouthData; return eciDataProcessed; }

  //@}
  
protected:

  short *m_audioBuffer;
  char  *m_phonemeBuffer;
  
private:

  ECIHand m_handle;
  
  static ECICallbackReturn m_callback(ECIHand handle, ECIMessage message,
				      long lparam, void *data);

  void m_error(const char *info) const
  {
    char errorMsg[100];
    eciErrorMessage(m_handle,errorMsg);
    DI_error(DI_ERR_UNKNOWN, info, errorMsg);
  }
  
};

#endif  // DI_TTS_BASE_H
