/* -*- C++ -*- */

//
// DIVA Library. Copyright (C) Jarmo Hiipakka, 1999-2000.
//
// $Log: di_speech_reco_base.C,v $
// Revision 1.3  2002/11/12 11:53:29  tilmonen
// *** empty log message ***
//
// Revision 1.2  2002/10/17 08:25:47  tilmonen
// Bug fixes.
//
// Revision 1.1  2002/10/16 08:45:55  tilmonen
// Wonderful Speech recognition system by
//  - IBM: ViaVoice.
//  - Jarmo Hiipakka: C++ wrapping
//  - Tommi Ilmonen: Fluid wrapping
//

#include <iostream>
using namespace std;

#include <string.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/errno.h>
#include <unistd.h>

#include "di_error.h"

#include "di_speech_reco_base.h"

// ---- Constructors ----------------------------------------------------------

DI_SpeechRecoBase::DI_SpeechRecoBase() :
  m_smapi_socket(0),
  m_smapi_fn(NULL),
  m_smapi_data(NULL),
  m_trans_fsg(NULL),
  m_initialized(false),
  m_connected(false)
{
  char *version;
  int rc(SmApiVersionCheck(SM_API_VERSION_STRING, &version));

  if (rc == SM_RC_WRONG_SM_VERSION)
  {
    DI_fatal(DI_ERR_INTERNAL,"Wrong SMAPI version");
  }
}

DI_SpeechRecoBase::~DI_SpeechRecoBase()
{
}

// ---- Initialization --------------------------------------------------------

bool DI_SpeechRecoBase::open(const char *applicationName, int connectID)
{
  const char *name("DI_SpeechRecoBase::open");
  int counter(0);
  int rc(0);
  SmArg args[30];

  if (!m_initialized)
  {
    SmSetArg(args[counter], SmNapplicationName, applicationName);
    counter++;
    SmSetArg(args[counter], SmNconnectionId, connectID);
    counter++;
    SmSetArg(args[counter], SmNexternalNotifier, notifier);
    counter++;
    SmSetArg(args[counter], SmNexternalNotifierData, this);
    counter++;
    
    rc = SmOpen(counter, args);

    if (rc != SM_RC_OK)
    {
      DI_error(DI_ERR_SYSTEM, name, "SmOpen() failed");
      return false;
    }

    SmAddCallback(SmNconnectCallback,            connectCB,       this);
    SmAddCallback(SmNdisconnectCallback,         disconnectCB,    this);
    SmAddCallback(SmNsetCallback,                setCB,           this);
    SmAddCallback(SmNrequestFocusCallback,       requestFocusCB,  this);
    SmAddCallback(SmNfocusGrantedCallback,       focusGrantedCB,  this);
    SmAddCallback(SmNsetCallback,                setCB,           this);
    SmAddCallback(SmNmicOnCallback,              micOnCB,         this);
    SmAddCallback(SmNmicOffCallback,             micOffCB,        this);
    SmAddCallback(SmNdefineGrammarCallback,      defineGramCB,    this);
    SmAddCallback(SmNdefineVocabCallback,        defineVocCB,     this);
    SmAddCallback(SmNenableVocabCallback,        enableVocCB,     this);
    SmAddCallback(SmNrecognizeNextWordCallback,  nextWordCB,      this);
    SmAddCallback(SmNrecognizedWordCallback,     recoWordCB,      this);
    SmAddCallback(SmNrecognizedPhraseCallback,   recoPhraseCB,    this);
    SmAddCallback(SmNrecognizedTextCallback,     recoTextCB,      this);
    SmAddCallback(SmNutteranceCompletedCallback, utteranceCB,     this);

    m_initialized = true;
  }

  if (m_connected)
    disconnect();

  counter = 0;
  
  SmSetArg(args[counter], SmNuserId, SM_USE_CURRENT);
  counter++;
  SmSetArg(args[counter], SmNenrollId, SM_USE_CURRENT);
  counter++;
  SmSetArg(args[counter], SmNtask, SM_USE_CURRENT);
  counter++;
  SmSetArg(args[counter], SmNrecognize, true);
  counter++;
  SmSetArg(args[counter], SmNoverrideLock, true);
  counter++;

  rc = SmConnect(counter,args,SmAsynchronous);
  if (rc != SM_RC_OK)
  {
    DI_error(DI_ERR_UNKNOWN,"SmConnect failed",SmReturnRcDescription(rc));
    return false;
  }

  m_connected = true;
  
  return true;
}

bool DI_SpeechRecoBase::disconnect()
{
  SmArg args[5];
  SM_MSG reply;
  int counter(0);

  int rc(SmDisconnect(counter,args,&reply));

  if (rc != SM_RC_OK)
  {
    DI_error(DI_ERR_UNKNOWN,"SmDisconnect failed",SmReturnRcDescription(rc));
    return false;
  }

  m_connected = false;
  return true;
}

bool DI_SpeechRecoBase::close()
{
  int rc(SmClose());
  
  if (rc == SM_RC_OK)
  {
    return true;
  }
  else if (rc == SM_RC_SM_NOT_OPEN)
  {
    DI_error(DI_ERR_UNINITIALIZED,"SmClose failed","SMAPI not open");
  }
  else
  {
    DI_error(DI_ERR_UNKNOWN,"SmClose failed");
  }
  return false;
}

// ---- Vocabularies and Grammars ---------------------------------------------

bool DI_SpeechRecoBase::defineVocab(const char *name, int numWords,
				   const char **words)
{
  SM_VOCWORD **vocPtrs;

  vocPtrs = new SM_VOCWORD *[numWords];

  for (int i(0); i<numWords; i++)
  {
    vocPtrs[i] = new SM_VOCWORD;
    vocPtrs[i]->flags = 0;
    vocPtrs[i]->spelling_size = strlen(words[i])+1;
    vocPtrs[i]->spelling = (char *)words[i];
  }

  int rc(SmDefineVocab((char *)name,numWords,vocPtrs,SmAsynchronous));

  if (rc != SM_RC_OK)
  {
    DI_error(DI_ERR_UNKNOWN, "SmDefineVocab failed",SmReturnRcDescription(rc));
  }

  for (int i(0); i<numWords; i++)
  {
    delete vocPtrs[i];
  }
  delete [] vocPtrs;

  return (rc == SM_RC_OK);
}

bool DI_SpeechRecoBase::undefineVocab(const char *name)
{
  int rc(SmUndefineVocab((char *)name, SmAsynchronous));

  if (rc != SM_RC_OK)
  {
    DI_error(DI_ERR_UNKNOWN,"SmUndefineVocab failed",
	     SmReturnRcDescription(rc));
    return false;
  }
  return true;
}

bool DI_SpeechRecoBase::addToVocab(const char *name, int numWords,
				  const char **words)
{
  SM_VOCWORD **vocPtrs;

  vocPtrs = new SM_VOCWORD *[numWords];

  for (int i(0); i<numWords; i++)
  {
    vocPtrs[i] = new SM_VOCWORD;
    vocPtrs[i]->flags = 0;
    vocPtrs[i]->spelling_size = strlen(words[i])+1;
    vocPtrs[i]->spelling = (char *)words[i];
  }

  int rc(SmAddToVocab((char *)name,(short)numWords,vocPtrs,SmAsynchronous));

  if (rc != SM_RC_OK)
  {
    DI_error(DI_ERR_UNKNOWN, "SmAddToVocab failed", SmReturnRcDescription(rc));
  }

  for (int i(0); i<numWords; i++)
  {
    delete vocPtrs[i];
  }
  delete [] vocPtrs;

  return (rc == SM_RC_OK);
}

bool DI_SpeechRecoBase::removeFromVocab(const char *name, int numWords,
				       const char **words)
{
  SM_VOCWORD **vocPtrs;

  vocPtrs = new SM_VOCWORD *[numWords];

  for (int i(0); i<numWords; i++)
  {
    vocPtrs[i] = new SM_VOCWORD;
    vocPtrs[i]->flags = 0;
    vocPtrs[i]->spelling_size = strlen(words[i])+1;
    vocPtrs[i]->spelling = (char *)words[i];
  }

  int rc =
    SmRemoveFromVocab((char *)name,(short)numWords,vocPtrs,SmAsynchronous);

  if (rc != SM_RC_OK)
  {
    DI_error(DI_ERR_UNKNOWN, "SmRemoveFromVocab failed",
	     SmReturnRcDescription(rc));
  }

  for (int i(0); i<numWords; i++)
  {
    delete vocPtrs[i];
  }
  delete [] vocPtrs;

  return (rc == SM_RC_OK);
}

bool DI_SpeechRecoBase::defineGrammar(const char *name, const char *filename)
{
  int rc(SmDefineGrammar((char *)name,(char *)filename,0,SmAsynchronous));

  if (rc != SM_RC_OK)
  {
    DI_error(DI_ERR_UNKNOWN, 
	     "DI_SpeechRecoBase::defineGrammar # SmDefineGrammar failed",
	     SmReturnRcDescription(rc));
  }

  return (rc == SM_RC_OK);
}

bool DI_SpeechRecoBase::loadTranslationGrammar(const char *name)
{
  if (m_trans_fsg)
    unloadTranslationGrammar();
  
  int rc(VtLoadFSG((char *)name, &m_trans_fsg));

  if (rc != 0)
  {
    char *errorMsg;
    VtGetMessage(&errorMsg);
    DI_error(DI_ERR_UNKNOWN,"Error in VtLoadFSG",errorMsg);
  }
  return (rc == 0);
}

bool DI_SpeechRecoBase::unloadTranslationGrammar()
{
  int rc(VtUnloadFSG(m_trans_fsg));

  if (rc != 0)
  {
    char *errorMsg;
    VtGetMessage(&errorMsg);
    DI_error(DI_ERR_UNKNOWN,"Error in VtUnloadFSG",errorMsg);
  }

  m_trans_fsg = NULL;

  return (rc == 0);
}

bool DI_SpeechRecoBase::enableVocab(const char *name)
{
  int rc(SmEnableVocab((char *)name, SmAsynchronous));

  if (rc != SM_RC_OK)
  {
    DI_error(DI_ERR_ARGUMENT, "SmEnableVocab failed",
	     SmReturnRcDescription(rc));
    return false;
  }
  
  return true;
}

bool DI_SpeechRecoBase::disableVocab(const char *name)
{
  int rc(SmDisableVocab((char *)name, SmAsynchronous));

  if (rc != SM_RC_OK)
  {
    DI_error(DI_ERR_ARGUMENT, "SmDisableVocab failed",
	     SmReturnRcDescription(rc));
    return false;
  }
  
  return true;
}

// ---- Recognition -----------------------------------------------------------

bool DI_SpeechRecoBase::requestFocus()
{
  int rc(SmRequestFocus(SmAsynchronous));

  if (rc != SM_RC_OK)
  {
    DI_error(DI_ERR_UNKNOWN,"SmRequestFocus failed");
    return false;
  }
  return true;
}

bool DI_SpeechRecoBase::releaseFocus()
{
  int rc(SmReleaseFocus(SmAsynchronous));

  if (rc != SM_RC_OK)
  {
    DI_error(DI_ERR_UNKNOWN,"SmReleaseFocus failed");
    return false;
  }
  return true;
}

bool DI_SpeechRecoBase::micOn()
{
  int rc(SmMicOn(SmAsynchronous));

  if (rc == SM_RC_OK)
    return true;
  else
  {
    DI_error(DI_ERR_UNKNOWN,"SmMicOn failed");
    return false;
  }
}

bool DI_SpeechRecoBase::micOff()
{
  int rc(SmMicOff(SmAsynchronous));

  if (rc == SM_RC_OK)
    return true;
  else
  {
    DI_error(DI_ERR_UNKNOWN,"SmMicOff failed");
    return false;
  }
}

bool DI_SpeechRecoBase::recognizeNextWord()
{
  int rc(SmRecognizeNextWord(SmAsynchronous));

  if (rc != SM_RC_OK)
  {
    DI_error(DI_ERR_INTERNAL, "SmRecognizeNextWord failed",
	     SmReturnRcDescription(rc));
    return false;
  }
  return true;
}

bool DI_SpeechRecoBase::dispatch()
{
  fd_set rfds;
  int    retval;
  int    rc;
  int    n;

  n = m_smapi_socket + 1;

  /*------------------------------------------------------------------------*/
  /* Check the socket that SMAPI passed in, processing whatever comes in... */
  /*------------------------------------------------------------------------*/
  FD_ZERO ( & rfds );
  FD_SET ( m_smapi_socket, & rfds );
  
  retval = select ( n, & rfds, NULL, NULL, ( struct timeval * ) NULL );
  
  if ( retval < 0 )
  {
    cout << "Retval is " << retval << ", " << errno << endl;
  }
  
  /*------------------------------------------------------------------------*/
  /* Look to see if anything came in from the reco engine.  If so, then     */
  /* call the function that smapi passed in (smapi_fn) with its data        */
  /*------------------------------------------------------------------------*/
  if ( FD_ISSET ( m_smapi_socket, & rfds ) )
  {
    rc = ( m_smapi_fn ) ( m_smapi_data );
    if ( rc != SM_RC_OK )
    {
      cout << ">>  Return code is " << rc << endl;
      return false;
    }
  }
  return true;
}


// -- Static members ----------------------------------------------------------

int DI_SpeechRecoBase::notifier(int socket_handle, int ( * recv_fn ) ( ... ),
				void * recv_data, void * client_data )
{
  ((DI_SpeechRecoBase *)client_data)->m_smapi_socket = socket_handle;
  ((DI_SpeechRecoBase *)client_data)->m_smapi_fn     = recv_fn;
  ((DI_SpeechRecoBase *)client_data)->m_smapi_data   = recv_data;

  return 0;
}

SmHandler DI_SpeechRecoBase::connectReply(SM_MSG *reply, void * call_d)
{
  int rc(SmReturnRc(*reply));
  if (rc != SM_RC_OK)
  {
    DI_error(DI_ERR_UNKNOWN,"SmConnect failed",SmReturnRcDescription(rc));
    return -1;
  }

  return SM_RC_OK;
}

SmHandler DI_SpeechRecoBase::disconnectReply(SM_MSG *reply, void * call_d)
{
  cout << "disconnectReply" << endl;
  return SM_RC_OK;
}

SmHandler DI_SpeechRecoBase::setReply(SM_MSG *reply, void * call_d)
{
  cout << "setReply" << endl;
  return SM_RC_OK;
}

SmHandler DI_SpeechRecoBase::requestFocusReply(SM_MSG *reply, void * call_d)
{
  int rc(SmReturnRc(*reply));
  if (rc != SM_RC_OK)
  {
    if (rc == SM_RC_FOCUS_GRANTED)
    {
      cout << "Remark: " << SmReturnRcDescription(rc) << endl;
      return SM_RC_OK;
    }
    else
    {
      DI_error(DI_ERR_UNKNOWN, "SmRequestFocus failed",
	       SmReturnRcDescription(rc));
      return -1;
    }
  }
  
  return SM_RC_OK;
}

SmHandler DI_SpeechRecoBase::focusGrantedReply(SM_MSG *reply, void * call_d)
{
  int rc(SmReturnRc(*reply));
  if (rc != SM_RC_OK)
  {
    DI_error(DI_ERR_UNKNOWN,"SmRequestFocus failed",SmReturnRcDescription(rc));
    return -1;
  }

  return SM_RC_OK;
}

SmHandler DI_SpeechRecoBase::micOnReply(SM_MSG *reply, void * call_d)
{
  int rc(SmReturnRc(*reply));
  if (rc != SM_RC_OK)
  {
    DI_error(DI_ERR_UNKNOWN,"SmMicOn failed",SmReturnRcDescription(rc));
    return -1;
  }

  return SM_RC_OK;
}

SmHandler DI_SpeechRecoBase::micOffReply(SM_MSG *reply, void * call_d)
{
  cout << "micOffReply" << endl;
  return SM_RC_OK;
}

SmHandler DI_SpeechRecoBase::defineGramReply(SM_MSG *reply, void * call_d)
{
  int rc(SmReturnRc(*reply));
  if (rc != SM_RC_OK)
  {
    DI_error(DI_ERR_UNKNOWN,"SmDefineGrammar failed",
	     SmReturnRcDescription(rc));
    return -1;
  }

  return SM_RC_OK;
}

SmHandler DI_SpeechRecoBase::defineVocReply(SM_MSG *reply, void * call_d)
{
  cout << "defineVocReply" << endl;
  return SM_RC_OK;
}

SmHandler DI_SpeechRecoBase::enableVocReply(SM_MSG *reply, void * call_d)
{
  int rc(SmReturnRc(*reply));
  if (rc != SM_RC_OK)
  {
    DI_error(DI_ERR_UNKNOWN,"SmEnableVocab failed",
	     SmReturnRcDescription(rc));
    return -1;
  }

  return SM_RC_OK;
}

SmHandler DI_SpeechRecoBase::nextWordReply(SM_MSG *reply, void * call_d)
{
  return SM_RC_OK;
}

SmHandler DI_SpeechRecoBase::recoWordReply(SM_MSG *reply, void * call_d)
{
  recognizeNextWord();
  return SM_RC_OK;
}

SmHandler DI_SpeechRecoBase::recoPhraseReply(SM_MSG *reply, void * call_d)
{
  recognizeNextWord();
  return SM_RC_OK;
}

SmHandler DI_SpeechRecoBase::recoTextReply(SM_MSG *reply, void * call_d)
{
  return SM_RC_OK;
}

SmHandler DI_SpeechRecoBase::utteranceReply(SM_MSG *reply, void * call_d)
{
  return SM_RC_OK;
}
