

#include "SerialMouse.hpp"

#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <termios.h>
#include <string.h>
#include <strings.h>


SerialMouse::SerialMouse()
  : m_fd(-1)
{
}

SerialMouse::~SerialMouse() 
{ 
  shutdown(); 
}

bool SerialMouse::init(const char * device)
{
  errno = 0;

  m_buttonState = 0;
  m_memory = 0;
  m_middle = false;

  m_fd = open(device, O_RDWR);

  printf("m_fd = %d\n", m_fd);

  if(m_fd <= 0) {
    puts(strerror(errno));
    return false;
  }

  struct termio sttyArgs;
  ioctl(m_fd, TCGETA , &sttyArgs);

  sttyArgs.c_cflag &= ~CBAUD;
  sttyArgs.c_cflag |=  B1200;
  
  sttyArgs.c_cflag &= ~CSIZE;
  sttyArgs.c_cflag |= CSIZE & CS8;
  
  sttyArgs.c_cflag |= CREAD;	/* Allow reading */
  sttyArgs.c_cflag |= CLOCAL;	/* No modem between us and device */
  
  sttyArgs.c_cc[VMIN]  = 0;	/* Return read even if no chars */
  sttyArgs.c_cc[VTIME] = 0;	/* Return without waiting */
  
  sttyArgs.c_oflag = 0;                  /* Raw output, leave tabs alone */
  sttyArgs.c_lflag = 0;              /* Raw input (no KILL, etc.), no echo */

  if ( ioctl(m_fd, TCSETA, &sttyArgs) == -1 ) {
    printf("ioctl failed (%s)", strerror(errno));
    close(m_fd);
  }

  int iFlags = fcntl(m_fd,F_GETFL);

  iFlags |= ~O_NONBLOCK;

  if (fcntl(m_fd,F_SETFL,iFlags) == (-1))
    printf("SerialMouse::init:fcntl (blocking) %s\n",strerror(errno));

  return true;
}

bool SerialMouse::shutdown()
{
  if(m_fd <= 0)
    return false;

  close(m_fd);

  m_fd = -1;

  return true;
}

bool SerialMouse::nextState(SerialMouseState *state, 
			    int timeoutMs)
{
  int n;
  bool message;
  bool first;
  unsigned char res[4];

  unsigned const bits01  = 0x03; // X6 & X7 bits
  unsigned const bits23  = 0x04 | 0x08; // Y6 & Y7 bits
  unsigned const bits0_5 = 0x3F;// X0-X5 & Y0-Y5 bits

  unsigned char tmp;

  message = false;
  first = false;
  tmp = 0;

  int sleptMs = 0;

  if(m_memory) {
    res[0] = m_memory;
    n = 1;
    first = true;
  }

  m_memory = 0;

  while(!first) {
    n = read(m_fd, res, 1);
    if(res[0] & 0x40)
      first = true;

    usleep(10000);
    sleptMs += 10;
    if(sleptMs > timeoutMs && !first)
      return false;
  }
  
  while(n < 3) {
    n += read(m_fd, res + n, 3 - n);
    if(n != 3)
      usleep(10000);
  }

  // buttons
  if(res[0] & 0x20)
    tmp |= 0x1;  // left button
  if (res[0] & 0x10)
    tmp |= 0x4;  // right button

  // motion
  unsigned char x = (res[0] & bits01) << 6;
  state->m_xMotion = (char) (x | (res[1] & bits0_5));

  unsigned char y = (res[0] & bits23) << 4;
  state->m_yMotion = (char) (y | (res[2] & bits0_5));

  usleep(10000);
  n += read(m_fd, res + 3, 1); //reading 4th byte
  if(n == 4) {
    if(res[3] & 0x40)
      m_memory = res[3];
    else
      m_middle = (res[3] & 0x20) != 0;
  }

  if(m_middle)
    tmp |= 0x2;

  state->m_buttonStates = tmp;

  m_buttonState = tmp;

  printf("%d,%d %x\n", state->m_xMotion, state->m_yMotion, 
	 state->m_buttonStates);
  fflush(0);

  return true;
}
