/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#include <Fluid/Fistiness.hpp>

#include <Fluid/Error.hpp>
#include <Fluid/Processors.hpp>

namespace Fluid {

Fistiness::Fistiness(Processors * host, 
		     const char * id,
		     const std::string &sourceId, 
		     float memorySeconds,
		     ProcessorNodeRequest *req)
  : ProcessorNodeContinuous(host, id),
    m_input(0),
    m_sourceId(sourceId),
    m_memorySeconds(memorySeconds),
    m_sourceRequest(req)
{}

Fistiness::~Fistiness()
{}

bool Fistiness::start()
{
  IONode * node = m_host->findIONode(m_sourceId);
  
  if(!node) {
    if(!m_sourceRequest.ptr())
      return false;

    m_host->requestNode(m_sourceRequest.ptr());

    node = m_host->findIONode(m_sourceId);

    if(!node)
      return false;
  }
  
  m_input = dynamic_cast<IONodeBuffer<GloveSample> *> (node);
  
  if(!m_input) {
    Error("Fistiness::start # Could not get input \"%s\""
	  "# Initialization failed",
	  m_sourceId.c_str());
    return false;
  }

  m_buffer.setSampleRate(m_input->getSampleRate());

  m_buffer.resizeMemory(m_memorySeconds, 3);

  return true;
}

bool Fistiness::stop()
{
  m_input = 0;
  
  return true;
}

void Fistiness::update()
{
  if(!m_input) 
    Fatal("Fistiness::update # There is no glove");

  const float scale = 1.0 / (float) (Glove::FINGERS * Glove::FINGERJOINTS);

  // Trace("Fistiness::update # Source samples = %d", (int) m_glove->getLatestIndex());

  while(processedSamples() < m_input->getSampleCount()) {
    
    const GloveSample & sample = 
      m_input->getIndex(processedSamples());

    float sum = 0.0;

    for(uint i=0; i < Glove::FINGERS; i++) 
      for(uint j=0; j < Glove::FINGERJOINTS; j++) 
	sum += sample.getFingerJoint(i, j);

    sum *= scale;
    m_buffer.put(sum);
    // printf("New fistiness = %f\n", sum);
  }
  m_buffer.copyState(*m_input);

  /* Trace("Fistiness::update # %f %ld", 
     m_buffer.getSampleRate(), m_buffer.getSampleCount()); */
}

IONode *Fistiness::ioGetOutput(const char *id)
{
  return strcmp(id, "output") == 0 ? & m_buffer : 0;
}

std::string Fistiness::ioGetOutputName(unsigned i)
{
  return i == 0 ? "output" : "";
}


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
  
FistinessRequest::FistinessRequest(const std::string &id, 
				   const std::string &sourceId, 
				   float memorySeconds,
				   ProcessorNodeRequest * req)
  : m_id(id),
    m_sourceId(sourceId),
    m_memorySeconds(memorySeconds),
    m_sourceRequest(req)
{}

FistinessRequest::~FistinessRequest()
{}
  
bool FistinessRequest::isEqual(ProcessorNode *node)
{
  Fistiness * f = dynamic_cast<Fistiness *> (node);

  if(!f) return false;

  bool ok = f->sourceId() == m_sourceId;

  if(ok) f->m_buffer.resizeMemory(m_memorySeconds);
  
  return ok;
}

ProcessorNode *FistinessRequest::createNew(Processors *p)
{
  Fistiness * f = new Fistiness
    (p, m_id.c_str(), m_sourceId, m_memorySeconds, m_sourceRequest.clone());

  f->start();
  
  if(f->isOk())
    return f;

  delete f;

  return 0;
}

ProcessorNodeRequest *FistinessRequest::clone() const
{
  return new FistinessRequest(*this);
}

} // namespace Fluid
