/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#include <Fluid/GloveNavigator.hpp>

#include <Fluid/Error.hpp>
#include <Fluid/MessageString.hpp>
#include <Fluid/MessageInt.hpp>
#include <Fluid/Processors.hpp>

namespace Fluid {

  static const char * NavigationGestures[GloveNavigator::FG_SIZEOF] = {
    "none",
    "point-index",
    "open",
    "half-closed",
    "fist"
  };
  
  GloveNavigator::GloveNavigator(Processors * host, 
				 const std::string & id, 
				 const std::string & headLocationStr,
				 const std::string & sensorLocationStr,
				 const std::string & sensorRotationStr,
				 const std::string & gloveSourceStr)
    : ProcessorNodeContinuous(host, id),
      m_headLocationStr(headLocationStr),
      m_sensorLocationStr(sensorLocationStr),
      m_sensorRotationStr(sensorRotationStr),
      m_gloveSourceStr(gloveSourceStr)
  {
    stop();
  }

  GloveNavigator::~GloveNavigator()
  {
    stop();
  }

  IONode *GloveNavigator::ioGetOutput(const char *id)
  {
    if(strcmp(id, "location") == 0)
      return & m_location;
    if(strcmp(id, "rotation") == 0)
      return & m_rotation;
    if(strcmp(id, "scale") == 0)
      return & m_scale;

    return 0;
  }
  
  std::string GloveNavigator::ioGetOutputName(unsigned i)
  {
    if(i == 0) return "location";
    if(i == 1) return "rotation";
    if(i == 2) return "scale";
    return "";
  }
 
  /** We expect to receive messages from the glove gesture source. */

  bool GloveNavigator::msgProcess(const char *id, Message *m)
  {
    if(strcmp(id, "gesture") == 0) {

      if(!m_enabled)
	return true;
      
      MessageString * ms = dynamic_cast<MessageString *> (m);

      if(!ms) return false;

      Trace("GloveNavigator::msgProcess # gesture = %s", ms->string().c_str());
    
      for(uint i=0; i < FG_SIZEOF; i++) {

	if(strcmp(ms->string().c_str(), NavigationGestures[i]) == 0) {

	  m_prevGesture = m_currentGesture;
	  m_currentGesture = (FingerGesture) i;
	  i = FG_SIZEOF;

	  TimeStamp tmp = m_currentTime;
	  m_currentTime = m->getTime();
	  m_deltaTime = m_currentTime - tmp;

	  if(palmFlap(m_prevGesture, m_currentGesture, m_deltaTime)) {
	    m_flaps++;
	  }
	}
      }
      
      return true;
    }
    else if(strcmp(id, "enable") == 0) {

      MessageInt * mi = dynamic_cast<MessageInt *> (m);

      if(!mi) return false;
      
      m_enabled = mi->value();
      
      return true;
    }
    
    return false;
  }
 
  bool GloveNavigator::start()
  {
    const char * fname = "GloveNavigator::start";

    IONode * node = m_host->findIONode(m_sensorLocationStr);
    m_sensorLocation = dynamic_cast<IONodeBuffer<Vector3> *> (node);

    if(m_sensorLocation) {
      m_count = m_sensorLocation->getSampleCount();
    }
    else
      Error("%s # No hand location %s", fname, m_sensorLocationStr.c_str());
      

    node = m_host->findIONode(m_headLocationStr);
    m_headLocation = dynamic_cast<IONodeBuffer<Vector3> *> (node);

    if(!m_headLocation)
      Error("%s # No head location %s", fname, m_headLocationStr.c_str());
    
    node = m_host->findIONode(m_sensorRotationStr);
    m_sensorRotation = dynamic_cast<IONodeBuffer<Matrix3> *> (node);

    if(!m_sensorRotation)
      Error("%s # No rotation %s", fname, m_sensorRotationStr.c_str());
    
    MessageSource * gestures = m_host->findMessageSource(m_gloveSourceStr);

    if(gestures) {
      gestures->msgAddSink("gesture", "gesture", this);
    }
    else {
      Error("%s # No gesture recognizer %s", fname, m_gloveSourceStr.c_str());
    }
    
    bool ok =  m_headLocation && m_sensorLocation && 
      m_sensorRotation && gestures;

    Trace("GloveNavigator::start # %d", (int) ok);

    if(ok) {
      m_location.copyParameters(*m_headLocation);
      m_rotation.copyParameters(*m_headLocation);
      m_scale.copyParameters(*m_headLocation);

      uint size = m_headLocation->size() > 20 ? m_headLocation->size() : 20;

      m_location.resize(size);
      m_rotation.resize(size);
      m_scale.resize(size);

      Matrix3 m;
      m.identity();
      m_rotation.setAll(m);      
      
      Fluid::Vector3 v;
      v.clear();
      m_location.setAll(v);

      m_scale.setAll(1.0);
    }

    return ok;
  }

  bool GloveNavigator::stop()
  {
    m_headLocation = 0;
    m_sensorLocation = 0;
    m_sensorRotation = 0;

    m_count   = 0;
    m_enabled = true;

    m_accelerationScale = 2.0;
    m_rotationScale = 1.0;

    m_accelerationDescale = 0.97;
    m_rotationDescale = 0.97;

    m_flaps   = 0;

    m_vel.clear();
    m_loc.clear();

    m_aziVel = m_azi = 0.0;

    m_worldRot.identity();
    m_dragStartRot.identity();
    m_dragStartLoc.clear();
    m_onDrag = false;

    m_currentGesture = m_prevGesture = FG_NONE;

    m_currentTime = m_deltaTime = 0;


    return true;
  }

  void GloveNavigator::update()
  {
    const char * fname = "GloveNavigator::update";

    // Trace("GloveNavigator::update");

    float frameDuration = 1.0f / m_sensorLocation->getSampleRate();

    for(; m_count < m_sensorLocation->getSampleCount() &&
	  m_count < m_sensorRotation->getSampleCount(); m_count++) {
      
      if(m_flaps) {
	flapMotion();
	m_flaps--;
      }

      if(m_currentGesture == FG_POINTING_INDEX) {
	if(!m_onDrag) {
	  // Start recording a drag
	  m_dragStartRot = m_sensorRotation->getIndex(m_count);
	  m_dragStartLoc = m_sensorLocation->getIndex(m_count);
	  m_onDrag = true;
	  Trace("%s # Began a drag", fname);
	}
	else {
	  dragMotion();
	}
      }
      else if(m_onDrag) {
	m_aziVel = 0.0;
	m_vel.clear();
	m_onDrag = false;
      }

      m_vel *= m_accelerationDescale;
      m_aziVel *= m_rotationDescale;

      if(Math::Abs(m_aziVel) < 0.01 || m_currentGesture == FG_FIST) {
	if(m_aziVel != 0.0)
	  Trace("%s # Rotation stopped", fname);
	m_aziVel = 0.0;
      }
      if(m_vel.length() < 0.01 || m_currentGesture == FG_FIST) {
	if(m_vel.length() != 0.0)
	  Trace("%s # Translation stopped", fname);
	m_vel.clear();
      }
      
      if(m_enabled) {
	m_loc += m_vel * frameDuration;
	m_azi += m_aziVel * frameDuration;
      }
      else {
	m_vel.clear();
	m_aziVel = 0.0;
	m_onDrag = false;
      }

      if(m_azi > Math::TWO_PI)  m_azi -= Math::TWO_PI;
      if(m_azi < -Math::PI) m_azi += Math::TWO_PI;

      TimeStamp time = m_sensorLocation->getTimeStamp(m_count);

      Matrix3 rot;
      rot.rotateY(-m_azi);

      m_worldRot = rot;

      m_location.put(m_loc);
      m_rotation.put(rot);
      m_scale.put(1.0);

      m_location.setLatestTimeStamp(time);
      m_rotation.setLatestTimeStamp(time);
      m_scale.setLatestTimeStamp(time);


      if(m_count % 120 == 0) {
	Vector3 l =  m_location.getNewest();
	Trace("%s # t = %f \n loc = [%f %f %f] m_loc = [%f %f %f] %p", fname, 
	      frameDuration, l[0], l[1], l[2], m_loc[0], m_loc[1], m_loc[2],
	      & m_location.getNewest());
      }

      /* while(m_flaps) {
	
	m_flaps--;
	}*/
      
    }
  }

  void GloveNavigator::flapMotion()
  {
    Matrix3 rot  = m_sensorRotation->getIndex(m_count);
    Vector3 hand = m_sensorLocation->getIndex(m_count);
    Vector3 head = m_headLocation->getIndex(m_count);
    
    // Now we calculate the direction of the palm
    
    Vector3 diff = hand - head;

    // float horizontalDiff = sqrtf(diff.x * diff.x + diff.z * diff.z);
    
    float azi = atan2f(-diff.z, diff.x);
    
    /* Transform the rotation matrix to local coordinates. The local
       coordinates are defined by head = origin and hand =
       z-direction. */
    Matrix3 mainRot;
    mainRot.rotateY(-azi + Math::HALF_PI);
    
    Matrix3 lrot = mainRot * rot;

    Vector3 unit_x(1.0, 0.0, 0.0);
    Vector3 mapped_x = lrot * unit_x;
      
    /* Vector3 unit_y(0.0, 1.0, 0.0);
       Vector3 mapped_y = lrot * unit_y; */

    Vector3 unit_z(0.0, 0.0, 1.0);
    Vector3 mapped_z = lrot * unit_z;
      
    Direction d = DIR_UNDEFINED;
      
    const double s2p2 = Math::SQRT2_PER2;
    
    int hit = -1;
    
    float threshold = 1.0;
    
    // Palm in:
    if((mapped_z - Vector3( 0.0, 1.0, 0.0)).length() < threshold &&
       (mapped_x - Vector3( 0.0, 0.0,-1.0)).length() < threshold) {
      d = DIR_PALM_IN;
      hit = 0;
    }
    if((mapped_z - Vector3(0.0, s2p2, s2p2)).length() < threshold &&
       (mapped_x - Vector3(0.0, s2p2,-s2p2)).length() < threshold) {
      d = DIR_PALM_IN;
      hit = 1;
    }
    if((mapped_z - Vector3(0.0, 0.0, 1.0)).length() < threshold &&
       (mapped_x - Vector3(0.0, 1.0, 0.0)).length() < threshold) {
      d = DIR_PALM_IN;
      hit = 2;
    }

    // Palm out:
    if((mapped_z - Vector3( 0.0,-1.0, 0.0)).length() < threshold &&
       (mapped_x - Vector3( 0.0, 0.0,-1.0)).length() < threshold) {
      d = DIR_PALM_OUT;
      hit = 3;
    }
    if((mapped_z - Vector3(0.0,-s2p2,-s2p2)).length() < threshold &&
       (mapped_x - Vector3(0.0, s2p2,-s2p2)).length() < threshold) {
      d = DIR_PALM_OUT;
      hit = 4;
    }
    if((mapped_z - Vector3(0.0, 0.0,-1.0)).length() < threshold &&
       (mapped_x - Vector3(0.0, 1.0, 0.0)).length() < threshold) {
      d = DIR_PALM_OUT;
      hit = 5;
    }
    // Palm left:
    if((mapped_z - Vector3(-1.0, 0.0, 0.0)).length() < threshold &&
       (mapped_x - Vector3( 0.0, 0.0,-1.0)).length() < threshold) {
      d = DIR_PALM_LEFT;
      hit = 6;
    }
    // Palm right:
    if((mapped_z - Vector3( 1.0, 0.0, 0.0)).length() < threshold &&
       (mapped_x - Vector3( 0.0, 0.0,-1.0)).length() < threshold) {
      d = DIR_PALM_RIGHT;
      hit = 6;
    }
    
    Vector3 direction = m_worldRot * diff;
    direction[1] = 0.0;
    direction.normalize();
    
    if(d == DIR_PALM_IN) {
      m_vel += direction * m_accelerationScale;
    }
    else if(d == DIR_PALM_OUT) {
      m_vel -= direction * m_accelerationScale;
    }
    else if(d == DIR_PALM_LEFT) {
      m_aziVel += m_rotationScale;
    }
    else if(d == DIR_PALM_RIGHT) {
      m_aziVel -= m_rotationScale;
    }
    
    // if(m_count % 70 == 0) {
    Trace("GloveNavigator::flapMotion # d = %d, h = %d "
	  "flaps = %d azi = %f m_azi = %f\n mz [%f %f %f] mx [%f %f %f] \n"
	  " vel [%.3f %.3f %.3f] dir [%.3f %.3f %.3f]", 
	  (int) d, hit, m_flaps, azi, m_azi,
	  mapped_z[0], mapped_z[1], mapped_z[2], 
	  mapped_x[0], mapped_x[1], mapped_x[2],
	  m_vel[0], m_vel[1], m_vel[2], 
	  direction[0], direction[1], direction[2]);
    // }
  }

  void  GloveNavigator::dragMotion()
  {
    Matrix3 rotInv = m_dragStartRot;
    rotInv.transpose();
    
    Matrix3 rot = rotInv * m_sensorRotation->getIndex(m_count);
    Vector3 dir = m_sensorLocation->getIndex(m_count) - m_dragStartLoc;
    
    Vector3 mapped_z = rot * Vector3(0.0, 0.0, 1.0);
    
    float angle = atan2f(-mapped_z.x, mapped_z.z);

    m_aziVel = angle * 0.4;

    m_vel = m_worldRot * dir;
  }

  bool GloveNavigator::palmFlap
  (FingerGesture g1, FingerGesture g2, TimeStamp delta)
  {
    if(((g1 == FG_PALM_OPEN && g2 == FG_PALM_HALF_CLOSED) ||
	(g2 == FG_PALM_OPEN && g1 == FG_PALM_HALF_CLOSED)) &&
       delta.secondsF() < 1.0)
      return true;

    return false;
  }
}
