/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#include <Fluid/GloveRecognizer.hpp>

#include <Fluid/Debug.hpp>
#include <Fluid/Error.hpp>
#include <Fluid/MessageString.hpp>
#include <Fluid/Processors.hpp>

#include <stdio.h>


namespace Fluid {


  void FingerTest::setMode(const char *greater)
  {
    /* int len = strlen(greater);
       for(int i=0; i < len; i++) {
       if(!isspace(greater[i])) {
       greater += i;
       break;
       }
       }*/

    m_greaterThan = (strcmp(greater, "greater") == 0);
    printf("FingerTest::setMode %d \"%s\"\n", (int) m_greaterThan, greater);
  }

  void FingerTest::write(std::ostream *s)
  {
    (*s) << " { " << m_finger << "  " << m_joint << "  "
	 << m_threshold << "  " << (int) m_greaterThan << " }";
  }

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  /** Test if the new sample triggers an event.  */

  bool GloveGesture::trigger(const GloveSample &s)
  {
    std::vector<FingerTest> *test;

    if(m_onGesture && m_exitTest.size())
      test = & m_exitTest;
    else 
      test = & m_entryTest;

    if(!test->size()) return false;

    bool match = true;

    for(uint i = 0; i < test->size() && match; i++) {
      match = match && (*test)[i].accept(s);
    }
  
    bool r = match && !m_onGesture;

    m_onGesture = match;

    return r;
  }
  
  void GloveGesture::write(std::ostream *s)
  {
    (*s) << m_name << " { \n";

    for(uint i = 0; i < m_entryTest.size(); i++) {
      m_entryTest[i].write(s);
      (*s) << "\n";
    }
  
    (*s) << "} \n";
  }

  /////////////////////////////////////////////////////////////////////////////  
  /////////////////////////////////////////////////////////////////////////////  



  GloveRecognizer::GloveRecognizer(Processors * host, 
				   const std::string &id, 
				   bool selectOnlyOne)
    : ProcessorNodeContinuous(host, id),
      m_input(0),
      m_processed(1),
      m_lastGesture(-1),
      m_selectOnlyOne(selectOnlyOne)
  {}

  GloveRecognizer::~GloveRecognizer()
  {}

  void GloveRecognizer::update()
  {
    if(m_selectOnlyOne)
      update1();
    else
      update2();
  }


  void GloveRecognizer::addGesture(const GloveGesture &gesture)
  { 
    m_gestures.push_back(gesture); 
    logStream << "GloveRecognizer::addGesture " << this 
	      << "  " << m_gestures.size() 
	      << std::endl;
  }


  bool GloveRecognizer::setSource(const std::string &name)
  {
    IONode * node = m_host->findIONode(name);

    m_input = dynamic_cast<IONodeBuffer<GloveSample> *> (node);

    return m_input != 0;
  }

  void GloveRecognizer::write(std::ostream *s)
  {
    for(uint i=0; i < m_gestures.size(); i++) {
      m_gestures[i].write(s);
    }
  }

  void GloveRecognizer::update1()
  {
    assert(m_input);

    if(!m_gestures.size())
      return;

    while(m_processed <= m_input->getSampleCount()) {

      const GloveSample &sample = m_input->getIndex(m_processed);
    
      int oldGesture = m_lastGesture;

      if(m_lastGesture != -1) {
	/* First check if the good old gesture would still be valid. */
      
	GloveGesture *gg = & m_gestures[m_lastGesture];
	gg->trigger(sample);

	if(!gg->onGesture())
	  m_lastGesture = -1;
      }
      if(m_lastGesture == -1) {
	/* Try to match any gestyre: */
	for(uint i = 0; i < m_gestures.size(); i++) {

	  GloveGesture *gg = & m_gestures[i];
	  gg->trigger(sample);
	  if(gg->onGesture()) {
	    m_lastGesture = i;
	    break;
	  }
	}
      }


      /*    
	    logStream << "GloveRecognizer::update " << m_lastGesture 
	    << " of " << m_gestures.size() 
	    << " round = " << m_processed << std::endl;
      */
      if(oldGesture != m_lastGesture) {

	//Trace("GloveRecognizer::update # %d -> %d", oldGesture, m_lastGesture);

	MessageString msg;

	if(m_lastGesture == -1)
	  msg.setString("unknown");
	else 
	  msg.setString(m_gestures[m_lastGesture].name());

	msgSend("gesture", & msg);
      }

      m_processed++;
    }
  }

  void GloveRecognizer::update2()
  {
    assert(m_input);

    // Trace("GloveRecognizer::update2");
  
    while(m_processed <= m_input->getSampleCount()) {

      const GloveSample &sample = m_input->getIndex(m_processed);
    
      /* Trace("GloveRecognizer::update2 # %f  %f  %f  %f  %f", 
	 sample.fingerJoints[0][1], sample.fingerJoints[1][1], 
	 sample.fingerJoints[2][1], sample.fingerJoints[3][1], 
	 sample.fingerJoints[4][1]);*/

      for(uint i = 0; i < m_gestures.size(); i++) {

	// Trace("GloveRecognizer::update2 # %u", i);

	GloveGesture *gg = & m_gestures[i];

	bool b1 = gg->onGesture();
	gg->trigger(sample);
	bool b2 = gg->onGesture();

	if(b2 && !b1) {
	  MessageString msg;
	  msg.setString(gg->name());
	  msgSend("gesture", & msg);
	}
      
	/* Trace("GloveRecognizer::update2 # Gesture %s = %d", 
	   gg->name().c_str(), (int) b2); */
      }
      m_processed ++;
    }
  }

  /////////////////////////////////////////////////////////////////////////////  
  /////////////////////////////////////////////////////////////////////////////  


  GloveRecognizerRequest::GloveRecognizerRequest(const std::string &id,
						 const std::string &configFile,
						 const std::string &sourceName,
						 bool selectOnlyOne)
    : m_id(id),
      m_configFile(configFile),
      m_sourceName(sourceName),
      m_selectOnlyOne(selectOnlyOne)
  {}

  GloveRecognizerRequest::~GloveRecognizerRequest()
  {}
  
  bool GloveRecognizerRequest::isEqual(ProcessorNode *node)
  {
    GloveRecognizer *gr = dynamic_cast<GloveRecognizer *> (node);

    if(!gr) return false;
  
    return (m_configFile == gr->configFile()) && 
      (m_sourceName == gr->sourceName()) &&
      (m_selectOnlyOne == gr->selectOnlyOne());
  }

  ProcessorNode *GloveRecognizerRequest::createNew(Processors *p)
  {
    GloveRecognizer *gr = new GloveRecognizer(p, m_id, m_selectOnlyOne);

    bool ok1 = gr->readConfiguration(m_configFile);

    logStream << "GloveRecognizerRequest::createNew " << gr->nofGestures()
	      << std::endl;

    bool ok2 = gr->setSource(m_sourceName);
  
    if(ok1 && ok2) {
      return gr;
    }
  
    errStream << "GloveRecognizerRequest::createNew failed " 
	      << ok1 << "  " << ok2 << std::endl;
  
    delete gr;
  
    return 0;
  }

  ProcessorNodeRequest *GloveRecognizerRequest::clone() const
  {
    return new GloveRecognizerRequest(*this);
  }

} // namespace Fluid
