/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#ifndef FLUID_GLOVE_FINGER_RECOGNIZER_HPP
#define FLUID_GLOVE_FINGER_RECOGNIZER_HPP

#include <Fluid/ProcessorNode.hpp>
#include <Fluid/MessageSource.hpp>

#include <Fluid/Glove.hpp>

#include <iostream>

namespace Fluid {

class FingerTest
{
public:
  FingerTest(Glove::Finger f,
	     Glove::FingerJoint j,
	     float threshold,
	     bool greaterThan)
    : m_finger(f), m_joint(j), 
      m_threshold(threshold), m_greaterThan(greaterThan) {}
  
  FingerTest() : m_finger(0), m_joint(0), 
		 m_threshold(0.0), m_greaterThan(false) {}
  
  bool accept(const GloveSample &s)
  {
    bool isGreater = s.fingerJoints[m_finger][m_joint] > m_threshold;
    if(m_greaterThan) {
      return isGreater;
    }
    return !isGreater;
  }

  void setFinger(Glove::Finger f)     { m_finger = f; }
  void setJoint(Glove::FingerJoint j) { m_joint = j; }
  void setThreshold(float threshold)  { m_threshold = threshold; } 
  void setMode(const char *greater);
  void setMode(bool greaterThanTest) { m_greaterThan = greaterThanTest; }

  Glove::Finger finger() const { return (Glove::Finger) m_finger; }
  Glove::FingerJoint joint() const { return (Glove::FingerJoint) m_joint; }
  float threshold() const { return m_threshold; }
  bool  greaterThan() const { return m_greaterThan; }

  void write(std::ostream *);
  
protected:
  short m_finger;
  short m_joint;
  float m_threshold;
  bool  m_greaterThan;
};


  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

class GloveGesture
{
public:
  GloveGesture(const char *name) : m_name(name), m_onGesture(false) {}
  GloveGesture() : m_onGesture(false) {}
  ~GloveGesture() {}

  bool trigger(const GloveSample &s);
  bool onGesture() const { return m_onGesture; }
  
  std::string & name() { return m_name; }
  void setName(const char *name) { m_name = name; }
  // const char * nameStr() const { return m_name; }
  
  void addEntryTest(const FingerTest &test)
  { m_entryTest.push_back(test); }

  void addExitTest(const FingerTest &test)
  { m_exitTest.push_back(test); }

  void clearTests() { m_entryTest.resize(0); m_exitTest.resize(0); }

  void write(std::ostream *);

protected:

  std::string             m_name;
  bool                    m_onGesture;
  std::vector<FingerTest> m_entryTest;
  std::vector<FingerTest> m_exitTest;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** Recognize the finger gestures of the user. */

class GloveRecognizer : public ProcessorNodeContinuous,
			public MessageSource
{
public:
  GloveRecognizer(Processors * host, const std::string &id, 
		  bool selectOnlyOne);
  virtual ~GloveRecognizer();
  virtual void update();

  void addGesture(const GloveGesture &gesture);

  uint nofGestures() const { return m_gestures.size(); }
  
  bool setSource(const std::string &);
  bool readConfiguration(const std::string &);

  const std::string & configFile() const { return m_configFile; }
  const std::string & sourceName() const { return m_sourceName; }

  /** Returns the current gesture or NULL. */
  GloveGesture *currentGesture()
  {
    return m_lastGesture == -1 ? 0 : & m_gestures[m_lastGesture];
  }

  void write(std::ostream *);

  bool selectOnlyOne() const { return m_selectOnlyOne; }
protected:

  void update1();
  void update2();

  IONodeBuffer<GloveSample> *m_input;

  ulong   m_processed;

  std::vector<GloveGesture> m_gestures;
  int m_lastGesture;

  std::string m_configFile;
  std::string m_sourceName;

  bool        m_selectOnlyOne;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
  
class GloveRecognizerRequest : public ProcessorNodeRequest
{
public:
  GloveRecognizerRequest(const std::string &id,
			 const std::string &configFile,
			 const std::string &sourceName,
			 bool selectOnlyOne);

  virtual ~GloveRecognizerRequest();
  
  virtual bool isEqual(ProcessorNode *);
  virtual ProcessorNode *createNew(Processors *);
  virtual ProcessorNodeRequest *clone() const;

protected:

  std::string m_id;
  std::string m_configFile;
  std::string m_sourceName;  
  bool        m_selectOnlyOne;
};


} // namespace Fluid

#endif
