/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#include <Fluid/MotionNavigator.hpp>

#include <Fluid/Processors.hpp>

namespace Fluid {

  
  MotionNavigator::MotionNavigator(Processors * host, 
				   const std::string & id, 
				   const std::string & trackerSensor,
				   const Vector3 & center,
				   float forwardAzi,
				   const Vector3 & ellipsoidDimensions,
				   const std::string & style)
    : ProcessorNodeContinuous(host, id),
      m_trackerSensor(trackerSensor),
      m_center(center),
      m_forwardAzi(forwardAzi),
      m_ellipsoid(ellipsoidDimensions),
      m_style(style)
  {
    stop();
  }

  MotionNavigator::~MotionNavigator()
  {
    stop();
  }

  IONode *MotionNavigator::ioGetOutput(const char *id)
  {
    if(strcmp(id, "location") == 0)
      return & m_location;
    if(strcmp(id, "rotation") == 0)
      return & m_rotation;
    if(strcmp(id, "scale") == 0)
      return & m_scale;

    return 0;
  }
  
  std::string MotionNavigator::ioGetOutputName(unsigned i)
  {
    if(i == 0) return "location";
    if(i == 1) return "rotation";
    if(i == 2) return "scale";
    return "";
  }
 
 
  bool MotionNavigator::start()
  {
    const char * fname = "MotionNavigator::start";

    stop();

    IONode * node = m_host->findIONode(m_trackerSensor);
    m_sensorLocation = dynamic_cast<IONodeBuffer<Vector3> *> (node);
    
    if(m_sensorLocation) {
      m_count = (long) m_sensorLocation->getSampleCount();
    }
    else {
      Error("%s # No sensor location \"%s\"", fname, m_trackerSensor.c_str());
      return false;
    }

    m_location.copyParameters(*m_sensorLocation);
    m_rotation.copyParameters(*m_sensorLocation);
    m_scale.copyParameters(*m_sensorLocation);
    
    uint size = m_sensorLocation->size() > 20 ? 
      m_sensorLocation->size() : 20;
    
    m_location.resize(size);
    m_rotation.resize(size);
    
    Fluid::Matrix3 m;
    m.identity();
    m_rotation.setAll(m);      
    
    Fluid::Vector3 v;
    v.clear();
    m_location.setAll(v);
    
    m_basicRot.rotateY(m_forwardAzi);
  
    return true;
  }

  bool MotionNavigator::stop()
  {
    m_sensorLocation = 0;

    m_count   = 0;

    /* m_accelerationScale = 2.0;
       m_rotationScale = 4.0; */

    m_loc.clear();

    m_azi = 0.0;

    return true;
  }

  void MotionNavigator::update()
  {
    const char * fname = "MotionNavigator::update";

    // Trace("MotionNavigator::update");

    float frameDuration = 1.0f / m_sensorLocation->getSampleRate();

    for(; m_count < (long) m_sensorLocation->getSampleCount(); m_count++) {

      // Sensor location
      Vector3 sloc = m_sensorLocation->getIndex(m_count);

      // Location in space
      Vector3 loc = m_basicRot * (sloc - m_center);
      Vector3 relDist(loc);
      relDist.descale(m_ellipsoid);

      if(m_style == "angles") {
	float angle = atan2(relDist.x, -relDist.z);

	// Trace("%s # angle = %f", fname, angle);

	// if(Math::Abs(angle) > Math::QUARTER_PI) {
	if(Math::Abs(angle) > Math::PI * 0.75) {

	  float aziVel = angle > 0.0f ? 1.0f : -1.0f;
	  
	  aziVel *= -frameDuration;
	  
	  Matrix3 rot, rot2;
	  rot.rotateY(m_azi);
	  rot2.rotateY(aziVel);
	  Vector3 fromRotCenter = rot * Vector3(0, 0, -2.0) * m_basicRot;
	  // Vector3 fromRotCenter = m_basicRot * rot * Vector3(0, 0, -40.0);
	  // Vector3 toRotCenter = loc - rotationCenter;
	  
	  m_azi += aziVel;
	  m_loc += rot2 * fromRotCenter - fromRotCenter;
	  
	  Trace("%s # aziVel = %f", fname, aziVel);
	  
	}
	
      }
      else {
	if(Math::Abs(relDist.x) > 1.0f) {
	  float aziVel = relDist.x > 0.0f ? 
	    relDist.x - 1.0f : relDist.x + 1.0f;
	  
	  aziVel *= frameDuration * -0.2f;
	  
	  Matrix3 rot, rot2;
	  rot.rotateY(m_azi);
	  rot2.rotateY(aziVel);
	  Vector3 fromRotCenter = rot * Vector3(0, 0, 2.0);
	  // Vector3 toRotCenter = loc - rotationCenter;
	  
	  m_azi += aziVel;
	  m_loc += rot2 * fromRotCenter - fromRotCenter;
	  
	  Trace("%s # aziVel = %f", fname, aziVel);
	}
      }

      Matrix3 rot;
      rot.rotateY(m_azi);

      /* if(Math::Abs(relDist.z) > 1.0) {
	
	float zVel = relDist.x > 0.0f ? 
	  relDist.z - 1.0f : relDist.z + 1.0f;
	
	zVel *= frameDuration * 0.5;
	m_loc += rot.column(2) * -zVel;

	Trace("%s # zVel = %f", fname, zVel);
	}*/
      
      m_location.put(m_loc);
      m_rotation.put(rot);
      m_scale.put(1.0);

      TimeStamp time = m_sensorLocation->getTimeStamp(m_count);
      m_location.setLatestTimeStamp(time);
      m_rotation.setLatestTimeStamp(time);
      m_scale.setLatestTimeStamp(time);
    }
  }

  void MotionNavigator::moveTo(const Vector3 &loc, float azi)
  {
    m_loc = loc;
    m_azi = azi;
  }

}
