/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#include <Fluid/MouseNavigator.hpp>

#include <Fluid/MessageMouse.hpp>
#include <Fluid/MessageInt.hpp>
#include <Fluid/Processors.hpp>

namespace Fluid {

  
  MouseNavigator::MouseNavigator(Processors * host, 
				 const std::string & id, 
				 const std::string & sensorLocationStr,
				 const std::string & sensorLocation2Str,
				 const std::string & sensorRotationStr,
				 const std::string & mouseSourceStr,
				 int button)
    : ProcessorNodeContinuous(host, id),
      m_sensorLocationStr(sensorLocationStr),
      m_sensorLocation2Str(sensorLocation2Str),
      m_sensorRotationStr(sensorRotationStr),
      m_mouseSourceStr(mouseSourceStr),
      m_button(button)
  {
    stop();
  }

  MouseNavigator::~MouseNavigator()
  {
    stop();
  }

  IONode *MouseNavigator::ioGetOutput(const char *id)
  {
    if(strcmp(id, "location") == 0)
      return & m_location;
    if(strcmp(id, "rotation") == 0)
      return & m_rotation;
    if(strcmp(id, "scale") == 0)
      return & m_scale;

    return 0;
  }
  
  std::string MouseNavigator::ioGetOutputName(unsigned i)
  {
    if(i == 0) return "location";
    if(i == 1) return "rotation";
    if(i == 2) return "scale";
    return "";
  }
 
  /** We expect to receive messages from the mouse gesture source. */

  bool MouseNavigator::msgProcess(const char *id, Message *m)
  {
    if(strcmp(id, "mouse") == 0) {

      if(!m_enabled)
	return true;
      
      MessageMouse * ms = dynamic_cast<MessageMouse *> (m);

      if(!ms) return false;

      Trace("MouseNavigator::msgProcess # state = %x", ms->currentState());
      
      if(ms->pressed(m_button)) {
	m_startDrag = true;
      }
      else if(ms->released(m_button)) {
	m_startDrag = false;
	m_onDrag    = false;
	m_vel.clear();
	m_aziVel = 0.0;
      }

      return true;
    }
    else if(strcmp(id, "enable") == 0) {

      MessageInt * mi = dynamic_cast<MessageInt *> (m);

      if(!mi) return false;
      
      m_enabled = mi->value();
      
      return true;
    }
    
    return false;
  }
 
  bool MouseNavigator::start()
  {
    const char * fname = "MouseNavigator::start";

    stop();

    IONode * node = m_host->findIONode(m_sensorLocationStr);
    m_sensorLocation = dynamic_cast<IONodeBuffer<Vector3> *> (node);

    if(m_sensorLocation) {
      m_count = (long) m_sensorLocation->getSampleCount();
    }
    else
      Error("%s # No hand location %s", fname, m_sensorLocationStr.c_str());
      
    node = m_host->findIONode(m_sensorLocation2Str);
    m_sensorLocation2 = dynamic_cast<IONodeBuffer<Vector3> *> (node);

    if(m_sensorLocation2) {
      m_count = (long) m_sensorLocation2->getSampleCount();
    }
    else
      Error("%s # No hand location %s", fname, m_sensorLocation2Str.c_str());

    node = m_host->findIONode(m_sensorRotationStr);
    m_sensorRotation = dynamic_cast<IONodeBuffer<Matrix3> *> (node);

    if(!m_sensorRotation)
      Error("%s # No rotation %s", fname, m_sensorRotationStr.c_str());
    
    MessageSource * buttons = m_host->findMessageSource(m_mouseSourceStr);

    if(buttons) {
      buttons->msgAddSink("mouse", "mouse", this);
    }
    else {
      Error("%s # No button recognizer %s", fname, m_mouseSourceStr.c_str());
    }
    
    bool ok =  m_sensorLocation && m_sensorRotation && buttons;

    Trace("MouseNavigator::start # %d", (int) ok);

    if(ok) {
      m_location.copyParameters(*m_sensorLocation);
      m_rotation.copyParameters(*m_sensorLocation);
      m_scale.copyParameters(*m_sensorLocation);

      uint size = m_sensorLocation->size() > 20 ? 
	m_sensorLocation->size() : 20;

      m_location.resize(size);
      m_rotation.resize(size);
      m_scale.resize(size);

      Fluid::Matrix3 m;
      m.identity();
      m_rotation.setAll(m);      
      
      Fluid::Vector3 v;
      v.clear();
      m_location.setAll(v);

      m_scale.setAll(1.0);
    }

    m_scaleVal = 1.0;

    return ok;
  }

  bool MouseNavigator::stop()
  {
    m_sensorLocation = 0;
    m_sensorRotation = 0;

    m_count   = 0;
    m_enabled = true;

    /* m_accelerationScale = 2.0;
       m_rotationScale = 4.0; */

    m_vel.clear();
    m_loc.clear();

    m_aziVel = m_azi = 0.0;
    m_scaleVal = 1.0;

    m_worldRot.identity();
    m_dragStartRot.identity();
    m_dragStartLoc.clear();
    m_startDrag = false;
    m_onDrag    = false;
    m_onScaling = false;

    return true;
  }

  void MouseNavigator::update()
  {
    const char * fname = "MouseNavigator::update";

    // Trace("MouseNavigator::update");

    float frameDuration = 1.0f / m_sensorLocation->getSampleRate();

    for(; m_count < (long) m_sensorLocation->getSampleCount() &&
	  m_count < (long) m_sensorRotation->getSampleCount(); m_count++) {

      if(m_startDrag) {
	// Start recording a drag
	m_dragStartRot = m_sensorRotation->getIndex(m_count);

	Vector3 loc1 = m_sensorLocation->getIndex(m_count);
	Vector3 loc2 = m_sensorLocation2->getIndex(m_count);

	m_onScaling = (loc1 - loc2).length() < 0.35;

	m_dragStartLoc = m_onScaling ? loc2 : loc1;

	m_onDrag = true;
	m_startDrag = false;
	Trace("%s # Began a drag", fname);
      }
      else if(m_onDrag) {
	dragMotion();
      }
      
      if(m_enabled) {
	m_loc += m_vel * frameDuration;
	m_azi += m_aziVel * frameDuration;
      }
      else {
	m_onDrag = false;
	m_vel.clear();
	m_aziVel = 0.0;
      }

      if(m_azi > Math::TWO_PI)  m_azi -= Math::TWO_PI;
      if(m_azi < -Math::PI) m_azi += Math::TWO_PI;

      TimeStamp time = m_sensorLocation->getTimeStamp(m_count);

      Matrix3 rot;
      rot.rotateY(-m_azi);

      m_worldRot = rot;

      m_location.put(m_loc);
      m_rotation.put(rot);
      m_scale.put(m_scaleVal);

      // Trace("%s # scale = %f %f", fname, m_scaleVal, m_scale.getNewest());

      m_location.setLatestTimeStamp(time);
      m_rotation.setLatestTimeStamp(time);
      m_scale.setLatestTimeStamp(time);

      /*
      if(m_count % 120 == 0) {
	Vector3 l =  m_location.getNewest();
	Trace("%s # t = %f \n loc = [%f %f %f] m_loc = [%f %f %f] %p", fname, 
	      frameDuration, l[0], l[1], l[2], m_loc[0], m_loc[1], m_loc[2],
	      & m_location.getNewest());
      }
      */
    }
  }

  void MouseNavigator::moveTo(const Vector3 &loc, float azi)
  {
    m_loc = loc;
    m_azi = azi;
    m_vel.clear();
    m_aziVel    = 0.0;
    m_onDrag    = false;
    m_startDrag = false;
  }

  void MouseNavigator::dragMotion()
  {    
    if(m_onScaling) {

      Vector3 diff = m_sensorLocation2->getIndex(m_count) - m_dragStartLoc;

      float ychange = diff.y;

      // float rescale = pow(2.0, ychange * 0.01);
      float rescale = 1.0 + ychange * 0.05;

      float scale = m_scaleVal * rescale;

      if(scale < 0.01)  scale = 0.01;
      if(scale > 100.0) scale = 100.0;

      m_scaleVal = scale;

      m_aziVel = 0.0f;
      m_vel.clear();

      return;
    }


    Matrix3 rot = /*rotInv **/ m_sensorRotation->getIndex(m_count);
    Vector3 dir = m_sensorLocation->getIndex(m_count) - m_dragStartLoc;
    
    Vector3 mapped_z1 = m_dragStartRot * Vector3(1.0, 0.0, 0.0);
    Vector3 mapped_z2 = rot * Vector3(1.0, 0.0, 0.0);
    
    float angle1 = atan2f(-mapped_z1.z, mapped_z1.x);
    float angle2 = atan2f(-mapped_z2.z, mapped_z2.x);

    float angle = angle1 - angle2;

    if(angle > Math::PI) angle -= Math::TWO_PI;
    if(angle <-Math::PI) angle += Math::TWO_PI;

    m_aziVel = angle * -0.5;

    m_vel = m_worldRot * dir * -1.8f;

    float l = m_vel.length();

    if(l > 1.0e-6) {
      l = pow(l, 1.5f);
      m_vel.normalize(l / m_scaleVal);
    }

    // m_vel = dir * 30.0;

    /* Trace("MouseNavigator::dragMotion # v = %f %f %f,  azi = %f",
       m_vel.x, m_vel.y, m_vel.z, m_azi); */

  }

}
