/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#include <Fluid/SensorContainment.hpp>

#include <Fluid/Error.hpp>
#include <Fluid/MessageInt.hpp>
#include <Fluid/Processors.hpp>
#include <Fluid/Tracker.hpp>

namespace Fluid {

SensorBoxContainment::SensorBoxContainment(Processors *host,
					   const std::string &id,
					   const std::string &nodeId,
					   const Vector3 &low,
					   const Vector3 &high,
					   float blockSeconds,
					   ProcessorNodeRequest *req)
  : ProcessorNodeContinuous(host, id),
    m_input(0),
    m_low(low),
    m_high(high),
    m_count(0),
    m_within(false),
    m_blockSeconds(blockSeconds),
    m_releaseTime(0),
    m_nodeId(nodeId),
    m_sourceRequest(req)
{}

SensorBoxContainment::~SensorBoxContainment()
{}

bool SensorBoxContainment::start()
{
  IONode * node = m_host->findIONode(m_nodeId);

  m_input = dynamic_cast<IONodeBuffer<Vector3> *> (node);

  if(!m_input) {
    Trace("SensorBoxContainment::SensorBoxContainment # No IO-node \"%s\"",
	  m_nodeId.c_str());
    return false;
  }
  
  m_count = m_input->getSampleCount();  
  return true;
}

bool SensorBoxContainment::stop()
{
  m_input = 0;
  return true;
}
    
void SensorBoxContainment::update()
{
  if(!m_input)
    Fatal("SensorBoxContainment::update # Not operational");
  
  for(; m_count < m_input->getSampleCount(); m_count++) {

    TimeStamp t = m_input->getTimeStamp(m_count);

    if((t-m_releaseTime).secondsF() < m_blockSeconds)
      continue;

    Vector3 v = m_input->getIndex(m_count);
    
    bool within;
    
    /*
    if(m_count % 30 == 0)
      Trace("SensorBoxContainment::update # %.2f %.2f %.2f || "
      "%.2f %.2f %.2f || %.2f %.2f %.2f",
	    v[0], v[1], v[2], m_low[0], m_low[1], m_low[2], 
	    m_high[0], m_high[1], m_high[2]);
    */
    if(v[0] < m_high[0] && 
       v[1] < m_high[1] && 
       v[2] < m_high[2] && 
       v[0] > m_low[0] && 
       v[1] > m_low[1] && 
       v[2] > m_low[2]) {

      within = true;      
    }
    else
      within = false;

    if(within && !m_within) {
      Trace("SensorBoxContainment::update # Hit");
      MessageInt e(1, t); // 1 -> enter box
      msgSend("hit", &e);
    }
    else if(!within && m_within) {
      Trace("SensorBoxContainment::update # Unhit");
      MessageInt e(0, t); // 0 -> exit box
      msgSend("hit", &e);
      m_releaseTime = t;
    }
    
    m_within = within;
  }
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

SensorBoxContainmentRequest::SensorBoxContainmentRequest
(const std::string &id,
 const std::string &nodeId,
 const Vector3 &low,
 const Vector3 &high,
 float blockSeconds,
 ProcessorNodeRequest *req)
  : m_id(id),
    m_nodeId(nodeId),
    m_low(low),
    m_high(high),
    m_blockSeconds(blockSeconds),
    m_sourceRequest(req)
{}

SensorBoxContainmentRequest::~SensorBoxContainmentRequest()
{}

bool SensorBoxContainmentRequest::isEqual(ProcessorNode *node)
{
  SensorBoxContainment *sbc = dynamic_cast<SensorBoxContainment *> (node);

  if(!sbc) return false;

  return (sbc->low() == m_low) && (sbc->high() == m_high) && 
    (sbc->nodeId() == m_nodeId) && (sbc->id() == m_id);
}

ProcessorNode *SensorBoxContainmentRequest::createNew(Processors *host)
{
  SensorBoxContainment *sbc = 
    new SensorBoxContainment(host,
			     m_id,
			     m_nodeId,
			     m_low,
			     m_high,
			     m_blockSeconds,
			     m_sourceRequest.clone());
  
  return sbc;
}

ProcessorNodeRequest *SensorBoxContainmentRequest::clone() const
{
  return new SensorBoxContainmentRequest(*this);
}

}

