
/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#include <Fluid/SensorHitDetector.hpp>

#include <Fluid/Error.hpp>
#include <Fluid/Processors.hpp>
#include <Fluid/Tracker.hpp>

// #include <MgcVector3.h>

namespace Fluid {

SensorHitDetector::SensorHitDetector(Processors *host,
				     const std::string &node1,
				     const std::string &node2,
				     float distance,
				     float accelerationThreshold)
  : ProcessorNodeContinuous(host),
    m_distance(distance),
    m_acceleration(accelerationThreshold),
    m_hitEnergy(0.0),
    m_hitLocation(0.0, 0.0, 0.0),
    m_onHit(false)
{
  m_nodeNames[0] = node1;
  m_nodeNames[1] = node2;

  m_inputs[0] = m_inputs[1] = 0;

  start();
}

SensorHitDetector::~SensorHitDetector()
{}

bool SensorHitDetector::start()
{
  m_inputs[0] = m_inputs[1] = 0;

  for(int i=0; i < 2; i++) {
    IONode * node = m_host->findIONode(m_nodeNames[i]);

    m_inputs[i] = dynamic_cast<IONodeBuffer<Vector3> *> (node);

    if(!m_inputs[i]) {
      Error("SensorHitDetector::SensorHitDetector # No node \"%s\"",
	    m_nodeNames[i].c_str());
      return false;
    }
    else
      Trace("SensorHitDetector::SensorHitDetector # Got source %d", i);
  }

  m_sampleIndex = m_inputs[0]->getSampleCount();
  if(m_sampleIndex < 3) m_sampleIndex = 3;

  return true;
}

bool SensorHitDetector::stop()
{
  m_inputs[0] = m_inputs[1] = 0;
  return true;
}
    
void SensorHitDetector::update()
{
  /* Trace("SensorHitDetector::update # %lu %lu %lu", 
     m_sampleIndex, m_inputs[0]->getSampleCount(), 
     m_inputs[1]->getSampleCount());*/

  float sampleRate = m_inputs[0]->getSampleRate();

  while(m_sampleIndex < m_inputs[0]->getSampleCount() &&
	m_sampleIndex < m_inputs[1]->getSampleCount()) {

    /* if(m_sampleIndex % 128 == 127) {
      
      printf("History size = %d\n", (int) m_tracker->getHistorySize());

      for(long i=0; i < 30; i++) {
	const TrackerSample &s1 = m_tracker->getSample(m_sampleIndex - i);
	Vector3 p = s1.getLocation(m_sensorsIds[0]);
	printf("delayed pos[%d] = [ %f %f %f ] \n", i, p[0], p[1], p[2]);
      }
      }*/

    m_sampleIndex++;

    // Accelerations of the sensors:
    Vector3 acc[2];
    const Vector3 *pos[2];

    // const TrackerSample &s1 = m_tracker->getSample(m_sampleIndex);
    // const TrackerSample &s2 = m_tracker->getSample(m_sampleIndex - 1);
    // const TrackerSample &s3 = m_tracker->getSample(m_sampleIndex - 2);

    for(uint i=0; i < 2; i++) {

      IONodeBuffer<Vector3> * input = m_inputs[i];

      const Vector3 &p1 = input->getIndex(m_sampleIndex);
      const Vector3 &p2 = input->getIndex(m_sampleIndex - 1);
      const Vector3 &p3 = input->getIndex(m_sampleIndex - 2);

      Vector3 v1 = p1 - p2;
      Vector3 v2 = p2 - p3;

      acc[i] = (v1 - v2) * (sampleRate * sampleRate);

      // Trace("p1 = [ %f %f %f ]", p1[0], p1[1], p1[2]);
      
      /* if(i)
	Trace("p1 = [ %f %f %f ] p2 = [ %f %f %f ] p3 = [ %f %f %f ] "
	      "v1 = [ %f %f %f ] v2 = [ %f %f %f ] a = [ %f %f %f ]",
	      p1[0], p1[1], p1[2], 
	      p2[0], p2[1], p2[2],
	      p3[0], p3[1], p3[2],
	      v1[0], v1[1], v1[2],
	      v2[0], v2[1], v2[2],
	      acc[i][0], acc[i][1], acc[i][2]); */
      
      pos[i] = & p1;
    }

    float accsum = acc[0].length() + acc[1].length();
    
    float distance = ((*pos[0]) - (*pos[1])).length();
    
    /* Trace("SensorHitDetector::update # d = %3.3f, acc = %3.4f bl = %lu", 
       distance, accsum, m_inputs[0]->size()); */

    // Deliver the event:

    if(!m_onHit) {
      if(accsum > m_acceleration && distance < m_distance) {
	Trace("SensorHitDetector::update # EVENT 0x%p "
	      "acceleration = %f distance = %f", this, accsum, distance);
 
	m_hitEnergy = accsum;
	m_hitLocation = ((*pos[0]) + (*pos[1])) * 0.5;

	MessageHit mh(accsum, 
		      m_hitLocation, 
		      m_inputs[0]->getTimeStamp((long) m_sampleIndex));

	// Message m;
	msgSend(0, &mh);
	m_onHit = true;
      }
    }
    else if(distance > m_distance) {
      Trace("SensorHitDetector::update # NOT on hit anymore, distance = %f",
	    distance);
      m_onHit = false;
    }

    /*else 
       Trace("SensorHitDetector::update # accsum = %f distance = %f",
	 accsum, distance); */
  }
  
  // Trace("SensorHitDetector::update # %p distance = %f", this, distance);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

SensorHitDetectorRequest::SensorHitDetectorRequest
(const std::string &node1,
 const std::string &node2,
 float distance,
 float accelerationThreshold)
  : m_node1(node1), m_node2(node2),
    m_distance(distance), m_acceleration(accelerationThreshold)
{}

SensorHitDetectorRequest::~SensorHitDetectorRequest()
{}

bool SensorHitDetectorRequest::isEqual(ProcessorNode *node)
{
  SensorHitDetector *shd = dynamic_cast<SensorHitDetector *> (node);

  if(!shd) return false;

  return 
    (shd->node1() == m_node1) &&
    (shd->node2() == m_node2) &&
    (shd->distanceThreshold() == m_distance) &&
    (shd->accelerationThreshold() == m_acceleration);
}

ProcessorNode *SensorHitDetectorRequest::createNew(Processors *host)
{
  SensorHitDetector *detector = 
    new SensorHitDetector(host, m_node1, m_node2, 
			  m_distance, m_acceleration);

  if(!detector->ok()) {
    Error("SensorHitDetectorRequest::createNew # detector is not ok %s %s",
	  detector->node1().c_str(), detector->node2().c_str());
    delete detector;
    return 0;
  }
  return detector;
}

ProcessorNodeRequest *SensorHitDetectorRequest::clone() const
{
  return new SensorHitDetectorRequest(*this);
}

} // namespace Fluid

