/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#ifndef FLUID_SENSOR_HIT_DETECTOR_HPP
#define FLUID_SENSOR_HIT_DETECTOR_HPP

#include <Fluid/Message.hpp>
#include <Fluid/MessageSource.hpp>
#include <Fluid/ProcessorNode.hpp>
#include <Fluid/Tracker.hpp>

namespace Fluid {

  class Tracker;

  class MessageHit : public Message
  {
  public:
    MessageHit(float hitEnergy, 
	       const Vector3 & hitLocation, 
	       const TimeStamp &t)
      : Message(t), m_hitEnergy(hitEnergy), m_hitLocation(hitLocation) {}
    virtual ~MessageHit() {}

    /// Energy of the latest hit
    float hitEnergy() const { return m_hitEnergy; }
    /// Location of the latest hit
    const Vector3 & hitLocation() { return m_hitLocation; }

    
  protected:
    float m_hitEnergy;

    Vector3 m_hitLocation;
  };

  /** Detect the colision of two tracker sensors. This class emits
      events as collision is detected. */

  class SensorHitDetector : public MessageSource,
			    public ProcessorNodeContinuous
  {
  public:
    SensorHitDetector(Processors *host,
		      const std::string &node1,
		      const std::string &node2,
		      float distance,
		      float accelerationThreshold);
    ~SensorHitDetector();

    virtual bool start();
    virtual bool stop();
    
    virtual void update();
    
    bool ok() { return m_inputs[0] && m_inputs[1]; }

    const std::string & node1() const { return m_nodeNames[0]; }
    const std::string & node2() const { return m_nodeNames[1]; }

    float distanceThreshold() const { return m_distance; }
    float accelerationThreshold() const { return m_acceleration; }

    /// Energy of the latest hit
    float hitEnergy() const { return m_hitEnergy; }
    /// Location of the latest hit
    const Vector3 & hitLocation() { return m_hitLocation; }

  protected:
    IONodeBuffer<Vector3> *m_inputs[2];

    float m_distance;
    float m_acceleration;
    float m_hitEnergy;

    Vector3 m_hitLocation;

    /// The number of processed samples.
    ulong  m_sampleIndex;

    bool  m_onHit;

    std::string m_nodeNames[2];
  };

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

  class SensorHitDetectorRequest : public ProcessorNodeRequest
  {
  public:
    SensorHitDetectorRequest(const std::string &sensor1,
			     const std::string &sensor2,
			     float distance,
			     float accelerationThreshold);

    virtual ~SensorHitDetectorRequest();

    virtual bool isEqual(ProcessorNode *);
    virtual ProcessorNode *createNew(Processors *);
  virtual ProcessorNodeRequest *clone() const;

  protected:
    std::string m_node1, m_node2;
    float m_distance, m_acceleration;
  };

} // namespace Fluid


#endif
