/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#include <Fluid/TrackerSensorSelector.hpp>

#include <Fluid/Error.hpp>
#include <Fluid/InputDevices.hpp>
#include <Fluid/Processors.hpp>

namespace Fluid {
  
  TrackerSensorSelector::TrackerSensorSelector(Processors * host, 
					       const std::string &id,
					       const std::string &trackerName,
					       const std::string &sensorName)
    : ProcessorNodeContinuous(host, id),
      m_trackerName(trackerName),
      m_sensorName(sensorName)
  {
  }

  TrackerSensorSelector::~TrackerSensorSelector()
  {}
  
  bool TrackerSensorSelector::start()
  {
    m_input = 0;
    m_index = -1;

    Tracker * tracker = InputDevices::getTracker(m_trackerName);
    
    if(!tracker) {
      Error("TrackerSensorSelector::start # No tracker \"%s\"", 
	    m_trackerName.c_str());
      return false;
    }

    m_input = & tracker->outputBuffer();
    m_index =   tracker->findSensor(m_sensorName);

    m_locations.copyParameters(*m_input);
    m_rotations.copyParameters(*m_input);    

    Trace("TrackerSensorSelector::start # index = %d bl = %lu", 
	  m_index, m_locations.size());

    return true;
  }

  bool TrackerSensorSelector::stop()
  {
    m_input = 0;
    return true;
  }
    
  void TrackerSensorSelector::update()
  {
    assert(m_input);

    /* Trace("TrackerSensorSelector::update # %ul %ul",
       m_locations.getSampleCount(), m_input->getSampleCount()); */

    while(m_locations.getSampleCount() < m_input->getSampleCount()) {
      
      const TrackerSample * sample = 
	& m_input->getIndex(m_locations.getSampleCount());
      
      m_locations.put(sample->getLocation(m_index));
      m_rotations.put(sample->getRotation(m_index));
    }
    
    m_locations.setLatestTimeStamp(m_input->getLatestTimeStamp());
    m_rotations.setLatestTimeStamp(m_input->getLatestTimeStamp());
  }

  IONode *TrackerSensorSelector::ioGetOutput(const char *id)
  {
    if(strcmp(id, "rotation") == 0)
      return & m_rotations;
    if(strcmp(id, "location") == 0)
      return & m_locations;

    return 0;
  }

  std::string TrackerSensorSelector::ioGetOutputName(unsigned i)
  {
    if(i == 0) return "rotation";
    if(i == 1) return "location";
    return "";
  }

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////
  
  TrackerSensorSelector2::TrackerSensorSelector2
  (Processors * host, 
   const std::string &id,
   const std::string &trackerName,
   const std::string &sensorName,
   const std::string &rotationName,
   const std::string &locationName,
   const std::string &scaleName,
    const Vector3 & offset)
    : ProcessorNodeContinuous(host, id),
      m_trackerName(trackerName),
      m_sensorName(sensorName),
      m_rotationName(rotationName),
      m_locationName(locationName),
      m_scaleName(scaleName),
      m_offset(offset)
  {}

  TrackerSensorSelector2::~TrackerSensorSelector2()
  {}
  
  bool TrackerSensorSelector2::start()
  {
    m_input = 0;
    m_index = -1;

    Tracker * tracker = InputDevices::getTracker(m_trackerName);
    
    if(!tracker) {
      Error("TrackerSensorSelector2::start # No tracker \"%s\"", 
	    m_trackerName.c_str());
      return false;
    }

    m_input = & tracker->outputBuffer();
    m_index =   tracker->findSensor(m_sensorName);

    m_locations.copyParameters(*m_input);
    m_rotations.copyParameters(*m_input);

    IONode * node = m_host->findIONode(m_rotationName);
    m_wrotation = dynamic_cast<IONodeBuffer<Matrix3> *> (node);

    node = m_host->findIONode(m_locationName);
    m_wlocation = dynamic_cast<IONodeBuffer<Vector3> *> (node);

    node = m_host->findIONode(m_scaleName);
    m_wscale = dynamic_cast<IONodeBuffer<float> *> (node);

    Trace("TrackerSensorSelector2::start # %p index = %d bl = %lu ws = %p n = %p",
	  this, m_index, m_locations.size(), m_wscale, node);

    if(!m_wscale)
      Error("TrackerSensorSelector2::start # %p No scale node \"%s\"",
	    this, m_scaleName.c_str());

    return m_input && m_wrotation && m_wlocation && m_wscale;
  }

  bool TrackerSensorSelector2::stop()
  {
    Trace("TrackerSensorSelector2::stop # index = %d", m_index);
    m_input = 0;
    return true;
  }
    
  void TrackerSensorSelector2::update()
  {
    assert(m_input);

    /* Trace("TrackerSensorSelector2::update # %ul %ul",
       m_locations.getSampleCount(), m_input->getSampleCount()); */

    while(m_locations.getSampleCount() < m_input->getSampleCount() && 
	  m_locations.getSampleCount() < m_wrotation->getSampleCount() && 
	  m_locations.getSampleCount() < m_wlocation->getSampleCount() && 
	  m_locations.getSampleCount() < m_wscale->getSampleCount()) {
      
      const TrackerSample * sample = 
	& m_input->getIndex(m_locations.getSampleCount());
      
      const Fluid::Matrix3 & r = 
	m_wrotation->getIndex(m_locations.getSampleCount());
      const Fluid::Vector3 & l = 
	m_wlocation->getIndex(m_locations.getSampleCount());

      float scale;

      if(m_wscale)
	scale = 1.0f / m_wscale->getIndex(m_locations.getSampleCount());
      else
	scale = 1.0f;

      m_locations.put(scale * (r * (sample->getLocation(m_index) + m_offset)) 
		      + l);
      m_rotations.put(r * sample->getRotation(m_index));
    }
    
    m_locations.setLatestTimeStamp(m_input->getLatestTimeStamp());
    m_rotations.setLatestTimeStamp(m_input->getLatestTimeStamp());
  }
    
  IONode *TrackerSensorSelector2::ioGetOutput(const char *id)
  {
    if(strcmp(id, "rotation") == 0)
      return & m_rotations;
    if(strcmp(id, "location") == 0)
      return & m_locations;

    return 0;
  }

  std::string TrackerSensorSelector2::ioGetOutputName(unsigned i)
  {
    if(i == 0) return "rotation";
    if(i == 1) return "location";
    return "";
  }


}
