/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Authors:   Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa 2001-2002
 * Copyright: Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa 2001-2002
 *
 * Tommi.Ilmonen@hut.fi
 * Tuukka.Heikura@hut.fi
 * Marko.Myllymaa@hut.fi
 *
 */

#ifndef FLUID_CLONABLE_POINTER_H
#define FLUID_CLONABLE_POINTER_H

namespace Fluid {

/** This class implements pointer-carrier/cloning functionality. When
    you need to store and \b copy objects of abstract type then this
    class is handy. The base class that is the type "T" must have a
    method "T * clone() const" that is used to copy (or clone)
    objects.

    @author Tommi Ilmonen
*/

template <typename T>
class ClonablePointer
{
public:
  /// Construct pointer carrier with NULL pointer
  ClonablePointer() : m_ptr(0) {}
  /// Takes hold of a pointer (does not clone)
  ClonablePointer(T *ptr) : m_ptr(ptr) {}
  /// Clone a pointer from another carrier
  ClonablePointer(const ClonablePointer &that) 
    : m_ptr(that.m_ptr ? that.m_ptr->clone() : 0) {}
  
  /// Deletes the hidden pointer
  ~ClonablePointer() { delete m_ptr; }
    
  /// Clone a pointer from another carrier
  ClonablePointer & operator = (const ClonablePointer &that)
  { delete m_ptr; m_ptr = that.clone(); return *this; }

  /// Takes hold of a pointer (does not clone)
  ClonablePointer & operator = (T * that)
  { delete m_ptr; m_ptr = that ; return *this; }

  /// Returns the pointer (potentially NULL).
  T * ptr() { return m_ptr; }

  /// Returns clone of the pointer (or NULL).
  T * clone() const { return m_ptr ? m_ptr->clone() : 0; }

  /// Clones the source pointer into this object
  void cloneFrom(const T * ptr) { delete m_ptr; m_ptr = ptr ? ptr->clone() : 0; }

  /// Forgets the pointer without deleting it.
  void forget() { m_ptr = 0; }

protected:
  T * m_ptr;
};

}

#endif
