/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#include <Fluid/Matrix3.hpp>

#include <Fluid/Error.hpp>
#include <Fluid/InputDevice.hpp>
#include <Fluid/InputDevices.hpp>
#include <Fluid/IONode.hpp>
#include <Fluid/MessageSource.hpp>
#include <Fluid/Processors.hpp>

#include <algorithm>
#include <typeinfo>

#include <string.h>

/* namespace Fluid {

  static void Foo()
  {
    Fluid::Matrix3 m1, m2, m3;
    m3 = m1 * m2;
  }

}

static void Fee()
{
  Fluid::Matrix3 m1, m2, m3;
  m3 = m1 * m2;
}
*/

namespace Fluid {

Processors::Processors()
  : m_running(false)
{}

/** Empty destructor. <B>This destructor does not delete any of the
    managed child objects.</B>*/

Processors::~Processors()
{
  if(m_running) stop();  
}

/** Clear the list of processors. The processor nodes are optionally
    deleted. */

void Processors::clear(bool deleteChildren)
{
  if(deleteChildren) {
    ProcessorNode **node = &* m_nodes.begin();
    ProcessorNode **sentinel  = &* m_nodes.end();

    while(node != sentinel) {
      (*node)->setHost(0);
      delete (*node);
      node++;
    }
  }
  m_nodes.resize(0);
  m_cnodes.resize(0);
}

/** Add a new processor node to the collection. If the node is derived
    from the ProcessorNodeContinuous -base class, then the inserted
    node will be updated when the processor collection is updated. */

void Processors::addNode(ProcessorNode *node)
{
  m_nodes.push_back(node);

  ProcessorNodeContinuous *continuous = 
    dynamic_cast<ProcessorNodeContinuous *> (node);
    
  if(continuous) {
    m_cnodes.push_back(continuous);
  }
}

/** Remove a processor from the list. The processor is not deleted. 

    @return This methods returns true if the node was successfully
    removed and false otherwise.
*/
bool Processors::removeNode(ProcessorNode *node)
{
  std::vector<ProcessorNode *>::iterator it = 
    std::find(m_nodes.begin(), m_nodes.end(), node);

  if(it != m_nodes.end()) m_nodes.erase(it);
  else return false;

  ProcessorNodeContinuous *continuous = 
    dynamic_cast<ProcessorNodeContinuous *> (node);
    
  if(!continuous)
    return true;
  
  std::vector<ProcessorNodeContinuous *>::iterator it2 = 
    std::find(m_cnodes.begin(), m_cnodes.end(), continuous);
  
  if(it2 != m_cnodes.end()) m_cnodes.erase(it2);
  return true;
}


IONode * Processors::findIONode(const char * sourceId)
{
  const char * fname = "Processors::findIONode";

  Trace("%s # \"%s\"", fname, sourceId);

  IONode * node = 0;
  IOSource * source = 0;

  const char * nodeId = 0;

  if(strncmp(sourceId, "../", 3) == 0) {

    sourceId += 3;

    if(strncmp(sourceId, "inputdevices/", 13) == 0) {
      Trace("%s # Getting IONode from InputDevices", fname);
      sourceId += 13;

      std::string tmp(sourceId);
      
      const char * mark = strchr(sourceId, '/');
      if(mark)
	tmp.resize(mark - sourceId);
	
      InputDeviceBase *device = InputDevices::getInputDevice(tmp);

      if(!device) {
	Error("%s # No input device \"%s\"", fname, tmp.c_str());
	return 0;
      }

      source = device;

      nodeId = strchr(sourceId, '/');

      if(nodeId) nodeId++;
    }
    else 
      Error("%s # Could not locate IOSource", fname, sourceId - 3);
    
  }
  else {
    const char * mark = strchr(sourceId, '/');

    if(!mark) {
      Error("%s # Could not locate IOSource (separator missing)", fname);      
      return 0;
    }
    std::string sourceName = sourceId;
    
    sourceName.resize(mark - sourceId); // Truncate to erase the IONode name

    ProcessorNode * processor = findNode(sourceName.c_str());

    source = dynamic_cast<IOSource *> (processor);

    if(processor && !source) {
      Error("%s # Processor %s is not source", 
	    fname, typeid(*processor).name());      
    }

    nodeId = mark + 1;
  }

  if(!source) {
    Error("%s # Could not locate IOSource (source missing)", fname);      
    return 0;
  }

  node = source->ioGetOutput(nodeId);
  
  Trace("%s # Got node %p \"%s\" from source %p %s", 
	fname, node, nodeId, source, typeid(*source).name());

  return node;
}

ProcessorNode *Processors::findNode(const char *id)
{
  Trace("Processors::findNode # \"%s\"", id);

  uint i, n = m_nodes.size();

  for(i=0; i < n; i++) {
    ProcessorNode *node = m_nodes[i];
    if(node->id() == id)
      return node;
  }

  Error("Processors::findNode # FAILED \"%s\"", id);
  
  return 0;
}

MessageSource *Processors::findMessageSource(const char *id)
{
  ProcessorNode * node = findNode(id);

  if(node)
    return dynamic_cast<MessageSource *> (node);

  return 0;
}

/** Get (request) a processor node. This method is of critical
    importance if you wish to use Fluid::Processors objects. The
    strategy to get a new node is to:

    <OL>

    <LI> Create a request that is capable of recognizing and
    manufacturing a new processor.

    <LI> Pass this request object to the processors -object.

    <LI> This method then returns the new processor or NULL if the
    processor could not be created.

    </OL>

*/

ProcessorNode *Processors::requestNode(ProcessorNodeRequest *request)
{
  for(uint i = m_nodes.size(); i--; ) {
    if(request->isEqual(m_nodes[i])) {
      return m_nodes[i];
    }
  }
  
  ProcessorNode *newNode = request->createNew(this);
  
  if(newNode) addNode(newNode);
  
  return newNode;
}
  
bool Processors::start()
{
  if(m_running) return false;

  bool ok = true;

  for(uint i = 0; i < m_nodes.size(); i++)
    ok = ok && m_nodes[i]->start();

  m_running = true;

  return ok;
}

/** Updates all the nodes that need continuous updating. */
void Processors::update()
{
  if(!m_running) {
    Error("Processors::update # Should call \"start\" before \"update\"");
    return;
  }

  for(uint i = 0; i < m_cnodes.size(); i++)
    m_cnodes[i]->update();
}
 
bool Processors::stop()
{
  if(!m_running) return false;

  bool ok = true;

  for(uint i = 0; i < m_nodes.size(); i++)
    ok = ok && m_nodes[i]->stop();

  m_running = false;

  return ok;
}

} // namespace Fluid
