/* COPYRIGHT
 *
 * This file is part of the Geometry Engine - GEE
 *
 * Read the "GEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen 2003
 *
 * firstname.lastname@hut.fi
 *
 */

#include <CaveProj.hpp>

#include <stdio.h>

namespace CaveProj
{

  Wall::Wall(void)
    : m_index	(-1),
      m_eyes	(0x1 << EYE_LEFT),
      m_active	(false),
      m_tracked	(false),
      m_pleq	(0.f, 0.f, 0.f, 0.f),
      m_origin	(0.f, 0.f),
      m_size	(0.f, 0.f),
      m_border	(false),
      m_name    ("Unkown"),
      m_display (":0.0")
  {

    m_corner[CORNER_LOWERLEFT]. make(0, 0, 0);
    m_corner[CORNER_LOWERRIGHT].make(1, 0, 0);
    m_corner[CORNER_UPPERLEFT]. make(0, 1, 0);
    m_corner[CORNER_UPPERRIGHT].make(1, 1, 0);

    m_matrix[0].identity();
    m_matrix[1].identity();
  }

  Wall::~Wall(void)
  {
    // empty
  }

  void Wall::resetPleq(void)
  {
    const Vector3 & a = m_corner[CORNER_LOWERLEFT];
    const Vector3 & b = m_corner[CORNER_LOWERRIGHT];
    const Vector3 & c = m_corner[CORNER_UPPERLEFT];

    float x1 = b.x - a.x;
    float y1 = b.y - a.y;
    float z1 = b.z - a.z;
    float x2 = c.x - a.x;
    float y2 = c.y - a.y;
    float z2 = c.z - a.z;
    float nx = (y1 * z2) - (z1 * y2);
    float ny = (z1 * x2) - (x1 * z2);
    float nz = (x1 * y2) - (y1 * x2);

    float d = nx * nx + ny * ny + nz * nz;
    if (d == 0.f)
      return;		// three corners are collinear

    d = 1.f / sqrtf(d);

    m_pleq.x = nx * d;
    m_pleq.y = ny * d;
    m_pleq.z = nz * d;
    m_pleq.w = -a.x * m_pleq.x - a.y * m_pleq.y - a.z * m_pleq.z;
  }
  
  void Wall::dump(std::ostream &os)
  {
    char buf[512];
    sprintf(buf, "Wall %d (%s):\n Display: %s\n"
	    " Eyes: %x\n Active: %d\n Tracked: %d\n Corners:\n", 
	    m_index, m_name.c_str(), m_display.c_str(),
	    m_eyes, (int) m_active, (int) m_tracked);
    os << buf;
    int i;

    for(i = 0; i < 4; i++)
      os << "  " << m_corner[i] << "\n";

    os << " Plane equation: " << m_pleq 
       << "\n Origin: " << m_origin
       << "\n Size: " << m_size
       << "\n Border: " << (int) m_border
       << "\n Matrices:\n  ";
    os << m_matrix[0] << "\n  " << m_matrix[1] << "\n";
  }
  

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////


  Cave * Cave::s_instance = 0;

  Cave::Cave(void)
    : m_leftEye(0.f, 0.f, 0.f),
      m_rightEye(0.f, 0.f, 0.f),
      m_headPosition(0.f, 0.f, 0.f),
      m_headRight   (1.f, 0.f, 0.f),
      m_interocularDistance(0.16f),
      m_nearClip(0.01f),
      m_farClip (100000.f)
  {
    if(!s_instance)
      s_instance = this;
  }

  Cave::~Cave(void)
  {
    if(s_instance == this)
      s_instance = 0;
  }

  bool Cave::init(const char* filename)
  {
    if(filename)
      parseJugglerConfig(filename);

    for (int i = 0; i < wallCount(); i++)
      m_walls[i].setIndex(i);

    return m_walls.size();
  }


  //------------------------------------------------------------------------

  void Cave::setHeadPosition(const Vector3& headPosition, const Vector3& right)
  {
    Vector3 r = right;
    r.normalize();

    m_headPosition = headPosition;
    m_headRight    = r;
    m_leftEye  = headPosition - r * m_interocularDistance/2.0f;
    m_rightEye = headPosition + r * m_interocularDistance/2.0f;

    // std::cout << "Eyes are: " << m_leftEye << " and " << m_rightEye << "\n";
  }

  void Cave::setInterocularDistance (float d)
  {
    m_interocularDistance=d;
    setHeadPosition(m_headPosition, m_headRight);
  }

  void Cave::setEyePositions(const Vector3& left,const Vector3& right)
  {
    m_leftEye  = left;
    m_rightEye = right;
    m_headPosition = (left+right)/2.f;
    m_headRight    = right-left;
  }

  void Cave::setClipDistances(float nearClip, float farClip)
  {
    m_nearClip = nearClip;
    m_farClip  = farClip;
  }

  bool Cave::calculateProjections()
  {
    for(int w = 0; w < wallCount(); w++) {
      Wall & wall = getWall(w);

      for(int i = 0; i < 2; i++) {
	if(wall.m_eyes & (0x1 << i))
	  calculateProjection(w, i);
      }
    }
    return true;
  }  

  //-----------------------------------------------------
  // Compute projection matrix
  //-----------------------------------------------------

  bool Cave::calculateProjection(int w, int ieye)
  {
    assert((uint) w < m_walls.size());

    // get current context and eye position
    Wall & wall = m_walls[w];
    // Vector3    eye = wall.isStereo() ? (ieye ? m_rightEye : m_leftEye) : (m_rightEye + m_leftEye) * 0.5f;
    Vector3    eye = ieye ? m_rightEye : m_leftEye;

    // project eye point onto the wall
    Vector4 eyeh(eye.x, eye.y, eye.z, 1.f);
    const Vector4& pleq = wall.getPleq();
    float eyed = dot(eyeh, pleq);
    Vector3 eyew = eye - eyed * Vector3(pleq.x, pleq.y, pleq.z);

    /* printf("Matrix %d %d %f (%f %f %f)\n ", w, ieye, eyed,
       eye.x, eye.y, eye.z); */

    // get coordinate vectors for wall coordinate system
    Vector3 wx = (wall.getCorner(CORNER_LOWERRIGHT) - 
		  wall.getCorner(CORNER_LOWERLEFT)).normalize();
    Vector3 wy = (wall.getCorner(CORNER_UPPERLEFT)  - 
		  wall.getCorner(CORNER_LOWERLEFT)).normalize();
    Vector3 wz = cross(wx, wy);

    // generate the matrix that transforms the wall into unit space
    Matrix4 wallMtx;
    // wallMtx.identity();
    wallMtx.identity();
    wallMtx.setColumn3b(0, wx);
    wallMtx.setColumn3b(1, wy);
    wallMtx.setColumn3b(2, wz);
    wallMtx.setColumn3b(3, eyew);
    bool ok = true;
    wallMtx = wallMtx.inverse(&ok);

    Matrix4 unitMtx;
    unitMtx.identity();
    unitMtx[2][3] = eyed;
    
    // std::cout << pleq << "\n " << eyeh << "\n " << unitMtx << std::endl ;

    unitMtx = unitMtx.inverse(&ok);
    
    // std::cout << unitMtx << "\n " ;

    wallMtx *= unitMtx;
    // wallMtx = wallMtx * unitMtx;
    Matrix4 wallToUnit = wallMtx;

    Vector3 p0 = wallToUnit * wall.getCorner(CORNER_LOWERLEFT);
    Vector3 p1 = wallToUnit * wall.getCorner(CORNER_UPPERRIGHT);

    // compute the standard matrix for fov 90/90 degrees
    float fnear   = Cave::m_nearClip;
    float ffar    = Cave::m_farClip;
    float fleft   = -(p0.x / p0.z) * fnear;
    float fright  = -(p1.x / p1.z) * fnear;
    float ftop    = -(p1.y / p1.z) * fnear;
    float fbottom = -(p0.y / p0.z) * fnear;
    Matrix4 mtx;
    mtx.identity();

    mtx[0][0] = 2 * fnear / (fright-fleft);
    mtx[1][1] = 2 * fnear / (ftop-fbottom);
    mtx[0][2] = (fright+fleft) / (fright-fleft);
    mtx[1][2] = (ftop+fbottom) / (ftop-fbottom);
    mtx[2][2] = -(ffar+fnear) / (ffar-fnear);
    mtx[2][3] = -(2 * ffar * fnear) / (ffar-fnear);
    mtx[3][2] = -1;
    mtx[3][3] = 0;

    // take the wall coordinate system into account
    wall.m_matrix[ieye] = mtx * wallToUnit;

    /* 
    std::cout << wallToUnit << "\n " 
	      << mtx << "\n "
	      << wall.m_matrix[ieye] << std::endl;
    */

    return true;
  }


  //------------------------------------------------------------------------
  // Configuration file parser
  //------------------------------------------------------------------------

  static const char* findParam(const char* string, int p)
  {
    int n = 0;
    p = 2 * p - 1; // 1->1, 2->3, 3->5, ...
    uint len = strlen(string);
    if(len > 1024)
      len = 1024;

    for(uint i = 0; i < len; i++) {
      if(string[i] == '\"') n++;
      if(n == p)    return & string[i];
    }

    puts("returning NULL");

    return NULL;
  }

  void Cave::parseJugglerConfig(const char * filename)
  {
    char tmp[1024], tmp2[1024];
    FILE * fp = fopen(filename, "rb");

    if(!fp) {
      fprintf(stderr, 
	      "CaveProj::Cave::parseJugglerConfig # Could not open \"%s\"\n",
	      filename);
      fflush(stderr);
      return;
    }

    while(fgets(tmp, 1024, fp)) {

      if(strstr(tmp, "surfaceDisplay")) { // display -> parse data
	Wall wall;

	while(fgets(tmp, 1024, fp)) { // read line

	  if(strstr(tmp, "end")) { // NOTE: less than bullet proof...
	    m_walls.push_back(wall);
	    break;
	  }
	  
	  if(strstr(tmp, "Name")) {
	    tmp2[0] = '\0';
	    sscanf(findParam(tmp, 1), "\"%s\"", tmp2);
	    wall.m_name = tmp2;
	    cleanString(wall.m_name);
	  }
	  
	  else if(strstr(tmp, "Display")) {
	    tmp2[0] = '\0';
	    sscanf(findParam(tmp, 1), "\"%s\"", tmp2);
	    wall.m_display = tmp2;
	    cleanString(wall.m_display);
	  }

	  else if(strstr(tmp, "origin")) {

	    Vector2 origin;
	    sscanf(findParam(tmp, 1), "\"%f\"",  &origin.x);
	    sscanf(findParam(tmp, 2), "\"%f\"",  &origin.y);
	    wall.setOrigin(origin);
	  }

	  else if(strstr(tmp, "size")) {
	    Vector2 size;
	    sscanf(findParam(tmp, 1), "\"%f\"",  &size.x);
	    sscanf(findParam(tmp, 2), "\"%f\"",  &size.y);
	    wall.setSize(size);
	  }

	  else if(strstr(tmp, "view")) {
	    if(strstr(tmp, "stereo"))
	      wall.setEyes(0x3);
	    if(strstr(tmp, "left"))
	      wall.setEyes(0x1);
	    if(strstr(tmp, "right"))
	      wall.setEyes(0x2);
	  }
	  else if(strstr(tmp, "border"))
	    wall.setBorder(strstr(tmp, "True") ? true : false);

	  else if(strstr(tmp, "tracked"))
	    wall.setTracked(strstr(tmp, "True") ? true : false);

	  else if(strstr(tmp, "active"))
	    wall.setActive(strstr(tmp, "True") ? true : false);


	  else if(strstr(tmp, "corners")) {

	    int nesting = (strstr(tmp, "{")) ? 1 : 0;
	    int corner  = -1;
	    Vector3 pos;

	    while(fgets(tmp, 1024, fp)) { // read line

	      if(strstr(tmp, "{"))  nesting++;
	      if(strstr(tmp, "}"))  nesting--;
	      if(nesting==0)
		break;    // NOTE: seems to be an exception...

	      int component = -1;

	      if(strstr(tmp, "Name")) {

		if(strstr(tmp, "lower left"))  corner = CORNER_LOWERLEFT;
		if(strstr(tmp, "lower right")) corner = CORNER_LOWERRIGHT;
		if(strstr(tmp, "upper left"))  corner = CORNER_UPPERLEFT;
		if(strstr(tmp, "upper right")) corner = CORNER_UPPERRIGHT;
	      }
	      else if(strstr(tmp, "x"))
		component = 0;
	      else if(strstr(tmp, "y"))
		component = 1;
	      else if(strstr(tmp, "z"))
		component = 2;

	      if (corner >= 0 && component >= 0) {

		float p = 0.0f;
		if(sscanf(findParam(tmp, 1), "\"%f\"",  &p) == 1) {
		  
		  Vector3 pos = wall.getCorner(corner);
		  pos[component] = p;
		  wall.setCorner(corner,  pos);
		}
	      }
	    }
	  }
	}
      }
      else { // something else -> skip

	int nesting = 0;
	while(fgets(tmp, 1024, fp)) { // read line

	  // HACK HACK
	  if(strstr(tmp, "interocular_distance"))
	    sscanf(findParam(tmp, 1), "\"%f\"",  &m_interocularDistance);

	  if(strstr(tmp, "{"))
	    nesting++;

	  if(strstr(tmp, "}"))
	    nesting--;

	  assert(nesting>=0);

	  if(nesting==0)   // on top level
	    if(strstr(tmp, "end")) // NOTE: less than bullet proof...
	      break;
	}
      }
    }

    fclose(fp);

    // recompute plane equations for walls

    for (int i = 0; i < wallCount(); i++)
      m_walls[i].resetPleq();
  }

  void Cave::clear()
  { 
    m_walls.clear(); 
  }

  void Cave::dump(std::ostream &os)
  {
    os << "Cave:\n Left eye = " << m_leftEye 
       << "\n Right eye = " << m_rightEye 
       << "\n Interocular distance = " << m_interocularDistance << "\n";

    for(uint i=0; i < m_walls.size(); i++)
      m_walls[i].dump(os);
  }

  /** Seeks a wall with given name. Return -1 if the wall cannot be
      found. */

  int Cave::getWallIndex(const char * name)
  {
    for(uint i = 0; i < m_walls.size(); i++)
      if(m_walls[i].name() == name)
	return i;

    return -1;
  }

  void Cave::cleanString(std::string & str)
  {
    uint tmp = str.size();

    if(tmp) {
      tmp--;
      if(str[tmp] == '\"')
	str.resize(tmp);
    }
  }

}
