/* COPYRIGHT
 *
 * This file is part of the Geometry Engine - GEE
 *
 * Read the "GEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen 2003
 *
 * firstname.lastname@hut.fi
 *
 */

#ifndef _CAVEPROJ_HPP_
#define _CAVEPROJ_HPP_

#include <Fluid/Matrix4.hpp>

#include <string>
#include <vector>

namespace CaveProj
{
  using namespace Fluid;
  
  enum {
    CORNER_LOWERLEFT  = 0,
    CORNER_LOWERRIGHT = 1,
    CORNER_UPPERRIGHT = 2,
    CORNER_UPPERLEFT  = 3
  };

  enum {
    EYE_LEFT  = 0,
    EYE_RIGHT = 1,
    NUM_CONTEXTS = 4
  };

  /** Originalsource code from Timo Aila and Matias. */
  
  class Wall
  {
  public:
    Wall();
    ~Wall();

    void resetPleq   (void);

    int     getIndex (void) const  { return m_index; }
    bool    isStereo (void) const  { return m_eyes == 0x3; }
    bool    isActive (void) const  { return m_active; }
    bool    isTracked (void) const  { return m_tracked; }
    const Vector3& getCorner (uint corner) const 
    { assert(corner < 4); return m_corner[corner]; }
    const Vector4& getPleq  (void) const  { return m_pleq; }
    const Vector2& getOrigin (void) const  { return m_origin; }
    const Vector2& getSize  (void) const  { return m_size; }
    bool    isBorder (void) const  { return m_border; }

    void    setSize  (const Vector2& s) { m_size = s; }
    void    setBorder (bool border)   { m_border = border; }
    void    setActive (bool active)   { m_active = active; }
    void    setTracked (bool tracked)   { m_tracked = tracked; }
    void    setCorner (uint corner, const Vector3& pos) 
    { assert(corner < 4); m_corner[corner] = pos;}

    void    setOrigin (const Vector2& o) { m_origin = o; }
    void    setIndex (int id)    { m_index = id; }
    void    setEyes (int mask)  { m_eyes = mask; }

    const Matrix4 & matrix(int eye) const { return m_matrix[eye]; }

    void dump(std::ostream &os);

    const std::string & name()    const { return m_name; }
    const std::string & display() const { return m_display; }

  protected:

    friend class Cave;

    int     m_index;   // set when created

    // These parameters are parsed from Juggler's config file
    int     m_eyes;
    bool    m_active;   // is it enabled?
    bool    m_tracked;
    Vector3   m_corner[4];  // coordinates of the corners
    Vector4   m_pleq;    // plane equation

    // Legacy data from Juggler config
    Vector2   m_origin;   // Window origin (pixels)
    Vector2   m_size;     // Window size (pixels)
    bool      m_border;   // Window border
    Matrix4   m_matrix[2]; // Matrices for both eyes
    std::string m_name;
    std::string m_display;
  };

  class Cave
  {
  public:
    Cave  (void);
    ~Cave  (void);


    /// calls parseJugglerConfig is filename!=NULL
    bool    init    (const char* filename);
    void    parseJugglerConfig (const char* filename);

    /// Position of the eyes. NOTE: the unit is arbitrary
    void    setEyePositions   (const Vector3& left, 
			       const Vector3& right);
    void    setHeadPosition   (const Vector3& headPosition, 
			       const Vector3& right);
    /// default value parsed from Juggler config
    void    setInterocularDistance (float d); 

    // Clip range, Znear and Zfar. 
    void    setClipDistances (float nearClip, float farClip);

    Wall &  getWall  (int index) { return m_walls[index]; }
    int     wallCount() { return m_walls.size(); }

    bool    calculateProjections();
    bool    calculateProjection(int wall, int eye);

    void    clear();
    void    dump(std::ostream &os);

    int     getWallIndex(const char * name);

    static Cave * instance() { return s_instance; }

  private:

    void cleanString(std::string & str);

    std::vector<Wall> m_walls;

    // set by application (possibly from a tracker):
    Vector3  m_leftEye; 
    Vector3  m_rightEye;

    Vector3  m_headPosition;
    Vector3  m_headRight;

    // from Juggler config:
    float   m_interocularDistance;  

    float   m_nearClip;
    float   m_farClip;

    static Cave * s_instance;
  };
}

#endif
