/* COPYRIGHT
 *
 * This file is part of the Geometry Engine - GEE
 *
 * Read the "GEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen 2003
 *
 * firstname.lastname@hut.fi
 *
 */

#include <VrSmuggler.hpp>

#include <Fluid/InputDevices.hpp>
#include <Fluid/Tracker.hpp>

#include <di_error.h>

#include <list>
#include <map>

#include <GL/glu.h>

using namespace CaveProj;

// #define DEBUG 1

#ifdef DEBUG
inline void OUTPUT(const char * buf)
{
  printf(buf); 
  fflush(stdout);
}
#else
#define OUTPUT(x)
#endif

static bool VRS_debug = false;

VRS_Sync::VRS_Sync()
  : m_threads(0),
    m_drawn(0),
    m_swapped(0),
    m_allowDrawExit(false),
    m_allowSwapExit(false)
{}

VRS_Sync::~VRS_Sync()
{}


/** Indicates to the master system that one frame was drawn. The call
    returns only after all threads have called this method.
    
    Called by renderer. 
*/

void VRS_Sync::oneWasDrawn()
{
  if(VRS_debug)
    DI_trace2("VRS_Sync::oneWasDrawn # %u %u", m_drawn, m_threads);

  OUTPUT(":");

  m_mutex.lock();

  m_drawn++;

  m_cond.broadcast();

  while(m_drawn != m_threads || !m_allowDrawExit)
    m_cond.wait(m_mutex);

  m_mutex.unlock();
  m_mutex.lock();

  m_drawnCleared++;
  m_cond.broadcast();

  m_mutex.unlock();
}

/** Indicates to the master system that one frame was swapped. The
    call returns only after all threads have called this method.

    Also used after context init.
    
    Called by renderer. 
*/

void VRS_Sync::oneWasSwapped()
{
  if(VRS_debug)
    DI_trace2("VRS_Sync::oneWasSwapped # %u %u", m_swapped, m_threads);

  OUTPUT(";");

  m_mutex.lock();
  m_swapped++;

  m_cond.broadcast();

  while(m_swapped != m_threads || !m_allowSwapExit)
    m_cond.wait(m_mutex);

  m_mutex.unlock();
  m_mutex.lock();

  m_swappedCleared++;
  m_cond.broadcast();
  
  m_mutex.unlock();  
}

/// Called by master
void VRS_Sync::waitDrawEnd()
{
  if(VRS_debug)
    DI_trace2("VRS_Sync::waitDrawEnd # %u %u", m_drawn, m_threads);

  OUTPUT(".");

  m_mutex.lock();

  m_swapped = 0;
  m_drawnCleared = 0;
  m_allowSwapExit = false;
  m_allowDrawExit = true;

  m_cond.broadcast();

  while(m_drawn != m_threads)
    m_cond.wait(m_mutex);

  OUTPUT("<");

  while(m_drawnCleared != m_threads)
    m_cond.wait(m_mutex);

  OUTPUT(">");

  m_mutex.unlock();  
}

/// Called by master
void VRS_Sync::waitSwapEnd()
{
  if(VRS_debug)
    DI_trace2("VRS_Sync::waitSwapEnd # %u %u", m_swapped, m_threads);

  OUTPUT(",");

  m_mutex.lock();

  m_drawn = 0;
  m_swappedCleared = 0;
  m_allowDrawExit = false;
  m_allowSwapExit = true;

  m_cond.broadcast();

  while(m_swapped != m_threads)
    m_cond.wait(m_mutex);
  
  OUTPUT("<");

  while(m_swappedCleared != m_threads)
    m_cond.wait(m_mutex);

  OUTPUT(">");

  m_mutex.unlock();  
}

/// Called by master
void VRS_Sync::waitSwaps()
{
  if(VRS_debug)
    DI_trace2("VRS_Sync::waitSwaps # %u %u", m_swapped, m_threads);

  OUTPUT("#");

  m_mutex.lock();
  while(m_swapped != m_threads)
    m_cond.wait(m_mutex);
  m_mutex.unlock();  
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

VRS_RenderThread::VRS_RenderThread(VRS_Sync * sync, VRS_GlApp * app)
  : m_sync(sync),
    m_app(app),
    m_run(true)
{}

VRS_RenderThread::~VRS_RenderThread()
{}

void VRS_RenderThread::childLoop()
{
  const char * fname = "VRS_RenderThread::childLoop";

  uint i, n = m_windows.size();

  DI_trace2("%s # %p %u", fname, this, n);

  // First context init:
  for(i = 0; i < n; i++) {
    
    VRS_Window & window = * m_windows[i];
    Wall * wall = window.m_wall;
    
    assert(wall);

    Vector2 size = wall->getSize();
    Vector2 origin = wall->getOrigin();

    int flags = GLX_LOCAL;
    if(wall->isStereo())
      flags |= GLX_USE_STEREO;

    if(!wall->isBorder())
      flags |= GLX_NODECOR;

    GLX_Open(window.m_glx, wall->display().c_str(), 
	     (int) origin.x, (int) origin.y,
	     (int) size.x, (int) size.y, flags);
    GLX_SetActive(window.m_glx);

    // Now draw a demo screen that shows that we are alive:
    renderDemo(&window);
    GLX_Swap(window.m_glx);
    renderDemo(&window);
    GLX_Swap(window.m_glx);

    m_app->contextInit( & window);
  }

  DI_trace2("%s # %p Init done", fname, this);

  m_sync->oneWasSwapped();

  DI_trace2("%s # %p to loop", fname, this);

  Matrix4 tmp;

  uint frame = 0;

  while(m_run) {

    frame++;

    // Render one frame:
    for(i = 0; i < n; i++) {
      
      VRS_Window & window = * m_windows[i];
      
      GLX_SetActive(window.m_glx);

      const Vector2 & size = window.m_wall->getSize();
      glViewport (0, 0, (int) size.x, (int) size.y);
      glDepthFunc (GL_LESS);
      glEnable (GL_DEPTH_TEST);
      
      // Set up projections:
      
      Wall * wall = window.m_wall;

      glMatrixMode(GL_PROJECTION);
      if(wall->isStereo()) {
	renderWithMatrix(GL_BACK_LEFT,  EYE_LEFT,  & window);
	renderWithMatrix(GL_BACK_RIGHT, EYE_RIGHT, & window);
      }
      else {
	renderWithMatrix(GL_BACK, EYE_LEFT, & window);
      }

      GLX_PreSwap(window.m_glx);
      
      /* Get events every once in a while. */
      if(frame % 5 == i % 5)
	if(GLX_GetEvents(window.m_glx))
	  m_app->escPressed();
    }

    if(VRS_debug)
      DI_trace2("%s # %p Drawn once", fname, this);

    m_sync->oneWasDrawn();
    
    // Swap buffers:
    for(i = 0; i < n; i++)
      
      GLX_Swap(m_windows[i]->m_glx);
    
    /* Implicitly stops here until the master calls us to render yet
       another frame. */
    m_sync->oneWasSwapped();
  } 
 
  DI_trace2("%s # %p Exit loop -> clean up contexts", fname, this);

  for(i = 0; i < n; i++) {
    VRS_Window * window = m_windows[i];
    
    GLX_SetActive(window->m_glx);
    m_app->contextCleanup(window);
    
    GLX_Close(window->m_glx);

    delete window;
  }

  m_windows.clear();

  DI_trace2("%s # %p EXIT", fname, this);
}

void VRS_RenderThread::renderWithMatrix(GLenum buffer, 
					 int eye,
					 VRS_Window *window)
{
  const Fluid::Matrix4 & projMtx = window->m_wall->matrix(eye);

  glDrawBuffer(buffer);
  glMatrixMode(GL_PROJECTION);
  Matrix4 tmp = projMtx;
  tmp.transpose();
  
  glLoadMatrixf(tmp.data());

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  m_app->draw(window, eye);
}

void VRS_RenderThread::renderDemo(VRS_Window *window)
{
  const Vector2 & size = window->wall()->getSize();
  glViewport (0, 0, (int) size.x, (int) size.y);  

  glMatrixMode(GL_PROJECTION);
  gluOrtho2D(-1, 1, -1, 1);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  float gray = 0.1f;
  
  glClearColor(gray, gray, gray, 1.0);
  glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  glEnable(GL_COLOR_MATERIAL);
  glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);

  static Vector3 colors[3] = {
    Vector3(0.6, 0.2, 0.2),
    Vector3(0.5, 0.5, 0.2),
    Vector3(0.3, 0.5, 0.5)
  };

  static Vector3 vertices[3] = {
    Vector3(0.0, 0.5, 0.0),
    Vector3(0.5, -0.5, 0.0),
    Vector3(-0.5, -0.5, 0.0)
  };

  glBegin(GL_TRIANGLES);
  
  for(int i = 0; i < 3; i++) {
    glColor3fv(colors[i].data());
    glVertex3fv(vertices[i].data());
  }

  glEnd();  
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

VRS_GlApp::VRS_GlApp()
  : m_run(false)
{}

VRS_GlApp::~VRS_GlApp()
{
  if(m_run)
    fprintf(stderr, "VRS_GlApp::~VRS_GlApp # Called while still running\n");
  fflush(stderr);
}

bool VRS_GlApp::loadVrConfig(const char * filename)
{
  m_cave.init(filename);
  // m_cave.dump(std::cout);
  return m_cave.wallCount() != 0;
}

bool VRS_GlApp::exec(bool threaded)
{
  const char * fname = "VRS_GlApp::exec";

  if(!m_cave.wallCount()) {
    DI_error2(DI_ERR_UNINITIALIZED, "%s # no walls", fname);
    return false;
  }

  m_run = true;

  if(threaded)
    run();
  else
    childLoop();

  return true;
}

/** Shuts down the application. If the application calls this method
    (in preFrame etc.) then the argument "inThread" should be set to
    "true". */

bool VRS_GlApp::quit(bool inThread)
{
  m_run = false;

  if(inThread)
    return true;

  waitEnd();

  return true;
}

void VRS_GlApp::childLoop()
{
  const char * fname = "VRS_GlApp::childLoop";

  uint i, walls = m_cave.wallCount();

  std::vector<VRS_RenderThread *> threads;

  DI_trace2("%s # %u walls", fname, walls);

  // Calculate the relevant number of threads to use:
  {
    typedef std::map<std::string, std::list<uint> > displayMap;
    displayMap displays;
    
    for(i=0; i < walls; i++) {
      Wall & wall = m_cave.getWall(i);
      displays[wall.display()].push_back(i);
    }

    m_sync.setThreadNum(displays.size());

    for(displayMap::iterator it = displays.begin(); it != displays.end();it++){

      VRS_RenderThread * thread = new VRS_RenderThread(&m_sync, this);

      std::list<uint> & wallList = (*it).second;
      
      for(std::list<uint>::iterator it2 = wallList.begin();
	  it2 != wallList.end(); it2++) {
	uint index = (*it2);

	VRS_Window * rw = new VRS_Window();
	rw->m_wall = & m_cave.getWall(index);
	thread->addWindow(rw);
      }

      threads.push_back(thread);
      // Span thread:
      thread->run();
    }
  }
  
  DI_trace2("%s # %u threads", fname, (uint) threads.size());  

  /* Threads are now running and will eventually get their contexts
     correct in the first waitSwapEnd. We need to wait until hat
     happens. */
  
  m_sync.waitSwaps();

  DI_trace2("%s # Start loop", fname);

  uint frames = 0;

  while(m_run) {
    preFrame();

    setHeadPosition();
    m_cave.calculateProjections();

    m_sync.waitSwapEnd();
    /* As we release the threads with this function they immediately
       go to swap-buffers. */
    m_sync.waitDrawEnd();
    frames++;
  }
  
  DI_trace2("%s # %u frames", fname, frames);  
  
  /* Now try to make a clean exit: */

  for(i=0; i < threads.size(); i++) {
    VRS_RenderThread * thread = threads[i];
    thread->quit();
  }

  // Release the threads:
  m_sync.waitSwapEnd();

  // Collect the garbage:
  for(i=0; i < threads.size(); i++) {
    VRS_RenderThread * thread = threads[i];
    thread->waitEnd();
    delete thread;
  }

  DI_trace2("%s # EXIT", fname);  
}

void VRS_GlApp::contextInit(VRS_Window *)
{}

void VRS_GlApp::contextCleanup(VRS_Window *)
{}

void VRS_GlApp::draw(VRS_Window *, int)
{}
  
void VRS_GlApp::preFrame()
{}

void VRS_GlApp::setHeadPosition()
{
  m_cave.setHeadPosition(Vector3(0, 0, 0), Vector3(1, 0, 0));
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


VRS_FluidGlApp::VRS_FluidGlApp()
  : m_tracker(0),
    m_headSensor(0)
{}

VRS_FluidGlApp::~VRS_FluidGlApp()
{}

bool VRS_FluidGlApp::exec(bool threaded)
{
  const char * fname = "VRS_FluidGlApp::exec";

  m_tracker = InputDevices::getTracker("tracker");

  if(!m_tracker) {
    DI_error2(DI_ERR_UNINITIALIZED, "%s # No tracker", fname);
    return false;
  }
  
  m_headSensor = m_tracker->findSensor("head");

  if(m_headSensor < 0) {
    DI_error2(DI_ERR_UNINITIALIZED, "%s # No head sensor", fname);
    return false;
  }

  return VRS_GlApp::exec(threaded);
}
  
void VRS_FluidGlApp::setHeadPosition()
{
  InputDevices::update();
  TrackerSample & sample = m_tracker->outputBuffer().getNewest();
  
  if(m_headSensor < sample.getSensorCount()) {
    const Matrix3 & r = sample.getRotation(m_headSensor);
    const Vector3 & l = sample.getLocation(m_headSensor);
    Vector3 right = r * Vector3(0, 0, 1);
    cave().setHeadPosition(l, right);
    // m_cave.setHeadPosition(Vector3(0, 0, 0), Vector3(1, 0, 1));
  }
  else {
    DI_trace2("VRS_FluidGlApp::setHeadPosition # "
	      "head sensor %d not present in tracker sample", 
	      m_headSensor);
  }
}
