/* COPYRIGHT
 *
 * This file is part of the Geometry Engine - GEE
 *
 * Read the "GEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen 2003
 *
 * firstname.lastname@hut.fi
 *
 */

#ifndef _VRSMUGGLER_HPP_
#define _VRSMUGGLER_HPP_

#include <CaveProj.hpp>
#include <glx_wrapper.h>

#include <di_thread.h>
#include <di_condition.h>

#include <GL/gl.h>

#include <vector>

#include <strings.h>

namespace Fluid {
  class Tracker;
}

class VRS_GlApp;

class VRS_Sync
{
public:
  VRS_Sync();
  ~VRS_Sync();

  void oneWasDrawn();
  void oneWasSwapped();

  void waitDrawEnd();
  void waitSwapEnd();
  void waitSwaps();

  void setThreadNum(uint threads) { m_threads = threads; }

protected:

  volatile uint m_threads;
  volatile uint m_drawn;
  volatile uint m_drawnCleared;
  volatile uint m_swapped;
  volatile uint m_swappedCleared;

  volatile bool m_allowDrawExit;
  volatile bool m_allowSwapExit;

  DI_MutexAuto m_mutex;
  DI_Condition m_cond;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


class VRS_Window
{
public:

  VRS_Window() : m_wall(0), m_glx(GLX_Create()) { }
  ~VRS_Window() {  GLX_Free(m_glx); }

  CaveProj::Wall * wall() { return m_wall; }


  CaveProj::Wall * m_wall;

  GLX  * m_glx;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


class VRS_RenderThread : public DI_Thread
{
public:
  VRS_RenderThread(VRS_Sync *, VRS_GlApp *);
  virtual ~VRS_RenderThread();

  void addWindow(VRS_Window * w) { m_windows.push_back(w); }

  void quit() { m_run = false; }

protected:

  virtual void childLoop();

  void renderWithMatrix(GLenum buffer, int eye,
			VRS_Window *window);

  void renderDemo(VRS_Window *window);
  
  std::vector<VRS_Window *> m_windows;

  VRS_Sync  * m_sync;
  VRS_GlApp * m_app;

  bool m_run;
};


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

class VRS_GlApp : public DI_Thread
{
public:
  friend class VRS_RenderThread;

  VRS_GlApp();
  virtual ~VRS_GlApp();

  bool loadVrConfig(const char * filename);
  virtual bool exec(bool threaded = true);
  virtual bool quit(bool inThread = false);
  void escPressed() { quit(true); }

  CaveProj::Cave & cave() { return m_cave; }

  uint windowCount() { return m_cave.wallCount(); }

protected:

  virtual void childLoop();

  /// Should be overridden if you want to initialize things
  virtual void contextInit(VRS_Window * context);

  /// Should be overridden if you need to clean up things
  virtual void contextCleanup(VRS_Window * context);

  /// Should be overridden to draw something. 
  virtual void draw(VRS_Window * context, int eye);

  /// Should be overridden to do per-frame calculations
  virtual void preFrame();

  /// Should be overridden to define the head position
  virtual void setHeadPosition();

private:
  VRS_Sync      m_sync;

  CaveProj::Cave m_cave;

  bool m_run;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

class VRS_FluidGlApp : public VRS_GlApp
{
public:
  VRS_FluidGlApp();
  virtual ~VRS_FluidGlApp();

  virtual bool exec(bool threaded = true);
  
protected:

  virtual void setHeadPosition();

private:
  Fluid::Tracker * m_tracker;
  int              m_headSensor;
};



#endif
