/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */


#ifndef AC_ATTENUATION_H
#define AC_ATTENUATION_H

/** Base class for calculating distance-based attenuation. The child
    classes implement all kinds of typical attenuation functions that
    might be useful in VR / AR / MR / *R systems.*/

class AC_Attenuation
{
public:
  AC_Attenuation() {}
  virtual ~AC_Attenuation() {}

  virtual float metersToGain(float meters) = 0;

  virtual AC_Attenuation * clone() const = 0;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** Distance attenuation based on division by polynomial. 

    Given distance d:

<PRE>
                1.0
tmp =  ----------------------
        d * c1 + d ^ 2 * c2

gain = limit(tmp, mingain, maxgain)

</PRE>

    You can set the coefficients c1 and c2. You can also set distance
    limits after which the sound will not fade or get louder any more.

    Default values:

    <UL>

    <LI> c1 = 0
    <LI> c2 = 1
    <LI> mindist = 0.5
    <LI> maxdist = 50.0

    </UL>

*/

class AC_AttenuationPolynomial : public AC_Attenuation
{
public:
  AC_AttenuationPolynomial();
  virtual ~AC_AttenuationPolynomial();

  virtual float metersToGain(float meters);
  virtual AC_Attenuation * clone() const;

  void setMinMaxDist(float minDist, float maxDist)
  { m_minDist = minDist; m_maxDist = maxDist; }
  void setCoefficients(float c1, float c2)
  { m_c1 = c1; m_c2 = c2; }

  float maxDist() const { return m_maxDist; }
  float minDist() const { return m_minDist; }
  float c1() const { return m_c1; }
  float c2() const { return m_c2; }

protected:
  float m_maxDist;
  float m_minDist;
  float m_c1, m_c2;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** Attenuation by given power. 

    Given distance d:

<PRE>
          1.0
gain =  -------
         d ^ p

</PRE>

    You can set the power p.
    
*/

class AC_AttenuationPower : public AC_Attenuation
{
public:
  AC_AttenuationPower(float power = 2.0);
  virtual ~AC_AttenuationPower();

  virtual float metersToGain(float meters);
  virtual AC_Attenuation * clone() const;

  void setMinMaxDist(float minDist, float maxDist)
  { m_minDist = minDist; m_maxDist = maxDist; }

  /** Sets the power. Default is 2.0. Useful range is 0.1 - 3.0. */
  void setPower(float p) { m_power = p; }

protected:
  float m_maxDist;
  float m_minDist;

  float m_power;
};

#endif
