/* -*- C++ -*- */

//
// DIVA Library. Copyright (C) Jarmo Hiipakka & Tommi Ilmonen, 1999-2000.
// 
// AC_MT_START and AC_MT_STOP added 11.07.2001, mtg
//
// $Log: ac_buffer.h,v $
// Revision 1.9  2003/06/13 10:10:52  tilmonen
// Added contiguous message transfers.
//
// Revision 1.8  2002/06/05 10:08:50  tilmonen
// Updates.
//
// Revision 1.7  2002/04/17 11:57:33  tilmonen
// Added distande and distance attenuation to the panoration system.
//
// Revision 1.6  2002/03/12 10:20:05  tilmonen
// Cosmetic clean-ups and minor updates.
//
// Revision 1.5  2002/02/18 14:54:16  tilmonen
// Added a possibility to control the gain and pitch of an active note.
//
// Revision 1.4  2001/12/21 15:44:08  tilmonen
// Additions for the pitched audio sample voices.
//
// Revision 1.3  2001/10/30 12:12:20  tilmonen
// Updates for better sound synthesis control.
//
// Revision 1.2  2001/07/27 12:57:37  tilmonen
// Removed dependencies to divabase (and therefore pthread).
//
// Revision 1.1  2001/07/26 08:57:48  tilmonen
// New auralization control toolkit.
//

#ifndef AC_BUFFER_H
#define AC_BUFFER_H

#include "di_data_buffer.h"

/// Auralization: source fade types.
enum AC_FadeType {
  AC_FT_LINEAR = 0
};

enum AC_PanningMethod {
  AC_PAN_UNKNOWN,
  AC_PAN_VBAP,
  AC_PAN_HRTF,
  AC_PAN_METHODS
};

/// Auralization control message types. 
enum AC_MsgType {
  /** Play a sample, arguments: sample name (string), gain, azimuth,
      elevation, (floats).*/
  AC_MT_PLAY_SAMPLE,
  /** Play an input channel: intput channel (int), gain, azimuth,
      elevation, (floats). */
  AC_MT_PLAY_INPUT,
  /** Note off command, arguments: note id (int) */
  AC_MT_NOTE_OFF,
  /** Change the source location, arguments:interpolation time
      (seconds), azimuth, elevation, distance, gain (floats). */
  AC_MT_CHANGE_SOURCE_LOCATION,
  /** Change the source orientation, arguments: interpolation time
      (seconds), azimuth and elevation (floats). */
  AC_MT_CHANGE_SOURCE_ORIENTATION,
  /// Set material, argument: material name (char string).
  AC_MT_SET_MATERIAL,
  /// Set the source channel, argument: channel ID (int32).
  AC_MT_SET_CHANNEL,
  /** Fade the source in or out, arguments: 0 (off) or 1 (on) (int32),
      fading time in ms (float), and fading type (int32, see AC_FadeType) */
  AC_MT_FADE,
  /** Request for an acknowledgment message, argument: isok (int32,
      optional). */
  AC_MT_IS_OK,
  /// Close the client connection.
  AC_MT_CLOSE,
  /// Start sound 
  AC_MT_START,
  /// Start sound 
  AC_MT_STARTED,
  /// Stop sound
  AC_MT_STOP,
  /// Stop sound
  AC_MT_STOPPED,
  /** Play a sample, arguments: sample name (string), gain, pitch, azimuth,
      elevation, (floats).*/
  AC_MT_PLAY_SAMPLE_PITCHED,
  /** Sets the gain of a note. Arguments: gain (float). */
  AC_MT_NOTE_GAIN,
  /** Sets the relative pitch of the note. */
  AC_MT_NOTE_RELATIVE_PITCH,
  /** Query panning method. */
  AC_MT_QUERY_PANNING,
  /** Return panning method. Argyments: int (AC_Control::PanningMethod). */
  AC_MT_PANNING,

  /** Mustajuuri message. Arguments: A message address (string), and
      MJ message. */
  AC_MT_MJ_MSG,
  /** Begin a contiguous message group. */
  AC_MT_BEGIN_CONTIGUOUS,
  /** Mustajuuri message. Arguments: A message address (string), and
      MJ message. */
  AC_MT_END_CONTIGUOUS,
  
  /// The number of auralization control message types.
  AC_MT_SIZEOF
};

/// Character strings that describe message types
extern const char * AC_MsgTypeKeys[AC_MT_SIZEOF];

class AC_BufferHeader
{
public:
  
  void clear()
  { 
    m_timeStamp = 0; m_msgId = 0; 
    m_type = AC_MT_CHANGE_SOURCE_LOCATION; m_sourceId = 0; 
  }

  void setTimeStamp (long long timeStamp) { m_timeStamp = timeStamp; }
  void setMsgId(int id) { m_msgId = id; }
  void setType(AC_MsgType type) { m_type = type; }
  void setSourceId(int id) { m_sourceId = id; }
  
  long long  timeStamp() const { return m_timeStamp; }
  int        msgId() const { return m_msgId; }
  AC_MsgType type() const { return m_type; }
  int        sourceId() const { return m_sourceId; }

  long long  * timeStampPtr() { return & m_timeStamp; }
  int        * msgIdPtr() { return & m_msgId; }
  AC_MsgType * typePtr() { return & m_type; }
  int        * sourceIdPtr() { return & m_sourceId; }

protected:

  long long  m_timeStamp;
  int        m_msgId;
  AC_MsgType m_type;
  int        m_sourceId;
};

/** Buffer with AC-header read/write capabilities.

    The header has the following fields:
    
    <OL>
    <LI> Time stamp (64-bit unsigned int)
    <LI> Message ID (32-bit int)
    <LI> Message type (32-bit int) (see AC_MsgType)
    <LI> Source ID (32-bit int)
    </OL>
 */
class AC_Buffer : public DI_DataBuffer2
{
public:
  AC_Buffer() {}
  ~AC_Buffer() {}

  /** Writes the header and updates the write access pointer. */
  void writeHeader(long long timestamp, int msgID, AC_MsgType type,
		   int sourceID);

  void writeHeader(const AC_BufferHeader &h)
  { writeHeader(h.timeStamp(), h.msgId(), h.type(), h.sourceId()); }
  
  /** Reads the header and updates the write access pointer. */
  bool readHeader(long long *timestamp, int *msgID, AC_MsgType *type,
		  int *sourceID);

  bool readHeader(AC_BufferHeader *h)
  { return readHeader(h->timeStampPtr(), h->msgIdPtr(), 
		      h->typePtr(), h->sourceIdPtr()); }
  
  /// Length of the header section
  inline unsigned int headerLength() const { return 20; }

  // Length of the message data section
  // unsigned int msgLength() const { return bytesUsed()-headerLength(); }
  
};

#endif
