
/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <ac_control.h>

#include <mj_message_multi.h>
#include <mj_messages.h>

#include <sys/types.h>
#include <unistd.h>

AC_Control::AC_Control()
  : m_state(AC_STATE_UNINITIALIZED),
    m_error(AC_ERR_NONE)
{}

/** Shut down the controller. */
AC_Control::~AC_Control()
{
  shutdown();
}

/** Initialize the auralization controller. 

    @arg address The network address for TCP/IP socket connection.

    @return This method returns true if the sound server was
    successfully contacted, otherwise false is returned.
*/

bool AC_Control::init(const char * address)
{
  const char * fname = "AC_Control::init";

  m_error = AC_ERR_NONE;

  AC_BufferHeader h;
  h.clear();

  if(m_state != AC_STATE_UNINITIALIZED) {
    AC_error(AC_ERR_RC_DENIED, fname, "Could not open socket");
    m_error = AC_ERR_RC_DENIED;
    return false;
  }
  
  bool ok = m_socket.open(address, false, true);

  if(!ok) {
    AC_error(AC_ERR_ARGUMENT, fname, "Could not open socket connection");
    return false;
  }

  AC_trace(fname, "Socket opened");

  h.setType(AC_MT_START);

  m_buffer.writeHeader(h);
  
  if(!DI_writeToSocket(&m_socket, &m_buffer)) {
    AC_error(AC_ERR_RC_DENIED, fname, "Write to socket failed");
    m_error = AC_ERR_RC_DENIED;
    return false;
  }

  DI_readFromSocket(&m_socket, &m_buffer);
  m_buffer.readHeader(&h);

  if(h.type() != AC_MT_STARTED) {
    AC_error(AC_ERR_RC_DENIED, fname, "Could not start");
    m_error = AC_ERR_RC_DENIED;
    return false;
  }

  m_state = AC_STATE_RUNNING;

  m_idCount = 0;
  
  return true;
}

/** Spans a server with given script. */

void AC_Control::spanServer(const char * script)
{
  pid_t pid = fork();

  if(!pid) {
    system(script);
    exit(0);
  }
}

/** Shut the server down. */
bool AC_Control::shutdown()
{
  const char * fname = "AC_Control::shutdown";

  AC_trace(fname);

  if(m_state != AC_STATE_RUNNING) {
    return false;
  }
  
  AC_BufferHeader h;
  h.clear();

  h.setType(AC_MT_STOP);

  m_buffer.writeHeader(0, 0, AC_MT_STOP, 0);
  DI_writeToSocket(&m_socket, &m_buffer);
  
  DI_readFromSocket(&m_socket, &m_buffer);
  m_buffer.readHeader(&h);

  m_socket.close();

  if(h.type() != AC_MT_STOPPED) {
    AC_error(AC_ERR_RC_DENIED, fname, "No AC_MT_STOPPED reply");
    m_error = AC_ERR_RC_DENIED;
    return false;
  }

  m_state = AC_STATE_UNINITIALIZED;
  
  return true;  
}

/** Tells the system to delay the application of messages until
    "endContiguous" command is issued. In practice all messages are
    transmitted as usual, but the sound server collects them until it
    is instructed to apply them. 

    The server will react normally to status queries etc. even while
    contiguous mode is on.

    @see AC_Control::beginContiguous
*/

void AC_Control::beginContiguous()
{
  if(m_state != AC_STATE_RUNNING) return;
  
  m_buffer.writeHeader(0, 0, AC_MT_BEGIN_CONTIGUOUS, 0);
  DI_writeToSocket(&m_socket, &m_buffer);
}

/** Ends a contiguous message segment.

    @see AC_Control::beginContiguous
*/

void AC_Control::endContiguous()
{
  if(m_state != AC_STATE_RUNNING) return;
  
  m_buffer.writeHeader(0, 0, AC_MT_END_CONTIGUOUS, 0);
  DI_writeToSocket(&m_socket, &m_buffer);
}

/** Determines sound panning method. */

AC_PanningMethod AC_Control::panningMethod()
{
  const char * fname = "AC_Control::panningMethod";

  if(m_state != AC_STATE_RUNNING) return AC_PAN_UNKNOWN;

  
  AC_BufferHeader h;
  h.clear();

  m_buffer.writeHeader(0, 0, AC_MT_QUERY_PANNING, 0);
  DI_writeToSocket(&m_socket, &m_buffer);
  
  DI_readFromSocket(&m_socket, &m_buffer);
  m_buffer.readHeader(&h);

  if(h.type() != AC_MT_PANNING) {
    AC_error(AC_ERR_UNKNOWN, fname, "h.type() != AC_MT_PANNING");
    return AC_PAN_UNKNOWN;
  }
  
  return (AC_PanningMethod) m_buffer.readInt32();
}

/** Play a sample.

    @arg sample Name of the sample to be played
    @arg gain Linear gain (0-1)
    @arg azi Source azimuth in degrees
    @arg ele Source elevation in degrees
    @arg loop The looping of the voice
    @arg sourceId Id of this voice

    @return This method returns the <B>note id</B> of the new note. If
    the connection was broken, then -1 is returned.
 */
long AC_Control::playSample
(int sourceId, const char *sample, float gain, 
 float azi, float ele, float dist, float dgain, bool loop, 
 float skipSeconds, float interleaveSeconds)
{
  if(m_state != AC_STATE_RUNNING) return -1;

  m_buffer.writeHeader(0, 0, AC_MT_PLAY_SAMPLE, sourceId);
  m_buffer.writeString(sample);
  m_buffer.writeFloat32(gain);
  m_buffer.writeFloat32(azi);
  m_buffer.writeFloat32(ele);
  m_buffer.writeFloat32(dist);
  m_buffer.writeFloat32(dgain);
  m_buffer.writeInt32((int) loop);
  m_buffer.writeFloat32(skipSeconds);
  m_buffer.writeFloat32(interleaveSeconds);
  DI_writeToSocket(&m_socket, &m_buffer);

  return m_idCount++;
}

/** Play a sample at given pitch.

    @arg sample Name of the sample to be played
    @arg gain Linear gain (0-1)
    @arg Pitch Linear pitch (0-inf)
    @arg azi Source azimuth in degrees
    @arg ele Source elevation in degrees
    @arg loop The looping of the voice
    @arg sourceId Id of this voice

    @return This method returns the <B>note id</B> of the new note. If
    the connection was broken, then -1 is returned.
 */
long AC_Control::playSamplePitched
(int sourceId, const char *sample, float gain, float pitch, 
 float azi, float ele, float dist, float dgain, bool loop, float skipSeconds)
{
  if(m_state != AC_STATE_RUNNING) return -1;

  m_buffer.writeHeader(0, 0, AC_MT_PLAY_SAMPLE_PITCHED, sourceId);
  m_buffer.writeString(sample);
  m_buffer.writeFloat32(gain);
  m_buffer.writeFloat32(pitch);  
  m_buffer.writeFloat32(azi);
  m_buffer.writeFloat32(ele);
  m_buffer.writeFloat32(dist);
  m_buffer.writeFloat32(dgain);
  m_buffer.writeInt32((int) loop);
  m_buffer.writeFloat32(skipSeconds);
  DI_writeToSocket(&m_socket, &m_buffer);

  return m_idCount++;
}

/** Play sound from one of synthesizer's input channels. */

long AC_Control::playInput(int sourceId, int inputChannel, 
			   float gain, float azi, float ele, 
			   float dist, float dgain)
{
  if(m_state != AC_STATE_RUNNING) return -1;
  
  m_buffer.writeHeader(0, 0, AC_MT_PLAY_INPUT, sourceId);

  m_buffer.writeInt32(inputChannel);
  m_buffer.writeFloat32(gain);
  m_buffer.writeFloat32(azi);
  m_buffer.writeFloat32(ele);  
  m_buffer.writeFloat32(dist);
  m_buffer.writeFloat32(dgain);
  
  DI_writeToSocket(&m_socket, &m_buffer);

  return m_idCount++;
}

/** Plays granular voice. */
long AC_Control::playGranular(int sourceId, const char * sample, float gain)
{
  MJ_MessageMulti mm;
  mm.setAutoDelete(false);

  MJ_MessageStlString msample(sample);
  MJ_MessageStlString mtype("granular");
  MJ_MessageDouble mgain(gain);
  MJ_MessageDouble mrate(199.0);
  MJ_MessageDouble mpitch(445.5);
  MJ_MessageInt mchan(sourceId);
  
  mm.addChild("sample", &msample);
  mm.addChild("voicetype", &mtype);
  mm.addChild("gain", &mgain);
  mm.addChild("rate", &mrate);
  mm.addChild("pitch", &mpitch);
  mm.addChild("channel", &mchan);
  
  return noteOn(&mm);
}

long AC_Control::playNoise(int sourceId, float gain, float hz, float octaves)
{
  MJ_MessageMulti mm;
  mm.setAutoDelete(false);

  MJ_MessageStlString mtype("noise");

  MJ_MessageDouble mgain(gain);
  MJ_MessageDouble mhz(hz);
  MJ_MessageDouble moctaves(octaves);
  MJ_MessageInt mchan(sourceId);
  
  mm.addChild("voicetype", &mtype);
  mm.addChild("gain", &mgain);
  mm.addChild("pitch", &mhz);
  mm.addChild("octaves", &moctaves);
  mm.addChild("channel", &mchan);

  return noteOn(&mm);
}

/** Plays a single "shake" on a shaker. */ 

long AC_Control::playShaker(int sourceId, int type, float amplitude)
{
  MJ_MessageMulti mm;
  mm.setAutoDelete(false);

  MJ_MessageStlString mtype("shakers-stk");
  MJ_MessageDouble mampl(amplitude);
  MJ_MessageInt minstrument(type);
  MJ_MessageInt mchan(sourceId);
  
  mm.addChild("amplitude", &mampl);
  mm.addChild("instrument", &minstrument);
  mm.addChild("channel", &mchan);
  mm.addChild("voicetype", &mtype);
  
  return noteOn(&mm);
}

/** Plays a lasting shake on a shaker. The note plays until it is
    silenced with noteoff. */
long AC_Control::playShaker2
(int sourceId, int type, float amplitude, float rate)
{
  MJ_MessageMulti mm;
  mm.setAutoDelete(false);

  MJ_MessageStlString mtype("shakers2-stk");
  MJ_MessageDouble mampl(amplitude);
  MJ_MessageDouble mrate(rate);
  MJ_MessageInt minstrument(type);
  MJ_MessageInt mchan(sourceId);
  
  mm.addChild("amplitude", &mampl);
  mm.addChild("instrument", &minstrument);
  mm.addChild("channel", &mchan);
  mm.addChild("voicetype", &mtype);
  mm.addChild("rate", &mrate);
  
  return noteOn(&mm);
  
}

/** Sends a note-on message to the synth. You can use this to send any
    kind of note-on message. Most of the "play*" methods use this as a
    back-end for sending the control message to the synth engine. */  
long AC_Control::noteOn(MJ_Message *m)
{
  if(m_state != AC_STATE_RUNNING) return -1;

  const char * address = "$(SYNTH)/noteon";
  sendMjMessage(address, m);
  
  return m_idCount++;  
}

/* long AC_Control::noteOn(int sourceId, MJ_MessageMulti *mm)
{
  if(m_state != AC_STATE_RUNNING) return -1;

  const char * address = "$(SYNTH)/noteon";

  
  
  sendMjMessage(address, m);

  return m_idCount++;  
  }*/

/** Change the orientation of a source. */
void AC_Control::changeSourceOrientation
(int sourceId, float interpolationTimeSeconds, float azi, float ele)
{
  if(m_state != AC_STATE_RUNNING) return;

  m_buffer.writeHeader(0, 0, AC_MT_CHANGE_SOURCE_ORIENTATION, sourceId);
  m_buffer.writeFloat32(interpolationTimeSeconds);
  m_buffer.writeFloat32(azi);
  m_buffer.writeFloat32(ele);

  DI_writeToSocket(&m_socket, &m_buffer);
}

void AC_Control::changeSourceLocation
(int sourceId, float interpolationTimeSeconds, 
 float azi, float ele, float dist, float gain)
{
  if(m_state != AC_STATE_RUNNING) return;

  m_buffer.writeHeader(0, 0, AC_MT_CHANGE_SOURCE_LOCATION, sourceId);
  m_buffer.writeFloat32(interpolationTimeSeconds);
  m_buffer.writeFloat32(azi);
  m_buffer.writeFloat32(ele);
  m_buffer.writeFloat32(dist);
  m_buffer.writeFloat32(gain);

  DI_writeToSocket(&m_socket, &m_buffer);
}
 
/** Sends a note-off -command to a particular note. */

void AC_Control::noteOff(long noteId)
{
  if(m_state != AC_STATE_RUNNING) return;

  m_buffer.writeHeader(0, 0, AC_MT_NOTE_OFF, noteId);

  DI_writeToSocket(&m_socket, &m_buffer);
}

void AC_Control::allNotesOff(bool forceNow)
{
  MJ_MessageInt mi(forceNow);

  sendMjMessage("$(SYNTH)/allnotesoff", & mi);
}


void AC_Control::setNoteGain(long noteId, float gain)
{
  if(m_state != AC_STATE_RUNNING) return;

  m_buffer.writeHeader(0, 0, AC_MT_NOTE_GAIN, -1);
  m_buffer.writeInt64(noteId);
  m_buffer.writeFloat32(gain);
  
  DI_writeToSocket(&m_socket, &m_buffer);
}

void AC_Control::setNoteRelativePitch(long noteId, float relativePitch)
{
  if(m_state != AC_STATE_RUNNING) return;

  m_buffer.writeHeader(0, 0, AC_MT_NOTE_RELATIVE_PITCH, -1);
  m_buffer.writeInt64(noteId);
  m_buffer.writeFloat32(relativePitch);
  
  DI_writeToSocket(&m_socket, &m_buffer);
}

void AC_Control::setNoteParameter
(long noteId, const char *parameter, float value)
{
  if(noteId < 0) {
    AC_error(AC_ERR_ARGUMENT, 
	     "AC_Control::setNoteParameter # noteId must be positive");
    return;
  }

  char buf[128];
  sprintf(buf, "$(SYNTH)/note/%ld/%s", noteId, parameter);
  m_buffer.writeString(buf);

  MJ_MessageDouble md(value);
  md.write2(&m_buffer);
  
  sendMjMessage(buf, &md);
}

/** Sets a parameter for an STK note. */
void AC_Control::setStkNoteParameter(long noteId, int param, float value)
{
  if(noteId < 0) {
    AC_error(AC_ERR_ARGUMENT, 
	     "AC_Control::setStkNoteParameter # noteId must be positive");
    return;
  }
  
  char buf[128];
  sprintf(buf, "$(SYNTH)/note/%ld/controller_%d", noteId, param);
  m_buffer.writeString(buf);

  MJ_MessageDouble md(value);
  md.write2(&m_buffer);
  
  sendMjMessage(buf, &md);
}

void AC_Control::sendMjMessage(const char *address, MJ_Message *m)
{
  if(m_state != AC_STATE_RUNNING) return;

  m_buffer.writeHeader(0, 0, AC_MT_MJ_MSG, 0);  
  m_buffer.writeString(address);
  m->write2(&m_buffer);

  DI_writeToSocket(&m_socket, &m_buffer);
}

/** Sends a message to control any parameter of a particular note. */
void AC_Control::sendMjNoteControl
(long noteId, const char * parameter, MJ_Message *m)
{
  // AC_trace2("AC_Control::sendMjNoteControl # %p %p", this, m);

  if(m_state != AC_STATE_RUNNING) return;

  if(noteId < 0) {
    AC_error(AC_ERR_ARGUMENT, 
	     "AC_Control::sendMjNoteControl # noteId must be positive");
    return;
  }

  // AC_trace("AC_Control::sendMjNoteControl # init");

  m_buffer.writeHeader(0, 0, AC_MT_MJ_MSG, 0);  

  char buf[256];
  sprintf(buf, "$(SYNTH)/note/%ld/%s", noteId, parameter);
  m_buffer.writeString(buf);

  // AC_trace("AC_Control::sendMjNoteControl # half");

  m->write2(&m_buffer);

  // AC_trace("AC_Control::sendMjNoteControl # Nearly done");

  DI_writeToSocket(&m_socket, &m_buffer);
}

/** Sends a message to control any parameter of a particular note with
    a floating point value. */
void AC_Control::sendMjNoteControl
(long noteId, const char * parameter, double value)
{
  MJ_MessageDouble md(value);
  sendMjNoteControl(noteId, parameter, &md);
}

/** Check that the server is still ok. */
bool AC_Control::serverOk()
{
  const char * fname = "AC_Control::serverOk";

  if(m_state != AC_STATE_RUNNING) return false;

  AC_BufferHeader h;
  h.clear();

  m_buffer.writeHeader(0, 0, AC_MT_IS_OK, 0);
  DI_writeToSocket(&m_socket, &m_buffer);
  
  DI_readFromSocket(&m_socket, &m_buffer);
  m_buffer.readHeader(&h);

  if(h.type() != AC_MT_IS_OK) {
    AC_error(AC_ERR_UNKNOWN, fname, "h.type() != AC_MT_IS_OK");
    return false;
  }
  
  return m_buffer.readInt32() == 1;
}
