/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef AC_CONTROL_H
#define AC_CONTROL_H

#include <ac_buffer.h>
#include <ac_error.h>

#include <di_socket.h>

class MJ_Message;

/** State of an AC_Control object. */

enum AC_State {
  AC_STATE_UNINITIALIZED,
  AC_STATE_RUNNING,
  AC_STATE_CLOSING,
  AC_STATE_SIZEOF
};

/** Auralization control class. This class is meant to be used in
    client applications that require auralization services. 

    The auralization engine typically runs in a separate computer. The
    server is controlled via a network. This class wraps the network
    protocols behind a simple C++ API.

    This class aims to be as robust as possible. Therefore you should
    not be able to cause core dumps with this class, no matter what
    you do. Please report bugs to Tommi.Ilmonen@hut.fi.

    <B>The business with identifiers:</B> There are two kinds of
    identifiers:

    <OL>

    <LI> Source identifiers - this corresponds to a particular voice
    source. You can select the identifiers freely (but you must have
    the backend configured to match the number of sources). You can
    attach any number of notes to one source.

    <LI> Note identifiers. Note identifiers identify a particular
    note. A source may play multiple voices at the same time. Note
    identifiers are generated as you start a note. You cannot select
    the note identifiers yourself. Instead you must store the
    identifiers that a returned by the note-initializer functions.

    </OL>

    It is necessary to use the correct identifier when controlling the
    sound parameters. For example you can attach several notes to
    particular source. As you move the source in space all the notes
    will move automatically. The notes in turn can and must be
    controlled individually.

    @author Tommi Ilmonen
*/
class AC_Control
{
public:

  AC_Control();
  ~AC_Control();
  
  bool init(const char * address);
  void spanServer(const char * script = "ssh a2 './bin/vr-audio-server'");
  bool shutdown();

  void beginContiguous();
  void endContiguous();

  AC_PanningMethod panningMethod();

  long playSample(int sourceId, const char *sample, float gain, 
		  float azi, float ele, float dist, float dgain, bool loop,
		  float skipSeconds = 0.0, float interleaveSeconds = 0.0);
  long playSamplePitched(int sourceId, const char *sample, float gain, 
			 float pitch, float azi, float ele, 
			 float dist, float dgain, bool loop,
			 float skipSeconds = 0.0);
  long playInput(int sourceId, int inputChannel, 
		 float gain, float azi, float ele, float dist, float dgain);
  long playGranular(int sourceId, const char * sample, float gain);
  long playNoise(int sourceId, float gain, float hz, float octaves);

  long playShaker(int sourceId, int type, float amplitude);
  long playShaker2(int sourceId, int type, float amplitude, float rate = 10);
  
  long noteOn(MJ_Message *);

  void noteOff(long noteId);
  void allNotesOff(bool forceNow);

  void setNoteGain(long noteId, float gain);
  void setNoteRelativePitch(long noteId, float relativePitch);
  void setNoteParameter(long noteId, const char *parameter, float value);
  void setStkNoteParameter(long noteId, int param, float value);

  void changeSourceOrientation(int sourceId, float interpolationTimeSeconds, 
			       float azi, float ele);
  void changeSourceLocation(int sourceId, float interpolationTimeSeconds, 
			    float azi, float ele, float dist, float gain);

  void sendMjMessage(const char *address, MJ_Message *m);
  void sendMjNoteControl(long noteId, const char * param, MJ_Message *m);
  void sendMjNoteControl(long noteId, const char * param, double val);
  
  bool clientOk() const { return m_error == AC_ERR_NONE; }
  
  /// Check that the sound server is ok.
  bool serverOk();
  void flush() { m_socket.flush(); }

  AC_State clientState() { return m_state; }


protected:

  DI_Socket     m_socket;
  AC_Buffer     m_buffer;

  AC_State      m_state;
  AC_ErrorCode  m_error;

  long          m_idCount;

private:
  AC_Control(const AC_Control &) {}  
  AC_Control &operator = (const AC_Control &) { return *this; }
};

#endif
