/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */
#include "ac_data_buffer.h"
#include "ac_config.h"  // for WORDS_BIGENDIAN definition

#include <di_types.h>
#include <netinet/in.h> // for ntohl() etc.
#include <stdlib.h>
#include <stdio.h>
#include <string.h>


#ifndef WORDS_BIGENDIAN

static unsigned long long ntohll(unsigned long long x)
{
  return ((x << 56) | ((x & 0xff00ll) << 40) | ((x & 0xff0000ll) << 24) |
	  ((x & 0xff000000ll) << 8) | ((x >> 8) & 0xff000000ll) |
	  ((x >> 24) & 0xff0000ll) | ((x >> 40) & 0xff00ll) | (x >> 56));
}
#define htonll(x)    ntohll(x)

#else

#define ntohll(x)    (x)
#define htonll(x)    (x)

#endif  // not defined WORDS_BIGENDIAN

// -- Constructors ------------------------------------------------------------

AC_DataBuffer::AC_DataBuffer(unsigned int size) :
  m_buf(NULL),
  m_size(size+4), // There must be room for the length field.
  m_index(0)
{
  resize(m_size);
  writeInt32(0);
}

AC_DataBuffer::~AC_DataBuffer()
{
  if (m_buf) free(m_buf);
}

void AC_DataBuffer::writeInt32(int i)
{
  if(!writeNetworkOrder(&i, 4)) failIn("writeInt32");
}

void AC_DataBuffer::writeInt64(long long i)
{
  /* if(sizeof(long long) != 8) {
     puts("sizeof(long long) != 8");
     }*/

  if(!writeNetworkOrder(&i, 8)) failIn("writeInt64");
}

void AC_DataBuffer::writeFloat32(float i)
{
  if(!writeNetworkOrder(&i, 4)) failIn("writeFloat32");
}

void AC_DataBuffer::writeFloat64(double i)
{
  if(!writeNetworkOrder(&i, 8)) failIn("writeFloat64");
}

void AC_DataBuffer::writeString(const char *str)
{
  const char * fname = "writeString";

  int len = strlen(str) + 1; // Include ending NULL
  int rlen = len + (4 - len & 0x3); // Pad to four

  // Do not accept too long strings:
  if ((len >= AC_MAX_DATA_STRING) || (m_index + rlen > m_size)) {
    failIn(fname);
    return;
  }
  
  writeInt32(len);
  memcpy(& m_buf[m_index], str, len);

  m_index += rlen;
  *(unsigned int *)m_buf = m_index - 4;
}

int AC_DataBuffer::readInt32()
{
  int i;
  if(!readNetworkOrder(&i, 4)) failIn("readInt32");
  return i;
}

long long AC_DataBuffer::readInt64()
{
  long long i;
  if(!readNetworkOrder(&i, 8)) failIn("readInt64");
  return i;
}

float AC_DataBuffer::readFloat32()
{
  float i;
  if(!readNetworkOrder(&i, 4)) failIn("readFloat32");
  return i;
}

double AC_DataBuffer::readFloat64()
{
  double i;
  if(!readNetworkOrder(&i, 8)) failIn("readFloat64");
  return i;
}

void AC_DataBuffer::readString(char *str)
{
  int len = readInt32();
  int rlen = len + (4 - len & 0x3); // Pad to four

  if(m_index + rlen  > m_size)
  {
    failIn("readString");
    return;
  }

  memcpy(str, & m_buf[m_index], len);

  m_index += rlen;
}

bool AC_DataBuffer::resize(uint s)
{
  m_buf = (char *) realloc(m_buf,s);

  if(!s)
  {
    m_buf = 0;
    m_size = 0;
  }

  if (m_buf)
  {
    m_size = s;
    return true;
  }
  else
    return false;
}

bool AC_DataBuffer::writeNetworkOrder(const void * data, uint bytes)
{
  // Pad to eight if necessary
  if (!(bytes & 0x7) && (m_index & 0x7))
    m_index += 4;

  if (m_index + bytes > m_size) return false;

  switch(bytes)
  {
   case 4:
    *(u_long *)(m_buf+m_index) = htonl(*(u_long *)data);
    break;
   case 8:
    *(unsigned long long *)(m_buf+m_index) =
      htonll(*(unsigned long long *)data);
    break;
   default:
    return false;
  }
  
  m_index += bytes;
  *(unsigned int *)m_buf = htonl(m_index - 4);
  return true;
}

bool AC_DataBuffer::readNetworkOrder(void * data, uint bytes)
{
  // Pad to eight if necessary
  if(!(bytes & 0x7) && (m_index & 0x7))
    m_index += 4;

  if(m_index + bytes > m_size) return false;

  switch(bytes)
  {
   case 4:
    *(u_long *)data = ntohl(*(u_long *)(m_buf+m_index));
    break;
   case 8:
    *(unsigned long long*)data =
      ntohll(*(unsigned long long *)(m_buf+m_index));
    break;
   default:
    return false;
  }

  m_index += bytes;
  return true;
}

/// Simple error reporting to stderr.
void AC_DataBuffer::failIn(const char *str)
{
  fprintf(stderr, "AC_DataBuffer::%s # error\n", str);
  fflush(0);
  int *tmp = 0;
  *tmp = 1000;
}
