/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

//
// DIVA Library. Copyright (C) Jarmo Hiipakka & Tommi Ilmonen, 1999-2001.
//
// $Log: ac_data_buffer.h,v $
// Revision 1.6  2004/11/30 13:51:43  tilmonen
// Copyright upodates...
//
// Revision 1.5  2002/05/22 12:41:11  tilmonen
// Minor changes.
//
// Revision 1.4  2001/12/05 14:19:25  tilmonen
// Updates.
//
// Revision 1.3  2001/10/30 12:12:20  tilmonen
// Updates for better sound synthesis control.
//
// Revision 1.2  2001/08/22 07:54:01  tilmonen
// Updates.
//
// Revision 1.1  2001/07/27 12:57:37  tilmonen
// Removed dependencies to divabase (and therefore pthread).
//
// Revision 1.4  2001/07/27 08:23:16  tilmonen
// License text updates (=cosmetic changes).
//
// Revision 1.3  2001/07/11 09:35:22  tilmonen
// Bug fixes.
//
// Revision 1.2  2001/07/09 14:41:31  tilmonen
// Important bug fixes (endian issue in ac_data_buffer).
//
// Revision 1.1.1.1  2001/06/04 09:04:23  tilmonen
// Mustajuuri 4.6.2001, initial CVS cersion
//

#ifndef AC_DATA_BUFFER_H
#define AC_DATA_BUFFER_H

#include <di_endian.h>

#include <di_types.h>
#include <netinet/in.h>

/// The maximum length of an argument string.
const int AC_MAX_DATA_STRING = 512;


/** Data buffer base class for portable communication.

    This class knows how to read/write data from/to a character buffer
    in network byte order. The buffer contains the length of the data and
    the data itself. The length field is in the beginning of the buffer,
    and it is 4 bytes long. The write and read methods advance an internal
    pointer automatically; all you have to do to create a buffer full of
    data, is to construct the buffer and call the write functions in the
    desired order. Reading the buffer must be done in exactly the same
    order.

    You can send the buffer over, e.g., a socket connection, and the
    receiver will be able to read the data regardless of the byte orders
    of the different hosts.

    This class was designed to be independent of the rest of the 'libdibase'
    library (i.e., the basic library used by the DIVA team), so that it
    can be extracted from the library by copying the relevant files. For
    easy transportation over a socket connection, the 'ac_socket.h' header
    file contains helper functions.

    Authors: Jarmo Hiipakka & Tommi Ilmonen, HUT, 2000.
    
*/

class AC_DataBuffer
{
public:

  /**@name Constructors */
  
  //@{
  
  /// Creates a buffer with desired size.

  AC_DataBuffer(/// Data size (in bytes), excluding the length field.
		unsigned int size = 1024);

  virtual ~AC_DataBuffer();

  //@}
  
  /** @name Write methods. */

  //@{

  /// Write a two-byte integer.
  
  void writeInt16(short input) { writeInt32((int)input); }

  /// Write a four-byte integer.
  
  void writeInt32(int input);

  /// Write an eight-byte integer.
  
  void writeInt64(long long input);

  /// Write a four-byte floating point number.
  
  void writeFloat32(float input);

  /// Write an eight-byte floating point number.
  
  void writeFloat64(double input);

  /// Write a NULL-terminated character string (max length AC_MAX_DATA_STRING).
  
  void writeString(const char * input);

  //@}

  /** @name Read methods. */

  //@{

  /// Read a two-byte integer.
  
  short readInt16() { return (short)readInt32(); }

  /// Read a four-byte integer.
  
  int readInt32();

  /// Read an eight-byte integer.

  long long readInt64();

  /// Read a four-byte floating point number.

  float readFloat32();

  /// Read an eight-byte floating point number.

  double readFloat64();

  /// Read a string.
  
  void readString(char *string);

  //@}

  /** @name Access methods. */

  //@{

  /// Move read/write head.

  inline bool moveIndex(unsigned int i)
  { if(i >= m_size) return false; m_index = i; return true; }

  /// Get the position of the read/write head

  inline uint index() const { return m_index; }

  /** Resize the buffer.

      The buffer must be resized to sufficient length before write can
      take place. The contents of the buffer will be unchanged up to
      the lesser of the new and old sizes.
  */

  bool resize(unsigned int size);

  /// Return a pointer to the internal data buffer.

  char * buffer() const { return m_buf; }

  /// Return the size of the data buffer (in bytes).

  unsigned int size() const { return m_size-4; }

  /// Return the number of bytes actually in use.

  unsigned int bytesUsed() const { return DI_ntohl(*(unsigned int *)m_buf); }

  //@}

protected:

  /** Write bytes to the internal buffer with network byte-order
      conversion. */
  bool writeNetworkOrder(const void * data, unsigned int bytes);
  /** Read bytes from the internal buffer with network byte-order
      conversion. */
  bool readNetworkOrder(void * data, unsigned int bytes);

  static void failIn(const char *name);

  char         *m_buf;
  unsigned int  m_size;
  unsigned int  m_index;

private:
  /// Disabled constructor:
  AC_DataBuffer(const AC_DataBuffer &) {}
  /// Disabled copy operator:
  AC_DataBuffer & operator = (const AC_DataBuffer &) { return *this; }
};


#endif  // AC_DATA_BUFFER_H
