/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include "ac_error.h"

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <iostream>

const char *AC_ErrorLevels[AC_ERRL_SIZEOF] = {
  "Trace",
  "Information",
  "Warning",
  "Error",
  "Fatal error"
};

const char *AC_ErrorCodes[AC_ERR_SIZEOF] = {
  "No error",
  "Unkown error",
  "Out of memory",
  "Data corrupted",
  "Internal error",
  "System error",
  "Array size exceeded",
  "List size exceeded",
  "Bad argument",
  "NULL argument",
  "Bad match",
  "End of file",
  "Resource denied",
  "Virtual function called",
  "Do not call this function",
  "Not implemented yet",
  "Not initialized yet",
  "Real-time performance failure",
  "Floating point error",
  "CPU load too high",
  "Numeric overflow",
  "Input problem",
  "Output problem"
};


std::ostream *AC_traceFile = &std::cout;
bool     AC_errorsToTraceFile = false;

const char *AC_errorLevelStr(AC_ErrorLevel level)
{
  if(level < AC_ERRL_SIZEOF) return AC_ErrorLevels[level];
  return "Bad error level";
}

const char *AC_errorStr(AC_ErrorCode code)
{
  if(code < AC_ERR_SIZEOF) return AC_ErrorCodes[code];
  return "Bad error type";
}

std::ostream * AC_getErrorStream(AC_ErrorLevel level)
{
  std::ostream *stream = AC_traceFile;

  if(AC_errorsToTraceFile || level == AC_ERRL_TRACE) return stream;
  if((level == AC_ERRL_ERROR) || (level == AC_ERRL_FATAL))
    return & std::cerr;

  return &std::cout;
}

extern void AC_failure(AC_ErrorLevel level,
		       AC_ErrorCode  code, 
		       const char *c1, const char *c2, const char *c3)
{
  std::ostream *s = AC_getErrorStream(level);
  (*s) << AC_errorLevelStr(level);

  if(level != AC_ERRL_TRACE)
    (*s) <<  " # " << AC_errorStr(code);

  if(c1) (*s) << " # " << c1;
  if(c2) (*s) << " # " << c2;
  if(c3) (*s) << " # " << c3;
  (*s) << std::endl;
}

void AC_warning(AC_ErrorCode  code, const char *c1, const char *c2)
{
  AC_failure(AC_ERRL_ERROR, code, c1, c2);
}

void AC_error(AC_ErrorCode  code, const char *c1, const char *c2)
{
  AC_failure(AC_ERRL_ERROR, code, c1, c2);
}

void AC_error(AC_ErrorCode  code, const char *c1, 
	      const char *c2, 
	      const char *c3)
{
  AC_failure(AC_ERRL_ERROR, code, c1, c2, c3);
}

void AC_error2(AC_ErrorCode  code, const char *msg, ...)
{
  char buf[8196];
  va_list args;
  va_start(args, msg);
  vsprintf(buf, msg, args);
  AC_error(code, buf);
  va_end(args);
}

void AC_fatal(AC_ErrorCode  code, const char *c1, const char *c2)
{
  AC_failure(AC_ERRL_FATAL, code, c1, c2);
  exit(0);
}

void AC_trace(AC_ErrorCode  code, const char *c1, const char *c2)
{
  AC_failure(AC_ERRL_TRACE, code, c1, c2);
}

void AC_trace(const char *c1, const char *c2)
{
  AC_trace(AC_ERR_NONE, c1, c2);
}

void AC_trace(const char *c1, const char *c2, int i)
{
  char * tmp = (char *) malloc(strlen(c1) + strlen(c2) + 64);
  sprintf(tmp, "%s # %s # %d", c1, c2, i);
  AC_trace(AC_ERR_NONE, tmp);
  free(tmp);
}

void AC_trace(const char *c1, int i)
{
  char buf[128];
  
  sprintf(buf, "%d", i);
  AC_trace(c1, buf);
}

void AC_trace2(const char *msg, ...)
{
  char buf[8196];
  va_list args;
  va_start(args, msg);
  vsprintf(buf, msg, args);
  AC_trace(buf);
  va_end(args);
}
