/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef AC_MATRIX3_H
#define AC_MATRIX3_H

#include <ac_vector3.h>

class pfMatrix;

/** 3 x 3 matrix class. Much of the functionality is ripped from class
    Matrix3T by Janne Kontkanen. 

    @author Tommi Ilmonen
*/

class AC_Matrix3
{
public:
  AC_Matrix3() {}
  AC_Matrix3(const AC_Vector3& a, const AC_Vector3& b, const AC_Vector3& c) 
  { m[0] = a; m[1] = b; m[2] = c; }

  AC_Matrix3(float v11, float v12, float v13, 
	     float v21, float v22, float v23, 
	     float v31, float v32, float v33)
  { 
    m[0].make(v11, v12, v13); 
    m[1].make(v21, v22, v23); 
    m[2].make(v31, v32, v33); 
  }


  template <class S>
  AC_Matrix3(const S * p)
  {
    m[0].make(p[0], p[1], p[2]);
    m[1].make(p[3], p[4], p[5]);
    m[2].make(p[6], p[7], p[8]);
  }
  
  AC_Vector3&       row(int i)             { return m[i]; }
  const AC_Vector3& row(int i) const       { return m[i]; }
  AC_Vector3        column(int i) const    
  { return AC_Vector3(m[0][i],m[1][i],m[2][i]); }

  AC_Vector3&       operator[](int i)      { return row(i); }
  const AC_Vector3& operator[](int i) const{ return row(i); }
  void              transpose();
  void              clear()  { m[0].clear(); m[1].clear(); m[2].clear(); } 
  void              identity(); 

  AC_Matrix3&       operator*=(const AC_Matrix3& that);

  inline void       copyFromPerformer(const pfMatrix *);

  inline void       copyFromRowMatrix34(const float *);

protected:
  static void swap(float &a, float& b);
  AC_Vector3 m[3];
};

inline void AC_Matrix3::swap(float &a, float& b)
{
  float t = a;
  a = b;
  b = t;
}

inline void AC_Matrix3::transpose()
{
  swap(m[0][1],m[1][0]);
  swap(m[0][2],m[2][0]);
  swap(m[1][2],m[2][1]);
}

inline void AC_Matrix3::identity()
{
  for(int i = 0; i < 3; i++) m[i][i] = 1.0f;
  m[1][0] = m[2][0] = m[2][1] = m[0][1] = m[0][2] = m[1][2] = 0.0f;
}

inline AC_Matrix3 operator*(const AC_Matrix3& m1, const AC_Matrix3& m2)
{
  AC_Matrix3 res;

  for(int i = 0; i < 3; i++) {
    AC_Vector3 t = m2.column(i);
    res[0][i] = dot(m1.row(0),t);
    res[1][i] = dot(m1.row(1),t);
    res[2][i] = dot(m1.row(2),t);
  }

  return res;
}

inline AC_Matrix3& AC_Matrix3::operator*= (const AC_Matrix3& that)
{
  *this = *this * that;
  return *this;
}

/** Takes in a Performer matrix and creates a valid AC rotation matrix
    from it. 

    Note that this operation typically does not result in valid
    rotation from Performer coordinates to AC coordinates.

    Instead you must still multiply the resulting matrix with matrix
    that turns performer coordinates to AC coordinates (usually
    multiplication from *right* is the correct choice. . */

inline void AC_Matrix3::copyFromPerformer(const pfMatrix *pfm)
{
  copyFromRowMatrix34((const float *)pfm);

  /* CRAP:

  1  0  0     a  b  c     a -g  d
  0  0 -1  *  d  e  f  =  b -h  e
  0  1  0     g  h  i     c -i  f

  a  b  c     1  0  0     a  d  g
  d  e  f  *  0  0 -1  =  c  f  i
  g  h  i     0  1  0    -b -e -h

  */

  /*
#if 0
  m[0][0] = that[0]; //  a
  m[0][1] =-that[8]; // -g
  m[0][2] = that[4]; //  d

  m[1][0] = that[1]; //  b
  m[1][1] =-that[9]; // -h
  m[1][2] = that[5]; //  e

  m[2][0] = that[2]; //  c
  m[2][1] =-that[10];// -i
  m[2][2] = that[6]; //  f
#elsif 0
  m[0][0] = that[0]; //  a
  m[0][1] = that[4]; //  d
  m[0][2] = that[8]; //  g


  m[1][0] = that[2]; //  c
  m[1][1] = that[6]; //  f
  m[1][2] = that[10];//  i

  m[2][0] =-that[2]; // -b
  m[2][1] =-that[5]; // -e
  m[2][2] =-that[9]; // -h 

#endif*/
}

/** Copy matrix from 3x4 (or 4x4) row matrix. The 3x4 matrix holds 12
    floats in row-major ordering (not OpenGL format, but the
    other). This method only gets the rotation elements out of the
    source matrix. */
inline void AC_Matrix3::copyFromRowMatrix34(const float *that)
{
  m[0][0] = that[0];
  m[0][1] = that[1];
  m[0][2] = that[2];

  m[1][0] = that[4];
  m[1][1] = that[5];
  m[1][2] = that[6];

  m[2][0] = that[8];
  m[2][1] = that[9];
  m[2][2] = that[10];
}

inline AC_Vector3 operator*(const AC_Matrix3& m1,const AC_Vector3& m2)
{
  AC_Vector3 res;
  for(int i = 0; i < 3; i++)
    res[i] = dot(m1.row(i),m2);
  return res;
}

inline AC_Vector3 operator*(const AC_Vector3& m2, const AC_Matrix3& m1)
{
  AC_Vector3 res;
  for(int i = 0; i < 3; i++)
    res[i] = dot(m1.column(i),m2);
  return res;
}

#endif
