/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef AC_VECTOR3_H
#define AC_VECTOR3_H

#include <math.h>

#include <stdio.h>

class pfVec3;

/** Minimalistic vector class. AC_Vector3 has the usual vector
    functions and operators. Besides those is has conversion routines
    so you can easily convert other vector types into AC_Vector3. 

    AC_Vector3 is not intended to be a full-featured vector class. You
    probably have your own vector classes that you like so there is
    little point in writing yet another one.

    @author Tommi Ilmonen
*/
class AC_Vector3
{
public:
  /// Default constructor (does not initialize values)
  AC_Vector3() { }
  /// Copy values from an array of floats
  AC_Vector3(const float *that) { *this = *(const AC_Vector3 *) that; }
  /// Copy values from an array of doubles
  AC_Vector3(const double *s) { v[0]=s[0]; v[1]=s[1]; v[2]=s[2]; }
  AC_Vector3(float x, float y, float z)
  { v[0] = x; v[1] = y; v[2] = z; }

  inline float lengthSquared() const 
  { return v[0] * v[0] + v[1] * v[1] + v[2] * v[2]; }

  /// The length of the vector
  inline float length() const { return sqrtf(lengthSquared()); }

  inline float distanceSquared(const AC_Vector3 &that) const;
  inline float distance(const AC_Vector3 &that) const;
  
  /// Set all the elements of the vector.
  inline void set(float x, float y, float z)
  { v[0] = x; v[1] = y; v[2] = z; }
  /// Set all the elements of the vector.
  inline void make(float x, float y, float z)
  { v[0] = x; v[1] = y; v[2] = z; }

  /// Set all vector elements to zero
  inline void zero()
  { v[0] = v[1] = v[2] = 0.0f; }

  inline void clear() { zero(); }

  /// Access the elements via a pointer
  inline float * data() { return v; }
  /// Access the elements via a pointer
  inline const float * data() const { return v; }

  /// Access operator
  inline float & operator[] (uint i) { return v[i]; }
  /// Access operator (const)
  inline const float & operator[] (uint i) const { return v[i]; }
  
  inline AC_Vector3 operator - (const AC_Vector3 &) const;
  inline AC_Vector3 operator + (const AC_Vector3 &) const;

  inline AC_Vector3 &operator = (const float *);

  inline void copyFromPerformer(const pfVec3 *that);

  inline int write(char *buf) const
  { return sprintf(buf, "[ %f %f %f ]", v[0], v[1], v[2]); }

protected:
  float v[3];
};

/// Squared distance between two vectors.
inline float AC_Vector3::distanceSquared(const AC_Vector3 &that) const
{
  float xd = v[0] - that[0];
  float yd = v[1] - that[1];
  float zd = v[2] - that[2];

  return xd * xd + yd * yd + zd * zd;
}

/// Distance between two vectors.
inline float AC_Vector3::distance(const AC_Vector3 &that) const
{
  return sqrtf(distanceSquared(that));
}

inline AC_Vector3 AC_Vector3::operator - (const AC_Vector3 &that) const
{
  return AC_Vector3(v[0] - that[0], v[1] - that[1], v[2] - that[2]);
}

inline AC_Vector3 AC_Vector3::operator + (const AC_Vector3 &that) const
{
  return AC_Vector3(v[0] + that[0], v[1] + that[1], v[2] + that[2]);
}

/// Copy operator (from float)
inline AC_Vector3 &AC_Vector3::operator = (const float *that)
{ *this = *(const AC_Vector3 *) that; return * this; }

inline void AC_Vector3::copyFromPerformer(const pfVec3 *that)
{
  const float *tmp = (const float *) that;

  v[0] =  tmp[0];
  v[1] =  tmp[1];
  v[2] =  tmp[2];
  /* v[0] =  tmp[0];
  v[1] =  tmp[2];
  v[2] = -tmp[1];*/
}

inline float dot(const AC_Vector3& a, const AC_Vector3& b)
{
  return a[0]*b[0] + a[1]*b[1] + a[2]*b[2];
}


#endif

