/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <ac_vr_control.h>

#include <ac_attenuation.h>

#define AC_PI 3.14159265359

#define AC_MAX_SOURCES 4096

bool AC_VrControl::m_debug = false;

AC_VrControl::AC_VrControl(AC_Control *ac)
  : m_defaultAttenuation(new AC_AttenuationPower(1.5)),
    m_listenerLocation(0.0f, 0.0f, 0.0f),
    m_panningMethod(AC_PAN_UNKNOWN),
    m_ac(ac),
    m_distanceScale(1.0f),
    m_delayTimeScale(1.0f)
{
  m_worldRotation.identity();

  if(!ac) AC_error(AC_ERR_ARGUMENT, "AC_VrControl::AC_VrControl # ac is NULL");
}

AC_VrControl::~AC_VrControl()
{
  delete m_defaultAttenuation;
}
  
long AC_VrControl::playSample(int sourceId, const char * sample, float gain,
			      const AC_Vector3 &location, bool loop,
			      float skipSeconds, float interleaveSeconds)
{
  if(m_debug)
    AC_trace2("AC_VrControl::playSample # id = %d, sample = \"%s\"",
	      sourceId, sample);

  /* if((uint) sourceId >= m_sourceLocations.size()) 
     expandSources(sourceId + 1); */

  m_sourceLocations[sourceId].location() = location;

  AC_Vector3 aziEleDist = locationToAziEleDist(location);

  
  AC_Attenuation *att = m_sourceLocations[sourceId].attenuation();

  if(!att) att = m_defaultAttenuation;

  float dgain = att ? att->metersToGain(aziEleDist[2]) : 1.0;


  return m_ac->playSample(sourceId, sample, gain, 
			  aziEleDist[0], aziEleDist[1], 
			  aziEleDist[2], dgain,
			  loop, skipSeconds, interleaveSeconds);
}
  
long AC_VrControl::playSamplePitched
(int sourceId, const char *sample, float gain, 
 float pitch, const AC_Vector3 &location, bool loop)
{
  if(m_debug)
    AC_trace2("AC_VrControl::playSamplePitched # id = %d, sample = \"%s\"",
	      sourceId, sample);

  /* if((uint) sourceId >= m_sourceLocations.size()) 
     expandSources(sourceId + 1); */

  m_sourceLocations[sourceId].location() = location;

  AC_Vector3 aziEleDist = locationToAziEleDist(location);

  AC_Attenuation *att = m_sourceLocations[sourceId].attenuation();

  if(!att) att = m_defaultAttenuation;

  float dgain = att ? att->metersToGain(aziEleDist[2]) : 0;

  return m_ac->playSamplePitched(sourceId, sample, gain, pitch,
				 aziEleDist[0], aziEleDist[1], 
				 aziEleDist[2], dgain,
				 loop);
}

long AC_VrControl::playInput(int sourceId, int inputChannel, 
			     float gain, const AC_Vector3 &location)
{
  if(m_debug)
    AC_trace2("AC_VrControl::playSamplePitched # id = %d, ichan = %d",
	      sourceId, inputChannel);

  /* if((uint) sourceId >= m_sourceLocations.size()) 
     expandSources(sourceId + 1); */

  m_sourceLocations[sourceId].location() = location;

  AC_Vector3 aziEleDist = locationToAziEleDist(location);

  AC_Attenuation *att = m_sourceLocations[sourceId].attenuation();

  if(!att) att = m_defaultAttenuation;

  float dgain = att ? att->metersToGain(aziEleDist[2]) : 1.0;

  return m_ac->playInput(sourceId, inputChannel, gain, 
			 aziEleDist[0], aziEleDist[1],
			 aziEleDist[2], dgain);
}
 
void AC_VrControl::moveListener(const AC_Vector3 &location, 
				float interpolationTimeSeconds)
{
  if(m_debug)
    AC_trace2("AC_VrControl::moveListener # [%f %f %f]  time = %f",
	      location[0], location[1], location[2], interpolationTimeSeconds);

  m_listenerLocation = location;

  for(iterator it = m_sourceLocations.begin(); 
      it != m_sourceLocations.end(); it++) {
    // for(uint i=0; i < m_sourceLocations.size(); i++) {
    moveSource((*it).first, interpolationTimeSeconds, (*it).second.location());
  }
}

void AC_VrControl::moveListener(const AC_Vector3 &location, 
				const AC_Matrix3 & rotation,
				float interpolationTimeSeconds)
{
  if(m_debug)
    AC_trace2("AC_VrControl::moveListener # [%f %f %f]  time = %f",
	      location[0], location[1], location[2], interpolationTimeSeconds);

  m_listenerLocation = location;
  m_listenerRotationInverse = rotation;
  /* For valid rotation matrices, transposition equals inversion. */
  m_listenerRotationInverse.transpose();

  for(iterator it = m_sourceLocations.begin(); 
      it != m_sourceLocations.end(); it++) {
    // for(uint i=0; i < m_sourceLocations.size(); i++) {
    moveSource((*it).first, interpolationTimeSeconds, (*it).second.location());
  }
}

/** Virtual world rotation relative to physical world. */
void AC_VrControl::setWorldRotation(const AC_Matrix3 &rot, 
				    float interpolationTimeSeconds)
{
  if(m_debug)
    AC_trace2("AC_VrControl::setWorldRotation #  time = %f\n"
	      "  rot = [%3.3f %3.3f %3.3f]\n"
	      "        [%3.3f %3.3f %3.3f]\n"
	      "        [%3.3f %3.3f %3.3f]",
	      interpolationTimeSeconds,
	      rot[0][0], rot[0][1], rot[0][2],
	      rot[1][0], rot[1][1], rot[1][2],
	      rot[2][0], rot[2][1], rot[2][2]);
  
  m_worldRotation = rot;

  for(iterator it = m_sourceLocations.begin(); 
      it != m_sourceLocations.end(); it++) {
    // for(uint i=0; i < m_sourceLocations.size(); i++) {
    moveSource((*it).first, interpolationTimeSeconds, (*it).second.location());
  }  
}

void AC_VrControl::setTransformations(const AC_Vector3 &listenerLocation,  
				      const AC_Matrix3 &listenerRotation,
				      const AC_Matrix3 &worldRotation,
				      float interpolationTimeSeconds)
{
  if(m_debug)
    AC_trace2("AC_VrControl::setTransformations");

  m_worldRotation = worldRotation;
  m_listenerLocation = listenerLocation;
  m_listenerRotationInverse = listenerRotation;
  m_listenerRotationInverse.transpose();

  for(iterator it = m_sourceLocations.begin(); 
      it != m_sourceLocations.end(); it++) {
    moveSource((*it).first, interpolationTimeSeconds, (*it).second.location());
  }
}

void AC_VrControl::moveSource(uint sourceId, 
			      float interpolationTimeSeconds, 
			      const AC_Vector3 &location)
{ 
  
  if(m_debug)
    AC_trace2
      ("AC_VrControl::moveSource # %d [%3.3f %3.3f %3.3f] time = %3.3f", 
       sourceId,
       location[0], location[1], location[2], interpolationTimeSeconds);

  // Resize the source array if necessary:
  /* if(m_sourceLocations.size() <= sourceId)
     expandSources(sourceId + 1); */
  
  m_sourceLocations[sourceId].location() = location;

  AC_Vector3 aziEleDist = locationToAziEleDist(location);

  AC_Attenuation *att = m_sourceLocations[sourceId].attenuation();

  if(!att) att = m_defaultAttenuation;

  float gain = att ? att->metersToGain(aziEleDist[2]) : 1.0;

  aziEleDist[2] *= m_delayTimeScale;  

  if(m_debug)
    AC_trace2
      ("AC_VrControl::moveSource # azi = %f, ele %f, dist = %f, gain = %f", 
       aziEleDist[0], aziEleDist[1], aziEleDist[2], gain);

  m_ac->changeSourceLocation(sourceId, interpolationTimeSeconds, 
			     aziEleDist[0], aziEleDist[1], 
			     aziEleDist[2], gain);
}

/** Quits using a particular sound source. Use this method if some
    of your sound sources is not active any more. 
    
    @author Eric Klein.
*/
void AC_VrControl::quitUsingSource(int sourceId)
{
  m_sourceLocations.erase(sourceId);
}

void AC_VrControl::setAttenuation(uint source, AC_Attenuation *att)
{
  /* expandSources(source + 1); */

  m_sourceLocations[source].setAttenuation(att);
}

/** Resize the source locations array. */
void AC_VrControl::resizeSources(uint)
{
  /* uint oldSize = m_sourceLocations.size();

  if(m_debug)
    AC_trace2("AC_VrControl::resizeSources # %d -> %d", oldSize, n);

  m_sourceLocations.resize(n);
  */

  AC_trace2("AC_VrControl::resizeSources # USE DEPRECATED (not need to call)");
}

void AC_VrControl::setDefaultAttenuation(AC_Attenuation *a)
{
  delete m_defaultAttenuation;
  m_defaultAttenuation = a;
}

void AC_VrControl::write(FILE *f)
{
  char buf[128];
  
  m_listenerLocation.write(buf);

  fprintf(f, "Listener location = %s\n"
	  "Panning method = %d\n"
	  "AC_Control = %p\n"
	  "Source count = %d\n{\n", 
	  buf, (int) m_panningMethod, m_ac, m_sourceLocations.size());

  for(iterator it = m_sourceLocations.begin(); 
      it != m_sourceLocations.end(); it++) {

    fprintf(f, "  Source[%d] : ", (*it).first);
    (*it).second.write(f);
    fprintf(f, "\n");
  }
  fprintf(f, "}\n");
}

/* void AC_VrControl::expandSources(uint n)
{
  if(n > m_sourceLocations.size())
    resizeSources(n);
}
*/

AC_Vector3 AC_VrControl::locationToAziEleDist(const AC_Vector3 &loc)
{
  AC_Vector3 v(loc - m_listenerLocation);

  if(m_panningMethod == AC_PAN_HRTF)
    v = m_listenerRotationInverse * v;

  v = m_worldRotation * v;
  
  float azi, ele, l = v.length();
  
  if(l > 1e-4) {
    azi = 180.0 * atan2f(v[0], -v[2]) / AC_PI;
    ele = 180.0 * atan2f(v[1], sqrtf(v[0] * v[0] + v[2] * v[2])) / AC_PI;
  }
  else {
    azi = 0.0;
    ele = 0.0;
  }

  l *= m_distanceScale;
  
  if(m_debug)
    AC_trace2("AC_VrControl::locationToAziEleDist # [%3.3f %3.3f %3.3f] -> "
	      "[%3.3f %3.3f %3.3f] -> "
	      "azi = %3.3f, ele = %3.3f, dist = %3.3f", 
	      loc[0], loc[1], loc[2], v[0], v[1], v[2], azi, ele, l);

  return AC_Vector3(azi, ele, l);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


AC_VrControl::SourceParams::SourceParams()
  : m_attenuation(0) 
{ 
  m_location.clear(); 
}

AC_VrControl::SourceParams::SourceParams(const AC_VrControl::SourceParams &sp)
  : m_location(sp.m_location),
    m_attenuation(sp.m_attenuation ? sp.m_attenuation->clone() : 0)
{}

AC_VrControl::SourceParams::~SourceParams()
{
  delete m_attenuation;
}

void AC_VrControl::SourceParams::write(FILE *f)
{
  char buf[128];

  m_location.write(buf);

  fprintf(f, "Location = %s, attenuation = %p", buf, m_attenuation);
}

AC_VrControl::SourceParams & AC_VrControl::SourceParams::operator = 
(const AC_VrControl::SourceParams &sp)
{
  m_location    = sp.m_location;
  m_attenuation = sp.m_attenuation ? sp.m_attenuation->clone() : 0;

  return *this;
}

void AC_VrControl::SourceParams::setAttenuation(AC_Attenuation *a)
{
  delete m_attenuation;
  m_attenuation = a;
}
