/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef AC_VR_CONTROL_H
#define AC_VR_CONTROL_H

#include <ac_control.h>
#include <ac_matrix3.h>

#include <map>
#include <stdio.h>

class AC_Attenuation;

/** A virtual reality -directed wrapper to handle auralization. This
    class offers the user the possibility to simply set the locations
    of the sound sources and listener. This class takes care of
    converting these parameters to auralization parameters (azimuth,
    elevation, distance).

    Most of the functions look very much like the functions in
    AC_Control. This is no accident -- the methods here are mostly
    very simple wrappers.
    
    There are a number of functions in AC_Control that are NOT
    replicated in AC_VrControl. Use AC_Control directly instead.

    @see AC_Control

    @author Tommi Ilmonen
 */

class AC_VrControl
{
public:

  AC_VrControl(AC_Control *);
  ~AC_VrControl();

  /** 
   * Plays a sample.
   * @arg sourceId Id of this voice
   * @arg sample Name of the sample to be played
   * @arg gain Linear gain (0.0 - 1.0)
   * @arg location Location of the sound source, in world coordinates.
   * @arg loop Should the sound play in a loop or not?
   * @return This method returns the <B>note id</B> of the new note. If
   * the connection was broken, then -1 is returned.
   * @see #moveListener, for updating the position of the listener.
   */
  long playSample(int sourceId, const char * sample, float gain,
		  const AC_Vector3 &location, bool loop,
		  float skipSeconds = 0.0, float interleaveSeconds = 0.0);

  /** Play a sample at given pitch.*/
  long playSamplePitched(int sourceId, const char *sample, float gain, 
			 float pitch, const AC_Vector3 &location, bool loop);

  /** Play sound from one of synthesizer's input channels. */  
  long playInput(int sourceId, int inputChannel, 
		 float gain, const AC_Vector3 &location);
  
  /** Plays granular voice. */
  void moveListener(const AC_Vector3 &location, 
		    float interpolationTimeSeconds);

  /** Moves listener to given location. */
  void moveListener(const AC_Vector3 &location, 
		    const AC_Matrix3 &rotation,
		    float interpolationTimeSeconds);

  /** Set virtual world rotation relative to physical world. */
  void setWorldRotation(const AC_Matrix3 &, 
			float interpolationTimeSeconds);

  /** Set the listener location and world rotation in one pass. This
    method is useful if you need to adjust both the listener location
    and world rotation frequently. You could get the same results with
    using the functions that do these tasks independently, but using
    this method causes less computation and less traffic between your
    app and the auralization engine -> better performance. */
  void setTransformations(const AC_Vector3 &listenerLocation, 
			  const AC_Matrix3 &listenerRotation,
			  const AC_Matrix3 &worldRotation,
			  float interpolationTimeSeconds);

  /** Move sound source to given location. */
  void moveSource(uint source, 
		  float interpolationTimeSeconds, 
		  const AC_Vector3 &location);

  /** Quits using a particular sound source. Use this method if some
    of your sound sources is not active any more. */
  void quitUsingSource(int sourceId);

  /** Assign attenuation parameters to given sound source. */
  void setAttenuation(uint source, AC_Attenuation *att);

  /** 
   * Resize the source locations array. 
   * @deprecated Doesn't do anything anymore!
   */
  void resizeSources(uint);

  /** Sets the default attenuation calculator. 

      Example:
      \code

      #include <ac_attenuation.h>

      AC_VrControl *avr;
      avr->setDefaultAttenuation(new AC_AttenuationPower(1.0));
      \endcode
  */
  void setDefaultAttenuation(AC_Attenuation *);

  /// Access the actual controller
  AC_Control * ac() { return m_ac;}

  /** Dump AC_VrControl state to a text file. Use for debugging. */
  void write(FILE *);

  void setDistanceScale(float scale)
  { m_distanceScale = scale; }

  void setDelayTimeScale(float scale)
  { m_delayTimeScale = scale; }

  static void setDebug(bool debug) { m_debug = debug; }

protected:

  AC_Vector3 locationToAziEleDist(const AC_Vector3 &location);  

  class SourceParams {
  public:
    SourceParams();
    SourceParams(const SourceParams &);
    ~SourceParams();

    void setAttenuation(AC_Attenuation *);

    AC_Vector3 & location() { return m_location; }
    AC_Attenuation *attenuation() { return m_attenuation; }

    void write(FILE *);

    SourceParams & operator = (const SourceParams &);

    AC_Vector3 m_location;
    AC_Attenuation *m_attenuation;

  };

  std::map<int, SourceParams> m_sourceLocations;
  typedef std::map<int, SourceParams>::iterator iterator;

  AC_Attenuation *m_defaultAttenuation;

  AC_Vector3              m_listenerLocation;
  /** World rotation relative to physical world coordinates. */
  AC_Matrix3              m_worldRotation;
  /** World rotation relative to physical world coordinates. */
  AC_Matrix3              m_listenerRotationInverse;

  AC_PanningMethod        m_panningMethod;

  AC_Control             *m_ac;

  float                   m_distanceScale;
  float                   m_delayTimeScale;

  static bool             m_debug;
};

#endif
