/* COPYRIGHT
 *
 * This file is part of the DIVA IO library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000.
 * Jarmo.Hiipakka@hut.fi Tommi.Ilmonen@hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tcm.hut.fi/Research/DIVA/
 *
 * DIVA IO library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */
#include "di_audio_file_threaded.h"

#include "di_error.h"

#include <sched.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

DI_AudioFileThreaded::DI_AudioFileThreaded()
  : m_tmpBuffer(0),
    m_fileName(0),
    m_blockSize(0),
    m_userFormat(DI_ASF_FLOAT32),
    m_end(false),
    m_ok(true),
    m_mode(DI_IO_NONE),
    m_status(DI_AFT_EMPTY)
{
  for(int i=0; i < 3; i++) 
    m_mutexes[i].init(false, false, false);
}

DI_AudioFileThreaded::~DI_AudioFileThreaded()
{
  if(m_fileName) free(m_fileName);
}

bool DI_AudioFileThreaded::startRead(const char *file, 
				     DI_AudioSampleFormat userFormat,
				     ulong maxBlock,
				     long startFrame,
				     long endFrame,
				     bool loop,
				     int priority)
{
  if(m_fileName) free(m_fileName);
  m_fileName = strdup(file);

  m_userFormat = userFormat;
  m_blockSize = maxBlock;
  m_startFrame = startFrame;
  m_endFrame = endFrame;
  m_threadFrames = 0;

  m_loop = loop;
  m_userBuffer = 0;
  m_threadBuffer = 0;
  m_end = false;
  m_oneFill = false;
  m_userEnter = false;
  m_mode = DI_IO_INPUT;

  m_status = DI_AFT_OPENING;

  m_userFrames = 0;
  m_threadFrames = 0;
  m_priority = priority;

  return run(true);
}

bool DI_AudioFileThreaded::startWrite(const char *file, 
				      DI_AudioSampleFormat userFormat,
				      uint channels,
				      ulong maxBlock,
				      int priority,
				      DI_AudioFileFormat fileFormat,
				      uint sampleWidth,
				      DI_AudioFileSampleFormat sampleFormat,
				      DI_AudioFileByteOrder byteOrder)
{  
  if(m_fileName) free(m_fileName);
  m_fileName = strdup(file);

  m_userFormat = userFormat;
  m_blockSize = maxBlock;
  
  m_startFrame = 0;
  m_endFrame = 0;
  m_threadFrames = 0;
  m_chans = channels;
  m_loop = false;
  m_userBuffer = 0;
  m_threadBuffer = 0;
  m_end = false;
  m_oneFill = false;
  m_userEnter = false;
  m_mode = DI_IO_OUTPUT;

  m_status = DI_AFT_OPENING;

  m_userFrames = 0;
  m_threadFrames = 0;
  m_priority = priority;
  m_mutexes[0].lock();

  m_file.setFileFormat(fileFormat);
  m_file.setSampleWidth(sampleWidth);
  m_file.setSampleFormat(sampleFormat);
  m_file.setByteOrder(byteOrder);
  m_file.setChannels(m_chans);
  
  m_ok = true;

  return run(true);
}

/** Prepare the file for operation. 

    This method may block if called too early. */
bool DI_AudioFileThreaded::userPrepare()
{
  if(m_mode == DI_IO_OUTPUT) {
    while(m_status != DI_AFT_READY && m_ok) {
      DI_Sleep s;
      s.sleepMs(20);
    }
    return m_ok;
  }

  while(!m_oneFill && m_ok) {
    DI_Sleep s;
    s.sleepMs(20);
  }
  m_userBuffer = 0;
  m_mutexes[0].lock();
  m_userEnter = true;
  return m_ok;
}

int DI_AudioFileThreaded::readFrames(void *buffer, ulong frames)
{
  // const char * fname = "DI_AudioFileThreaded::readFrames";
  
  // DI_trace(fname);

  ulong currPlace = m_userFrames % (m_blockSize * 2);
  char * currPtr = & m_tmpBuffer[currPlace * m_userSWidth * m_chans];

  memcpy(buffer, currPtr, frames * m_userSWidth * m_chans);

  m_userFrames += frames;

  if(!(m_userFrames % m_blockSize)) {

    // DI_trace(fname, "Block switch");
    
    int oldBuf = m_userBuffer;
    m_userBuffer = (oldBuf + 1) % 2;

    m_mutexes[m_userBuffer].lock();
    m_mutexes[oldBuf].unlock();

    m_mutexes[2].lock();
    m_mutexes[2].unlock();
  }

  return (int) frames;
}

int DI_AudioFileThreaded::writeFrames(void *buffer, ulong frames)
{
  // DI_trace("DI_AudioFileThreaded::writeFrames");

  /** Block while the write thread has not written the first block
      (unfinished mutex locks). We should use condition variables
      here. The current setup is highly non-optimal. */

  while((m_userFrames >= m_blockSize && m_status != DI_AFT_WRITING)) {
    DI_trace("DI_AudioFileThreaded::writeFrames BLOCK");
    DI_Sleep s;
    s.sleepMs(20);
  }

  ulong currPlace = m_userFrames % (m_blockSize * 2);
  char * currPtr = & m_tmpBuffer[currPlace * m_userSWidth * m_chans];

  memcpy(currPtr, buffer, frames * m_userSWidth * m_chans);
  
  // DI_trace("DI_AudioFileThreaded::writeAudio", 3);
  
  m_userFrames += frames;
  
  if(!(m_userFrames % m_blockSize)) {
    
    int oldBuf = m_userBuffer;
    m_userBuffer = (oldBuf + 1) & 1;

    m_userEnter = true;

    m_mutexes[m_userBuffer].lock();
    m_mutexes[oldBuf].unlock();

    m_mutexes[2].lock();
    m_mutexes[2].unlock();   
  }
  // DI_trace("DI_AudioFileThreaded::writeFrames EXIT");
  return (int) frames;
}

void DI_AudioFileThreaded::requestEnd()
{
  m_end = true;
  m_mutexes[m_userBuffer].unlock();
}

bool DI_AudioFileThreaded::allocateBuffer
(DI_AudioSampleFormat userFormat, ulong maxBlock, ulong ch)
{
  if(!maxBlock) return false;

  if(m_tmpBuffer) free(m_tmpBuffer);  

  switch(userFormat) {
  case DI_ASF_INT16:
    m_userSWidth = 2;
    break;

  case DI_ASF_INT32:
  case DI_ASF_FLOAT32:
    m_userSWidth = 4;
    break;

  case DI_ASF_FLOAT64:
    m_userSWidth = 8;
    break;
  }

  m_blockChars = m_userSWidth * ch * m_blockSize;

  m_tmpBuffer = (char *) malloc(m_blockChars * 2);

  // DI_trace("DI_AudioFileThreaded::allocateBuffer", m_blockSize);
  // DI_trace("DI_AudioFileThreaded::allocateBuffer", m_tmpBuffer);

  return m_tmpBuffer != 0;
}

void DI_AudioFileThreaded::childLoop()
{
  const char * fname = "DI_AudioFileThreaded::childLoop";

  DI_trace(fname, m_fileName);
  
  if(m_priority >= 0) {
    sched_param param;
    param.sched_priority = m_priority;
    pid_t pid = getpid();
    sched_setscheduler(pid, SCHED_FIFO, &param);
  }

  bool ok = m_file.open(m_fileName, m_mode, m_userFormat, m_blockSize);
  m_ok = m_ok && ok;

  m_mutexes[2].lock();

  if(ok) {
    DI_trace(fname, "IS OPEN");

    if(m_mode == DI_IO_INPUT)
      m_chans = m_file.inChannels();
    else
      m_chans = m_file.outChannels();

    ok = allocateBuffer(m_userFormat, m_blockSize, m_chans);

    m_ok = m_ok && ok;

    DI_trace(fname, "IS ALLOCATED");

    m_status = DI_AFT_READY;
    
    if(m_mode == DI_IO_INPUT)
      while(!m_end && ok)
	readAudio();
    else 
      while(!m_end && ok)
	writeAudio();
  }
  
  m_mutexes[2].unlock();
  m_file.close();
  // DI_trace(fname, "EXIT");
}

/** File-thread method. */
void DI_AudioFileThreaded::readAudio()
{
  // DI_trace("DI_AudioFileThreaded::readAudio", m_threadFrames);

  if(m_end) return;

  // User has not read a single frame
  while(m_threadFrames && !m_userEnter && !m_threadBuffer) {
    DI_Sleep s;
    s.sleepMs(20);
  }

  int threadBuffer = m_threadBuffer;

  m_mutexes[threadBuffer].lock();
  m_mutexes[2].unlock();

  m_status = DI_AFT_READING;

  char * ptr = &m_tmpBuffer[m_blockChars * threadBuffer];

  long framesLeft = m_file.frameCount() - m_threadFrames;
  long blockSize = framesLeft < (long) m_blockSize ? framesLeft : m_blockSize;

  int readFrames;

  if(blockSize)
    readFrames = m_file.readFrames(ptr, blockSize);
  else 
    readFrames = 0;

  // Fill with zero:
  if(readFrames != (int) m_blockSize) {
    
    // Advance ptr;
    ptr += (readFrames * m_userSWidth * m_chans);
    ulong fillFrames = m_blockSize - readFrames;
    
    if(m_userFormat == DI_ASF_INT16 || m_userFormat == DI_ASF_INT32) {
      bzero(ptr, fillFrames * m_userSWidth * m_chans);
    }
    else if (m_userFormat == DI_ASF_FLOAT32) {

      float *fptr = (float *) ptr;
      uint samples = fillFrames * m_chans;

      for(uint i=0; i < samples; i++) 
	fptr[i] = 0.0;
    }
    else if (m_userFormat == DI_ASF_FLOAT64) {

      double *dptr = (double *) ptr;
      uint samples = fillFrames * m_chans;

      for(uint i=0; i < samples; i++) 
	dptr[i] = 0.0;
    } 
    // DI_trace("DI_AudioFileThreaded::readAudio fillzero", fillFrames);

  }

  m_oneFill = true;

  m_mutexes[2].lock();
  m_mutexes[threadBuffer].unlock();

  m_threadBuffer = (threadBuffer + 1) & 1;
  m_threadFrames += readFrames;
}

/** File-thread method. */

void DI_AudioFileThreaded::writeAudio()
{
  // DI_trace("DI_AudioFileThreaded::writeAudio", m_threadFrames);
  
  if(m_end) return;
  
  // Block until user has written one data block
  if(!m_userEnter) {
    m_mutexes[0].lock();
    m_mutexes[0].unlock();
  }

  int threadBuffer = m_threadBuffer;

  m_mutexes[threadBuffer].lock();
  m_mutexes[2].unlock();

  m_status = DI_AFT_WRITING;

  m_threadBuffer = (threadBuffer + 1) & 1;
  
  m_mutexes[2].lock();
  m_mutexes[threadBuffer].unlock();

  int writeFrames = m_file.writeFrames
    (&m_tmpBuffer[m_blockChars * threadBuffer], m_blockSize);
  
  if((uint) writeFrames != m_blockSize)
    DI_error(DI_ERR_INTERNAL, 
	     "DI_AudioFileThreaded::writeAudio # write failed");

  m_threadFrames += writeFrames;

}
