/* COPYRIGHT
 *
 * This file is part of the DIVA IO library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000.
 * Jarmo.Hiipakka@hut.fi Tommi.Ilmonen@hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tcm.hut.fi/Research/DIVA/
 *
 * DIVA IO library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */
//
// DIVA Library. Copyright (C) Tommi Ilmonen 2000
//
// $Log: di_audio_port.C,v $
// Revision 1.1.1.1  2001/06/04 09:04:19  tilmonen
// Mustajuuri 4.6.2001, initial CVS cersion
//


#include "di_audio_port.h"
#include "di_error.h"
#include "di_sleep.h"
#include "di_numconv.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <sys/time.h>
#include <unistd.h>

#define USE_BLOCK_MODE 1

#define USE_GO 1

// 2^15 - 1
#define DI_ALSA_SSCALE 32767
// 2^31 - 1
#define DI_ALSA_ISCALE 2147483647

// --- Constructors -----------------------------------------------------------

DI_AudioPort::DI_AudioPort()
  : m_phandle(0),
    m_chandle(0),
    m_tmpBuffer(0),
    m_tmpBufferFrames(0),
    m_bytesPerSample(0),
    m_mode(DI_IO_NONE)    
{
}

DI_AudioPort::~DI_AudioPort(void)
{
  DI_trace("DI_AudioPort::~DI_AudioPort");

  close();
}

// --- Operations -------------------------------------------------------------

bool DI_AudioPort::open(const char *device,
			DI_IoMode mode,
			DI_AudioSampleFormat userSample, 
			ulong bufferSize,
			ulong samplerate,
			ulong channels,
			ulong maxBlock,
			bool interleave)
{
  const char * fname = "DI_AudioPort::open";
  char buf[512];

  DI_trace(fname, (int) mode);

  m_userSample = userSample;

  int cardId = -1;
  int dev = -1;
  int subDev = -1;

  if(device) 
    sscanf(device, "%d %d %d", &cardId, &dev, &subDev);
  
  sprintf(buf, "card = %d, dev = %d, subdev = %d",
	  cardId, dev, subDev);
  
  if(cardId < 0) cardId = snd_defaults_pcm_card();
  if(dev < 0) dev = 0;
  int e = 0;

  DI_trace(fname, buf);

  if(mode & DI_IO_INPUT) {
    if(subDev < 0)
      e = snd_pcm_plug_open_device
	(&m_chandle, cardId, dev, SND_PCM_STREAM_CAPTURE, 0);
    else
      e = snd_pcm_plug_open_subdevice(&m_chandle, cardId, dev, subDev, 
				      SND_PCM_STREAM_CAPTURE, 0);

    if(!m_chandle || e) {
      DI_error(DI_ERR_SYSTEM, fname, "Could not open input");
      close();
      return false;
    }
  }
  DI_trace(fname, "0.6");
  if(mode & DI_IO_OUTPUT) {
    if(subDev < 0)
      e = snd_pcm_plug_open_device
	(&m_phandle, cardId, dev, SND_PCM_STREAM_PLAYBACK, 0);
    else
      e = snd_pcm_plug_open_subdevice(&m_phandle, cardId, dev, subDev, 
				      SND_PCM_STREAM_PLAYBACK, 0);
    if(!m_phandle || e) {
      DI_error(DI_ERR_SYSTEM, fname, "Could not open output");
      DI_error(DI_ERR_SYSTEM, fname, snd_strerror(e));
      close();
      return false;
    }
    
    snd_pcm_info_t info;
    bzero(&info, sizeof(info));
    snd_pcm_info(m_phandle, &info);
    sprintf(buf, "device = %d, stream = %d, subdevice = %d type = %d\n\
name = \"%s\"\n\
subname  = \"%s\"\n\
subdevices_count = %d, subdevices_avail = %d\n",
	    info.device, info.stream, info.subdevice, info.type,
	    info.name, 
	    info.subname,
	    info.subdevices_count, info.subdevices_avail);
    DI_trace(fname, buf);
  }   

 
  DI_trace(fname, 1);

  bzero(&m_params, sizeof(m_params));

  switch(userSample) {
  case DI_ASF_INT16:
    m_params.format.sfmt = SND_PCM_SFMT_S16_LE;
    break;
  case DI_ASF_INT32:
    m_params.format.sfmt = SND_PCM_SFMT_S32_LE;
    break;
  case DI_ASF_FLOAT32:
    m_params.format.sfmt = SND_PCM_SFMT_FLOAT_LE;
    break;
  case DI_ASF_FLOAT64:
    m_params.format.sfmt = SND_PCM_SFMT_FLOAT64_LE;
    break;
  default:
    DI_error(DI_ERR_ARGUMENT, "Bad sample format");
    close();
    return false;
  }    

  m_params.format.rate = samplerate;
  m_params.format.channels = channels;

  m_params.ready_mode = SND_PCM_READY_FRAGMENT;
  m_params.start_mode = SND_PCM_START_DATA;
  m_params.xrun_mode = SND_PCM_XRUN_NONE;
  m_params.xfer_mode = interleave ? 
    SND_PCM_XFER_INTERLEAVED : SND_PCM_XFER_NONINTERLEAVED;
  // params.xfer_mode = SND_PCM_XFER_INTERLEAVED;
  m_params.mmap_shape = SND_PCM_MMAP_NONINTERLEAVED;
  
#ifdef USE_BLOCK_MODE
  int frag_size = 1;

  while(frag_size < (int) (maxBlock * 2 * channels) && frag_size < 10000) {
    frag_size = frag_size << 1;
  }
    
  /* if(frag_size > chaninfo.max_fragment_size)
    frag_size = chaninfo.max_fragment_size;
  else if(frag_size < chaninfo.min_fragment_size)
    frag_size = chaninfo.min_fragment_size;
  */

  // int totalBytes = bufferSize * 2 * channels;

  m_params.frag_size = frag_size;
  m_params.avail_min = m_params.frag_size;  
  m_params.buffer_size = m_params.frag_size * 3;
#else
  m_params.mode = SND_PCM_MODE_STREAM;
  m_params.buf.stream.queue_size = bufferSize;
  m_params.buf.stream.fill = SND_PCM_FILL_SILENCE;
  m_params.buf.stream.max_fill = bufferSize / 2;
#endif
  
  DI_trace(fname, 2);

  if(m_phandle) {
    
    if(!alsaSeekSampleFormat(m_phandle, &m_params)) {
      
      DI_error(DI_ERR_ARGUMENT, fname);
      close();
      return false;
    }

    DI_trace(fname, snd_pcm_format_description(m_params.format.sfmt));
  }

  DI_trace(fname, 3);

  if(m_chandle) {

    if(!alsaSeekSampleFormat(m_chandle, &m_params)) {
      
      DI_error(DI_ERR_ARGUMENT, fname);
      close();
      return false;
    }
  }

  DI_trace(fname, 4);

  if(m_phandle) {

    e = snd_pcm_prepare(m_phandle);

    if(e < 0) {
      DI_error(DI_ERR_ARGUMENT, fname, snd_strerror(e));
      close();
      return false;
    }
  }

  if(m_chandle) {
    
    e = snd_pcm_prepare(m_chandle);
    
    if(e < 0) {
      DI_error(DI_ERR_ARGUMENT, fname, snd_strerror(e));
      close();
      return false;
    }
  }

  if(!allocBuffer(&m_params, maxBlock)) {
    close();
    return false;
  }

  m_underFlows = 0;
  m_mode = mode;

  if(mode & DI_IO_INPUT)  m_inChannels  = m_params.format.channels;
  if(mode & DI_IO_OUTPUT) m_outChannels = m_params.format.channels;
  
  DI_trace(fname, 5);
  
  return true;

}

bool DI_AudioPort::start()
{
  const char * fname = "DI_AudioPort::start";

  if(!m_phandle && !m_chandle) return false;

#ifndef USE_GO
  if(m_rmeHacks)
    return true;
#endif

  return true; 

  // Does not work yet (!)

  if(m_phandle) {
    int e = snd_pcm_start(m_phandle);
    if(e < 0) {

      if(e == -EBADFD) {
	DI_error(DI_ERR_SYSTEM, fname, "-BADFD");
      }

      DI_error(DI_ERR_SYSTEM, fname, snd_strerror(e));
      return false;      
    }
  }
  if(m_chandle) {
    int e = snd_pcm_start(m_chandle);
    if(e < 0) {
      DI_error(DI_ERR_SYSTEM, fname, snd_strerror(e));
      return false;      
    }    
  }
  return true;
}

bool DI_AudioPort::close(void)
{
  const char *fname = "DI_AudioPort::close";
  int e;

  DI_trace(fname, m_mode);

  // If outputting, wait until the output queue has drained
  
  
  // Close the port and free all resources
  
  if(m_phandle) {
    DI_trace(fname, "pclose 1");

    e = snd_pcm_close(m_phandle);
    if(e) {
      DI_error(DI_ERR_SYSTEM, fname, snd_strerror(e));
    }
    m_phandle = 0;
    DI_trace(fname, "pclose 2");
  }
  if(m_chandle) {
    DI_trace(fname, "cclose 1");

    e = snd_pcm_close(m_chandle);
    if(e) {
      DI_error(DI_ERR_SYSTEM, fname, snd_strerror(e));
    }
    m_chandle = 0;
    DI_trace(fname, "cclose 2");
  }

  m_mode = DI_IO_NONE;

  if(m_tmpBuffer) {
    free(m_tmpBuffer);
    m_tmpBuffer = 0;
    m_tmpBufferFrames = 0;
  }
  return true;
}


// --- I/O operations ---------------------------------------------------------

int DI_AudioPort::readFrames(void *buffer, ulong frames)
{
  if(frames > m_tmpBufferFrames) return -1;

  uint samples(frames * m_params.format.channels);

  int e ;

  e = snd_pcm_read(m_chandle, m_tmpBuffer, samples * m_bytesPerSample);

  if(e < 0 ) return -1;

  frames = e / (m_bytesPerSample * m_params.format.channels);
  
  if(m_params.format.sfmt == SND_PCM_SFMT_S16_LE) {
    switch(m_userSample) {
    case DI_ASF_INT16:
      memcpy(buffer, m_tmpBuffer, 2 * samples);
      break;
    case DI_ASF_INT32:
      DI_short2Int((int *)buffer, (short *)m_tmpBuffer, 16, 0, samples);
      break;
    case DI_ASF_FLOAT32:
      DI_short2Float((float *)buffer, (short *)m_tmpBuffer,
		     1.0 / DI_ALSA_SSCALE, 0.0, samples);
      break;
    case DI_ASF_FLOAT64:
      DI_short2Double((double *)buffer, (short *)m_tmpBuffer,
		      1.0 / DI_ALSA_SSCALE, 0.0, samples);
      break;
    default:
      return -1;
    }
  }
  else if(m_params.format.sfmt == SND_PCM_SFMT_S32_LE) {
    switch(m_userSample) {
    case DI_ASF_INT16:
      DI_int2Short((short *)buffer, (int *)m_tmpBuffer, -16, 0, samples);
      break;
    case DI_ASF_INT32:
      memcpy(buffer, m_tmpBuffer, 4*samples);
      break;
    case DI_ASF_FLOAT32:
      DI_int2Float((float *)buffer, (int *)m_tmpBuffer, 1.0 / DI_ALSA_ISCALE,
		   0.0, samples);
      break;
    case DI_ASF_FLOAT64:
      DI_int2Double((double *)buffer, (int *)m_tmpBuffer, 1.0 / DI_ALSA_ISCALE,
		    0.0, samples);
      break;
     default:
       return -1;
    }
  }
  else if(m_params.format.sfmt == SND_PCM_SFMT_FLOAT_LE) {
    // User sample *SHOULD BE* float
    switch(m_userSample) {
    case DI_ASF_FLOAT32:
      memcpy(buffer, m_tmpBuffer, 4*samples);
      break;
    default:
      return -1;
    }
  }
  else if(m_params.format.sfmt == SND_PCM_SFMT_FLOAT64_LE) {
    // User sample *SHOULD BE* double
    switch(m_userSample) {
    case DI_ASF_FLOAT64:
      memcpy(buffer, m_tmpBuffer, 8*samples);
      break;
    default:
      return -1;
    }
  }

  return (int) frames;
}
int DI_AudioPort::readBuffers(void **buffers, ulong frames)
{
  
  if(frames > m_tmpBufferFrames) return -1;

  uint bufbytes (frames * m_bytesPerSample);
  uint chans = m_params.format.channels, i = 0;

  int e;

  e = snd_pcm_readi(m_chandle, m_tmpBuffer, bufbytes * chans);

  if(e < 0 ) return -1;
  
  frames = e / (m_bytesPerSample * chans);
  
  if(m_params.format.sfmt == SND_PCM_SFMT_S16_LE) {
    switch(m_userSample) {
    case DI_ASF_INT16:
      for(; i < chans; i++)
	memcpy((buffers[i]), (& m_tmpBuffer[i*bufbytes]), 2 * frames);
      break;
    case DI_ASF_INT32:
      for(; i < chans; i++)
	DI_short2Int((int *)(buffers[i]), 
		     (short *)(& m_tmpBuffer[i*bufbytes]), 16, 0, frames);
      break;
    case DI_ASF_FLOAT32:
      for(; i < chans; i++)
	DI_short2Float((float *)(buffers[i]), 
		       (short *)(& m_tmpBuffer[i*bufbytes]),
		       1.0 / DI_ALSA_SSCALE, 0.0, frames);
      break;
    case DI_ASF_FLOAT64:
      for(; i < chans; i++)
	DI_short2Double((double *)(buffers[i]), 
			(short *)(& m_tmpBuffer[i*bufbytes]),
			1.0 / DI_ALSA_SSCALE, 0.0, frames);
      break;
    default:
      return -1;
    }
  }
  
  else if(m_params.format.sfmt == SND_PCM_SFMT_S32_LE) {
    switch(m_userSample) {
    case DI_ASF_INT16:
      for(; i < chans; i++)
	DI_int2Short((short *)(buffers[i]), 
		     (int *)(& m_tmpBuffer[i*bufbytes]), -16, 0, frames);
      break;
    case DI_ASF_INT32:
      for(; i < chans; i++)
	memcpy((buffers[i]), (& m_tmpBuffer[i*bufbytes]), 4*frames);
      break;
    case DI_ASF_FLOAT32:
      for(; i < chans; i++)
	DI_int2Float((float *)(buffers[i]), 
		     (int *)(& m_tmpBuffer[i*bufbytes]), 1.0 / DI_ALSA_ISCALE,
		     0.0, frames);
      break;
    case DI_ASF_FLOAT64:
      for(; i < chans; i++)
	DI_int2Double((double *)(buffers[i]), 
		      (int *)(& m_tmpBuffer[i*bufbytes]), 1.0 / DI_ALSA_ISCALE,
		      0.0, frames);
      break;
    default:
      return -1;
    }
  }
  else if(m_params.format.sfmt == SND_PCM_SFMT_FLOAT_LE) {
    // User sample *SHOULD BE* float
    switch(m_userSample) {
    case DI_ASF_FLOAT32:
      for(; i < chans; i++)
	memcpy((buffers[i]), (& m_tmpBuffer[i*bufbytes]), 4*frames);
      break;
    default:
      return -1;
    }
  }
  else if(m_params.format.sfmt == SND_PCM_SFMT_FLOAT64_LE) {
    // User sample *SHOULD BE* double
    switch(m_userSample) {
    case DI_ASF_FLOAT64:
      for(; i < chans; i++)
	memcpy((buffers[i]), (& m_tmpBuffer[i*bufbytes]), 8*frames);
      break;
    default:
      return -1;
    }
  }

  return (int) frames;

}


int DI_AudioPort::writeFrames(const void *buffer, ulong frames)
{
  if(frames > m_tmpBufferFrames) return -1;

  uint samples(frames * m_params.format.channels);

  if(m_params.format.sfmt == SND_PCM_SFMT_S16_LE) {
    switch(m_userSample) {
    case DI_ASF_INT16:
      memcpy(m_tmpBuffer, buffer, samples * 2);
      break;
    case DI_ASF_INT32:
      DI_int2Short((short *)m_tmpBuffer, (int *) buffer, -16, 0, samples);
      break;
     case DI_ASF_FLOAT32:
       // DI_trace("Good");
       DI_float2Short((short *)m_tmpBuffer, (float *) buffer,
		      DI_ALSA_SSCALE, 0, samples);
       break;
     case DI_ASF_FLOAT64:
       DI_double2Short((short *)m_tmpBuffer, (double *) buffer,
		       DI_ALSA_SSCALE, 0, samples);
       break;
     default:
      return -1;
    }
  }
  else if(m_params.format.sfmt == SND_PCM_SFMT_S32_LE) {
    switch(m_userSample) {
    case DI_ASF_INT16:
      DI_short2Int((int *)m_tmpBuffer, (short *)buffer, 16, 0, samples);
      break;
    case DI_ASF_INT32:
      memcpy(m_tmpBuffer, buffer, 4*samples);
      break;
    case DI_ASF_FLOAT32:
      DI_float2Int((int *)m_tmpBuffer, (float *)buffer, DI_ALSA_ISCALE,
		   0, samples);
      break;
    case DI_ASF_FLOAT64:
      DI_double2Int((int *)m_tmpBuffer, (double *)buffer, DI_ALSA_ISCALE,
		    0, samples);
      break;
    default:
      return -1;
    }
  }
  else if(m_params.format.sfmt == SND_PCM_SFMT_FLOAT_LE) {
    // User sample *SHOULD BE* float
    switch(m_userSample) {
    case DI_ASF_FLOAT32:
      memcpy(m_tmpBuffer, buffer, 4*samples);
      break;
    default:
      return -1;
    }
  }
  else if(m_params.format.sfmt == SND_PCM_SFMT_FLOAT64_LE) {
    // User sample *SHOULD BE* double
    switch(m_userSample) {
    case DI_ASF_FLOAT64:
      memcpy(m_tmpBuffer, buffer, 8*samples);
      break;
    default:
      return -1;
    }
  }

  return snd_pcm_writei(m_phandle, m_tmpBuffer, samples * m_bytesPerSample) / 
    (m_bytesPerSample * m_params.format.channels);
}

int DI_AudioPort::writeBuffers(const void **buffers, ulong frames)
{
  // DI_trace("DI_AudioPort::writeBuffers");

  if(frames > m_tmpBufferFrames) return -1;

  uint chans = m_params.format.channels;

  // uint samples (frames * chans);
  uint bufbytes (frames * m_bytesPerSample);
  uint i = 0;

  if(m_params.format.sfmt == SND_PCM_SFMT_S16_LE) {
    switch(m_userSample) {
    case DI_ASF_INT16:
      for(; i < chans; i++)
	memcpy(& m_tmpBuffer[i*bufbytes],  (buffers[i]), frames * 2);
      break;
      
    case DI_ASF_INT32:
      for(; i < chans; i++)
	DI_int2Short((short *) (& m_tmpBuffer[i*bufbytes]), 
		     (int *) (buffers[i]), -16, 0, frames);
      break;

    case DI_ASF_FLOAT32:
      // DI_trace(",");
      for(; i < chans; i++)
	DI_float2Short((short *) ( & m_tmpBuffer[i*bufbytes]), 
		       (float *) (buffers[i]),
		       DI_ALSA_SSCALE, 0, frames);
      // DI_trace(";");
      break;
     case DI_ASF_FLOAT64:
       DI_double2Short((short *) (& m_tmpBuffer[i*bufbytes]), 
		       (double *) (buffers[i]),
		       DI_ALSA_SSCALE, 0, frames);
       break;
     default:
       return -1;
    }
  }
  
  else if(m_params.format.sfmt == SND_PCM_SFMT_S32_LE) {
    switch(m_userSample) {
    case DI_ASF_INT16:
      for(; i < chans; i++)
	DI_short2Int((int *)(& m_tmpBuffer[i*bufbytes]), 
		     (short *)(buffers[i]), 16, 0, frames);
      break;
    case DI_ASF_INT32:
      for(; i < chans; i++)
	memcpy((&m_tmpBuffer[i*bufbytes]), (buffers[i]), 4*frames);
      break;
    case DI_ASF_FLOAT32:
      for(; i < chans; i++)
	DI_float2Int((int *)(&m_tmpBuffer[i*bufbytes]), 
		     (float *)(buffers[i]), DI_ALSA_ISCALE,
		     0, frames);
      break;
    case DI_ASF_FLOAT64:
      for(; i < chans; i++)
	DI_double2Int((int *)(&m_tmpBuffer[i*bufbytes]), 
		      (double *)(buffers[i]), DI_ALSA_ISCALE,
		      0, frames);
      break;
    default:
      return -1;
    }
  }
  
  else if(m_params.format.sfmt == SND_PCM_SFMT_FLOAT_LE) {
    // User sample *SHOULD BE* float
    switch(m_userSample) {
    case DI_ASF_FLOAT32:
      memcpy((& m_tmpBuffer[i*bufbytes]), (buffers[i]), 4*frames);
      break;
    default:
      return -1;
    }
  }
  else if(m_params.format.sfmt == SND_PCM_SFMT_FLOAT64_LE) {
    // User sample *SHOULD BE* double
    switch(m_userSample) {
    case DI_ASF_FLOAT64:
      memcpy((& m_tmpBuffer[i*bufbytes]), (buffers[i]), 8*frames);
      break;
    default:
      return -1;
    }
  }

  return snd_pcm_write(m_phandle, m_tmpBuffer, bufbytes * chans) / 
    (m_bytesPerSample * chans);
}

int DI_AudioPort::inFilledFrames(void) const 
{ 
  snd_pcm_status_t *cs = ( snd_pcm_status_t *) & m_status;

  int e = snd_pcm_status(m_chandle, cs);
  
  if(e) {
    DI_error(DI_ERR_SYSTEM, "DI_AudioPort::inFilledFrames",
	     snd_strerror(e));
    return -1; 
  }

  return cs->avail;
}

int DI_AudioPort::outFilledFrames(void) const 
{ 
  // const char * fname = "DI_AudioPort::outFilledFrames";

  snd_pcm_status_t *cs = ( snd_pcm_status_t *) & m_status;

  int e = snd_pcm_status(m_phandle, cs);
  
  if(e) {
    DI_error(DI_ERR_SYSTEM, "DI_AudioPort::outFilledFrames",
	     snd_strerror(e));
    return -1; 
  }

  return cs->avail;
}

int DI_AudioPort::waitInputFilled(ulong frames, int timeOut)
{
  struct timeval tv;
  int r;
  while(true) {
    // DI_trace("DI_AudioPort::waitOutputFilled", outFilledFrames());
    if((r = inFilledFrames()) < (int) frames) {
      // gettimeofday(&tv, 0);
      tv.tv_sec = 0;      // Zero full seconds.
      tv.tv_usec = 2000;  // two millisecs.
      select(0, 0, 0, 0, & tv); // Sleep
    }
    else
      break;
  }
  return r;
}

int DI_AudioPort::waitOutputFilled(ulong frames, int timeOut)
{
  struct timeval tv;
  int r;

  while(true) {
    // DI_trace("DI_AudioPort::waitOutputFilled", outFilledFrames());
    if((r = outFilledFrames()) > (int) frames) {
      // gettimeofday(&tv, 0);
      tv.tv_sec = 0;      // Zero full seconds.
      tv.tv_usec = 2000;  // two millisecs.
      select(0, 0, 0, 0, & tv);
    }
    else
      break;
  }
  return r;
}


void DI_AudioPort::alsaDescribePlaybackSetup(char *)
{
  /* snd_pcm_channel_setup_t setup;

  bzero(&setup, sizeof(setup));
  setup.mode = SND_PCM_MODE_BLOCK;
  setup.channel = SND_PCM_CHANNEL_PLAYBACK;

  if(m_rmeHacks)
    snd_pcm_plugin_setup(m_phandle, &setup);
  else
    snd_pcm_channel_setup(m_phandle, &setup);

  sprintf(b, "Voices = %d, rate = %d, fraqs = %d, frag_size = %d",
	  setup.format.channels, setup.format.rate, 
	  setup.buf.block.frags, setup.buf.block.frag_size); */
}

void DI_AudioPort::alsaDescribeChannelSetup(char *buf, 
					    snd_pcm_info_t *info)
{
  sprintf(buf, "device = %d, stream = %d, subdevice = %d type = %d\n\
name = \"%s\"\n\
subname  = \"%s\"\n\
subdevices_count = %d, subdevices_avail = %d\n",
	 info->device, info->stream, info->subdevice, info->type,
	 info->name, 
	 info->subname,
	 info->subdevices_count, info->subdevices_avail);
}

bool DI_AudioPort::alsaSeekSampleFormat(snd_pcm_t *pcmh, 
					snd_pcm_params_t *cph)
{
  // int e = snd_pcm_plugin_params(pcmh, cph);
  int e;
  
  e = snd_pcm_params(pcmh, cph);

  if(e >= 0) return true;

  while(e < 0) {
    cph->format.sfmt = alsaDownGradeFormat(cph->format.sfmt);

    e = snd_pcm_params(pcmh, cph);

    if(e >= 0) return true;
    // Failed to get decent sample format
    if(cph->format.sfmt == SND_PCM_SFMT_S16_LE)
      break;
  }

  DI_error(DI_ERR_ARGUMENT, "DI_AudioPort::alsaSeekSampleFormat", 
	   snd_strerror(e));

  return false;
}

int DI_AudioPort::alsaDownGradeFormat(int alsaSampleFormat)
  {
  // return alsaSampleFormat - 1;
  int f;

  switch(alsaSampleFormat) {
  case SND_PCM_SFMT_FLOAT64_LE:
    f = SND_PCM_SFMT_S32_LE;
    // f =  SND_PCM_SFMT_FLOAT_LE;
    break;
  case SND_PCM_SFMT_FLOAT_LE:
    f = SND_PCM_SFMT_S32_LE;
    break;

  case SND_PCM_SFMT_S32_LE:
    f = SND_PCM_SFMT_S16_LE;
    break;
    /* case SND_PCM_SFMT_S24_LE: 
    f = SND_PCM_SFMT_S16_LE;
    break;*/
    
  default:
    f = SND_PCM_SFMT_S16_LE;
  }
  return f;
}

int DI_AudioPort::alsaBytesPerSample(int alsaSampleFormat)
{
  int f;

  switch(alsaSampleFormat) {
  case SND_PCM_SFMT_FLOAT64_LE:
    f = 8;
    break;
  case SND_PCM_SFMT_FLOAT_LE:
  case SND_PCM_SFMT_S32_LE:
    // case SND_PCM_SFMT_S24_LE:
    f = 4;
    break;
  case SND_PCM_SFMT_S16_LE:
    f = 2;
    break;
  default:
    f = 0;
    break;
  }
  return f;
}


// --- Predicates -------------------------------------------------------------

bool DI_AudioPort::isOK(void) const
{
  return true;
}

/** Allocate memory buffer and set "certain" class members. */

bool DI_AudioPort::allocBuffer(const snd_pcm_params_t *params, 
			       int maxBlock)
{
  m_alsaSample = params->format.sfmt;
  m_bytesPerSample = alsaBytesPerSample(m_alsaSample);


  int bufferBytes = m_bytesPerSample * params->format.channels * maxBlock;
  
  if(!bufferBytes) {
    DI_error(DI_ERR_INTERNAL, "DI_AudioPort::allocBuffer");
    return false;
  }

  m_tmpBuffer = (char *) malloc(bufferBytes);
  m_tmpBufferFrames = maxBlock;
  DI_trace("DI_AudioPort::allocBuffer", bufferBytes);
  return true;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

DI_AudioDevice::DI_AudioDevice(ulong id,
			       bool input,
			       ulong channels,
			       ulong rate)
{
  DI_IoMode mode;

  if(input) mode = DI_IO_INPUT;
  else mode = DI_IO_OUTPUT;

  char buf[DI_AUDIO_MAX_NAME_LENGTH];

  alsaDupCardName(id, buf);
  
  if(!rate) rate = 44100;

  m_ok = m_port.open(buf, mode, DI_ASF_INT16, 1024 * 4, rate, channels, 1024);
}
  
DI_AudioDevice::DI_AudioDevice(const char *name,
			       bool input,
			       ulong channels,
			       ulong rate)
{
  DI_IoMode mode;

  if(input) mode = DI_IO_INPUT;
  else mode = DI_IO_OUTPUT;

  if(!rate) rate = 44100;

  m_ok = m_port.open(name, mode, DI_ASF_INT16, 1024 * 4, rate, channels, 1024);
  if(m_ok) m_port.start();
}
  
DI_AudioDevice::~DI_AudioDevice()
{
  m_port.close();
}

ulong DI_AudioDevice::getChannels(void)
{
  if(!m_ok) return 0;

  switch(m_port.mode()) {
  case DI_IO_INPUT:
  case DI_IO_INPUT_OUTPUT:
    return m_port.inChannels();
  case DI_IO_OUTPUT:
    return m_port.outChannels();
  case DI_IO_NONE:
    break;
  }
  return 0;
}



void DI_AudioDevice::getDeviceLabel(ulong devID, char *label)
{
  char *tmp(0);

  int e = snd_card_get_longname(devID, &tmp);

  if(e) return;

  if(!tmp || strlen(tmp) >= DI_AUDIO_MAX_NAME_LENGTH)
    return;
  
  strcpy(label, tmp);

  // free(tmp);
}


int DI_AudioDevice::alsaNumberOfCards()
{
  return snd_cards();
}

/** Copies name of the sound card to argument name (allocates new
    memory). On success, returns zero, otherwise an error. */

int DI_AudioDevice::alsaGetCardName(int card, char **name)
{
  return snd_card_get_name(card, name);
}

bool DI_AudioDevice::alsaDupCardName(int card, char *name)
{
  char * tmp(0);

  int e = alsaGetCardName(card, & tmp);
  if(e) return false;

  if(!tmp || strlen(tmp) >= DI_AUDIO_MAX_NAME_LENGTH)
    return false;

  strcpy(name, tmp);
  
  // free(tmp);

  return true;
}
