/* COPYRIGHT
 *
 * This file is part of the DIVA IO library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000.
 * Jarmo.Hiipakka@hut.fi Tommi.Ilmonen@hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tcm.hut.fi/Research/DIVA/
 *
 * DIVA IO library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */

//
// DIVA Library. Copyright (C) Jarmo Hiipakka and Tommi Ilmonen, 1999-2001.
//
// $Log: di_audio_port.h,v $
// Revision 1.1.1.1  2001/06/04 09:04:19  tilmonen
// Mustajuuri 4.6.2001, initial CVS cersion
//

#ifndef DI_AUDIO_PORT_H
#define DI_AUDIO_PORT_H

extern "C" {
#include <sys/asoundlib.h>
}

#include <stropts.h>
#include <poll.h>

#include "di_audio_base.h"

#define DI_AUDIO_IS_ALSA 1

/// Maximum length of audio library names and labels.
const ulong DI_AUDIO_MAX_NAME_LENGTH = 128;

/// Audio library negative infinity gain.
const double DI_AUDIO_NEG_INFINITY = -10000.0;


/** Audio signal I/O for ALSA. The ALSA wrappers are still rather
    untested. They seem to work on Liunx-2.2.16 with ALSA
    sriver/library 0.5.8 and via686a driver. No others have been
    tested to date. 

    This ALSA wrapper is still rather experimental.

    \b Note: ALSA requires that samples be written to the port in
    chunks of 2^n (or something like this). Thus writing 30 sample
    frames at a time is going to be problem, but writing 32 samples at
    a time is 100 % ok.

    DI_AudioPort is declared in "di_audio_port.h" and defined in libdiio.
    
*/

class DI_AudioPort : public DI_AudioBase
{
public:

  /**@name Constructors */
  
  //@{

  /** Create a new audio port.
    
      This constructor creates a new audio port. The port is unconnected
      and is made operational only by using the open method. It is
      recommended that the port is opened only when it is needed, and
      closed when not needed anymore.
  */

  DI_AudioPort();
  
  ///

  ~DI_AudioPort(void);

  //@}
  
  /**@name Access */

  //@{

  /// Return the audio input device name.

  const char *inName(void) const { return (const char *)m_inName; }

  /// Return the audio output device name.

  const char *outName(void) const { return (const char *)m_outName; }

  /// Return the audio input device label.

  const char *inLabel(void) const { return (const char *)m_inLabel; }

  /// Return the audio output device label.

  const char *outLabel(void) const { return (const char *)m_outLabel; }

  
  //@}

  /**@name Operations */
  
  //@{

  /** Open the audio port.

      This method opens an actual audio port to a named audio device.
      The audio port may be opened for recording or playback or for both.
      However, the supported modes may differ on different platforms.

      The 'bufferSize' parameter sets the audio queue buffer size
      (in sample frames) that  will be used as a temporary data storage
      during recording or playing.

      The 'maxBlock' parameters sets the maximum block size (in sample frames)
      that can be read or written on a single call to either readFrames or
      writeFrames.

  */
  
  bool open(/// Audio device.
	    const char *device,
	    /// IO mode: none, input, output, or input&output.
	    DI_IoMode mode,
	    /// Audio port sample format.
	    DI_AudioSampleFormat userSample, 
	    /// Audio port buffer size (sample frames).
	    ulong bufferSize,
	    /// Audio sampling rate.
	    ulong samplerate,
	    /// Number of channels.
	    ulong channels,
	    /// Maximum block size.
	    ulong maxBlock,
	    bool interleave = true);


  /// Close the port.

  bool close(void);

  /** Start recording/playback. On IRIX and OSS this method does
      nothing, on ALSA it is compulsory. */

  bool start();


  //@}

  /**@name I/O operations */
  
  //@{

  /** Read sample frames recorded by audio hardware.

      This method reads 'frames' sample frames from audio hardware recording
      device, and stores them into buffer pointed by 'buffer'.

      Please note that one sample may take more than one byte, so
      'buffer' must point to a buffer of capable of holding 'frames'
      * N * M bytes, where N is the size of one sample and M is number of
      channels.
     
      The number of sample frames actually recorded is
      returned, or a negative value if the operation failed.

  */

  int readFrames(void *buffer, ulong frames);

  int readBuffers(void **buffers, ulong frames);

  /** Write sample frames to be played by audio hardware.

      This method writes 'frames' sample frames to audio hardware playback
      device, reading them from buffer pointer by 'buffer'. The number
      of sample frames actually played is returned, or a negative value if
      the operation failed. NOTE! This method may change the contents of
      the buffer.
      
      writeFrames() may, and usually does, return before the playback
      sequence is completed, storing the sample data into a temporary
      buffer and leaving the playback subsystem running after write()
      returns. In these cases, the returned value shows how many sample frames
      the the audio subsystem is going to try to play.

  */

  int writeFrames(const void *buffer,ulong frames);

  int writeBuffers(const void **buffers, ulong frames);

  /// Get number of filled sample frames in the input sample buffer

  int inFilledFrames(void) const;
  
  /** Get number of filled sample frames in the output sample buffer. 

      Also checks the number of underflows in the buffer.
   */

  int outFilledFrames(void) const;

  /** Wait until 'frames' # of frames can be read.
      
      Return value tells how many frames are filled as the call
      returns. Negative values indicate error of some sort.

  */
  
  int waitInputFilled(ulong frames, int timeOut = -1);

  /// Wait until output has only the number of frames specified filled.  

  int waitOutputFilled(ulong frames, int timeOut = -1);

  //@}

  /**@name ALSA specials */
  
  //@{

  uint underFlows() { return m_underFlows; }

  void alsaDescribePlaybackSetup(char *buf);
  void alsaDescribeChannelSetup(char *buf, snd_pcm_info_t *info);

  /** Try to get a decent sample format based on the one in params. 

      Works by downgrading the format until the driver accepts the format.
   */

  bool alsaSeekSampleFormat(snd_pcm_t *, snd_pcm_params_t *);

  static int alsaDownGradeFormat(int alsaSampleFormat);

  static int alsaBytesPerSample(int alsaSampleFormat);

  //@}

  /**@name Predicates */
  
  //@{

  ///

  bool isOK(void) const;
  
  ///
  
  bool isEOF(void) const { return false; }

  //@}

protected:
  
  bool allocBuffer(const snd_pcm_params_t *params, 
		   int maxBlock);


  inline DI_AudioPort &operator = (const DI_AudioPort &xOriginal)
  { (void)xOriginal; return *this; }

  snd_pcm_t            *m_phandle;
  snd_pcm_t            *m_chandle;

  // snd_pcm_channel_status_t m_channelStatus;

  snd_pcm_params_t      m_params; // Playback
  snd_pcm_status_t      m_status;
  snd_pcm_setup_t       m_setup;

  char                 *m_tmpBuffer;
  uint                  m_tmpBufferFrames;
  uint                  m_bytesPerSample;

  int                   m_card;
  int                   m_device;

  /// Alsa mode.
  int                   m_amode; 
  int                   m_alsaSample;
  DI_AudioSampleFormat  m_userSample;

  uint                  m_underFlows;

  DI_IoMode             m_mode;

  pollfd                m_inPFD;
  pollfd                m_outPFD;
  bool                  m_deleteDev;
  // bool                  m_rmeHacks;

  char                  m_inName[DI_AUDIO_MAX_NAME_LENGTH];
  char                  m_outName[DI_AUDIO_MAX_NAME_LENGTH];
  char                  m_inLabel[DI_AUDIO_MAX_NAME_LENGTH];
  char                  m_outLabel[DI_AUDIO_MAX_NAME_LENGTH];
  
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
/** Audio hardware setup.

    DI_AudioDevice is declared in "di_audio_port.h" and defined in divaio.

    <B> Very experimental - may hang your driver !! </B>
    
    see: DI_AudioPort
*/

class DI_AudioDevice
{
  friend class DI_AudioPort;

public:

  /** Open audio control device. */
  DI_AudioDevice(/// ID of the desired device.
		 ulong ID,
		 /// Create input (true) or output (false) device.
		 bool input,
		 /// Number of device audio channels.
		 ulong channels,
		 /// Device sampling rate (defaults to current rate).
		 ulong rate=0);

  /** Open audio control device. */
  DI_AudioDevice(/// Name of the desired device.
		 const char *name,
		 /// Create input (true) or output (false) device.
		 bool input,
		 /// Number of device audio channels.
		 ulong channels,
		 /// Device sampling rate (defaults to current rate).
		 ulong rate=0);


  ~DI_AudioDevice();

  bool isOK() { return m_ok; }
  
  ulong getChannels(void);

  /**@name Utilities */

  //@{

  /// Return the number of audio subsystem input devices.

  static ulong getNumberOfInputDevices() { return alsaNumberOfCards(); }

  /// Return the number of audio subsystem output devices.

  static ulong getNumberOfOutputDevices() { return alsaNumberOfCards(); }

  
  static void getInputDevices(ulong *IDArray)
    {
      ulong cards = alsaNumberOfCards();

      for(ulong i=0; i < cards; i++)
	IDArray[i] = i;
    }

  static void getOutputDevices(ulong *IDArray)
    { getInputDevices(IDArray); }

  static void getInputDeviceName(ulong devID, char *name)
    { alsaDupCardName(devID, name); }

  static void getOutputDeviceName(ulong devID, char *name)
    { alsaDupCardName(devID, name); }

  static void getDeviceLabel(ulong devID, char *label);
  //@}
  
  /**@name ALSA specific methods */

  //@{

  static int alsaNumberOfCards();

  static int alsaGetCardName(int card, char **name);

  static bool alsaDupCardName(int, char *);

  //@}

protected:
  
  DI_AudioPort m_port;

  bool m_ok;
};

#endif  // DI_AUDIO_PORT_H
