/* COPYRIGHT
 *
 * This file is part of Mustajuuri.
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * http://www.tcm.hut.fi/~tilmonen/mustajuuri/
 *
 * Mustajuuri is licensed under GNU Lesser General Public License
 * (LGPL). The LGPL conditions can be found in file "LGPL" that is
 * distributed with this source package. The Mustajuuri license
 * interpretation that further clarifes LGPL can be found in file
 * "MUSTAJUURI_LICENSE" that is distributed with the Mustajuuri source code.
 *    
 */

#include "mj_debug.h"
#include "mj_dsp_plugin.h"
#include "mj_error.h"
#include "mj_fpu.h"
#include "mj_global.h"
#include "mj_kernel.h"
#include "mj_knob_test.h"
#include "mj_level_display.h"
#include "mj_menu_test.h"
#include "mj_midi_sequence.h"
#include "mj_misc.h"
#include "mj_module_dsp.h"
#include "mj_mutex.h"
#include "mj_prefs.h"
#include "mj_root_powers.h"
#include "mj_run_status.h"
#include "mj_thread.h"
#include "mj_version.h"

#include <qapplication.h>
#include <qfileinfo.h> 

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <unistd.h>

#ifdef HAVE_PCAPS
#include <sys/capability.h>
#endif

extern void MJ_FilterTest();

enum MJ_CmdLineOption {   
  MJ_CMDL_AUTOSTART, 
  MJ_CMDL_HELP, 
  MJ_CMDL_DEBUG, 
  MJ_CMDL_KERNEL_RUN,
  MJ_CMDL_MINUS,
  MJ_CMDL_NO_PRIORITY,
  MJ_CMDL_TEST_FIELDS, 
  MJ_CMDL_TEST_FILTER, 
  MJ_CMDL_TEST_FPU, 
  MJ_CMDL_TEST_IO_DEFS, 
  MJ_CMDL_TEST_LCD_CLOCK, 
  MJ_CMDL_TEST_LEVEL_QG, 
  MJ_CMDL_TEST_KNOB, 
  MJ_CMDL_TEST_MENU, 
  MJ_CMDL_TEST_MIDI_FILE, 
  MJ_CMDL_TEST_MUTEX, 
  MJ_CMDL_TEST_PLUGIN_LOAD, 
  MJ_CMDL_TEST_ROOT_POWERS, 
  MJ_CMDL_TEST_TIMING, 
  MJ_CMDL_TEST_THREAD, 
  MJ_CMDL_TEST_VERSION, 
  MJ_CMDL_SIZEOF  
};

static const char *LC_cpaCmdLineOptions[MJ_CMDL_SIZEOF] = {
  "--autostart",
  "-h",
  "--debug",
  "--kernel_run",
  "-",
  "--no_priority",
  "--test_fields",
  "--test_filter",
  "--test_fpu",
  "--test_io_defs",
  "--test_lcd_clock",
  "--test_level_qg",
  "--test_knob",
  "--test_menu",
  "--test_midi_file",
  "--test_mutex",
  "--test_plugin_load",
  "--test_root_powers",
  "--test_timing",
  "--test_thread",
  "--test_version"
};

void MJ_help(const char *programName, 
	     const char *badOption = 0,
	     bool exit_now = true)
{
  if(badOption) 
    fprintf(stderr, "%s : bad option: \"%s\"\n", programName, badOption);

  fprintf(stderr, "Usage: %s [options]\nOptions:\n", programName);

  fflush(0);
  
  for(int i=0; i < MJ_CMDL_SIZEOF; i++) {
    fprintf(stderr, "%s\n", LC_cpaCmdLineOptions[i]);
  }

  fflush(0);

  if(exit_now) exit(0);
}

/* Get necessary privileges and drop all others.  This function is
   defined here so it would *not* be dynamically linked. */

void MJ_getPrivileges(int argc, char **argv)
{
  const char * fname = "MJ_getPrivileges";

  bool mlock(true);

  for(int option = 0; option < argc; option++) {
    
    const char *cpOpt = argv[option];
    MJ_CmdLineOption o = 
      (MJ_CmdLineOption) MJ_findStr(cpOpt, 
				    LC_cpaCmdLineOptions, 
				    MJ_CMDL_SIZEOF);
    switch(o) {
      
    case MJ_CMDL_NO_PRIORITY: // Already handled

      mlock = false;
      break;
    default:
      break;
    }

  }

  if(!mlock) return;

  if(!mlockall(MCL_CURRENT | MCL_FUTURE)) {
    // We seem to have super-high priorities :-)
    MJ_trace(fname, "mlockall successful");
  }
  else
    errno = 0;

  uid_t uid = getuid();
  uid_t euid = geteuid();

  // Get the necessary capabilities:
#ifdef HAVE_PCAPS

  // Rename the capabilites
#ifdef __sgi

#define MJ_CAPS 1
  // With this capability system threads are enabled.
#define MJ_CAP_SCHED CAP_SCHED_MGT
  cap_value_t capList[MJ_CAPS] = { MJ_CAP_SCHED };

#else

#define MJ_CAPS 2
#define MJ_CAP_SCHED CAP_SYS_NICE
  cap_value_t capList[MJ_CAPS] = { MJ_CAP_SCHED, CAP_SYS_RESOURCE} ;

#endif // __sgi

  cap_t caps;
  
  caps = cap_init();
  cap_clear(caps);

  cap_set_flag(caps, CAP_EFFECTIVE, MJ_CAPS, capList , CAP_SET);
  cap_set_flag(caps, CAP_INHERITABLE, MJ_CAPS, capList , CAP_SET);
  cap_set_flag(caps, CAP_PERMITTED, MJ_CAPS, capList , CAP_SET);

  char * capstr = cap_to_text(caps, 0);
  MJ_trace(fname, capstr);

  capstr = cap_to_text(cap_get_proc(), 0);
  MJ_trace(fname, capstr);

  int e = cap_set_proc(caps);
  if(!e) {
    MJ_trace(fname, "cap_set_proc successfull !!");
  }
  
  // cap_free(& caps);
  // free(capstr);

  // Drop setuid root.
  if(uid != euid) {
    if(!setuid(uid))
      MJ_trace(fname, "setuid successfull !!");
  }

  capstr = cap_to_text(cap_get_proc(), 0);
  MJ_trace(fname, capstr);

#else  // HAVE_PCAPS

  if(uid != euid) {
    MJ_trace(fname, "WARNING !!! Mustajuuri is setuid,\
 but not compiled with capabilities - WILL KEEP ROOT UID !!!");
  }
#endif // HAVE_PCAPS
  
  errno = 0;
  
  // putenv("LD_LIBRARY_PATH=/usr/lib");
  // MJ_trace("LD_LIBRARY_PATH", getenv("LD_LIBRARY_PATH"));

  QFile file("/root/foo");
  bool root = file.open(IO_WriteOnly);
  if(root)
    MJ_trace(fname, "Managed to open /root/foo !!!");

}
  
int main(int argc, char **argv)
{
  int option, tmp;
  bool autostart = false;

  /* printf("FOO is: %x", (0xab00 | 0x12));
     fflush(0); */
  
  MJ_getPrivileges(argc, argv);

  QApplication qa(argc, argv);
  QFileInfo fi;

  MJ_prefs.scanArgs(argc, argv);

  if(!MJ_prefs.allowFpe())
    MJ_Fpu::defaultOperation();

  for(option = 1; option < argc; option++) {
    const char *cpOpt = argv[option];
    MJ_CmdLineOption o = 
      (MJ_CmdLineOption) MJ_findStr(cpOpt, 
				    LC_cpaCmdLineOptions, 
				    MJ_CMDL_SIZEOF);
    
    switch(o) {
    case MJ_CMDL_AUTOSTART:
      autostart = true;
      break;

    case MJ_CMDL_HELP:
      MJ_help(argv[0], 0, true);
      break;
      
    case MJ_CMDL_DEBUG:
      if(option + 1 == argc) MJ_Debug::debugType(MJ_DEBUG_GENERIC);
      else if(argv[option+1][0] == '-') {
	MJ_Debug::debugType(MJ_DEBUG_GENERIC);
      }
      else {
	tmp = 1;
	
	while (++option < argc && tmp) {
	  
	  if(argv[option][0] == '-') { // If this is a new option, break loop.
	    
	    option -= 2;
	    tmp = 0;
	  }
	  else MJ_Debug::debugName(argv[option]);
	}
      }
      break;

    case MJ_CMDL_KERNEL_RUN:
      MJ_KernelGui::run(&qa, argc, argv, autostart);
#ifdef MJ_MEM_DEBUG
      MJ_memDebugReport();
#endif
      return 0;

    case MJ_CMDL_NO_PRIORITY: // Already handled 
    case MJ_CMDL_MINUS: // Null argument
      break;

    case MJ_CMDL_TEST_FIELDS:
      MJ_fieldTest();
      return 0;

    case MJ_CMDL_TEST_FILTER:
      MJ_FilterTest();
      return 0;

    case MJ_CMDL_TEST_FPU:
      if(++option < argc)
	MJ_Fpu::test(argv[option]);
      else
	MJ_Fpu::test(0);
      return 0;

    case MJ_CMDL_TEST_IO_DEFS:
      MJ_ModuleDsp::testIoDefRw();
      return 0;

    case MJ_CMDL_TEST_LCD_CLOCK:
      MJ_LcdClock::test(&qa);
      return 0;

    case MJ_CMDL_TEST_LEVEL_QG:
      MJ_LevelDisplay::test(&qa, argc, argv);
      return 0;

    case MJ_CMDL_TEST_KNOB:
      MJ_testKnob(&qa);
      return 0;

    case MJ_CMDL_TEST_MENU:
      MJ_MenuTest::test(&qa);
      return 0;

    case MJ_CMDL_TEST_MIDI_FILE:
      MJ_MidiSequence::test();
      return 0;

    case MJ_CMDL_TEST_MUTEX:
      MJ_Mutex::test();
      return 0;

    case MJ_CMDL_TEST_PLUGIN_LOAD:
      MJ_PluginDbeDsp::test(&qa, argc, argv);
      return 0;

    case MJ_CMDL_TEST_ROOT_POWERS:
      MJ_RootPowers::test();
      return 0;

    case MJ_CMDL_TEST_TIMING:
      MJ_RootPowers::testClock();
      return 0;

    case MJ_CMDL_TEST_THREAD:
      MJ_Thread::test();
      return 0;

    case MJ_CMDL_TEST_VERSION:
      MJ_Version::test();
      return 0;

    default:

      // Check if it appears to be a file name.
      fi.setFile(cpOpt);
      if(fi.exists())
	break;
      
      MJ_help(argv[0], cpOpt, true);
      break;
    }
  }
  
  MJ_KernelGui::run(&qa, argc, argv, autostart);

#ifdef MJ_MEM_DEBUG
  MJ_memDebugReport();
#endif

  return 0;
}
