/* COPYRIGHT
 *
 * This file is part of DIVA base library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA base library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */
#include "di_data_buffer.h"
#include "di_error.h"
#include "di_config.h"  // for WORDS_BIGENDIAN definition
#include "di_types.h"

#ifndef WIN32
#include <di_types.h>
#include <netinet/in.h> // for ntohl() etc.
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>

#ifndef WORDS_BIGENDIAN

static ullong ntohll(ullong x)
{
  return ((x << 56) | ((x & 0xff00) << 40) | ((x & 0xff0000) << 24) |
	  ((x & 0xff000000) << 8) | ((x >> 8) & 0xff000000) |
	  ((x >> 24) & 0xff0000) | ((x >> 40) & 0xff00) | (x >> 56));
}
#define htonll(x)    ntohll(x)

#else

#define ntohll(x)    (x)
#define htonll(x)    (x)

#endif  // not defined WORDS_BIGENDIAN

// -- Constructors ------------------------------------------------------------

DI_DataBuffer::DI_DataBuffer(unsigned int size) :
  m_buf(NULL),
  m_size(size+4), // There must be room for the length field.
  m_index(0)
{
  resize(m_size);
  writeInt32(0);
}

DI_DataBuffer::~DI_DataBuffer()
{
  free(m_buf);
}

void DI_DataBuffer::writeInt32(int i)
{
  if(!writeNetworkOrder(&i, 4)) failIn("writeInt32");
}



void DI_DataBuffer::writeInt64(llong i)
{
  /* if(sizeof(llong) != 8) {
     puts("sizeof(llong) != 8");
     }*/

  if(!writeNetworkOrder(&i, 8)) failIn("writeInt64");
}

void DI_DataBuffer::writeFloat32(float i)
{
  if(!writeNetworkOrder(&i, 4)) failIn("writeFloat32");
}

void DI_DataBuffer::writeFloat64(double i)
{
  if(!writeNetworkOrder(&i, 8)) failIn("writeFloat64");
}

void DI_DataBuffer::writeString(const char *str)
{
  const char * fname = "writeString";

  int len = strlen(str) + 1; // Include ending NULL
  int rlen = len + (4 - len & 0x3); // Pad to four

  // Do not accept too long strings:
  if ((len >= DI_MAX_DATA_STRING) || (m_index + rlen > m_size)) {
    failIn(fname);
    return;
  }
  
  writeInt32(len);
  memcpy(& m_buf[m_index], str, len);

  m_index += rlen;
  *(unsigned int *)m_buf = m_index - 4;
}

int DI_DataBuffer::readInt32()
{
  int i;
  if(!readNetworkOrder(&i, 4)) {
    failIn("readInt32");
    i = 0;
  }
  return i;
}

llong DI_DataBuffer::readInt64()
{
  llong i;
  if(!readNetworkOrder(&i, 8)) {
    failIn("readInt64");
    i = 0;
  }
  return i;
}

float DI_DataBuffer::readFloat32()
{
  float i;
  if(!readNetworkOrder(&i, 4)) {
    failIn("readFloat32");
    i = 0;
  }
  return i;
}

double DI_DataBuffer::readFloat64()
{
  double i;
  if(!readNetworkOrder(&i, 8)) {
    failIn("readFloat64");
    i = 0;
  }
  return i;
}

void DI_DataBuffer::readString(char *str)
{
  int len = readInt32();
  int rlen = len + (4 - len & 0x3); // Pad to four

  if(m_index + rlen  > m_size)
  {
    failIn("readString");
    str[0] = '\0';
    return;
  }

  memcpy(str, & m_buf[m_index], len);

  m_index += rlen;
}

bool DI_DataBuffer::resize(uint s)
{
  m_buf = (char *) realloc(m_buf,s ? s + 4 : 0);

  if(!s)
  {
    m_buf = 0;
    m_size = 0;
  }

  if (m_buf)
  {
    m_size = s + 4;
    return true;
  }
  else
    return false;
}

bool DI_DataBuffer::writeNetworkOrder(const void * data, uint bytes)
{
  // Pad to eight if necessary
  if (!(bytes & 0x7) && (m_index & 0x7))
    m_index += 4;

  if (m_index + bytes > m_size) return false;

  switch(bytes)
  {
   case 4:
    *(u_long *)(m_buf+m_index) = htonl(*(u_long *)data);
    break;
   case 8:
    *(ullong *)(m_buf+m_index) =
      htonll(*(ullong *)data);
    break;
   default:
    return false;
  }
  
  m_index += bytes;
  *(unsigned int *)m_buf = htonl(m_index - 4);
  return true;
}

bool DI_DataBuffer::readNetworkOrder(void * data, uint bytes)
{
  // Pad to eight if necessary
  if(!(bytes & 0x7) && (m_index & 0x7))
    m_index += 4;

  if(m_index + bytes > m_size) {
    char buf[32];
    sprintf(buf, "readNetworkOrder # m_size = %d, m_index = %d, bytes = %d",
	    m_size, m_index, bytes);
    failIn(buf);
    return false;
  }

  switch(bytes)
  {
   case 4:
    *(u_long *)data = ntohl(*(u_long *)(m_buf+m_index));
    break;
   case 8:
    *(ullong*)data =
      ntohll(*(ullong *)(m_buf+m_index));
    break;
   default:
    return false;
  }

  m_index += bytes;
  return true;
}

/// Simple error reporting to stderr.
void DI_DataBuffer::failIn(const char *str)
{
  fprintf(stderr, "DI_DataBuffer::%s # error\n", str);
  fflush(0);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

DI_DataBuffer2::DI_DataBuffer2()
  : m_buf((unsigned  char *) malloc(512)),
    m_size(512),
    m_position(0),
    m_overrun(false),
    m_littleEndian(true)
{}

DI_DataBuffer2::DI_DataBuffer2(unsigned size)
  : m_buf((unsigned  char *) malloc(size)), 
    m_size(size),
    m_position(0),
    m_overrun(false),
    m_littleEndian(true)
{}

DI_DataBuffer2::DI_DataBuffer2(const DI_DataBuffer2 &b)
  : m_buf((unsigned  char *) malloc(b.m_size)), 
    m_size(b.m_size),
    m_position(b.m_position),
    m_overrun(false),
    m_littleEndian(b.m_littleEndian)
{
  memcpy(m_buf, b.m_buf, b.m_size);
}

DI_DataBuffer2::~DI_DataBuffer2()
{
  free(m_buf);
}
  
void DI_DataBuffer2::resize(unsigned size)
{
  if(size == m_size) return;

  m_buf = (unsigned char *) realloc(m_buf, size);
  m_size = size;

  m_overrun = (m_position > m_size);

  if(!m_buf && size) {
    DI_fatal2(DI_ERR_MEMORY, "DI_DataBuffer2::resize # %u", size);
  }
}

void DI_DataBuffer2::writeByte(char v)
{
  if(m_position + 1 > m_size)
    resize(m_size + 128);

  m_buf[m_position++] = v;
}

/** Writes a vector of bytes to the buffer.*/

void DI_DataBuffer2::writeBytes(const void *ptr, unsigned n)
{
  if(m_position + n > m_size)
    resize(m_size + n + 128);
  
  memcpy(m_buf + m_position, ptr, n);
  m_position += n;
}

/** Reads a vector of bytes from the stream. If there are not enough
    bytes to be read, then nothing is read and the buffer is filled
    with zero.  */
void DI_DataBuffer2::readBytes(void *ptr, unsigned n)
{
  if(m_position + n > m_size) {
    readErr();
    bzero(ptr, n);
    return;
  }
  memcpy(ptr, m_buf + m_position, n);
  m_position += n;
}

void DI_DataBuffer2::writeString(const char *str)
{
  // String length including the terminating NULL.
  int l = str ? strlen(str) + 1: 0;

  if(m_position + l + 4 > m_size)
    resize(m_size + l + 128);

  write4(l);

  if(l) {
    memcpy(m_buf + m_position, str, l);
    m_position += l;
  }
}

void DI_DataBuffer2::writeString(const std::string *str)
{
  // String length including the terminating NULL.
  
  int l = str->length();

  if(m_position + l + 4 > m_size)
    resize(m_size + l + 128);

  write4(l);

  if(l) {
    memcpy(m_buf + m_position, & * str->begin(), l);
    m_position += l;
  }
}

char  DI_DataBuffer2::readByte()
{
  if(m_position + 1 > m_size) {
    readErr();
    return 0;
  }
  
  return m_buf[m_position++];
}

/** Reads a string from the buffer. The target string must be large
    enough to hold all the necessary data. */
void DI_DataBuffer2::readString(char *str)
{
  int l = readInt32();

  if(l) {
    memcpy(str, m_buf + m_position, l);
    m_position += l;
  }
  else
    str[0] = '\0';
}

/** Reads a string to an STL string. */
void DI_DataBuffer2::readString(std::string *str)
{
  int l = readInt32();

  if(l) {

    if(m_position + l > m_size) {
      readErr();
      return;
    }

    str->resize(l);
    memcpy(& * str->begin(), m_buf + m_position, l);
    m_position += l;
  }
  else
    str->resize(0);
}

void DI_DataBuffer2::write2(unsigned short source)
{
  if(m_position + 2 > m_size)
    resize(m_size + 128);

  unsigned char * target = m_buf + m_position;

  m_position += 2;

  if(m_littleEndian) {
    *target++ = source;
    *target   = source >> 8;
  }
  else {
    *target++ = source >> 8;
    *target   = source;

  }
}

void DI_DataBuffer2::write4(unsigned int source)
{
  if(m_position + 4 > m_size)
    resize(m_size + 128);

  unsigned char * target = m_buf + m_position;

  m_position += 4;

  if(m_littleEndian) {
    *target++ = source;
    *target++ = source >> 8;
    *target++ = source >> 16;
    *target   = source >> 24;
  }
  else {
    *target++ = source >> 24;
    *target++ = source >> 16;
    *target++ = source >> 8;
    *target   = source;
  }
}

void DI_DataBuffer2::write8(ullong source)
{
  if(m_position + 8 > m_size)
    resize(m_size + 128);

  unsigned char * target = m_buf + m_position;
  
  m_position += 8;
  
  if(m_littleEndian) {
    *target++ = source;
    *target++ = source >> 8;
    *target++ = source >> 16;
    *target++ = source >> 24;
    *target++ = source >> 32;
    *target++ = source >> 40;
    *target++ = source >> 48;
    *target   = source >> 56;
  }
  else {
    *target++ = source >> 56;
    *target++ = source >> 48;
    *target++ = source >> 40;
    *target++ = source >> 32;
    *target++ = source >> 24;
    *target++ = source >> 16;
    *target++ = source >> 8;
    *target   = source;
  }
}

short DI_DataBuffer2::read2()
{
  if(m_position + 2 > m_size) {
    readErr();
    return 0;
  }

  unsigned const char * s = m_buf + m_position;
  
  m_position += 2;
  
  if(m_littleEndian)
    return s[0] + (s[1] << 8);
  else
    return s[1] + (s[0] << 8);
    
}

int DI_DataBuffer2::read4()
{
  if(m_position + 4 > m_size) {
    readErr();
    return 0;
  }

  unsigned const char * s = m_buf + m_position;
  
  m_position += 4;
  
  if(m_littleEndian)
    return s[0] + (s[1] << 8) + (s[2] << 16) + (s[3] << 24);
  else
    return s[3] + (s[2] << 8) + (s[1] << 16) + (s[0] << 24);
}

llong DI_DataBuffer2::read8()
{
  typedef llong ll;

  if(m_position + 8 > m_size) {
    readErr();
    return 0;
  }

  unsigned const char * s = m_buf + m_position;
  
  m_position += 8;
  
  if(m_littleEndian)
    return (ll) s[0] + ((ll) s[1] << 8) + 
      ((ll) s[2] << 16) + ((ll) s[3] << 24) + 
      ((ll) s[4] << 32) + ((ll) s[5] << 40) + 
      ((ll) s[6] << 48) + ((ll) s[7] << 56);
  else
    return (ll) s[7] + ((ll) s[6] << 8) + 
      ((ll) s[5] << 16) + ((ll) s[4] << 24) + 
      ((ll) s[3] << 32) + ((ll) s[2] << 40) + 
      ((ll) s[1] << 48) + ((ll) s[0] << 56);
}

/** Load a file into memory. */
bool DI_DataBuffer2::loadFile(const char *filename)
{
  FILE *file = fopen(filename, "rb");

  if(!file) return false;

  fseek(file,0,SEEK_END);

  int size = ftell(file);
  fseek(file,0,SEEK_SET);
  resize(size);
  bool ok = fread(m_buf,1,size,file) == 1;
  fclose(file);

  return ok;
}

int DI_DataBuffer2::write(FILE *fp, bool flush)
{
  int tmp = positionLE();
  int written = fwrite(&tmp, 4, 1, fp);

  if(written != 1) return 0;

  written = fwrite(m_buf, position(), 1, fp);

  if(flush) fflush(fp);
  
  if(written != 1) return 0;

  return position() + 4;
}

int DI_DataBuffer2::read(FILE *fp)
{
  int tmp;
  int r = fread(&tmp, 4, 1, fp);

  if(r != 1) return 0;

  setPositionLE(tmp);
  expandTo(position());
  r = fread(m_buf, position(), 1, fp);

  if(r != 1) return 0;

  return position() + 4;
}

DI_DataBuffer2 & DI_DataBuffer2::operator = (const DI_DataBuffer2 &b)
{
  m_buf = (unsigned char *) realloc(m_buf, b.m_size); 
  m_size = b.m_size;
  m_position = b.m_position;
  memcpy(m_buf, b.m_buf, b.m_size);
  return *this;
}

void DI_DataBuffer2::readErr()
{
  DI_error(DI_ERR_UNKNOWN, 
	   "DI_DataBuffer2 # Tried to read past the end of the buffer");
  m_position = m_size;
  m_overrun = true;
}

int DI_match(const char * str, const char ** strs, uint nstrs)
{
  for(uint i = 0; i < nstrs; i++) 
    if(strcmp(str, strs[i]) == 0)
      return i;

  return -1;
}



