/* COPYRIGHT
 *
 * This file is part of DIVA base library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA base library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */

#include <di_error.h>
#include <di_mutex.h>

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

void (*DI_errorHandle) (DI_ErrorLevel, const char *) = 0;

const char *DI_ErrorLevels[DI_ERRL_SIZEOF] = {
  "Trace",
  "Information",
  "Warning",
  "Error",
  "Fatal error"
};

const char *DI_ErrorCodes[DI_ERR_SIZEOF] = {
  "No error",
  "Unkown error",
  "Out of memory",
  "Data corrupted",
  "Internal error",
  "System error",
  "Array size exceeded",
  "List size exceeded",
  "Bad argument",
  "NULL argument",
  "Bad match",
  "End of file",
  "Resource denied",
  "Virtual function called",
  "Do not call this function",
  "Not implemented yet",
  "Not initialized yet",
  "Real-time performance failure",
  "Floating point error",
  "CPU load too high",
  "Numeric overflow",
  "Input problem",
  "Output problem"
};

static DI_MutexAuto DI_FailureDiagnose_Mutex(true, false);

FILE *DI_traceFile = stdout;
bool  DI_errorsToTraceFile = false;

const char *DI_errorLevelStr(DI_ErrorLevel level)
{
  if(level < DI_ERRL_SIZEOF) return DI_ErrorLevels[level];
  return "Bad error level";
}

const char *DI_errorStr(DI_ErrorCode code)
{
  if(code < DI_ERR_SIZEOF) return DI_ErrorCodes[code];
  return "Bad error type";
}

FILE * DI_getErrorStream(DI_ErrorLevel level)
{
  FILE *stream = DI_traceFile;

  if(DI_errorsToTraceFile || level == DI_ERRL_TRACE) return stream;
  if((level == DI_ERRL_ERROR) || (level == DI_ERRL_FATAL))
    return stderr;

  return stdout;
}

extern void DI_failure(DI_ErrorLevel level,
		       DI_ErrorCode  code, 
		       const char *c1, const char *c2, const char *c3)
{
  char buf[8196];
  char * ptr = buf;

  ptr += sprintf(ptr, DI_errorLevelStr(level));
  
  
  if(level != DI_ERRL_TRACE)
    ptr += sprintf(ptr, " # %s ", DI_errorStr(code));

  if(c1) ptr += sprintf(ptr, " # %s", c1);
  if(c2) ptr +=  sprintf(ptr, " # %s", c2);
  if(c3) ptr += sprintf(ptr, " # %s", c3);

  sprintf(ptr, "\n");

  DI_FailureDiagnose_Mutex.lock();

  FILE *s = DI_getErrorStream(level);
  fprintf(s, buf);
  fflush(s);

  DI_FailureDiagnose_Mutex.unlock();

  if(DI_errorHandle)
    (*DI_errorHandle) (level, buf);
}

void DI_warning(DI_ErrorCode  code, const char *c1, const char *c2)
{
  DI_failure(DI_ERRL_ERROR, code, c1, c2);
}

void DI_error(DI_ErrorCode  code, const char *c1, const char *c2)
{
  DI_failure(DI_ERRL_ERROR, code, c1, c2);
}

void DI_error(DI_ErrorCode  code, const char *c1, 
	      const char *c2, 
	      const char *c3)
{
  DI_failure(DI_ERRL_ERROR, code, c1, c2, c3);
}

void DI_error2(DI_ErrorCode  code, const char *msg, ...)
{
  char buf[8196];
  va_list args;
  va_start(args, msg);
  vsprintf(buf, msg, args);
  DI_error(code, buf);
  va_end(args);
}

void DI_fatal(DI_ErrorCode  code, const char *c1, const char *c2)
{
  DI_failure(DI_ERRL_FATAL, code, c1, c2);
  exit(0);
}

void DI_fatal2(DI_ErrorCode  code, const char *msg, ...)
{
  char buf[8196];
  va_list args;
  va_start(args, msg);
  vsprintf(buf, msg, args);
  DI_fatal(code, buf);
  va_end(args);
}

void DI_trace(DI_ErrorCode  code, const char *c1, const char *c2)
{
  DI_failure(DI_ERRL_TRACE, code, c1, c2);
}

void DI_trace(const char *c1, const char *c2)
{
  DI_trace(DI_ERR_NONE, c1, c2);
}

void DI_trace(const char *c1, const char *c2, int i)
{
  char * tmp = (char *) malloc(strlen(c1) + strlen(c2) + 64);
  sprintf(tmp, "%s # %s # %d", c1, c2, i);
  DI_trace(DI_ERR_NONE, tmp);
  free(tmp);
}

void DI_trace(const char *c1, int i)
{
  char buf[128];
  
  sprintf(buf, "%d", i);
  DI_trace(c1, buf);
}

void DI_trace2(const char *msg, ...)
{
  char buf[8196];
  va_list args;
  va_start(args, msg);
  vsprintf(buf, msg, args);
  DI_trace(buf);
  va_end(args);
}
