/* COPYRIGHT
 *
 * This file is part of DIVA base library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA base library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */

#include "di_midi_event.h"

#include <di_data_buffer.h>

#include <math.h>
#include <string.h>
#include <stdio.h>

const char *DI_MidiEvent::m_label= "DI_MIDI_EVENT";

DI_MidiEvent::DI_MidiEvent(const DI_MidiEvent &e)
  : m_sysexData(0),
    m_sysexSize(0)
{
  *this = e;
}

bool DI_MidiEvent::setSysex(const char *data, unsigned l)
{
  if(m_sysexSize != l) {
  
    m_sysexData = (unsigned char *) realloc(m_sysexData, l);
    m_sysexSize = l;
  }

  if(data) memcpy(m_sysexData, data, l);
  return true;
}



void DI_MidiEvent::clear()
{
  m_msg = 0;
  free(m_sysexData);
  m_sysexData = 0;
  m_sysexSize = 0;
}


DI_MidiEvent& DI_MidiEvent::operator = (const DI_MidiEvent &e)
{
  if(m_sysexData) {
    free(m_sysexData);
    m_sysexData = 0;
    m_sysexSize = 0;
  }

  m_ticks = e.m_ticks;
  m_time  = e.m_time;
  m_msg   = e.m_msg;

  if(e.m_sysexSize) {
    m_sysexData = (unsigned char *) malloc(e.m_sysexSize);
    if(!m_sysexData) return *this;
    memcpy(m_sysexData, e.m_sysexData, e.m_sysexSize);
    m_sysexSize = e.m_sysexSize;
  }
  return *this;
}

void DI_MidiEvent::noteOn
(unsigned char key, unsigned char velocity, unsigned char channel, unsigned ticks)
{ 
  m_ticks = ticks; 
  set(0, 0x90 | channel); 
  set(1, key); 
  set(2, velocity); 
}

void DI_MidiEvent::noteOff
(unsigned char key, unsigned char velocity, unsigned char channel, unsigned ticks)
{ 
  m_ticks = ticks; 
  set(0, 0x80 | channel); 
  set(1, key); 
  set(2, velocity); 
}

double DI_MidiEvent::keyNumberToHz(int key)
{
  double dOff = (double) (key - 69);
  double hz = pow(2.0, dOff / 12.0) * 440.0;
  return hz;
}

/** The number of significant bytes in main message. The sysex bytes
    are not included in this count.


    \return Returns a value between zero and three.
*/

int DI_MidiEvent::messageBytes()
{
  unsigned char tmp = a() & 0xF0;
  
  if((tmp == 0x80) ||  // Note off.
     (tmp == 0x90) ||  // Note on.
     (tmp == 0xA0) ||  // Polyphonic Key
     (tmp == 0xB0) ||  // Controller Change 
     (tmp == 0xE0))    // Pitch Bend
    return 3;
  else if((tmp == 0xC0) ||  // Program Change
	  (tmp == 0xD0))    // Channel Key Pressure 
    return 2;

  else if(a() == 0xF0) {  // Sysex
    return 1;
    /*
    if(b() == 0) // Two ids
      return 4;
    else
    return 2;*/
  }
  else if(a() >= 0xF8) // Only some status data
    return 1;

  return 0;
}

bool DI_MidiEvent::write(DI_DataBuffer2 *buf)
{
  buf->writeByte(a());
  buf->writeByte(b());
  buf->writeByte(c());

  buf->writeInt32(m_ticks);
  buf->writeInt64(m_time);

  
  buf->writeInt32(m_sysexSize);

  if(m_sysexSize)
    buf->writeBytes(m_sysexData, m_sysexSize);

  return true;
}

bool DI_MidiEvent::read(DI_DataBuffer2 *buf)
{
  set(0, buf->readByte());
  set(1, buf->readByte());
  set(2, buf->readByte());
  
  m_ticks = buf->readInt32();
  m_time = (unsigned long) buf->readInt64();

  int sysexSize = buf->readInt32();

  if(sysexSize) {
    resizeSysex(sysexSize);
    buf->readBytes(m_sysexData, m_sysexSize);
  }

  return true;
  
}
/* void DI_MidiEvent::write(ostream *s) const
{
  char buf[128];

  sprintf(buf, "0x %x %x %x %x", a(), b(), c(), d());

  (*s) << m_label << " " << m_ticks << " " << m_time << " " << buf;
  }*/
