/* COPYRIGHT
 *
 * This file is part of DIVA base library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA base library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */

#include "di_sleep.h"

#include "di_mutex.h"
#include "di_time_stamp.h"


/** Sleep until time indicated in the stamp object is passed.  The
    initialization time of this object is added to the stamp, before
    calculating sleeping period.  
      
    The second argument should contain a locked mutex. The mutex
    will be freed while the (potential) sleeping is taking place and
    locked again.  */
long DI_SleepSync::sleepTo(const DI_TimeStamp *stamp, DI_Mutex *mutex)
{
  struct timeval tmp;

  gettimeofday(&tmp, 0);
  m_timing.tv_sec = (long) stamp->secondsUl();
  m_timing.tv_usec = (long) stamp->subSecondsUs();

  DI_addTime(& m_timing, & m_initial);

  long diffUs = DI_timeDiffUs(&tmp, &m_timing);
  if(diffUs < 0) return diffUs;

  if(mutex) mutex->unlock();
  DI_Sleep::sleepUs(diffUs);
  if(mutex) mutex->lock();

  gettimeofday(&tmp, 0);
  return DI_timeDiffUs(&tmp, &m_timing);
}

/** Sleep in synchronous mode. The argument value is added to
    current time value. The return value tells how many microseconds
    early or late the function call returns. Zero is optimal,
    negative values indicate the call returned early and positive
    values indicate the call returned too late. */
long DI_SleepSync::sleepSynchroUs(long us)
{
  struct timeval tmp;

  DI_addTimeUs(&m_timing, us);
  gettimeofday(&tmp, 0);

  long diffUs = DI_timeDiffUs(&tmp, &m_timing);
  if(diffUs < 0) return diffUs;

  DI_Sleep::sleepUs(diffUs);
  gettimeofday(&tmp, 0);
  return DI_timeDiffUs(&tmp, &m_timing);
}
