/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of DIVA base library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000-2001.
 * Jarmo.Hiipakka@hut.fi Tommi.Ilmonen@hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA base library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */
//
// DIVA Library. Copyright (C) Jarmo Hiipakka, 1999-2001.
// DIVA Library. Copyright (C) Tommi Ilmonen,  1999-2001.
//
// $Log: di_socket.C,v $
// Revision 1.4  2002/05/22 12:41:59  tilmonen
// Changes due to port to Windows.
//
// Revision 1.3  2002/04/24 07:15:16  tilmonen
// Minor updates.
//
// Revision 1.2  2002/04/19 10:13:11  tilmonen
// Bug fixes.
//
// Revision 1.7  2002/02/22 11:28:53  tilmonen
// New (and possibly better) DI_DataBuffer2 class for binary
// data management.
//
// Revision 1.6  2002/01/18 12:18:29  tilmonen
// Cleanups.
//
// Revision 1.5  2001/10/08 12:25:53  tilmonen
// Updates.
//
// Revision 1.4  2001/07/27 08:23:17  tilmonen
// License text updates (=cosmetic changes).
//
// Revision 1.3  2001/07/26 08:49:23  tilmonen
// Updates is isPendingInput.
//
// Revision 1.2  2001/07/09 14:41:31  tilmonen
// Important bug fixes (endian issue in di_data_buffer).
//
// Revision 1.1.1.1  2001/06/04 09:04:23  tilmonen
// Mustajuuri 4.6.2001, initial CVS cersion
//

#include "di_socket.h"

#include <qsocketdevice.h> 

#include <errno.h>
#include <unistd.h>
#include <di_types.h>
#include <string.h>
#include <sys/time.h>

// --- Constructors -----------------------------------------------------------

DI_Socket::DI_Socket(void)
  : m_qsocket(0)
{
  init();
}

DI_Socket::~DI_Socket(void)
{
  close();
}

// --- Access -----------------------------------------------------------------

void DI_Socket::setBlocking(DI_BlockingMode blocking)
{
  bool bb = (blocking == DI_BLOCK_NOT) ? false : true;

  m_qsocket->setBlocking(bb);
}

ulong DI_Socket::getBufferSize(DI_SocketStatus which) const
{
  if(!m_qsocket) {
    DI_error(DI_ERR_ARGUMENT,"DI_Socket:getBufferSize # Socket not open");
    return 0;
  }

  switch(which) {
  case DI_SS_INPUT:
    return m_qsocket->receiveBufferSize();
  case DI_SS_OUTPUT:
    return m_qsocket->sendBufferSize();
  default:
    DI_error(DI_ERR_ARGUMENT,"DI_Socket:getBufferSize");
  }
  return 0;

}

void DI_Socket::setBufferSize(DI_SocketStatus which, ulong size)
{
  if(!m_qsocket) {
    DI_error(DI_ERR_ARGUMENT,"DI_Socket::setBufferSize # Socket not open");
    return;
  }

  switch(which) {
  case DI_SS_INPUT:
    m_qsocket->setReceiveBufferSize(size);
  case DI_SS_OUTPUT:
    m_qsocket->setSendBufferSize(size);
   case DI_SS_INPUT_OUTPUT:
    setBufferSize(DI_SS_INPUT,size);
    setBufferSize(DI_SS_OUTPUT,size);
  default:
    DI_error(DI_ERR_ARGUMENT,"DI_Socket::setBufferSize");
  }
  return;

}

// --- Operations -------------------------------------------------------------

void DI_Socket::clear(void)
{
  
}

bool DI_Socket::close(void)
{
  delete m_qsocket;
  m_qsocket = 0;

  /* if (m_socketID >= 0)
  {
    ::close(m_socketID);
    m_socketID = -1;
  }

  m_blocking = DI_BLOCK_FULL;
  m_type  = DI_SS_UNINITIALIZED;
  m_error = DI_ERR_NONE;*/
  return true;
}

DI_Socket *DI_Socket::listen(void)
{
  if(!m_qsocket) {
    DI_error(DI_ERR_ARGUMENT,"DI_Socket::listen # Socket not open");
    return 0;
  }

  int sock = m_qsocket->accept();

  if(sock < 0) return 0;

  DI_Socket *rval = new DI_Socket;

  QSocketDevice *qs = new QSocketDevice(sock, m_qsocket->type());

  rval->m_qsocket = qs;

  return rval;

}

bool DI_Socket::open(const char * const address, bool server, bool stream)
{
  const char * fname = "DI_Socket::open";

  close();

  int slen = strlen(address);

  if(slen < 1) {
    DI_error2(DI_ERR_ARGUMENT, "%s # address \"%s\" is too short", 
	      fname, address);
    return false;
  }

  // Determine the port number
  int portstart = slen - 1;

  for(bool loop = true; portstart > 0 && loop;) {
    if(!isdigit(address[portstart])) {
      portstart++;
      loop = false;
    }
    else
      portstart--;
  }

  short port = atoi(address + portstart);

  if(port <= 0) {
    DI_error2(DI_ERR_ARGUMENT, "%s # bad port: %d (\"%s\") %d", 
	      fname, (int) port, address, portstart);
    return false;
  }

  QSocketDevice::Type type = stream ?  QSocketDevice::Stream :  
    QSocketDevice::Datagram;
  
  m_qsocket = new QSocketDevice(type);
  
  bool ok;

  QString tmp(address);
  // tmp.truncate(address[portstart] == ':' ? portstart - 1 : portstart);
  tmp.truncate(portstart ? portstart - 1 : 0);

  QHostAddress addr;
  addr.setAddress(tmp);
  
  if(server) {
    ok = m_qsocket->bind(addr, port);
    if(ok) m_qsocket->listen(50);
  }
  else 
    ok = m_qsocket->connect(addr, port);

  if(!ok) {
    DI_error2(DI_ERR_UNKNOWN, 
	      "%s # socket to \"%s:%d\" (%s) could not be opened",
	      fname, addr.toString().ascii(), (int) port, tmp.ascii());
    close();
  }

  return ok;
}

int DI_Socket::read(void *buffer, ulong bytes)
{
  if (bytes == 0) return 0;

  if(!m_qsocket) {
    DI_error(DI_ERR_ARGUMENT,"DI_Socket::read # Socket not open");
    return -1;
  }

  return m_qsocket->readBlock((char *) buffer, bytes);
  
}

int DI_Socket::write(const void *buffer, ulong bytes)
{
  if (bytes == 0) return 0;

  if(!m_qsocket) {
    DI_error(DI_ERR_ARGUMENT,"DI_Socket::write # Socket not open");
    return -1;
  }

  return m_qsocket->writeBlock((const char *) buffer, bytes);
}

// --- Predicates -------------------------------------------------------------



bool DI_Socket::isDatagram(void) const
{
  if(!m_qsocket) return false;
  
  return m_qsocket->type() == QSocketDevice::Datagram;
}

bool DI_Socket::isEOF(void) const
{
  //  DI_error(DI_ERR_UNIMPLEMENTED,"DI_Socket:IsEOF");
  // return (m_error == DI_ERR_EOF);
  
  if(!m_qsocket) return true;

  bool timeout = false;

  Q_LONG l = m_qsocket->waitForMore(0, &timeout);

  return l < 0;
}

bool DI_Socket::isOK(void) const
{
  if(!m_qsocket) return false;

  return m_qsocket->error() == QSocketDevice::NoError;
}

bool DI_Socket::isPendingInput(unsigned waitMicroSeconds)
{
  if(!m_qsocket) {
    DI_error(DI_ERR_INTERNAL,"DI_Socket::isPendingInput # Socket not open");
    return false;
  }

  /* THIS WOULD PURE QT IMPLEMENTATION. IT DOES NOT QUITE WORK.


  if(m_qsocket->bytesAvailable() > 0)
  return true;
  
  return waitMicroSeconds ? 
  m_qsocket->waitForMore(waitMicroSeconds / 1000) > 0: false;*/
  
  bool isPending = false;

  fd_set xReadFds;
  struct timeval xTimeout;

  xTimeout.tv_sec  = 0;
  xTimeout.tv_usec = (long) waitMicroSeconds;

  int socketID = m_qsocket->socket();

  FD_ZERO(&xReadFds);
  FD_SET(socketID,&xReadFds);

  switch(select(socketID+1,&xReadFds,(fd_set *)0,(fd_set *)0,&xTimeout))
  {
  case -1:
    DI_error(DI_ERR_SYSTEM,"DI_Socket:IsPendingInput:select",strerror(errno));
    break;
  case 0:
    break;
  default:
    if (FD_ISSET(socketID,&xReadFds)) isPending = true;
    break;
  }

  return isPending;
}

// --- Private members --------------------------------------------------------

void DI_Socket::init(void)
{
  m_qsocket = 0;
  /* m_blocking = DI_BLOCK_FULL;

  m_type     = DI_SS_UNINITIALIZED;
  m_socketID = -1;
  m_error    = DI_ERR_NONE;*/
}

// --- Helper functions -------------------------------------------------------

int DI_writeToSocket(DI_Socket *sock, const DI_DataBuffer *data)
{
  if (!sock->isDatagram())
  {
    return sock->write(data->buffer(), data->bytesUsed()+4);
  }
  else  // Datagram socket
  {
    int written(0);
    written += sock->write(data->buffer(), 4);  // Write the data size
    written += sock->write(data->buffer()+4, (ulong)data->bytesUsed());
    return written;
  }
}

int DI_readFromSocket(DI_Socket *sock, DI_DataBuffer *data)
{
  int read(0);
  read += sock->read(data->buffer(), 4); // Read the data size
  // DI_trace("DI_readFromSocket", data->bytesUsed());
  read += sock->read(data->buffer()+4, (ulong)data->bytesUsed());
  return read;
}

int DI_writeToSocket(DI_Socket *sock, const DI_DataBuffer2 *data)
{
  int tmp = data->positionLE();
  tmp = sock->write(&tmp, 4);
  return tmp + sock->write(data->buffer(), data->position());
}

int DI_readFromSocket(DI_Socket *sock, DI_DataBuffer2 *data)
{
  int read(0), tmp;
  read += sock->read(&tmp, 4);
  data->setPositionLE(tmp);
  tmp = data->position();
  data->expandTo(tmp);
  // DI_trace("DI_readFromSocket", data->bytesUsed());
  read += sock->read(data->buffer(), tmp);
  return read;
}
