/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of DIVA base library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000-2001.
 * Jarmo.Hiipakka@hut.fi Tommi.Ilmonen@hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA base library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */

//
// DIVA Library. Copyright (C) Jarmo Hiipakka, 1999-2001.
//
// $Log: di_socket.h,v $
// Revision 1.2  2002/05/22 12:42:00  tilmonen
// Changes due to port to Windows.
//
// Revision 1.1  2002/04/19 09:54:21  tilmonen
// Socket class based on QSocketDevice.
//
// Revision 1.4  2002/02/22 11:28:53  tilmonen
// New (and possibly better) DI_DataBuffer2 class for binary
// data management.
//
// Revision 1.3  2001/07/27 08:23:17  tilmonen
// License text updates (=cosmetic changes).
//
// Revision 1.2  2001/07/26 08:49:24  tilmonen
// Updates is isPendingInput.
//
// Revision 1.1.1.1  2001/06/04 09:04:23  tilmonen
// Mustajuuri 4.6.2001, initial CVS cersion
//

#ifndef DI_SOCKET_H
#define DI_SOCKET_H

#include <di_types.h>
// #include <sys/socket.h>
#include <netinet/in.h>
#include <sys/un.h>

#include "di_config.h"
#include "di_error.h"
#include "di_data_buffer.h"

class QSocketDevice;

/// Blocking modes.

enum DI_BlockingMode
{
  /// Full blocking mode.
  DI_BLOCK_FULL,
  /// Transmit/receive at least one.
  DI_BLOCK_ONE,
  /// Non-blocking mode.
  DI_BLOCK_NOT
};

/** Socket connection type or status.

    @see DI_Socket
*/

enum DI_SocketStatus
{
  /// Error indicator.
  DI_SS_ERROR = -2,
  /// End-of-file indicator.
  DI_SS_EOF = -1,
  /// Initial status.
  DI_SS_UNINITIALIZED = 0,
  /// Server node.
  DI_SS_SERVER = 1,
  /// Data node.
  DI_SS_DATA = 2,
  /// Apply to input.
  DI_SS_INPUT = 3,
  /// Apply to output.
  DI_SS_OUTPUT = 4,
  /// Apply to both input and output.
  DI_SS_INPUT_OUTPUT = 5  
};

/** Connection address. @see DI_Socket */

union DI_SocketAddress
{
  struct sockaddr_in in;
  struct sockaddr_un un;
};

/** Communication services.

    DI_Socket is declared in "di_socket.h" and defined in divabase.

    <B>Introduction</B>

    DI_Socket offers an easy-to-use, portable front-end layer to
    socket-like communication services. All system-dependend technical
    implementation details of socket-based communication have been
    hidden under the DI_Socket class interface.

    DI_Socket instances, or nodes, come in two flavors: data nodes and
    server nodes. A data node receives and transmits data over a
    network, while a server node monitors and accepts incoming
    calls.

    <B>Data Nodes</B>

    Nodes are data nodes by default. A data node is connected to another
    data node using the open() method call. After data node is connected,
    information can be exchanged with the other data node with 'read' and
    'write' methods.

    <B>Server Nodes</B>

    A server node is created when a data node is bound to a network
    address that the server node will monitor. This can be done with the
    open(...,true) method. After a server node has been bound to an
    address, it can listen the network and accept new connections with
    the listen() method.

    <B>Addressing</B>

    DI_Socket addresses are passed to DI_Socket methods as text strings,
    much like file paths are passed to file functions. The format of the
    string determines the connection type, while the information in the
    string itself defines the actual connection address. See 'open()'
    for details.

    <B>Examples</B>

\code
    // Open a listening socket for incoming calls:

    DI_Socket *listen = new DI_Socket();
    listen->open("127.64.33.7:10000",true);

    // Check status:

    if(!listen || !(*listen)) exit(1);

    // Accept an incoming call:

    DI_Socket *input = listen->listen();

    // Check status:

    if(!input) exit(1);

    // Read data:

    while(input->read(buffer,1000) > 0) doSomething(buffer);

    // Close connection:

    delete input;

    // Close listener:

    delete listen;
\endcode

    @see DI_SocketStatus DI_SocketAddress

  */

class DI_Socket
{

public:

  /**@name Constructors */

  //@{

  /// Create an unconnected data node.

  DI_Socket(void);

  /// Close the current connection, flushing any buffered data.

  virtual ~DI_Socket(void);

  //@}

  /**@name Access */

  //@{

  /// Return current blocing mode.

  DI_BlockingMode getBlocking(void);

  /** Set node blocking mode.

      This methods sets the new blocking mode. The three blocking
      modes available are:
      
      <UL>
      <LI>[DI_BLOCK_FULL] Full blocking mode.
      <LI>[DI_BLOCK_ONE] Partial blocking mode.
      <LI>[DI_BLOCK_NOT] Non-blocking mode.
      </UL>

      The default blocking mode for all nodes is DI_BLOCK_FULL.

  */

  void setBlocking(DI_BlockingMode blocking);

  /** Get buffer size.

      This method returns the size of the system input or output buffer
      for current socket connection. The buffer size is expressed in
      bytes. 'which' determines which buffer size is returned, and it
      must have the value DI_SS_INPUT or DI_SS_OUTPUT.

  */

  ulong getBufferSize(DI_SocketStatus which) const;

  /** Set buffer size.

      This method sets the size of the system input and output buffers
      for current socket connection. The new buffer size 'size' is
      expressed in bytes. 'which' determines which buffer sizes are
      affected, and it must have the value DI_SS_INPUT,
      DI_SS_OUTPUT, or DI_SS_INPUT_OUTPUT.

  */

  void setBufferSize(DI_SocketStatus which, ulong size);

  //@}

  /**@name Operations */

  //@{

  /// Clear node error flag.

  void clear(void);

  /** Close current connection.

      Close current connection, flush all pending data, and return the
      connection back to it's initial state. If closing operation was
      successfull, 'true' is returned, otherwise 'false'.

  */

  bool close(void);

  /** Enter listening mode, return next caller or NULL.

      Listen is an operation that can be used with server nodes only.
      It returns a data node that can be used to communicate with
      next caller that tries to connect to the address the server node
      is currently bound, or NULL if for some reason no callers were
      available.

      Depending of the current blocking mode, the server node will
      either wait until there is a caller (DI_BLOCK_FULL,DI_BLOCK_ONE),
      or return immediately (DI_BLOCK_NOT).
      
  */

  DI_Socket *listen(void);

  /** Open connection to 'address'.

      This method tries to connect or bind the node to 'address', and
      returns 'true' if this succeeded. The node is opened as data or
      server node, depending on the value of 'server'. A node that
      will be used to listen the network and accept incoming calls must
      be opened in the server mode. The value of 'stream' affects the type
      of the socket 'true' results in a stream socket, whereas 'false'
      results in a datagram socket.

      The following address formats are recogniced. The address format
      determines also the type of connection:

      <UL>
      <LI>[a.b.c.d:port] IP connection to the given IP address
      <LI>[name:port] IP connection to the given host
      <LI>[port] IP connection to local host (127.0.0.1:port)
      <LI>[name] Unix domain file connection
      </UL>
  
  */

  bool open(const char * const address,
	    bool server = false,
	    bool stream = true);

  /** Read data from an open connection.

      'read' reads 'bytes' bytes from an open connection and stores
      them into 'buffer'. If the operation succeedes, the number of read
      bytes is returned. If an error occurs, DI_SS_ERROR, a negative
      value is returned. If the sender has closed the connection,
      DI_SS_EOF, an another negative value is returned. 'read' will
      never attempt to read more than 'bytes' bytes from the input
      source.

      If full blocking mode is enabled (DI_BLOCK_FULL), 'read' does not
      return until the requested number of bytes have been read, an
      error occurs, or the data sender has closed the connection.
      If the connection was closed after at least one but less than
      'bytes' bytes were successfully read, the number of read bytes is
      returned, and the next 'read' call will return DI_SS_EOF.

      If partial blocking mode is enabled (DI_BLOCK_ONE), 'read' returns
      when at least one byte has been read, an error occurs, or the data
      sender has closed the connection.

      If blocking is disabled (DI_BLOCK_NOT), 'read' returns always
      immediately.

  */

  int read(void *buffer, ulong bytes);

  /** Write data into an open connection.

      'write' writes 'bytes' bytes to an open connection, reading them
      from 'buffer'. If the operation succeedes, the number of written
      bytes is returned. If an error occurs, a negative value is
      returned. If the receiver closes the connection, zero is
      returned. 'write' will never attemp to write more than 'bytes' 
      bytes to the output sink.

      If full blocking mode is enabled (DI_BLOCK_FULL), 'write' does not
      return until the requested number of bytes have been written, an
      error occurs or the data receiver closes the connection.

      If partial blocking mode is denabled (DI_BLOCK_ONE), 'write' returns
      when at least one byte has been written, an error occurs or the
      data receiver closes the connection.

      If blocking is disabled (DI_BLOCK_NOT), 'write' returns always
      immediately.

  */

  int write(const void *buffer, ulong bytes);

  //@}

  /**@name Predicates */

  //@{


  /// Return 'true' if stream is a datagram connection.

  bool isDatagram(void) const;

  /// Return 'true' if the socket contains no more data.

  bool isEOF(void) const;

  /// Return 'true' if the socket is fully functional

  bool isOK(void) const;

  /// Return 'true' if readable data is pending.

  bool isPendingInput(unsigned waitMicroSeconds = 0);


  //@}

  /**@name Operators */

  //@{

  /// Return 'true' if 'IsOk()' returns 'true'.

  inline operator bool (void) { return isOK(); }

  /// Return 'true' if 'IsOk()' return 'false'.

  inline bool operator ! (void) { return !isOK(); }

  //@}

private:

  void init(void);

  QSocketDevice *m_qsocket;

};

/// Write a DI_DataBuffer to a socket
int DI_writeToSocket(DI_Socket *sock, const DI_DataBuffer *data);

inline DI_Socket &operator<<(DI_Socket &sock, const DI_DataBuffer &data)
{ DI_writeToSocket(&sock, &data); return sock; }

/// Read a DI_DataBuffer from a socket
int DI_readFromSocket(DI_Socket *sock, DI_DataBuffer *data);

inline DI_Socket &operator>>(DI_Socket &sock, DI_DataBuffer &data)
{ DI_readFromSocket(&sock, &data); return sock; }

/// Write a DI_DataBuffer2 to a socket
int DI_writeToSocket(DI_Socket *sock, const DI_DataBuffer2 *data);

inline DI_Socket &operator<<(DI_Socket &sock, const DI_DataBuffer2 &data)
{ DI_writeToSocket(&sock, &data); return sock; }

/// Read a DI_DataBuffer2 from a socket
int DI_readFromSocket(DI_Socket *sock, DI_DataBuffer2 *data);

inline DI_Socket &operator>>(DI_Socket &sock, DI_DataBuffer2 &data)
{ DI_readFromSocket(&sock, &data); return sock; }

#endif  // DI_SOCKET_H
