/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of DIVA base library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000-2001.
 * Jarmo.Hiipakka@hut.fi Tommi.Ilmonen@hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA base library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */

#include "di_thread.h"
#include "di_mutex.h"
#include "di_sleep.h"

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <di_types.h>

bool DI_Thread::m_threadDebug = false;
bool DI_Thread::m_threadWarnings = false;

DI_Thread::DI_Thread()
{
   m_qthread.init(this);
}

DI_Thread::~DI_Thread()
{}

bool DI_Thread::run(bool prefer_system)
{
  if(m_threadDebug) {
    printf("DI_Thread::run %d %p\n", (int) prefer_system, this);
    fflush(0);
  }
  if(prefer_system) {
    if(!runSystem()) {
      return runProcess();
    }
  }
  else if(!runProcess())
    return runSystem();

  return true;
}

bool DI_Thread::runSystem()
{
  if(m_threadDebug) {
    printf("DI_Thread::runSystem %p\n", this);
    fflush(0);
  }

  m_qthread.start();

  return true;
}

bool DI_Thread::runProcess()
{
  if(m_threadDebug) {
    printf("DI_Thread::runProcess %p\n", this);
    fflush(0);
  }

  m_qthread.start();

  return true;
}

bool DI_Thread::waitEnd()
{
  if(m_threadDebug) {
    printf("DI_Thread::waitEnd %p\n", this);
    fflush(0);
  }

  bool ok = m_qthread.wait();
  
  if(!ok) {
    if(m_threadDebug || m_threadWarnings) {
      fprintf(stderr, "DI_Thread::waitEnd failed %p\n", this);
      fflush(0);
    }
    return false;
  }
  return true;
}

void DI_Thread::kill()
{
  fprintf(stderr, "DI_Thread::kill # NOT IMPLEMENTED IN QTHREADS\n");
  fflush(0);
}

void DI_Thread::test()
{
  bool debug = m_threadDebug;
  m_threadDebug = true;  // Debugging texts on.

  const char name[] = "DI_Thread::test (QT)";

  DI_Sleep sleep;

#ifndef WIN32
  printf("%s # pid = %d TEST 1\n", name, getpid()); fflush(0);
#endif

  // cout << name << " pid: " << getpid() << " test 1" << endl;

  {
    printf("%s # creating thread that runs for 4 seconds\n", name); fflush(0);

    DI_ThreadTest t;
    printf("%s # starting thread...\n", name); fflush(0);
    t.run();
    sleep.sleepUs(100000);
    printf("%s # ...thread running, sleeping 4 seconds...\n", name); fflush(0);
    DI_Sleep::sleepS(4);
    printf("%s # ending thread...\n", name); fflush(0);
    t.requestEnd();
    bool ok = t.waitEnd();
    printf("%s # success = %d\n", name, (int) ok); fflush(0);
  }

  printf("%s # TEST 2\n", name); fflush(0);
  {
    printf("%s # creating thread that runs for 2 seconds\n", name); fflush(0);

    DI_ThreadTest t;
    printf("%s # starting thread...\n", name); fflush(0);
    t.run();
    printf("%s # ...thread running, sleeping 2 seconds...\n", name); fflush(0);
    DI_Sleep::sleepS(2);
    printf("%s # deliberate deadlock now, USE CTRL+C to exit\n", name); 
    fflush(0);

    /* "requestEnd" is not called. Therefore we deadlock. */ 
    bool ok = t.waitEnd();
    printf("%s # Failed to deadlock %d\n", name, (int) ok); 

  }
  m_threadDebug = debug; // Return the original value
}


void DI_Thread::threadExit()
{
  if(m_threadDebug) {
    printf("DI_Thread::threadExit # %p\n", this); fflush(0);
  }
}

void DI_Thread::mainLoop()
{
  if(m_threadDebug) {
    printf("DI_Thread::mainLoop # %p id = %d\n", this, (int) myThreadId()); 
    fflush(0);
  }

  childLoop();
}

void *DI_Thread::entry(void *t)
{
  if(m_threadDebug)
    printf("DI_Thread::entry # %p\n", t); fflush(0);
  
  ((DI_Thread *) t)->mainLoop();

  return 0;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

void DI_Thread::QThreadWrap::run()
{
  DI_Thread::entry(m_host);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

bool DI_ThreadAvailable::m_systemLevel = false;
bool DI_ThreadAvailable::m_processLevel = false;
bool DI_ThreadAvailable::m_defined = false;

bool DI_ThreadAvailable::systemLevelWorks()
{
  if(!m_defined) check();
  return m_systemLevel;
}

bool DI_ThreadAvailable::processLevelWorks()
{
  if(!m_defined) check();
  return m_processLevel;
}

void DI_ThreadAvailable::check()
{
  {
    DI_ThreadTest t;
    t.requestEnd();
    m_systemLevel = t.runSystem();
    if(m_systemLevel)
      t.waitEnd();
  }
  {
    DI_ThreadTest t;
    t.requestEnd();
    m_processLevel = t.runProcess();
    if(m_processLevel)
      t.waitEnd();
  }
  m_defined = true;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

DI_ThreadTest::DI_ThreadTest()
  : m_mutex(new DI_MutexAuto),
    m_continue(true),
    m_counter(0)
{}

int DI_ThreadTest::count()
{
  m_mutex->lock();
  int i = m_counter;
  m_mutex->unlock();
  return i;
}

void DI_ThreadTest::requestEnd()
{
  m_mutex->lock();
  m_continue = false;
  m_mutex->unlock();
}

void DI_ThreadTest::childLoop()
{
  m_mutex->lock();
  DI_Sleep s;

  while(m_continue) {

    m_mutex->unlock();
    s.sleepUs(500000); // Half second sleep.
    m_mutex->lock();

    printf("DI_ThreadTest::childLoop %p %d\n", this,  m_counter++);
    fflush(0);
  }
  m_mutex->unlock();
}
