/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the DIVA IO library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000-2001.
 * Jarmo.Hiipakka@hut.fi Tommi.Ilmonen@hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA IO library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */
//
// DIVA Library. Copyright (C) Jarmo Hiipakka, 1999.
//
// $Log: di_audio_port.C,v $
// Revision 1.5  2004/02/26 15:20:56  tilmonen
// Fixes.
//
// Revision 1.4  2004/02/06 15:47:58  tilmonen
// *** empty log message ***
//
// Revision 1.3  2003/08/11 13:18:19  tilmonen
// Fixes for the MipsPro 7.4 compiler/headers.
//
// Revision 1.2  2001/07/27 08:23:27  tilmonen
// License text updates (=cosmetic changes).
//
// Revision 1.1.1.1  2001/06/04 09:04:21  tilmonen
// Mustajuuri 4.6.2001, initial CVS cersion
//

#include <stdlib.h>
#include <errno.h>

extern "C" {
/* Some SGI systems require this crap (compiler/header mismatch ??). */
#include <sys/types.h>
}

#include "di_audio_port.h"
#include "di_error.h"
#include "di_sleep.h"
#include "di_numconv.h"

// --- Constructors -----------------------------------------------------------

DI_AudioPort::DI_AudioPort() :
  m_inDevice(NULL),
  m_outDevice(NULL),
  m_deleteDev(false)
{
}

DI_AudioPort::~DI_AudioPort(void)
{
  if ((m_inDevice) || (m_outDevice)) // Not yet closed
  {
    close();
  }
}

// --- Access -----------------------------------------------------------------

uint DI_AudioPort::inSampleRate(void) const
{
  if (((m_mode == DI_IO_INPUT) || (DI_IO_INPUT_OUTPUT)) && (m_inDevice))
    return m_inDevice->getRate();
  else
    return DI_AUDIO_UNSPECIFIED_RATE;
}

uint DI_AudioPort::outSampleRate(void) const
{
  if (((m_mode == DI_IO_OUTPUT) || (DI_IO_INPUT_OUTPUT)) && (m_outDevice))
    return m_outDevice->getRate();
  else
    return DI_AUDIO_UNSPECIFIED_RATE;
}

uint DI_AudioPort::inChannels(void) const
{
  if (((m_mode == DI_IO_INPUT) || (DI_IO_INPUT_OUTPUT)) && (m_inDevice))
    return m_inDevice->getChannels();
  else
    return 0;
}

uint DI_AudioPort::outChannels(void) const
{
  if (((m_mode == DI_IO_OUTPUT) || (DI_IO_INPUT_OUTPUT)) && (m_outDevice))
    return m_outDevice->getChannels();
  else
    return 0;
}

int DI_AudioPort::inFilledFrames(void) const
{
  if ((m_mode & DI_IO_INPUT) && (m_inPort))
    return alGetFilled(m_inPort);
  else
    return 0;
}

int DI_AudioPort::outFilledFrames(void) const
{
  if ((m_mode & DI_IO_OUTPUT) && (m_outPort))
    return alGetFilled(m_outPort);
  else
    return 0;
}

// --- Operations -------------------------------------------------------------

bool DI_AudioPort::open(const char *device,
			DI_IoMode mode,
			DI_AudioSampleFormat userSample, 
			uint bufferSize,
			uint samplerate,
			uint channels,
			uint maxBlock)
{
  m_deleteDev = true;
  // On SGI, one device can only be for input or for output, not for both.
  if ((mode == DI_IO_NONE) || ((mode == DI_IO_INPUT_OUTPUT) && (device)))
  {
    DI_error(DI_ERR_ARGUMENT, "Can't open the device for this IO mode");
    return false;
  }
  if (mode == DI_IO_INPUT)
  {
    m_inDevice = new DI_AudioDevice(device, true, channels, samplerate);
    if (!((m_inDevice->isOK()) &&
	  open(*m_inDevice,mode,userSample,bufferSize,maxBlock)))
      return false;
  }
  if (mode == DI_IO_OUTPUT)
  {
    m_outDevice = new DI_AudioDevice(device, false, channels, samplerate);
    if (!((m_outDevice->isOK()) &&
	  open(*m_outDevice,mode,userSample,bufferSize,maxBlock)))
      return false;
  }
  if (mode == DI_IO_INPUT_OUTPUT)
  {    
    m_inDevice = new DI_AudioDevice(device, true, channels, samplerate);
    m_outDevice = new DI_AudioDevice(device, false, channels, samplerate);
    if (!((m_inDevice->isOK()) && (m_outDevice->isOK()) &&
	  open(*m_inDevice, *m_outDevice, userSample, userSample, bufferSize,
	       bufferSize, maxBlock, maxBlock)))
    {
      return false;
    }
  }
  return true;
}

bool DI_AudioPort::open(DI_AudioDevice &device,
			DI_IoMode mode,
			DI_AudioSampleFormat userSample, 
			uint bufferSize,
			uint maxBlock)
{
  if ((mode != DI_IO_INPUT) && (mode != DI_IO_OUTPUT))
  {
    DI_error(DI_ERR_ARGUMENT,"Invalid IO mode for this device");
    return false;
  }
  
  ALconfig config = alNewConfig();
  
  switch(userSample)
  {
   case DI_ASF_INT16:
    alSetSampFmt(config,AL_SAMPFMT_TWOSCOMP);
    alSetWidth(config,2);
    break;
   case DI_ASF_INT32:
    alSetSampFmt(config,AL_SAMPFMT_TWOSCOMP);
    alSetWidth(config,4);
    break;
   case DI_ASF_FLOAT32:
    alSetSampFmt(config,AL_SAMPFMT_FLOAT);
    alSetFloatMax(config,1.0);
    break;
   case DI_ASF_FLOAT64:
    alSetSampFmt(config,AL_SAMPFMT_DOUBLE);
    alSetFloatMax(config,1.0);
    break;
   default:
    DI_error(DI_ERR_ARGUMENT, "Invalid audio sample format");
    return false;
  }
  
  alSetChannels(config, (int)device.getChannels());
  alSetQueueSize(config, (int)bufferSize);
  alSetDevice(config, device.getDevice());
  
  ALport port(alOpenPort("DI_AudioPort",((mode == DI_IO_INPUT) ? "r" : "w"),
			 config));
  if(port == 0)
  {
    DI_error(DI_ERR_INTERNAL,"DI_AudioPort:alOpenPort failed");
    return false;
  }

  int FD(alGetFD(port));
  if(FD < 0)
  {
    DI_error(DI_ERR_INTERNAL,"DI_AudioPort:alGetFD failed");
    return false;
  }

  m_mode = mode;
  m_inFormat = userSample;
  m_outFormat = userSample;
  m_inBufSize = bufferSize;
  m_outBufSize = bufferSize;
  m_inBlock = maxBlock; // This is not needed for anything on SGI platform
  m_outBlock = maxBlock; // This is not needed for anything on SGI platform
  m_inPFD.events = POLLIN;
  m_outPFD.events = POLLOUT;

  if (mode == DI_IO_INPUT)
  {
    m_inDevice = &device;
    m_outDevice = NULL;
    m_inFD = FD;
    m_outFD = 0;
    m_inPFD.fd = FD;
    m_outPFD.fd = 0;
    device.getName(m_inName);
    m_outName[0] = '\0';
    device.getLabel(m_inLabel);
    m_outLabel[0] = '\0';
    m_inConfig = config;
    m_inPort = port;
    m_outPort = 0;
  }
  else
  {
    m_inDevice = NULL;
    m_outDevice = &device;
    m_inFD = 0;
    m_outFD = FD;
    m_inPFD.fd = 0;
    m_outPFD.fd = FD;
    m_inName[0] = '\0';
    device.getName(m_outName);
    m_inLabel[0] = '\0';
    device.getLabel(m_outLabel);
    m_outConfig = config;
    m_inPort = 0;
    m_outPort = port;
  }

  return true;
}

bool DI_AudioPort::open(const char *inDevice,
			const char *outDevice,
			DI_AudioSampleFormat inFormat, 
			DI_AudioSampleFormat outFormat, 
			uint inBufSize,
			uint outBufSize,
			uint inSamplerate,
			uint outSamplerate,
			uint inChannels,
			uint outChannels,
			uint inMaxBlock,
			uint outMaxBlock)
{
  m_deleteDev = true;
  m_inDevice = new DI_AudioDevice(inDevice,true,inChannels,inSamplerate);
  m_outDevice = new DI_AudioDevice(outDevice,false,outChannels,outSamplerate);
  if (!((m_inDevice->isOK()) && (m_outDevice->isOK()) &&
	open(*m_inDevice, *m_outDevice, inFormat, outFormat, inBufSize,
	     outBufSize, inMaxBlock, outMaxBlock)))
  {
    return false;
  }
  return true;
}

bool DI_AudioPort::open(DI_AudioDevice &inDevice,
			DI_AudioDevice &outDevice,
			DI_AudioSampleFormat inFormat, 
			DI_AudioSampleFormat outFormat, 
			uint inBufSize,
			uint outBufSize,
			uint inMaxBlock,
			uint outMaxBlock)
{
  m_mode = DI_IO_INPUT_OUTPUT;
  
  m_inConfig = alNewConfig();
  m_outConfig = alNewConfig();
  
  switch(inFormat)
  {
   case DI_ASF_INT16:
    alSetSampFmt(m_inConfig,AL_SAMPFMT_TWOSCOMP);
    alSetWidth(m_inConfig,2);
    break;
   case DI_ASF_INT32:
    alSetSampFmt(m_inConfig,AL_SAMPFMT_TWOSCOMP);
    alSetWidth(m_inConfig,4);
    break;
   case DI_ASF_FLOAT32:
    alSetSampFmt(m_inConfig,AL_SAMPFMT_FLOAT);
    alSetFloatMax(m_inConfig,1.0);
    break;
   case DI_ASF_FLOAT64:
    alSetSampFmt(m_inConfig,AL_SAMPFMT_DOUBLE);
    alSetFloatMax(m_inConfig,1.0);
    break;
   default:
    DI_error(DI_ERR_ARGUMENT, "Invalid audio sample format");
    return false;
  }
  switch(outFormat)
  {
   case DI_ASF_INT16:
    alSetSampFmt(m_outConfig,AL_SAMPFMT_TWOSCOMP);
    alSetWidth(m_outConfig,2);
    break;
   case DI_ASF_INT32:
    alSetSampFmt(m_outConfig,AL_SAMPFMT_TWOSCOMP);
    alSetWidth(m_outConfig,4);
    break;
   case DI_ASF_FLOAT32:
    alSetSampFmt(m_outConfig,AL_SAMPFMT_FLOAT);
    alSetFloatMax(m_outConfig,1.0);
    break;
   case DI_ASF_FLOAT64:
    alSetSampFmt(m_outConfig,AL_SAMPFMT_DOUBLE);
    alSetFloatMax(m_outConfig,1.0);
    break;
   default:
    DI_error(DI_ERR_ARGUMENT, "Invalid audio sample format");
    return false;
  }
  
  alSetChannels(m_inConfig, (int)inDevice.getChannels());
  alSetQueueSize(m_inConfig, (int)inBufSize);
  alSetDevice(m_inConfig, inDevice.getDevice());
  alSetChannels(m_outConfig, (int)outDevice.getChannels());
  alSetQueueSize(m_outConfig, (int)outBufSize);
  alSetDevice(m_outConfig, outDevice.getDevice());
  
  m_inPort = alOpenPort("DI_AudioPort", "r", m_inConfig);
  m_outPort = alOpenPort("DI_AudioPort", "w", m_outConfig);
  if((m_inPort == 0) || (m_outPort == 0))
  {
    DI_error(DI_ERR_INTERNAL,"DI_AudioPort:alOpenPort failed");
    return false;
  }

  m_inFD = alGetFD(m_inPort);
  m_outFD = alGetFD(m_outPort);
  if((m_inFD < 0) || (m_outFD < 0))
  {
    DI_error(DI_ERR_INTERNAL,"DI_AudioPort:alGetFD failed");
    return false;
  }
  
  m_inDevice = &inDevice;
  m_outDevice = &outDevice;
  m_inFormat = inFormat;
  m_outFormat = outFormat;
  m_inBufSize = inBufSize;
  m_outBufSize = outBufSize;
  m_inBlock = inMaxBlock; // This is not needed for anything on SGI platform
  m_outBlock = outMaxBlock; // This is not needed for anything on SGI platform
  m_inPFD.fd = m_inFD;
  m_outPFD.fd = m_outFD;
  m_inPFD.events = POLLIN;
  m_outPFD.events = POLLOUT;
  inDevice.getName(m_inName);
  outDevice.getName(m_outName);
  inDevice.getLabel(m_inLabel);
  outDevice.getLabel(m_outLabel);
  return true;
}

bool DI_AudioPort::close(void)
{
  // If outputting, wait until the output queue has drained
  
  DI_Sleep sleeper;
  while(outFilledFrames()) sleeper.sleepMs(10);
  
  // Close the port and free all resources
  
  if (m_inPort)
  {
    alClosePort(m_inPort);
    alFreeConfig(m_inConfig);
  }
  if (m_outPort)
  {
    alClosePort(m_outPort);
    alFreeConfig(m_outConfig);
  }
  m_mode = DI_IO_NONE;

  if (m_deleteDev)
  {
    delete m_inDevice;
    delete m_outDevice;
  }
  m_deleteDev = false;
  m_inDevice = NULL;
  m_outDevice = NULL;

  return true;
}

// --- I/O operations ---------------------------------------------------------

int DI_AudioPort::readFrames(void *buffer, uint frames)
{
  alReadFrames(m_inPort,buffer,(int)frames);
  if ((alGetSampFmt(m_inConfig) == AL_SAMPFMT_TWOSCOMP) &&
      (alGetWidth(m_inConfig) == AL_SAMPLE_24))
  {
    DI_int2Int((int *)buffer,(int *)buffer,8,0,
	       frames*alGetChannels(m_inConfig));
  }
  return (int)frames;
}

int DI_AudioPort::writeFrames(const void *buffer, uint frames)
{
  alWriteFrames(m_outPort, (void *)buffer, (int)frames);
  if ((alGetSampFmt(m_outConfig) == AL_SAMPFMT_TWOSCOMP) &&
      (alGetWidth(m_inConfig) == AL_SAMPLE_24))
  {
    DI_int2Int((int *)buffer,(int *)buffer,-8,0,
	       frames*alGetChannels(m_inConfig));
  }
  return (int)frames;
}

int DI_AudioPort::waitInputFilled(uint frames, int timeOut)
{
  bool OK(true);

  if(frames > m_inBufSize)
  {
    DI_error(DI_ERR_ARGUMENT,
	     "DI_AudioPort:WaitInputFilled:'frames' > frame buffer size");
    OK = false;
  }
  else
  {
    alSetFillPoint(m_inPort, (int)frames);

    if(poll(&m_inPFD,1,timeOut) < 0)
    {
      DI_error(DI_ERR_INTERNAL,"DI_AudioPort:Wait:poll failed");
      OK = false;
    }
  }
  return OK ? alGetFilled(m_inPort) : -alGetFilled(m_inPort);
}

int DI_AudioPort::waitOutputFilled(uint frames, int timeOut)
{
  bool OK(true);

  if(frames > m_outBufSize)
  {
    DI_error(DI_ERR_ARGUMENT,
	     "DI_AudioPort:WaitOutputFilled:'frames' > frame buffer size");
    OK = false;
  }
  else
  {
    alSetFillPoint(m_outPort, (int)frames);

    if(poll(&m_outPFD,1,timeOut) < 0)
    {
      DI_error(DI_ERR_INTERNAL,"DI_AudioPort:Wait:poll failed");
      OK = false;
    }
  }
  return OK ? alGetFilled(m_outPort) : -alGetFilled(m_outPort);
}

// --- Predicates -------------------------------------------------------------

bool DI_AudioPort::isOK(void) const
{
  if ((m_mode == DI_IO_NONE) ||
      ((m_mode == DI_IO_INPUT) && (!m_inDevice)) ||
      ((m_mode == DI_IO_OUTPUT) && (!m_outDevice)) ||
      ((m_mode == DI_IO_INPUT_OUTPUT) && ((!m_inDevice) || (!m_outDevice))))
    return false;
  return true;
}




// --- Audio Device Class -----------------------------------------------------

/* NOTE!

   This class breaks the nice idea of having one header file for each class.
   However, this class is only available on SGI platform, and to keep things
   simple, it is included here.
*/

// --- Utilities --------------------------------------------------------------

uint DI_AudioDevice::getNumberOfInputDevices()
{
  return (uint)alQueryValues(AL_SYSTEM, AL_DEFAULT_INPUT, NULL, 0, NULL, 0);
}

uint DI_AudioDevice::getNumberOfOutputDevices()
{
  return (uint)alQueryValues(AL_SYSTEM, AL_DEFAULT_OUTPUT, NULL, 0, NULL, 0);
}

void DI_AudioDevice::getInputDevices(ulong *IDArray)
{
  ulong   nN;
  ulong   nIndex;
  ALvalue *xpValues;

  nN = getNumberOfInputDevices();
  xpValues = new ALvalue[nN];
  alQueryValues(AL_SYSTEM, AL_DEFAULT_INPUT, xpValues, (int)nN, NULL, 0);
  for (nIndex=0; nIndex<nN; nIndex++)
  {
    IDArray[nIndex] = (ulong)(xpValues[nIndex].i);
  }
  delete [] xpValues;
}

void DI_AudioDevice::getOutputDevices(ulong *xpIDArray)
{
  ulong   nN;
  ulong   nIndex;
  ALvalue *xpValues;

  nN = getNumberOfOutputDevices();
  xpValues = new ALvalue[nN];
  alQueryValues(AL_SYSTEM, AL_DEFAULT_OUTPUT, xpValues, (int)nN, NULL, 0);
  for (nIndex=0; nIndex<nN; nIndex++)
  {
    xpIDArray[nIndex] = (ulong)(xpValues[nIndex].i);
  }
  delete [] xpValues;
}

// --- Constructors -----------------------------------------------------------

DI_AudioDevice::DI_AudioDevice(bool input, ulong channels, ulong rate) :
  m_input(input),
  m_channels(0),
  m_rate(0),
  m_OK(false)
{
  // Do configuration here

  if (input)
    m_device = AL_DEFAULT_INPUT;
  else
    m_device = AL_DEFAULT_OUTPUT;
  
  if (!setChannels(channels))
  {
    DI_error(DI_ERR_ARGUMENT,"Invalid channel setting");
    m_OK = false;
    return;
  }

  if (rate == 0)
  {
    m_rate = getRate();
  }
  else
  {
    setRate(rate);
  }
  m_OK = true;
}

DI_AudioDevice::DI_AudioDevice(ulong ID,bool input,ulong channels,ulong rate) :
  m_input(input),
  m_channels(0),
  m_rate(0),
  m_OK(false)
{
  // Do configuration here

  m_device = (int)ID;
  if (alIsSubtype(AL_OUTPUT_DEVICE_TYPE, m_device) && m_input)
  {
    DI_error(DI_ERR_ARGUMENT, "Device can't be used for input");
    return;
  }
  else if (alIsSubtype(AL_INPUT_DEVICE_TYPE, m_device) && !m_input)
  {
    DI_error(DI_ERR_ARGUMENT, "Device can't be used for output");
    return;
  }
  
  if (!setChannels(channels))
  {
    DI_error(DI_ERR_ARGUMENT,"Invalid channel setting");
    m_OK = false;
    return;
  }

  if (rate == 0)
  {
    m_rate = getRate();
  }
  else
  {
    setRate(rate);
  }
  m_OK = true;
}

DI_AudioDevice::DI_AudioDevice(const char *name,bool input,ulong channels,
			       ulong rate) :
  m_input(input),
  m_channels(0),
  m_rate(0),
  m_OK(false)
{
  // Do configuration here

  if (name == NULL)
  {
    if (input)
      m_device = AL_DEFAULT_INPUT;
    else
      m_device = AL_DEFAULT_OUTPUT;
  }
  else
  {
    if (!(m_device =
	  alGetResourceByName(AL_SYSTEM, (char *)name, AL_DEVICE_TYPE)))
    {
      DI_error(DI_ERR_ARGUMENT, "Could not get audio resource by name");
      return;
    }
    if (alIsSubtype(AL_OUTPUT_DEVICE_TYPE, m_device) && m_input)
    {
      DI_error(DI_ERR_ARGUMENT, "Device can't be used for input");
      return;
    }
    else if (alIsSubtype(AL_INPUT_DEVICE_TYPE, m_device) && !m_input)
    {
      DI_error(DI_ERR_ARGUMENT, "Device can't be used for output");
      return;
    }
  }
  
  if (!setChannels(channels))
  {
    DI_error(DI_ERR_ARGUMENT,"Invalid channel setting");
    m_OK = false;
    return;
  }

  if (rate == 0)
  {
    m_rate = getRate();
  }
  else
  {
    setRate(rate);
  }
  m_OK = true;
}

DI_AudioDevice::~DI_AudioDevice(void)
{
}

// --- Operations -------------------------------------------------------------

ulong DI_AudioDevice::getChannels(void)
{
  // Check true number of channels here

  ALpv xParams;
  
  xParams.param = AL_CHANNELS;
  if (alGetParams(m_device, &xParams ,1) == 1)
  {
    // System supports channel mode setting
    m_channels = xParams.value.i;
  }
  else
  {
    // System does not support channel mode setting => assuming stereo
    m_channels = 2;
  }

  return m_channels;
}

ulong DI_AudioDevice::getRate(void)
{
  // Check true sample rate here
  
  ALpv xParams;

  xParams.param = AL_RATE;
  
  if (alGetParams(m_device,&xParams,1) != 1)
  {
    DI_error(DI_ERR_INTERNAL, "Could not get sample rate");
    return DI_AUDIO_UNSPECIFIED_RATE;
  }

  double dTemp = alFixedToDouble(xParams.value.ll);
  if (dTemp < 0)
    return DI_AUDIO_UNSPECIFIED_RATE;
  m_rate = (ulong)dTemp;
    
  return m_rate;
}

bool DI_AudioDevice::setChannels(ulong channels)
{
  // Set true number of channels here
  
  ALpv xParams;

  xParams.param = AL_CHANNELS;
  xParams.value.i = (int)channels;

  if (alSetParams(m_device, &xParams, 1) != 1)
  {
    DI_error(DI_ERR_INTERNAL, "Could not change channel setting");
    return false;
  }

  // Diagnostics
  if (xParams.sizeOut == AL_INVALID_PARAM)
  {
    DI_error(DI_ERR_INTERNAL, "Parameter not recognized");
    return false;
  }
  if (xParams.sizeOut == AL_INVALID_VALUE)
  {
    DI_error(DI_ERR_ARGUMENT, "Invalid number of channels");
  }

  m_channels = channels;
  return true;
}

bool DI_AudioDevice::setRate(ulong rate)
{
  // Set true rate here

  ALpv xaParams[2];

  xaParams[0].param = AL_MASTER_CLOCK;
  xaParams[0].value.i = AL_CRYSTAL_MCLK_TYPE;
  xaParams[1].param = AL_RATE;
  xaParams[1].value.ll = alDoubleToFixed((double)rate);
  
  if (alSetParams(m_device, xaParams, 2) != 2)
  {
    DI_error(DI_ERR_INTERNAL, "Could not change sample rate setting");
    return false;
  }
  
  m_rate = rate;
  return true;
}

// --- Predicates -------------------------------------------------------------

bool DI_AudioDevice::isInput(void) const
{
  // Check actual input status here
  return m_input;
}

// --- Interface Operations ---------------------------------------------------

ulong DI_AudioDevice::getNumberOfInterfaces() const
{
  return (ulong)alQueryValues(m_device, AL_INTERFACE, NULL, 0, NULL, 0);
}

void DI_AudioDevice::getInterfaces(ulong *xpIDArray) const
{
  ulong   nN;
  ulong   nIndex;
  ALvalue *xpValues;

  nN = getNumberOfInterfaces();
  xpValues = new ALvalue[nN];
  alQueryValues(m_device, AL_INTERFACE, xpValues, (int)nN, NULL, 0);
  for (nIndex=0; nIndex<nN; nIndex++)
  {
    xpIDArray[nIndex] = (ulong)(xpValues[nIndex].i);
  }
  delete [] xpValues;
}

bool DI_AudioDevice::selectInterface(ulong interface)
{
  ALpv xParams;

  xParams.param = AL_INTERFACE;
  xParams.value.i = (int)interface;
  
  if (alSetParams(m_device, &xParams, 1) != 1)
  {
    DI_error(DI_ERR_INTERNAL, "Could not change interface setting");
    return false;
  }
  return true;
}

ulong DI_AudioDevice::getInterface() const
{
  ALpv xParams;
  
  xParams.param = AL_INTERFACE;
  if (alGetParams(m_device, &xParams ,1) == 1)
  {
    return (ulong)(xParams.value.i);
  }
  else
  {
    DI_error(DI_ERR_INTERNAL, "Could not get active interface");
    return 0;
  }
}

bool DI_AudioDevice::setInterfaceGains(ulong interface, const double *gains)
{
  ulong nChannels = getChannels();
  ALpv xParams;
  long long *llpGains = new long long[nChannels];

  for (ulong nIndex=0; nIndex<nChannels; nIndex++)
    llpGains[nIndex] = alDoubleToFixed(gains[nIndex]);

  xParams.param = AL_GAIN;
  xParams.value.ptr = llpGains;
  xParams.sizeIn = (short)nChannels;
  
  if (alSetParams((int)interface, &xParams, 1) != 1)
  {
    DI_error(DI_ERR_INTERNAL, "Could not change interface gains");
    return false;
  }
  return true;
}

bool DI_AudioDevice::getInterfaceGains(ulong interface, double *gains)
{
  ulong nChannels = getChannels();
  ALpv xParams;
  long long *llpGains = new long long[nChannels];
  
  xParams.param = AL_GAIN;
  xParams.value.ptr = llpGains;
  xParams.sizeIn = (short)nChannels;
  if (alGetParams((int)interface, &xParams ,1) == 1)
    for (ulong nIndex=0; nIndex<nChannels; nIndex++)
      gains[nIndex] = alFixedToDouble(llpGains[nIndex]);
  else
  {
    DI_error(DI_ERR_INTERNAL, "Could not get interface gains");
    return false;
  }
  return true;
}

bool DI_AudioDevice::setInterfaceMuting(ulong interface, bool mute)
{
  ALpv xParams;

  xParams.param = AL_MUTE;
  xParams.value.i = (mute ? 1 : 0);
  
  if (alSetParams((int)interface, &xParams, 1) != 1)
  {
    DI_error(DI_ERR_INTERNAL, "Could not change muting setting");
    return false;
  }
  return true;
}
  
bool DI_AudioDevice::getInterfaceMuting(ulong interface)
{
  ALpv xParams;
  
  xParams.param = AL_MUTE;
  if (alGetParams((int)interface, &xParams ,1) == 1)
  {
    return (bool)(xParams.value.i);
  }
  else
  {
    DI_error(DI_ERR_INTERNAL, "Could not get muting setting");
    return false;
  }
}

// --- Private methods --------------------------------------------------------

void DI_AudioDevice::getResourceName(int ID, char *name)
{
  ALpv xParam;

  xParam.param = AL_NAME;
  xParam.value.ptr = name;
  xParam.sizeIn = DI_AUDIO_MAX_NAME_LENGTH;

  alGetParams(ID, &xParam, 1);
}

void DI_AudioDevice::getResourceLabel(int ID, char *label)
{
  ALpv xParam;

  xParam.param = AL_LABEL;
  xParam.value.ptr = label;
  xParam.sizeIn = DI_AUDIO_MAX_NAME_LENGTH;

  alGetParams(ID, &xParam, 1);
}
