/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the DIVA IO library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000-2001.
 * Jarmo.Hiipakka@hut.fi Tommi.Ilmonen@hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA IO library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */

//
// DIVA Library. Copyright (C) Jarmo Hiipakka, 1999-2001.
//
// $Log: di_audio_port.h,v $
// Revision 1.4  2004/02/06 15:47:59  tilmonen
// *** empty log message ***
//
// Revision 1.3  2002/11/25 09:22:42  tilmonen
// Fake debug option.
//
// Revision 1.2  2001/07/27 08:23:27  tilmonen
// License text updates (=cosmetic changes).
//
// Revision 1.1.1.1  2001/06/04 09:04:21  tilmonen
// Mustajuuri 4.6.2001, initial CVS cersion
//

#ifndef DI_AUDIO_PORT_H
#define DI_AUDIO_PORT_H

#include <audio.h>
#include <stropts.h>
#include <poll.h>

#include "di_audio_base.h"

#define DI_AUDIO_IS_SGI

/// Maximum length of audio library names and labels.
const uint DI_AUDIO_MAX_NAME_LENGTH = 32;

/// Audio library negative infinity gain.
const double DI_AUDIO_NEG_INFINITY = -10000.0;

class DI_AudioDevice;

/** Audio signal I/O.

    DI_AudioPort is declared in "di_audio_port.h" and defined in libdiio.

    {\large {\bf Introduction}}

    DI_AudioPort provides an easy, stream-like access to platform audio
    hardware. Theoretically, DI_AudioPort provides complete hardware
    independency, in practice, this remains to be seen as the current
    version supports only SGI and OSS audio libraries.

*/

class DI_AudioPort : public DI_AudioBase
{
public:

  /**@name Constructors */
  
  //@{

  /** Create a new audio port.
    
      This constructor creates a new audio port. The port is unconnected
      and is made operational only by using the open method. It is
      recommended that the port is opened only when it is needed, and
      closed when not needed anymore.
  */

  DI_AudioPort();
  
  ///

  ~DI_AudioPort(void);

  //@}
  
  /**@name Access */

  //@{

  /// Return the audio input device name.

  const char *inName(void) const { return (const char *)m_inName; }

  /// Return the audio output device name.

  const char *outName(void) const { return (const char *)m_outName; }

  /// Return the audio input device label.

  const char *inLabel(void) const { return (const char *)m_inLabel; }

  /// Return the audio output device label.

  const char *outLabel(void) const { return (const char *)m_outLabel; }

  /// Return a pointer to the input device. Not available on all platforms!
  
  DI_AudioDevice *inDevice(void) const { return m_inDevice; }

  /// Return a pointer to the output device. Not available on all platforms!
  
  DI_AudioDevice *outDevice(void) const { return m_outDevice; }

  /// Return input sample rate.

  uint inSampleRate(void) const;

  /// Return output sample rate.

  uint outSampleRate(void) const;

  /// Return number of input channels.

  uint inChannels(void) const;

  /// Return number of output channels.

  uint outChannels(void) const;

  /// Get number of filled sample frames in the input sample buffer

  int inFilledFrames(void) const;
  
  /// Get number of filled sample frames in the output sample buffer

  int outFilledFrames(void) const;
  
  //@}

  /**@name Operations */
  
  //@{

  /** Open the audio port.

      This method opens an actual audio port to a named audio device.
      The audio port may be opened for recording or playback or for both.
      However, the supported modes may differ on different platforms.

      The 'bufferSize' parameter sets the audio queue buffer size
      (in sample frames) that  will be used as a temporary data storage
      during recording or playing.

      The 'maxBlock' parameters sets the maximum block size (in sample frames)
      that can be read or written on a single call to either readFrames or
      writeFrames.

  */
  
  bool open(/// Audio device.
	    const char *device,
	    /// IO mode: none, input, output, or input&output.
	    DI_IoMode mode,
	    /// Audio port sample format.
	    DI_AudioSampleFormat userSample, 
	    /// Audio port buffer size (sample frames).
	    uint bufferSize,
	    /// Audio sampling rate.
	    uint samplerate,
	    /// Number of channels.
	    uint channels,
	    /// Maximum block size.
	    uint maxBlock);

  /** Open the audio port to a pre-defined audio device.

      This method is similar to the previous open method. The only
      difference is that using this method, you can initialize the
      audio device beforehand. This method is only available on SGI
      platform.

  */

  bool open(/// Audio device the port is connected to.
	    DI_AudioDevice &device,
	    /// IO mode: none, input, output, or input&output.
	    DI_IoMode mode,
	    /// Audio port sample format.
	    DI_AudioSampleFormat userSample, 
	    /// Audio port buffer size (sample frames).
	    uint bufferSize,
	    /// Maximum block size.
	    uint maxBlock);
  
  /** Open the audio port.

      This method opens an actual audio port to named audio devices.
      The first device is opened for input, and the second is opened for
      output. The whole port structure is in the input&output state.

  */
  
  bool open(/// Audio input device.
	    const char *inDevice,
	    /// Audio output device.
	    const char *outDevice,
	    /// Audio port sample input format.
	    DI_AudioSampleFormat inFormat, 
	    /// Audio port sample output format.
	    DI_AudioSampleFormat outFormat, 
	    /// Audio port input buffer size (sample frames).
	    uint inBufSize,
	    /// Audio port output buffer size (sample frames).
	    uint outBufSize,
	    /// Input sampling rate.
	    uint inSamplerate,
	    /// Output sampling rate.
	    uint outSamplerate,
	    /// Number of input channels.
	    uint inChannels,
	    /// Number of output channels.
	    uint outChannels,
	    /// Maximum input block size.
	    uint inMaxBlock,
	    /// Maximum output block size.
	    uint outMaxBlock);

  /** Open the audio port to pre-defined audio devices.

      This method is similar to the previous open method. The only
      difference is that using this method, you can initialize the
      audio devices beforehand. This method is only available on SGI
      platform.

  */

  bool open(/// Audio input device the port is connected to.
	    DI_AudioDevice &inDevice,
	    /// Audio output device the port is connected to.
	    DI_AudioDevice &outDevice,
	    /// Audio port sample input format.
	    DI_AudioSampleFormat inFormat, 
	    /// Audio port sample output format.
	    DI_AudioSampleFormat outFormat, 
	    /// Audio port input buffer size (sample frames).
	    uint inBufSize,
	    /// Audio port output buffer size (sample frames).
	    uint outBufSize,
	    /// Maximum input block size.
	    uint inMaxBlock,
	    /// Maximum output block size.
	    uint outMaxBlock);
  
  /// Close the port.

  bool close(void);

  //@}

  /**@name I/O operations */
  
  //@{

  /** Read sample frames recorded by audio hardware.

      This method reads 'frames' sample frames from audio hardware recording
      device, and stores them into buffer pointed by 'buffer'.

      Please note that one sample may take more than one byte, so
      'buffer' must point to a buffer of capable of holding 'frames'
      * N * M bytes, where N is the size of one sample and M is number of
      channels.
     
      The number of sample frames actually recorded is
      returned, or a negative value if the operation failed.

  */

  int readFrames(void *buffer, uint frames);

  /** Write sample frames to be played by audio hardware.

      This method writes 'frames' sample frames to audio hardware playback
      device, reading them from buffer pointer by 'buffer'. The number
      of sample frames actually played is returned, or a negative value if
      the operation failed. NOTE! This method may change the contents of
      the buffer.
      
      writeFrames() may, and usually does, return before the playback
      sequence is completed, storing the sample data into a temporary
      buffer and leaving the playback subsystem running after write()
      returns. In these cases, the returned value shows how many sample frames
      the the audio subsystem is going to try to play.

  */

  int writeFrames(const void *buffer,uint frames);
  
  /** Wait until 'frames' # of frames can be read.
      
      Return value tells how many frames are filled as the call
      returns. Negative values indicate error of some sort.

  */
  
  int waitInputFilled(uint frames, int timeOut = -1);

  /// Wait until output has only the number of frames specified filled.  

  int waitOutputFilled(uint frames, int timeOut = -1);

  /** Start recording/playback. On IRIX and OSS this method does
      nothing, on ALSA it is compulsory. */
  bool start() { return true; }

  //@}

  /**@name Predicates */
  
  //@{

  ///

  bool isOK(void) const;

  ///

  bool isEOF(void) const { return false; }

  //@}

  static void setDebug(bool) {}

private:

  inline DI_AudioPort &operator = (const DI_AudioPort &xOriginal)
  { (void)xOriginal; return *this; }

  DI_AudioDevice       *m_inDevice;
  DI_AudioDevice       *m_outDevice;
  int                   m_inFD;
  int                   m_outFD;
  pollfd                m_inPFD;
  pollfd                m_outPFD;
  bool                  m_deleteDev;

  char                  m_inName[DI_AUDIO_MAX_NAME_LENGTH];
  char                  m_outName[DI_AUDIO_MAX_NAME_LENGTH];
  char                  m_inLabel[DI_AUDIO_MAX_NAME_LENGTH];
  char                  m_outLabel[DI_AUDIO_MAX_NAME_LENGTH];
  
  ALconfig              m_inConfig;
  ALport                m_inPort;
  ALconfig              m_outConfig;
  ALport                m_outPort;
};



// --- Audio Device Class -----------------------------------------------------

/* NOTE!

   This class breaks the nice idea of having one header file for each class.
   However, this class is only available on SGI platform, and to keep things
   simple, it is included here.
*/


/** Audio hardware setup.

    DI_AudioDevice is declared in "di_audio_port.h" and defined in divaio.

    {\large {\bf Introduction}}

    DI_AudioDevice provides a hardware independent access to platform audio
    hardware. Theoretically, DI_AudioDevice provides complete hardware
    independency, in practice, this remains to be seen as the current
    version supports only SGI audio library.
    
    see: DI_AudioPort
*/

class DI_AudioDevice
{
  friend class DI_AudioPort;

public:

  /**@name Utilities */

  //@{

  /// Return the number of audio subsystem input devices.

  static uint getNumberOfInputDevices();

  /// Return the number of audio subsystem output devices.

  static uint getNumberOfOutputDevices();

  /** Get the ID's of the audio subsystem input devices.

      Returns the ID's of the audio subsystem input devices. The ID's are
      returned in an array of 'getNumberOfInputDevices()' uint's, which
      the user must allocate before calling this method.

  */

  static void getInputDevices(ulong *IDArray);

  /** Get the ID's of the audio subsystem output devices.

      Returns the ID's of the audio subsystem output devices. The ID's are
      returned in an array of 'getNumberOfOutputDevices()' ulong's, which
      the user must allocate before calling this method.

  */

  static void getOutputDevices(ulong *IDArray);

  /** Get the name of an audio input device.

      Returns the name of the audio device with ID 'devID'. The name can
      be DI_AUDIOMAXNAMELENGTH (incl. '\0') chars long at most and the
      array must be allocated by the caller. The name is intended to be
      used internally in programs; see 'getDeviceLabel' for human readable
      labels.

  */

  static void getInputDeviceName(ulong devID, char *name)
  { DI_AudioDevice::getResourceName((int)devID, name); }

  /** Get the name of an audio output device.

      Returns the name of the audio device with ID 'devID'. The name can
      be DI_AUDIOMAXNAMELENGTH (incl. '\0') chars long at most and the
      array must be allocated by the caller. The name is intended to be
      used internally in programs; see 'getDeviceLabel' for human readable
      labels.

  */

  static void getOutputDeviceName(ulong devID, char *name)
  { DI_AudioDevice::getResourceName((int)devID, name); }

  /** Get the label of the audio device.

      Returns the (human readable) label of the audio device with ID 'devID'.
      The label can be DI_AUDIOMAXNAMELENGTH chars long at most and the array
      must be allocated by the caller.

  */

  static void getDeviceLabel(ulong devID, char *label)
  { DI_AudioDevice::getResourceLabel((int)devID, label); }

  //@}

  /**@name Constructors */

  //@{

  /** Create a new default audio device.

      This constructor creates a new default audio device. On systems
      supporting only one physical audio device, this is the constructor
      to use. On systems capable of handling many different audio i/o
      devices, this constructor could create an interface to any of such
      devices (usually a reasonable default device).

  */

  DI_AudioDevice(/// Create input (true) or output (false) device.
		 bool input,
		 /// Number of device audio channels.
		 ulong channels,
		 /// Device sampling rate (defaults to current rate).
		 ulong rate=0);

  /** Create a new interface to a specific audio device.

      This constuctor creates a new interface to audio hardware device
      with ID 'ID'. On systems supporting only one physical audio
      device, this constructor always creates a new default audio device.
      On systems capable of handling many different audio i/o devices,
      'ID' must be a valid device identifier.

  */
  
  DI_AudioDevice(/// ID of the desired device.
		 ulong ID,
		 /// Create input (true) or output (false) device.
		 bool input,
		 /// Number of device audio channels.
		 ulong channels,
		 /// Device sampling rate (defaults to current rate).
		 ulong rate=0);

  /** Create a new interface to a named audio device.

      This constuctor creates a new interface to audio hardware device
      called '*name'. On systems supporting only one physical audio
      device, this constructor always creates a new default audio device.
      On systems capable of handling many different audio i/o devices,
      'name' must point to a name, the system recognises as one of
      the audio devices.

  */
  
  DI_AudioDevice(/// Name of the desired device.
		 const char *name,
		 /// Create input (true) or output (false) device.
		 bool input,
		 /// Number of device audio channels.
		 ulong channels,
		 /// Device sampling rate (defaults to current rate).
		 ulong rate=0);

  /// 

  ~DI_AudioDevice(void);

  //@}

  /**@name Operations */
  
  //@{

  /** Get the name of the audio device.

      Returns the name of the audio device. The name can be at most
      DI_AUDIOMAXNAMELENGTH (incl. '\0') chars long and the array must
      be allocated by the caller. The name is intended to be used
      internally in programs; see 'getLabel' for human-readable labels.

  */

  void getName(char *name) const
  { DI_AudioDevice::getResourceName(m_device, name); }

  /** Get the label of the audio device.

      Returns the (human readable) label of the audio device.
      The label can be DI_AUDIOMAXNAMELENGTH chars long at most and the array
      must be allocated by the caller.

  */

  void getLabel(char *label) const
  { DI_AudioDevice::getResourceLabel(m_device, label); }
  
protected:

  /// Return device number

  inline int getDevice(void) const { return m_device; }
  
public:

  /// Return the number of channels.

  ulong getChannels(void);

  /// Return the current sample rate.

  ulong getRate(void);

  /// Set the number of channels.

  bool setChannels(ulong channels);

  /// Set the sample rate.

  bool setRate(ulong rate);

  //@}

  /**@name Predicates */
  
  //@{

  /// Return 'true' if device was opened for recording.

  bool isInput(void) const;

  /// Return 'true' if device was opened for playback.

  inline bool isOutput(void) const { return !isInput(); }

  /// Return 'true' if device is OK.

  bool isOK(void) const { return m_OK; }

  //@}

  /**@name Interface Operations */

  //@{

  /// Return the # of possible interfaces (physical jacks) for this device.

  ulong getNumberOfInterfaces() const;

  /** Get the ID's of the possible interfaces for this device.

      Returns the ID's of the possible interfaces for this device. The ID's
      are returned in an array of 'getNumberOfInterfaces()' ulong's, which
      the user must allocate before calling this method.

  */

  void getInterfaces(ulong *IDArray) const;

  /** Get the name of the interface.

      Returns the name of the interface with ID 'interface'. The name can
      be DI_AUDIOMAXNAMELENGTH (incl. '\0') chars long at most and the
      array must be allocated by the caller. The name is intended to be
      used internally in programs; see 'GetInterfaceLabel' for human-readable
      labels.

  */

  void getInterfaceName(ulong interface, char *name) const
  { DI_AudioDevice::getResourceName((int)interface, name); }

  /** Get the label of the interface.

      Returns the (human readable) label of the interface with ID 'interface'.
      The label can be DI_AUDIOMAXNAMELENGTH chars long at most and the array
      must be allocated by the caller.

  */

  void getInterfaceLabel(ulong interface, char *label) const
  { DI_AudioDevice::getResourceLabel((int)interface, label); }

  /** Set the active interface.

      In systems capable of using just one interface at a time, this method
      sets the interface to be used.
  */
  
  bool selectInterface(ulong interface);

  /// Get the ID of the active interface.
  
  ulong getInterface() const;

  /** Set the gains of a specific interface.

      Sets the channel gains of a specific interface (in dBs). Positive values
      in array xpGains represent gain, negative values represent attenuation.
      Some interfaces also accept the special value of DI_AUDIONEGINFINITY,
      which sets the gain to negative infinity desibels (mute).
  */
  
  bool setInterfaceGains(ulong interface, const double *gains);

  /** Get the gains of a specific interface

      Returns the channel gains of a specific interface (in dBs) in array
      'xpGains'. 'xpGains' must point to an array of doubles. The array
      must be at least 'GetChannels()' values long and it must be allocated
      by the caller.
  */

  bool getInterfaceGains(ulong interface, double *gains);

  /** Mute (mute=true) or unmute (mute=false) a specific interface.

      This method sets or unsets the muting of a specific interface. Muting
      does not affect the value of the interface gain. If an applications
      needs to know the true gain, it should call both 'getInterfaceGains()'
      and 'getInterfMute()'.
  */
  
  bool setInterfaceMuting(ulong interface, bool mute);

  /// Get the muting setting of an interface.
  
  bool getInterfaceMuting(ulong interface);
  
  //@}

private:

  static void getResourceName(int ID, char *name);
  static void getResourceLabel(int ID, char *label);

  inline DI_AudioDevice &operator = (const DI_AudioDevice &original)
  { (void)original; return *this; }

  int               m_device;
  bool              m_input;
  ulong             m_channels;
  ulong             m_rate;
  bool              m_OK;
};

#endif  // DI_AUDIO_PORT_H
