/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the DIVA IO library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000-2001.
 * Jarmo.Hiipakka@hut.fi Tommi.Ilmonen@hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA IO library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */

#include "di_midi_io.h"

#include "di_error.h"

#include <dmedia/dmedia.h>

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>

#include <iostream>

const char *DI_defaultMidiDev()
{
  return "Software Synth";
}

bool DI_irixMdToMjMidi(DI_MidiEvent *me, const MDevent *md)
{
  memcpy(me->message(), md->msg, 4);
  
  me->setTime((long) (md->stamp / 100000));

  if(md->sysexmsg) {
    me->setSysex(md->sysexmsg, md->msglen);
  }
  return true;
}


bool DI_mjMidiToIrixMd(MDevent *md, DI_MidiEvent *me)
{
  memcpy(md->msg, me->message(), 4);
  
  md->stamp = ((long long) me->time()) * 100000;

  /* md->stamp = 0;
     md->msglen = 0; */

  /* char buf[128];
     sprintf(buf, "0x%x", * (int *) md->msg);
     cout << buf << endl; */


  if(me->sysexSize()) {
    if(md->sysexmsg)
      mdFree(md->sysexmsg);

    md->sysexmsg = (char *) mdMalloc(me->sysexSize());
    
    memcpy(md->sysexmsg, me->sysexData(), me->sysexSize());
    
    md->msglen = me->sysexSize();
  }
   
  /* char buf2[128];
     mdPrintEvent(buf2, md, 1);
     cout << buf2 << endl; */
    
  return true;
}


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

DI_MidiIo::DI_MidiIo()
  : m_portName(0),
    m_mode(DI_IO_NONE),
    m_inPort(0),
    m_outPort(0)
{
  mdInit();  // Make sure this function gets called.
  bzero(&m_event, sizeof(m_event));
}

DI_MidiIo::~DI_MidiIo()
{
  close();
}

bool DI_MidiIo::open(const char * port,
		     DI_IoMode mode)
{
  errno = 0;
  const char * fname = "DI_MidiIo::open";

  DI_trace(fname);
  
  if(port) m_portName = strdup(port);
  else port = DI_defaultMidiDev();
  
  if(m_inPort || m_outPort) {
    DI_error(DI_ERR_INTERNAL, fname, "Already open");
    return false;
  }

  if((mode != DI_IO_INPUT) && 
     (mode != DI_IO_OUTPUT) &&
     (mode != DI_IO_INPUT_OUTPUT) ) {
    DI_error(DI_ERR_INTERNAL, fname, "Bad mode");
    return false;
  }  

  // Open ports:
  if((int) mode & (int) DI_IO_INPUT) {
    m_inPort = mdOpenInPort(port);
    if(!m_inPort) {
      close();
      return false;
    }
  }

  if((int) mode & (int) DI_IO_OUTPUT) {
    m_outPort = mdOpenOutPort(port);
    if(!m_outPort) {
      close();
      return false;
    }
  }
  
  if(errno) { 
    /* cerr << "Errno detected 0: " << strerror(errno)
       << "\nCould not open MIDI, port=" << port << endl; */

    errno = 0;
    /* close();
       return false; */
  }
  
  // Stamp modes:
  if(m_inPort) {
    if(mdSetStampMode(m_inPort, MD_RELATIVESTAMP) < 0) {
      DI_error(DI_ERR_SYSTEM, "Unable to set MIDI in stamp mode");
      return false;
    }
  }

  if(m_outPort) {
    if(mdSetStampMode(m_outPort, MD_RELATIVESTAMP) < 0) {
      DI_error(DI_ERR_SYSTEM, "Unable to set MIDI out stamp mode");
      return false;
    }
  }

  if(m_inPort) {
    m_pollFd.fd = mdGetFd(m_inPort);
    m_pollFd.events = POLLIN;
    m_pollFd.revents = 0;
  }
  
  m_mode = mode;
  
  // Final preparations:  
  resetTiming();

  return true;
}

bool DI_MidiIo::close()
{
  const char *fname = "DI_MidiIo::close";
  
  DI_trace(fname);
  
  if(m_inPort) {
    mdClosePort(m_inPort);
    m_inPort = 0;
  }
  if(m_outPort) {
    mdClosePort(m_outPort);
    m_outPort = 0;
  }

  // Remove the port name
  if(m_portName) {
    free(m_portName);
    m_portName = 0;
  }

  m_mode = DI_IO_NONE;

  return true;
}

void DI_MidiIo::resetTiming()
{
  unsigned long long ust;

  dmGetUST(&ust);
  if(m_inPort)  mdSetStartPoint(m_inPort, ust, 0);
  if(m_outPort) mdSetStartPoint(m_outPort, ust, 0);
}

bool DI_MidiIo::eventPending()
{
  return poll(&m_pollFd, 1, 0) != 0;
}

bool DI_MidiIo::readEvent(DI_MidiEvent *event)
{
  if(mdReceive(m_inPort, &m_event, 1) <= 0) return false;

  DI_irixMdToMjMidi(event, &m_event);    

  if(m_event.sysexmsg) mdFree(m_event.sysexmsg);
  
  return true;
}

bool DI_MidiIo::writeEvent(DI_MidiEvent *event)
{
  DI_mjMidiToIrixMd(&m_event, event);

  return mdSend(m_outPort, &m_event, 1) > 0;
}


