/* COPYRIGHT
 *
 * This file is part of the DIVA IO library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000.
 * Jarmo.Hiipakka@hut.fi Tommi.Ilmonen@hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA IO library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */


#include "di_audio_port.h"
#include "di_error.h"
#include "di_sleep.h"
#include "di_numconv.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <sys/time.h>
#include <unistd.h>

// 2^15 - 1
#define DI_ALSA_SSCALE 32767
// 2^31 - 1
#define DI_ALSA_ISCALE 2147483647

/** Template class that handles data type conversions for ALSA version
    of DI_audioPort. 
    
    This class provides static conversion functions that are used as
    function pointers by the DI_AudioPort. */

template <class TUser, class TAlsa> 
class DI_apBufferConversion
{
public:  
  /// Interleaved user data to interleaved ALSA data
  static void iUserToiAlsa(char**, const void *, 
			   uint, uint, uint, uint);
  /// Interleaved ALSA data to interleaved user data
  static void ialsaToIUser(const char**, void *, 
			   uint, uint, uint, uint);

  /// Interleaved user data to non-interleaved ALSA data
  static void iUserToAlsa(char**, const void *, 
			  uint, uint, uint, uint);
  /// Non-interleaved ALSA data to interleaved user data
  static void alsaToIUser(const char**, void *, 
			  uint, uint, uint, uint);

  static inline TAlsa toAlsa(const TUser &x) { return (TAlsa) x; }
  static inline TUser toUser(const TAlsa &x) { return (TUser) x; }
};

// Specialized conversion functions.

// Float & int
inline int DI_apBufferConversion<float,int>::toAlsa(const float &x)
{ return (int) (x * DI_ALSA_ISCALE); }

inline float DI_apBufferConversion<float,int>::toUser(const int &x)
{ return ((float) x) * (1.0 / (DI_ALSA_ISCALE + 1)); }

// Float & short
/* The strange bit-shift operation is done to minimize rounding
   errors. Direct casting always happens towards zero which not not
   very nice -> nonlinearity. In case of 32-bit ints that is hardly
   important, but with 16-bit samples all bits really count. */
inline short DI_apBufferConversion<float,short>::toAlsa(const float &x)
{ return (short) (((int) (x * DI_ALSA_ISCALE)) >> 16); }

inline float DI_apBufferConversion<float,short>::toUser(const short &x)
{ return ((float) x) * (1.0 / (DI_ALSA_SSCALE + 1)); }

// double & int
inline int DI_apBufferConversion<double,int>::toAlsa(const double &x)
{ return (int) (x * DI_ALSA_ISCALE); }

inline double DI_apBufferConversion<double,int>::toUser(const int &x)
{ return (double) (x * (1.0 / DI_ALSA_ISCALE)); }

// double & short
inline short DI_apBufferConversion<double,short>::toAlsa(const double &x)
{ return (short) (((int) (x * DI_ALSA_ISCALE)) >> 16); }

inline double DI_apBufferConversion<double,short>::toUser(const short &x)
{ return ((double) x) * (1.0 / (DI_ALSA_SSCALE + 1)); }

// Int & short
inline short DI_apBufferConversion<int,short>::toAlsa(const int &x)
{ return (short) (x >> 16); }

inline int DI_apBufferConversion<int,short>::toUser(const short &x)
{ return ((int) x) << 16; }

// Short & int
inline int DI_apBufferConversion<short,int>::toAlsa(const short &x)
{ return ((int) x) << 16; }

inline short DI_apBufferConversion<short,int>::toUser(const int &x)
{ return (short) (x >> 16); }



template <class TUser, class TAlsa> 
void DI_apBufferConversion<TUser,TAlsa>::iUserToiAlsa(char**addrs, 
						      const void *buffer, 
						      uint frames, 
						      uint channels, 
						      uint offset,
						      uint systemChannels)
{
  for(uint i=0; i < channels; i++) {

    const TUser *source = ((const TUser *) buffer) + i;
    TAlsa * target = (TAlsa *) addrs[i] + offset * systemChannels;

    /* DI_trace2("DI_apBufferConversion::iUserToiAlsa %d %p -> %p",
       i, source, target); */
    
    for(uint j=0; j < frames; j++) {
      *target = toAlsa(source[j *  channels]);
      target += systemChannels;
    }
  }
}

template <class TUser, class TAlsa> 
void DI_apBufferConversion<TUser,TAlsa>::ialsaToIUser(const char**addrs, 
						      void *buffer, 
						      uint frames, 
						      uint channels, 
						      uint offset,
						      uint systemChannels)
{
  TUser *target = (TUser *) buffer;

  for(uint i=0; i < channels; i++) {
    const TAlsa * source = (const TAlsa *) addrs[i] + systemChannels * offset;
    
    for(uint j=0; j < frames; j++) {
      target[i + j *  channels] = toUser(*source);
      source += systemChannels;
    }
  }  
}

template <class TUser, class TAlsa> 
void DI_apBufferConversion<TUser,TAlsa>::iUserToAlsa(char**addrs, 
						      const void *buffer, 
						      uint frames, 
						      uint channels, 
						      uint offset,
						      uint systemChannels)
{
  const TUser *source = (const TUser *) buffer;
  
  
  for(uint i=0; i < channels; i++) {
    TAlsa * target = (TAlsa *) addrs[i] + offset;
    
    for(uint j=0; j < frames; j++)
      *target++ = toAlsa(source[i + j *  channels]);
  }
}

template <class TUser, class TAlsa> 
void DI_apBufferConversion<TUser,TAlsa>::alsaToIUser(const char**addrs, 
						     void *buffer, 
						     uint frames, 
						     uint channels, 
						     uint offset,
						     uint systemChannels)
{
  TUser *target = (TUser *) buffer;
  
  for(uint i=0; i < channels; i++) {
    const TAlsa * source = (const TAlsa *) addrs[i] + offset;
    
    for(uint j=0; j < frames; j++)
      target[i + j *  channels] = toUser(*source++);
  }
}

template class DI_apBufferConversion<float, int>;
template class DI_apBufferConversion<float, short>;

template class DI_apBufferConversion<double, int>;
template class DI_apBufferConversion<double, short>;

template class DI_apBufferConversion<int, int>;
template class DI_apBufferConversion<int, short>;

template class DI_apBufferConversion<short, int>;
template class DI_apBufferConversion<short, short>;

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

#define USE_BLOCK_MODE 1


// --- Constructors -----------------------------------------------------------

bool DI_AudioPort::m_debug = false;

DI_AudioPort::DI_AudioPort()
  : m_phandle(0),
    m_chandle(0),
    m_phw(0),
    m_chw(0),
    m_psw(0),
    m_csw(0),
    m_paddr(0),
    m_caddr(0),
    m_bytesPerSample(0),
    m_mode(DI_IO_NONE),
    m_name(0)
{
}

DI_AudioPort::~DI_AudioPort(void)
{
  close();
}

// --- Operations -------------------------------------------------------------

bool DI_AudioPort::open(const char *device,
			DI_IoMode mode,
			DI_AudioSampleFormat userSample, 
			ulong bufferSize,
			ulong samplerate,
			ulong channels,
			ulong maxBlock,
			bool interleave)
{
  const char * fname = "DI_AudioPort::open";
  // char buf[512];

  if(!device) device = "hw:0,0";

  if(m_debug) DI_trace2("%s # %s, bufSize = %d, maxBlock = %d",
			fname, device, bufferSize, maxBlock);
  
  m_userSample = userSample;

  switch(m_userSample) {
  case DI_ASF_INT16:
    m_userSampleBytes = 2;
    break;
  case DI_ASF_INT32:
  case DI_ASF_FLOAT32:
    m_userSampleBytes = 4;
    break;
  case DI_ASF_FLOAT64:
    m_userSampleBytes = 8;
    break;
  default:
    DI_error(DI_ERR_ARGUMENT, fname, "bad user sample type");
    return false;
  }

  m_inChannels  = channels;
  m_outChannels = channels;

  m_inRate = samplerate;
  m_outRate = samplerate;

  m_fragments = 0;
  m_bufferFlushes = 0;

  if(!samplerate) samplerate = 44100;

  if(mode & DI_IO_OUTPUT) {

    bool openOk = !snd_pcm_open(&m_phandle, 
				device, 
				SND_PCM_STREAM_PLAYBACK, 
				0);
    
    if((snd_pcm_hw_params_malloc(&m_phw) < 0) ||
       (snd_pcm_sw_params_malloc(&m_psw) < 0) ||
       !openOk) {

      if(m_debug) DI_trace(fname, "Playback open");

      DI_error(DI_ERR_MEMORY, fname, "playback init failed");
      close();
      return false;
    }

    if(m_debug) DI_trace(fname, "Output opened");
  }

  if(mode & DI_IO_INPUT) {
    if((snd_pcm_hw_params_malloc(&m_chw) < 0) ||
       (snd_pcm_sw_params_malloc(&m_csw) < 0) ||
       snd_pcm_open(&m_chandle, 
		    device, 
		    SND_PCM_STREAM_CAPTURE, 
		    0)) {

      DI_error(DI_ERR_MEMORY, fname, "capture init failed");
      close();
      return false;
    }
    if(m_debug) DI_trace(fname, "Input opened");
  }
  
  uint fragmentFrames = 1;
  
  while(fragmentFrames < maxBlock)
    fragmentFrames = fragmentFrames << 1;
  
  if(m_phandle) {
    m_outChannels = channels;
    m_systemOutChannels = channels;

    if(!configure(m_phandle, 
		  m_phw, 
		  m_psw, 
		  fragmentFrames, 
		  bufferSize, 
		  & m_systemOutChannels)) {
      close();
      return false;
    }

    m_outBlock = snd_pcm_hw_params_get_period_size(m_phw, 0);  
    m_framesPerFragment = m_outBlock;
    if(m_debug) DI_trace(fname, "Playback configured");
  }


  if(m_chandle) {
    m_inChannels = channels;
    m_systemInChannels = channels;

    if(!configure(m_chandle, 
		  m_chw, 
		  m_csw, 
		  fragmentFrames, 
		  bufferSize, 
		  &m_systemInChannels)) {
      close();
      return false;
    }
    
    m_inBlock = snd_pcm_hw_params_get_period_size(m_chw, 0);
    m_framesPerFragment = m_inBlock;

    if(m_debug) DI_trace(fname, "Capture configured");
  }

  if(m_chandle && m_phandle) {

    m_linkStart = ! snd_pcm_link(m_chandle, m_phandle);

    if(m_inBlock != m_outBlock) {
      DI_error(DI_ERR_UNKNOWN, fname, "non-equal fragment sizes");
      close();
      return false;
    }

    if(snd_pcm_hw_params_get_format(m_phw) != 
       snd_pcm_hw_params_get_format(m_chw)) {
      DI_error(DI_ERR_UNKNOWN, fname, "different sample formats");
      close();
      return false;
    }
  }
  else m_linkStart = false;

  m_fragmentChannelBytes = m_bytesPerSample * m_framesPerFragment;

  if(m_phandle)
    m_bytesPerFragment = m_fragmentChannelBytes * m_systemOutChannels;
  else if(m_chandle)
    m_bytesPerFragment = m_fragmentChannelBytes * m_systemInChannels;

  if(m_debug) DI_trace(fname, "Getting mmap areas");

  if(!getMmapAreas()) {
    close();
    return false;
  }

  if(m_debug) DI_trace(fname, "Getting function pointers");

  if(!configureFunctionPointers()) {
    DI_error(DI_ERR_UNKNOWN, fname, "Could not set function pointers");
    return false;
  }

  m_bufferWrittenFrames = 0;
  m_bufferReadFrames = m_framesPerFragment;
  // m_bufferReadFrames = 0;
  m_xrun = false;
  m_name = strdup(device);

  if(m_debug) DI_trace(fname, "EXIT OK");

  return true;
}

bool DI_AudioPort::start()
{
  // return true;

  const char * fname = "DI_AudioPort::start";

  int err;

  /* RRR: 0 may not be the right value */

  m_pfrag = 0;
  m_cfrag = 0;

  if(m_debug) DI_trace(fname);
  
  if(m_phandle) {
    if ((err = snd_pcm_prepare (m_phandle)) < 0) {
      DI_error(DI_ERR_INTERNAL, fname, "Failed to prepare playback",
	       snd_strerror(err));
      return false;
    }
  }

  if (!m_linkStart && m_chandle) {
    if ((err = snd_pcm_prepare (m_chandle)) < 0) {
      DI_error(DI_ERR_INTERNAL, fname, "Failed to prepare capture",
	       snd_strerror(err));
      return false;
    }
  }
  
  if(m_debug) {
    DI_trace(fname, m_fragments);
    DI_trace(fname, m_framesPerFragment);
    
    if(m_phandle && m_chandle)
      DI_trace(fname, m_linkStart ? "LINKED START" : "NONLINKED START");
  }


  if(m_phandle) {
    snd_pcm_mmap_commit(m_phandle, 
			m_poffset, 
			m_fragments * m_framesPerFragment);

    if ((err = snd_pcm_start (m_phandle)) < 0) {
      DI_error(DI_ERR_UNKNOWN, fname, "could not start playback",
	       snd_strerror (err));
      return false;
    }
  }

  if (!m_linkStart && m_chandle) {
    if ((err = snd_pcm_start (m_chandle)) < 0) {
      DI_error(DI_ERR_UNKNOWN, fname, "could not start capture",
	       snd_strerror (err));
      return false;
    }
  }

  if(m_debug) DI_trace(fname, "EXIT OK");
  
  return true;
}

bool DI_AudioPort::stop()
{
  const char * fname = "DI_AudioPort::stop";

  int err;

  if(m_debug) DI_trace(fname);

  if(m_phandle) {
    if ((err = snd_pcm_drop (m_phandle)) < 0) {
      DI_error(DI_ERR_UNKNOWN, fname, "playback drop failed", 
	       snd_strerror(err));
      return false;
    }
  }

  if (!m_linkStart && m_chandle) {
    if ((err = snd_pcm_drop (m_chandle)) < 0) {
      DI_error(DI_ERR_UNKNOWN, fname, "capture drop failed", 
	       snd_strerror(err));
      return false;
    }
  }

  if(m_debug) DI_trace(fname, "EXIT OK");

  return true;
}

bool DI_AudioPort::close(void)
{
  const char *fname = "DI_AudioPort::close";
  int e;
  bool ok = true;

  if(m_phandle) {
    snd_pcm_unlink(m_phandle);

    snd_pcm_drop(m_phandle);
    
    e = snd_pcm_close(m_phandle);
    if(e) {
      DI_error(DI_ERR_SYSTEM, fname, snd_strerror(e));
      ok = false;
    }
    m_phandle = 0;
  }

  if(m_chandle) {
    snd_pcm_unlink(m_chandle);

    snd_pcm_drop(m_chandle);

    e = snd_pcm_close(m_chandle);
    if(e) {
      DI_error(DI_ERR_SYSTEM, fname, snd_strerror(e));
      ok = false;
    }
    m_chandle = 0;
  }

  /* freeMmappArea(m_paddr);
  freeMmappArea(m_caddr);

  // Unallocate hardware params:

  if(m_phw) {
    snd_pcm_hw_params_free(m_phw);
    m_phw = 0;
  }

  if(m_chw) {
    snd_pcm_hw_params_free(m_chw);
    m_chw = 0;
  }

  // Unallocate software params:

  if(m_psw) {
    snd_pcm_sw_params_free(m_psw);
    m_psw = 0;
  }

  if(m_csw) {
    snd_pcm_sw_params_free(m_csw);
    m_csw = 0;
  }

  if(m_name) {
    free(m_name);
    m_name = 0;
    }*/

  m_fragments = 0;
  m_fragmentChannelBytes = 0;

  m_bytesPerSample = 0;
  m_bytesPerFragment = 0;

  m_systemOutChannels = 0;
  m_systemInChannels = 0;

  m_mode = DI_IO_NONE;
  
  return ok;
}


// --- I/O operations ---------------------------------------------------------

int DI_AudioPort::readFrames(void *buffer, ulong frames)
{
  const char *fname = "DI_AudioPort::readFrames";

  if(!m_chandle) 
    DI_fatal(DI_ERR_INTERNAL, fname, "m_chandle = NULL");
  
  uint handled = 0;

  while(frames) {   
    
    if(m_bufferReadFrames == m_framesPerFragment) { 
      int e = snd_pcm_mmap_commit(m_chandle, 
				  m_coffset, 
				  m_framesPerFragment);      

      m_coffset = (m_coffset + m_framesPerFragment) % 
	(m_fragments * m_framesPerFragment);

      if(e <= 0) {
	DI_error(DI_ERR_UNKNOWN, fname, snd_strerror(e));
	return 0;
      }

      m_cfrag = (m_cfrag + 1) % m_fragments;
      m_bufferReadFrames = 0;
    }
    
    uint avail = m_framesPerFragment - m_bufferReadFrames;

    uint readFrames = (frames < avail) ? frames : avail;

    (*m_mmapToFrames)((const char **)m_caddr + m_cfrag * m_systemInChannels, 
		      & (((char *) buffer)
			 [handled * m_outChannels * m_userSampleBytes]), 
		      readFrames, 
		      m_inChannels, 
		      m_bufferReadFrames, 
		      m_systemInChannels);
    
    m_bufferReadFrames += readFrames;
    handled += readFrames;
    frames -= readFrames;
  }


  return frames;
}

int DI_AudioPort::writeFrames(const void *buffer, ulong frames)
{
  const char *fname = "DI_AudioPort::writeFrames";

  uint handled = 0;

  while(frames) {
    uint avail = m_framesPerFragment - m_bufferWrittenFrames;
    
    uint writeFrames = (frames < avail) ? frames : avail;
    
    /* DI_trace2("%s  %d %d %d %d", 
	      fname , handled, (int) m_outChannels, 
	      m_userSampleBytes, writeFrames);  */
    
    (*m_framesToMmap)(m_paddr + m_pfrag * m_systemOutChannels, 
		      & (((const char *) buffer)
			 [handled * m_outChannels * m_userSampleBytes]), 
		      writeFrames, 
		      m_outChannels, 
		      m_bufferWrittenFrames, 
		      m_systemOutChannels);
    
    // DI_trace("saved");
    
    m_bufferWrittenFrames += writeFrames;
    handled += writeFrames;

    if(m_bufferWrittenFrames == m_framesPerFragment) {
      
      // DI_trace("write");
      int e = snd_pcm_mmap_commit(m_phandle, 
				  m_poffset, 
				  m_framesPerFragment);

      m_poffset = (m_poffset + m_framesPerFragment) % 
	(m_fragments * m_framesPerFragment);

      if(e <= 0) {
	DI_error(DI_ERR_UNKNOWN, fname, snd_strerror(e));
	return 0;
      }

      m_bufferFlushes++;
      m_bufferWrittenFrames = 0;
      m_pfrag = (m_pfrag + 1) % m_fragments;
    }
    
    frames -= writeFrames;
  }

  return handled;
}

int DI_AudioPort::inFilledFrames(void) const 
{ 
  if(m_phandle && !m_bufferFlushes) return 0;

  int f = snd_pcm_avail_update (m_chandle);
  if(f < 0)  {
    ((DI_AudioPort *) this)->m_xrun = true; // CONST !!!
    DI_error(DI_ERR_UNKNOWN, "DI_AudioPort::inFilledFrames");
  }
  return f;
}

int DI_AudioPort::outFilledFrames(void) const 
{ 
  int f = snd_pcm_avail_update(m_phandle);
  if(f < 0) {
    ((DI_AudioPort *) this)->m_xrun = true;
    DI_error(DI_ERR_UNKNOWN, "DI_AudioPort::outFilledFrames");
  }

  return m_fragments * m_framesPerFragment - f;
}

int DI_AudioPort::outEmptyFrames()
{
  int f = snd_pcm_avail_update (m_phandle);
  if(f < 0) {
    m_xrun = true;
    DI_error(DI_ERR_UNKNOWN, "DI_AudioPort::outEmptyFrames", snd_strerror(f));
  }

  if(m_fragments > 2) {
    if((uint) f <= m_framesPerFragment) f = 0;
    else f -= m_framesPerFragment;
  }

  return f;
}


bool DI_AudioPort::check()
{
  // #ifdef ALSA_IS_0_9
#if 0

  if(!m_xrun) return true;
  
  if(!xrunRecovery()) {
    m_inAvail = 0;
    m_outAvail = 0;
    m_outWritten = 0;
    return false;
  }
  m_xrun = false;
#endif

  return true;
}


int DI_AudioPort::xrunRecovery ()
{
  if(m_debug) DI_trace("DI_AudioPort::xrunRecovery");
  return stop() && start();
}

int DI_AudioPort::waitInputFilled(ulong frames, int timeOut)
{
  ulong iff;

  while((iff = inFilledFrames()) < frames)
    ;
  
  return iff;
}

int DI_AudioPort::waitOutputFilled(ulong frames, int timeOut)
{
  struct timeval tv;
  int r;

  while(true) {
    // DI_trace("DI_AudioPort::waitOutputFilled", outFilledFrames());
    if((r = outFilledFrames()) > (int) frames) {
      // gettimeofday(&tv, 0);
      tv.tv_sec = 0;      // Zero full seconds.
      tv.tv_usec = 2000;  // two millisecs.
      select(0, 0, 0, 0, & tv);
    }
    else
      break;
  }
  return r;
}




// --- Predicates -------------------------------------------------------------

bool DI_AudioPort::isOK(void) const
{
  return true;
}


bool DI_AudioPort::configure(snd_pcm_t *handle,
			     snd_pcm_hw_params_t *hwp,
			     snd_pcm_sw_params_t *swp,
			     uint fragmentFrames,
			     uint bufferSize,
			     uint *channels)
{
  const char * fname = "DI_AudioPort::configure";

  int e;

  if(m_debug) DI_trace(fname, fragmentFrames);

  // Hardware parameters:
  
  if((e = snd_pcm_hw_params_any(handle, hwp)) < 0) {
    DI_error(DI_ERR_INTERNAL, fname, "No param combination available");
    return false;
  }

  if(m_debug) hwDump(fname, hwp);

  if(!seekFormat(handle, hwp))
    return false;
  
  // Access mode:
  if((e = snd_pcm_hw_params_set_access
      (handle, hwp, SND_PCM_ACCESS_MMAP_NONINTERLEAVED)) < 0) {

    if((e = snd_pcm_hw_params_set_access
	(handle, hwp, SND_PCM_ACCESS_MMAP_INTERLEAVED)) < 0) {
    
      DI_error(DI_ERR_INTERNAL, fname, "No mmapped noninterleaved access");
      return false;
    }
    m_interleave = true;
  }
  else
    m_interleave = false;
  
  // Channels
  uint origChannels = *channels;
  *channels = snd_pcm_hw_params_get_channels_max (hwp);

  if(origChannels > *channels) {
    DI_error(DI_ERR_INTERNAL, fname, "Requested too many channels");
    return false;
  }

  if((e = snd_pcm_hw_params_set_channels(handle, hwp, *channels)) < 0) {
    DI_error(DI_ERR_INTERNAL, fname, "Could not set channels");
    hwParamsTryExplainFailure(handle, hwp);
    return false;    
  }

  // Sampling rate:
  e = snd_pcm_hw_params_set_rate(handle, hwp, m_inRate, 0);
  
  if(e) {
    if ((e = snd_pcm_hw_params_set_rate_near(handle, hwp, m_inRate, 0)) < 0) {
      DI_error(DI_ERR_INTERNAL, fname, "Could not set sampling rate");
      hwParamsTryExplainFailure(handle, hwp);
      return false;
    }
  }

  uint hwRate = snd_pcm_hw_params_get_rate(hwp, 0);
  uint diff;
  if(hwRate > m_inRate) diff = hwRate - m_inRate;
  else diff = m_inRate - hwRate;
  
  if(diff > 100) {
    DI_error2(DI_ERR_SYSTEM, 
	      "%s # Failed to get required sampling rate %d != %d",
	      fname, m_inRate, hwRate);
  }
  if(m_debug) DI_trace2("%s # rate = %d", fname, (int) hwRate);
  m_inRate = hwRate;

  // Fragment size

  uint minFrag = snd_pcm_hw_params_get_period_size_min(hwp, 0);
  uint maxFrag = snd_pcm_hw_params_get_period_size_max(hwp, 0);
  
  while(fragmentFrames < minFrag) fragmentFrames = fragmentFrames << 1;
  while(fragmentFrames > maxFrag) fragmentFrames = fragmentFrames >> 1;
  
  if(!snd_pcm_hw_params_set_period_size_near(handle, hwp, fragmentFrames, 0)) {
    DI_error(DI_ERR_UNKNOWN, fname, "Unable to set any frame size");
    return false;
  }
  
  fragmentFrames = snd_pcm_hw_params_get_period_size(hwp, 0);

  // Number of fragments
  uint frags;
  
  if(m_fragments) {
    frags = m_fragments;
  }
  else{
    frags = bufferSize / fragmentFrames;

    if(frags < 3) frags = 3;

    uint minFrags = snd_pcm_hw_params_get_periods_min(hwp, 0);
    uint maxFrags = snd_pcm_hw_params_get_periods_max(hwp, 0);
  
    if(frags < minFrags) frags = minFrags;
    if(frags > maxFrags) frags = maxFrags;
  
  }

  if(!snd_pcm_hw_params_set_periods_near(handle, hwp, frags, 0)) {
    DI_error(DI_ERR_UNKNOWN, fname, "Unable to set any buffer size");
    return false;
  }

  m_fragments = snd_pcm_hw_params_get_periods(hwp, 0);

  bufferSize = m_fragments * fragmentFrames;

  // Buffer size (what is it ?)
  if(!snd_pcm_hw_params_set_buffer_size_near(handle, hwp, bufferSize)) {
    DI_error(DI_ERR_UNKNOWN, fname, "Unable to set any buffer size");
    return false;
  }

  if(m_debug) hwDump(fname, hwp);

  if ((e = snd_pcm_hw_params (handle, hwp)) < 0) {
    DI_error(DI_ERR_UNKNOWN, "Failed to set hw params");
    return false;    
  }

  // return true;

  // Software parameters:
  snd_pcm_sw_params_current(handle, swp);
  m_explicitStart = snd_pcm_sw_params_set_start_threshold 
    (handle, swp, ~0U) >= 0;
  
  snd_pcm_sw_params_set_stop_threshold(handle, swp, ~0U);  
  snd_pcm_sw_params_set_avail_min(handle, swp, fragmentFrames);

  // snd_pcm_sw_params_set_avail_min(handle, swp, 0);
  // snd_pcm_sw_params_set_period_step(handle, swp, 1);

  if ((e = snd_pcm_sw_params (handle, swp)) < 0) {
    DI_error(DI_ERR_UNKNOWN, "Failed to set sw params");
    return false;    
  }
  if(m_debug) {
    swDump(fname, swp);
    pcmDump(fname, handle);
  }

  return true;
}
  

void DI_AudioPort::hwParamsTryExplainFailure(snd_pcm_t *handle,
					     snd_pcm_hw_params_t *hwp)
{  
#if 0
  snd_pcm_hw_params_t *hwp2 = 0;
  snd_output_t *out = 0;
  char *buf = 0;

  snd_pcm_hw_params_alloca (&hwp2);
  snd_output_buffer_open (&out);
  
  snd_pcm_hw_params_try_explain_failure (handle, hwp, hwp2, 10, out);
  snd_output_buffer_string (out, &buf);

  DI_error(DI_ERR_UNKNOWN, buf);
  
  // free (buf);
  snd_output_close (out);
#endif
}

void DI_AudioPort::hwDump(const char *fname, snd_pcm_hw_params_t *hwp)
{
  char *buf = 0;
  snd_output_t *out = 0;
  
  snd_output_buffer_open (&out);
  snd_pcm_hw_params_dump(hwp, out);
  snd_output_buffer_string (out, &buf);
  
  DI_trace(fname, buf);
  
  // free (buf);
  snd_output_close (out);
}

void DI_AudioPort::swDump(const char *fname, snd_pcm_sw_params_t *swp)
{
  char *buf = 0;
  snd_output_t *out = 0;
  
  snd_output_buffer_open (&out);
  snd_pcm_sw_params_dump(swp, out);
  snd_output_buffer_string (out, &buf);
  
  DI_trace(fname, buf);
  
  // free (buf);
  snd_output_close (out);
}

void DI_AudioPort::pcmDump(const char *fname, snd_pcm_t *handle)
{
  char *buf = 0;
  snd_output_t *out = 0;
  
  snd_output_buffer_open (&out);
  snd_pcm_dump(handle, out);
  snd_output_buffer_string (out, &buf);
  
  DI_trace(fname, buf);
  
  // free (buf);
  snd_output_close (out);
}

bool DI_AudioPort::seekFormat(snd_pcm_t *handle,
			      snd_pcm_hw_params_t *hwp)
{
  const char * fname = "DI_AudioPort::seekFormat";

  snd_pcm_format_t alsaSample;
  
  // Determine the optimal sample format:
  switch(m_userSample) {
  case DI_ASF_INT16:
    alsaSample = SND_PCM_FORMAT_S16_LE;
    break;
  case DI_ASF_INT32:
    alsaSample = SND_PCM_FORMAT_S32_LE;
    break;
  case DI_ASF_FLOAT32:
    alsaSample = SND_PCM_FORMAT_FLOAT_LE;
    break;
  case DI_ASF_FLOAT64:
    alsaSample = SND_PCM_FORMAT_FLOAT64_LE;
    break;
  default:
    DI_error(DI_ERR_ARGUMENT, fname, "Bad sample format");
    return false;
  }
  
  // Iterate to get the best possible sample format
  while(snd_pcm_hw_params_set_format(handle, hwp, alsaSample) < 0) {

    // Downgrade sample format:
    switch(alsaSample) {

    case SND_PCM_FORMAT_FLOAT64_LE:
      alsaSample = SND_PCM_FORMAT_FLOAT_LE;
      break;

    case SND_PCM_FORMAT_FLOAT_LE:
      alsaSample = SND_PCM_FORMAT_S32_LE;
      break;

    case SND_PCM_FORMAT_S32_LE:
      alsaSample = SND_PCM_FORMAT_S16_LE;
      break;

    default: 
      DI_error(DI_ERR_UNKNOWN, fname, "unable to set sample format");
      return false;
    }
  }

  m_alsaSample = alsaSample;
  m_bytesPerSample = snd_pcm_format_physical_width(alsaSample) / 8;

  return true;
}

bool DI_AudioPort::getMmapAreas()
{
  if(m_phandle) {
    if(!getMmapArea(m_phandle, & m_poffset, m_paddr, m_systemOutChannels))
      return false;
  }
  if(m_chandle) {
    if(!getMmapArea(m_chandle, & m_coffset, m_caddr, m_systemInChannels))
      return false;
  }
  return true;
}

bool DI_AudioPort::getMmapArea(snd_pcm_t *handle, 
			       snd_pcm_uframes_t *offset, 
			       char **&addrs, 
			       uint chans)
{
  const char * fname = "DI_AudioPort::getMmapArea";

  *offset = 0;

  const snd_pcm_channel_area_t *areas;

  snd_pcm_uframes_t avail = 0;

  int err = snd_pcm_mmap_begin(handle, & areas, offset, &avail);
  
  if (err) {
    DI_error(DI_ERR_UNKNOWN, fname, "snd_pcm_mmap_begin failed");
    return false;
  }

  uint bytes = sizeof(char *) * m_fragments * chans;

  addrs = (char **) malloc(bytes);
  bzero(addrs, bytes);

  if(!addrs) {
    DI_error(DI_ERR_MEMORY, fname, "Memory allocation failed");
    return false;
  }

  for(uint frag = 0; frag < m_fragments; frag++) {
    char **tmpAddr = addrs + (frag * chans);

    uint fragJump;
    
    if(m_interleave)
      fragJump = frag * m_fragmentChannelBytes * chans;
    else
      fragJump = frag * m_fragmentChannelBytes;
    
    if(m_debug) DI_trace(fname, fragJump);

    for(uint chan = 0; chan < chans; chan++) {
      tmpAddr[chan] = ((char *) areas[chan].addr) + areas[chan].first / 8;
      tmpAddr[chan] += fragJump;

      zeroAlsaBuffer(tmpAddr[chan], m_framesPerFragment, chans);

      if(m_debug) DI_trace2("New pointer, frag %d, chan %d, ptr = %p", 
			    frag, chan, tmpAddr[chan]);
    }
  }
  return true;
}

void DI_AudioPort::freeMmappArea(char **&addrs)
{
  if(addrs) free(addrs);
  addrs = 0;
}

bool DI_AudioPort::configureFunctionPointers()
{
  if(m_interleave) {
    if(m_userSample ==  DI_ASF_FLOAT64) {
      if(m_alsaSample == SND_PCM_FORMAT_S32_LE) {
	m_framesToMmap = DI_apBufferConversion<double, int>::iUserToiAlsa;
	m_mmapToFrames = DI_apBufferConversion<double, int>::ialsaToIUser;
      }
      else if(m_alsaSample == SND_PCM_FORMAT_S16_LE) {
	m_framesToMmap = DI_apBufferConversion<double, short>::iUserToiAlsa;
	m_mmapToFrames = DI_apBufferConversion<double, short>::ialsaToIUser;
      }
      else
	return false;
    }
    else if(m_userSample ==  DI_ASF_FLOAT32) {
      if(m_alsaSample == SND_PCM_FORMAT_S32_LE) {
	m_framesToMmap = DI_apBufferConversion<float, int>::iUserToiAlsa;
	m_mmapToFrames = DI_apBufferConversion<float, int>::ialsaToIUser;
      }
      else if(m_alsaSample == SND_PCM_FORMAT_S16_LE) {
	m_framesToMmap = DI_apBufferConversion<float, short>::iUserToiAlsa;
	m_mmapToFrames = DI_apBufferConversion<float, short>::ialsaToIUser;
      }
      else
	return false;
    }
    else if(m_userSample ==  DI_ASF_INT32) {
      if(m_alsaSample == SND_PCM_FORMAT_S32_LE) {
	m_framesToMmap = DI_apBufferConversion<int, int>::iUserToiAlsa;
	m_mmapToFrames = DI_apBufferConversion<int, int>::ialsaToIUser;
      }
      else if(m_alsaSample == SND_PCM_FORMAT_S16_LE) {
	m_framesToMmap = DI_apBufferConversion<int, short>::iUserToiAlsa;
	m_mmapToFrames = DI_apBufferConversion<int, short>::ialsaToIUser;
      }
      else
	return false;
    }
    else 
      return false;
  }
  else {
    if(m_userSample ==  DI_ASF_FLOAT32) {
      if(m_alsaSample == SND_PCM_FORMAT_S32_LE) {
	m_framesToMmap = DI_apBufferConversion<float, int>::iUserToAlsa;
	m_mmapToFrames = DI_apBufferConversion<float, int>::alsaToIUser;
      }
      else if(m_alsaSample == SND_PCM_FORMAT_S16_LE) {
	m_framesToMmap = DI_apBufferConversion<float, short>::iUserToAlsa;
	m_mmapToFrames = DI_apBufferConversion<float, short>::alsaToIUser;
      }
      else
	return false;
    }
    else 
      return false;
  }
  return true;
}

/** This method clears the mmapped sound card driver buffers. */

void DI_AudioPort::zeroAlsaBuffer(void *buf, uint frames, uint totalChans)
{
  const char * fname = "DI_AudioPort::zeroAlsaBuffer";

  if(m_interleave) {

    if(m_alsaSample == SND_PCM_FORMAT_S16_LE || 
       m_alsaSample == SND_PCM_FORMAT_S16_BE) {
      short *sbuf = (short *) buf;
      for(uint i=0; i < frames; i++)
	sbuf[i * totalChans] = 0;
    }
    else if(m_alsaSample == SND_PCM_FORMAT_S32_LE||
	    m_alsaSample == SND_PCM_FORMAT_S32_BE) {
      int *ibuf = (int *) buf;
      for(uint i=0; i < frames; i++)
	ibuf[i * totalChans] = 0;
    }
    else 
      DI_error(DI_ERR_INTERNAL, fname, "Unsupported ALSA sample format");
    
  }
  else {
    if(m_alsaSample == SND_PCM_FORMAT_S16_LE || 
       m_alsaSample == SND_PCM_FORMAT_S16_BE) {
      bzero(buf, frames * 2);
    }
    else if(m_alsaSample == SND_PCM_FORMAT_S32_LE ||
	    m_alsaSample == SND_PCM_FORMAT_S32_BE) {
      bzero(buf, frames * 4);
    }
    else 
      DI_error(DI_ERR_INTERNAL, fname, "Unsupported ALSA sample format");
  }
}
