/* COPYRIGHT
 *
 * This file is part of the DIVA IO library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000-2001.
 * Jarmo.Hiipakka@hut.fi Tommi.Ilmonen@hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA IO library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */

#include <di_midi_io.h>

#include <di_error.h>

const char *DI_defaultMidiDev()
{
  return "UNDEFINED";
}

static void DI_midiAlsaToDiva(const snd_seq_event_t * ev,
			 DI_MidiEvent * e, uint ticks)
{
  switch(ev->type) {
  case SND_SEQ_EVENT_NOTEOFF:
    e->noteOff(ev->data.note.channel,
	       ev->data.note.note,
	       ev->data.note.velocity, ticks);
    break;
  case SND_SEQ_EVENT_NOTEON:
    e->noteOn(ev->data.note.channel,
	      ev->data.note.note,
	      ev->data.note.velocity, ticks);
    break;
  case SND_SEQ_EVENT_CONTROLLER:
    e->controller(ev->data.control.channel, 
		  ev->data.control.param,
		  ev->data.control.value, ticks);
    break;

  case SND_SEQ_EVENT_PITCHBEND:
    e->pitchBend(ev->data.control.channel, 
		 ev->data.control.param,
		 ev->data.control.value, ticks);
    break;

  default:
    e->clearMessage();
    
  }

}


DI_MidiIo::DI_MidiIo()
  : m_seq(0)
{}

DI_MidiIo::~DI_MidiIo()
{
  close();
}

/// Open the port.
bool DI_MidiIo::open(const char * port, DI_IoMode)
{
  const char * fname = "DI_MidiIo::open";

  int portid;

  if (snd_seq_open(&m_seq, "default", SND_SEQ_OPEN_INPUT, 0) < 0) {
    DI_error2(DI_ERR_SYSTEM, "%s # Error opening ALSA sequencer.", fname);
    return false;
  }

  snd_seq_set_client_name(m_seq, "DIVA IO");
  if ((portid = snd_seq_create_simple_port(m_seq, "DIVA IO",
            SND_SEQ_PORT_CAP_WRITE|SND_SEQ_PORT_CAP_SUBS_WRITE,
            SND_SEQ_PORT_TYPE_APPLICATION)) < 0) {
    DI_error2(DI_ERR_SYSTEM, "%s # Error creating sequencer port.", fname);
    return false;
  }

  return true;
}

/// Close the port.
bool DI_MidiIo::close()
{
  if(!m_seq)
    return false;

  snd_seq_close(m_seq);
  m_seq = 0;

  return true;
}

/// Resets the base timing against which events are compared.
void DI_MidiIo::resetTiming()
{

}

/// Returns true if there is an incoming event pending.
bool DI_MidiIo::eventPending()
{
  if(m_mode & DI_IO_INPUT)
    return snd_seq_event_input_pending(m_seq, 0) > 0;

  return false;
}

/** Reads a MIDI event. Blocks until the event can be read. */
bool DI_MidiIo::readEvent(DI_MidiEvent *e)
{
  snd_seq_event_t *ev = 0;
  snd_seq_event_input(m_seq, &ev);

  DI_midiAlsaToDiva(ev, e, 0);

  snd_seq_free_event(ev);

  return true;
}

/** Writes the event to the port. On Linux/OSS we use raw MIDI
    interface -> The message is written immediately rather than at
    the correct time. */
bool DI_MidiIo::writeEvent(DI_MidiEvent *)
{

  return false;
}
