/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the DIVA IO library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000-2001.
 * Jarmo.Hiipakka@hut.fi Tommi.Ilmonen@hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA IO library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */

//
// DIVA Library. Copyright (C) Jarmo Hiipakka and Tommi Ilmonen, 1999-2001.
//
// $Log: di_audio_port.h,v $
// Revision 1.1  2004/02/06 15:47:59  tilmonen
// *** empty log message ***
//
// Revision 1.4  2003/08/18 06:56:18  tilmonen
// Fixes.
//
// Revision 1.3  2002/11/21 16:00:47  tilmonen
// *** empty log message ***
//
// Revision 1.2  2002/04/24 07:15:33  tilmonen
// Minor updates (changed one include).
//
// Revision 1.1  2001/10/02 06:08:21  tilmonen
// Wrappers for alsa 0.9 beta 7 (sigh).
//
// Revision 1.3  2001/07/27 15:48:59  tilmonen
// Bug fix.
//
// Revision 1.2  2001/07/27 08:23:28  tilmonen
// License text updates (=cosmetic changes).
//
// Revision 1.1.1.1  2001/06/04 09:04:22  tilmonen
// Mustajuuri 4.6.2001, initial CVS cersion
//

#ifndef DI_AUDIO_PORT_H
#define DI_AUDIO_PORT_H

#include <alsa/asoundlib.h>

#include <stropts.h>
#include <poll.h>

#include "di_audio_base.h"

#define DI_AUDIO_IS_ALSA 1
#define ALSA_IS_0_9 1

/// Maximum length of audio library names and labels.
const ulong DI_AUDIO_MAX_NAME_LENGTH = 128;

/// Audio library negative infinity gain.
const double DI_AUDIO_NEG_INFINITY = -10000.0;


/** 
    Audio port.
    
*/

class DI_AudioPort : public DI_AudioBase
{
public:

  /** Create a new audio port.
    
      This constructor creates a new audio port. The port is unconnected
      and is made operational only by using the open method. It is
      recommended that the port is opened only when it is needed, and
      closed when not needed anymore.
  */

  DI_AudioPort();
  
  virtual ~DI_AudioPort(void);

  virtual const char *inName(void) const { return m_name; }
  virtual const char *outName(void) const { return m_name; }
  virtual const char *inLabel(void) const { return m_name; }
  virtual const char *outLabel(void) const { return m_name; }

  /** Open the audio port.

      This method opens an actual audio port to a named audio device.
      The audio port may be opened for recording or playback or for both.
      However, the supported modes may differ on different platforms.

      The 'bufferSize' parameter sets the audio queue buffer size
      (in sample frames) that  will be used as a temporary data storage
      during recording or playing.

      The 'maxBlock' parameters sets the maximum block size (in sample frames)
      that can be read or written on a single call to either readFrames or
      writeFrames.

  */
  
  bool open(/// Audio device.
	    const char *device,
	    /// IO mode: none, input, output, or input&output.
	    DI_IoMode mode,
	    /// Audio port sample format.
	    DI_AudioSampleFormat userSample, 
	    /// Audio port buffer size (sample frames).
	    ulong bufferSize,
	    /// Audio sampling rate.
	    ulong samplerate,
	    /// Number of channels.
	    ulong channels,
	    /// Maximum block size.
	    ulong maxBlock,
	    bool interleave = true);


  /// Close the port.

  bool close(void);

  /** Start recording/playback. On IRIX and OSS this method does
      nothing, on ALSA it is compulsory. */

  bool start();

  bool stop();

  /**@name I/O operations */
  
  //@{

  /** Read sample frames recorded by audio hardware.

      This method reads 'frames' sample frames from audio hardware recording
      device, and stores them into buffer pointed by 'buffer'.

      Please note that one sample may take more than one byte, so
      'buffer' must point to a buffer of capable of holding 'frames'
      * N * M bytes, where N is the size of one sample and M is number of
      channels.
     
      The number of sample frames actually recorded is
      returned, or a negative value if the operation failed.

  */

  int readFrames(void *buffer, uint frames);

  /** Write sample frames to be played by audio hardware.

      This method writes 'frames' sample frames to audio hardware playback
      device, reading them from buffer pointer by 'buffer'. The number
      of sample frames actually played is returned, or a negative value if
      the operation failed. NOTE! This method may change the contents of
      the buffer.
      
      writeFrames() may, and usually does, return before the playback
      sequence is completed, storing the sample data into a temporary
      buffer and leaving the playback subsystem running after write()
      returns. In these cases, the returned value shows how many sample frames
      the the audio subsystem is going to try to play.

  */

  int writeFrames(const void *buffer,uint frames);

  /// Get number of filled sample frames in the input sample buffer

  int inFilledFrames(void) const;
  
  /** Get number of filled sample frames in the output sample buffer. 

      Also checks the number of underflows in the buffer.
   */

  int outFilledFrames(void) const;

  int outEmptyFrames();

  /** Check that everything is ok, may restart the audio io to make it
      ok. */
  bool check();
  
  int xrunRecovery ();

  /** Wait until 'frames' # of frames can be read.
      
      Return value tells how many frames are filled as the call
      returns. Negative values indicate error of some sort.

  */
  
  int waitInputFilled(uint frames, int timeOut = -1);

  /// Wait until output has only the number of frames specified filled.  

  int waitOutputFilled(uint frames, int timeOut = -1);

  //@}

  /**@name ALSA specials */
  
  //@{

  uint underFlows() { return m_underFlows; }

  bool xrun() const { return m_xrun; }

  uint minimumBlock() const { return m_framesPerFragment; }
  
  uint bufferFlushes() const { return  m_bufferFlushes; }

  //@}

  /**@name Predicates */
  
  //@{

  ///

  bool isOK(void) const;
  
  ///
  
  bool isEOF(void) const { return false; }

  static void setDebug(bool b) { m_debug = b; }

  //@}

protected:
  
  bool configure(snd_pcm_t *handle,
		 snd_pcm_hw_params_t *hwp,
		 snd_pcm_sw_params_t *swp,
		 uint fragmentFrames,
		 uint bufferSize,
		 uint *channels);
  
  void hwParamsTryExplainFailure(snd_pcm_t *handle,
				 snd_pcm_hw_params_t *hwp);
  
  void hwDump(const char *fname, snd_pcm_hw_params_t *hwp);
  void swDump(const char *fname, snd_pcm_sw_params_t *swp);
  void pcmDump(const char *fname, snd_pcm_t *swp);

  bool seekFormat(snd_pcm_t *handle,
		  snd_pcm_hw_params_t *hwp);

  bool getMmapAreas();

  bool getMmapArea(snd_pcm_t *handle, 
		   snd_pcm_uframes_t *offset, 
		   char **&addrs, 
		   uint chans);

  void freeMmappArea(char **&addrs);

  bool configureFunctionPointers();
  
  void zeroAlsaBuffer(void *buf, uint frames, uint totalChans);
  
  inline DI_AudioPort &operator = (const DI_AudioPort &xOriginal)
  { (void)xOriginal; return *this; }

  snd_pcm_t            *m_phandle;
  snd_pcm_t            *m_chandle;

  snd_pcm_hw_params_t  *m_phw;
  snd_pcm_hw_params_t  *m_chw;

  snd_pcm_sw_params_t  *m_psw;
  snd_pcm_sw_params_t  *m_csw;

  /// Playback offset
  snd_pcm_uframes_t     m_poffset;
  /// Capture offset
  snd_pcm_uframes_t     m_coffset;
  
  /// Playback addresses
  char                **m_paddr;
  /// Capture addresses
  char                **m_caddr;

  /* Pointer to a function that transforms user samples to mmapped
     memory areas. 

     Arguments: 
     
     1) Pointers to mmapped region
     2) Pointer to user data
     3) Number of sample-frames to write
     4) Number of channels
     5) Offset (number of samples already written)
  */
  void  (*m_framesToMmap) (char**, const void *, uint, uint, uint, uint);
  void  (*m_mmapToFrames) (const char**, void *, uint, uint, uint, uint);

  // Playback and capture fragments
  uint                  m_pfrag;
  uint                  m_cfrag;

  uint                  m_fragments;
  uint                  m_framesPerFragment;
  uint                  m_fragmentChannelBytes;

  uint                  m_bytesPerSample;
  uint                  m_bytesPerFragment;
  
  uint                  m_bufferReadFrames;
  uint                  m_bufferWrittenFrames;

  uint                  m_systemInChannels;
  uint                  m_systemOutChannels;

  uint                  m_bufferFlushes;
  
  bool                  m_interleave;
  bool                  m_explicitStart;
  bool                  m_linkStart;
  bool                  m_xrun;

  /// Alsa mode.
  snd_pcm_format_t      m_alsaSample;
  DI_AudioSampleFormat  m_userSample;
  uint                  m_userSampleBytes;
  
  uint                  m_underFlows;
  
  DI_IoMode             m_mode;

  char                 *m_name;

  static bool           m_debug;
};

#endif  // DI_AUDIO_PORT_H





