/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the DIVA IO library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000-2001.
 * Jarmo.Hiipakka@hut.fi Tommi.Ilmonen@hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA IO library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */
//
// DIVA Library. Copyright (C) Tommi Ilmonen 2001
//
// $Log: di_audio_port.C,v $
// Revision 1.2  2001/07/27 08:23:28  tilmonen
// License text updates (=cosmetic changes).
//
// Revision 1.1.1.1  2001/06/04 09:04:21  tilmonen
// Mustajuuri 4.6.2001, initial CVS cersion
//

#include "di_audio_port.h"
#include "di_error.h"
#include "di_sleep.h"
#include "di_numconv.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <sys/time.h>
#include <unistd.h>

#define USE_BLOCK_MODE 1

// 2^15 - 1
#define DI_ALSA_SSCALE 32767
// 2^31 - 1
#define DI_ALSA_ISCALE 2147483647

// --- Constructors -----------------------------------------------------------

DI_AudioPort::DI_AudioPort()
  : m_phandle(0),
    m_chandle(0),
    m_tmpReadBuffer(0),
    m_tmpWriteBuffer(0),
    m_tmpBufferRead(0),
    m_tmpBufferWriteCached(0),
    m_tmpBufferFrames(0),
    
    m_bytesPerSample(0),
    m_mode(DI_IO_NONE)    
{
}

DI_AudioPort::~DI_AudioPort(void)
{
  close();
}

// --- Operations -------------------------------------------------------------

bool DI_AudioPort::open(const char *device,
			DI_IoMode mode,
			DI_AudioSampleFormat userSample, 
			ulong bufferSize,
			ulong samplerate,
			ulong channels,
			ulong maxBlock,
			bool interleave)
{
  const char * fname = "DI_AudioPort::open";
  char buf[512];

  DI_trace(fname, (int) mode);

  m_userSample = userSample;

  if(mode == DI_IO_INPUT) 
    m_amode = SND_PCM_OPEN_CAPTURE;
  else if(mode == DI_IO_OUTPUT) 
    m_amode = SND_PCM_OPEN_PLAYBACK;
  else if(mode == DI_IO_INPUT_OUTPUT) 
    m_amode = SND_PCM_OPEN_DUPLEX;
  else {
    DI_error(DI_ERR_ARGUMENT, fname, "Bad mode");
    return false;
  }

  int cardId = -1;
  int dev = -1;
  int subDev = -1;

  if(!device) device = "0 0";

  sscanf(device, "%d %d %d", &cardId, &dev, &subDev);
  
  /* #if 0
  if(device) cardId = snd_card_name(device);
  else cardId = snd_defaults_pcm_card();
#else
  cardId = snd_defaults_pcm_card();
  cardId = 1;
  #endif */

  sprintf(buf, "card = %d, dev = %d, subdev = %d",
	  cardId, dev, subDev);
  
  if(cardId < 0) cardId = snd_defaults_pcm_card();
  if(dev < 0) dev = 0;
  int e = 0;

  DI_trace(fname, buf);

  if(mode & DI_IO_INPUT) {
    if(subDev < 0)
      e = snd_pcm_open(&m_chandle, cardId, dev, SND_PCM_OPEN_CAPTURE);
    else
      e = snd_pcm_open_subdevice(&m_chandle, cardId, dev, subDev, 
				 SND_PCM_OPEN_CAPTURE);
    if(!m_chandle || e) {
      DI_error(DI_ERR_SYSTEM, fname, "Could not open input");
      close();
      return false;
    }
  }
  DI_trace(fname, "0.6");
  if(mode & DI_IO_OUTPUT) {
    if(subDev < 0)
      e = snd_pcm_open(&m_phandle, cardId, dev, SND_PCM_OPEN_PLAYBACK);
    else
      e = snd_pcm_open_subdevice(&m_phandle, cardId, dev, subDev, 
				 SND_PCM_OPEN_PLAYBACK);
    if(!m_phandle || e) {
      DI_error(DI_ERR_SYSTEM, fname, "Could not open output");
      DI_error(DI_ERR_SYSTEM, fname, snd_strerror(e));
      close();
      return false;
    }
    
    snd_pcm_info_t pinfo;
    bzero(&pinfo, sizeof(pinfo));
    snd_pcm_info(m_phandle, &pinfo);
    
    sprintf(buf, "PINFO: type = %x, playback = %d capture = %d ",
	    pinfo.type, pinfo.playback, pinfo.capture);
    DI_trace(fname, buf);
  }    

  if((m_rmeHacks = alsaEnableRmeHacks(cardId)))
    DI_trace(fname, "RME hacks"); 

  snd_pcm_channel_info_t chaninfo;
  bzero(&chaninfo, sizeof(chaninfo));
  chaninfo.subdevice = (subDev >= 0) ? subDev : 0;

  if(m_rmeHacks)
    snd_pcm_plugin_info(m_phandle, &chaninfo);
  else
    snd_pcm_channel_info(m_phandle, &chaninfo);
  
  alsaDescribeChannelSetup(buf, & chaninfo);
  DI_trace(fname, buf);

  DI_trace(fname, 1);

  bzero(&m_pparams, sizeof(m_pparams));

  m_pparams.channel = SND_PCM_CHANNEL_PLAYBACK;
  m_pparams.format.interleave = interleave;
  
  switch(userSample) {
  case DI_ASF_INT16:
    m_pparams.format.format = SND_PCM_SFMT_S16_LE;
    break;
  case DI_ASF_INT32:
    m_pparams.format.format = SND_PCM_SFMT_S32_LE;
    break;
  case DI_ASF_FLOAT32:
    m_pparams.format.format = SND_PCM_SFMT_FLOAT_LE;
    break;
  case DI_ASF_FLOAT64:
    m_pparams.format.format = SND_PCM_SFMT_FLOAT64_LE;
    break;
  default:
    DI_error(DI_ERR_ARGUMENT, "Bad sample format");
    close();
    return false;
  }

  m_pparams.format.rate = samplerate;
  m_pparams.format.voices = channels;
  m_pparams.start_mode = m_rmeHacks ? SND_PCM_START_FULL : SND_PCM_START_GO;
  m_pparams.stop_mode = SND_PCM_STOP_ROLLOVER;
  // m_pparams.time = 1;

#ifdef USE_BLOCK_MODE
  int fraq_size = 1;

  while(fraq_size < (int) (maxBlock * 2 * channels) && fraq_size < 10000) {
    fraq_size = fraq_size << 1;
  }

  if(fraq_size > chaninfo.max_fragment_size)
    fraq_size = chaninfo.max_fragment_size;
  else if(fraq_size < chaninfo.min_fragment_size)
    fraq_size = chaninfo.min_fragment_size;


  int totalBytes = bufferSize * 2 * channels;

  m_pparams.mode = SND_PCM_MODE_BLOCK;
  m_pparams.buf.block.frag_size = fraq_size ;
  m_pparams.buf.block.frags_min = 2;
  m_pparams.buf.block.frags_max = totalBytes / fraq_size + 1;
  // m_pparams.buf.block.frags_max = 1024;

  sprintf(buf, "maxBlock = %d, fraq_size = %d, frags_max = %d", 
	  (int) maxBlock, fraq_size, m_pparams.buf.block.frags_max);
  DI_trace(fname, buf);

#else
  m_pparams.mode = SND_PCM_MODE_STREAM;
  m_pparams.buf.stream.queue_size = bufferSize;
  m_pparams.buf.stream.fill = SND_PCM_FILL_SILENCE;
  m_pparams.buf.stream.max_fill = bufferSize / 2;
#endif
  
  m_cparams = m_pparams;
  m_cparams.channel = SND_PCM_CHANNEL_CAPTURE;
  
  DI_trace(fname, 2);

  if(m_phandle) {
    
    if(!alsaSeekSampleFormat(m_phandle, &m_pparams)) {
      
      DI_error(DI_ERR_ARGUMENT, fname);
      close();
      return false;
    }
    
    if(m_pparams.format.format == SND_PCM_SFMT_S16_LE)
      DI_trace(fname, "16-bit int");
    if(m_pparams.format.format == SND_PCM_SFMT_S32_LE)
      DI_trace(fname, "32-bit int");
  }

  DI_trace(fname, 3);

  if(m_chandle) {

    m_cparams.channel = SND_PCM_CHANNEL_CAPTURE;
    
    if(!alsaSeekSampleFormat(m_chandle, &m_cparams)) {
      
      DI_error(DI_ERR_ARGUMENT, fname);
      close();
      return false;
    }
  }

  DI_trace(fname, 4);

  if(m_phandle) {

    e = snd_pcm_playback_prepare(m_phandle);
  
    if(e < 0) {
      DI_error(DI_ERR_ARGUMENT, fname, snd_strerror(e));
      close();
      return false;
    }
  }

  if(m_chandle) {
    
    e = snd_pcm_capture_prepare(m_chandle);
    
    if(e < 0) {
      DI_error(DI_ERR_ARGUMENT, fname, snd_strerror(e));
      close();
      return false;
    }
  }
  m_alsaSample = m_pparams.format.format;

  m_bytesPerSample = alsaBytesPerSample(m_alsaSample);
  m_bytesPerUserSample = DI_sampleWidth(userSample);

  m_tmpBufferFrames = fraq_size / (channels * m_bytesPerSample);

  int bufferBytes = m_bytesPerSample * channels * m_tmpBufferFrames;
  
  if(!bufferBytes) DI_fatal(DI_ERR_INTERNAL, fname);

  m_tmpReadBuffer = (char *) malloc(bufferBytes);
  m_tmpWriteBuffer = (char *) malloc(bufferBytes);
  
  m_underFlows = 0;
  m_mode = mode;

  if(mode & DI_IO_INPUT)  m_inChannels  = m_cparams.format.voices;
  if(mode & DI_IO_OUTPUT) m_outChannels = m_pparams.format.voices;
  
  m_tmpBufferRead = m_tmpBufferFrames;
  m_tmpBufferWriteCached = 0;
  m_written = 0;

  DI_trace(fname, 5);

  return true;
}

bool DI_AudioPort::start()
{
  const char * fname = "DI_AudioPort::start";

  if(!m_phandle && !m_chandle) return false;

  if(m_rmeHacks) {
    return true;
  }

  if(m_phandle) {
    int e = snd_pcm_playback_go(m_phandle);
    if(e < 0) {

      if(e == -EBADFD) {
	DI_error(DI_ERR_SYSTEM, fname, "-BADFD");
      }

      DI_error(DI_ERR_SYSTEM, fname, snd_strerror(e));
      return false;      
    }
  }
  if(m_chandle) {
    int e = snd_pcm_capture_go(m_chandle);
    if(e < 0) {
      DI_error(DI_ERR_SYSTEM, fname, snd_strerror(e));
      return false;      
    }    
  }
  return true;
}

bool DI_AudioPort::close(void)
{
  const char *fname = "DI_AudioPort::close";
  int e;

  // If outputting, wait until the output queue has drained
  
  // DI_Sleep sleeper;
  // while(outFilledFrames()) sleeper.sleepMs(10);
  
  // Close the port and free all resources
  
  if(m_phandle) {
    e = snd_pcm_close(m_phandle);
    if(e) {
      DI_error(DI_ERR_SYSTEM, fname, snd_strerror(e));
    }
    m_phandle = 0;
  }
  if(m_chandle) {
    e = snd_pcm_close(m_chandle);
    if(e) {
      DI_error(DI_ERR_SYSTEM, fname, snd_strerror(e));
    }
    m_chandle = 0;
  }

  m_mode = DI_IO_NONE;

  if(m_tmpReadBuffer) {
    free(m_tmpReadBuffer);
    m_tmpReadBuffer = 0;
  }
  if(m_tmpWriteBuffer) {
    free(m_tmpWriteBuffer);
    m_tmpWriteBuffer = 0;
  }

  m_tmpBufferFrames = 0;
  return true;
}


// --- I/O operations ---------------------------------------------------------

int DI_AudioPort::readFrames(void *buffer, ulong frames)
{
  ulong currentIndex = 0;

  char * cbuffer = (char *) buffer;

  while(frames) {    
    
    if(m_tmpBufferRead == m_tmpBufferFrames) {
      int e = 
	snd_pcm_read(m_chandle, m_tmpReadBuffer, 
		     m_tmpBufferFrames * m_inChannels * m_bytesPerSample);
      if(e < 0 ) return -1;
      m_tmpBufferRead = 0;

    }
    
    uint readable = m_tmpBufferFrames - m_tmpBufferRead;

    if(readable > frames) readable = frames;

    getReadFrames(cbuffer + 
		  currentIndex * m_bytesPerUserSample * m_inChannels, 
		  readable);
    
    currentIndex += readable;
    frames -= readable;
  }

  return (int) frames;
}

/* int DI_AudioPort::readBuffers(void **buffers, ulong frames)
{
  
  if(frames > m_tmpBufferFrames) return -1;

  uint bufbytes (frames * m_bytesPerSample);
  uint chans = m_pparams.format.voices, i = 0;

  int e;

  if(m_rmeHacks) 
    e = snd_pcm_plugin_read(m_chandle, m_tmpBuffer, bufbytes * chans);
  else
    e = snd_pcm_read(m_chandle, m_tmpBuffer, bufbytes * chans);

  if(e < 0 ) return -1;
  
  frames = e / (m_bytesPerSample * chans);
  
  if(m_pparams.format.format == SND_PCM_SFMT_S16_LE) {
    switch(m_userSample) {
    case DI_ASF_INT16:
      for(; i < chans; i++)
	memcpy((buffers[i]), (& m_tmpBuffer[i*bufbytes]), 2 * frames);
      break;
    case DI_ASF_INT32:
      for(; i < chans; i++)
	DI_short2Int((int *)(buffers[i]), 
		     (short *)(& m_tmpBuffer[i*bufbytes]), 16, 0, frames);
      break;
    case DI_ASF_FLOAT32:
      for(; i < chans; i++)
	DI_short2Float((float *)(buffers[i]), 
		       (short *)(& m_tmpBuffer[i*bufbytes]),
		       1.0 / DI_ALSA_SSCALE, 0.0, frames);
      break;
    case DI_ASF_FLOAT64:
      for(; i < chans; i++)
	DI_short2Double((double *)(buffers[i]), 
			(short *)(& m_tmpBuffer[i*bufbytes]),
			1.0 / DI_ALSA_SSCALE, 0.0, frames);
      break;
    default:
      return -1;
    }
  }
  
  else if(m_pparams.format.format == SND_PCM_SFMT_S32_LE) {
    switch(m_userSample) {
    case DI_ASF_INT16:
      for(; i < chans; i++)
	DI_int2Short((short *)(buffers[i]), 
		     (int *)(& m_tmpBuffer[i*bufbytes]), -16, 0, frames);
      break;
    case DI_ASF_INT32:
      for(; i < chans; i++)
	memcpy((buffers[i]), (& m_tmpBuffer[i*bufbytes]), 4*frames);
      break;
    case DI_ASF_FLOAT32:
      for(; i < chans; i++)
	DI_int2Float((float *)(buffers[i]), 
		     (int *)(& m_tmpBuffer[i*bufbytes]), 1.0 / DI_ALSA_ISCALE,
		     0.0, frames);
      break;
    case DI_ASF_FLOAT64:
      for(; i < chans; i++)
	DI_int2Double((double *)(buffers[i]), 
		      (int *)(& m_tmpBuffer[i*bufbytes]), 1.0 / DI_ALSA_ISCALE,
		      0.0, frames);
      break;
    default:
      return -1;
    }
  }
  else if(m_pparams.format.format == SND_PCM_SFMT_FLOAT_LE) {
    // User sample *SHOULD BE* float
    switch(m_userSample) {
    case DI_ASF_FLOAT32:
      for(; i < chans; i++)
	memcpy((buffers[i]), (& m_tmpBuffer[i*bufbytes]), 4*frames);
      break;
    default:
      return -1;
    }
  }
  else if(m_pparams.format.format == SND_PCM_SFMT_FLOAT64_LE) {
    // User sample *SHOULD BE* double
    switch(m_userSample) {
    case DI_ASF_FLOAT64:
      for(; i < chans; i++)
	memcpy((buffers[i]), (& m_tmpBuffer[i*bufbytes]), 8*frames);
      break;
    default:
      return -1;
    }
  }

  return (int) frames;

  } */


int DI_AudioPort::writeFrames(const void *buffer, ulong frames)
{
  ulong currentIndex = 0;

  const char * cbuffer = (const char *) buffer;
  
  while(frames) {
    
    uint fillable = m_tmpBufferFrames - m_tmpBufferWriteCached;

    if(fillable > frames) fillable = frames;

    storeWriteFrames(cbuffer + 
		     currentIndex * m_bytesPerUserSample * m_outChannels, 
		     fillable);
    
    currentIndex += fillable;
    
    if(m_tmpBufferWriteCached == m_tmpBufferFrames) {
      
      int e = snd_pcm_write
	(m_phandle, m_tmpWriteBuffer, 
	 m_tmpBufferFrames * m_bytesPerSample * m_outChannels);
      
      if(e < 0) {
	DI_error(DI_ERR_UNKNOWN, "DI_AudioPort::writeFrames", 
		 snd_strerror(e));
	return -1;
      }
      m_written++;
      
      m_tmpBufferWriteCached = 0;
    }
    frames -= fillable;
  }
  return (int) currentIndex;
}

/* int DI_AudioPort::writeBuffers(const void **buffers, ulong frames)
{
  // DI_trace("DI_AudioPort::writeBuffers");

  if(frames > m_tmpBufferFrames) return -1;

  uint chans = m_pparams.format.voices;

  // uint samples (frames * chans);
  uint bufbytes (frames * m_bytesPerSample);
  uint i = 0;

  if(m_pparams.format.format == SND_PCM_SFMT_S16_LE) {
    switch(m_userSample) {
    case DI_ASF_INT16:
      for(; i < chans; i++)
	memcpy(& m_tmpBuffer[i*bufbytes],  (buffers[i]), frames * 2);
      break;
      
    case DI_ASF_INT32:
      for(; i < chans; i++)
	DI_int2Short((short *) (& m_tmpBuffer[i*bufbytes]), 
		     (int *) (buffers[i]), -16, 0, frames);
      break;

    case DI_ASF_FLOAT32:
      // DI_trace(",");
      for(; i < chans; i++)
	DI_float2Short((short *) ( & m_tmpBuffer[i*bufbytes]), 
		       (float *) (buffers[i]),
		       DI_ALSA_SSCALE, 0, frames);
      // DI_trace(";");
      break;
     case DI_ASF_FLOAT64:
       DI_double2Short((short *) (& m_tmpBuffer[i*bufbytes]), 
		       (double *) (buffers[i]),
		       DI_ALSA_SSCALE, 0, frames);
       break;
     default:
       return -1;
    }
  }
  
  else if(m_pparams.format.format == SND_PCM_SFMT_S32_LE) {
    switch(m_userSample) {
    case DI_ASF_INT16:
      for(; i < chans; i++)
	DI_short2Int((int *)(& m_tmpBuffer[i*bufbytes]), 
		     (short *)(buffers[i]), 16, 0, frames);
      break;
    case DI_ASF_INT32:
      for(; i < chans; i++)
	memcpy((&m_tmpBuffer[i*bufbytes]), (buffers[i]), 4*frames);
      break;
    case DI_ASF_FLOAT32:
      for(; i < chans; i++)
	DI_float2Int((int *)(&m_tmpBuffer[i*bufbytes]), 
		     (float *)(buffers[i]), DI_ALSA_ISCALE,
		     0, frames);
      break;
    case DI_ASF_FLOAT64:
      for(; i < chans; i++)
	DI_double2Int((int *)(&m_tmpBuffer[i*bufbytes]), 
		      (double *)(buffers[i]), DI_ALSA_ISCALE,
		      0, frames);
      break;
    default:
      return -1;
    }
  }
  
  else if(m_pparams.format.format == SND_PCM_SFMT_FLOAT_LE) {
    // User sample *SHOULD BE* float
    switch(m_userSample) {
    case DI_ASF_FLOAT32:
      memcpy((& m_tmpBuffer[i*bufbytes]), (buffers[i]), 4*frames);
      break;
    default:
      return -1;
    }
  }
  else if(m_pparams.format.format == SND_PCM_SFMT_FLOAT64_LE) {
    // User sample *SHOULD BE* double
    switch(m_userSample) {
    case DI_ASF_FLOAT64:
      memcpy((& m_tmpBuffer[i*bufbytes]), (buffers[i]), 8*frames);
      break;
    default:
      return -1;
    }
  }

  return snd_pcm_write(m_phandle, m_tmpBuffer, bufbytes * chans) / 
    (m_bytesPerSample * chans);
    }*/

int DI_AudioPort::inFilledFrames(void) const 
{ 
  if(m_phandle) {
    if(m_written < 3) {
      DI_trace("DI_AudioPort::inFilledFrames", m_written);
      return 0;

    }
    // DI_trace("DI_AudioPort::inFilledFrames # written", (int) m_written);
    // Very stupid thing to return:
    return m_cparams.buf.block.frag_size;
  }

  snd_pcm_channel_status_t *cs = 
    (snd_pcm_channel_status_t *) & m_channelStatus;

  cs->channel = SND_PCM_CHANNEL_CAPTURE;
  
  int e;
  
  if(m_rmeHacks) e = snd_pcm_plugin_status(m_chandle, cs);
  else e = snd_pcm_channel_status(m_chandle, cs);
  
  if(e) {
    DI_error(DI_ERR_SYSTEM, "DI_AudioPort::inFilledFrames",
	     snd_strerror(e));
    return -1; 
  }

  // return (cs->count - cs->free) / 
  return (cs->count - cs->free) / 
    (m_bytesPerSample * m_pparams.format.voices);
}

int DI_AudioPort::outFilledFrames(void) const 
{ 
  const char * fname = "DI_AudioPort::outFilledFrames";

  snd_pcm_channel_status_t *cs = 
    (snd_pcm_channel_status_t *) & m_channelStatus;

  bzero(cs, sizeof(*cs));
  cs->channel = SND_PCM_CHANNEL_PLAYBACK;
  
  int e;
  
  if(m_rmeHacks) e = snd_pcm_plugin_status(m_phandle, cs);
  else e = snd_pcm_channel_status(m_phandle, cs);

  if(e) {
    DI_error(DI_ERR_SYSTEM, fname, snd_strerror(e));
    return -1; 
  }

  if((cs->status != SND_PCM_STATUS_RUNNING) && !m_rmeHacks) {

    char buf[128];
    sprintf(buf, "%s # status = %d", fname, cs->status);
    DI_fatal(DI_ERR_SYSTEM, buf);
  }

  // DI_trace(fname, cs->status);

  * ((uint *) &m_underFlows) += cs->underrun;

  // return (cs->count - cs->free) / 
  return cs->count / 
    (m_bytesPerSample * m_pparams.format.voices);
}

int DI_AudioPort::waitInputFilled(ulong frames, int timeOut)
{
  struct timeval tv;
  int r;
  while(true) {
    // DI_trace("DI_AudioPort::waitOutputFilled", outFilledFrames());
    if((r = inFilledFrames()) < (int) frames) {
      // gettimeofday(&tv, 0);
      tv.tv_sec = 0;      // Zero full seconds.
      tv.tv_usec = 2000;  // two millisecs.
      select(0, 0, 0, 0, & tv); // Sleep
    }
    else
      break;
  }
  return r;
}

int DI_AudioPort::waitOutputFilled(ulong frames, int timeOut)
{
  struct timeval tv;
  int r;

  while(true) {
    // DI_trace("DI_AudioPort::waitOutputFilled", outFilledFrames());
    if((r = outFilledFrames()) > (int) frames) {
      // gettimeofday(&tv, 0);
      tv.tv_sec = 0;      // Zero full seconds.
      tv.tv_usec = 2000;  // two millisecs.
      select(0, 0, 0, 0, & tv);
    }
    else
      break;
  }
  return r;
}


void DI_AudioPort::alsaDescribePlaybackSetup(char *b)
{
  snd_pcm_channel_setup_t setup;

  bzero(&setup, sizeof(setup));

  snd_pcm_channel_setup(m_phandle, &setup);

  sprintf(b, "Voices = %d, rate = %d, fraqs = %d, fraq_size = %d",
	  setup.format.voices, setup.format.rate, 
	  setup.buf.block.frags, setup.buf.block.frag_size);
}

void DI_AudioPort::alsaDescribeChannelSetup(char *buf, 
					    snd_pcm_channel_info_t *info)
{
  sprintf(buf, "subdevice = %d subname = \"%s\"\n\
min_rate = %d, max_rate = %d, \n\
min_voices = %d max_voices = %d\n \
min_fragment_size = %d max_fragment_size = %d\n \
flags = 0x%x formats = 0x%x \n\
formats = \"%s\"", 
	  info-> subdevice, info->subname, 
	  info->min_rate, info->max_rate, 
	  info->min_voices, info->max_voices,
	  info->min_fragment_size, info->max_fragment_size,
	  info->flags, info->formats, 
	  snd_pcm_get_format_name(1 << info->formats));
}

bool DI_AudioPort::alsaSeekSampleFormat(snd_pcm_t *pcmh, 
					snd_pcm_channel_params_t *cph)
{
  // int e = snd_pcm_plugin_params(pcmh, cph);
  int e;
  
  if(m_rmeHacks) e = snd_pcm_plugin_params(pcmh, cph);
  else e = snd_pcm_channel_params(pcmh, cph);

  if(e >= 0) return true;

  while(e < 0) {
    cph->format.format = alsaDownGradeFormat(cph->format.format);
    
    DI_trace("DI_AudioPort::alsaSeekSampleFormat", 
	     snd_pcm_get_format_name(cph->format.format));

    if(m_rmeHacks) e = snd_pcm_plugin_params(pcmh, cph);
    else e = snd_pcm_channel_params(pcmh, cph);

    if(e >= 0) return true;
    // Failed to get a decent sample format
    if(cph->format.format == SND_PCM_SFMT_S16_LE) {
      break;
    }
  }

  DI_error(DI_ERR_ARGUMENT, "DI_AudioPort::alsaSeekSampleFormat", 
	   snd_strerror(e));

  return false;
}

int DI_AudioPort::alsaDownGradeFormat(int alsaSampleFormat)
{
  // return alsaSampleFormat - 1;
  int f;

  switch(alsaSampleFormat) {
  case SND_PCM_SFMT_FLOAT64_LE:
    f = SND_PCM_SFMT_S32_LE;
    // f =  SND_PCM_SFMT_FLOAT_LE;
    break;
  case SND_PCM_SFMT_FLOAT_LE:
    f = SND_PCM_SFMT_S32_LE;
    break;

  case SND_PCM_SFMT_S32_LE:
    f = SND_PCM_SFMT_S16_LE;
    break;
    /* case SND_PCM_SFMT_S24_LE: 
    f = SND_PCM_SFMT_S16_LE;
    break;*/
    
  default:
    f = SND_PCM_SFMT_S16_LE;
  }
  return f;
}

int DI_AudioPort::alsaBytesPerSample(int alsaSampleFormat)
{
  int f;

  switch(alsaSampleFormat) {
  case SND_PCM_SFMT_FLOAT64_LE:
    f = 8;
    break;
  case SND_PCM_SFMT_FLOAT_LE:
  case SND_PCM_SFMT_S32_LE:
    // case SND_PCM_SFMT_S24_LE:
    f = 4;
    break;
  case SND_PCM_SFMT_S16_LE:
    f = 2;
    break;
  default:
    f = 0;
    break;
  }
  return f;  
}

bool DI_AudioPort::alsaEnableRmeHacks(int card)
{
  char cardName[DI_AUDIO_MAX_NAME_LENGTH];
  const char rmeStr[] = "RME";
  int  rmeLen = 3;

  DI_AudioDevice::alsaDupCardName(card, cardName);
  int nameLen = strlen(cardName);
  for(int i=0; i < nameLen; i++) {
    if(!strncmp(rmeStr, &cardName[i], rmeLen))
      return true;
  }
  return false;
}

// --- Predicates -------------------------------------------------------------

bool DI_AudioPort::isOK(void) const
{
  return true;
}

void DI_AudioPort::getReadFrames(void *buffer, ulong frames)
{
  uint samples(frames * m_pparams.format.voices);

  char * tmpBuffer = m_tmpReadBuffer + 
    m_tmpBufferRead * m_bytesPerSample * m_inChannels;
  
  if(m_pparams.format.format == SND_PCM_SFMT_S16_LE) {
    switch(m_userSample) {
    case DI_ASF_INT16:
      memcpy(buffer, tmpBuffer, 2 * samples);
      break;
    case DI_ASF_INT32:
      DI_short2Int((int *)buffer, (short *)tmpBuffer, 16, 0, samples);
      break;
    case DI_ASF_FLOAT32:
      DI_short2Float((float *)buffer, (short *)tmpBuffer,
		     1.0 / DI_ALSA_SSCALE, 0.0, samples);
      break;
    case DI_ASF_FLOAT64:
      DI_short2Double((double *)buffer, (short *)tmpBuffer,
		      1.0 / DI_ALSA_SSCALE, 0.0, samples);
      break;
    default:
      return;
    }
  }
  else if(m_pparams.format.format == SND_PCM_SFMT_S32_LE) {
    switch(m_userSample) {
    case DI_ASF_INT16:
      DI_int2Short((short *)buffer, (int *)tmpBuffer, -16, 0, samples);
      break;
    case DI_ASF_INT32:
      memcpy(buffer, tmpBuffer, 4*samples);
      break;
    case DI_ASF_FLOAT32:
      DI_int2Float((float *)buffer, (int *)tmpBuffer, 1.0 / DI_ALSA_ISCALE,
		   0.0, samples);
      break;
    case DI_ASF_FLOAT64:
      DI_int2Double((double *)buffer, (int *)tmpBuffer, 1.0 / DI_ALSA_ISCALE,
		    0.0, samples);
      break;
     default:
       return;
    }
  }
  else if(m_pparams.format.format == SND_PCM_SFMT_FLOAT_LE) {
    // User sample *SHOULD BE* float
    switch(m_userSample) {
    case DI_ASF_FLOAT32:
      memcpy(buffer, tmpBuffer, 4*samples);
      break;
    default:
      return;
    }
  }
  else if(m_pparams.format.format == SND_PCM_SFMT_FLOAT64_LE) {
    // User sample *SHOULD BE* double
    switch(m_userSample) {
    case DI_ASF_FLOAT64:
      memcpy(buffer, tmpBuffer, 8*samples);
      break;
    default:
      return;
    }
  }
  m_tmpBufferRead += (uint) frames;
}

void DI_AudioPort::storeWriteFrames(const void *buffer, ulong frames)
{
  uint samples(frames * m_pparams.format.voices);

  char * tmpBuffer = m_tmpWriteBuffer + 
    m_tmpBufferWriteCached * m_bytesPerSample * m_outChannels;

  if(m_pparams.format.format == SND_PCM_SFMT_S16_LE) {
    switch(m_userSample) {
    case DI_ASF_INT16:
      memcpy(tmpBuffer, buffer, samples * 2);
      break;
    case DI_ASF_INT32:
      DI_int2Short((short *)tmpBuffer, (int *) buffer, -16, 0, samples);
      break;
     case DI_ASF_FLOAT32:
       // DI_trace("Good");
       DI_float2Short((short *)tmpBuffer, (float *) buffer,
		      DI_ALSA_SSCALE, 0, samples);
       break;
     case DI_ASF_FLOAT64:
       DI_double2Short((short *)tmpBuffer, (double *) buffer,
		       DI_ALSA_SSCALE, 0, samples);
       break;
     default:
      return;
    }
  }
  else if(m_pparams.format.format == SND_PCM_SFMT_S32_LE) {
    switch(m_userSample) {
    case DI_ASF_INT16:
      DI_short2Int((int *)tmpBuffer, (short *)buffer, 16, 0, samples);
      break;
    case DI_ASF_INT32:
      memcpy(tmpBuffer, buffer, 4*samples);
      break;
    case DI_ASF_FLOAT32:
      DI_float2Int((int *)tmpBuffer, (float *)buffer, DI_ALSA_ISCALE,
		   0, samples);
      break;
    case DI_ASF_FLOAT64:
      DI_double2Int((int *)tmpBuffer, (double *)buffer, DI_ALSA_ISCALE,
		    0, samples);
      break;
    default:
      return;
    }
  }
  else if(m_pparams.format.format == SND_PCM_SFMT_FLOAT_LE) {
    // User sample *SHOULD BE* float
    switch(m_userSample) {
    case DI_ASF_FLOAT32:
      memcpy(tmpBuffer, buffer, 4*samples);
      break;
    default:
      return;
    }
  }
  else if(m_pparams.format.format == SND_PCM_SFMT_FLOAT64_LE) {
    // User sample *SHOULD BE* double
    switch(m_userSample) {
    case DI_ASF_FLOAT64:
      memcpy(tmpBuffer, buffer, 8*samples);
      break;
    default:
      return;
    }
  }

  m_tmpBufferWriteCached += (uint) frames;
}


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

DI_AudioDevice::DI_AudioDevice(ulong id,
			       bool input,
			       ulong channels,
			       ulong rate)
{
  DI_IoMode mode;

  if(input) mode = DI_IO_INPUT;
  else mode = DI_IO_OUTPUT;

  char buf[DI_AUDIO_MAX_NAME_LENGTH];

  alsaDupCardName(id, buf);
  
  if(!rate) rate = 44100;

  m_ok = m_port.open(buf, mode, DI_ASF_INT16, 1024 * 4, rate, channels, 1024);
}
  
DI_AudioDevice::DI_AudioDevice(const char *name,
			       bool input,
			       ulong channels,
			       ulong rate)
{
  DI_IoMode mode;

  if(input) mode = DI_IO_INPUT;
  else mode = DI_IO_OUTPUT;

  if(!rate) rate = 44100;

  m_ok = m_port.open(name, mode, DI_ASF_INT16, 1024 * 4, rate, channels, 1024);
  if(m_ok) m_port.start();
}
  
DI_AudioDevice::~DI_AudioDevice()
{
  m_port.close();
}

ulong DI_AudioDevice::getChannels(void)
{
  if(!m_ok) return 0;

  switch(m_port.mode()) {
  case DI_IO_INPUT:
  case DI_IO_INPUT_OUTPUT:
    return m_port.inChannels();
  case DI_IO_OUTPUT:
    return m_port.outChannels();
  case DI_IO_NONE:
    break;
  }
  return 0;
}



void DI_AudioDevice::getDeviceLabel(ulong devID, char *label)
{
  char *tmp(0);

  int e = snd_card_get_longname(devID, &tmp);

  if(e) return;

  if(!tmp || strlen(tmp) >= DI_AUDIO_MAX_NAME_LENGTH)
    return;
  
  strcpy(label, tmp);

  // free(tmp);
}


int DI_AudioDevice::alsaNumberOfCards()
{
  return snd_cards();
}

/** Copies name of the sound card to argument name (allocates new
    memory). On success, returns zero, otherwise an error. */

int DI_AudioDevice::alsaGetCardName(int card, char **name)
{
  return snd_card_get_name(card, name);
}

bool DI_AudioDevice::alsaDupCardName(int card, char *name)
{
  char * tmp(0);

  int e = alsaGetCardName(card, & tmp);
  if(e) return false;

  if(!tmp || strlen(tmp) >= DI_AUDIO_MAX_NAME_LENGTH)
    return false;

  strcpy(name, tmp);
  
  // free(tmp);

  return true;
}
