/* COPYRIGHT
 *
 * This file is part of the DIVA IO library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA IO library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */

#include <di_audio_file_handler.h>

#include <di_error.h>
#include <di_sleep.h>

#include <strings.h>

DI_AudioFileHandler::Handle::Handle
(DI_AudioFileHandler * host, 
 const char * filename, 
 DI_IoMode mode, 
 long startFrame)
  : m_host(host),
    m_fileName(filename),
    m_ioMode(mode),
    m_startFrame(startFrame),
    m_status(OPEN_NOT),
    m_file(0),
    m_userDone(false)
{
  bzero(&m_info, sizeof(m_info));
}

DI_AudioFileHandler::Handle::~Handle()
{
  close();
}

bool DI_AudioFileHandler::Handle::waitOpen()
{
  if(m_status == OPEN_DONE)
    return true;
  
  lock();
  while(m_status == OPEN_NOT)
    waitCond();
  unlock();

  return m_status == OPEN_DONE;
}

int DI_AudioFileHandler::Handle::writeFrames(DI_Sample * data, int frames)
{
  if(m_status != OPEN_DONE)
    return -1;

  int r = frames;

  while(frames) {

    /* DI_trace2("DI_AudioFileHandler::Handle::writeFrames # %ld %ld", 
       m_userFrames, m_fileFrames); */

    long blockleft = m_blockSize - (m_userFrames % m_blockSize);

    int avail = frames > blockleft ? blockleft : frames;

    lock();
    while(m_userFrames + avail > m_fileFrames + m_blockSize * (m_blocks - 1))
      waitCond();
    unlock();

    // DI_trace("DI_AudioFileHandler::Handle::writeFrames >", frames);
    
    int samples = avail * channels();

    memcpy(ptr(m_userFrames), data, sizeof(DI_Sample) * samples);
    
    data += samples;
    m_userFrames += avail;
    frames -= avail;
  }

  return r;
}

int DI_AudioFileHandler::Handle::readFrames(DI_Sample * data, int frames)
{
  if(m_status != OPEN_DONE)
    return -1;

  int r = frames;

  while(frames) {

    long blockleft = m_blockSize - (m_userFrames % m_blockSize);

    long avail = frames > blockleft ? blockleft : frames;

    /* DI_trace2("DI_AudioFileHandler::Handle::readFrames %ld %ld %ld %ld", 
       frames, avail, m_userFrames, m_fileFrames); */

    lock();
    while(m_userFrames + avail > m_fileFrames)
      waitCond();
    unlock();

    int samples = avail * channels();

    memcpy(data, ptr(m_userFrames), sizeof(DI_Sample) * samples);
    
    data += samples;
    m_userFrames += avail;
    frames -= avail;
  }
  
  return r;
}

/** Update the file structure. */

bool DI_AudioFileHandler::Handle::update()
{
  bool something = false;

  if(m_userDone) {

    // puts("Closing file");

    if(m_status == OPEN_DONE && m_ioMode == DI_IO_OUTPUT)
      flushWrite();
      
    close();
    
    m_status = OPEN_CLOSED;

    something = true;
  }
  else if(m_status == OPEN_NOT) {

    if(open())
      m_status = OPEN_DONE;
    else
      m_status = OPEN_FAILED;

    something =  true;
  }
  else if(m_status == OPEN_DONE) {
    if(m_ioMode == DI_IO_INPUT)
      something = fileRead();
    else
      something = fileWrite();
  }
  else if(m_status == OPEN_EOF) {
    close();
    something =  true;
    m_status = OPEN_CLOSED;
  }

  return something;
}

bool DI_AudioFileHandler::Handle::open()
{
  close();

  int mode = (m_ioMode == DI_IO_INPUT) ? SFM_READ : SFM_WRITE;

  m_file = sf_open(m_fileName.c_str(), mode, & m_info);

  m_blocks = 6;
  m_blockSize = m_info.samplerate / m_blocks;

  int buffersamples = m_info.channels * m_blockSize * m_blocks;

  m_data.resize(buffersamples);

  bzero(& m_data[0], buffersamples * sizeof(DI_Sample));

  m_fileFrames = 0;
  m_userFrames = 0;
  
  return m_file != 0;
}

bool DI_AudioFileHandler::Handle::close()
{
  if(!m_file)
    return false;

  sf_close(m_file);

  return true;
}

bool DI_AudioFileHandler::Handle::fileRead()
{
  // DI_trace("DI_AudioFileHandler::Handle::fileRead", m_fileFrames);

  if(m_fileFrames >= m_userFrames + m_blocks * m_blockSize)
    return false;

  DI_Sample * data = ptr(m_fileFrames);
  
  long avail = m_info.frames - m_fileFrames;

  if(avail > m_blockSize)
    avail = m_blockSize;
  
  sf_read_float(m_file, data, avail);

  m_fileFrames += avail;

  if(m_fileFrames == m_info.frames)
    m_status = OPEN_EOF;

  return true;
}

bool DI_AudioFileHandler::Handle::fileWrite()
{
  /* DI_trace2("DI_AudioFileHandler::Handle::fileWrite # HELLO %ld %ld", 
     m_fileFrames, m_userFrames); */

  if(m_fileFrames + m_blockSize > m_userFrames)
    return false;

  /* DI_trace2("DI_AudioFileHandler::Handle::fileWrite # %ld %ld", 
     m_fileFrames, m_userFrames); */

  while(m_fileFrames + m_blockSize <= m_userFrames) {
    DI_Sample * data = ptr(m_fileFrames);
    sf_write_float(m_file, data, m_blockSize);
    m_fileFrames += m_blockSize;
  }

  return true;
}

bool DI_AudioFileHandler::Handle::flushWrite()
{
  /* DI_trace2("DI_AudioFileHandler::Handle::flushWrite # %ld %ld", 
     m_fileFrames, m_userFrames); */

  while(m_fileFrames < m_userFrames) {

    long avail = m_userFrames - m_fileFrames;
    long n = avail > m_blockSize ? m_blockSize : avail;

    DI_Sample * data = ptr(m_fileFrames);
    sf_write_float(m_file, data, n);
    m_fileFrames += n;
  }

  return true;
}


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

DI_AudioFileHandler::DI_AudioFileHandler()
  : m_done(true)
{

}

DI_AudioFileHandler::~DI_AudioFileHandler()
{
  if(!m_done)
    stop();
}

DI_AudioFileHandler::Handle * DI_AudioFileHandler::readFile
(const char * filename, long startFrame)
{
  Handle * h = new Handle(this, filename, DI_IO_INPUT, startFrame);

  m_mutex2.lock();
  m_files.push_back(h);
  m_mutex2.unlock();

  return h;
}

/** Prepare to write to a file.  */

DI_AudioFileHandler::Handle * DI_AudioFileHandler::writeFile
(const char * filename, 
 /// Number of channels
 int channels, 
 /// Sample rate (Hz)
 int samplerate, 
 /// file format in sndfile format
 int sfFormat)
{
  Handle * h = new Handle(this, filename, DI_IO_OUTPUT, 0);

  h->m_info.channels   = channels;
  h->m_info.samplerate = samplerate;
  h->m_info.format     = sfFormat;

  m_mutex2.lock();
  m_files.push_back(h);
  m_mutex2.unlock();

  return h;
}

/** Return a file handle to the system, implying that its work is
    done. */

void DI_AudioFileHandler::done(Handle * h)
{
  /* iterator it = std::find(m_files.begin(), m_files.end(), h);

  if(it == m_files.end())
  return; */

  h->m_userDone = true;
}

void DI_AudioFileHandler::start()
{
  if(!m_done)
    return;

  m_done = false;

  run();
}

void DI_AudioFileHandler::stop()
{
  if(m_done)
    return;

  m_done = true;
  waitEnd();
}

void DI_AudioFileHandler::childLoop()
{  
  while(!m_done) {
    if(!update()) {
      DI_Sleep::sleepMs(20);
    }
  }

  for(iterator it = m_files.begin(); it != m_files.end(); it++) {
    Handle * h = (*it);
    h->m_userDone = true;
    h->update();
    delete h;
  }

  m_files.clear();
}

bool DI_AudioFileHandler::update()
{
  bool something = false;

  m_mutex2.lock();

  for(iterator it = m_files.begin(); it != m_files.end(); ) {

    Handle * h = (*it);

    something = something || h->update();

    if(h->m_status == OPEN_CLOSED) {
      /* DI_trace2("DI_AudioFileHandler::update # Deleting file handle %s",
	 h->m_fileName.c_str()); */
      delete h;
      iterator tmp = it;
      tmp++;
      m_files.erase(it);
      it = tmp;
    }
    else
      it++;
  }

  m_mutex2.unlock();

  if(something)
    signalCond();

  // DI_trace2("DI_AudioFileHandler::update # %d", (int) something);

  return something;
}
