/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the DIVA IO library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA IO library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */

#ifndef DI_AUDIO_FILE_HANDLER_H
#define DI_AUDIO_FILE_HANDLER_H

#include <di_condition.h>
#include <di_config.h>
#include <di_io_defs.h>
#include <di_thread.h>

#include <sndfile.h>

#include <list>
#include <string>
#include <vector>

/** Read/write multiple audio files at the same time in one
    background-thread. */

class DI_AudioFileHandler : public DI_Thread
{
  friend class Handle;

public:

  enum OpenStatus {
    OPEN_NOT,
    OPEN_DONE,
    OPEN_FAILED,
    OPEN_EOF,
    OPEN_CLOSED
  };

  /** A handle that offers access to the audio files. */
  class Handle
  {
  public:
    friend class DI_AudioFileHandler;

    bool waitOpen();
    bool isOpen() const { return m_file != 0; }

    int writeFrames(DI_Sample * data, int frames);
    int readFrames(DI_Sample * data, int frames);

    int channels()   const { return m_info.channels; }
    int sampleRate() const { return m_info.samplerate; }
    long frames()    const { return (long) m_info.frames; }

  protected:

    void lock()   { m_host->lock(); }
    void unlock() { m_host->unlock(); }
    void waitCond()   { m_host->waitCond(); }
    void signalCond() { m_host->signalCond(); }

    Handle(DI_AudioFileHandler * host,
	   const char * filename, DI_IoMode mode, long startFrame);
    ~Handle();

    bool update();

    bool open();
    bool close();

    bool fileRead();
    bool fileWrite();
    bool flushWrite();

    DI_Sample * ptr(long frame)
    { return & m_data[(frame * channels()) % m_data.size()]; }

    DI_AudioFileHandler * m_host;

    std::string m_fileName;
    DI_IoMode   m_ioMode;

    long        m_blocks;
    long        m_blockSize;
    long        m_startFrame;

    volatile OpenStatus m_status;
    volatile long       m_fileFrames;
    volatile long       m_userFrames;

    SNDFILE * m_file;
    SF_INFO   m_info;

    std::vector<DI_Sample> m_data;
    bool m_userDone;
  };


  DI_AudioFileHandler();
  ~DI_AudioFileHandler();

  Handle * readFile(const char * filename, long startFrame);
  Handle * writeFile(const char * filename, 
		     int channels, 
		     int samplerate, 
		     int sfFormat);

  void done(Handle *);

  void start();
  void stop();

protected:

  virtual void childLoop();
  
  bool update();

  void lock()   { m_mutex.lock(); }
  void unlock() { m_mutex.unlock(); }
  void waitCond()   { m_cond.wait(m_mutex); }
  void signalCond() { m_cond.signal2(m_mutex); }

  typedef std::list<Handle *> container;
  typedef container::iterator iterator;

  container m_files;
  
  DI_MutexAuto m_mutex, m_mutex2;
  DI_Condition m_cond;

  volatile bool m_done;
};


#endif
