/* COPYRIGHT
 *
 * This file is part of the DIVA IO library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA IO library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */
#include "di_audio_file_threaded.h"

#include "di_error.h"

#include <errno.h>
#include <sched.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

DI_AudioFileThreaded::DI_AudioFileThreaded()
  : m_tmpBuffer(0),
    m_fileName(0),
    m_blockSize(0),
    m_userFormat(DI_ASF_FLOAT32),
    m_end(false),
    m_ok(true),
    m_mode(DI_IO_NONE),
    m_status(DI_AFT_EMPTY)
{
  m_mutex.init(false, false, false);
}

DI_AudioFileThreaded::~DI_AudioFileThreaded()
{
  free(m_fileName);
}

/** Span a thread for reading from the file. */
bool DI_AudioFileThreaded::startRead(const char *file, 
				     DI_AudioSampleFormat userFormat,
				     ulong maxBlock,
				     long startFrame,
				     long endFrame,
				     bool loop,
				     int priority)
{
  free(m_fileName);

  m_fileName = strdup(file);

  m_userFormat = userFormat;
  m_blockSize = maxBlock;
  m_startFrame = startFrame;
  m_endFrame = endFrame;
  m_threadFrames = 0;

  m_loop = loop;
  m_threadBuffer = 0;
  m_buffers = 0;
  m_end = false;
  m_mode = DI_IO_INPUT;

  m_status = DI_AFT_OPENING;

  m_ok = true;
  m_fills = m_drains = 0;

  m_userFrames = 0;
  m_threadFrames = 0;
  m_priority = priority;

  return run(true);
}

/** Span a thread for writing to the file. */
bool DI_AudioFileThreaded::startWrite(const char *file, 
				      DI_AudioSampleFormat userFormat,
				      uint sampleRate,
				      uint channels,
				      ulong maxBlock,
				      int priority,
				      DI_AudioFileFormat fileFormat,
				      DI_AudioSampleFormat sampleFormat,
				      DI_AudioFileByteOrder byteOrder)
{  
  const char * fname = "DI_AudioFileThreaded::startWrite";

  free(m_fileName);
  m_fileName = strdup(file);

  m_userFormat = userFormat;
  m_blockSize = maxBlock;
  
  m_startFrame = 0;
  m_endFrame = 0;
  m_threadFrames = 0;
  m_chans = channels;
  m_loop = false;
  m_threadBuffer = 0;
  m_buffers = 0;
  m_end = false;
  m_mode = DI_IO_OUTPUT;

  m_status = DI_AFT_OPENING;

  m_userFrames = 0;
  m_threadFrames = 0;
  m_priority = priority;

  if(!m_file.setFileSampleFormat(sampleFormat))
    DI_error(DI_ERR_ARGUMENT, fname, "Bad file sample format");

  m_file.setFileFormat(fileFormat);
  m_file.setSampleRate(sampleRate);
  // m_file.setSampleWidth(DI_sampleWidth(sampleFormat));
  // m_file.setSampleFormat(afformat);
  m_file.setByteOrder(byteOrder);
  m_file.setChannels(m_chans);
  
  m_ok = true;

  m_fills = m_drains = 0;

  return run(true);
}

/** Prepare the file for operation. 

    This method may block if called too early. */
bool DI_AudioFileThreaded::userPrepare()
{
  if(m_mode == DI_IO_OUTPUT) {
    
    m_mutex.lock();
    while(m_status != DI_AFT_READY && m_ok)
      m_cond.wait(m_mutex);
    m_mutex.unlock();

    return m_ok;
  }
  else {
    
    m_mutex.lock();
    while(m_fills < (m_buffers - 2) && m_ok)
      m_cond.wait(m_mutex);
    m_mutex.unlock();

    return m_ok;    
  }
}

int DI_AudioFileThreaded::readFrames(void *buffer, ulong frames)
{
  if(!m_ok) return -1;

  char * cbuffer = (char *) buffer;

  ulong frameBytes = m_userSWidth * m_chans;

  while(frames) {
    
    /* Current location of the client process in frames. Normalized to
       range 0-bufferFrames. */

    ulong currPlace = m_userFrames % (m_blockSize * m_buffers);

    /* Get pointer to the current read location. */

    char * currPtr = & m_tmpBuffer[currPlace * frameBytes];

    /* Frames available before the end of segment. */

    ulong framesAvailable = (m_blockSize - (currPlace % m_blockSize));
    if(!framesAvailable) framesAvailable = m_blockSize;

    /* Number of frames to read in this pass. */

    ulong readFrames = (frames > framesAvailable) ? framesAvailable : frames;
    
    memcpy(cbuffer, currPtr, readFrames * frameBytes);

    /* DI_trace2("DI_AudioFileThreaded::readFrames %d %d %d # %d %d %d # %d %d",
       m_userFrames % m_blockSize, readFrames, m_fills, 
       m_drains, framesAvailable, m_blockSize, 
       frames, currPlace); */
    

    cbuffer += readFrames * frameBytes;
    m_userFrames += readFrames;
    frames -= readFrames;

    if(!(m_userFrames % m_blockSize)) {
    
      /* If the client hits the end of this segment it tells the
	 read-thread its new segment. */
  
      /* DI_trace2("DI_AudioFileThreaded::readFrames Block switch %d %d %d",
	 m_fills, m_drains, m_buffers); */

      /* Increment the number of segments that have been read by the
	 client. */      
      m_drains++;
      
      /* Wait until the read-thread has got enough stuff (avoid
	 overrunning it). */
      m_mutex.lock();
      while(m_drains >= (m_fills + 1))
	m_cond.wait(m_mutex);
      m_mutex.unlock();

      /* Signal the mutex to tell the read-thread that it can fill the
	 previously locked segment. */
      m_cond.signal2(m_mutex);

    }
  }

  return m_ok ? (int) frames : -1;
}

int DI_AudioFileThreaded::writeFrames(void *buffer, ulong frames)
{
  // DI_trace("DI_AudioFileThreaded::writeFrames");

  if(!m_ok) return -1;

  char *cbuffer = (char *) buffer;

  ulong r = frames;

  ulong frameBytes = m_userSWidth * m_chans;

  while(frames) {

    ulong currPlace = m_userFrames % (m_blockSize * m_buffers);
    char * currPtr = & m_tmpBuffer[currPlace * frameBytes];

    ulong framesAvailable = (m_blockSize - (currPlace % m_blockSize));
    // ulong framesAvailable = (m_blockSize - currPlace) % m_blockSize;
    if(!framesAvailable) framesAvailable = m_blockSize;
    ulong writeFrames = (frames > framesAvailable) ? framesAvailable : frames;

    memcpy(currPtr, cbuffer, writeFrames * frameBytes);
    
    cbuffer += writeFrames * frameBytes;
    m_userFrames += writeFrames;
    frames -= writeFrames;

    /* DI_trace2("DI_AudioFileThreaded::writeFrames %d %d %d %d %d", 
       frames, writeFrames, framesAvailable, currPlace, m_buffers); */
  
    if(!(m_userFrames % m_blockSize)) {

      /* DI_trace2("DI_AudioFileThreaded::writeFrames %d %d %d", 
	 m_fills, m_drains, m_buffers); */

      m_fills++; 
      
      m_mutex.lock();
      while(m_fills >= (m_drains + m_buffers - 1))
	m_cond.wait(m_mutex);
      m_mutex.unlock();
      
      m_cond.signal2(m_mutex);
      
    }

  }
  // DI_trace("DI_AudioFileThreaded::writeFrames EXIT");
  return m_ok ? (int) r : -1;
}

void DI_AudioFileThreaded::requestEnd()
{
  m_end = true;

  m_cond.signal2(m_mutex);
}

bool DI_AudioFileThreaded::allocateBuffer
(DI_AudioSampleFormat userFormat, ulong maxBlock, ulong ch)
{
  if(!(maxBlock / 8)) return false;

  switch(userFormat) {
  case DI_ASF_INT16:
    m_userSWidth = 2;
    break;

  case DI_ASF_INT32:
  case DI_ASF_FLOAT32:
    m_userSWidth = 4;
    break;

  case DI_ASF_FLOAT64:
    m_userSWidth = 8;
    break;
  }

  m_blockSize = maxBlock / 8;

  m_buffers = 4;

  while((m_buffers-3) * m_blockSize < maxBlock)
    m_buffers++;

  // ulong totalFrames = m_buffers * m_blockSize;

  m_blockChars = m_userSWidth * ch * m_blockSize;
  ulong bufBytes = m_blockChars * m_buffers;

  m_tmpBuffer = (char *) realloc(m_tmpBuffer, bufBytes);
  if(m_tmpBuffer) bzero(m_tmpBuffer, bufBytes);

  // DI_trace("DI_AudioFileThreaded::allocateBuffer", m_blockSize);
  /* DI_trace2("DI_AudioFileThreaded::allocateBuffer %d %d %d %d %d %d %d %d %p", 
	    (int) bufBytes, (int) m_blockSize, (int) ch, 
	    (int) m_buffers, (int) m_userSWidth, (int) totalFrames, 
	    (int) m_blockChars, (int) (m_blockChars * m_buffers),
	    m_tmpBuffer);*/

  return m_tmpBuffer != 0;
}

void DI_AudioFileThreaded::childLoop()
{
  const char * fname = "DI_AudioFileThreaded::childLoop";

  DI_trace(fname, m_fileName);
  
  if(m_priority >= 0) {
    sched_param param;
    param.sched_priority = m_priority;
    pid_t pid = getpid();
    sched_setscheduler(pid, SCHED_FIFO, &param);
  }

  bool ok = m_file.open(m_fileName, m_mode, m_userFormat, m_blockSize);
  m_ok = m_ok && ok;

  if(errno) {
    DI_error2(DI_ERR_UNKNOWN, "%s # \"%s\"", fname, strerror(errno));
    errno = 0;
  }

  if(ok) {
    // DI_trace(fname, "IS OPEN");

    if(m_mode == DI_IO_INPUT) {
      m_file.seekFrame(m_startFrame);
      m_chans = m_file.inChannels();
    }
    else
      m_chans = m_file.outChannels();

    ok = allocateBuffer(m_userFormat, m_blockSize, m_chans);

    m_ok = m_ok && ok;

    // DI_trace(fname, "IS ALLOCATED");

    m_status = DI_AFT_READY;
    
    m_cond.signal2(m_mutex);

    // DI_trace2("%s # ok = %d", fname, (int) ok);

    if(m_mode == DI_IO_INPUT)
      while(!m_end && m_ok)
	readAudio();
    else {
      while(!m_end && m_ok)
	writeAudio();
      writeRemaining();
    }
  }

  // DI_trace(fname, "CLOSE FILE");
  m_file.close();

  // DI_trace(fname, "FREEING MEMORY");
  
  // Free memory.
  free(m_tmpBuffer);
  m_tmpBuffer = 0;

  // DI_trace(fname, "EXIT");
}

/** Read the frames from the audio file into the temporary
    buffer. 
    
    File-thread method. */
void DI_AudioFileThreaded::readAudio()
{
  /* DI_trace2("DI_AudioFileThreaded::readAudio %d %d %d", 
     m_fills, m_drains, m_threadFrames); */
  
  /* Check that we do not everrun the client read-head. */
  m_mutex.lock();
  while(m_fills >= (m_drains + m_buffers - 1) && !m_end)
    m_cond.wait(m_mutex);
  m_mutex.unlock();
  
  if(m_end) return;

  m_status = DI_AFT_READING;
  
  char * ptr = & m_tmpBuffer[m_blockChars * m_threadBuffer];

  long frameCount = (long) m_file.frameCount();
  // Number of frames left in the file.
  long framesLeft;
  if(m_loop)
    framesLeft = (long) (frameCount - (m_threadFrames % frameCount));
  else 
    framesLeft = (long) (frameCount - m_threadFrames);
    
  long blockSize = framesLeft < (long) m_blockSize ? 
    framesLeft : (long) m_blockSize;

  int readFrames = blockSize ? m_file.readFrames(ptr, blockSize) : 0;

  // Print crap to the screen:
  /* float * samples = (float *) ptr;

  for(uint c = 0; c < blockSize * m_chans; c++)
  printf(" # T = %f", samples[c]); */

  // Do the looping
  if(m_loop) {
    
    /* The purpose of "framesLeft" changes now: 
       
       It is the number of frames available for filling.
     */
    framesLeft = (long) (m_blockSize - readFrames);
    ptr += m_userSWidth * m_chans * readFrames;

    while(framesLeft) {

      // DI_trace("DI_AudioFileThreaded::readAudio rewinding to start");
      
      m_file.seekFrame(m_startFrame);
      
      blockSize = framesLeft < frameCount ? framesLeft : frameCount;

      readFrames += m_file.readFrames(ptr, blockSize);
      
      framesLeft -= blockSize;
      ptr += m_userSWidth * m_chans * readFrames;
    }
  }
  // Fill with zero:
  else if(readFrames != (int) m_blockSize) {
    
    // Advance ptr;
    ptr += (readFrames * m_userSWidth * m_chans);
    ulong fillFrames = m_blockSize - readFrames;
    
    if(m_userFormat == DI_ASF_INT16 || m_userFormat == DI_ASF_INT32) {
      bzero(ptr, fillFrames * m_userSWidth * m_chans);
    }
    else if (m_userFormat == DI_ASF_FLOAT32) {

      float *fptr = (float *) ptr;
      uint samples = fillFrames * m_chans;

      for(uint i=0; i < samples; i++) 
	fptr[i] = 0.0;
    }
    else if (m_userFormat == DI_ASF_FLOAT64) {

      double *dptr = (double *) ptr;
      uint samples = fillFrames * m_chans;

      for(uint i=0; i < samples; i++) 
	dptr[i] = 0.0;
    } 
    // DI_trace("DI_AudioFileThreaded::readAudio fillzero", fillFrames);

  }


  m_threadBuffer = (m_threadBuffer + 1) % m_buffers;
  m_threadFrames += readFrames;
  
  m_fills++;

  m_cond.signal2(m_mutex);

  // DI_trace("DI_AudioFileThreaded::readAudio exit");
}

/** Write the audio from the temporary buffer to the file. File-thread
    method. */

void DI_AudioFileThreaded::writeAudio()
{
  /* DI_trace2("DI_AudioFileThreaded::writeAudio %d %d %d", 
     (int) m_threadFrames, (int) m_drains, (int) m_fills); */
  
  // Block until user has written one data block

  m_mutex.lock();
  while((m_drains + 1) >= m_fills && !m_end)
    m_cond.wait(m_mutex);
  m_mutex.unlock();

  if(m_end) return;
  
  m_status = DI_AFT_WRITING;
  
  int writeFrames = m_file.writeFrames
    (&m_tmpBuffer[m_blockChars * m_threadBuffer], m_blockSize);
  
  if((uint) writeFrames != m_blockSize) {
    DI_error(DI_ERR_INTERNAL, 
	     "DI_AudioFileThreaded::writeAudio # write failed");
    m_ok = false;
  }
  
  m_threadBuffer = (m_threadBuffer + 1) % m_buffers;
  m_threadFrames += writeFrames;
  m_drains++;
  
  m_cond.signal2(m_mutex);
}

/** Write any remaining data to the output file. */
void DI_AudioFileThreaded::writeRemaining()
{
  DI_trace("DI_AudioFileThreaded::writeRemaining");

  while(m_userFrames > m_threadFrames) {

    const char * data = & m_tmpBuffer[m_blockChars * m_threadBuffer];

    ulong frames = m_userFrames - m_threadFrames;
    
    if(frames > m_blockSize) frames = m_blockSize;

    m_file.writeFrames(data, frames);

    m_threadFrames += frames;
    m_threadBuffer = (m_threadBuffer + 1) % m_buffers;
  }
}

