/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the DIVA IO library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA IO library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */
//
// DIVA Library. Copyright (C) Jarmo Hiipakka, 1999.
//
// $Log: di_virtual_audio.C,v $
// Revision 1.7  2004/11/30 13:51:48  tilmonen
// Copyright upodates...
//
// Revision 1.6  2004/11/30 13:14:40  tilmonen
// Copyright updates.
//
// Revision 1.5  2004/11/30 12:58:02  tilmonen
// Updated license text.
//
// Revision 1.4  2004/02/06 15:47:53  tilmonen
// Moved many data types from "ulong" to "uint".
//
// Revision 1.3  2002/05/22 12:42:46  tilmonen
// Changes due to port to Windows.
//
// Revision 1.2  2001/07/27 08:23:26  tilmonen
// License text updates (=cosmetic changes).
//
// Revision 1.1.1.1  2001/06/04 09:04:21  tilmonen
// Mustajuuri 4.6.2001, initial CVS cersion
//

#include "di_virtual_audio.h"
#include "di_error.h"
#include "di_numconv.h"

#include <stdlib.h>
#include <stdio.h>  // Using I/O functions from std C for file access
#include <string.h>
#include <errno.h>
#include <di_types.h>
#include <sys/param.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>

static const short DI_SEM_READABLE = 0;
static const short DI_SEM_WRITABLE = 1;

// ---- Constructors ----------------------------------------------------------

DI_VirtualAudio::DI_VirtualAudio() :
  m_name(NULL),
  m_bufBegin(NULL),
  m_bufBeginS(NULL),
  m_bufBeginI(NULL),
  m_bufBeginF(NULL),
  m_bufBeginD(NULL),
  m_info(NULL),
  m_shmID(-1),
  m_semID(-1)
{
}

DI_VirtualAudio::~DI_VirtualAudio(void)
{
  if (m_shmID != -1)
    close();
}

// ---- Operations ------------------------------------------------------------

bool DI_VirtualAudio::open(const char *interface,
			   DI_IoMode mode,
			   DI_AudioSampleFormat format, 
			   uint bufferSize,
			   uint channels)
{
  const char *method = "DI_VirtualAudio::open";
  /* We need to limit the number of samples in the buffer
     because of the way we use (short) type semafores. */
  if (bufferSize > 32767) bufferSize = 32767;
  int size((int)(bufferSize*channels));
  if (format == DI_ASF_INT16)
    size *= 2;
  else if ((format == DI_ASF_INT32) || (format == DI_ASF_FLOAT32))
    size *= 4;
  else if (format == DI_ASF_FLOAT64)
    size *= 8;
  else
  {
    DI_error(DI_ERR_ARGUMENT,method,"Invalid sample format");
    return false;
  }
  
  switch(mode)
  {

   case DI_IO_INPUT:
    if ((bufferSize == 0) || (channels == 0))
    {
      DI_error(DI_ERR_ARGUMENT,method,"Buffer size or channel count equals 0");
      return false;
    }
    m_shmID = shmget(getKey(interface),size+sizeof(DI_VaiInfo),
		     IPC_CREAT|IPC_EXCL|0666);
    if (m_shmID == -1)
    {
      DI_error(DI_ERR_INTERNAL,"shmget failed",strerror(errno));
      return false;
    }
    m_semID = semget(getKey(interface),2,IPC_CREAT|IPC_EXCL|0666);
    if (m_semID == -1)
    {
      DI_error(DI_ERR_INTERNAL,"semget failed",strerror(errno));
      close();
      return false;
    }
    
    m_info = (DI_VaiInfo *)shmat(m_shmID,NULL,0);
    if ((int)m_info == -1)
    {
      DI_error(DI_ERR_INTERNAL,method,"shmat failed");
      close();
      return false;
    }
    m_info->format = format;
    m_info->bufferSize = bufferSize;
    m_info->channels = channels;
    m_info->writers = 0;
    m_info->write = 0;
    m_info->read = 0;
    m_info->isEOF = 0;
    m_name = strdup(interface);
    m_mode = mode;
    m_inFormat = format;
    m_outFormat = format;
    m_bufBegin = (void *)(m_info+1);
    m_bufBeginS = (short *)m_bufBegin;
    m_bufBeginI = (int *)m_bufBegin;
    m_bufBeginF = (float *)m_bufBegin;
    m_bufBeginD = (double *)m_bufBegin;
    break;

   case DI_IO_OUTPUT:
    m_shmID = shmget(getKey(interface),size+sizeof(DI_VaiInfo),0);
    if (m_shmID == -1)
    {
      DI_error(DI_ERR_INTERNAL,"shmget failed",strerror(errno));
      return false;
    }
    m_semID = semget(getKey(interface),2,0);
    if (m_semID == -1)
    {
      DI_error(DI_ERR_INTERNAL,"semget failed",strerror(errno));
      close();
      return false;
    }
    m_info = (DI_VaiInfo *)shmat(m_shmID,NULL,0);
    if ((int)m_info == -1)
    {
      DI_error(DI_ERR_INTERNAL,method,"shmat failed");
      close();
      return false;
    }
    if (m_info->writers == 1)
    {
      DI_error(DI_ERR_INTERNAL,method,"Interface already in use");
      close();
      return false;
    }
    semOp(DI_SEM_WRITABLE,(short)(m_info->bufferSize));
    m_info->writers = 1;
    m_info->isEOF = 0;
    m_name = strdup(interface);
    m_mode = mode;
    m_inFormat = format;
    m_outFormat = format;
    m_bufBegin = (void *)(m_info+1);
    m_bufBeginS = (short *)m_bufBegin;
    m_bufBeginI = (int *)m_bufBegin;
    m_bufBeginF = (float *)m_bufBegin;
    m_bufBeginD = (double *)m_bufBegin;
    break;

   default:
    DI_error(DI_ERR_ARGUMENT,method,"I/O mode not supported");
    return false;
  }
  return true;
}

bool DI_VirtualAudio::close(void)
{
  m_bufBegin = NULL;
  m_bufBeginS = NULL;
  m_bufBeginI = NULL;
  m_bufBeginF = NULL;
  m_bufBeginD = NULL;
  if (m_name)
  {
    free(m_name);
    m_name = NULL;
  }
  m_info->writers = 0;
  m_info->isEOF = 1;
  shmdt((void *)m_info);
  m_info = NULL;
  if (m_mode == DI_IO_INPUT)
  {
    shmctl(m_shmID,IPC_RMID,NULL);
    semctl(m_semID,0,IPC_RMID);
  }
  m_shmID = -1;
  m_semID = -1;
  m_mode = DI_IO_NONE;
  return true;
}

// ---- I/O operations --------------------------------------------------------

int DI_VirtualAudio::readFrames(void *buffer,uint frames)
{
  frames = (frames < (m_info->bufferSize>>1)) ? frames : m_info->bufferSize>>1;
  int filled(filledFrames());
  if (m_info->isEOF) frames = ((uint)filled < frames ? (uint)filled : frames);

  // Wait for a sufficient number of filled frames
  semOp(DI_SEM_READABLE,-((short)frames));

  /* Format dependent copy. Note that the shared memory buffer always has
     the format that we are reading. Format conversion only at the writer */
  if (m_info->read + frames > m_info->bufferSize)
  {
    uint n1(m_info->bufferSize - m_info->read);
    switch(m_inFormat)
    {
     case DI_ASF_INT16:
      memcpy(buffer, (void *)(m_bufBeginS + m_info->channels*m_info->read),
	     n1 * m_info->channels * 2);
      memcpy((void *)(((short *)buffer) + n1*m_info->channels), m_bufBegin,
	     (frames-n1) * m_info->channels * 2);
      break;
     case DI_ASF_INT32:
      memcpy(buffer, (void *)(m_bufBeginI + m_info->channels*m_info->read),
	     n1 * m_info->channels * 4);
      memcpy((void *)(((int *)buffer) + n1*m_info->channels), m_bufBegin,
	     (frames-n1) * m_info->channels * 4);
      break;
     case DI_ASF_FLOAT32:
      memcpy(buffer, (void *)(m_bufBeginF + m_info->channels*m_info->read),
	     n1 * m_info->channels * 4);
      memcpy((void *)(((float *)buffer) + n1*m_info->channels), m_bufBegin,
	     (frames-n1) * m_info->channels * 4);
      break;
     case DI_ASF_FLOAT64:
      memcpy(buffer, (void *)(m_bufBeginD + m_info->channels*m_info->read),
	     n1 * m_info->channels * 8);
      memcpy((void *)(((double *)buffer) + n1*m_info->channels), m_bufBegin,
	     (frames-n1) * m_info->channels * 8);
      break;
    }
    m_info->read = frames-n1;
  }
  else
  {
    switch(m_inFormat)
    {
     case DI_ASF_INT16:
      memcpy(buffer, (void *)(m_bufBeginS + m_info->channels*m_info->read),
	     frames * m_info->channels * 2);
      break;
     case DI_ASF_INT32:
      memcpy(buffer, (void *)(m_bufBeginI + m_info->channels*m_info->read),
	     frames * m_info->channels * 4);
      break;
     case DI_ASF_FLOAT32:
      memcpy(buffer, (void *)(m_bufBeginF + m_info->channels*m_info->read),
	     frames * m_info->channels * 4);
      break;
     case DI_ASF_FLOAT64:
      memcpy(buffer, (void *)(m_bufBeginD + m_info->channels*m_info->read),
	     frames * m_info->channels * 8);
      break;
    }
    m_info->read += frames;
  }
  semOp(DI_SEM_WRITABLE, (short)frames);
  return (int)frames;
}

int DI_VirtualAudio::writeFrames(const void *buffer,uint frames)
{
  frames = (frames < (m_info->bufferSize>>1)) ? frames : m_info->bufferSize>>1;

  // Wait for a sufficient number of empty frames
  semOp(DI_SEM_WRITABLE,-((short)frames));

  /// Format dependent copy or conversion.
  if (m_info->write + frames > m_info->bufferSize)
  {
    uint n1(m_info->bufferSize - m_info->write);
    switch(m_outFormat)
    {
     case DI_ASF_INT16:
      switch(m_info->format)
      {
       case DI_ASF_INT16:
	memcpy((void *)(m_bufBeginS + m_info->channels*m_info->write), buffer,
	       n1 * m_info->channels * 2);
	memcpy(m_bufBegin, (void *)(((short *)buffer) + n1*m_info->channels), 
	       (frames-n1) * m_info->channels * 2);
	break;
       case DI_ASF_INT32:
	DI_short2Int(m_bufBeginI + m_info->channels*m_info->write,
		     (short *)buffer, 16, 0, n1 * m_info->channels);
	DI_short2Int(m_bufBeginI, ((short *)buffer) + n1*m_info->channels,
		     16, 0, (frames-n1) * m_info->channels);
	break;
       case DI_ASF_FLOAT32:
	DI_short2Float(m_bufBeginF + m_info->channels*m_info->write,
		       (short *)buffer, 1.0/32768.0, 0, n1 * m_info->channels);
	DI_short2Float(m_bufBeginF, ((short *)buffer) + n1*m_info->channels,
		       1.0/32768.0, 0, (frames-n1) * m_info->channels);
	break;
       case DI_ASF_FLOAT64:
	DI_short2Double(m_bufBeginD + m_info->channels*m_info->write,
			(short *)buffer, 1.0/32768.0, 0,
			n1 * m_info->channels);
	DI_short2Double(m_bufBeginD, ((short *)buffer) + n1*m_info->channels,
			1.0/32768.0, 0, (frames-n1) * m_info->channels);
	break;
      }
      break;
     case DI_ASF_INT32:
      switch(m_info->format)
      {
       case DI_ASF_INT16:
	DI_int2Short(m_bufBeginS + m_info->channels*m_info->write,
		     (int *)buffer, -16, 0, n1 * m_info->channels);
	DI_int2Short(m_bufBeginS, ((int *)buffer) + n1*m_info->channels,
		     -16, 0, (frames-n1) * m_info->channels);
	break;
       case DI_ASF_INT32:
	memcpy((void *)(m_bufBeginI + m_info->channels*m_info->write), buffer,
	       n1 * m_info->channels * 4);
	memcpy(m_bufBegin, (void *)(((int *)buffer) + n1*m_info->channels), 
	       (frames-n1) * m_info->channels * 4);
	break;
       case DI_ASF_FLOAT32:
	DI_int2Float(m_bufBeginF + m_info->channels*m_info->write,(int *)buffer,
		     1.0/2147483648.0, 0, n1 * m_info->channels);
	DI_int2Float(m_bufBeginF, ((int *)buffer) + n1*m_info->channels,
		     1.0/2147483648.0, 0, (frames-n1) * m_info->channels);
	break;
       case DI_ASF_FLOAT64:
	DI_int2Double(m_bufBeginD + m_info->channels*m_info->write,
		      (int *)buffer, 1.0/2147483648.0, 0,
		      n1 * m_info->channels);
	DI_int2Double(m_bufBeginD, ((int *)buffer) + n1*m_info->channels,
		      1.0/2147483648.0, 0, (frames-n1) * m_info->channels);
	break;
      }
      break;
     case DI_ASF_FLOAT32:
      switch(m_info->format)
      {
       case DI_ASF_INT16:
	DI_float2Short(m_bufBeginS + m_info->channels*m_info->write,
		       (float *)buffer, 32767.0, 0, n1 * m_info->channels);
	DI_float2Short(m_bufBeginS, ((float *)buffer) + n1*m_info->channels,
		       32767.0, 0, (frames-n1) * m_info->channels);
	break;
       case DI_ASF_INT32:
 	DI_float2Int(m_bufBeginI + m_info->channels*m_info->write,
		     (float *)buffer, 2147483647.0, 0, n1 * m_info->channels);
	DI_float2Int(m_bufBeginI, ((float *)buffer) + n1*m_info->channels,
		     2147483647.0, 0, (frames-n1) * m_info->channels);
	break;
       case DI_ASF_FLOAT32:
	memcpy((void *)(m_bufBeginF + m_info->channels*m_info->write), buffer,
	       n1 * m_info->channels * 4);
	memcpy(m_bufBegin, (void *)(((float *)buffer) + n1*m_info->channels), 
	       (frames-n1) * m_info->channels * 4);
	break;
       case DI_ASF_FLOAT64:
	DI_float2Double(m_bufBeginD + m_info->channels*m_info->write,
			(float *)buffer, n1 * m_info->channels);
	DI_float2Double(m_bufBeginD, ((float *)buffer) + n1*m_info->channels,
			(frames-n1) * m_info->channels);
	break;
      }
      break;
     case DI_ASF_FLOAT64:
      switch(m_info->format)
      {
       case DI_ASF_INT16:
	DI_double2Short(m_bufBeginS + m_info->channels*m_info->write,
			(double *)buffer, 32767.0, 0, n1 * m_info->channels);
	DI_double2Short(m_bufBeginS, ((double *)buffer) + n1*m_info->channels,
			32767.0, 0, (frames-n1) * m_info->channels);
	break;
       case DI_ASF_INT32:
	DI_double2Int(m_bufBeginI + m_info->channels*m_info->write,
		      (double *)buffer, 2147483647.0, 0,
		      n1 * m_info->channels);
	DI_double2Int(m_bufBeginI, ((double *)buffer) + n1*m_info->channels,
		      2147483647.0, 0, (frames-n1) * m_info->channels);
	break;
       case DI_ASF_FLOAT32:
	DI_double2Float(m_bufBeginF + m_info->channels*m_info->write,
			(double *)buffer, n1 * m_info->channels);
	DI_double2Float(m_bufBeginF, ((double *)buffer) + n1*m_info->channels,
			(frames-n1) * m_info->channels);
	break;
       case DI_ASF_FLOAT64:
	memcpy((void *)(m_bufBeginD + m_info->channels*m_info->write), buffer,
	       n1 * m_info->channels * 8);
	memcpy(m_bufBegin, (void *)(((double *)buffer) + n1*m_info->channels), 
	       (frames-n1) * m_info->channels * 8);
	break;
      }
      break;
    }
    m_info->write = frames-n1;
  }

  else
  {
    switch(m_outFormat)
    {
     case DI_ASF_INT16:
      switch(m_info->format)
      {
       case DI_ASF_INT16:
	memcpy((void *)(m_bufBeginS + m_info->channels*m_info->write), buffer,
	       frames * m_info->channels * 2);
	break;
       case DI_ASF_INT32:
	DI_short2Int(m_bufBeginI + m_info->channels*m_info->write,
		     (short *)buffer, 16, 0, frames * m_info->channels);
	break;
       case DI_ASF_FLOAT32:
	DI_short2Float(m_bufBeginF + m_info->channels*m_info->write,
		       (short *)buffer, 1.0/32768.0, 0,
		       frames * m_info->channels);
	break;
       case DI_ASF_FLOAT64:
	DI_short2Double(m_bufBeginD + m_info->channels*m_info->write,
			(short *)buffer, 1.0/32768.0, 0,
			frames * m_info->channels);
	break;
      }
      break;
     case DI_ASF_INT32:
      switch(m_info->format)
      {
       case DI_ASF_INT16:
	DI_int2Short(m_bufBeginS + m_info->channels*m_info->write,
		     (int *)buffer, -16, 0, frames * m_info->channels);
	break;
       case DI_ASF_INT32:
	memcpy((void *)(m_bufBeginI + m_info->channels*m_info->write), buffer,
	       frames * m_info->channels * 4);
	break;
       case DI_ASF_FLOAT32:
	DI_int2Float(m_bufBeginF + m_info->channels*m_info->write,
		     (int *)buffer, 1.0/2147483648.0, 0,
		     frames*m_info->channels);
	break;
       case DI_ASF_FLOAT64:
	DI_int2Double(m_bufBeginD + m_info->channels*m_info->write,
		      (int *)buffer, 1.0/2147483648.0, 0,
		      frames * m_info->channels);
	break;
      }
      break;
     case DI_ASF_FLOAT32:
      switch(m_info->format)
      {
       case DI_ASF_INT16:
	DI_float2Short(m_bufBeginS + m_info->channels*m_info->write,
		       (float *)buffer, 32767.0, 0, frames * m_info->channels);
	break;
       case DI_ASF_INT32:
 	DI_float2Int(m_bufBeginI + m_info->channels*m_info->write,
		     (float *)buffer, 2147483647.0, 0,
		     frames * m_info->channels);
	break;
       case DI_ASF_FLOAT32:
	memcpy((void *)(m_bufBeginF + m_info->channels*m_info->write), buffer,
	       frames * m_info->channels * 4);
	break;
       case DI_ASF_FLOAT64:
	DI_float2Double(m_bufBeginD + m_info->channels*m_info->write,
			(float *)buffer, frames * m_info->channels);
	break;
      }
      break;
     case DI_ASF_FLOAT64:
      switch(m_info->format)
      {
       case DI_ASF_INT16:
	DI_double2Short(m_bufBeginS + m_info->channels*m_info->write,
			(double *)buffer, 32767.0, 0,
			frames * m_info->channels);
	break;
       case DI_ASF_INT32:
	DI_double2Int(m_bufBeginI + m_info->channels*m_info->write,
		      (double *)buffer, 2147483647.0, 0,
		      frames * m_info->channels);
	break;
       case DI_ASF_FLOAT32:
	DI_double2Float(m_bufBeginF + m_info->channels*m_info->write,
			(double *)buffer, frames * m_info->channels);
	break;
       case DI_ASF_FLOAT64:
	memcpy((void *)(m_bufBeginD + m_info->channels*m_info->write), buffer,
	       frames * m_info->channels * 8);
	break;
      }
      break;
    }
    m_info->write += frames;
  }
  semOp(DI_SEM_READABLE, (short)frames);
  return (int)frames;
}

int DI_VirtualAudio::waitInputFilled(uint frames, int timeOut)
{
  (void)timeOut;
  int c(0);
  if (m_mode & DI_IO_INPUT)
    while ((c = filledFrames()) < (int)frames)
    {
      semOp(DI_SEM_READABLE, -((short)frames));
      semOp(DI_SEM_READABLE, (short)frames);
    }
  return c;
}

int DI_VirtualAudio::waitOutputFilled(uint frames, int timeOut)
{
  (void)timeOut;
  int c(0);
  if (m_mode & DI_IO_OUTPUT)
    while ((c = filledFrames()) > (int)frames)
    {
      semOp(DI_SEM_WRITABLE, -((short)frames));
      semOp(DI_SEM_WRITABLE, (short)frames);
    }
  return c;
}

// ---- Predicates ------------------------------------------------------------

bool DI_VirtualAudio::isOK(void) const
{
  return ((m_mode == DI_IO_INPUT) || (m_mode == DI_IO_OUTPUT)) &&
    (m_shmID != -1) && (m_info != NULL);
}

// ---- Private methods -------------------------------------------------------
key_t DI_VirtualAudio::getKey(const char *name)
{
  // Compose the path.
  char path[MAXPATHLEN];
  strncpy(path,getenv("DIVAPATH"),MAXPATHLEN);
  strncat(path,"/etc/config/DI_VirtualAudio.cfg",MAXPATHLEN-strlen(path));

  // Open file.
  FILE *config(fopen(path,"r"));
  if (!config)
  {
    DI_error(DI_ERR_INTERNAL, "Unable to open config file");
    return -1;
  }

  const int MAXLEN(1000);
  char tempLine[MAXLEN];
  char tempName[MAXLEN];
  int id;
  char *notEOF;

  do
  {
    while ((notEOF = fgets(tempLine,MAXLEN,config)) && (tempLine[0] == '#'))
      ;
    sscanf(tempLine, "%d: %s", &id, tempName);
    if (tempName[0] == '"')
    {
      int start(0),stop(strlen(tempLine)-1);
      while (tempLine[start++] != '"')
	;
      while (tempLine[stop--] != '"')
	;
      tempLine[stop+1] = '\0';
      strcpy(tempName,tempLine+start);
    }
  }
  while ((strcmp(tempName,name) != 0) && notEOF);

  if (!notEOF)
  {
    DI_error(DI_ERR_INTERNAL, "Error reading config file");
    fclose(config);
    return -1;
  }

  fclose(config);
  return ftok(path,id+1);
}

int DI_VirtualAudio::filledFrames(void) const
{
  return (m_info->write >= m_info->read) ?
    (int)(m_info->write - m_info->read) :
    (int)(m_info->bufferSize + m_info->write - m_info->read);
}
