/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the DIVA IO library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000-2001.
 * Jarmo.Hiipakka@hut.fi Tommi.Ilmonen@hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA IO library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */

//
// DIVA Library. Copyright (C) Jarmo Hiipakka, Tommi Ilmonen, 1999-2001.
//
// $Log: di_audio_port.C,v $
// Revision 1.5  2004/02/23 13:17:20  tilmonen
// ulong -> uint.
//
// Revision 1.4  2003/06/09 11:40:08  tilmonen
// Fixes.
//
// Revision 1.3  2002/11/25 09:22:48  tilmonen
// Fake debug option.
//
// Revision 1.2  2001/07/27 08:23:29  tilmonen
// License text updates (=cosmetic changes).
//
// Revision 1.1.1.1  2001/06/04 09:04:21  tilmonen
// Mustajuuri 4.6.2001, initial CVS cersion
//

#include "di_audio_port.h"
#include "di_error.h"
#include "di_numconv.h"

#include <stdio.h>  // Using I/O functions from std C for /dev/sndstat access
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <math.h>  // Needed for the log() function

#include <sys/ioctl.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>

#ifdef WORDS_BIGENDIAN

/* This code only works on little endian machines. To add support
   for big endians, please modify the data conversion routines
   accordingly.
*/

#pragma error 

DO_NOT_COMPILE();

#endif


// ---- Audio port class ------------------------------------------------------

// 2^15 - 1
#define DI_OSS_SSCALE 32767
// 2^31 - 1
#define DI_OSS_ISCALE 2147483647

bool DI_AudioPort::m_debug = false;

DI_AudioPort::DI_AudioPort() : 
  m_inDevice(NULL),
  m_outDevice(NULL),
  m_mode(DI_IO_NONE),
  m_inFormat(DI_ASF_FLOAT32),
  m_outFormat(DI_ASF_FLOAT32),
  m_inChannels(0),
  m_outChannels(0),
  m_inBufSize(0),
  m_outBufSize(0),
  m_inRate(0),
  m_outRate(0),
  m_inBlock(0),
  m_outBlock(0),
  m_tmpBuffer(0),
  m_tmpBufferFrames(0),
  //  m_maxInBytes(0),
  m_maxOutBytes(0),
  m_inFD(-1),
  m_outFD(-1),
  m_in32Bit(false),
  m_out32Bit(false)
{
}

DI_AudioPort::~DI_AudioPort()
{
  close();
  delete [] (char *) m_tmpBuffer;
}

int DI_AudioPort::inFilledFrames(void) const
{
  if (ioctl(m_inFD, SNDCTL_DSP_GETISPACE, &m_bufInfo) == -1)
    return -1;
  
  return m_bufInfo.bytes / ((m_in32Bit ? 4 : 2) * m_inChannels);
}

int DI_AudioPort::outFilledFrames(void) const
{
  if (!m_maxOutBytes) return 0;
  if (ioctl(m_outFD, SNDCTL_DSP_GETOSPACE, &m_bufInfo) == -1)
    return -1;

  int r(((int) m_maxOutBytes - (int) m_bufInfo.bytes) / 
	(int) ((m_out32Bit ? 4 : 2) * m_outChannels));
  return (r < 0 ? 0 : r);
}

bool DI_AudioPort::open(/// Audio device.
			const char *device,
			/// IO mode: none, input, output, or input&output.
			DI_IoMode mode, 
			/// Audio port sample format.
			DI_AudioSampleFormat userSample, 
			/// Audio port buffer size (sample frames).
			uint bufferSize,
			/// Audio sampling rate.
			uint samplerate,
			/// Number of channels.
			uint channels,
			/// Maximum block size.
			uint maxBlock)
{
  const char * fname = "DI_AudioPort::open";
  
  const char *dev(device ? device : "/dev/dsp");
  int imode(0);
  m_mode = mode;
  switch(mode) 
  {
   case DI_IO_NONE:
    close();
    return false;
   case DI_IO_INPUT:
    imode = O_RDONLY;
    break;
   case DI_IO_OUTPUT:
    imode = O_WRONLY;
    break;
   case DI_IO_INPUT_OUTPUT:
    imode = O_RDWR;
    break;
   default:
    DI_error(DI_ERR_INTERNAL, "Invalid I/O mode");
    close();
    return false;
  }

  errno = 0;

  int fd(-1);
  fd = ::open(dev, imode, 0);
  
  if (fd == -1) 
  {
    DI_error(DI_ERR_INTERNAL, fname, strerror(errno));
    DI_error(DI_ERR_INTERNAL, "Failed to open audio device", dev);
    m_inFD = -1;
    m_outFD = -1;
    close();
    return false;
  }

  m_inFormat = userSample;
  m_outFormat = userSample;

  int ch = (int)channels;
  if ((ioctl(fd, SNDCTL_DSP_CHANNELS, &ch) == -1) || (ch != (int)channels))
  {
    DI_error(DI_ERR_INTERNAL, fname, "Could not set the number of channels");
    close();
    return false;
  }
  m_inChannels = (uint)ch;
  m_outChannels = (uint)ch;

  m_inRate = samplerate;
  if (ioctl(fd, SNDCTL_DSP_SPEED, &samplerate) == -1)
  {
    DI_error(DI_ERR_INTERNAL, fname, strerror(errno));
  }
  if (samplerate != m_inRate)
  {
    if (m_inRate != 0)
    {
      DI_error(DI_ERR_ARGUMENT, fname, "Unable to set the desired samplerate");
      close();
      return false;
    }
  }
  
#if defined(AFMT_S32_LE)
  if ((m_inFormat == DI_ASF_INT16) && (m_outFormat == DI_ASF_INT16))
  {
    m_in32Bit = false;
    m_out32Bit = false;
  }
  else
  {
    m_in32Bit = true;
    m_out32Bit = true;
  }
#else
  m_in32Bit = false;
  m_out32Bit = false;
#endif
  
  // Resize data buffer.
  if(m_tmpBuffer)
  {
    delete [] (char * ) m_tmpBuffer;
    m_tmpBuffer = NULL;
  }
  if (m_in32Bit || m_out32Bit)
    m_tmpBuffer = new char[maxBlock * channels * 4];
  else
    m_tmpBuffer = new char[maxBlock * channels * 2];

  m_tmpBufferFrames = maxBlock;
  m_inBlock = maxBlock;
  m_outBlock = maxBlock;
  
  // Sample format:
  int format(AFMT_S16_LE);
#if defined(AFMT_S32_LE)
  if (m_in32Bit || m_out32Bit) format = AFMT_S32_LE;
#endif
  int dformat(format);
  if (ioctl(fd, SNDCTL_DSP_SETFMT, &format) == -1)
  {
    DI_error(DI_ERR_INTERNAL, fname, strerror(errno));
    close();
    return false;
  }  
  if (format != dformat)
  {
    if (m_in32Bit || m_out32Bit)   // requested AFMT_S32_LE
    {
      format = AFMT_S16_LE;
      if (ioctl(fd, SNDCTL_DSP_SETFMT, &format) == -1)
      {
	DI_error(DI_ERR_INTERNAL, fname, strerror(errno));
	close();
	return false;
      }
      if (format != AFMT_S16_LE)
      {
	DI_error(DI_ERR_INTERNAL,fname,"AFMT_S16_LE format is not available");
	close();
	return false;
      }
      m_in32Bit = false;
      m_out32Bit = false;
    }
    else
    {
      DI_error(DI_ERR_INTERNAL, fname, "The desired format is not available");
      close();
      return false;
    }
  }

  // Set fragment sizes and number (desired = 64 samples / fragment)
  int desFragSize(64 * channels * ((m_in32Bit||m_out32Bit) ? 4 : 2));
  int desFragNum(bufferSize/64);
  // See OSS documentation for block configuration
  int desBlockConf((desFragNum<<16) +
		   (int)rint(log((double)desFragSize)/log(2)));
  ioctl(fd, SNDCTL_DSP_SETFRAGMENT, &desBlockConf);
  int trueFragSize(desFragSize);
  // Get the true fragment size from the driver
  if (ioctl(fd, SNDCTL_DSP_GETBLKSIZE, &trueFragSize) == -1) 
    DI_error(DI_ERR_INTERNAL, fname, "Cannot determine block size");
  else
  {
    // Set the number of fragments according to the true fragment size
    desFragNum = channels * ((m_in32Bit||m_out32Bit) ? 4 : 2) * bufferSize /
      trueFragSize;
    // The number of fragments cannot be smaller than two.
    if (desFragNum < 2)
      desFragNum = 2;
    desBlockConf = (desFragNum<<16) +
      (int)rint(log((double)desFragSize)/log(2));
    ioctl(fd, SNDCTL_DSP_SETFRAGMENT, &desBlockConf);
  }
  
  // Determine how many bytes fit to driver IO buffers:
  if((int)mode  & (int)DI_IO_INPUT)
  {
    ioctl(fd, SNDCTL_DSP_GETISPACE, &m_bufInfo);
    m_inBufSize = (m_bufInfo.fragstotal * m_bufInfo.fragsize) /
      (channels * (m_in32Bit ? 4 : 2));
  }
  if((int)mode  & (int)DI_IO_OUTPUT)
  {
    ioctl(fd, SNDCTL_DSP_GETOSPACE, &m_bufInfo);
    m_maxOutBytes = m_bufInfo.fragstotal * m_bufInfo.fragsize - 
      m_bufInfo.fragsize - m_bufInfo.fragsize / 2;
    m_outBufSize = (m_bufInfo.fragstotal * m_bufInfo.fragsize) /
      (channels * (m_out32Bit ? 4 : 2));
  }
  
  switch(m_mode)
  {
   case DI_IO_NONE:
    close();
    m_inFD = -1;
    m_outFD = -1;
    m_inRate = 0;
    m_outRate = 0;
    break;
   case DI_IO_INPUT:
    m_inFD = fd;
    m_outFD = -1;
    m_inDevice = (device ? strdup(device) : NULL);
    m_outDevice = NULL;
    m_inRate = samplerate;
    m_outRate = 0;
    break;
   case DI_IO_OUTPUT:
    m_inFD = -1;
    m_outFD = fd;
    m_inDevice = NULL;
    m_outDevice = (device ? strdup(device) : NULL);
    m_inRate = 0;
    m_outRate = samplerate;
    break;
   case DI_IO_INPUT_OUTPUT:
    m_inFD = fd;
    m_outFD = fd;
    m_inDevice = (device ? strdup(device) : NULL);
    m_outDevice = (device ? strdup(device) : NULL);
    m_inRate = samplerate;
    m_outRate = samplerate;
    break;
   default:
    close();
    return false;
  }
  getLabel(m_inDevice, true);
  getLabel(m_outDevice, false);
  
  return true;
}

bool DI_AudioPort::open(const char *inDevice,
			const char *outDevice,
			DI_AudioSampleFormat inFormat, 
			DI_AudioSampleFormat outFormat, 
			uint inBufSize,
			uint outBufSize,
			uint inSamplerate,
			uint outSamplerate,
			uint inChannels,
			uint outChannels,
			uint inMaxBlock,
			uint outMaxBlock)
{
  if (strcmp(inDevice, outDevice) == 0)
  {
    bool res(open(inDevice,DI_IO_INPUT_OUTPUT,inFormat,inBufSize,inSamplerate,
		  inChannels,inMaxBlock));
    m_outFormat = outFormat;
    return res;
  }
  
  const char * fname = "DI_AudioPort::open";

  m_mode = DI_IO_INPUT_OUTPUT;
  m_inFormat = inFormat;
  m_outFormat = outFormat;

  if (inDevice)
    m_inDevice = strdup(inDevice);
  else
    m_inDevice = NULL;
  if (outDevice)
    m_outDevice = strdup(outDevice);
  else
    m_outDevice = NULL;
  getLabel(m_inDevice, true);
  getLabel(m_outDevice, false);
  
  char *dev = m_inDevice;
  if (!dev) dev = "/dev/dsp";
  m_inFD = ::open(dev, O_RDONLY, 0);
  
  if (m_inFD == -1) 
  {
    DI_error(DI_ERR_INTERNAL, fname, strerror(errno));
    DI_error(DI_ERR_INTERNAL, "Failed to open audio device", dev);
    close();
    return false;
  }
  
  dev = m_outDevice;
  if (!dev) dev = "/dev/dsp";
  m_outFD = ::open(dev, O_WRONLY, 0);

  if (m_outFD == -1)
  {
    DI_error(DI_ERR_INTERNAL, fname, strerror(errno));
    DI_error(DI_ERR_INTERNAL, "Failed to open audio device", dev);
    close();
    return false;
  }

  int ch = (int)inChannels;
  if ((ioctl(m_inFD,SNDCTL_DSP_CHANNELS,&ch)==-1) || (ch!=(int)inChannels))
  {
    DI_error(DI_ERR_INTERNAL, fname, "Could not set the number of channels");
    close();
    return false;
  }
  m_inChannels = (uint)ch;
  ch = (int)outChannels;
  if ((ioctl(m_outFD,SNDCTL_DSP_CHANNELS,&ch)==-1) || (ch!=(int)outChannels))
  {
    DI_error(DI_ERR_INTERNAL, fname, "Could not set the number of channels");
    close();
    return false;
  }
  m_outChannels = (uint)ch;

  m_inRate = inSamplerate;
  if (ioctl(m_inFD, SNDCTL_DSP_SPEED, &inSamplerate) == -1)
  {
    DI_error(DI_ERR_INTERNAL, fname, strerror(errno));
  }
  if (inSamplerate != m_inRate)
  {
    if (m_inRate != 0)
    {
      DI_error(DI_ERR_ARGUMENT, fname, "Unable to set the desired samplerate");
      close();
      return false;
    }
    else
    {
      m_inRate = inSamplerate;
    }
  }

  m_outRate = outSamplerate;
  if (ioctl(m_outFD, SNDCTL_DSP_SPEED, &outSamplerate) == -1)
  {
    DI_error(DI_ERR_INTERNAL, fname, strerror(errno));
  }
  if (outSamplerate != m_outRate)
  {
    if (m_outRate != 0)
    {
      DI_error(DI_ERR_ARGUMENT, fname, "Unable to set the desired samplerate");
      close();
      return false;
    }
    else
    {
      m_outRate = outSamplerate;
    }
  }
  
#if defined(AFMT_S32_LE)
  if (m_inFormat == DI_ASF_INT16) m_in32Bit = false;
  else m_in32Bit = true;
  if (m_outFormat == DI_ASF_INT16) m_out32Bit = false;
  else m_out32Bit = true;
#else
  m_in32Bit = false;
  m_out32Bit = false;
#endif

  // Resize data buffer.
  if(m_tmpBuffer)
  {
    delete (char *) m_tmpBuffer;
    m_tmpBuffer = NULL;
  }
  uint maxBlock(inMaxBlock>outMaxBlock ? inMaxBlock : outMaxBlock);
  uint maxChan(inChannels>outChannels ? inChannels : outChannels);
  if (m_in32Bit || m_out32Bit)
    m_tmpBuffer = new int[maxBlock * maxChan];
  else
    m_tmpBuffer = new short[maxBlock * maxChan];
  m_tmpBufferFrames = maxBlock;
  m_inBlock = maxBlock;
  m_outBlock = maxBlock;
  
  // Sample format:
  int format(AFMT_S16_LE);
#if defined(AFMT_S32_LE)
  if (m_in32Bit) format = AFMT_S32_LE;
#endif
  int dformat(format);
  if (ioctl(m_inFD, SNDCTL_DSP_SETFMT, &format) == -1)
  {
    DI_error(DI_ERR_INTERNAL, fname, strerror(errno));
    close();
    return false;
  }  
  if (format != dformat)
  {
    if (m_in32Bit)   // requested AFMT_S32_LE
    {
      format = AFMT_S16_LE;
      if (ioctl(m_inFD, SNDCTL_DSP_SETFMT, &format) == -1)
      {
	DI_error(DI_ERR_INTERNAL, fname, strerror(errno));
	close();
	return false;
      }
      if (format != AFMT_S16_LE)
      {
	DI_error(DI_ERR_INTERNAL,fname,"AFMT_S16_LE format is not available");
	close();
	return false;
      }
      m_in32Bit = false;
    }
    else
    {
      DI_error(DI_ERR_INTERNAL, fname, "The desired format is not available");
      close();
      return false;
    }
  }
  format = AFMT_S16_LE;
#if defined(AFMT_S32_LE)
  if (m_out32Bit) format = AFMT_S32_LE;
#endif
  dformat = format;
  if (ioctl(m_outFD, SNDCTL_DSP_SETFMT, &format) == -1)
  {
    DI_error(DI_ERR_INTERNAL, fname, strerror(errno));
    close();
    return false;
  }  
  if (format != dformat)
  {
    if (m_out32Bit)   // requested AFMT_S32_LE
    {
      format = AFMT_S16_LE;
      if (ioctl(m_outFD, SNDCTL_DSP_SETFMT, &format) == -1)
      {
	DI_error(DI_ERR_INTERNAL, fname, strerror(errno));
	close();
	return false;
      }
      if (format != AFMT_S16_LE)
      {
	DI_error(DI_ERR_INTERNAL,fname,"AFMT_S16_LE format is not available");
	close();
	return false;
      }
      m_out32Bit = false;
    }
    else
    {
      DI_error(DI_ERR_INTERNAL, fname, "The desired format is not available");
      close();
      return false;
    }
  }

  // Set fragment sizes and number (desired = 64 samples / fragment)
  int desFragSize(64 * inChannels * (m_in32Bit ? 4 : 2));
  int desFragNum(inBufSize/64);
  // See OSS documentation for block configuration
  int desBlockConf((desFragNum<<16) +
		   (int)rint(log((double)desFragSize)/log(2)));
  ioctl(m_inFD, SNDCTL_DSP_SETFRAGMENT, &desBlockConf);
  int trueFragSize(desFragSize);
  // Get the true fragment size from the driver
  if(ioctl(m_inFD, SNDCTL_DSP_GETBLKSIZE, &trueFragSize) == -1) 
    DI_error(DI_ERR_INTERNAL, fname, "Cannot determine block size");
  else
  {
    // Set the number of fragments according to the true fragment size
    desFragNum = inChannels * (m_in32Bit ? 4 : 2) * inBufSize / trueFragSize;
    // The number of fragments cannot be smaller than two.
    if (desFragNum < 2)
      desFragNum = 2;
    desBlockConf = (desFragNum<<16) +
      (int)rint(log((double)desFragSize)/log(2));
    ioctl(m_inFD, SNDCTL_DSP_SETFRAGMENT, &desBlockConf);
  }
  desFragSize = 64 * outChannels * (m_out32Bit ? 4 : 2);
  printf("%d\n",desFragSize);
  desFragNum = outBufSize/64;
  printf("%d\n",desFragNum);
  desBlockConf = (desFragNum<<16) +
    (int)rint(log((double)desFragSize)/log(2));
  printf("%x\n",desBlockConf);
  ioctl(m_outFD, SNDCTL_DSP_SETFRAGMENT, &desBlockConf);
  trueFragSize = desFragSize;
  if(ioctl(m_outFD, SNDCTL_DSP_GETBLKSIZE, &trueFragSize) == -1) 
    DI_error(DI_ERR_INTERNAL, fname, "Cannot determine block size");
  else
  {
    desFragNum = outChannels * (m_out32Bit ? 4 : 2) * outBufSize/trueFragSize;
    if (desFragNum < 2)
      desFragNum = 2;
    desBlockConf = desFragNum<<16 +
      (int)rint(log((double)desFragSize)/log(2));
    ioctl(m_outFD, SNDCTL_DSP_SETFRAGMENT, &desBlockConf);
  }
  
  // Determine how many bytes fit to driver IO buffers:
  ioctl(m_inFD, SNDCTL_DSP_GETISPACE, &m_bufInfo);
  m_inBufSize = (m_bufInfo.fragstotal * m_bufInfo.fragsize) /
    (inChannels * (m_in32Bit ? 4 : 2));
  ioctl(m_outFD, SNDCTL_DSP_GETOSPACE, &m_bufInfo);
  m_maxOutBytes = m_bufInfo.fragstotal * m_bufInfo.fragsize - 
    m_bufInfo.fragsize - m_bufInfo.fragsize / 2;
  m_outBufSize = (m_bufInfo.fragstotal * m_bufInfo.fragsize) /
    (outChannels * (m_out32Bit ? 4 : 2));
  
  return true;
}

bool DI_AudioPort::close()
{
  // Delete names and set empty labels
  if (m_inDevice) free(m_inDevice);
  if (m_outDevice) free(m_outDevice);
  m_inDevice = NULL;
  m_outDevice = NULL;
  getLabel(NULL, true);
  getLabel(NULL, false);

  m_mode = DI_IO_NONE;
  m_inBufSize = 0;
  m_outBufSize = 0;
  m_maxOutBytes = 0;
  
  if ((m_inFD < 0) && (m_outFD < 0)) return false;

  if (m_inFD == m_outFD)
  {
    return (::close(m_inFD) != -1);
  }
  
  if (m_inFD < 0)
  {
    return (::close(m_outFD) != -1);
  }
  else if (m_outFD < 0)
  {
    return (::close(m_inFD) != -1);
  }
  else
  {
    return ((::close(m_inFD) != -1) && (::close(m_outFD) != -1));
  }
}

int DI_AudioPort::readFrames(void *buffer, uint frames)
{
  if(frames > m_tmpBufferFrames) return -1;

  uint samples(frames * m_inChannels);

  int e(::read(m_inFD, m_tmpBuffer, samples * (m_in32Bit ? 4 : 2)));
  if (e == -1) return -1;
  
  if (m_in32Bit)
  {
    switch(m_inFormat)
    {
     case DI_ASF_INT16:
      DI_int2Short((short *)buffer, (int *)m_tmpBuffer, -16, 0, samples);
      break;
     case DI_ASF_INT32:
      memcpy(buffer, m_tmpBuffer, 4*samples);
      break;
     case DI_ASF_FLOAT32:
      DI_int2Float((float *)buffer, (int *)m_tmpBuffer, 1.0 / DI_OSS_ISCALE,
		   0.0, samples);
      break;
     case DI_ASF_FLOAT64:
      DI_int2Double((double *)buffer, (int *)m_tmpBuffer, 1.0 / DI_OSS_ISCALE,
		    0.0, samples);
     default:
      return -1;
    }
    return e/(4*m_inChannels);
  }
  else
  {
    switch(m_inFormat)
    {
     case DI_ASF_INT16:
      memcpy(buffer, m_tmpBuffer, 2 * samples);
      break;
     case DI_ASF_INT32:
      DI_short2Int((int *)buffer, (short *)m_tmpBuffer, 16, 0, samples);
      break;
     case DI_ASF_FLOAT32:
      DI_short2Float((float *)buffer, (short *)m_tmpBuffer,
		     1.0 / DI_OSS_SSCALE, 0.0, samples);
      break;
     case DI_ASF_FLOAT64:
      DI_short2Double((double *)buffer, (short *)m_tmpBuffer,
		      1.0 / DI_OSS_SSCALE, 0.0, samples);
      break;
     default:
      return -1;
    }
    return e/(2*m_inChannels);
  }
}

int DI_AudioPort::writeFrames(void *buffer, uint frames)
{
  if(frames > m_tmpBufferFrames) return -1;

  uint samples(frames * m_outChannels);

  if (m_out32Bit)
  {
    switch(m_outFormat)
    {
     case DI_ASF_INT16:
      DI_short2Int((int *)m_tmpBuffer, (short *)buffer, 16, 0, samples);
      break;
     case DI_ASF_INT32:
      memcpy(m_tmpBuffer, buffer, 4*samples);
      break;
     case DI_ASF_FLOAT32:
      DI_float2Int((int *)m_tmpBuffer, (float *)buffer, DI_OSS_ISCALE,
		   0, samples);
      break;
     case DI_ASF_FLOAT64:
      DI_double2Int((int *)m_tmpBuffer, (double *)buffer, DI_OSS_ISCALE,
		    0, samples);
      break;
     default:
      return -1;
    }
    return ::write(m_outFD, m_tmpBuffer, samples*4) / (4*m_outChannels);
  }
  else
  {
    switch(m_outFormat)
    {
     case DI_ASF_INT16:
      memcpy(m_tmpBuffer, buffer, 2 * samples);
      break;
     case DI_ASF_INT32:
      DI_int2Short((short *)m_tmpBuffer, (int *) buffer, -16, 0, samples);
      break;
     case DI_ASF_FLOAT32:
      DI_float2Short((short *)m_tmpBuffer, (float *) buffer,
		     DI_OSS_SSCALE, 0, samples);
      break;
     case DI_ASF_FLOAT64:
      DI_double2Short((short *)m_tmpBuffer, (double *) buffer,
		      DI_OSS_SSCALE, 0, samples);
      break;
     default:
      return -1;
    }
    return ::write(m_outFD, m_tmpBuffer, samples*2) / (2*m_outChannels);
  }
}

int DI_AudioPort::waitInputFilled(uint frames, int timeOut)
{
  int slept(0);
  timespec sleepTime;
  sleepTime.tv_sec = 0;
  sleepTime.tv_nsec = 10000000;  // 10 ms
  int filled(0);

  do
  {
    nanosleep(&sleepTime, NULL);
    slept += 10;
    filled = inFilledFrames();
  }
  while(((uint)filled<frames) || ((slept < timeOut) && (timeOut>=0)));
  
  return filled;
}

int DI_AudioPort::waitOutputFilled(uint frames, int timeOut)
{
  int slept(0);
  timespec sleepTime;
  sleepTime.tv_sec = 0;
  sleepTime.tv_nsec = 10000000;  // 10 ms
  int filled(0);

  do
  {
    nanosleep(&sleepTime, NULL);
    slept += 10;
    filled = inFilledFrames();
  }
  while(((uint)filled>frames) || ((slept < timeOut) && (timeOut>=0)));
  
  return filled;
}

bool DI_AudioPort::isOK(void) const
{
  return ((((m_mode & DI_IO_INPUT) && (m_inFD != -1)) ||
	   (!(m_mode & DI_IO_INPUT))) &&
	  (((m_mode & DI_IO_OUTPUT) && (m_outFD != -1)) ||
	   (!(m_mode & DI_IO_OUTPUT))));
}

// ---- Private methods -------------------------------------------------------

void DI_AudioPort::getLabel(const char *device, bool input)
{
  char *tempLabel(input ? m_inLabel : m_outLabel);
  
  if (!device)
  {
    tempLabel[0] = '\0';
    return;
  }

  int dev(0), dev_len(strlen(device));
  if ((dev_len >= 1) && isdigit(device[dev_len-1]))
  {
    dev = (int)(device[dev_len-1] - '0');
    if ((dev_len >= 2) && isdigit(device[dev_len-2]))
    {
      dev = dev + 10 * (int)(device[dev_len-2] - '0');
      if ((dev_len >= 3) && isdigit(device[dev_len-3]))
	dev = dev + 100 * (int)(device[dev_len-3] - '0');
    }
  }
  else
  {
    strncpy(tempLabel,device,DI_AUDIOMAXNAMELENGTH);
    return;
  }

  FILE *sndstat(fopen("/dev/sndstat","r"));
  if (!sndstat)
  {
    DI_error(DI_ERR_INTERNAL, "Unable to open '/dev/sndstat'");
    tempLabel[0] = '\0';
    return;
  }

  // Now search for the "Audio devices" part of the /dev/sndstat
  const int MAXLINE(1000);
  char tempLine[MAXLINE];
  int size(MAXLINE);
  while (fgets(tempLine,size,sndstat) &&
	 (strncmp(tempLine,"Audio devices:",14) != 0))
    ;

  // Read the lines that preceed the device we are interested in.
  // This assumes that devices are numbered sequentially from 0 to N-1.
  // Please, tell us, if you now this to be not true.
  for (int l(0); l<=dev; l++)
    fgets(tempLine,size,sndstat);

  int offset(0);
  if (dev < 10)
    offset = 3;
  else if (dev < 100)
    offset = 4;
  else if (dev < 1000)
    offset = 5;
  else
  {
    DI_error(DI_ERR_INTERNAL, "Error processing '/dev/sndstat'");
    tempLabel[0] = '\0';
    fclose(sndstat);
    return;
  }

  strncpy(tempLabel,tempLine+offset,DI_AUDIOMAXNAMELENGTH);
  for (unsigned i(0); i<DI_AUDIOMAXNAMELENGTH; i++)
    if (tempLabel[i] == '\n') tempLabel[i] = '\0';
  tempLabel[DI_AUDIOMAXNAMELENGTH] = '\0';
  fclose(sndstat);
  return;
}
