/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the DIVA IO library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000-2001.
 * Jarmo.Hiipakka@hut.fi Tommi.Ilmonen@hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA IO library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */

//
// DIVA Library. Copyright (C) Jarmo Hiipakka, Tommi Ilmonen, 1999-2001.
//
// $Log: di_audio_port.h,v $
// Revision 1.5  2004/02/23 13:17:21  tilmonen
// ulong -> uint.
//
// Revision 1.4  2002/11/25 09:22:48  tilmonen
// Fake debug option.
//
// Revision 1.3  2002/05/22 12:42:53  tilmonen
// Changes due to port to Windows.
//
// Revision 1.2  2001/07/27 08:23:29  tilmonen
// License text updates (=cosmetic changes).
//
// Revision 1.1.1.1  2001/06/04 09:04:21  tilmonen
// Mustajuuri 4.6.2001, initial CVS cersion
//

#ifndef DI_AUDIO_PORT_H
#define DI_AUDIO_PORT_H

#include <sys/soundcard.h>
#include <di_types.h>

#include "di_config.h"
#include "di_io_defs.h"

#define DI_AUDIO_IS_OSS

/// Unspecified audio library sample rate.
const ulong DI_AUDIOUNSPECIFIEDRATE = 0;

/// Maximum length of audio library names and labels.
const ulong DI_AUDIOMAXNAMELENGTH = 32;

/** Audio signal I/O.

    DI_AudioPort is declared in "di_audio_port.h" and defined in libdiio.

    {\large {\bf Introduction}}

    DI_AudioPort provides an easy, stream-like access to platform audio
    hardware. Theoretically, DI_AudioPort provides complete hardware
    independency, in practice, this remains to be seen as the current
    version supports only SGI and OSS audio libraries.

*/

class DI_AudioPort
{
public:

  /**@name Constructors */
  
  //@{

  /** Create a new audio port.
    
      This constructor creates a new audio port. The port is unconnected
      and is made operational only by using the open method. It is
      recommended that the port is opened only when it is needed, and
      closed when not needed anymore.
  */

  DI_AudioPort();

  ///

  ~DI_AudioPort(void);

  //@}

  /**@name Access */

  //@{

  /// Return the audio input device name.

  const char *inDeviceName(void) const { return (const char *)m_inDevice; }

  /// Return the audio output device name.

  const char *outDeviceName(void) const { return (const char *)m_outDevice; }

  /// Return the audio input device label.

  const char *inLabel(void) const { return (const char *)m_inLabel; }

  /// Return the audio output device label.

  const char *outLabel(void) const { return (const char *)m_outLabel; }

  /// Return the IO mode.

  DI_IoMode mode(void) const { return m_mode; }

  /// Return the input sample format.

  DI_AudioSampleFormat inSampleFormat(void) const { return m_inFormat; }
  
  /// Return the output sample format.

  DI_AudioSampleFormat outSampleFormat(void) const { return m_outFormat; }
  
  /// Return the input audio buffer size.

  uint inBufSize(void) const { return m_inBufSize; }

  /// Return the output audio buffer size.

  uint outBufSize(void) const { return m_outBufSize; }

  /// Return input sample rate.

  uint inSampleRate(void) const { return m_inRate; }

  /// Return output sample rate.

  uint outSampleRate(void) const { return m_outRate; }

  /// Return number of input channels.

  uint inChannels(void) const { return m_inChannels; }

  /// Return number of output channels.

  uint outChannels(void) const { return m_outChannels; }

  /// Return maximum input block size.

  uint inMaxBlock(void) const { return m_inBlock; }
  
  /// Return maximum output block size.

  uint outMaxBlock(void) const { return m_outBlock; }
  
  /// Get number of filled sample frames in the input sample buffer

  int inFilledFrames(void) const;
  
  /// Get number of filled sample frames in the output sample buffer

  int outFilledFrames(void) const;
  
  //@}

  /**@name Operations */
  
  //@{

  /** Open the audio port.

      This method opens an actual audio port to a named audio device.
      The audio port may be opened for recording or playback or for both.
      However, the supported modes may differ on different platforms.
      On Linux platform, the name of the device may only be of format
      "/dev/dsp*", i.e., the human-readable labels returned by inDeviceLabel
      and outDeviceLabel methods, cannot be used when calling open.

      The 'bufferSize' parameter sets the audio queue buffer size
      (in sample frames) that  will be used as a temporary data storage
      during recording or playing.

      The 'maxBlock' parameters sets the maximum block size (in sample frames)
      that can be read or written on a single call to either readFrames or
      writeFrames.

  */

  bool open(/// Audio device.
	    const char *device,
	    /// IO mode: none, input, output, or input&output.
	    DI_IoMode mode, 
	    /// Audio port sample format.
	    DI_AudioSampleFormat userSample, 
	    /// Audio port buffer size (sample frames).
	    uint bufferSize,
	    /// Audio sampling rate.
	    uint samplerate,
	    /// Number of channels.
	    uint channels,
	    /// Maximum block size.
	    uint maxBlock);

  /** Open the audio port.

      This method opens an actual audio port to named audio devices.
      The first device is opened for input, and the second is opened for
      output. The whole port structure is in the input&output state.

  */
  
  bool open(/// Audio input device.
	    const char *inDevice,
	    /// Audio output device.
	    const char *outDevice,
	    /// Audio port sample input format.
	    DI_AudioSampleFormat inFormat, 
	    /// Audio port sample output format.
	    DI_AudioSampleFormat outFormat, 
	    /// Audio port input buffer size (sample frames).
	    uint inBufSize,
	    /// Audio port output buffer size (sample frames).
	    uint outBufSize,
	    /// Input sampling rate.
	    uint inSamplerate,
	    /// Output sampling rate.
	    uint outSamplerate,
	    /// Number of input channels.
	    uint inChannels,
	    /// Number of output channels.
	    uint outChannels,
	    /// Maximum input block size.
	    uint inMaxBlock,
	    /// Maximum output block size.
	    uint outMaxBlock);

  /// Close the port.
  
  bool close(void);

  //@}

  /**@name I/O operations */
  
  //@{

  /** Read sample frames recorded by audio hardware.

      This method reads 'frames' sample frames from audio hardware recording
      device, and stores them into buffer pointed by 'buffer'.

      Please note that one sample may take more than one byte, so
      'buffer' must point to a buffer of capable of holding 'frames'
      * N * M bytes, where N is the size of one sample and M is number of
      channels.
     
      The number of sample frames actually recorded is
      returned, or a negative value if the operation failed.

  */

  int readFrames(void *buffer,uint frames);

  /** Write sample frames to be played by audio hardware.

      This method writes 'frames' sample frames to audio hardware playback
      device, reading them from buffer pointer by 'buffer'. The number
      of sample frames actually played is returned, or a negative value if the
      operation failed. NOTE! This method may change the contents of
      the buffer.
      
      writeFrames() may, and usually does, return before the playback
      sequence is completed, storing the sample data into a temporary
      buffer and leaving the playback subsystem running after write()
      returns. In these cases, the returned value shows how many sample frames
      the the audio subsystem is going to try to play.

  */

  int writeFrames(void *buffer,uint frames);
  
  /** Wait until 'frames' # of frames can be read.
      
      Return value tells how many frames are filled as the call
      returns. Negative values indicate error of some sort.
      Timeout is specified as milliseconds.
  */
  
  int waitInputFilled(uint frames, int timeOut);

  /// Wait until output has only the number of frames specified filled.  

  int waitOutputFilled(uint frames, int timeOut);

  /** Start recording/playback. On IRIX and OSS this method does
      nothing, on ALSA it is compulsory. */
  bool start() { return true; }


  //@}

  /**@name Predicates */
  
  //@{

  ///

  bool isOK(void) const;

  //@}

  //

  // int outputResolution() { return -1; }

  static void setDebug(bool debug) { m_debug = debug; }

private:

  void getLabel(const char *device, bool input);

  char                 *m_inDevice;
  char                 *m_outDevice;
  char                  m_inLabel[DI_AUDIOMAXNAMELENGTH];
  char                  m_outLabel[DI_AUDIOMAXNAMELENGTH];
  DI_IoMode             m_mode;
  DI_AudioSampleFormat  m_inFormat;
  DI_AudioSampleFormat  m_outFormat;
  uint                 m_inChannels;
  uint                 m_outChannels;
  uint                 m_inBufSize;
  uint                 m_outBufSize;
  uint                 m_inRate;
  uint                 m_outRate;
  uint                 m_inBlock;
  uint                 m_outBlock;

  void           *m_tmpBuffer;
  uint            m_tmpBufferFrames;
  
  audio_buf_info  m_bufInfo;
  //  int             m_maxInBytes;
  int             m_maxOutBytes;
  int             m_inFD;
  int             m_outFD;
  bool            m_in32Bit;  // 32 bit (true) or 16 bit (false) input used
  bool            m_out32Bit; // 32 bit (true) or 16 bit (false) output used

  static bool     m_debug;
};

#endif
