/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the DIVA IO library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000-2001.
 * Jarmo.Hiipakka@hut.fi Tommi.Ilmonen@hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA IO library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */

#include "di_midi_io.h"

#include "di_error.h"
// #include "di_time_utils.h"

#define DI_MILLION 1000000

#include <fcntl.h>
#include <sys/soundcard.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <unistd.h>

#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>


SEQ_DEFINEBUF(2048);

const char *DI_defaultMidiDev()
{
  return "/dev/midi";
}


bool DI_MidiIo::m_debug = false;

DI_MidiIo::DI_MidiIo()
  : m_portName(0),
    m_open(false),
    m_status1(0),
    // m_status2(0),
    m_mode(DI_IO_NONE),
    m_midiFd(-1)
{}

DI_MidiIo::~DI_MidiIo()
{
  if(m_midiFd != -1) close();
}

bool DI_MidiIo::open(const char * port,
		     DI_IoMode mode)
{
  const char *fname = "DI_MidiIo::open";

  int flags = 0;

  if(m_midiFd != -1) {
    DI_error(DI_ERR_INTERNAL, fname, "Port already open.");
    return false;
  }
  
  if(port) m_portName = strdup(port);
  else port = DI_defaultMidiDev();

  if(mode == DI_IO_INPUT) flags = O_RDONLY;
  else if(mode == DI_IO_OUTPUT) flags = O_WRONLY;
  else if(mode == DI_IO_INPUT_OUTPUT) flags = O_RDWR;
  else return false;

  m_midiFd = ::open(port, flags, 0);
  
  if(m_midiFd < 0) {
    return false;
  }
  
  if(m_debug) DI_trace(fname, "Device opened successfully");  
  
  FD_ZERO(&m_rfds);
  FD_SET(m_midiFd, &m_rfds);
  
  m_tv.tv_sec = 0;
  m_tv.tv_usec = 0;
  m_mode = mode;
  m_open = true;

  resetTiming();
  return true;
}

bool DI_MidiIo::close()
{
  const char *fname = "DI_MidiIo::close";
  
  if(m_debug) DI_trace(fname);
  
  if(!m_open) DI_error(DI_ERR_INTERNAL, fname);
  m_open = false;

  ::close(m_midiFd);
  m_midiFd = -1;
  
  m_mode = DI_IO_NONE;
  if(m_portName) {
    free(m_portName);
    m_portName = 0;
  }

  return true;
}

void DI_MidiIo::resetTiming()
{
  gettimeofday(& m_origin, 0);
}

bool DI_MidiIo::eventPending()
{
  FD_ZERO(&m_rfds);
  FD_SET(m_midiFd, &m_rfds);

  int retVal = select(m_midiFd + 1, &m_rfds, 0, 0, &m_tv);

  if(retVal) {
    FD_ZERO(&m_rfds);
    FD_SET(m_midiFd, &m_rfds);
    return true;
  }
  return false;
}

bool DI_MidiIo::readEvent(DI_MidiEvent *event)
{
  const char *fname = "DI_MidiIo::readEvent";

  unsigned char b = 0;

  event->clearMessage();
  
  int r = read(m_midiFd, &b, 1);

  if(r != 1) {
    DI_error2(DI_ERR_SYSTEM, "%s # Broken MIDI pipe", fname);
    return false;
  }

  /* Number of bytes to skip from the beginning of the message during
     the second read period. */
  int forget;
  
  /* Fill in the missing header info. All real headers (header = first
     message byte) have the most significant bit set.*/
  if(!(b & 0x80)) {
    
    event->set(0, m_status1);
    event->set(1, b);
    
    forget = 2;
    /* DI_trace2("Scanning run-time status byte (%x %x)", 
       (int) m_status1, (int) b); */
  }
  else {
    /* DI_trace2("No run-time status (%x %x)", 
       (int) m_status1, (int) b); */

    event->set(0, b);
    m_status1 = b;
    forget = 1;
  }
  
  int totalBytes = event->messageBytes();    
  
  if(totalBytes > forget)
    read(m_midiFd, & event->message()[forget], totalBytes - forget); 
  
  // m_status2 = event->b();

  if(event->a() == 0xf0) {    
    if(m_debug) DI_trace2("%s # a = 0x%x", fname, (int) event->a());
    readSysex(event);
  }
  else 
    event->freeSysex();
  
  // Make sure the event is properly formatted.
  event->realNoteOff();

  // Adjust event time.
  gettimeofday(&m_time, 0);

  // long t = DI_timeDiff(&m_origin, &m_time, DI_MIDI_TIME_RESOLUTION);

  long t = (m_time.tv_sec - m_origin.tv_sec) * DI_MIDI_TIME_RESOLUTION + 
    (m_time.tv_usec - m_origin.tv_usec) / 
    (DI_MILLION / DI_MIDI_TIME_RESOLUTION);

  event->setTime((ulong)t);
  
  if(m_debug) { // Debug info...
    DI_trace2("%s # a = 0x%x b = 0x%x c = 0x%x, "
	      "total bytes = %d, forget = %d, m_time.tv_usec = %d",
	      fname, (int) event->a(), (int) event->b(), (int) event->c(), 
	      totalBytes, forget, m_time.tv_usec);
  }
  
  return true;
}

bool DI_MidiIo::writeEvent(DI_MidiEvent *event)
{
  int mainBytes = event->messageBytes();
  int tmp = ::write(m_midiFd, event->message(), mainBytes);

  if(tmp && event->isSysex())
    ::write(m_midiFd, event->sysexData(), event->sysexSize());
    
  return tmp;
}

void DI_MidiIo::readSysex(DI_MidiEvent *event)
{
  if(m_debug)
    DI_trace("DI_MidiIo::readSysex # enter");

  uint buflen = 1024;
  uint messageBytes;
  char buf[buflen]; // Lets hope that nobody has messages longer than this...

  unsigned char lastbyte = 0;

  // Scan the data until we hit "0xf7" = end of sysex
  for(messageBytes = 0; lastbyte != 0xf7 && messageBytes < buflen; 
      messageBytes++) {
    read(m_midiFd, buf + messageBytes, 1); 
    lastbyte = buf[messageBytes];
  } 
  
  event->setSysex(buf, messageBytes);

  if(m_debug) 
    DI_trace2("DI_MidiIo::readSysex # read %d bytes", event->sysexSize());
}

int DI_MidiIo::read(int fd, void *p, int bytes)
{
  unsigned char * up = (unsigned char *) p;
  int count = 0;
  
  while(count < bytes) {
    int r = ::read(fd, up, bytes);
    if(r < 0) return -1;
    count += r;
    up += r;
  }
  if(m_debug) {
    for(int i=0; i < bytes; i++)
      printf("byte %d/%d: %x\n", i + 1, bytes, (int) ((unsigned char *) p)[i]);
    fflush(0);
  }
  
  return bytes;
}
