/* COPYRIGHT
 *
 * This file is part of Mustajuuri.
 *
 * Author: Tommi Ilmonen, 2001
 * Tommi.Ilmonen@hut.fi
 *
 * http://www.tcm.hut.fi/~tilmonen/mustajuuri/
 *
 * Mustajuuri is licensed under GNU Lesser General Public License
 * (LGPL). The LGPL conditions can be found in file "LGPL" that is
 * distributed with this source package. The Mustajuuri license
 * interpretation that further clarifes LGPL can be found in file
 * "MUSTAJUURI_LICENSE" that is distributed with the Mustajuuri source code.
 *    
 */

#include <mj_ladspa_export.h>

#include <mj_debug.h>
#include <mj_dsp_plugin_loader.h>
#include <mj_error.h>
#include <mj_global.h>
#include <mj_kernel.h>
#include <mj_mapper.h>
#include <mj_parameters.h>
#include <mj_to_ladspa_plugin.h>

#include <mj_arraytmpl.C>

#include <qapplication.h> 

#include <stdlib.h>

static bool MJ_ladspaExportDebug = false;

static QApplication *MJ_ladspaQtApp = 0;

/* If Mustajuuri has not been initialized, then the "local" kernel
   will be put here. This is deleted as this shared library is
   closed. */

static MJ_Kernel * MJ_ladspaExportKernel = 0;

MJ_LadspaPlugin::MJ_LadspaPlugin(MJ_ModuleDsp *m, MJ_LadspaExport *le)
  : m_module(m),
    m_lastSampleCount(0),
    m_ladspaSource(le)
{
  if(MJ_ladspaExportDebug) MJ_trace("MJ_LadspaPlugin::MJ_LadspaPlugin", this);

  MJ_Parameters *params = & le->params();

  uint size = params->count();

  m_inMessagesNew.resize(size);
  m_inMessagesNew.setAll(0);
  m_inMessagesOld.resize(size);
  m_inMessagesOld.setAll(0.0);

  m_timeStamp.setHz(m->dspSamplingRate());
  m->dspSetTimer(&m_timeStamp);
}

MJ_LadspaPlugin::~MJ_LadspaPlugin()
{
  if(MJ_ladspaExportDebug) MJ_trace("MJ_LadspaPlugin::~MJ_LadspaPlugin", this);
  if(m_module) delete m_module;
}

void MJ_LadspaPlugin::activate(LADSPA_Handle instance)
{
  const char * fname = "MJ_LadspaPlugin::activate";

  if(MJ_ladspaExportDebug) MJ_trace("MJ_LadspaPlugin::activate", instance);

  MJ_LadspaPlugin *mlp = (MJ_LadspaPlugin *) instance;

  mlp->m_lastSampleCount = 0;
  mlp->m_timeStamp.setStamp(0);

  MJ_ModuleDsp *m = mlp->m_module;

  m->dspClearOutputs();

  // Update parameters:
  uint params = mlp->m_inMessagesNew.size();
  for(uint i=0; i < params; i++) {
    LADSPA_Data newData = *mlp->m_inMessagesNew.getFast(i);
    LADSPA_Data *oldData = & mlp->m_inMessagesOld.getFast(i);

    MJ_trace2("%s # param %d changed to %f", fname, i, newData);
    *oldData = newData;
    
    MJ_Parameter *p = mlp->m_ladspaSource->params().at(i);
    const char *id = p->id();
    m->msgProcessDouble(id, newData);
  }

  mlp->m_timeStamp.setRealTime(true);  

  if(!m->dspPrepare()) {
    MJ_error(MJ_ERR_UNKNOWN, fname, "dspPrepare failed");
    return;
  }
  if(!m->dspPrepareFast()) {
    MJ_error(MJ_ERR_UNKNOWN, fname, "dspPrepareFast failed");
    return;
  }
}

void MJ_LadspaPlugin::run(LADSPA_Handle instance,
			  unsigned long sampleCount)
{
  MJ_LadspaPlugin *mlp = (MJ_LadspaPlugin *) instance;
  
  MJ_ModuleDsp *m = mlp->m_module;

  uint i;

  // Update buffer lengths if necessary:
  if(sampleCount != mlp->m_lastSampleCount) {

    mlp->m_lastSampleCount = sampleCount;

    uint chans = m->dspInChannels();
    
    for(i=0; i < chans; i++) {
      m->dspInSignal(i).samples().setLength(sampleCount);
    }
    chans = m->dspOutChannels();

    for(i=0; i < chans; i++) {
      m->dspOutSignal(i).samples().setLength(sampleCount);
    }
  }

  // Check for parameter changes:
  uint params = mlp->m_inMessagesNew.size();
  // bool changed = false;
  for(i=0; i < params; i++) {
    LADSPA_Data newData = *mlp->m_inMessagesNew.getFast(i);
    LADSPA_Data *oldData = & mlp->m_inMessagesOld.getFast(i);
    
    if(newData != *oldData) {
      //MJ_trace2("MJ_LadspaPlugin::run # param %d changed to %f", i, newData);
      *oldData = newData;

      MJ_Parameter *p = mlp->m_ladspaSource->params().at(i);
      const char *id = p->id();
      m->msgProcessDouble(id, newData);
      /* MJ_trace2("MJ_LadspaPlugin::run # param %d done", i);
	 changed = true;*/
    }
  }
  
  m->dspProcessPeriod();

  mlp->m_timeStamp.incrementStamp(sampleCount);

  // if(changed) MJ_trace("MJ_LadspaPlugin::run exit after change");
}

void MJ_LadspaPlugin::deactivate(LADSPA_Handle instance)
{
  if(MJ_ladspaExportDebug) MJ_trace("MJ_LadspaPlugin::deactivate", instance);

  MJ_LadspaPlugin *mlp = (MJ_LadspaPlugin *) instance;

  MJ_ModuleDsp *m = mlp->m_module;

  m->dspStopFast();
  m->dspStop();

  mlp->m_timeStamp.setRealTime(false);  

  /* Output pointers need to be explicitly dropped. Input pointers are
     forgotten automatically. */
  m->dspDropOutputPtrs();
}

void MJ_LadspaPlugin::cleanup(LADSPA_Handle instance)
{
  const char * fname = "MJ_LadspaPlugin::cleanup";

  if(MJ_ladspaExportDebug) MJ_trace(fname, instance);

  MJ_LadspaPlugin *mlp = (MJ_LadspaPlugin *) instance;

  MJ_ModuleDsp *m = mlp->m_module;

  if(m) {
    // Just to be sure...
    m->dspDropOutputPtrs();
  }

  if(MJ_ladspaExportDebug) MJ_trace(fname, "half-way");  

  delete mlp;

  if(MJ_ladspaExportDebug) MJ_trace(fname, "EXIT");  
}

void MJ_LadspaPlugin::connectPort(LADSPA_Handle instance,
				  unsigned long port,
				  LADSPA_Data * data)
{
  const char * fname = "MJ_LadspaPlugin::connectPort";

  if(MJ_ladspaExportDebug) MJ_trace(fname, (int) port);

  MJ_LadspaPlugin *mlp = (MJ_LadspaPlugin *) instance;
  MJ_ModuleDsp *m = mlp->m_module;
  
  MJ_Array<MJ_LadspaPortMapping> *mappings = 
    & mlp->m_ladspaSource->mappings();

  MJ_LadspaPortMapping pt(mappings->get(port));

  /* MJ_trace2("%s # Type = %d, index = %d module = %p data = %p", 
     fname, pt.type(), pt.index(), m, data);
     m->toCout(); */
  
  switch(pt.type()) {

  case MJ_IOT_SIGNAL_INPUT:
    MJ_trace(fname, "G1");
    m->dspInSignal(pt.index()).samples().adoptPointer(data);
    MJ_trace(fname, "G2");
    break;

  case MJ_IOT_SIGNAL_OUTPUT:
    m->dspOutSignal(pt.index()).samples().adoptPointer(data);
    break;

  case MJ_IOT_MESSAGE_INPUT:
    mlp->m_inMessagesNew.get(pt.index()) = data;
    break;

  case MJ_IOT_MESSAGE_OUTPUT:
    break;

  default:
    MJ_error(MJ_ERR_ARGUMENT, fname, "Bad type");
    break;
  }

  if(MJ_ladspaExportDebug) MJ_trace(fname, "EXIT");
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_LadspaExport::MJ_LadspaExport(MJ_PluginDbeDsp *dspPlugin, 
				 LADSPA_Descriptor *lad, 
				 uint index)
  : m_dspPlugin(dspPlugin),
    m_index(index),
    m_ladspaDescriptor(lad)
{
  m_params.setAutoDelete(true);

  MJ_ModuleDsp *m = newModule();
  constructMappings(m);
  delete m;
}

LADSPA_Handle MJ_LadspaExport::instantiate
(const LADSPA_Descriptor * descriptor,
 unsigned long             sampleRate)
{
  const char * fname = "MJ_LadspaExport::instantiate";

  if(MJ_ladspaExportDebug) MJ_trace(fname, descriptor);

  MJ_LadspaExport *le = (MJ_LadspaExport *) descriptor->ImplementationData;

  MJ_cpf(le, fname);

  MJ_ModuleDsp *m = le->newModule();

  uint i, chans = m->dspOutChannels();

  for(i=0; i < chans; i++) {
    MJ_SignalSink *ss = & m->dspOutSignal(i);
    ss->sampleRate().set((int) sampleRate, 1, 1);
  }

  MJ_LadspaPlugin *lp = new MJ_LadspaPlugin(m, le);

  return lp;
}

LADSPA_Descriptor *MJ_LadspaExport::getLadspaDescriptor
(MJ_PluginDbeDsp *p, uint i)
{
  if(MJ_ladspaExportDebug) MJ_trace("MJ_LadspaExport::getLadspaDescriptor", p);

  MJ_LadspaData *d = p->ladspaDatas()->datas().at(i);

  LADSPA_Descriptor *lad = (LADSPA_Descriptor *) 
    calloc(1, sizeof(LADSPA_Descriptor));

  if(d->name()) {
    lad->Name = strdup(d->name());
  }
  else if(i){
    char buf[256];
    sprintf(buf, "%s (%d)", (const char *) p->menuPosition(), i + 1);
    lad->Name = strdup(buf);
  }
  else
    lad->Name = strdup(p->menuPosition());

  lad->UniqueID = d->uniqueId();
  lad->Label = d->label();
  lad->Maker = p->parameter(MJ_PP_AUTHORS_LATIN1);
  lad->Copyright = p->parameter(MJ_PP_COPYRIGHT);

  lad->instantiate  = MJ_LadspaExport::instantiate;
  lad->connect_port = MJ_LadspaPlugin::connectPort;
  lad->activate = MJ_LadspaPlugin::activate;
  lad->run = MJ_LadspaPlugin::run;
  lad->deactivate = MJ_LadspaPlugin::deactivate;
  lad->cleanup = MJ_LadspaPlugin::cleanup;

  MJ_LadspaExport *le = new MJ_LadspaExport(p, lad, i);
  lad->ImplementationData = le;

  return lad;
}

void MJ_LadspaExport::constructMappings(MJ_ModuleDsp *m)
{  
  if(MJ_ladspaExportDebug) 
    MJ_trace("MJ_LadspaExport::constructMappings", this);

  m_params.clear();
  
  m->msgGetInputs(&m_params);

  uint inChans = m->dspInChannels();
  uint outChans = m->dspOutChannels();
  uint inMsgs = m_params.count();
  uint i, index = 0;

  uint ports = inChans + outChans + inMsgs;

  m_mappings.resize(ports);
  m_portDescriptors.resize(ports);
  m_portNames.resize(ports);
  m_portNamePtrs.resize(ports);
  m_portRangeHints.resize(ports);

  // QCString str;

  LADSPA_PortRangeHint audioPortHint = { 0, -1.0, 1.0 };

  m_portRangeHints.setRange(audioPortHint, 0, inChans + outChans);

  // Signal input
  for(i = 0; i < inChans; i++) {
    m_mappings.get(index).set(i, MJ_IOT_SIGNAL_INPUT);
    m_portDescriptors.get(index) = LADSPA_PORT_AUDIO | LADSPA_PORT_INPUT;
    m_portNames.get(index).sprintf("Audio input %d", i);
    index++;
  }

  // Signal output
  for(i = 0; i < outChans; i++) {
    m_mappings.get(index).set(i, MJ_IOT_SIGNAL_OUTPUT);
    m_portDescriptors.get(index) = LADSPA_PORT_AUDIO | LADSPA_PORT_OUTPUT;
    m_portNames.get(index).sprintf("Audio output %d", i);
    index++;
  }
  // Control input
  for(i = 0; i < inMsgs; i++) {
    m_mappings.get(index).set(i, MJ_IOT_MESSAGE_INPUT);
    m_portDescriptors.get(index) = LADSPA_PORT_CONTROL | LADSPA_PORT_INPUT;
    
    MJ_Parameter * param = m_params.at(i);

    m_portNames.get(index) = param->name();
    m_portRangeHints.get(index) = parameterToHint(param);
    // m_portNames.get(index)
    index++;
  }

  for(i=0; i < ports; i++) {
    m_portNamePtrs.getFast(i) = m_portNames.getFast(i);
  }

  m_ladspaDescriptor->PortCount = ports;
  m_ladspaDescriptor->PortDescriptors = m_portDescriptors.dataConst();
  m_ladspaDescriptor->PortNames = m_portNamePtrs.dataConst();
  m_ladspaDescriptor->PortRangeHints = m_portRangeHints.dataConst();
}

MJ_ModuleDsp *MJ_LadspaExport::newModule()
{
  if(MJ_ladspaExportDebug) MJ_trace("MJ_LadspaExport::newModule", this);

  MJ_ModuleDsp *m = m_dspPlugin->newPlugin();
  MJ_LadspaDatas *ld = m_dspPlugin->ladspaDatas();

  MJ_LadspaData *d = ld->datas().at(m_index);

  d->apply(m);

  return m;
}

LADSPA_PortRangeHint MJ_LadspaExport::parameterToHint(MJ_Parameter *p)
{
  // MJ_trace2("MJ_LadspaExport::parameterToHint # 0x%p", p);

  LADSPA_PortRangeHint r = { 0, 0, 0};

  MJ_ParameterDouble * pd = dynamic_cast<MJ_ParameterDouble *> (p);
  if(pd) {

    r.HintDescriptor = LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE;
    
    if(pd->mapper()) {
      if(dynamic_cast<MJ_MapperLog *> (pd->mapper()))
	r.HintDescriptor = r.HintDescriptor | LADSPA_HINT_LOGARITHMIC;
    }

    r.LowerBound = pd->min();
    r.UpperBound = pd->max();
    return r;
  }

  MJ_ParameterInt * pi = dynamic_cast<MJ_ParameterInt *> (p);
  if(pi) {

    r.HintDescriptor = LADSPA_HINT_INTEGER | LADSPA_HINT_BOUNDED_BELOW |
      LADSPA_HINT_BOUNDED_ABOVE;

    if(pi->mapper()) {
      if(dynamic_cast<MJ_MapperLog *> (pi->mapper()))
	r.HintDescriptor = r.HintDescriptor | LADSPA_HINT_LOGARITHMIC;
    }

    r.LowerBound = pi->min();
    r.UpperBound = pi->max();
    return r;
  }
  MJ_ParameterBool * pb = dynamic_cast<MJ_ParameterBool *> (p);
  if(pb) {
    
    r.HintDescriptor = LADSPA_HINT_TOGGLED;
    
    r.LowerBound = 0.0;
    r.UpperBound = 1.0;

    return r;
  }

  MJ_error(MJ_ERR_ARGUMENT, "MJ_LadspaExport::parameterToHint");

  return r;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


#if 0
typedef QList<LADSPA_Descriptor> MJ_ladspaList;
#else

// This used to fail on Linux (!) But bugs were fixed...

class MJ_ladspaList : public QList<LADSPA_Descriptor>
{
public:
  MJ_ladspaList() {} 
  ~MJ_ladspaList();
};

MJ_ladspaList::~MJ_ladspaList()
{
  const char * fname = "MJ_ladspaList::~MJ_ladspaList";

  if(MJ_ladspaExportDebug) MJ_trace(fname);

  while(count()) {
    LADSPA_Descriptor *lad = first();
    MJ_LadspaExport *le = (MJ_LadspaExport *) lad->ImplementationData;
    
    delete le;
    free((void *) lad->Name);
    free(lad);

    removeFirst();
  }

  if(MJ_ladspaExportDebug && MJ_ladspaExportKernel) 
    MJ_trace(fname, "Deleting local kernel");
  delete MJ_ladspaExportKernel;
  MJ_ladspaExportKernel = 0;

  delete MJ_ladspaQtApp;
  MJ_ladspaQtApp = 0;
}

#endif

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

static MJ_ladspaList MJ_mjLadspaPlugins;

template class QList<LADSPA_Descriptor>;

template class MJ_Array<MJ_LadspaPortMapping>;
template class MJ_Array<LADSPA_PortDescriptor>;
template class MJ_Array<LADSPA_PortRangeHint>;
template class MJ_Array<QCString>;
template class MJ_Array<const char *>;

template class MJ_Array<float *>;

/** Load all Mustajuuri plugins and try to transform them into LADSPA
    plugins. 

    TODO: Clean up (=unload) plugins that are not actually used.
*/

static void MJ_loadMjPluginsAsLadspa()
{
  const char * fname = "mj_ladspa_export.C:MJ_loadMjPluginsAsLadspa";

  MJ_Kernel *mjk;

  /* This is broken with Qt3 - wonder why...

  if(!qApp) {
    if(MJ_ladspaExportDebug) MJ_trace(fname, "New qApp");

    char * foo = "/unknown/app";
    char ** argv = &foo;
    int argc = 1;

    MJ_ladspaQtApp = new QApplication(argc, argv, QApplication::Tty);
    if(MJ_ladspaExportDebug) MJ_trace(fname, "Created qApp");
    }

    if(MJ_ladspaExportDebug) MJ_trace(fname, "Got qApp"); */

  // Get the kernel:

  if(MJ_kernel) {
    MJ_trace(fname, "Mustajuuri is already initialized");
    // return;
    mjk = MJ_kernel;
  }
  else if(MJ_ladspaExportKernel) {
    MJ_error(MJ_ERR_UNKNOWN, 
	     fname, "Mustajuuri plugins already loaded.");
    return;
  }
  else{
    MJ_trace(fname, "Creating local MJ_kernel");
    mjk = new MJ_Kernel;
    MJ_ladspaExportKernel = mjk;
  }

  if(MJ_ladspaExportDebug) MJ_trace(fname, "Got MJ_kernel");

  MJ_PluginDbeList *mjlist = & mjk->dspLoader()->entryList();

  uint mods = mjlist->count();

  if(MJ_ladspaExportDebug) MJ_trace(fname, mods);

  for(uint i=0; i < mods; i++) {
    MJ_PluginDbeDsp *dbe = ( MJ_PluginDbeDsp *) mjlist->at(i);

    if(dbe->ladspaDatas()) {
      if(MJ_ladspaExportDebug) MJ_trace(fname, "Has LADSPA");

      QList<MJ_LadspaData> *lds = & dbe->ladspaDatas()->datas();

      for(uint j=0; j < lds->count(); j++) {
	LADSPA_Descriptor *lde = 
	  MJ_LadspaExport::getLadspaDescriptor(dbe, j);
	MJ_mjLadspaPlugins.append(lde);
      }
    }
  }

  if(MJ_ladspaExportDebug) MJ_trace(fname, MJ_mjLadspaPlugins.count());
}


const LADSPA_Descriptor * ladspa_descriptor(unsigned long index)
{
  const char * debugLevel = getenv("DEBUG_MJLADSPA");

  if(debugLevel) {
    int level = atoi(debugLevel);

    if(level > 1) {
      MJ_Debug::setDebugAll(level);
    }
  }

  if(MJ_Debug::debugAll() || 
     MJ_Debug::debugNeeded("ladspaexport") || debugLevel) {
    
    MJ_ladspaExportDebug = true;    
  }
 

  if(MJ_ladspaExportDebug) 
    MJ_trace("mj_ladspa_export.C:ladspa_descriptor", (int) index);

  if(!index) {
    MJ_loadMjPluginsAsLadspa();
  }

  if(index < MJ_mjLadspaPlugins.count())
    return MJ_mjLadspaPlugins.at(index);
  
  return 0;
}
