/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_rtc_hz.h>
#include <mj_mutex.h>

#include <linux/version.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 0)
#include <linux/rtc.h>
#else 
#include <linux/mc146818rtc.h>
#endif

#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>

MJ_RtcHz * MJ_RtcHz::m_rtc = 0;

static MJ_MutexAuto MJ_rtcMutex;

MJ_RtcHz::MJ_RtcHz()
  : m_fd(-1)
{}

MJ_RtcHz::~MJ_RtcHz()
{
  if(m_fd != -1) close();
}


/** Inits the device (opens for reading). 
 */
bool MJ_RtcHz::open(
		    /// Device name; if this is NULL default device wil be used
		    const char *device)
{
  if(!device) device = "/dev/rtc";

  MJ_rtcMutex.lock();

  if(m_rtc) {
    MJ_rtcMutex.unlock();
    return false;
  }

  bool ok = ((m_fd = ::open(device, O_RDONLY)) != -1);

  if(ok) m_rtc = this;
  else {
    close();
  }
  MJ_rtcMutex.unlock();

  return ok;
}
bool MJ_RtcHz::startPeriodic()
{
  return (ioctl(m_fd, RTC_PIE_ON, 0) != -1);
}

/** Closes the device. */
bool MJ_RtcHz::close()
{
  bool ok = true;
  
  MJ_rtcMutex.lock();

  if(m_fd == -1) {
    MJ_rtcMutex.unlock();
    return false;
  }

  ok = (::close(m_fd) != -1);
  
  m_fd = -1;
  if(m_rtc == this) m_rtc = 0;

  MJ_rtcMutex.unlock();

  return ok;
}

/** Sets the operating frequency. 

    Unfortunately Linux kernel does not permit normal users to use
    frequencies greater than 64 Hz. For real-time clock-keeping we
    usually need more than 1024 :-( As a result you may/will need root
    privileges to set the frequency. Alternatively you may need to
    modify your kernel.
*/

bool MJ_RtcHz::setHz(int hz)
{
  return (ioctl(m_fd, RTC_IRQP_SET, hz) != -1);
}

/** Returns current frequency. Upon error, returns -1. */
int MJ_RtcHz::getHz()
{
  unsigned long hz = 0;

  if(ioctl(m_fd, RTC_IRQP_READ, &hz) == -1)
    return -1;
  return hz;
}

/** Wait for n ticks. Uses "read" to block to the device. 
    
    Ad the process blocks to the device it will effectively sleep.  */
bool MJ_RtcHz::waitTicks(int n)
{
  unsigned long data;

  bool ok = true;

  for(int i=0; i < n; i++) {
    
    if(read(m_fd, &data, sizeof(unsigned long)) == -1) {
      return false;
    }
    if(i) {
      unsigned long ticks = (data >> 8) & 0xffffff;
      if(ticks != 1) {
	i += ticks - 1;  // Update the loop counter.
	ok = false;
      }
    }
  }

  return ok;
}

/** Returns a new real-time clock object. If the device cannot be
    opened or the frequency set to the desired value NULL is
    returned. */

MJ_RtcHz *MJ_RtcHz::newRtc(
			   /// Device name
			   const char *device, 
			   /// Clock frequnecy
			   int Hz)
{
  if(m_rtc) return 0;

  MJ_RtcHz *rtc = new MJ_RtcHz;

  bool rtcOk = rtc->open(device);

  if(!rtcOk) {
    delete rtc;
    return 0;
  }
  rtcOk = rtc->setHz(Hz);
  
  if(!rtcOk || (rtc->getHz() != Hz)) {
    delete rtc;
    return 0;
  }
  return rtc;
}

/* void MJ_RtcHz::setRtc(MJ_RtcHz *r)
{
  
}*/
