/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */


#include <mj_debug.h>
#include <mj_display_type.h>
#include <mj_dsp_plugin.h>
#include <mj_error.h>
#include <mj_fpu.h>
#include <mj_global.h>
#include <mj_iostream.h>
#include <mj_kernel.h>
#include <mj_level_display.h>
#include <mj_menu_test.h>
#include <mj_message_factory.h>
#include <mj_message_multi.h>
#include <mj_message_qio.h>
#include <mj_message_xml.h>
#include <mj_midi_sequence.h>
#include <mj_misc.h>
#include <mj_module_dsp.h>
#include <mj_mutex.h>
#include <mj_pixmap_loader.h>
#include <mj_prefs.h>
#include <mj_root_powers.h>
#include <mj_run_status.h>
#include <mj_thread.h>
#include <mj_version.h>

#include <qapplication.h>
//#include <qfileinfo.h> 

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <sys/mman.h>
#include <unistd.h>
#endif

#ifdef HAVE_PCAPS
#include <sys/capability.h>

#ifdef __OS_LINUX__
#define MJ_USE_LINUX_CAPS 1
#endif // __linux__

#endif // HAVE_PCAPS
/*
msvcrt.lib d3dx.lib LIBCPMT.LIB ddraw.lib winmm.lib dsound.lib kernel32.lib user32.lib gdi32.lib 
winspool.lib comdlg32.lib advapi32.lib shell32.lib ole32.lib oleaut32.lib uuid.lib odbc32.lib odbccp32.lib 
quartz.lib amstrmid.lib dinput.lib strmbase.lib strmiids.lib dxguid.lib dxerr8.lib ogg_static.lib vorbis_static.lib 
vorbisfile_static.lib

msvcrtd.lib MSVCPRTd.LIB  
*/
  //extern void MJ_FilterTest();

enum MJ_CmdLineOption {   
  MJ_CMDL_AUTOSTART, 
  MJ_CMDL_HELP, 
  MJ_CMDL_DEBUG, 
  MJ_CMDL_MINUS,
  MJ_CMDL_NO_PRIORITY,
  MJ_CMDL_SILENCE,
  MJ_CMDL_PATHS, 
  MJ_CMDL_TEST_FIELDS, 
  MJ_CMDL_TEST_FILTER, 
  MJ_CMDL_TEST_IO_DEFS, 
  MJ_CMDL_TEST_LCD_CLOCK, 
  MJ_CMDL_TEST_LEVEL_QG, 
  //   MJ_CMDL_TEST_KNOB, 
  MJ_CMDL_TEST_MENU, 
  MJ_CMDL_TEST_MIDI_FILE, 
  MJ_CMDL_TEST_MUTEX, 
  MJ_CMDL_TEST_MSG_BUFFER_IO, 
  MJ_CMDL_TEST_PLUGIN_LOAD, 
  MJ_CMDL_TEST_ROOT_POWERS, 
  MJ_CMDL_TEST_TIMING, 
  MJ_CMDL_TEST_THREAD, 
  MJ_CMDL_TEST_VERSION, 
  MJ_CMDL_TEST_XML, 
  MJ_CMDL_VERSION, 
  MJ_CMDL_SIZEOF  
};

static const char *LC_cpaCmdLineOptions[MJ_CMDL_SIZEOF] = {
  "--autostart",
  "-h",
  "--debug",
  "-",
  "--no_priority",
  "--silence",
  "--paths",
  "--test_fields",
  "--test_filter",
  "--test_io_defs",
  "--test_lcd_clock",
  "--test_level_qg",
  //   "--test_knob",
  "--test_menu",
  "--test_midi_file",
  "--test_mutex",
  "--test_msg_buffer_io",
  "--test_plugin_load",
  "--test_root_powers",
  "--test_timing",
  "--test_thread",
  "--test_version",
  "--test_xml",
  "--version"
};

void MJ_help(const char *programName, 
	     const char *badOption = 0,
	     bool exit_now = true)
{
  if(badOption) 
    fprintf(stderr, "%s : bad option: \"%s\"\n", programName, badOption);

  fprintf(stderr, "Usage: %s [options]\nOptions:\n", programName);

  fflush(0);
  
  for(int i=0; i < MJ_CMDL_SIZEOF; i++) {
    fprintf(stderr, "%s\n", LC_cpaCmdLineOptions[i]);
  }

  fflush(0);

  if(exit_now) exit(0);
}

/* Get necessary privileges and drop all others.  This function is
   defined here so it would *not* be dynamically linked. */

void MJ_getPrivileges(int argc, char **argv)
{
  const char * fname = "MJ_getPrivileges";

  // MJ_trace(fname, "ENTER");

  bool mlock = true;

  for(int option = 0; option < argc; option++) {
    
    const char *cpOpt = argv[option];
    MJ_CmdLineOption o = 
      (MJ_CmdLineOption) MJ_findStr(cpOpt, 
				    LC_cpaCmdLineOptions, 
				    MJ_CMDL_SIZEOF);
    switch(o) {
      
    case MJ_CMDL_NO_PRIORITY: // Already handled
      mlock = false;
      break;
    default:
      break;
    }
  }

  if(!mlock)
    return;


#ifdef __OS_LINUX__
  MJ_RootPowers::getCapabilities();

  if(!mlockall(MCL_CURRENT | MCL_FUTURE)) {
    // This may succeed on IRIX even without extra privileges

    // We seem to have super-high priorities :-)
    MJ_trace(fname, "MLOCKALL SUCCESSFUL, HAS REALTIME CAPABILITIES, "
	     "THIS IS GOOD, CONGRATULATIONS");
    MJ_prefs.setLatencies(15.0, 20.0);
  }
  else {
    errno = 0;
  }
#endif // __OS_LINUX__

#if defined(HAVE_PCAPS) && defined(__OS_LINUX__)

  cap_t caps = cap_get_proc();
  // capstr = cap_to_text(cap_get_proc(), 0);
  if(caps) {
    MJ_trace(fname, cap_to_text(caps, 0));
    cap_free(& caps);
  }
#elif defined(WIN32)
//TODO?

#else  // HAVE_PCAPS
  
  MJ_trace(fname, "No capabilities");

  uid_t uid = getuid();
  uid_t euid = geteuid();

  if(uid != euid) {
    MJ_trace(fname, "WARNING !!! Mustajuuri is setuid,\
 but not compiled with capabilities - WILL KEEP SETUID !!!");
  }
#endif // HAVE_PCAPS
  
  errno = 0;
  
  // See if we have really powerful capabilities:
  const char * fileName = "/root/mustajuuri.foo";

  QFile file(fileName);
  bool isRoot = file.open(IO_WriteOnly);
  if(isRoot)
    MJ_trace(fname, "Managed to open /root/mustajuuri.foo I AM ROOT !!!");

  // MJ_trace(fname, "EXIT");
}


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
  
int main(int argc, char **argv)
{
  int option, tmp;
  bool autostart = false;

#ifdef MJ_HAVE_FPU_MANAGER
  MJ_Fpu::fast();
#endif
  MJ_MessageQio::initMessageFactory();
  MJ_MessageXml::initMessageXml();
  // MJ_MessageFactory::init();
  
  MJ_getPrivileges(argc, argv);

  MJ_Kernel::setEnv();

  // new char[3333]; // Intentional memory leak

  // qa->setFont(QFont("helvetica"));

  QFileInfo fi;

  MJ_prefs.scanArgs(argc, argv);

  QApplication *qa = new QApplication(argc, argv, !MJ_prefs.noGui());
  
  MJ_DisplayType::init();

  for(option = 1; option < argc; option++) {

    const char *cpOpt = argv[option];
    MJ_CmdLineOption o = 
      (MJ_CmdLineOption) MJ_findStr(cpOpt, 
				    LC_cpaCmdLineOptions, 
				    MJ_CMDL_SIZEOF);
    
    switch(o) {
    case MJ_CMDL_AUTOSTART:
      autostart = true;
      break;

    case MJ_CMDL_HELP:
      MJ_help(argv[0], 0, true);
      break;
      
    case MJ_CMDL_DEBUG:
      if(option + 1 == argc) MJ_Debug::debugType(MJ_DEBUG_ALL);
      else if(argv[option+1][0] == '-') {
	MJ_Debug::debugType(MJ_DEBUG_ALL);
      }
      else {
	tmp = 1;
	
	while (++option < argc && tmp) {
	  
	  if(argv[option][0] == '-') { // If this is a new option, break loop.
	    
	    option -= 2;
	    tmp = 0;
	  }
	  else MJ_Debug::debugName(argv[option]);
	}
      }
      break;

    case MJ_CMDL_NO_PRIORITY: // Already handled 
    case MJ_CMDL_MINUS: // Null argument
      break;

    case MJ_CMDL_SILENCE:
      MJ_silenceErrors = true;
      break;

    case MJ_CMDL_PATHS:
      MJ_trace2("MUSTAJUURI_PATH=%s", getenv("MUSTAJUURI_PATH"));
      MJ_trace2("MUSTAJUURI_PLUGIN_PATH=%s", getenv("MUSTAJUURI_PLUGIN_PATH"));
      break;

    case MJ_CMDL_TEST_FIELDS:
      MJ_fieldTest();
      return 0;

    case MJ_CMDL_TEST_FILTER:
      //MJ_FilterTest();
      return 0;

    case MJ_CMDL_TEST_IO_DEFS:
      MJ_ModuleDsp::testIoDefRw();
      return 0;

    case MJ_CMDL_TEST_LCD_CLOCK:
      MJ_LcdClock::test(qa);
      return 0;

    case MJ_CMDL_TEST_LEVEL_QG:
      MJ_LevelDisplay::test(qa);
      return 0;

      /* case MJ_CMDL_TEST_KNOB:
      MJ_testKnob(qa);
      return 0; */

    case MJ_CMDL_TEST_MENU:
      MJ_MenuTest::test(qa);
      return 0;

    case MJ_CMDL_TEST_MIDI_FILE:
      MJ_MidiSequence::test();
      return 0;

    case MJ_CMDL_TEST_MUTEX:
      MJ_Mutex::test();
      return 0;

    case MJ_CMDL_TEST_MSG_BUFFER_IO:
      MJ_MessageQio::testBufferIO();
      return 0;

    case MJ_CMDL_TEST_PLUGIN_LOAD:
      MJ_PluginDbeDsp::test(qa, argc, argv);
      return 0;

    case MJ_CMDL_TEST_ROOT_POWERS:
      MJ_RootPowers::test();
      return 0;

    case MJ_CMDL_TEST_TIMING:
      MJ_RootPowers::testClock();
      return 0;

    case MJ_CMDL_TEST_THREAD:
      MJ_Thread::test();
      return 0;

    case MJ_CMDL_TEST_VERSION:
      MJ_Version::test();
      return 0;

    case MJ_CMDL_TEST_XML:
      MJ_MessageXml::test();
      return 0;

    case MJ_CMDL_VERSION:
      MJ_cout() << MJ_describeVersion() << "\n";
      return 0;

    default:

      // Check if it appears to be a file name.
      fi.setFile(cpOpt);
      if(fi.exists())
	break;
      
      MJ_help(argv[0], cpOpt, true);
      break;
    }
  }
 
  MJ_trace("main.C:main # Starting Mustajuuri kernel");

  if(!MJ_prefs.noGui()) {
    MJ_KernelGui::exec(qa, argc, argv, autostart);
    
    MJ_PixmapLoader::clear();
  }
  else {
    MJ_Kernel::exec(qa, argc, argv, autostart);
  }

#ifdef MJ_MEM_DEBUG
  MJ_memDebugReport();
#endif

  delete qa;
  qa = 0;

  return 0;
}

