/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

/* Contains bug fixes from Eric Klein *at* ait.nrl.navy.mil. */

#include <mj_app_action.h>
#include <mj_app_actions.h>

#include <mj_debug.h>
#include <mj_error.h>
#include <mj_module_dsp.h>

#include <typeinfo>

#ifdef __OS_LINUX__
#include <mj_realtime_scheduling.h>
#endif

bool MJ_AppAction::isIdentical(MJ_AppAction *)
{
  return false;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_AppActions *MJ_AppActions::m_act = 0;

MJ_AppActions::MJ_AppActions()
  : m_currentBuffer(0)
{
  m_actionBuffers[0].setAutoDelete(true);
  m_actionBuffers[1].setAutoDelete(true);
  m_actions.setAutoDelete(true);
  m_act = this;
}

MJ_AppActions::~MJ_AppActions()
{
  if(m_act == this) m_act = 0;
}

/** Adds an action to be executed. This method does not block for a
    long time. This method can be called from any thread.*/

void MJ_AppActions::addAction(MJ_AppAction *a)
{
  m_mutex.lock();
  m_actionBuffers[m_currentBuffer].append(a);
  m_mutex.unlock();
}
  
/** Executes the actions that were added. Before an action is
    performed it is checked that it does not equal any other action
    that is currently waiting to be executed.

    This method must be called from the main application thread.
*/

void MJ_AppActions::operate()
{
  const char * fname = "MJ_AppActions::operate";
  uint i, j;

#ifdef __OS_LINUX__
  // Grab super-high priority so the DSP thread will not block here.
  MJ_realTimeScheduling(MJ_RTP_AUDIO);
#endif

  m_mutex.lock();
  uint operationsBuffer = m_currentBuffer;
  m_currentBuffer = m_currentBuffer ? 0 : 1;
  m_mutex.unlock();

#ifdef __OS_LINUX__
  MJ_realTimeScheduling(MJ_RTP_NOPRIORITY);
#endif


  // MJ_trace2("%s # %d", fname, m_actions.count());
  
  // Try to merge the "new" actions into the list of pending actions

  while(m_actionBuffers[operationsBuffer].count()) {
    MJ_AppAction *a = m_actionBuffers[operationsBuffer].take(0);
    
    bool found = false;

    for(j = 0; j < m_actions.count() && !found; j++) {

      if(a->isIdentical(m_actions.at(j))) {
	delete a;
	found = true;
      }
    }
    
    if(!found) m_actions.append(a);
  }

  // Do the actions:

  for(i = 0; i < m_actions.count(); ) {

    MJ_AppAction *a = m_actions.at(i);

    a->operate();

    if(a->isDone()) {
      MJ_trace3(MJ_Debug::debugAll(), 
		"%s # Remove %s %d %d", fname, typeid(*a).name(), 
	 m_actions.count(), i);
      m_actions.remove(i);
    }
    else
      i++;
  }
}

/// Stops the DSP processing
void MJ_AppActions::stopDsp()
{
  if(!m_act) return;
  m_act->m_mutex.lock();
  m_act->m_actionBuffers[m_act->m_currentBuffer].append
    (new MJ_AppActionStopDsp);
  m_act->m_mutex.unlock();
}

/// Starts the DSP processing
void MJ_AppActions::startDsp(MJ_ModuleDsp *m)
{
  if(!m_act) return;
  m_act->m_mutex.lock();
  m_act->m_actionBuffers[m_act->m_currentBuffer].append
    (new MJ_AppActionStartDsp(m));
  m_act->m_mutex.unlock();
}

/// Informs the UI that a module was modified.
void MJ_AppActions::moduleModified(MJ_ModuleDsp *module, const char *parameter)
{
  if(!m_act) return;
  m_act->m_mutex.lock();
  m_act->m_actionBuffers[m_act->m_currentBuffer].append
    (new MJ_AppActionModuleModified(module, parameter));
  m_act->m_mutex.unlock();
}

/// Informs the UI that the IO setup of a module was modified.
void MJ_AppActions::moduleIOModified(MJ_ModuleDsp *module, uint flags)
{
  if(!m_act) return;
  m_act->m_mutex.lock();
  m_act->m_actionBuffers[m_act->m_currentBuffer].append
    (new MJ_AppActionModuleModified(module, flags));
  m_act->m_mutex.unlock();
}

void MJ_AppActions::reportError(const char *str)
{
  if(!m_act) return;
  m_act->m_mutex.lock();
  m_act->m_actionBuffers[m_act->m_currentBuffer].append
    (new MJ_AppActionErrorStr(str));
  m_act->m_mutex.unlock();
}

void MJ_AppActions::newAction(MJ_AppAction *a)
{
  if(!m_act) {
    delete a;
    return;
  }
  m_act->addAction(a);
  
}

template class QList<MJ_AppAction>;
