/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_audio_conf.h>

#include <mj_error.h>
#include <mj_global.h>
#include <mj_iostream.h>
#include <mj_misc.h>

#include <stdlib.h>
#include <qtextstream.h> 

const char * MJ_AudioDeviceTypeKeys[MJ_ADT_SIZEOF] = {
  "Input",
  "Output"
};

const char * MJ_AudioDeviceTypeKeys2[MJ_ADT_SIZEOF] = {
  "Input device",
  "Output device"
};

const char * MJ_AudioEnvVarName = "MJ_AUDIO_CONF";

const char *MJ_AudioConf::m_identifier = "MJ_AUDIO_CONFIG";

MJ_AudioConf::MJ_AudioConf()
{
  m_devices.setAutoDelete(true);
}

MJ_AudioConf::~MJ_AudioConf()
{}

/** Adds a new device to the device list. 

    @return If the device already exists false is returned. Else true
    is returned.*/

bool MJ_AudioConf::addDevice(MJ_AudioDeviceType type, int number)
{
  if(findDevice(type, number) >= 0) return false;
  
  MJ_2Int *tmp = new MJ_2Int;
  tmp->m_v[0] = (int) type;
  tmp->m_v[1] = number;
  m_devices.append(tmp);

  MJ_trace2("MJ_AudioConf::addDevice %s %d",
	    MJ_AudioDeviceTypeKeys[tmp->m_v[0]], tmp->m_v[1]);

  return true;
}

/** Returns a device from the list. If the device is not found false
    is returned. */

bool MJ_AudioConf::removeDevice(MJ_AudioDeviceType type, int number)
{
  int index;
  if((index = findDevice(type, number)) == -1) return false;
  m_devices.remove(index);
  return true;
}

/** Seeks a device in the device list of this object (not the global
    list). 

    @return If the device is not found, -1 is returned. */
int MJ_AudioConf::findDevice(MJ_AudioDeviceType type, int number)
{
  uint count = m_devices.count();
  MJ_2Int tmp;
  tmp.m_v[0] = (int) type;
  tmp.m_v[1] = number;
  
  for(uint i = 0; i < count; i++) {
    if(MJ_equals(tmp, * m_devices.at(i) ))
      return (int) i;
  }
  return -1;
}

/** Returns the number of input channels in this setup. */
int MJ_AudioConf::inputChannels()
{
  uint count = m_devices.count();
  int sum = 0;

  for(uint i = 0; i < count; i++) {
    const MJ_2Int *dev = m_devices.at(i);
    if(((MJ_AudioDeviceType) dev->m_v[0]) == MJ_ADT_INPUT)
      sum += deviceChannels((MJ_AudioDeviceType) dev->m_v[0], dev->m_v[1]);
  }
  return sum;
}


/** Returns the number of output channels in this setup. */
int  MJ_AudioConf::outputChannels()
{
  uint count = m_devices.count();
  int sum = 0;

  for(uint i = 0; i < count; i++) {
    const MJ_2Int *dev = m_devices.at(i);
    if(((MJ_AudioDeviceType) dev->m_v[0]) == MJ_ADT_OUTPUT)
      sum += deviceChannels((MJ_AudioDeviceType) dev->m_v[0], dev->m_v[1]);
  }
  return sum;
}

/** Returns name of a particular audio device. */
QString MJ_AudioConf::deviceName(uint i)
{
  MJ_2Int *mi = m_devices.at(i);
  return audioDevice((MJ_AudioDeviceType) mi->m_v[0], mi->m_v[1]);
}

/** Write decsription of the audio configuration to a stream. */
bool MJ_AudioConf::write(QTextStream *s)
{
  uint siz = m_devices.count();
  
  (*s) << m_identifier << " " << siz << "\n";
  
  for(uint i=0; i < siz; i++) {
    const MJ_2Int &mi = *m_devices.at(i);

    if((uint) mi.m_v[0] >= MJ_ADT_SIZEOF) {
      MJ_error(MJ_ERR_INTERNAL, "MJ_AudioConf::write");
      return false;
    }

    (*s) << MJ_AudioDeviceTypeKeys[mi.m_v[0]] << "\t" << mi.m_v[1] << "\n";
  }
  return true;
}

/** Creates stereo I/O configuration. If there aren't enough devices
    this method creates the device it can create. 

    Possible configurations after this method has been called

    <OL> 

    <LI>Stereo (or greater) input and output are activated
    <LI>Stereo (or greater) input is activated
    <LI>Stereo (or greater) output is activated
    <LI>Empty configuration

    </OL>

    In the first case this method returns true. In all other cases
    false is returned.

*/

bool MJ_AudioConf::defaultStereoIo()
{
  m_devices.clear();

  bool ok = false;

  int i, indevs = audioDevices(MJ_ADT_INPUT);

  for(i = 0; i < indevs; i++) {
    if(deviceChannels(MJ_ADT_INPUT, i) >= 2) {
      addDevice(MJ_ADT_INPUT, i);
      ok = true;
      break;
    }
  }

  if(!ok) return false;
     
  int outdevs = audioDevices(MJ_ADT_OUTPUT);

  for(i = 0; i < outdevs; i++) {    
    if(deviceChannels(MJ_ADT_OUTPUT, i) >= 2) {
      addDevice(MJ_ADT_OUTPUT, i);
      break;
    }
  }
       
  return ok;
}

/** This method works very much like defaultStereoIo except that it
    tries to get more than 2 channels. */

bool MJ_AudioConf::defaultMultiChannelIo()
{
  m_devices.clear();

  bool ok = false;
  
  int i, indevs = audioDevices(MJ_ADT_INPUT);

  for(i = 0; i < indevs; i++) {
    if(deviceChannels(MJ_ADT_INPUT, i) > 2) {
      addDevice(MJ_ADT_INPUT, i);
      ok = true;
      break;
    }
  }

  if(!ok) return false;
     
  int outdevs = audioDevices(MJ_ADT_OUTPUT);

  for(i = 0; i < outdevs; i++) {
    if(deviceChannels(MJ_ADT_OUTPUT, i) > 2) {
      addDevice(MJ_ADT_OUTPUT, i);
      break;
    }
  }
    
  return ok;
}

/** Read decsription of the audio configuration from a stream. */
bool MJ_AudioConf::read(QTextStream *s)
{
  const char * fname = "MJ_AudioConf::read";

  int siz, t1, t2;
  bool ok = true;
  QCString str;
  (*s) >> str;

  if(str != m_identifier) {
    MJ_error2(MJ_ERR_DATA_CORRUPTED, "%s # Got \"%s\" instead of \"%s\"",
	      fname, str.data(), m_identifier);
    return false;
  }

  (*s) >> str;
  siz = str.toInt(&ok);

  if(!ok || siz > 1000) {
    MJ_error(MJ_ERR_DATA_CORRUPTED, fname, 2);
    return false;
  }
  
  m_devices.clear();

  for(uint i=0; i < (uint) siz; i++) {

    (*s) >> str >> t2;
    
    if((t1 = MJ_findStr(str, MJ_AudioDeviceTypeKeys, MJ_ADT_SIZEOF)) 
       == -1) {
      MJ_error(MJ_ERR_DATA_CORRUPTED, fname, str);
      return false;
    }
    MJ_2Int *mi = new MJ_2Int();
    mi->m_v[0] = t1;
    mi->m_v[1] = t2;
    m_devices.append(mi);
  }
  // Read the end-of-line
  s->readLine();
  return true;
}

/** Returns the device description of the queried audio device.

    The device name format depends on your system. 

    On Linux/OSS device names are names of the devices in /dev/dsp.

    On IRIX the names are typically "Stereo input" etc.
*/
QString MJ_AudioConf::audioDeviceDescription(MJ_AudioDeviceType type,
					     int number)
{
  QString tmp, tmp2, envDef(getenv(MJ_AudioEnvVarName));

 int fields = MJ_fields(envDef, "|");

  int hits = 0;

  for(int i=0; i < fields; i++) {
    tmp = MJ_field(envDef, i, "|");
    
    if(tmp.find(MJ_AudioDeviceTypeKeys[type], 0) == 0) {
      if(hits == number) {

	return tmp;
      }
      hits++;
    }
  }
    
  return tmp2;
}

/** Returns the audio device name of the audio device descriptor. */

QString MJ_AudioConf::audioDevice(const QString &descriptor)
{
  if(MJ_fields(descriptor, "=") != 3) {
    MJ_error(MJ_ERR_ENVIRONMENT, 
	     "MJ_AudioConf::audioDevice, malformed device information",
	     (const char *) descriptor);
  }
  return MJ_field(descriptor, 2, "=");
}


QString MJ_AudioConf::audioDevice(MJ_AudioDeviceType type,
				  int number)
{
  return audioDevice(audioDeviceDescription(type, number));
}


/** Returns the number of audio devices of the desired type.*/

int MJ_AudioConf::audioDevices(MJ_AudioDeviceType type)
{
  QString envDef(getenv(MJ_AudioEnvVarName));
  
  int n = 0, start = 0;

  while((start = envDef.find(MJ_AudioDeviceTypeKeys[type],
			     start)) >= 0) {
    n++;
    start++;
  }
  return n;
}

/** Returns the number of channels in given audio device type. */
int MJ_AudioConf::deviceChannels(MJ_AudioDeviceType type, uint index)
{
  QString descriptor(audioDeviceDescription(type, index));

  QString number(MJ_field(descriptor, 1, "="));

  bool ok = true;
  
  int channels = number.toInt(&ok);

  if(!ok) 
    MJ_error(MJ_ERR_ENVIRONMENT, 
	     "MJ_AudioConf::deviceChannels, malformed device information",
	     (const char *) descriptor);

  return channels;
}

