/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_base.h>
#include <mj_error.h>
#include <mj_failure_diagnose.h>

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>


/**  Returns parent object (if any). The default implementation
     returns NULL. */
MJ_Base* MJ_Base::mjParent()
{ return 0; }

/// Writes a string or two to the trace stream.
void MJ_Base::trace(const char *fname, const char *c2) const
{
  MJ_FailureDiagnose *d = new MJ_FailureDiagnose;
  d->failure(this, MJ_ERR_NONE, MJ_ERRL_TRACE);
  d->addMessage(fname);
  if(c2) d->addMessage(c2);

  failure(d);
}

/// Writes a string and an int to the trace stream
void MJ_Base::trace(const char *fname, int arg) const
{
  MJ_FailureDiagnose *d = new MJ_FailureDiagnose;
  d->failure(this, MJ_ERR_NONE, MJ_ERRL_TRACE);
  
  QString str(fname);
  QString tmp;
  tmp.setNum(arg);
  str += " " + tmp;
  d->addMessage(str);
  failure(d);
}

/// Writes a string and a pointer to the trace stream.
void MJ_Base::trace(const char *fname, void *v) const
{
  char *tmp = (char *) malloc(strlen(fname) + 64);
  
  sprintf(tmp, "%s # %p", fname, v);
  trace(tmp);
  free(tmp);
}

/// Writes a string and two ints to the trace stream
void MJ_Base::trace(const char *fname, int arg, int arg2) const
{
  MJ_FailureDiagnose *d = new MJ_FailureDiagnose;
  d->failure(this, MJ_ERR_NONE, MJ_ERRL_TRACE);
  
  QString str(fname);
  QString tmp;
  tmp.setNum(arg);
  str += " " + tmp;
  tmp.setNum(arg2);
  str += " " + tmp;
  d->addMessage(str);
  failure(d);
}

/** Trace with format string conversion. The arguments of this
    function work like the arguments to printf-family of
    functions. 

    This function does not check for argument sanity. A fixed-length
    buffer is used to store the string. The buffer is fairly large
    (8192 bytes), so you should seldom run out of memory. If in doubt,
    then do not use this method.    
 */
void MJ_Base::trace2(const char *msg, ...) const
{
  char buf[8196];
  va_list args;
  va_start(args, msg);
  vsprintf(buf, msg, args);
  trace(buf);
  va_end(args);
}

/** This is like MJ_Base::trace2, but the boolean is used to check if
    the message should be printed or not.*/
void MJ_Base::trace3(bool show, const char *msg, ...) const
{
  if(!show) return;

  char buf[8196];
  va_list args;
  va_start(args, msg);
  vsprintf(buf, msg, args);
  trace(buf);
  va_end(args);
}

/// Reports a string or two to the error stream.
void MJ_Base::error(MJ_ErrorCode errorCode, const char *s, const char *s2) const
{
  MJ_FailureDiagnose *d = new MJ_FailureDiagnose;
  d->failure(this, errorCode);
  if(s) d->addMessage(s);
  if(s2) d->addMessage(s2);
  failure(d);
}

/// Reports a string and int to the error stream.
void MJ_Base::error(MJ_ErrorCode errorCode, const char *s, int arg) const
{
  MJ_FailureDiagnose *d = new MJ_FailureDiagnose;
  d->failure(this, errorCode, MJ_ERRL_ERROR);
  
  QString str(s);
  QString tmp;
  tmp.setNum(arg);
  str += " " + tmp;
  d->addMessage(str);
  failure(d);
}

/** Error messaging with format string conversion. Works like the
    MJ_Base::trace2 method. */

void MJ_Base::error2(MJ_ErrorCode errorCode, const char *msg, ...) const
{
  char buf[8196];
  va_list args;
  va_start(args, msg);
  vsprintf(buf, msg, args);
  error(errorCode, buf);
  va_end(args);
}

/** Error messaging with format string conversion. Works like the
    MJ_Base::error2 method. */

void MJ_Base::error3
(bool show, MJ_ErrorCode errorCode, const char *msg, ...) const
{
  if(!show) return;

  char buf[8196];
  va_list args;
  va_start(args, msg);
  vsprintf(buf, msg, args);
  error(errorCode, buf);
  va_end(args);
}

/// Takes the failure somewhere and DELETES it.
void MJ_Base::failure(MJ_FailureDiagnose *e) const
{
  const MJ_Base* parent = (((MJ_Base*)this) ->mjParent());

  if(parent) parent->failure(e);
  else {
    e->toUser();
    delete e;
  }
}
