/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_dsp_options.h>
#include <mj_error.h>

#include <di_types.h>
#include <stdio.h>

/// Keywords for MJ_DspOption
const char *MJ_DspOptionWords[MJ_DSP_SIZEOF] = {
  "DSP_EQUAL_SAMPLE_RATE_IO",
  "DSP_MAX_SAMPLE_RATE_IN",
  "DSP_MAX_SAMPLE_RATE_OUT",
  "DSP_MIN_SAMPLE_RATE_IN",
  "DSP_MIN_SAMPLE_RATE_OUT",
  "DSP_EQUAL_CHANNEL_IO",
  "DSP_MAX_CHANNELS_IN",
  "DSP_MAX_CHANNELS_OUT",
  "DSP_MIN_CHANNELS_IN",
  "DSP_MIN_CHANNELS_OUT"
};

/// Reaable keywords for MJ_DspOption
const char *MJ_DspOptionWords2[MJ_DSP_SIZEOF] = {
  "Equal sampling rates",
  "Maximum input sampling rate",
  "Maximum output sampling rate",
  "Minimum input sampling rate",
  "Minimum output sampling rate",

  "Equal number of in and out chans",
  "Maximum input channels",
  "Maximum output channels",
  "Minimum input channels",
  "Minimum output channels",
};

MJ_DspOptions::MJ_DspOptions()
{
  undefineAll();
}

MJ_DspOptions::~MJ_DspOptions()
{}

bool MJ_DspOptions::setValue(MJ_DspOption o, int v)
{
  if(!checkArg(o)) return false;
  m_dataArray[o] = v;
  m_defined[o] = true;
  return true;
}

void MJ_DspOptions::undefine(MJ_DspOption o)
{
  if(!checkArg(o)) return;
  m_dataArray[o] = 0;
  m_defined[o] = false;
  return;
}

/** Value of a parameter. */
int MJ_DspOptions::value(MJ_DspOption o) const
{
  if(!checkArg(o)) return 0;
  return m_dataArray[o];
}

/** Tells if a parameter is defined. */
bool MJ_DspOptions::isDefined(MJ_DspOption o) const
{
  if(!checkArg(o)) return false;
  return m_defined[o];
}

/** Undefine all parameters.*/
void MJ_DspOptions::undefineAll()
{
  for(int i=0; i < MJ_DSP_SIZEOF; i++) {
    m_dataArray[i] = 0;
    m_defined[i] = false;
  }
}

/// Limits the unmber of input and output channels to this one value.
void MJ_DspOptions::defaultChans(int chans)
{
  setValue(MJ_DSP_MAX_CHANNELS_IN, chans);
  setValue(MJ_DSP_MAX_CHANNELS_OUT, chans);
  setValue(MJ_DSP_MIN_CHANNELS_IN, chans);
  setValue(MJ_DSP_MIN_CHANNELS_OUT, chans);
}


/// Forces this one combination of input/output hannel pair.
void MJ_DspOptions::defaultChans2(int chansIn, int chansOut)
{
  setValue(MJ_DSP_MAX_CHANNELS_IN, chansIn);
  setValue(MJ_DSP_MAX_CHANNELS_OUT, chansOut);
  setValue(MJ_DSP_MIN_CHANNELS_IN, chansIn);
  setValue(MJ_DSP_MIN_CHANNELS_OUT, chansOut);
}

/** Set the IO channel range. If you pass -1 as any argument the value
    for that parameter will be ignored. */
void MJ_DspOptions::setChans(int minIn, int maxIn, int minOut, int maxOut)
{
  setInChans(minIn, maxIn);
  setOutChans(minOut, maxOut);
}

/** Set the input channel range. If you pass -1 as any argument the value
    for that parameter will be ignored.*/
void MJ_DspOptions::setInChans(int minIn, int maxIn)
{
  if(minIn == -1)
    undefine(MJ_DSP_MIN_CHANNELS_IN);
  else
    setValue(MJ_DSP_MIN_CHANNELS_IN, minIn);

  if(maxIn == -1)
    undefine(MJ_DSP_MAX_CHANNELS_IN);
  else
    setValue(MJ_DSP_MAX_CHANNELS_IN, maxIn);
}

/** Set the output channel range. If you pass -1 as any argument the value
    for that parameter will be ignored.*/
void MJ_DspOptions::setOutChans(int minOut, int maxOut)
{
  if(minOut == -1)
    undefine(MJ_DSP_MIN_CHANNELS_OUT);
  else
    setValue(MJ_DSP_MIN_CHANNELS_OUT, minOut);

  if(maxOut == -1)
    undefine(MJ_DSP_MAX_CHANNELS_OUT);
  else
    setValue(MJ_DSP_MAX_CHANNELS_OUT, maxOut);
}

bool MJ_DspOptions::checkArg(MJ_DspOption o) const
{
  bool r = ((uint) o < (uint) MJ_DSP_SIZEOF);
  if(!r) MJ_error(MJ_ERR_ARGUMENT, "MJ_DspOptions::checkArg");
  return r;
}

bool MJ_DspOptions::acceptIoChans(int in, int out)
{
  bool r = true;

  if(in != newInChans(in)) r = false;
  if(out != newOutChans(out)) r = false;

  if(m_defined[MJ_DSP_EQUAL_CHANNEL_IO] && 
     m_dataArray[MJ_DSP_EQUAL_CHANNEL_IO]) {
    if(in != out) r = false;
  }

  return r;
}

bool MJ_DspOptions::acceptInputChans(int in)
{
  bool r = true;

  if(exceeds(MJ_DSP_MAX_CHANNELS_IN, in)) r = false;
  if(below(MJ_DSP_MIN_CHANNELS_IN, in)) r = false;

  return r;
}

int MJ_DspOptions::newInChans(int propose)
{
  int val = propose;

  if((m_defined[MJ_DSP_MIN_CHANNELS_IN]) && 
     (val < m_dataArray[MJ_DSP_MIN_CHANNELS_IN]))
    val = m_dataArray[MJ_DSP_MIN_CHANNELS_IN];

  if((m_defined[MJ_DSP_MAX_CHANNELS_IN]) && 
     (val > m_dataArray[MJ_DSP_MAX_CHANNELS_IN]))
    val = m_dataArray[MJ_DSP_MAX_CHANNELS_IN];

  return val;
}

int MJ_DspOptions::newOutChans(int propose)
{
  // const char *fname = "MJ_DspOptions::newOutChans";

  int val = propose;

  if((m_defined[MJ_DSP_MIN_CHANNELS_OUT]) && 
     (val < m_dataArray[MJ_DSP_MIN_CHANNELS_OUT])) {
    val = m_dataArray[MJ_DSP_MIN_CHANNELS_OUT];
  }

  if((m_defined[MJ_DSP_MAX_CHANNELS_OUT]) && 
     (val > m_dataArray[MJ_DSP_MAX_CHANNELS_OUT]))
    val = m_dataArray[MJ_DSP_MAX_CHANNELS_OUT];

  /* if(val != propose) {
     char buf[128];
     sprintf(buf, "%s %d -> %d", fname, propose, val);
     MJ_trace(buf);
     }*/
  
  return val;
}

bool MJ_DspOptions::acceptIoSr(int in, int out)
{
  bool r = true;

  if(in != newInChans(in)) r = false;
  if(out != newOutChans(out)) r = false;
  
  if(m_defined[MJ_DSP_EQUAL_SAMPLE_RATE_IO] && 
     m_dataArray[MJ_DSP_EQUAL_SAMPLE_RATE_IO]) {
    if(in != out) r = false;
  }

  return r;
}

bool MJ_DspOptions::exceeds(MJ_DspOption o, int v) const
{
  if(!checkArg(o)) return false;

  if(!m_defined[o]) return true;
  return m_dataArray[o] < v;
}

bool MJ_DspOptions::below(MJ_DspOption o, int v)
{
  if(!checkArg(o)) return false;
  
  if(!m_defined[o]) return true;
  return m_dataArray[o] > v;
}
