/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_dsp_plugin.h>

#include <mj_dsp_plugin_loader.h>
#include <mj_error.h>
#include <mj_iostream.h>
#include <mj_module_dsp.h>
#include <mj_module_dsp_gui_default.h>
#include <mj_plugin_db_tree.h>
#include <mj_plugin_menu.h>
#include <mj_so_ref_count.h>
#include <mj_to_ladspa_plugin.h>

#include <qapplication.h>

#include <dlfcn.h>

const char * MJ_DspPluginFunctionalityKeys[] = {
  "DSP_FUNCTION_AUDIO_IN",
  "DSP_FUNCTION_AUDIO_OUT",
  "DSP_FUNCTION_FILE_IN",
  "DSP_FUNCTION_FILE_OUT",
  "DSP_FUNCTION_MIDI_IN",
  "DSP_FUNCTION_MIDI_OUT"
};

const char * MJ_DspPluginMetaFunctionalityKeys[] = {
  "DSP_META_FUNCTION_STAND_ALONE",
  "DSP_META_FUNCTION_CHAINABLE"
};

/** Holds internal data of MJ_PluginDbeDsp. This class is created to
    enable us to expand the MJ_PluginDbeDsp with new members and
    methods without breaking binary compatibility with previous
    releases.

    The members are seldom needed. This member object is not created
    unless it is needed.
*/

class MJ_PluginDbeDsp::__Internal
{
public:

  __Internal() : m_worksAsTopLevel(false),
		 m_mayConflict(false),
		 m_ladspaDatas(0),
		 m_translator(0) {}
    
  ~__Internal() 
  { 
    delete m_ladspaDatas; 
    if(qApp) qApp->removeTranslator(m_translator);
    delete m_translator;
  }
  
  bool m_worksAsTopLevel;
  bool m_mayConflict;
  MJ_LadspaDatas *m_ladspaDatas;

  QTranslator    *m_translator;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_PluginDbeDsp::MJ_PluginDbeDsp()
  : m_internal(0)
{
  for(int i=0; i < MJ_GUI_SIZEOF; i++) {
    m_pluginGuiCreator[i] = 0;
  }
  // By default all modules are chainable
  setParameter(MJ_PP_METAFUNCTIONALITY, 
	       MJ_DspPluginMetaFunctionalityKeys[MJ_DPMF_CHAINABLE]);
}

MJ_PluginDbeDsp::~MJ_PluginDbeDsp()
{
  delete m_internal;
}

bool MJ_PluginDbeDsp::isEmpty() const
{
  return isValid() && m_pluginCreator;
}


MJ_ModuleDsp *MJ_PluginDbeDsp::newPlugin()
{
  const char * fname = "MJ_PluginDbeDsp::newPlugin";

  // Just in case....
  if(!this) 
    MJ_fatal(MJ_ERR_UNINITIALIZED, fname);

  if(!m_pluginCreator) {
    if(m_debug) 
      MJ_error(MJ_ERR_INTERNAL, fname, "No creator");
    return 0;
  }
  // cout << "Addr of creator" << (void *) m_pluginCreator << endl;

  MJ_ModuleDsp *m = (*m_pluginCreator)();
  if(!m) {
    MJ_error(MJ_ERR_MEMORY, fname,
	     "Could not create a new module");
    return 0;
  }

  exportCreator(m);

  if(m->dspInChannels() || m->dspOutChannels()) return m;
  
  const int limit = 32;

  bool ok;

  // Try to guess some acceptable IO config:

  for(int i=0; i < limit; i++) {
    for(int j=0; j < limit; j++) {
      if(m_options.acceptIoChans(i,j)) {
	ok = m->dspSetInChannels(i);
	ok = ok && m->dspSetOutChannels(j);
	if(ok)
	  return m;
      }
    }
  }
  return m;
}


MJ_ModuleDspGui *MJ_PluginDbeDsp::newGui(MJ_RtControl *rtparent,
					 MJ_ModuleDsp *module,
					 MJ_GuiSize s,
					 QWidget *parent, 
					 const char *name)
{
  const char *fname = "MJ_PluginDbeDsp::newGui";

  if(m_debug) MJ_trace(fname);
  
  MJ_ModuleDspGui *g = 0;

  if(m_pluginGuiCreator[s]) g = ((*m_pluginGuiCreator[s])
				 (rtparent, module, parent, name));

  if(!g) {
    if(s == MJ_GUI_STAMP) 
      g = new MJ_ModuleDspGuiPostCard
	(rtparent, module, true, 1, MJ_GUI_STAMP, parent, name);
    // g = new MJ_ModuleDspGuiStamp(parent, name);
    else if(s == MJ_GUI_POSTCARD) 
      g = new MJ_ModuleDspGuiPostCard
	(rtparent, module, true, 5, MJ_GUI_POSTCARD, parent, name);
    else {
      if(m_pluginGuiCreator[MJ_GUI_POSTCARD]) 
	g = (*m_pluginGuiCreator[MJ_GUI_POSTCARD])(rtparent, module, 0, name);
      else
	g = new MJ_ModuleDspGuiPostCard
	  (rtparent, module, true, (uint) -1, MJ_GUI_OWN_WINDOW, 0, name);
    }
  }

  if(g) g->setDbe(this);

  return g;
}

void MJ_PluginDbeDsp::setGuiFactory
(MJ_GuiSize s, MJ_ModuleDspGui *(*f)
 (MJ_RtControl *, MJ_ModuleDsp *, QWidget *, const char *))
{
  if(s >= MJ_GUI_SIZEOF)
    MJ_fatal(MJ_ERR_ARGUMENT, "MJ_PluginDbeDsp::setGuiFactory");
  
  m_pluginGuiCreator[s] = f;
}


void MJ_PluginDbeDsp::setWorksAsTopLevel(bool b) 
{ 
  internalData()->m_worksAsTopLevel = b; 
}

bool MJ_PluginDbeDsp::worksAsTopLevel() 
{ 
  if(!m_internal) return false;

  return m_internal->m_worksAsTopLevel; 
}

bool MJ_PluginDbeDsp::mayConflict() const 
{ 
  if(!m_internal) return false;

  return m_internal->m_mayConflict;
}

void MJ_PluginDbeDsp::setMayConflict(bool b) 
{ 
  internalData()->m_mayConflict = b; 
}

bool MJ_PluginDbeDsp::conflicts(MJ_PluginDbeDsp *)
{
  return false;
}

MJ_LadspaDatas *MJ_PluginDbeDsp::ladspaDatas() 
{ 
  if(!m_internal) return 0;

  return m_internal->m_ladspaDatas;
}
/** Sets the ladspa datas. If old data existed, it is deleted. */
void MJ_PluginDbeDsp::setLadspaDatas(MJ_LadspaDatas *ld)
{ 
  MJ_LadspaDatas *& ladspaDatas = internalData()->m_ladspaDatas;

  if(ladspaDatas) {
    delete ladspaDatas;
  }
  ladspaDatas = ld;
}


void MJ_PluginDbeDsp::setTranslator(QTranslator *t)
{
  QTranslator *& translator = internalData()->m_translator;

  if(translator) {
    if(qApp) qApp->removeTranslator(translator);
    delete translator;
  }
  translator = t;
}


void MJ_PluginDbeDsp::test(QApplication *xpQA, int iArgc, char **cppArgv)
{
  MJ_PluginDbeDspLoader loader;

  for(int i=1; i < iArgc; i++) {
    MJ_trace("MJ_PluginDbeDsp::test, loading dir ", cppArgv[i]);
    loader.loadDir(QDir(cppArgv[i]), "");
  }
  loader.writeNames(MJ_cout());
  
  MJ_trace("MJ_PluginDbeDsp::test - NAMES WRITTEN");

  // Form a tree out of the bastards:
  MJ_PluginDbTreeNodeMany tree;
  tree.formTree(loader.entryList());
  tree.print(MJ_cout());

  MJ_trace("MJ_PluginDbeDsp::test Plugins printed");
  MJ_PluginMenu pm(0, 0, &loader);
  pm.blockSignals(true);
  xpQA->setMainWidget(&pm);
  pm.show();
  xpQA->exec();
}

bool MJ_PluginDbeDsp::read
(const QFileInfo &xFileInfo, 
 const QString   &funcName,
 MJ_PluginDbeDspLoader *loader,
 bool             WarnOnErrors,
 QString         *errorStr)
{
  const char * fname = "MJ_PluginDbeDsp::read";

  if(m_debug)
    MJ_trace(fname, (const char *) xFileInfo.absFilePath());
  
  void *handle = dlopen((const char *) xFileInfo.absFilePath(), RTLD_NOW);
  if(!handle) {

    QCString dle(dlerror());    
    if(errorStr) (*errorStr) = dle;

    if(WarnOnErrors) {
      QCString dle(dlerror());
      
      MJ_error2(MJ_ERR_ARGUMENT, 
		"%s # Could not open dynamic library \"%s\" # %s", fname,
		  (const char *) xFileInfo.absFilePath(), (const char *) dle);
    }
    return 0;
  }
    
  if(m_debug) {
    MJ_trace2("%s # Assuming function name: \"%s\"", 
	      fname, (const char *)funcName);
  }
  
  /* We should have a function returning the templated class, but
     buggy g++ (prior to 2.95.x ?) does not allow this. Instead return
     void pointer and cast it to proper type later on....
     
     For some reason this works properly in the previous method "read".  */
  // void *(*xpCreator)(int ) = 0;
  const char * funcNameStr = (const char *) funcName;
  MJ_PluginDbeDsp *(*funcPtr)(int) = (MJ_PluginDbeDsp*(*)(int)) 
    dlsym(handle, funcNameStr);
  // xpCreator =  (void *((*)(int))) tmpPtr;
  
  if(!funcPtr) {

    QCString dle(dlerror());    
    if(errorStr) (*errorStr) = dle;

    if(WarnOnErrors) {
      MJ_error2(MJ_ERR_ARGUMENT, 
		"%s # Could not load function \"%s\" # %s", fname,
		funcNameStr, (const char *) dle);
      dlclose(handle);
    }  
    return 0;
  }  

  // if(m_debug) cout << "Got function" << endl;
  
  MJ_SoRefCount * socounter = 
    new MJ_SoRefCount(handle, xFileInfo.absFilePath());

  for(int iArg=0; true; iArg++) {

    if(m_debug) MJ_trace("Calling function",  (const char *)funcName, iArg);

    MJ_PluginDbeDsp *xpRes = (*funcPtr)(iArg);

    // (MJ_PluginDbeDsp *) (*xpCreator)(iArg);

    if(xpRes) {
      xpRes->setRefCounter(socounter);
      if(!loader->add(xpRes))
	delete xpRes;
    }
    else {
      if(!iArg) {
	// Not a single plugin in the library.
	MJ_trace3(WarnOnErrors, "%s # no modules", fname);
	delete socounter;
	return false;
      }
      else
	return true;
    }
  }
}

void MJ_PluginDbeDsp::exportCreator(MJ_ModuleDsp *m)
{
  int pluginCount = m_pluginRefCount;
  m->setDbe(this); 
  if(pluginCount == m_pluginRefCount) {
    MJ_error(MJ_ERR_INTERNAL, 
	     "MJ_PluginDbeDsp::exportCreator # \
The plugin should increase the reference count of the creator", 
	     menuPosition());
  }
}

MJ_PluginDbeDsp::__Internal *MJ_PluginDbeDsp::internalData() const
{
  MJ_PluginDbeDsp *thi = ( MJ_PluginDbeDsp *) this;

  if(!thi->m_internal) thi->m_internal = new __Internal;
  
  return thi->m_internal;
}
