/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_dsp_plugin_filters.h>

#include <mj_dsp_plugin.h>
#include <mj_dsp_options.h>
#include <mj_error.h>

#include <string.h>
#include <stdlib.h>

/** Returns always true. */

bool MJ_DspPluginFilterAny::accept(MJ_PluginDbeBase *)
{
  return true;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

bool MJ_DspPluginFilterNoInput::accept(MJ_PluginDbeBase *dbe)
{
  MJ_PluginDbeDsp *dpsDbe = dynamic_cast<MJ_PluginDbeDsp *> (dbe);

  if(!dpsDbe) return false;

  MJ_DspOptions *dspOpt = dpsDbe->options();
  
  bool ret = ! dspOpt->newInChans(0);

  /* if(ret) {
    MJ_trace("MJ_DspPluginFilterNoInput::accept", dpsDbe->menuPosition());
    }*/
  
  return ret;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

bool MJ_DspPluginFilterNoOutput::accept(MJ_PluginDbeBase *dbe)
{
  MJ_PluginDbeDsp *dpsDbe = dynamic_cast<MJ_PluginDbeDsp *> (dbe);

  if(!dpsDbe) return false;

  MJ_DspOptions *dspOpt = dpsDbe->options();

  if(dspOpt->newOutChans(0)) return false;
  
  return true;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_DspPluginFilterMetaFunc::MJ_DspPluginFilterMetaFunc(const char *metaFunc,
						       bool negate)
  : m_negate(negate)
{
  if(metaFunc) m_str = strdup(metaFunc);
  else m_str = 0;
}

MJ_DspPluginFilterMetaFunc::~MJ_DspPluginFilterMetaFunc()
{
  if(m_str) free(m_str);
}
  
bool MJ_DspPluginFilterMetaFunc::accept(MJ_PluginDbeBase *dbe)
{
  bool v = dbe->containsMetaFunctionality(m_str);

  return m_negate ? !v : v;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_DspPluginFilterParam::MJ_DspPluginFilterParam(const char *paramStr, 
						 MJ_PLuginParameter param,
						 bool negate)
  : MJ_DspPluginFilterMetaFunc(paramStr, negate),
    m_param(param)
{}

MJ_DspPluginFilterParam::~MJ_DspPluginFilterParam()
{}
  
bool MJ_DspPluginFilterParam::accept(MJ_PluginDbeBase *dbe)
{
  bool v = dbe->containsFeature(m_param, m_str);

  return m_negate ? !v : v;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_PluginFilterMulti::MJ_PluginFilterMulti(MJ_Logic logic, bool autoDelete)
  : m_logic(logic)
{
  if((uint) logic >= MJ_LOGIC_SIZEOF)
    MJ_error(MJ_ERR_ARGUMENT, 
	     "MJ_PluginFilterMulti::MJ_PluginFilterMulti", (int) logic);
  
  m_filters.setAutoDelete(autoDelete);
}

MJ_PluginFilterMulti::~MJ_PluginFilterMulti()
{}
  
bool MJ_PluginFilterMulti::accept(MJ_PluginDbeBase *dbe)
{
  const char * fname = "MJ_PluginFilterMulti::accept";

  uint count = m_filters.count();
  uint sum = 0;

  if(!count) {
    MJ_error(MJ_ERR_INTERNAL, fname, "No filters");
    return false;
  }

  for(uint i=0; i < count; i++) {
    bool ok = m_filters.at(i)->accept(dbe);
    if(ok) sum++;
  }

  switch(m_logic) {
  case MJ_LOGIC_AND:
    return sum == count;
  case MJ_LOGIC_OR:
    return sum != 0;
  case MJ_LOGIC_XOR:
    return sum == 1;
  case MJ_LOGIC_NOT:
    return sum == 0;
  case MJ_LOGIC_SIZEOF:
  default:
    MJ_error(MJ_ERR_INTERNAL, fname, m_logic);
  }
  return false;
}

void MJ_PluginFilterMulti::addFilter(MJ_PluginFilter *f)
{
  if(!f) MJ_error(MJ_ERR_ARGUMENT, "MJ_PluginFilterMulti::addFilter");
  else m_filters.append(f);
}
