/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_dsp_plugin_info_gui.h>

#include <mj_dsp_plugin.h>
#include <mj_dsp_plugin_loader.h>
#include <mj_error.h>
#include <mj_so_ref_count.h>
#include <mj_tr.h>

#include <qcheckbox.h> 
#include <qcombobox.h> 
#include <qlabel.h> 
#include <qlayout.h> 

MJ_DspPLuginInfoGui::MJ_DspPLuginInfoGui(MJ_PluginDbeDsp *dbe,
					 QWidget      *parent,
					 const char   *name)
  : QWidget(parent, name, Qt::WDestructiveClose),
    m_dbe(0),
    m_debug(0)
{
  drawChildWidgets();

  if(dbe) setDbe(dbe);
}

MJ_DspPLuginInfoGui::~MJ_DspPLuginInfoGui()
{
  qDebug("MJ_DspPLuginInfoGui::~MJ_DspPLuginInfoGui");
}

/** Sets the plugin to be displayed. 
    
    @arg dbe The plugin. This argument must not be null.
*/
void MJ_DspPLuginInfoGui::setDbe(MJ_PluginDbeDsp *dbe)
{
  if(!dbe)
    MJ_fatal(MJ_ERR_NULL_ARGUMENT, "MJ_DspPLuginInfoGui::setDbe");

  m_dbe = dbe;

  m_about[1]->setText(dbe->parameter(MJ_PP_ABOUT));
  m_copyright[1]->setText(dbe->parameter(MJ_PP_COPYRIGHT));
  m_menuPos[1]->setText(dbe->menuPosition2());
  m_functionality[1]->setText(dbe->parameter(MJ_PP_FUNCTIONALITY));
  m_metafunctionality[1]->setText(dbe->parameter(MJ_PP_METAFUNCTIONALITY));
  m_authors[1]->setText(dbe->parameter(MJ_PP_AUTHORS));
  m_library[1]->setText(dbe->soRefCounter()->fileName());
  m_compileTime[1]->setText(dbe->parameter(MJ_PP_COMPILE_TIME));

  if(dbe->pluginDebug()) {
    m_debug->setDown(*dbe->pluginDebug());
    m_debug->setEnabled(true);
  }
  else m_debug->setEnabled(false);

  MJ_DspOptions * opt = dbe->options();

  QString str;
  QString na(MJ_Tr::tr("N/A"));

  for(uint i=0; i < MJ_DSP_SIZEOF; i++) {

    if(opt->isDefined((MJ_DspOption) i))
      str.setNum(opt->value((MJ_DspOption) i));
    else str = na;

    m_optParams[i]->setText(str);
  }
}

void MJ_DspPLuginInfoGui::setDebug(bool debug)
{
  if(!m_dbe) {
    MJ_error(MJ_ERR_INTERNAL, 
	     "MJ_DspPLuginInfoGui::setDebug # Cannot debug NULL");
    return;
  }

  if(m_dbe->pluginDebug())
    *(m_dbe->pluginDebug()) = debug;
}

void MJ_DspPLuginInfoGui::drawChildWidgets()
{
  QGridLayout *grid = new QGridLayout(this, 8 + MJ_DSP_SIZEOF, 5, 5);
  
  QString pinfo(MJ_Tr::tr("Plugin information"));

  if(!parent()) {
    setCaption(pinfo);
    setIconText(pinfo);
  }

  QLabel *label = new QLabel(pinfo, this);
  QFont font = label->font();
  font.setPointSizeFloat(font.pointSizeFloat() * 1.5);  // Bigger font
  font.setBold(true);  // Bold font
  label->setFont(font);
  
  int row = 0;

  grid->addMultiCellWidget(label, row, row, 0, 1, AlignCenter);
  row++;

  label->show();

  addLabels(grid, m_about, row++, 
	    MJ_Tr::tr(MJ_PLuginParameterKeys[MJ_PP_ABOUT]));
  addLabels(grid, m_copyright, row++, 
	    MJ_Tr::tr(MJ_PLuginParameterKeys[MJ_PP_COPYRIGHT]));
  addLabels(grid, m_menuPos, row++, 
	    MJ_Tr::tr(MJ_PLuginParameterKeys[MJ_PP_MENU_POSITION]));
  addLabels(grid, m_functionality, row++, 
	    MJ_Tr::tr(MJ_PLuginParameterKeys[MJ_PP_FUNCTIONALITY]));
  addLabels(grid, m_metafunctionality, row++, 
	    MJ_Tr::tr(MJ_PLuginParameterKeys[MJ_PP_METAFUNCTIONALITY]));
  addLabels(grid, m_authors, row++, 
	    MJ_Tr::tr(MJ_PLuginParameterKeys[MJ_PP_AUTHORS]));
  addLabels(grid, m_library, row++, 
	    MJ_Tr::tr("Library"));
  addLabels(grid, m_compileTime, row++, 
	    MJ_Tr::tr("Compilation time"));

  m_debug = new QCheckBox(MJ_Tr::tr("Debug"), this);
  m_debug->setEnabled(false);
  connect(m_debug, SIGNAL(toggled(bool)), this, SLOT(setDebug(bool)));
  grid->addWidget(m_debug, row++, 0);

  grid->addRowSpacing(row++, 15);
  
  for(uint i=0; i < MJ_DSP_SIZEOF; i++) {

    QLabel * label= new QLabel(MJ_Tr::tr(MJ_DspOptionWords2[i]), this);
    grid->addWidget(label, row, 0);
    m_optParams[i] = new QLabel(this);
    grid->addWidget(m_optParams[i], row, 1);

    row++;
  }
  grid->setRowStretch(row, 60); // Make the empty rows take all the space
}
  
void MJ_DspPLuginInfoGui::addLabels(QGridLayout *grid, 
				    QLabel **labels, 
				    int row, 
				    const QString & name)
{
  labels[0] = new QLabel(name, this);
  labels[1] = new QLabel(this);
  grid->addWidget(labels[0], row, 0);
  grid->addWidget(labels[1], row, 1);
  labels[0]->show();
  labels[1]->show();
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


MJ_PluginsInfo::MJ_PluginsInfo(MJ_PluginDbeDspLoader *loader,
			       QWidget      *parent,
			       const char   *name)
  : QWidget(parent, name, Qt::WDestructiveClose),
    m_loader(loader),
    m_combo(new QComboBox(this)),
    m_pluginGui(new MJ_DspPLuginInfoGui(0, this, "plugin-info-gui"))
{
  QString pinfo(MJ_Tr::tr("Plugin information"));

  if(!parent) {
    setCaption(pinfo);
    setIconText(pinfo);
  }

  MJ_PluginDbeList *plist = & loader->entryList();
  for(uint i=0; i < plist->count(); i++) {
    QString id;
    id.setNum(i + 1);
    id += ": " + plist->at(i)->menuPosition2();
    m_combo->insertItem(id);
  }
  connect(m_combo, SIGNAL(activated(int)),
	  this, SLOT(selectDbe(int)));

  if(plist->count()) {
    m_combo->setCurrentItem(0);
    selectDbe(0);
  }

  QVBoxLayout *layout = new QVBoxLayout(this, 4, 5);
  layout->addWidget(m_combo, 0, AlignLeft);
  layout->addWidget(m_pluginGui);
}

/* void MJ_PluginsInfo::hide()
{
  if(!parent())
    delete this;
  // QWidget::hide();
  }*/

void MJ_PluginsInfo::selectDbe(int i)
{
  MJ_PluginDbeDsp *dbe = (MJ_PluginDbeDsp *) m_loader->entryList().at(i);
  m_pluginGui->setDbe(dbe);
}
