/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_dsp_plugin_loader.h>

#include <mj_debug.h>
#include <mj_error.h>
#include <mj_multi_info_gui.h>
#include <mj_paths.h>
#include <mj_so_ref_count.h>
#include <mj_tr.h>

#include <qapplication.h>
#include <qtextstream.h>

#include <errno.h>

bool MJ_PluginDbeDspLoader::m_debug=false;

MJ_PluginDbeDspLoader::MJ_PluginDbeDspLoader()
{}

MJ_PluginDbeDspLoader::~MJ_PluginDbeDspLoader()
{
  // Clear up the list.
  while(m_list.count()) {
    MJ_PluginDbeBase *dbe = m_list.first();
    m_list.removeFirst();

    deleteDbe(dbe);
  }
}

bool MJ_PluginDbeDspLoader::loadTranslations(const QString &paths, 
					     const QString &lang)
{
  const char * fname = "MJ_PluginDbeDspLoader::loadTranslations";

  if(paths.isEmpty()) return false;

  bool ok = true;

  uint i, len = m_list.count();

  if(m_debug) MJ_trace(fname, len);

  MJ_Paths p;

  p.addPath(paths);
  
  QString tmp, tmp2;

  for(i=0; i < len; i++) {

    MJ_PluginDbeDsp *dbe = (MJ_PluginDbeDsp *) getEntry(i);

    // Load any translation files:
    const QString name = dbe->parameter(MJ_PP_TRANSLATION_FILE);

    if(!name.isEmpty()) {
      
      if(m_debug) MJ_trace(fname, name);
      
      bool loop = true;

      QString tmpLang(lang);
      
      while(loop) {
	QString trfile;
	
	if(tmpLang.isEmpty())
	  trfile = name + ".qm";
	else
	  trfile = name + "_" + tmpLang + ".qm";
	
	if(m_debug)
	  MJ_trace(fname, trfile);

	QString trfilePath = p.findFile(trfile);

	if(!trfilePath.isEmpty() && qApp) {
	  
	  QTranslator *qtr = new QTranslator(qApp);
	  
	  int ok = qtr->load(trfilePath);
	  
	  //QWidget::qDebug("%s tried to load %s success = %d", 
	  //		fname, trfilePath, ok);
	  if(!ok) 
	    delete qtr;
	  else {
	    qApp->installTranslator(qtr);
	    dbe->setTranslator(qtr);
	  }
	  loop = false;
	}
	else {
	  /* Failed to load translation, change language string
             shorter. Maybe we will find something.  */
	  if(tmpLang.isEmpty()) {

	    if(!lang.isEmpty() && MJ_Debug::debugAll())
	      MJ_trace2("%s # Could not load translations for \"%s\"", 
			fname, (const char *) name);
	    loop = false;
	    ok = false;
	  }
	  else {
	    if(MJ_fields(tmpLang, "_"))
	      tmpLang = MJ_stripLastField(tmpLang, "_");
	    else
	      tmpLang.truncate(0);
	  }

	}
      }
    }

    // Translate menu position
    const QString &menu = getEntry(i)->menuPosition();
    // Try to translate to full menu path
    tmp = MJ_Tr::tr(menu);

    if(tmp == menu) {
      /** Apparently it was not possibleto translate full menu
          path. Instead, translate each menu field. */
      int fs = MJ_fields(menu, MJ_pluginSeparator);
      tmp.truncate(0);

      for(int j=0; j < fs; j++) {
	tmp2 = MJ_Tr::tr(MJ_field(menu, j, MJ_pluginSeparator));
	tmp += MJ_pluginSeparator + tmp2;
      }
      // Now set this new menu path:

    }

    getEntry(i)->setMenuPosition(tmp);      

    // Translate the "about" text:
    const QString &about = getEntry(i)->about();
    tmp = MJ_Tr::tr(about);

    if(tmp != about)
      getEntry(i)->setAbout(tmp);

  }
  m_list.sort();

  return ok;
}

bool MJ_PluginDbeDspLoader::loadDir
(QDir xDir, 
 const QString &funcName, 
 bool bRecursive, 
 int iDepth,
 MJ_MultiInfoGui *errorGui)
{
  const char * fname = "MJ_PluginLoaderTmpl::loadDir";

  MJ_trace3(m_debug, "%s # Directory \"%s\"", fname, xDir.path().ascii());

#ifndef WIN32
  const char * libraryPostFix = "so";
#else
  const char * libraryPostFix = "dll";
#endif


  if(iDepth > 20) {
    if(m_debug) MJ_error(MJ_ERR_INTERNAL,
			 fname, "Too many directories traversed");
    return false;
  }

  if(!xDir.exists()) {
    if(m_debug) 
      MJ_error(MJ_ERR_ARGUMENT, fname, "No such directory: ", 
	       (const char *) xDir.path());
    return false;
  }

  const QFileInfoList *xpInfoList = xDir.entryInfoList();

  if(xpInfoList->isEmpty()) {
    if(m_debug) 
      MJ_error(MJ_ERR_ARGUMENT, fname, "No modules in: ", 
	       (const char *) xDir.path());
    return false;
  }
  
  QFileInfoListIterator xIterator(*xpInfoList);
  QFileInfo *xpInfo;

  QString err1, err2;

  for(;(xpInfo = xIterator.current()); ++xIterator) {
    err2.truncate(0);
    
    // MJ_trace(fname, xpInfo->fileName());

    if(strcmp(xpInfo->extension(), libraryPostFix) == 0) {
      
      if(m_debug) MJ_trace(fname, "to read");

      MJ_PluginDbeDsp::read
	(*xpInfo, funcName, this, m_debug, & err2);

      if(!err2.isEmpty() && errorGui) {
	QString str(xpInfo->filePath() + " \n\n " + err2);
	errorGui->addInfo(str);
      }
      
    }
    else if(xpInfo->isDir()) {
      // Do not reload the directory.
      if(strcmp(xpInfo->fileName().data(), ".") != 0 &&
	 strcmp(xpInfo->fileName().data(), "..") != 0 && 
	 bRecursive)
	loadDir(QDir(xpInfo->absFilePath()), funcName, bRecursive, iDepth+1);
    }
    else if(m_debug) {
      MJ_error(MJ_ERR_UNKNOWN, fname, "Unclassified filename ",
	       xpInfo->absFilePath());
      return false;
    }
  }
  if(m_debug)
    MJ_trace(fname, "EXIT");

  errno = 0;

  return true;
}
  

void MJ_PluginDbeDspLoader::writeNames(QTextStream &xStream)
{
  QListIterator<MJ_PluginDbeBase> 
    xIterator(m_list);
  
  MJ_PluginDbeBase *entry2 = NULL;
  
  for(;(entry2 = xIterator.current()); ++xIterator) {
    if(entry2)
      xStream << entry2->menuPosition() << "\n";
  }
}

void MJ_PluginDbeDspLoader::writeHWNames(QTextStream &xStream)
{
  (void) xStream;

  /* QListIterator<MJ_PluginDbeBase> 
    xIterator(m_list);
  
  MJ_PluginDbeDsp *entry2(0);
  for(;(entry2 = 
	(MJ_PluginDbeDsp *)xIterator.current()); 
      ++xIterator) {
    if(entry2->options()->isHardWare())
      xStream << entry2->menuPosition() << "\n";
      }*/
}

MJ_PluginDbeDsp  * MJ_PluginDbeDspLoader::findEntry
(const QString & id, 
 MJ_PLuginParameter searchCriterion)
{
  MJ_PluginDbeBase *dbe = m_list.findDbe(id, searchCriterion);

  if(dbe) return dynamic_cast<MJ_PluginDbeDsp  *> (dbe);
  return 0;
}


bool  MJ_PluginDbeDspLoader::add(MJ_PluginDbeDsp *entry)
{
  if(entry->identifier().isEmpty() ||
     entry->menuPosition2().isEmpty()) {
    MJ_error2(MJ_ERR_ARGUMENT, 
	      "MJ_PluginDbeDspLoader::add # plugin lacks id or menupos");
    return false;
  }
  
  QString menuBase = entry->menuPosition();
  QString menuPos = menuBase;
  int index = 0;
  MJ_PluginDbeDsp *old;

  while((old = findEntry(menuPos, MJ_PP_MENU_POSITION))) {

    MJ_trace2("MJ_PluginDbeDspLoader::add # Plugin \"%s\" (id: %s )"
	      "was loaded twice. Second version ignored."
	      "First is defined in library \"%s\", the second in \"%s\"",
	      (const char *) entry->menuPosition2(),
	      (const char *) entry->identifier(),
	      old->soRefCounter()->fileName().ascii(),
	      entry->soRefCounter()->fileName().ascii());

    if(index)
      menuPos = menuBase + " " + QString().setNum(index);

    index++;
    
  }
  
  entry->setMenuPosition(menuPos);

  // Switch debugging on if can/need
  bool *g = entry->pluginDebug();
  if(g)
    *g = *g || MJ_Debug::debugNeeded(entry->parameter(MJ_PP_IDENTIFIER)) || 
      MJ_Debug::debugAll();

  // Add to the list of plugins.
  m_list.inSort(entry);
  return true;
}

void MJ_PluginDbeDspLoader::deleteDbe(MJ_PluginDbeBase *dbe)
{
  MJ_SoRefCount *socount = dbe->soRefCounter();

  delete dbe;
  
  if(!socount->count()) delete socount;
}
