/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_failure_diagnose.h>

#include <mj_app_action.h>
#include <mj_base.h>
#include <mj_error_streams.h>
#include <mj_iostream.h>
#include <mj_mutex.h>

#include <qbuffer.h> 
#include <qtextstream.h> 

#include <stdlib.h>

#ifndef WIN32
#include <unistd.h>
#endif

#include <typeinfo>

static MJ_MutexAuto MJ_FailureDiagnose_Mutex(true, false);

MJ_FailureDiagnose::MJ_FailureDiagnose()
{
  clear();
}

MJ_FailureDiagnose::~MJ_FailureDiagnose()
{
  clear();
}

void MJ_FailureDiagnose::failure(const MJ_Base *origin,
				 MJ_ErrorCode  code,
				 MJ_ErrorLevel level)
{ 
  m_origin = origin; 
  setCode(code);
  setLevel(level);
}

void MJ_FailureDiagnose::addMessage(const QString &m)
{
  QString *s = new QString;
  (*s) = m.copy();
  m_messages.append(s);
}

void MJ_FailureDiagnose::addMessage(const char *m)
{
  QString s(m);
  addMessage(s);
}

bool MJ_FailureDiagnose::isEmpty() const
{
  if(m_code == MJ_ERR_NONE && 
     !m_origin && !m_messages.count()) return true;
  return false;
}

void MJ_FailureDiagnose::clear()
{
  m_origin = 0;
  m_code   = MJ_ERR_NONE;
  m_level  = MJ_ERRL_ERROR;
#ifndef WIN32
  m_pid    = (int) getpid();
#endif
  m_messages.setAutoDelete(true);
  m_messages.clear();
}

void MJ_FailureDiagnose::setCode(MJ_ErrorCode code)
{
  if((uint) code < (uint) MJ_ERR_SIZEOF)
    m_code = code; 
  else {
    MJ_cerr() << "MJ_FailureDiagnose::setCode " 
	    << MJ_ErrorCodes[MJ_ERR_ARGUMENT] << "\n";
    MJ_cerr().device()->flush();
    m_code = MJ_ERR_UNKNOWN;
  }
}

void MJ_FailureDiagnose::setLevel(MJ_ErrorLevel level)
{
  if((uint) level < (uint) MJ_ERRL_SIZEOF)
    m_level = level; 
  else {
    MJ_cerr() << "MJ_FailureDiagnose::setLevel " 
	    << MJ_ErrorCodes[MJ_ERR_ARGUMENT] << "\n";
    MJ_cerr().device()->flush();
  }
}

void MJ_FailureDiagnose::write(QTextStream *s)
{
  QString tmpStr;
  QTextOStream tmpStream(&tmpStr);

  tmpStream << "MJ:" << MJ_ErrorLevels[m_level] ;

  if(m_level != MJ_ERRL_TRACE) {
    tmpStream << ":" << MJ_ErrorCodes[m_code];
    tmpStream << ":pid=" << m_pid;
  }

  if(m_origin) {
    tmpStream << "(" << typeid(*m_origin).name() << ")";
  }

  tmpStream << " origin: " << (void *) m_origin << flush;
  
  for(uint i=0; i < m_messages.count(); i++) {
    tmpStream << " # " << *m_messages.at(i);
  }
  tmpStream << "\n";

  (*s) << tmpStr;

  s->device()->flush();

  if(m_level != MJ_ERRL_TRACE)
    MJ_AppActions::reportError(tmpStr);
}

void MJ_FailureDiagnose::toUser()
{
  if(MJ_silenceErrors)
    return;

  MJ_FailureDiagnose_Mutex.lock();

  QBuffer b;
  b.open(IO_ReadWrite);
  QTextStream s(&b);

  write(&s);

  FILE * errStream = MJ_getErrorStream(m_level);
  
  /* errStream->writeRawBytes(b.buffer(), b.buffer().size());
     errStream->device()->flush(); */

  fwrite(b.buffer(), b.buffer().size(), 1, errStream);

  if(m_level == MJ_ERRL_FATAL) {
#ifndef WIN32
    int *badAddr(0);
#else
    int *badAddr=NULL;
#endif
    *badAddr = 0;
    exit(0);
  }

  MJ_FailureDiagnose_Mutex.unlock();
}

void MJ_FailureDiagnose::setTraceFile(FILE *f)
{
  MJ_traceFile = (FILE *) f;
}
