/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_kernel.h>

#include <mj_app_action.h>
#include <mj_dsp_plugin_loader.h>
#include <mj_dsp_plugin_filters.h>
#include <mj_error.h>
#include <mj_global.h>
#include <mj_iostream.h>
#include <mj_latency_gui.h>
#include <mj_misc.h>
#include <mj_module_dsp_gui.h>
#include <mj_module_dsp_ui.h>
#include <mj_multi_info_gui.h>
#include <mj_dsp_plugin_info_gui.h>
#include <mj_plugin_db_tree.h>
#include <mj_plugin_menu.h>
#include <mj_prefs.h>
#include <mj_rt_driver.h>
#include <mj_rt_thread.h>
#include <mj_satellite.h>
#include <mj_strs.h>
#include <mj_tr.h>
#include <mj_watch_dog.h>

#include <di_error.h>

#include <qapplication.h>
#include <qdatetime.h> 
#include <qdir.h> 
#include <qfile.h> 
#include <qfileinfo.h> 
#include <qfiledialog.h>
#include <qkeycode.h>
#include <qlayout.h> 
#include <qmenubar.h>  
#include <qmessagebox.h> 
#include <qpopupmenu.h>
#include <qtextstream.h> 

#include <dlfcn.h>

#ifndef WIN32
#include <sched.h>
#endif

#include <errno.h>
#include <stdlib.h>

#ifndef WIN32
#include <sys/utsname.h>
#include <unistd.h>
#endif

#include <typeinfo>

#ifdef __OS_LINUX__
#include <mj_rt_test.h>
#endif // __OS_LINUX__


class MJ_Kernel::Internal 
{
public:
  Internal() : m_actions(new MJ_AppActions),
	       m_driver(new MJ_RtDriver),
	       m_dspThread(0),
	       m_rootUi(0),
	       m_dspLoader(new MJ_PluginDbeDspLoader),
	       m_dspPluginTree(new MJ_PluginDbTreeNodeMany),
	       m_translator(new QTranslator(0)),
	       m_translatorTriv(0),
	       m_status(MJ_KERNEL_STOP),
	       m_runWidget(0)
  {
    m_setups.setAutoDelete(true);
  }

  ~Internal()
  {
    delete m_dspLoader; 
    delete m_dspPluginTree;
    
    delete m_runWidget;
    m_setups.clear();

    if(qApp) {
      if(m_translator) qApp->removeTranslator(m_translator);
      if(m_translatorTriv) qApp->removeTranslator(m_translatorTriv);
    }

    delete m_translator;
    delete m_translatorTriv;
  }

  QList<MJ_ModuleDsp>    m_setups;

  MJ_AppActions         *m_actions;
  MJ_RtDriver           *m_driver;
  MJ_RtThread           *m_dspThread;
  MJ_ModuleDspUi        *m_rootUi;

  MJ_Paths               m_paths[MJ_PATH_SIZEOF];

  MJ_PluginDbeDspLoader   *m_dspLoader;
  MJ_PluginDbTreeNodeMany *m_dspPluginTree;

  /** Translator for generic Mustajuuri stuff. Translates texts in
      both "mj" and "mjmodules". */
  QTranslator             *m_translator;
  QTranslator             *m_translatorTriv;

  MJ_KernelStatus          m_status;

  MJ_RunStatus            *m_runWidget;

  ulong                    m_lastFrame;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

enum MJ_KernelOpt {
  MJ_KOPT_SIZEOF
};

const char *MJ_KernelOpt[] = {
  ""
};

const char *MJ_kernelStatusWords[MJ_KERNEL_SIZEOF] = {
  "STOP",
  "NOTHREAD_PLAY",
  "THREADED_PLAY"
};

bool MJ_Kernel::m_debug = false;

/** Creates a new kernel object. 

    @pre No other MJ_Kernel object should exist.

    @post All plugins will be loaded and a plugin tree will be
    formed. If the system is run with Qt GUI (qApp is not NULL), then
    also the translations will be loaded.
*/

MJ_Kernel::MJ_Kernel(bool autoInit)
  : m_internal(new Internal)
{
  const char *fname = "MJ_Kernel::MJ_Kernel";
  
  MJ_trace3(m_debug, "%s # %p", fname, this);
  
  if(MJ_kernel)
    MJ_fatal(MJ_ERR_INTERNAL, 
	     fname, 
	     "Cannot have two kernels");
  
  MJ_kernel = this;

  if(!autoInit) return;
  
  init();

  MJ_trace3(m_debug, "%s # EXIT", fname);
}

/** Deletes the kernel and frees all resources. 

    If the application is still active then it will be stopped first.
    
    This method should be called from the user interface thread.
*/
MJ_Kernel::~MJ_Kernel()
{
  MJ_trace3(m_debug, "MJ_Kernel::~MJ_Kernel # %p", this);

  Internal *internal = m_internal;

  if(internal->m_dspThread) stopPlaying();

  delete internal->m_actions;
  
  if(internal->m_dspThread) killThread();
  if(internal->m_driver)    killDriver();

  delete internal;

  MJ_kernel = 0;
}

void MJ_Kernel::addModule(MJ_ModuleDsp *m) 
{ 
  m_internal->m_setups.append(m); 
}

/** Tries to set up the mustajuuri paths. This method is useful for
    any application that wants to use mustajuuri plugins, translations
    or icons. */

void MJ_Kernel::setEnv()
{
  const char *fname = "MJ_Kernel::setEnv";

  static QCString pathHolder;
  static QCString pluginPathHolder;

  // This is coming from the compiler:
#ifndef WIN32
  const char *defaultMustajuuriPath = PREFIX "/lib/mustajuuri";
#else
  const char *defaultMustajuuriPath = " ";
#endif

  // Check if the user has defined the location of Mustajuuri
  QString path = getenv(MJ_PathKeys[MJ_PATH_MUSTAJUURI]);

  if(!path.isEmpty()) return;
  
  QString localPath(getenv("HOME"));
    
  localPath += "/.mustajuuri-conf";
    
  QFile configFile(localPath);
  
  bool ok = configFile.open(IO_ReadOnly);
  
  if(!ok) {
    configFile.setName("/etc/mustajuuri-conf");
    ok = configFile.open(IO_ReadOnly);
  }

  // Read the config file:
  if(ok) {
    MJ_trace2("%s # Loading configuration from \"%s\"", 
	      fname, configFile.name().ascii());
    
    QTextStream config(&configFile);

    while(!configFile.atEnd()) {
      QString line = config.readLine();
      if(line.ascii()[0] == '#')
	continue;
      
      int fields = MJ_fields(line, "=");

      QString field1;

      if(fields) field1 = MJ_field(line, 0, "=");
      else field1 = line;

      if(field1 == MJ_PathKeys[MJ_PATH_MUSTAJUURI]) {
	path = MJ_field(line, 1, "=");

	MJ_trace2("%s # Configuration tells that Mustajuuri is located under"
		  "\"%s\"", fname, path.ascii());
	pathHolder.sprintf("%s=%s", MJ_PathKeys[MJ_PATH_MUSTAJUURI], 
			   path.ascii());
	putenv(pathHolder.data());
      }
    }
  }

  /* Make an ad-hoc guess that Mustajuuri might be located under
     PREFIX/lib/mustajuuri if the path has not been defined yet: */
  if(path.isEmpty())
    path = defaultMustajuuriPath;
    
  QFileInfo fi(path);
  
  if(fi.exists()) {
    pathHolder.sprintf("%s=%s", MJ_PathKeys[MJ_PATH_MUSTAJUURI], 
		       path.ascii());
    putenv(pathHolder.data());
    
    QCString pluginPath = getenv(MJ_PathKeys[MJ_PATH_MUSTAJUURI_PLUGIN]);
    
    if(pluginPath.isEmpty()) {
      /* The path for plugins. The path does NOT contain
	 platform-information . We assume that Mustajuuri has been
	 installed as a single-platform entity with the normal
	 installation methods. */

      pluginPathHolder = MJ_PathKeys[MJ_PATH_MUSTAJUURI_PLUGIN];
      pluginPathHolder += "=" + path + "/plugins";
      putenv(pluginPathHolder.data());

      MJ_trace2("%s # Assuming Mustajuuri plugins are located under \"%s\"", 
		fname, pluginPathHolder.data());
    }
  }
  else {
    /* Make the last desperate guess that Mustajuuri is in the place
       where it was compiled to. The plugin path will be generated
       once we hit MJ_Kernel::loadDefaultDirs. */
    pathHolder.sprintf("%s=%s", MJ_PathKeys[MJ_PATH_MUSTAJUURI], 
		       defaultMustajuuriPath);
    putenv(pathHolder.data());
    MJ_trace2("%s # Mustajuuri is assumed to be under \"%s\"", fname,
	      defaultMustajuuriPath);
  }
}

void MJ_Kernel::init()
{
  initPaths();
  loadDefaultDirs();
  formPLuginTree();
}

/** Removes a module from the list of modules. The module <B>is
    not</B> deleted. 

    @return Returns the module.
*/

MJ_ModuleDsp *MJ_Kernel::forgetModule(uint index)
{ 
  Internal *internal = m_internal;

  MJ_ModuleDsp *m = internal->m_setups.take(index);
  if(internal->m_runWidget && m)
    internal->m_runWidget->moduleDestroyed(m);
  return  m;
}

/** Removes a module from the list of modules. The module <B>is</B>
    deleted. 
    
    @return Returns true if the module is deleted correctly.
*/
bool MJ_Kernel::removeModule(uint index)
{   
  MJ_trace("MJ_Kernel::removeModule", index);

  Internal *internal = m_internal;

  MJ_ModuleDsp *m = internal->m_setups.at(index);
  if(internal->m_runWidget && m)
    internal->m_runWidget->moduleDestroyed(m);

  return internal->m_setups.remove(index); 
}

/** Starts to play a module.

    @pre the kernel must be in MJ_KERNEL_STOP-state.

    @param rootUi The user interface of the module to be played. 

    @param method The method of playing. Should be one either
    MJ_KERNEL_NOTHREAD_PLAY or MJ_KERNEL_THREADED_PLAY.

    @return Returns true if the operation was successful. Note that if
    you are using a non-threaded mode, this method will return only
    after the signal processing is finished.
*/
bool MJ_Kernel::playModule(MJ_ModuleDspUi *rootUi, 
			   MJ_KernelStatus method, 
			   ulong firstFrame, ulong lastFrame)
{
  const char * fname = "MJ_Kernel::playModule";

  Internal *internal = m_internal;

  if(internal->m_status != MJ_KERNEL_STOP) {
    MJ_error(MJ_ERR_UNKNOWN, fname, "Already operational (!?)");
    return false;
  }

  if(!lastFrame) lastFrame = (ulong) -100000;

  /** If the user seems to be running a GUI and there is not (yet) a
      run-status object, make one. */
  if(qApp && !internal->m_runWidget && MJ_kernelGui)
    internal->m_runWidget = new MJ_RunStatus;

  if(method == MJ_KERNEL_NOTHREAD_PLAY) {
    return runNonThreaded(rootUi);
  }
  else if(method == MJ_KERNEL_THREADED_PLAY) {
    bool ok = startThreaded(rootUi, firstFrame, lastFrame);
    if(ok && internal->m_runWidget) {
      internal->m_runWidget->started(rootUi, internal->m_dspThread);
      internal->m_runWidget->show();
      internal->m_runWidget->raise();
    }
    return ok;
  }
  
  MJ_error(MJ_ERR_ARGUMENT, fname, "Unknown method");
  
  return false;
}
 
/** Stops playing of the previous module. 

    @param rtThread This argument tells whether the method is called
    within the real-time thread or the user interface thread.

    @returns Returns true if the operation was successfull.
*/
 
bool MJ_Kernel::stopPlaying(bool rtThread)
{
  if(m_debug) MJ_trace("MJ_Kernel::stopPlaying");

  Internal *internal = m_internal;

  if(internal->m_runWidget && !rtThread) internal->m_runWidget->stopped();

  if(internal->m_status == MJ_KERNEL_NOTHREAD_PLAY) {
    internal->m_driver->requestStop();
    return true;
  }
  else if(internal->m_status == MJ_KERNEL_THREADED_PLAY) {
    return stopThreaded(rtThread);
  } 
  return false;
}

/** Try to stop the processing. If the real-time thread is active and
    GUI exists, then the user is prompted to verify that operation
    should be stopped.

    @pre This method should only be called from the user interface
    thread.

    @return Returns true if the processing is stopped.
*/
bool MJ_Kernel::tryToStop()
{
  if(m_debug) MJ_trace("MJ_Kernel::tryToStop");

  if(m_internal->m_status == MJ_KERNEL_STOP) return true;

  if(m_internal->m_rootUi) {
    int choice = QMessageBox::warning
      (0, 
       MJ_Tr::tr("Stop Mustajuuri ?"),
       MJ_Tr::tr("Mustajuuri is active, really stop it ?"),
       MJ_Tr::tr("Yes"),
       MJ_Tr::tr("No"),
       QString::null,
       1);
    if(choice == 1) return false;
  }
  return stopPlaying();
}

/** Panic exit. This method does its very best to forcefully kill
    any DSP-threads and exit as fast as possible. Only to be called
    in emergency. */
void MJ_Kernel::panic()
{
  const char * fname = "MJ_Kernel::panic";
  

  if(m_internal->m_dspThread) {
    MJ_error(MJ_ERR_UNKNOWN, fname, "Killing dsp thread");
    m_internal->m_dspThread->kill();
  }
  
  MJ_fatal(MJ_ERR_UNKNOWN, fname);
}

MJ_RtThread *MJ_Kernel::dspThread()
{ 
  return m_internal->m_dspThread; 
}

/** Create a new DSP module. 

    @param identifier String describing the <B>feature value</B> based
    on which the plugin will bo sought.

    @param searchCriterion Enum describing the feature based on
    which the plugin will bo sought.
*/
MJ_ModuleDsp *MJ_Kernel::newDspModule(const QString &identifier,
				      MJ_PLuginParameter searchCriterion)
{
  MJ_PluginDbeDsp *dbe = findDspPlugin(identifier, searchCriterion);
  
  MJ_trace3(m_debug, "MJ_Kernel::newDspModule # %p %s", dbe,
	    identifier.ascii());

  if(!dbe) return 0;

  return dbe->newPlugin();
}

MJ_ModuleDsp *MJ_Kernel::newFunctionalModule(const QString &function)
{
  return newDspModule(function, MJ_PP_FUNCTIONALITY);
}

MJ_PluginDbeDsp *MJ_Kernel::findDspPlugin
(const QString &identifier,
 MJ_PLuginParameter searchCriterion)
{
  const char * fname = "MJ_Kernel::findDspPlugin";

  if(!MJ_kernel) 
    MJ_fatal(MJ_ERR_INTERNAL, fname);

  MJ_PluginDbeDspLoader *dbl = MJ_kernel->dspLoader();
  if(!dbl) return 0;

  return MJ_kernel->dspLoader()->findEntry(identifier, searchCriterion);
}

/// Status of the kernel.
MJ_KernelStatus MJ_Kernel::status() const 
{ 
  return m_internal->m_status; 
}

MJ_Paths &MJ_Kernel::path(uint i) 
{ 
  return m_internal->m_paths[i];
}

MJ_AppActions *MJ_Kernel::actions()
{
  MJ_cpf(MJ_kernel, "MJ_Kernel::actions # MJ_kernel = NULL !!");
  return MJ_kernel->m_internal->m_actions;
}

/// Object that is used to lead the plugins.
MJ_PluginDbeDspLoader *MJ_Kernel::dspLoader()
{ 
  if(m_debug) MJ_trace("MJ_Kernel::dspLoader", this);
  return m_internal->m_dspLoader; 
}

/// A ready-made tree of the plugins.
MJ_PluginDbTreeNodeMany *MJ_Kernel::dspPluginTree()
{ 
  const char *fname = "MJ_Kernel::dspPluginTree";
  if(m_debug) MJ_trace(fname, this);
  return m_internal->m_dspPluginTree; 
}

/** Load plugins and translations from default directories. */
void MJ_Kernel::loadDefaultDirs()
{
  const char *fname = "MJ_Kernel::loadDefaultDirs";
  // Load generic translations

  Internal *internal = m_internal;

  MJ_Paths &basePath = internal->m_paths[MJ_PATH_MUSTAJUURI];
  
  const char * lang = getenv("MUSTAJUURI_LANG");
  if(!lang) lang = getenv("LANG");

  QString etcPath = basePath.str() + "/etc/";

  loadDefaultTranslation(etcPath, lang);
  
  // Load plugins

  // Get path to the plugins
  MJ_Paths &paths = internal->m_paths[MJ_PATH_MUSTAJUURI_PLUGIN];
  uint fields = paths.paths();

  // if(true) {
  if(!fields) {
    
    /* User has not defined MUSTAJUURI_PLUGIN_PATH. Guess the correct
       one. */

    QString version;
#ifndef WIN32
    struct utsname ubuf;
    uname(&ubuf);
#ifdef __OS_IRIX__
    version.sprintf("%s-%s", ubuf.sysname, ubuf.release);
#else
    version.sprintf("%s-%s", ubuf.sysname, ubuf.machine);
#endif

#else
    version = "Windows";
#endif
    QString mjpath(basePath.path(0));

    if(mjpath.isEmpty()) {
      MJ_error(MJ_ERR_ENVIRONMENT, fname, "MUSTAJUURI_PATH is empty");
      return;
    }

    QString newPluginPath(mjpath + "/plugins/" + version);
    paths.addPath(newPluginPath);
    fields = 1;
    MJ_trace2("%s plugin path is now: %s", fname, newPluginPath.ascii());
  }

  QString funcName(MJ_Strs::pluginFactoryFunction);

  QString path;

  MJ_MultiInfoGui * errorGui = 0;

  // Display errors to the user
  if(qApp) {
    QString text(MJ_Tr::tr("Erratic plugin archives"));
    errorGui = new MJ_MultiInfoGui(MJ_Tr::tr("Close"), text);
    // errorGui->setWFlags(Qt::WDestructiveClose);
    errorGui->setIconText(text);
    errorGui->setCaption(text);
  }

  // Load plugins along each path directory.
  for(uint i=0; i < fields; i++) {
    path = paths.path(i);

    if(m_debug) MJ_trace(fname, path);

    QDir d;
    d.setPath(path);
    internal->m_dspLoader->loadDir(d, funcName, false, 0, errorGui);
  }
  
  if(errorGui) {
    if(errorGui->elements()) errorGui->show();
    else delete errorGui;
  }

  QCString tmp(dlerror());
  
  if(!tmp.isEmpty()) {    
    MJ_error(MJ_ERR_INTERNAL, fname, tmp);
  }
  if(errno) {
    MJ_error(MJ_ERR_INTERNAL, fname, strerror(errno));
    errno = 0;
  }

  // Load plugins' translations if possible
  if(qApp) {
    internal->m_dspLoader->loadTranslations(etcPath, lang);
  }
}


MJ_ModuleDspUi *MJ_Kernel::rootUi() 
{ 
  return m_internal->m_rootUi; 
}

void MJ_Kernel::exec(QApplication *qa, int argc, char **argv, bool autostart)
{
  const char * fname = "MJ_Kernel::exec";

  MJ_trace(fname);
  
  MJ_Kernel k(true);

  const char * name = "foo.mj";

  // Open all files that were given as an argument.
  for(int i=1; i < argc; i++) {
    name = argv[i];
  }
  
  QFile file(name);
  
  if(!file.open(IO_ReadOnly)) {
    MJ_error(MJ_ERR_ARGUMENT, fname, name);
    return;
  }
  
  QTextStream stream(&file);
  MJ_ModuleDsp *m = MJ_ModuleDsp::readFromStream(&stream);
  
  if(!m) {
    
    QString base = MJ_Tr::tr("Could not read file");
    QString text = base + ": " + name;

    MJ_error(MJ_ERR_ARGUMENT, fname, text.ascii());
    return;
  }
  else if(m->satellite())
    m->satellite()->pushMruFile(name);

  k.addModule(m);

  MJ_ModuleDspUi ui(0, m);

  if(autostart) {
    k.runNonThreaded(&ui);
  }
  else {
    MJ_AppActionRunner ar;
    QTimer t;
    QObject::connect(&t, SIGNAL(timeout()), &ar, SLOT(runAppActions));
    t.start(50);
    qa->exec();
  }
}

bool MJ_Kernel::runNonThreaded(MJ_ModuleDspUi *rootUi)
{
  Internal *internal = m_internal;

  MJ_ModuleDsp *root = rootUi->module();

  internal->m_status = MJ_KERNEL_NOTHREAD_PLAY;
  root->dspGetTimer()->setRealTime(true);
  internal->m_driver->setModule(root);
  // Very long time ~ infinite
  internal->m_driver->runPeriods((ulong) - 500000); 
  internal->m_driver->setModule(0);
  internal->m_status = MJ_KERNEL_STOP;
  root->dspGetTimer()->setRealTime(false);
  return true;
}

bool MJ_Kernel::startThreaded(MJ_ModuleDspUi *rootUi, 
			      ulong firstFrame, ulong lastFrame)
{
  const char * fname = "MJ_Kernel::startThreaded";

  Internal *internal = m_internal;

  MJ_ModuleDsp *root = rootUi->module();
  
  internal->m_driver->setModule(root);
  internal->m_dspThread = new MJ_RtThread;
  internal->m_dspThread->setDriver(internal->m_driver);
  root->dspGetTimer()->setRealTime(true);
  rootUi->rtStatusChange(MJ_RT_THREADED_RT, &internal->m_dspThread->moduleAccesMutex());
  
  if(!internal->m_dspThread->rtRun()) {
    MJ_error(MJ_ERR_UNKNOWN, fname, "Could not run the thread");
    return false;
  }
  if(!internal->m_dspThread->start(firstFrame, lastFrame)) {
    MJ_error(MJ_ERR_UNKNOWN, fname, "Could not start the thread");
    return false;
  }
  internal->m_status = MJ_KERNEL_THREADED_PLAY;
  internal->m_rootUi = rootUi;
  return true;
}

bool MJ_Kernel::stopThreaded(bool rtThread)
{
  const char * fname = "MJ_Kernel::stopThreaded";

  if(m_debug) MJ_trace(fname);

  Internal *internal = m_internal;

  if(!internal->m_dspThread) {
    return false;
    
  }

  internal->m_dspThread->requestEnd(rtThread);
  if(rtThread) {
    // internal->m_dspThread->moduleAccesMutex().unlock();
    return true;
  }

  internal->m_dspThread->waitEnd();

  internal->m_rootUi->rtStatusChange(MJ_RT_NOT, 0);
  internal->m_rootUi = 0;
  
  delete internal->m_dspThread; 
  internal->m_dspThread = 0;

  internal->m_driver->setModule(0);  
  internal->m_status = MJ_KERNEL_STOP;

  if(m_debug) MJ_trace(fname, "EXIT 2 OK");

  return true;
}

bool MJ_Kernel::killThread()
{
  Internal *internal = m_internal;

  if(!internal->m_dspThread) return false;

  internal->m_dspThread->requestEnd();
  internal->m_dspThread->waitEnd();
  delete internal->m_dspThread; 
  internal->m_dspThread = 0;

  return true;
}

bool MJ_Kernel::killDriver()
{
  Internal *internal = m_internal;

  if(!internal->m_driver) return false; 

  internal->m_driver->setModule(0);

  delete internal->m_driver;
  internal->m_driver = 0;
  return true;
}

/// Init path variables.
void MJ_Kernel::initPaths()
{
  for(uint i=0; i < MJ_PATH_SIZEOF; i++) {
    m_internal->m_paths[i].addPathVariableN((MJ_Path) i);
  }
}

/** Form the plugin tree.
    
    @pre Class members m_internal->m_dspPluginTree and m_internal->m_dspLoader must exist.
*/
void MJ_Kernel::formPLuginTree()
{
  Internal *internal = m_internal;

  internal->m_dspPluginTree->formTree(internal->m_dspLoader->entryList());
  if(qApp)
    internal->m_dspPluginTree->setName(MJ_Tr::tr("Modules"));
  else
    internal->m_dspPluginTree->setName("Modules");
    
  if(m_debug) internal->m_dspPluginTree->print(MJ_cout());
}

void MJ_Kernel::loadDefaultTranslation(const QString &path, const char *lang)
{
  const char *fname = "MJ_Kernel::loadDefaultTranslation";

  Internal *internal = m_internal;

  if(!qApp) return;

  QString file = "mj_base_" + QString(lang) + ".qm";
  QString file2 = "mj_trivial_" + QString(lang) + ".qm";
  
  MJ_Paths p;
  p.addPath(path);

  MJ_trace3(m_debug, "%s # \"%s\"", fname, (const char *) p.str());

  // First load the trivial translations.
  QString trfilePath = p.findFile(file2);

  if(!trfilePath.isEmpty()) {

    MJ_trace3(m_debug, "%s # HALF", fname);

    internal->m_translatorTriv = new QTranslator(0);

    if(!internal->m_translatorTriv->load(trfilePath))
      MJ_error2(MJ_ERR_ARGUMENT, 
		"%s # Could not load trivial translations from \"%s\"",
		fname, trfilePath.ascii());
    else
      qApp->installTranslator(internal->m_translatorTriv);
  }

  MJ_trace3(m_debug, "%s # HALF", fname, (const char *) p.str());

  // Now the non-trivial translations.
  trfilePath = p.findFile(file);

  if(trfilePath.isEmpty()) return;
  
  if(internal->m_translator->load(trfilePath) && qApp) 
    qApp->installTranslator(internal->m_translator);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_AppActionRunner::MJ_AppActionRunner()
{}

MJ_AppActionRunner::~MJ_AppActionRunner()
{

}

void MJ_AppActionRunner::runAppActions()
{
  MJ_Kernel::actions()->operate();
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

class MJ_KernelGui::Internal
{
public:

  Internal(MJ_Kernel *k) : m_kernel(k),
			   m_latencyGui(0),
			   m_errorGui(0),
			   m_errors(0) 
  {
    m_setupGuis.setAutoDelete(false);
    m_guiThread = DI_Thread::myThreadId();
  }

  MJ_Kernel *m_kernel;

  MJ_LatengyGui *m_latencyGui;

  MJ_MultiInfoGui *m_errorGui;

  uint             m_errors;

  QList<QWidget> m_setupGuis;

  QString     m_lastFile;
  QTimer      m_guardTimer;
  QTimer      m_actionTimer;
  bool        m_guardVar;

  static void di_errorHandle(DI_ErrorLevel, const char *);
  static void mj_errorHandle(MJ_ErrorLevel, const char *);
  static DI_Thread::id_t m_guiThread;
};

DI_Thread::id_t MJ_KernelGui::Internal::m_guiThread = 0;

void MJ_KernelGui::Internal::di_errorHandle(DI_ErrorLevel l, const char *str)
{
  if((int) l < DI_ERRL_ERROR)
    return;

  if(DI_Thread::myThreadId() == m_guiThread)
    MJ_KernelGui::addError(str);
  else
    MJ_AppActions::reportError(str);
}

void MJ_KernelGui::Internal::mj_errorHandle(MJ_ErrorLevel l, const char *str)
{
  di_errorHandle((DI_ErrorLevel) l, str);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

bool MJ_KernelGui::m_debug = false;

MJ_KernelGui::MJ_KernelGui(MJ_Kernel    *kernel,
			   QWidget      *parent,
			   const char   *name,
			   bool showNow)
  : QWidget(parent, name),
    m_internal(new Internal(kernel))
{
  const char * fname = "MJ_KernelGui::MJ_KernelGui";

  if(MJ_kernelGui) MJ_fatal(MJ_ERR_INTERNAL, fname);

  if(m_debug) MJ_trace2(fname);

  MJ_kernelGui = this;

  // draw();

  if(!parent && showNow) show();

  setIconText(MJ_mustajuuriStr);
  setCaption(MJ_mustajuuriStr);

  if(MJ_WatchDog::dog()) {
    connect(&m_internal->m_guardTimer, SIGNAL(timeout()), this, SLOT(guardVarUpdate()));
    m_internal->m_guardTimer.start(500);
    m_internal->m_guardVar = true;
    MJ_WatchDog::dog()->setGuard2(&m_internal->m_guardVar);
  }

  connect(&m_internal->m_actionTimer, SIGNAL(timeout()), this, SLOT(runAppActions()));
  m_internal->m_actionTimer.start(111);

  DI_errorHandle = Internal::di_errorHandle;
  MJ_errorHandle = Internal::mj_errorHandle;
}

MJ_KernelGui::~MJ_KernelGui()
{
  MJ_kernelGui = 0;

  if(MJ_WatchDog::dog()) {
    MJ_WatchDog::dog()->setGuard2(0);
  }

  delete m_internal->m_errorGui;
  delete m_internal;

  /* Exit greetings. It is important that the human being who sat by
     Mustajuuri feels happy. Therefore we give him/her a few words as
     we depart. */

  /* 
  QDate date(QDate::currentDate());
  QTime time(QTime::currentTime());

  int day = date.dayOfWeek();

  const char * byeText = 0;

  if(day > 4)
    byeText = "Have a nice weekend";
  else if(time.hour() > 19) 
    byeText = "Good night";
  else 
    byeText = "Bye-bye";

  if(byeText) {
    QString byeStr(MJ_Tr::tr(byeText));

    MJ_cout() << "\n\n\t***  " << byeStr << " ***\n\n";
  }
  */
}

void MJ_KernelGui::exec(QApplication *xpQA, 
			int argc, char **argv, bool autorun)
{
  const char * fname = "MJ_KernelGui::run";

  if(m_debug) MJ_trace2(fname);

  MJ_WatchDog *wd;

  if(MJ_prefs.noWatchDog())
    wd = 0; 
  else {
    wd = new MJ_WatchDog;
    wd->run(true);
  }
  
  if(m_debug) MJ_trace2("%s # Watch dog = %p", fname, wd);

  QString tmp;

  MJ_Kernel *k = new MJ_Kernel(false);

  MJ_KernelGui *kg = new MJ_KernelGui(k, 0, "mj-kernelgui", false);
  k->init();

  kg->draw();
  kg->show();

  xpQA->setMainWidget(kg);
  
  // Open all files that were given as an argument.
  for(int i=1; i < argc; i++) {
    tmp = argv[i];
    
    QFileInfo finfo(tmp);
    if(finfo.exists()) {
      kg->open(tmp);
    }
  }

  if(autorun && kg->m_internal->m_setupGuis.count()) {
    MJ_ModuleDspGui * mg = (MJ_ModuleDspGui *) kg->m_internal->m_setupGuis.first();
    mg->startThreaded();
  }

  xpQA->exec();

  kg->cleanUp();

  delete kg;
  delete k;

  if(wd){
    wd->requestEnd();
    wd->waitEnd();
    delete wd;
  }
}

void MJ_KernelGui::moduleMenu
(QObject *host, const char *slotName, QPopupMenu *menu, 
 MJ_PluginFilter *filter)
{
  if(!MJ_kernel)
    MJ_fatal(MJ_ERR_INTERNAL, "MJ_KernelGui::moduleMenu", "No kernel");

  MJ_PluginDbTree_GQ moduleMenuBuilder;
  moduleMenuBuilder.setNode(MJ_kernel->dspPluginTree());

  moduleMenuBuilder.formTree(host,  slotName, menu, filter);
}


void MJ_KernelGui::open(const QString &name)
{
  const char * fname = "MJ_KernelGui::newModule";

  if(m_debug) MJ_trace(fname);

  m_internal->m_lastFile = name;

  QFile file(name);

  if(!file.open(IO_ReadOnly)) {
    MJ_error(MJ_ERR_ARGUMENT, fname, name);
    return;
  }

  QTextStream stream(&file);
  MJ_ModuleDsp *m = MJ_ModuleDsp::readFromStream(&stream);

  if(!m) {
    
    QString base = MJ_Tr::tr("Could not read file");
    QString text = base + ": " + name;

    MJ_error(MJ_ERR_ARGUMENT, fname, text.ascii());
    
    // QMessageBox::warning(0, base, text, MJ_Tr::tr("Bummer"));
    return;
  }
  else if(m->satellite())
    m->satellite()->pushMruFile(name);
  appendModule(m);
}

void MJ_KernelGui::cleanUp()
{
  uint guis = m_internal->m_setupGuis.count();

  for(uint i=0; i < guis; i++) {
    QWidget * w = m_internal->m_setupGuis.at(0);
    delete w;
  }

  if(m_internal->m_setupGuis.count())
    MJ_error2(MJ_ERR_UNKNOWN, 
	      "MJ_KernelGui::cleanUp # Failed to clear GUIs");
}


MJ_ModuleDspGui *MJ_KernelGui::findGui(MJ_ModuleDsp *m)
{
  if(!MJ_kernelGui) return 0;

  for(uint i=0; i < MJ_kernelGui->m_internal->m_setupGuis.count(); i++) {
    MJ_ModuleDspGui * mg = (MJ_ModuleDspGui *) 
      MJ_kernelGui->m_internal->m_setupGuis.at(i);
    if(mg->module() == m) 
      return mg;
  }
  return 0;
}

bool MJ_KernelGui::addError(const QString &errorString)
{
  if(!MJ_kernelGui || MJ_prefs.noGuiWarnings()) return false;
  
  MJ_kernelGui->addErrorReal(errorString);

  return true;
}

void MJ_KernelGui::open()
{
  const char * fname = "MJ_KernelGui::open";

  if(m_debug) MJ_trace(fname);

  QString dir(QDir::currentDirPath());

  QFileInfo fi(m_internal->m_lastFile);
  if(fi.exists())
    dir = fi.dirPath();

  QStringList s
    (QFileDialog::getOpenFileNames("*.mj", dir));
  // do something with the files in s.
  // s.setAutoDelete(true);
  
  for(uint i=0; i < s.count(); i++) {
    if(m_debug) {
      char buf[128];
      sprintf(buf, "%s # file", fname);
      MJ_trace(buf, s[i]);
    }
    open(s[i]);
  }
}

void MJ_KernelGui::quit()
{
  if(m_debug) MJ_trace("MJ_KernelGui::quit");

  if(!m_internal->m_kernel->tryToStop()) return;
  
  qApp->quit(); 
}

/** Adds a new module to the list of managed modules. A GUI will be
    created for the module. */
void  MJ_KernelGui::appendModule(MJ_ModuleDsp *newBlock)
{
  m_internal->m_kernel->addModule(newBlock);

  MJ_ModuleDspGui *newGui = 
    newBlock->newGui(0, MJ_GUI_OWN_WINDOW, 0, 0);

  if(newGui) {
    newGui->show();
    m_internal->m_setupGuis.append(newGui);
    if(m_debug) 
      MJ_trace("MJ_KernelGui::appendModule # GUI class name",
	       typeid(*newGui).name());
  }
}

/** Deletes a setup. Th GUI <B>is not</B> deleted. The module <B>is</B>
    deleted. */
void MJ_KernelGui::deleteSetup(QWidget *w)
{
  const char * fname = "MJ_KernelGui::deleteSetup";

  MJ_trace3(m_debug, "%s # %p", fname, w);

  int index = m_internal->m_setupGuis.find(w);

  if(index == -1) {
    MJ_error2(MJ_ERR_ARGUMENT, "%s # No GUI %p", fname, w);
    return;
  }

  m_internal->m_setupGuis.remove(index);
  // m_internal->m_setupGuis.take(index);
  m_internal->m_kernel->removeModule(index);
  
  MJ_trace3(m_debug, "%s # EXIT", fname);
}

/** Reports memory management status of various object types. 
    
    This method is used mainly for memory management debugging.
*/
void MJ_KernelGui::memReport()
{
  MJ_memDebugReport();
}

/** Bring up the latency editor window. */

void MJ_KernelGui::editLatency()
{
  if(m_internal->m_latencyGui) {
    m_internal->m_latencyGui->show();
    m_internal->m_latencyGui->raise();
  }
  else {
    m_internal->m_latencyGui = new MJ_LatengyGui(MJ_prefs.minimumLatency(), 
				     MJ_prefs.maximumLatency(),
				     0);

    connect(m_internal->m_latencyGui, SIGNAL(newConf(float, float)), 
	    this, SLOT(latencyConf(float, float)));
    m_internal->m_latencyGui->show();
  }
}

/** Set the target latency. */
void MJ_KernelGui::latencyConf(float minLat, float maxLat)
{
  MJ_prefs.setLatencies(minLat, maxLat);
}

/** Show a window displaying information about all loaded plugins.*/
void MJ_KernelGui::pluginInfos()
{
  MJ_PluginsInfo *pi = new MJ_PluginsInfo(m_internal->m_kernel->dspLoader());
  pi->show();
}

/** Reports the priority of the calling thread. */
void MJ_KernelGui::myprio()
{
#ifndef WIN32
  pid_t pid = getpid();

  int policy = sched_getscheduler(pid);
  const char * pstr = "foo";

  switch(policy) {
  case SCHED_FIFO:
    pstr = "SCHED_FIFO";
    break;
  case SCHED_RR:
    pstr = "SCHED_RR";
    break;
  case SCHED_OTHER:
    pstr = "SCHED_OTHER";
    break;
  }

  MJ_trace2("MJ_KernelGui::myprio priority = %s, this is %s", 
	    pstr, policy == SCHED_OTHER ? "good" : "bad");
  
  // pthread_t pt = pthread_self();

  struct sched_param sp;
  sp.sched_priority = 5;
  
  sched_setscheduler(pid, SCHED_OTHER, &sp);
#endif
}

void MJ_KernelGui::addModule(const QString &id)
{
  const char * fname = "MJ_KernelGui::addModule";

  if(m_debug) MJ_trace(fname, id);
  
  MJ_PluginDbeBase *dbe = 
    m_internal->m_kernel->dspLoader()->entryList().findDbe(id, MJ_PP_MENU_POSITION);
    
  if(!dbe) {
    MJ_error(MJ_ERR_ARGUMENT, fname, "NULL dbe");
    return;
  }
    
  MJ_PluginDbeDsp *dbeDsp = dynamic_cast<MJ_PluginDbeDsp *> (dbe);
  // MJ_PluginDbeDsp *dbeDsp = (MJ_PluginDbeDsp *) (dbe);
    
  if(!dbeDsp) {
    MJ_error(MJ_ERR_ARGUMENT, fname, "Bad dbe");
    return;
  }
    
  if(m_debug) MJ_trace(fname, "Found entry. Creating block");

  MJ_ModuleDsp *newBlock = dbeDsp->newPlugin();
    
  if(!newBlock) return;
  // newBlock->toCout();
  appendModule(newBlock);
}

/** Updates the guard variable. The guard variable is updated to let
    the watch-dig thread know that the user interface is still
    alive.
*/
void MJ_KernelGui::guardVarUpdate()
{
  m_internal->m_guardVar = true;
}

/** Reports the number of plugin entries in the data-base. */
void MJ_KernelGui::nofPlugins()
{
  MJ_trace2("MJ_KernelGui::nofPlugins %d", 
	    MJ_kernel->dspLoader()->entryListConst().count());
}

/** Dumps a HTML-formatted tree of the plugins to the screen.*/
void MJ_KernelGui::htmlDump()
{
  MJ_kernel->dspPluginTree()->print(MJ_cout(), 0, true);
}

void MJ_KernelGui::runAppActions()
{
  MJ_Kernel::actions()->operate();
}

void MJ_KernelGui::errorGuiDeleted()
{
  m_internal->m_errorGui = 0;
}

void MJ_KernelGui::addErrorReal(const QString &errorString)
{
  m_internal->m_errors++;

  if(!m_internal->m_errorGui) {
    m_internal->m_errorGui = 
      new MJ_MultiInfoGui(MJ_Tr::tr("Close"), MJ_Tr::tr("Error"));
    connect(m_internal->m_errorGui, SIGNAL(destroyed()), 
	    this, SLOT(errorGuiDeleted()));
  }
  
  QString totalError(MJ_Tr::tr("Failure %1 :\n").arg(m_internal->m_errors));
  // QString totalError(MJ_Tr::tr("Failure"));

  totalError += errorString;

  m_internal->m_errorGui->addInfo(totalError);
  
  m_internal->m_errorGui->raise();
  m_internal->m_errorGui->show();
}

void MJ_KernelGui::draw()
{
  if(m_debug) MJ_trace("MJ_KernelGui::draw");

  QVBoxLayout *layout = new QVBoxLayout(this);

  QMenuBar *menu = new QMenuBar( this );

  QPopupMenu *setups = new QPopupMenu(menu);
  MJ_PluginDbTree_GQ moduleMenuBuilder;
  moduleMenuBuilder.setNode(m_internal->m_kernel->dspPluginTree());
  
  MJ_DspPluginFilterMetaFunc 
    filter(MJ_DspPluginMetaFunctionalityKeys[MJ_DPMF_STAND_ALONE]);

  moduleMenuBuilder.formTree
    (this, SLOT(addModule(const QString &)), setups, & filter);
  
  QPopupMenu *file = new QPopupMenu(menu);
  
  file->insertItem( MJ_Tr::tr("&New"), setups );  
  file->insertItem( MJ_Tr::tr("&Open..."), this, SLOT(open()), CTRL + Key_O);
  file->insertItem( MJ_Tr::tr("&Quit"), this, SLOT(quit()), CTRL + Key_Q);
    
#ifdef MJ_MEM_DEBUG
  if(m_debug) 
    file->insertItem( MJ_Tr::tr("Memory report"), this, SLOT(memReport())); 
#endif

  menu->insertItem( MJ_Tr::tr("&File"), file );  

  // General options
  QPopupMenu *options = new QPopupMenu(menu);
  options->insertItem(MJ_Tr::tr("&Latency..."), this, SLOT(editLatency()));  
  options->insertItem(MJ_Tr::tr("&Plugin info..."), this, SLOT(pluginInfos()));
  options->insertItem(MJ_Tr::tr("&Nof plugins"), this, SLOT(nofPlugins()));
  options->insertItem(MJ_Tr::tr("&HTML dump"), this, SLOT(htmlDump()));

#ifdef __OS_LINUX__
  options->insertItem(MJ_Tr::tr("My prio ? (for debug only)"), 
		      this, SLOT(myprio()));
#endif // __OS_LINUX__
  
  menu->insertItem( MJ_Tr::tr("&Setup"), options );  

  layout->setMenuBar(menu);
  layout->activate();
}

void MJ_KernelGui::closeEvent(QCloseEvent * e)
{
  if(m_debug) MJ_trace("MJ_KernelGui::closeEvent");  

  if(!m_internal->m_kernel->tryToStop()) return;

  QWidget::closeEvent(e);
}
