/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */


#ifndef MJ_KERNEL_H
#define MJ_KERNEL_H

#include <mj_plugin_base.h>
#include <mj_module_dsp.h>
#include <mj_module_dsp_ui.h>
#include <mj_paths.h>
#include <mj_run_status.h>

#include <qwidget.h>

class MJ_AppActions;
class MJ_LatengyGui;
class MJ_MultiInfoGui;
class MJ_RtDriver;
class MJ_RtThread;
class MJ_PluginDbeDspLoader;
class MJ_PluginDbTreeNodeMany;
class MJ_PluginFilter;

class QApplication;
class QPopupMenu;
class QTranslator;

/** Mustajuuri state. */
enum MJ_KernelStatus {
  MJ_KERNEL_STOP,
  MJ_KERNEL_NOTHREAD_PLAY,
  MJ_KERNEL_THREADED_PLAY,
  MJ_KERNEL_SIZEOF
};

extern const char *MJ_kernelStatusWords[MJ_KERNEL_SIZEOF];

/** The heart of MJ. The true driver. The ultimate top-level. There
    can only be one. 

    It is recommended that you initialize qApp before creating an
    object of this kind.  */
class MJ_Kernel
{
public:
  MJ_Kernel(bool autoInit = true);

  virtual ~MJ_Kernel();

  static void setEnv();

  void init();

  void addModule(MJ_ModuleDsp *m);
  
  MJ_ModuleDsp *forgetModule(uint index);

  bool removeModule(uint index);

  bool playModule(MJ_ModuleDspUi *rootUi, 
		  MJ_KernelStatus method, 
		  ulong firstFrame = 0, ulong lastFrame = 0);
  
  bool stopPlaying(bool rtThread = false);

  bool tryToStop();

  void panic();

  /// The current DSP thread (NULL if inactive)
  MJ_RtThread *dspThread();

  static MJ_ModuleDsp *newDspModule
  (const QString &identifier,
   MJ_PLuginParameter searchCriterion);

  /// A new module with given functionality requirements
  static MJ_ModuleDsp *newFunctionalModule(const QString &);

  static MJ_PluginDbeDsp *findDspPlugin
    (const QString &identifier,
     MJ_PLuginParameter searchCriterion);

  MJ_KernelStatus status() const;

  /// Get one of the path variables.
  MJ_Paths &path(uint i);


  static MJ_AppActions *actions();
  MJ_PluginDbeDspLoader *dspLoader();

  MJ_PluginDbTreeNodeMany *dspPluginTree();

  void loadDefaultDirs();

  MJ_ModuleDspUi *rootUi();
  
  /// Sets the debugging mode.
  static void setDebug(bool b) { m_debug = b; }

  static void exec(QApplication *, int argc, char **argv, bool autostart);

protected:

  bool runNonThreaded(MJ_ModuleDspUi *);
  bool startThreaded(MJ_ModuleDspUi *, 
		     ulong firstFrame, ulong lastFrame);
  bool stopThreaded(bool);
  bool killThread();
  bool killDriver();

  void initPaths();
  void formPLuginTree();
  void loadDefaultTranslation(const QString &path, const char *lang);

  class Internal;
  Internal *m_internal;
  
  static bool m_debug;
private:
  /// Disabled.
  MJ_Kernel(MJ_Kernel &) {}
  MJ_Kernel & operator = (const MJ_Kernel &) { return * this; }
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

class MJ_AppActionRunner : public QObject
{
  Q_OBJECT;
public:
  MJ_AppActionRunner();
  virtual ~MJ_AppActionRunner();

public slots:
  void runAppActions();
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/// GUI for MJ_Kernel
class MJ_KernelGui : public QWidget
{
  Q_OBJECT
public:
  MJ_KernelGui(MJ_Kernel    *kernel,
	       QWidget      *parent = 0,
	       const char   *name = 0,
	       bool showNow = true);

  virtual ~MJ_KernelGui();

  /** Executes the application. */

  static void exec(QApplication *, 
		   int argc, char **argv, bool autorun = false);
  static void setDebug(bool b) { m_debug = b; }

  /** Buils a module menu based on the tree of plugins. Also connects
      signals from menu entries to the target object.
      
      The modules may also be filtered. If no filter is set default
      filter (MJ_DspPluginFilterChanable) will be used.  */
  static void moduleMenu(
			 /// Host object (receives the signals
			 QObject *host,
			 /// Slot to connect the signals to
			 const char *slotName, 
			 /// Target menu
			 QPopupMenu *menu,
			 /// Filter to be used to select modules
			 MJ_PluginFilter *filter = 0);
  

  void appendModule(MJ_ModuleDsp *);

  void open(const QString &);

  void cleanUp();

  static MJ_ModuleDspGui *findGui(MJ_ModuleDsp *);

  static bool addError(const QString &);

public slots:
  void open();
  void quit();

  void addModule(const QString &);

  void deleteSetup(QWidget *);
  
  void memReport();

  void editLatency();

  void latencyConf(float, float);

  void pluginInfos();

  void myprio();

  void guardVarUpdate();

  void nofPlugins();

  void htmlDump();

  void runAppActions();

  void errorGuiDeleted();

protected:

  void addErrorReal(const QString &);

  void draw();

  void closeEvent(QCloseEvent * e);

  class Internal;
  Internal * m_internal;

  static bool m_debug;
};

#endif
