/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_malloc.h>

#include <mj_error.h>

#include <stdlib.h>
#include <string.h>

/** Create a MJ_MallocUnit and allocate the requested amount of
    memory. Also fills the allocated memory with zero. */

MJ_Malloc::Unit::Unit(ulong bytes)
  : m_data(0),
    m_used(0)
{
  // Pad to fill eight-byte segmenting.
  if(bytes % 8)
    bytes += (8 - bytes & 0x7);

  if(bytes) {
    m_data = malloc(bytes);
    if(!m_data) {
      MJ_error2(MJ_ERR_MEMORY, "MJ_Malloc::Unit::Unit # %d", (int) bytes);
      m_bytes = 0;
    }
    else {
      memset(m_data,0,bytes);
      m_bytes = bytes;
    }
  }
}

/** Frees all allocated memory. */
MJ_Malloc::Unit::~Unit()
{
  ::free(m_data);
}

/** Allocate memory. The memory is allocated in 8-byte blocks. */
void *MJ_Malloc::Unit::allocate(ulong bytes)
{
  ulong overBytes = bytes & 0x7;
  
  // Pad to fill eight-byte segmenting.
  bytes += (8 - overBytes);

  if(bytes > available()) return 0;

  char *ptr = & (((char *) m_data) [m_used]);
  m_used += bytes;
  return ptr;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** Allocate memory. If the memory cannot be allocated a fatal error
    is generated. 

    If the argument "bytes" is zero then NULL-pointer is returned. 
*/
void *MJ_Malloc::allocate(ulong bytes)
{
  if(!bytes) return 0;

  // Seek from already allocated memory.
  for(uint i=0; i < m_units.count(); i++) {
    void *ptr = m_units.at(i)->allocate(bytes);
    if(ptr) return ptr;
  }
  
  ulong allocBytes = Unit::defaultSize();
  
  while (allocBytes < bytes) {
    allocBytes += (Unit::defaultSize() / 2);
  }
  Unit *mu = new Unit(allocBytes);

  uint gotBytes = mu->totalBytes();

  if(gotBytes < allocBytes) {
    MJ_error2(MJ_ERR_MEMORY, "MJ_Malloc::allocate # %d < %d (%d)", 
	      gotBytes, (int) allocBytes, (int) bytes);

    MJ_fatal(MJ_ERR_MEMORY, "MJ_Malloc::allocate");
  }

  /* Instead of appending the unit to the list we prepend it. It is
     highly likely that this unit contains extra memory that can be
     used during the next call. */
  m_units.insert(0, mu); 

  return mu->allocate(bytes);
}
