/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_message_carrier.h>

#include <mj_error.h>
#include <mj_global.h>
#include <mj_iostream.h>
#include <mj_message_factory.h>
#include <mj_message_qio.h>
#include <mj_message_xml.h>
#include <mj_messages.h>
#include <mj_module_dsp.h>
#include <mj_misc.h>
#include <mj_parameters.h>
#include <mj_stream_utils.h>

#include <di_mutex.h>

#include <typeinfo>

static DI_MutexAuto MJ_msgCountMutex;

const char *MJ_MessageCarrier::m_label = "MJ_MESSAGE_ID";

int MJ_MessageCarrier::m_counter = 0;

MJ_MessageCarrier::MJ_MessageCarrier()
  : m_id(0),
    m_msg(0)
{
#ifdef MJ_MEM_DEBUG
  incrCount();
#endif
}

MJ_MessageCarrier::MJ_MessageCarrier(const char *id)
  : m_id(0),
    m_msg(0)
{
  setId(id);
#ifdef MJ_MEM_DEBUG
  incrCount();
#endif
}

MJ_MessageCarrier::MJ_MessageCarrier(const char *id, MJ_Message *m)
  : m_id(0),
    m_msg(m->duplicate())
{
  setId(id);
#ifdef MJ_MEM_DEBUG
  incrCount();
#endif  
}

MJ_MessageCarrier::MJ_MessageCarrier(const MJ_MessageCarrier &i)
  : m_id(0)
{ 
  setId(i.m_id);

  if(i.m_msg) m_msg = i.m_msg->duplicate();
  else m_msg = 0;

#ifdef MJ_MEM_DEBUG
  incrCount();
#endif
}

MJ_MessageCarrier::~MJ_MessageCarrier()
{
#ifdef MJ_MEM_DEBUG
  decrCount();
#endif
  delete m_msg;

  free(m_id);
}


void  MJ_MessageCarrier::setId(const char *s)
{ 
  MJ_delAndDup(& m_id, s); 
}

void MJ_MessageCarrier::setMessage(MJ_Message *msg)
{ 
  delete m_msg;
  m_msg = msg;
}

void MJ_MessageCarrier::dropData()
{
  m_id = 0;
  m_msg = 0;
}

/// Duplicate this message
MJ_MessageCarrier *MJ_MessageCarrier::duplicate()
{
  return new MJ_MessageCarrier(*this);
}

/// Write to a steam
bool MJ_MessageCarrier::write(QTextStream *s)
{
  (*s) << m_label << " " << (int) (m_msg != 0) << " " 
       << m_id << "\n";
  if(m_msg) return MJ_MessageFactory::write(s, m_msg);
  return true;
}

/// Read from stream
bool MJ_MessageCarrier::read(QTextStream *s)
{
  bool ok = true;
  QString str;

  (*s) >> str;

  if(str !=  m_label) {
    MJ_error(MJ_ERR_DATA_CORRUPTED, "MJ_MessageCarrier::read", str);
    return false;
  }
  
  (*s) >> str;
  bool hasMessage = (bool) str.toInt(&ok);
  if(!ok) return false;

  // MJ_trace(str);

  (*s) >> str;
  
  if(str.isEmpty())
    ok = false;
  else {
    MJ_delAndDup(&m_id, str);
    // MJ_trace("MJ_MessageCarrier::read New ID", m_id);
  }

  if(hasMessage) {
    if(m_msg) delete m_msg;
    m_msg = MJ_MessageQio::readFromStream(s);
    if(!m_msg) ok = false;
  }
  return ok;
}

QDomElement MJ_MessageCarrier::write(QDomDocument &doc)
{
  QDomElement e = doc.createElement("MessageCarrier");
  QDomElement e2 = doc.createElement("Msg");
  
  MJ_DomUtil::addTextChild(e, doc, "id", m_id);

  if(!m_msg)
    return e;

  e2.appendChild(MJ_MessageXml::writeMsg(doc, m_msg));
  e.appendChild(e2);

  return e;
}

bool MJ_MessageCarrier::read(QDomElement &e)
{
  if(e.tagName() != "MessageCarrier")
    return false;

  delete m_msg;

  setId(MJ_DomUtil::getString(e, "id").ascii());

  QDomElement message = e.namedItem("Msg").toElement();

  m_msg = MJ_MessageXml::readMsg(message);

  return m_msg != 0;
}

void MJ_MessageCarrier::toCout()
{
  write(&MJ_cout());
  MJ_cout().device()->flush();
}

MJ_MessageCarrier & MJ_MessageCarrier::operator = (const MJ_MessageCarrier &m)
{
  delete m_msg;

  if(m.m_msg) m_msg = m.m_msg->duplicate();
  else m_msg = 0;

  free(m_id);

  m_id = m.m_id ? strdup(m.m_id) : 0;
  
  return *this;
}

void MJ_MessageCarrier::incrCount()
{
  MJ_msgCountMutex.lock();
  m_counter++;
  MJ_msgCountMutex.unlock();
}

void MJ_MessageCarrier::decrCount()
{
  MJ_msgCountMutex.lock();
  m_counter--;

  if(m_counter < 0)
    MJ_fatal(MJ_ERR_INTERNAL, "MJ_MessageCarrier::decrCount");

  MJ_msgCountMutex.unlock();
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

const char *MJ_MessageList::m_label = "MJ_MESSAGE_LIST";

bool MJ_MessageList::m_debug = false;

bool MJ_MessageList::copyValues(MJ_Parameters *ps)
{
  QListIterator<MJ_Parameter> iter(*ps);

  MJ_Parameter *p;

  while((p = iter.current())) {
    MJ_MessageCarrier *m = p->valueMessage();
    if(m) append(m);
    ++iter;
  }

  return true;
}

bool MJ_MessageList::writeMessages(QTextStream *s)
{
  QListIterator<MJ_MessageCarrier> iter(*this);  
 
  MJ_MessageCarrier *p;

  (*s) << m_label << " " << count() << "\n";

  while((p = iter.current())) {
    if(!p->write(s)) return false;
    ++iter;
  }

  return true;
}

bool MJ_MessageList::readMessages(QTextStream *s)
{  
  const char *fname = "MJ_MessageList::readMessages";

  if(m_debug) MJ_trace(fname);
  
  bool ok = true;
  int n;
  QString str;
  
  (*s) >> str;
  if(str != m_label) return false;
  
  str = s->readLine();
  n = str.toInt(&ok);
  if(!ok) return false;
  
  for(int i=0; i < n; i++) {
    MJ_MessageCarrier *mi = new MJ_MessageCarrier;
    if(!mi) return false;
    bool miOk = mi->read(s);

    if(m_debug) mi->toCout();

    if(!miOk) {
      MJ_error2(MJ_ERR_DATA_CORRUPTED, "%s # index = %d, near line %d",
		fname, i, MJ_currentLine(s->device()));
      delete mi;
      return false;
    }

    append(mi);
  }
  return true;
}

bool MJ_MessageList::writeMessages(QDomElement &e, QDomDocument &doc)
{
  for(uint i=0; i < count(); i++)
    e.appendChild(at(i)->write(doc));

  return true;
}

bool MJ_MessageList::readMessages(QDomElement &e)
{
  QDomNode n = e.firstChild();
  
  while ( !n.isNull() ) {
    QDomElement n2 = n.toElement();
    
    MJ_MessageCarrier * mc = new MJ_MessageCarrier;

    if(mc->read(n2))
      append(mc);
    else
      delete mc;
    
    n = n.nextSibling();
  }

  return true;
}

bool MJ_MessageList::deliverMessages(MJ_ModuleDsp *m)
{
  QListIterator<MJ_MessageCarrier> iter(*this);  
 
  MJ_MessageCarrier *p;
  bool ok = true;

  while((p = iter.current())) {

#if 0
    cout << "Delivering message:\n";
    p->toCout();
#endif

    bool ok2 = m->msgProcess(p->id(), p->message());
    
    if(!ok2 && m_debug) 
      MJ_error(MJ_ERR_ARGUMENT, "MJ_MessageList::deliverMessages",
	       typeid(*m).name(), p->id());
    
    ok = ok && ok2;
    ++iter;
  }

  return ok;
}

int MJ_MessageList::compareItems (void *v1, void *v2)
{
  MJ_Message *m1 = ((MJ_MessageCarrier *) v1)->message();
  MJ_Message *m2 = ((MJ_MessageCarrier *) v2)->message();

  if(!m1 || !m2) return 0;  // Undefined
  
  if(m1->timeStamp() > m2->timeStamp()) return -1;
  if(m1->timeStamp() == m2->timeStamp()) return 0;

  return 1;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

const char *MJ_MessageArray::m_label = "MJ_MESSAGE_ARRAY";

bool MJ_MessageArray::m_debug = false;

MJ_MessageArray::~MJ_MessageArray()
{
  empty();
}

bool MJ_MessageArray::resize(uint size, bool fatalOnFailure)
{
  if(m_debug) MJ_trace("MJ_MessageArray::resize");

  for(uint i = size; i < m_size; i++) {
    if(m_array[i]) {
      delete m_array[i];
      m_array[i] = 0;
    }
  }

  uint oldSize = m_size;

  bool b = MJ_Array<MJ_MessageCarrier *>::resize(size, fatalOnFailure);

  if(b) {
    for(uint i = oldSize; i < size; i++) {
      m_array[i] = 0;
    }
  }
  return b;
}

bool MJ_MessageArray::copyValues(MJ_Parameters *ps)
{
  if(m_debug) MJ_trace("MJ_MessageArray::copyValues");  

  QListIterator<MJ_Parameter> iter(*ps);

  MJ_Parameter *p;
  uint i=0; 

  if(m_size) empty();  
  resize(ps->count());
  setAll(0);

  while((p = iter.current())) {

    if(m_debug) MJ_trace("MJ_MessageArray::copyValues", (int) i);  

    MJ_MessageCarrier *m = p->valueMessage();

    if(m) {
      MJ_MessageCarrier *&old = get(i);
      if(old) delete(old);
      old = m;
    }

    ++iter;
    ++i;
  }

  return true;
}
  
/** Duplicates the given message to the array element.  If the old
    object is non-null it is deleted. */
void MJ_MessageArray::setMessage(uint index, MJ_MessageCarrier *msg)
{
  MJ_MessageCarrier *&old = get(index);
  if(old) delete(old);
  
  if(!msg) old = 0;
  else {
    old = msg->duplicate();
  }
}
/// Delete all elements, but do not resize the array.
void MJ_MessageArray::empty() 
{
  if(m_debug) MJ_trace("MJ_MessageArray::empty", (int) m_size);
  
  MJ_MessageCarrier **array = m_array;
  MJ_MessageCarrier **arrayMax = &array[m_size];

  while(array < arrayMax) {
    if(*array) {
      delete *array;
      *array = 0;
    }
    array++;
  }
}
