/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_message_pass.h>

#include <mj_app_action.h>
#include <mj_message.h>
#include <mj_module_dsp.h>
#include <mj_strs.h>

#include <qtextstream.h> 
#include <qstring.h> 

#include <stdlib.h> 

const char *MJ_MessageRemap::m_key = "MSG_REMAP";

static const char * m_nil = "NULL";


MJ_MessageRemap::MJ_MessageRemap()
 : m_from(0), 
  m_to(0), 
  m_target(0), 
  m_sourceId(0),
  m_targetId(0)
{}

/** Sets all data necessary for this remap object. */
void MJ_MessageRemap::set(const char * from, 
			  const char * to, 
			  const char * sourceId, 
			  const char * targetId)
{
  if(MJ_Strs::equals(from, m_nil))     from = 0;
  if(MJ_Strs::equals(to, m_nil))       to = 0;
  if(MJ_Strs::equals(sourceId, m_nil)) sourceId = 0;
  if(MJ_Strs::equals(targetId, m_nil)) targetId = 0;

  MJ_delAndDup(&m_from, from);
  MJ_delAndDup(&m_to, to);
  MJ_delAndDup(&m_sourceId, sourceId);
  MJ_delAndDup(&m_targetId, targetId);
}

MJ_MessageRemap::~MJ_MessageRemap()
{
  free(m_from);
  free(m_to);
  free(m_sourceId);
  free(m_targetId);
}

void MJ_MessageRemap::setTarget(MJ_ModuleDsp *t)
{ 
  m_target = t; 
  free(m_targetId);
  if(t)
    m_targetId = strdup(t->id());
  else
    m_targetId = 0;
}

/** Delivers the message if the "id" matches the "from" -address and
    there is a non-null target module. If the conditions are not met
    (fully ok), then nothing happens.

    This method also sends a notification to the user interface thread
    via MJ_AppActions::moduleModified -function. The assumption is
    that otherwise GUI would not know anything about what parameter
    changes take place in the DSP engine.
*/
bool MJ_MessageRemap::deliver(const char * id, MJ_Message *m)
{
  if(!m_target) return false;

  bool ok = false;

  if(!id) {
    bool ok = m_target->msgProcess(m_to, m);
    
    if(ok) 
      MJ_AppActions::moduleModified(m_target, m_to);    
  }
  if(!m_from) {
    bool ok = m_target->msgProcess(m_to ? m_to : id, m);
    
    if(ok) 
      MJ_AppActions::moduleModified(m_target, m_to ? m_to : id);
  }
  else if(MJ_Strs::equals(id, m_from)) {
    bool ok = m_target->msgProcess(m_to ? m_to : id, m);
    
    if(ok) 
      MJ_AppActions::moduleModified(m_target, m_to ? m_to : id);
  }

  return ok;
}

/** Tests if the objects equal. Compares all member variables
    besides the pointers. */
bool MJ_MessageRemap::equals(const MJ_MessageRemap &r)
{

  return (MJ_Strs::equals(m_from, r.m_from) &&
	  MJ_Strs::equals(m_to, r.m_to) &&
	  MJ_Strs::equals(m_sourceId, r.m_sourceId) &&
	  MJ_Strs::equals(m_targetId, r.m_targetId));
}

/** Write to a steam. */
bool MJ_MessageRemap::write(QTextStream *s)
{
  (*s) << m_key << " " 
       << (m_sourceId ? m_sourceId : m_nil) << " " 
       << (m_targetId ? m_targetId : m_nil) << " "
       << (m_from ? m_from : m_nil) << " " 
       << (m_to ? m_to : m_nil) << "\n";
  
  return true;
}


/** Read from a stream. If the read fails the object keeps its
    original state. */
bool MJ_MessageRemap::read(QTextStream *s)
{
  QCString sourceId, targetId, from, to;

  QString tmp;

  tmp = s->readLine();

  if(MJ_field(tmp, 0) != m_key) return false;

  sourceId = MJ_field(tmp, 1);
  targetId = MJ_field(tmp, 2);
  from = MJ_field(tmp, 3);
  to = MJ_field(tmp, 4);

  if(sourceId.isEmpty() || targetId.isEmpty() || 
     from.isEmpty() || to.isEmpty())
    return false;

  set(from, to, sourceId, targetId);

  return true;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

int MJ_MessageRemapList::findTarget(MJ_ModuleDsp *m)
{
  if(!m) return -1;

  for(uint i=0; i < count(); i++) {
    if(at(i)->target() == m)
      return (int) i;
  }
  return -1;
}

/* int MJ_MessageRemapList::findTarget(int id)
{
  for(uint i=0; i < count(); i++) {
    if(at(i)->sourceId() == id)
      return (int) i;
  }
  return -1;  
  }*/

int MJ_MessageRemapList::findRemap(MJ_MessageRemap *rm)
{
  if(!rm) return -1;

  for(uint i=0; i < count(); i++) {
    if(at(i)->equals(*rm))
      return (int) i;
  }
  return -1;
}

