/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_message_qio.h>

#include <mj_error.h>
#include <mj_iostream.h>
#include <mj_message_factory.h>
#include <mj_message_multi2.h>
#include <mj_message_string.h>
#include <mj_messages.h>
#include <mj_midi.h>
#include <mj_stream_utils.h>

#include <di_data_buffer.h>

#include <qtextstream.h>

bool MJ_MessageQio::writeMsg(QTextStream *s, MJ_Message *m)
{
  (*s) << MJ_MessageFactory::determineName(m->messageType()) << " " 
       << m->timeStamp() << " ";
  return true;
}

bool MJ_MessageQio::readMsg(QTextStream *s, MJ_Message *m, int readFields)
{
  const char * fname = "MJ_Message::read";

  if(s->eof()) return false;

  bool ok = true;
  QString c;

  if(readFields < 1) {
    (*s) >> c;
    const char * name = MJ_MessageFactory::determineName(m->messageType());
    if(c != name) {
      MJ_error2(MJ_ERR_ARGUMENT, "%s # Type mismatch %s != %s",
		fname, c.ascii(), name);
      return false;
    }
  }

  if(readFields < 2) {
    (*s) >> c;
    m->setTimeStamp(c.toLong(&ok));
  }

  // Skip a char:
  // QIODevice * d = s->device();
  // if(!d->at(d->at() + 1)) return false;

  return ok;
}

bool MJ_MessageQio::writeMsgBool(QTextStream *s, MJ_Message *m)
{
  if(!MJ_MessageQio::writeMsg(s, m)) return false;
  
  MJ_MessageBool *mb = (MJ_MessageBool *) (m);

  (*s) << (int) mb->value() << "\n";
  return true;  
}

bool MJ_MessageQio::readMsgBool(QTextStream *s, MJ_Message *m, int readFields)
{
  if(!MJ_MessageQio::readMsg(s, m, readFields)) return false;
  
  int tmp;
  (*s) >> tmp;
  
  MJ_MessageBool *mb = (MJ_MessageBool *) (m);

  mb->setValue((bool) tmp);
  s->readLine();

  // MJ_trace2("MJ_MessageBool::read m_val = %d", (int) m_val);

  return true;
}


bool MJ_MessageQio::writeMsgDouble(QTextStream *s, MJ_Message *msg)
{
  if(!MJ_MessageQio::writeMsg(s, msg)) return false;

  MJ_MessageDouble *tmp = (MJ_MessageDouble *) (msg);

  (*s) << tmp->value() << "\n";
  return true;
}

bool MJ_MessageQio::readMsgDouble(QTextStream *s, MJ_Message *msg, int readFields)
{
  if(!MJ_MessageQio::readMsg(s, msg, readFields)) return false;
  
  QString tmp(s->readLine());

  MJ_MessageDouble *mi = (MJ_MessageDouble *) (msg);
  
  bool ok = true;
  mi->setValue(tmp.toDouble(&ok));

  return ok;
}

bool MJ_MessageQio::writeMsgInt(QTextStream *s, MJ_Message *msg)
{
  if(!MJ_MessageQio::writeMsg(s, msg)) return false;

  MJ_MessageInt *tmp = (MJ_MessageInt *) (msg);

  (*s) << tmp->value() << "\n";
  return true;
}

bool MJ_MessageQio::readMsgInt(QTextStream *s, MJ_Message *msg, int readFields)
{
  if(!MJ_MessageQio::readMsg(s, msg, readFields)) return false;
  
  QString tmp(s->readLine());

  MJ_MessageInt *mi = (MJ_MessageInt *) (msg);
  
  bool ok = true;
  mi->setValue(tmp.toInt(&ok));

  return ok;
}

bool MJ_MessageQio::writeMsgMidi(QTextStream *s, MJ_Message *m)
{
  if(!MJ_MessageQio::writeMsg(s, m)) return false;
  
  MJ_MessageMidiEvent *mm = (MJ_MessageMidiEvent *) (m);

  (*s) << mm->midiMsg() << "\n";
  return true;

}

bool MJ_MessageQio::readMsgMidi(QTextStream *s, MJ_Message *m, int readFields)
{
  if(!MJ_MessageQio::readMsg(s, m, readFields)) return false;
  
  MJ_MessageMidiEvent *mm = (MJ_MessageMidiEvent *) (m);

  (*s) >> mm->midiMsg();

  s->readLine();

  return true;
}

bool MJ_MessageQio::writeMsgMulti(QTextStream *s, MJ_Message *m)
{  
  if(!MJ_MessageQio::writeMsg(s, m)) return false;

  MJ_MessageMulti * mm = (MJ_MessageMulti *) (m);
  
  (*s) << mm->childCount() << "\n";
  
  MJ_MessageMulti::iterator it = mm->m_children.begin();

  while(it != mm->m_children.end()) {
    (*s) << (*it).first.c_str() << "\n";
    MJ_MessageFactory::write(s, (*it).second);
    ++it;
  }
  
  return true;
}

bool MJ_MessageQio::readMsgMulti(QTextStream *s, MJ_Message *m, int readFields)
{
  if(!MJ_MessageQio::readMsg(s, m, readFields)) return false;

  QString tmp(s->readLine());

  bool ok = true;
  
  uint count = tmp.toInt(&ok);

  if(!ok) return false;

  MJ_MessageMulti * mm = (MJ_MessageMulti *) (m);

  mm->clearChildren();

  for(uint i=0; i < count; i++) {
    tmp = s->readLine();
    
    MJ_Message *m = MJ_MessageQio::readFromStream(s);

    if(!m) return false;

    mm->addChild(tmp.ascii(), m);
  }

  return true;
}

bool MJ_MessageQio::writeMsgStlString(QTextStream *s, MJ_Message *m)
{
  if(!MJ_MessageQio::writeMsg(s, m)) return false;

  MJ_MessageStlString *ms = (MJ_MessageStlString *) m;

  (*s) << ms->string().length() << "\n";
  if(ms->string().length()) (*s) << ms->string().c_str() << "\n";

  return true;
}


bool MJ_MessageQio::readMsgStlString
(QTextStream *s, MJ_Message *m, int readFields)
{
  const char *fname = "MJ_MessageQio::readMsgStlString";

  if(!MJ_MessageQio::readMsg(s, m, readFields)) {
    MJ_error(MJ_ERR_ARGUMENT, fname);
    return false;
  }

  // QCString tmp;
  QString tmp;
  tmp = s->readLine();

  bool ok(true);

  int len = tmp.toInt(&ok);

  if(!len) {
    return ok;
  }
  
  MJ_MessageString *ms = (MJ_MessageString *) m;

  ms->setString(s->readLine().ascii());

  return ok;
}

bool MJ_MessageQio::writeMsgString(QTextStream *s, MJ_Message *m)
{
  if(!MJ_MessageQio::writeMsg(s, m)) return false;

  MJ_MessageString *ms = (MJ_MessageString *) m;

  (*s) << ms->string().length() << "\n";
  if(ms->string().length()) (*s) << ms->string() << "\n";

  /* if(m_string != QString::fromUtf8(m_string.utf8()))
     MJ_error(MJ_ERR_INTERNAL, "MJ_MessageString::write"); */

  return true;
}

bool MJ_MessageQio::readMsgString
(QTextStream *s, MJ_Message *m, int readFields)
{
  const char *fname = "MJ_MessageQio::readMsgString";

  if(!MJ_MessageQio::readMsg(s, m, readFields)) {
    MJ_error(MJ_ERR_ARGUMENT, fname);
    return false;
  }

  // QCString tmp;
  QString tmp;
  tmp = s->readLine();

  bool ok(true);

  int len = tmp.toInt(&ok);

  if(!len) {
    return ok;
  }
  
  MJ_MessageString *ms = (MJ_MessageString *) m;

  ms->setString(s->readLine());

  return ok;
}

void MJ_MessageQio::writeCout(MJ_Message *m)
{
  MJ_MessageFactory::write(&MJ_cout(), m);
  MJ_cout().device()->flush();
}

MJ_Message *MJ_MessageQio::readFromStream(QTextStream *s)
{
  const char * fname = "MJ_MessageQio::readFromStream";
  
  QString str;
  QCString tmp;

  (*s) >> str;

  MJ_Message * msg = MJ_MessageFactory::create(str.ascii());

  if(!msg) {
    MJ_error2(MJ_ERR_DATA_CORRUPTED, 
	      "%s # Bad message string \"%s\" # Around line %d", 
	      fname, (const char *) str, MJ_currentLine(s->device()));
    return 0;
  }

  if(!MJ_MessageFactory::read(s, msg, 1)) {
    delete msg;
    msg = 0;
  }
  
  return msg;
}

void MJ_MessageQio::initMessageFactory()
{
  MJ_MessageFactory::addFactoryFunction(MJ_messageTypes[MJ_MSG_EMPTY],
					MJ_MSG_EMPTY,
					MJ_Message::createNew,
					readMsg,
					writeMsg);
  MJ_MessageFactory::addFactoryFunction(MJ_messageTypes[MJ_MSG_BOOL],
					MJ_MSG_BOOL,
					MJ_MessageBool::createNew,
					readMsgBool,
					writeMsgBool);
  MJ_MessageFactory::addFactoryFunction(MJ_messageTypes[MJ_MSG_DOUBLE],
					MJ_MSG_DOUBLE,
					MJ_MessageDouble::createNew,
					readMsgDouble,
					writeMsgDouble);
  MJ_MessageFactory::addFactoryFunction(MJ_messageTypes[MJ_MSG_INT],
					MJ_MSG_INT,
					MJ_MessageInt::createNew,
					readMsgInt,
					writeMsgInt);
  MJ_MessageFactory::addFactoryFunction(MJ_messageTypes[MJ_MSG_STL_STRING],
					MJ_MSG_STL_STRING,
					MJ_MessageStlString::createNew,
					readMsgStlString,
					writeMsgStlString);
  MJ_MessageFactory::addFactoryFunction(MJ_messageTypes[MJ_MSG_STRING],
					MJ_MSG_STRING,
					MJ_MessageString::createNew,
					readMsgString,
					writeMsgString);
  MJ_MessageFactory::addFactoryFunction(MJ_messageTypes[MJ_MSG_MIDI_EVENT],
					MJ_MSG_MIDI_EVENT,
					MJ_MessageMidiEvent::createNew,
					readMsgMidi,
					writeMsgMidi);
  /* Note that we prefer MJ_MessageMulti2 over MJ_MessageMulti. They
     are nearly identical, but MJ_MessageMulti2 offers the possibility
     to query Qt strings -- not just the disgusting STL strings. */
  MJ_MessageFactory::addFactoryFunction(MJ_messageTypes[MJ_MSG_MULTI],
					MJ_MSG_MULTI,
					MJ_MessageMulti2::createNew,
					readMsgMulti,
					writeMsgMulti);

}


/** This is test function. It checks whether messages can suffessfully
    stored to DI_DataBuffer2. It should print out two identical
    MJ_MessageMulti objects. */
void MJ_MessageQio::testBufferIO()
{
  const char * fname = "MJ_MessageMulti::testIO";

  MJ_MessageMulti tmp(111), tmp2(123);

  DI_DataBuffer2 buf;

  tmp.addChild("double1", new MJ_MessageDouble(0.1231324));
  tmp.addChild("double2", new MJ_MessageDouble(0.899898231324));
  tmp.addChild("int1", new MJ_MessageInt(333666));
  tmp.addChild("e-midi", MJ_MessageFactory::create(MJ_MSG_MIDI_EVENT));
  tmp.addChild("bool1", MJ_MessageFactory::create(MJ_MSG_BOOL));
  
  tmp.addChild("StlString", new MJ_MessageStlString("This is the real Value"));
  tmp.addChild("fool2", new MJ_MessageBool(true, 888));
  tmp.addChild("fool3", new MJ_MessageBool(false, 777));

  MJ_MessageFactory::write(&MJ_cout(), &tmp);
  
  MJ_cout().device()->flush();

  tmp.write2(&buf);

  MJ_trace2("%s # About to read from buffer (%d bytes used)", 
	    fname, buf.position());

  buf.rewind();

  bool ok = tmp2.read2(&buf, 0);

  if(!ok)
    MJ_error(MJ_ERR_INTERNAL, fname, "Read failed");

  MJ_trace(fname, "Read from buffer (did not crash :-) :");

  MJ_MessageFactory::write(&MJ_cout(), &tmp2);
  
  MJ_cout().device()->flush();
}
