/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_message_xml.h>

#include <mj_error.h>
#include <mj_iostream.h>
#include <mj_message_factory.h>
#include <mj_message_multi2.h>
#include <mj_message_string.h>
#include <mj_messages.h>
#include <mj_midi.h>

#include <qstringlist.h> 

#include <typeinfo>

// #include <di_data_buffer.h>

// #include <qtextstream.h>

class __MJ_MessageXml
{
public:
  
  __MJ_MessageXml() : m_writer(0), m_reader(0) {}

  QDomElement (*m_writer) (QDomDocument &, MJ_Message *);
  bool        (*m_reader) (QDomElement &,  MJ_Message *);

};

static std::map<std::string, __MJ_MessageXml> __MJ_XmlMaps;

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

void MJ_DomUtil::addTextChild(QDomElement &base, QDomDocument &doc,
			      const QString &tag,  const QString &value)
{
  QDomElement si = doc.createElement(tag);
  si.appendChild(doc.createTextNode(value));
  base.appendChild(si);
}

void MJ_DomUtil::addTextChild(QDomElement &base, QDomDocument &doc,
			      const QString &tag,  int value)
{
  addTextChild(base, doc, tag, QString().setNum(value));
}

void MJ_DomUtil::addTextChild(QDomElement &base, QDomDocument &doc,
			      const QString &tag,  double value)
{
  addTextChild(base, doc, tag, QString().setNum(value));
}

double           MJ_DomUtil::getDouble(QDomElement &e, const QString &tag, 
				       double def)
{
  QDomNode c = e.namedItem(tag);
  
  QDomText t = c.toElement().firstChild().toText();
  
  bool ok = true;
    
  double val = t.data().toDouble(&ok);
  
  return ok ? val : def;
}

int              MJ_DomUtil::getInt(QDomElement &e, const QString &tag, int def)
{
  QDomNode c = e.namedItem(tag);
  
  QDomText t = c.toElement().firstChild().toText();
  
  bool ok = true;
  
  int val = t.data().toInt(&ok);
  
  return ok ? val : def;
}

QString          MJ_DomUtil::getString(QDomElement &e, const QString &tag, 
				       const char * def)
{
  QString str =  e.namedItem(tag).toElement().firstChild().toText().data();

  if(str.isEmpty())
    str = def;

  return str;
}



/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

QDomElement MJ_MessageXml::writeMsg(QDomDocument &doc, MJ_Message *m)
{
  QString id = MJ_MessageFactory::determineName(m->messageType());

  if(id.isEmpty())
    return QDomElement();

  std::map<std::string, __MJ_MessageXml>::iterator it = 
    __MJ_XmlMaps.find(id.ascii());

  if(it == __MJ_XmlMaps.end()) {
    MJ_error2(MJ_ERR_UNKNOWN,
	      "MJ_MessageXml::writeMsg # No writer function for %s",
	      typeid(*m).name());
    return QDomElement();
  }

  return (*(*it).second.m_writer)(doc, m);
}

MJ_Message * MJ_MessageXml::readMsg(QDomElement &e)
{
  QString id = e.tagName();

  MJ_Message * m = MJ_MessageFactory::create(id.ascii());

  if(!m)
    return 0;

  return m;
}

QDomElement MJ_MessageXml::writeMsgBool(QDomDocument &doc, MJ_Message *m)
{
  QDomElement e = writeMsgBasics(doc, m);

  MJ_MessageBool *mb = (MJ_MessageBool *) (m);
  
  addTextChild(e, doc, "value", (int) mb->value());
  
  return e;
}

bool MJ_MessageXml::readMsgBool(QDomElement &e, MJ_Message *m)
{
  MJ_MessageBool *mb = (MJ_MessageBool *) (m);

  mb->setValue(getInt(e, "value", 0));

  return true;
}

QDomElement MJ_MessageXml::writeMsgDouble(QDomDocument &doc, MJ_Message *m)
{
  MJ_trace("MJ_MessageXml::writeMsgDouble");

  QDomElement e = writeMsgBasics(doc, m);

  MJ_MessageDouble *md = (MJ_MessageDouble *) (m);
  
  addTextChild(e, doc, "value", md->value());
  
  return e;  
}

bool MJ_MessageXml::readMsgDouble(QDomElement &e, MJ_Message *m)
{
  MJ_MessageDouble *md = (MJ_MessageDouble *) (m);

  md->setValue(getDouble(e, "value", 0.0));

  return true;
}

QDomElement MJ_MessageXml::writeMsgInt(QDomDocument &doc, MJ_Message *m)
{
  QDomElement e = writeMsgBasics(doc, m);

  MJ_MessageInt *mi = (MJ_MessageInt *) (m);
  
  addTextChild(e, doc, "value", mi->value());
  
  return e;
}


bool MJ_MessageXml::readMsgInt(QDomElement &e, MJ_Message *m)
{
  MJ_MessageInt *mb = (MJ_MessageInt *) (m);

  mb->setValue(getInt(e, "value", 0));

  return true;
}

QDomElement MJ_MessageXml::writeMsgMidi(QDomDocument &doc, MJ_Message *m)
{
  QDomElement e = writeMsgBasics(doc, m);

  MJ_MessageMidiEvent * me = (MJ_MessageMidiEvent *) m;

  DI_MidiEvent &event = me->midiMsg();

  // Write the header bytes:
  
  QString tmp;
  QString bytes;
  uint i;

  for(i=0; i < 4; i++) {
    tmp.sprintf("%d", (int) event.get(i));
    bytes += tmp + " ";
  }

  addTextChild(e, doc, "header", bytes);
  addTextChild(e, doc, "ticks", (int) event.ticks());
  addTextChild(e, doc, "time", (int) event.time());

  // Write the sysex data:

  if(!event.sysexSize())
    return e;

  bytes.truncate(0);

  for(i=0; i < event.sysexSize(); i++) {
    tmp.sprintf("%x", (int) event.sysexDataConst()[i]);
    bytes += tmp + " ";
  }

  addTextChild(e, doc, "sysex", bytes);

  return e;
}

bool MJ_MessageXml::readMsgMidi(QDomElement &e, MJ_Message *m)
{
  uint i;

  MJ_MessageMidiEvent * me = (MJ_MessageMidiEvent *) m;

  DI_MidiEvent &event = me->midiMsg();

  event.clear();

  QString header = getString(e, "header");
  
  for(i = 0; i < 4; i++) {
    QString byte = header.section(' ', i, i);
    event.set(i, byte.toUInt());
  }

  event.setTicks(getInt(e, "ticks", 0));
  event.setTime(getInt(e, "time", 0));

  QStringList sysex =  QStringList::split(' ', getString(e, "sysex"));

  uint n = sysex.count();

  if(n == 0)
    return true;

  event.resizeSysex(n);

  uchar * sx = event.sysexData();
  
  for(i=0; i < n; i++)
    sx[i] = (*sysex.at(i)).toUInt();
  
  return true;
}

QDomElement MJ_MessageXml::writeMsgMulti(QDomDocument &doc, MJ_Message *m)
{
  QDomElement e = writeMsgBasics(doc, m);

  MJ_MessageMulti * mm = (MJ_MessageMulti *) m;

  if(!mm->size())
    return e;

  QDomElement children = doc.createElement("messages");

  for(MJ_MessageMulti::iterator it = mm->begin(); it != mm->end(); it++) {
    QDomElement tmp = doc.createElement((*it).first.c_str());
    QDomElement tmp2 = writeMsg(doc, (*it).second);
    tmp.appendChild(tmp2);
    children.appendChild(tmp);
  }

  e.appendChild(children);

  return e;
}

bool MJ_MessageXml::readMsgMulti(QDomElement &e, MJ_Message *m)
{
  MJ_MessageMulti * mm = (MJ_MessageMulti *) m;
  
  mm->clearChildren();

  QDomElement children = e.namedItem("messages").toElement();

  QDomNode n = children.firstChild();
  
  while ( !n.isNull() ) {

    QDomElement n2 = n.toElement();

    QString path = n2.tagName();
    QDomElement elem = n2.firstChild().toElement();
    MJ_Message * m2 = readMsg(elem);

    if(m2) {
      mm->addChild(path.ascii(), m2);
    }
    
    n = n.nextSibling();
  }

  return true;
}

QDomElement MJ_MessageXml::writeMsgStlString(QDomDocument &doc, MJ_Message *m)
{
  QDomElement e = writeMsgBasics(doc, m);

  MJ_MessageStlString *mi = (MJ_MessageStlString *) (m);
  
  addTextChild(e, doc, "str", mi->string().c_str());
  
  return e;
}

bool MJ_MessageXml::readMsgStlString(QDomElement &e, MJ_Message *m)
{
  MJ_MessageStlString *mb = (MJ_MessageStlString *) (m);

  mb->setString(getString(e, "str").ascii());

  return true;
}

QDomElement MJ_MessageXml::writeMsgString(QDomDocument &doc, MJ_Message *m)
{

  QDomElement e = writeMsgBasics(doc, m);

  MJ_MessageString *mi = (MJ_MessageString *) (m);
  
  addTextChild(e, doc, "str", mi->string());
  
  return e;
}

bool MJ_MessageXml::readMsgString(QDomElement &e, MJ_Message *m)
{
  MJ_MessageString *mb = (MJ_MessageString *) (m);

  mb->setString(getString(e, "str"));

  return true;
}

void MJ_MessageXml::initMessageXml()
{
  __MJ_MessageXml tmp;

  tmp.m_writer = writeMsgBool;
  tmp.m_reader = readMsgBool;
  __MJ_XmlMaps[MJ_messageTypes[MJ_MSG_BOOL]] = tmp;

  tmp.m_writer = writeMsgInt;
  tmp.m_reader = readMsgInt;
  __MJ_XmlMaps[MJ_messageTypes[MJ_MSG_INT]] = tmp;

  tmp.m_writer = writeMsgDouble;
  tmp.m_reader = readMsgDouble;
  __MJ_XmlMaps[MJ_messageTypes[MJ_MSG_DOUBLE]] = tmp;

  tmp.m_writer = writeMsgMidi;
  tmp.m_reader = readMsgMidi;
  __MJ_XmlMaps[MJ_messageTypes[MJ_MSG_MIDI_EVENT]] = tmp;

  tmp.m_writer = writeMsgString;
  tmp.m_reader = readMsgString;
  __MJ_XmlMaps[MJ_messageTypes[MJ_MSG_STRING]] = tmp;

  tmp.m_writer = writeMsgStlString;
  tmp.m_reader = readMsgStlString;
  __MJ_XmlMaps[MJ_messageTypes[MJ_MSG_STL_STRING]] = tmp;

  tmp.m_writer = writeMsgMulti;
  tmp.m_reader = readMsgMulti;
  __MJ_XmlMaps[MJ_messageTypes[MJ_MSG_MULTI]] = tmp;

}

void MJ_MessageXml::test()
{
  initMessageXml();

  MJ_trace("MJ_MessageXml::test");

  MJ_MessageDouble md(0.87);
  MJ_MessageInt mi(56);
  MJ_MessageString ms("FOO is GREAT");
  MJ_MessageMulti mm;
  MJ_MessageMidiEvent midi;
  
  midi.midiMsg().noteOn(64, 77, 16, 200);
  
  mm.addChild("md1", new MJ_MessageDouble(0.6768));
  mm.addChild("md2", new MJ_MessageDouble(777));
  mm.addChild("mi1", new MJ_MessageInt(66));

  QDomDocument doc( "mj_test" );
  QDomElement root = doc.createElement( "mj_root" );
  doc.appendChild( root );

  root.appendChild(writeMsg(doc, &md));
  root.appendChild(writeMsg(doc, &mi));
  root.appendChild(writeMsg(doc, &ms));
  root.appendChild(writeMsg(doc, &mm));
  root.appendChild(writeMsg(doc, &midi));

  QString xml = doc.toString();
  
  MJ_trace2("MJ_MessageXml::test # About to write xml (%d characters)",
	    (int) xml.length());
  
  puts(xml.ascii());
}


QDomElement MJ_MessageXml::writeMsgBasics(QDomDocument &doc, MJ_Message *m)
{
  QString id = MJ_MessageFactory::determineName(m->messageType());
  QDomElement elem = doc.createElement(id);
  return elem;
}

bool MJ_MessageXml::readMsgBasics(QDomElement &, MJ_Message *)
{
  return true;
}
