/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_midi_conf.h>

#include <mj_error.h>
#include <mj_global.h>
#include <mj_iostream.h>
#include <mj_misc.h>

#include <stdlib.h>
#include <qtextstream.h> 

const char * MJ_MidiDeviceTypeKeys[MJ_MIDID_SIZEOF] = {
  "Input",
  "Output"
};

const char * MJ_MidiDeviceTypeKeys2[MJ_MIDID_SIZEOF] = {
  "Input device",
  "Output device"
};

const char * MJ_MidiEnvVarName = "MJ_MIDI_CONF";

const char *MJ_MidiConf::m_identifier = "MJ_MIDI_CONFIG";

MJ_MidiConf::MJ_MidiConf()
{
  m_devices.setAutoDelete(true);
}

MJ_MidiConf::~MJ_MidiConf()
{}

/** Adds a new device to the device list. 

    @return If the device already exists false is returned. Else true
    is returned.*/

bool MJ_MidiConf::addDevice(MJ_MidiDeviceType type, int number)
{
  if(findDevice(type, number) >= 0) return false;
  
  MJ_2Int *tmp = new MJ_2Int;
  tmp->m_v[0] = (int) type;
  tmp->m_v[1] = number;
  m_devices.append(tmp);

  MJ_trace2("MJ_MidiConf::addDevice %s %d",
	    MJ_MidiDeviceTypeKeys[tmp->m_v[0]], tmp->m_v[1]);

  return true;
}

/** Returns a device from the list. If the device is not found false
    is returned. */

bool MJ_MidiConf::removeDevice(MJ_MidiDeviceType type, int number)
{
  int index;
  if((index = findDevice(type, number)) == -1) return false;
  m_devices.remove(index);
  return true;
}

/** Seeks a device in the device list of this object (not the global
    list). 

    @return If the device is not found, -1 is returned. */
int MJ_MidiConf::findDevice(MJ_MidiDeviceType type, int number)
{
  uint count = m_devices.count();
  MJ_2Int tmp;
  tmp.m_v[0] = (int) type;
  tmp.m_v[1] = number;
  
  for(uint i = 0; i < count; i++) {
    if(MJ_equals(tmp, * m_devices.at(i) ))
      return (int) i;
  }
  return -1;
}

uint MJ_MidiConf::numDevices(MJ_MidiDeviceType type)
{
  uint sum = 0;

  for(uint i=0; i < m_devices.count(); i++) {
    if(m_devices.at(i)->m_v[0] == (int) type)
      sum++;
  }

  return sum;
}


/** Returns name of a particular midi device. */
QString MJ_MidiConf::deviceName(uint i)
{
  MJ_2Int *mi = m_devices.at(i);
  return midiDevice((MJ_MidiDeviceType) mi->m_v[0], mi->m_v[1]);
}

/** Write decsription of the midi configuration to a stream. */
bool MJ_MidiConf::write(QTextStream *s)
{
  uint siz = m_devices.count();
  
  (*s) << m_identifier << " " << siz << "\n";
  
  for(uint i=0; i < siz; i++) {
    const MJ_2Int &mi = *m_devices.at(i);

    if((uint) mi.m_v[0] >= MJ_MIDID_SIZEOF) {
      MJ_error(MJ_ERR_INTERNAL, "MJ_MidiConf::write");
      return false;
    }

    (*s) << MJ_MidiDeviceTypeKeys[mi.m_v[0]] << "\t" << mi.m_v[1] << "\n";
  }
  return true;
}


/** Read decsription of the midi configuration from a stream. */
bool MJ_MidiConf::read(QTextStream *s)
{
  const char * fname = "MJ_MidiConf::read";

  int siz, t1, t2;
  bool ok = true;
  QCString str;
  (*s) >> str;

  if(str != m_identifier) {
    MJ_error(MJ_ERR_DATA_CORRUPTED, fname, 1);
    return false;
  }

  (*s) >> str;
  siz = str.toInt(&ok);

  if(!ok || siz > 1000) {
    MJ_error(MJ_ERR_DATA_CORRUPTED, fname, 2);
    return false;
  }
  
  m_devices.clear();

  for(uint i=0; i < (uint) siz; i++) {

    (*s) >> str >> t2;
    
    if((t1 = MJ_findStr(str, MJ_MidiDeviceTypeKeys, MJ_MIDID_SIZEOF)) 
       == -1) {
      MJ_error(MJ_ERR_DATA_CORRUPTED, fname, str);
      return false;
    }
    MJ_2Int *mi = new MJ_2Int();
    mi->m_v[0] = t1;
    mi->m_v[1] = t2;
    m_devices.append(mi);
  }
  // Read the end-of-line
  s->readLine();
  return true;
}

/** Returns the device description of the queried midi device.

    The device name format depends on your system. 

    On Linux/OSS device names are names of the devices in /dev/dsp.

    On IRIX the names are typically "Stereo input" etc.
*/
QString MJ_MidiConf::midiDeviceDescription(MJ_MidiDeviceType type,
					     int number)
{
  QString tmp, tmp2, envDef(getenv(MJ_MidiEnvVarName));

 int fields = MJ_fields(envDef, "|");

  int hits = 0;

  for(int i=0; i < fields; i++) {
    tmp = MJ_field(envDef, i, "|");
    
    if(tmp.find(MJ_MidiDeviceTypeKeys[type], 0) == 0) {
      if(hits == number) {

	return tmp;
      }
      hits++;
    }
  }
    
  return tmp2;
}

/** Returns the midi device name of the midi device descriptor. */

QString MJ_MidiConf::midiDevice(const QString &descriptor)
{
  if(MJ_fields(descriptor, "=") != 2) {
    MJ_error(MJ_ERR_ENVIRONMENT, 
	     "MJ_MidiConf::midiDevice, malformed device information",
	     (const char *) descriptor);
  }
  return MJ_field(descriptor, 1, "=");
}


QString MJ_MidiConf::midiDevice(MJ_MidiDeviceType type,
				int number)
{
  return midiDevice(midiDeviceDescription(type, number));
}


/** Returns the number of midi devices of the desired type.*/

int MJ_MidiConf::midiDevices(MJ_MidiDeviceType type)
{
  QString envDef(getenv(MJ_MidiEnvVarName));
  
  int n = 0, start = 0;

  while((start = envDef.find(MJ_MidiDeviceTypeKeys[type],
			     start)) >= 0) {
    n++;
    start++;
  }
  return n;
}
