/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

// Kludke to get access to QLNode:
#define private public

#include <mj_error.h>
#include <mj_midi_sequence.h>
#include <mj_listtmpl.C>


#include <qdatastream.h> 
#include <qiodevice.h> 
#include <qfile.h> 
#include <qtextstream.h> 



template class QList<MJ_MidiEvent>;
template class MJ_List<MJ_MidiEvent>;

static const char s_mthd[] = "MThd";
static const char s_mtrk[] = "MTrk";

bool MJ_skipBytes(QDataStream *s, unsigned l)
{
  QIODevice *d = s->device();
  ulong p = (ulong) d->at();
  return d->at(p+l);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

int MJ_MidiEventList::compareItems(void *v1, void *v2)
{
  MJ_MidiEvent *e1 = (MJ_MidiEvent *) v1;
  MJ_MidiEvent *e2 = (MJ_MidiEvent *) v2;

  if(e1->ticks() == e2->ticks()) return 0;
  if(e1->ticks() > e2->ticks()) return 1;
  return -1;
}


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

const char *MJ_MidiTrack::m_label = "MJ_MIDI_TRACK";


bool MJ_MidiTrack::m_debug = false;


MJ_MidiTrack::MJ_MidiTrack()
  : m_current(0),
    m_tempoStartTime(0),
    m_tempoStartTicks(0),
    m_tempo(2.0),  // 120 bpm
    m_ticksPerQuarter(256.0)
{
  // puts("MJ_MidiTrack::MJ_MidiTrack");
  setAutoDelete(true);
}

MJ_MidiTrack::MJ_MidiTrack(const MJ_MidiTrack &r)
{
  setAutoDelete(true);
  *this = r;
}

MJ_MidiTrack::~MJ_MidiTrack()
{}

bool MJ_MidiTrack::readTrack(QDataStream *s, int n)
{
  const char *fname = "MJ_MidiTrack::readTrack";

  if(m_debug)
    MJ_trace(fname, n);

  char b[4];
  Q_INT32 len;

  s->readRawBytes(b, 4);

  if(strncmp(b, s_mtrk, 4)) {
    MJ_error(MJ_ERR_ARGUMENT, fname, "Bad track header");
    return false;
  }

  (*s) >> len;
  QIODevice *d = s->device();

  ulong beginPos = d->at();
  int endPos = beginPos + len;
  
  clear();
  
  while(endPos > (int) d->at()) {
    if(!readEvent(s)) {
      MJ_error(MJ_ERR_ARGUMENT, fname, "Event read failed");
      return false;
    }
  }

  if(m_debug) 
    MJ_trace(fname, "SUCCESS");
  return true;
}

bool MJ_MidiTrack::writeTrack(QTextStream *s)
{

  (*s) << m_label << " " << count() << "\n";

  QListIterator<MJ_MidiEvent> iter(*this);
  
  MJ_MidiEvent *e;
  while((e = iter.current())) {
    (*s) << (*e);
    (*s) << "\n";
    ++iter;
  }
  return true;
}

uint MJ_MidiTrack::endTicks()
{
  if(count())
    return last()->ticks();
  return 0;
}

MJ_MidiEvent *MJ_MidiTrack::findNoteOff(MJ_MidiEvent *noteOn)
{
  QLNode *ln = currentNode();

  if(ln->getData() != noteOn) {
    if(m_debug)
      MJ_error(MJ_ERR_INTERNAL, "MJ_MidiTrack::findNoteOff # corrupted");
    return 0;
  }

  char chan = noteOn->channel();
  char key  = noteOn->keyNumber();

  // while(ln.getData()) {
  while(ln) {
    void *vtmp = ln->getData();
    if(!vtmp) return 0;

    MJ_MidiEvent *etmp = (MJ_MidiEvent *) vtmp;

    if(etmp->isNoteOff() && 
       (etmp->channel() == chan) &&
       (etmp->keyNumber() == key)) 
      return etmp;
    
    ln = ln->next;
  }
  return 0;
}


void MJ_MidiTrack::getEvents(QList<MJ_MidiEvent> *events, ulong ticks)
{
  // MJ_trace("MJ_MidiTrack::getEvents", (int) ticks);

  if((uint) m_current >= count()) {
    return;
  }

  MJ_MidiEvent *tmp = at(m_current);
  
  if(!tmp) MJ_fatal(MJ_ERR_INTERNAL, "MJ_MidiTrack::getEvents");

  while(tmp->ticks() < ticks) {

    // MJ_trace("MJ_MidiTrack::getEvents # Added", tmp->ticks());

    // events->append(new MJ_MidiEvent(*tmp));
    events->inSort(tmp);
    m_current++;
    if((uint) m_current >= count()) return;

    tmp = at(m_current);
    if(!tmp) MJ_fatal(MJ_ERR_INTERNAL, "MJ_MidiTrack::getEvents");
  }
}

void MJ_MidiTrack::prepare()
{ 
  m_current = 0; 

  if(!count())
    MJ_trace("MJ_MidiTrack::prepare # no events");
  else
    MJ_trace("MJ_MidiTrack::prepare # Start ticks ", at(m_current)->ticks());
}

MJ_MidiTrack &MJ_MidiTrack::operator = (const MJ_MidiTrack &t)
{
  clear();

  QListIterator<MJ_MidiEvent> iter(t);
  MJ_MidiEvent *event;
  
  while((event = iter.current())) {
    MJ_MidiEvent *newEvent = new MJ_MidiEvent(*event);
    if(newEvent)
      append(newEvent);
    ++iter;
  }
  return *this;
}

bool MJ_MidiTrack::readEvent(QDataStream *s)
{
  const char * fname = "MJ_MidiTrack::readEvent";

  if(m_debug)
    MJ_trace(fname, this);

  int delta = readDeltaTime(s);

  if(delta == -1) return false;
  uint currTime;
  if(count())
    currTime = delta + last()->ticks();
  else
    currTime = delta;

  MJ_MidiEvent *e = new MJ_MidiEvent;
  if(!e) return false;

  if(m_debug)
    MJ_trace(fname, "Normal event");
  
  e->setTicks(currTime);
  e->setTime(ticksToTimeStamp(currTime));
  
  unsigned char status;
  // s->readRawBytes(&status, 1);
  (*s) >> status;
  e->set(0, status);
 
  if(m_debug) {
    char buf[64];
    sprintf(buf, "event, delta: %d status: %X\n", delta, (unsigned) status);
    MJ_trace(fname, buf);
  }
  
  // if(status == 0xFF) return readSpecial(s, e);

  if(status == 0xF0 || status == 0xF7 || status == 0xFF)
    return readSysex(s, e);
  
  return readNormal(s, e);
}

bool MJ_MidiTrack::readSysex(QDataStream *s, MJ_MidiEvent *e)
{
  if(m_debug) 
    MJ_trace("MJ_MidiTrack::readSysex");

  char c = 0;
  s->readRawBytes(&c, 1);

  e->set(1, c);
  
  int l = readDeltaTime(s);

  /*  Q_INT32 l;
      (*s) >> l;*/

  if(!e->setSysex(0, l)) {
    delete e;
    return false;
  }

  if(l)
    s->readRawBytes((char *) e->sysexData(), l);

  append(e);

  return true;
}

bool MJ_MidiTrack::readNormal(QDataStream *s, MJ_MidiEvent *e)
{
  int toRead = 0;

  unsigned char tmp = e->a() & 0xF0;
  
  if((tmp == 0x90) ||  // Note off.
     (tmp == 0x80) ||  // Note on.
     (tmp == 0xA0) ||  // Polyphonic Key
     (tmp == 0xB0) ||  // Controller Change 
     (tmp == 0xE0))    // Pitch Bend
    toRead = 2;
  else if((tmp == 0xC0) ||  // Program Change
	  (tmp == 0xD0))    // Channel Key Pressure 
    toRead = 1;
 
  /* if(m_debug) 
     cout << " toRead: " << toRead << " e: " << e<< endl; */
  
  if(toRead)
    s->readRawBytes((char *) & e->message()[1], toRead);

  append(e);
  return true;
}

int  MJ_MidiTrack::readDeltaTime(QDataStream *s)
{
  int t = 0;
  
  for(int i=0; i < 4; i++) {
    unsigned char c = 0;
    (*s) >> c;
    
    t = (t << 7) + (int) (c & 0x7f);
    if(!(c & 0x80)) {
      return t;
    }
  }
  return -1;
}

ulong MJ_MidiTrack::ticksToTimeStamp(uint ticks)
{
  uint diff = ticks - m_tempoStartTicks;

  double quarters = (double) diff / m_ticksPerQuarter;
  double timeSeconds = quarters / m_tempo;
  return m_tempoStartTime + (ulong) (timeSeconds * MJ_MIDI_TIME_RESOLUTION);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

const char *MJ_MidiSequence::m_label = "MJ_MIDI_SEQUENCE";

bool MJ_MidiSequence::m_debug = true;

MJ_MidiSequence::MJ_MidiSequence()
  : m_ticksPerQuarter(256), // Just a guess
  m_tempo(1000000),
  m_tempoBeginUs(0),
  m_tempoBeginTicks(0)
{}

MJ_MidiSequence::~MJ_MidiSequence()
{}

bool MJ_MidiSequence::readSequence(const QString &fileName)
{
  if(m_debug) 
    MJ_trace("MJ_MidiSequence::readSequence", fileName);

  QFile midiFile(fileName);
  if(!midiFile.open(IO_ReadOnly)) return false;
  QDataStream s(&midiFile);
  
  if(!readHeader(&s)) return false;
  for(uint i=0; i < size(); i++) {
    (*this)[i].setTicksPerQuarter(m_ticksPerQuarter);
    if(! (*this)[i].readTrack(&s, i)) return false;

    if(m_debug) {
      QFile f;
      f.open( IO_WriteOnly, stdout );
      QTextStream sout(&f);
      (*this)[i].writeTrack(&sout);
    }
  }
  m_fileName = fileName.copy();
  return true;
}

bool MJ_MidiSequence::writeSequence(QTextStream *s)
{
  (*s) << m_label << " " << size() << "\n";

  for(uint i=0; i < size(); i++) {
    if(!get(i).writeTrack(s))
      return false;
  }
  return true;
}

uint MJ_MidiSequence::endTicks()
{
  uint et = 0;

  for(uint i=0; i < size(); i++) {
    uint tmp = get(i).endTicks();

    if(tmp > et) et = tmp;
  }
  return et;
}

void MJ_MidiSequence::prepare()
{
  for(uint i=0; i < size(); i++) {
    get(i).prepare();
  }
}

void MJ_MidiSequence::getEvents(QList<MJ_MidiEvent> *events, ulong timeUs)
{

  /* #if 0
     ulong ticks = (timeUs - m_tempoBeginUs) * m_ticksPerQuarter / m_tempo +
     m_tempoBeginTicks;
     #elsif 0
     ulong ticks = timeUs / 1000 / 2; // 500 ticks per second
     #else*/
  ulong time = timeUs / (1000000 / MJ_MIDI_TIME_RESOLUTION);
  // #endif

  for(uint i=0; i < size(); i++) {
    get(i).getEvents(events, time);
  }
  
  // cout << " timeUs " << timeUs << " events " << events->count() << endl;
}

void MJ_MidiSequence::copyTo(MJ_MidiTrack *track)
{
  // uint current = 0;

  for(uint i=0; i < size(); i++) {
    QListIterator<MJ_MidiEvent> iter(*track);
    MJ_MidiEvent *e;

    while((e = iter.current())) {

      MJ_MidiEvent *enew=new MJ_MidiEvent(*e);

      if(!track->count()) {
	track->append(enew);
      }
      else {
	track->inSort(enew);
      }
    }
  }
}

void MJ_MidiSequence::test(const char *filename)
{
  const char * fname = "MJ_MidiSequence::test";

  MJ_MidiSequence ms;
  if(!ms.readSequence(filename))
    MJ_error(MJ_ERR_ARGUMENT, fname, filename);
  else 
    MJ_trace(fname, filename);
  
  QFile f;
  f.open( IO_WriteOnly, stdout );
  QTextStream sout(&f);
  ms.writeSequence(&sout);
  fflush(0);
}
  
bool MJ_MidiSequence::readHeader(QDataStream *s)
{

  const char *fname = "MJ_MidiSequence::readHeader";

  if(m_debug) MJ_trace(fname, "enter");

  char b[4];
  Q_INT32 tmp;
  Q_INT16 format, trackNum, ticksPerQuarter;

  s->readRawBytes(b, 4);

  if(strncmp(b, s_mthd, 4)) return false;
  (*s) >> tmp;
  
  if(tmp != 6) {
    MJ_error(MJ_ERR_ARGUMENT, fname, tmp);
    return false;
  }

  (*s) >> format >> trackNum >> ticksPerQuarter;

  if(format > 3) return false; // Bad format.

  resize(0);  // Resize to zero to avoid copying data.

  if(format > 0) {
    
    if(m_debug)  MJ_trace(fname, "Tracks ", trackNum);

    resize(trackNum);
  }
  else 
    resize(1);
  
  m_ticksPerQuarter = (ulong) ticksPerQuarter;

  if(m_debug) MJ_trace(fname, "SUCCESS");

  return true;
}

