/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_array.h>
#include <mj_cstr_remap.h>
#include <mj_misc.h>
#include <mj_error.h>
#include <mj_iostream.h>

#include <mj_message.h>
#include <mj_message_carrier.h>
#include <mj_parameter.h>
#include <mj_plugin_base.h>
#include <mj_so_ref_count.h>
#include <mj_tr.h>
#include <mj_version.h>

#include <stdlib.h>


const char *MJ_mustajuuriStr = "Mustajuuri";

const char *MJ_UnitKeys[MJ_UNIT_SIZEOF] = {
  "blank",
  "radians",
  "degrees",
  "beats per minute",
  "Decibels",
  "Hertz",
  "meters",
  "omega",
  "quarters",
  "samples",
  "seconds",
  "milliseconds"
};

const char *MJ_UnitKeysShort[MJ_UNIT_SIZEOF] = { 
  "",
  "rad",
  "deg",
  "bpm",
  "db",
  "Hz",
  "m",
  "w",
  "quarters",
  "samps",
  "secs",  
  "millisecs"
};

const char *MJ_StereoKeys[MJ_STEREO_SIZEOF] = {
  "left",
  "right"
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

// NULL-POINTER TO FATAL

void MJ_cpf(void *vpP)
{
  if(!vpP) MJ_fatal(MJ_ERR_MEMORY, "MJ_cpf");
}

void MJ_cpf(void *vpP1, void *vpP2)
{
  if(!vpP1 || !vpP2) MJ_fatal(MJ_ERR_MEMORY, "MJ_cpf2");
}

void MJ_cpf(void *vpP1, void *vpP2, void *vpP3)
{
  if(!vpP1 || !vpP2 || vpP3) MJ_fatal(MJ_ERR_MEMORY, "MJ_cpf3");
}

void MJ_cpf(void *vpP1, const char *cpP1)
{
  if(!vpP1) MJ_fatal(MJ_ERR_MEMORY, cpP1);
}

void MJ_cpf(void *vpP1, void *vpP2, const char *cpP1)
{
  if(!vpP1 || !vpP2) MJ_fatal(MJ_ERR_MEMORY, cpP1);
}

void MJ_cpf(void *vpP1, void *vpP2, void *vpP3, const char *cpP1)
{
  if(!vpP1 || !vpP2 || !vpP3) MJ_fatal(MJ_ERR_MEMORY, cpP1);
}


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** Finds which string str is of strs. There should be nofStrs
    pointers in strs. 


    @return This method returns zero or positive if the string was
    found. In other cases it returns zero.
*/
int MJ_findStr(const char *str, const char **strs, int nofStrs)
{
  if(!str || !strs) return -1;

  for(int i=0; i < nofStrs; i++) {
    if(strcmp(str, strs[i]) == 0)
      return i;
  }
  return -1;
}

/** Only checks to the shorter of the strings. */
int MJ_findStr2(const char *str, const char **strs, int nofStrs)
{
  static const int maxLoop = 1000000;

  if(!str || !strs) return -1;

  char end = '\0';
  
  for(int i=0; i < nofStrs; i++) {

    const char *str2 = strs[i];

    for(int j=0; j < maxLoop; j++) {

      if(str[j] == end || str2[j] == end)
	return i;
      else if(str[j] != str2[j])
	j = maxLoop; // Break loop.
    }
  }
  return -1;
}

/** Copies one string on top of another. The old target string is
    deleted if it is non-null. */
void MJ_delAndDup(char **target, const char *str)
{
  char * tt = *target;

  free(tt);

  if(str) {
    if(strlen(str)) {
      tt = strdup(str);
      MJ_cpf(tt, "MJ_delAndDup");
    }
    else
      tt = 0;
  }
  else 
    tt = 0;
  *target = tt;
}


QCString MJ_copyIfZero(const char *str, int i)
{
  QCString s;
  if(!i) s = str;
  return s;
}

/** Returns the number of fields in a string. Useful for path handling etc. */
int MJ_fields(const QString &s, const char *separators)
{
  QString seps;
  if(separators) seps = separators;
  else seps = " \n\t";

  bool onField = false;
  int count = 0;

  for(int i=0; i <= (int) s.length(); i++) {
    bool nowBlank = false;

    // Try to match all separators
    for(int j=0; j <= (int) seps.length(); j++) {
      if(s[i] == seps[j]) {
	nowBlank = true;
	j = seps.length();
      }
    }

    bool nowField = ! nowBlank;

    if(nowField) {
      if(!onField) {
	count++;
      }
      onField = true;
    }
    else onField = false;
    
  }
  return count;
}

/** Returns the number of fields in a string. Useful for path handling etc. */
int MJ_fields(const char *s, const char *seps)
{
  if(!seps) seps = " \n\t";

  bool onField = false;
  int count = 0;

  int sepCount = strlen(seps);

  for(int i=0; s[i] != '\0'; i++) {
    bool nowBlank = false;
    
    // Try to match all separators
    for(int j=0; j <= sepCount; j++) {
      if(s[i] == seps[j]) {
	nowBlank = true;
	j = sepCount;
      }
    }

    bool nowField = ! nowBlank;

    if(nowField) {
      if(!onField) {
	count++;
      }
      onField = true;
    }
    else onField = false;
    
  }
  return count;
}

/** Retuns a field from a unicode string. */
QString MJ_field(const QString &s, int field, const char *separators)
{
  
  QString r;

  QString seps;
  if(separators) seps = separators;
  else seps = " \n\t";

  bool onField = false;
  int count = -1;
  int lastBegan = 0;

  for(uint i=0; i <= s.length(); i++) {
    bool nowBlank = false;

    // Try to match all separators
    for(uint j=0; j <= seps.length(); j++) {
      if(s[(int)i] == seps[(int)j]) {
	nowBlank = true;
	j = seps.length();
      }
    }

    bool nowField = ! nowBlank;

    if(nowField) {
      if(!onField) {
	count++;
	lastBegan = i;
      }
      onField = true;
    }
    else {
      if(onField) {
	if(count == field) {
	  // On target !!!
	  r = s;
	  r.remove(i, r.length() - i);
	  r.remove(0, lastBegan);
	  return r;
	}
      }
      onField = false;
    }
  }
  return r;
}

/** Retuns a field from a character string. */
QCString MJ_fieldC(const QCString &s, int field, const char *separators)
{
  QCString r;

  if(!separators)  separators = " \n\t";

  bool onField = false;
  int count = -1;
  int lastBegan = 0;

  uint sepslen = (uint) strlen(separators);
  
  for(uint i=0; i <= s.length(); i++) {
    bool nowBlank = false;

    // Try to match all separators
    for(uint j=0; j <= sepslen; j++) {
      if(s[(int)i] == separators[(int)j]) {
	nowBlank = true;
	j = sepslen;
      }
    }

    bool nowField = ! nowBlank;

    if(nowField) {
      if(!onField) {
	count++;
	lastBegan = i;
      }
      onField = true;
    }
    else {
      if(onField) {
	if(count == field) {
	  // On target !!!
	  r = s;
	  r.remove(i, r.length() - i);
	  r.remove(0, lastBegan);
	  return r;
	}
      }
      onField = false;
    }
  }
  return r;
}

/** Returns the beginning of one field. */

const char *MJ_fieldStart(const char *s, int field, const char * separators)
{
  if(!separators) separators = " \n\t";

  bool onField = false;
  int count = 0;

  int sepCount = strlen(separators);

  for(int i=0; s[i] != '\0'; i++) {
    bool nowBlank = false;
    
    // Try to match all separators
    for(int j=0; j <= sepCount; j++) {
      if(s[i] == separators[j]) {
	nowBlank = true;
	j = sepCount;
      }
    }

    bool nowField = ! nowBlank;

    if(nowField) {
      if(!onField) {

	if(count == field) return s + i;

	count++;
      }
      onField = true;
    }
    else onField = false;
    
  }
  return 0;
}

/** Returns the last field of a string. */
QString MJ_lastField(const QString &s, const char *separators)
{
  int f = MJ_fields(s, separators);
  QString r;
  if(f) r = MJ_field(s, f-1, separators);
  return r;
}

/** Tests field functions. */
void MJ_fieldTest()
{
  QString s1;
  s1 = "foo paina JUURES 0 /auto/punajuuri ";
  int i, f = MJ_fields(s1);

  MJ_cout() << "TEST 1\n" << s1 << "\nfields " << f << "\n";
  
  for(i=0; i < f; i++) {
    MJ_cout() << ":" << MJ_field(s1, i) << flush;
  }
  MJ_cout() << ":\n";

  f = MJ_fields(s1,"/ ");
  
  MJ_cout() << "TEST 2\n" << s1 << "\nfields " << f << "\n";
  for(i=0; i < f; i++) {
    MJ_cout() << ":" << MJ_field(s1, i, "/ ") << flush;
  }
  MJ_cout() << ":\n";
  
  s1 = "/usr/people/tilmonen/vsop/conductor/omt/src/mj";
  MJ_cout() << "SHORTEN: " << s1 << "\n"
	  << MJ_shorten(s1, 20) << "\n";
  MJ_cout().device()->flush();
}

QString MJ_stripLastField(const QString &s, const char * separator)
{
  QString tmp(s);
  int ind = tmp.findRev(separator);
  if(ind >= 0)
    tmp.truncate(ind);
  else
    tmp.truncate(0);
  return tmp;
}

/** Shorten a string. When a string is too long it can be shortened
    with this function. Middle part of the string id removed and
    replaced with string "...". 

    The operation looks like this:
    \begin{verbatim}
    QString s = MJ_shorten("/usr/people/tilmonen/vsop/conductor/omt/src/mj");
    // s = "/usr/peo.../src/mj"
    \end{verbatim}
    
*/
QString MJ_shorten(const QString &s, uint l)
{
  uint ss = s.length();

  if(l >= ss) return s;

  if(l < 5) {
    MJ_error(MJ_ERR_ARGUMENT, "MJ_shorten too short string asked");
    return s;
  }
  
  uint takeBlock = (l - 3) / 2;

  QString tmp1(s), tmp2(s);

  tmp1.truncate(takeBlock);
  tmp2.remove(0, ss - takeBlock);
  tmp1 += "..." + tmp2;
  return tmp1;
}

/** Adds unit text to the string in parenthesis. Also separated with
    space.  */
QString MJ_appendUnit(const QString &str, MJ_Unit u)
{
  QString ret;
  
  if(u != MJ_UNIT_BLANK) {
    ret = str + " (" + MJ_Tr::tr(MJ_UnitKeys[u]) + ")";
  }
  else ret = str;

  return ret;
}

/** Translates absolute file paths to relative paths. 
    
 */

QString MJ_absToRelPath(
			/// The absolute file path
			const QString &absPath,
			/// The current directory
			const QString &dotPath)
{
  MJ_trace(absPath, dotPath);

  char * dirLim = "/";

  int i, absLevels = MJ_fields(absPath, dirLim) - 1;
  int j, relLevels =  MJ_fields(dotPath, dirLim);
  
  bool diff = false;

  QString tmp1, tmp2;

  for(i = 0; i < absLevels && !diff; i++) {
    tmp1 =  MJ_field(absPath, i, dirLim);
    tmp2 =  MJ_field(dotPath, i, dirLim);

    if(tmp1 != tmp2) diff = true;
  }
  i--;

  if(!diff || !i) return absPath;

  tmp2 = QString::null;

  for(j = i; j <= absLevels; j++) {
    tmp2 += MJ_field(absPath, j, dirLim);

    if(j < absLevels) tmp2 += dirLim;
  }
  
  for(j = i; j < relLevels; j++) {
    tmp2 =  "../" + tmp2;
  }
  return tmp2;
}

/** Shift strings in an array. A quick hack for command line
    argument processing. */
void MJ_shiftBack(int startInd, int skip, int *size, char ** argv)
{
  for(int i=startInd + skip; i < *size; i++) {
    argv[i-skip] = argv[i];
  }
  (*size) -= skip;
}

/** Plugin menu path separator. 
    
 */
const char * MJ_pluginSeparator = "_";

/** Reports memory leak debugging info. Writes the number of
    MJ_Message and MJ_Parameter instances in the system. 

    @see MJ_Message MJ_Parameter
*/

void MJ_memDebugReport()
{
#ifdef MJ_MEM_DEBUG
  MJ_trace("MJ_memDebugReport # Number of messages in the system: ", 
	   MJ_Message::msgCount());
  MJ_trace("MJ_memDebugReport # Number of message ids in the system: ", 
	   MJ_MessageCarrier::msgIdCount());
  MJ_trace("MJ_memDebugReport # Number of parameters in the system: ", 
	   MJ_Parameter::paramCount());  
  MJ_trace("MJ_memDebugReport # Number of shared objects loaded: ", 
	   MJ_SoRefCount::soCount());    
  MJ_trace("MJ_memDebugReport # Number of MJ_CStrRemap objects: ", 
	   MJ_CStrRemap::count());    
  MJ_trace("MJ_memDebugReport # Number of plugin dbes: ", 
	   MJ_PluginDbeBase::pluginObjects());    
#endif
}

/** This function allocates and frees some memory. Effectively it does
    nothing, but causes work to memory management. Used to (sort of)
    guarantee the system has memory preallocated. */

void MJ_allocFree(uint bytes)
{
  free(calloc(bytes, 1));

  // MJ_Array<int> a;
  // a.resize(bytes / 4, false); 
  /* Fill in the memory. If the memory is not used the kernel may be
     smart enough to not actually allocate the memory (=delayed
     allocate). We want to make sure the memory is there so as it is
     needed we do not need to allocate it then. This is a dirty hack
     at its best, but such is life. */
  // a.setAll(0); 
}

/** Describe the version number of this Mustajuuri release. */

QString MJ_describeVersion()
{
  QString str;

  QTextOStream s(&str);

  MJ_Version v(MJ_VERSION_MAJOR, MJ_VERSION_MINOR, MJ_VERSION_SUBMINOR);

  v.write(& (s << "Mustajuuri "));
  s << "Copyright (C) Tommi Ilmonen 1999-2004\n"
    "Compiled on " DATETIME;


  return str;
}
