/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_module_dsp.h>

#include <mj_dsp_plugin.h>
#include <mj_error.h>
#include <mj_failure_diagnose.h>
#include <mj_global.h>
#include <mj_iostream.h>
#include <mj_kernel.h>
#include <mj_misc.h>
#include <mj_messages.h>
#include <mj_message_string.h>
#include <mj_message_xml.h>
#include <mj_mutex.h>
#include <mj_module_dsp_gui_default.h>
#include <mj_module_dsp_tui_default.h>
#include <mj_parameter.h>
#include <mj_parameters.h>
#include <mj_rt_driver.h>
#include <mj_satellite.h>

#include <stdlib.h>

#include <qbuffer.h> 
#include <qfile.h> 
#include <qtextstream.h> 

#include <typeinfo>

static MJ_MutexAuto MJ_Module_Mutex;

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

static const char MJ_SampleRateKey[] = "MJ_SAMPLE_RATE_DEFINITION";


MJ_SampleRateDef::MJ_SampleRateDef()
  : m_rate(44100), m_masterRate(44100), m_nominator(1), m_deNominator(1)
{}

/// Set master sampling rate.
void MJ_SampleRateDef::setMasterRate(int hz)
{ 
  m_masterRate = hz; 
  calculate();

  // MJ_trace2("MJ_SampleRateDef::setMasterRate # rate now %d", hz);
}

/// Fractional value describing the rate multiplier
void MJ_SampleRateDef::setRatio(int nominator, int deNominator)
{
  m_nominator   = nominator; 
  m_deNominator = deNominator;
  calculate();
}


/** Returns true if the number samples (in master rate) can be
    expressed exactly with this sampling rate. */
bool MJ_SampleRateDef::validLength(uint l)
{
  uint tmp = l * m_nominator;
  uint rem = tmp % m_deNominator;  
  return rem == 0;
}

/** Transfers master sampling rate sample count to local rate sample
    count. */
uint MJ_SampleRateDef::toThisLength(uint l)
{
  uint tmp = (l * m_nominator) / m_deNominator;
  return tmp;  
}

/** Transfers local sampling rate sample count to master rate sample
    count. */
uint MJ_SampleRateDef::fromThisLength(uint l)
{
  uint tmp = (l * m_deNominator) / m_nominator;
  return tmp;  
}

/// Writes the definition to a stream
bool MJ_SampleRateDef::write(QTextStream *s)
{
  (*s) << MJ_SampleRateKey << " " << m_masterRate << " " 
       << m_nominator << " " 
       << m_deNominator << "\n";
  return true;
}

/// Reads the definition from a stream
bool MJ_SampleRateDef::read(QTextStream *s)
{
  bool ok = true;
  QString str;

  str = s->readLine();

  if(MJ_field(str, 0) !=  MJ_SampleRateKey) {
    MJ_error(MJ_ERR_DATA_CORRUPTED, "MJ_SampleRateDef::read");
    return false;
  }
  
  m_masterRate  = MJ_field(str,1).toInt(&ok);
  m_nominator   = MJ_field(str,2).toInt(&ok);
  m_deNominator = MJ_field(str,3).toInt(&ok);

  calculate();

  return ok;
}

QDomElement MJ_SampleRateDef::write(QDomDocument &doc)
{
  QDomElement e = doc.createElement("SampleRateDef");

  MJ_DomUtil::addTextChild(e, doc, "rate", m_rate);
  MJ_DomUtil::addTextChild(e, doc, "master", m_masterRate);
  MJ_DomUtil::addTextChild(e, doc, "nominator", m_nominator);
  MJ_DomUtil::addTextChild(e, doc, "denominator", m_deNominator);

  return e;
}

bool MJ_SampleRateDef::read(QDomElement &e)
{
  if(e.tagName() != "SampleRateDef")
    return false;

  m_rate        = MJ_DomUtil::getDouble(e, "rate", 44100.0); 
  m_masterRate  = MJ_DomUtil::getInt(e, "master", 44100); 
  m_nominator   = MJ_DomUtil::getInt(e, "nominator", 1); 
  m_deNominator = MJ_DomUtil::getInt(e, "denominator", 1); 

  calculate();

  return true;
}


/// Updates the sampling rate value
void MJ_SampleRateDef::calculate()
{
  if(m_deNominator < 1 || m_masterRate < 1 || m_nominator < 1) {
    MJ_error(MJ_ERR_INTERNAL, "MJ_SampleRateDef::calculate");
    return;
  }
  m_rate = (double) m_masterRate * (double) m_nominator / 
    (double)m_deNominator; 
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


static const char MJ_SignalSourceDefKey[] = "MJ_SIGNAL_SOURCE_DEFINITION";

MJ_SignalSourceDef::MJ_SignalSourceDef()
  : m_sourceModule(0),
    m_sourceChannel(-1),
    m_sourceId(0)
{}

MJ_SignalSourceDef::~MJ_SignalSourceDef()
{
  if(m_sourceId) free(m_sourceId);
}

void MJ_SignalSourceDef::setModule(MJ_ModuleDsp *m)
{ 
  m_sourceModule = m; 
  setSourceId(m->id());
}

void MJ_SignalSourceDef::setSourceId(const char *id)
{ 
  MJ_delAndDup(& m_sourceId, id);
}


bool MJ_SignalSourceDef::validChannel()
{
  bool ok = true;
  if(m_sourceChannel < 0 || !m_sourceModule) ok = false;
  else if(m_sourceChannel >= m_sourceModule->dspOutChannels()) ok = false;
  return ok;
}


bool MJ_SignalSourceDef::connectSource(MJ_SignalSource *s)
{
  MJ_SignalSink *sink = & m_sourceModule->dspGetOutSignal(m_sourceChannel);
  s->connectSink(sink);
  return true;
}

bool MJ_SignalSourceDef::write(QTextStream *s)
{
  (*s) << MJ_SignalSourceDefKey << " " 
       << m_sourceId << " " << m_sourceChannel << "\n"
       << MJ_ModuleKeyWords[MJ_MKW_PLAINEND] << "\n";
  return true;
}

bool MJ_SignalSourceDef::read(QTextStream *s)
{
  QString str;

  str = s->readLine();

  if(MJ_field(str, 0) !=  MJ_SignalSourceDefKey) {
    MJ_error(MJ_ERR_DATA_CORRUPTED, "MJ_SignalSourceDef::read");
    return false;
  }
  MJ_delAndDup(& m_sourceId, MJ_field(str,1));
  m_sourceChannel = MJ_field(str,2).toInt();

  // Skip lines until we get end-word.
  for(int i=0; i < 100; i++) {
    str = s->readLine();
    if(str == MJ_ModuleKeyWords[MJ_MKW_PLAINEND])
      return true;
  }
  return false;
}


QDomElement MJ_SignalSourceDef::write(QDomDocument &doc)
{
  QDomElement e = doc.createElement("SignalSourceDef");

  MJ_DomUtil::addTextChild(e, doc, "source_channel", m_sourceChannel);
  MJ_DomUtil::addTextChild(e, doc, "source_id", m_sourceId);
  
  return e;
}

bool MJ_SignalSourceDef::read(QDomElement &e)
{
  if(e.tagName() != "SignalSourceDef")
    return false;

  m_sourceChannel = MJ_DomUtil::getInt(e, "source_channel"); 
  QString id      = MJ_DomUtil::getString(e, "source_id"); 
  
  setSourceId(id.ascii());

  return true;
}


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


MJ_SignalSource::MJ_SignalSource()
  : m_sink(0),
    m_sourceChannel(0)
{}

MJ_SignalSource::~MJ_SignalSource()
{}

/// Connects this source pointer to sink pointer.
void MJ_SignalSource::connectSink(MJ_SignalSink *xpSink)
{ 
  m_sink = xpSink;
  m_samples = xpSink->samples(); 
  if(!m_samples.size()) {
    MJ_error(MJ_ERR_ARGUMENT, "MJ_SignalSource::connectSink");
  }
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

static const char MJ_SignalSinkDefKey[] = "MJ_SIGNAL_SINK_DEFINITION";

MJ_SignalSink::MJ_SignalSink()
  : m_sinkChannel(0)
{}

MJ_SignalSink::~MJ_SignalSink()
{}


bool MJ_SignalSink::setLength(uint)
{
  return true;

  /* if(!m_sampleRate.validLength(masterRateSamples))
    return false;
  
  uint l = m_sampleRate.toThisLength(masterRateSamples);
  m_samples.resize(l);
  return true;*/
}

bool MJ_SignalSink::write(QTextStream *s)
{
  bool ok;
  (*s) << MJ_SignalSinkDefKey << "\n";
  ok = m_sampleRate.write(s);
  (*s) << MJ_ModuleKeyWords[MJ_MKW_PLAINEND] << "\n";
  return ok;
}

bool MJ_SignalSink::read(QTextStream *s)
{
  bool ok;
  QString str;

  str = s->readLine();
  
  if(MJ_field(str, 0) !=  MJ_SignalSinkDefKey) {
    MJ_error(MJ_ERR_DATA_CORRUPTED, "MJ_SignalSink::read");
    return false;
  }
  ok = m_sampleRate.read(s);
  
  if(!ok) return false;

  // Skip lines until we get end-word.
  for(int i=0; i < 100; i++) {
    str = s->readLine();
    if(str == MJ_ModuleKeyWords[MJ_MKW_PLAINEND])
      return true;
  }
  return false;
}
  
QDomElement MJ_SignalSink::write(QDomDocument &doc)
{
  QDomElement e = doc.createElement("SignalSink");
  
  e.appendChild(m_sampleRate.write(doc));

  return e;
}

bool MJ_SignalSink::read(QDomElement &e)
{
  if(e.tagName() != "SignalSink")
    return false;

  QDomElement e2 = e.firstChild().toElement();

  return m_sampleRate.read(e2);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

const char *MJ_ModuleKeyWords[MJ_MKW_SIZEOF] = {
  "MODULE_DSP_ID",
    "DSP_IO_DEFINITIONS",
    "DSP_INPUT_CHANNELS",
    "DSP_OUTPUT_CHANNELS",
    "DSP_OUTPUT_MESSAGES",
    "SIZE",
    "MODULE_DSP_END",
    "END"
};

bool MJ_ModuleDsp::m_moduleDspDebug = false;

/** Constructor. Creates a DSP module with a satellite. The module
    does not have any input or output channels (signal or message). 

    Member list classes have their autodelete behaviour enabled.
*/

MJ_ModuleDsp::MJ_ModuleDsp()
  : m_parent(0),
    m_dbe(0),
    m_satellite(new MJ_Satellite(this)),
    m_currentTime(0),
    m_id(0)
{
  const char *fname = "MJ_ModuleDsp::MJ_ModuleDsp";

  if(m_moduleDspDebug) trace(fname);

  if(!m_satellite) MJ_fatal(MJ_ERR_MEMORY, fname);

  m_signalSourceDefs.setAutoDelete(true);

  m_outputMsgRemap.setAutoDelete(true);
}

/** Destructor. Disconnects the message links and frees all output
    buffers. Hosting MJ_RtDriver must be null before this method is
    called. This method should not be called while the module is in
    real-time mode.*/

MJ_ModuleDsp::~MJ_ModuleDsp()
{
  const char *fname = "MJ_ModuleDsp::~MJ_ModuleDsp";
  if(m_moduleDspDebug) trace(fname);
  
  MJ_RtDriver * rtd = dynamic_cast<MJ_RtDriver *> (m_parent);

  if(rtd) {
    // Things may be so badly screwed that calls via "error" might go astray.
    MJ_fatal(MJ_ERR_INTERNAL, fname, "m_rtDriver must be NULL !!!");
  }

  msgDisconnectAll();

  // dspDropOutputPtrs(); // Forget memory.
  
  if(m_moduleDspDebug) trace(fname, "HALF");

  if(m_dbe) m_dbe->decrementReferenceCounter();

  delete m_satellite;
  free(m_id);

  if(m_moduleDspDebug) trace(fname, "EXIT");
}

/** Finds the top-level real-time driver. If the driver cannot be
    found, returns NULL.*/
MJ_RtDriver * MJ_ModuleDsp::topRtDriver()
{
  MJ_ModuleDsp *m = this;

  MJ_Base * parent = m->m_parent;

  while(parent) {
    MJ_RtDriver * rtd = dynamic_cast<MJ_RtDriver *> (parent);
    if(rtd) return rtd; 

    MJ_ModuleDsp *m = dynamic_cast<MJ_ModuleDsp *> (parent);
    if(m) parent = m->m_parent;
    else {
      parent = 0;
    }
  }
  return 0;
}


/** True class of the argument is MJ_PluginDbeDsp. Due to templates
    and inheritance etc it is easier to have this here. */

void MJ_ModuleDsp::setDbe(MJ_PluginDbeBase *dbe)
{ 
  m_dbe = (MJ_PluginDbeDsp *) dbe; 
  if(!m_dbe) MJ_fatal(MJ_ERR_ARGUMENT,"MJ_ModuleDsp::setDbe");
  m_dbe->incrementReferenceCounter();
}

/** Handle conflict situations. Most plugins need not take care of
    this. The usual reason for aa plugin to conflict with another is
    that they would share some common resource (such as access to
    certain audio hardware). 

    The default implementation returns false (=no conflict). */

bool MJ_ModuleDsp::conflicts(MJ_ModuleDsp *)
{
  return false;
}

/** Reads a new DSP module from a stream. Relies on the plugin
    database to create the module. */

MJ_ModuleDsp *MJ_ModuleDsp::readFromStream(QTextStream *s)
{
  const char *fname = "MJ_ModuleDsp::readFromStream";

  QIODevice *device = s->device();
  ulong p = device->at();

  if(m_moduleDspDebug)
    MJ_trace(fname, 0);

  // Plenty of global data to access. Use mutex locks.

  if(!MJ_kernel) return 0;

  QString str = s->readLine();

  if(str.isEmpty())
    str = s->readLine();

  QString str1 = MJ_field(str, 0);
  QString str2 = MJ_field(str, 1);
  
  MJ_Module_Mutex.lock();

  MJ_ModuleDsp *m = MJ_kernel->newDspModule(str1, MJ_PP_IDENTIFIER);
  if(!m) m = MJ_kernel->newDspModule(str2, MJ_PP_FUNCTIONALITY);

  MJ_Module_Mutex.unlock();

  if(!m) {
    MJ_error(MJ_ERR_ARGUMENT, fname, str);
    return 0;
  }

  if(m_moduleDspDebug)
    m->trace(fname, 2);

  device->at(p); // Return so the header may be rescanned.

  m->read(s);
  
  return m;
}

/** Write signal and message IO definitions to a stream. */

bool MJ_ModuleDsp::writeIODefs(QTextStream *s)
{
  uint i;

  (*s) << MJ_ModuleKeyWords[MJ_MKW_IODEFS] << "\n"
       << MJ_ModuleKeyWords[MJ_MKW_ID] << " " << m_id << "\n"
       << MJ_ModuleKeyWords[MJ_MKW_ICHANS] 
       << " " << m_signalSourceDefs.count() << "\n";
  
  for( i=0; i < m_signalSourceDefs.count(); i++) {
    m_signalSourceDefs.at(i)->write(s);
  }
  
  (*s) << MJ_ModuleKeyWords[MJ_MKW_OCHANS] 
       << " " << m_signalOutput.size() << "\n";

  for(i=0; i < m_signalOutput.size(); i++) {
    m_signalOutput[i].write(s);
  }

  (*s) << MJ_ModuleKeyWords[MJ_MKW_OMESSAGES] 
       << " " << m_outputMsgRemap.count() << "\n";
  
  for(i=0; i < m_outputMsgRemap.count(); i++) {
    m_outputMsgRemap.at(i)->write(s);
  }

  return true;
}

/** Read signal and message IO definitions from a stream. */

bool MJ_ModuleDsp::readIODefs(QTextStream *s, bool skipIODefs)
{
  QString str;
  uint tmp = 0;
  uint i = 0;

  if(skipIODefs) {
    /* Simply scan in IO defs to a temporary module, but forget the
       module. */
    MJ_ModuleDsp tmp;
    return tmp.readIODefs(s, false);
  }

  str = s->readLine();

  if(MJ_field(str, 0) != MJ_ModuleKeyWords[MJ_MKW_IODEFS]) {
    error(MJ_ERR_DATA_CORRUPTED, "readIODefs 1", str);
    return false;
  }

  str = s->readLine();
  
  if(MJ_field(str, 0) !=  MJ_ModuleKeyWords[MJ_MKW_ID]) {
    error(MJ_ERR_DATA_CORRUPTED, "readIODefs 1.5", str);
    return false;
  }

  MJ_delAndDup(&m_id,  MJ_field(str,1));

  str = s->readLine();
  
  if(MJ_field(str,0) != MJ_ModuleKeyWords[MJ_MKW_ICHANS]) {
    error(MJ_ERR_DATA_CORRUPTED, "readIODefs 2", str);
    return false;
  }
  
  tmp = MJ_field(str,1).toInt();

  if(!dspSetInChannels(tmp)) {
    error(MJ_ERR_DATA_CORRUPTED, "readIODefs 2.1", tmp);    
    return false;
  }
  
  for(i = 0; i < tmp; i++) {
    if(!m_signalSourceDefs.at(i)->read(s)) {
      error(MJ_ERR_DATA_CORRUPTED, "readIODefs 2.5", i);
      return false;
    }
  }
  
  str = s->readLine();
  
  if(MJ_field(str,0) != MJ_ModuleKeyWords[MJ_MKW_OCHANS]) {
    error(MJ_ERR_DATA_CORRUPTED, "readIODefs 3", str);
    return false;
  }

  tmp = MJ_field(str,1).toInt();

  if(!dspSetOutChannels(tmp)) {
    error(MJ_ERR_DATA_CORRUPTED, "readIODefs 2.1", tmp);
    return false;
  }

  for( i=0; i < m_signalOutput.size(); i++) {
    if(!m_signalOutput[i].read(s)) {
      error(MJ_ERR_DATA_CORRUPTED, "readIODefs", 4);
      return false;
    }
  }
  str = s->readLine();
  
  if(MJ_field(str,0) != MJ_ModuleKeyWords[MJ_MKW_OMESSAGES]) {
    error(MJ_ERR_DATA_CORRUPTED, "readIODefs 5", str);
    return false;
  }
  tmp = MJ_field(str,1).toInt();

  for(i=0; i < tmp; i++) {
    MJ_MessageRemap *rm = new MJ_MessageRemap;
    if(!rm->read(s)) {
      error(MJ_ERR_DATA_CORRUPTED, "readIODefs", 6);
      return false;
    }
    m_outputMsgRemap.append(rm);
  }

  return true;
}

/// Write a header to a stream.
bool MJ_ModuleDsp::writeHeader(QTextStream *s)
{
  if(!m_dbe) 
    (*s) << typeid(*this).name() << "\n";
  else 
    (*s) << m_dbe->parameter(MJ_PP_IDENTIFIER) << "  " 
	 << m_dbe->parameter(MJ_PP_FUNCTIONALITY) << "\n";

  return true;
}

/// Read module header from a stream.
bool MJ_ModuleDsp::readHeader(QTextStream *s)
{
  const char * fname = "MJ_ModuleDsp::readHeader";

  if(m_moduleDspDebug) trace(fname);

  QString correctStr;

  if(!m_dbe) 
    correctStr = typeid(*this).name();
  else
    correctStr = m_dbe->parameter(MJ_PP_IDENTIFIER);

  QString str;
  str = s->readLine();
  
  QString str1 = MJ_field(str, 0);
  QString str2 = MJ_field(str, 1);

  bool matchFirst(str1 == correctStr);
  bool matchSecond(false);

  if(!str2.isEmpty()) 
    matchSecond = (str2 == m_dbe->parameter(MJ_PP_FUNCTIONALITY));

  if(!matchFirst && !matchSecond) {

    if(m_moduleDspDebug) {
      QString tmp;
      tmp.sprintf("%s # %s != %s", fname, 
		  (const char *) str, 
		  (const char *) correctStr);
      error(MJ_ERR_DATA_CORRUPTED, tmp);
    }
    return false;
  }

  return true;
}

/// Write the module to a steam
bool MJ_ModuleDsp::write(QTextStream *s)
{
  bool ok1 = writeHeader(s);
  bool ok2 = writeIODefs(s);
  bool ok3 = writeParameters(s);
  return ok1 && ok2 && ok3;
}

/// Read the module from a stream
bool MJ_ModuleDsp::read(QTextStream *s, bool skipIODefs)
{
  if(!readHeader(s)) return false;
  if(!readIODefs(s, skipIODefs)) return false;
  // if(!readParameters(s)) return false;
  readParameters(s);
  return true;
}

/** Write the current parameters to a steam.
      
    The default implementation tries to get the input parameters via
    the normal method and save them. This is usually good enough for
    simple plugins.

    Reasons to override this method: If your class contain complex
    structures, you may wish save in the format of your choice
    (including binary formats). By overriding this method you can put
    any parameters you wish to the file/stream. You may also create a
    temporary binary stream object to write binary data. */

bool MJ_ModuleDsp::writeParameters(QTextStream *s)
{
  MJ_Parameters params;
  params.setAutoDelete(true);
  msgGetInputs(&params);

  MJ_MessageList msgs;
  msgs.setAutoDelete(true);
  if(!msgs.copyValues(&params)) return false;

  msgs.writeMessages(s);
  return true;
}

/** Read parameters from a stream. Complements writeParameters. */
bool MJ_ModuleDsp::readParameters(QTextStream *s)
{
  const char *fname = "MJ_ModuleDsp::readParameters";

  if(m_moduleDspDebug) trace(fname);
  
  MJ_MessageList msgs;
  msgs.setAutoDelete(true);
  msgs.readMessages(s);
  
  bool ok = msgs.deliverMessages(this);
  
  if(!ok && m_moduleDspDebug) 
    error(MJ_ERR_INTERNAL, fname);
  
  return ok;
}

QDomElement MJ_ModuleDsp::writeIODefs(QDomDocument &doc)
{
  uint i;

  QDomElement e = doc.createElement("IODefs");
  
  MJ_DomUtil::addTextChild(e, doc, "id", m_id);
  
  // Source definitions:

  QDomElement sources = doc.createElement("Sources");

  for(i=0; i < m_signalSourceDefs.count(); i++)
    sources.appendChild(m_signalSourceDefs.at(i)->write(doc));

  e.appendChild(sources);

  // Output definitions:

  QDomElement sinks = doc.createElement("Sinks");

  for(i=0; i < m_signalOutput.size(); i++)
    sinks.appendChild(m_signalOutput[i].write(doc));

  e.appendChild(sinks);

  return e;
}

bool MJ_ModuleDsp::readIODefs(QDomElement &e)
{
  if(e.tagName() != "IODefs")
    return false;
  
  setId(MJ_DomUtil::getString(e, "id"));

  QDomElement sources = e.namedItem("Sources").toElement();

  QDomNode n = sources.firstChild();
  
  while ( !n.isNull() ) {

    QDomElement n2 = n.toElement();
    MJ_SignalSourceDef * ssd = new MJ_SignalSourceDef();

    if(ssd->read(n2))
      m_signalSourceDefs.append(ssd);
    else
      delete ssd;

    n = n.nextSibling();
  }

  QDomElement sinks = e.namedItem("Sinks").toElement();

  uint i, c = sinks.childNodes().count();

  m_signalOutput.resize(c);

  n = sources.firstChild();

  for(i=0; i < c; i++) {
    QDomElement n2 = n.toElement();
    m_signalOutput[i].read(n2);
    n = n.nextSibling();
  }

  return true;
}

QDomElement MJ_ModuleDsp::writeHeader(QDomDocument &doc)
{
  QDomElement e = doc.createElement("Type");
  
  if(!m_dbe)
    MJ_DomUtil::addTextChild(e, doc, "cpp_class", typeid(*this).name());
  else {
    MJ_DomUtil::addTextChild(e, doc, "type", 
			     m_dbe->parameter(MJ_PP_IDENTIFIER));
    MJ_DomUtil::addTextChild(e, doc, "functionality", 
			     m_dbe->parameter(MJ_PP_FUNCTIONALITY));
  }

  return e;
}

bool MJ_ModuleDsp::readHeader(QDomElement &e)
{
  if(e.tagName() != "Type")
    return false;
  
  if(m_dbe) {
    if((MJ_DomUtil::getString(e, "type") != 
	m_dbe->parameter(MJ_PP_IDENTIFIER)) &&
       (MJ_DomUtil::getString(e, "functionality") != 
	m_dbe->parameter(MJ_PP_FUNCTIONALITY)))
      return false;
  }
  
  return true;
}

QDomElement MJ_ModuleDsp::write(QDomDocument &doc)
{
  QDomElement e = doc.createElement("DSP");

  e.appendChild(writeHeader(doc));
  e.appendChild(writeIODefs(doc));
  e.appendChild(writeParameters(doc));

  return e;
}

bool MJ_ModuleDsp::read(QDomElement &e)
{
  if(e.tagName() != "DSP")
    return false;
  
  QDomElement e2 = e.namedItem("Type").toElement();

  bool ok = readHeader(e2);

  e2 = e.namedItem("IODefs").toElement();
  ok = ok && readIODefs(e2);

  e2 = e.namedItem("Parameters").toElement();
  ok = ok && readIODefs(e2);

  return ok;
}

QDomElement MJ_ModuleDsp::writeParameters(QDomDocument &doc)
{
  QDomElement e = doc.createElement("Parameters");

  MJ_Parameters params;
  params.setAutoDelete(true);
  msgGetInputs(&params);
  
  MJ_MessageList msgs;
  msgs.setAutoDelete(true);

  if(!msgs.copyValues(&params)) return e;
  
  msgs.writeMessages(e, doc);

  return e;
}

bool MJ_ModuleDsp::readParameters(QDomElement &e)
{
  const char *fname = "MJ_ModuleDsp::readParameters(QDom)";

  if(e.tagName() != "Parameters")
    return false;
  
  MJ_MessageList msgs;
  msgs.setAutoDelete(true);

  msgs.readMessages(e);

  bool ok = msgs.deliverMessages(this);
  
  if(!ok && m_moduleDspDebug) 
    error(MJ_ERR_INTERNAL, fname);
  
  return ok;
}


/** Write the object to standard output. This is a utility method for
    debugging etc. */

void MJ_ModuleDsp::toCout()
{
  write(&MJ_cout());
  MJ_cout().device()->flush();
}

/** Test that IO definition read and writing works. Writes the
    definitions to a file, reads them from the file and prints
    everything to the screen. */

void MJ_ModuleDsp::testIoDefRw()
{
  const char *filename="foo.io";

  MJ_ModuleDsp m1, m2;

  m1.setId("m15");
  m1.dspSetInChannels(2);
  m1.dspSetOutChannels(2);
  m1.dspConnectInput("m1", 0, 0);
  m1.dspConnectInput("m2", 4, 1);

  {
    QFile f1(filename);
    f1.open(IO_WriteOnly);
    QTextStream s1(&f1);
    m1.writeIODefs(&s1);
  }
  MJ_cout() << "WRITTEN\n";
  MJ_cout().device()->flush();

  {
    QFile f2(filename);
    f2.open(IO_ReadOnly);
    QTextStream s2(&f2);
    m2.readIODefs(&s2);
  }
  m2.toCout();
}

/** Set identifier string. */
void MJ_ModuleDsp::setId(const char * id)
{
  if(m_moduleDspDebug) trace("MJ_ModuleDsp::setId", id);
  MJ_delAndDup(& m_id, id);
}

/** Check that given I/O configuration is valid. 
    
    \b NOTE: The "in" and "out" values may be changed if need be. 

    The default implementation uses the creator DBE to determine a new
    IO conf (see MJ_DspOptions). 
    
    <B> When to override:</B> Some plugins do not have a fixed set of
    I/O channels. For example file input plugins produce n channels of
    audio and there is no way to tell the number of channels before
    the file is named.  */

bool MJ_ModuleDsp::dspAcceptIoChans(int * in, int * out)
{
  if(m_dbe) {
    
    MJ_DspOptions *opt = m_dbe->options();

    if(opt->acceptIoChans(*in, *out))
      return true;

    *in = opt->newInChans(*in);
    *out = opt->newOutChans(*out);

    return opt->acceptIoChans(*in, *out);
  }
  return true;
}

/** Tries to get the input buffer pointers correct. */

bool MJ_ModuleDsp::dspCompileInputs()
{
  const char *fname = "MJ_ModuleDsp::compileDspInputs";

  if(m_moduleDspDebug) trace(fname);
  bool ok = true;

  uint insize = m_signalSourceDefs.count();

  if(m_signalInput.size() != insize)
    ok = m_signalInput.resize(insize, false);

  if(!ok) return false;

  for(uint i=0; i < insize; i++) {
    MJ_SignalSourceDef *d = m_signalSourceDefs.at(i);
    MJ_ModuleDsp *source = findModule(d->sourceId());

    if(!source) {
      // if(m_moduleDspDebug) {
      error2(MJ_ERR_INTERNAL, "%s # Unfound ID \"%s\"", 
	     fname, d->sourceId());
      // }
      return false;
    }
    d->setModule(source);

    if(!d->validChannel()) {
      // if(m_moduleDspDebug) {
      error2(MJ_ERR_INTERNAL, "%s %s this->m_id = %s inchan = %d "
	     "source->m_id = %s source channel = %d",
	     fname, typeid(*this).name(), 
	     m_id, i, source->m_id, d->channel());
      // }
      return false;
    }

    d->connectSource(&m_signalInput[i]);
  }
  
  if(m_moduleDspDebug) trace(fname, "successfull");
  return true;
}

/// Returns the number of current input channels

int MJ_ModuleDsp::dspInChannels()
{
  return m_signalSourceDefs.count();
}

/** Sets the number of input channels. Both the input definitions and
    input pointer structures are updated. */
bool MJ_ModuleDsp::dspSetInChannels(int chans)
{
  if(m_moduleDspDebug) trace("MJ_ModuleDsp::setInChannels", chans);
  // uint uchans = (uint) chans;
  // if(uchans == m_signalSourceDefs.count()) return true;

  bool ok = m_signalSourceDefs.resize(chans) && 
    m_signalInput.resize(chans, false);

  return ok;
}

/** Connects a source module signal output to signal input of this
    module. 

    The physical signal links are not created at this time. Instead
    the link definition is saved. Afterwards (before signal processing
    can start) the links need to be made with
    MJ_ModuleDsp::dspCompileInputs. */
bool MJ_ModuleDsp::dspConnectInput
(const char* sourceId, int sourceChannel, int inputChannel)
{
  const char *fname = "MJ_ModuleDsp::connectDspInput";

  if(!sourceId) {
    error(MJ_ERR_ARGUMENT, fname, "NULL sourceId");
    return false;
  }

  if(m_moduleDspDebug) {
    trace2("%s # %s sourceId = %s %d -> %d", 
	   fname, typeid(*this).name(), 
	   sourceId, sourceChannel, inputChannel);
  }

  if((uint) inputChannel >= m_signalSourceDefs.count()) {
    if(m_moduleDspDebug) {
      trace2("%s # Available channels = %d", 
	      fname, (int) m_signalSourceDefs.count());
    }
    return false;
  }

  m_signalSourceDefs.at(inputChannel)->set(sourceId, sourceChannel);

  return true;
}

/// Returns the number of current output channels.

int MJ_ModuleDsp::dspOutChannels()
{
  return (int) m_signalOutput.size();
}

/// Sets the number of output channels. 

bool MJ_ModuleDsp::dspSetOutChannels(int chans)
{
  if(m_moduleDspDebug) trace("MJ_ModuleDsp::dspSetOutChannels", chans);
  // trace("MJ_ModuleDsp::setOutChannels", chans);
  
  if(chans == (int) m_signalOutput.size()) return true;
  
  dspDropOutputPtrs();

  bool ok = m_signalOutput.resize((uint) chans, false);
  return ok;
}

/** Sets the values of samples in output to zero. 
      
    This is done between real-time usage sessions. Otherwise old
    values would be present as the processing starts.  */

void MJ_ModuleDsp::dspClearOutputs()
{
  // if(m_moduleDspDebug) trace("MJ_ModuleDsp::clearOutputs");
  for(uint i=0; i < m_signalOutput.size(); i++)
    m_signalOutput.getFast(i).samples().setAll(0.0);
}

/** Drop all output pointers. If the pointers were created with an
    external allocator, then they are also freed with that method. */
void MJ_ModuleDsp::dspDropOutputPtrs()
{
  for(uint i=0; i < m_signalOutput.size(); i++)
    m_signalOutput.getFast(i).samples().dropData();
}

/** Finds a module with given id. The default implementation searches
    the module from the parent module (if the parent exists). */
MJ_ModuleDsp *MJ_ModuleDsp::findModule(const char * id)
{
  const char * fname = "MJ_ModuleDsp::findModule";
  
  if(m_moduleDspDebug) trace(fname, id);
  
  MJ_ModuleDsp *m=0;
  
  MJ_ModuleDsp *parentDsp = dynamic_cast<MJ_ModuleDsp *> (m_parent);

  if(parentDsp)
    m = parentDsp->findModule(id);

  if(!m) error2(MJ_ERR_INTERNAL, 
		"%s # Could not find module \"%s\", parent = %p, %p - \"%s\"",
		fname, id, m_parent, parentDsp,
		m_parent ? typeid(*m_parent).name() : "");

  return m;
}

/** Sets the time stamp object (master clock). 

    Besides setting the time stamp this method also resizes the output
    signal buffers. This needs to be done as the buffer size depends
    on the time stamp parameters. 

    This method uses MJ_Malloc to allocate memory for signal
    buffers. */
bool MJ_ModuleDsp::dspSetTimer(MJ_TimeStamp *t)
{ 
  const char * fname = "MJ_ModuleDsp::dspSetTimer";

  if(!t)
    MJ_fatal(MJ_ERR_NULL_ARGUMENT, fname);

  m_currentTime = t; 

  MJ_Allocator *allocator = mjDspAllocator();

  dspDropOutputPtrs();
  
  uint masterRateSamples = t->frameSize();

  // trace2("%s # new master rate: %d", fname, (int)t->hz());
  
  for(uint i=0; i < m_signalOutput.size() && allocator; i++) {
    
    MJ_SignalSink *sink = & m_signalOutput.getFast(i);
    
    sink->sampleRate().setMasterRate((int)t->hz());
    
    if(!sink->sampleRate().validLength(masterRateSamples)) {
      error(MJ_ERR_ARGUMENT, fname);
      return false;
    }

    uint l = sink->sampleRate().toThisLength(masterRateSamples);

    if(m_moduleDspDebug) trace2("%s # Resizing output %u to %u samples", 
				fname, i, l);

    sink->samples().allocatorResize(allocator, l);
  }

 
  return true;
}

/** Check if the module is in realtime mode. It is always safe to call
    this method. */
bool MJ_ModuleDsp::dspIsRealTime() const
{
  if(m_currentTime)
    if(m_currentTime->isRealTime())
      return true;
  return false;
}

/** Determines current sampling rate. This utility method works by
    finding the first input or output buffer and querying its sampling
    rate. If the rate cannot be determined this method returns the
    default sampling rate = 44.1 kHz.

    It is always safe to call this method.

    You should use this method only in cases when the module has
    fairly simple IO conf - with multirate systems this method simply
    returns the first value it can get. */

int MJ_ModuleDsp::dspSamplingRate() const
{
  if(m_signalOutput.size())
    return m_signalOutput.getFastConst(0).sampleRateConst().rateUInt();

  for(uint i=0; i < m_signalInput.size(); i++) {

    const MJ_SignalSink *sink = m_signalInput.getFastConst(i).sinkConst();

    if(sink)
      return sink->sampleRateConst().rateUInt();
  }

  // trace("MJ_ModuleDsp::dspSamplingRate # No real sampling rate");

  return 44100;
}

/** Prepare for real-time processing. Fairly often it is
    necessary to do some preparations prior to signal processing. Such
    tasks are memory allocation and cleaning (filling any old buffers
    with zero etc). 
    
    By the time this method is called certain initializetions have
    alrady been done by the host system. Such tasks are creation and
    linking of audio and message buffers. Also the time-stamp should
    be in place. A module can at this stage check that everything is
    correctly initialized.

    This method is called n milliseconds before the signal processing
    starts.  

    If this method returns false the host will not start signal
    processing.

    The default implementation returns true. */
bool MJ_ModuleDsp::dspPrepare()
{
  const char * fname = "MJ_ModuleDsp::dspPrepare";

  if(m_moduleDspDebug) trace(fname);

  if(m_dbe) {
    int chansIn = m_signalInput.size();
    int chansOut = m_signalOutput.size();

    if(!m_dbe->options()->acceptIoChans(chansIn, chansOut)) {
      error(MJ_ERR_INTERNAL, fname, "Bad I/O setup");
      return false;
    }
  }

  return true;
}

/** QUICK preparations for real-time processing. 

    This method is called just before the signal processing starts.  

    The default implementation returns true. */

bool MJ_ModuleDsp::dspPrepareFast()
{
  if(m_moduleDspDebug) trace("MJ_ModuleDsp::dspPrepareFast");
  return true;
}

/** Process one frame of audio. ALL modules need to reimplement this
    is they are to be worth anything. The default implementation
    writes a warning message to the user.
    
    The audio frame is measured in samples of master rate.  */
void MJ_ModuleDsp::dspProcessPeriod()
{
  error(MJ_ERR_INTERNAL, "MJ_ModuleDsp::processFrame needs to be overridden");
}

/** QUICKLY stop the processing. This method is called just before the
    signal processing starts. 

    The default implementation returns true. */

bool MJ_ModuleDsp::dspStopFast()
{
  if(m_moduleDspDebug) trace("MJ_ModuleDsp::dspStopFast");
  return true;
}

/** Stop the processing and free resources. This method should be
    overridden if the plugin needs to do some cleanup after signal
    processing (close files, free memory etc.).
    
    The default implementation returns true.
*/

bool MJ_ModuleDsp::dspStop()
{
  if(m_moduleDspDebug) trace("MJ_ModuleDsp::dspStop");
  return true;
}

/** This method is called as the host is running out of CPU. If the
    plugin can do something to lower the CPU-load it should return
    true.

    The default implementation returns false. */
bool MJ_ModuleDsp::dropLoad(MJ_Load load)
{
  if(m_moduleDspDebug) trace("MJ_ModuleDsp::dropLoad");
  (void) load;
  return false;
}

/** A utility method that delivers a string message to the module. */
bool MJ_ModuleDsp::msgProcessString(const char * i, const QString &s)
{
  MJ_MessageString ms(s);
  return msgProcess(i, &ms);
}

/** A utility method that delivers a double message to the module. */
bool MJ_ModuleDsp::msgProcessDouble(const char * i, double d)
{
  MJ_MessageDouble md(d);
  return msgProcess(i, &md);
}

/** A utility method that delivers a double message to the module. */
bool MJ_ModuleDsp::msgProcessInt(const char * i, int d)
{
  MJ_MessageInt mi(d);
  return msgProcess(i, &mi);
}

/** A utility method that delivers a blank message to the module. */
bool MJ_ModuleDsp::msgProcessEmpty(const char * i)
{
  MJ_Message m;
  return msgProcess(i, &m);
}

/** Process a message. If the module wants to receive messages it
    should override this method. Messages are timestamped information
    packets that are sent asynchronously between modules (and user
    interface).*/
bool MJ_ModuleDsp::msgProcess(const char *, MJ_Message *)
{
  error(MJ_ERR_UNIMPLEMENTED, "MJ_ModuleDsp::processMessageNow");
  return false;
}

/** Tells the messages that that can be used for
    input. Reimplementaion is seldom needed. The default
    implmentaion simply calls "msgGetInputName" with increasing
    count (starting from zero) and then "msgGetInput" with the found
    name. Acquired parameters are appended to the list. 
    
    You might want to reimplement this method if you have a strange
    combination of parameter names (for example a hiearchical
    structure that cannot be presented as a list too naturally). */

void MJ_ModuleDsp::msgGetInputs(MJ_Parameters *p)
{
  const char * fname = "MJ_ModuleDsp::msgGetInputs";

  if(m_moduleDspDebug) trace(fname);

  QCString str;

  for(uint i=0; true; i++) {
    str = msgGetInputName(i);

    if(str.isEmpty()) return;

    MJ_Parameter *param = msgGetInput(str);

    if(!param) return;
    p->append(param);

    if(i == 10000) {
      error(MJ_ERR_INTERNAL, fname, "Extremely many input messages");
    }
  }
}

/** Get one of the current input parameters. Override this method to
    declare the parameter space of your object. The returned parameter
    must be a new object, the system will eventually delete the
    parameter object. If the argument exceedthe parameter range this
    method should simply return null.
      
    Default implementaion returns null. */

MJ_Parameter *MJ_ModuleDsp::msgGetInput(const char *)
{
  return 0;
}

/** A utility method that can be used to get a particular input
    parameter value. 
    
    Relies on "msgGetInputName" and "msgGetInput". */

MJ_Parameter *MJ_ModuleDsp::msgGetInputIndex(int i)
{
  QCString s(msgGetInputName(i));
  if(s.isEmpty()) return 0;
  return msgGetInput(s);
}

/** Returns the string associated with particular integer index. 

    It is not absolutely necessary to use this method even if the
    plugin does have input parameters. An alternative to overriding
    this method is to override msgGetInputs.
    
    Default implementation returns an empty string. 

    The index starts from zero and goes linearly up. As the first
    empty string is returned Mustajuuri stops seeking new
    parameters.  */

QCString MJ_ModuleDsp::msgGetInputName(int)
{
  QCString s;
  return s;
}

/// Tells the messages that that appear at output.
void MJ_ModuleDsp::msgGetOutputs(MJ_Parameters *p)
{
  QCString str;

  for(uint i=0; true; i++) {
    str = msgGetOutputName(i);
    MJ_Parameter *param = msgGetOutput(str);
    if(!param) return;
    p->append(param);
  }
}

/** Get one of the current output parameters. Similar to
    "getInputMessage". 
    
    There is seldom real need to override this method.  */

MJ_Parameter *MJ_ModuleDsp::msgGetOutput(const char *)
{
  return 0;
}

MJ_Parameter *MJ_ModuleDsp::msgGetOutputIndex(int i)
{
  QCString s(msgGetOutputName(i));
  if(s.isEmpty()) return 0;
  return msgGetOutput(s);
}

/** Returns name of the nth input parameter. */

QCString MJ_ModuleDsp::msgGetOutputName(int)
{
  QCString s;
  return s;
}

/** Removes given object from list of message receivers. */

void MJ_ModuleDsp::msgDisconnectTarget(MJ_ModuleDsp *m)
{
  if(m_moduleDspDebug) trace("MJ_ModuleDsp::disconnectMsgTarget");

  int tmp;

  while((tmp = m_outputMsgRemap.findTarget(m)) != -1) {
    m_outputMsgRemap.remove((uint) tmp);
  }
}

/** Removes given object from list of message sources. */

void MJ_ModuleDsp::msgDisconnectSource(MJ_ModuleDsp *m)
{
  if(m_moduleDspDebug) trace("MJ_ModuleDsp::disconnectMsgSource");

  int tmp;
  
  while((tmp = m_inputMsgSources.find(m)) != -1) {
    m_inputMsgSources.remove((uint) tmp);
  }
}


    
/** Disconnects all message inputs and outputs.
    
    This can be used as one is about to delete a DSP object. Prior to
    deleting the object one should use this method to unlink the
    object from srouce/target message receivers. */
void MJ_ModuleDsp::msgDisconnectAll()
{
  if(m_moduleDspDebug) trace("MJ_ModuleDsp::disconnectMessages");

  msgDisconnectInput();
  msgDisconnectOutput();
  if(m_moduleDspDebug) trace("MJ_ModuleDsp::disconnectMessages EXIT");
}

/** Disconnect all message links from other DSP modules to this module. */

void MJ_ModuleDsp::msgDisconnectInput()
{
  for(uint i=0; i < m_inputMsgSources.count(); i++) {
    m_inputMsgSources.at(i)->msgDisconnectTarget(this);
  }
  m_inputMsgSources.clear();
}  

/** Disconnect all message links from this module to other DSP modules. */
void MJ_ModuleDsp::msgDisconnectOutput()
{  
  for(uint i=0; i < m_outputMsgRemap.count(); i++) {
    MJ_MessageRemap *tmp = m_outputMsgRemap.at(i);
    if(tmp->target()) tmp->target()->msgDisconnectSource(this);
  }
  m_outputMsgRemap.clear();
}

/** Create message links from this module to receiving DSP modules. */
bool MJ_ModuleDsp::msgCompileOutputs()
{
  if(m_moduleDspDebug) 
    trace2("MJ_ModuleDsp::msgCompileOutputs # %d", m_outputMsgRemap.count());  

  bool ok = true;

  for(uint i=0; i < m_outputMsgRemap.count(); i++) {

    MJ_MessageRemap *rm = m_outputMsgRemap.at(i);
    MJ_ModuleDsp *m = findModule(rm->targetId());

    rm->setTarget(m); // NULL argument does not hurt.

    if(m) {
      if(m->m_inputMsgSources.find(this) < 0) {
	m->m_inputMsgSources.append(this);
      }
    }
    else {
      ok = false;
    }
  }
  return ok;
}

/** Connects messages between objects. The information about the
    linkage is saved by the source module. If the source module cannot
    be found this method returns false.
    
    Saving the information to the source rather than target is a bit
    illogical. This is done to make this method resemble
    MJ_ModuleDsp::dspConnectInput.

    In case of success this method returns true.  */

bool MJ_ModuleDsp::msgConnectInput(const char * sourceId, 
				   const char * sourceMsg,
				   const char * targetMsg)
{
  if(m_moduleDspDebug) trace("MJ_ModuleDsp::msgConnectInput", sourceId);

  MJ_ModuleDsp *m = findModule(sourceId);

  if(!m) return false;

  MJ_MessageRemap *rm = new MJ_MessageRemap;
  rm->set(sourceMsg, targetMsg, sourceId, m_id);
  rm->setTarget(this);
  
  bool ok = m->msgAddOutputRemap(rm);

  if(!ok) delete rm;

  return ok;
}

/** Adds a remapper to message output handler. Once the remapper is in
    place all new messages will be delivered to the target defined by
    the remapper. 

    If a remapper that equals the given remap is already in place this
    method returns true.
*/

bool MJ_ModuleDsp::msgAddOutputRemap(MJ_MessageRemap *rm)
{
  if(m_moduleDspDebug) trace("MJ_ModuleDsp::msgAddOutputRemap");

  if(m_outputMsgRemap.findRemap(rm) >= 0) 
    return false;
  
  m_outputMsgRemap.append(rm);
  return true;
}

/** Create a new GUI for this object. The GUI objects method
    "setModule" is NOT called. Set the mutexes before you do it.*/

MJ_ModuleDspGui *MJ_ModuleDsp::newGui
(MJ_RtControl *rtparent,MJ_GuiSize s, QWidget *parent, const char *name)
{
  if(m_moduleDspDebug) trace("MJ_ModuleDsp::newGui");

  MJ_ModuleDspGui *g = 0;
  
  if(m_dbe) g = m_dbe->newGui(rtparent, this, s, parent, name);
  
  if(!g) {
    switch (s) {
    case MJ_GUI_STAMP:
      g = new MJ_ModuleDspGuiPostCard
	(rtparent, this, true, 1, MJ_GUI_STAMP, parent, name);
      // g = new MJ_ModuleDspGuiStamp(parent, name);
      break;
    case MJ_GUI_POSTCARD:
      g = new MJ_ModuleDspGuiPostCard
	(rtparent, this, true, 5, MJ_GUI_POSTCARD, parent, name);
      break;
    case MJ_GUI_OWN_WINDOW:
      g = new MJ_ModuleDspGuiPostCard
	(rtparent, this, true, (uint) -1, MJ_GUI_OWN_WINDOW, 0, name);
      break;
    case MJ_GUI_SIZEOF:
      break;
    }
  }
  return g;
}

/** Create a new TUI (Text-mode User Interface) for this object. */

MJ_ModuleDspTui *MJ_ModuleDsp::newTui(MJ_RtControl *parent,
				      const char *name)
{
  return new MJ_ModuleDspTuiDefault(parent, this, name);
}

/** Gets all children of this module. As Mustajuuri plugins can be
    recursive a module can be responsible for many child modules. In
    some cases it is necessary to access the child modules. For these
    cases it is necessary to be able to get listing of them. */

void MJ_ModuleDsp::getChildren(QList<MJ_ModuleDsp> &l, bool recursive)
{
  (void) l;
  (void) recursive;
}

/** Push messages to receivers. This is a utility method. */

void MJ_ModuleDsp::msgPushOutput(const char * id, MJ_Message *msg)
{
  // Timing undefined, make a good guess:
  if(msg->timeStamp() == -1 && m_currentTime) {
    msg->setTimeStamp((long) m_currentTime->stamp());
  }

  m_outputMsgRemap.first();
  uint count = m_outputMsgRemap.count();

  MJ_MessageRemap *rm;

  for(uint i=0; i < count; i++) {
    rm = m_outputMsgRemap.current();
    rm->deliver(id, msg);
    m_outputMsgRemap.next();
  }
}

/** Allocate memory for DSP activity. If you need memory area that is
    accessed very frequently, then it you should use this method to
    allocate the memory.
    
    This method is provided for optimized memory-behaviour. The memory
    allocated with this method exists between dspPrepare and dspStop
    calls. After dspStop the allocator will free the memory. 

    This method is a short-cut to find the current MJ_Malloc object
    from top-level MJ_RtDriver and allocate memory from it.

    <B>Warning:</B> Only use this method to allocate memory if you
    *know* what you are doing. 

    <B> Why ?</B> It seems Mustajuuri has very different performance
    at different runs. It is assumed non-optimal memory lauout is the
    cause for these problems. To counter such behaviour an internal
    large-block memory allocator (MJ_Malloc) is used.  */
MJ_Allocator * MJ_ModuleDsp::mjDspAllocator()
{
  MJ_ModuleDsp *m = this;

  MJ_Base * parent = m->m_parent;

  while(parent) {
    MJ_RtDriver * rtd = dynamic_cast<MJ_RtDriver *> (parent);
    if(rtd) return rtd->dspMalloc(); 

    MJ_ModuleDsp *m = dynamic_cast<MJ_ModuleDsp *> (parent);
    if(m) parent = m->m_parent;
    else {
      parent = 0;
    }
  }
  // MJ_error(MJ_ERR_INTERNAL, "MJ_ModuleDsp::mjDspMalloc FAILED");
  return 0;
}

/** Overridden from MJ_Base. Returns member m_parentModule. */
MJ_Base * MJ_ModuleDsp::mjParent()
{
  return m_parent;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


MJ_ModuleDspNull::~MJ_ModuleDspNull()
{}

void MJ_ModuleDspNull::dspProcessPeriod()
{}
