/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_MODULE_DSP_H
#define MJ_MODULE_DSP_H

#include <mj_base.h>
#include <mj_config.h>
#include <mj_gui_enums.h>
#include <mj_list.h>
#include <mj_message.h>
#include <mj_message_pass.h>
#include <mj_time_stamp.h>
#include <mj_vector.h>

#include <qdom.h>

// Empty declaration to decrease include links here.

class MJ_ModuleDsp;

class MJ_Allocator;
class MJ_ModuleDspGui;
class MJ_ModuleDspTui;
class MJ_SignalSource;
class MJ_SignalSink;
class MJ_Parameter;
class MJ_Parameters;
class MJ_PluginDbeBase;
class MJ_PluginDbeDsp;
class MJ_RtDriver;
class MJ_RtControl;
class MJ_Satellite;

class QWidget;

/** Specifify some part of module IO setup. */

enum MJ_ModuleIO {
  MJ_MIO_SIGNAL_INPUT        = 0x1,
  MJ_MIO_SIGNAL_OUTPUT       = 0x2,
  MJ_MIO_SIGNAL_SAMPLERATE   = 0x4,
  MJ_MIO_MESSAGE_INPUT       = 0x10,
  MJ_MIO_MESSAGE_OUTPUT      = 0x20
};

/** DSP module state. */

enum MJ_ModuleDspState {
  MJ_MDS_STOPPED,
  MJ_MDS_PREPARED,
  MJ_MDS_PREPARED_FAST,
  MJ_MDS_PROCESS,
  MJ_MDS_STOPPED_FAST,
  MJ_MDS_SIZEOF
};

/** CPU loads. These tend to be self-explanatory. */
enum MJ_Load {
  MJ_LOAD_NONE,
  MJ_LOAD_INSIGNIFICANT,
  MJ_LOAD_LOW,
  MJ_LOAD_MEDIOCRE,
  MJ_LOAD_HIGH,
  MJ_LOAD_PANIC,
  MJ_LOAD_SIZEOF
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** Defines sample rate. 
    
    The sample rate is defined with "master" rate and fractional
    multiplier that determines the final sample rate.  With multirate
    systems is is necessary that we can get exact rational ratio
    between sampling rates.

    In general one set of modules will have the same master sampling
    rate.  The exact sampling rates of each signal can be adjusted by
    adjusting the nominator/denominator pair.  

    @author Tommi Ilmonen */

class MJ_SampleRateDef 
{
public:
  MJ_SampleRateDef();
  ~MJ_SampleRateDef() {}

  /// Set the parameters of this object.
  void set(int masterHz, int nominator, int deNominator)
    { 
      m_masterRate = masterHz; 
      m_nominator = nominator; 
      m_deNominator = deNominator; 
      calculate();
    }
  
  void setMasterRate(int hz);
  void setRatio(int nominator, int deNominator);

  /// Sampling rate as double
  inline const double &rate() const { return m_rate; }

  /// Sampling rate as an unsigned integer
  inline uint rateUInt() const 
  { return (m_masterRate * m_nominator) / m_deNominator; }

  /// Returns master sampling rate
  inline int masterRate()  const { return m_masterRate; }
  inline int nominator()   const { return m_nominator; }
  inline int deNominator() const { return m_deNominator; }

  bool validLength(uint);
  uint toThisLength(uint);
  uint fromThisLength(uint);

  bool write(QTextStream *s);
  bool read(QTextStream *s);

  QDomElement write(QDomDocument &);
  bool read(QDomElement &);

protected:

  void calculate();

  double m_rate;
  int m_masterRate, m_nominator, m_deNominator;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** Defines a signal source. This is not an actual sginal source.
    
    The signal is defined by a source module id and its output
    channel.

    @author Tommi Ilmonen
    @see MJ_SignalSource
*/
class MJ_SignalSourceDef
{
public:
  MJ_SignalSourceDef();
  virtual ~MJ_SignalSourceDef();
  
  /** Returns the source module. */
  inline const MJ_ModuleDsp *module() const { return m_sourceModule; }
  /** Sets the source module. */
  void setModule(MJ_ModuleDsp *m);

  /// Returns the source channel.
  inline int channel () const { return m_sourceChannel; }
  /// Sets the source channel
  inline void setChannel(int c) { m_sourceChannel = c; }
  
  
  /** Returns source id. */
  const char * sourceId() const { return m_sourceId; }
  /** Sets source id. */
  void setSourceId(const char *id);

  /** Sets the parameters that describe the source. */
  void set(const char * sourceId, int sourceChannel) 
  { setSourceId(sourceId); m_sourceChannel = sourceChannel; }

  /// Check that the channel really exists.
  bool validChannel();

  /** Connects the actual source. Before calling this method you
      should set the source module. */
  bool connectSource(MJ_SignalSource *);
  
  /** @name Stream IO. */
  //@{
  /// Write to a steam
  bool write(QTextStream *s);
  /// Read from a stream
  bool read(QTextStream *s);
  //@}

  QDomElement write(QDomDocument &);
  bool read(QDomElement &);

protected:

  /// Default is NULL.
  MJ_ModuleDsp *m_sourceModule;
  int           m_sourceChannel;
  /// Id of the source module
  char *        m_sourceId;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** A signal source. Contains pointers to source signal channels. An
    array of these objects contains input signal needed by a
    particular MJ_ModuleDsp.

    @author Tommi Ilmonen
    @see MJ_ModuleDsp MJ_SignalSourceDef MJ_SignalSink */
class MJ_SignalSource 
{
public:
  MJ_SignalSource();
  virtual ~MJ_SignalSource();

  inline int channel() const { return m_sourceChannel; }
  inline void setChannel(int c) { m_sourceChannel = c; }

  MJ_Vector2<float> &samples() { return m_samples; }
  const MJ_Vector2<float> &samplesConst() const { return m_samples; }

  void connectSink(MJ_SignalSink *xpSink);

  /// The signal sink this source is connected to.
  MJ_SignalSink *sink() { return m_sink; }

  /// The signal sink this source is connected to.
  const MJ_SignalSink *sinkConst() const { return m_sink; }
protected:
  ///
  MJ_Vector2<MJ_Sample> m_samples;
  ///
  MJ_SignalSink        *m_sink;
  ///
  int                   m_sourceChannel;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** A signal sink. The signal processing objects push their resulting
    data to an array of these objects.
    
    @author Tommi Ilmonen
    @see MJ_ModuleDsp MJ_SignalSource */
class MJ_SignalSink
{
public:
  MJ_SignalSink();
  virtual ~MJ_SignalSink();
  
  inline int channel() const { return m_sinkChannel; }
  inline void setChannel(int c) { m_sinkChannel = c; }

  /// Check that the channel really exists.
  bool validChannel();

  /// The sample vector.
  MJ_Vector<MJ_Sample> &samples() { return m_samples; }
  
  /// Current sample rate definition.
  MJ_SampleRateDef &sampleRate() { return m_sampleRate; }
  ///
  const MJ_SampleRateDef &sampleRateConst() const { return m_sampleRate; }
  
  /** Resizes the buffer to match the sampling rate. */
  bool setLength(uint masterRateSamples);

  /// Write to a steam
  virtual bool write(QTextStream *s);
  /// Read from a stream
  virtual bool read(QTextStream *s);

  virtual QDomElement write(QDomDocument &);
  virtual bool read(QDomElement &);

protected:

  MJ_Vector2<MJ_Sample> m_samples;
  int                  m_sinkChannel;
  MJ_SampleRateDef     m_sampleRate;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/// Keywords for MJ_ModuleDsp
enum MJ_ModuleKeyWord {
  MJ_MKW_ID,
  MJ_MKW_IODEFS,
  MJ_MKW_ICHANS,
  MJ_MKW_OCHANS,
  MJ_MKW_OMESSAGES,
  // Plain "SIZE" word.
  MJ_MKW_SIZE,
  MJ_MKW_END,
  MJ_MKW_PLAINEND,
  MJ_MKW_SIZEOF  
};

///
extern const char *MJ_ModuleKeyWords[MJ_MKW_SIZEOF];

/** Base class for DSP-capable modules. 

    To make new DSP object types you need to inherit this class.  

    <B> Audio IO:</B> To improve performance these objects operate on a
    buffer of samples at a time. Typically the buffer size is less
    than 100 samples. A DSP module can have multiple input and output
    signal (or audio) channels. Each channel input channel is
    contained in MJ_SignalSource object. Each output channel is
    contained in one MJ_SignalSink object. The input and output
    source/sink obejcts contain one channel each (no multichannel
    streams). For multichannel operation you need use multiple signal
    source/sink objects. 

    To write stuff to audio output you modify the values of the output
    samples. To get input you read values from input samples. 
    
    <B>Message IO:</B> Another way to pass information between modules
    (besides audio) is the message interface. Messages are
    encapsulated into message objects (class MJ_Message) that may
    contain any data. Messages may appear at random times e.g. there
    is no fixed-rate for them. A practical massage type is the MIDI
    message which can transfer MIDI messages from one module to
    another. The GUI components can also use the messages to give
    information to the DSP object. The messages are also used to store
    undo/redo information for the GUI.
    
    @author Tommi Ilmonen
    @see MJ_SignalSource MJ_SignalSink MJ_SignalSourceDef MJ_Message MJ_ModuleDspGui MJ_ModuleDspUi MJ_Satellite
*/
class MJ_ModuleDsp : public MJ_Base
{

public:

  MJ_ModuleDsp();
  
  virtual ~MJ_ModuleDsp();

  /** Sets the real-time driver that is responsible for this object.
   */
  void setRtDriver(MJ_RtDriver *d) { m_parent = (MJ_Base *) d; }
  
  /** Sets the parent module. A module can host child modules (mixer
      module for example). Children modules (and other objects as
      well) may need to access their parent module. this is the method
      that does it. */
  void setParentModule(MJ_ModuleDsp *m) { m_parent = m; }

  MJ_RtDriver * topRtDriver();

  /** Queries the parent module. */
  MJ_ModuleDsp *parentModule() 
  { return dynamic_cast<MJ_ModuleDsp *> (m_parent); }

  void setDbe(MJ_PluginDbeBase *dbe);

  /** The database entry that created this module. If this module was
      not created by a database entry then the return value will be
      null. */
  MJ_PluginDbeDsp *dbe() { return m_dbe; }

  virtual bool conflicts(MJ_ModuleDsp *);

  /** @name Stream IO. */
  //@{

  static MJ_ModuleDsp *readFromStream(QTextStream *s);

  bool writeIODefs(QTextStream *s);
  bool readIODefs(QTextStream *s, bool skipIODefs = false);

  bool writeHeader(QTextStream *s);
  bool readHeader(QTextStream *s);

  bool write(QTextStream *s);
  bool read(QTextStream *s, bool skipIODefs = false);

  virtual bool writeParameters(QTextStream *s);
  virtual bool readParameters(QTextStream *s);

  QDomElement writeIODefs(QDomDocument &);
  bool readIODefs(QDomElement &);

  QDomElement writeHeader(QDomDocument &);
  bool readHeader(QDomElement &);

  QDomElement write(QDomDocument &);
  bool read(QDomElement &);

  virtual QDomElement writeParameters(QDomDocument &);
  virtual bool readParameters(QDomElement &);

  void toCout();

  static void testIoDefRw();
  //@}

  /** @name ID access. */
  //@{
  /// Get identifier string.
  const char * id() const { return m_id; }
  void setId(const char * id);
  //@}

  /** @name DSP IO control. */
  //@{
  virtual bool dspAcceptIoChans(int * in, int * out);

  bool dspCompileInputs();

  int dspInChannels();

  bool dspSetInChannels(int); 

  bool dspConnectInput(const char * sourceId, 
		       int sourceChannel, int inputChannel);

  /// Returns an input signal
  MJ_SignalSource &dspInSignal(uint i) { return m_signalInput.get(i); }

  /// Returns input signal definitions.
  MJ_SignalSourceDef &dspInSignalDef(uint i) 
  { return * m_signalSourceDefs.at(i); }
  
  /// Returns an output signal
  MJ_SignalSink &dspOutSignal(uint i) { return m_signalOutput.get(i); }

  int dspOutChannels();

  bool dspSetOutChannels(int); 

  void dspClearOutputs();

  void dspDropOutputPtrs();
  
  virtual MJ_ModuleDsp *findModule(const char *id);
  
  /** Gets one output signal */
  MJ_SignalSink &dspGetOutSignal(uint i) { return m_signalOutput[i]; }
    
  virtual bool dspSetTimer(MJ_TimeStamp *t);

  /** Gets the time stamp object (master clock). */
  inline MJ_TimeStamp *dspGetTimer() { return m_currentTime; }
  
  inline uint dspFrameSize() const
  { return m_currentTime ? m_currentTime->frameSize() : 0; }

  inline ulong dspMasterRate() const
  { return m_currentTime ? m_currentTime->hz() : 44100; }

  bool dspIsRealTime() const;

  int dspSamplingRate() const;

  //@}

  /** @name DSP control. 
      
      The default implementations do nothing. */
  //@{

  virtual bool dspPrepare();
  virtual bool dspPrepareFast();
  virtual void dspProcessPeriod();
  virtual bool dspStopFast();
  virtual bool dspStop();
  //@}
  
  /** Tell module, it should reduce CPU load.  The higher the load,
      the more the module should try to reduce its load. */
  virtual bool dropLoad(MJ_Load load);

  /** @name Message-based control. */
  //@{
  
  bool msgProcessString(const char *, const QString &);
  bool msgProcessDouble(const char *, double);
  bool msgProcessInt(const char *, int);
  bool msgProcessEmpty(const char *);

  virtual bool msgProcess(const char *id, MJ_Message *);

  virtual void msgGetInputs(MJ_Parameters *);
  virtual MJ_Parameter *msgGetInput(const char *id);
  MJ_Parameter *msgGetInputIndex(int);
  virtual QCString msgGetInputName(int);

  virtual void msgGetOutputs(MJ_Parameters *);
  virtual MJ_Parameter *msgGetOutput(const char *);
  MJ_Parameter *msgGetOutputIndex(int);
  virtual QCString msgGetOutputName(int);

  void msgDisconnectTarget(MJ_ModuleDsp *);

  void msgDisconnectSource(MJ_ModuleDsp *);

  void msgDisconnectAll();

  void msgDisconnectInput();

  void msgDisconnectOutput();

  bool msgCompileOutputs();

  bool msgConnectInput(const char * sourceId, 
		       const char * sourceMsg,
		       const char * targetMsg);
  bool msgAddOutputRemap(MJ_MessageRemap *);
  //@}

  /** @name Satellite Access. */
  //@{
  /** Get a pointer to the satellite. The GUI thread does \b not need
      to lock the module when calling this method. */
  MJ_Satellite *satellite() { return m_satellite; }
  //@}

  virtual void getChildren(QList<MJ_ModuleDsp> &l, bool recursive);
  
  virtual MJ_ModuleDspGui *newGui
  (MJ_RtControl *rtparent,MJ_GuiSize s, QWidget *parent, const char *name = 0);

  virtual MJ_ModuleDspTui *newTui(MJ_RtControl *parent = 0,
				  const char *name = 0);

  /** Switch debugging on or off. */
  static void setDspDebug(bool b) { m_moduleDspDebug = b; }

protected:

  void msgPushOutput(const char * id, MJ_Message *msg);
  
  MJ_Allocator *mjDspAllocator();
  
  MJ_Base* mjParent(); 

  /** Parent. This pointer has two possible data types: MJ_ModuleDsp
      and MJ_RtDriver. 

      When using hiearchical systems it often necesary to
      access the parent modules.

      The MJ_RtDriver is needed if the module wants to allocate
      memory from the current MJ_Allocator object. 

      Another purpose for having MJ_RtDriver is the need for error
      reporting. Remember, that when running in threaded mode the DSP
      components cannot throw warning to screen directly. The messages
      are instead routed to the object responsible for the
      operation. Hopefully some GUI object will eventually see this
      message and report it to the user.  */
  MJ_Base                  *m_parent;

  /** The Data Base Entry that created this module. */
  MJ_PluginDbeDsp          *m_dbe;

  /** Satellite. In general, the module should seldom modify the
      satellite. The only place to do the modifications are in the
      constructor. A typical modification is to delete the (default)
      satellite and replace it with class-specific custom satellite
      (that is of different class). */
  MJ_Satellite             *m_satellite;
  
  /** List of input source module and associated channels. 

      MJ_List<MJ_SignalSourceDef>

      This information is compiled to the input buffer array.

      @see MJ_SignalSourceDef MJ_List
   */
  MJ_List<MJ_SignalSourceDef> m_signalSourceDefs;

  /** Array of input signals.
 
      MJ_Array<MJ_SignalSource>

      @see  MJ_SignalSource MJ_Array */
  MJ_Array<MJ_SignalSource> m_signalInput;

  /** Array of output signals. 
      
      MJ_Array<MJ_SignalSink>

      @see  MJ_SignalSink MJ_Array */
  MJ_Array<MJ_SignalSink>   m_signalOutput;

  ///
  MJ_MessageRemapList       m_outputMsgRemap;

  /** List of input message sources.

      QList<MJ_ModuleDsp>
      
  */
  QList<MJ_ModuleDsp>       m_inputMsgSources;

  /** Current time. Usually, many modules share the same time by
      linking to the same timer. The timer in turn is usually
      controlled by the top-level DSP module. */
  MJ_TimeStamp             *m_currentTime;

  /// Identifier
  char                    * m_id;
    
  /// Static debugging variable
  static bool m_moduleDspDebug;

private:
  /// Disabled
  MJ_ModuleDsp(const MJ_ModuleDsp &) {}
  MJ_ModuleDsp &operator = (const MJ_ModuleDsp &) { return *this; }
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/// DSP class that does nothing.
class MJ_ModuleDspNull : public MJ_ModuleDsp
{
public:
  virtual ~MJ_ModuleDspNull();
  virtual void dspProcessPeriod();
};

#endif
