/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_module_dsp_array.h>

#include <mj_debug.h>
#include <mj_error.h>
#include <mj_iostream.h>
#include <mj_strs.h>

#include <qtextstream.h> 

#include <stdio.h>

#include <typeinfo>

const char *MJ_ModuleFuncKeys[] = {
  "dspPrepare",
  "dspPrepareFast",
  "dspProcessPeriod",
  "dspStopFast",
  "dspStop"
};

MJ_ModuleDspArray::MJ_ModuleDspArray()
  : m_childTimer(0)
{}

MJ_ModuleDspArray::~MJ_ModuleDspArray()
{}

MJ_ModuleDsp *MJ_ModuleDspArray::findModule(const char * id)
{
  MJ_ModuleDsp **m = m_array.data();
  MJ_ModuleDsp **sentinel = m + m_array.size();

  while(m < sentinel) {
    if(!strcmp((*m)->id(), id)) return *m;
    m++;
  }
  return 0;
}

bool MJ_ModuleDspArray::dspSetTimer(MJ_TimeStamp *t)
{
  // trace2("MJ_ModuleDspArray::dspSetTimer # %d", (int) t->hz());

  MJ_ModuleDsp **m = m_array.data();
  MJ_ModuleDsp **sentinel = m + m_array.size();

  bool ok = true;

  m_childTimer = t;

  while(m < sentinel) {
    ok = ok && (*m)->dspSetTimer(m_childTimer);
    m++;
  }
  return ok;
}
  
/** Calls dspPrepare of all children. */

bool MJ_ModuleDspArray::dspPrepare()
{
  const char * fname = "MJ_ModuleDspArray::dspPrepare";

  m_childTimer->setStamp(m_currentTime->stamp());

  trace2("%s # Timer is %p (frame %lu)", 
	 fname, m_childTimer, m_childTimer->stamp());

  MJ_ModuleDsp **m = m_array.data();
  MJ_ModuleDsp **sentinel = m + m_array.size();

  MJ_ModuleDsp *md;
  
  // m_childTimer->setStamp(0);

  while(m < sentinel) {
    md = *m;
    md->dspSetTimer(m_childTimer);
    md->dspClearOutputs();
    m++;
  }

  m = m_array.data();

  /** Recompile all inputs. */
  
  while(m < sentinel) {
    md = *m;

    if(!md->dspCompileInputs() || 
       !md->msgCompileOutputs()) {
      
      QString str;
      QTextStream s(&str, IO_WriteOnly);
      
      s << fname << " # io compilation failed, bad module id = " 
	<< md->id() << " type = " 
	<< typeid(*m).name();
      
      error(MJ_ERR_INTERNAL, str);
      
      toCout();
      // return false;
    }
    
    m++;
  }

  return dspOperate(MJ_MF_PREPARE);
}

/** Calls dspPrepareFast of all children. */

bool MJ_ModuleDspArray::dspPrepareFast()
{
  return dspOperate(MJ_MF_PREPARE_FAST);
}

/** Calls dspProcessPeriod of all children. */
void MJ_ModuleDspArray::dspProcessPeriod()
{
  MJ_ModuleDsp **m = m_array.data();
  MJ_ModuleDsp **sentinel = m + m_array.size();

  while(m < sentinel) {
    (*m)->dspProcessPeriod();
    m++;
  }
  
  if(m_childTimer != m_currentTime)
    m_childTimer->incrementStamp(m_childTimer->frameSize());

  if(!parentModule()) {
    // trace("FOO");
    m_currentTime->incrementStamp(m_currentTime->frameSize());
  }
}

/** Calls dspStopFast of all children. */

bool MJ_ModuleDspArray::dspStopFast()
{
  return dspOperate(MJ_MF_STOP_FAST);
}

/** Calls dspStop of all children. */
bool MJ_ModuleDspArray::dspStop()
{
  return dspOperate(MJ_MF_STOP);
}

bool MJ_ModuleDspArray::msgProcess(const char *id, MJ_Message *msg)
{
  const char * slashAt = MJ_Strs::findNext(id, '/');

  if(!slashAt) return false;

  char buf[256];

  strncpy(buf, id, slashAt - id);

  MJ_ModuleDsp *m = findModule(buf);
  
  if(!m) return false;

  return m->msgProcess(slashAt + 1, msg);
}

void MJ_ModuleDspArray::getChildren(QList<MJ_ModuleDsp> &l, bool recursive)
{
  MJ_ModuleDsp **m = m_array.data();
  MJ_ModuleDsp **sentinel = m + m_array.size();

  while(m < sentinel) {
    l.append(*m);
    if(recursive) (*m)->getChildren(l, true);
    m++;
  }  
}

void MJ_ModuleDspArray::clearArray(bool alsoDelete)
{
  if(alsoDelete) {
    for(uint i=0; i < m_array.size(); i++)
      delete m_array.getFast(i);
  }

  m_array.resize(0);
}

/** Adds the given module to the module array. The array size is
    increased by one. */

void MJ_ModuleDspArray::addToArray(MJ_ModuleDsp *m, bool append)
{
  const char * fname = "MJ_ModuleDspArray::addToArray";

  MJ_cpf(m, fname);

  if(!m->id()) m->setId(allocateId());
  
  int s = m_array.size();
  m_array.resize(s + 1);

  // MJ_trace(fname, "HALF");

  m->setParentModule(this);

  // MJ_trace(fname, "PARENT SET");

  m->dspSetTimer(m_childTimer);

  // MJ_trace(fname, "TIMER SET");

  if(append) {
    m_array.get(s) = m;
  }
  else {
    m_array.shiftUp(0, 1);
    *m_array.data() = m;
  }

  // MJ_trace(fname, "EXIT");
}

bool MJ_ModuleDspArray::removeModule(MJ_ModuleDsp *m)
{
  for(uint i=0; i < m_array.size(); i++) {
    if(m_array.getFast(i) == m) {
      m_array.removeFast(i);
      return true;
    }
  }
  return false;
}

QCString MJ_ModuleDspArray::allocateId()
{
  QCString id;
  int toTry = 0;
  bool found = false;
  
  while(!found) {
    id.sprintf("mod_%d", toTry);
    
    if(!findModule(id)) 
      return id;
    toTry++;
  }

  return id;
}

/** Does one of the operations defined in argument "func".  The method
    dspProcessPeriod is not available how-ever. That method is
    implemented directly to get maximum performance.
*/

bool MJ_ModuleDspArray::dspOperate(MJ_ModuleFunc func, uint start, uint end)
{
  const char * fname = "MJ_ModuleDspArray::dspOperate";

  bool ok = true;

  bool debug = MJ_Debug::debugAll();

  end = end ? end : m_array.size();

  for(uint i = start; i < end; i++) {
    
    bool tmpOk = true;
    
    MJ_ModuleDsp *m = m_array.getFast(i);

    if(debug) {
      m->trace(fname, MJ_ModuleFuncKeys[func]);
    }

    switch(func) {
    case MJ_MF_PREPARE:
      tmpOk = m->dspPrepare();
      break;
    case MJ_MF_PREPARE_FAST:
      tmpOk = m->dspPrepareFast();
      break;
    case MJ_MF_STOP_FAST:
      tmpOk = m->dspStopFast();
      break;
    case MJ_MF_STOP:
      tmpOk = m->dspStop();
      break;
    default:
      error(MJ_ERR_ARGUMENT, fname);
      tmpOk = false;
      break;
    }

    if(!tmpOk) {
      char buf[128];

      sprintf(buf, "Module \"%s\" of type %s -- %s failed",
	      m->id(), typeid(*m).name(), MJ_ModuleFuncKeys[func]);
      error(MJ_ERR_INTERNAL, fname, buf);
      m->write(&MJ_cerr());
      MJ_cerr().device()->flush();
      
      for(uint j = MJ_MF_STOP_FAST; j <= MJ_MF_STOP; j++) {
	dspOperate((MJ_ModuleFunc) j, 0, i);
      }

      return false;
    }

    ok = ok && tmpOk;
    m++;
  }
  return ok;  
}

