/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_module_dsp_gui.h>

#include <mj_dsp_plugin.h>
#include <mj_module_dsp.h>
#include <mj_satellite.h>

#include <mj_error.h>

#include <typeinfo>

const char *MJ_GuiSizes[MJ_GUI_SIZEOF] = {
  "MJ_GUI_STAMP",
  "MJ_GUI_POSTCARD",
  "MJ_GUI_OWN_WINDOW"  
};

bool MJ_ModuleDspGui::m_dspGuiDebug = false;

/** Create a GUI object. The object is not initialized at this
    stage.

    <B>In future this method will be the sole initializer !</B> */

MJ_ModuleDspGui::MJ_ModuleDspGui(MJ_RtControl *rtparent,
				 MJ_ModuleDsp *module,
				 QWidget      *parent,
				 const char   *name)
  : QFrame(parent, name, Qt::WDestructiveClose),
    MJ_ModuleDspUi(rtparent, module)
{
  if(m_dspGuiDebug) trace("MJ_ModuleDspGui::MJ_ModuleDspGui", module);
  defaultConnections();
}
  
/** Deletes the GUI object and frees all resources. */
MJ_ModuleDspGui::~MJ_ModuleDspGui()
{
  if(m_dspGuiDebug) trace("MJ_ModuleDspGui::~MJ_ModuleDspGui");
  emit emitDestroyed(this);
}

/** Sets the data-base entry that created this module. If the object
    has no parent widget the icon and window names are automagically
    created unless they already exist. */
void MJ_ModuleDspGui::setDbe(MJ_PluginDbeDsp *dbe)
{
  m_dbe = dbe;

  if(!parent() && dbe) {

    // If window name is undefined, redifine:
    QString tmp(caption()), tmp2(dbe->objectName());
    if(tmp.isEmpty()) {
      setCaption(tmp2);
    }

    tmp = iconText();
    if(tmp.isEmpty()) {
      setIconText(tmp2);
    }
  }
}

/** Hide the GUI. If this module has no parent and if it is not the
    main widget it is deleted. In other cases QWidget::hide is
    called. */

void MJ_ModuleDspGui::hide()
{
  trace3(m_dspGuiDebug, "MJ_ModuleDspGui::hide");

  if((qApp->mainWidget() != this) && !parent()) 
    close();
  else QWidget::hide();

  trace3(m_dspGuiDebug, "MJ_ModuleDspGui::hide EXIT");
}

/** Tells the weight of this widget. The more weight, the more the
    system tries to allocate screen space for it (compared to other
    widgets). 

    The weight is passed directly to Qt layout managers. It should be
    in range 1-100.

    The default implementatiion returns 10.  */

int MJ_ModuleDspGui::weight()
{
  return 10;
}

/** Called as the module has been modified. By default this method is
    called as the object's satellite's \b modified method is called
    (MJ_Satellite::modified). */

void MJ_ModuleDspGui::moduleModified(const char *)
{}

/** Called as the module's IO setup has been changed. By default this
    method is called as the object's satellite's \b ioModified method
    is called (MJ_Satellite::ioModified). 

    @arg flags This variable holds information about the parts of IO
    setup that were changed (see MJ_ModuleIO). Yuo can test this
    variable against \b MJ_MIO_* macros to see if a particular IO
    feature was changed.

*/

void MJ_ModuleDspGui::moduleIoModified(uint flags)
{
  (void) flags;
}

/** Used to check when a module has been destroyed. This method should
    be never called (the GUI \b must be deleted before the
    object). This method is only used as a safeguard against very
    broken application behaviour. */

void MJ_ModuleDspGui::moduleDestroyed()
{
  /* The fact that this method has been called implies 
     a great problem somewhere... */
  if(m_module || !this)
    MJ_fatal2(MJ_ERR_INTERNAL, 
	      "MJ_ModuleDspGui::moduleDestroyed # \"%s\", %p %p", 
	      typeid(*this).name(), m_module, this);
}

/** Creates default signal/slot connectoins between the GUI object and
    modules' satellite.*/

void MJ_ModuleDspGui::defaultConnections()
{
  MJ_cpf(m_module, "MJ_ModuleDspGui::defaultConnections");
  
  MJ_Satellite *s = m_module->satellite();

  connect(s, SIGNAL(emitModified(const char *)), 
	  this, SLOT(moduleModified(const char *)));

  connect(s, SIGNAL(emitIoModified(uint)), 
	  this, SLOT(moduleIoModified(uint)));
  
  connect(s, SIGNAL(destroyed()), this, SLOT(moduleDestroyed()));
}

MJ_Base* MJ_ModuleDspGui::mjParent()
{
  QObject *p = parent();

  while (p) {
    MJ_Base *b = dynamic_cast<MJ_Base *> (p);
    if(b) return b;
    p = p->parent();
  }
  return 0;
}
