/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_module_dsp_gui_default.h>

#include <mj_cstr_remap.h> 
#include <mj_dial.h>
#include <mj_dsp_plugin.h>
#include <mj_gui_utils.h>
#include <mj_gui_numeric.h>
#include <mj_gui_int.h>
#include <mj_gui_string.h>
#include <mj_error.h>
#include <mj_header_label.h>
#include <mj_mapper.h>
#include <mj_math.h>
#include <mj_messages.h>
#include <mj_message_string.h>
#include <mj_module_dsp.h>
#include <mj_module_menubar.h>
#include <mj_parameters.h>
#include <mj_satellite.h>
#include <mj_slider.h>
#include <mj_tr.h>

#include <qcheckbox.h> 
#include <qcombobox.h> 
#include <qlayout.h> 
#include <qlineedit.h> 
#include <qpopupmenu.h> 
#include <qscrollview.h> 
#include <qtooltip.h> 

#include <typeinfo>

bool MJ_ModuleDspGuiStamp::m_debug = false;

MJ_ModuleDspGuiStamp::MJ_ModuleDspGuiStamp(MJ_RtControl *rtparent,
					   MJ_ModuleDsp *module,
					   QWidget      *parent,
					   const char   *name)
  : MJ_ModuleDspGui(rtparent, module, parent, name),
    m_label(0)
{
  if(m_debug) trace("MJ_ModuleDspGuiStamp::MJ_ModuleDspGuiStamp");
  
  draw();
}

MJ_ModuleDspGuiStamp::~MJ_ModuleDspGuiStamp()
{
  if(m_debug) trace("MJ_ModuleDspGuiStamp::~MJ_ModuleDspGuiStamp");
}

MJ_ModuleDspGui *MJ_ModuleDspGuiStamp::createNew(MJ_RtControl *rtparent,
						 MJ_ModuleDsp *module,
						 QWidget *parent, 
						 const char *name)
{
  return new MJ_ModuleDspGuiStamp(rtparent, module, parent, name);
}



void MJ_ModuleDspGuiStamp::draw()
{
  if(m_debug) trace("MJ_ModuleDspGuiStamp::draw");

  QGridLayout *layout = new QGridLayout(this, 3,3, 5, 0);
  MJ_equalStretch(layout);
  
  m_label = new MJ_HeaderLabel(this, "MJ stamp-gui label");

  if(parent()) {
    m_label->enableClicking(true);  
    connect(m_label, SIGNAL(pressed()), this, SLOT(showModule()));
  }
  if(m_module->dbe()) {
    m_label->setText(MJ_lastField(m_module->dbe()->menuPosition(), 
				  MJ_pluginSeparator));
    const QString & about = m_module->dbe()->parameter(MJ_PP_ABOUT);
    if(!about.isEmpty())
      QToolTip::add(m_label, about);
  }
  else
    m_label->setText(typeid(*m_module).name());

  m_label->setAlignment(AlignLeft | AlignTop);
  MJ_limMinSize(m_label);

  layout->addWidget(m_label, 0,0);

  realTimeChange(isRealTime());
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

bool MJ_ModuleDspGuiPostCard::m_debug = false;

MJ_ModuleDspGuiPostCard::MJ_ModuleDspGuiPostCard(MJ_RtControl *rtparent,
						 MJ_ModuleDsp *module,
						 bool          drawThis,
						 uint maxDisplayParameters,
						 MJ_GuiSize    guiSize,
						 QWidget      *parent,
						 const char   *name)
  : MJ_ModuleDspGui(rtparent, module, parent, name),
    m_main(0),
    m_layout(0),
    m_scrollView(0),
    m_scrollChild(0),
    m_guiSize(guiSize),
    m_maxDisplayParameters(maxDisplayParameters),
    m_rowsUsed(0)
{
  if(m_debug) 
    trace("MJ_ModuleDspGuiPostCard::MJ_ModuleDspGuiPostCard", 
	  maxDisplayParameters);

  m_parameters.setAutoDelete(true);
  // m_needToDelete.setAutoDelete(true);
  m_children.setAutoDelete(true);

  setModuleInternal();

  if(drawThis)
    draw();
}

MJ_ModuleDspGuiPostCard::~MJ_ModuleDspGuiPostCard()
{
  if(m_debug) trace("MJ_ModuleDspGuiPostCard::~MJ_ModuleDspGuiPostCard");
}


void MJ_ModuleDspGuiPostCard::moduleModified(const char *id)
{
  if(m_debug) trace("MJ_ModuleDspGuiPostCard::moduleModified", id);
  if(!id) update();
  if(m_editParameter == id) return;
  update2(id);
}

void MJ_ModuleDspGuiPostCard::setValue(const char * id,double d)
{
  if(m_debug) 
    trace("MJ_ModuleDspGuiPostCard::setValue(const char *,double)", id);

  bool ok = true;

  if(id == m_editParameter) return;
  
  int i = m_parameters.findParameter(id);

  MJ_Parameter *p = m_parameters.at(i);
  
  MJ_ParameterDouble *pd = dynamic_cast<MJ_ParameterDouble *> (p);

  if(pd) {
    pd->setValue(d);
    d = pd->value();
  }
  else return;

  if(!ok) return;

  MJ_MessageCarrier msg;
  msg.setId(id);
  MJ_MessageDouble *md = new MJ_MessageDouble(d);
  msg.setMessage(md);
  
  deliverMessage(&msg);
  
  m_newParameters.setMessage(i, &msg);

  /* Now the module is modified. If there are any other GUIs open they
     will caontain wrong (old) values. They can be informed via the
     use of a "satellite" that the state of the module has changed. 
     
     The satellite distributes the change information to other GUIs
     via Qt's signal /slot mechanism.
     
     We have a class member that is used to make sure we do not need
     to change this GUI as the change occurs. */

  m_editParameter = id;
  m_module->satellite()->modified(id);
  m_editParameter.truncate(0);
  
  if(p->affectsOthers())
    update();
}

void MJ_ModuleDspGuiPostCard::getValue(const char * id, double *d)
{
  if(m_debug) 
    trace("MJ_ModuleDspGuiPostCard::getValue(const char *,double)", id); 

  lock();
  MJ_Parameter * p = m_module->msgGetInput(id);
  unlock();

  if(!p) return;

  MJ_ParameterDouble *pd = dynamic_cast<MJ_ParameterDouble *> (p);

  if(pd) {
    *d = pd->value();
  }
  
  delete p;
}


void MJ_ModuleDspGuiPostCard::setValue(const char *id,int d)
{
  if(m_debug) trace("MJ_ModuleDspGuiPostCard::setValue(const char *,int)", id);

  bool ok = true;

  if(id == m_editParameter) return;
  
  int i = m_parameters.findParameter(id);

  MJ_Parameter *p = m_parameters.at(i);
  
  MJ_ParameterInt *pd = dynamic_cast<MJ_ParameterInt *> (p);

  if(pd) {
    pd->setValue(d);
    d = pd->value();
  }
  else return;

  if(!ok) return;

  MJ_MessageCarrier msg;
  msg.setId(p->id());
  MJ_MessageInt *md = new MJ_MessageInt(d);
  msg.setMessage(md);
  
  deliverMessage(&msg);
  
  m_newParameters.setMessage(i, &msg);

  /* Now the module is modified. If there are any other GUIs open they
     will contain wrong (old) values. They can be informed via the
     use of a "satellite" that the state of the module has changed. 
     
     The satellite distributes the change information to other GUIs
     via Qt's signal/slot mechanism.
     
     We have a class member that is used to make sure we do not need
     to change this GUI as the change occurs. */

  m_editParameter = id;
  m_module->satellite()->modified(id);
  m_editParameter.truncate(0);

  if(p->affectsOthers())
    update();
}
 
void MJ_ModuleDspGuiPostCard::getValue(const char *id, int *d)
{
  if(m_debug) 
    trace("MJ_ModuleDspGuiPostCard::getValue(const char *,int)", id);  

  lock();
  MJ_Parameter * p = m_module->msgGetInput(id);
  unlock();
  
  if(!p) return;

  MJ_ParameterInt *pd = dynamic_cast<MJ_ParameterInt *> (p);

  if(pd) {
    *d = pd->value();
  }
  
  delete p;
}

void MJ_ModuleDspGuiPostCard::setValue(const char *id, bool b)
{
  const char *fname = "MJ_ModuleDspGuiPostCard::setValue(const char *,bool)";

  if(m_debug) trace2("%s # \"%s\" %d", fname, id, (int) b);
  
  if(id == m_editParameter) return;
  
  int i = m_parameters.findParameter(id);

  MJ_Parameter *p = m_parameters.at(i);
  
  MJ_ParameterBool *pb = dynamic_cast<MJ_ParameterBool *> (p);

  if(!pb) {
    error(MJ_ERR_ARGUMENT, "%s # NULL POInter", fname);
    return;
  }

  MJ_MessageCarrier msg;
  msg.setId(p->id());
  
  MJ_MessageBool *mb = new MJ_MessageBool(b);
  msg.setMessage(mb);
  
  deliverMessage(&msg);
  
  m_newParameters.setMessage(i, &msg);

  m_editParameter = id;
  m_module->satellite()->modified(id);
  m_editParameter.truncate(0);

  doneValue(id);
  
  if(p->affectsOthers())
    update();
}

void MJ_ModuleDspGuiPostCard::doneValue(const char * id)
{
  static const char *fname = "MJ_ModuleDspGuiPostCard::doneValue";

  if(m_debug) trace(fname, id);

  int i = m_parameters.findParameter(id);
  
  if((uint) i >= m_newParameters.size()) {
    error(MJ_ERR_ARGUMENT, fname);
    return;
  }

  MJ_MessageCarrier *oldMsg = m_oldParameters[i];
  MJ_MessageCarrier *newMsg = m_newParameters[i];

  QString name = m_parameters.at(i)->name();

  MJ_Unit unit = m_parameters.at(i)->unit();

  if(oldMsg && newMsg)
    addUndoInfo(newMsg, oldMsg, name, unit);

  if(oldMsg) delete oldMsg;
  
  m_oldParameters[i] = newMsg;
  m_newParameters[i] = 0;
}

void MJ_ModuleDspGuiPostCard::setValueString(const char *id, const QString &s)
{
  const char * fname = "MJ_ModuleDspGuiPostCard::setValueString";

  if(m_debug) {
    trace(fname, id);
  }
  
  int index = m_parameters.findParameter(id);

  if(index < 0) {
    error(MJ_ERR_ARGUMENT, fname, id);
    return;
  }
  // msgDeliverString(id, s);

  MJ_MessageCarrier msg;
  msg.setId(id);
  MJ_MessageString *ms = new MJ_MessageString(s);
  msg.setMessage(ms);
  
  deliverMessage(&msg);
  
  m_newParameters.setMessage(index, &msg);

  m_editParameter = id;
  m_module->satellite()->modified(id);
  m_editParameter.truncate(0);

  doneValue(id);
}

void MJ_ModuleDspGuiPostCard::realTimeChange(bool nowRt)
{
  emit enterNonRealTime(!nowRt);
}

void MJ_ModuleDspGuiPostCard::draw()
{
  if(m_debug) trace("MJ_ModuleDspGuiPostCard::draw");

  int params = MJ_min(m_parameters.count(), m_maxDisplayParameters);

  QVBoxLayout *mainLayout = new QVBoxLayout(this, 6, 3, "mainlayout");
  
  MJ_HeaderLabel *mainLabel= new MJ_HeaderLabel(this, "MJ card-gui label");
  // mainLabel->setAlignment(AlignCenter);
  

  if(m_guiSize != MJ_GUI_OWN_WINDOW) {
    mainLabel->enableClicking(true);  
    connect(mainLabel, SIGNAL(pressed()), this, SLOT(showModule()));
  }
  else {
    mainLayout->setMenuBar(new MJ_ModuleMenubar(this, "mj_menubar"));
  }

  // Guess a label text for the module.
  if(m_module->dbe()) {
    mainLabel->setText(MJ_lastField(m_module->dbe()->menuPosition(), 
				    MJ_pluginSeparator));
    const QString & about = m_module->dbe()->parameter(MJ_PP_ABOUT);
    if(!about.isEmpty())
      QToolTip::add(mainLabel, about);
  }
  else
    mainLabel->setText(typeid(*m_module).name());

  mainLayout->addWidget(mainLabel, 0);
  
  if((int) m_guiSize >= (int) MJ_GUI_POSTCARD) {
    MJ_limMinSize(mainLabel);
    if(params > 15) {
      m_scrollView = new QScrollView(this, "postcard scroll");
      // m_scrollView->enableClipper(true);
      m_scrollView->resizeContents(700, 3000);
      mainLayout->addWidget(m_scrollView, 90);
      m_scrollChild = new QWidget(m_scrollView->viewport(), "m_scrollChild");
      m_scrollView->addChild(m_scrollChild);
      m_scrollChild->resize(700,3000);
      resize(400, 400);
    }
  }
  
  int border = 5;
  int space = 4;

  if(m_scrollChild)
    m_layout= new QGridLayout(m_scrollChild, params + 1, 3, border, space);
  else {
    m_layout = new QGridLayout(params + 1, 3, space);
    mainLayout->addLayout(m_layout);
  }

  MJ_equalStretch(m_layout);
  m_layout->setRowStretch(params, 10);
 
  // Add each parameter one at a time.
  for(int i=0; (i < params) && ((uint) i < m_maxDisplayParameters); i++) {
    addParam(i);
  }
  realTimeChange(isRealTime());
}

/** This method updates all possible GUIs in the screen. */

void MJ_ModuleDspGuiPostCard::update()
{
  const char * fname = "MJ_ModuleDspGuiPostCard::update";

  if(m_debug) trace(fname);
  
  m_parameters.clear();
  lock();
  m_module->msgGetInputs(&m_parameters);
  unlock();

  for(uint i=0; i < m_parameters.count() && i < m_children.count(); i++) {

    // The relevant child object
    QObject *o = m_children.at(i);
    MJ_Parameter *p = m_parameters.at(i);

    trace3(m_debug, "%s # Processing parameter %d \"%s\" obj = %p", 
	   fname, i, p->id(), o);

    // Double:
    MJ_ParameterDouble *pd = dynamic_cast<MJ_ParameterDouble *> (p);
    // Int:
    MJ_ParameterInt *pi = dynamic_cast<MJ_ParameterInt *> (p);
    
    if(pd) {
      MJ_GuiDouble *gd = dynamic_cast<MJ_GuiDouble *> (o);
      if(gd) {
	trace3(m_debug, "%s # Adjusting double GUI \"%s\"", fname, pd->id());
	gd->setRange(pd->min(), pd->max());
	gd->setValue(pd->value());
      }
    }
    else if(pi) {

      MJ_GuiInt *gi = dynamic_cast<MJ_GuiInt *> (o);
      
      // trace2("%s # gi = %p", fname, gi);

      if(gi) {
	trace3(m_debug, "%s # Adjusting int GUI \"%s\" to %d", 
	       fname, pi->id(), pi->value());
	gi->setRange(pi->min(), pi->max());
	gi->setValue(pi->value());
      }
      // trace(fname, "INT DONE");
    }
  }
  if(m_debug) trace(fname, "EXIT");
}

/** This method takes care of doing the real parameter update. This
    method is called as some parameter has been changed in some other
    user interface, but this UI instance is not yet aware of the
    change. */

void MJ_ModuleDspGuiPostCard::update2(const char *id)
{
  const char * fname = "MJ_ModuleDspGuiPostCard::update2";

  if(m_debug) trace(fname, id);
  
  lock();
  MJ_Parameter * p = m_module->msgGetInput(id);
  unlock();
  
  if(!p) return;

  if(p->affectsOthers()) {
    update();
    delete p;
    return;
  }
  
  int i = m_parameters.findParameter(id);

  if(i < 0 || (uint) i >= m_children.count()) {
    delete p;
    return;
  }
  // MJ_error(MJ_ERR_INTERNAL, fname);

  // The relevant child object
  QObject *o = m_children.at(i);

  MJ_MessageType mt = p->type();

  if(mt == MJ_MSG_DOUBLE) {

    MJ_ParameterDouble *pd = (MJ_ParameterDouble *) (p);
    
    if(pd) {
      double d = pd->value();
      MJ_GuiDouble *gd = dynamic_cast<MJ_GuiDouble *> (o);
      if(gd) {
	gd->setValue(d);
      }
    }
  }
  else if(mt == MJ_MSG_INT) {
    MJ_ParameterInt *pd = (MJ_ParameterInt *) (p);
    
    if(pd) {
      int d = pd->value();
      MJ_GuiInt *gd = dynamic_cast<MJ_GuiInt *> (o);
      if(gd) {
	gd->setValue(d);
      }
    }
  }
  else if(mt == MJ_MSG_STRING) {
    
    MJ_ParameterString *ps = (MJ_ParameterString *) (p);
    
    MJ_GuiString *gs = dynamic_cast<MJ_GuiString *> (o);
    if(gs) {
      gs->lineEdit()->setText(ps->string());
    }
  }
  else if (mt == MJ_MSG_BOOL) {
    
    MJ_ParameterBool *pb = (MJ_ParameterBool *) (p);
    
    QCheckBox *cb = dynamic_cast<QCheckBox *> (o);
    if(cb) {
      cb->setChecked(pb->value());
    }
  }

  delete p;
  if(m_debug) trace(fname, "EXIT");
}

void MJ_ModuleDspGuiPostCard::addParam(int ind)
{
  const char *fname = "MJ_ModuleDspGuiPostCard::addParam";

  if(m_debug) trace(fname, ind);

  MJ_Parameter *p = m_parameters.at(ind);

  MJ_MessageType mt = p->type();

  if(mt == MJ_MSG_DOUBLE) {
    // QComboBox with doubles.
    MJ_ParameterDouble *pd = dynamic_cast<MJ_ParameterDouble *> (p);

    if(!pd) MJ_fatal(MJ_ERR_INTERNAL, fname);

    addParamDouble(m_rowsUsed++, pd);
  }
  else if(mt == MJ_MSG_INT) {
    MJ_ParameterInt *pd = dynamic_cast<MJ_ParameterInt *> (p);

    MJ_cpf(pd, fname);

    /* if(pd->allowedValues().count())
      addParamIntCombo(ind, pd);
      else  */
    if(pd->allowedValues().count())
      addParamIntCombo(m_rowsUsed++, pd);
    else 
      addParamIntSlider(m_rowsUsed++, pd);
  }
  else if(mt == MJ_MSG_STRING) {
    MJ_ParameterString *ps = dynamic_cast<MJ_ParameterString *> (p);

    MJ_cpf(ps, fname);
    
    if(ps->allowedStrs())
      addParamStringCombo(m_rowsUsed++, ps);
    else
      addParamString(m_rowsUsed++, ps);
  }
  else if(mt == MJ_MSG_BOOL) {
    MJ_ParameterBool *pb = dynamic_cast<MJ_ParameterBool *> (p);
    
    MJ_cpf(pb, fname);

    addParamBool(m_rowsUsed++, pb);
  }
  else {
    return;
  }

  if(!parent()) show();
}

void MJ_ModuleDspGuiPostCard::addParamDouble
(int ind, MJ_ParameterDouble *pd)
{
  uint i;
  
  MJ_GuiDouble *gd = 0;
  MJ_GuiDoubleDial *gdd = 0;
  MJ_GuiDoubleSlider *gds = 0;
  MJ_GuiDoubleCombo *gdc = 0;

  if(pd->allowedValues().count()) {

    gdc = new MJ_GuiDoubleCombo;
  
    for(i=0; i < pd->allowedValues().count(); i++) {
      gdc->addValue(*pd->allowedValues().at(i));
    }
    gd = gdc;

    connect(gd, SIGNAL(emitGetValue2(const char *,double *)),
	    this, SLOT(getValue(const char *,double *)));
  }
  else if(m_guiSize == MJ_GUI_STAMP) {
    gd = gdd = new MJ_GuiDoubleDial;
    m_layout->setColStretch(2, 50);
  }
  else {
    gds = new MJ_GuiDoubleSlider;
    m_layout->setColStretch(1, 30);
    gd = gds;
  }

  gd->createObjs(childParent());

  if(gds && pd->mapper())
      gds->slider()->setMapper(pd->mapper()->duplicate());

  if(gdd) {
    QFont f = font();
    f.setPointSizeFloat(f.pointSize() * 0.6f);
    gdd->display()->setFont(f);
  }

  // if(m_scrollView) gdc->addChildren(m_scrollView);

  gd->setId(pd->id());
  gd->setRange(pd->min(), pd->max());
  gd->setValue(pd->value());
  
  gd->label()->setText(pd->name((int) m_guiSize));
  
  connect(gd, SIGNAL(emitValue2(const char *,double)),
	  this, SLOT(setValue(const char *,double)));
  connect(gd, SIGNAL(emitDone2(const char *)),
	  this, SLOT(doneValue(const char *)));
  
  if(!pd->realTimeAdjustable())
    connect(this, SIGNAL(enterNonRealTime(bool)), 
	    gdc, SLOT(setEnabled(bool)));

  if(gdd) {
    m_layout->addWidget(gdd->label(), ind, 0, AlignLeft);
    m_layout->addMultiCellWidget(gdd->dial(), 
				 ind, ind + 1,
				 1,2, AlignRight);
    m_layout->addWidget(gdd->display(), ind+1, 0, AlignLeft);
    m_rowsUsed ++;
  }
  else
    gd->toGridHorizontal(m_layout, ind, 0);

  if((int) m_guiSize >= (int) MJ_GUI_POSTCARD)
    gd->limMinSizes();

  m_children.append(gd);
  gd->show();

}

void MJ_ModuleDspGuiPostCard::addParamIntCombo
(int ind, MJ_ParameterInt *pd)
{
  uint i;
  
  MJ_GuiIntCombo *gdc = new MJ_GuiIntCombo;
  
  for(i=0; i < pd->allowedValues().count(); i++) {
    gdc->addValue(*pd->allowedValues().at(i));
  }
  
  gdc->createObjs(childParent());

  // if(m_scrollView) gdc->addChildren(m_scrollView);

  gdc->setId(pd->id());
  gdc->setValue(pd->value());
  
  gdc->label()->setText(pd->name((int) m_guiSize));
  
  if((int) m_guiSize >= (int) MJ_GUI_POSTCARD)
    gdc->limMinSizes();

  connect(gdc, SIGNAL(emitValue2(const char *,int)),
	  this, SLOT(setValue(const char *,int)));
  /* connect(gdc, SIGNAL(emitGetValue2(const char *,int *)),
     this, SLOT(getValue(const char *,int *))); */
  connect(gdc, SIGNAL(emitDone2(const char *)),
	  this, SLOT(doneValue(const char *)));
  
  if(!pd->realTimeAdjustable())
    connect(this, SIGNAL(enterNonRealTime(bool)), 
	    gdc, SLOT(setEnabled(bool)));

  gdc->toGridHorizontal(m_layout, ind, 0);
  m_children.append(gdc);
  gdc->show();

}

void MJ_ModuleDspGuiPostCard::addParamIntSlider(int ind, MJ_ParameterInt *pi)
{
  if(m_debug) 
    trace("MJ_ModuleDspGuiPostCard::addParamIntSlider", pi->id());  

  MJ_GuiIntSlider *gis = new MJ_GuiIntSlider;

  bool withEdit;
  
  if(m_guiSize == MJ_GUI_STAMP) withEdit = false;
  else withEdit = true;

  gis->createObjs(childParent(), withEdit);

  // if(m_scrollView) gis->addChildren(m_scrollView);

  gis->setId(pi->id());

  gis->setRange(pi->min(), pi->max());
  gis->setValue(pi->value());


  gis->label()->setText(pi->name((int) m_guiSize));

  if((int) m_guiSize >= (int) MJ_GUI_POSTCARD)
    gis->limMinSizes();

  connect(gis, SIGNAL(emitValue2(const char *,int)),
	  this, SLOT(setValue(const char *,int)));
  connect(gis, SIGNAL(emitGetValue2(const char *,int *)),
	  this, SLOT(getValue(const char *,int *)));
  connect(gis, SIGNAL(emitDone2(const char *)),
	  this, SLOT(doneValue(const char *)));

  if(!pi->realTimeAdjustable())
    connect(this, SIGNAL(enterNonRealTime(bool)), 
	    gis, SLOT(setEnabled(bool)));

  gis->toGridHorizontal(m_layout, ind, 0);
  m_layout->setColStretch(1, 30);
  m_children.append(gis);
  gis->show();
}

void MJ_ModuleDspGuiPostCard::addParamString(int ind, MJ_ParameterString *ps)
{
  if(m_debug) trace("MJ_ModuleDspGuiPostCard::addParamString");  
  
  MJ_GuiString *gs = new MJ_GuiString;

  gs->createObjs(childParent(), ps->name((int) m_guiSize), ps->string());
  
  // if(m_scrollView) gs->addChildren(m_scrollView);

  gs->setId(ps->id());
  connect(gs, SIGNAL(textEmit(const char *, const QString &)), 
	  this, SLOT(setValueString(const char *, const QString &)));

  MJ_ParameterFilename *pn = dynamic_cast<MJ_ParameterFilename *> (ps);

  int ew; // Entry width

  if(pn && m_guiSize !=  MJ_GUI_STAMP) {
    gs->createBrowseButton(childParent(), MJ_Tr::tr("Browse"), 
			   pn->filter(), (QFileDialog::Mode) pn->mode());
    ew = 1;
  }
  else
    ew = 2;

  gs->toGridHorizontal(m_layout, ind, 0, ew);
  m_layout->setColStretch(1, 30);
  gs->show();
  m_children.append(gs);
}

void MJ_ModuleDspGuiPostCard::addParamStringCombo(int ind, 
						  MJ_ParameterString *ps)
{
  if(m_debug) trace("MJ_ModuleDspGuiPostCard::addParamStringCombo");  
  
  MJ_GuiStringSelector *gs = new MJ_GuiStringSelector;
  
  gs->createObjs(childParent(), ps->name((int) m_guiSize));

  gs->box()->insertStringList(*ps->allowedStrs());

  int index = ps->allowedStrs()->findIndex(ps->string());

  if(index >= 0)
    gs->box()->setCurrentItem(index);

  // if(m_scrollView) gs->addChildren(m_scrollView);

  gs->setId(ps->id());
  connect(gs, SIGNAL(textEmit(const char *, const QString &)), 
	  this, SLOT(setValueString(const char *, const QString &)));
  
  gs->toGridHorizontal(m_layout, ind, 0);
  gs->show();
  m_children.append(gs);  
}

void MJ_ModuleDspGuiPostCard::addParamBool(int ind, MJ_ParameterBool *pb)
{
  const char * fname = "MJ_ModuleDspGuiPostCard::addParamBool";

  if(m_debug) trace2("%s # %p %d", fname, pb, (int) pb->value());

  QCheckBox *cb = new QCheckBox(pb->name(), childParent(), pb->id());
  
  cb->setChecked(pb->value());

  MJ_CStrRemap *remap = new MJ_CStrRemap(pb->id());
  // m_needToDelete.append(remap);

  connect(cb, SIGNAL(toggled(bool)), remap, SLOT(callBool(bool)));
  connect(remap, SIGNAL(emitCallBool(const char *,bool)),
	  this, SLOT(setValue(const char *,bool)));

  m_layout->addMultiCellWidget(cb, ind, ind, 0, 2, AlignLeft);
  
  cb->show();
  m_children.append(cb); 
}

void MJ_ModuleDspGuiPostCard::setModuleInternal()
{
  if(m_debug) trace("MJ_ModuleDspGuiPostCard::setModuleInternal");  

  m_parameters.clear();

  lock();
  m_module->msgGetInputs(&m_parameters);
  unlock();

  m_oldParameters.copyValues(&m_parameters);
  m_newParameters.copyValues(&m_parameters);
}

QWidget *MJ_ModuleDspGuiPostCard::childParent()
{
  if(m_scrollChild) return m_scrollChild;
  return this;
}
