/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_module_dsp_tui_default.h>

#include <mj_dsp_plugin.h>
#include <mj_error.h>
#include <mj_global.h>
#include <mj_iostream.h>
#include <mj_module_dsp.h>
#include <mj_tr.h>

#include <stdio.h>

#include <typeinfo>

// These strings are as short as possible.
const char * MJ_tuiKeyWords[MJ_TUC_SIZEOF] = {
  "load",
  "panic",
  "param",
  "up",
  "save"
};

bool MJ_ModuleDspTuiDefault::m_tuiDebug = false;

MJ_ModuleDspTuiDefault::MJ_ModuleDspTuiDefault(MJ_RtControl *parent,
					       MJ_ModuleDsp *module,
					       const char *name)
  : MJ_ModuleDspTui(parent, module, name)
{
  m_parameters.setAutoDelete(true);

  // Loop as long as the user wishes
  while(oneEvent())
    ;
}

MJ_ModuleDspTuiDefault::~MJ_ModuleDspTuiDefault()
{}
  
bool MJ_ModuleDspTuiDefault::oneEvent()
{
  if(m_tuiDebug) trace("MJ_ModuleDspTuiDefault::oneEvent");

  describeModule();
  return handleMessage();
}

void MJ_ModuleDspTuiDefault::describeModule()
{
  if(m_tuiDebug) trace("MJ_ModuleDspTuiDefault::describeModule");

  uint i;

  m_parameters.clear();

  lock();
  m_module->msgGetInputs(&m_parameters);
  unlock();

  uint params = m_parameters.count();
  
  QString objName;

  MJ_PluginDbeDsp *dbe = m_module->dbe();
  
  if(dbe) objName = dbe->MJ_PluginDbeBase::objectName();
  else objName = typeid(*m_module).name();
  
  // Write a header:
  MJ_cout() << "\n" << MJ_Tr::tr("Module") <<  " " << objName << "\n";
  
  // Write each parameter:
  for(i=0; i < params; i++) {
    MJ_cout() << m_parameters.at(i)->writeClean().join(" - ");
    MJ_cout() << "\n";
  }

  MJ_cout() << "keywords: [  ";

  for(i=0; i < MJ_TUC_SIZEOF; i++) {
    MJ_cout() << MJ_tuiKeyWords[i] << "  ";
  }

  MJ_cout() << "]\n";
  
  // Write the "command prompt".
  MJ_cout() << MJ_Tr::tr("Command") << ">";
  MJ_cout().device()->flush();
}

bool MJ_ModuleDspTuiDefault::handleMessage()
{
  QCString cmd;

  MJ_cin() >> cmd;

  MJ_TuiCommand c = (MJ_TuiCommand) 
    MJ_findStr(cmd, MJ_tuiKeyWords, MJ_TUC_SIZEOF);

  switch(c) {
  case MJ_TUC_LOAD: 
    load();
    break;

  case MJ_TUC_PANIC: 
    MJ_fatal(MJ_ERR_UNKNOWN, "MJ_ModuleDspTuiDefault::handleMessage");

  case MJ_TUC_PARAM:
    getParam();
    break;

  case MJ_TUC_UP: 
    return false;
    
  case MJ_TUC_SAVE:
    save();
    break;
    
  default:
    inputWarning(cmd);
    break;
  }

  return true;
}

void MJ_ModuleDspTuiDefault::getParam()
{
  QCString id, cmd;

  bool ok = true;
  double tmpd;
  
  MJ_cin() >> id;

  if(!ok) {
    inputWarning(id);
    return;
  }
  
  MJ_cin() >> cmd;

  MJ_MessageType mt = (MJ_MessageType) 
    MJ_findStr(cmd.data(), MJ_messageTypesShort, MJ_MSG_SIZEOF);

  switch(mt) {
  case MJ_MSG_EMPTY:
    msgDeliverEmpty(id);
    break;
    
  case MJ_MSG_DOUBLE:
    // Forget safety checks...
    MJ_cin() >> tmpd;
    msgDeliverDouble(id, tmpd);
    break;

  case MJ_MSG_STRING:
    MJ_cin() >> cmd;
    msgDeliverString(id, cmd);
    break;

  default:
    inputWarning(cmd);
    break;
  }
}

void MJ_ModuleDspTuiDefault::inputWarning(QCString &str)
{
  MJ_cout() << MJ_Tr::tr("Unexpected string") << " \"" << str << "\"\n";
  MJ_cout().device()->flush();
}

void MJ_ModuleDspTuiDefault::save()
{
  QString fileName;
  
  MJ_cin() >> fileName;

  bool ok = saveModule(fileName);

  QString tmp = " \"" + fileName + "\"\n";

  if(!ok) {
    MJ_cout() << MJ_Tr::tr("Failed to save to file") 
	    << tmp;
  }
  else MJ_cout() << MJ_Tr::tr("Saved module to file") 
	       << tmp;

  MJ_cout().device()->flush();
}

void MJ_ModuleDspTuiDefault::load()
{
  QString fileName;
  
  MJ_cin() >> fileName;

  bool ok = loadModule(fileName);

  if(!ok) {
    MJ_cout() << MJ_Tr::tr("Failed to load from file") 
	    << " \"" << fileName << "\"";
  }
  MJ_cout().device()->flush();  
}
