/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_module_dsp_ui.h>

#include <mj_error.h>
#include <mj_global.h>
#include <mj_kernel.h>
#include <mj_message_carrier.h>
#include <mj_module_dsp.h>
#include <mj_parameters.h>
#include <mj_satellite.h>
#include <mj_undo_info.h>

#include <qfile.h> 
#include <qtextstream.h>

#include <set>
#include <typeinfo>

static std::set<MJ_ModuleDspUi *> m_uis;

bool MJ_ModuleDspUi::m_dspUiDebug = false;

MJ_ModuleDspUi::MJ_ModuleDspUi(MJ_RtControl *rtparent,
			       MJ_ModuleDsp *module)
  : MJ_RtControl(rtparent),
    m_dbe(0),
    m_module(module)
{
  if(m_dspUiDebug) trace("MJ_ModuleDspUi::MJ_ModuleDspUi", module);

  if(module) {
    MJ_Satellite *s = module->satellite();
    if(s) s->incrementUiCount();
  }

  m_uis.insert(this);
}

MJ_ModuleDspUi::~MJ_ModuleDspUi()
{
  if(m_module) m_module->satellite()->decrementUiCount();

  std::set<MJ_ModuleDspUi *>::iterator it = m_uis.find(this);
  
  if(it != m_uis.end())
    m_uis.erase(it);
}


bool MJ_ModuleDspUi::deliverMessage
(MJ_MessageCarrier *msg, MJ_MessageCarrier *counterMsg)
{
  if(m_dspUiDebug) trace("MJ_ModuleDspUi::deliverMessage");

  lock();
  bool b = m_module->msgProcess(msg->id(), msg->message());
  unlock();

  if(counterMsg) {
    MJ_UndoInfo *ui = new MJ_UndoInfo;
    
    if(ui) {
      ui->set(msg, counterMsg);
      m_module->satellite()->addUndoInfo(ui);
    }
    else 
      return false;
  }

  return b;
}

bool MJ_ModuleDspUi::msgDeliverString(const char * i, const QString &s)
{
  lock();
  bool r = m_module->msgProcessString(i, s);
  unlock();
  return r;  
}

bool MJ_ModuleDspUi::msgDeliverDouble(const char * i, double d)
{
  lock();

  // 1) Try to find the range limitations for the value.
  MJ_Parameter *p = m_module->msgGetInput(i);

  if(p) {
    MJ_ParameterDouble *pd = dynamic_cast<MJ_ParameterDouble *> (p);
    
    // Limit the value to allowed range
    if(pd) d = pd->clamp(d);
    
    delete p;
  }
  
  // 2) Deliver value. 
  bool r = m_module->msgProcessDouble(i, d);
  unlock();
  return r;
}

bool MJ_ModuleDspUi::msgDeliverEmpty(const char * i)
{
  lock();
  bool r = m_module->msgProcessEmpty(i);
  unlock();
  return r;
}

void MJ_ModuleDspUi::addUndoInfo
(MJ_MessageCarrier *msg, 
 MJ_MessageCarrier *counterMsg, 
 const QString &name,
 MJ_Unit unit)
{
  if(!counterMsg) return;

  if(typeid(*msg) != typeid(*counterMsg)) {
    error(MJ_ERR_ARGUMENT, "MJ_ModuleDspUi::addUndoInfo", "Type mismatch");
    return;
  }

  MJ_UndoInfo *ui = new MJ_UndoInfo;

  if(ui) {
    ui->set(msg, counterMsg);
    ui->setParamName(name);
    ui->setUnit(unit);
    m_module->satellite()->addUndoInfo(ui);
  }
}

bool MJ_ModuleDspUi::saveModule(const QString & f)
{
  QFile file(f);

  if(!file.open(IO_WriteOnly)) return false;

  QTextStream stream(&file);
    
  QString suffix = f.section('.', -1, -1);

  bool ok = false;

  if(suffix == "mj") {
    
    ok = m_module->write(&stream);
  }
  else if(suffix == "mjx") {
    // trace2("MJ_ModuleDspUi::saveModule # This is XML");
    QDomDocument doc("MUSTAJUURI");
    doc.appendChild(m_module->write(doc));
    stream << doc.toCString();
    ok = file.at() > 10; // Assume the write was successful :-)
  }
  else
    error2(MJ_ERR_ARGUMENT, "MJ_ModuleDspUi::saveModule # Bad suffix \"%s\"",
	   suffix.ascii());

  if(ok && m_module->satellite()) 
    m_module->satellite()->pushMruFile(f);
    
  file.close();
    
  return ok;
}

bool MJ_ModuleDspUi::loadModule(const QString &f)
{
  QFile file(f);
  
  if(!file.open(IO_ReadOnly)) return false;
  
  QTextStream stream(&file);
  
  // Read the parameters, but ignore IO defs.
  bool ok = m_module->read(&stream, true);

  if(!ok) return false;
  
  MJ_Satellite *s = m_module->satellite();

  QCString str;

  for(int i = 0; ok; i++) {
    str = m_module->msgGetInputName(i);
    if(str.isEmpty())
      ok = false;
    else s->modified(str);
  }

  if(m_module->satellite()) 
    m_module->satellite()->pushMruFile(f);

  return true;
}


bool MJ_ModuleDspUi::startThreaded(ulong firstFrame, ulong lastFrame)
{
  if(MJ_kernel) 
    return MJ_kernel->playModule(this, MJ_KERNEL_THREADED_PLAY, 
				 firstFrame, lastFrame);

  MJ_error(MJ_ERR_INTERNAL, 
	   "MJ_ModuleDspUi::startThreaded # MJ_kernel is NULL");

  return false;
}

bool MJ_ModuleDspUi::stopPlaying()
{
  if(MJ_kernel) 
    return MJ_kernel->stopPlaying();

  return false;  
}

MJ_ModuleDspUi * MJ_ModuleDspUi::findUi(const MJ_ModuleDsp *m)
{
  for(std::set<MJ_ModuleDspUi *>::iterator it = m_uis.begin();
      it != m_uis.end(); it++) {
    MJ_ModuleDspUi * ui = (*it);
    if(ui->m_module == m)
      return ui;
  }

  return 0;
}
