/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_module_menubar.h>

#include <mj_dsp_plugin_info_gui.h>
#include <mj_error.h>
#include <mj_module_dsp_gui.h>
#include <mj_module_dsp.h>
#include <mj_parameter.h>
#include <mj_satellite.h>
#include <mj_signal_frame_gui.h>

#include <mj_tr.h>

#include <qcursor.h> 
#include <qdir.h> 
#include <qfiledialog.h>
#include <qlistview.h> 
// #include <qlabel.h> 
#include <qmessagebox.h> 

MJ_ModuleMenubar::MJ_ModuleMenubar(MJ_ModuleDspGui * parent, const char * name)
  : QMenuBar(parent, name),
    MJ_RtControl(parent),
    m_dspGui(parent),
    m_sfg(0),
    m_file(0),
    m_edit(0),
    m_pluginInfo(0)
{
  // "File" menu
  QPopupMenu *file = new QPopupMenu(this);
  
  m_fileOpen = file->insertItem(MJ_Tr::tr("&Load..."), this, 
				SLOT(open()), CTRL + Key_L);

  m_fileSave = file->insertItem(MJ_Tr::tr("&Save"), this, 
				SLOT(save()), CTRL + Key_S);

  if(!m_dspGui->module()->satellite()->mruFiles().count())
    file->setItemEnabled(m_fileSave, false);
  
  m_fileSaveAs = file->insertItem(MJ_Tr::tr("Save as..."), this, 
				  SLOT(saveAs()));

  file->insertItem(MJ_Tr::tr("Close"), parent, SLOT(hide()), CTRL + Key_W);

  insertItem( MJ_Tr::tr("&File"), file);
  m_file = file;

  // "Edit" menu
  QPopupMenu *edit = new QPopupMenu(this);
  m_undoControl.setMenu(edit);
  m_undoControl.addSatellite(m_dspGui->module()->satellite());

  insertItem( MJ_Tr::tr("&Edit"), edit);  
  m_edit = edit;

  if(!m_dspGui->module()->parentModule()) {
    edit->insertSeparator();
    edit->insertItem(MJ_Tr::tr("Signal configuration..."), 
		     this, SLOT(rateConf()));
  }

  // "Help" menu

  QPopupMenu *help = new QPopupMenu(this);
  insertItem( MJ_Tr::tr("&Help"), help);
  
  help->insertItem(MJ_Tr::tr("&Plugin information"), this, SLOT(pluginInfo()));
  help->insertItem(MJ_Tr::tr("P&arameters"), this, SLOT(parameters()));

  if(isRealTime())
    MJ_ModuleMenubar::realTimeChange(true);
}

MJ_ModuleMenubar::~MJ_ModuleMenubar()
{
  // MJ_trace("MJ_ModuleMenubar::~MJ_ModuleMenubar", this);
  delete m_sfg;
  delete m_pluginInfo;
}

void MJ_ModuleMenubar::open()
{
  if(isRealTime()) return;

  QString f = QFileDialog::getOpenFileName( 0, "*.mj", this );

  if (f.isEmpty()) return;

  f = MJ_absToRelPath(f, QDir::currentDirPath());
  
  if(!m_dspGui->loadModule(f)) {

    QString text(MJ_Tr::tr("Could not load module from file"));
    text += ": \"" + f + "\"";
    QString text2("Mustajuuri");
    QMessageBox::warning(0, text2, text, MJ_Tr::tr("Close window"));
  }
}

void MJ_ModuleMenubar::save()
{
  QString f(m_dspGui->module()->satellite()->mruFiles().first());
  
  // f = MJ_absToRelPath(f, QDir::currentDirPath());

  internalSave(f);
}

void MJ_ModuleMenubar::saveAs()
{
  if(isRealTime()) return;

  // Most Recently used files:
  const QStringList &mruList = m_dspGui->module()->satellite()->mruFiles();

  QString initially(QDir::currentDirPath());
  
  if(mruList.count()) {
    initially = mruList.first();
  }

  QString f = QFileDialog::getSaveFileName(initially, "*.mj", this );
  
  // f = MJ_absToRelPath(f, QDir::currentDirPath());

  if(f.isEmpty()) return;

  internalSave(f);
}

/** Enable and disable menu entries based on the argument. */
void MJ_ModuleMenubar::realTimeChange(bool nowRt)
{
  bool notRt = !nowRt;
  
  m_file->setItemEnabled(m_fileOpen, notRt);  
  if(m_dspGui->module()->satellite()->mruFiles().count())
    m_file->setItemEnabled(m_fileSave, notRt);
  m_file->setItemEnabled(m_fileSaveAs, notRt);
}

void MJ_ModuleMenubar::rateConf()
{
  if(m_sfg) {
    m_sfg->show();
    m_sfg->raise();
  }
  else {
    lock();
    MJ_TimeStamp *ts = m_dspGui->module()->dspGetTimer();
    int sr = (int) ts->hz();
    int fs = ts->frameSize();
    unlock();
    
    MJ_SignalFrameGui *sfg = new MJ_SignalFrameGui(sr, fs, 4096, this);
    sfg->show();
    connect(sfg, SIGNAL(masterRate(int)), this, SLOT(setRate(int)));
    connect(sfg, SIGNAL(emitFrameSize(int)), this, SLOT(setFrameSize(int)));
    connect(sfg, SIGNAL(emitGetFrameSize(int *)), 
	    this, SLOT(getFrameSize(int *)));
    m_sfg = sfg;
  }
}

void MJ_ModuleMenubar::setRate(int hz)
{
  lock();
  MJ_TimeStamp *ts = m_dspGui->module()->dspGetTimer();
  ts->setHz(hz);
  unlock();
  // MJ_trace("MJ_ModuleMenubar::setRate", hz);
}

void MJ_ModuleMenubar::setFrameSize(int s)
{
  lock();
  MJ_TimeStamp *ts = m_dspGui->module()->dspGetTimer();
  ts->setFrameSize(s);
  unlock();
  // MJ_trace("MJ_ModuleMenubar::setFrameSize", s);
}

void MJ_ModuleMenubar::getFrameSize(int *s)
{
  lock();
  const MJ_TimeStamp *ts = m_dspGui->module()->dspGetTimer();
  *s = ts->frameSize();
  unlock();
}

/** This slot should be invoked as the MRU list of the module is
    changed. 

    @see MJ_Satellite::mruChange()
*/
void MJ_ModuleMenubar::newMru()
{
  if(!isRealTime())
    m_file->setItemEnabled(m_fileSave, true);
}

/** Displays plugin information in a separate window. */

void MJ_ModuleMenubar::pluginInfo()
{
  if(m_pluginInfo) {
    m_pluginInfo->move(QCursor::pos());
  }
  else {
    m_pluginInfo = new MJ_DspPLuginInfoGui(m_dspGui->module()->dbe());
    connect(m_pluginInfo, SIGNAL(destroyed()), 
	    this, SLOT(pluginInfoDeleted()));
  }
  m_pluginInfo->show();
  m_pluginInfo->raise();
}

/** Displays the parameters in a separate window. */

void MJ_ModuleMenubar::parameters()
{
  MJ_Parameters params;
  params.setAutoDelete(true);

  m_dspGui->module()->msgGetInputs(&params);

  QString str;
  QTextOStream stream(&str);
  QValueList<QStringList> pl;
  params.writeParamsClean(&pl);

  QListView * lv = new QListView();

  lv->addColumn(MJ_Tr::tr("Index"));
  lv->addColumn(MJ_Tr::tr("Id"));
  lv->addColumn(MJ_Tr::tr("Display"));
  lv->addColumn(MJ_Tr::tr("Description"));
  lv->addColumn(MJ_Tr::tr("Value"));
  lv->addColumn(MJ_Tr::tr("Range"));

  int count = 1;

  for(QValueList<QStringList>::iterator it = pl.begin(); 
      it != pl.end(); it++) {

    QStringList & sl = (*it);

    QListViewItem * lvi = new QListViewItem(lv);

    lvi->setText(0, QString().setNum(count++));

    int i=1;
    for(QStringList::iterator it2 = sl.begin(); it2 != sl.end(); it2++) {
      lvi->setText(i, (*it2));
      i++;
    }
  }

  lv->setCaption("Parameters");
  lv->show();
}

void MJ_ModuleMenubar::pluginInfoDeleted()
{
  m_pluginInfo = 0;
}

void MJ_ModuleMenubar::internalSave(const QString &f)
{
  if(f.isEmpty()) return;
  
  if(!m_dspGui->saveModule(f)) {

    QString text(MJ_Tr::tr("Could not write module to file"));
    text += ": \"" + f + "\"";
    QString text2("Mustajuuri");
    QMessageBox::warning(0, text2, text, MJ_Tr::tr("Close window"));
  }  
  else newMru();
}
