/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_parameter.h>

#include <mj_error.h>
#include <mj_gui_enums.h>
#include <mj_message_factory.h>
#include <mj_message.h>
#include <mj_mutex.h>
#include <mj_misc.h>
#include <mj_strs.h>
#include <mj_tr.h>

#include <qapplication.h> 
#include <qtextstream.h> 
#include <qobject.h> 

#include <stdlib.h> 

static MJ_MutexAuto MJ_paramCountMutex;

int MJ_Parameter::m_counter = 0;

MJ_Parameter::MJ_Parameter()
  : m_nameC(0),
    m_nameShortC(0),
    m_translator(0),
    m_id(0),
    m_unit(MJ_UNIT_BLANK),
    m_flags(defaultFlags())
{
#ifdef MJ_MEM_DEBUG
  incrCount();
#endif
}

MJ_Parameter::MJ_Parameter(const QString &name,
			   const char *id,
			   MJ_Unit unit)
  :  m_nameC(0),
     m_nameShortC(0),
     m_translator(0),
     m_name(name.copy()),
     m_id(0),
     m_unit(unit),
     m_flags(defaultFlags())
{
#ifdef MJ_MEM_DEBUG
  incrCount();
#endif  
  setId(id);
}

MJ_Parameter::MJ_Parameter(const char *name,
			   const char *id,
			   MJ_Unit unit)
  :  m_nameC(0),
     m_nameShortC(0),
     m_translator(0),
     m_id(0),
     m_unit(unit),
     m_flags(defaultFlags())
{
#ifdef MJ_MEM_DEBUG
  incrCount();
#endif  

  setId(id);
  setName(name);
}


MJ_Parameter::MJ_Parameter(const char *name,
			   const char *nameShort,
			   const char *id,
			   MJ_Unit unit)
  : m_nameC(0),
    m_nameShortC(0),
    m_translator(0),
    m_id(0),
    m_unit(unit),
    m_flags(defaultFlags())
{
#ifdef MJ_MEM_DEBUG
  incrCount();
#endif 
  setId(id);

  setName(name);
  setNameShort(nameShort);
}

MJ_Parameter::MJ_Parameter(const MJ_Parameter &p)
  : m_nameC(0),
    m_nameShortC(0),
    m_translator(p.m_translator),
    m_name(p.m_name.copy()),
    m_nameShort(p.m_nameShort.copy()),
    m_id(0),
    m_unit(p.m_unit),
    m_flags(defaultFlags())
{
#ifdef MJ_MEM_DEBUG
  incrCount();
#endif  

  if(p.m_nameC) m_nameC = strdup(p.m_nameC);
  if(p.m_nameShortC) m_nameShortC = strdup(p.m_nameShortC);
}

MJ_Parameter::~MJ_Parameter()
{
  if(m_nameC) free(m_nameC);
  if(m_nameShortC) free(m_nameShortC);
  if(m_id) free(m_id);
  
#ifdef MJ_MEM_DEBUG
  decrCount();
#endif
}

void MJ_Parameter::setId(const char * id)
{
  MJ_delAndDup(&m_id, id);
}

void MJ_Parameter::setName(const char *s)
{
  MJ_delAndDup(&m_nameC, s);
}

void MJ_Parameter::setNameShort(const char *s)
{
  MJ_delAndDup(&m_nameShortC, s);
}

const QString &MJ_Parameter::name()
{ 
  if(m_nameC && m_name.isEmpty()) {
    MJ_TRANSLATOR(translator) = m_translator;
    if(!translator)
      translator = & MJ_Tr::tr;

#if QT_VERSION < 300
    if(qApp) m_name = (*translator) (m_nameC);
    else m_name = m_nameC;
#else
    m_name = (*translator) (m_nameC, 0);
#endif
  }
  else if(m_name.isEmpty()) {
    if(qApp) m_name = MJ_Tr::tr(MJ_Strs::Parameter);
    else m_name = MJ_Strs::Parameter;
  }

  return m_name;
}

const QString &MJ_Parameter::nameShort()
{
  if(m_nameShortC && m_nameShort.isEmpty()) {
    MJ_TRANSLATOR(translator) = m_translator;

    if(!translator)
      translator = & MJ_Tr::tr;

#if QT_VERSION < 300
    if(qApp) m_nameShort = (*translator) (m_nameShortC);
    else m_nameShort = m_nameShortC;
#else
    m_name = (*translator) (m_nameShortC, 0);
#endif
  }
  else if(m_nameShort.isEmpty()) {
    m_nameShort = name();
    m_nameShort.truncate(5); // Cut the string short.
  }
  return m_nameShort;
}


QString MJ_Parameter::name(int guiSize)
{
  QString r;

  if(guiSize == MJ_GUI_STAMP)
    r = nameShort();
  else 
    r = name();

  if(r.isEmpty())
    r = MJ_Tr::tr("Parameter");
    
  return r;
}

bool MJ_Parameter::write(QTextStream *s)
{
  (*s) << MJ_MessageFactory::determineName(type())
       << " " << " " << m_name.length() << " " << m_name << "\n";
  
  return true;
}

QStringList MJ_Parameter::writeClean()
{
  QStringList l;

  l.append(m_id);
  l.append( MJ_messageTypesShort[type()]);
  
  return l;
}

bool MJ_Parameter::read(QTextStream *s)
{
  bool ok = true;

  int n;
  QString str;

  (*s) >> str;
  if(str != MJ_MessageFactory::determineName(type())) return false;
  
  (*s) >> str;
  n = str.toInt(&ok);
  if(!ok) return false;

  if(n) m_name = s->readLine();
  
  return true;
}

MJ_MessageCarrier *MJ_Parameter::valueMessage()
{
  return 0;
}

void MJ_Parameter::incrCount()
{
  MJ_paramCountMutex.lock();
  m_counter++;
  MJ_paramCountMutex.unlock();
}

void MJ_Parameter::decrCount()
{
  MJ_paramCountMutex.lock();
  m_counter--;
  MJ_paramCountMutex.unlock();
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


const char *MJ_Parameters::m_label = "MJ_PARAMETERS";

int MJ_Parameters::findParameter(const char *id)
{
  QListIterator<MJ_Parameter> iter(*this);  
  MJ_Parameter *p;

  int i=0;

  while((p = iter.current())) {
    
    if(strcmp(id, p->id()) == 0)
      return i;
    
    // MJ_trace("MJ_Parameters::findParameter", p->id(), id);

    ++iter;
    i++;
  }
  return -1;
}

bool MJ_Parameters::writeParams(QTextStream *s)
{
  QListIterator<MJ_Parameter> iter(*this);  
  MJ_Parameter *p;

  (*s) << m_label << " " << count() << "\n";

  while((p = iter.current())) {
    if(!p->write(s)) return false;
    ++iter;
  }

  return true;
}

bool MJ_Parameters::writeParamsClean(QValueList<QStringList> *lists)
{
  QListIterator<MJ_Parameter> iter(*this);  
  MJ_Parameter *p;

  while((p = iter.current())) {
    lists->append(p->writeClean());
    ++iter;
  }

  return true;
}
