/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_pixmap_loader.h>

#include <mj_error.h>
#include <mj_paths.h>

#include <qimage.h>

#include <map>

static std::map<QString, QPixmap> MJ_pixmaps;
static std::map<QString, QIconSet> MJ_iconSets;

typedef std::map<QString, QPixmap>::iterator pixmapIterator;
typedef std::map<QString, QIconSet>::iterator iconsetIterator;

/** Load and cache a pixmap. */
QPixmap MJ_PixmapLoader::getPixmap(const QString &file)
{
  QString path(getenv(MJ_PathKeys[MJ_PATH_MUSTAJUURI]));
  QString suffix(".xpm");
  
  QString fullPath = path + "/icons/" + file + suffix;

  pixmapIterator it = MJ_pixmaps.find(fullPath);

  if(it != MJ_pixmaps.end())
    return (*it).second;

  QPixmap px(fullPath);

  MJ_trace2("MJ_PixmapLoader::getPixmap # %s: \"%s\"", 
	    px.isNull() ? "failure" : "success", fullPath.ascii());

  MJ_pixmaps[fullPath] = px;

  return px;
}

/** Load and cache an icon set. */
QIconSet MJ_PixmapLoader::getIconSet(const QString &name, const QColor &bg)
{
  iconsetIterator it = MJ_iconSets.find(name);

  if(it != MJ_iconSets.end())
    return (*it).second;

  QPixmap px1(getPixmap(name));
  QPixmap px2(getPixmap(name + "-disabled"));
  
  QIconSet is;
  
  if(!px1.isNull()) {
    is.setPixmap(px1, QIconSet::Automatic, QIconSet::Normal);

    if(px2.isNull()) {
      
      // Manufacture the px2 from px1

      QImage im = px1.convertToImage().convertDepth(32);
      
      int w = im.width();
      
      for(int y=0; y < im.height(); y++) {
	uchar * data = im.scanLine(y);
	
	for(int x = 0; x < w; x++) {
	  for(int j=1; j < 4; j++) {
	    uchar *component = data + j;

	    int bgcomponent = (bg.rgb() >> ((3-j) * 8)) & 0xff;
	    int tmp = (bgcomponent + (int) *component) / 2;
	    *component = (uchar) ((tmp > 0xFF) ? 0xFF : tmp);
	  }
	  data += 4;
	}
      }
      px2.convertFromImage(im);
    }

    /* qDebug("MJ_PixmapLoader::getIconSet # w1 = %d w2 = %d",
       px1.width(), px2.width());*/

    is.setPixmap(px2, QIconSet::Automatic, QIconSet::Disabled);
    // is.setPixmap(px1, QIconSet::Automatic, QIconSet::Disabled);
  }

  MJ_iconSets[name] = is;

  return is;
}

/// Removes all loaded (and therefore cached) pixmaps
void MJ_PixmapLoader::clear()
{
  MJ_pixmaps.clear();
  MJ_iconSets.clear();
}
