/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_config.h>
#include <mj_error.h>
#include <mj_misc.h>
#include <mj_version.h>
#include <mj_so_ref_count.h>
#include <mj_plugin_base.h>

#include <qtextstream.h> 

const char *MJ_PLuginParameterKeys[MJ_PP_SIZEOF] = {
  "About",
  "Authors",
  "Authors-latin1",
  "Classname",
  "Copyright",
  "Functionality",
  "Help",
  "Identifier",
  "Menu location",
  "Metafunctionality",
  "Name",
  "Translation file",
  "Help URL",
  "Homepage URL"
};

bool MJ_PluginDbeBase::m_debug = false;

static long MJ_plugins = 0;

MJ_PluginDbeBase::MJ_PluginDbeBase()
  : m_pluginDebug(0),
    m_pluginRefCount(0),
    m_refCount(0),
    m_mjVersion(new MJ_Version(MJ_VERSION_MAJOR,
			       MJ_VERSION_MINOR,
			       MJ_VERSION_SUBMINOR)),
    m_pluginVersion(new MJ_Version(0,0,0))
{
  const char *fname = "MJ_PluginDbeBase::MJ_PluginDbeBase";

  MJ_plugins++;

  if(m_debug) 
    MJ_trace2("%s # %p, %d", fname, this, (int) MJ_plugins);

  MJ_cpf(m_mjVersion, m_pluginVersion, fname);

}

/** Delete this plugin data-base entry. 
    
    Note that this method does not delete shared object reference
    counter. Typically users must delete it themselves \b after the
    plugin has been deleted.
*/

MJ_PluginDbeBase::~MJ_PluginDbeBase()
{
  const char *fname = "MJ_PluginDbeBase::~MJ_PluginDbeBase";

  if(m_debug) 
    MJ_trace2("%s # %s # %p, %d", fname, (const char *) menuPosition(),
	      this, (int) MJ_plugins);

  if(--MJ_plugins < 0) {

    MJ_trace2("%s # \"%s\"", fname, (const char *) m_refCount->fileName());

    if(!m_debug)
      MJ_fatal(MJ_ERR_INTERNAL, fname, "Overdelete");
    else
      MJ_error(MJ_ERR_INTERNAL, fname, "Overdelete");
  }
  
  if(m_refCount) {
    m_refCount->decrementReferenceCounter();
    // if(m_refCount->count()) delete m_refCount;
  }

  if(m_mjVersion)     delete m_mjVersion;
  if(m_pluginVersion) delete m_pluginVersion;

  if(m_pluginRefCount) {
    MJ_error(MJ_ERR_INTERNAL, 
	     "MJ_PluginDbeBase::~MJ_PluginDbeBase",
	     "Not all plugins deleted", menuPosition());
  }

}


const QString &MJ_PluginDbeBase::parameter(MJ_PLuginParameter p) const
{ 
  uint tmp = (uint) p;

  if(tmp >= MJ_PP_SIZEOF) {

    MJ_error(MJ_ERR_ARGUMENT, "MJ_PluginDbeBase::parameter");

    return QString::null;
  }

  return m_params.contains(p) ? m_params[p] : QString::null;
}


/** Gets menu position as an unmangled string. */
QString MJ_PluginDbeBase::menuPosition2()
{
  char c = MJ_pluginSeparator[0];

  QString res(menuPosition());

  if(res.isEmpty()) return res;

  if(res.find(c) == 0) res.remove(0,1);

  int i;

  while((i = res.find(c)) > 0) {
    QString left(res.left(i));
    QString right(res.right(res.length() - i - 1));
    res = left + " => " + right;
  }
  return res;
}

/** Sets one parameter string. Setting the latin1 version of
    authors field automatically sets the same string to "native"
    authors field if that field was empty. 
    
    In most cases the strings should be standard English (they will
    be translated when showing to the user).  */
void MJ_PluginDbeBase::setParameter(MJ_PLuginParameter p, const QString &s)
{
  if(m_debug) {
    QCString tmp;
    QCString stmp(s);

    tmp.resize(s.length() + 128);
    tmp.sprintf("MJ_PluginDbeBase::setParameter %p %d \"%s\"",
		(void *) this, (int) p, (const char *) stmp);
    MJ_trace(tmp);
  }

  m_params[p] = s;

  if(p == MJ_PP_AUTHORS_LATIN1 && parameter(MJ_PP_AUTHORS).isEmpty()) {
    setParameter(MJ_PP_AUTHORS, s);
  }
}

void MJ_PluginDbeBase::setParameter(MJ_PLuginParameter p, const char *str)
{
  if(!str) str = "";

  setParameter(p, QString(str));
}

QString MJ_PluginDbeBase::objectName()
{
  QString s;

  int fields = MJ_fields(menuPosition(), MJ_pluginSeparator);

  if(fields <= 0) s = "unknown";
  else s = MJ_field(menuPosition(), fields - 1, MJ_pluginSeparator);
  
  return s;
}


bool MJ_PluginDbeBase::isValid() const
{
  if(!parameter(MJ_PP_MENU_POSITION).length() ||
     !parameter(MJ_PP_IDENTIFIER).length()) return false;
  return true;
}


void MJ_PluginDbeBase::addMetaFunctionality(const QString &s)
{
  m_params[MJ_PP_METAFUNCTIONALITY] += " " + s;
}

/** Returns true if the plugin has required metafunctionality. Also
    returns true if the string is empty.
    
*/
bool MJ_PluginDbeBase::containsMetaFunctionality(const QString &s)
{
  return containsFeature(MJ_PP_METAFUNCTIONALITY, s);
}

bool MJ_PluginDbeBase::containsFeature
(MJ_PLuginParameter p, const QString &s)
{
  return (parameter(p).find(s) != -1) || s.isEmpty();
}

bool MJ_PluginDbeBase::setRefCounter(MJ_SoRefCount *RefCount)
{
  if(m_refCount && !m_refCount->count())
    delete m_refCount;

  m_refCount = RefCount;

  if(!RefCount) {
    MJ_error(MJ_ERR_ARGUMENT, 
	     "MJ_PluginDbeBase::SetRefCounter","NULL argument");
    return false;
  }

  RefCount->incrementReferenceCounter();

  return true;
}

int MJ_PluginDbeBase::pluginObjects()
{
  return (int) MJ_plugins;
}

void  MJ_PluginDbeBase::incrementReferenceCounter()
{ 
  // m_refCount->incrementReferenceCounter(); 
  m_pluginRefCount++; 
}

void MJ_PluginDbeBase::decrementReferenceCounter()
{ 
  // m_refCount->decrementReferenceCounter(); 
  m_pluginRefCount--; 
}

void MJ_PluginDbeBase::setSoHandle(void *soHandle)
{
  if(!m_refCount) setRefCounter(new MJ_SoRefCount(soHandle, QString::null));
  else m_refCount->setHandle(soHandle);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_PluginDbeBase *MJ_PluginDbeList::findDbe
(const QString &matchStr, MJ_PLuginParameter matchParameter)
{
  QListIterator<MJ_PluginDbeBase> iter(*this);
  MJ_PluginDbeBase *dbe;
  
  while((dbe = iter.current())) {
    if(dbe->parameter(matchParameter) == matchStr)
      return dbe;
    ++iter;
  }
  return 0;
}

MJ_PluginDbeBase *MJ_PluginDbeList::findFindDbe
(const QString &matchStr, MJ_PLuginParameter matchParameter)
{
  QListIterator<MJ_PluginDbeBase> iter(*this);
  MJ_PluginDbeBase *dbe;
  
  while((dbe = iter.current())) {
    if(dbe->parameter(matchParameter).find(matchStr) != -1)
      return dbe;
    ++iter;
  }
  return 0;
}

void MJ_PluginDbeList::writeParams
(QTextStream *s, 
 MJ_PLuginParameter param, 
 bool writeIndexes)
{
  QListIterator<MJ_PluginDbeBase> iter(*this);
  MJ_PluginDbeBase *dbe;
  int i = 1;

  while((dbe = iter.current())) {
    if(writeIndexes) (*s) << i++ << "  ";
    (*s) << dbe->parameter(param) << "\n";
    ++iter;
  }
}

void MJ_PluginDbeList::sort()
{
  MJ_PluginDbeList tmp;
  
  bool ad = autoDelete();
  setAutoDelete(false);

  tmp = *this;
  clear();

  for(uint i = 0; i < tmp.count(); i++) {
    inSort(tmp.at(i));
  }

  setAutoDelete(ad);
}

int MJ_PluginDbeList::compareItems(void *xp1, void *xp2)
{
  const QString &s1 = ((MJ_PluginDbeBase *)xp1)->menuPosition();
  const QString &s2 = ((MJ_PluginDbeBase *)xp2)->menuPosition();

  return s1.compare(s2);
}
