/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */
#include <mj_root_powers.h>

#include <mj_error.h>

#include <mj_iostream.h>
#include <mj_paths.h>
#include <mj_random.h>
#include <mj_realtime_scheduling.h>
#include <mj_rtc_hz.h>
#include <mj_system_signals.h>

#include <errno.h>

#ifndef WIN32
#include <sched.h>
#include <unistd.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>


#include <qdatetime.h> 
#include <qfile.h> 
#include <qfileinfo.h> 

/* These two symbols needed for the test method. */

static bool MJ_rootHup = false;

static void MJ_rootHupHandler(int)
{
  MJ_rootHup = true;
}

/** The path to the application that gives us the necessary
    capabilities. The "capability application" should be able to pass
    this application the necessary capabilities without any command
    line arguments.

    If this variable is null then the system tries to get the path of
    the application from environment variable "GIVERTCAP". If the
    variable is not defined, then default path is used:
    "/usr/local/bin/givertcap". 

    This variable is used in MJ_RootPowers::getCapabilities.
*/

const char *MJ_RootPowers::m_capApplication = 0;

/** Get real-time capabilities to this process. This method is only
    useful on Linux. This method is also defined on other systems but
    it does nothing on them.

    @see MJ_RootPowers::m_capApplication.
*/

void MJ_RootPowers::getCapabilities()
{
#ifdef __linux__
  const char * app = m_capApplication;

  if(!app) app = getenv("GIVERTCAP");

  if(!app) app = "/usr/local/bin/givertcap";
  
  QFileInfo appInfo(app);

  if(appInfo.isExecutable())
    system(app);
  else
    MJ_trace2("Cannot gain higher capabilities since \"%s\" is missing\n"
	      "You can set environment variable GIVERTCAP to "
	      "point to the givertcap application",
	      app);
#endif // __linux__
}
  

/* Root powers test.

   This method tries to get real-time priority and use the real-time
   clock (MJ_RtcHz) to determine scheduling jitter.

   Matlab commands for evaluating results:

<PRE>
   load foo.scheduling
   semilogy(foo, '*-'); grid on; zoom on;
</PRE> */

void MJ_RootPowers::test()
{
#ifndef WIN32
  MJ_realTimeScheduling(50);

  MJ_SystemSignals::setHupHandler(MJ_rootHupHandler);

  MJ_Random r;

  ulong rounds;
  ulong r1 = 250000;
  ulong sum = 0;

  uint low = (ulong) -1;
  uint high = 0;

  MJ_RtcHz rtc;
  if(!rtc.open()) {
    MJ_error(MJ_ERR_UNKNOWN, "rtc.open() failed: %s", strerror(errno));
    return;
  }

  rtc.setHz(2048);

  if(!rtc.startPeriodic()) {
    MJ_error(MJ_ERR_UNKNOWN, 
	     "rtc.startPeriodic() failed: " , strerror(errno));

    if(rtc.setHz(64)) {
      if(!rtc.startPeriodic()) 
	MJ_error(MJ_ERR_UNKNOWN, "rtc.startPeriodic() failed again.");
    }
    else {
      MJ_error(MJ_ERR_UNKNOWN, "rtc.setHz(64) failed");
      return;
    }
  }

  QTime t;
  ulong hz = rtc.getHz();
  r1 = r1 / (hz / 64);
  // rounds = hz * 20;
  rounds = hz * 60 * 60 * 5; // Five hours
  ulong i;
  const ulong maxInterval = 300;
  ulong histogram[maxInterval];

  for(i=0; i < maxInterval; i++) histogram[i] = 0;

  printf("RTC freq = %d \nUse 'kill -HUP %d' to interrupt\n", 
	 (int) hz, (int) getpid());
  fflush(0);

  sleep(1);
  
  t.start();

  for(i=0; i < rounds && !MJ_rootHup; i++) {
    t.restart();

    for(ulong j=0; j < r1; j++) { // Load the CPU.
      sum += r.get();
    }
    rtc.waitTicks(1);
    uint tmp = (uint) t.elapsed();

    if(i > (hz / 4)) {

      if(tmp < low) low = tmp;
      if(tmp > high) {
	high = tmp;
	MJ_cout() << "New high = " << high << "\n";
	MJ_cout().device()->flush();	
      }
      
      if(tmp < maxInterval) histogram[tmp]++;
      else histogram[maxInterval-1]++;
      
      if(i % (hz * 4) == 0) {
	printf("x");
	fflush(0);
      }
    }
  }
  
  MJ_cout() << "low (ms) = " << low << " high (ms) = " << high 
	  << " total time = " << (i / hz) << " secs." << "\n";
  
  MJ_cout().device()->flush();

  QFile f("foo.scheduling");
  if(!f.open(IO_WriteOnly))
    MJ_fatal(MJ_ERR_UNKNOWN, "could not open file");
  
  QTextStream s(&f);

  for(i=0; i < maxInterval; i++)
    s << histogram[i] << " " ;
#endif	//#ifndef WIN32
}

/** Test the resolution of the OS clock. Fairly often the system
    clock is granular. Clock jitter of 10 ms is common. To check the
    situation on you system, use this method. 
    
    It polls the system clock for two seconds (with real-time
    priorities). The largest interval indicates the resolution of the
    clock. This method does not report resolutions better than one
    millisecond (even if the system sports such reso). */

void MJ_RootPowers::testClock()
{
  MJ_realTimeScheduling(50);

  MJ_cout() << "Machine halt for 2 seconds :-)\n";
  MJ_cout().device()->flush();

  QTime t;
 
  t.start();

  int el = 0;
  int last = 0;

  uint low = (uint) -1;
  uint high = 0;

  while(el < 2000) {
    el = t.elapsed();
    uint tmp = el - last;

    if(!tmp) continue;

    if(tmp < low) low = tmp;
    if(tmp > high) high = tmp;

    last = el;
  }
  MJ_cout() << "low (ms) = " << low << " high (ms) = " << high << "\n";
  MJ_cout().device()->flush();
}
