/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_rt_control.h>

#include <mj_error.h>
#include <mj_mutex.h>

#include <di_time.h>

#include <stdio.h>

#ifdef __OS_LINUX__
#include <mj_realtime_scheduling.h>
#endif

static DI_Time timer;

MJ_RtControl::MJ_RtControl()
  : m_rtStatus(MJ_RT_NOT),
    m_rtMutex(0),
    m_rtLocks(0),
    m_rtParent(0),
    m_rtChildren(0)
{}

MJ_RtControl::MJ_RtControl(MJ_RtControl *parent)
  : m_rtMutex(0),
    m_rtLocks(0),
    m_rtParent(parent),
    m_rtChildren(0)
{
  if(parent) {
    parent->addRtChild(this);
  }
  else {
    m_rtStatus = MJ_RT_NOT;
  }
}

MJ_RtControl::~MJ_RtControl()
{
  if(m_rtLocks != 0) {
    MJ_error(MJ_ERR_INTERNAL, "MJ_RtControl::~MJ_RtControl", "Non-zero lock count");
    
    if(m_rtLocks > 0)
      while(m_rtLocks)
	unlock();
  }
  if(m_rtParent) m_rtParent->removeRtChild(this);
  
  if(m_rtChildren) {
    
    { 
      QListIterator<MJ_RtControl> xIter(*m_rtChildren);
      MJ_RtControl *rtg;
      for(;(rtg = xIter.current());++xIter) {
	rtg->rtStatusChange(MJ_RT_NOT, 0);
	rtg->m_rtParent = 0;
      }
    }
    delete m_rtChildren;
  }
}

void MJ_RtControl::rtStatusChange(MJ_RtStatus status, MJ_Mutex *mutex)
{
  const char *fname = "MJ_RtControl::statusChange";

  if(status == MJ_RT_THREADED_RT && !mutex)
    MJ_error(MJ_ERR_ARGUMENT, fname, "Threaded, but no mutex");

  if(m_rtLocks) 
    MJ_error(MJ_ERR_ARGUMENT, fname, "Lock count not zero");

  m_rtMutex = mutex;
  m_rtStatus = status;

  // Distribute to children
  if(m_rtChildren) {
    uint chs = m_rtChildren->count();
    for(uint i = 0; i < chs; i++) {
      m_rtChildren->at(i)->rtStatusChange(status, mutex);
    }
  }

  bool rt = (bool) status;
  realTimeChange(rt);
}

void MJ_RtControl::setRtParent(MJ_RtControl *xpParent)
{
  if(m_rtParent) m_rtParent->removeRtChild(this);
  m_rtParent = xpParent;
  if(xpParent) xpParent->addRtChild(this);
}


void MJ_RtControl::addRtChild(MJ_RtControl *child)
{
  if(!m_rtChildren) m_rtChildren = new QList<MJ_RtControl>;

  if(m_rtChildren->find(child) != -1) {
    MJ_error(MJ_ERR_ARGUMENT, "MJ_RtControl::AddChild", "Child is already");
    return;
  }
  m_rtChildren->append(child);
  child->rtStatusChange(m_rtStatus, m_rtMutex);
} 

void MJ_RtControl::removeRtChild(MJ_RtControl *child)
{
  if(!m_rtChildren) {
    MJ_error(MJ_ERR_INTERNAL, "MJ_RtControl::removeRtChild");
    return;
  }

  if(!m_rtChildren->remove(child)) {
    char buf[256];

    sprintf(buf, "MJ_RtControl::RemoveChild # No such child, children: %d",
	    (int) m_rtChildren->count());
    MJ_error(MJ_ERR_ARGUMENT, buf);
  }
}

bool MJ_RtControl::lock()
{
  timer.getTimeOfDay();

  m_rtLocks++;

  if(!m_rtMutex) return true;

#ifdef __OS_LINUX__
  /* Try to increase priority, while the mutex is locked.  */  
  if(m_rtStatus == MJ_RT_THREADED_RT) {
    // This may fail, but that is not a problem.
    MJ_realTimeScheduling(MJ_RTP_AUDIO);
  }
#endif

  bool bLock = m_rtMutex->lock();

  if(!bLock) {
    MJ_warning(MJ_ERR_INTERNAL, "MJ_RtControl::lock # Failed to lock mutex");
    m_rtLocks--;
  }
  return bLock;
}

void MJ_RtControl::unlock()
{
  m_rtLocks--;
  if(m_rtLocks < 0) 
    MJ_error(MJ_ERR_INTERNAL, "MJ_RtControl::unlock # Negative lock count");
  if(!m_rtMutex) return;
  m_rtMutex->unlock();

#ifdef __OS_LINUX__

  /* Try to drop priority as the mutex is freed.  */  
  if(m_rtStatus == MJ_RT_THREADED_RT) {

    double lockTime = timer.secsSince();
    if(lockTime > 0.5) {
      MJ_error2(MJ_ERR_INTERNAL, 
		"MJ_RtControl::unlock # long lock period = %g",
		lockTime);
    }

    // This may fail, but that is not a problem.
    MJ_realTimeScheduling(MJ_RTP_NOPRIORITY);
  }
#endif
}

void MJ_RtControl::realTimeChange(bool)
{}
