/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_rt_driver.h>

#include <mj_cpu_count.h>
#include <mj_dsp_plugin.h>
#include <mj_error.h>
#include <mj_failure_diagnose.h>
#include <mj_prefs.h>
#include <mj_rtc_hz.h>
#include <mj_sleep.h>

#include <di_sleep.h>

#include <qapplication.h> 

#include <typeinfo>

bool MJ_RtDriver::m_debug = false;

MJ_RtDriver::MJ_RtDriver(MJ_ModuleDsp *m)
  : m_mainModule(0),
    m_stop(false),
    m_firstFrame(0),
    m_lastFrame((ulong) -100000),
    m_minBufferMS(MJ_prefs.minimumLatency())
{
  if(m_debug) MJ_trace("MJ_RtDriver::MJ_RtDriver", this);
  setModule(m);
  m_errors.setAutoDelete(true);
}

MJ_RtDriver::~MJ_RtDriver()
{
  if(m_debug) MJ_trace("MJ_RtDriver::~MJ_RtDriver", this);
  setModule(0);
}

void MJ_RtDriver::setModule(MJ_ModuleDsp *m)
{
  if(m_debug) MJ_trace("MJ_RtDriver::setModule", this);

  if(m_mainModule) m_mainModule->setRtDriver(0);

  m_mainModule = m;
  if(m) m->setRtDriver(this);
}

bool MJ_RtDriver::runPeriods(ulong frames)
{
  const char *fname = "MJ_RtDriver::runFrames";
  
  m_stop = false;
  if(m_debug) MJ_trace(fname, this);

  if(!prepare(0)) {
    MJ_error(MJ_ERR_INTERNAL, fname, "Prepare failed");
    return false;
  }

  MJ_CpuCount cc;

  if(m_audioIos.size())
    processRt(frames);
  
  else if(!qApp) {
    for(ulong i=0; (i < frames) && !m_stop; i++) 
      processPeriod();
  }

  else {
    // Check GUI 10 times per second.
    double stepSize = (double) m_mainModule->dspFrameSize() /  
      (double) m_mainModule->dspMasterRate();
    ulong interval = (ulong) (0.1 / stepSize);

    for(ulong i=0; (i < frames) && !m_stop; i++) { 
      processPeriod();
      if((i % interval) == 0) qApp->processEvents(0);
    }
  }
  
  cc.update();
  double cpuTime = (double) cc.cpuTimeMS() / 1000.0;
  double dspTime = m_mainModule->dspGetTimer()->secondsD();
  MJ_trace2("%s # CPU time: %3.3f, DSP time = %3.3f  -> %3.3f speed, ",
	    fname, cpuTime, dspTime, dspTime / cpuTime);
  
  return stop();
}

void MJ_RtDriver::addError(MJ_FailureDiagnose *f)
{
  if(!f) MJ_fatal(MJ_ERR_ARGUMENT, "MJ_RtDriver::addError");

#if 0
  m_errors.append(f); // RRR This is what should happen.
#else
  f->toUser();
#endif  
}

/** Fills the audio output buffers to prescribed length. Returns the
    highest buffer length reached. */

float MJ_RtDriver::fillOutputs(float fillMs)
{
  const char *fname = "MJ_RtDriver::fillOutputs";

  uint aum = m_audioIos.size(); // Number of audio modules.
  float lbuf;
  int count = 0;

  float fillS = fillMs / 1000.0;
  float maxFill = 0.0;

  while((lbuf = lowestBuf(true)) < fillS && aum) {

    if (lbuf > maxFill) maxFill = lbuf;
    
    for(uint j=0; j < aum; j++) {
      MJ_ModuleRtAudio *tmp = m_audioIos.getFast(j);
      
      if(tmp->dspInChannels() || !count)
	tmp->dspProcessPeriod();
    }
    count ++;

    if(m_debug && ((count & 0xFF) == 0)) { // Give info to user...
      MJ_trace(fname, (int) (lbuf * 1000));
    }
    if(count > 1000) {
      if(m_debug) 
	MJ_trace(fname, "Breaking from output filling.");
      break;
    }
  }
  if (lbuf > maxFill) maxFill = lbuf;
  return maxFill * 1000.0;
}

float MJ_RtDriver::fillInputs(float fillMs)
{
  float fillS = fillMs / 1000.0;
  
  for(uint i = 0; i < 100; i++) {

    float fill = lowestBuf(false);
    
    if(fill >= fillS) return fill * 1000.0;

#ifdef __OS_LINUX__
    MJ_RtcHz *rtc = MJ_RtcHz::rtc();  
    if(rtc) rtc->waitTicks(2);
    else DI_Sleep::sleepMs(10);
#else
    DI_Sleep::sleepMs(5);
#endif

  }
  
  return -1.0;
}

bool MJ_RtDriver::prepare(long firstFrame, long lastFrame)
{
  const char * fname = "MJ_RtDriver::prepare";

  if(m_debug) MJ_trace(fname);

  m_firstFrame = firstFrame;
  m_lastFrame  = lastFrame ? lastFrame : (ulong) -100000;

  m_minBufferMS = MJ_prefs.minimumLatency();

  if(!prepareInternal()) {
    if(m_mainModule)
      m_mainModule->dspGetTimer()->setRealTime(false);

    return false;
  }

  return true;
}

bool MJ_RtDriver::prepareFast()
{
  if(!m_mainModule->dspPrepareFast()) {
    m_mainModule->dspStopFast();
    m_mainModule->dspStop();
    return false;
  }

  return true;
}

void MJ_RtDriver::processPeriod()
{
  m_mainModule->dspProcessPeriod();
}

bool MJ_RtDriver::stop()
{
  const char * fname = "MJ_RtDriver::stop";

  if(m_debug) MJ_trace(fname, this);

  if(!m_mainModule) return false;
  
  m_mainModule->dspGetTimer()->setRealTime(false);

  bool ok = m_mainModule->dspStopFast();
  ok = ok && m_mainModule->dspStop();
  
  if(m_debug) MJ_trace(fname, m_malloc.units());

  m_malloc.free(); // Free all memory.

  return ok;
}

double MJ_RtDriver::lowestBuf(bool out)
{
  double low = 1e+5;

  uint auIos = m_audioIos.size();
  
  for(uint j=0; j < auIos; j++) {
    MJ_ModuleRtAudio *tmp = m_audioIos.getFast(j);
    
    if(out) {
      if(!tmp->dspInChannels())
	continue;
    }
    else if(tmp->dspInChannels())
      continue;

    int bufLen = tmp->bufferLength();
    int sr = (int) tmp->dspMasterRate();
    
    double bufLenS = (double) bufLen / (double) sr;

    if(low > bufLenS) //  && tmp->dspInChannels())
      low = bufLenS;
    
    // MJ_trace("MJ_RtDriver::lowestBuf", bufLen);
  }

  return low;
}

bool MJ_RtDriver::prepareInternal()
{
  const char * fname = "MJ_RtDriver::prepareInternal";

  if(m_debug) MJ_trace(fname);
  
  if(!m_mainModule) return false;

  getChildren();

  // m_mainModule->dspSetTimer(m_mainModule->dspGetTimer());

  if(!checkConflicts()) {
    if(m_debug) MJ_error(MJ_ERR_INTERNAL, fname, "Conflicts");
    return false;
  }

  MJ_trace2("%s # Beginning at frame %lu", fname, m_firstFrame);

  m_mainModule->dspGetTimer()->setStamp(m_firstFrame);

  if(!m_mainModule->dspPrepare()) {
    if(m_debug) MJ_error(MJ_ERR_INTERNAL, fname, "Prepare failed");
    return false;
  }

  m_mainModule->dspGetTimer()->setRealTime(true);

  return true;  
}

void MJ_RtDriver::processRt(ulong periods)
{
  const char *fname = "MJ_RtDriver::processRt";

  if(m_debug) MJ_trace(fname, "Enter");

  MJ_Sleep s;

  // Check GUI 10 times per second.
  double stepSize = (double) m_mainModule->dspFrameSize() /  
    (double) m_mainModule->dspMasterRate();
  ulong interval = (ulong) (0.1 / stepSize);
 
  double dLimit = 30.0;

  if(m_debug) MJ_trace(fname, "Fill< output");

  fillOutputs(dLimit);
  
  if(m_debug) MJ_trace(fname, "Enter RT");

  for(ulong i=0; (i < periods) && !m_stop; i++) {
    processPeriod();

    if(qApp)
      if((i % interval) == 0) qApp->processEvents(0);

    
    double dLowestOutBuf = lowestBuf();
    // cout << " l: " << dLowestOutBuf ;

    // Sleep:
    if(dLowestOutBuf > 0.03) // 30 ms
      s.sleepMs((uint) (1000.0 * (dLowestOutBuf - 0.03)));
  }
}

void MJ_RtDriver::getChildren()
{
  const char *fname = "MJ_RtDriver::getChildren";

  if(m_debug) MJ_trace(fname);
  
  m_children.clear();
  m_audioIos.resize(0);

  if(m_mainModule) m_mainModule->getChildren(m_children, true);

  m_children.append(m_mainModule);
  
  QListIterator<MJ_ModuleDsp> iter(m_children);
  MJ_ModuleDsp *m;

  if(m_debug) MJ_trace(fname);

  while((m = iter.current())) {
    
    MJ_ModuleRtAudio *ma = dynamic_cast<MJ_ModuleRtAudio *> (m);

    if(ma) {
      m_audioIos.resize(m_audioIos.size() + 1);

      int sr = ma->dspSamplingRate();
      ma->setMinimumBuffer((uint) (sr * m_minBufferMS * 0.001));

      /* First come the *outputs*. This bad behaviour is needed for
         alsa-cvs-nov-01 RME drivers. */
      /* if(ma->dspInChannels()) {
	m_audioIos.shiftUp(0, 1);
	m_audioIos.get(0) = ma;
	}
      else*/
      m_audioIos.getFast(m_audioIos.size() - 1) = ma;
    }
    ++iter;
  }
 
  
  if(m_debug) {
    MJ_trace2("%s # Children: %d, audio modules: %d",
	      fname, (int)m_children.count(), (int) m_audioIos.size());
  }
}

bool MJ_RtDriver::checkConflicts()
{
  const char * fname = "MJ_RtDriver::checkConflicts";

  if(m_debug) MJ_trace(fname);

  uint i;

  MJ_PluginDbeDsp *dbe;

  QList<MJ_ModuleDsp> conflicts;

  for(i = 0; i < m_children.count(); i++) {
    MJ_ModuleDsp *m =  m_children.at(i);
    dbe = m->dbe();
    if(dbe) {
      if(dbe->mayConflict()) {
	conflicts.append(m);
      }
    }
  }  
  
  QListIterator<MJ_ModuleDsp> iter(conflicts);

  uint max =  conflicts.count();

  bool ok = true;

  if(m_debug) MJ_trace(fname, (int) max);

  /* Why "i+1": The last item will be checked anyhow, no
     double-checks. */
  for(i = 0; i + 1 < max; i++) { 
    MJ_ModuleDsp *m1 = conflicts.at(i);

    // Iterate over the remaining modules:
    iter.toFirst();
    for(iter += i; !iter.atLast();) {
      ++iter; 

      MJ_ModuleDsp *m2 = iter.current();
      bool problem = m2->conflicts(m1);
      
      if(problem) {
	char buf[512];
	sprintf(buf, "%s # Unable to solve a conflict # %s # %s",
		fname, typeid(*m1).name(), typeid(*m2).name());
	
	MJ_error(MJ_ERR_INTERNAL, buf);
	ok = false;
      }
    }
  }
  return ok;
}
