/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_rt_thread.h>

#include <mj_global.h>
#include <mj_kernel.h>
#include <mj_math.h>
#include <mj_prefs.h>
#include <mj_realtime_scheduling.h>
#include <mj_rt_driver.h>
#include <mj_rtc_hz.h>
#include <mj_root_powers.h>
#include <mj_sleep.h>
#include <mj_time_monitor.h>
#include <mj_watch_dog.h>

#include <stdio.h>

bool MJ_RtThread::m_debug = false;

MJ_RtThread::MJ_RtThread()
  : m_moduleAccess(),
    m_halt(false, false, false),
    m_driver(0),
    m_playing(false),
    m_stopPlaying(false),
    m_exit(false),
    m_firstFrame(0),
    m_lastFrame((ulong) -100000), // MAX_ULONG - 100000
    m_load(0.0),
    m_bufferFills(16), // Buffer histogram bins
    m_underflows(0),
    m_watchDog(0)
{
  m_halt.lock();
}

MJ_RtThread::~MJ_RtThread()
{
  if(m_watchDog) m_watchDog->setGuard(0);
}

bool MJ_RtThread::rtRun()
{
  bool systemThreads = DI_ThreadAvailable::systemLevelWorks();  
  m_tryRtPrio = (systemThreads && MJ_prefs.tryRtPrio());

  if(systemThreads) return runSystem();
  else return runProcess();
}

bool MJ_RtThread::start(ulong firstFrame, ulong lastFrame)
{
  if(m_debug) MJ_trace("MJ_RtThread::start");

  if(m_playing) return false;

  if(!lastFrame) lastFrame = (ulong) -100000;

  m_firstFrame = firstFrame;
  m_lastFrame = lastFrame;

  m_halt.unlock();
  return true;
}

bool MJ_RtThread::stop()
{
  const char *fname = "MJ_RtThread::stop ";

  if(m_debug) MJ_trace(fname);

  // cout << fname << "enter 1" << endl;
  // m_halt.lock();

  // cout << fname << "enter 2" << endl;
  
  m_moduleAccess.lock();
  m_stopPlaying = true;
  m_exit = true;
  m_moduleAccess.unlock();

  // cout << fname << "enter 3" << endl;

  m_halt.unlock();

  MJ_Sleep s;

  while(m_playing) {
    s.sleepMs(300);
    MJ_trace(fname, "waiting");
  }

  return true;
}

void MJ_RtThread::requestEnd(bool rtThread)
{
  if(m_debug) MJ_trace("MJ_RtThread::requestEnd");

  if(rtThread) 
    m_moduleAccess.lock();

  m_stopPlaying = true;
  m_exit = true;

  if(rtThread) {
    // m_moduleAccess.lock();
    m_moduleAccess.unlock();
    m_halt.unlock(); // The DSP thread *must* be in stopper.
  }
}

/** Returns the number of underflows in the system. The underflow
    counter is reset to zero by this method. */

uint MJ_RtThread::underflows()
{
  return m_underflows;
}

void MJ_RtThread::childLoop()
{
  const char * fname = "MJ_RtThread::childLoop";

  if(m_debug) MJ_trace(fname, "ENTER");

  bool ok = true; 

  m_halt.lock(); // Block here.
  m_halt.unlock();

  while(!m_exit && ok) {
    ok = starter();
    if(!ok) {
      MJ_error(MJ_ERR_UNKNOWN, fname, "Start failed");
      break;
    }

    ok = processor();

#ifdef __OS_LINUX__
    if(MJ_RtcHz::rtc()) delete MJ_RtcHz::rtc();
#endif

    ok = ok && stopper();
    
    if(m_debug) MJ_trace(fname, "Loop end");
  }
  // Make sure the thread requesting end can break away:
  m_playing = false; 

  if(!ok) MJ_error(MJ_ERR_UNKNOWN, fname);

  if(m_watchDog) {
    m_watchDog->setGuard(0);
    m_watchDog = 0;
  }
  
  if(m_debug) MJ_trace(fname, "EXIT");
}

/// Start processing
bool MJ_RtThread::starter()
{
  const char *fname = "MJ_RtThread::starter";

  if(m_debug) MJ_trace(fname);

  m_playing = true;
  m_stopPlaying = false;

  m_bufferFills.setAll(0);
  m_underflows = 0;

  if(!m_driver) return false;

  // Do basic preparations for DSP:
  if(!m_driver->prepare((long)m_firstFrame)) {
    MJ_error(MJ_ERR_UNKNOWN, fname, "prepare failed");
    return false;
  }
  
  // Get higher priority
  if(m_tryRtPrio) {

#ifdef __OS_IRIX___
    struct sched_param sp;
    sp.sched_priority = MJ_RTP_AUDIO;

    int e = pthread_setschedparam(m_thread, SCHED_FIFO, &sp);

    if(e) {
      MJ_error(MJ_ERR_UNKNOWN, fname, "Could not get high priority");
    }
    else if(m_debug) 
      MJ_trace(fname, "Got RT priority");
#else
    
    if(!MJ_realTimeScheduling(MJ_RTP_AUDIO))
      MJ_error(MJ_ERR_UNKNOWN, fname, "Could not get high priority");
    else if(m_debug) 
      MJ_trace(fname, "Got RT priority");
#endif
  }
  else if(m_debug) {
    MJ_trace(fname, "Did not try RT prio");
  }

  m_guard = true;

  // Set up the watch dog:
  MJ_WatchDog *dog = MJ_WatchDog::dog();
  if(dog) dog->setGuard(& m_guard);
  m_watchDog = dog;

  // Do final preparations:
  if(!m_driver->prepareFast()) {
    MJ_error(MJ_ERR_UNKNOWN, fname, "prepare failed");
    return false;
  }
  
  // MJ_allocFree();

  return true;
}

/// Process frames forever.
bool MJ_RtThread::processor()
{
  const char *fname = "MJ_RtThread::processor";

  if(m_debug) MJ_trace(fname, "ENTER");

  // return true;

  m_guard = true;

#ifdef __OS_LINUX__
  /* On Linux sleeping very short time periods is a bit difficult.  It
     seems the only reliable approach is to use the real-time clock
     with frequency greater than 1024 Hz...  */

  MJ_RtcHz * rtc = MJ_RtcHz::newRtc(0, 2048);
  if(rtc) {
    if(!rtc->startPeriodic()) {
      delete rtc;
      rtc = 0;
    }
    else {
      MJ_trace(fname, "Got RTC @ 2048 Hz");
    }
  }

  /* Note that the object is not deleted in this function, but it MUST
     be deleted in the hosting function. */

#endif

  MJ_SleepSync s;
  MJ_TimeMonitor tm;
  m_abp.initRun();

  bool independentSync = m_driver->nofAudioIos() == 0;
  ulong count = 0;
  ulong lastSleep = 0;

  MJ_TimeStamp *stamp;
  MJ_ModuleDsp *m = m_driver->module();

  if(m && independentSync) stamp = m->dspGetTimer();
  else stamp = 0;

  ulong periods = 0;
  // How often to check for CPU overload.
  ulong checkInterval = 1 + (ulong) 
    (2.0 / m->dspGetTimer()->frameDurationS()); 
  ulong forceSleepInterval = checkInterval / 4 + 1;
  ulong afterSleepUnderflows = 0;

  if(!m_driver) {

    return false;
  }

  m_moduleAccess.lock();

  if(m_debug) {
    if(stamp) MJ_trace(fname, "Internal sync");
    else MJ_trace(fname, "External sync");
  }

  bool syncWithOutput = true;

  float maxBuf = m_driver->fillOutputs(m_abp.minBuffer());

  if(maxBuf > 1000.0) {
    syncWithOutput = false;
    MJ_trace2("%s # Trying to fill input buffers to %f", 
	      fname, (float) m_abp.minBuffer());
    maxBuf = m_driver->fillInputs(m_abp.minBuffer());
  }

  if(maxBuf < 0.0)
    maxBuf = 300.0;

  m_currAbp = m_abp;

  if(maxBuf < m_currAbp.maxBuffer() || 
     maxBuf > m_currAbp.maxBuffer() * 1.5) {

    if(maxBuf < 1.0)
      MJ_error(MJ_ERR_UNKNOWN, fname, "Failed to fill buffers");
    
    m_currAbp.setRangesMs(maxBuf * 0.6, maxBuf * 0.9);
  }
  else if(m_debug) {
    MJ_trace(fname, "Buffers filled");
  }

  MJ_trace2("%s # Current buffering parameters [%3.2f, %3.2f]",
	    fname, m_currAbp.minBuffer(), m_currAbp.maxBuffer());
  
  while(!m_stopPlaying) {    

    // MJ_trace(fname, "Loop 1");

    /* if(m_debug && ((periods % 128) == 0))
       MJ_trace(fname, (int) periods+1);*/

    m_driver->processPeriod();

    m_guard = true;

    periods++;

    if(m->dspGetTimer()->stamp() >= m_lastFrame) {

      if(MJ_kernel) MJ_kernel->stopPlaying(true);
      else {
	m_stopPlaying = true;
	m_exit = true;
      }
    }
    /* Avoid CPU overload (we may hang the machine if this is not taken
       care of). */
    if((periods % checkInterval) == 0) {
      tm.update();
      m_load = tm.cpuLoad();

      if(m_load > 0.9) {
	m_moduleAccess.unlock();
	MJ_trace(fname, "CPU overload");
	s.sleepMs(20);
	m_moduleAccess.lock();
      }
    }
    
    // MJ_trace(fname, "Loop 2.5");

    // Try to sleep:
    if(stamp) { // Not a single output module that would sync us.
      m_moduleAccess.unlock();
#ifndef WIN32
      s.sleepTo(stamp);
#else
      s.sleepMs(10);
#endif
      m_moduleAccess.lock();
    }
    else {
      
      // MJ_trace(fname, "Loop 2.6");

      float lb = m_driver->lowestBuf(syncWithOutput) * 1000.0;

      if(periods > checkInterval) {
	m_currAbp.addRunValue(lb);
      }

      if((lb > m_currAbp.minBuffer() && syncWithOutput) ||
	 (lb < m_currAbp.minBuffer())) { 

	m_moduleAccess.unlock();

	float sleepMs = lb - m_currAbp.minBuffer() - 1;
	
	if(!syncWithOutput)
	  sleepMs *= -1.0;

	// MJ_trace2("%s # Sleep ms %f (%f)", fname, sleepMs, lb);
	
#ifdef __OS_LINUX__
	struct timeval tv1, tv2;
	gettimeofday(&tv1, 0); 
	
	float sleptMs = 0;

	if(rtc) {
	  while(sleptMs < sleepMs) {
	  
	    rtc->waitTicks(2);
	    gettimeofday(&tv2, 0); 
	    long diff = tv2.tv_usec - tv1.tv_usec + 
	      (tv2.tv_sec - tv1.tv_sec) * 1000000;

	    if(diff > 4000 + sleepMs * 1000)
	      MJ_trace("MJ_RtThread::processor blocked !", diff);

	    sleptMs = (float) diff * (float) 0.001;
	  }
	}
	else s.sleepMs(10);
#else
	s.sleepMs(sleepMs);
#endif

	// MJ_trace("MJ_RtThread::processor, sleep 1");
	m_moduleAccess.lock();

	lastSleep = count;
      }
      else { // Make sure we do not block the CPU
	count++;
	if((lastSleep + forceSleepInterval) < count && 
	   periods > checkInterval * 2) {
	  m_moduleAccess.unlock();
	  s.sleepMs(20);
	  MJ_trace("MJ_RtThread::processor, sleep 2");
	  m_moduleAccess.lock();
	  lastSleep = count;
	}
      }


      /* Save buffering parameters. */
      int bin = (int) 
	(lb / m_currAbp.maxBuffer() * (float) m_bufferFills.size() * 0.5);

      if(bin < 0) bin = 0;
      else if(bin >= (int)m_bufferFills.size()) bin = m_bufferFills.size() - 1;
      m_bufferFills.get(bin)++;
      
      m_underflows = m_currAbp.underFlows();
      
      // MJ_trace(fname, "Loop 3");
    }

    /* if(m_debug && ((periods % 128) == 0))
       MJ_trace(fname, (int) periods);*/
  }
  
  MJ_trace(fname, "preparing EXIT");

  m_moduleAccess.unlock();

  if(m_debug) MJ_trace(fname, "EXIT");

  if(!stamp) {
    MJ_trace2("%s # Audio buffers, lowest = %f, highest = %f,"
	      " total periods = %ld, frames = %lu,"
	      " underflows = %lu, asus = %lu",
	      fname, m_currAbp.lowestBufMs(), 
	      m_currAbp.highestBufMs(), 
	      periods, m->dspGetTimer()->stamp(),
	      m_currAbp.underFlows(),
	      afterSleepUnderflows);
  }
  

  return true;
}

/// Wait here until the master mutex is released. 
bool MJ_RtThread::stopper()
{
  if(m_debug) MJ_trace("MJ_RtThread::stopper", this);
  
  if(!m_driver) return false;
  
  bool ok = m_driver->stop();
  
  m_playing = false;

  // ok = ok && m_halt.lock(); // Block here.
  // m_halt.unlock();

  if(m_debug) MJ_trace("MJ_RtThread::stopper # EXIT");

  return ok;
}
