/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_RT_THREAD_H
#define MJ_RT_THREAD_H

#include <mj_array.h>
#include <mj_thread.h>

#include <mj_audio_buf_params.h>

#include <di_mutex.h>

class MJ_RtDriver;
class MJ_WatchDog;

/** A thread that runs the DSP.
    
    Using this class DOES take some care.

    Objects of this class span a new thread - the DSP thread (DT).
    The spanning thread is the User Thread (UT). One must be careful
    with synchronizing these two threads.  

    The DT runs the cycle starter->processor->stopper. 
    
    For examples see file <mj_module_dsp_row.C", function:
    MJ_ModuleDspRow_QG::startThreaded and
    MJ_ModuleDspRow_QG::stopThreaded.  */

class MJ_RtThread : public MJ_Thread
{
public:
  MJ_RtThread();
  virtual ~MJ_RtThread();

  DI_MutexAuto & moduleAccesMutex() { return m_moduleAccess; }
  
  /** The method to be used to start the processing. */
  bool rtRun();

  /// Starts the processing. UT calls 
  bool start(ulong firstFrame = 0, ulong lastFrame = 0);
  /// Stops the processing. UT calls
  bool stop();

  void requestEnd(bool rtThread = false);

  /// Returns current CPU load
  float load() const { return m_load; }

  const MJ_Array<uint> & bufferFills() const { return m_bufferFills; }
  void clearBufferFills() { m_bufferFills.setAll(0); }
  
  uint underflows();

  bool playing() { return m_playing; }
  
  void setDriver(MJ_RtDriver *driver) { m_driver = driver; }

  void setwatchDog(MJ_WatchDog *watchDog) {  m_watchDog = watchDog; }

  bool *guardPtr() { return & m_guard; }

  const MJ_AudioBufParams & currentAbp() const { return m_currAbp; }

  void clearAbp() { m_currAbp.initRun(); }

  /* void setPlayFrames(ulong firstFrame, ulong lastFrame)
     { m_firstFrame = firstFrame; m_lastFrame = lastFrame; } */

  static void setDebug(bool b) { m_debug = b; }
  
protected:
  
  /** @name DSP-thread methods: */
  //@{
  virtual void childLoop();

  bool starter();
  bool processor();
  bool stopper();
  //@}

  

  DI_MutexAuto m_moduleAccess;
  DI_MutexAuto m_halt;

  MJ_RtDriver *m_driver;
  bool         m_playing;
  bool         m_stopPlaying;
  bool         m_exit;
  bool         m_guard;

  bool         m_tryRtPrio;
  bool         m_hasRtPrio;

  MJ_AudioBufParams m_abp, m_currAbp;

  ulong        m_firstFrame;
  ulong        m_lastFrame;

  /// Current CPU load.
  float        m_load;

  /// Current buffer fill status
  MJ_Array<uint> m_bufferFills;

  /// Number of underflows since last check
  uint           m_underflows;

  MJ_WatchDog   *m_watchDog;

  static bool  m_debug;
};

#endif
