/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_run_status.h>

#include <mj_error.h>

#include <mj_global.h>
#include <mj_graphic_polygon.h> 
#include <mj_graphic_text.h>
#include <mj_graphics_area.h> 
#include <mj_graphics_context_simple.h>
#include <mj_gui_utils.h>
#include <mj_kernel.h>
#include <mj_level_display.h> 
#include <mj_module_dsp.h>
#include <mj_module_dsp_gui.h>
#include <mj_pixmap_loader.h>
#include <mj_rt_driver.h>
#include <mj_rt_thread.h>
#include <mj_tr.h>
#include <mj_vector.h>

#include <qapplication.h> 
#include <qlayout.h> 
#include <qlabel.h> 
#include <qpushbutton.h> 
#include <qtoolbutton.h> 
#include <qtooltip.h> 

#include <stdio.h>

#ifdef __linux__
#include <mj_realtime_scheduling.h>
// #include <sched.h>
#define MJ_GET_RT_PRIO() MJ_realTimeScheduling(MJ_RTP_AUDIO)
#define MJ_DROP_RT_PRIO() MJ_realTimeScheduling(MJ_RTP_NOPRIORITY)
#else
#define MJ_GET_RT_PRIO() 
#define MJ_DROP_RT_PRIO() 
#endif

MJ_LcdClock::MJ_LcdClock(QWidget      *parent,
			 const char   *name)
  : QWidget(parent, name),
    m_fragMax(100),
    m_fragDigits(2)
{
  QHBoxLayout *layout = new QHBoxLayout(this, 1);
  int i;

  m_tmp = ":";

  for(i = 0; i < MJ_LCD_SIZEOF; i++) {
    m_timeDisplays[i] = new QLabel(this);
    layout->addWidget(m_timeDisplays[i], 30);
    if(i != MJ_LCD_FRAGMENTS)
      layout->addWidget(new QLabel(m_tmp, this), 10);

    m_time[i] = -1;
  }

  // setMinimumSize(10, 10);
  // setMaximumSize(10, 10);
  clear();
}

MJ_LcdClock::~MJ_LcdClock()
{}

void MJ_LcdClock::clear()
{
  set(0,0,0,0);
}

void MJ_LcdClock::set(int hour, int min, int sec, int frag)
{
  setNumber(MJ_LCD_HOURS, hour);
  setNumber(MJ_LCD_MINUTES, min);
  setNumber(MJ_LCD_SECONDS, sec);
  setNumber(MJ_LCD_FRAGMENTS, frag);
}

void MJ_LcdClock::setNumber(int i, int v)
{
  const char *fname = "MJ_LcdClock::setNumber";

  char buf[32];

  if(i >= MJ_LCD_SIZEOF) MJ_fatal(MJ_ERR_ARGUMENT, fname);
  
  if(m_time[i] == v) return;  // No need to update.
  m_time[i] = v;

  int max = 100;
  if(i == MJ_LCD_FRAGMENTS) max = m_fragMax;

  int k, j = v;

  if(j < 1) j = 1;

  j *= 10;
  for(k=0; j < max; k++) {
    j *= 10;
    buf[k] = '0';
  }

  sprintf(&buf[k], "%d", v);
  // puts(buf);
  m_tmp = buf;
  m_timeDisplays[i]->setText(m_tmp);
}

void MJ_LcdClock::setMaxFrag(int frag)
{
  m_fragMax = frag;
  int i, j = 1;

  for(i=0; j < frag; i++) {
    j *= 10;
  }
  if(!i) i = 1;  
  m_fragDigits = i;
}

void MJ_LcdClock::test(QApplication *qa)
{
  MJ_LcdClock lcdc;
  lcdc.set(45,12,2,79);
  qa->setMainWidget(&lcdc);
  lcdc.show();
  qa->exec();
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_RunStatus::MJ_RunStatus(QWidget      *parent,
			   const char   *name)
  : QWidget(parent, name),
    m_time(new MJ_LcdClock(this)),
    m_mainModule(0),
    m_mainUi(0),
    m_thread(0)
{
  m_time->setSizePolicy(QSizePolicy(QSizePolicy::Minimum, 
				    QSizePolicy::Minimum));
  QToolTip::add(m_time, MJ_Tr::tr("Processing time"));
  
  // QPixmap px;
  QIconSet px;

  if(!parent) {
    QString str(MJ_Tr::tr("Mustajuuri play controls"));
    setCaption(str);
    setIconText(str);
  }

  QGridLayout *layout = new QGridLayout(this, 3, 5, 2);
  MJ_equalStretch(layout, 10, 10);
  layout->setColStretch(0, 90);

  m_panic = new QPushButton(MJ_Tr::tr("Panic"), this, "panic_b");
  QToolTip::add(m_panic, MJ_Tr::tr("Shut down immediately(!)"));
  m_panic->setAccel(CTRL + Key_P);
  connect(m_panic, SIGNAL(pressed()), this, SLOT(panic()));

  // px = MJ_PixmapLoader::getPixmap("stop");

  // px = MJ_PixmapLoader::getIconSet("stop", backgroundColor());
  px = MJ_PixmapLoader::getPixmap("stop");

  if(px.pixmap(QIconSet::Small, QIconSet::Normal).isNull())
    m_stop = new QPushButton(MJ_Tr::tr("Stop"), this, "stop_b");
  else {
    QToolButton * tb = new QToolButton(this, "stop_b");
    tb->setIconSet(px);
    tb->setTextLabel(MJ_Tr::tr("Stop"), true);
    m_stop = tb;
    // tb->setMinimumHeight(30);
  }

  connect(m_stop, SIGNAL(pressed()), this, SLOT(stop()));
  m_stop->setAccel(Key_Escape);

  px = MJ_PixmapLoader::getPixmap("play");

  if(px.pixmap(QIconSet::Small, QIconSet::Normal).isNull())
    m_restart = new QPushButton(MJ_Tr::tr("Restart"), this, "restart_b");
  else {
    QToolButton * tb = new QToolButton(this, "restart_b");
    tb->setIconSet(px);
    tb->setTextLabel(MJ_Tr::tr("Restart"), true);
    m_restart = tb;
  }

  connect(m_restart, SIGNAL(pressed()), this, SLOT(restart()));

  layout->addWidget(m_panic, 0, 0);
  layout->addWidget(m_stop, 0, 1);
  layout->addWidget(m_restart, 0, 2);

  layout->addMultiCellWidget(m_time, 1, 1, 0, 2);

  connect(&m_timer, SIGNAL(timeout()), this, SLOT(update()));

  layout->addColSpacing(2, 15);
  layout->setColStretch(3, 1); // Minimal stretch

  // Load display
  m_load = new MJ_LevelDisplay(this, "cpuload");
  m_load->setMinimumSize(30, 30);
  m_load->defaultColors(60, QColor(204, 148, 79), QColor(204, 61, 63));
  QToolTip::add(m_load, MJ_Tr::tr("CPU load"));

  m_loadText = new QLabel(this, "cpuload-label");
  m_loadText->setSizePolicy(QSizePolicy(QSizePolicy::Minimum, 
					QSizePolicy::Minimum));
  m_loadText->setText("CPU\nload\n0.0%%");
  
  QVBoxLayout * loadlayout = new QVBoxLayout(layout, 0);
  
  loadlayout->addWidget(m_loadText, 1, AlignHCenter);
  loadlayout->addWidget(m_load, 100);
  
  layout->addMultiCellLayout(loadlayout, 0, 2, 4, 4);

  // layout->addColSpacing(4, 15);
  // layout->setColStretch(4, 1); // Minimal stretch
  layout->setRowStretch(2, 80); // Maximal stretch

  // Buffer display
  m_bufArea = new MJ_GraphicsArea(this, "bufview");
  QToolTip::add(m_bufArea, MJ_Tr::tr("Output audio buffer fill timer"));
  m_bufArea->setSizePolicy(QSizePolicy(QSizePolicy::Expanding, 
				       QSizePolicy::Expanding));
  connect(m_bufArea, SIGNAL(mousePress(QMouseEvent *, float, float)), 
	  this, SLOT(grMousePress(QMouseEvent *, float, float)));
  m_bufArea->setMinimumSize(30, 30);
  m_bufArea->setBackgroundColor(QColor(150, 150, 150));
  layout->addMultiCellWidget(m_bufArea, 2, 2, 0, 2);

  m_bufRange = new MJ_GraphicPolygon();
  m_bufRange->setGraphicsContext
    (new MJ_GraphicsContextPenBrush(QPen(QColor(100, 60, 60), 2),
				    QBrush(QColor(200, 100, 100))));
  m_bufArea->appendGraphic(m_bufRange);
  
  m_bufCurve = new MJ_GraphicPolygon();
  m_bufCurve->setGraphicsContext
    (new MJ_GraphicsContextPenBrush(QPen(QColor(50, 50, 100), 2),
				    QBrush(QColor(150, 150, 250))));
  m_bufArea->appendGraphic(m_bufCurve);
  
  m_underflowWarn = new MJ_GraphicText(QString::null, 0.5, 0.6);
  m_underflowWarn->setGraphicsContext
    (new MJ_GraphicsContextPen(QPen(Qt::red)));
  m_bufArea->appendGraphic(m_underflowWarn);
  // setPaletteBackgroundColor(red);
}

MJ_RunStatus::~MJ_RunStatus()
{
  const char * fname = "MJ_RunStatus::~MJ_RunStatus";

  MJ_trace(fname);

  if(m_thread) stop();
}

bool MJ_RunStatus::startThreaded(MJ_ModuleDspUi *m)
{
  const char * fname = "MJ_RunStatus::startThreaded";

  if(!m) MJ_fatal(MJ_ERR_ARGUMENT, fname);
  if(!MJ_kernel) MJ_fatal(MJ_ERR_INTERNAL, fname);
  
  if(MJ_kernel->status() != MJ_KERNEL_STOP) {
    MJ_error(MJ_ERR_ARGUMENT, fname, "Already playing");
    return false;
  }
  // Should call "started".
  if(!MJ_kernel->playModule(m, MJ_KERNEL_THREADED_PLAY))
    return false;

  m_mainUi = m;

  return true;
}

void MJ_RunStatus::started(MJ_ModuleDspUi *m, MJ_RtThread *t)
{
  const int interval = 1000 / 25; // Update interval in ms.
  
  m_mainModule = m->module();
  m_mainUi = m;
  m_thread = t;
  m_timer.start(interval);

  m_updates = 0;

  m_underflowWarn->setText(QString::null);

  m_stop->setEnabled(true);
  m_restart->setEnabled(false);
}

void MJ_RunStatus::stopped()
{
  m_timer.stop();
  m_thread = 0;

  m_stop->setEnabled(false);
  m_restart->setEnabled(true);
}


void MJ_RunStatus::moduleDestroyed(MJ_ModuleDsp *m)
{
  MJ_trace("MJ_RunStatus::moduleDestroyed", m);
  

  if(m == m_mainModule) {
    m_mainModule = 0;
    m_mainUi = 0;
    m_restart->setEnabled(false);
  }
}

void MJ_RunStatus::stop()
{  
  const char * fname = "MJ_RunStatus::stop";

  MJ_trace(fname);

  // Should call "stopped".
  MJ_kernel->stopPlaying();
}

void MJ_RunStatus::panic()
{
  MJ_kernel->panic();
}

void MJ_RunStatus::restart()
{
  const char * fname = "MJ_RunStatus::restart";

  MJ_trace(fname);

  if(m_mainUi) {
    if(!m_mainUi->startThreaded())
      MJ_error(MJ_ERR_UNKNOWN, fname, "m_mainUi->startThreaded() failed");
  }
  else MJ_error(MJ_ERR_INTERNAL, fname, "NULL GUI");
}

void MJ_RunStatus::update()
{
  const char * fname = "MJ_RunStatus::update";

  MJ_GET_RT_PRIO();
  m_thread->moduleAccesMutex().lock();
  
  MJ_TimeStamp *timer = m_mainModule->dspGetTimer();

  ulong seconds = timer->secondsUl();
  double subSeconds = timer->subSeconds();
  float load = m_thread->load();
  
  m_thread->moduleAccesMutex().unlock();
  MJ_DROP_RT_PRIO();

  // MJ_trace(fname, (int) seconds);

  int hrs = (int) (seconds / 3600);
  int minutes = ((int) (seconds / 60)) % 60;
  int dispSeconds = (int) (seconds % 60);
  int fragments = (int) (subSeconds * 100.0);
  if(fragments >= 100) fragments = 99;

  m_time->set(hrs, minutes, dispSeconds, fragments);

  m_load->setLevel(load);
  
  m_loadText->setText(QString().sprintf("CPU\nload\n%2.0f%%", load * 100.0f));

  if(!(m_updates % 25)) updateBuffer();

  // Final checks:
  if(!timer->isRealTime()) {
    if(!MJ_kernel->stopPlaying())
      MJ_error(MJ_ERR_UNKNOWN, fname, "MJ_kernel->stopPlaying() failed");
    stopped();
  }
  m_updates++;
}


void MJ_RunStatus::updateBuffer()
{
  uint i;

  MJ_cpf(m_thread, "MJ_RunStatus::updateBuffer");

  MJ_GET_RT_PRIO();
  m_thread->moduleAccesMutex().lock();

  MJ_Array<uint> bufHistogram(m_thread->bufferFills());
  m_thread->clearBufferFills();
  ulong uflows = m_thread->underflows();

  MJ_AudioBufParams abp(m_thread->currentAbp());
  
  m_thread->moduleAccesMutex().unlock();
  MJ_DROP_RT_PRIO();

  // Buffer range:
  MJ_PointArray2D parray;
  parray.resize(4);

  float val = abp.lowestBufMs() / abp.maxBuffer() * 0.5;
  parray.setPoint(0, val, 0.0);
  parray.setPoint(1, val, 1.5);
  val = abp.highestBufMs() / abp.maxBuffer() * 0.5;
  parray.setPoint(2, val, 1.5);
  parray.setPoint(3, val, 0.0);

  m_bufRange->setGraph(parray.qarray().copy());

  // Buffer histogram:
  MJ_Vector<float> hist(bufHistogram.size());

  for(i=0; i < bufHistogram.size(); i++) {
    hist.getFast(i) = bufHistogram.getFast(i);
  }

  // hist.mulThis(1.0 / hist.cumSum());
  hist.mulThis(0.8 / hist.max());

  parray.resize(hist.size() + 2);
  parray.setPoint(0, 0.0, 0.0);
  
  uint s = hist.size();

  for(i=0; i < s; i++) {
    parray.setPoint(i + 1, i / (float) (s-1), hist.getFast(i));
  }
  parray.setPoint(s + 1, 1.0, 0.0);
  
  m_bufCurve->setGraph(parray.qarray());
  
  QString s1(MJ_Tr::tr("Underflows"));
  QString s2;
  s2.sprintf(" %lu", uflows);
  s1 += s2;
  
  if(uflows) m_underflowWarn->setText(s1);

  m_bufArea->redraw(true);
}

void MJ_RunStatus::grMousePress(QMouseEvent *e, float, float)
{
  if(e->stateAfter() && m_thread) {
    MJ_GET_RT_PRIO();
    m_thread->moduleAccesMutex().lock();
    m_thread->clearAbp();
    m_thread->moduleAccesMutex().unlock();
    MJ_DROP_RT_PRIO();
  }
}
