/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_satellite.h>

#include <mj_error.h>
#include <mj_message_carrier.h>
#include <mj_module_dsp.h>

#include <typeinfo>

MJ_Satellite::MJ_Satellite(MJ_Base *h)
  : m_maxUndos(15),
    m_host(h),
    m_guiCount(0)
{
  m_undos.setAutoDelete(true);
  m_redos.setAutoDelete(true);
}

/** Delete the satellite and free all resources. 

    First all undo- and redo objects are deleted.

    Finally this method emits "destroyed" signal.

    If the number of UI objects is non-zero this destructor issues a
    warning.  */
MJ_Satellite::~MJ_Satellite()
{
  if(m_guiCount)
    MJ_error2(MJ_ERR_INTERNAL, 
	      "MJ_Satellite::~MJ_Satellite # %d GUIs still left to \"%s\"", 
	      m_guiCount, typeid(*m_host).name());

  limitUndos(0);
  limitRedos(0);
  emit destroyed(this);
}

/** Add an undo item. Once this method is called the caller must not
    delete the undo packet. The responsibiloty is transferred
    MJ_Satellite.
    
    If the number of undo items exceeds m_maxUndos the oldest undo
    packet is deleted. */

void MJ_Satellite::addUndoInfo(MJ_UndoInfo *i)
{
  if(!i) MJ_fatal(MJ_ERR_ARGUMENT, "MJ_Satellite::addUndoInfo");

  if(!m_maxUndos) {
    delete i;
    return;
  }

  i->setHost(this);
  m_undos.append(i);
  emit emitUndoAdded(i);

  limitUndos(m_maxUndos);
}

/** Add a redo item. If the number of redo items exceeds m_maxUndos
    the oldest redo packet is deleted. */

void MJ_Satellite::addRedoInfo(MJ_UndoInfo *i)
{
  if(!i) MJ_fatal(MJ_ERR_ARGUMENT, "MJ_Satellite::addRedoInfo");

  if(!m_maxUndos) {
    delete i;
    return;
  }

  i->setHost(this);
  m_redos.append(i);
  emit emitRedoAdded(i);

  limitRedos(m_maxUndos);
}

/// Undo an operation. 
bool MJ_Satellite::undo(MJ_UndoInfo *info)
{
  if(!m_host) return false;

  MJ_ModuleDsp *md = dynamic_cast<MJ_ModuleDsp *> (m_host);

  if(!md || !info)
    MJ_fatal(MJ_ERR_INTERNAL, "MJ_Satellite::undo");
  
  for(uint i=0; i < info->nofUndoMsgs(); i++) {
    
    MJ_MessageCarrier *mi = info->counterMsgIndex(i);
    
    bool ok = md->msgProcess(mi->id(), mi->message());
    if(!ok) return false;

    emit modified(mi->id());
  }
  undoToRedo(info);
  return true;
}

/// Redo an operation. 
bool MJ_Satellite::redo(MJ_UndoInfo *info)
{
  if(!m_host) return false;

  MJ_ModuleDsp *md = dynamic_cast<MJ_ModuleDsp *> (m_host);

  if(!md || !info)
    MJ_fatal(MJ_ERR_INTERNAL, "MJ_Satellite::redo");
  
  for(uint i=0; i < info->nofUndoMsgs(); i++) {
    
    MJ_MessageCarrier *mi = info->msgIndex(i);
    
    bool ok = md->msgProcess(mi->id(), mi->message());
    if(!ok) return false;
    
    emit modified(mi->id());
  }
  redoToUndo(info);

  return true;
}

/// Tell that a new UI has been created for the object.
void MJ_Satellite::incrementUiCount()
{
  m_guiCount++;
}

/// Tell that a UI has been deleted.
void MJ_Satellite::decrementUiCount()
{
  if(--m_guiCount < 0) {
    MJ_error(MJ_ERR_INTERNAL, 
	     "MJ_Satellite::decrementUiCount # GUI count below zero", 
	     typeid(*m_host).name());
  }
}

/** This method adds the file to the MRU list. The MRU (Most Recently
    Used) list holds all pasts filenames. Based on this list it is
    easy to build stuff like autosaving etc. 

    The new file name is prepended to the list. Other instances of the
    same string are removeds from the list (=each filename appears
    only once). If the list is changed MJ_Satellite::mruChange signal
    is emitted. 
    
    Quite often the user pushes same file name repeatedly to the
    satellite. In these the list does not change and no signal is
    emitted. */

void MJ_Satellite::pushMruFile(const QString &f)
{
  int currInd = m_mruFiles.findIndex(f);

  if(currInd == 0) return;
  if(currInd > 0)
    m_mruFiles.remove(f);

  m_mruFiles.prepend(f);
  emit mruChange();
  
  MJ_trace("MJ_Satellite::pushMruFile", f);
}

/// Called just prior the object is deleted.
void MJ_Satellite::nextDelete()
{
  emit emitNextDelete();
}

/// Call this method to indicate the host object was modified
void MJ_Satellite::modified(const char *id)
{
  emit emitModified(id);
}

void MJ_Satellite::ioModified(uint flags)
{
  emit emitIoModified(flags);
}

/// Delete extra undo info packets starting from the oldest
void MJ_Satellite::limitUndos(uint l)
{
  while(m_undos.count() > l) {
    MJ_UndoInfo *tmp = m_undos.take(0);
    emit emitUndoRemoved(tmp);
    delete tmp;
  }
}


/// Delete extra redo info packets starting from the oldest
void MJ_Satellite::limitRedos(uint l)
{
  while(m_redos.count() > l) {
    MJ_UndoInfo *tmp = m_redos.take(0);
    emit emitRedoRemoved(tmp);
    delete tmp;
  }
}

/// Transfer a particular undo packet to a redo packet.
void MJ_Satellite::undoToRedo(MJ_UndoInfo *i)
{
  int index = m_undos.find(i);

  if(index == -1) return;

  emit emitUndoRemoved(i);
  m_undos.take(index);
  
  m_redos.append(i);
  emit emitRedoAdded(i);
}

/// Transfer a particular redo packet to a undo packet.
void MJ_Satellite::redoToUndo(MJ_UndoInfo *i)
{
  int index = m_redos.find(i);

  if(index == -1) return;

  emit emitRedoRemoved(i);
  m_redos.take(index);
  
  m_undos.append(i);
  emit emitUndoAdded(i);
}
