/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */
#include <mj_error.h>
#include <mj_satellite.h>
#include <mj_undo_center.h>
#include <mj_tr.h>

#include <qkeycode.h>
#include <qpopupmenu.h>


int MJ_UndoCenter::m_debug = 0;

MJ_UndoCenter::MJ_UndoCenter()
  : m_undoMenu(0),
    m_redoMenu(0),
    m_edit(0),
    m_undoId(-1),
    m_redoId(-1)
{
  m_undoRemaps.setAutoDelete(true);
  m_redoRemaps.setAutoDelete(true);
}

MJ_UndoCenter::~MJ_UndoCenter()
{}

void MJ_UndoCenter::setMenu(QPopupMenu *editMenu)
{
  m_edit = editMenu;

  m_undoMenu = new QPopupMenu(editMenu);
  m_undoId   = m_edit->insertItem(MJ_Tr::tr("Undo"), m_undoMenu);
  m_edit->setItemEnabled(m_undoId, false);
  
  m_undoLatestId = 
    m_edit->insertItem(MJ_Tr::tr("Undo latest"), 
		       this, SLOT(undoLatest()), CTRL+Key_Z);
  m_edit->setItemEnabled(m_undoLatestId, false);

  m_redoMenu = new QPopupMenu(editMenu);
  m_redoId   = m_edit->insertItem(MJ_Tr::tr("Redo"), m_redoMenu);
  m_edit->setItemEnabled(m_redoId, false);

  m_redoLatestId = 
    m_edit->insertItem(MJ_Tr::tr("Redo latest"), 
		       this, SLOT(redoLatest()), CTRL+Key_U);
  m_edit->setItemEnabled(m_redoLatestId, false);
}

void MJ_UndoCenter::addSatellite(MJ_Satellite *s)
{
  uint i, count;

  count = s->nofUndos();

  for(i = 0; i < count; i++) {
    addUndo(s->undoInfo(i));
  }

  count = s->nofRedos();

  for(i = 0; i < count; i++) {
    addRedo(s->redoInfo(i));
  }
  
  connect(s, SIGNAL(emitUndoAdded(MJ_UndoInfo *)),
	  this, SLOT(addUndo(MJ_UndoInfo *)));
  connect(s, SIGNAL(emitUndoRemoved(MJ_UndoInfo *)),
	  this, SLOT(removeUndo(MJ_UndoInfo *)));

  connect(s, SIGNAL(emitRedoAdded(MJ_UndoInfo *)),
	  this, SLOT(addRedo(MJ_UndoInfo *)));
  connect(s, SIGNAL(emitRedoRemoved(MJ_UndoInfo *)),
	  this, SLOT(removeRedo(MJ_UndoInfo *)));
}

void MJ_UndoCenter::addUndo(MJ_UndoInfo *i)
{
  m_edit->setItemEnabled(m_undoId, true);
  m_edit->setItemEnabled(m_undoLatestId, true);
  
  addAction(m_undos, m_undoRemaps, m_undoMenu, SLOT(undo(int)), i);
}

void MJ_UndoCenter::removeUndo(MJ_UndoInfo *i)
{
  int ind = m_undos.find(i);
  if(ind < 0) return;
  m_undos.remove(ind);

  rebuildMenu(m_undos, m_undoRemaps, m_undoMenu, SLOT(undo(int)));
}

void MJ_UndoCenter::addRedo(MJ_UndoInfo *i)
{
  if(m_debug) trace("MJ_UndoCenter::addRedo");
  m_edit->setItemEnabled(m_redoId, true);
  m_edit->setItemEnabled(m_redoLatestId, true);
  addAction(m_redos, m_redoRemaps, m_redoMenu, SLOT(redo(int)), i);
}

void MJ_UndoCenter::removeRedo(MJ_UndoInfo *i)
{
  int ind = m_redos.find(i);
  if(ind < 0) return;
  m_redos.remove(ind);

  rebuildMenu(m_redos, m_redoRemaps, m_redoMenu, SLOT(redo(int)));
}

void MJ_UndoCenter::undo(int i)
{
  static const char *fname = "MJ_UndoCenter::undo";
  if(m_debug) trace(fname, i);

  if((uint) i >= m_undos.count()) {
    error(MJ_ERR_ARGUMENT, fname);
    return;
  }
  if(!m_undos.at(i)->undo())
    error(MJ_ERR_UNKNOWN, fname);
}

void MJ_UndoCenter::redo(int i)
{
  static const char *fname = "MJ_UndoCenter::redo";
  if(m_debug) trace(fname, i);

  if((uint) i >= m_redos.count()) {
    error(MJ_ERR_ARGUMENT, fname);
    return;
  }
  if(!m_redos.at(i)->redo())
    error(MJ_ERR_UNKNOWN, fname);
}

void MJ_UndoCenter::undoLatest()
{
  if(m_undos.count()) {
    undo(m_undos.count() - 1);
  }
}

void MJ_UndoCenter::redoLatest()
{
  if(m_redos.count()) {
    redo(m_redos.count() - 1);
  }
}

void MJ_UndoCenter::addAction(MJ_UndoList &doList, 
			      MJ_List<MJ_IntRemap> &remaps,
			      QPopupMenu *menu,
			      const char *slotName,
			      MJ_UndoInfo *i)
{
  if(m_debug) trace("MJ_UndoCenter::addAction");

  if(doList.find(i) >= 0) return;

  doList.append(i);

  QString text(i->describe());

  MJ_IntRemap *m = new MJ_IntRemap(remaps.count());
  if(!m) return;
  remaps.append(m);

  menu->insertItem(text, m, SLOT(call()), 0, -1, 0);

  connect(m, SIGNAL(emitCall(int)),
	  this, slotName);
}

void MJ_UndoCenter::rebuildMenu(MJ_UndoList &doList, 
				MJ_List<MJ_IntRemap> &remaps,
				QPopupMenu *menu,
				const char *slotName)
{
  remaps.clear();

  int id, idLatest;
  
  if(menu == m_undoMenu) {
    id = m_undoId;
    idLatest = m_undoLatestId;
  }
  else {
    id = m_redoId;
    idLatest = m_redoLatestId;
  }

  while(menu->count())
    menu->removeItemAt(0);

  QListIterator<MJ_UndoInfo> iter(doList);
  MJ_UndoInfo *i;
  
  while((i = iter.current())) {
    QString text;
    text = i->describe();
    MJ_IntRemap *m = new MJ_IntRemap;
    
    if(!m) return;
    remaps.append(m);
    
    menu->insertItem(text, m, SLOT(call()));
    
    m->setId(remaps.count() - 1);
    
    connect(m, SIGNAL(emitCall(int)), this, slotName);
    ++iter;
  }
  bool enable = menu->count() != 0;
  m_edit->setItemEnabled(id, enable);
  m_edit->setItemEnabled(idLatest, enable);
}
