/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_undo_info.h>

#include <mj_message_carrier.h>
#include <mj_message.h>
#include <mj_error.h>
#include <mj_satellite.h>
#include <mj_tr.h>

#include <qtextstream.h> 

const char *MJ_UndoInfo::m_label = "MJ_UNDO_INFO";

unsigned long MJ_UndoInfo::m_units = 0;

MJ_UndoInfo::MJ_UndoInfo()
  : m_firstCounterMsg(0),
    m_host(0),
    m_unit(MJ_UNIT_BLANK),
    m_count(m_units++)
{
  m_msgs.setAutoDelete(true);
  if(m_units == (unsigned long) -1) {
    MJ_error(MJ_ERR_NUMERIC_OVERFLOW, "MJ_UndoInfo::MJ_UndoInfo");
  }
}

MJ_UndoInfo::~MJ_UndoInfo()
{
}

void MJ_UndoInfo::set(MJ_MessageCarrier *msg, MJ_MessageCarrier *counterMsg)
{
  m_msgs.clear();

  m_msgs.append(msg->duplicate());

  if(counterMsg)
    m_msgs.append(counterMsg->duplicate());

  m_firstCounterMsg = 1;
}


void MJ_UndoInfo::addMsg(MJ_MessageCarrier *msg, bool copy)
{  
  if(copy) msg = msg->duplicate();
  m_msgs.insert(m_firstCounterMsg++, msg);
}

void MJ_UndoInfo::addCounterMsg(MJ_MessageCarrier *msg, bool copy)
{
  if(copy) msg = msg->duplicate();
  m_msgs.append(msg);
}


void MJ_UndoInfo::clear()
{
  m_msgs.clear();
  m_firstCounterMsg = 0;
}

/** Returns pointer to the first message. If there are no messages,
    this methods returns NULL. */
MJ_MessageCarrier *MJ_UndoInfo::msg() 
{ 
  return m_msgs.count() ? m_msgs.first() : 0;
}  

/** Returns pointer to the nth message. If there are no messages,
    this methods returns NULL. */
MJ_MessageCarrier *MJ_UndoInfo::msgIndex(uint i)
{
  if(i < nofRedoMsgs()) 
    return m_msgs.at(i);
  
  MJ_error(MJ_ERR_ARGUMENT, "MJ_UndoInfo::msgIndex", i);
  return 0;
}

/** Returns pointer to the first countermessage. If there are no messages,
    this methods returns NULL. */
MJ_MessageCarrier *MJ_UndoInfo::counterMsg()
{ 
  return (m_firstCounterMsg <= m_msgs.count()) ? 
    m_msgs.at(m_firstCounterMsg) : 0;
}  

MJ_MessageCarrier *MJ_UndoInfo::counterMsgIndex(uint i)
{
  if(i < nofUndoMsgs()) 
    return m_msgs.at(i + m_firstCounterMsg);
  
  MJ_error(MJ_ERR_ARGUMENT, "MJ_UndoInfo::counterMsgIndex", i);
  return 0;
}

QString MJ_UndoInfo::describe()
{
  // MJ_trace("MJ_UndoInfo::describe");
  if(nofUndoMsgs() <= 1 && nofRedoMsgs() <= 1)
    return describeSimple();

  return describeMulti();
}

bool MJ_UndoInfo::undo()
{
  if(!m_host) return false;

  MJ_Satellite *satellite = dynamic_cast<MJ_Satellite *> (m_host);

  if(satellite) {
    return satellite->undo(this);
  }
  else {
    MJ_error(MJ_ERR_INTERNAL, "MJ_UndoInfo::undo");
    return false;
  }
}

bool MJ_UndoInfo::redo()
{
  if(!m_host) return false;

  MJ_Satellite *satellite = dynamic_cast<MJ_Satellite *> (m_host);

  if(satellite) {
    return satellite->redo(this);
  }
  else {
    MJ_error(MJ_ERR_INTERNAL, "MJ_UndoInfo::redo");
    return false;
  }
}

bool MJ_UndoInfo::write(QTextStream *s)
{
  bool ok = true;
  
  (*s) << m_label << " " <<  m_msgs.count() 
       << " " << m_firstCounterMsg << "\n";
  
  for(uint i=0; i < m_msgs.count(); i++)
    ok = ok && m_msgs.at(i)->write(s);

  return ok;
}

bool MJ_UndoInfo::read(QTextStream *s)
{
  QString str;
  int msgs, firstCounter;
  bool ok = true;

  clear();

  (*s) >> str;
  if(str != m_label) return false;

  (*s) >> str;
  msgs = str.toInt(&ok);

  str = s->readLine();
  firstCounter = str.toInt(&ok);
  
  if(!ok) return false;

  for(int i=0; i < msgs; i++) {
    MJ_MessageCarrier *mi = new MJ_MessageCarrier;

    ok = mi->read(s);
    if(!ok) {
      delete mi;
      return false;
    }
    m_msgs.append(mi);
  }

  m_firstCounterMsg = firstCounter;

  return true;
}

QString MJ_UndoInfo::describeSimple()
{
  QString s;

  if(m_paramName.length() > 1)
    s = m_paramName;
  else
    s = MJ_Tr::tr("Parameter change");

  if(m_unit != MJ_UNIT_BLANK) {
    s += " (" + QString(MJ_Tr::tr(MJ_UnitKeysShort[m_unit])) + ")";
  }
  
  s += ":";

  MJ_MessageCarrier *m = msg();
  MJ_MessageCarrier *mc = counterMsg();
  

  if(m && mc) {
    
    QString oldVal = mc->message()->describe().c_str();
    QString newVal = m->message()->describe().c_str();

    s += " " + oldVal + " -> " + newVal;
  }
  return s;

}

QString MJ_UndoInfo::describeMulti()
{
  QString s;

  if(m_paramName.length() > 1)
    s = m_paramName;
  else
    s = MJ_Tr::tr("Multiparameter change");

  if(m_unit != MJ_UNIT_BLANK) {
    s += " (" + QString(MJ_Tr::tr(MJ_UnitKeysShort[m_unit])) + ")";
  }
  
  return s;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

int MJ_UndoList::compareItems(void *v1, void *v2)
{
  MJ_UndoInfo *i1 = (MJ_UndoInfo *) v1;
  MJ_UndoInfo *i2 = (MJ_UndoInfo *) v2;

  if(i1->count() == i2->count()) return 0;
  if(i1->count() > i2->count()) return 1;
  return -1;
}
