/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_message.h>

#include <mj_error.h>
#include <mj_message_factory.h>
#include <mj_mutex.h>

#include <di_data_buffer.h>

/* #include <qbuffer.h> 
   #include <qtextstream.h> */
 
#include <typeinfo>

#include <stdio.h>
#include <stdlib.h>

const char *MJ_messageTypes[MJ_MSG_SIZEOF] = {
  "MJ_MSG_EMPTY",
  "MJ_MSG_BOOLEAN",
  // "MJ_MSG_FLOAT",
  "MJ_MSG_DOUBLE",
  "MJ_MSG_INT",
  // "MJ_MSG_LONG",
  "MJ_MSG_STRING",
  "MJ_MSG_STL_STRING",
  "MJ_MSG_MIDI_EVENT",
  "MJ_MSG_MULTI"
};
const char *MJ_messageTypesShort[MJ_MSG_SIZEOF] = {
  "empty",
  "boolean",
  // "float",
  "double",
  "int",
  // "long",
  "Qt-string",
  "STL-string",
  "MIDI",
  "multi"
};

static MJ_MutexAuto MJ_msgCountMutex(false, false, false);

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

int MJ_Message::m_counter = 0;

bool MJ_Message::m_debug = false;

/// Returns the type of this message
MJ_MessageType MJ_Message::messageType() const
{
  return MJ_MSG_EMPTY;
}

MJ_Message *MJ_Message::duplicate()
{
  return new MJ_Message(*this);
}

// float MJ_Message::toFloat(bool *ok)
// {
//   if(ok) *ok = false;
//   return 0.0;
// }

double MJ_Message::toDouble(bool *ok)
{
  if(ok) *ok = false;
  return 0.0;
}

int MJ_Message::toInt(bool *ok)
{
  if(ok) *ok = false;
  return 0;
}

bool MJ_Message::toBool(bool *ok)
{
  if(ok) *ok = false;
  return false;
}

// long   MJ_Message::toLong(bool *ok)
// {
//   if(ok) *ok = false;
//   return 0;
// }

// MJ_MessageFloat *MJ_Message::toMessageFloat()
// {
//   if(messageType() == MJ_MSG_FLOAT) return (MJ_MessageFloat *) this;
//   return 0;
// }

MJ_MessageDouble *MJ_Message::toMessageDouble()
{
  if(messageType() == MJ_MSG_DOUBLE) return (MJ_MessageDouble *) this;
  return 0;
}

MJ_MessageInt *MJ_Message::toMessageInt()
{
  if(messageType() == MJ_MSG_INT) return (MJ_MessageInt *) this;
  return 0;
}

// MJ_MessageLong *MJ_Message::toMessageLong()
// {
//   if(messageType() == MJ_MSG_LONG) return (MJ_MessageLong *) this;
//   return 0;
// }

MJ_MessageString *MJ_Message::toMessageString()
{
  if(messageType() == MJ_MSG_STRING) return (MJ_MessageString *) this;
  return 0;
}

bool MJ_Message::write2(DI_DataBuffer2 *b)
{
  b->writeInt32(messageType());
  b->writeInt64(m_timeStamp);
  return true;
}

bool MJ_Message::read2(DI_DataBuffer2 *b, int readFields)
{
  if(readFields < 1) {
    MJ_MessageType type = (MJ_MessageType) b->readInt32();
    MJ_MessageType mytype = messageType();
    if(mytype != type) {
      MJ_error2(MJ_ERR_ARGUMENT, "MJ_Message::read # Wrong type %d != %d",
		(int) mytype, (int) type);
      return false;
    }
  }
  
  if(readFields < 2) {
    m_timeStamp = (stamp_t) b->readInt64();
  }
  
  return true;
}

MJ_Message *MJ_Message::readFromBuffer(DI_DataBuffer2 *b)
{
  const char * fname = "MJ_Message::readFromBuffer";

  int t = b->readInt32();

  MJ_Message *m = MJ_MessageFactory::create(t);

  if(!m) {
    MJ_error2(MJ_ERR_ARGUMENT, "%s # Failed to create object of type %d", 
	     fname, t);
    return 0;
  }

  bool ok = m->read2(b, 1);

  if(!ok || b->isOverrun()) {
    MJ_error2(MJ_ERR_ARGUMENT, 
	      "%s # Read failed, object of type %d (%d %d, %d %d)", 
	     fname, t, (int) ok,(int)b->isOverrun(), b->size(), b->position());
    delete m;
    return 0;
  }
  
  return m;
}

std::string MJ_Message::describe()
{
  std::string tmp;
  return tmp;
}


MJ_Message *MJ_Message::createNew()
{
  return new MJ_Message;
}

void MJ_Message::incrCount()
{
  MJ_msgCountMutex.lock();
  int count = ++m_counter;
  MJ_msgCountMutex.unlock();

  if(m_debug) MJ_trace("MJ_Message::incrCount", count);
}

void MJ_Message::decrCount()
{
  const char * fname = "MJ_Message::decrCount";

  MJ_msgCountMutex.lock();
  int count = --m_counter;
  MJ_msgCountMutex.unlock();
  
  if(count < 0) {
    char buf[64];
    sprintf("%s # count = %d", fname, count);
    MJ_fatal(MJ_ERR_INTERNAL, buf);
  }

  if(m_debug) MJ_trace(fname, count);
  
}
