/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_message_factory.h>

#include <mj_error.h>
#include <mj_messages.h>
#include <mj_message_multi.h>
#include <mj_stream_utils.h>

#include <map>

class __MJ_MessageFactory
{
public:
  __MJ_MessageFactory() : m_factory(0), m_reader(0), m_writer(0) {}

  MJ_Message *create() { return (*m_factory)(); }

  void set(MJ_Message *(*factory)(),
	   bool (*reader)(QTextStream *, MJ_Message *, int),
	   bool (*writer)(QTextStream *, MJ_Message *),
	   int id,
	   const std::string & str)
  { m_factory = factory; m_reader = reader; m_writer = writer; m_id = id;
  m_str = str; }

  MJ_Message * (*m_factory)();

  bool (*m_reader)(QTextStream *, MJ_Message *, int);
  bool (*m_writer)(QTextStream *, MJ_Message *);

  int m_id;
  std::string m_str;
};

static std::map<std::string, __MJ_MessageFactory> m_factories;
static std::map<int, __MJ_MessageFactory>         m_factories2;
// typedef std::map<std::string, MJ_Message * (*)()>::iterator iterator;


static __MJ_MessageFactory * __MJ_FindFactory(int id)
{
  std::map<int, __MJ_MessageFactory>::iterator it = 
    m_factories2.find(id);

  if(it == m_factories2.end()) return 0;

  return & (*it).second;
}

static __MJ_MessageFactory * __MJ_FindFactory(const std::string &name)
{
  std::map<std::string, __MJ_MessageFactory>::iterator it = 
    m_factories.find(name);

  if(it == m_factories.end()) return 0;

  return & (*it).second;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** YOU MUST CALL THIS METHOD. Initialize the factory table to contain
    the built-in message types. This method must be called before the
    application starts. It does not hurt to call it multiple times. */
void MJ_MessageFactory::init()
{
  addFactoryFunction(MJ_messageTypes[MJ_MSG_EMPTY],
		     MJ_MSG_EMPTY,
		     MJ_Message::createNew);
  addFactoryFunction(MJ_messageTypes[MJ_MSG_BOOL],
		     MJ_MSG_BOOL,
		     MJ_MessageBool::createNew);
  addFactoryFunction(MJ_messageTypes[MJ_MSG_DOUBLE],
		     MJ_MSG_DOUBLE,
		     MJ_MessageDouble::createNew);
  addFactoryFunction(MJ_messageTypes[MJ_MSG_INT],
		     MJ_MSG_INT,
		     MJ_MessageInt::createNew);
  addFactoryFunction(MJ_messageTypes[MJ_MSG_STRING],
		     MJ_MSG_STRING,
		     MJ_MessageStlString::createNew);
  addFactoryFunction(MJ_messageTypes[MJ_MSG_MIDI_EVENT],
		     MJ_MSG_MIDI_EVENT,
		     MJ_MessageMidiEvent::createNew);
  addFactoryFunction(MJ_messageTypes[MJ_MSG_MULTI],
		     MJ_MSG_MULTI,
		     MJ_MessageMulti::createNew);
}

MJ_Message * MJ_MessageFactory::create(int id)
{
  std::map<int, __MJ_MessageFactory>::iterator it = 
    m_factories2.find(id); 
  
  if(it != m_factories2.end()) 
    return (*it).second.create();
  
  return 0;
}

MJ_Message * MJ_MessageFactory::create(const char *str)
{
  return create(std::string(str));
}

/** Creates a new message of argument type. */
MJ_Message * MJ_MessageFactory::create(const std::string & str)
{
  std::map<std::string, __MJ_MessageFactory>::iterator it = 
    m_factories.find(str); 

  if(it != m_factories.end()) 
    return (*it).second.create();
  
  return 0;
}

int MJ_MessageFactory::determineId(const char *name)
{
  __MJ_MessageFactory *f = __MJ_FindFactory(name);

  return f ? f->m_id : -1;
}

const char * MJ_MessageFactory::determineName(int id)
{
  __MJ_MessageFactory *f = __MJ_FindFactory(id);

  if(!f) return 0;

  return f->m_str.c_str();
}


bool MJ_MessageFactory::read(QTextStream *s, MJ_Message*m, int readFields)
{
  const char * fname = "MJ_MessageFactory::read";

  int id = m->messageType();

  __MJ_MessageFactory *f = __MJ_FindFactory(id);

  if(!f) {
    MJ_error2(MJ_ERR_ARGUMENT, "%s # No factory for message type %d",
	      fname, id);
    return false;
  }
  if(!f->m_reader) {
    MJ_error2(MJ_ERR_ARGUMENT, "%s # No reader for message type %d",
	      fname, id);
    return false;
  }
  
  return (*f->m_reader)(s, m, readFields);
}

bool MJ_MessageFactory::write(QTextStream *s, MJ_Message * m)
{
  __MJ_MessageFactory *f = __MJ_FindFactory(m->messageType());

  if(!f) return false;
  if(!f->m_writer) return false;
  
  return (*f->m_writer)(s, m);
}

void MJ_MessageFactory::addFactoryFunction
(const char *str,
 int id,
 MJ_Message * (*f)(),
 bool (*reader)(QTextStream *,MJ_Message*, int),
 bool (*writer)(QTextStream *,MJ_Message*))
{
  // MJ_trace2("MJ_MessageFactory::addFactoryFunction \"%s\"", str);
  
  std::string tmp(str);

  __MJ_MessageFactory ftmp;
  ftmp.set(f, reader, writer, id, tmp);

  m_factories[tmp] = ftmp; 
  m_factories2[id] = ftmp; 
}

/** Removes a factory function. */
void MJ_MessageFactory::removeFactoryFunction(const std::string & str, int id)
{ 
  {
    std::map<std::string, __MJ_MessageFactory>::iterator it = 
      m_factories.find(str); 
    
    if(it != m_factories.end()) 
      m_factories.erase(it); 
  }
  
  {
    std::map<int, __MJ_MessageFactory>::iterator it = 
      m_factories2.find(id); 
    
    if(it != m_factories2.end()) 
      m_factories2.erase(it); 
  }
}

