/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_message_multi.h>

#include <mj_error.h>
// #include <mj_iostream.h>
#include <mj_message_factory.h>
#include <mj_messages.h>
// #include <mj_misc.h>

#include <di_data_buffer.h>

MJ_MessageMulti::~MJ_MessageMulti()
{
  clearChildren();
}

MJ_MessageType MJ_MessageMulti::messageType() const
{
  return MJ_MSG_MULTI;
}

MJ_Message *MJ_MessageMulti::duplicate()
{
  MJ_MessageMulti *mm = new MJ_MessageMulti;

  *mm = *this;

  return mm;
}

/** Write to a DI_DataBuffer2 object. 

    About the formatting: The usual message header plus:
    <UL>

    <LI> Number of children (int32)
    <LI> Child1
    <LI> Child2
    <LI> ...
    <LI> ChildN

    </UL>

    The Children are stored in the following format:

    <UL>
    
    <LI> Number of consumed bytes (int32)
    <LI> Name of child (string)
    <LI> The child

    </UL>

    The number of consumed bytes includes the sizes of of the name of
    the child and and the child data. It tells the number of bytes
    between the mark and the next message (= it does not include
    itself).

*/
  
bool MJ_MessageMulti::write2(DI_DataBuffer2 *b)
{
  if(!MJ_Message::write2(b)) return false;

  b->writeInt32(m_children.size());

  iterator it = m_children.begin();

  while(it != m_children.end()) {

    b->movePosition(4);
    int pos = b->position();

    b->writeString(& (*it).first);
    (*it).second->write2(b);

    int currPos = b->position();

    b->setPosition(pos - 4);
    b->writeInt32(currPos - pos);
    b->setPosition(currPos);
    
    ++it;
  }

  return true;
}

/** Try to read values from the buffer. If the operation fails all
    child messages are cleared. */

bool MJ_MessageMulti::read2(DI_DataBuffer2 *b, int readFields)
{
  const char * fname = "MJ_MessageMulti::read2";

  clearChildren();

  if(!MJ_Message::read2(b, readFields)) return false;

  uint n = b->readInt32();

  std::string name;

  for(uint i=0; i < n; i++) {

    int consumeBytes = b->readInt32();
    int pos = b->position();

    b->readString(&name);

    int type = b->readInt32();

    MJ_Message * child = MJ_MessageFactory::create(type);
    
    if(!child) {
      MJ_error2(MJ_ERR_ARGUMENT, 
		"%s # Could not create child %d \"%s\" of type %d", 
		fname, (int) i, name.c_str(), type);
      return false;
    }

    bool ok = child->read2(b, 1);

    if(ok)
      addChild(name, child);    
    else {
      MJ_error2(MJ_ERR_ARGUMENT,
		"%s # Failed to read child %u ... skipping", fname, i);
      delete child;
    }
    b->setPosition(pos + consumeBytes);
  }

  return true;
}

/** Get a pointer to a child message. If there is no message, then
    null is returned. */
MJ_Message *MJ_MessageMulti::getChild(const std::string &id)
{
  iterator it = m_children.find(id);

  if(it == m_children.end())
    return 0;
  
  return (*it).second;
}

/** A utility method to get a double from some child message. If there
    is no corresponding child message, then the user-supplied default
    value is used. 

    @arg id Name of the child
    @arg def Default value
*/
double MJ_MessageMulti::getDouble(const std::string &id, double def)
{
  MJ_Message *msg = getChild(id);

  if(!msg) return def;

  bool ok = true;

  double v = msg->toDouble(&ok);

  return ok ? v : def;
}
				
/** Returns an int from a child message. Works like "getDouble". */  
int MJ_MessageMulti::getInt(const std::string &id, int def)
{
  MJ_Message *msg = getChild(id);

  if(!msg) return def;

  bool ok = true;

  int v = msg->toInt(&ok);

  return ok ? v : def;
}

/** Returns a QString from a child message. Works like "getDouble". */  
std::string MJ_MessageMulti::getStlString
(const std::string &id, const char * def)
{
  MJ_Message *msg = getChild(id);

  if(!msg) return def;

  MJ_MessageStlString *ms = dynamic_cast<MJ_MessageStlString *> (msg);
  
  return ms ? ms->string() : std::string(def);
}

/** Adds a new child. The new child must be non-null. */
void MJ_MessageMulti::addChild(const std::string &id, MJ_Message *m)
{
  if(!m)
    MJ_fatal(MJ_ERR_NULL_ARGUMENT, "MJ_MessageMulti::addChild");
  
  iterator it = m_children.find(id);

  if(it == m_children.end()) {
    m_children[id] = m;
  }
  else {
    delete (*it).second;
    (*it).second = m;
  }
}

/** Removes all children. If auto-deletion is enabled, then the child
    messages are deleted. */
void MJ_MessageMulti::clearChildren()
{

  if(m_autoDelete) {
    while(m_children.size()) {
      delete (*m_children.begin()).second;
      m_children.erase(m_children.begin());
    }
  }
  else {
    m_children.clear();
  }
}
  
/// Factory function
MJ_Message *MJ_MessageMulti::createNew()
{
  return new MJ_MessageMulti;
}

MJ_MessageMulti & MJ_MessageMulti::operator = (const MJ_MessageMulti &that)
{
  clearChildren();

  MJ_MessageMulti *tmp = (MJ_MessageMulti *) &that;
  
  iterator it = tmp->begin();

  while(it != tmp->end()) {
    addChild((*it).first, (*it).second->duplicate());
    ++it;
  }

  return *this;
}
