/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_filter_iir2.h>
#include <mj_filter_iir.h>
#include <mj_filter_fir.h>
#include <mj_fpu.h>
#include <mj_ramp.h>
#include <mj_resampler.h>

#include <mj_config.h>

#include <string.h>
#include <stdio.h>
#include <time.h>

/*

Just a stupid check to see that the endianness has been correctly
determined:
 
#ifdef WORDS_BIGENDIAN
# warning "Hey, this is BIG endian"
#else
# error "Hey, this is LITTLE endian"
#endif
*/

typedef float testSample;
// typedef float testSample;

// Check that IIRs do not crash...
void MJ_iirTest()
{
  clock();

  const int order = 15;

  MJ_FilterCoeffsIIR<testSample> coeffs(order);
  MJ_FilterDFII<testSample> filter(order);
  MJ_FilterDFII<testSample> filter2, filter3;

  filter.update(coeffs);
  filter2.update(coeffs);
  filter3.update(&coeffs);

  filter3.setOrder(123);
  filter3.setOrder(100);
  filter3.setOrder(50);
  coeffs.setOrder(15);
  filter3.update(&coeffs);
  filter2.update(&coeffs);

  filter.put(0.4);
  filter.put(0.2);
  filter.put(0.1);

  ulong secs = 3;
  ulong chans = 15;
  ulong rounds = 44100 * secs * chans;
  
  puts("Enter performance test");

  clock_t t1 = clock();

  for(uint i=0; i < rounds; i++) {
    filter2.put(0.2);
  }
  
  float t = (float) (clock()-t1) / (float) CLOCKS_PER_SEC;

  printf
    ("IIR calculation took %.3f seconds time, order = %d, \
chans = %d, secs = %d MFlops = %3.3f\n", 
     t,
     (int) coeffs.getOrder(),
     (int) chans,
     (int) secs,
     (float) (rounds * (ulong) coeffs.getOrder() * 4) / t / 1000000.0);
}

// Check that FIRs do not crash...
void MJ_firTest()
{
  clock();

  const int order = 15;

  MJ_FilterCoeffsFIR<testSample> coeffs(order);
  MJ_FilterFIR<testSample> filter(order);
  MJ_FilterFIR<testSample> filter2, filter3;

  filter.update(coeffs);
  filter2.update(coeffs);
  filter3.update(&coeffs);

  filter3.setOrder(123);
  filter3.setOrder(100);
  filter3.setOrder(50);
  coeffs.setOrder(15);
  filter3.update(&coeffs);
  filter2.update(&coeffs);

  filter.put(0.4);
  filter.put(0.2);
  filter.put(0.1);

  ulong secs = 3;
  ulong chans = 15;
  ulong rounds = 44100 * secs * chans;
  
  puts("Enter performance test");

  clock_t t1 = clock();

  for(uint i=0; i < rounds; i++) {
    filter2.put(0.2);
  }
  
  float t = (float) (clock()-t1) / (float) CLOCKS_PER_SEC;

  printf
    ("FIR calculation took %.3f seconds time, order = %d, \
chans = %d, secs = %d MFlops = %3.3f\n", 
     t,
     (int) coeffs.getOrder(),
     (int) chans,
     (int) secs,
     (float) (rounds * (ulong) coeffs.getOrder() * 2) / t / 1000000.0);
}

void MJ_firDesignTest()
{
  uint order = 33;

  MJ_FilterCoeffsFIR<float> coeffs(order);
  
  // coeffs.designLowPass(MJ_pi * 0.33);
  coeffs.designLowPassNormalized(MJ_pi * 0.33, MJ_WINDOW_HANNING);
  
  printf("b = [");
  for(uint i=0; i < order; i++) {
    printf(" %f ", coeffs.getNumerator(i));
  }
  printf("]\n");
  fflush(0);
}

void MJ_resamplerTest()
{
  MJ_Resampler<float> resample;

  resample.setParameters(0.3, 4, 8);
  
  float phase = 0.0;
  float step = MJ_2pi * 0.11;
  float phase2 = 0.0;
  float step2 = MJ_2pi * 0.021;
  double ratio = 2.3;
  
  uint i;
  uint rounds = 100;

  printf("\nFirst resampler test\n out1 = [");

  for(i=0; i < rounds; i++) {

    while(resample.needNewSample()) {
      phase += step;
      if(phase > MJ_2pi) phase -= MJ_2pi;

      phase2 += step2;
      if(phase2 > MJ_2pi) phase2 -= MJ_2pi;

      float in = cosf(phase);
      float in2 = cosf(phase2);
      // printf("%3.3f ", in);
      resample.put(in + in2);
    }
    
    float out = resample.getResampled(ratio);

    if(i + 60 > rounds) {
      printf("%3.3f ", out);
    }
  }
  puts("]");

  ratio = 0.23;

  printf("\nSecond resampler test\n out2 = [");

  for(i=0; i < rounds; i++) {

    while(resample.needNewSample()) {

      phase += step;
      if(phase > MJ_2pi) phase -= MJ_2pi;

      phase2 += step2;
      if(phase2 > MJ_2pi) phase2 -= MJ_2pi;

      float in = cosf(phase);
      float in2 = cosf(phase2);
      // printf("%3.3f ", in);
      resample.put(in + in2);
    }
    
    float out = resample.getResampled(ratio);

    if(i + 60 > rounds) {
      printf("%3.3f ", out);
      // printf("out =  %3.3f\n", out);
    }
  }
  puts("]");
}

/* Prints the precalculated 3:rd order Lagrangian interpolation
   coefficients to the screen so that the user can verify that they
   are ok. */
static void printCoeffs()
{
  for(int i=0; i < MJ_LAG3_INTCOEFFS; i++) {

    float *ptr = & MJ_lag3IntCoeffs[i * 4];

    printf("Coeffs %d (%3.3f) = [%4.4f %4.4f %4.4f %4.4f]\n",
	   i, i / (float) MJ_LAG3_INTCOEFFS, ptr[0], ptr[1], ptr[2], ptr[3]);
  }
  fflush(0);
}

static void rampTest()
{
  MJ_Ramp<double> ramp1, ramp2;

  ramp1.reset(1.0);
  ramp2 = ramp1;

  uint samples = 10000;
  double target = 12.0;

  ramp1.setTarget(samples, target, MJ_RAMP_LINEAR);
  ramp2.setTarget(samples, target, MJ_RAMP_EXPONENTIAL);

  for(uint i=0; i < samples; i++) {
    ramp1.advance();
    ramp2.advance();
  }

  printf("Ramp values are = %lf %lf\n", ramp1.get(), ramp2.get());
}

static void precisionTest(float mul)
{
  MJ_RingBuffer2<float> buf;

  uint size = 256;

  buf.resize(size);
  buf.setAll(1.0);

  clock_t t1 = clock();

  for(uint i=0; i < 10000000; i++) {
    buf.put(buf.getNewest(size - 2) * mul);
  }
  
  float t = (float) (clock()-t1) / (float) CLOCKS_PER_SEC;

  printf("Feedback calculations took %.3f seconds time (mul = %f)\n", 
	 t, mul);
    
}

int main(int argc, char **argv)
{
  for(int i=1; i < argc; i++) {
    if(strcmp(argv[i], "--print-interp-coeffs") == 0) {
      printCoeffs();
      return 0;
    }
  }

#ifdef MJ_HAVE_FPU_MANAGER
  MJ_Fpu::fast();
#endif

  printf("test sample is %s\n", sizeof(testSample) == 8 ? "double" : "float");

  puts("Running IIR test");
  MJ_iirTest();
  puts("IIR test did not crash - GREAT !!! \n");

  puts("Running FIR test");
  MJ_firTest();
  puts("FIR test did not crash - GREAT !!! \n");

  MJ_FilterIIR2<float> iir;
  char buf[128];

  float sr = 44100;
  float hz = 2340;
  float gainDb = 4.5;
  // float q = 1.0;
  float slope = 1.0;
  
  fprintf(stdout, "params = %f, %f, %f, %f\n", hz, sr, gainDb, slope);
  fflush(0);

  iir.designHighShelve2(hz, sr, gainDb, slope);
  
  iir.writeTo(buf);
  fprintf(stdout, "High shelving:\n%s\n", buf);
  fflush(0);
  
  iir.designLowShelve2(hz, sr, gainDb, slope);
  
  iir.writeTo(buf);
  fprintf(stdout, "Low shelving:\n%s\n", buf);
  fflush(0);
  
  iir.designBandPassB2(hz, sr, 1.0);
  
  iir.writeTo(buf);
  fprintf(stdout, "Band pass:\n%s\n", buf);
  fflush(0);
  
  iir.designLowPass2(hz, sr, 1.0);
  
  iir.writeTo(buf);
  fprintf(stdout, "Low pass:\n%s\n", buf);
  fflush(0);

  MJ_firDesignTest();
  
  // Lets leak some memory:
  // new char[1080];
  MJ_resamplerTest();

  rampTest();
  precisionTest(1.0);
  precisionTest(0.95);
  precisionTest(0.9);
  precisionTest(0.8);
  precisionTest(0.2);

  return 0;
}
