/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_ALL_PASS_OPERATOR_H
#define MJ_ALL_PASS_OPERATOR_H

#include <mj_ringbuffer.h>

/** An all-pass delay operator. This class is meant to be a helper for
    all the people, who need the classical all-pass delay line:

\code
                 -d
            a + z  
    H(z) = ----------.
                  -d
            1 + az
\endcode


   This class implements an operator that should be used in
   conjunction with MJ_RingBufferDelay delay buffer. If you need
   cascaded all-pass filters working on the same delay line, then yuo
   can simply use multiple all-pass operators.

 */

template <class TElem> class MJ_AllPassOperator
{
public:  
  MJ_AllPassOperator(TElem a = 0.5, uint t1 = 1, uint t2 = 3)
    : m_a(a), m_t1(t1), m_t2(t2) {}
  
  /// Sets all the parameters. 
  void set(TElem a, uint t1, uint t2)
  { m_a = a; m_t1 = t1; m_t2 = t2; }

  /** Operate on the delay buffer. */
  inline void operate(MJ_RingBufferDelay<TElem> &buf)
  { 
    TElem &v1 = buf.getNewest(m_t1);
    TElem &v2 = buf.getNewest(m_t2);
    v2 += v1 * m_a;
    v1 -= v2 * m_a;
  }

protected:
  /// Feedback gain, range 0-1.
  TElem m_a;
  /// Delay times.
  uint  m_t1, m_t2;
};

#endif
