/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_filter_coeffs.h>

#include <mj_math.h>

#include <string.h>

template <class TElem>
MJ_FilterCoeffsIIR<TElem>::MJ_FilterCoeffsIIR()
  : m_a(0), m_b(0), m_order(0), m_delete(false) 
{}

/// Creates coefficients of order "order"
template <class TElem>
MJ_FilterCoeffsIIR<TElem>::MJ_FilterCoeffsIIR(uint order)
  : m_a(0), m_b(0), m_order(0), m_delete(false) 
{
  setOrder(order);
}

/// Copies the argument coefficients
template <class TElem>
MJ_FilterCoeffsIIR<TElem>::MJ_FilterCoeffsIIR(const MJ_FilterCoeffsIIR &c)
  : m_a(0), m_b(0), m_order(0), m_delete(false) 
{
  copy(c);
}

/// Links to the argument coefficients
template <class TElem>
MJ_FilterCoeffsIIR<TElem>::MJ_FilterCoeffsIIR(MJ_FilterCoeffsIIR *c)
  : m_a(0), m_b(0), m_order(0), m_delete(false) 
{
  link(*c);
}

template <class TElem>
MJ_FilterCoeffsIIR<TElem>::~MJ_FilterCoeffsIIR()
{
  empty();
}

/** Clears the coefficients. If the coefficients were linked, the
    coefficient buffers are not deleted. */
template <class TElem>
void MJ_FilterCoeffsIIR<TElem>::empty()
{
  if(m_delete) {
    if(m_a) delete [] m_a;
    if(m_b) delete [] m_b;
  }
  m_a = 0;
  m_b = 0;
  m_order = 0;
}

/** Sets the order of the filter. Both "a" and "b" coefficient buffers
    have order+1 elements. If the "reset" argument is "true" then all
    filter coefficients are initialized. 
    
    This function returns false if the memory allocation failed. 

    \b Note: Calls to this method always cause memory reallocation ->
    old filter coefficients are lost.

    \b Note: If the order is zero the filter will still allocate
    memory - both "a" and "b" buffers will have one element. The only
    way to free all memory is to use the "empty"-method. 

    @arg order This argument gives the filter order

    @arg reset This argument tells whether the filter coefficients
    should be reset to default all-pass filter. If you are adjusting
    the parameters yourself anyway there is little point in resetting
    the coefficients here.

    @return This method returns true if memory was successfully
    allocated and false if memory allocation failed.
*/

template <class TElem>
bool MJ_FilterCoeffsIIR<TElem>::setOrder(uint order, bool reset)
{
  empty();

  m_order = order;
  order++;
  
  m_a = new TElem[order];
  m_b = new TElem[order];
  m_delete = true;

  if(!m_a || !m_b) return false;

  if(reset) {
    m_a[0] = m_b[0] = 1.0;
    
    for(uint i=1; i < order; i++) {
      m_a[i] = m_b[i] = 0.0;
    }
  }
  return true;
}

/** Set a numerator ("b") coefficient. If the index greater than the
    order this method does nothign. */

template <class TElem>
void MJ_FilterCoeffsIIR<TElem>::setNumerator(uint index, TElem coeff)
{
  if(index <= m_order) m_b[index] = coeff;
}

/** Set a denominator ("a") coefficient. If the index is zero or
    greater than the order this method does nothign. */
template <class TElem>
void MJ_FilterCoeffsIIR<TElem>::setDenominator(uint index, TElem coeff)
{
  if(index && (index <= m_order)) m_a[index] = coeff;
}

/** Copies the elements from the source. Does memory allocation. 

    @param c Filter coeffiecints to be copied.
*/
template <class TElem>
void MJ_FilterCoeffsIIR<TElem>::copy(const MJ_FilterCoeffsIIR &c)
{
  if(this == &c)
    return;

  if(c.m_order != m_order || !m_delete)
    setOrder(c.m_order, false);

  if(m_order == 0) return;

  uint bytes = (m_order + 1) * sizeof(TElem);

  memcpy(m_a, c.m_a, bytes);
  memcpy(m_b, c.m_b, bytes);
}

/** Links the elements from the source. Does \b no memory allocation,
    but may free previous pointers. 

    @param c Filter coeffiecints to be linked.
*/
template <class TElem>
void MJ_FilterCoeffsIIR<TElem>::link(MJ_FilterCoeffsIIR &c)
{
  empty();

  m_a = c.m_a;
  m_b = c.m_b;
  m_order = c.m_order;
  m_delete = false;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

template <class TElem>
MJ_FilterCoeffsFIR<TElem>::MJ_FilterCoeffsFIR()
  : m_b(0), m_order(0), m_delete(false) 
{}

/// Creates coefficients of order "order"
template <class TElem>
MJ_FilterCoeffsFIR<TElem>::MJ_FilterCoeffsFIR(uint order)
  : m_b(0), m_order(0), m_delete(false) 
{
  setOrder(order);
}

/// Copies the argument coefficients
template <class TElem>
MJ_FilterCoeffsFIR<TElem>::MJ_FilterCoeffsFIR(const MJ_FilterCoeffsFIR &c)
  : m_b(0), m_order(0), m_delete(false) 
{
  copy(c);
}

/// Links to the argument coefficients
template <class TElem>
MJ_FilterCoeffsFIR<TElem>::MJ_FilterCoeffsFIR(MJ_FilterCoeffsFIR *c)
  : m_b(0), m_order(0), m_delete(false) 
{
  link(*c);
}

template <class TElem>
MJ_FilterCoeffsFIR<TElem>::~MJ_FilterCoeffsFIR()
{
  empty();
}

/** Clears the coefficients. If the coefficients were linked, the
    coefficient buffers are not deleted. */
template <class TElem>
void MJ_FilterCoeffsFIR<TElem>::empty()
{
  if(m_delete) {
    delete [] m_b;
  }
  m_b = 0;
  m_order = 0;
}

/** Sets the order of the filter. The coefficient buffer has order+1
    elements. If the "reset" argument is "true" then all filter
    coefficients are initialized.
    
    This function returns false if the memory allocation failed. 

    \b Note: Calls to this method always cause memory reallocation ->
    old filter coefficients are lost.

    \b Note: If the order is zero the filter will still allocate
    memory - the memory buffer will have one element. The only way to
    free all memory is to use the "empty"-method. 

    @param order This argument gives the filter order

    @param reset This argument tells whether the filter coefficients
    should be reset to default all-pass filter. If you are adjusting
    the parameters yourself anyway there is little point in resetting
    the coefficients here.

    @return This method returns true if memory was successfully
    allocated and false if memory allocation failed.
*/

template <class TElem>
bool MJ_FilterCoeffsFIR<TElem>::setOrder(uint order, bool reset)
{
  empty();

  m_order = order;

  if(!order) return true;

  order++;
  
  m_b = new TElem[order];
  m_delete = true;

  if(!m_b) return false;

  if(reset) {
    m_b[0] = 1.0;
    
    for(uint i=1; i < order; i++) {
      m_b[i] = 0.0;
    }
  }
  return true;
}

/** Set a numerator ("b") coefficient. If the index greater than the
    order this method does nothign. */

template <class TElem>
void MJ_FilterCoeffsFIR<TElem>::setNumerator(uint index, TElem coeff)
{
  if(index <= m_order) m_b[index] = coeff;
}

/** Copies the elements from the source. Does memory allocation. */
template <class TElem>
void MJ_FilterCoeffsFIR<TElem>::copy(const MJ_FilterCoeffsFIR &c)
{
  if(this == &c)
    return;

  if(c.m_order != m_order || !m_delete)
    setOrder(c.m_order, false);

  if(m_order == 0)
    return;

  uint bytes = (m_order + 1) * sizeof(TElem);

  memcpy(m_b, c.m_b, bytes);
}

/** Links the elements from the source. Does \b no memory allocation,
    but may free previous pointers. */
template <class TElem>
void MJ_FilterCoeffsFIR<TElem>::link(MJ_FilterCoeffsFIR &c)
{
  empty();

  m_b = c.m_b;
  m_order = c.m_order;
  m_delete = false;
}

/** Design a linear-phase low-pass filter. 

    No safety checks. 
*/

template <class TElem>
void MJ_FilterCoeffsFIR<TElem>::designLowPass(TElem frequency)
{
  uint i;
  uint order = m_order;
  
  /* If the filter order is even, then there are non-even number of
     coefficients. In those cases the center element needs special
     care. */
  int center = (order & 0x1) ? -1 : (order + 1) >> 1;

  TElem *b = m_b;
  TElem tmp;
  TElem tcenter = (TElem) order * 0.5;

  for (i=0; i <= order; i++) {
    tmp = (TElem) i - tcenter;

    if((int) i != center)
      b[i] = MJ_sin(frequency * tmp) / (MJ_pi * tmp);
  }
  
  if(center > 0)
    b[center] = frequency / MJ_pi;
}

/** Apply a particular window on the filter coefficients. 
    
    The filter coefficients should be already initialized.
*/

template <class TElem>
void MJ_FilterCoeffsFIR<TElem>::applyWindow(MJ_WindowFunction window)
{
  if(window == MJ_WINDOW_RECTANGULAR) return;

  uint i, order = m_order;
  TElem *b = m_b;

  if(window == MJ_WINDOW_HANNING) {

    TElem scale = 1.0 / (TElem) (order + 2);

    for (i=0; i <= order; i++) {
      b[i] *= (0.5 * (1.0 - MJ_cos(2.0 * MJ_pi * (i + 1) * scale)));
    }
  }
  else if(window == MJ_WINDOW_HAMMING) {
    
    TElem scale = 1.0 / (TElem) order;
    
    for (i=0; i <= order; i++) {
      b[i] *= ((0.54 - 0.46 * MJ_cos(2.0 * MJ_pi * (TElem) i * scale) + 
		0.08 * MJ_cos(4.0 * MJ_pi * (TElem) i * scale)));
    }
  }
}

template <class TElem>
void MJ_FilterCoeffsFIR<TElem>::normalizeLowPass()
{
  TElem gain = 0.0;

  
  uint order = m_order;
  TElem *b = m_b;
  TElem *sentinel = b + order + 1;
  
  while(b != sentinel)
    gain += *b++;

  gain = (gain > 0.0) ? 1.0 / gain : -1.0 / gain;

  b = m_b;
  while(b != sentinel)
    *b++ *= gain;
}
