/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_FILTER_IIR_H
#define MJ_FILTER_IIR_H

#include <mj_filter_coeffs.h>
#include <mj_math.h>
#include <mj_ringbuffer.h>

/** Direct form II linear IIR filter (nth order).

    This class implements a fairly optimized linear digital IIR
    (Inifinite Impulse Response) filter. 

    @see MJ_FilterCoeffsIIR MJ_FilterIIR2 MJ_FilterAllPole1

    @author Tommi Ilmonen */
template <class TElem> class MJ_FilterDFII
{
public:
  MJ_FilterDFII();
  MJ_FilterDFII(uint order);
  ~MJ_FilterDFII();

  /** Copy coefficients. Automatically resizes memory buffer. */
  void update(MJ_FilterCoeffsIIR<TElem> &coeffs) 
  { m_coeffs.copy(coeffs); m_buffer.resize(coeffs.getOrder() ? 
					   coeffs.getOrder() : 1); }
  /** Link coefficients. Automatically resizes memory buffer. */
  void update(MJ_FilterCoeffsIIR<TElem> *coeffs) 
  { m_coeffs.link(*coeffs); m_buffer.resize(coeffs->getOrder() ? 
					    coeffs->getOrder() : 1); }

  /** Sets the order of the filter. */
  inline void setOrder(uint order)
  { m_coeffs.setOrder(order); m_buffer.resize(order ? order : 1); }

  inline uint getOrder() const { return m_coeffs.getOrder();}

  inline void put(TElem s);
  /// Get the current output value.
  inline const TElem & get() const { return m_val; }

  /// Resets filter buffer memory to zero.
  inline void resetMemory() { m_buffer.setAll(0.0); m_val = 0.0; }

  /// Access the coefficients directly (use discouraged)
  MJ_FilterCoeffsIIR<TElem> & coeffs() { return m_coeffs; }
  /// Access the feedback buffer directly (use discouraged)
  MJ_RingBufferDelay<TElem> & buffer() { return m_buffer; }

protected:
  TElem m_val;
  MJ_FilterCoeffsIIR<TElem> m_coeffs;
  MJ_RingBufferDelay<TElem> m_buffer;
};

/** Put a new input sample to the filter. This method calculates the
    new output value of the filter using the direct form II scheme. */
template <class TElem>
inline void MJ_FilterDFII<TElem>::put(TElem s)
{
  /* This implementation is fairly optimized and may look a bit
     strange...*/

  const TElem *aptr = m_coeffs.aConst() + 1; // a[0] = 1.0 -> ignore
  const TElem *bptr = m_coeffs.bConst() + 1; // Use b[0] in the last line

  uint i, order = m_coeffs.getOrder();
  TElem bsum = 0.0;

  /* These calculations are done prior to adding the new sample the
     the buffer. Thus the apparent mismatch between filter
     coefficients and delayed values. */
  for(i=0; i < order; i++) {
    TElem tmp = m_buffer.getNewest(i);
    s -= *aptr++ * tmp;
    bsum += *bptr++ * tmp;
  }

#ifdef __x86__
  /* Flush underflows manually to zero. This is not necessary if you
     have a decent FPU (Intel x86 systems do not)... */

  if(MJ_abs(s) < 1.0e-8)
    s = 0.0;
#endif  // __x86__

  m_buffer.put(s);
  
  m_val = bsum + s * (*m_coeffs.bConst());
}

#endif
