/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_ramp.h>

#include <stdlib.h>
#include <string.h>

/** Advance the ramp and perform multiply and accumulate (MAC). This
    method works on a number of vector elements at a time.

    target += source * ramp.

*/

template <class TElem>
void MJ_Ramp<TElem>::advanceLinearMac
(TElem *target, const TElem *source, unsigned n)
{
  const TElem *end = source + n;
  
  if(n < m_counter) {
    while(source < end) {
      *target++ += *source++ * m_current;
      m_current += m_step;
    }
    m_counter -= n;
  }
  else if(!m_counter) {
    while(source < end) {
      *target++ += *source++ * m_current;
    }
  }
  else {
    for(const TElem *end2 = source + m_counter; source < end2; ) {
      *target++ += *source++ * m_current;
      m_current += m_step;
    }

    m_current = m_target;
    m_counter = 0;

    while(source < end) {
      *target++ += *source++ * m_current;
    }
  }
}

/** Advance the ramp and perform multiply. This method works on a
    number of vector elements at a time.

    target = source * ramp.

*/

template <class TElem>
void MJ_Ramp<TElem>::advanceLinearMul
(TElem *target, const TElem *source, unsigned n)
{
  const TElem *end = source + n;
  
  if(n < m_counter) {
    while(source < end) {
      *target++ = *source++ * m_current;
      m_current += m_step;
    }
    m_counter -= n;
  }
  else if(!m_counter) {
    while(source < end) {
      *target++ = *source++ * m_current;
    }
  }
  else {
    for(const TElem *end2 = source + m_counter; source < end2; ) {
      *target++ = *source++ * m_current;
      m_current += m_step;
    }

    m_current = m_target;
    m_counter = 0;

    while(source < end) {
      *target++ = *source++ * m_current;
    }
  }
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

template <class TElem>
MJ_MultiRamp<TElem>::~MJ_MultiRamp()
{
  free(m_points);
}

/** Change the number of interpolation points. */
template <class TElem>
bool MJ_MultiRamp<TElem>::resize(unsigned points)
{
  if(points == m_nofPoints) return true;

  if(!points) return false;

  m_points = (MJ_RampPoint<TElem> *) 
    realloc(m_points, points * sizeof(MJ_RampPoint<TElem>));

  m_nofPoints = points;

  m_counter = 0;

  return m_points != 0;
}

/** Advance the interpolator to the next sample. */
template <class TElem>
void MJ_MultiRamp<TElem>::advance()
{ 
  m_ramp.advance(); 
  
  m_counter++;

  if(!m_ramp.isActive() && m_pointsLeft) {
    nextPoint();
  } 
}

/** Adds a new target point. */

template <class TElem>
bool MJ_MultiRamp<TElem>::addTarget(long time, TElem value, MJ_RampType type)
{
  if(m_pointsLeft >= m_nofPoints) return false;
  
  MJ_RampPoint<TElem> *rp = & m_points[m_pointsLeft++];

  rp->m_time  = time;
  rp->m_value = value;
  rp->m_type  = type;

  if(m_pointsLeft == 1 && !m_ramp.isActive())
    nextPoint();

  return true;
}

template <class TElem>
MJ_MultiRamp<TElem> &MJ_MultiRamp<TElem>::operator = (const MJ_MultiRamp &r)
{
  if(resize(r.m_nofPoints)) {

    memcpy(m_points, r.m_points, sizeof(MJ_RampPoint<TElem>) * r.m_nofPoints);

    m_ramp = r.m_ramp;
    m_counter = r.m_counter;
    m_nofPoints = r.m_nofPoints;
    m_pointsLeft = r.m_pointsLeft;
  }
  return *this;
}

/** Advance to the next point. You must be careful with this method --
    it does not do any safety checks. */

template <class TElem>
void MJ_MultiRamp<TElem>::nextPoint()
{
  m_pointsLeft--;
  
  memmove(m_points, 
	  m_points + 1, 
	  (m_nofPoints - 1) * sizeof(MJ_RampPoint<TElem>));
  
  m_ramp.setTarget(m_points->m_time - m_counter, 
		   m_points->m_value,
		   m_points->m_type);  
}
