/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * Tommi.Ilmonen(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_wave_editor.h>

#include <mj_audio_conf.h>
#include <mj_kernel.h>
#include <mj_message_string.h>
#include <mj_module_level.h>
#include <mj_pixmap_loader.h>
#include <mj_satellite.h>
#include <mj_strs.h>
#include <mj_str_remap.h>
#include <mj_tr.h>
#include <mj_wave_operations.h>
#include <mj_wave_player.h>
#include <mj_wave_recorder.h>

#include <qawl_view.h>

#include <qcombobox.h> 
#include <qcursor.h>
#include <qfiledialog.h>
#include <qlayout.h> 
#include <qmenubar.h>
#include <qmessagebox.h> 
#include <qpixmap.h> 
#include <qpopupmenu.h>
#include <qpushbutton.h> 
#include <qtoolbutton.h> 
#include <qtoolbar.h>
#include <qtooltip.h> 
#include <qslider.h>

#include <vector>
#include <map>

#include <math.h>

static QString tmpFile("/tmp/zeb-record.wav");

enum MJE_SampleFormat {
  MJE_SF_INT16,
  MJE_SF_INT24,
  MJE_SF_INT32,
  MJE_SF_FLOAT,
  MJE_SF_SIZEOF
};

class MJ_WaveEditor::Internal
{
  Internal() :
    m_zoom(1000.0),
    m_magnify(1.0),
    m_player(new MJ_WavePlayer),
    m_outputDevice(0),
    m_inputDevice(0),
    m_fileSampleFormat(0), 
    m_recordChannels(2),
    m_state(MJ_WaveEditor::Inactive),
    m_needToSave(false),
    m_tmpIndex(0),
    m_inLevels(0),
    m_outLevels(0)

  {}

  ~Internal()
  {
    delete m_inLevels;
    delete m_outLevels;
    m_player->satellite()->decrementUiCount();
    delete m_player;
  }

  friend class MJ_WaveEditor;

  int newTmpIndex() { return ++m_tmpIndex; }

  double      m_zoom;
  double      m_magnify;

  QAWL_View2      *m_view;
  MJ_WavePlayer   *m_player;

  QPopupMenu *m_fileMenu;
  int m_saveSegmentAsId;

  QPopupMenu *m_listenMenu;
  int m_playId;
  int m_stopId;

  QPopupMenu *m_outputDeviceMenu;
  QPopupMenu *m_inputDeviceMenu;
  int m_outputDevice;
  int m_inputDevice;

  QPopupMenu *m_editMenu;
  int m_cropToSelection;
  
  QPopupMenu *m_fileFormatMenu;
  int m_fileFormats[MJE_SF_SIZEOF];

  std::vector<int> m_pastFiles;
  std::map<QString, QString> m_pastFileNames;

  QWidget *m_tools[ToolCount];

  QPtrList<MJ_StrRemap> m_remaps;


  QTimer m_timer;
  
  int    m_fileSampleFormat;
  int    m_recordChannels;

  MJ_WaveEditor::State m_state;

  // Do we need to save the file
  bool   m_needToSave;

  // The index at which we are running with the temporary files.
  int m_tmpIndex;

  QString m_currentRealFile;

  MJ_ModuleDspGui * m_inLevels;
  MJ_ModuleDspGui * m_outLevels;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


static const char *__MJ_sampleFormats[MJE_SF_SIZEOF] = {
  "16 bit integer",
  "24 bit integer",
  "32 bit integer",
  "32 bit floating point"
};

MJ_WaveEditor::MJ_WaveEditor(QWidget * parent, const char * name)
  : QWidget(parent, name),
    MJ_ModuleDspUi(0, 0),
    m_internal(new Internal)
{
  uint i;

  QPixmap px;

  px = MJ_PixmapLoader::getPixmap("zeb");
  
  if(!px.isNull()) {
    // qDebug("GOT THE MASTER ICON");
    setIcon(px);
  }
    /* else
       qDebug("Failed to load master ICON"); */

  m_module = m_internal->m_player;
  m_internal->m_player->satellite()->incrementUiCount();

  QVBoxLayout *layout = new QVBoxLayout(this);

  // Create menus:
  QMenuBar *menuBar = new QMenuBar(this);
  // QMenuBar *menuBar = menuBar();
  layout->setMenuBar(menuBar);

  // File menu:
  QPopupMenu *fmenu = new QPopupMenu(menuBar);
  menuBar->insertItem("&File", fmenu);

  fmenu->insertItem("&Open...", this, SLOT(open()));
  fmenu->insertItem("Save &as...", this, SLOT(saveAs()));
  m_internal->m_saveSegmentAsId = 
    fmenu->insertItem("Save selection as...", this, SLOT(saveSelectionAs()));
  fmenu->insertItem("&Quit", this, SLOT(quit()), CTRL + Key_Q);
  fmenu->insertSeparator();

  m_internal->m_fileMenu = fmenu;

  // File format menu
  
  QPopupMenu *ffmenu = new QPopupMenu(fmenu);
  fmenu->insertItem("Output file format", ffmenu);

  for(i=0; i < MJE_SF_SIZEOF; i++) {
    m_internal->m_fileFormats[i] = ffmenu->insertItem(__MJ_sampleFormats[i]);
  }

  connect(ffmenu, SIGNAL(activated(int)), 
	  this, SLOT(selectFileSampleFormat(int)));
  m_internal->m_fileFormatMenu = ffmenu;
  
  selectFileSampleFormat(m_internal->m_fileFormats[1]);

  fmenu->insertSeparator();
  
  // Edit menu:
  
  QPopupMenu *emenu = new QPopupMenu(menuBar);
  menuBar->insertItem("Edit", emenu);
  emenu->insertItem("Select All", this, SLOT(selectAll()));
  m_internal->m_cropToSelection = 
    emenu->insertItem("Crop to selection", this, SLOT(cropToSelection()));
  
  m_internal->m_editMenu = emenu;

  // Listen menu:

  QPopupMenu *lmenu = new QPopupMenu(menuBar);
  menuBar->insertItem("Listen", lmenu);
  m_internal->m_playId = lmenu->insertItem("Play", this, SLOT(togglePlay()));
  m_internal->m_stopId = lmenu->insertItem("Stop", this, SLOT(togglePlay()));
  
  m_internal->m_listenMenu = lmenu;

  // Audio output device menu:

  QPopupMenu *dmenu = new QPopupMenu(lmenu);
  lmenu->insertItem("Output device", dmenu);
  
  uint devCount = MJ_AudioConf::audioDevices(MJ_ADT_OUTPUT);

  for(i=0; i < devCount; i++) {
    QString device(MJ_AudioConf::audioDevice(MJ_ADT_OUTPUT, i));
    dmenu->insertItem(device, i);
    dmenu->setItemChecked(i, i == (uint) m_internal->m_outputDevice);
  }
  connect(dmenu, SIGNAL(activated(int)), this, SLOT(selectOutput(int)));
  m_internal->m_outputDeviceMenu = dmenu;
  
  // Audio input device menu:

  dmenu = new QPopupMenu(lmenu);
  lmenu->insertItem("Input device", dmenu);
  
  devCount = MJ_AudioConf::audioDevices(MJ_ADT_INPUT);

  for(i=0; i < devCount; i++) {
    QString device(MJ_AudioConf::audioDevice(MJ_ADT_INPUT, i));
    dmenu->insertItem(device, i);
    dmenu->setItemChecked(i, i == (uint) m_internal->m_inputDevice);
  }
  connect(dmenu, SIGNAL(activated(int)), this, SLOT(selectInput(int)));
  m_internal->m_inputDeviceMenu = dmenu;
  
  // Create toolbar:

  QHBoxLayout *toolLayout = new QHBoxLayout(layout, 4);
  
  const char * slotNames[ToolCount] = 
    { SLOT(record()), SLOT(magnifyIn()), SLOT(magnifyOut()), 
      SLOT(zoomIn()), SLOT(zoomOut()), SLOT(zoomSelection()) };
  const char * labels[ToolCount] = 
    { "Record", "Magnify In", "Magnify Out", 
      "Zoom In", "Zoom Out", "Zoom Selection" };

  const char * pixmaps[ToolCount] = 
    { "record", "magnify-in", "magnify-out", "zoom-in", "zoom-out", 0 };
  
  QButton *button;
  QToolButton *tbutton;

  for(i=0; i < ToolCount; i++) {

    px = MJ_PixmapLoader::getPixmap(pixmaps[i]);

    QString label(labels[i]);
    QString name(label.stripWhiteSpace());
    
    if(!px.isNull()) {
      tbutton = new QToolButton(this, name.ascii());
      tbutton->setIconSet(px);
      tbutton->setTextLabel (label, true);
      button = tbutton;
    }
    else
      button = new QPushButton(label, this, name.ascii());

    connect(button, SIGNAL(pressed()), this, slotNames[i]);
    button->setAutoRepeat(true);
    toolLayout->addWidget(button);

    m_internal->m_tools[i] = button;    
  }

  // toolLayout->addWidget(new QLabel("Axis unit", this));

  QComboBox *cb = new QComboBox(this);

  cb->insertItem("Seconds");
  cb->insertItem("Frames");
  toolLayout->addWidget(cb);
  QToolTip::add(cb, "Axis unit");

  
  toolLayout->addStretch(90);

  
  QGridLayout *layout2 = new QGridLayout(layout, 2, 2, 4);

  // Create audio viewer:
  m_internal->m_view = new QAWL_View2(this);
  m_internal->m_view->setKeyboardHandler(this);

  connect(m_internal->m_view, SIGNAL(isSelected(bool)), 
	  this, SLOT(selectionAvailable(bool)));

  m_internal->m_inLevels = 
    m_internal->m_player->inLevel()->
    newGui(this, MJ_GUI_POSTCARD, this, "inlevels");

  /* m_internal->m_outLevels = 
    m_internal->m_player->outLevel()->
    newGui(this, MJ_GUI_POSTCARD, this, "outlevels");*/

  layout2->addMultiCellWidget(m_internal->m_inLevels, 0, 1, 0, 0);
  // layout2->addWidget(m_internal->m_inLevels, 0, 0);
  // layout2->addWidget(m_internal->m_outLevels, 1, 0);
  layout2->addMultiCellWidget(m_internal->m_view, 0, 1, 1, 1);
  layout2->setColStretch(0, 10);
  layout2->setColStretch(1, 90);
  // layout->addLayout(layout2, 90);

  selectionAvailable(false);

  m_internal->m_remaps.setAutoDelete(true);

  realTimeChange(false);

  connect(& m_internal->m_timer, SIGNAL(timeout()), this, SLOT(update()));

  connect(cb, SIGNAL(activated(int)), 
	  m_internal->m_view, SLOT(setScaleType(int)));
}

MJ_WaveEditor::~MJ_WaveEditor()
{
  m_module = 0;
  delete m_internal;
}

void MJ_WaveEditor::viewKeyPressEvent (QAWL_View *, QKeyEvent * e)
{
  if(e->key() == Key_Space) {
    qDebug("MJ_WaveEditor::viewKeyPressEvent # Space pressed");

    togglePlay();
  }
}

void MJ_WaveEditor::viewKeyReleaseEvent (QAWL_View *, QKeyEvent * )
{
}

bool MJ_WaveEditor::close ( bool alsoDelete)
{
  if(m_internal->m_needToSave) {
    
    int choice = QMessageBox::warning
      (0, 
       MJ_Tr::tr("Close editor ?"),
       MJ_Tr::tr("File is modified, really close it?"),
       MJ_Tr::tr("Yes"),
       MJ_Tr::tr("No"),
       QString::null,
       1);
    if(choice == 1) return false;
  }

  return QWidget::close(alsoDelete);
}

void MJ_WaveEditor::setFile(const QString &fileName)
{
  m_internal->m_view->setFile(fileName);
  addMruItem(fileName);
  redoMruMenu();

  QString caption("ZEB: \"");
  caption += fileName + "\"";
  setCaption(caption);
  setIconText(caption);
}

void MJ_WaveEditor::open()
{
  QString fileName = QFileDialog::getOpenFileName
    (QString::null, "*.wav *.aif *.aiff", this, "Select audio file");

  if(fileName.isEmpty()) return;

  setFile(fileName);
}

void MJ_WaveEditor::saveAs()
{
  QString fileName = QFileDialog::getSaveFileName
    (QString::null, "*.wav *.aif *.aiff", this, "Select audio file");

  if(fileName.isEmpty()) return;
  
  addMruItem(fileName);
  redoMruMenu();

  MJ_WaveOperations::copySegment
    (m_internal->m_view->fileName(), fileName, 0, -1, 
     MJ_WaveOperations::UserFile, 
     (DI_AudioSampleFormat) m_internal->m_fileSampleFormat);
  m_internal->m_needToSave = false;
}

void MJ_WaveEditor::saveSelectionAs()
{
  QString fileName = QFileDialog::getSaveFileName
    (QString::null, "*.wav *.aif *.aiff", this, "Select audio file");

  if(fileName.isEmpty()) return;

  addMruItem(fileName);
  redoMruMenu();

  bool ok = MJ_WaveOperations::copySegment
    (m_internal->m_view->fileName(), fileName, 
     m_internal->m_view->selectionStart(), 
     m_internal->m_view->selectionEnd(), 
     MJ_WaveOperations::UserFile, 
     (DI_AudioSampleFormat) m_internal->m_fileSampleFormat);

  if(!ok)
    error(MJ_ERR_UNKNOWN, "MJ_WaveEditor::saveSelectionAs # Operation failed");
}

void MJ_WaveEditor::quit()
{
  close(true);
}

void MJ_WaveEditor::record()
{
  if(!isRealTime()) {

    // m_module = m_internal->m_recorder;
    MJ_WavePlayer   * player = m_internal->m_player;
    MJ_WaveRecorder * recorder = player->recorder();

    recorder->setFileName(tmpFile);
    recorder->setInputDevice(m_internal->m_inputDevice);
    recorder->setChannels(m_internal->m_recordChannels);

    player->setPlayBack(false);
    player->connectPlayItems();

    startThreaded();
    m_internal->m_state = Recording;

  }
  else {
    endRecording();
  }
}

void MJ_WaveEditor::selectionAvailable(bool available)
{
  trace("MJ_WaveEditor::selectionAvailable");

  m_internal->m_fileMenu->setItemEnabled
    (m_internal->m_saveSegmentAsId, available);

  m_internal->m_tools[ZoomSelection]->setEnabled(available);

  m_internal->m_editMenu->
    setItemEnabled(m_internal->m_cropToSelection, available);
}

static const double maxZoom = 1e+6;
static const double minZoom = 0.1;

void MJ_WaveEditor::zoomIn()
{
  double zoom = m_internal->m_zoom;

  zoom /= 1.5;
  
  if(zoom < minZoom) {
    zoom = minZoom;
    m_internal->m_tools[ZoomIn]->setEnabled(false);
  }
  m_internal->m_tools[ZoomOut]->setEnabled(true);

  long ff = m_internal->m_view->firstFrame();

  double w = m_internal->m_view->width();

  double centerFrame = ff + 0.5 * m_internal->m_view->samplesPerPixel() * w;

  double newFirst = centerFrame - zoom * w * 0.5;
  double viewSpanHalf = w * zoom * 0.5;
  double maxFirst = m_internal->m_view->fileFrames() - viewSpanHalf;

  if(newFirst > maxFirst) newFirst = maxFirst;

  if(newFirst < 0.0 || ff == 0.0) newFirst = 0.0;

  m_internal->m_view->setArea((long) newFirst, zoom);
  m_internal->m_zoom = zoom;
  
  // m_view->setScale(m_zoom);
}


void MJ_WaveEditor::zoomOut()
{
  double zoom = m_internal->m_zoom;

  zoom *= 1.5;

  if(zoom > maxZoom) {
    zoom = maxZoom;
    m_internal->m_tools[ZoomOut]->setEnabled(false);
  }

  m_internal->m_tools[ZoomIn]->setEnabled(true);

  double w = m_internal->m_view->width();

  double centerFrame = m_internal->m_view->firstFrame() + 
    0.5 * m_internal->m_view->samplesPerPixel() * w;

  double newFirst = centerFrame - zoom * w * 0.5;

  if(newFirst < 0.0) newFirst = 0.0;

  m_internal->m_view->setArea((long) newFirst, zoom);
  m_internal->m_zoom = zoom;
}

void MJ_WaveEditor::zoomSelection()
{
  long ss = m_internal->m_view->selectionStart();
  long se = m_internal->m_view->selectionEnd();

  int w = m_internal->m_view->width();

  if(ss < 0 || ss == se) return;

  m_internal->m_zoom = (double) (se - ss) / (double) w;

  if(m_internal->m_zoom > maxZoom) {
    m_internal->m_zoom = maxZoom;
    m_internal->m_tools[ZoomOut]->setEnabled(false);
    m_internal->m_tools[ZoomIn]->setEnabled(true);    
  }
  else if(m_internal->m_zoom < minZoom) {
    m_internal->m_zoom = minZoom;

    m_internal->m_tools[ZoomOut]->setEnabled(true);
    m_internal->m_tools[ZoomIn]->setEnabled(false);
  }

  // qDebug("MJ_WaveEditor::zoomSelection # %ld %ld", ss, se);

  m_internal->m_view->setArea(ss, m_internal->m_zoom);
}

void MJ_WaveEditor::magnifyIn()
{
  m_internal->m_magnify *= 1.5;

  if(m_internal->m_magnify > 1000.0) {
    m_internal->m_magnify = 1000.0;
    m_internal->m_tools[MagnifyIn]->setEnabled(false);
  }

  m_internal->m_tools[MagnifyOut]->setEnabled(true);

  m_internal->m_view->setMagnification(m_internal->m_magnify);
}

void MJ_WaveEditor::magnifyOut()
{
  m_internal->m_magnify /= 1.5;

  if(m_internal->m_magnify < 1.0) {
    m_internal->m_magnify = 1.0;
    m_internal->m_tools[MagnifyOut]->setEnabled(false);
  }
  m_internal->m_tools[MagnifyIn]->setEnabled(true);

  m_internal->m_view->setMagnification(m_internal->m_magnify);
}


void MJ_WaveEditor::selectAll()
{
  m_internal->m_view->selectAll();
}

void MJ_WaveEditor::cropToSelection()
{
  QString tmpFileName;
  tmpFileName.sprintf("/tmp/zeb-tmp-%d.wav", m_internal->newTmpIndex());

  // QString realFile = m_internal->m_view->fileName();
  
  bool ok = MJ_WaveOperations::copySegment
    (m_internal->m_view->fileName(), tmpFileName, 
     m_internal->m_view->selectionStart(), m_internal->m_view->selectionEnd(),
     MJ_WaveOperations::TmpFile);

  if(ok) {
    m_internal->m_view->setFile(tmpFileName);
    m_internal->m_needToSave = true;
  }
  else
    error(MJ_ERR_UNKNOWN, 
	  "MJ_WaveEditor::cropToSelection # Operation failed");
}

void MJ_WaveEditor::togglePlay()
{
  if(!isRealTime()) {
    
    MJ_WavePlayer * player = m_internal->m_player;

    MJ_MessageString ms(m_internal->m_view->fileName());
    player->msgProcess(MJ_Strs::filename, &ms);

    player->setOutputDevice(m_internal->m_outputDevice);

    player->setPlayBack(true);
    player->connectPlayItems();

    long startFrame = m_internal->m_view->selectionStart();
    if(startFrame < 0) {
      startFrame = m_internal->m_view->firstFrame();
    }

    long endFrame = m_internal->m_view->selectionEnd();
    if(endFrame < 0)
      endFrame = 0;
    
    qDebug("MJ_WaveEditor::togglePlay # %ld %ld", startFrame, endFrame);

    m_internal->m_view->setCursorFrame(startFrame);

    m_module = player;


    startThreaded(startFrame, endFrame);

    m_internal->m_state = Playing;
  }
  else {
    if(m_internal->m_state == Recording)
      endRecording();
    else {
      stopPlaying();
      m_internal->m_state = Inactive;
    }
  }
}

void MJ_WaveEditor::selectOutput(int dev)
{
  for(int i=0; i < (int) m_internal->m_outputDeviceMenu->count(); i++) {
    m_internal->m_outputDeviceMenu->setItemChecked(i, i == dev);
  }
  m_internal->m_outputDevice = dev;
}

void MJ_WaveEditor::selectInput(int dev)
{
  for(int i=0; i < (int) m_internal->m_inputDeviceMenu->count(); i++) {
    m_internal->m_inputDeviceMenu->setItemChecked(i, i == dev);
  }
  m_internal->m_inputDevice = dev;
}

void MJ_WaveEditor::selectFileSampleFormat(int f)
{
  trace2("MJ_WaveEditor::selectFileSampleFormat # %d", f);

  for(int i=0; i < MJE_SF_SIZEOF; i++) {
    int tmp = m_internal->m_fileFormats[i];
    m_internal->m_fileFormatMenu->setItemChecked(tmp, f == tmp);
    if(f == tmp) 
      m_internal->m_fileSampleFormat = i;
  }
}

/*void MJ_WaveEditor::selectXAxis(int f)
{
  m_internal->m_view->setScaleType(f);
  }*/

void MJ_WaveEditor::update()
{
  lock();
  long frame = (long) m_module->dspGetTimer()->stamp();
  unlock();

  m_internal->m_view->setCursorFrame(frame);
}

void MJ_WaveEditor::addMruItem(const QString &f)
{
  QFileInfo fi(f);

  m_internal->m_pastFileNames[fi.fileName()] = f;
}

void MJ_WaveEditor::redoMruMenu()
{
  uint i;

  for(i=0; i < m_internal->m_pastFiles.size(); i++)
    m_internal->m_fileMenu->removeItem(m_internal->m_pastFiles[i]);

  m_internal->m_pastFiles.resize(0);
  m_internal->m_remaps.clear();

  for(std::map<QString, QString>::iterator it = 
	m_internal->m_pastFileNames.begin();
      it != m_internal->m_pastFileNames.end(); it++) {

    MJ_StrRemap * remap = new MJ_StrRemap((*it).second);
    m_internal->m_remaps.append(remap);

    connect(remap, SIGNAL(emitCall(const QString &)), 
	    this, SLOT(setFile(const QString &)));

    int tmp = m_internal->m_fileMenu->insertItem
      ((*it).first, remap, SLOT(call()));

    m_internal->m_pastFiles.push_back(tmp);
  }
}

void MJ_WaveEditor::realTimeChange(bool nowRt)
{
  bool notRt = !nowRt;

  m_internal->m_listenMenu->setItemEnabled(m_internal->m_playId, notRt);
  m_internal->m_listenMenu->setItemEnabled(m_internal->m_stopId, nowRt);

  if(nowRt) m_internal->m_timer.start(20);
  else { 
    m_internal->m_timer.stop();
    long frame = (long) m_module->dspGetTimer()->stamp();
    m_internal->m_view->setCursorFrame(frame);
  }

  int chans = m_internal->m_player->inLevel()->dspInChannels();
  
  int levelWidth = chans * 10;

  if(levelWidth < 80) levelWidth = 80;
  else if(levelWidth > 250)
    levelWidth = 250;

  QWidget * inLevels = m_internal->m_inLevels;

  inLevels->setMinimumWidth(levelWidth);
  inLevels->setMaximumWidth(levelWidth);
}

void MJ_WaveEditor::endRecording()
{
  trace("MJ_WaveEditor::endRecording");

  if(m_internal->m_state != Recording) {
    error(MJ_ERR_INTERNAL, "MJ_WaveEditor::endRecording # Not recording");
  }

  stopPlaying();
  m_internal->m_state = Inactive;
  m_internal->m_needToSave = true;
  setFile(tmpFile);
}
