/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * Tommi.Ilmonen(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_wave_player.h>

#include <mj_audio_conf.h>
#include <mj_dsp_plugin.h>
#include <mj_kernel.h>
#include <mj_message_numeric.h>
#include <mj_message_string.h>
#include <mj_module_gain.h>
#include <mj_module_level.h>
#include <mj_strs.h>
#include <mj_wave_recorder.h>

#include <di_io_defs.h>

#include <typeinfo>

inline void MJ_WavePlayer::DspItem::dspProcessPeriod()
{
  if(m_module) m_module->dspProcessPeriod();

  if(m_wetDryRamp.get() == 1.0 && ! m_wetDryRamp.isActive())
    return;

  uint chans = m_module->dspInChannels();

  if(!chans || !m_module->dspOutChannels()) return;
  
  MJ_SignalSource * source = & m_module->dspInSignal(0);
  MJ_SignalSink * sink = & m_module->dspOutSignal(0);

  uint samples = source->samples().size();
  
  for(uint chan=0; chan < chans; chan++) {
    MJ_Ramp<MJ_Sample> tmpRamp(m_wetDryRamp);

    MJ_Sample *in = source->samples().data();
    MJ_Sample *out = sink->samples().data();

    for(uint j=0; j < samples; j++) {
      MJ_Sample wet = tmpRamp.get();

      tmpRamp.advance();

      *out = (*out) * wet + (1.0 - wet) * (*in);
      in++;
      out++;
    }
    
    source++;
    sink++;
  }
  
  m_wetDryRamp.advanceN(samples);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

bool MJ_WavePlayer::m_debug = true;

MJ_WavePlayer::MJ_WavePlayer()
  : m_recorder(new MJ_WaveRecorder),
    m_fileReader(0),
    m_gain(new MJ_ModuleGain),
    m_audioOut(0),
    m_inLevel(new MJ_ModuleLevel),
    m_outLevel(new MJ_ModuleLevel),
    m_audioDevId(0),
    m_recordChannels(0),
    m_playBack(true)
{
  MJ_ModuleDspArray::dspSetTimer(&m_timer);
}

MJ_WavePlayer::~MJ_WavePlayer()
{
  delete m_recorder;
  delete m_gain;
  delete m_fileReader;
  delete m_audioOut;
  delete m_inLevel;
  delete m_outLevel;
}

bool MJ_WavePlayer::dspPrepare()
{
  // const char * fname = "MJ_WavePlayer::dspPrepare";

  if(!connectPlayItems())
    return false;  

  for(dspIterator it = m_dspItems.begin(); it != m_dspItems.end(); it++)
    (*it).prepare();

  return MJ_ModuleDspArray::dspPrepare();
}

void MJ_WavePlayer::dspProcessPeriod()
{

  if(m_playBack) {
    m_fileReader->dspProcessPeriod();  
    m_inLevel->dspProcessPeriod();  

    for(dspIterator it = m_dspItems.begin(); it != m_dspItems.end(); it++) {
      (*it).dspProcessPeriod();
    }
    m_gain->dspProcessPeriod();
    m_outLevel->dspProcessPeriod();  
    m_audioOut->dspProcessPeriod();
  }
  else {
    m_recorder->dspProcessPeriod();
    m_inLevel->dspProcessPeriod();
  }

  if(m_childTimer != m_currentTime)
    m_childTimer->incrementStamp(m_childTimer->frameSize());
  if(!parentModule())
    m_currentTime->incrementStamp(m_currentTime->frameSize());

  // m_graphTimer.incrementStamp(m_graphTimer.frameSize());
}

bool MJ_WavePlayer::msgProcess(const char *id, MJ_Message *m)
{
  const char * fname = "MJ_WavePlayer::msgProcess";

  trace3(m_debug, "%s # \"%s\", %p", fname, id, m);
  
  if(strcmp(id, MJ_Strs::filename) == 0) {
    MJ_MessageString *ms = dynamic_cast<MJ_MessageString *> (m);

    if(!ms)
      return false;

    m_fileName = ms->string();
    return true;
  }
  
  return false;
}

bool MJ_WavePlayer::connectPlayItems()
{
  const char * fname = "MJ_WavePlayer::connectPlayItems";

  dspIterator it;

  // FIRST WE CREATE THE I/O MODULES:

  // Create the file reader module:

  if(!m_fileReader)
    m_fileReader = MJ_Kernel::newFunctionalModule
      (MJ_DspPluginFunctionalityKeys[MJ_DPF_FILE_IN]);
  
  if(!m_fileReader && m_playBack) {
    error2(MJ_ERR_UNKNOWN, "%s # Could not create file reader module", fname);
    return false;
  }

  // Create the audio output module:
  if(!m_audioOut)
    m_audioOut = MJ_Kernel::newFunctionalModule
      (MJ_DspPluginFunctionalityKeys[MJ_DPF_AUDIO_OUT]);

  if(!m_audioOut) {
    error2(MJ_ERR_UNKNOWN, "%s # Could not create audio output module", fname);
    return false;
  }
  
  clearArray();

  // NOW WE DO THE CONNECTIONS:

  MJ_ModuleDsp *prev, *curr = 0;

  prev = m_playBack == true ? m_fileReader : m_recorder;

  appendModule(prev);
  
  appendModule(m_inLevel);
  
  for(it = m_dspItems.begin(); it != m_dspItems.end() && m_playBack; it++) {
    MJ_ModuleDsp * m = (*it).module();
    if(m) appendModule(m);
  }

  appendModule(m_gain);
  appendModule(m_outLevel);

  if(m_playBack)
    appendModule(m_audioOut);

  MJ_MessageString ms(m_fileName);
  
  m_fileReader->msgProcess(MJ_Strs::filename, &ms);
  
  int foo = 0, channels = 1;

  // Determine the number of channels
  prev->dspAcceptIoChans(&foo, &channels);
  
  trace3(m_debug, "%s # %d channels from %s (%d)", 
	 fname, channels, typeid(*prev).name(), (int) m_playBack);

  prev->dspSetOutChannels(channels);

  m_inLevel->dspSetInChannels(channels);
  for(int i = 0; i < channels; i++) {
    m_inLevel->dspConnectInput(prev->id(), i, i);
  }

  for(it = m_dspItems.begin(); it != m_dspItems.end(); it++) {
    curr = (*it).module();
    
    int tmp1 = channels, tmp2 = channels;

    curr->dspAcceptIoChans(&tmp1, &tmp2);

    if(tmp1 != channels || tmp2 != channels) {
      error2(MJ_ERR_UNKNOWN, 
	     "%s # Module %s could not be configured to %d channels", 
	     fname, typeid(*curr).name(), channels);
      return false;
    }
    
    curr->dspSetOutChannels(channels);
    curr->dspSetInChannels(channels);

    for(int i=0; i < channels; i++)
      curr->dspConnectInput(prev->id(), i, i);

    prev = curr;
  }

  // Connect the gain to the previous output
  m_gain->dspSetOutChannels(channels);
  m_gain->dspSetInChannels(channels);
  
  for(int i = 0; i < channels; i++) {
    m_gain->dspConnectInput(prev->id(), i, i);
  }

  // Connect the level meter after the gain:
  m_outLevel->dspSetInChannels(channels);
  for(int i = 0; i < channels; i++) {
    m_outLevel->dspConnectInput(m_gain->id(), i, i);
  }
  
  // Configure and connect the audio output:

  QString devName = MJ_AudioConf::audioDevice(MJ_ADT_OUTPUT, m_audioDevId);
  ms.setString(devName);
  bool ok = m_audioOut->msgProcess(MJ_Strs::device, &ms);

  if(!ok) {
    error2(MJ_ERR_UNKNOWN, 
	   "%s # Audio output configured failed", fname);    
    return false;
  }
  else
    trace2("%s # Got audio device \"%s\"", fname, devName.ascii());

  int audioDevChannels = MJ_AudioConf::deviceChannels
    (MJ_ADT_OUTPUT, m_audioDevId);

  m_audioOut->dspSetInChannels(audioDevChannels);
  
  for(int i = 0; i < audioDevChannels; i++) {
    m_audioOut->dspConnectInput(m_gain->id(), i % channels, i);
  }

  return true;
}

