/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * Tommi.Ilmonen(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_wave_recorder.h>

#include <mj_audio_conf.h>
#include <mj_message_string.h>
#include <mj_realtime_scheduling.h>
#include <mj_strs.h>


#ifdef __linux__
#include <mj_rtc_hz.h>
#endif

bool MJ_WaveRecorder::m_debug = true;

MJ_WaveRecorder::MJ_WaveRecorder()
  : m_device(MJ_AudioConf::audioDevice(MJ_ADT_INPUT, 0)),
    m_deviceId(0),
    m_fileName("record.wav"),
    m_channels(-1),
    m_fileChannels(0)
{
  trace3(m_debug, "MJ_WaveRecorder::MJ_WaveRecorder");
}

MJ_WaveRecorder::~MJ_WaveRecorder()
{
  trace3(m_debug, "MJ_WaveRecorder::~MJ_WaveRecorder");
}

int  MJ_WaveRecorder::bufferLength()
{
  int b = m_input.inFilledFrames();

  // trace3(m_debug, "MJ_WaveRecorder::bufferLength # %d", b);

  return b;
}

bool MJ_WaveRecorder::dspAcceptIoChans(int *in, int *out)
{
  bool changed = false;

  if(*in) {
    *in = 0;
    changed = true;
  }

  int rout = fileChans();
  if(*out != rout) {
    *out = rout;
    changed = true;
  }

  return changed;
}

bool MJ_WaveRecorder::dspPrepare()
{
  const char * fname = "MJ_WaveRecorder::dspPrepare";

  trace3(m_debug, fname);

  int sr = dspSamplingRate();

  int minBlock = dspFrameSize();

  if(minBlock < m_minimumBuffer / 2) {

    minBlock = m_minimumBuffer / 2;

    if(m_debug) trace2("%s # minBlock set to %d", fname, (int) minBlock);
  }
  
  int chans = MJ_AudioConf::deviceChannels(MJ_ADT_INPUT, m_deviceId);

  bool ok = m_input.open(m_device.ascii(),
			 DI_IO_INPUT, 
			 DI_ASF_INT32,
			 minBlock * 5,
			 sr, chans, minBlock);

  if(!ok) {
    error2(MJ_ERR_UNKNOWN, "%s # Could not open input \"%d\"", 
	   fname, m_device.ascii());
    return false;
  }

  m_buffer.resize(chans * dspFrameSize());

  m_fileChannels = fileChans();

  m_buffer2.resize(m_fileChannels * dspFrameSize());

  m_file.startWrite(m_fileName.ascii(),
		    DI_ASF_INT32,
		    sr, m_fileChannels, sr * 2, 
		    MJ_RTP_FILE_IO, DI_AFF_WAVE, DI_ASF_INT32);

  ok = m_file.userPrepare();

  trace2("%s # Opened file %s for writing with %d channels",
	 fname, m_fileName.ascii(), m_fileChannels);
  
  return ok;
}

bool MJ_WaveRecorder::dspPrepareFast()
{
  return m_input.start();
}

void MJ_WaveRecorder::dspProcessPeriod()
{
  // trace("MJ_WaveRecorder::dspProcessPeriod");

  uint frames = dspFrameSize();

  m_input.readFrames(&* m_buffer.begin(), frames);

  int * data;

  if(m_buffer2.size() != m_buffer.size()) {

    int inchans = m_input.inChannels();
    uint fileChannels = (uint) m_fileChannels;
    
    /*trace2("MJ_WaveRecorder::dspProcessPeriod # HALF %d %u %u", inchans, 
      fileChannels, frames); */

    for(uint j=0; j < fileChannels; j++) {

      MJ_Sample *out =  m_signalOutput[j].samples().data();

      for(uint i=0; i < frames; i++) {
	int v = m_buffer[i * inchans + j % inchans];

	m_buffer2[i * fileChannels + j] = v;
	*out++ = v / 2147483648.0;
      }
    }

    data = &* m_buffer2.begin();
  }
  else
    data = &* m_buffer.begin();

  m_file.writeFrames(data, frames);

  m_currentTime->incrementStamp(m_currentTime->frameSize());
}

bool MJ_WaveRecorder::dspStop()
{
  trace3(m_debug, "MJ_WaveRecorder::dspStop");

  m_input.close();

  m_file.requestEnd();
  m_file.waitEnd();

  return true;
}

bool MJ_WaveRecorder::msgProcess(const char *id, MJ_Message *m)
{
  const char * fname = "MJ_WaveRecorder::msgProcess";

  trace3(m_debug, "%s # \"%s\"", fname , id);

  if(!strcmp(id, MJ_Strs::device)) {
    
    MJ_MessageString *ms = dynamic_cast<MJ_MessageString *> (m);

    if(ms) {
      m_device = ms->string();
      if(m_debug) trace(fname, ms->string());
      return true;
    }
    else {
      error(MJ_ERR_ARGUMENT, fname, "Not a string message");
      return false;
    }
  }
  if(!strcmp(id, MJ_Strs::channels)) {
    
    bool ok = true;

    int chans = m->toInt(&ok);

    if(!ok) {
      return false;
    }
    
    m_channels = chans;
  }

  return false;
}

void MJ_WaveRecorder::setInputDevice(int id)
{
  const char * fname = "MJ_WaveRecorder::setInputDevice";

  int maxDev = MJ_AudioConf::audioDevices(MJ_ADT_INPUT);

  if(id >= maxDev) {
    error2(MJ_ERR_ARGUMENT, "%s # Only %d devices (required %d)", 
	   fname, maxDev, id);
    return;
  }

  MJ_MessageString ms(MJ_AudioConf::audioDevice(MJ_ADT_INPUT, id));

  msgProcess(MJ_Strs::device, &ms);
  m_deviceId = id;
}

int MJ_WaveRecorder::fileChans()
{
  int fileChans = m_channels;

  if(fileChans <= 0)
    fileChans = MJ_AudioConf::deviceChannels(MJ_ADT_INPUT, m_deviceId);

  return fileChans;
}
