/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */


#include <mj_midi_thru.h>
#include <mj_module_dsp_row_gui.h>
#include <mj_module_gain.h>
#include <mj_module_midiseq.h>
#include <mj_module_minisynth.h>
#include <mj_module_mixer_gui.h>
#include <mj_module_monodelay.h>
#include <mj_module_print_messages.h>
#include <mj_module_printer.h>
#include <mj_module_protector.h>

#include <mj_debug.h>
#include <mj_global.h>
#include <mj_fpu.h>
#include <mj_prefs.h>

#include <stdlib.h>

#include <qapplication.h>

enum MJ_CmdLineOption {
  MJ_CMDL_HELP, 
  MJ_CMDL_HELP_LONG, 
  MJ_CMDL_BECHMARK_DSP_ROW, 
  MJ_CMDL_DEBUG, 
  MJ_CMDL_MIDI_THRU_GUI, 
  MJ_CMDL_TEST_DSP_ROW, 
  MJ_CMDL_TEST_DSP_ROW_GUI, 
  MJ_CMDL_TEST_GAIN_WRITE, 
  MJ_CMDL_TEST_MIDI_SEQ_GUI, 
  MJ_CMDL_TEST_MINI_SYNTH_GUI, 
  MJ_CMDL_TEST_MSG_PRINTER, 
  MJ_CMDL_TEST_MIXER_GUI, 
  MJ_CMDL_TEST_PRINTER_GUI, 
  MJ_CMDL_TEST_PROTECTOR_GUI, 
  MJ_CMDL_TEST_TUI, 
  MJ_CMDL_SIZEOF  
};


static const char *LC_cpaCmdLineOptions[MJ_CMDL_SIZEOF] = {
  "-h",
  "--help",
  "--bechmark_dsp_row",
  "--debug",
  "--midi_thru_gui",
  "--test_dsp_row",
  "--test_dsp_row_gui",
  "--test_gain_write",
  "--test_midi_seq_gui",
  "--test_mini_synth_gui",
  "--test_msg_printer",
  "--test_mixer_gui",
  "--test_printer_gui",
  "--test_protector_gui",
  "--test_tui"
};

void MJ_help(const char *programName, 
	     const char *badOption = 0,
	     bool exit_now = true)
{
  if(badOption) 
    fprintf(stderr, "%s : bad option: \"%s\"\n", programName, badOption);

  fprintf(stderr, "Usage: %s [options]\nOptions:\n", programName);

  fflush(0);
  
  for(int i=0; i < MJ_CMDL_SIZEOF; i++) {
    fprintf(stderr, "%s\n", LC_cpaCmdLineOptions[i]);
  }

  fflush(0);

  if(exit_now) exit(0);
}

int main(int argc, char **argv)
{
  int option, tmp, tmp2, tmp3, tmp4;

#ifdef MJ_HAVE_FPU_MANAGER
  MJ_Fpu::fast();
#endif
  
  // First determine if we want qApp.
  bool useQApp = true;
  for(option = 1; option < argc; option++) {
    if(strcmp(argv[option], "--no_qapp") == 0) {
      useQApp = false;
      MJ_shiftBack(option, 1, &argc, argv);
    }
  }

  QApplication *qa(0);
  if(useQApp)
    qa = new QApplication(argc, argv);
  
  MJ_prefs.scanArgs(argc, argv);

  for(option = 1; option < argc; option++) {
    const char *cpOpt = argv[option];
    MJ_CmdLineOption o = 
      (MJ_CmdLineOption) MJ_findStr(cpOpt, 
				    LC_cpaCmdLineOptions, 
				    MJ_CMDL_SIZEOF);
    
    switch(o) {
    case MJ_CMDL_HELP:
    case MJ_CMDL_HELP_LONG:
      MJ_help(argv[0], 0, true);
      break;

    case MJ_CMDL_MIDI_THRU_GUI:
      MJ_MidiThru(qa, argc, argv);
      return 0;

    case MJ_CMDL_BECHMARK_DSP_ROW:
      tmp  = 30; // Modules
      tmp2 = 64; // Frame size
      tmp3 = 44100 * 5; // Samples
      tmp4 = 0;

      if(++option < argc) tmp  = atoi(argv[option]);
      if(++option < argc) tmp2 = atoi(argv[option]);
      if(++option < argc) tmp3 = atoi(argv[option]);
      if(++option < argc) tmp4 = atoi(argv[option]);

      MJ_ModuleDspRow::runBenchmark(tmp, tmp2, tmp3, tmp4);
      return 0;

    case MJ_CMDL_DEBUG:
      if(option + 1 == argc) MJ_Debug::debugType(MJ_DEBUG_ALL);
      else if(argv[option+1][0] == '-') {
	MJ_Debug::debugType(MJ_DEBUG_ALL);
      }
      else {
	tmp = 1;
	
	while (++option < argc && tmp) {
	  
	  if(argv[option][0] == '-') { // If this is a new option, break loop.
	    
	    option -= 2;
	    tmp = 0;
	  }
	  else MJ_Debug::debugName(argv[option]);
	}
      }
      break;

    case MJ_CMDL_TEST_DSP_ROW:
      MJ_ModuleDspRow::test();
      return 0;

    case MJ_CMDL_TEST_DSP_ROW_GUI:
      MJ_ModuleDspRow_QG::test(qa, argc, argv);
      return 0;

    case MJ_CMDL_TEST_GAIN_WRITE:
      MJ_ModuleGain::testStreamIo();
      return 0;

    case MJ_CMDL_TEST_MIDI_SEQ_GUI:
      MJ_ModuleMidiseq_QG::test(qa, argc, argv);
      return 0;

    case MJ_CMDL_TEST_MINI_SYNTH_GUI:
      MJ_ModuleMiniSynth_QG::test(qa, argc, argv);
      return 0;

    case MJ_CMDL_TEST_MSG_PRINTER:
      MJ_ModulePrintMessages::test(argc, (const char **)argv);
      return 0;

    case MJ_CMDL_TEST_MIXER_GUI:
      MJ_ModuleMixer_QG::test(qa, argc, argv);
      return 0;
      
    case MJ_CMDL_TEST_PRINTER_GUI:
      MJ_ModulePrinter_QG::test(qa, argc, argv);
      return 0;


    case MJ_CMDL_TEST_PROTECTOR_GUI:
      MJ_ModuleProtector::test(qa, argc, argv);
      return 0;

    case MJ_CMDL_TEST_TUI:
      MJ_ModuleMonoDelay::testTui(argc - option - 1, &argv[option+1]);
      return 0;

    default:
      MJ_help(argv[0], cpOpt, true);
      break;
    }
  }

  if(qa) delete qa;
  MJ_help(argv[0], 0, true);
  return 0;
}
