/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_mixer_strip_qg.h>

#include <mj_error.h>
#include <mj_math.h>
#include <mj_module_mixer_source.h>
#include <mj_tr.h>


#include <qapplication.h>
#include <qlayout.h>
#include <qmenubar.h> 
#include <qpopupmenu.h>
#include <qscrollview.h> 

#include <stdio.h>

int MJ_MixerStrip_QG::m_debug = 0;

MJ_MixerStrip_QG::MJ_MixerStrip_QG(MJ_RtControl *rtparent,
				   MJ_ModuleDsp *module,
				   uint          strip,
				   MJ_ModuleMixer_QG *mixerGui,
				   QWidget      *parent,
				   const char   *name)
  : MJ_ModuleDspGui(rtparent, module, parent, name),
    m_strip(strip),
    m_mixer(0), 
    m_mixerGui(mixerGui),
    m_layout1(0),
    m_layout2(0),
    m_scroll(new QScrollView(this)),
    m_base(new QWidget(m_scroll->viewport()))
{
  m_layout1 = new QVBoxLayout(this);

  QMenuBar *menu = new QMenuBar( this );
  m_layout1->setMenuBar(menu);

  QPopupMenu *edit = new QPopupMenu(this);
  menu->insertItem(MJ_Tr::tr("&Edit"), edit);
  m_undoControl.setMenu(edit);

  m_layout1->addWidget(m_scroll);

  m_scroll->addChild(m_base);

  m_children.setAutoDelete(true);

  m_base->setBackgroundColor(QColor(40, 40, 40));

  if(!parent) {
    QString tmp(MJ_Tr::tr("Strip")), tmp2;
    tmp2.sprintf(" %d", (int) strip + 1);
    tmp += tmp2;
    setIconText(tmp);
    setCaption(tmp);
  }
  
  // Make sure this widget does not start too small:
  resize(MJ_max(300, width()), MJ_max(300, height()));

  m_mixer = (MJ_ModuleMixer *) module;

  MJ_Satellite *s = m_mixer->satellite();
  
  connect(s, SIGNAL(emitRemove(int,int)), this, SLOT(remove(int,int)));
  connect(s, SIGNAL(emitUpdate(int,int)), this, SLOT(update(int,int)));
  
  draw();
}
  
MJ_MixerStrip_QG::~MJ_MixerStrip_QG()
{
  // trace("MJ_MixerStrip_QG::~MJ_MixerStrip_QG");
}

void MJ_MixerStrip_QG::hide()
{
  QWidget::hide();

  /* if(this != qApp->mainWidget())
     kill(); */
}

void MJ_MixerStrip_QG::moduleModified4(int i1, int i2, int i3)
{
  (void) i1;

  switch(i1) {
  case MJ_MICH_RESIZE_PREPARE:

    if((uint)i2 <= m_strip) hide();
    limitSize(i3);

    break;
    
  case MJ_MICH_RESIZE_FINAL:
    expandTo(i3);
    break;

  default:
    break;
  }
  
}

void MJ_MixerStrip_QG::remove(int strip, int row)
{
  if(strip != (int) m_strip) return;

  if(m_debug) trace("MJ_MixerStrip_QG::remove", row);

  m_children.remove(row);
}

void MJ_MixerStrip_QG::update(int strip, int row)
{
  if(strip != (int) m_strip) return;
  
  if(m_debug) trace("MJ_MixerStrip_QG::update", row);

  QWidget *w = m_mixer->unit(m_strip, row)->newGui
    (MJ_GUI_POSTCARD, strip, row, m_mixerGui, m_base);
  
  m_children.insert(row, w);
  
  lock();
  MJ_ModuleDsp *m = m_mixer->module(strip, row);
  unlock();

  if(m)
    m_undoControl.addSatellite(m->satellite());
  
  redoLayout();
}

void MJ_MixerStrip_QG::kill()
{ 
  if(m_debug) trace("MJ_MixerStrip_QG::kill");
  
  delete this; 
}

void MJ_MixerStrip_QG::resizeEvent(QResizeEvent *e)
{
  QWidget::resizeEvent(e);
  determineSizes();
}

void MJ_MixerStrip_QG::limitSize(int rows)
{
  while(rows < (int) m_children.count()) {
    m_children.removeLast();
  }
  redoLayout();
}

void MJ_MixerStrip_QG::expandTo(int rows)
{
  uint tmp;

  while(rows > (int) (tmp = m_children.count()))
    fillSlot(tmp);

  redoLayout();
}

void MJ_MixerStrip_QG::fillSlot(uint rowInd)
{
  int strip = m_strip;
  
  QWidget *w = m_mixer->unit(strip, rowInd)->newGui
    (MJ_GUI_POSTCARD, strip, rowInd, m_mixerGui, m_base);
  
  m_children.append(w);
  
  lock();
  MJ_ModuleDsp *m = m_mixer->module(strip, rowInd);
  unlock();
  
  if(m)
    m_undoControl.addSatellite(m->satellite());
}


void MJ_MixerStrip_QG::draw()
{
  lock();
  uint rows = m_mixer->rows();
  unlock();

  for(uint i=0; i < rows; i++)
    fillSlot(i);

  redoLayout();
}

void MJ_MixerStrip_QG::redoLayout()
{
  const char * fname = "MJ_MixerStrip_QG::redoLayout";

  if(m_layout2) delete m_layout2;
  
  m_layout2 = new QVBoxLayout(m_base, 20, 20);
  if(!m_layout2) MJ_fatal(MJ_ERR_MEMORY, fname);

  const uint max = m_children.count();
  
  for(uint i=0; i < max; i++) {
    QWidget * w = m_children.at(i);

    uint stretch = 10;
    
    MJ_ModuleDspGui *gd = dynamic_cast<MJ_ModuleDspGui *> (w);
    MJ_ModuleMixerSource_QG *ms = dynamic_cast<MJ_ModuleMixerSource_QG *> (w);

    if(gd) stretch = gd->weight();
    if(ms) stretch = 40;
    
    m_layout2->addWidget(w, stretch);
    w->show();
  }
  determineSizes();
}

void MJ_MixerStrip_QG::determineSizes()
{
  const char *fname = "MJ_MixerStrip_QG::determineSizes";

  QSize sizeHint = m_base->minimumSizeHint();
  
  sizeHint *= 1.5;

  int w, h;

  if(sizeHint.width() < m_scroll->visibleWidth())
    w = m_scroll->visibleWidth();
  else 
    w = sizeHint.width();
  
  if(sizeHint.height() < m_scroll->visibleHeight())
    h = m_scroll->visibleHeight();
  else 
    h = sizeHint.height();

  // sizeHint = m_scroll->viewportSize(w, h);
  // m_scroll->setVScrollBarMode(QScrollView::AlwaysOn);

  // m_base->resize(w,h);
  m_base->setMinimumSize(w,h);
  m_base->setMaximumSize(w,h);
  m_scroll->resizeContents(w,h);
  // m_scroll->show();

  if(m_debug) {
    char buf[128];
    sprintf(buf, "%s w = %d h = %d", fname, w, h);
    MJ_trace(buf);
  }

  // m_base->resize(sizeHint);
}
