/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_automation.h>

#include <mj_error.h>
#include <mj_message_numeric.h>

#include <qlayout.h> 

#include <math.h>

MJ_ModuleAutomation::MJ_ModuleAutomation()
{}

MJ_ModuleAutomation::~MJ_ModuleAutomation()
{}

void MJ_ModuleAutomation::dspProcessPeriod()
{
  if(!m_currentTime)
    MJ_fatal(MJ_ERR_INTERNAL, 
	     "MJ_ModuleAutomation::dspProcessPeriod # time = NULL");

  MJ_AutomationSingle::Time time;

  time.setSeconds(m_currentTime->secondsD());
  
  MJ_MessageDouble md;

  for(iterator it = m_automations.begin(); it != m_automations.end(); it++) {

    MJ_AutomationSingle * automation = & (*it).second;

    if(!automation->size())
      continue;
    
    const char * address = (*it).first.c_str();

    double val = automation->value(time);

    md.setValue(val);

    msgPushOutput(address, &md);
  }
}
  
MJ_ModuleDsp *MJ_ModuleAutomation::createNew()
{
  return new MJ_ModuleAutomation();
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_AutomationGraphs::MJ_AutomationGraphs(MJ_ModuleAutomation_QG * host,
					 QWidget      *parent,
					 const char   *name)
  : QGLWidget(parent, name),
    m_host(host),
    m_activeParam("undefined"),
    m_showHit(false)
{
  m_left.setSeconds(0.0);
  m_right.setSeconds(100.0);
  setMouseTracking(true);
}

MJ_AutomationGraphs::~MJ_AutomationGraphs()
{}

void MJ_AutomationGraphs::mousePressEvent ( QMouseEvent * e )
{
  double y;
  MJ_AutomationSingle::Time t;
  windowToReal(e->x(), e->y(), &t, &y);

  if((e->stateAfter() & LeftButton) /* && !(e->state() & LeftButton) */) {

    if(m_showHit) {
      // Start to move the current vertex

      MJ_AutomationSingle::container * graph = &
	m_host->automation()->automations()[m_activeParam].graph();

      MJ_AutomationSingle::iterator it = graph->find(m_hitTime);

      MJ_AutomationSingle::Time maxTime;
      MJ_AutomationSingle::Time minTime(0);

      {
	// calculate the limits:

	MJ_AutomationSingle::iterator next = it;
	next++;
	if(next == graph->end())
	  maxTime.makeMaximum();
	else
	  maxTime = (*next).first;

	MJ_AutomationSingle::iterator prev = it;
	if(prev == graph->begin())
	  minTime = -1;
	else
	  minTime = (* --prev).first;
      }
      
      if(t <= minTime)
	t = minTime + 1;
      if(t >= maxTime)
	t = maxTime - 1;
      
      /* DANGEROUS: We bypass the normal "const"
	 limitations. Assumption is made that since we do not move the
	 new time value beyond the previous and next time values, then
	 the map ordering remains valid. C++ rocks. STL shakes. */
      * ((MJ_AutomationSingle::Time *) & (*it).first)  = t;
      (*it).second = y;
    }
    else {      
      m_host->automation()->automations()[m_activeParam].addPoint(t, y);
      MJ_trace2("MJ_AutomationGraphs::mousePressEvent # %lf %lf",
		t.secondsD(), y);
    }

    m_hitTime = t;
    m_hitVal = y;
    m_showHit = true;

    updateGL();
  }
}

void MJ_AutomationGraphs::mouseReleaseEvent ( QMouseEvent * e )
{}

void MJ_AutomationGraphs::mouseMoveEvent ( QMouseEvent * e )
{
  if(e->stateAfter() & LeftButton) {
    mousePressEvent(e);
  }
  else {
    // Check if we should highlight some vertex:

    MJ_AutomationSingle::Time time;
    double v;
    windowToReal(e->x(), e->y(), &time, &v);

    // Seek if we find a point to shoot at:
    MJ_AutomationSingle::container * graph = &
      m_host->automation()->automations()[m_activeParam].graph();

    MJ_AutomationSingle::iterator node = graph->lower_bound(time);

    const int sentinel = 15;
    double closest = sentinel;

    MJ_AutomationSingle::Time tsentinel = windowToRealX(e->x() - sentinel);

    for(; node != graph->begin() && (*node).first >= tsentinel; node--)
      ;

    m_showHit = false;

    while(node != graph->end()) {
      int x, y;
      realToWindow((*node).first, (*node).second, &x, &y);
      
      int dx = x - e->x();
      int dy = e->y() - y;

      double dist = sqrt((double) (dx * dx + dy * dy));

      if(closest > dist) {
	closest = dist;
	m_hitTime = (*node).first;
	m_hitVal  = (*node).second;
	m_showHit = true;
      }
      else if(dx > sentinel)
	break; // No need to scan the rest.
      
      node++;
    }

    MJ_trace2("MJ_AutomationGraphs::mouseMoveEvent # %d", (int) m_showHit);
    updateGL();
  }

}

void MJ_AutomationGraphs::initializeGL()
{
  glClearColor(0.3, 0.3, 0.3, 1.0);
}

void MJ_AutomationGraphs::resizeGL(int w, int h)
{
  QGLWidget::resizeGL(w, h);
  glViewport (0, 0, w, h);
}

void MJ_AutomationGraphs::paintGL()
{
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  glMatrixMode (GL_PROJECTION);

  glLoadIdentity ();
  gluOrtho2D(0.0, width(), 0.0, height());

  glMatrixMode (GL_MODELVIEW);

  glPushAttrib(~0);

  glPointSize(3);
  glLineWidth(1);

  {
    // Draw some crap:
    glColor3f(1.0, 1.0, 0.0);
    
    glBegin(GL_LINE_STRIP);
    
    vertex2f(MJ_AutomationSingle::Time().setSeconds(0.0), 0.0);
    vertex2f(MJ_AutomationSingle::Time().setSeconds(30.0), 1.0);
    vertex2f(MJ_AutomationSingle::Time().setSeconds(100.0), 0.5);
    
    glEnd();
  }

  MJ_ModuleAutomation * automations = m_host->automation();

  glColor3f(1.0, 1.0, 1.0);

  for(MJ_ModuleAutomation::iterator it = automations->begin();
      it != automations->end(); it++) {

    MJ_AutomationSingle::container * graph = & (*it).second.graph();

    if(!graph->size())
      continue;

    MJ_AutomationSingle::iterator node = graph->lower_bound(m_left);

    if((*node).first > m_left && node != graph->begin())
      node--;

    glBegin(GL_LINE_STRIP);

    // Put line to the left of the graph (before first point)
    if((*node).first > m_left)
      vertex2d(m_left, (*node).second);
    
    // Draw the curve that fits inside the window.
    while((*node).first < m_right && node != graph->end()) {
      vertex2d((*node).first, (*node).second);
      // MJ_trace2("MJ_AutomationGraphs::paintGL # %lf, %lf", x, y);
      node++;
    }
    
    if(node == graph->end()) {
      // Fill the right side if the curve ran out before the window.
      node--;
      if((*node).first < m_right)
	vertex2d(m_right, (*node).second);
    }
    else
      vertex2d((*node).first, (*node).second);
      
    glEnd();

    // Render the knots with points:

    glBegin(GL_POINTS);
    
    node = graph->lower_bound(m_left);

    if((*node).first > m_left && node != graph->begin())
      node--;

    while((*node).first < m_right && node != graph->end()) {
      double y = (*node).second;
      vertex2d((*node).first, y);
      // MJ_trace2("MJ_AutomationGraphs::paintGL # %lf, %lf", x, y);
      node++;
    }

    glEnd();
    

  }

  if(m_showHit) {

    glBegin(GL_LINE_STRIP);

    int x, y;
    realToWindow(m_hitTime, m_hitVal, &x, &y);

    y = height() - y;

    int radius = 4;

    glVertex2f(x - radius, y - radius);
    glVertex2f(x + radius, y - radius);
    glVertex2f(x + radius, y + radius);
    glVertex2f(x - radius, y + radius);
    glVertex2f(x - radius, y - radius);
    
    glEnd();
  }

  glPopAttrib();
}

/** Map window coordinates to real coordinates. */

void MJ_AutomationGraphs::windowToReal
(int wx, int wy,MJ_AutomationSingle::Time *t, double * ry)
{
  *ry = (height() - wy) / (double) height();
  *t = (m_right - m_left) * (wx / (double) width()) + m_left;
}

MJ_AutomationSingle::Time MJ_AutomationGraphs::windowToRealX(int wx)
{
  return (m_right - m_left) * (wx / (double) width()) + m_left;
}

void MJ_AutomationGraphs::realToWindow
(MJ_AutomationSingle::Time t, double ry, int * wx, int *wy)
{
  *wy = height() - (int) (ry * height());
  *wx = (int) 
    (width() * (t - m_left).secondsD() / (m_right - m_left).secondsD());
}

void MJ_AutomationGraphs::vertex2f(const MJ_AutomationSingle::Time &t, float v)
{
  glVertex2f
    ((width() * (t - m_left).secondsD() / (m_right - m_left).secondsD()),
     v * height());
	     
}

void MJ_AutomationGraphs::vertex2d(const MJ_AutomationSingle::Time &t,double v)
{
  glVertex2d
    ((width() * (t - m_left).secondsD() / (m_right - m_left).secondsD()),
     v * height());
}


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_ModuleAutomation_QG::MJ_ModuleAutomation_QG(MJ_RtControl *rtparent,
					       MJ_ModuleDsp *module,
					       QWidget      *parent,
					       const char   *name)
  : MJ_ModuleDspGuiPostCard(rtparent, module, true, 0, 
			    parent == 0  ? MJ_GUI_OWN_WINDOW : MJ_GUI_POSTCARD,
			    parent, name),
    m_automation(dynamic_cast<MJ_ModuleAutomation *> (module)),
    m_graphs(new MJ_AutomationGraphs(this, this, "Graph display"))
{
  if(!m_automation)
    MJ_fatal(MJ_ERR_ARGUMENT, 
	     "MJ_ModuleAutomation_QG::MJ_ModuleAutomation_QG # m_automation == null");
  
  m_graphs->setMinimumSize(50, 50);

  m_layout->expand(1, 1);
  m_layout->addMultiCellWidget(m_graphs, 0, 0, 0, 2);

  // m_layout->activate();

  m_graphs->show();
}

MJ_ModuleAutomation_QG::~MJ_ModuleAutomation_QG()
{}
  
MJ_ModuleDspGui *MJ_ModuleAutomation_QG::createNew(MJ_RtControl *rtparent,
						   MJ_ModuleDsp *module,
						   QWidget *parent, 
						   const char *name)
{
  return new MJ_ModuleAutomation_QG(rtparent, module, parent, name);
}
