/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#ifndef MJ_MODULE_AUTOMATION_H
#define MJ_MODULE_AUTOMATION_H

#include <mj_module_dsp.h>
#include <mj_module_dsp_gui_default.h>

#include <map>
#include <string>

#include <qgl.h>

class MJ_AutomationSingle
{
public:

  /** Time stamp. In practice a 64 bit integer. 24 LSB are fractions
      of second and 40 MSB are full seconds. */

  class Time
  {
  public:
    Time() {}
    Time(llong t) : m_time(t) {}

    Time & setSeconds(double v) 
    { m_time = (llong) (v * 0xFFFFFF); return *this;}

    double secondsD() const { return m_time / (double) 0xFFFFFF; }

    llong stamp() const { return m_time; }
    void makeMaximum() { m_time = 0x7FFFFFFF; }

    bool operator <  (const Time & that) const{ return m_time < that.m_time; }
    bool operator >  (const Time & that) const{ return m_time > that.m_time; }
    bool operator <= (const Time & that) const{ return m_time <= that.m_time; }
    bool operator >= (const Time & that) const{ return m_time >= that.m_time; }
    bool operator == (const Time & that) const{ return m_time == that.m_time; }

    Time operator - (const Time & that) const{ return m_time - that.m_time; }
    Time operator + (const Time & that) const{ return m_time + that.m_time; }
    Time operator * (double x) const{ return Time((llong) (m_time * x)); }

  protected:
    llong m_time;

  };

  typedef std::map<Time, double> container;
  typedef container::iterator iterator;

  MJ_AutomationSingle() { m_current = m_graph.begin(); }

  uint size() const { return m_graph.size(); }

  /// Seeks a given time-stamp
  void seek(const Time &t)
  {
    m_current = m_graph.lower_bound(t);

    if(m_current == m_graph.begin() || !m_graph.size())
      return;

    if((*m_current).first.stamp() >= t.stamp())
      m_current--;
  }

  /** Returns the interpolated value associated with time t. */

  double value(const Time &t)
  {
    // check that there is more than one control point
    if(m_graph.size() <= 1)
      return (*m_graph.begin()).second;

    // First seek the proper interval:

    iterator plusOne = m_current;
    plusOne++;

    while((*plusOne).first.stamp() < t.stamp() && plusOne != m_graph.end())
      plusOne++;
    
    m_current = plusOne;

    // Is at end:
    if((*plusOne).first.stamp() < t.stamp())
      return (*plusOne).second;

    // Move "current" iterator;
    m_current--;
    
    // Then interpolate the value:

    llong interval = (*plusOne).first.stamp() - (*m_current).first.stamp();

    llong distance = t.stamp() - (*m_current).first.stamp();

    double w1 = (double) distance / (double) interval;

    return (1.0 - w1) * (*m_current).second + w1 * (*plusOne).second;
  }

  container & graph() { return m_graph; }

  void addPoint(Time t, double v) { m_graph[t] = v; }

  //  void add

protected:

  container m_graph;
  iterator  m_current;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** Automation module. */
class MJ_ModuleAutomation : public MJ_ModuleDsp
{
public:
  
  typedef std::map<std::string, MJ_AutomationSingle> container;
  typedef container::iterator iterator;


  MJ_ModuleAutomation();
  virtual ~MJ_ModuleAutomation();

  virtual void dspProcessPeriod();
  
  static MJ_ModuleDsp *createNew();

  container & automations() { return m_automations; }

  iterator begin() { return m_automations.begin(); }
  iterator end()   { return m_automations.end(); }

protected:

  container m_automations;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

class MJ_ModuleAutomation_QG;

class MJ_AutomationGraphs : public QGLWidget
{
  Q_OBJECT
public:
  MJ_AutomationGraphs(MJ_ModuleAutomation_QG * host,
		      QWidget      *parent = 0,
		      const char   *name = 0);
  virtual ~MJ_AutomationGraphs();

protected:
  virtual void mousePressEvent ( QMouseEvent * e );
  virtual void mouseReleaseEvent ( QMouseEvent * e );
  virtual void mouseMoveEvent ( QMouseEvent * e );
    
  virtual void initializeGL ();
  virtual void resizeGL ( int width, int height );
  virtual void paintGL ();

  void windowToReal(int wx, int wy, MJ_AutomationSingle::Time *t, double * ry);
  MJ_AutomationSingle::Time windowToRealX(int wx);
  void realToWindow(MJ_AutomationSingle::Time t, double ry, int * wx, int *wy);

  void vertex2f(const MJ_AutomationSingle::Time &t, float v);
  void vertex2d(const MJ_AutomationSingle::Time &t, double v);

  MJ_ModuleAutomation_QG * m_host;
  // The left and right border in seconds:
  MJ_AutomationSingle::Time m_left, m_right;

  std::string m_activeParam;

  MJ_AutomationSingle::Time m_hitTime;
  double                    m_hitVal;
  bool                      m_showHit;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

class MJ_ModuleAutomation_QG : public MJ_ModuleDspGuiPostCard
{
  Q_OBJECT
public:
  MJ_ModuleAutomation_QG(MJ_RtControl *rtparent,
			 MJ_ModuleDsp *module,
			 QWidget      *parent = 0,
			 const char   *name = 0);
  virtual ~MJ_ModuleAutomation_QG();
  
  static MJ_ModuleDspGui *createNew(MJ_RtControl *rtparent,
				    MJ_ModuleDsp *module,
				    QWidget *parent, 
				    const char *name);

  MJ_ModuleAutomation * automation() { return m_automation; }
protected:
  MJ_ModuleAutomation * m_automation;
  MJ_AutomationGraphs * m_graphs;
};

#endif
