/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_compressor.h>

#include <mj_mapper.h> 
#include <mj_math.h>
#include <mj_parameters.h> 
#include <mj_strs.h>

const char * MJ_ParamsCompressorKeys[MJ_PARCOMP_SIZEOF] = {
  MJ_Strs::threshold,
  MJ_Strs::ratio,
  MJ_Strs::attack
};

MJ_ModuleCompressor::MJ_ModuleCompressor()
  : m_threshold(-10.0),
    m_ratio(2.0),
    m_attack(5),
    m_syncChannels(false)
{
  dspSetInChannels(1);
  dspSetOutChannels(1);
}

MJ_ModuleCompressor::~MJ_ModuleCompressor()
{
  dspStop();
}
  
bool MJ_ModuleCompressor::dspPrepare()
{
  uint inchans = dspInChannels();

  if(!inchans || inchans != (uint) dspOutChannels()) {
    error(MJ_ERR_INTERNAL, "MJ_ModuleCompressor::dspPrepare");
    return false;
  }

  uint samples = m_signalInput.get(0).samples().size();

  MJ_Allocator *allocator = mjDspAllocator();

  m_compressVect.customResize(allocator, samples);
  m_compressVect.setAll(0.0);

  m_lastSamples.customResize(allocator, inchans);
  m_lastSamples.setAll(0.0);

  MJ_Sample sr = dspSamplingRate();

  m_diffMul = sr * 0.3;

  m_levelFilters.resize(inchans);

  for(uint i =0; i < inchans; i++) {
    MJ_FilterAP1<MJ_Sample> *f = &m_levelFilters.get(i);
    f->resetMemory();
  }

  redesignFilters();

  m_compressGains.resize(inchans);

  // trace2("MJ_ModuleCompressor::dspPrepare # %d Hz", dspSamplingRate());
  
  return true;
}
 
void MJ_ModuleCompressor::dspProcessPeriod()
{
  if(m_syncChannels) processSync();
  else processNonSync();
}

bool MJ_ModuleCompressor::dspStop()
{  
  MJ_Allocator *allocator = mjDspAllocator();

  m_compressVect.customResize(allocator, 0);
  m_lastSamples.customResize(allocator, 0);
  return true;
}

bool MJ_ModuleCompressor::msgProcess(const char *id, MJ_Message *m)
{
  MJ_ParamsCompressor pc = (MJ_ParamsCompressor)
    MJ_findStr(id, MJ_ParamsCompressorKeys, MJ_PARCOMP_SIZEOF);
  
  if((int) pc == -1) {
    error(MJ_ERR_ARGUMENT, "MJ_ModuleCompressor::msgProcess");
    return false;
  }

  bool ok(true);

  MJ_Sample val = m->toDouble(&ok);

  if(!ok) {
    return false;
  }

  switch (pc) {
  case MJ_PARCOMP_TRESHOLD:
    m_threshold = val;
    break;
  case MJ_PARCOMP_RATIO:
    m_ratio = val;
    break;
  case MJ_PARCOMP_ATTACK:
    m_attack = val;
    redesignFilters();
    break;
  case MJ_PARCOMP_SIZEOF:
  default:
    break;
  }
  return true;
}

MJ_Parameter * MJ_ModuleCompressor::msgGetInput(const char *id)
{
  MJ_ParamsCompressor pc = (MJ_ParamsCompressor)
    MJ_findStr(id, MJ_ParamsCompressorKeys, MJ_PARCOMP_SIZEOF);

  if(pc == MJ_PARCOMP_TRESHOLD) {
    return new MJ_ParameterDouble
      (MJ_Strs::Threshold, MJ_Strs::threshold, m_threshold, -80.0, 10, 
       MJ_UNIT_DECIBEL);
  }
  else if(pc == MJ_PARCOMP_RATIO) {
    MJ_ParameterDouble *pd = new MJ_ParameterDouble
      (MJ_Strs::Ratio, MJ_Strs::ratio, m_ratio, 1.0, 50);
    pd->setMapper(new MJ_MapperLog);
    return pd;
  }
  else if(pc == MJ_PARCOMP_ATTACK) {
    MJ_ParameterDouble *pd = new MJ_ParameterDouble
      (MJ_Strs::Attack, MJ_Strs::attack, m_attack, 1, 300, 
       MJ_UNIT_MILLISECONDS);
    pd->setMapper(new MJ_MapperLog);
    return pd;
  }
  return 0;
}

QCString MJ_ModuleCompressor::msgGetInputName(int i)
{
  QCString s;
  if(i < MJ_PARCOMP_SIZEOF) s = MJ_ParamsCompressorKeys[i];
  return s;
}

void MJ_ModuleCompressor::processSync()
{
  m_compressVect.setAll(0.0);

  const uint chans   = m_signalInput.size();
  const uint samples = m_signalInput.get(0).samples().size();
  uint i;

  m_compressVect.setAll(0.0);

  if(m_lastSamples.size() != chans) {
    m_lastSamples.customResize(mjDspAllocator(), chans);
  }

  if(m_compressVect.size() != samples) {
    m_compressVect.customResize(mjDspAllocator(), samples);
  }

  // Determine amplitudes sums
  for(i=0; i < chans; i++) {
    
    const MJ_Sample *inSamples = m_signalInput.getFast(i).samples().data();

    MJ_FilterAP1<MJ_Sample> *f = &m_levelFilters.get(i);
    
    MJ_Sample *last = & m_lastSamples.get(i);

    for(uint j=0; j < samples; j++) {
      
      MJ_Sample newSample = inSamples[j];

      MJ_Sample energy = MJ_abs(newSample) + MJ_abs(newSample - *last);
      
      MJ_Sample level = f->putGetSample(energy); // Linear level
      
      m_compressVect.get(j) += level;
    }
  }
  // Map maximum amplitudes to gain values:

  for(i=0; i < samples; i++) {
    MJ_Sample level = m_compressVect.get(i);
    
    if(level < 1e-10) level = 1e-10;
    else level = MJ_absToDb(level);  // Log level.
      
    MJ_Sample gain;

    if(level <= m_threshold) gain = 1.0;
    else {
      MJ_Sample compress = ( m_threshold - level) * (1.0 - 1.0 / m_ratio);
      
      gain = MJ_dbToAbs(compress);
    }
    m_compressVect.getFast(i) = gain;
  }

  // Calculate output.
  MJ_Vector<MJ_Sample>::vmlBegin();

  for(i=0; i < chans; i++) {
    const MJ_Vector<MJ_Sample> *invect = & m_signalInput.get(i).samples();
    MJ_Vector<MJ_Sample> *outvect = & m_signalOutput.get(i).samples();
    invect->mulVml(m_compressVect, outvect); 
  }
  MJ_Vector<MJ_Sample>::vmlEnd();
}

void MJ_ModuleCompressor::processNonSync()
{
  const uint chans   = m_signalInput.size();
  const uint samples = m_signalInput.get(0).samples().size();

  for(uint i=0; i < chans; i++) {

    const MJ_Sample *inSamples = m_signalInput.getFast(i).samples().data();
    MJ_Sample *outSamples = m_signalOutput.getFast(i).samples().data();
    
    MJ_Sample *last = & m_lastSamples.get(i);

    MJ_FilterAP1<MJ_Sample> *f = &m_levelFilters.get(i);

    MJ_Sample gain; // Stupid g++ warnings

    for(uint j=0; j < samples; j++) {
      MJ_Sample newSample = inSamples[j];

      MJ_Sample energy = MJ_abs(newSample) + MJ_abs(newSample - *last);

      MJ_Sample level = f->putGetSample(energy); // Linear level
      
      if(level < 1e-10) level = 1e-10;
      else level = MJ_absToDb(level);  // Log level.
      
      if(level <= m_threshold) gain = 1.0;
      else {
	MJ_Sample compress = ( m_threshold - level) * (1.0 - 1.0 / m_ratio);
	
	gain = MJ_dbToAbs(compress);
      }
      outSamples[j] = newSample * gain;
    }

    m_compressGains.getFast(i) = gain;
  }
}

void MJ_ModuleCompressor::redesignFilters()
{
  MJ_Sample sr = dspSamplingRate();

  MJ_Sample rate = (2000.0 / m_attack) / sr; // 1000 * 2 to scale between 0,1

  for(uint i =0; i < m_levelFilters.size(); i++) {
    MJ_FilterAP1<MJ_Sample> *f = &m_levelFilters.getFast(i);
    f->designLowPass(rate);
  }
}
