/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_delay.h>
#include <mj_error.h>
#include <mj_iostream.h>
#include <mj_mapper.h> 
#include <mj_math.h>
#include <mj_module_keys.h>

#include <mj_module_dsp_tui_default.h>
#include <mj_strs.h>
#include <mj_tr.h>

#include <mj_parameters.h>
#include <mj_global.h>
#include <mj_kernel.h>

#include <qobject.h>

static const char *MJ_maxTimeStr = "maxtime";

MJ_ModuleDelay::MJ_ModuleDelay()
  : m_targetTime(100.0),
    m_targetWetRatio(0.5),
    m_targetDryRatio(0.5),
    m_targetFeedBack(0.5),
    m_maxDelayMs(2000.0),
    m_interpSamples(0)
{
  reset();

  m_signalSourceDefs.resize(1);
  m_signalOutput.resize(1);
}

MJ_ModuleDelay::~MJ_ModuleDelay()
{
  // cleanMemory();
}

bool MJ_ModuleDelay::dspPrepare()
{
  uint chans = m_signalInput.size();

  if(!chans || chans != m_signalOutput.size()) return false;

  // Custom memory allocation (NOT)
  const float sr = (float) dspSamplingRate();
  const ulong len = (ulong) (sr * maxTimeMs() / 1000.0) + 1;
  // const ulong bufSize = MJ_RingBuffer<float>::targetSize(len);
  // const ulong bytes = bufSize * sizeof(float);
  
  m_delayBuffers.resize(chans);

  for(uint i=0; i < chans; i++) {
    MJ_RingBuffer2<float> *rb = & m_delayBuffers.get(i);
    rb->resize(len);
    rb->setAll(0.0); 
  }
  m_interpSamples = 0;
  reset();
  return true;
}

void MJ_ModuleDelay::dspProcessPeriod()
{
  ulong samples = m_signalInput.data()->samples().size();
  ulong interpSamples = MJ_min(samples, m_interpSamples);
  uint calcSamples = samples - interpSamples;
  uint chans = m_signalInput.size();
  uint i, j;

  // Interpolated samples
  if(interpSamples) {

    double scale = 1.0 / (double) m_interpSamples;

    double dryRate = (m_targetDryRatio - m_dryRatio) * scale;
    double wetRate = (m_targetWetRatio - m_wetRatio) * scale;
    double feedRate = (m_targetFeedBack - m_feedBack) * scale;
    double weightRate = (1.0 - m_newWeight) * scale;
    
    uint targetTime = m_targetTimeSamples;
    uint currTime = m_timeSamples;

    for(i=0; i < chans; i++) {

      double dryRatio = m_dryRatio;
      double wetRatio = m_wetRatio;
      double feedBack = m_feedBack;
      double newWeight = m_newWeight;

      const MJ_Sample *inb = m_signalInput.get(i).samples().data();
      MJ_Sample *outb = m_signalOutput.get(i).samples().data();
      MJ_RingBuffer2<float> *rb = & m_delayBuffers.get(i);

      for(j=0; j < interpSamples; j++) {

	dryRatio += dryRate;
	wetRatio += wetRate;
	feedBack += feedRate;
	newWeight += weightRate;

	float delay = rb->getNewest(currTime) * (1.0 - newWeight) + 
	  rb->getNewest(targetTime) * newWeight; 
	
	const float inSample = *inb++;
	
	*outb++ = inSample * dryRatio + delay * wetRatio; // output
	rb->getNewest() = 
	  delay * feedBack + inSample; // feedback
	rb->advance();
      }
      if(i == chans - 1) {
	m_dryRatio = dryRatio;
	m_wetRatio = wetRatio;
	m_feedBack = feedBack;
	m_newWeight = newWeight;
      }
    }
    m_interpSamples -= interpSamples;

    if(!m_interpSamples)
      m_timeSamples = m_targetTimeSamples;
  }
  
  // Direct calculation:
  if(calcSamples) {
    
    float dryRatio = m_dryRatio;
    float wetRatio = m_wetRatio;
    float feedBack = m_feedBack;
    uint time = m_timeSamples;

    // printf("PARS = %f %f %f\n", dryRatio, wetRatio, feedBack);

    for(i=0; i < chans; i++) {

      const MJ_Sample *inb = m_signalInput.get(i).samples().data();
      MJ_Sample *outb = m_signalOutput.get(i).samples().data();
      MJ_RingBuffer2<float> *rb = & m_delayBuffers.get(i);
     
      for(j=0; j < calcSamples; j++) {
	
	float delay = rb->getNewest(time);
	
	/* if(MJ_abs(delay) > 0.00001)
	   trace("MJ_ModuleDelay::dspProcessPeriod", time); */
	
	const float inSample = *inb++;
	*outb++ = inSample * dryRatio + delay * wetRatio; // output
	rb->getNewest() = 
	  delay * feedBack + inSample; // feedback
	rb->advance();
      }      
    }
  }
}

bool MJ_ModuleDelay::dspStop()
{
  cleanMemory();

  return true;
}

bool MJ_ModuleDelay::msgProcess(const char * i, MJ_Message *msg)
{
  bool ok = true;

  float val = msg->toDouble(&ok);

  if(!ok) {
    error(MJ_ERR_ARGUMENT, "MJ_ModuleDelay::msgProcess");
    return false;
  }

  ulong interpSamples;
  int sr = dspSamplingRate();

  if(dspIsRealTime()) {
    if((long) msg->timeStamp() < (long) m_currentTime->stamp()) {
      // The number of interpolation samples, slow interpolation
      interpSamples = (ulong) (sr * 0.1) + 1;
    }
    else {
      interpSamples = msg->timeStamp() - m_currentTime->stamp();
    }
  }
  else interpSamples = 0;

  m_interpSamples = interpSamples;

  if(!strcmp(i, MJ_Strs::delay)) {
    m_targetTime = MJ_clampF(val, 1.0, maxTimeMs()); // Limit the range
    m_targetTimeSamples = (int) (m_targetTime * (float) sr / (float) 1000.0);
    return true;
  }
  else if(!strcmp(i, MJ_Strs::feedback)) {
    m_targetFeedBack = MJ_clampF(val, 0.0, 0.99);
    return true;
  }
  else if(!strcmp(i, MJ_Strs::wetratio)) {
    m_targetWetRatio = MJ_clampF(val, 0.0, 1.0);
    m_targetDryRatio = (float) 1.0 - (float) val;
    return true;
  }
  else if(!strcmp(i, MJ_maxTimeStr)) {
    m_maxDelayMs = MJ_clampF(val, 0.1, 10.0) * 1000.0;
    return true;
  }

  m_newWeight = 0.0;
  
  return false;
}

MJ_Parameter * MJ_ModuleDelay::msgGetInput(const char * i)
{
  if(!strcmp(i, MJ_Strs::delay)) {
    return new 
      MJ_ParameterDouble("Delay time", i, 
			 m_targetTime, 1, maxTimeMs(), MJ_UNIT_MILLISECONDS);
  }
  else if(!strcmp(i, MJ_Strs::feedback)) {    
    return new MJ_ParameterDouble("Feedback ratio", i,
				  m_feedBack, 0.0, 0.99, MJ_UNIT_BLANK);
  }
  else if(!strcmp(i, MJ_Strs::wetratio)) {    
    return new MJ_ParameterDouble("Wet ratio", i,
				  m_wetRatio, 0.0, 1.0, MJ_UNIT_BLANK);
  }
  else if(!strcmp(i, MJ_maxTimeStr)) {    
    MJ_ParameterDouble * pd = 
      new MJ_ParameterDouble("Maximum time", i,
			     m_maxDelayMs * 0.001, 0.1, 10.0, MJ_UNIT_SECONDS);
    pd->setMapper(new MJ_MapperLog);
    pd->setFlags(MJ_PFLAG_AFFECTS); // 
    return pd;
  }
  return 0;
}

QCString MJ_ModuleDelay::msgGetInputName(int i)
{
  QCString s;

  if(i == 0) s = MJ_Strs::delay;
  else if(i == 1) s = MJ_Strs::feedback;
  else if(i == 2) s = MJ_Strs::wetratio;
  else if(i == 3) s = MJ_maxTimeStr;

  return s;
}


