/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_dsp_row.h>

#include <mj_module_compressor.h>
#include <mj_module_dsp_row_gui.h>
#include <mj_module_dsp_row_tui.h>
#include <mj_dsp_plugin.h>
#include <mj_error.h>
#include <mj_global.h>
#include <mj_iostream.h>
#include <mj_kernel.h>
#include <mj_parameter.h>
#include <mj_satellite.h>

#include <mj_module_bypass.h>
#include <mj_module_gain.h>
#include <mj_module_impulsetrain.h>
#include <mj_module_level.h>
#include <mj_module_monodelay.h>
#include <mj_module_printer.h>

#include <stdio.h>


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

bool MJ_ModuleDspRow::m_rowDebug = true;

MJ_ModuleDspRow::MJ_ModuleDspRow()
{
  if(m_rowDebug) trace("MJ_ModuleDspRow::MJ_ModuleDspRow");
}

MJ_ModuleDspRow::~MJ_ModuleDspRow()
{
  if(m_rowDebug) trace("MJ_ModuleDspRow::~MJ_ModuleDspRow");
}

MJ_ModuleDspTui *MJ_ModuleDspRow::newTui(MJ_RtControl *rtparent,
					 const char *name)
{
  return new MJ_ModuleDspRow_CT(rtparent, this, name);
}

void MJ_ModuleDspRow::connectModules()
{
  const char *fname = "MJ_ModuleDspRow::connectModules";

  if(m_rowDebug) trace(fname, m_array.size());

  uint i,j;

  char buf[256];

  for(i=1; i < m_array.size(); i++) {

    int source = i-1;

    MJ_ModuleDsp *m1 = m_array.get(source);
    MJ_ModuleDsp *m2 = m_array.get(i);

    // Connect signals if possible:

    while(m2->dspInChannels() !=  m1->dspOutChannels()) {
      if(!source) {
	toCout();
	MJ_fatal(MJ_ERR_INTERNAL, "MJ_ModuleDspRow::connectModules");
      }

      m1 = m_array.get(--source);
    }

    if(m_rowDebug) {
      sprintf(buf, "%s # connecting %d to %d", fname, source, i);
      trace(buf);
    }

    for(j=0; (int) j < m1->dspOutChannels(); j++) {
      if(!m2->dspConnectInput(m1->id(), j, j)) {
	if(m_rowDebug) {
	  sprintf(buf, "%s # bad module index %d input %d", fname, i, j);
	  error(MJ_ERR_INTERNAL, buf);
	}
      }
    }

    // Connect messages if possible:
    
    source = i-1;
    m1 = m_array.get(source);

    MJ_Parameter *p1=0, *p2=0;
    bool loop = true;

    while(loop) {

      for(j=0; (p1 = m1->msgGetOutputIndex(j)) && 
	    (p2 = m2->msgGetInputIndex(j)); j++) {

	if(p1->type() == p2->type()) {

	  if(m_rowDebug) {
	    sprintf(buf, "%s # connecting messages %d to %d, mid = %d", 
		    fname, source, i, j);
	    trace(buf);
	  }

	  // Found something to connect.
	  m2->msgConnectInput(m1->id(), p1->id(), p2->id());
	  loop = false;
	}
	else
	  warning(fname);

	delete p1; 
	p1 = 0;
	delete p2;
	p2 = 0;
      }
      if(p1) delete p1;
      if(p2) delete p2;

      if(!source) loop = false;
      else m1 = m_array.get(--source);
    }
  }
}

void MJ_ModuleDspRow::testRow(int i)
{
  const char * fname = "MJ_ModuleDspRow::testRow";

  clearArray(); 

  appendModule(new MJ_ModuleImpulseTrain);
  appendModule(new MJ_ModuleBypass);
  appendModule(new MJ_ModuleGain);
  appendModule(new MJ_ModuleMonoDelay);
  appendModule(new MJ_ModuleLevel);

  MJ_ModuleDsp *d = 0;

  if(i == 1 && MJ_kernel) {
    d = MJ_kernel->newFunctionalModule
      (QString(MJ_DspPluginFunctionalityKeys[MJ_DPF_AUDIO_OUT]));
    
    if(!d->dspSetInChannels(1)) 
      MJ_fatal(MJ_ERR_INTERNAL, fname);
    
    appendModule(d);
  }
  else
    appendModule(new MJ_ModulePrinter);

  connectModules();
  if(d) {
    char buf[128];
    sprintf(buf, "In chans in dpwe = %d", d->dspInChannels());
    trace(fname, buf);
  }
}

void MJ_ModuleDspRow::benchMark(int sources, int version)
{
  const char * fname = "MJ_ModuleDspRow::benchMark";

  trace2("%s #  %d sources, version = %d", fname, sources, version);

  int i;
  MJ_ModuleDsp *tmp1, *tmp2, *tmp3;
  
  tmp3 = new MJ_ModuleBypass;  // Sum them all.
  tmp3->dspSetInChannels(sources);
  tmp3->dspSetOutChannels(1);
 
  for(i=0; i < sources; i++) {
    tmp1 = new MJ_ModuleImpulseTrain;
    appendModule(tmp1);

    if(version == 0)
      tmp2 = new MJ_ModuleMonoDelay;
    else if(version == 1)
      tmp2 = new MJ_ModuleGain;
    else if(version == 2)
      tmp2 = new MJ_ModuleCompressor;
    else {
      error2(MJ_ERR_ARGUMENT, "%s # Bad benchmark version %d", fname, version);
      return;
    }

    tmp2->dspSetInChannels(1);
    tmp2->dspSetOutChannels(1);

    tmp2->dspConnectInput(tmp1->id(), 0, 0);
    appendModule(tmp2);

    tmp3->dspConnectInput(tmp2->id(), 0, i);
  }

  appendModule(tmp3);

  MJ_ModulePrinter *p = new MJ_ModulePrinter;
  p->setSilence(true);

  appendModule(p);
  p->dspConnectInput(tmp3->id(), 0, 0);
}

void MJ_ModuleDspRow::playbackRow()
{
  const char *fname = "MJ_ModuleDspRow::playbackRow";

  if(!MJ_kernel) MJ_fatal(MJ_ERR_INTERNAL, fname, "No MJ_kernel");

  // File reader.
  MJ_ModuleDsp *mfr = MJ_kernel->newFunctionalModule
    (QString(MJ_DspPluginFunctionalityKeys[MJ_DPF_FILE_IN]));
  
  if(!mfr) MJ_fatal(MJ_ERR_INTERNAL, fname, "No file input module");

  appendModule(mfr);

  MJ_ModuleDsp * mg = new MJ_ModuleGain;

  mg->dspSetInChannels(1);  
  mg->dspSetOutChannels(1);
  mg->dspConnectInput(mfr->id(), 0, 0);

  appendModule(mg);
  appendModule(new MJ_ModuleLevel);

  mfr->dspSetOutChannels(1);
  
  MJ_ModuleDsp *maw = MJ_kernel->newFunctionalModule
    (MJ_DspPluginFunctionalityKeys[MJ_DPF_AUDIO_OUT]);
  
  if(!maw) MJ_fatal(MJ_ERR_INTERNAL, fname, "No audio output module");

  maw->dspSetInChannels(1);
  maw->dspConnectInput(mg->id(), 0, 0);

  appendModule(maw);
  connectModules();
}

void MJ_ModuleDspRow::test()
{
  const char *fname = "MJ_ModuleDspRow::test ";

  MJ_trace(fname);

  MJ_ModuleDspRow r;
  
  r.testRow(1);
  r.runShort();

  MJ_trace(fname, "EXIT");
}

void MJ_ModuleDspRow::runBenchmark
(int modules, int frameSize, int samples, int version)
{  
  // Disable debugging partially.
  MJ_ModuleDsp::setDspDebug(false);
  m_rowDebug = false;

  MJ_ModuleDspRow r;
  
  r.benchMark(modules, version); // (100 * 2 + 2) * 10000 * 30 

  /// Set the processing period:
  MJ_TimeStamp ts(frameSize);
  r.dspSetTimer(&ts);

  r.runShort(samples / frameSize);
}
