/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_dsp_row_tui.h>

#include <mj_dsp_plugin_loader.h>
#include <mj_error.h>
#include <mj_global.h>
#include <mj_iostream.h>
#include <mj_kernel.h>

#include <typeinfo>

const char * MJ_DspRowTuiCommandKeys[MJ_DRT_SIZEOF] = {
  "append",
  "edit",
    "load",
    "panic",
    "save",
    "start",
    "stop",
    "up"
};

MJ_ModuleDspRow_CT::MJ_ModuleDspRow_CT(MJ_RtControl *parent,
				       MJ_ModuleDsp *module,
				       const char *name)
  : MJ_ModuleDspTuiDefault(parent, module, name)
{
  const char *fname = "MJ_ModuleDspRow_CT::MJ_ModuleDspRow_CT";

  if(!MJ_kernel) MJ_fatal(MJ_ERR_INTERNAL, fname, "Needs MJ_kernel");

  m_dspRow = dynamic_cast<MJ_ModuleDspRow *> (module);

  if(!m_dspRow) MJ_fatal(MJ_ERR_ARGUMENT, fname);

  while(doEvent())
    ;
}

MJ_ModuleDspRow_CT::~MJ_ModuleDspRow_CT()
{}
  
bool MJ_ModuleDspRow_CT::doEvent()
{
  /* const char * fname = "MJ_ModuleDspRow_CT::doEvent";
     trace(fname); */

  describeRow();

  uint i;

  MJ_cout() << "keywords: [  ";

  for(i=0; i < MJ_DRT_SIZEOF; i++) {
    MJ_cout() << MJ_DspRowTuiCommandKeys[i] << "  ";
  }

  MJ_cout() << "]\n";
  


  MJ_cout() << "Command > ";
  MJ_cout().device()->flush();

  QCString cmd;
  MJ_cin() >> cmd;

  MJ_DspRowTuiCommand c = (MJ_DspRowTuiCommand) 
    MJ_findStr(cmd, MJ_DspRowTuiCommandKeys, MJ_DRT_SIZEOF);
  
  switch(c) {
  case MJ_DRT_APPEND_MODULE:
    appendModule();
    break;

  case MJ_DRT_EDIT_MODULE:
    edit();
    break;
    
  case MJ_DRT_LOAD: 
    load();  // Parent class
    break;

  case MJ_DRT_PANIC:
    panic(); // Never returns
    
  case MJ_DRT_SAVE:
    save();  // Parent class
    break;

  case MJ_DRT_START:
    startThreaded();  // Parent class
    break;

  case MJ_DRT_STOP:
    stopPlaying();  // Parent class
    break;

  case MJ_DRT_UP: 
    return false;

  default:
    inputWarning(cmd); // Parent class
    break;
  }
  return true;
}


void MJ_ModuleDspRow_CT::describeRow()
{
  const char * fname = "MJ_ModuleDspRow_CT::describeRow";
  // trace(fname);

  char buf[512];

  MJ_Array<MJ_ModuleDsp *> *modules = &m_dspRow->moduleArray();

  sprintf(buf, "\nDSP row has %d modules. Modules are:\n", 
	  (int) m_dspRow->moduleArray().size());
  
  MJ_cout() << buf; 
  
  QString s;
  for(uint i=0; i < modules->size(); i++) {
    MJ_ModuleDsp * m = modules->get(i);
    if(!m) MJ_fatal(MJ_ERR_INTERNAL, fname);

    if(m->dbe()) {
      s = m->dbe()->objectName();
    }
    else 
      s = typeid(*m).name();

    MJ_cout() << i + 1 << ".  " << s << "\n";
  }
}

void MJ_ModuleDspRow_CT::appendModule()
{
  if(isRealTime()) {
    MJ_cout() << "Cannot add modules while operating";
    return;
  }

  QCString cmd;
  bool ok = true;

  // Display available plugins:
  MJ_cout() << "Available plugins:\n";
  MJ_kernel->dspLoader()->entryList().writeParams(&MJ_cout());
  MJ_cout() << "Choice > ";
  MJ_cout().device()->flush();

  MJ_cin() >> cmd;

  int choiceIndex = cmd.toInt(&ok) - 1;

  if(!ok) {
    inputWarning(cmd); 
    return;
  }
  
  MJ_PluginDbeDsp *dbe =
    MJ_kernel->dspLoader()->getEntry(choiceIndex);
  
  if(!dbe) {
    MJ_cout() << "Bad index\n";
    return;
  }

  MJ_ModuleDsp *m = dbe->newPlugin();
  
  if(!m) {
    MJ_cout() << "Could not create module\n";
    return;
  }
  
  m_dspRow->appendModule(m);
  m_dspRow->connectModules();
}

void MJ_ModuleDspRow_CT::panic()
{
  if(MJ_kernel) MJ_kernel->panic();
  MJ_fatal(MJ_ERR_UNKNOWN, "MJ_ModuleDspRow_CT::panic");
}

void MJ_ModuleDspRow_CT::edit()
{
  QCString cmd;
  bool ok = true;

  MJ_cin() >> cmd;
  int choiceIndex = cmd.toInt(&ok) - 1;
  
  if(!ok) {
    inputWarning(cmd); 
    return;
  }
  
  MJ_ModuleDsp *m = m_dspRow->moduleArray().get(choiceIndex);

  if(!m) {
    MJ_cout() << "Could not access module " << choiceIndex << "\n";
    return;
  }
  
  MJ_ModuleDspTui *mt = m->newTui(this);

  delete mt;
}
