/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

/* Include the header of this file. */ 
#include <mj_module_gain.h>

#include <mj_error.h>
#include <mj_math.h>
#include <mj_parameters.h>
#include <mj_strs.h>
#include <mj_tr.h>

#include <qfile.h> 
#include <qtextstream.h> 
#include <qobject.h> 

#include <math.h>

bool MJ_ModuleGain::m_debug = false;

MJ_ModuleGain::MJ_ModuleGain()
  : m_gain(1.0),
    m_targetTime(0),
    m_targetGain(1.0)
{
  /* Make one input and one output channels (default). */
  m_signalSourceDefs.resize(1);
  m_signalOutput.resize(1);
}

MJ_ModuleGain::~MJ_ModuleGain()
{}

bool MJ_ModuleGain::dspPrepare()
{
  const char * fname = "MJ_ModuleGain::dspPrepare";

  m_targetTime = 0;

  /* Check that the number of input channels equals the number of
     output channels. */
  if(m_signalInput.size() != m_signalOutput.size()) {

    error2(MJ_ERR_INTERNAL, "%s # I/O range mismatch %d != %d", 
	   fname, m_signalInput.size(), m_signalOutput.size());
    return false;
  }

  if(!m_signalInput.size()) {
    error(MJ_ERR_INTERNAL, fname, "No channels");    
    return false;    
  }

  m_ramp.resize(m_signalInput.getFast(0).samples().size());

  if(m_debug) trace2("%s # cycle samples = %d", fname, m_ramp.size());

  return true;
}

void MJ_ModuleGain::dspProcessPeriod()
{
  if(m_targetTime) {
    // Interpolate
    
    // First create ramp vector.
    long until = m_targetTime - (long) m_currentTime->stamp();

    MJ_Sample rate = (m_targetGain - m_gain) / (MJ_Sample) until;
    
    uint rsize = m_signalInput.getFast(0).samples().size();
    
    if(m_ramp.size() < rsize) 
      m_ramp.resize(rsize * 2); // This might be enough also in the next round.

    if(until >= (long) rsize) {
      m_ramp.ramp(m_gain, rate);
      m_gain = m_ramp.getFast(rsize - 1) + rate;
    }
   
    else {
      m_ramp.ramp(0, until, m_gain, rate);
      if(until < (long) (rsize - 1)) {
	m_ramp.setRange(m_targetGain, until + 1, (uint) -1);
      }
      m_gain = m_targetGain;
      m_targetTime = 0;
    }

    // Use the ramp vector for smooth interpolation
    MJ_Vector<MJ_Sample>::vmlBegin();
    
    for(uint i=0; i < m_signalInput.size(); i++) {
      
      const MJ_Vector<MJ_Sample> *invect = & m_signalInput.get(i).samples();
      MJ_Vector<MJ_Sample> *outvect = & m_signalOutput.get(i).samples();
      
      invect->mulVml(m_ramp, outvect); // Vector elementwise multiplication
    }
    
    MJ_Vector<MJ_Sample>::vmlEnd();

  }
  else {
    // No interpolation
    MJ_Vector<MJ_Sample>::vmlBegin();
    
    for(uint i=0; i < m_signalInput.size(); i++) {
      
      const MJ_Vector<MJ_Sample> *in = & m_signalInput[i].samples();
      MJ_Vector<MJ_Sample> *out = & m_signalOutput[i].samples();
      
      in->mulVml(m_gain, out); // Scalar multiplication
      // in->mul(m_gain, out); // Scalar multiplication
    }
    
    MJ_Vector<MJ_Sample>::vmlEnd();
  }
}

bool MJ_ModuleGain::msgProcess(const char * id, MJ_Message *m)
{
  const char *fname = "MJ_ModuleGain::msgProcess";
  bool ok = true;

  /* Optionally trace program execution. */
  if(m_debug) trace(fname);
  
  if(!id) {
    if(m_debug) error(MJ_ERR_ARGUMENT, fname, "NULL ID");
    return false;
  }

  if(!strcmp(id, MJ_Strs::gain)) {
    
    /* Transform the message to double regardless of its type.  If the
       cast fails (eg. the message cannot be turned to numeric
       value) the boolean argument is modified. */
    double tmp = m->toDouble(&ok);
    
    if(m_debug) {
      char buf[128];
      sprintf(buf, "%s - Value = %f", fname, tmp);
      trace(buf);
    }

    if(ok) {
      if(m_debug) trace(fname, "Success 1");
      /* Calculate gain coefficient. It is assumed the input value 
       was expressed in the decibel scale. */
      m_targetGain = MJ_dbToAbs(tmp);

      if(dspIsRealTime()) {

	long mstamp = m->timeStamp();
	long cstamp = (long) m_currentTime->stamp();

	if(mstamp > cstamp)
	  m_targetTime = mstamp;
	else
	  m_targetTime = cstamp + 
	    (long) ((float) dspSamplingRate() / (float) 30.0) + 1;
      }
      else
	m_gain = m_targetGain;
    }
    else if(m_debug) 
      trace(fname, "Failure 1");
  }
  else ok = false;

  return ok;
}

MJ_Parameter * MJ_ModuleGain::msgGetInput(const char *id)
{
  if(m_debug) trace("MJ_ModuleGain::msgGetInput");

  if(!id) return 0;
  
  if(!strcmp(id, MJ_Strs::gain)) { 

    /* This module has only one parameter (gain) that can be edited
       via the message interface. The following three lines declare
       the parameter and give its limiting values (minimum/maximum),
       plaintext name (for GUI) and units (of general interest, also
       for GUI). */
    MJ_ParameterDouble *pd = new MJ_ParameterDouble;
    pd->set(MJ_Strs::Gain, MJ_Strs::gain,
	    MJ_absToDb(m_targetGain), -50, 25, MJ_UNIT_DECIBEL);
    return pd;
  }
  return 0;
}

QCString MJ_ModuleGain::msgGetInputName(int i)
{
  QCString s;
  if(i == 0) s = MJ_Strs::gain;
  return s;
}
  
void MJ_ModuleGain::testStreamIo()
{
  const char *filename="foo.gain";

  MJ_ModuleGain m1, m2;
  m1.setGain(0.6565);

  {
    QFile f1(filename);
    f1.open(IO_WriteOnly);
    QTextStream s1(&f1);
    m1.write(&s1);
  }
  MJ_trace("GAIN WRITTEN");

  {
    QFile f2(filename);
    f2.open(IO_ReadOnly);
    QTextStream s2(&f2);
    m2.read(&s2);
  }
  m2.toCout();  
}
