/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

/* Include the header of this file. */ 
#include <mj_module_gain_multi.h>

#include <mj_error.h>
#include <mj_math.h>
#include <mj_parameters.h>
#include <mj_strs.h>

bool MJ_ModuleGainMulti::m_debug = false;

MJ_ModuleGainMulti::MJ_ModuleGainMulti()
{
  /* Make one input and one output channels (default). */
  m_signalSourceDefs.resize(1);
  m_signalOutput.resize(1);
}

MJ_ModuleGainMulti::~MJ_ModuleGainMulti()
{}

bool MJ_ModuleGainMulti::dspPrepare()
{
  const char * fname = "MJ_ModuleGainMulti::dspPrepare";

  /* Check that the number of input channels equals the number of
     output channels. */
  if(m_signalInput.size() != m_signalOutput.size()) {

    error2(MJ_ERR_INTERNAL, "%s # I/O range mismatch %d != %d", 
	   fname, m_signalInput.size(), m_signalOutput.size());
    return false;
  }

  if(!m_signalInput.size()) {
    error(MJ_ERR_INTERNAL, fname, "No channels");    
    return false;    
  }

  m_gains.resize(m_signalInput.size());

  if(m_debug) trace2("%s # cycle samples = %d", fname, m_gains.size());

  return true;
}

void MJ_ModuleGainMulti::dspProcessPeriod()
{  
  for(uint i=0; i < m_signalInput.size(); i++) {
    
    const MJ_Vector<MJ_Sample> *in = & m_signalInput[i].samples();
    MJ_Vector<MJ_Sample> *out = & m_signalOutput[i].samples();

    MJ_Ramp<MJ_Sample> *ramp = & m_gains[i];

    if(ramp->isActive()) {
      // Interpolate
      ramp->advanceLinearMul(out->data(), in->data(), in->size());
    }
    else {
      // No interpolation
      MJ_Vector<MJ_Sample>::vmlBegin();
      in->mulVml(ramp->get(), out);   // Scalar multiplication
      MJ_Vector<MJ_Sample>::vmlEnd();
    }
  }
}

bool MJ_ModuleGainMulti::msgProcess(const char * id, MJ_Message *m)
{
  const char *fname = "MJ_ModuleGainMulti::msgProcess";
  bool ok = true;

  /* Optionally trace program execution. */
  if(m_debug) trace(fname);
  
  if(!id) {
    if(m_debug) error(MJ_ERR_ARGUMENT, fname, "NULL ID");
    return false;
  }

  if(strncmp(id, MJ_Strs::gain, 4))
     return false;
  
  if(id[4] != '_')
    return false;
     
  double tmp = m->toDouble(&ok);
    
  if(m_debug) {
    trace2("%s # Value = %f", fname, tmp);
  }

  int index = atoi(id+5);

  expand(index+1);
    
  if(ok) {
    if(m_debug) trace(fname, "Success 1");
    /* Calculate gain coefficient. It is assumed the input value 
       was expressed in the decibel scale. */
    double gainTarget = MJ_dbToAbs(tmp);

    if(dspIsRealTime()) {

      long mstamp = m->timeStamp();
      long cstamp = (long) m_currentTime->stamp();
      uint interpTime;

      if(mstamp > cstamp)
	interpTime = mstamp - cstamp;
      else
	interpTime = (long) ((float) dspSamplingRate() / (float) 30.0) + 1;
      
      m_gains[index].setTarget(interpTime, gainTarget, MJ_RAMP_LINEAR);
    }
    else 
      m_gains[index].reset(gainTarget);
  }
  else ok = false;

  return ok;
}

MJ_Parameter * MJ_ModuleGainMulti::msgGetInput(const char *id)
{
  const char * fname = "MJ_ModuleGainMulti::msgGetInput";

  if(m_debug) trace(fname);

  if(!id) {
    MJ_fatal(MJ_ERR_ARGUMENT, fname);
  }
  
  if(!strncmp(id, MJ_Strs::gain, 4)) { 

    if(id[4] != '_')
      return false;
    
    int index = atoi(id+5);

    expand(index+1);
    
    char buf[32];
    sprintf(buf, "Gain %d", index + 1);

    return new MJ_ParameterDouble
      (buf, id, MJ_absToDb(m_gains[index].target()), -70, 25, MJ_UNIT_DECIBEL);
  }
  return 0;
}

QCString MJ_ModuleGainMulti::msgGetInputName(int i)
{
  QCString s;

  if((uint) i < m_signalInput.size()) {
    s.sprintf("gain_%d", i);
  }
  return s;
}
  
/** Expand the gain array to contain n elements. N should equal or
    exceed the number of channels. */
void MJ_ModuleGainMulti::expand(uint n)
{
  uint oldSize = m_gains.size();

  if(n <= oldSize)
    return;

  m_gains.resize(n);

  for(uint i=oldSize; i < n; i++)
    m_gains[i].reset(1.0);
}
