/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_level.h>

#include <mj_arraytmpl.C>

#include <mj_error.h>
#include <mj_math.h>
#include <mj_tr.h>

#include <qlabel.h> 
#include <qlayout.h> 
#include <qpushbutton.h> 

MJ_ModuleLevel::MJ_ModuleLevel()
  : // m_peak(0),
    m_timeMs(30.0),
    m_peakHoldMs(1500.0)
{ 
  m_signalSourceDefs.resize(1);
  m_levels.resize(1);

  zeroLevels();
}

MJ_ModuleLevel::~MJ_ModuleLevel()
{}

bool MJ_ModuleLevel::dspPrepare()
{
  if((uint)m_signalInput.size() != m_signalSourceDefs.count())  
    MJ_fatal(MJ_ERR_INTERNAL, "MJ_ModuleLevel::dspPrepare");

  m_levels.resize(m_signalInput.size());
  
  /* m_levels.adoptData(mjDspMalloc(m_levels.dataSize()),
     m_levels.size()); */

  
  zeroLevels();

  // m_peak = 0.0;

  MJ_Sample sr = dspSamplingRate();

  m_timeSamples = (uint) (sr * m_timeMs / 1000.0) + 1;
  m_peakHoldSamples = (uint) (sr * m_peakHoldMs / 1000.0) + 1;
  
  return true;
}

void MJ_ModuleLevel::dspProcessPeriod()
{
  // MJ_Sample peak = m_peak;

  for(uint i=0; i < m_signalInput.size(); i++) {
    
    const MJ_Vector<MJ_Sample> *in = & m_signalInput.getFast(i).samples();
    const MJ_Sample * data = in->dataConst();

    const uint l = in->size();
    const MJ_Sample *sentinel = & data[l];
    
    //MJ_LevelMeter<MJ_Sample> & uis1 = m_levels.get(i);
    //MJ_LevelMeter<MJ_Sample> & uis2 = m_levels.get(i + m_signalInput.size());

    Level * level = & m_levels.getFast(i);

    while(data < sentinel) {
      
      level->put(*data++, m_timeSamples, m_peakHoldSamples);
      /* MJ_Sample vol = MJ_abs();

      uis1.put(vol, m_timeSamples);
      uis2.put(vol, m_peakHoldSamples);*/
      
      // if(peak < vol) peak = vol;
    }
  }
  // m_peak = peak;
}

bool MJ_ModuleLevel::dspStop()
{
  zeroLevels();

  return true;
}

MJ_ModuleDspGui *MJ_ModuleLevel::newGui
(MJ_RtControl *rtparent, MJ_GuiSize s, QWidget *parent, const char *name)
{
  return new MJ_ModuleLevel_QG(rtparent, s, this, parent, name);
}

void MJ_ModuleLevel::zeroLevels()
{
  Level meter;
  m_levels.setAll(meter);
}

template class MJ_Array<MJ_ModuleLevel::Level>;

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

bool MJ_ModuleLevel_QG::m_debug = false;

MJ_ModuleLevel_QG::MJ_ModuleLevel_QG(MJ_RtControl *rtparent,
				     MJ_GuiSize    size,
				     MJ_ModuleDsp *module,
				     QWidget      *parent,
				     const char   *name)
  : MJ_ModuleDspGui(rtparent, module, parent, name),
    m_size(size),
    m_peak(0),
    m_reset(0)
{
  if(m_debug) trace("MJ_ModuleLevel_QG::MJ_ModuleLevel_QG");
  
  connect(&m_timer, SIGNAL(timeout()),
	  this, SLOT(update()));

  MJ_ModuleLevel *ml = dynamic_cast<MJ_ModuleLevel *> (module);
  
  if(!ml) MJ_fatal(MJ_ERR_ARGUMENT, "MJ_ModuleLevel_QG::MJ_ModuleLevel_QG");
  
  draw();
  realTimeChange(isRealTime());
  update();
}

MJ_ModuleLevel_QG::~MJ_ModuleLevel_QG()
{}

int MJ_ModuleLevel_QG::weight()
{
  return 40;
}

void MJ_ModuleLevel_QG::update()
{
  const char * fname = "MJ_ModuleLevel_QG::update";

  // if(m_debug) trace("MJ_ModuleLevel_QG::update");

  MJ_ModuleLevel *ml = dynamic_cast<MJ_ModuleLevel *> (m_module);
  if(!ml) MJ_fatal(MJ_ERR_ARGUMENT, fname);
  
  lock();
  uint i, s = ml->levels().size();
  unlock();
  
  m_levels.resize(s);

  lock();
  for(i=0; i < s; i++) 
    m_levels[i].copyPeaks(ml->levels().get(i));
  unlock();

  uint bars = m_levels.size();
  
  if(!bars) {
    // MJ_fatal(MJ_ERR_INTERNAL, fname);
    return;
  }

  if(bars != (uint) m_ldpy->nofBars())
    m_ldpy->setNofBars(bars);

  // float * levels = m_levels.data();

  m_levelDb.resize(bars * 2);

  float * dbs = m_levelDb.data();
  
  for(i=0; i < bars; i++) {

    Level * level = & m_levels[i];

    MJ_Sample tmp  = level->m_level;;
    MJ_Sample tmp2 = level->m_currentPeak;

    MJ_Sample db = MJ_absToDb(tmp);
    MJ_Sample peak = MJ_absToDb(tmp2);
    
    /* qDebug("MJ_ModuleLevel_QG::update %d %f -> %f %f", 
       i, m_levels.getConst(i), db, peak); */

    MJ_Sample dbLevel = db / 60 + 1.0;
    MJ_Sample peakLevel = peak / 60 + 1.0;

    
    if(tmp2 < 1e-8)
      peakLevel = -1.0;

    if(tmp < 1e-8)
      dbLevel = -1.0;

    dbs[i] = peakLevel;
    dbs[i + bars] = dbLevel;
  }
  
  m_ldpy->set(dbs, dbs + bars);

  if(m_peak) {

    QString tmp(MJ_Tr::tr("Peak:")), tmp2;

    lock();
    MJ_Sample peak = ml->peak();
    unlock();

    if(peak > 1e-10)
      tmp2.sprintf("%.2f", (float) MJ_absToDb(peak));
    else 
      tmp2 = "-inf";
    
    if(m_size == MJ_GUI_OWN_WINDOW)
      tmp += " " + MJ_Tr::tr(MJ_UnitKeysShort[MJ_UNIT_DECIBEL]) + " " + tmp2;
    else
      tmp = tmp2;
  
    if(tmp != m_peak->text())
      m_peak->setText(tmp);
    
    if(isVisible())
      m_peak->show();
  }
}

void MJ_ModuleLevel_QG::resetPeak()
{
  lock();
  ((MJ_ModuleLevel *) m_module)->resetPeak();
  unlock();

  update();
}

void MJ_ModuleLevel_QG::realTimeChange (bool nowRt)
{
  if(nowRt) m_timer.start(50);
  else m_timer.stop();
}

void MJ_ModuleLevel_QG::resizeDisplays(uint s)
{
  if(m_debug) trace("MJ_ModuleLevel_QG::resizeDisplays", s);
}
  
void MJ_ModuleLevel_QG::draw()
{
  if(m_debug) trace("MJ_ModuleLevel_QG::draw");
  
  uint bars = m_levels.size() / 2;
  
  resizeDisplays(bars);
  
  QVBoxLayout * vbox = new QVBoxLayout(this, 3, 3);
  
  m_ldpy = new MJ_LevelDisplay(this, 0, MJ_LevelDisplay::BottomUp, 
			       (MJ_LevelDisplay::Clickable|
				MJ_LevelDisplay::FadeToBackground));
  connect(m_ldpy, SIGNAL(clicked()), this, SLOT(showModule()));
  m_ldpy->setMinimumSize(30, 50);
  vbox->addWidget(m_ldpy, 80);

  if(m_size != MJ_GUI_STAMP) {
    if(m_debug) trace("MJ_ModuleLevel_QG::draw WITH PEAK");
    QHBoxLayout *tmplayout = new QHBoxLayout(vbox);

    m_peak = new QLabel(this);
    tmplayout->addWidget(m_peak, 50, AlignLeft);

    m_reset = new QPushButton
      (MJ_Tr::tr(m_size == MJ_GUI_OWN_WINDOW ? "Reset" : "R"), this);
    
    tmplayout->addWidget(m_reset, 50, AlignRight);
    m_reset->show();
    connect(m_reset, SIGNAL(pressed()), this, SLOT(resetPeak()));
  }
}

template class MJ_Array<MJ_ModuleLevel_QG::Level>;
