/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_midiseq.h>

#include <mj_dsp_plugin.h>
#include <mj_error.h>
#include <mj_gui_utils.h>
#include <mj_message_carrier.h>
#include <mj_message_string.h>

#include <mj_parameters.h>
#include <mj_strs.h>
#include <mj_tr.h>

#include <qapplication.h>
#include <qcheckbox.h> 
#include <qfiledialog.h> 
#include <qlabel.h> 
#include <qlayout.h> 
#include <qmenubar.h> 
#include <qpainter.h> 
#include <qpopupmenu.h> 
#include <qscrollview.h> 

bool MJ_ModuleMidiseq::m_debug = false;

MJ_ModuleMidiseq::MJ_ModuleMidiseq()
{
  if(m_debug) trace("MJ_ModuleMidiseq::MJ_ModuleMidiseq");
}

MJ_ModuleMidiseq::~MJ_ModuleMidiseq()
{
  if(m_debug) trace("MJ_ModuleMidiseq::~MJ_ModuleMidiseq");
}

bool MJ_ModuleMidiseq::dspPrepare()
{
  if(m_debug) trace("MJ_ModuleMidiseq::dspPrepare");

  m_currEvents.clear();
  m_sequence.prepare();
  return m_currentTime != 0; // We need explicit timing info
}

void MJ_ModuleMidiseq::dspProcessPeriod()
{
  const char * fname = "MJ_ModuleMidiseq::dspProcessPeriod";
  
  if(!m_currentTime) MJ_fatal(MJ_ERR_INTERNAL, fname);
  
  m_currEvents.clear();
  m_sequence.getEvents(&m_currEvents, m_currentTime->stamp());

  uint max = m_currEvents.count();

  for(uint i=0; i < max; i++) {
    
    // m_currEvents.at(i)->writeCout(); // debug
    m_eventMsg.midiMsg() = * m_currEvents.at(i);
    msgPushOutput(MJ_Strs::MIDI, & m_eventMsg);

    if(m_debug) trace(fname, "One event");
  }
}

bool MJ_ModuleMidiseq::msgProcess(const char * id, MJ_Message *m)
{
  const char * fname = "MJ_ModuleMidiseq::msgProcess";

  if(m_debug) trace(fname, id);

  if(!strcmp(id,MJ_Strs::filename)) {
    
    MJ_MessageString *fileName = dynamic_cast<MJ_MessageString *> (m);
    if(!fileName) {
      if(m_debug) 
	error(MJ_ERR_ARGUMENT, fname);

      return false;
    }
    
    return readFile(fileName->stringConst());
  }
  return false;
}

MJ_Parameter *MJ_ModuleMidiseq::msgGetInput(const char * id)
{
  if(!strcmp(id,MJ_Strs::filename)) {
    MJ_ParameterString *ps = new MJ_ParameterString(m_sequence.fileName());
    if(!ps) return 0;
    ps->setName(MJ_Tr::tr("File name"));
    ps->setId(MJ_Strs::filename);
    return ps;
  }
  return 0;
}

QCString MJ_ModuleMidiseq::msgGetInputName(int i)
{
  QCString s;
  if(i == 0) s = MJ_Strs::filename;
  return s;
}


MJ_Parameter *MJ_ModuleMidiseq::msgGetOutput(const char * i)
{
  if(!strcmp(MJ_Strs::MIDI, i)) {
    MJ_ParameterMidi *pm = new MJ_ParameterMidi();
    pm->setName("MIDI events");
    pm->setId(i);
    return pm;
  }  
  return 0;  
}

QCString MJ_ModuleMidiseq::msgGetOutputName(int i)
{
  QCString s;
  if(i == 0) s = MJ_Strs::MIDI;
  return s;
}


MJ_ModuleDspGui *MJ_ModuleMidiseq::newGui
(MJ_RtControl *rtparent, MJ_GuiSize s, QWidget *parent, const char *name)
{
  if(m_dbe) return m_dbe->newGui(rtparent, this, s, parent, name);
  
  if(s != MJ_GUI_STAMP)
    return new MJ_ModuleMidiseq_QG(rtparent, this, parent, name);
  
  return MJ_ModuleDsp::newGui(rtparent, s, parent, name);
}


bool MJ_ModuleMidiseq::readFile(const QString &fileName)
{
  if(!m_sequence.readSequence(fileName)) {

    if(m_debug) error(MJ_ERR_ARGUMENT, 
		      "MJ_ModuleMidiseq::readFile",
		      fileName);

    return false;
  }
  return true;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


MJ_MidiView::MJ_MidiView(QWidget      *parent,
			 const char   *name)
  : QWidget(parent, name),
    m_midiSeq(0)
{}

MJ_MidiView::~MJ_MidiView()
{}

void MJ_MidiView::setSequence(MJ_MidiSequence *s)
{ 
  m_midiSeq = s; 
  if(s) draw();
}

void MJ_MidiView::paintEvent(QPaintEvent *e)
{
  (void) e;

  if(*MJ_ModuleMidiseq::debugPtr()) 
    MJ_trace("MJ_MidiView::paintEvent");

  int h = height();
  
  if(!m_midiSeq) return;

  QPainter paint(this);

  paint.eraseRect(0,0,width(), h);

  for(uint i=0; i < m_midiSeq->size(); i++) {
    MJ_MidiTrack *tr = & m_midiSeq->get(i);

    for(uint j=0; j < tr->count(); j++) {
      
      lock();
      MJ_MidiEvent *me = tr->at(j);

      if(me->isNoteOn()) {
      
	int ticks = me->ticks();
	int key = me->keyNumber();
	int x2, x1 = ticks / 8;
	int h1 = h - key;
      
	MJ_MidiEvent *noteOff = tr->findNoteOff(me);
	unlock();
	
#if 1
	if(noteOff) x2 = noteOff->ticks() / 8;
	else x2 = x1 + 10;
#else
	x2 = x1 + 10;
#endif
	paint.drawLine(x1, h1, x2, h1);
      }
      else
	unlock();
    }
  }
}

void MJ_MidiView::draw()
{
  lock();
  uint w = m_midiSeq->endTicks() / 8;
  unlock();

  setMinimumSize(w, 128);
  
  paintEvent(0);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_ModuleMidiseq_QG::MJ_ModuleMidiseq_QG(MJ_RtControl *rtparent,
					 MJ_ModuleDsp *module,
					 QWidget      *parent,
					 const char   *name)
  : MJ_ModuleDspGui(rtparent, module, parent, name),
    m_midiseq(0),
    m_midiView(0)
{
  MJ_ModuleMidiseq *mp = dynamic_cast<MJ_ModuleMidiseq *> (module);
  if(!mp) {
    error(MJ_ERR_ARGUMENT, "MJ_ModuleMidiseq_QG::MJ_ModuleMidiseq_QG");
    return;
  }
  m_midiseq = mp;

  if(m_midiView) {
    m_midiView->setSequence(& m_midiseq->sequence());
  }
  else
    draw();
}

MJ_ModuleMidiseq_QG::~MJ_ModuleMidiseq_QG()
{}

void MJ_ModuleMidiseq_QG::moduleModified(const char *id) 
{
  if(strcmp(id, MJ_Strs::filename) == 0)
    m_midiView->draw();
}

void MJ_ModuleMidiseq_QG::test(QApplication *xpQA, int iArgc, char **cppArgv)
{
  (void) iArgc;
  (void) cppArgv;

  MJ_ModuleMidiseq p;
  p.readFile("bach.mid");
  
  MJ_ModuleMidiseq_QG pg(0, &p);

  xpQA->setMainWidget(&pg);
  pg.show();
  xpQA->exec();
}

MJ_ModuleDspGui *MJ_ModuleMidiseq_QG::createNew(MJ_RtControl *rtparent,
						MJ_ModuleDsp *module,
						QWidget *parent, 
						const char *name)
{
  return new MJ_ModuleMidiseq_QG(rtparent, module, parent, name);
}

void MJ_ModuleMidiseq_QG::open()
{
  QString fileName = QFileDialog::getOpenFileName();
  
  if(fileName.isEmpty()) return;

  lock();
  QString oldFileName = m_midiseq->sequence().fileName();
  unlock();

  QFileInfo fi(fileName);
  if(!fi.exists()) {
    return;
  }

  fileName = fi.dirPath() + "/" + fi.fileName();
  
  // if(m_debug)
  MJ_trace("Trying to load MIDI file", fileName);
  
  if(!m_midiseq->readFile(fileName)) {
    // error(MJ_ERR_ARGUMENT, "MJ_ModuleMidiseq_QG::open", fileName);
    
    return;
  }
  
  // Undo info:
  { 
    MJ_MessageString *m1 = new MJ_MessageString();
    MJ_MessageString *m2 = new MJ_MessageString();
    
    m1->setString(oldFileName);
    m2->setString(fileName);

    MJ_MessageCarrier mi1(MJ_Strs::filename, m1), 
      mi2(MJ_Strs::filename, m2);
    addUndoInfo(&mi2, &mi1, "File name");
  }

  m_midiseq->satellite()->modified(MJ_Strs::filename);
}

void MJ_ModuleMidiseq_QG::draw()
{
  QHBoxLayout *l = new QHBoxLayout(this);

  QPopupMenu *tests = new QPopupMenu(this);
  tests->insertItem( MJ_Tr::tr("&Open..."), this, SLOT(open()));
  QMenuBar *menu = new QMenuBar( this );
  menu->insertItem( MJ_Tr::tr("&File"), tests );

  l->setMenuBar(menu);

  QScrollView *scroll = new QScrollView(this);
  l->addWidget(scroll, 50);

  scroll->setMinimumSize(250, 128);

  m_midiView = new MJ_MidiView(this);
  m_midiView->setRtParent(this);
  scroll->addChild(m_midiView);
  m_midiView->setSequence(& m_midiseq->sequence());
}
