/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_audio_conf.h>
#include <mj_dsp_plugin.h>
#include <mj_error.h>
#include <mj_global.h>
#include <mj_gui_utils.h>
#include <mj_iostream.h>
#include <mj_kernel.h>
#include <mj_math.h>
#include <mj_messages.h>
#include <mj_module_dsp_row_gui.h>
#include <mj_module_level.h>
#include <mj_parameters.h>
#include <mj_strs.h>

#include <mj_module_midiseq.h>
#include <mj_module_minisynth.h>

#include <qapplication.h>
#include <qlayout.h> 
#include <qlabel.h> 
#include <qlcdnumber.h> 


#include <math.h>
#include <stdlib.h>


MJ_SimpleNote::MJ_SimpleNote()
  : m_gain(1.0),
    m_phase(0),
    m_phaseVel(0),
    m_channel(0)
{}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

bool MJ_ModuleMiniSynth::m_debug = false;

MJ_ModuleMiniSynth::MJ_ModuleMiniSynth()
  : m_activeNotes(0),
    m_eventCount(0),
    m_gain(0.005)
{
  m_sampleBuffer.resize(100); // Size will we 128.
  m_sampleBuffer.setAll(0.0);

  double pi = atan2(1.0, 1.0) * 4.0;
  double cycle = m_sampleBuffer.size();

  for(uint i = 0; i < m_sampleBuffer.size(); i++) {
    double phase = (double) i / cycle * pi * 2.0;
    double tmp = cos(phase);
    double sign;

    if(tmp > 0) sign = 1.0;
    else sign = -1.0;

    m_sampleBuffer.getIndex(i) = (MJ_Sample) sqrt(MJ_abs(tmp)) * sign;
  }
  
  m_signalOutput.resize(1); // Mono output.
}

MJ_ModuleMiniSynth::~MJ_ModuleMiniSynth()
{}

bool MJ_ModuleMiniSynth::dspPrepare()
{
  m_activeNotes = 0;
  m_eventCount = 0;
  return m_signalOutput.size() == 1;
}

void MJ_ModuleMiniSynth::dspProcessPeriod()
{
  MJ_Vector<MJ_Sample> *out = & m_signalOutput[0].samples();
  
  out->setAll(0.0);

  float bufSize = (float) m_sampleBuffer.size();

  for(int i=0; i < m_activeNotes; i++) {
    MJ_SimpleNote *note = & m_notes[i];

    for(uint j=0; j < out->size(); j++) {

      note->m_phase += note->m_phaseVel;
      
      if(note->m_phase > bufSize) {
	note->m_phase -= bufSize;
      }
      
#if 1
      out->get(j) += (m_sampleBuffer.getIndex((ulong)note->m_phase) *
		      note->m_gain);
#endif

    }
  }
  out->mulThis(m_gain);
}

bool MJ_ModuleMiniSynth::dspStop()
{
  m_activeNotes = 0;
  if(m_debug)
    MJ_trace("MJ_ModuleMiniSynth::dspStop # event count", (int)m_eventCount);
  return true;
}

MJ_ModuleDspGui *MJ_ModuleMiniSynth::newGui
(MJ_RtControl *rtparent, MJ_GuiSize s, QWidget *parent, const char *name)
{
  QWidget *p = 0;

  if(s != MJ_GUI_OWN_WINDOW)
    p = parent;

  return new MJ_ModuleMiniSynth_QG(rtparent, this, p, name);
}


double MJ_ModuleMiniSynth::gainDb() const 
{ 
  return MJ_absToDb(m_gain);
}

void MJ_ModuleMiniSynth::setGainDb(double db) 
{ 
  if(db > 30) db = 30.0;
  m_gain = MJ_dbToAbs(db);
}

bool MJ_ModuleMiniSynth::msgProcess(const char *id, MJ_Message *m)
{
  const char *fname = "MJ_ModuleMiniSynth::msgProcess";
  
  if(m_debug) trace(fname, id);
  
  if(!strcmp(MJ_Strs::MIDI, id)) {
    
    MJ_MessageMidiEvent *mm = dynamic_cast<MJ_MessageMidiEvent *> (m);
    
    if(!mm) return false;
    
    m_eventCount++;
    
    MJ_MidiEvent *me = & mm->midiMsg();
    
    if(m_debug) trace(fname, 2);
    
    if(me->isNoteOn()) 
      noteOn(me);
    else if(me->isNoteOff())
      noteOff(me);
    else if(m_debug) {
      MJ_cout() << (*me);
      
      // me->write(&cout);
    }
    return true;
  }
  return false;
}

MJ_Parameter *MJ_ModuleMiniSynth::msgGetInput(const char * i)
{
  if(!strcmp(MJ_Strs::MIDI, i)) {
    MJ_ParameterMidi *pm = new MJ_ParameterMidi();
    pm->setName("MIDI input");
    pm->setId(i);
    return pm;
  }  
  return 0;
}


QCString MJ_ModuleMiniSynth::msgGetInputName(int i)
{
  QCString s;
  if(i == 0) s = MJ_Strs::MIDI;
  return s;
}

void MJ_ModuleMiniSynth::noteOn(MJ_MidiEvent *e)
{
  const char * fname = "MJ_ModuleMiniSynth::noteOn";

  if(m_debug) trace(fname);

  if(m_activeNotes == MJ_MINISYNTH_MAX_POLY) return;

  int keyNum   = e->keyNumber();
  int velocity = e->noteVelocity();
  
  MJ_SimpleNote *note = & m_notes[m_activeNotes];

  note->m_phase = 0.0;
  note->m_phaseVel = (float) MJ_MidiEvent::keyNumberToHz(keyNum) * 
    (float) m_sampleBuffer.size() / 
    (float) dspSamplingRate();

  note->m_gain = (double) velocity / 256.0;

  note->m_channel = e->channel();
  note->m_keyNumber = keyNum;
  
  m_activeNotes++;

  if(m_debug || true) {

    float gainSum = 0;
    for(int i=0; i < m_activeNotes; i++)
      gainSum += m_notes[i].m_gain;

    char buf[128];

    sprintf(buf, "gain = %f # phaseVel = %f # gainSum = %f",
	    note->m_gain, note->m_phaseVel, gainSum);
    trace(fname, buf);
  }
}

void MJ_ModuleMiniSynth::noteOff(MJ_MidiEvent *e)
{
  const char *fname = "MJ_ModuleMiniSynth::noteOff";

  if(m_debug) trace(fname);

  int n = findNote(e);
  if(n >= 0) {
    removeNote(n);
  }
  else {
    error(MJ_ERR_INTERNAL, fname, "note not found");
  }
}

void MJ_ModuleMiniSynth::removeNote(int n)
{
  if(m_debug) trace("MJ_ModuleMiniSynth::removeNote");

  for(int i = n; i+1 < MJ_MINISYNTH_MAX_POLY; i++) {
    m_notes[i] = m_notes[i+1];
  }
  m_activeNotes--;
}

int MJ_ModuleMiniSynth::findNote(MJ_MidiEvent *e)
{
  for(int i=0; i < m_activeNotes; i++) {
    MJ_SimpleNote *note = & m_notes[i];
    if((note->m_channel == e->channel()) &&
       (note->m_keyNumber == e->keyNumber()))
      return i;
  }
  return -1;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_ModuleMiniSynth_QG::MJ_ModuleMiniSynth_QG(MJ_RtControl *rtparent,
					     MJ_ModuleDsp *module,
					     QWidget      *parent,
					     const char   *name)
  : MJ_ModuleDspGui(rtparent, module, parent, name),
    m_lcd(0),
    m_synth(0)
{
  connect(&m_timer, SIGNAL(timeout()),
	  this, SLOT(update()));
  m_timer.start(100);

  MJ_ModuleMiniSynth *ms = dynamic_cast<MJ_ModuleMiniSynth *> (module);
  if(!ms) return;

  m_synth = ms;
  draw();
}

MJ_ModuleMiniSynth_QG::~MJ_ModuleMiniSynth_QG()
{}


void MJ_ModuleMiniSynth_QG::test(QApplication *qa, int argc, char **argv)
{
  const char *fname = "MJ_ModuleMiniSynth_QG::test";

  MJ_Kernel k;

  bool audioOut = false;
  bool rtMidi = false;

  static const char * audioOpt = "--audio";
  static const char * midiOpt = "--rtmidi";

  char buf[128];

  for(int i=0; i < argc; i++) {
    if(strcmp(argv[i], audioOpt) == 0)
      audioOut = true;
    if(strcmp(argv[i], midiOpt) == 0)
      rtMidi = true;
  }
  
  MJ_trace(fname);

  if(!audioOut) {
    sprintf(buf, "Use %s to get real-time audio output", audioOpt);
    MJ_trace(buf);
  }

  if(!rtMidi) {
    sprintf(buf, "Use %s to get real-time MIDI input", midiOpt);
    MJ_trace(buf);
  }


  MJ_ModuleDspRow dr;

  MJ_ModuleDsp *seq=0;
  
  if(!rtMidi) {
    MJ_ModuleMidiseq * mms = new MJ_ModuleMidiseq;
    mms->readFile("bach.mid");
    seq = mms;
  }
  else {
    seq = MJ_kernel->newFunctionalModule
      (QString(MJ_DspPluginFunctionalityKeys[MJ_DPF_MIDI_IN]));
    if(!seq) {
      MJ_fatal(MJ_ERR_INTERNAL, fname, "No MIDI in plugin");
    }
  }

  dr.appendModule(seq);

  MJ_ModuleMiniSynth *synth = new MJ_ModuleMiniSynth;
  MJ_ModuleLevel *level     = new MJ_ModuleLevel;

  dr.appendModule(synth);
  dr.appendModule(level);

  level->dspConnectInput(synth->id(), 0, 0);

  bool ok = synth->msgConnectInput(seq->id(), MJ_Strs::MIDI, MJ_Strs::MIDI);
  
  if(!ok) {
    MJ_fatal(MJ_ERR_UNKNOWN, fname, "Failed to connect");
  }

  if(audioOut) {
    MJ_ModuleDsp *ao = MJ_kernel->newFunctionalModule
      (QString(MJ_DspPluginFunctionalityKeys[MJ_DPF_AUDIO_OUT]));

    if(!ao) {
      MJ_fatal(MJ_ERR_INTERNAL, fname, "No audio out plugin");
    }

    QString dev(MJ_AudioConf::audioDevice(MJ_ADT_OUTPUT, 0));
    if(!dev.isEmpty())
      ao->msgProcessString(0, dev);

    MJ_trace(fname, "Using audio out");

    if(!ao->dspSetInChannels(1)) 
      MJ_fatal(MJ_ERR_INTERNAL, "MJ_ModuleDspRow::testRow");
    
    ao->dspConnectInput(synth->id(), 0, 0);
    dr.appendModule(ao);
  }
  
  MJ_ModuleDspRow_QG rg(0, &dr, 0,"minisynth test");
  rg.setTestMode(true);

  qa->setMainWidget(&rg);
  qa->exec();
}

MJ_ModuleDspGui *MJ_ModuleMiniSynth_QG::createNew(MJ_RtControl *rtparent,
						  MJ_ModuleDsp *module,
						  QWidget *parent, 
						  const char *name)
{
  return new MJ_ModuleMiniSynth_QG(rtparent, module, parent, name);
}

void MJ_ModuleMiniSynth_QG::update()
{
  if(!m_synth || !m_lcd) return;

  lock();
  int notes = m_synth->activeNotes();
  unlock();

  m_lcd->display(notes);
}

void MJ_ModuleMiniSynth_QG::allNotesOff()
{
  lock();
  m_synth->allNotesOff();
  unlock();
}

void MJ_ModuleMiniSynth_QG::setVolume(double d)
{
  lock();
  m_synth->setGainDb(d);
  unlock();
}

void MJ_ModuleMiniSynth_QG::getVolume(double *d)
{
  lock();
  *d = m_synth->gainDb();
  unlock();
}

void MJ_ModuleMiniSynth_QG::draw()
{
  QGridLayout *l = new QGridLayout(this, 3, 3, 5);

  QLabel *label = new QLabel(this);
  label->setText("Notes");
  MJ_limMinSize(label);
  l->addWidget(label, 0, 0);

  m_lcd = new QLCDNumber(this, "note count");
  m_lcd->setNumDigits(3); // Enough for 128 notes
  m_lcd->display(0);
  MJ_limMinSize(m_lcd);
  l->addWidget(m_lcd, 0, 1);
  
  m_gainSlider.createObjs(this);
  m_gainSlider.label()->setText("Gain (db)");
  m_gainSlider.setRange(-40.0, 10.0);

  connect(&m_gainSlider, SIGNAL(emitValue(double)),
	  this, SLOT(setVolume(double)));

  connect(&m_gainSlider, SIGNAL(emitGetValue(double *)),
	  this, SLOT(getVolume(double *)));
  
  double v;
  getVolume(&v);
  m_gainSlider.setValue(v);
  m_gainSlider.limMinSizes();
  m_gainSlider.toGridHorizontal(l, 1, 0);

  MJ_equalStretch(l);
  l->setColStretch(1, 50);
  l->setRowStretch(2, 50);
  l->activate();
}
