/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_mixer.h>

#include <di_sleep.h>
#include <mj_app_action.h>
#include <mj_dsp_plugin.h>
#include <mj_error.h>
#include <mj_global.h>
#include <mj_iostream.h>
#include <mj_kernel.h>
#include <mj_math.h>
#include <mj_messages.h>
#include <mj_message_string.h>
#include <mj_module_keys.h>
#include <mj_module_menubar.h>
#include <mj_module_mixer_gui.h>
#include <mj_module_mixer_source.h>
#include <mj_parameter_numeric.h>
#include <mj_parameters.h>
#include <mj_stream_utils.h>
#include <mj_strs.h>
#include <mj_tr.h>

#include <qtextstream.h>

#include <typeinfo>

const char *MJ_MixerUnitTypes[MJ_MIXER_UNIT_SIZEOF] = {
  "MJ_MIXER_UNIT_BLANK",
  "MJ_MIXER_UNIT_SEND",
  "MJ_MIXER_UNIT_SOURCE",
  "MJ_MIXER_UNIT_MODULEDSP"
};

int MJ_MixerUnit::m_debug = 0;

MJ_MixerUnit::MJ_MixerUnit() 
  : m_host(0), 
  m_obj(0) 
{
  if(m_debug) MJ_trace("MJ_MixerUnit::MJ_MixerUnit", this);
}

MJ_MixerUnit::~MJ_MixerUnit() 
{ 
  if(m_debug) MJ_trace("MJ_MixerUnit::~MJ_MixerUnit", this);
  clear(); 
}
    

void MJ_MixerUnit::set(MJ_ModuleDspGroup *host, MJ_Base *obj)
{
  clear();

  if(obj && !host) MJ_fatal(MJ_ERR_ARGUMENT, "MJ_MixerUnit::set");

  m_host = host;
  m_obj = obj;
}

void MJ_MixerUnit::clear()
{
  if(!m_obj) return;

  MJ_ModuleDsp *m = dynamic_cast<MJ_ModuleDsp *> (m_obj);

  if(m)
    m_host->removeModule(m);
  
  delete m_obj;
  
  m_obj = 0;
}

QWidget *MJ_MixerUnit::newGui(MJ_GuiSize   s, 
			      uint         strip,
			      uint         row,
			      MJ_ModuleMixer_QG  *mixerGui, 
			      QWidget *parent,
			      const char  *name)
{
  const char * fname = "MJ_MixerUnit::newGui";

  if(m_debug) MJ_trace(fname, row);

  if(!m_obj) return new MJ_MixerBlank_QG(mixerGui, strip, row, parent, name);

  MJ_ModuleDsp *mdsp = dynamic_cast<MJ_ModuleDsp *> (m_obj);
  
  QWidget * w = 0;

  if(mdsp) {
    MJ_ModuleDspGui *mg = mdsp->newGui(mixerGui, s, parent, name);
    w = mg;
    mixerGui->menuBar()->undoCenter()->addSatellite(mdsp->satellite());
    
    QObject::connect(mg, SIGNAL(emitShowModule(MJ_ModuleDsp *)),
		     mixerGui, SLOT(showModule(MJ_ModuleDsp *)));
  }
  else {
    MJ_error(MJ_ERR_INTERNAL, fname, "No DSP module");
  }

  if(m_debug && w) MJ_trace(fname, typeid(*w).name());

  return w;
}


bool MJ_MixerUnit::write(QTextStream *s)
{
  if(!m_obj) {
    (*s) << MJ_MixerUnitTypes[MJ_MIXER_UNIT_BLANK] << "\n";
    return true;
  }

  MJ_ModuleDsp *mdsp = dynamic_cast<MJ_ModuleDsp *> (m_obj);
  if(mdsp) {
    (*s) << MJ_MixerUnitTypes[MJ_MIXER_UNIT_MODULEDSP] << "\n";
    return mdsp->write(s);
  }

  return false;
}

/** Read units data from a stream. Creates child objects if possible. */

bool MJ_MixerUnit::read(QTextStream *s)
{
  const char *fname = "MJ_MixerUnit::read";

  if(m_debug) MJ_trace(fname, this);
  
  QCString cstr1, cstr2;

  cstr1 = s->readLine();

  MJ_MixerUnitType type = (MJ_MixerUnitType)
    MJ_findStr(cstr1, MJ_MixerUnitTypes, MJ_MIXER_UNIT_SIZEOF);
  
  if((int) type == -1) {
    if(m_debug) {

      char buf[128];
      // const char *str(0);
      
      sprintf(buf, "%s # Error on line %d", 
	      fname, MJ_currentLine(s->device()));

      if(cstr1.isEmpty())
	MJ_error(MJ_ERR_ARGUMENT, buf, "Empty string");
      else MJ_error(MJ_ERR_ARGUMENT, buf, cstr1);
    }
    return false;
  }
  else if(m_debug) MJ_trace(fname, MJ_MixerUnitTypes[type]);

  MJ_ModuleDsp *m;

  clear();

  if(m_debug) MJ_trace(fname, 2);

  switch(type) {
  case MJ_MIXER_UNIT_BLANK:
    return true;

  case MJ_MIXER_UNIT_MODULEDSP:
    m = MJ_ModuleDsp::readFromStream(s);
    if(!m) {
      if(m_debug) MJ_error(MJ_ERR_ARGUMENT, fname, "Could not load module");
      return false;
    }
    clear();
    
    m->setParentModule(m_host);

    if(dynamic_cast<MJ_ModuleMixerSource *> (m))
      ((MJ_ModuleMixerSource *) m)->makeLegal();
    
    m_obj = m;
    break;
    
  default:
    if(m_debug) MJ_error(MJ_ERR_ARGUMENT, fname, "Bad type");
    return false;
  }
  return true;
}

/** Determines name of the object. */
QString MJ_MixerUnit::objectName()
{
  QString s;

  MJ_ModuleDsp *mdsp = dynamic_cast<MJ_ModuleDsp *> (m_obj);
  
  if(mdsp) {
    if(mdsp->dbe()) {
      s = mdsp->dbe()->objectName();
    }
    else 
      s = typeid(*mdsp).name();
    return s;
  }

  return s;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

bool MJ_ModuleMixerStrip::m_debug = false;

MJ_ModuleMixerStrip::MJ_ModuleMixerStrip()
  : m_mixer(0),
    m_index(-1)
{
  setAutoDelete(true);
}

MJ_ModuleMixerStrip::~MJ_ModuleMixerStrip()
{}

void MJ_ModuleMixerStrip::setMixer(MJ_ModuleMixer *m) 
{ 
  m_mixer = m; 

  uint size = count();

  for(uint i=0; i < size; i++)
    at(i)->setHost(m);
}


/** Returns nth module, or null if there is no module. If the module
    is of type MJ_ModuleMixerSend, then NULL is returned. */
MJ_ModuleDsp *MJ_ModuleMixerStrip::module(uint row)
{
  MJ_MixerUnit *mu = at(row);
  MJ_ModuleDsp *m = dynamic_cast<MJ_ModuleDsp *> (mu->obj());
  if(m && dynamic_cast<MJ_ModuleMixerSend *> (m)) {
    return 0;
  }

  return m;
}

/** Returns nth module, or null if there is no module. */
MJ_ModuleDsp *MJ_ModuleMixerStrip::moduleAny(uint row)
{
  return dynamic_cast<MJ_ModuleDsp *> (at(row)->obj());
}

/// Returns nth send, or null if there is no module.
MJ_ModuleMixerSend *MJ_ModuleMixerStrip::send(uint row)
{
  MJ_MixerUnit *mu = at(row);
  return dynamic_cast<MJ_ModuleMixerSend *> (mu->obj());
}

/// Returns nth source, or null if there is no module.
MJ_ModuleMixerSource *MJ_ModuleMixerStrip::source(uint row)
{
  MJ_MixerUnit *mu = at(row);
  return dynamic_cast<MJ_ModuleMixerSource *> (mu->obj());
}

/** Tries to connect messages between modules. */
bool MJ_ModuleMixerStrip::connectMessages()
{
  uint size = count();

  bool ok = true;

  for(uint i=1; i < size; i++) {
    if(module(i))
      ok = ok && connectMessagesTo(i);
  }
  return ok;
}

bool MJ_ModuleMixerStrip::connectMessagesTo(uint i)
{
  const char * fname = "MJ_ModuleMixerStrip::connectMessagesTo";

  if(!i) return true; // Really, this should not happen

  MJ_ModuleDsp *mto = module(i);
  
  if(!mto) MJ_fatal(MJ_ERR_ARGUMENT, fname);

  bool loop = true;

  for(int j = (int) i-1; (j >= 0) && loop; j--) {
    MJ_ModuleDsp *mfrom = module(j);

    if(!mfrom) continue;

    mto->msgConnectInput(mfrom->id(), 0, 0);
    loop = false;
    
    /*
    MJ_Parameter *p1=0, *p2=0;

    for(uint k=0; (p1 = mfrom->msgGetOutputIndex(k)); k++) {
      for(uint l=0; (p2 = mto->msgGetInputIndex(l)); l++) {
      
	if(p1->type() == p2->type() && 
	   mfrom->msgGetOutputName(k) == mto->msgGetInputName(l)) {
	  
	  if(m_debug) {
	    MJ_trace2
	      ("%s # connecting messages from %s to %s, index = %d -> %d",
	       fname, mfrom->id(), mto->id(), k, l);
	  }
	  
	  // Found something to connect.
	  mto->msgConnectInput(mfrom->id(), p1->id(), p2->id());
	  loop = false;
	}

	delete p2;
	p2 = 0;
      }
      delete p1; 
      p1 = 0;
    }
    if(p1) delete p1;
    if(p2) delete p2;*/
  }

  return true;
}

void MJ_ModuleMixerStrip::exportModules(MJ_ModuleMixer *mixer)
{
  const char * fname = "MJ_ModuleMixerStrip::exportModules";

  if(m_debug) MJ_trace(fname);

  uint size = count();
  
  for(uint i=0; i < size; i++) {
    MJ_ModuleDsp *m = moduleAny(i);
    if(m) {
      if(m_debug) MJ_trace(fname, i);

      mixer->appendModule(m);
    }
  }
}

bool MJ_ModuleMixerStrip::rerouteStripAfter(int row)
{
  if(m_debug) MJ_trace("MJ_ModuleMixerStrip::rerouteStripAfter", row);

  bool ok = true;

  uint size = count();

  // if(!row) row = 1;
  
  for(uint i=row; i < size; i++) {
    if(m_debug) 
      MJ_trace("MJ_ModuleMixerStrip::rerouteStripAfter", at(i)->obj());
    
    if(moduleAny(i))
      ok = ok && rerouteDspInput(i);
    else if(send(i))
      ok = ok && rerouteSend(i);
  }

  return ok;    
}

bool MJ_ModuleMixerStrip::rerouteSend(int row)
{
  const char * fname = "MJ_ModuleMixerStrip::rerouteSend";

  if(m_debug) MJ_trace(fname, row);

  /* Resize send channels... */
  MJ_ModuleMixerSend *s = send(row);
  
  MJ_ModuleDsp *prev = prevDspSource(row);

  if(!s || !prev) {
    MJ_error2(MJ_ERR_INTERNAL, "%s # s = %p prev = %p", fname, s, prev);
    return false;
  }
  int chans = prev->dspOutChannels();
  s->dspSetInChannels(chans);
  s->dspSetOutChannels(chans);

  return true;
}

bool MJ_ModuleMixerStrip::rerouteDspInput(int row)
{
  const char * fname = "MJ_ModuleMixerStrip::rerouteDspInput";

  if(m_debug) MJ_trace(fname, row);

  bool ok = true;

  if(m_debug) MJ_trace(fname, row);

  MJ_ModuleDsp *curr;

  if(row) curr = moduleAny(row);
  else {
    curr = module(row);
    if(dynamic_cast<MJ_ModuleMixerSource *> (curr))
      return true;
  }

  if(!curr) {
    MJ_error(MJ_ERR_INTERNAL, fname, "NULL curr");
    return false;
  }

  MJ_DspOptions *currOpt = curr->dbe()->options();
  
  if(!currOpt) {
    MJ_error(MJ_ERR_INTERNAL, fname, "NULL currOpt");
    return false;
  }

  MJ_ModuleDsp *prev = prevDspSource(row);

  if(!prev) {
    int in = curr->dspInChannels(), out = curr->dspOutChannels();

    bool change = curr->dspAcceptIoChans(&in, &out);

    if(in) {
      m_mixer->error2(MJ_ERR_INTERNAL, " %s NULL prev, but prev needed %d",
		      fname, row);
      return false;
    }
    if(change) {
      curr->dspSetInChannels(in);
      curr->dspSetOutChannels(out);

      MJ_AppActions::moduleIOModified
	(curr, MJ_MIO_SIGNAL_INPUT | MJ_MIO_SIGNAL_OUTPUT);
    }

    return true;
  }
  
  if(!prev->dbe())
     MJ_fatal(MJ_ERR_INTERNAL, fname);
  
  // determine IO conf.
  
  int prevOut = prev->dspOutChannels();
  int in = 0, out = curr->dspOutChannels();

  bool equalIo = false;

  if(currOpt->isDefined(MJ_DSP_EQUAL_CHANNEL_IO)) {
    equalIo = (bool) currOpt->value(MJ_DSP_EQUAL_CHANNEL_IO);
    if(m_debug) MJ_trace2("%s # Equal IO %d", fname, (int) equalIo);
  }
  else
    if(m_debug) MJ_trace2("%s # Non-equal IO %d ", fname, (int) equalIo);

  in = currOpt->newInChans(prevOut);
    
  if(equalIo)
    out = in;
  else
    out = currOpt->newOutChans(out);

  int in1 = in, out1 = out;

  curr->dspAcceptIoChans(&in, &out);

  if(in != in1 && in1 && !prevOut) {
    
    if(m_debug) {
      MJ_trace2("%s # No IO pair # in = %d, in1 = %d, out1 = %d, row = %d",
		fname, in, in1, out1, row);
    }
    ok = false;
  }
  //}
  else {
    curr->dspSetInChannels(in);
    curr->dspSetOutChannels(out);
      
    MJ_AppActions::moduleIOModified
      (curr, MJ_MIO_SIGNAL_INPUT | MJ_MIO_SIGNAL_OUTPUT);

    const char * prevId = prev->id();
      
    if(in && !prevOut) {
      MJ_error(MJ_ERR_INTERNAL, fname, "No source");
      return false;
    }
      
    for(int k=0; k < in; k++) {
      if(!curr->dspConnectInput(prevId, k % prevOut, k)) {
	MJ_error(MJ_ERR_INTERNAL, fname, "Failed to connect");
	ok = false;
      }
    }
  }
  return ok;
}

bool MJ_ModuleMixerStrip::makeDefault()
{
  const char * fname = "MJ_ModuleMixerStrip::makeDefault";

  if(m_debug) MJ_trace(fname, m_index);

  if(!m_mixer || (m_index < 0)) 
    MJ_fatal(MJ_ERR_INTERNAL, fname);

  uint size = count();
  
  if(size < 2) return false;

  MJ_AudioConf *ac = & m_mixer->audioConf();

  int inChans = ac->inputChannels();
  
  MJ_ModuleDsp *m = 0;
  MJ_ModuleMixerSource *ms = 0;

  // Mixer source to the first row.
  if(inChans) {
    
    m = MJ_Kernel::newDspModule(MJ_mixerSourceIdentifier, MJ_PP_IDENTIFIER);

    if(!m) {
      MJ_error(MJ_ERR_INTERNAL, fname, "No module");
      return false; 
    }

    ms = dynamic_cast<MJ_ModuleMixerSource *> (m);

    if(!ms) {
      MJ_error(MJ_ERR_INTERNAL, fname, "Bad module");
      return false; 
    }
    
    ms->dspSetOutChannels(1);
    ms->setMappings(1);
    ms->setMapping(0, MJ_MIXERSOURCE_AUDIO_INPUT, m_index % inChans, 0);
    
    m_mixer->addModule(ms, m_index, 0);
  }
  else {
    if(m_debug)
      MJ_trace(fname, "No input");
    return false;
  }

  // Create a mixer send module to the last row
  m = MJ_Kernel::newDspModule(MJ_mixerSendIdentifier, MJ_PP_IDENTIFIER);
  MJ_ModuleMixerSend *send = dynamic_cast<MJ_ModuleMixerSend *> (m);

  if(!send)
    MJ_fatal(MJ_ERR_UNKNOWN, fname, "No MJ_ModuleMixerSend plugin");

  m_mixer->addModule(send, m_index, size - 1);
  
  int outChans = ac->outputChannels();

  if(outChans) {
    send->resizeMappings(1);
    send->setMapping(0, MJ_MIXERSOURCE_AUDIO_OUTPUT, 0, m_index % outChans);
  }

  if(!ms || (size < 5)) {
    return true;
  }
  
  // Level meter to the second-last row
  m = MJ_Kernel::newDspModule(MJ_dspVisualIdentifier, MJ_PP_IDENTIFIER);
  if(m) m_mixer->addModule(m, m_index, size - 2);

  if(size < 6) return true;

  // Gain controller to the third last row.
  m = MJ_Kernel::newDspModule(MJ_dspGainIdentifier, MJ_PP_IDENTIFIER);
  if(m) {
    m_mixer->addModule(m, m_index, size - 3);
    // Lower gain to eliminate potential feedback problems.
    m->msgProcessDouble(MJ_Strs::gain, -50.0);
  }
  return true;
}

/** Check all mixer sends and make the legal. */

void MJ_ModuleMixerStrip::checkSends()
{
  if(!m_mixer) MJ_fatal(MJ_ERR_INTERNAL, "MJ_ModuleMixerStrip::checkSends");

  uint s = count();

  for(uint i=0; i < s; i++) {
    MJ_ModuleMixerSend *s = send(i);
    if(s) s->makeLegal(m_mixer);
  }
}

/** Check all mixer sources and make the legal. */
void MJ_ModuleMixerStrip::checkSources()
{
  uint s = count();

  for(uint i=0; i < s; i++) {
    MJ_ModuleMixerSource *s = source(i);
    if(s) s->makeLegal();
  }
}

MJ_ModuleDsp *MJ_ModuleMixerStrip::prevDspSource(int row)
{
  for(int i = row - 1; i >= 0; i--) {
    MJ_ModuleDsp *m = module(i);

    if(m) {
      if(m->dspOutChannels())
	return m;
      else if(m_debug)
	MJ_trace2
	  ("MJ_ModuleMixerStrip::prevDspSource # row %d has no outputs", i);
    }
	
  }
  return 0;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_ModuleMixerSatellite::~MJ_ModuleMixerSatellite()
{}

void MJ_ModuleMixerSatellite::remove(int strip, int row)
{
  emit emitRemove(strip, row);
}

void MJ_ModuleMixerSatellite::update(int strip, int row)
{
  emit emitUpdate(strip, row);
}

void MJ_ModuleMixerSatellite::modified2(int i)
{
  emit emitModified2(i);
}

void MJ_ModuleMixerSatellite::modified4(int i,int j,int k)
{
  emit emitModified4(i,j,k);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** Keywords related to the mixer file format. */
const char *MJ_ModuleMixerWords[] = {
  "MIXER_AUDIOS",
    "MIXER_STRIPS",
    "MIXER_ROWS",
    "MIXER_BUSES",
    "MIXER_BUSWIDTH",
    "MIXER_UNIT"
    };

const char * MJ_ModuleMixer::addMidiDevStr = "addMidiDev";
const char * MJ_ModuleMixer::removeMidiDevStr = "removeMididev";

/** Debugging variable. */
bool MJ_ModuleMixer::m_debug = false;

/** Initializes a mixer with 2x2x0 structure. */
MJ_ModuleMixer::MJ_ModuleMixer(bool init)
  : m_mmcController(-1),
    m_rows(0)
{
  const char * fname = "MJ_ModuleMixer::MJ_ModuleMixer";

  if(m_debug) trace(fname);

  if(!MJ_kernel) 
    MJ_fatal(MJ_ERR_INTERNAL, fname, "MJ_kernel must not be NULL");

  updateTimer();

  m_buses = (MJ_ModuleMixerBus *) 
    MJ_Kernel::newDspModule(MJ_busIdentifier, MJ_PP_IDENTIFIER);

  if(!m_buses) 
    MJ_fatal(MJ_ERR_INTERNAL, fname, "Could not create bus module");

  m_buses->setId(MJ_Strs::buses);

  m_strips.setAutoDelete(true);

  m_midiIos.setAutoDelete(true);
  m_audioIos.setAutoDelete(true);
  m_audioOutSums.setAutoDelete(true);

  // Override default saellite type:
  delete m_satellite;
  m_satellite = new MJ_ModuleMixerSatellite(this);
  updateModuleList();
  
  if(init) {
    // Some kind of IO setup

    MJ_AudioConf tmpConf;

    if(!tmpConf.defaultStereoIo())
      tmpConf.defaultMultiChannelIo();

    for(uint i=0; i < tmpConf.numDevices(); i++) {
      MJ_2Int *dev = tmpConf.device(i);

      addAudioDev((MJ_AudioDeviceType) dev->m_v[0], dev->m_v[1]);
    }

    resize(6,6);
    setBuses(4);    
  }
}

MJ_ModuleMixer::~MJ_ModuleMixer()
{
  if(m_debug) trace("MJ_ModuleMixer::~MJ_ModuleMixer");

  allUnitsOk();

  resize(0,0);
  delete m_buses;
  // while(m_moduleList.count()) m_moduleList.take(0);
}
  
bool MJ_ModuleMixer::dspPrepare()
{
  const char * fname = "MJ_ModuleMixer::dspPrepare";

  if(m_debug) trace(fname);

  bool ok(true);

  checkRow0();

  if(!rerouteStrips()) {
    if(m_debug) error(MJ_ERR_UNKNOWN, fname,  "rerouteStrips failed");

    return false;
  }


  checkSends();
  checkSources();

  ok = ok && connectBuses();
  ok = ok && connectSources();
  ok = ok && updateModuleList();

  ok = ok && MJ_ModuleDspArray::dspPrepare();

  int index = m_midiConf.findDevice(MJ_MIDID_INPUT, m_mmcController);
  
  if(index >= 0 && ok) {
    MJ_ModuleDsp * m = m_midiIos.at(index);
    MJ_MessageRemap *remap = new MJ_MessageRemap;
    remap->set(MJ_Strs::MIDI,
	       MJ_Strs::MIDI,
	       m->id(), "../");
    remap->setTarget(this);
    m->msgAddOutputRemap(remap);
  }

  if(!ok) error(MJ_ERR_UNKNOWN, fname, 2);

  return ok;
}

bool MJ_ModuleMixer::dspPrepareFast()
{
  const char * fname = "MJ_ModuleMixer::dspPrepareFast";

  if(m_mmcController >= 0) {
    trace2("%s # Using MMC to trigger processing", fname);

    int index = m_midiConf.findDevice(MJ_MIDID_INPUT, m_mmcController);

    if(index >= 0) {
      MJ_ModuleDsp *m = m_midiIos.at(index);

      m_triggerStart = false;

      DI_Sleep s;
    
      uint rounds = 500;
    
      m->dspPrepareFast(); // Start the module (dirty).

      while(!m_triggerStart) {
	s.sleepMs(10);
	m->dspProcessPeriod();
	if(!rounds--) {
	  error(MJ_ERR_UNKNOWN, fname, "No MMC yet (?)");
	  break;
	}
      }
      if(m_triggerStart) {
	trace(fname, "Got MMC triggering");
      }
    }
  }
  return MJ_ModuleDspArray::dspPrepareFast();
}

bool MJ_ModuleMixer::writeParameters(QTextStream *s)
{
  MJ_ModuleDsp::writeParameters(s);

  // MJ_ModuleDspGroup::writeParameters(s);
  uint i, j;
  uint strips = m_strips.count();

  (*s) << MJ_ModuleMixerWords[MJ_MODULE_MIXER_STRIPS] << " " << strips
       << " " << MJ_ModuleMixerWords[MJ_MODULE_MIXER_ROWS] << " " << m_rows
       << "\n";

  m_audioConf.write(s);

  uint audios = m_audioIos.count();

  (*s) << MJ_ModuleMixerWords[MJ_MODULE_MIXER_AUDIOS] << " " << audios << "\n";
  
  for(i=0; i < audios; i++) {
    m_audioOutSums.at(i)->write(s);
    m_audioIos.at(i)->write(s);
  }

  m_buses->write(s);


  for(i=0; i < strips; i++) {

    for(j=0; j < m_rows; j++) {
      MJ_MixerUnit *u = unit(i,j);
      u->write(s);
      (*s) << "\n"; // Blank lines make the output more readable...
    }
  }

  return true;
}

bool MJ_ModuleMixer::readParameters(QTextStream *s)
{
  const char *fname="MJ_ModuleMixer::readParameters";

  if(m_debug) {
    trace(fname);
    allUnitsOk();
  }

  MJ_ModuleDsp::readParameters(s);

  bool ok = true;

  uint strips, rows;
  uint i, j;
  QCString cstr1, cstr2;

  (*s) >> cstr1 >> cstr2;

  if(cstr1 != MJ_ModuleMixerWords[MJ_MODULE_MIXER_STRIPS]) {
    error(MJ_ERR_ARGUMENT, fname, 1);
    return false;
  }
  
  strips = cstr2.toUInt(&ok);

  if(!ok) {
    error(MJ_ERR_ARGUMENT, fname, 2);
    return false;
  }

  (*s) >> cstr1;
  
  if(cstr1 != MJ_ModuleMixerWords[MJ_MODULE_MIXER_ROWS]) {
    error(MJ_ERR_ARGUMENT, fname, 3);
    return false;
  }
  
  cstr2 = s->readLine();

  rows = cstr2.toUInt(&ok);

  if(!ok) {
    error(MJ_ERR_ARGUMENT, fname, 4);
    return false;
  }
  
  if(!resize(strips, rows))  {
    error(MJ_ERR_ARGUMENT, fname, 5);
    return false;
  }

  if(m_debug) {
    trace(fname, "1/3-way");
    allUnitsOk();
  }

  // Now read the audio IO modules:

  if(!m_audioConf.read(s)) {
    error(MJ_ERR_ARGUMENT, fname, "Audio conf corrupted");
    return false;
  }
  
  m_audioOutSums.clear();
  m_audioIos.clear();

  updateModuleList();

  (*s) >> cstr1 >> cstr2;
  
  if(cstr1 != MJ_ModuleMixerWords[MJ_MODULE_MIXER_AUDIOS]) {
    error(MJ_ERR_ARGUMENT, fname, 6);
    return false;
  }

  uint audios = cstr2.toUInt(&ok);
  
  if(!ok) {
    error(MJ_ERR_ARGUMENT, fname, 7);
    return false;
  }

  s->readLine();

  for(i=0; i < audios; i++) {
    
    if(m_debug) {
      cstr1.sprintf("%s # Reading IO %d", fname, (int) i);
      trace(cstr1);
    }

    MJ_ModuleDsp *m  = MJ_ModuleDsp::readFromStream(s);
    if(!m) {
      error(MJ_ERR_ARGUMENT, fname, 8);
      return false;
    }

    MJ_ModuleMixerBus *mb = dynamic_cast<MJ_ModuleMixerBus *> (m);
    if(!mb) {
      error(MJ_ERR_ARGUMENT, fname, 9);
      delete m; // Do not leak memory.
      return false;
    }

    m_audioOutSums.append(mb);

    m = MJ_ModuleDsp::readFromStream(s);
    if(!m) {
      error(MJ_ERR_ARGUMENT, fname, 10);
      return false;
    }
    m_audioIos.append(m);

    updateModuleList();
  }
  
  setAudioDevNames();

  ok = ok && m_buses->read(s);

  if(!ok) {
    return false;
  }

  if(m_debug) trace(fname, "2/3-way");

  for(i=0; i < strips; i++) {

    for(j=0; j < rows; j++) {
      MJ_MixerUnit *u = unit(i,j);
      MJ_ModuleDsp *mtmp;
      if(u) {

	u->setHost(this);
	if(!u->read(s)) {
	  error(MJ_ERR_ARGUMENT, fname, 11);
	  MJ_errorNearLine(s->device(), fname);
	  u->clear();
	  ok = false;
	}
	else if((mtmp = moduleAny(i, j))) {
	  appendModule(mtmp);
	}
      }
      s->readLine();
    }
    rerouteStripAfter(i, 0);
  }
  
  if(m_debug) {
    trace(fname, "DONE");
    allUnitsOk();
  }
  return ok;
}

void MJ_ModuleMixer::msgGetInputs(MJ_Parameters *p)
{
  char buf[32];

  p->clear();
  p->append(msgGetInput(MJ_Strs::framesPerCycle));
  p->append(msgGetInput(MJ_Strs::sampleRate));

  for(uint i=0; i < m_midiConf.numDevices(); i++) {
    MJ_2Int dev = *m_midiConf.device(i);
    sprintf(buf, "%d %d", dev.m_v[0], dev.m_v[1]);
    p->append(new MJ_ParameterString(buf, "MIDI device", addMidiDevStr));
  }
}

bool MJ_ModuleMixer::msgProcess(const char *id, MJ_Message *m)
{
  bool ok = true;

  if(strcmp(id, MJ_Strs::framesPerCycle) == 0) {
    int v = m->toInt(&ok);

    if(ok) {
      m_timer.setFrameSize(v);
      updateTimer();
    }
  }
  if(strcmp(id, MJ_Strs::sampleRate) == 0) {
    int v = m->toInt(&ok);

    if(ok) {
      m_timer.setHz(v);
      updateTimer();
    }
  }
  if(strcmp(id, MJ_Strs::MIDI) == 0) {
    MJ_MessageMidiEvent *midiMsg = dynamic_cast<MJ_MessageMidiEvent *> (m);

    if(midiMsg) handleMidiMessage(& midiMsg->midiMsg());
  }
  if(strcmp(id, addMidiDevStr) == 0) {

    /* This is a bit tricky, we must rely on the same parameter ids
       regardless of how many devices there are. */

    MJ_MessageString *strMsg = dynamic_cast<MJ_MessageString *> (m);

    if(strMsg) {
      const QString &str = strMsg->string();
      int tmp1 = -1, tmp2 = -1;
      sscanf(str, "%d %d", &tmp1, &tmp2);

      ok = addMidiDev((MJ_MidiDeviceType) tmp1, tmp2);
    }
  }
  if(strcmp(id, removeMidiDevStr) == 0) {

    /* This is a bit tricky, we must rely on the same parameter ids
       regardless of how many devices there are. */

    MJ_MessageString *strMsg = dynamic_cast<MJ_MessageString *> (m);

    if(strMsg) {
      const QString &str = strMsg->string();
      int tmp1 = -1, tmp2 = -1;
      sscanf(str, "%d %d", &tmp1, &tmp2);
      
      ok = removeMidiDev((MJ_MidiDeviceType) tmp1, tmp2);
    }
  }
    
  return ok;
}

MJ_Parameter *MJ_ModuleMixer::msgGetInput(const char *id)
{
  
  if(strcmp(id, MJ_Strs::framesPerCycle) == 0) {
    return new MJ_ParameterInt("Frames per cycle", id, 
			       m_currentTime->frameSize(),
			       1, 4096);
  }
  else if(strcmp(id, MJ_Strs::sampleRate) == 0) {
    return new MJ_ParameterInt("Sampling rate", id, 
			       (int) m_currentTime->hz(),
			       1, 4096);
  }
  return 0;
}

MJ_ModuleDsp *MJ_ModuleMixer::findModule(const char *id)
{
  const char * fname = "MJ_ModuleMixer::findModule";

  if(m_debug) trace(fname, id);

  QCString idCStr(id);

  bool hasFirst = true, hasSecond = true;
  int strip = MJ_fieldC(idCStr, 1, "_").toInt(&hasFirst);
  int row = MJ_fieldC(idCStr, 2, "_").toInt(&hasSecond);

  if(!hasFirst || !hasSecond || 
     (MJ_fieldC(idCStr, 0, "_") != "module"))
    return MJ_ModuleDspGroup::findModule(id);

  if(strcmp(id, "../") == 0)
    return this;

  return moduleAny(strip, row);
}

MJ_Base *MJ_ModuleMixer::block(uint strip, uint row)
{
  MJ_MixerUnit *u = unit(strip, row);
  return u ? u->obj() : 0;
}

MJ_ModuleMixerSend *MJ_ModuleMixer::send(uint strip, uint row)
{
  MJ_Base *b = block(strip, row);
  return b ? dynamic_cast<MJ_ModuleMixerSend *> (b) : 0;
}

MJ_ModuleDsp *MJ_ModuleMixer::module(uint strip, uint row)
{
  MJ_Base *b = block(strip, row);
  return b ? dynamic_cast<MJ_ModuleDsp *> (b) : 0;
}

MJ_ModuleDsp *MJ_ModuleMixer::moduleAny(uint strip, uint row)
{
  return m_strips.at(strip)->moduleAny(row);
}

MJ_ModuleMixerSource *MJ_ModuleMixer::moduleMixerSource(uint strip, uint row)
{
  MJ_Base *b = block(strip, row);
  if(b) return dynamic_cast<MJ_ModuleMixerSource *> (b);
  return 0;
}

MJ_MixerUnit *MJ_ModuleMixer::unit(uint strip, uint row)
{
  if(strip >= m_strips.count()) return 0;
  MJ_ModuleMixerStrip *s = m_strips.at(strip);
  if(row >= s->count()) return 0;
  return s->at(row);
}

int MJ_ModuleMixer::prevModuleOutput(uint strip, uint row)
{
  for(int r = (int) row - 1; r >= 0; r--) {
    MJ_ModuleDsp *m = module(strip, r);
    if(m)
      if(m->dspOutChannels()) return r;
  }
  return -1;
}

bool MJ_ModuleMixer::locate(MJ_Base *obj, int *strip, int *row)
{
  int crows = rows();
  int cstrips = strips();

  for(int j=0; j < cstrips; j++) {
    for(int i = 0; i < crows; i++) {
      if(unit(j, i)->obj() == obj) {
	*strip = j;
	*row = i;
	return true;
      }
    }
  }

  return false;
}

/** Resize the mixer. */

bool MJ_ModuleMixer::resize(uint strips, uint rows, bool init)
{
  const char *fname ="MJ_ModuleMixer::resize";

  if(m_debug) trace(fname);
  
  uint i;
  
  uint oldStrips = m_strips.count();

  if(!m_strips.resize(strips)) return false;
  
  for(i=0; i < strips; i++) {
    MJ_ModuleMixerStrip *s = m_strips.at(i);

    if(!s->resize(rows)) return false;

    s->setMixer(this);
    s->setIndex(i);
    
  }

  m_rows = rows;

  for(i=oldStrips; i < strips && init; i++)
    m_strips.at(i)->makeDefault();

  return true;
}

bool MJ_ModuleMixer::addModule(MJ_ModuleDsp *m, uint strip, uint row)
{
  const char *fname ="MJ_ModuleMixer::addModule";

  char buf[128];

  if(m_debug) trace2("%s # strip = %d row = %d, module = %p", 
		     fname, (int) strip, (int) row, m);
  
  if((strip >= strips()) || (row >= rows())) {
    sprintf(buf, "strip = %d, row = %d", (int) strip, (int) row);
    error(MJ_ERR_ARGUMENT, fname, buf);
    return false;
  }

  unit(strip, row)->set(this, m);
  
  if(!m) return true; // No cause for panic.

  sprintf(buf, "module_%d_%d", (int) strip, (int) row);
  m->setId(buf);
  m->setParentModule(this);
  MJ_DspOptions *currOpt = m->dbe()->options();
  if(currOpt) {
    int in = currOpt->newInChans(1);
    int out = currOpt->newOutChans(1);
    m->dspSetInChannels(in);
    m->dspSetOutChannels(out);
  }
  else
    MJ_fatal(MJ_ERR_ARGUMENT, fname);

  appendModule(m);

  bool ok = rerouteStripAfter(strip, row);
  if(!ok) error(MJ_ERR_INTERNAL, fname, typeid(*m).name());

  allUnitsOk();

  if(m_debug) trace2(fname, "EXIT");

  return ok;
}

bool MJ_ModuleMixer::rerouteStripAfter(int strip, int row)
{
  const char *fname ="MJ_ModuleMixer::rerouteStripAfter";

  if(m_debug) trace(fname, strip);

  return m_strips.at(strip)->rerouteStripAfter(row);
}

bool MJ_ModuleMixer::rerouteStrips()
{
  bool ok(true);

  for(uint i=0; i < m_strips.count(); i++) {
    MJ_ModuleMixerStrip *ms =  m_strips.at(i);
    ok = ok && ms->rerouteStripAfter(0) && ms->connectMessages();
  }

  return ok;
}

void MJ_ModuleMixer::checkRow0()
{
  const char *fname ="MJ_ModuleMixer::checkRow";

  if(m_debug) trace(fname);

  for(uint i=0; i < m_strips.count(); i++) {
    MJ_ModuleDsp *m = moduleAny(i, 0);
    if(m) {
      MJ_DspOptions *currOpt = m->dbe()->options();
      int origChans = m->dspOutChannels();
      int outChans = currOpt->newOutChans(origChans);

      if(origChans != outChans) {

	trace3(m_debug, "%s # Resizing out channels %s to %d",
	       fname, typeid(*m).name(), outChans);

	m->dspSetOutChannels(outChans);

	MJ_AppActions::moduleIOModified(m, MJ_MIO_SIGNAL_OUTPUT);
      }
    }
  }
}

bool MJ_ModuleMixer::connectBuses()
{
  const char *fname ="MJ_ModuleMixer::connectBuses";

  if(m_debug) trace(fname);

  bool ok = true;
  uint i, j, l;
  int channel;

  MJ_ModuleDsp *mtmp;

  m_buses->clearMappings();

  for(i = 0; i < m_audioOutSums.count(); i++) {
    ((MJ_ModuleMixerBus *)m_audioOutSums.at(i))->clearMappings();
  }

  // Connect the buses to input channels.
  for(i=0; i < strips(); i++) {
    // The first row (index zero) cannot be a send.
    for(j=1; j < m_rows; j++) {
      MJ_ModuleMixerSend *s = send(i, j);

      if(!s)
	continue;

      // int chans = s->dspOutChannels();
      
      // Connect the module:
      const char * sid = s->id();           // Source id.
      for(l = 0; l < (uint) s->nofMappings(); l++) {

	if(!s->isActiveMapping(l, this)) {
	  continue;
	}

	const MJ_3Int *map = & s->mapping(l);
	
	// trace(fname, "Connecting send");
	
	if(map->m_v[0] == MJ_MIXERSOURCE_BUS) {
	  m_buses->addMapping
	    (sid, map->m_v[1], map->m_v[2], 1.0);
	}
	else if(map->m_v[0] == MJ_MIXERSOURCE_AUDIO_OUTPUT) {
	  
	  // trace(fname, "To audio out");
	  
	  if(audioOutId(map->m_v[2], &mtmp, &channel)) {
	    MJ_ModuleMixerBus *mb = (MJ_ModuleMixerBus *) mtmp;
	    mb->addMapping(sid, map->m_v[1], channel, 1.0);
	    
	    // mtmp->dspConnectInput(sid, map->m_v[1], channel);
	  }
	  else {
	    error(MJ_ERR_INTERNAL, fname, "Bad config");
	    ok = false;
	  }
	}
	else
	  error(MJ_ERR_INTERNAL, fname, "Unknown target");
      }
    }
  }
  return ok;
}

bool MJ_ModuleMixer::connectSources()
{
  const char * fname = "MJ_ModuleMixer::connectSources";

  uint i, j, strips = m_strips.count();
  const char * id = 0;
  int channel = -1;
    
  const char * busesId = m_buses->id();
  
  bool ok = true;

  char buf[256];

  for(i=0; i < strips; i++) {
    MJ_ModuleMixerSource *ms = moduleMixerSource(i, 0);

    if(!ms) continue;
      
    MJ_Array<MJ_3Int> * sds = & ms->sourceDefs();
    
    ms->dspSetInChannels(sds->size());

    MJ_AppActions::moduleIOModified(ms, MJ_MIO_SIGNAL_INPUT);
    
    MJ_Array<MJ_2Int> *mim = & ms->mappings2();
    mim->resize(sds->size());
    
    for(j = 0; j < sds->size(); j++) {
      
      MJ_3Int *map = & sds->getFast(j);
      MJ_2Int *intMap = & mim->get(j);

      intMap->m_v[0] = j;
      intMap->m_v[1] = map->m_v[2];
      
      MJ_MixerSourceType type = (MJ_MixerSourceType) map->m_v[0];
      
      if(type == MJ_MIXERSOURCE_BUS) {
	if(m_debug) {
	  sprintf(buf, "%s # from bus = %s", fname, busesId);
	  trace(buf);
	}
	
	ms->dspConnectInput(busesId, map->m_v[1], j);
	
      }
      else if(type == MJ_MIXERSOURCE_AUDIO_INPUT) {
	
	if(audioInId(map->m_v[1], &id, &channel))
	  ms->dspConnectInput(id, channel, j);
	else {
	  error(MJ_ERR_INTERNAL, fname, "No such audio in");
	  ok = false;
	}
      }
      else {
	if(m_debug)
	  error(MJ_ERR_INTERNAL, fname, type);
	ok = false;
      }
    }
  }
  return ok;
}

/** Check all mixer sends and make the legal. */

void MJ_ModuleMixer::checkSends()
{
  uint ss = strips();

  for(uint i=0; i < ss; i++)
    m_strips.at(i)->checkSends();
}

/** Check all mixer sources and make the legal. */
void MJ_ModuleMixer::checkSources()
{
  uint ss = strips();

  for(uint i=0; i < ss; i++)
    m_strips.at(i)->checkSources();
}


/** Updates the internal module list for real-time use. */

bool MJ_ModuleMixer::updateModuleList()
{
  const char *fname ="MJ_ModuleMixer::updateModuleList";

  if(m_debug) trace(fname);

  clearArray();

  uint i, strips = m_strips.count();
  
  for(i=0; i < strips; i++)
    m_strips.at(i)->exportModules(this);

  appendModule(m_buses);
  
  uint aios = m_audioIos.count();
  
  if(m_debug) trace(fname, "HALF");

  for(i=0; i < aios; i++) {

    MJ_ModuleDsp *m = m_audioOutSums.at(i);

    MJ_ModuleDsp *ma = m_audioIos.at(i);

    bool append = ma->dspOutChannels();

    // if(m->dspOutChannels())
    MJ_ModuleDspGroup::addToArray(m, append);
    MJ_ModuleDspGroup::addToArray(ma, append);
  }

  uint mios = m_midiIos.count();
  
  for(i=0; i < mios; i++) {
    MJ_ModuleDspGroup::addToArray(m_midiIos.at(i), true);
  }
  
  if(m_debug) trace(fname, m_array.size());
  return true;
}

bool MJ_ModuleMixer::allUnitsOk()
{
  if(m_debug) trace("MJ_ModuleMixer::allUnitsOk");

  uint i, j;
  bool ok = true;

  for(i = 0; i < strips(); i++) {
    for(j = 0; j < rows(); j++) {
      if(!unit(i, j)->ok()) {
	ok = false;
	char buf[512];
	sprintf(buf, "MJ_ModuleMixer::allUnitsOk %d %d", i, j);
	error(MJ_ERR_INTERNAL, buf);
      }
    }      
  }
  return ok;
}

bool MJ_ModuleMixer::validConnection(int sourceType, int index)
{
  int foo2;

  const char * foo1=0;

  if((sourceType == MJ_MIXERSOURCE_BUS && 
      index >= m_buses->dspOutChannels()) ||

     (sourceType == MJ_MIXERSOURCE_AUDIO_INPUT &&
      !audioId(index, &foo1, &foo2, true)) ||

     (sourceType == MJ_MIXERSOURCE_AUDIO_OUTPUT &&
      !audioId(index, &foo1, &foo2, false))){
    return false;
  }
  
  return true;
}

bool MJ_ModuleMixer::addAudioDev
(MJ_AudioDeviceType type, int number)
{
  const char * fname = "MJ_ModuleMixer::addAudioDev";

  trace3(m_debug, "%s # type = %d, number = %d", fname, (int) type, number);

  int channels = MJ_AudioConf::deviceChannels(type, number);

  int ind = m_audioConf.findDevice(type, number);

  int suchDevices = MJ_AudioConf::audioDevices(type);

  if(number >= suchDevices) {    
    error2(MJ_ERR_ARGUMENT, "%s # %d >= %d", fname, number, suchDevices);
    return false;
  }

  if(!channels || (ind >= 0) || !MJ_kernel) {
    error2(MJ_ERR_ARGUMENT, "%s # ind = %d, MJ_kernel = %p, channels = %d",
	   fname, ind, MJ_kernel, channels);
    return false;
  }

  QString devName(MJ_AudioConf::audioDevice(type, number));
  if(devName.isEmpty()) {
    error(MJ_ERR_ARGUMENT, fname, 2);
    return false;
  }

  bool input = MJ_AudioConf::isInput(type);

  int dpf; // Dsp Plugin Functionality

  if(input) dpf = MJ_DPF_AUDIO_IN;
  else dpf = MJ_DPF_AUDIO_OUT;

  MJ_ModuleDsp *maw = MJ_Kernel::newFunctionalModule
    (MJ_DspPluginFunctionalityKeys[dpf]);

  trace3(m_debug, 
	 "%s # Got new module %s %p", 
	 fname, MJ_DspPluginFunctionalityKeys[dpf], maw);

  if(!maw) {
    if(m_debug) error(MJ_ERR_ARGUMENT, fname, "No audio plugin available");
    return false;
  }

  if(input) maw->dspSetOutChannels(channels);
  else maw->dspSetInChannels(channels);
  
  // Set the device name:
  MJ_MessageString mstr;
  mstr.setString(devName);
  maw->msgProcess(MJ_Strs::device, &mstr);
  
  addAudioModule(maw);
  m_audioConf.addDevice(type, number);

  return true;
}

bool MJ_ModuleMixer::removeAudioDev(MJ_AudioDeviceType type, int number)
{
  const char * fname = "MJ_ModuleMixer::removeAudioDev";

  if(m_debug) trace2("%s # type = %d, number = %d", fname, (int) type, number);
  
  int ind = m_audioConf.findDevice(type, number);

  if(ind == -1) {
    error2(MJ_ERR_ARGUMENT, 
	   "%s #  No device %d %d, %d devices", fname,
	   (int) type, number, m_audioConf.numDevices());

    m_audioConf.write(&MJ_cerr());
    return false;
  }

  m_audioConf.removeDevice(type, number);

  m_audioIos.remove(ind);
  m_audioOutSums.remove(ind);

  updateModuleList();

  return true;
}


bool MJ_ModuleMixer::audioInId(int c, const char **id, int *channel)
{
  return audioId(c, id, channel, true);
}

bool MJ_ModuleMixer::audioOutId(int chan, MJ_ModuleDsp **mptr, int *channel)
{
  const char * id;
  if(!audioId(chan, &id, channel, false)) return false;

  MJ_ModuleDsp *m;
  *mptr = 0;

  for(uint i=0; i < m_audioIos.count(); i++) {
    m = m_audioIos.at(i);
    if(!strcmp(m->id(), id)) {
      *mptr = m_audioOutSums.at(i);
      return true;
    }
  }
  return false;
}

bool MJ_ModuleMixer::audioId(int c, const char **id, int *channel, bool input)
{
  uint counter = 0;

  for(uint i=0; i < m_audioIos.count(); i++) {
    MJ_ModuleDsp * m =  m_audioIos.at(i);
    
    uint newCounter;

    if(input) newCounter = counter + m->dspOutChannels();
    else newCounter = counter + m->dspInChannels();

    if(newCounter > (uint) c) {
      *channel = c - counter;
      *id =  m->id();
      return true;
    }
    counter = newCounter;
  }
  return false;
}

void MJ_ModuleMixer::setAudioDevNames()
{
  const char * fname = "MJ_ModuleMixer::setAudioDevNames";

  if(m_debug) trace(fname);

  uint nd = m_audioConf.numDevices();
  
  MJ_MessageString mstr;
  
  for(uint i= 0; i < nd; i++) {
    MJ_ModuleDsp *maw = m_audioIos.at(i);
    mstr.setString(m_audioConf.deviceName(i));

    if(mstr.string().isEmpty()) {
      error(MJ_ERR_INTERNAL, fname, "No audio device name");
    }
    else 
      maw->msgProcess(MJ_Strs::device, &mstr);
  }
}

uint MJ_ModuleMixer::audioChansIn()
{
  uint max = m_audioIos.count();

  uint count = 0;  
  for(uint i=0; i < max; i++)
    count += (uint) m_audioIos.at(i)->dspOutChannels();
  return count;
}

uint MJ_ModuleMixer::audioChansOut()
{
  uint max = m_audioIos.count();

  uint count = 0;  
  for(uint i=0; i < max; i++)
    count += (uint) m_audioIos.at(i)->dspInChannels();
  return count;
}

/** Adds a MIDI device to the mixer. */
bool MJ_ModuleMixer::addMidiDev(MJ_MidiDeviceType type, int number)
{
  const char * fname = "MJ_ModuleMixer::addMidiDev";
  
  int ind = m_midiConf.findDevice(type, number);
  
  int suchDevices = MJ_MidiConf::midiDevices(type);
  
  if((ind >= 0) || !MJ_kernel || (number >= suchDevices)) {
    error2(MJ_ERR_ARGUMENT, "%s # ind = %d, MJ_kernel = %p",
	   fname, ind, MJ_kernel);
    return false;
  }

  MJ_ModuleDsp *m = createMidiModule(type, number);

  if(!m) {
    return false;
  }
  
  addMidiModule(m);
  m_midiConf.addDevice(type, number);

  return true;
}

/** Removes a MIDI device from the mixer. */
bool MJ_ModuleMixer::removeMidiDev(MJ_MidiDeviceType type, int number)
{
  const char * fname = "MJ_ModuleMixer::removeMidiDev";

  if(m_debug) trace2("%s # type = %d, number = %d", fname, (int) type, number);
  
  int ind = m_midiConf.findDevice(type, number);

  if(ind == -1) {

    error2(MJ_ERR_ARGUMENT, 
	   "%s #  No device %d %d, %d devices", fname,
	   (int) type, number, m_midiConf.numDevices());

    m_midiConf.write(&MJ_cerr());
    return false;
  }

  m_midiConf.removeDevice(type, number);
  m_midiIos.remove(ind);

  updateModuleList();

  return true;
}

void MJ_ModuleMixer::setMidiDevNames()
{
  const char * fname = "MJ_ModuleMixer::setMidiDevNames";

  if(m_debug) trace(fname);

  uint nd = m_midiConf.numDevices();
  
  MJ_MessageString mstr;
  
  for(uint i= 0; i < nd; i++) {
    MJ_ModuleDsp *maw = m_midiIos.at(i);
    mstr.setString(m_midiConf.deviceName(i));

    if(mstr.string().isEmpty()) {
      error(MJ_ERR_INTERNAL, fname, "No midi device name");
    }
    else 
      maw->msgProcess(MJ_Strs::device, &mstr);
  }
}

MJ_ModuleDsp *MJ_ModuleMixer::createMidiModule
(MJ_MidiDeviceType type, int number)
{
  const char * fname = "MJ_ModuleMixer::createMidiModule";

  QString devName(MJ_MidiConf::midiDevice(type, number));
  if(devName.isEmpty()) {
    error(MJ_ERR_ARGUMENT, fname, 2);
    return false;
  }
  
  int dpf; // Dsp Plugin Functionality

  bool input = MJ_MidiConf::isInput(type);

  if(input) dpf = MJ_DPF_MIDI_IN;
  else dpf = MJ_DPF_MIDI_OUT;
  
  MJ_ModuleDsp *maw = MJ_Kernel::newFunctionalModule
    (MJ_DspPluginFunctionalityKeys[dpf]);
  
  trace3(m_debug, 
	 "%s # Got new module %s %p", 
	 fname, MJ_DspPluginFunctionalityKeys[dpf], maw);

  // Set the device name:
  MJ_MessageString mstr;
  mstr.setString(devName);
  maw->msgProcess(MJ_Strs::device, &mstr);

  return maw;
}

void MJ_ModuleMixer::addAudioModule(MJ_ModuleDsp *m)
{
  const char * fname = "MJ_ModuleMixer::addAudioModule";

  uint i, inChans = m->dspInChannels();

  // trace3(m_debug, "%s # ");

  appendModule(m);

  // if(!inChans)
  m_audioIos.append(m); 
  /* else 
     m_audioIos.prepend(m);  */

  MJ_ModuleDsp *mb = 
    MJ_Kernel::newDspModule(MJ_busIdentifier, MJ_PP_IDENTIFIER);
  
  if(!mb) {
    error(MJ_ERR_INTERNAL, fname, "No bus plugin !! ??");
    return;
  }
  
  mb->dspSetOutChannels(inChans);
  m_audioOutSums.append(mb);
  appendModule(mb);

  // Connect the audio module to the multiplexer module.
  for(i=0; i < inChans; i++) {
    m->dspConnectInput(mb->id(), i, i);
  }

  updateModuleList();  
}

void MJ_ModuleMixer::addMidiModule(MJ_ModuleDsp *m)
{
  // const char * fname = "MJ_ModuleMixer::addMidiModule";

  // trace3(m_debug, "%s # ");

  appendModule(m);
  m_midiIos.append(m);
  updateModuleList();  
}

void MJ_ModuleMixer::updateTimer()
{
  dspSetTimer(&m_timer);
}

void MJ_ModuleMixer::handleMidiMessage(const DI_MidiEvent *e)
{
  const char * fname = "MJ_ModuleMixer::handleMidiMessage";

  if(e->isMmc()) {
    DI_MidiEvent::MmcCommand cmd = e->mmcCommand();

    trace2("%s # MMC message, id = %x, command = %x", fname, 
	   e->mmcDeviceId(), (int) cmd);

    if(cmd == DI_MidiEvent::MmcPlay) {
      m_triggerStart = true;
    }
  }
  else {
    trace2("%s # Got non-MMC MIDI event 0x%x", 
	   fname, (int) e->a());
    // m_triggerStart = true;    
  }
}
