/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#ifndef MJ_MODULE_MIXER_H
#define MJ_MODULE_MIXER_H

#include <mj_array2d.h>
#include <mj_audio_conf.h>
#include <mj_midi_conf.h>
#include <mj_module_dsp_group.h>
#include <mj_module_mixer_bus.h>
#include <mj_module_mixer_send.h>
#include <mj_module_mixer_send.h>
#include <mj_multi_int.h>
#include <mj_satellite.h>

class DI_MidiEvent;
class MJ_ModuleMixerSource;
class MJ_ModuleMixer_QG;


enum MJ_MixerChanges {
  MJ_MICH_IO,
  MJ_MICH_RESIZE_PREPARE,
  MJ_MICH_RESIZE_FINAL,
  MJ_MICH_SIZEOF
};

enum MJ_MixerUnitType {
  MJ_MIXER_UNIT_BLANK,
  MJ_MIXER_UNIT_SEND,
  MJ_MIXER_UNIT_SOURCE,
  MJ_MIXER_UNIT_MODULEDSP,
  MJ_MIXER_UNIT_SIZEOF
};


/** Mixer unit types. */
extern const char *MJ_MixerUnitTypes[];

/** Mixer unit. */
class MJ_MixerUnit
{
public:
  MJ_MixerUnit();
  ~MJ_MixerUnit();
    
  /** Sets he current object. The old object is deleted if such
      exists. */
  void set(MJ_ModuleDspGroup *, MJ_Base *);
  void setHost(MJ_ModuleDspGroup *host)
  { m_host = host; }

  void clear();
  
  MJ_Base *obj() { return m_obj; }

  /** Checks that the pointers appear to be ok. */
  bool ok() const { return !(m_obj && !m_host); }

  QWidget *newGui(MJ_GuiSize   s, 
		  uint         strip,
		  uint         row,
		  MJ_ModuleMixer_QG  *mixerGui, 
		  QWidget *parent,
		  const char  *name = 0);
  
  bool write(QTextStream *s);
  bool read(QTextStream *s);

  QString objectName();

  static void setDebug(int debug) {  m_debug = debug; }
protected:
  MJ_ModuleDspGroup *m_host;

  /// Either NULL, MJ_ModuleMixerSend or MJ_ModuleDsp
  MJ_Base *m_obj;

  static int m_debug;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/// Mixer strip.
class MJ_ModuleMixerStrip : public MJ_List<MJ_MixerUnit>
{
public:
  MJ_ModuleMixerStrip();
  ~MJ_ModuleMixerStrip();

  const QString & name() { return m_name; }

  /// Set the host mixer of this strip object.
  void setMixer(MJ_ModuleMixer *m);

  /// Set the index number of this strip
  void setIndex(int i) { m_index = i; }

  MJ_ModuleDsp *module(uint row);
  MJ_ModuleDsp *moduleAny(uint row);
  MJ_ModuleMixerSend *send(uint row);
  MJ_ModuleMixerSource *source(uint row);

  bool connectMessages();

  bool connectMessagesTo(uint);

  void exportModules(MJ_ModuleMixer *mixer);

  bool rerouteStripAfter(int row);

  bool rerouteSend(int row);
  bool rerouteDspInput(int row);
  
  bool makeDefault();

  void checkSends();

  void checkSources();

  MJ_ModuleDsp *prevDspSource(int row);
  
  static bool *debugPtr() { return &m_debug; }
  
protected:

  /// Name of the strip.
  QString m_name;

  MJ_ModuleMixer *m_mixer;

  int             m_index;

  static bool m_debug;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** Satellite for the mixer class.  The mixer is fairly involved
    system so we need a specific satellite for it. */
class MJ_ModuleMixerSatellite  : public MJ_Satellite
{
  Q_OBJECT

public:
  MJ_ModuleMixerSatellite(MJ_Base *b)
    : MJ_Satellite(b) {}
  virtual ~MJ_ModuleMixerSatellite();

public slots:
  void remove(int,int);
  void update(int,int);
  void modified2(int);
  void modified4(int,int,int);

signals:
  void emitRemove(int,int);
  void emitUpdate(int,int);
  void emitModified2(int);
  void emitModified4(int,int,int);
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/// Mixer keywords.
enum MJ_ModuleMixerWord {
  MJ_MODULE_MIXER_AUDIOS,
  MJ_MODULE_MIXER_STRIPS,
  MJ_MODULE_MIXER_ROWS,
  MJ_MODULE_MIXER_BUSES,
  MJ_MODULE_MIXER_BUSWIDTH,
  MJ_MODULE_MIXER_UNIT,
  MJ_MODULE_MIXER_SIZEOF
};

extern const char *MJ_ModuleMixerWords[];

/** Implements the traditional mixer interface. 

    Very big and bloated class. Rewrite would not hurt.

 */
class MJ_ModuleMixer : public MJ_ModuleDspGroup
{
public:

  MJ_ModuleMixer(bool init = true);
  virtual ~MJ_ModuleMixer();

  bool dspPrepare();
  bool dspPrepareFast();

  bool writeParameters(QTextStream *s);
  bool readParameters(QTextStream *s);

  virtual void msgGetInputs(MJ_Parameters *);
  virtual bool msgProcess(const char *id, MJ_Message *);
  virtual MJ_Parameter *msgGetInput(const char *id);

  virtual MJ_ModuleDsp *findModule(const char *id);

  MJ_ModuleMixerBus & buses() { return *m_buses; }
  uint nofBuses() { return m_buses->dspOutChannels(); }
  void setBuses(uint number) { m_buses->dspSetOutChannels(number); }

  uint strips() const { return m_strips.count(); }
  uint rows() const { return m_rows; }

  /// Returns the satellite
  MJ_ModuleMixerSatellite *mixerSatellite()
  { return (MJ_ModuleMixerSatellite *) m_satellite; }

  MJ_Base *block(uint strip, uint row);

  MJ_ModuleMixerSend *send(uint strip, uint row);

  MJ_ModuleDsp *module(uint strip, uint row);
  MJ_ModuleDsp *moduleAny(uint strip, uint row);

  MJ_ModuleMixerSource *moduleMixerSource(uint strip, uint row);

  MJ_MixerUnit *unit(uint strip, uint row);

  MJ_ModuleMixerStrip * strip(uint i) { return m_strips.at(i); }
  
  /** Search the previous module with non-zero output. 
      Search starts from position "strip, row-1".
      Returns -1 upon failure. */
  int prevModuleOutput(uint strip, uint row);

  /** Locate a child module in grid. */
  bool locate(MJ_Base *obj, int *strip, int *row);

  bool resize(uint strips, uint rows, bool init = true);

  /// Add module to the mixer.
  bool addModule(MJ_ModuleDsp *m, uint strip, uint row);
  /// Rechannels the contents of a strip.
  bool rerouteStripAfter(int strip, int row);
  bool rerouteStrips();

  void checkRow0();
  
  /// Connect the bus objects.
  bool connectBuses();
  bool connectSources();

  bool updateModuleList();

  void checkSends();

  void checkSources();

  bool allUnitsOk();

  bool validConnection(int sourceType, int index);

  /** @name Audio I/O */
  //@{
  bool addAudioDev(MJ_AudioDeviceType type, int number);
  bool removeAudioDev(MJ_AudioDeviceType type, int number);

  uint nofAudioIos() const { return m_audioIos.count(); }
  QList<MJ_ModuleDsp> & audioIos() { return m_audioIos; }

  MJ_ModuleDsp *audioIo(uint i) { return m_audioIos.at(i); }
  bool audioInId(int c, const char **id, int *channel);
  bool audioOutId(int c, MJ_ModuleDsp **mptr, int *channel);
  bool audioId(int c, const char **id, int *channel, bool input);
  MJ_AudioConf &audioConf() { return m_audioConf; }
  void setAudioDevNames();

  uint audioChansIn();
  uint audioChansOut();

  //@}

  /** @name MIDI I/O and control. */
  //@{
  bool addMidiDev(MJ_MidiDeviceType type, int number);
  bool removeMidiDev(MJ_MidiDeviceType type, int number);
  void setMidiDevNames();
  MJ_ModuleDsp *createMidiModule(MJ_MidiDeviceType type, int number);

  uint nofMidiIos() const { return m_midiIos.count(); }
  MJ_MidiConf &midiConf() { return m_midiConf; }

  /** Sets the index of the MMC controller. Use -1 to disble MMC.  */
  void setMmcController(int index) { m_mmcController = index; }
  int mmcController() const { return m_mmcController; }
  //@}

  /// Access the debugging variables address.
  static bool *debugPtr() { return &m_debug; }

  static const char * addMidiDevStr;
  static const char * removeMidiDevStr;
    
protected:

  void addAudioModule(MJ_ModuleDsp *m);
  void addMidiModule(MJ_ModuleDsp *m);

  void updateTimer();

  void handleMidiMessage(const DI_MidiEvent *);

  MJ_List<MJ_ModuleMixerStrip> m_strips;

  MJ_ModuleMixerBus            *m_buses;

  QList<MJ_ModuleDsp>   m_audioIos;

  /** MJ_ModuleMixerBus is the real type. */
  QList<MJ_ModuleDsp>   m_audioOutSums;

  MJ_AudioConf          m_audioConf;

  QList<MJ_ModuleDsp>   m_midiIos;
  MJ_MidiConf           m_midiConf;
  int                   m_mmcController;
  bool                  m_triggerStart;

  uint m_rows;

  MJ_TimeStamp m_timer;

  static bool m_debug;  
};


#endif
