/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_mixer_gui.h>

#include <mj_audio_io_gui.h>
#include <mj_color.h>
#include <mj_debug.h>
#include <mj_dsp_plugin_loader.h>
#include <mj_dsp_plugin_filters.h>
#include <mj_error.h>
#include <mj_global.h>
#include <mj_gui_utils.h>
#include <mj_kernel.h>
#include <mj_message_string.h>
#include <mj_midi_conf_gui.h>
#include <mj_module_keys.h>
#include <mj_module_menubar.h>
#include <mj_module_mixer_source.h>
#include <mj_mixer_strip_qg.h>
#include <mj_plugin_menu.h>
#include <mj_rt_driver.h>
#include <mj_rt_thread.h>
#include <mj_strs.h>
#include <mj_tr.h>

#include <qapplication.h> 
#include <qbuffer.h> 
#include <qcombobox.h> 
#include <qcursor.h> 
#include <qlayout.h> 
#include <qlabel.h> 
#include <qmenubar.h> 
#include <qmessagebox.h> 
#include <qpopupmenu.h>
#include <qpushbutton.h> 
#include <qscrollview.h> 
#include <qtextstream.h>

const int MJ_ModuleMixer_QG::m_gdist= 8;
const int MJ_ModuleMixer_QG::m_gbord= 4;

static MJ_RtThread *s_thread = 0;

/** A function that returns ascending numbers. It has
    pseudo-exponential growth rate: 1 2 3 4 5 6 7 8 10 12 14 16 24 32
    40 48.... */

extern int MJ_mixerNonlinearPower(int i)
{
  if(i <= 8) return i; // 0-8
  i -= 8;
  if(i < 4) return 10 + i * 2; // 10-16

  i-= 1;
  return i * 8;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

bool MJ_MixerBlank_QG::m_debug = false;

MJ_MixerBlank_QG::MJ_MixerBlank_QG(MJ_ModuleMixer_QG *host,
				   uint strip, 
				   uint row,
				   QWidget      *parent,
				   const char   *name)
  : QLabel(parent, name),
    m_host(host),
    m_strip(strip),
    m_row(row)
{
  if(m_debug) {
    char buf[128];
    sprintf(buf, "MJ_MixerBlank_QG::MJ_MixerBlank_QG %d %d", strip, row);
    MJ_trace(buf, this);
  }

  setIndent(10);
  setAlignment(AlignTop | AlignLeft);

  if(!MJ_kernel) return;
  
  /* QPopupMenu *edit = new QPopupMenu;
  QMenuBar *menu = new QMenuBar( this );

  MJ_PluginDbTree_GQ moduleMenuBuilder;
  moduleMenuBuilder.setNode(MJ_kernel->dsPluginTree());
     moduleMenuBuilder.formTree(this, SLOT(newModule(const QString &)), edit);
  */
  QString s1(MJ_Tr::tr("Slot")), s2;
  s2.setNum(m_row + 1);

  s1 += " " + s2;
  
  setText(s1);

  MJ_limMinSize(this);

  // menu->insertItem(s1, edit);
}

MJ_MixerBlank_QG::~MJ_MixerBlank_QG()
{
  if(m_debug) MJ_trace("MJ_MixerBlank_QG::~MJ_MixerBlank_QG", this);
}

void MJ_MixerBlank_QG::mousePressEvent(QMouseEvent *)
{
  m_host->setActive(m_strip, m_row);
  m_host->pluginMenu(m_row)->exec(QCursor::pos());
}


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_ModuleMixerResize::MJ_ModuleMixerResize(MJ_ModuleMixer *mixer,
					   MJ_Mutex       *lock,
					   QWidget      *parent,
					   const char   *name)
  : QGroupBox(parent, name)
{
  if(lock) lock->lock();
  m_strips = mixer->strips();
  m_rows   = mixer->rows();
  m_buses  = mixer->nofBuses();
  if(lock) lock->unlock();

  int i, j, tmp;
  QString s;
  QLabel *label;
  QGridLayout *grid = new QGridLayout(this, 4, 3, 20, 3);

  const char *labels[3] = {"Strips", "Rows", MJ_Strs::Buses};
  const char *slotKeys[3] = 
  { SLOT(setStrips(int)), SLOT(setRows(int)), SLOT(setBuses(int)) };

  for(i=0; i < 3; i++) {
    label = new QLabel(this);
    label->setText(MJ_Tr::tr(labels[i]));
    MJ_limMinSize(label);
    grid->addWidget(label, i, 0);

    int v;
    tmp = 1;

    if(i == 0) v = m_strips;
    else if(i == 1) v = m_rows;
    else {
      v = m_buses;
      tmp = 0;
    }

    QComboBox *combo = new QComboBox(this, labels[i]);

    for(j=tmp; j < 20; j++) {
      int tmp2 = MJ_mixerNonlinearPower(j);
      s.setNum(tmp2);
      combo->insertItem(s);

      if(tmp2 == v)
	combo->setCurrentItem(j - tmp);

    }
    connect(combo, SIGNAL(activated(int)), this, slotKeys[i]);
    MJ_limSize(combo);
    grid->addWidget(combo, i, 1);
  }

  QWidget *wtmp = new QWidget(this);
  QHBoxLayout *hbox = new QHBoxLayout(wtmp);
  
  QPushButton *b = new QPushButton(wtmp, "ok button");
  b->setText(MJ_Tr::tr("Apply"));
  connect(b, SIGNAL(pressed()), this, SLOT(apply()));
  MJ_limSize(b);
  hbox->addWidget(b, 10);

  b = new QPushButton(wtmp, "cancel button");
  b->setText(MJ_Tr::tr("Close"));
  connect(b, SIGNAL(pressed()), this, SLOT(hide()));
  MJ_limSize(b);
  hbox->addWidget(b, 10);
  hbox->addStretch(50);
  
  QSize sizeCancel = b->minimumSize();

  resize(sizeCancel.width() * 3, sizeCancel.height() * 4);

  grid->addMultiCellWidget(wtmp, 3, 3, 0, 2);

  s = MJ_Tr::tr("Change mixer size");
  setTitle(s);

  if(!parent) {
    setCaption(s);
    setIconText(s);
    show();
  }
}
  
MJ_ModuleMixerResize::~MJ_ModuleMixerResize()
{}

/* void MJ_ModuleMixerResize::hide()
{
  close();
  }*/

void MJ_ModuleMixerResize::setStrips(int i)
{
  m_strips = MJ_mixerNonlinearPower(i + 1);
}

void MJ_ModuleMixerResize::setRows(int i)
{
  m_rows = MJ_mixerNonlinearPower(i + 1);
}

void MJ_ModuleMixerResize::setBuses(int i)
{
  m_buses = MJ_mixerNonlinearPower(i);
}

void MJ_ModuleMixerResize::apply()
{
  emit emitMixerConf(m_strips, m_rows, m_buses);
}


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

bool MJ_ModuleStripHeader::m_debug = false;

MJ_ModuleStripHeader::MJ_ModuleStripHeader(int             strip,
					   MJ_ModuleMixer_QG *mixer,
					   MJ_RtControl   *rtParent,
					   QWidget        *parent,
					   const char     *name)
  : QWidget(parent, name),
    MJ_RtControl(rtParent),
    m_strip(strip),
    m_mixer(mixer)
{
  m_remaps.setAutoDelete(true);

  QMenuBar *menuBar = new QMenuBar(this);

  m_menu = new QPopupMenu(this);
  m_delete = new QPopupMenu(this);
  m_viewModule = new QPopupMenu(this);

  m_deleteId = m_menu->insertItem(MJ_Tr::tr("Remove"), m_delete);
  m_menu->setItemEnabled(m_deleteId, false);

  m_viewModuleId = 
    m_menu->insertItem(MJ_Tr::tr("View module"), m_viewModule);
  m_menu->setItemEnabled(m_viewModuleId, false);

  m_menu->insertItem(MJ_Tr::tr("View strip"), this, SLOT(viewStrip()));
  m_defId = 
    m_menu->insertItem(MJ_Tr::tr("Default strip"), this, SLOT(defaultStrip()));
  m_clearId = 
    m_menu->insertItem(MJ_Tr::tr("Clear strip"), this, SLOT(clear()));

  QString ops(MJ_Tr::tr("Strip"));
  QString n;
  n.setNum(strip);
  ops += " &" + n;

  menuBar->insertItem(ops, m_menu);

  int minH = menuBar->heightForWidth(500);
  setMinimumHeight(minH);
  setMaximumHeight(minH);


}
  
MJ_ModuleStripHeader::~MJ_ModuleStripHeader()
{}

void MJ_ModuleStripHeader::addModule(const QString &name, int number)
{
  // First the delete part:
  MJ_IntRemap *remap = new MJ_IntRemap(number);

  if(!remap) MJ_fatal(MJ_ERR_MEMORY, "MJ_ModuleStripHeader::addModule");

  if(!isRealTime())
    m_menu->setItemEnabled(m_deleteId, true);

  m_remaps.append(remap);
  m_delete->insertItem(name, remap, SLOT(call()));
  connect(remap, SIGNAL(emitCall(int)),
	  this, SLOT(deleteModule(int)));

  // Now the detailed view:

  remap = new MJ_IntRemap(number);

  if(!remap) MJ_fatal(MJ_ERR_MEMORY, "MJ_ModuleStripHeader::addModule");

  m_menu->setItemEnabled(m_viewModuleId, true);

  m_remaps.append(remap);
  m_viewModule->insertItem(name, remap, SLOT(call()));
  connect(remap, SIGNAL(emitCall(int)),
	  this, SLOT(viewModule(int)));
}

void MJ_ModuleStripHeader::clearHeader()
{
  m_remaps.clear();
  m_delete->clear();
  m_viewModule->clear();
}

void MJ_ModuleStripHeader::rebuild()
{
  const char * fname = "MJ_ModuleStripHeader::rebuild";

  MJ_ModuleMixer * m = m_mixer->mixer();

  if((uint) m_strip >= m->strips()) return;

  // lock(); // Why bother...
  MJ_ModuleMixerStrip *strip = m->strip(m_strip); 
  // unlock();
  
  MJ_cpf(strip, fname);

  QString name;

  for(uint i=0; i < strip->count(); i++) {

    lock();
    MJ_MixerUnit *carrier = strip->at(i);
    unlock();

    if(carrier) {

      lock();
      name = carrier->objectName();
      unlock();
      
      if(!name.isEmpty())
	addModule(name, i);
    }
  }
  realTimeChange(isRealTime());
}

void MJ_ModuleStripHeader::deleteModule(int d)
{
  if(m_debug) MJ_trace("MJ_ModuleStripHeader::deleteModule", d);
  emit emitDelete(m_strip, d);
  if(m_debug) MJ_trace("MJ_ModuleStripHeader::deleteModule", "EXIT");
}

void MJ_ModuleStripHeader::moduleDeleted(int strip, int row)
{
  if(strip != m_strip) return;

  if(m_debug) MJ_trace("MJ_ModuleStripHeader::moduleDeleted");

  for(uint i=0; i < m_remaps.count(); i++) {
    if(((MJ_IntRemap *) m_remaps.at(i))->id() == row) {
      m_delete->removeItemAt(i);
      m_remaps.remove(i);
    }
  }
}

void MJ_ModuleStripHeader::viewModule(int i)
{ 
  emit emitViewModule(m_strip, i);
}

void MJ_ModuleStripHeader::clear()
{
  if(isRealTime()) 
    MJ_fatal(MJ_ERR_INTERNAL, "MJ_ModuleStripHeader::clear");

  uint rows = m_mixer->mixer()->rows();

  for(uint i=0; i < rows; i++) {
    emit emitDelete(m_strip, i);
  }
}

void MJ_ModuleStripHeader::defaultStrip()
{
  const char * fname = "MJ_ModuleStripHeader::defaultStrip";

  uint i;

  if(isRealTime()) 
    MJ_fatal(MJ_ERR_INTERNAL, fname);

  MJ_ModuleMixer *mixer = m_mixer->mixer();
  MJ_ModuleMixerSatellite *s = mixer->mixerSatellite();

  uint rows = mixer->rows();

  for(i=0; i < rows; i++) {
    m_mixer->removeModule(m_strip, i);
  }
  
  MJ_ModuleMixerStrip *ms = mixer->strip(m_strip);
  bool ok = ms->makeDefault();
  
  for(i=0; i < rows; i++) {
    s->update(m_strip, i);
  }
  
  if(!ok) 
    MJ_error(MJ_ERR_INTERNAL, fname);
}

/** Enable/disable menu entries as necessary. This method can be used
    even when there is really no change in RT status. Such cases are
    when the menus have changed and empty menus need to be
    disabled. */

void MJ_ModuleStripHeader::realTimeChange(bool nowRt)
{
  bool ena = !nowRt;

  m_menu->setItemEnabled(m_deleteId, ena && m_delete->count());
  m_menu->setItemEnabled(m_viewModuleId, m_viewModule->count());
  
  m_menu->setItemEnabled(m_clearId, ena);
  m_menu->setItemEnabled(m_defId, ena);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

bool MJ_ModuleMixer_QG::m_debug = false;
    
MJ_ModuleMixer_QG::MJ_ModuleMixer_QG(MJ_RtControl *rtparent,
				     MJ_ModuleDsp *module,
				     QWidget      *parent,
				     const char   *name)
  : MJ_ModuleDspGui(rtparent, module, parent, name),
    m_mixer(0),
    m_scrollView(new QScrollView(this)),
    m_menuBar(0),
    m_playMenu(0),
    m_audioIoGui(0),
    m_midiConfGui(0),
    m_pluginMenuFirstRow(0),
    m_pluginMenu(0),
    m_pluginMenuLastRow(0),
    m_activeStrip(-1),
    m_activeRow(-1),
    m_action(-1)
{
  // m_scrollView->enableClipper(true);

  setMinimumSize(250, 120);

  m_layoutEngine.objSize() = QSize(120, 80);

  m_scrollView->viewport()->setBackgroundColor(QColor(40, 40, 40));

  m_stripWindowGuis.setAutoDelete(true);

  /* In theory we do not need to iitialize these, but I'll do it
     anyways - makes debugging easier in case something goes wrong. */
  for(int i=0; i < MJ_MIXER_RT_SIZEOF; i++)
    m_menuIds[i] = -1;

  MJ_ModuleMixer *mr = dynamic_cast<MJ_ModuleMixer *> (module);

  if(!mr) {
    error(MJ_ERR_ARGUMENT, " MJ_ModuleMixer_QG::MJ_ModuleMixer_QG");
    return;
  }
  m_mixer = mr;

  MJ_Satellite *s = mr->mixerSatellite();

  connect(s, SIGNAL(emitRemove(int,int)), 
	  this, SLOT(forgetModule(int,int)));

  connect(s, SIGNAL(emitUpdate(int,int)), 
	  this, SLOT(updateSlot(int,int)));

  connect(s, SIGNAL(emitModified2(int)), 
	  this, SLOT(moduleModified2(int)));

  connect(s, SIGNAL(emitModified4(int,int,int)), 
	  this, SLOT(moduleModified4(int,int,int)));

  if(m_debug) m_mixer->allUnitsOk();

  draw();

  relayout();

  // Capture reasonable amount of screen space.
  QSize siz1(m_scrollView->contentsWidth(), m_scrollView->contentsHeight());
  
  QSize siz2(QApplication::desktop()->size());

  siz2 *= 0.7;

  siz1 = siz1.boundedTo(siz2);

  resize(siz1 * 1.1);

  // drawVertSeparators();
  showAll();
}

MJ_ModuleMixer_QG::~MJ_ModuleMixer_QG()
{
  trace3(m_debug, "MJ_ModuleMixer_QG::~MJ_ModuleMixer_QG");

  clearInternal();

  deleteChildren();

#ifdef MJ_MEM_DEBUG
  MJ_memDebugReport();
#endif
  delete m_menuBar;
  delete m_playMenu;

  delete m_audioIoGui;
  delete m_midiConfGui;

  m_stripWindowGuis.clear();

  // Remove the elements from the MJ_kernel(Gui):
    
  if(m_module) {
    // First we break the links
    m_module->satellite()->decrementUiCount();
    m_module = 0;
  }
  
  // Then we remove this GUI/DSP-pair from the list of elements.
  MJ_kernelGui->deleteSetup(this);
}

void MJ_ModuleMixer_QG::sendBusChanged(int bus, int s, int r)
{
  (void) bus;

  lock();
  MJ_Base *b = m_mixer->block(s, r);
  unlock();

  if(!b) return;
  MJ_ModuleMixerSend *ms = dynamic_cast<MJ_ModuleMixerSend *> (b);
  if(!ms) return;
}

void MJ_ModuleMixer_QG::placeModule(MJ_ModuleDsp *m, uint strip, uint row)
{
  const char *fname = "MJ_ModuleMixer_QG::placeModule";

  if(m_debug) trace(fname, (int) strip, (int) row);

  MJ_ModuleMixerSatellite *s = m_mixer->mixerSatellite();
  s->remove(strip, row);

  MJ_ModuleDsp *mold = m_mixer->module(strip, row);

  if(mold) mold->satellite()->nextDelete();

  // Put new module there are and kill the old.
  m_mixer->addModule(m, strip, row);

  s->update(strip, row);

  // drawVertSeparators();

  if(m_debug) trace(fname, "EXIT");
}

void MJ_ModuleMixer_QG::test(QApplication *xpQA, int iArgc, char **cppArgv)
{
  (void) iArgc;
  (void) cppArgv;

  const char *fname="MJ_ModuleMixer_QG::test";

  MJ_trace(fname);

  MJ_Kernel k;

  MJ_ModuleMixer m;
  m.resize(4, 5);
  m.setBuses(4);

  // First strip:
  {
    m.addModule(k.newDspModule("_source_noise_1", MJ_PP_IDENTIFIER) , 2, 0);
    m.addModule(k.newDspModule("_delay_mono_delay_1", MJ_PP_IDENTIFIER) , 2, 1);
    m.addModule(k.newDspModule("_panning_stereo_pan_1", MJ_PP_IDENTIFIER) , 2, 2);
    m.addModule(k.newDspModule(MJ_dspVisualIdentifier, MJ_PP_IDENTIFIER) , 2, 3);
  }
  // Second strip:
  {
    m.addModule(k.newDspModule("_routing_mixer_source_1", MJ_PP_IDENTIFIER) , 3, 0);

    MJ_ModuleMixerSource *ms = m.moduleMixerSource(3,0);

    if(!ms) MJ_fatal(MJ_ERR_INTERNAL, fname);

    ms->dspSetOutChannels(2);
    ms->setMappings(2);
    ms->setMapping(0, MJ_MIXERSOURCE_BUS, 0, 0);
    ms->setMapping(1, MJ_MIXERSOURCE_BUS, 1, 1);
    
    m.addModule(k.newDspModule(MJ_dspVisualIdentifier, MJ_PP_IDENTIFIER) , 3, 1);
  }

  MJ_MixerUnit *u = m.unit(2,4);
  MJ_ModuleMixerSend *send = new MJ_ModuleMixerSend;
  
  send->resizeMappings(2);
  send->setMapping(0, MJ_MIXERSOURCE_BUS, 0, 0);
  send->setMapping(1, MJ_MIXERSOURCE_BUS, 1, 1);

  u->set(&m, send);

  // Try get audio I/O
  m.addAudioDev(MJ_ADT_INPUT, 0);
  m.addAudioDev(MJ_ADT_OUTPUT, 0);

  // Connect all modules before giving this module to the GUI...
  m.rerouteStrips();

  MJ_ModuleMixer_QG mg(0, &m, 0, "Mixer test");
  
  xpQA->setMainWidget(&mg);
  mg.show();
  xpQA->exec();
}

MJ_ModuleDspGui *MJ_ModuleMixer_QG::createNew(MJ_RtControl *rtparent,
					      MJ_ModuleDsp *module,
					      QWidget *parent, 
					      const char *name)
{
  return new MJ_ModuleMixer_QG(rtparent, module, parent, name);
}

/** Returns the menu that corresponds to given row. The plugin menus
    differ between rows. In particular there is no point in having
    input modules in the last row or output modules in the first.*/

QPopupMenu *MJ_ModuleMixer_QG::pluginMenu(uint row)
{ 
  trace3(m_debug, "MJ_ModuleMixer_QG::pluginMenu %d %d", 
	 row, m_moduleGuis.rows());

  if(!row) return m_pluginMenuFirstRow;
  else if(row < m_moduleGuis.rows() - 1) return m_pluginMenu; 
  return m_pluginMenuLastRow;
}

QWidget *MJ_ModuleMixer_QG::childParent()
{
  return m_scrollView->viewport();
}

void MJ_ModuleMixer_QG::moduleModified2(int i)
{
  // if(i == m_action) return; // We created this, no panic.

  bool enable;

  switch(i) {
  case MJ_MICH_IO:
    updateAudioConf();

    lock();
    enable = m_mixer->midiConf().numDevices(MJ_MIDID_INPUT) != 0;
    unlock();
    m_menuBar->edit()->setItemEnabled(m_mmcId, enable);

  }
}

void MJ_ModuleMixer_QG::moduleModified4(int i1, int i2, int i3)
{
  if(m_debug) trace("MJ_ModuleMixer_QG::moduleModified4", i1);
  
  if(i1 == m_action) return; // We created this, no panic.

  switch(i1) {
  case MJ_MICH_RESIZE_PREPARE:
    // Delete old GUI stuff.
    resizeGrid(i2, i3);
    break;
    
  case MJ_MICH_RESIZE_FINAL:

    redraw(i2, i3);
    showAll();
    
    break;

  default:
    break;
  }
}


void MJ_ModuleMixer_QG::writeCout()
{
  m_mixer->toCout();
}

void MJ_ModuleMixer_QG::runShort()
{
  m_mixer->runShort(20000);
}

void MJ_ModuleMixer_QG::runShort2()
{
  if(s_thread) {
    error(MJ_ERR_INTERNAL, "MJ_ModuleMixer_QG::runShort", "Already active");
    return;
  }
  if(!MJ_kernel) new MJ_Kernel;

  MJ_kernel->playModule(this, MJ_KERNEL_NOTHREAD_PLAY);
}

void MJ_ModuleMixer_QG::stopShort2()
{
  if(MJ_kernel) {
    if(!MJ_kernel->stopPlaying())
      error(MJ_ERR_INTERNAL, "MJ_ModuleMixer_QG::dspStopShort2 0");
  }
  else error(MJ_ERR_INTERNAL, "MJ_ModuleMixer_QG::dspStopShort2 1");
}

void MJ_ModuleMixer_QG::resizeMixer()
{
  MJ_ModuleMixerResize *mr = new MJ_ModuleMixerResize(m_mixer, getMutex());

  connect(mr, SIGNAL(emitMixerConf(int,int,int)), 
	  this, SLOT(mixerConf(int,int,int)));

  // If "this" is deleted, also delete the child.
  connect(this, SIGNAL(destroyed()), 
	  mr, SLOT(hide()));
}

void MJ_ModuleMixer_QG::mixerConf(int s, int r, int b)
{
  if(m_debug) {
    char buf[128];
    sprintf(buf, "MJ_ModuleMixer_QG::mixerConf # %d %d %d", s, r, b);
    trace(buf);
  }

  if(isRealTime()) return;

  m_mixer->mixerSatellite()->modified4(MJ_MICH_RESIZE_PREPARE, s, r);
  m_mixer->resize(s, r);
  m_mixer->mixerSatellite()->modified4(MJ_MICH_RESIZE_FINAL, s, r);

  if(m_mixer->nofBuses() != (uint) b) {
    m_mixer->setBuses(b);
    m_action = MJ_MICH_IO;
    m_mixer->mixerSatellite()->modified2(MJ_MICH_IO);
    m_action = -1;
  }
}

void MJ_ModuleMixer_QG::removeModule(int strip, int row)
{
  if(m_debug) trace("MJ_ModuleMixer_QG::removeModule");
  
  if(isRealTime()) return;
  
  placeModule(0, strip, row);
  rebuildStripHeader(strip);
  if(m_debug) trace("MJ_ModuleMixer_QG::removeModule", "EXIT");
}

void MJ_ModuleMixer_QG::viewStrip(int strip)
{
  const char *fname = "MJ_ModuleMixer_QG::viewStrip";

  if(m_debug) trace(fname);

  /* First check if the GUI already exists. If it does, bring up the
     old one.*/
  uint swgs = m_stripWindowGuis.count();

  for(uint i=0; i < swgs; i++) {
    MJ_MixerStrip_QG *sq = (MJ_MixerStrip_QG *) m_stripWindowGuis.at(i);
    if((int) sq->strip() == strip) {
      sq->show();
      sq->raise();
      return;
    }
  }

  // No old GUI, generate a new one.

  MJ_MixerStrip_QG * msg = 
    new MJ_MixerStrip_QG(this, m_mixer, strip, this, 0, 0);

  connect(this, SIGNAL(destroyed()), msg, SLOT(hide()));
  connect(this, SIGNAL(reload()), msg, SLOT(hide()));

  connect(msg, SIGNAL(emitDestroyed(QWidget *)),
	  this, SLOT(forgetChild(QWidget *)));

  msg->show();
  m_stripWindowGuis.append(msg);
}

void MJ_ModuleMixer_QG::viewModule(int strip, int row)
{
  const char *fname = "MJ_ModuleMixer_QG::viewModule";

  if(m_debug) trace(fname);

  QWidget *&current = m_moduleWindowGuis.get(row, strip);
  
  if(current) {
    current->move(QCursor::pos());
    current->show();
    current->raise();
    return;
  }

  lock();
  MJ_MixerUnit *u = m_mixer->unit(strip, row);
  unlock();
  
  if(!u) {
    MJ_fatal(MJ_ERR_INTERNAL, fname, "No unit");
  }

  QWidget *ug = u->newGui(MJ_GUI_OWN_WINDOW, strip, row, this, 0);

  if(!ug) MJ_fatal(MJ_ERR_MEMORY, fname, "No gui");

  connect(this, SIGNAL(reload()), ug, SLOT(hide()));

  connect(ug, SIGNAL(emitDestroyed(QWidget *)), 
	  this, SLOT(forgetChild(QWidget *)));

  m_moduleWindowGuis.get(row, strip) = ug;
  
  // Determine icon/window name:
  
  QString tmp(tr("MJ")), tmp2;
  tmp += " ";
  tmp2 = MJ_Tr::tr("strip");
  tmp += tmp2;
  tmp2.sprintf(" %d ", strip + 1);
  tmp += tmp2;
  tmp2 = MJ_Tr::tr("row");
  tmp += tmp2;
  tmp2.sprintf(" %d", row + 1);
  tmp += tmp2;
  
  // Icon text:
  tmp2 = ug->iconText();
  if(!tmp2.isEmpty()) {
    tmp2 = tmp + " : " + tmp2;
  }
  else tmp2 = tmp;

  ug->setIconText(tmp2);

  // Caption:
  tmp2 = ug->caption();
  if(!tmp2.isEmpty()) {
    tmp2 = tmp + " : " + tmp2;
  }
  else tmp2 = tmp;

  ug->setCaption(tmp2);
  
  ug->show();
}

void MJ_ModuleMixer_QG::forgetModule(int strip, int row)
{
  // Delete previous GUI.
  QWidget **wcurr = & m_moduleGuis.get(row, strip);
  if(*wcurr) delete *wcurr;
  *wcurr = 0;
  
  wcurr = & m_moduleWindowGuis.get(row, strip);
  if(*wcurr) delete *wcurr;
  *wcurr = 0;
}

void MJ_ModuleMixer_QG::updateSlot(int strip,int row)
{
  QWidget *w = placeGui(m_mixer->unit(strip, row), strip, row);
  w->show();
  rebuildStripHeader(strip);
}

void MJ_ModuleMixer_QG::clear()
{
  if(isRealTime()) return;

  int choice = QMessageBox::warning(0, MJ_Tr::tr("Warning"),
				    MJ_Tr::tr("Empty the mixer ?"),
				    MJ_Tr::tr("Yes"),
				    MJ_Tr::tr("No"));
  
  if(choice == 1) return;

  clearInternal();
}

bool MJ_ModuleMixer_QG::loadModule(const QString &f)
{
  QFile file(f);
  if(!file.open(IO_ReadOnly)) return false;
  QTextStream stream(&file);

  emit reload(); // Should close all other GUIs related to this object.
  
  // Empty this GUI:
  resizeGrid(0, 0);

  bool ok = m_mixer->read(&stream);

  if(!ok)
    return false;

  int strips = m_mixer->strips();
  int rows = m_mixer->rows();

  resizeGrid(strips, rows);
  redraw(strips, rows);
  buildStripHeaders();

  showAll();

  if(m_module->satellite()) 
    m_module->satellite()->pushMruFile(f);

  return true;
}

void MJ_ModuleMixer_QG::startThreaded()
{
  if(!MJ_ModuleDspUi::startThreaded()) 
    error(MJ_ERR_INTERNAL, "MJ_ModuleMixer_QG::startThreaded");
}

void MJ_ModuleMixer_QG::stopThreaded()
{
  MJ_ModuleDspUi::stopPlaying();
}

void MJ_ModuleMixer_QG::viewAudioIo()
{
  if(m_audioIoGui) {
    m_audioIoGui->move(QCursor::pos());
    m_audioIoGui->show();
    m_audioIoGui->raise();
  }
  else {
    m_audioIoGui = new MJ_AudioIoGui(0, "Mixer aio");
    connect(m_audioIoGui, SIGNAL(enableDevice(bool,int,bool)),
	    this, SLOT(enableAudioIo(bool,int,bool)));
    m_audioIoGui->setEnabled(!isRealTime());
    updateAudioConf();
    m_audioIoGui->show();
  }
}

void MJ_ModuleMixer_QG::enableAudioIo(bool input, int index, bool enable)
{
  const char *fname = "MJ_ModuleMixer_QG::enableAudioIo";

  trace2("%s %d %d %d", fname, (int) input, index, (int) enable);

  if(isRealTime()) return;

  /* Translate to this relative system: */
  MJ_AudioDeviceType type = input ? MJ_ADT_INPUT : MJ_ADT_OUTPUT;
  
  trace2("%s #  type = %d, number = %d", fname, (int) type, index);
  
  bool ok;

  if(enable) {
    ok = m_mixer->addAudioDev(type, index);
  }
  else {
    ok = m_mixer->removeAudioDev(type, index);
  }
  
  if(!ok) {
    error2(MJ_ERR_UNKNOWN, 
	   "%s #  %s, %d", fname,
	   MJ_AudioDeviceTypeKeys[type], index);
  }

  if(m_debug) trace(fname, "HALF");

  if(MJ_AudioConf::isInput((MJ_AudioDeviceType) type))
    m_mixer->checkSources();
  else
    m_mixer->checkSends();

  m_mixer->mixerSatellite()->modified2(MJ_MICH_IO);

  if(m_debug) trace(fname, "EXIT");
}

void MJ_ModuleMixer_QG::viewMidiIo()
{
  if(m_midiConfGui) {
    m_midiConfGui->move(QCursor::pos());
    m_midiConfGui->show();
    m_midiConfGui->raise();
  }
  else {
    m_midiConfGui = new MJ_MidiConfGui(0, "Mixer mio");
    connect(m_midiConfGui, SIGNAL(enableDevice(bool,int,bool)),
	    this, SLOT(enableMidiIo(bool,int,bool)));
    m_midiConfGui->setEnabled(!isRealTime());
    updateMidiConf();
    m_midiConfGui->show();
  }
}
void MJ_ModuleMixer_QG::enableMidiIo(bool input, int index, bool enable)
{
  const char *fname = "MJ_ModuleMixer_QG::enableMidiIo";

  trace2("%s %d %d %d", fname, (int) input, index, (int) enable);

  if(isRealTime()) return;

  /* Translate to this relative system: */
  MJ_MidiDeviceType type = input ? MJ_MIDID_INPUT : MJ_MIDID_OUTPUT;
  
  trace2("%s #  type = %d, number = %d", fname, (int) type, index);
  
  bool ok;
  char buf[32];

  sprintf(buf, "%d %d", (int) type, index);

  MJ_MessageString strMsg(buf);
  
  if(enable) {
    ok = m_mixer->msgProcess(MJ_ModuleMixer::addMidiDevStr, &strMsg);
  }
  else {
    ok = m_mixer->msgProcess(MJ_ModuleMixer::removeMidiDevStr, &strMsg);
  }
  
  if(!ok) {
    error2(MJ_ERR_UNKNOWN, 
	   "%s #  %s, %d", fname,
	   MJ_MidiDeviceTypeKeys[type], index);
  }

  if(m_debug) trace(fname, "HALF");
  
  m_mixer->mixerSatellite()->modified2(MJ_MICH_IO);
    

  if(m_debug) trace(fname, "EXIT");
}

void MJ_ModuleMixer_QG::newModule(const QString &s)
{
  const char *fname = "MJ_ModuleMixer_QG::newModule";

  if(m_debug) MJ_trace(fname, s);

  if(isRealTime()) return;

  if((uint) m_activeStrip >= (uint) m_moduleGuis.cols() ||
     (uint) m_activeRow >= (uint) m_moduleGuis.rows())
    MJ_fatal(MJ_ERR_INTERNAL, fname);

  MJ_ModuleDsp *m = MJ_Kernel::newDspModule(s, MJ_PP_MENU_POSITION);

  if(!m) {
    MJ_error(MJ_ERR_ARGUMENT, fname, s);
    return; 
  }
  
  placeModule(m, m_activeStrip, m_activeRow); 
}

void MJ_ModuleMixer_QG::mixerSend()
{
  const char *fname = "MJ_ModuleMixer_QG::mixerSend";

  if(m_debug) trace(fname);

  if(isRealTime()) return;

  MJ_PluginDbeBase *dbe = MJ_kernel->dspLoader()->entryList().findFindDbe
    (MJ_mixerSendIdentifier);
    
  if(!dbe) MJ_fatal(MJ_ERR_INTERNAL, fname);
  
  newModule(dbe->menuPosition());
}

void MJ_ModuleMixer_QG::mixerSource()
{
  MJ_PluginDbeBase *dbe = MJ_kernel->dspLoader()->entryList().findFindDbe
    (MJ_mixerSourceIdentifier);

  if(!dbe) MJ_fatal(MJ_ERR_INTERNAL, "MJ_ModuleMixer_QG::mixerSource");

  newModule(dbe->menuPosition());
}

void MJ_ModuleMixer_QG::forgetChild(QWidget *w)
{
  trace3(m_debug, "MJ_ModuleMixer_QG::forgetChild %p", w);

  // Check if it is one the direct child modules
  for(uint i=0; i < m_moduleWindowGuis.rows(); i++) {
    for(uint j = 0; j < m_moduleWindowGuis.cols(); j++) {
      if(w == m_moduleWindowGuis.get(i, j)) {
	m_moduleWindowGuis.get(i, j) = 0;
	return;
      }
    }
  }

  // Check if it is one strip guis.
  int swg = m_stripWindowGuis.find((MJ_MixerStrip_QG *)w);
  
  if(swg == -1)
    error2(MJ_ERR_ARGUMENT, "MJ_ModuleMixer_QG::forgetChild %p", w);
  else
    m_stripWindowGuis.take(swg);
}

void MJ_ModuleMixer_QG::showModule(MJ_ModuleDsp *m)
{
  int strip, row;

  bool ok = m_mixer->locate(m, &strip, &row);
  if(!ok) {
    error(MJ_ERR_ARGUMENT, "MJ_ModuleMixer_QG::showModule");
    return;
  }
  viewModule(strip, row);
}

void MJ_ModuleMixer_QG::toggleMMC()
{
  int current = m_mixer->mmcController();

  int controller = current < 0 ? 0 : -1;

  m_mixer->setMmcController(controller);

  bool check = (controller >= 0);
  m_menuBar->edit()->setItemChecked(m_mmcId, check);
}

void MJ_ModuleMixer_QG::paintEvent(QPaintEvent *e)
{
  if(m_debug) trace("MJ_ModuleMixer_QG::paintEvent");

  QWidget::paintEvent(e);
  relayout();
  // drawVertSeparators();
  showAll();
}
void MJ_ModuleMixer_QG::showEvent(QShowEvent *)
{
  relayout();
  // drawVertSeparators();
  showAll(); 
}
void MJ_ModuleMixer_QG::drawVertSeparators()
{
  if(!m_moduleGuis.rows() || !isVisible()) return;

  QWidget *par = childParent();

  int h = par->height();
  int w = par->width();

  QPainter paint(par);

  paint.eraseRect(0,0,w,h);

  QColor color, colorBase(backgroundColor());
  int val = MJ_colorBrightness(colorBase);

  int scale;
  if(val > 128) scale = 70;
  else scale = 150;
  
  for(uint i = 1; i < m_moduleGuis.cols(); i++) {

    QWidget *wl = m_moduleGuis.get(0, i-1);
    QWidget *wr = m_moduleGuis.get(0, i);
    int leftMostLine = wl->frameGeometry().right();
    int rightMostLine = wr->frameGeometry().left();
    
    int shades = (rightMostLine - leftMostLine) / 2 + 1;

    color = colorBase;
    
    for(int j = 0; j < shades; j++) {
      paint.setPen(color);
      paint.drawLine(leftMostLine + j, 0, leftMostLine + j, h);
      paint.drawLine(rightMostLine - j, 0, rightMostLine - j, h);
      color = color.light(scale);
    }
  }
}

void MJ_ModuleMixer_QG::realTimeChange(bool nowRt)
{
  bool notRt = !nowRt;

  m_playMenu->setItemEnabled(m_menuIds[MJ_MIXER_RT_START_THREADED], notRt); 
  m_playMenu->setItemEnabled(m_menuIds[MJ_MIXER_RT_STOP_THREADED], nowRt);   

  if(m_audioIoGui) m_audioIoGui->setEnabled(notRt);
  
  uint i, count = m_pluginMenuFirstRow->count();

  for(i=0; i < count; i++) {
    int id = m_pluginMenuFirstRow->idAt(i);
    m_pluginMenuFirstRow-> setItemEnabled(id, notRt);
  }

  count = m_pluginMenu->count();
  for(i=0; i < count; i++) {
    int id = m_pluginMenu->idAt(i);
    m_pluginMenu-> setItemEnabled(id, notRt);
  }
  count = m_pluginMenuLastRow->count();
  for(i=0; i < count; i++) {
    int id = m_pluginMenuLastRow->idAt(i);
    m_pluginMenuLastRow-> setItemEnabled(id, notRt);
  }
}

void MJ_ModuleMixer_QG::hide()
{
  const char * fname = "MJ_ModuleMixer_QG::hide";

  if(m_debug) trace(fname);

  if(isRealTime()) {
    int res = QMessageBox::warning
      (0, 
       MJ_Tr::tr("Close mixer ?"),
       MJ_Tr::tr("The mixer is active, really close it ?"),
       MJ_Tr::tr("Yes"),
       MJ_Tr::tr("No"),
       QString::null,
       1);

    if(res == 0) stopThreaded();
    else return;
  }

  m_mixer->allUnitsOk();

  if(qApp->mainWidget() == this) {
    qApp->exit();
    return;
  }
  
  if(MJ_kernelGui) {
    // QWidget::hide();
    // MJ_kernelGui->deleteSetup(this);
    close();
  }
  else {
    error(MJ_ERR_INTERNAL, "MJ_ModuleMixer_QG::hide");
  }

  if(m_debug) trace(fname, "EXIT");
}

void MJ_ModuleMixer_QG::draw()
{
  if(m_debug) trace("MJ_ModuleMixer_QG::draw");

  // Create a filter that selects only the plugins we need.
  MJ_PluginFilterMulti fm(MJ_LOGIC_AND, false);

  // These child filters are needed
  MJ_DspPluginFilterNoInput noInput;
  MJ_DspPluginFilterNoInput noOutput;
  MJ_DspPluginFilterMetaFunc chainable
    (MJ_DspPluginMetaFunctionalityKeys[MJ_DPMF_CHAINABLE]);

  // Filter that hides audio I/O plugins.
  MJ_DspPluginFilterParam audioIn
    (MJ_DspPluginFunctionalityKeys[MJ_DPF_AUDIO_IN], MJ_PP_FUNCTIONALITY);
  MJ_DspPluginFilterParam audioOut
    (MJ_DspPluginFunctionalityKeys[MJ_DPF_AUDIO_OUT], MJ_PP_FUNCTIONALITY);

  MJ_PluginFilterMulti noAudio(MJ_LOGIC_NOT, false);
  noAudio.addFilter(&audioIn);
  noAudio.addFilter(&audioOut);

  // Build the plugin menus
  { // First row
    QPopupMenu *pluginMenu = new QPopupMenu(this);
    m_pluginMenuFirstRow = pluginMenu;
    
    fm.addFilter(&noInput);
    fm.addFilter(&chainable);
    fm.addFilter(&noAudio);
    
    // Build the menu
    MJ_KernelGui::moduleMenu(this, 
			     SLOT(newModule(const QString &)),
			     pluginMenu,
			     &fm);
    pluginMenu->insertSeparator();
    pluginMenu->insertItem(MJ_Tr::tr(MJ_Strs::AudioSource), 
			   this, SLOT(mixerSource()));
  }
  { // All middle rows
    QPopupMenu *pluginMenu = new QPopupMenu(this);
    m_pluginMenu = pluginMenu;
    
    fm.clear();
    fm.addFilter(&chainable);
    fm.addFilter(&noAudio);

    // Build the menu
    MJ_KernelGui::moduleMenu(this, 
			     SLOT(newModule(const QString &)),
			     pluginMenu,
			     &fm);
    pluginMenu->insertSeparator();
    pluginMenu->insertItem(MJ_Tr::tr(MJ_Strs::AudioSource), 
			   this, SLOT(mixerSource()));

    pluginMenu->insertItem(MJ_Tr::tr(MJ_Strs::AudioSend), 
			   this, SLOT(mixerSend()));
  }


  { // Final row
    QPopupMenu *pluginMenu = new QPopupMenu(this);
    m_pluginMenuLastRow = pluginMenu;
    
    fm.clear();
    // fm.addFilter(&noOutput);
    fm.addFilter(&chainable);
    fm.addFilter(&noAudio);
    
    // Build the menu
    MJ_KernelGui::moduleMenu(this, 
			     SLOT(newModule(const QString &)),
			     pluginMenu,
			     &fm);
    pluginMenu->insertSeparator();

    pluginMenu->insertItem(MJ_Tr::tr(MJ_Strs::AudioSend), 
			   this, SLOT(mixerSend()));
  }

  lock();
  uint rows =  m_mixer->rows();
  uint strips = m_mixer->strips();  
  unlock();

  m_playMenu = new QPopupMenu(this);

  m_menuIds[MJ_MIXER_RT_START_THREADED] = 
    m_playMenu->insertItem(MJ_Tr::tr("&Run"), this, 
			   SLOT(startThreaded()));

  m_menuIds[MJ_MIXER_RT_STOP_THREADED] = 
  m_playMenu->insertItem(MJ_Tr::tr("&Stop"), this, 
			 SLOT(stopThreaded()), Key_Escape);
  
  m_playMenu->setItemEnabled(m_menuIds[MJ_MIXER_RT_STOP_THREADED],
			     false);

  m_menuBar = new MJ_ModuleMenubar(this);

  QPopupMenu *edit = m_menuBar->edit();

  edit->insertSeparator();
  m_mmcId = edit->insertItem( MJ_Tr::tr("&Use MMC (experimental)"), 
			      this, SLOT(toggleMMC()));
  lock();
  bool enable = m_mixer->midiConf().numDevices(MJ_MIDID_INPUT);
  unlock();
  edit->setItemEnabled(m_mmcId, enable);

  edit->insertSeparator();
  edit->insertItem( MJ_Tr::tr("&Resize"), this, SLOT(resizeMixer()));
  edit->insertItem( MJ_Tr::tr("&Audio I/O"), this, SLOT(viewAudioIo()));
  edit->insertItem( MJ_Tr::tr("&MIDI I/O"), this, SLOT(viewMidiIo()));
  edit->insertItem( MJ_Tr::tr("&Clear mixer"), this, SLOT(clear()));
  
  m_menuBar->insertItem( MJ_Tr::tr("&Play"), m_playMenu);

  if(MJ_Debug::allowTestMenus()) { 
    QPopupMenu *tests = new QPopupMenu(m_menuBar);
    
    tests->insertItem( MJ_Tr::tr("Write"), this, SLOT(writeCout()));
    tests->insertItem( MJ_Tr::tr("Run short"), this, SLOT(runShort()));
    tests->insertItem( MJ_Tr::tr("&Start non-threaded"), 
		       this, SLOT(runShort2()));
    tests->insertItem( MJ_Tr::tr("&Stop non-threaded"), 
		       this, SLOT(stopShort2()));
    m_menuBar->insertItem( MJ_Tr::tr("&Tests"), tests);
  }
  
  QSize vs(m_layoutEngine.calculateSize(rows + 1, strips));
  m_scrollView->resizeContents(vs.width(), vs.height());

  /* if(m_debug) {
    QCString str;
    str.sprintf("CHILD SIZE = %d x %d", vs.width(), vs.height());
    MJ_trace(str);
    }*/


  QVBoxLayout *layout = new QVBoxLayout(this);

  m_moduleGuis.resize(rows, strips);
  m_moduleGuis.setAll(0);
  
  m_moduleWindowGuis.resize(rows, strips);
  m_moduleWindowGuis.setAll(0);

  layout->setMenuBar(m_menuBar);
  layout->addWidget(m_scrollView);

  redraw(strips, rows);

  m_stripHeaders.resize(strips);
  m_stripHeaders.setAll(0);
  buildStripHeaders();

  showAll();

  if(m_debug) m_mixer->allUnitsOk();
}

void MJ_ModuleMixer_QG::redraw(int strips, int rows)
{
  if(m_debug) trace("MJ_ModuleMixer_QG::redraw");

  for(uint i=0; i < (uint) strips; i++) {
    for(uint j=0; j < (uint) rows; j++) {
      QWidget *w = m_moduleGuis.get(j, i);

      if(!w) {
	lock();
	MJ_MixerUnit *u = m_mixer->unit(i, j);
	unlock();
	
	placeGui(u, i, j);
      }
    }
  }
  buildStripHeaders();
}

void MJ_ModuleMixer_QG::resizeGrid(int strips, int rows)
{
  if(m_debug) trace("MJ_ModuleMixer_QG::resizeGrid");

  const uint grows = m_moduleGuis.rows();
  const uint gstrips = m_moduleGuis.cols();
  
  uint i, j;

  QSize vs(m_layoutEngine.calculateSize(rows + 1, strips));
  m_scrollView->resizeContents(vs.width(), vs.height());
 
  // Resize the header strip;
  for(i = strips; i < gstrips; i++) 
    delete m_stripHeaders[i];
  m_stripHeaders.resize(strips);

  for(i = (uint) gstrips; (int) i < strips; i++)
    m_stripHeaders[i] = 0;

  // Delete the units that will be lost:
  for( i = 0; i < gstrips; i++) {
    for( j = 0; j < grows; j++) {
      if((i >= (uint) strips) || (j >= (uint) rows)) {
	forgetModule(i, j);
      }
    }
  }

  // Set NULL pointers to elements that are not there yet:
  m_moduleWindowGuis.resize(rows, strips);
  m_moduleGuis.resize(rows, strips);
  for(i = 0; i < (uint) strips; i++) {
    for(j = 0; j < (uint) rows; j++) {
      if((i >= gstrips) || (j >= grows)) {
	m_moduleGuis.get(j, i) = 0;
	m_moduleWindowGuis.get(j, i) = 0;
      }
    }
  }

  buildStripHeaders();
  m_mixer->allUnitsOk();
  showAll();
}

void MJ_ModuleMixer_QG::showAll()
{
  uint rows = m_moduleGuis.rows();
  uint cols = m_moduleGuis.cols();

  for(uint j = 0; j < cols; j++) {
    QWidget *w = m_stripHeaders.get(j);
    if(w) w->show();

    for(uint i=0; i < rows; i++) {
      w = m_moduleGuis.get(i, j);
      if(w) w->show();
    }
  }
}

void MJ_ModuleMixer_QG::relayout()
{
  QRect r;
  
  uint rows = m_moduleGuis.rows();
  uint cols = m_moduleGuis.cols();


  for(uint j = 0; j < cols; j++) {
    QWidget *w = m_stripHeaders.get(j);
    if(w) {
      r = m_layoutEngine.calculateGeometry(0, j);
      w->setGeometry(r);
    }

    for(uint i=0; i < rows; i++) {
      w = m_moduleGuis.get(i, j);
      if(w) {
	r = m_layoutEngine.calculateGeometry(i + 1, j);
	w->setGeometry(r);
      } 
    }
  }
}

void MJ_ModuleMixer_QG::buildStripHeaders()
{
  for(uint i=0; i < m_stripHeaders.size(); i++) {
    rebuildStripHeader(i);    
  }
}

void MJ_ModuleMixer_QG::rebuildStripHeader(uint strip)
{
  if(m_debug) trace("MJ_ModuleMixer_QG::rebuildStripHeader", strip);

  QWidget *&w = m_stripHeaders[strip];

  MJ_ModuleStripHeader *sh; 

  if(w) {
    sh = (MJ_ModuleStripHeader *) w;
    sh->clearHeader();
  }
  else {
    sh = new MJ_ModuleStripHeader(strip, this, this, m_scrollView->viewport(), 
				  "strip header");

    m_scrollView->addChild(sh);

    connect(sh, SIGNAL(emitDelete(int,int)),
	    this, SLOT(removeModule(int,int)));

    connect(sh, SIGNAL(emitViewStrip(int)),
	    this, SLOT(viewStrip(int)));

    connect(sh, SIGNAL(emitViewModule(int,int)),
	    this, SLOT(viewModule(int,int)));

  }

  sh->rebuild();
  
  QRect r = m_layoutEngine.calculateGeometry(0, strip);
  sh->setGeometry(r);
  m_scrollView->moveChild(sh, r.x(), r.y());
  sh->show();
  w = sh;
}

void MJ_ModuleMixer_QG::updateAudioConf()
{
  MJ_AudioIoGui *aig = m_audioIoGui;
  if(!aig) return;

  MJ_AudioConf *aic = & m_mixer->audioConf();

  bool blocked = aig->signalsBlocked();
  aig->blockSignals(true);
  
  uint count = aic->devices().count();
  
  // bool input;
  
  MJ_2Int mi;
    
  aig->uncheckAll();

  for(uint i=0; i < count; i++) {
    /* We should not need to lock. Lets do it anyway, the data *is*
       critical.*/
    lock();
    mi = *aic->device(i);
    unlock();
    
    aig->setChecked(MJ_AudioConf::isInput((MJ_AudioDeviceType) mi.m_v[0]), 
		    mi.m_v[1], true);

    if(m_debug) trace2("MJ_ModuleMixer_QG::updateAudioConf checked %s %d",
		       MJ_AudioDeviceTypeKeys2[mi.m_v[0]], mi.m_v[1]);
  }
  /* else if(m_debug)
     trace("MJ_ModuleMixer_QG::updateAudioConf no", absIndex); */

  aig->blockSignals(blocked);
}

void MJ_ModuleMixer_QG::updateMidiConf()
{
  MJ_MidiConfGui *mcg = m_midiConfGui;
  if(!mcg) return;

  MJ_MidiConf *mic = & m_mixer->midiConf();

  bool blocked = mcg->signalsBlocked();
  mcg->blockSignals(true);
  
  uint count = mic->devices().count();
  
  // bool input;
  
  MJ_2Int mi;
    
  mcg->uncheckAll();

  for(uint i=0; i < count; i++) {
    /* We should not need to lock. Lets do it anyway, the data *is*
       critical.*/
    lock();
    mi = *mic->device(i);
    unlock();
    
    mcg->setChecked(MJ_MidiConf::isInput((MJ_MidiDeviceType) mi.m_v[0]), 
		    mi.m_v[1], true);

    if(m_debug) trace2("MJ_ModuleMixer_QG::updateMidiConf checked %s %d",
		       MJ_MidiDeviceTypeKeys2[mi.m_v[0]], mi.m_v[1]);
  }
  /* else if(m_debug)
     trace("MJ_ModuleMixer_QG::updateMidiConf no", absIndex); */

  mcg->blockSignals(blocked);
}

void MJ_ModuleMixer_QG::deleteChildren()
{
  for(uint i=0; i < m_moduleWindowGuis.rows(); i++) {
    for(uint j = 0; j < m_moduleWindowGuis.cols(); j++) {
      QWidget *&w = m_moduleWindowGuis.get(i, j);
      if(w) {
	delete w;
	w = 0;
      }
    }
  }

  m_stripWindowGuis.clear();
}

void MJ_ModuleMixer_QG::clearInternal()
{
  lock();
  int strips = m_mixer->strips();
  int rows = m_mixer->rows();
  unlock();

  for(int i=0; i < strips; i++) {
    for(int j=0; j < rows; j++) {
      if(m_mixer->block(i,j))
	removeModule(i, j);
    }
  }
}

QWidget * MJ_ModuleMixer_QG::placeGui(MJ_MixerUnit *u, uint strip, uint row)
{
  const char * fname = "MJ_ModuleMixer_QG::placeGui";

  /* {
    char buf[32];
    sprintf(buf, "%d %d", strip, row);
    trace(fname, buf);
    }*/

  if(m_debug) trace2("%s # %u %u", fname, strip, row);

  QWidget *w = m_moduleGuis.get(row, strip);
  
  // Remove old:
  if(w) delete w;
  w = u->newGui(MJ_GUI_STAMP, strip, row, this, m_scrollView->viewport());

  if(!w) error2(MJ_ERR_INTERNAL, "%s # u->newGui failed", fname);

  m_moduleGuis.get(row, strip) = w;

  QRect r = m_layoutEngine.calculateGeometry(row + 1, strip);
  
  if(!w)
    return 0;

  w->setGeometry(r);
  m_scrollView->addChild(w, r.x(), r.y());
 
  if(m_debug) {
    trace2("%s # EXIT x = %d, y = %d", fname, w->x(), w->y());
  }

  return w;
}
