/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_mixer_send.h>

#include <mj_error.h>
#include <mj_math.h>
#include <mj_misc.h>
#include <mj_module_mixer.h>
#include <mj_module_mixer_source.h>
#include <mj_parameters.h>
#include <mj_satellite.h>
#include <mj_strs.h>
#include <mj_tr.h>

#include <qlayout.h> 
#include <qscrollview.h> 
#include <qtextstream.h>

const char MJ_ModuleMixerSend::m_keyWord[] = "MJ_MIXER_SEND";

const char *MJ_ModuleMixerSend::m_params[] = {
  "add",
  "remove"
};

MJ_ModuleMixerSend::MJ_ModuleMixerSend()
  // : m_channels(0)
{}

MJ_ModuleMixerSend::~MJ_ModuleMixerSend()
{
  // if(m_satellite) delete m_satellite;
}

/** Sets the parameters of a particular mapping element. */

void MJ_ModuleMixerSend::setMapping
(uint ind, int type, int sourceChannel, int targetChannel)
{
  MJ_3Int *map = & m_mappings.get(ind);
  map->m_v[0] = type;
  map->m_v[1] = sourceChannel;
  map->m_v[2] = targetChannel;
}


bool MJ_ModuleMixerSend::isActiveMapping(uint ind, MJ_ModuleMixer *mixer)
{
  MJ_3Int *map = & m_mappings.get(ind);

  if(map->m_v[1] >= dspOutChannels())
    return false;

  return mixer->validConnection(map->m_v[0], map->m_v[2]);
}

/** DOES NOTHING. This method used to be necessary. Today, it is
    not. It used to remove all the illegal mappings. One day I'll
    remove it. */

bool MJ_ModuleMixerSend::makeLegal(MJ_ModuleMixer *)
{
  /* const char * fname = "MJ_ModuleMixerSend::makeLegal";

  MJ_trace(fname);
  
  uint ind = 0;
  bool modified = false;
  
  while(ind  < m_mappings.size()) {
    
    MJ_3Int *map = & m_mappings.get(ind);

    if(mixer->validConnection(map->m_v[0], map->m_v[2]))
      ind++;
    else {
      forgetMapping(ind);
      MJ_trace2("%s # Removing map %d", fname, ind);
      modified = true;
    }
  }
  return modified; */
  return true;
}

bool MJ_ModuleMixerSend::writeParameters(QTextStream *s)
{
  if(!MJ_ModuleGain::writeParameters(s)) return false;

  uint size = m_mappings.size();
  
  (*s) << size << "\n";

  for(uint i = 0; i < size; i++) {
    MJ_3Int *map = & m_mappings.getFast(i);

    for(uint j= 0; j < 3; j++) 
      (*s) << map->m_v[j] << "  ";

    (*s) << "\n";
  }
  return true;
}

bool MJ_ModuleMixerSend::readParameters(QTextStream *s)
{
  if(!MJ_ModuleGain::readParameters(s)) return false;

  const char * fname = "MJ_ModuleMixerSend::read";

  bool ok = true;
  uint size;
  QCString cstr1;

  (*s) >> cstr1;
  size = cstr1.toInt(&ok);
  
  if(!ok) return false;

  m_mappings.resize(size); 

  for(uint i = 0; i < size; i++) {
    MJ_3Int *map = & m_mappings.getFast(i);
    for(uint j=0; j < 3; j++) {
      (*s) >> cstr1;
      map->m_v[j] = cstr1.toInt(&ok);
      if(!ok) {
	MJ_error(MJ_ERR_DATA_CORRUPTED, fname);
	return false;
      }
    }
  }

  s->readLine();
  return ok;
}



/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

bool MJ_ModuleMixerSend_QG::m_debug = true;

MJ_ModuleMixerSend_QG::MJ_ModuleMixerSend_QG(MJ_ModuleMixerSend *send,
					     MJ_ModuleMixer     *mixer,
					     MJ_RtControl       *rtparent,
					     QWidget      *parent,
					     const char   *name)
  : MJ_ModuleDspGuiPostCard(rtparent, send, true,
			    1, 
			    parent == 0  ? MJ_GUI_OWN_WINDOW : MJ_GUI_POSTCARD,
			    parent, name),
    m_mixerSend(send),
    m_mixer(mixer)
{
  const char * fname = "MJ_ModuleMixerSend_QG::MJ_ModuleMixerSend_QG";

  if(m_debug) MJ_trace(fname);

  MJ_cpf(mixer, fname);

  m_scroll = new QScrollView(this);
  MJ_cpf(m_scroll, fname);
  
  m_scroll->setMouseTracking(true);
  
  m_mapper = new MJ_ConnectView(m_scroll->viewport(), "source connect");
  MJ_cpf(m_mapper, fname);
  
  int rows = m_layout->numRows();
  m_layout->expand(rows + 1, m_layout->numCols());
  m_layout->addMultiCellWidget(m_scroll, rows, rows, 0, m_layout->numCols()-1);

  m_scroll->addChild(m_mapper);
  m_scroll->setMouseTracking(true);
  setMouseTracking(true);
  
  connect(m_mapper, SIGNAL(newConnection(int,int)), 
	  this, SLOT(newConnection(int,int)));
  connect(m_mapper, SIGNAL(deleteConnection(int)), 
	  this, SLOT(deleteConnection(int)));

  connect(send->satellite(), SIGNAL(emitModified(const char *)), 
	  this, SLOT(sendModified(const char *)));

  connect(mixer->satellite(), SIGNAL(emitModified2(int)), 
  this, SLOT(mixerModified(int)));
  
  updateView();

  m_mapper->setEnabled(!isRealTime());

  m_scroll->show();
  m_mapper->show();

  setMinimumSize(100, 100);

  if(!parent) resize(500, 300);
}

MJ_ModuleMixerSend_QG::~MJ_ModuleMixerSend_QG()
{
  if(m_debug) MJ_trace("MJ_ModuleMixerSend_QG::~MJ_ModuleMixerSend_QG");

  emit emitDestroyed(this); 
}

void MJ_ModuleMixerSend_QG::realTimeChange(bool nowRt)
{
  m_mapper->setEnabled(!nowRt);
}


MJ_ModuleDspGui *MJ_ModuleMixerSend_QG::createNew(MJ_RtControl *rtparent,
						  MJ_ModuleDsp *module,
						  QWidget *parent, 
						  const char *name)
{
  MJ_ModuleMixer *mixer = dynamic_cast<MJ_ModuleMixer *> 
    (module->parentModule());
  
  if(!mixer) {
    char buf[64];
    sprintf(buf, "MJ_ModuleMixerSend_QG::createNew # %p # %p", 
	    module, module->parentModule());
    MJ_fatal(MJ_ERR_ARGUMENT, buf);
  }

  return new MJ_ModuleMixerSend_QG((MJ_ModuleMixerSend *) module,
				   mixer, 
				   rtparent,
				   parent, 
				   name);
}

void MJ_ModuleMixerSend_QG::mixerModified(int ind)
{
  MJ_trace("MJ_ModuleMixerSend_QG::mixerModified", ind);

  if(ind == MJ_MICH_IO) {
    m_mixerSend->makeLegal(m_mixer);
 
    updateView();
  }
}
void MJ_ModuleMixerSend_QG::moduleIoModified(uint flags)
{
  if(flags & MJ_MIO_SIGNAL_INPUT)
    updateView();
}



void MJ_ModuleMixerSend_QG::sendModified(const char *id)
{
  // MJ_trace("MJ_ModuleMixerSend_QG::sendModified2", ind);

  int i = MJ_findStr(id, MJ_ModuleMixerSend::m_params, MJ_ModuleMixerSend::DO_SIZEOF);
  

  if(i >= 0) updateView();
}

void MJ_ModuleMixerSend_QG::newConnection(int from, int to)
{
  if(isRealTime()) return;
  
  // MJ_trace("MJ_MixerSource_QG::newConnection", from);

  uint achans = m_mixer->audioChansOut();
  
  int type;
  
  if((uint) to >= achans) {
    type = MJ_MIXERSOURCE_BUS;
    to -= achans;
  }
  else {
    type = MJ_MIXERSOURCE_AUDIO_OUTPUT;
  }
  
  m_mixerSend->addMappings(1);
  m_mixerSend->setMapping(m_mixerSend->nofMappings() - 1, type, from, to);

  m_mixerSend->satellite()->modified
    (MJ_ModuleMixerSend::m_params[MJ_ModuleMixerSend::DO_ADD]);
}

void MJ_ModuleMixerSend_QG::deleteConnection(int ind)
{
  if(isRealTime()) return;
  
  // MJ_trace("MJ_ModuleMixerSend_QG::deleteConnection", ind);
  
  m_mixerSend->forgetMapping(ind);
  m_mixerSend->satellite()->modified
    (MJ_ModuleMixerSend::m_params[MJ_ModuleMixerSend::DO_REMOVE]);
}

void MJ_ModuleMixerSend_QG::resizeEvent(QResizeEvent *)
{

  QSize contensSize(m_scroll->viewportSize(1, 1));
  
  const QSize hint = m_mapper->minimumSizeHint();
  
  int minW = MJ_max(hint.width(), contensSize.width());
  int minH = MJ_max(hint.height(), contensSize.height());

  if(minW != contensSize.width() || minH != contensSize.height()) {
    m_mapper->drawBase();
    m_mapper->resize(minW,minH);
    m_mapper->redraw();
  }
}

void MJ_ModuleMixerSend_QG::updateView()
{
  const char * fname = "MJ_ModuleMixerSend_QG::updateView";
  char buf[128];

  if(m_debug) MJ_trace(fname);
  
  std::vector<MJ_LabelDef> leftLabels;
  std::vector<MJ_LabelDef> rightLabels;
  std::vector<MJ_2Int>     connections;

  uint i, j, labelCount = 0;
  int send;

  lock();
  int auChans = m_mixer->audioChansOut();
  MJ_ModuleWeightedSum * buses = & m_mixer->buses();
  uint busCount = buses->dspOutChannels();
  QList<MJ_ModuleDsp> *audios = & m_mixer->audioIos();
  const uint nofAudios = audios->count();
  const uint mchannels = m_mixerSend->dspOutChannels();
  unlock();
  
  const uint totalCount = auChans + busCount;
  
  leftLabels.resize(mchannels);
  rightLabels.resize(totalCount);

  QString tmp1, tmp2;

  if(m_debug) {
    sprintf(buf, "%s # nofAudios = %d # busCount = %d # mchannels = %d", 
	    fname, (int) nofAudios, (int) busCount, (int) mchannels);
    MJ_trace(buf);
  }
  
  /* Get the audio channel names. */
  for(i = 0; i < nofAudios; i++) {
    
    lock();
    MJ_ModuleDsp *mdsp = audios->at(i);
    MJ_Parameter *p = mdsp->msgGetInput(MJ_Strs::device);
    const uint chans = mdsp->dspInChannels();
    unlock();
    
    if(!p) {
      MJ_error(MJ_ERR_INTERNAL, fname, "No message");
      return;
    }

    // This is not an input.
    if(!chans) {
      delete p;
      continue;
    }
    
    MJ_ParameterString *ps = dynamic_cast<MJ_ParameterString *> (p);
    if(!ps) {
      MJ_error(MJ_ERR_INTERNAL, fname, "No string");
      delete p;
      return;
    }
    
    for(j=0; j < chans; j++) {
      tmp1 = ps->string();
      if(tmp1.isEmpty()) tmp1 = MJ_Tr::tr("Audio output");
      
      tmp2.sprintf(" %d",j+1);
      tmp1 += tmp2;
      
      MJ_LabelDef * sl = & rightLabels[labelCount];
      sl->m_string = tmp1;
      
      labelCount++;
    }
    delete p;
  }

  /* Get the bus channel names. */
  for(i = 0; i < busCount; i++) {
    tmp1 = MJ_Tr::tr("Bus (internal channel)");
    tmp2.sprintf(" %d",i+1);
    tmp1 += tmp2;
    
    MJ_LabelDef * sl = & rightLabels[labelCount];
    
    sl->m_string = tmp1;

    labelCount++;
  }

  /* Get target channel names: */

  for(i = 0; i < mchannels; i++) {
    
    if(mchannels == 2)
      tmp1 = MJ_Tr::tr(MJ_StereoKeys[i]);
    else {
      tmp2.setNum(i+1);
      tmp1 = MJ_Tr::tr("Channel") + " " + tmp2;
    }
    
    MJ_LabelDef * sl = & leftLabels[i];
    sl->m_string = tmp1;
  }
  
  /* Get the connections: */

  lock();
  uint nofMaps = m_mixerSend->nofMappings();
  unlock();
  
  connections.resize(nofMaps);

  uint usedConnections = 0;

  for(i = 0; i < nofMaps; i++) {

    lock();
    bool activeMap = m_mixerSend->isActiveMapping(i, m_mixer);
    const MJ_3Int connections3 = m_mixerSend->mapping(i);
    unlock();
    
    if(!activeMap) continue;

    usedConnections++;

    // MJ_3Int *connections3 = & tmp.getFast(i);
    MJ_2Int *connections2 = & connections[i];

    send = connections3.m_v[2];

    if(connections3.m_v[0] == MJ_MIXERSOURCE_BUS) {
      send += auChans;
    }
    connections2->m_v[0] = connections3.m_v[1];
    connections2->m_v[1] = send;
  }

  connections.resize(usedConnections);

  m_mapper->resize(2000, 2000);
  m_mapper->set(leftLabels, rightLabels, connections);
}
