/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_mixer_source.h>

#include <mj_error.h>
#include <mj_iostream.h>
#include <mj_math.h>
#include <mj_module_mixer.h>
#include <mj_parameters.h>
#include <mj_strs.h>
#include <mj_tr.h>

#include <qcombobox.h> 
#include <qlabel.h> 
#include <qlayout.h> 
#include <qslider.h> 
#include <qscrollview.h> 

const char *MJ_MixerSourceTypes[] = {
  "Audio input",
  "Bus"
};

static const char *MJ_mmsKey = "MIXER_SOURCE";

MJ_ModuleMixerSource::MJ_ModuleMixerSource()
  : m_gain(1.0)
{}

MJ_ModuleMixerSource::~MJ_ModuleMixerSource()
{}

bool MJ_ModuleMixerSource::dspPrepare()
{
  /* trace("MJ_ModuleMixerSource::dspPrepare");
  for(uint i = 0; i < m_mappings2.size(); i++) {
    MJ_writeMultiInt(m_mappings2.getConst(i).m_v, 2, & MJ_cout());
    MJ_cout() << "\n";
    }*/
  
  return m_signalInput.size() == m_mappings2.size();
}

void MJ_ModuleMixerSource::dspProcessPeriod()
{
  dspClearOutputs();

  uint max = m_mappings2.size();
  MJ_2Int *mappings = m_mappings2.data();

  // MJ_Vector<MJ_Sample>::vmlBegin();
  for(uint i = 0; i < max; i++) {
    MJ_2Int *map = mappings++;
    
    const MJ_Vector<MJ_Sample> &inVect = m_signalInput[map->m_v[0]].samples();
    MJ_Vector<MJ_Sample> &outVect = m_signalOutput[map->m_v[1]].samples();

    // MJ_Vector<MJ_Sample>::vmlEnd();
    /* for(uint x = 0; x < inVect.size(); x++) {
      if(inVect.getConst(x) > 100) 
	trace2("TOO WIDE IN SIGNAL %d %d %f", i, x, inVect.getConst(x));
      if(outVect.get(x) > 100) 
	trace2("TOO WIDE OUT SIGNAL %d %d %f", i, x, outVect.get(x) );
	}*/

    outVect.macVml(inVect, m_gain);
  }
  MJ_Vector<MJ_Sample>::vmlEnd(); 
}

bool MJ_ModuleMixerSource::writeParameters(QTextStream *s)
{
  uint maps = m_mappings.size();

  (*s) << MJ_mmsKey << " " << maps << "\n";

  for(uint i=0; i < maps; i++) {
    MJ_writeMultiInt(m_mappings.getConst(i).m_v, 3, s);
    (*s) << "\n";
  }
  return true;
}

bool MJ_ModuleMixerSource::readParameters(QTextStream *s)
{
  const char * fname = "MJ_ModuleMixerSource::readParameters";

  QCString str;
  bool ok = true;
  uint maps;

  (*s) >> str;

  if(str != MJ_mmsKey) {
    error(MJ_ERR_DATA_CORRUPTED, fname, 1);
    return false;
  }
  
  (*s) >> str;
  maps = str.toUInt(&ok);

  if(!ok) {
    error(MJ_ERR_DATA_CORRUPTED, fname, 2);
    return false;
  }

  m_mappings.resize(maps);
  
  for(uint i=0; i < maps; i++) {
    MJ_readMultiInt(m_mappings.get(i).m_v, 3, s);
  }
  s->readLine();
  return true;
}


void MJ_ModuleMixerSource::setMapping
(uint ind, int type, int sourceChannel, int targetChannel)
{
  MJ_3Int *map = & m_mappings.get(ind);
  map->m_v[0] = type;
  map->m_v[1] = sourceChannel;
  map->m_v[2] = targetChannel;
}

void MJ_ModuleMixerSource::forgetMapping(int ind)
{
  m_mappings.removeFast(ind);
}

bool MJ_ModuleMixerSource::makeLegal()
{
  const char * fname = "MJ_ModuleMixerSource::makeLegal";

  // trace(fname);
  
  MJ_ModuleDsp *m = parentModule();

  if(!m) return false;

  uint ind = 0;
  bool modified = false;

  MJ_ModuleMixer *mixer = dynamic_cast<MJ_ModuleMixer *> (m);

  if(!mixer) return false;
  
  while(ind  < m_mappings.size()) {
    
    MJ_3Int *map = & m_mappings.get(ind);

    if(mixer->validConnection(map->m_v[0], map->m_v[1]) && 
       (uint) map->m_v[2] < m_signalOutput.size())
      ind++;
    else {
      forgetMapping(ind);
      trace(fname, "Removing map");
      modified = true;
    }
  }
  return modified;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_ModuleMixerSource_QG::MJ_ModuleMixerSource_QG(MJ_RtControl *rtparent,
						 MJ_ModuleDsp *module,
						 QWidget      *parent,
						 const char   *name)
  : MJ_ModuleDspGui(rtparent, module, parent, name),
    m_mixerSource(0),
    m_mixer(0),
    m_mapper(0),
    m_scroll(0)
{
  const char * fname = "MJ_ModuleMixerSource_QG::MJ_ModuleMixerSource_QG";

  m_newChans = module->dspOutChannels();

  draw();
  setMinimumSize(100, 100);

  m_mixerSource = dynamic_cast<MJ_ModuleMixerSource *> (module);

  if(!m_mixerSource) MJ_fatal(MJ_ERR_ARGUMENT, fname);
  
  lock();
  if(m_mixerSource->parentModule()) {
    m_mixer = dynamic_cast<MJ_ModuleMixer *> (m_mixerSource->parentModule());
  }
  int chans = module->dspOutChannels();
  unlock();

  setChannels(chans);

  if(!m_mixer) MJ_fatal(MJ_ERR_ARGUMENT, fname);
  
  connect(m_mixer->satellite(), SIGNAL(emitModified2(int)), 
	  this, SLOT(mixerModified(int)));
    
  m_mapper->setEnabled(!isRealTime());
  updateView();

  if(!parent) resize(500, 300);
}

MJ_ModuleMixerSource_QG::~MJ_ModuleMixerSource_QG()
{}

MJ_ModuleDspGui *MJ_ModuleMixerSource_QG::createNew(MJ_RtControl *rtparent,
						    MJ_ModuleDsp *module,
						    QWidget *parent, 
						    const char *name)
{
  return new MJ_ModuleMixerSource_QG(rtparent, module, parent, name);
}

void MJ_ModuleMixerSource_QG::moduleModified2(int ind)
{
  if(ind == CHANGE_CONNECTION) { 
    updateView();
    // redo();
  }
}


void MJ_ModuleMixerSource_QG::setChannels(int i)
{
  m_newChans = i;
  QString t;
  t.setNum(i);
  m_chanLabel->setText(t);
}

void MJ_ModuleMixerSource_QG::updateChannels()
{
  if(isRealTime()) return;
 
  if(m_mixerSource->dspOutChannels() == m_newChans)
    return;

  int strip, row;
 
  m_mixerSource->dspSetOutChannels(m_newChans);
  if(!m_mixer->locate(m_mixerSource, & strip, & row))
    return;
  
  m_mixer->rerouteStripAfter(strip, row);
  
  m_mixer->mixerSatellite()->modified2(MJ_MICH_IO);
}


void MJ_ModuleMixerSource_QG::mixerModified(int ind)
{
  m_mixerSource->makeLegal();

  if(ind == MJ_MICH_IO) updateView();
}

void MJ_ModuleMixerSource_QG::updateView()
{
  const char * fname = "MJ_ModuleMixerSource_QG::updateView";
  
  // trace(fname);

  std::vector<MJ_LabelDef> leftLabels;
  std::vector<MJ_LabelDef> rightLabels;
  std::vector<MJ_2Int>     connections;

  uint i, j, labelCount = 0;
  int source;
  int auChans = audioChansIn();

  lock();
  MJ_ModuleWeightedSum * buses = & m_mixer->buses();
  uint busCount = buses->dspOutChannels();
  QList<MJ_ModuleDsp> *audios = & m_mixer->audioIos();
  const uint nofAudios = audios->count();
  const uint mchannels = m_mixerSource->dspOutChannels();
  unlock();
  
  const uint totalCount = auChans + busCount;

  leftLabels.resize(totalCount);
  rightLabels.resize(mchannels);

  QString tmp1, tmp2;
  
  trace(fname);

  /* Get the audio channel names. */
  for(i = 0; i < nofAudios; i++) {

    lock();
    MJ_ModuleDsp *mdsp = audios->at(i);
    MJ_Parameter *p = mdsp->msgGetInput(MJ_Strs::device);
    const uint chans = mdsp->dspOutChannels();
    unlock();

    if(!p) {
      MJ_error(MJ_ERR_INTERNAL, fname, "No device message");
      return;
    }

    MJ_ParameterString *ps = dynamic_cast<MJ_ParameterString *> (p);
    if(!ps) {
      MJ_error(MJ_ERR_INTERNAL, fname, "No string");
      delete p;
      return;
    }
    
    for(j=0; j < chans; j++) {
      tmp1 = ps->string();
      if(tmp1.isEmpty()) tmp1 = MJ_Tr::tr("Audio input");
      
      tmp2.sprintf(" %d",j+1);
      tmp1 += tmp2;
      
      MJ_LabelDef * sl = & leftLabels[labelCount];
      sl->m_string = tmp1;
      
      labelCount++;
    }
    delete ps;
  }

  /* Get the bus channel names. */
  for(i = 0; i < busCount; i++) {
    tmp1 = MJ_Tr::tr("Bus (internal channel)");
    tmp2.sprintf(" %d",i+1);
    tmp1 += tmp2;
    
    MJ_LabelDef * sl = & leftLabels[labelCount];
    
    sl->m_string = tmp1;

    labelCount++;
  }

  /* Get target channel names: */

  for(i = 0; i < mchannels; i++) {
    
    if(mchannels == 2)
      tmp1 = MJ_Tr::tr(MJ_StereoKeys[i]);
    else {
      tmp2.setNum(i+1);
      tmp1 = MJ_Tr::tr("Channel") + " " + tmp2;
    }
    
    MJ_LabelDef * sl = & rightLabels[i];
    sl->m_string = tmp1;
  }
  
  /* Get the connections: */

  lock();
  MJ_Array<MJ_3Int> tmp(m_mixerSource->sourceDefs());
  unlock();
  
  uint size = tmp.size();
  connections.resize(size);

  for(i = 0; i < size; i++) {

    MJ_3Int *connections3 = & tmp.getFast(i);
    MJ_2Int *connections2 = & connections[i];

    source = connections3->m_v[1];

    if(connections3->m_v[0] == MJ_MIXERSOURCE_BUS) {
      source += auChans;
    }
    connections2->m_v[0] = source;
    connections2->m_v[1] = connections3->m_v[2];
  }
  // m_mapper->resize(2000, 2000);
  m_mapper->set(leftLabels, rightLabels, connections);
}

void MJ_ModuleMixerSource_QG::newConnection(int from, int to)
{
  if(isRealTime()) return;
  
  // trace("MJ_ModuleMixerSource_QG::newConnection", from);

  uint achans = m_mixer->audioChansIn();

  int type;
  
  if((uint) from >= achans) {
    type = MJ_MIXERSOURCE_BUS;
    from -= achans;
  }
  else {
    type = MJ_MIXERSOURCE_AUDIO_INPUT;
  }

  m_mixerSource->addMappings(1);
  m_mixerSource->setMapping(m_mixerSource->nofMappings() - 1, type, from, to);

  m_mixer->mixerSatellite()->modified2(CHANGE_CONNECTION);
}

void MJ_ModuleMixerSource_QG::deleteConnection(int ind)
{
  if(isRealTime()) return;
  
  trace("MJ_ModuleMixerSource_QG::deleteConnection", ind);

  m_mixerSource->forgetMapping(ind);

  m_mixer->mixerSatellite()->modified2(CHANGE_CONNECTION);
}

void MJ_ModuleMixerSource_QG::realTimeChange(bool nowRt)
{  
  m_mapper->setEnabled(!nowRt);
  m_chanSlider->setEnabled(!nowRt);
}

void MJ_ModuleMixerSource_QG::resizeEvent(QResizeEvent *)
{
  m_mapper->drawBase();

  QSize contensSize(m_scroll->viewportSize(1, 1));
  
  const QSize hint = m_mapper->minimumSizeHint();
  
  int minW = MJ_max(hint.width(), contensSize.width());
  int minH = MJ_max(hint.height(), contensSize.height());
       
  m_mapper->resize(minW, minH);
  m_mapper->redraw();
  /* 
     qDebug
     ("MJ_ModuleMixerSource_QG::resizeEvent w = %d  h = %d rw = %d rh = %d", 
     minW, minH, hint.width(), hint.height());
  */

  // m_scroll->viewport()->resize(minW, minH);
}

void MJ_ModuleMixerSource_QG::draw()
{
  trace("MJ_ModuleMixerSource_QG::draw");

  QVBoxLayout *hbox = new QVBoxLayout(this, 5, 5);

  QHBoxLayout *vbox = new QHBoxLayout(5);
  hbox->addLayout(vbox);

  QLabel *l = new QLabel(MJ_Tr::tr("Channels"), this);
  vbox->addWidget(l);

  m_chanSlider = new QSlider(1, MJ_MOD_MIXER_SOURCE_MAX_CHANNELS,
			     1, m_newChans,  
			     QSlider::Horizontal, this, "chanslider");

  vbox->addWidget(m_chanSlider);
  connect(m_chanSlider, SIGNAL(valueChanged(int)),
	  this, SLOT(setChannels(int)));
  
  connect(m_chanSlider, SIGNAL(sliderReleased()),
	  this, SLOT(updateChannels()));
  
  m_chanLabel = new QLabel(this);
  vbox->addWidget(m_chanLabel);

  
  m_scroll = new QScrollView(this);
  m_scroll->setMouseTracking(true);

  m_mapper = new MJ_ConnectView(m_scroll->viewport(), "source connect");
  
  hbox->addWidget(m_scroll, 50);

  m_scroll->addChild(m_mapper);
  m_scroll->setMouseTracking(true);
  setMouseTracking(true);

  connect(m_mapper, SIGNAL(newConnection(int,int)), 
	  this, SLOT(newConnection(int,int)));
  connect(m_mapper, SIGNAL(deleteConnection(int)), 
	  this, SLOT(deleteConnection(int)));
}

uint MJ_ModuleMixerSource_QG::audioChansIn()
{
  lock();
  uint count = m_mixer->audioChansIn();
  unlock();

  return count;
}
