/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_monodelay.h>
#include <mj_error.h>
#include <mj_iostream.h>
#include <mj_math.h>
#include <mj_module_keys.h>

#include <mj_module_dsp_tui_default.h>
#include <mj_strs.h>
#include <mj_tr.h>

#include <mj_parameters.h>
#include <mj_global.h>
#include <mj_kernel.h>

#include <qobject.h>



MJ_ModuleMonoDelay::MJ_ModuleMonoDelay()
  : m_time(30.0),
    m_wetRatio(0.5),
    m_dryRatio(0.5),
    m_feedBack(0.5)
{
  m_signalSourceDefs.resize(1);
  m_signalOutput.resize(1);
}

MJ_ModuleMonoDelay::~MJ_ModuleMonoDelay()
{}

bool MJ_ModuleMonoDelay::dspPrepare()
{
  if(m_signalInput.size()  != 1 || 
     m_signalOutput.size() != 1 || 
     m_time < 1.0) return false;

  m_delayBuffer.resize(100000);
  m_delayBuffer.setAll(0.0);
  return true;
}

void MJ_ModuleMonoDelay::dspProcessPeriod()
{
  const MJ_Vector<MJ_Sample> *in = & m_signalInput.data()->samples();
  MJ_Vector<MJ_Sample> *out = & m_signalOutput.data()->samples();

  ulong t = (ulong) m_time;

  const float dryRatio = m_dryRatio;
  const float wetRatio = m_wetRatio;
  const float feedBack = m_feedBack;
  
  const MJ_Sample *inb = in->dataConst();
  const MJ_Sample *inmax = & inb[in->size()];
  MJ_Sample *outb = out->data();

  while(inb < inmax) {

    float delay = m_delayBuffer.getNewest(t); 
    const float &inSample = *inb++;

    *outb++ = inSample * dryRatio + delay * wetRatio; // output
    m_delayBuffer.getNewest() = 
      delay * feedBack + inSample; // feedback
    m_delayBuffer.advance();
  }
}


bool MJ_ModuleMonoDelay::msgProcess(const char * i, MJ_Message *msg)
{
  bool ok = true;

  if(!strcmp(i, MJ_Strs::delay)) {
    double delay = msg->toDouble(&ok);
    if(ok) {
      m_time = MJ_clampF((float) delay, 1, 100000); // Limit the range
      return true;
    }
  }
  else if(!strcmp(i, MJ_Strs::feedback)) {
    double fb = msg->toDouble(&ok);    
    if(ok) {
      m_feedBack = MJ_clampF((float) fb, 0.0, 0.99);
      return true;
    }
  }
  else if(!strcmp(i, MJ_Strs::wetratio)) {
    double wr = msg->toDouble(&ok);
    if(ok) {
      m_wetRatio = MJ_clampF((float) wr, 0.0, 1.0);
      m_dryRatio = (float) 1.0 - (float) wr;
      return true;
    }
  }

  return false;
}

MJ_Parameter * MJ_ModuleMonoDelay::msgGetInput(const char * i)
{
  if(!strcmp(i, MJ_Strs::delay)) {
    return new MJ_ParameterInt("Delay time", i, 
			       (int) m_time, 1, 100000, MJ_UNIT_SAMPLES);

    /* MJ_ParameterDouble *pd = new MJ_ParameterDouble;
       pd->set("Delay time", i, 
       m_time, 1, 100000, MJ_UNIT_SAMPLES);
       return pd; */
  }
  else if(!strcmp(i, MJ_Strs::feedback)) {    
    MJ_ParameterDouble *pd = new MJ_ParameterDouble;
    pd->set("Feedback ratio", i,
	    m_feedBack, 0.0, 0.99, MJ_UNIT_BLANK);
    return pd;
  }
  else if(!strcmp(i, MJ_Strs::wetratio)) {    
    MJ_ParameterDouble *pd = new MJ_ParameterDouble;
    pd->set("Wet ratio", i,
	    m_wetRatio, 0.0, 1.0, MJ_UNIT_BLANK);
    return pd;
  }
  return 0;
}

QCString MJ_ModuleMonoDelay::msgGetInputName(int i)
{
  QCString s;

  if(i == 0) s = MJ_Strs::delay;
  else if(i == 1) s = MJ_Strs::feedback;
  else if(i == 2) s = MJ_Strs::wetratio;

  return s;
}


void MJ_ModuleMonoDelay::testTui(int argc, char **argv)
{
  MJ_ModuleDsp *m=0;

  MJ_Kernel k;

  const char * identifier = MJ_monoDelayIdentifier;

  if(argc) {
    identifier = argv[0];
    MJ_cout() << "Trying to get module " << identifier << "\n";
    MJ_cout().device()->flush();
  }

  // Create a new DSP module.
  if(MJ_kernel) m = MJ_Kernel::newDspModule(identifier, MJ_PP_IDENTIFIER);
  
  if(!m) {
    MJ_error(MJ_ERR_INTERNAL, "MJ_ModuleMonoDelay::testTui, module not in DB");
    m = new MJ_ModuleMonoDelay;
  }

  MJ_ModuleDspTui *mt =  m->newTui(0, "test TUI");

  delete mt;  // First delete the GUI
  delete m;   // Then delete the module
}
