/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <GL/gl.h>

#ifdef __OS_LINUX__

// NVidia fixes...
#define GL_GLEXT_PROTOTYPES 1

#ifndef GLAPI
#define GLAPI 
#endif

#ifdef APIENTRY
#undef APIENTRY
#endif

#include <GL/glext.h>

#endif

#include <mj_module_phase_display.h>

#include <mj_error.h>
#include <mj_math.h>
#include <mj_misc.h>
#include <mj_parameters.h>
#include <mj_strs.h>
#include <mj_tr.h>

#include <qlayout.h>

MJ_ModulePhaseDisplay::MJ_ModulePhaseDisplay()
  : m_gain(1.0)
{ 
}

MJ_ModulePhaseDisplay::~MJ_ModulePhaseDisplay()
{}

bool MJ_ModulePhaseDisplay::dspPrepare()
{
  if((uint)m_signalInput.size() != 2)  
    MJ_fatal(MJ_ERR_INTERNAL, "MJ_ModulePhaseDisplay::dspPrepare");

  m_buffer.resize(1024);

  MJ_DualSample s;
  s.clear();

  m_buffer.setAll(s);

  return true;
}

void MJ_ModulePhaseDisplay::dspProcessPeriod()
{
  const MJ_Vector<MJ_Sample> * in = & m_signalInput.getFast(0).samples();
  const uint l = in->size();

  const MJ_Sample * left = in->dataConst();
  const MJ_Sample * right = m_signalInput.getFast(1).samples().dataConst();
  
  const MJ_Sample *sentinel = & left[l];
    
  MJ_DualSample s;

  while(left < sentinel) {
    
    s.m_v[0] = *left;
    s.m_v[1] = *right;
    
    m_buffer.put(s);
    
    left++;
    right++;
  }
  // m_peak = peak;
}

bool MJ_ModulePhaseDisplay::msgProcess(const char * id, MJ_Message * m)
{
  bool ok = true;

  if(!strcmp(id, MJ_Strs::gain)) {
    double tmp = m->toDouble(&ok);
    if(ok)
      m_gain = tmp;
  }

  return ok;
}

MJ_Parameter * MJ_ModulePhaseDisplay::msgGetInput(const char *id)
{
  if(!strcmp(id, MJ_Strs::gain)) {
    MJ_ParameterDouble *pd = new MJ_ParameterDouble;
    pd->set(MJ_Strs::Gain, MJ_Strs::gain,
	    MJ_absToDb(m_gain), -10, 30, MJ_UNIT_DECIBEL);
    return pd;
  }

  return 0;
}

QCString MJ_ModulePhaseDisplay::msgGetInputName(int i)
{
  QCString s;
  if(i == 0) s = MJ_Strs::gain;
  return s;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_PhaseWidget::MJ_PhaseWidget(const QGLFormat & format,
			       MJ_RtControl *rtparent,
			       MJ_ModulePhaseDisplay * phase,
			       QWidget      *parent,
			       const char   *name)
  : QGLWidget(format, parent, name), 
    MJ_RtControl(rtparent), 
    m_phase(phase) 
{
  connect( & m_timer, SIGNAL(timeout()), this, SLOT(updateGL()));

  realTimeChange(isRealTime());
}

MJ_PhaseWidget::~MJ_PhaseWidget()
{}

void MJ_PhaseWidget::clearGl()
{
  makeCurrent();

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  
  doneCurrent();
}

void MJ_PhaseWidget::initializeGL ()
{
  clearGl();
}

void MJ_PhaseWidget::resizeGL ( int width, int height )
{
  clearGl();
}

void MJ_PhaseWidget::paintGL()
{
  int i, n;

  lock();
  m_buffer = m_phase->m_buffer;
  float scale = m_phase->m_gain;
  unlock();

  scale = MJ_dbToAbs(scale);

  int w = width();
  int h = height();

  glViewport(0, 0, w, h);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();

  if(w > h) {
    float wscale = (float) w / (float) h;
    gluOrtho2D(-wscale, wscale, -1, 1);
  }
  else {
    float wscale = (float) h / (float) w;
    gluOrtho2D(-1, 1, -wscale, wscale);
  }

  glMatrixMode(GL_MODELVIEW);

  glLoadIdentity();
  float ss = 1.0f / (1.2f * sqrtf(2.0f));
  glScalef(ss, ss, ss);
  glRotatef(45.0, 0.0, 0.0, 1.0);

  glDisable(GL_DEPTH_TEST);

  glEnable(GL_BLEND);
  
  glBlendFunc(GL_SRC_ALPHA, GL_ONE);

  glDisable(GL_LIGHTING);

  glBlendEquation(GL_FUNC_REVERSE_SUBTRACT);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE);
  glColor4f(1, 1, 1, 0.035);

  /* Draw a rectangle to fade out the old stuff: */
  glBegin(GL_QUADS);
  glVertex2f(-20, -20);
  glVertex2f( 20, -20);
  glVertex2f( 20,  20);
  glVertex2f(-20,  20);
  glEnd();

  // Now draw the surroundings:

  glBlendEquation(GL_FUNC_ADD);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  glColor4f(0.4, 0.4, 0.4, 0.3);

  glBegin(GL_TRIANGLE_STRIP);
  
  float b = 1.1f;

  glVertex2f(-b, -b);
  glVertex2f(-1, -1);
  glVertex2f( b, -b);
  glVertex2f( 1, -1);

  glVertex2f( b,  b);
  glVertex2f( 1,  1);
  glVertex2f(-b,  b);
  glVertex2f(-1,  1);
  
  glVertex2f(-b, -b);
  glVertex2f(-1, -1);

  glEnd();

  // Draw the axis marks:

  glColor4f(0.7, 0.7, 1, 0.3);
  glBegin(GL_LINES);
  
  n = 9;
  
  int half = (n - 1) / 2;
  for(i = 0; i < n; i++) {
    float y = (i - half) / (float) half;
    glVertex2f(-1, y);
    glVertex2f( 1, y);

    glVertex2f(y, -1);
    glVertex2f(y,  1);
  }

  glEnd();

  // Now draw the the phase display:

  glScalef(scale, scale, scale);

  glBlendEquation(GL_FUNC_ADD);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE);
  glColor4f(0.7, 0.7, 1, 0.3);

  n = m_buffer.size();

  glBegin(GL_LINE_STRIP);

  for(i = 0; i < n; i++)
    glVertex2fv(m_buffer.getIndex(i).m_v);

  glEnd();

}

void MJ_PhaseWidget::realTimeChange (bool nowRt)
{
  MJ_trace2("MJ_PhaseWidget::realTimeChange # %d", (int) nowRt);

  if(nowRt) {
    clearGl();
    m_buffer.resize(1024);
    m_timer.start(50);
  }
  else
    m_timer.stop();
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


MJ_ModulePhaseDisplayGui::MJ_ModulePhaseDisplayGui(MJ_RtControl *rtparent,
						   MJ_ModuleDsp *module,
						   QWidget      *parent,
						   const char   *name)
  : MJ_ModuleDspGuiPostCard(rtparent, module, true,
			    (uint) -1, 
			    parent ? MJ_GUI_STAMP : MJ_GUI_OWN_WINDOW,
			    parent, name)
{
  const char * fname = "MJ_ModulePhaseDisplayGui::MJ_ModulePhaseDisplayGui";

  QVBoxLayout * mainLayout = dynamic_cast<QVBoxLayout *> 
    (child("mainlayout"));
  
  MJ_cpf(mainLayout, fname);

  QGLFormat format;
  format.setDoubleBuffer(false);
  format.setDepth(false);

  MJ_PhaseWidget * phase = 
    new MJ_PhaseWidget(format, this, 
		       dynamic_cast<MJ_ModulePhaseDisplay *> (module), 
		       this, "PhaseGL");
  mainLayout->addWidget(phase, 80);
  phase->show();
  realTimeChange(isRealTime());
}

MJ_ModulePhaseDisplayGui::~MJ_ModulePhaseDisplayGui()
{}

MJ_ModuleDspGui *MJ_ModulePhaseDisplayGui::createNew(MJ_RtControl *rtparent,
						     MJ_ModuleDsp *module,
						     QWidget *parent, 
						     const char *name)
{
  return new MJ_ModulePhaseDisplayGui(rtparent, module, parent, name);
}
