/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_span.h>

#include <mj_parameters.h>
#include <mj_strs.h>
#include <mj_tr.h>

#include <qobject.h>

MJ_ModuleStereoPan::MJ_ModuleStereoPan()
  : m_targetTime(0)
{
  m_targetGains[0] = 0.5;
  m_targetGains[1] = 0.5;

  targetToCurr();

  setPan(0.0);
}

MJ_ModuleStereoPan::~MJ_ModuleStereoPan()
{}

bool MJ_ModuleStereoPan::dspPrepare()
{
  m_targetTime = 0;
  targetToCurr();
  return ((m_signalOutput.size() == 2) && 
	  (m_signalInput.size() >= 1) &&
	  (m_signalInput.size() <= 2));
}

void MJ_ModuleStereoPan::dspProcessPeriod()
{
  if(m_targetTime) {

    // Interpolate
    
    uint rsize = m_signalInput[0].samples().size();

    long until = m_targetTime - (long) m_currentTime->stamp();

    MJ_Sample suntil = (MJ_Sample) until;

    if(until > (long) rsize) {
      
      MJ_Sample rate =  (m_targetGains[0] - m_gains[0]) / suntil;
      
      const MJ_Vector<MJ_Sample> *in = & m_signalInput[0].samples();
      MJ_Vector<MJ_Sample> *out = & m_signalOutput[0].samples();    
      in->mulRamp(m_gains[0], rate, out);
      
      m_gains[0] += rate * rsize;
	
      rate =  (m_targetGains[1] - m_gains[1]) / suntil;    
      
      in = & m_signalInput[m_signalInput.size()-1].samples();
      out = & m_signalOutput[1].samples();
      in->mulRamp(m_gains[1], rate, out);
      
      m_gains[1] += rate * rsize;
    }
    else {
      
      // Interpolate the last frame
      
      for(int i=0; i < 2; i++) {

	const MJ_Sample * inp = 
	  m_signalInput[i % m_signalInput.size()].samples().data();

	MJ_Sample * outp = m_signalOutput[i].samples().data();
	
	MJ_Sample g = m_gains[i];
	MJ_Sample rate =  (m_targetGains[i] - g) / suntil;
	
	uint j;
	for(j=0; j < (uint) until; j++) { // Last ramp period
	  outp[j] = g * inp[j];
	  g += rate;
	}
	for(; j < rsize; j++)      // Back to fixed gain
	  outp[j] = g * inp[j];
      }
      // No more interpolation
      targetToCurr();

      m_targetTime = 0;
    }
  }
  else {

    // No interpolation.

    MJ_Vector<MJ_Sample>::vmlBegin();
    
    const MJ_Vector<MJ_Sample> *in = & m_signalInput[0].samples();
    MJ_Vector<MJ_Sample> *out = & m_signalOutput[0].samples();
    
    in->mulVml(m_gains[0], out);
    
    in = & m_signalInput[m_signalInput.size()-1].samples();
    out = & m_signalOutput[1].samples();
    in->mulVml(m_gains[1], out);
    
    MJ_Vector<MJ_Sample>::vmlEnd();
  }
}

bool MJ_ModuleStereoPan::msgProcess(const char * id, MJ_Message *m)
{
  bool ok = true;

  if(!strcmp(id, MJ_Strs::pan)) {
    double pan = m->toDouble(&ok);
    if(!ok) return false;
    
    setPan(pan);

    if(dspIsRealTime()) {

      // Set interpolation parameters.
      
      long mstamp = m->timeStamp();
      long cstamp = (long) m_currentTime->stamp();

      if(mstamp > cstamp)
	m_targetTime = mstamp;
      else {
	m_targetTime = cstamp + 
	  (long) ((float) dspSamplingRate() / (float) 30.0) + 1;
      }
    }
  }
  else ok = false;

  return ok;
}

MJ_Parameter * MJ_ModuleStereoPan::msgGetInput(const char * i)
{
  if(!strcmp(i, MJ_Strs::pan)) {
    return new MJ_ParameterDouble 
      (MJ_Strs::Pan, MJ_Strs::pan, getPan(), -1, 1);
  }
  return 0;
}

QCString MJ_ModuleStereoPan::msgGetInputName(int i)
{
  QCString s;
  if(i == 0) s = MJ_Strs::pan;
  return s;
}

void MJ_ModuleStereoPan::setPan(MJ_Sample pan)
{
  if(pan < -1.0) pan = -1.0;
  else if(pan > 1.0) pan = 1.0;

  m_targetGains[0] = (1.0 - pan) / 2.0;
  m_targetGains[1] = 1.0 - m_targetGains[0];
}
