/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */
#include <mj_module_wsum.h>
#include <mj_error.h>
#include <mj_math.h>
#include <mj_misc.h>
#include <mj_parameters.h>
#include <mj_strs.h>

#include <qbuffer.h> 
#include <qtextstream.h> 

#include <stdlib.h>

MJ_ModuleWeightedSum::MJ_ModuleWeightedSum()
  : m_defaultGain(1.0)
{}

MJ_ModuleWeightedSum::~MJ_ModuleWeightedSum()
{}

bool MJ_ModuleWeightedSum::dspSetInChannels(int chans)
{
  uint oldSize = m_gains.size();
  bool ok = MJ_ModuleDsp::dspSetInChannels(chans);
  if(ok) {
    m_gains.resize(chans);
    for(uint i= oldSize; i < m_gains.size(); i++) 
      m_gains[i] = m_defaultGain;
  }
  return ok;
}

bool MJ_ModuleWeightedSum::dspPrepare()
{
  if(m_moduleDspDebug) {

    QBuffer b;
    b.open(IO_ReadWrite);
    QTextStream s(&b);

    s << "MJ_ModuleWeightedSum::dspPrepare # GAINS = ";
    
    uint size = m_gains.size();
    const MJ_Sample *gains = m_gains.dataConst();

    for(uint i=0; i < size; i++) {
      s << gains[i] << "  ";
    }
    char * str = b.buffer().data();
    trace(str);
  }

  if(!m_signalOutput.size() && !m_gains.size() && !m_signalInput.size()) {
    return true;
  }

  bool ok = (m_signalInput.size() == m_gains.size()) && m_signalOutput.size();
  
  if(!ok)
    error2(MJ_ERR_INTERNAL, 
	   "MJ_ModuleWeightedSum::dspPrepare %d != %d \
|| m_signalOutput.size()",
	   m_signalInput.size(), m_gains.size(), m_signalOutput.size());

  return ok;
}

void MJ_ModuleWeightedSum::dspProcessPeriod()
{
  MJ_Vector<MJ_Sample> *out = & m_signalOutput[0].samples();
  out->setAll(0.0);

  for(uint i=0; i < m_signalInput.size(); i++) {

    const MJ_Vector<MJ_Sample> *in = & m_signalInput[i].samples();
    
    out->mac(*in, m_gains[i]);
  }
}

void MJ_ModuleWeightedSum::addInput(MJ_ModuleDsp *m, 
				    int channel, 
				    MJ_Sample gainDb)
{
  if(!m_signalSourceDefs.resize(m_signalSourceDefs.count() + 1)) return;
  
  m_gains.resize(m_signalSourceDefs.count());
  m_gains.last() = MJ_dbToAbs(gainDb);
  
  MJ_SignalSourceDef *sd = m_signalSourceDefs.last();
  sd->set(m->id(), channel);
  sd->setModule(m);
}


void MJ_ModuleWeightedSum::setGainDb(uint i, MJ_Sample gain)
{
  m_gains[i] = MJ_dbToAbs(gain);
}

bool MJ_ModuleWeightedSum::msgProcess(const char * id, MJ_Message *m)
{
  if(strncmp(id, MJ_Strs::gain, 4))
    return false;

  bool ok = true;

  char *endptr;

  long int index = strtol(&id[4], &endptr, 10);

  if(endptr) return false;

  if((uint) index < m_gains.size()) {
    double gain = m->toDouble(&ok);
    if(!ok) return false;
    setGainDb((uint) index, gain);
    return true;
  }
  return false;
}

MJ_Parameter * MJ_ModuleWeightedSum::msgGetInput(const char * id)
{
  trace("MJ_ModuleWeightedSum::msgGetInputs");

  if(strncmp(id, MJ_Strs::gain, 4))
    return 0;

  char *endptr;
  long int index = strtol(&id[4], &endptr, 10) - 1;

  if(endptr) return 0;
  
  if((uint) index < m_gains.size()) {
    MJ_ParameterDouble *pd = new MJ_ParameterDouble;
    pd->set(MJ_Strs::Gain, id,
	    MJ_absToDb(m_gains[index]), -100, 10, MJ_UNIT_DECIBEL);
    return pd;
  }
  return 0;
}

QCString MJ_ModuleWeightedSum::msgGetInputName(int i)
{
  QCString s;

  if((uint) i < m_gains.size()) {
    s.sprintf("%s%d", MJ_Strs::gain, i + 1);
  }
  return s;
}

/// NEVER CALL
MJ_ModuleWeightedSum &MJ_ModuleWeightedSum::operator=
(const MJ_ModuleWeightedSum &)
{
  MJ_fatal(MJ_ERR_DO_NOT_CALL, "MJ_ModuleWeightedSum::operator=");
  return *this;
}
