/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_modules_arithmetic.h>

#include <mj_math.h>
#include <mj_message.h>
#include <mj_parameter_numeric.h>


bool MJ_ModuleMul2::dspPrepare()
{
  return m_signalInput.size() == 2 && m_signalOutput.size() == 1;
}

void MJ_ModuleMul2::dspProcessPeriod()
{
  MJ_Vector<MJ_Sample>::vmlBegin();

  m_signalInput.getFast(0).samples().mulVml
    (m_signalInput.getFast(1).samples(), 
     & m_signalOutput.getFast(0).samples());
  
  MJ_Vector<MJ_Sample>::vmlEnd();
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

void MJ_ModuleAdd2::dspProcessPeriod()
{
  m_signalInput.getFast(0).samples().add
    (m_signalInput.getFast(1).samples(), 
     & m_signalOutput.getFast(0).samples());
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

void MJ_ModuleSub2::dspProcessPeriod()
{
  m_signalInput.getFast(0).samples().sub
    (m_signalInput.getFast(1).samples(), 
     & m_signalOutput.getFast(0).samples());
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

bool MJ_ModuleAbs::dspPrepare()
{
  return m_signalInput.size() == m_signalOutput.size();
}

void MJ_ModuleAbs::dspProcessPeriod()
{
  MJ_SignalSource *sources = m_signalInput.data();
  MJ_SignalSink *sinks = m_signalOutput.data();

  MJ_SignalSource *sentinel = sources + m_signalInput.size();

  while(sources < sentinel) {
    sources->samples().abs(& sinks->samples());
    sources++;
    sinks++;
  }
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

void MJ_ModuleSquare::dspProcessPeriod()
{
  MJ_SignalSource *sources = m_signalInput.data();
  MJ_SignalSink *sinks = m_signalOutput.data();

  MJ_SignalSource *sentinel = sources + m_signalInput.size();
  
  while(sources < sentinel) {
    sources->samples().square(& sinks->samples());
    sources++;
    sinks++;
  }
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

void MJ_ModuleSignedSquareRoot::dspProcessPeriod()
{
  MJ_SignalSource *sources = m_signalInput.data();
  MJ_SignalSink *sinks = m_signalOutput.data();

  MJ_SignalSource *sentinel = sources + m_signalInput.size();
  
  uint samples = sources->samples().size();


  while(sources < sentinel) {
    const MJ_Sample *in = sources->samples().dataConst();
    MJ_Sample *out = sinks->samples().data();

    for(uint i=0; i < samples; i++) {

      MJ_Sample s = in[i];

      if(s >= 0) *out = MJ_sqrt(s);
      else *out = MJ_sqrt(-s);

      out++;
    }

    sources++;
    sinks++;
  }
}


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_ModuleConst::~MJ_ModuleConst()
{}

bool MJ_ModuleConst::dspPrepare()
{
  return m_signalOutput.size() == 1;
}

void MJ_ModuleConst::dspProcessPeriod()
{
  m_signalOutput[0].samples().setAll(m_value);
}

bool MJ_ModuleConst::msgProcess(const char *id, MJ_Message *m)
{
  if(strcmp(id, "value") == 0) {
    bool ok = true;
    double v = m->toDouble(&ok);
    
    if(ok) 
      m_value = v;

    return ok;
  }
  
  return false;
}

MJ_Parameter * MJ_ModuleConst::msgGetInput(const char *id)
{
  if(strcmp(id, "value") == 0)
    return new MJ_ParameterDouble("Value", id, m_value, -100000.0, 100000);
  
  return 0;
}

QCString MJ_ModuleConst::msgGetInputName(int i)
{
  return QCString(i ? 0 : "value");
}
