/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */


#ifndef MJ_FREQ_RESP_H
#define MJ_FREQ_RESP_H

#include <mj_array.h>
#include <mj_config.h>

#include <mj_math.h>

/** Calculates the frequency response of a digital filter. 

    The frequencies are normalized - 1.0 = Nyquist.

    The frequnecy response is stored to an array that contains both
    real and imaginary elements related to a frequency point.  */

template <class TElem> class MJ_FreqResp
{
public:
  MJ_FreqResp() {}
  ~MJ_FreqResp() {}
  
  static void pointResponse(const MJ_Sample *b, uint bdegree,
			    const MJ_Sample *a, uint adegree,
			    TElem freq,
			    TElem *real, TElem * imag);

  void calculate(const MJ_Sample *b, uint bdegree,
		 const MJ_Sample *a, uint adegree,
		 uint points, 
		 TElem minFreq = 0.001,
		 TElem maxFreq = 1.0,
		 bool  logScale = true);

  /** Returns the frequency response. */
  const MJ_Array<TElem> & response() const { return m_response; }

  inline TElem realN(uint n) const { return m_response.getConst(n*2); }
  inline TElem imagN(uint n) const { return m_response.getConst(n*2+1); }
  inline TElem freqN(uint n) const { return m_freqs.getConst(n); }

  /// Returns the absolute value of Nth point.
  inline double absN(uint n) const
  { 
    const TElem &r = m_response.getConst(n*2); 
    const TElem &i = m_response.getConst(n*2+1); 
    return MJ_sqrt((r*r + i*i)); 
  }

  /// Returns the phase value of Nth point in radians
  inline double phaseN(uint n) const
  { 
    const TElem &r = m_response.getConst(n*2); 
    const TElem &i = m_response.getConst(n*2+1); 
    return atan2((double) i, (double) r);
  }

protected:
  MJ_Array<TElem> m_response;
  MJ_Array<TElem> m_freqs;
};

#endif
