/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_freq_resp.h>
#include <mj_freq_resp_gui.h>

#include <mj_error.h>
#include <mj_graphic_axes.h>
#include <mj_graphic_polyline.h>
#include <mj_graphics_context_simple.h> 

#include <mj_filter_iir2.h> 
#include <mj_misc.h>

#include <qapplication.h> 

/** Constructor for single-curve display. */

MJ_FreqRespGui::MJ_FreqRespGui(QWidget      *parent, 
			       const char   *name,
			       float         widthDb,
			       float         minFreq,
			       const QColor *graphColor,
			       int           graphWidth)
  : MJ_GraphicsArea(parent, name),
    m_curves(1)
{
  MJ_GraphicAxes *ax = new MJ_GraphicAxes();
  m_curve = new MJ_GraphicPolyLine[1];

  MJ_cpf(ax, m_curve, "MJ_FreqRespGui::MJ_FreqRespGui");

  if(graphColor || (graphWidth != 1)) {
    
    if(!graphColor) graphColor = & Qt::black;

    MJ_GraphicsContextPen *gp = 
      new MJ_GraphicsContextPen(QPen(*graphColor, graphWidth));
    
    m_curve->setGraphicsContext(gp);
  }

  appendGraphic(m_curve);

  ax->setAxDistance(50);
  appendGraphic(ax);

  setBoundingBox(minFreq, widthDb, 1.0, -widthDb);
}

/** Constructor for multi-band display. */

MJ_FreqRespGui::MJ_FreqRespGui(int filters,
			       QWidget      *parent, 
			       const char   *name,
			       float         widthDb,
			       float         minFreq)
  : MJ_GraphicsArea(parent, name),
    m_curve(0),
    m_curves(0),
    m_hasSum(false)
{
  MJ_GraphicAxes *ax = new MJ_GraphicAxes();
  
  MJ_cpf(ax, "MJ_FreqRespGui::MJ_FreqRespGui");

  ax->setAxDistance(50);
  appendGraphic(ax);
  
  resizeFilters(filters);
  
  setBoundingBox(minFreq, widthDb, 1.0, -widthDb);
}
  
MJ_FreqRespGui::~MJ_FreqRespGui()
{  
  clearFilters();
  if(m_curve) delete [] m_curve;
}


void MJ_FreqRespGui::resizeFilters(int filters)
{
  clearFilters();

  if(m_curve) delete [] m_curve;
  
  m_curves = filters + 1;
  m_curve = new MJ_GraphicPolyLine[m_curves];
  
  for(uint i=0; i < m_curves; i++) 
    appendGraphic(& m_curve[i]);
  
}

/** Remove all filter graphs. */
void MJ_FreqRespGui::clearFilters()
{
  m_hasSum = false;
  
  for(uint i=0; i < m_curves; i++) 
    removeGraphic(m_curve[i].id(), false);
}

/** Adds a filter in multi-filter mode. */
void MJ_FreqRespGui::setNthFilter(uint filter,
				  const MJ_Sample *b, uint bdegree,
				  const MJ_Sample *a, uint adegree)
{ 
  const char * fname = "MJ_FreqRespGui::addFilter";

  if(filter >= m_curves - 1)
    MJ_fatal(MJ_ERR_ARGUMENT, fname);
  
  MJ_GraphicPolyLine *curve = & m_curve[filter];
      
  formCurve(curve, b, bdegree, a, adegree);
}

/** Set the graphics context of the nth filter curve. */

void MJ_FreqRespGui::setNthCurveGC(uint curve, MJ_GraphicsContext *gc)
{
  const char * fname = "MJ_FreqRespGui::setNthCurveGC";

  if(curve >= m_curves)
    MJ_fatal(MJ_ERR_ARGUMENT, fname);
  
  m_curve[curve].setGraphicsContext(gc);
}

/** Brings a filter graph to top in multi-filter mode. This method
    only changes the data structures, it does not update the actual
    look immediately. */
void MJ_FreqRespGui::bringToTop(int filter)
{
  const char * fname = "MJ_FreqRespGui::bringToTop";

  if((uint) filter >= m_curves)
    MJ_fatal(MJ_ERR_ARGUMENT, fname);

  int ind = m_graphics.find(& m_curve[filter]);

  if(ind == -1) MJ_fatal(MJ_ERR_INTERNAL, fname);

  MJ_Graphic *g = m_graphics.take(ind);
  m_graphics.append(g);
}

/** Brings the filter sum graph to top in multi-filter mode. */
void MJ_FreqRespGui::sumToTop()
{
  int ind = m_graphics.find(& m_curve[m_curves - 1]);

  if(ind == -1) MJ_fatal(MJ_ERR_INTERNAL, "MJ_FreqRespGui::sumToTop");

  MJ_Graphic *g = m_graphics.take(ind);
  m_graphics.append(g);
}

/** Calculates filter sum in multi-filter mode. */
void MJ_FreqRespGui::calculateSum()
{
  m_hasSum = true;

  MJ_GraphicPolyLine *sumCurve = & m_curve[m_curves-1];
  sumCurve->points().qarray() = m_curve[0].points().qarray().copy();
  
  for(uint i=1; i < (m_curves - 1); i++) 
    sumCurve->points().addYValues(m_curve[i].points());
}

void MJ_FreqRespGui::setFilter(const MJ_Sample *b, uint bdegree,
			       const MJ_Sample *a, uint adegree)
{
  removeGraphic(m_curve->id(), false);
  formCurve(m_curve, b, bdegree, a, adegree);
  appendGraphic(m_curve);
  drawAll();
}

void MJ_FreqRespGui::test()
{
  MJ_FilterIIR2<float> iir;
  iir.designParametric1(0.1, MJ_pi2 , 0.3);

  MJ_FreqRespGui frg(0, 0, 20.0, 0.001, 0, 2);

  frg.setFilter(iir.bptr(), 2,
		iir.aptr(), 2);
  
  qApp->setMainWidget(&frg);
  frg.show();
  qApp->exec();
}

void MJ_FreqRespGui::formCurve(MJ_GraphicPolyLine *curve, 
			       const MJ_Sample *b, uint bdegree,
			       const MJ_Sample *a, uint adegree)
{
  m_b.copyData(b, bdegree + 1);
  m_a.copyData(a, adegree);

  MJ_FreqResp<MJ_Sample> fr;
  
  uint points = 500;

  fr.calculate(b, bdegree,
	       a, adegree,
	       points);

  MJ_PointArray2D *pa = & curve->points();
  pa->resize(points);
  for(uint i=0; i < points; i++) {
    float x = (i / (float) (points - 1)) * (1.0 - m_bbox[0]) + m_bbox[0];
    pa->setPoint(i, x, MJ_absToDb(fr.absN(i)));
  }
}


