#include <mj_freq_resp.h>
#include <mj_error.h>


#define MJ_PI 3.14159265358979

/** Calculates frequency response in one point. */

template <class TElem> 
void MJ_FreqResp<TElem>::pointResponse(const MJ_Sample *b, uint bdegree,
				       const MJ_Sample *a, uint adegree,
				       TElem freq,
				       TElem *real, TElem * imag)
{
  TElem numReal = 0.0, numImag = 0.0;
  TElem denomReal = 1.0, denomImag = 0.0;

  uint i;

  freq *= MJ_PI;

  for(i=0; i < bdegree; i++) {
    TElem tmp = freq * (TElem) i;
    numReal += cos(tmp) * b[i];
    numImag += sin(tmp) * b[i];
  }

  for(i=0; i < adegree; i++) {
    TElem tmp = freq * (TElem) (i + 1);
    denomReal += cos(tmp) * a[i];
    denomImag += sin(tmp) * a[i];
  }
  TElem numAmpl = sqrt(numReal * numReal + numImag * numImag);
  TElem denomAmpl = sqrt(denomReal * denomReal + denomImag * denomImag);

  TElem ampl = numAmpl / denomAmpl;
  TElem numPhase = MJ_atan2(numImag, numReal);
  TElem denomPhase = MJ_atan2(denomImag, denomReal);

  TElem phase = numPhase - denomPhase;

  *real = ampl * sin(phase);
  *imag = ampl * cos(phase);
}

/** Calculates the frequency response based on filter coefficients. */

template <class TElem> 
void MJ_FreqResp<TElem>::calculate(
				   /// The filter "b0-bn" coefficients
				   const MJ_Sample *b, 
				   /// The number of "b" coefficents
				   uint bdegree,
				   /// The filter "a1-an" coefficients
				   const MJ_Sample *a, 
				   /// The number of "a" coefficents
				   uint adegree,
				   /// Number of points to calculate
				   uint points, 
				   /// Minimum frequency - range [0, 1]
				   TElem minFreq,
				   /// Maximum frequency - range [0, 1]
				   TElem maxFreq,
				   /// Use logarithmic frequency scale
				   bool  logScale)
{
  const char *fname = "MJ_FreqResp::calculate";
  
  bdegree++;

  if(points < 2) {
    MJ_error(MJ_ERR_ARGUMENT, fname, "Need at least 2 points");
    return;
  }

  if(bdegree < 1 || adegree < 1) {
    MJ_error(MJ_ERR_ARGUMENT, fname, "Empty filter");
    return;
  }
  
  uint i;
  m_response.resize(points * 2); // Real and imag for each point.
  m_freqs.resize(points);
  
  TElem *data  = m_response.data();
  TElem *freqs = m_freqs.data();

  TElem range  = maxFreq - minFreq;
  TElem freq;

  TElem logMin = 0.0; // Stupid g++ warns here
  TElem logMax;
  TElem logStep = 0.0;
  
  if(logScale) {
    if(minFreq <= 0.0) MJ_error(MJ_ERR_ARGUMENT, fname, "too low log freq");

    logMax = log(maxFreq);
    logMin = log(minFreq);
    logStep = (logMax - logMin) / (TElem) (points - 1);
  }
  for(i=0; i < points; i++) {

    if(logScale)
      freq = MJ_exp((TElem) i * logStep + logMin);
    else
      freq = minFreq + range * ((TElem)i / (TElem) (points - 1));

    freqs[i] = freq;

    pointResponse(b, bdegree, a, adegree, freq, &data[i*2], &data[i*2+1]);
  }
}
