/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_connect_view.h>

#include <mj_math.h>

#include <mj_misc.h>

#include <qpainter.h> 

#define MJ_CONNECT_VIEW_DEBUG 0

MJ_ConnectView::MJ_ConnectView (QWidget      *parent,
				const char   *name)
  : QWidget(parent, name),
    m_startLeft(false),
    m_startIndex(-1),
    m_lastLeft(false),
    m_lastIndex(-1),
    m_connectionIndex(-1),
    m_leftBorder(0)
{
  setFocusPolicy(ClickFocus);

  setMouseTracking(true);
}

MJ_ConnectView::~MJ_ConnectView()
{}

void MJ_ConnectView::set
(const std::vector<MJ_LabelDef> &leftLabels,
 const std::vector<MJ_LabelDef> &rightLabels,
 const std::vector<MJ_2Int> &connections)
{
  m_sourceLabels = leftLabels;
  m_targetLabels = rightLabels;
  m_connections = connections; 
  
  // unhighlightConnection();

  redraw();
}

QSize MJ_ConnectView::minimumSizeHint() const
{ 
  return m_minimumSize;
}

void MJ_ConnectView::mousePressEvent(QMouseEvent *e)
{
#if MJ_CONNECT_VIEW_DEBUG
  const char * fname = "MJ_ConnectView::mousePressEvent";
#endif
  

  int connection;

  if((connection = onConnection(e->pos())) != -1) {
    
    unhighlightConnection();
    m_connectionIndex = connection;
    drawConnections();
    m_startIndex = -1;
  }
  else if(seekIndex(e->pos(), &m_startLeft, &m_startIndex)) {
    m_connectionIndex = -1;
#if MJ_CONNECT_VIEW_DEBUG
    qDebug("%s # on label %d", fname, m_startIndex);
#endif
    highlight(m_startLeft, true, m_startIndex);
  }
  else {
    m_startIndex = -1;
    unhighlightConnection();
    drawConnections();
  }
}

void MJ_ConnectView::mouseReleaseEvent(QMouseEvent *e)
{
  bool left;
  int index;
  
#if MJ_CONNECT_VIEW_DEBUG
  const char * fname = "MJ_ConnectView::mouseReleaseEvent";
  qDebug("%s %d %d %d", fname, e->pos().x(), e->pos().y(), m_startIndex);
#endif

  bool good = true;

  if(!seekIndex(e->pos(), &left, &index) || m_startIndex == -1)
    good = false;

  if(left == m_startLeft)
    good = false;

  int from, to;

  if(left) {
    from = index;
    to = m_startIndex;
  }
  else {
    from = m_startIndex;
    to = index;
  }
  
  if(isConnection(from, to))
    good = false;

  
  // addConnection(from, to);
  // drawConnections();
  if(good) {
#if MJ_CONNECT_VIEW_DEBUG
    qDebug("%s # new connection %d -> %d", fname, from, to);
#endif
    emit newConnection(from, to);
  }

  m_startIndex = -1;
  unhighlight();
}

void MJ_ConnectView::mouseMoveEvent(QMouseEvent *e)
{
  bool isleft;
  int index; 

#if MJ_CONNECT_VIEW_DEBUG
  qDebug("MJ_ConnectView::mouseMoveEvent");
#endif
  
  if(seekIndex(e->pos(), &isleft, &index)) {
    if(!(isleft == m_startLeft && m_startIndex != -1)) {

      if((m_lastIndex != -1) && 
	 !(m_lastIndex == m_startIndex && isleft == m_startLeft) &&
	 (isleft == m_lastLeft)) {
	highlight(m_lastLeft, false, m_lastIndex);
      }
      highlight(isleft, true, index);
    }
  }
  else if(m_startIndex == -1) 
    unhighlight();
  else if(m_lastIndex != -1 && m_lastIndex != m_startIndex) {
    highlight(m_lastLeft, false, m_lastIndex);
  }
}

void MJ_ConnectView::paintEvent(QPaintEvent *)
{
  redraw();
  // resize(m_minimumSize);
}

void MJ_ConnectView::keyPressEvent(QKeyEvent *e)
{
#if MJ_CONNECT_VIEW_DEBUG
  qDebug("MJ_ConnectView::keyPressEvent %d", (int) e->key());
#endif
  
  if(e->key() == Key_Backspace && m_connectionIndex != -1) {
    int tmp = m_connectionIndex;
    m_connectionIndex = -1;
    emit deleteConnection(tmp);
  }
}

void MJ_ConnectView::drawBase()
{
  QPainter painter(this);
  painter.eraseRect(0,0, width(), height());
}

void MJ_ConnectView::redraw()
{
  drawBase();
  alignLabels();
  drawLabels();
  drawConnections();
}

void MJ_ConnectView::addConnection(int from, int to)
{
  uint ind = m_connections.size();
  m_connections.resize(m_connections.size() + 1);
  MJ_2Int *tmp = & m_connections[ind];
  tmp->m_v[0] = from;
  tmp->m_v[1] = to;
}

void MJ_ConnectView::highlight
(bool leftSide, bool makeVisible, uint index)
{
  QPainter painter(this);
  QRect *rect;
  MJ_LabelDef *sl;

  if(leftSide) sl = & m_sourceLabels[index];
  else sl = & m_targetLabels[index];

  if(makeVisible) {
    m_lastLeft = leftSide;
    m_lastIndex = (int) index;
  }
  else if((uint) m_lastIndex == index && m_lastLeft == leftSide) {
    m_lastIndex = -1;
  }

  rect = & sl->m_rect;
  sl->m_highlighted = makeVisible;
  
  if(!makeVisible) painter.setPen(backgroundColor());
  painter.drawRect(*rect);
  if(!makeVisible) painter.setPen(foregroundColor());
}

void MJ_ConnectView::unhighlight()
{
#if MJ_CONNECT_VIEW_DEBUG
  qDebug("MJ_ConnectView::unhighlight");
#endif

  uint i, max = m_sourceLabels.size();

  for(i = 0 ; i < max; i++) {
    if(m_sourceLabels[i].m_highlighted)
      highlight(true, false, i);
  }

  max = m_targetLabels.size();

  for(i = 0 ; i < max; i++) {
    if(m_targetLabels[i].m_highlighted)
      highlight(false, false, i);
  }
}

bool MJ_ConnectView::seekIndex
(const QPoint &pos, bool *isleft, int *index)
{
  if(pos.x() < m_leftBorder) {
    uint smax = m_sourceLabels.size();

    for(uint i = 0; i < smax; i++) {
      if(m_sourceLabels[i].m_rect.contains(pos)) {
	*isleft = true;
	*index = (int) i;
	return true;
      }
    }
  }
  else {    
    uint smax = m_targetLabels.size();

    for(uint i = 0; i < smax; i++) {
      if(m_targetLabels[i].m_rect.contains(pos)) {
	*isleft = false;
	*index = i;
	return true;
      }
    }
  }
  return false;
}

void MJ_ConnectView::drawConnections()
{
  uint size = m_connections.size();

  if(!size || !m_sourceLabels.size() || !m_targetLabels.size()) return;

  m_points.resize(size * 2);

  int leftLim = m_leftBorder;
  int rightLim = m_targetLabels[0].m_rect.left() - 10;

  for(uint i = 0; i < size; i++) {

    MJ_2Int *map = & m_connections[i];

    if((uint) map->m_v[0] >= m_sourceLabels.size() || 
       (uint) map->m_v[1] >= m_targetLabels.size()) {
      qDebug("MJ_ConnectView::drawConnections # Bad connections %d %d",
	     map->m_v[0], map->m_v[1]);
      break;
    }
    
    const QRect *sr = & m_sourceLabels[map->m_v[0]].m_rect;
    const QRect *tr = & m_targetLabels[map->m_v[1]].m_rect;

    int sh = (sr->top() + sr->bottom()) / 2;
    
    m_points.setPoint(i * 2, leftLim, sh);
    m_points.setPoint(i * 2 + 1, rightLim, (tr->top() + tr->bottom()) / 2);
  }
  QPainter painter(this);
  painter.drawLineSegments(m_points);
  
  highlightConnection();
}

void MJ_ConnectView::alignLabels()
{
  int width = 0;
  uint i;

  QPainter painter(this);
  uint fontH = (uint) painter.fontInfo().pointSize();

  // If the size is notreported correctly:
  if (fontH == 0)
    fontH = 20;

  uint distance = fontH * 2;
  uint margin = distance;
  uint vertCount = distance;
  
  int minHeight = 0;

#if MJ_CONNECT_VIEW_DEBUG
  qDebug("MJ_ConnectView::alignLabels w = %d h = %d"
	 , (int) this->width(), this->height());
#endif

  for(i = 0; i < m_sourceLabels.size(); i++) {
    MJ_LabelDef * sl = & m_sourceLabels[i];

    QRect r = painter.boundingRect(0, 0, 1000, distance * 3, 
				     AlignLeft, sl->m_string);

    sl->m_rect.setRect(margin, vertCount, r.width(), r.height());

    vertCount += distance;
    
    if(r.width() > width) width = r.width();
    if((int) vertCount > minHeight) minHeight = vertCount;
  }
  
  // Resize the rects of the left labels:
  for(i = 0; i < m_sourceLabels.size(); i++) {
    MJ_LabelDef * sl = & m_sourceLabels[i];
    sl->m_rect.setWidth(width);
  }

  vertCount = distance;

  m_leftBorder = width + (int) distance;
  
  int rightBorder = m_leftBorder + 100 + (int) distance;
  int minWinWidth = rightBorder;

  for(i = 0; i < m_targetLabels.size(); i++) {
    
    MJ_LabelDef * sl = & m_targetLabels[i];

    QRect r = painter.boundingRect(0, 0, 1000, distance * 3, 
				     AlignLeft, sl->m_string);

    sl->m_rect.setRect(rightBorder, vertCount, r.width(), distance);

    vertCount += distance;
    int tmp = rightBorder + r.width();

    if(tmp > minWinWidth) minWinWidth = tmp;
    if((int) vertCount > minHeight) minHeight = vertCount;
    
#if MJ_CONNECT_VIEW_DEBUG
    qDebug("MJ_ConnectView::alignLabels right %d y = %d x = %d w = %d h = %d this->h = %d", 
	   (int) i, sl->m_rect.y(), sl->m_rect.x(), 
	   sl->m_rect.width(), sl->m_rect.height(), this->height());
#endif

  }

  m_minimumSize.setWidth(minWinWidth + margin);
  m_minimumSize.setHeight(minHeight + margin);
}

void MJ_ConnectView::drawLabels()
{
  uint i;
  
  QPainter painter(this);

  for(i=0; i < m_sourceLabels.size(); i++) {
    MJ_LabelDef * sl = & m_sourceLabels[i];
    sl->m_rect = painter.boundingRect(sl->m_rect, AlignLeft, sl->m_string);
    painter.drawText(sl->m_rect, AlignLeft, sl->m_string);

#if MJ_CONNECT_VIEW_DEBUG
    qDebug("MJ_ConnectView::drawLabels y = %d h = %d", 
	   sl->m_rect.y(), this->height());
#endif
  }

  for(i = 0; i < m_targetLabels.size(); i++) {    
    MJ_LabelDef * sl = & m_targetLabels[i];

#if MJ_CONNECT_VIEW_DEBUG
    qDebug("MJ_ConnectView::drawLabels right %d y = %d x = %d w = %d h = %d this->h = %d", 
	   (int) i, sl->m_rect.y(), sl->m_rect.x(), 
	   sl->m_rect.width(), sl->m_rect.height(), this->height());
#endif

    sl->m_rect = painter.boundingRect(sl->m_rect, AlignLeft, sl->m_string);
    painter.drawText(sl->m_rect, AlignLeft, sl->m_string);

#if MJ_CONNECT_VIEW_DEBUG
    qDebug("MJ_ConnectView::drawLabels right %d y = %d x = %d w = %d h = %d this->h = %d", 
	   (int) i, sl->m_rect.y(), sl->m_rect.x(), 
	   sl->m_rect.width(), sl->m_rect.height(), this->height());
#endif
  }
  
}

bool MJ_ConnectView::isConnection(int from, int to)
{
  uint max = m_connections.size();
  MJ_2Int *array = & * m_connections.begin();

  for(uint i = 0; i < max; i++) {
    MJ_2Int *tmp = & array[i];
    if((tmp->m_v[0] == from) && (tmp->m_v[1] == to))
      return true;
  }
  return false;
}

void MJ_ConnectView::unhighlightConnection()
{
  if(m_connectionIndex == -1) return;

  QPainter painter(this);
  
  QPen penOrig(painter.pen());
  
  QPen penNew(backgroundColor(), 3);

  painter.setPen(penNew);
  painter.drawLineSegments(m_points, m_connectionIndex * 2, 1);
  
  painter.setPen(penOrig);
  
  
  m_connectionIndex = -1;
}

void MJ_ConnectView::highlightConnection()
{
  if(m_connectionIndex == -1) return;

  QPainter painter(this);
  
  QPen penOrig(painter.pen());

  QPen penNew(foregroundColor(), 3);
  painter.setPen(penNew);
  painter.drawLineSegments(m_points, m_connectionIndex * 2, 1);
  painter.setPen(penOrig);
  
}

int MJ_ConnectView::onConnection(const QPoint &pos, int range)
{
  uint max = m_points.size() / 2;

  if(!max) return -1;

  int x, y1, y2;

  const int left = m_points.point(0).x();
  const int right = m_points.point(1).x();
  
  if(pos.x() > right || pos.x() < left) return -1;
  
  const float ratio = (float) (pos.x() - left) / (float) (right - left);

  const int ypos = pos.y();
  
  for(uint i = 0; i < max; i++) {
    uint tmp = i * 2;
    m_points.point(tmp, &x, &y1);
    m_points.point(tmp+1, &x, &y2);

    int ycmp = y1 + (int) ((float) (y2 - y1) * ratio);
    
    int diff = MJ_abs(ycmp - ypos);

    if(diff <= range) 
      return (int) i;
  }
  return -1;
}
