/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_dial.h>

#include <mj_mapper.h>
#include <mj_math.h>

MJ_Dial::MJ_Dial(double minValue, double maxValue, double value, 
		 QWidget *parent, const char * name,
		 int iminValue,
		 int imaxValue,
		 int lineStep,
		 int pageStep)
  : QDial(iminValue, imaxValue, pageStep, 0, parent, name),
    m_minValue(minValue),
    m_maxValue(maxValue),
    m_allowWrap(false),
    m_mapper(0)
{
  setValueDouble(value);
  setSteps(lineStep, pageStep);
}

MJ_Dial::~MJ_Dial()
{
  delete m_mapper;
}

void MJ_Dial::setRealRange(double minV, double maxV)
{
  m_minValue = minV;
  m_maxValue = maxV;

  setValueDouble(m_value);
}

void MJ_Dial::setSliderRange(int minR, int maxR)
{
  setRange(minR, maxR);
  
  setValueDouble(m_value);
}

void MJ_Dial::setMapper(MJ_Mapper  *m)
{
  delete m_mapper;
  m_mapper = m;
  setValueDouble(m_value);
}

void MJ_Dial::setValueDouble(double val)
{
  double localMin(m_minValue);
  double localMax(m_maxValue);
  
  if(val < localMin) val = localMin;
  if(val > localMax) val = localMax;

  m_value = val;

  if(m_mapper) {
    localMin = m_mapper->map(localMin);
    localMax = m_mapper->map(localMax);
    val = m_mapper->map(val);
  }

  double diff = localMax - localMin;
  
  int ivalue;

  if(MJ_abs(diff) < 1e-10)
    ivalue = (int) ((m_maxValue + m_minValue) * 0.5);
  else 
    ivalue = (int) ((val - m_minValue) / (diff) * 
		    (double) (maxValue() - minValue()) + minValue());

  m_oldVal = ivalue;
  
  /* qDebug("MJ_Dial::setValueDouble 0x%p %d %f  [%f %f]", this, value(), val,
     m_minValue, m_maxValue); */

  bool block = signalsBlocked();
  bool allowWrap = m_allowWrap;
  m_allowWrap = true;
  blockSignals(true);
  setValue(ivalue);
  blockSignals(block);
  m_allowWrap = allowWrap;

  m_value = val;
}

void MJ_Dial::valueChange()
{
  QDial::valueChange();

  int ival = value();
  int imin = minValue();
  int imax = maxValue();

  int ithird = (imax - imin) / 3;
  int change = ival - m_oldVal;

  if(change < 0) change = -change;

  if(!m_allowWrap && (change > ithird)) {
    setValue(m_oldVal);
    return;
  }

  m_oldVal = ival;

  double val = (ival - imin) / (double) (imax - imin);
  
  if(m_mapper) {
    double localMin = m_mapper->map(m_minValue);
    double localMax = m_mapper->map(m_maxValue);

    val = m_mapper->unmap(val * (localMax - localMin) + localMin);
    m_value = val;
  }
  else {
    m_value = val * (m_maxValue - m_minValue) + m_minValue;
  }

  emit valueChangedDouble(m_value);

  /* qDebug("MJ_Dial::valueChange 0x%p %d %f  [%f %f]", this, value(), m_value,
     m_minValue, m_maxValue); */
}
  
