/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_graphics_area.h>

#include <mj_graphics_context.h>
#include <mj_painter.h>

#include <qpixmap.h> 

MJ_GraphicsArea::MJ_GraphicsArea(QWidget *parent,
				 const char * name, 
				 bool autoDeleteChildren)
  : QWidget(parent, name),
    m_squarePixels(false)
{
  m_graphics.setAutoDelete(autoDeleteChildren);
  setMinimumSize(20, 20);
  setBoundingBox(0.0, 1.0, 1.0, 0.0);
}

/** Free all resources. Deletes the graphics objects if autodeletion
    is enabled. */

MJ_GraphicsArea::~MJ_GraphicsArea()
{}

/** Appends a new graphic element at the end of the list. 

    The graphic element is not drawn.
*/
int MJ_GraphicsArea::appendGraphic(MJ_Graphic *gr)
{
  if(!gr) {
    qDebug("MJ_GraphicsArea::appendGraphic");
    return -1;
  }
  int id = allocateId();
  gr->setId(id);
  gr->setWidget(this);
  m_graphics.append(gr);
  return id;
}

/** Remove the graphic element matched by the id.

    The graphics object is unrendered before being removed. 

    Often it is necessary to call MJ_GraphicsArea::drawAll after this
    method is called to rebuid the view.  */
bool MJ_GraphicsArea::removeGraphic(int id, bool del)
{
  if(!m_graphics.count()) return false;
  
  MJ_Graphic *gr = m_graphics.first();

  while(gr) {
    if(gr->id() == id) {
      m_graphics.take(m_graphics.at());

      MJ_Painter paint(this);
      makeTransform(&paint);
      
      MJ_GraphicsContext *gc = gr->graphicsContext();
      if(gc) gc->apply(& paint.qpainter());
      
      // Override the color.
      QPen pen = paint.qpainter().pen();
      pen.setColor(backgroundColor());
      paint.qpainter().setPen(pen);
  
      gr->unrender(&paint);

      if(del) delete gr;
      
      paintAll(&paint);

      return true;
    }
    gr = m_graphics.next();
  }

  qDebug("MJ_GraphicsArea::removeGraphic id %d not found", id);

  return false;
}


bool MJ_GraphicsArea::removeGraphicIndex(int index, bool del)
{
  MJ_Graphic *g = m_graphics.at(index);

  if(g) return removeGraphic(g->id(), del);
  else return false;
}

/** Sets the bounding box of the visible area. */
void MJ_GraphicsArea::setBoundingBox
(float leftB, float topB, float rightB, float bottomB)
{
  m_bbox[0] = leftB;
  m_bbox[1] = topB;
  m_bbox[2] = rightB;
  m_bbox[3] = bottomB;
}

void MJ_GraphicsArea::redraw(bool doubleBuffer)
{
  if(!m_graphics.count()) return;
  
  if(doubleBuffer) {
    QPixmap pxm(size());
    
    MJ_Painter paint;
    makeTransform(&paint);
    
    paint.qpainter().begin(&pxm, this);

    paint.qpainter().eraseRect(0, 0, width(), height());
    paintAll(&paint);

    paint.qpainter().end();
    
    /* QPainter p2(this);
       p2.drawPixmap(0,0, pxm); */
    bitBlt(this, 0, 0, &pxm);
    
  }
  else {
    MJ_Painter paint(this);
    makeTransform(&paint);
  
    paint.qpainter().eraseRect(0, 0, width(), height());
    paintAll(&paint);
  }
}

/** Draws all elements without clearing the area first. */
void MJ_GraphicsArea::drawAll()
{
  MJ_Painter paint(this);
  makeTransform(&paint);
  paintAll(&paint);
}

/** Draws one graphic element to the widget. */
void MJ_GraphicsArea::drawGraphic(MJ_Graphic *gr)
{
  MJ_Painter paint(this);
  makeTransform(&paint);
  
  MJ_GraphicsContext *gc = gr->graphicsContext();
  if(gc) gc->apply(& paint.qpainter());
      
  gr->render(&paint);
}


/// Unrenders one graphic element from widget. 
void MJ_GraphicsArea::undrawGraphic(MJ_Graphic *gr)
{  
  MJ_Painter paint(this);
  makeTransform(&paint);
  
  MJ_GraphicsContext *gc = gr->graphicsContext();
  if(gc) gc->apply(& paint.qpainter());
      
  // Override the color.
  QPen pen = paint.qpainter().pen();
  pen.setColor(backgroundColor());
  paint.qpainter().setPen(pen);
  
  gr->unrender(&paint);
}

void MJ_GraphicsArea::mousePressEvent(QMouseEvent * e)
{
  float x, y;
  unmapPoint(e->x(), e->y(), &x, &y);
  emit mousePress(e, x, y);
}

void MJ_GraphicsArea::mouseReleaseEvent(QMouseEvent * e)
{
  float x, y;
  unmapPoint(e->x(), e->y(), &x, &y);
  emit mouseRelease(e, x, y);
}

void MJ_GraphicsArea::mouseDoubleClickEvent(QMouseEvent * e)
{
  float x, y;
  unmapPoint(e->x(), e->y(), &x, &y);
  emit mouseDoubleClick(e, x, y);
}

void MJ_GraphicsArea::mouseMoveEvent(QMouseEvent * e)
{
  float x, y;
  unmapPoint(e->x(), e->y(), &x, &y);
  emit mouseMove(e, x, y);
}

/** Clears the area and repaints it. */
void MJ_GraphicsArea::paintEvent(QPaintEvent * ) 
{
  redraw();
}

/** Allocates new id for an object. Returns the new id. */
int MJ_GraphicsArea::allocateId()
{
  for(int i=0; true; i++) {
    MJ_Graphic *gr = m_graphics.first();

    bool found = false;

    while(gr) {
      if(gr->id() == i) {
	found = true;
	gr = 0;
      }
      else
	gr = m_graphics.next();
    }    
    if(!found) return i;
  }
}

/** Paints all ojects to the area. */
void MJ_GraphicsArea::paintAll(MJ_Painter *paint)
{
  // qDebug("MJ_GraphicsArea::paintAll");
  MJ_Graphic *gr = m_graphics.first();
  
  QPen defaultPen(Qt::black);

  while(gr) {
    paint->qpainter().setPen(foregroundColor());
    
    MJ_GraphicsContext *gc = gr->graphicsContext();
    if(gc) gc->apply(& paint->qpainter());
    else paint->qpainter().setPen(defaultPen);

    gr->render(paint);
    gr = m_graphics.next();
  }
}

/** Creates a proper transformation matrix for the painter. */
void MJ_GraphicsArea::makeTransform(MJ_Painter *paint)
{
  float xscale, yscale, xtrans, ytrans;

  int w = width();
  int h = height();

  if(!m_squarePixels) {
    xscale = (w-1) / (m_bbox[2] - m_bbox[0]);
    yscale = (h-1) / (m_bbox[3] - m_bbox[1]);
    
    xtrans = -xscale * m_bbox[0];
    ytrans = -yscale * m_bbox[1];
  }
  else {
    if(w > h) {
      xscale = (h-1) / (m_bbox[2] - m_bbox[0]);
      yscale = (h-1) / (m_bbox[3] - m_bbox[1]);
      
      xtrans = -xscale * m_bbox[0];
      ytrans = -yscale * m_bbox[1];
    }
    else {
      xscale = (w-1) / (m_bbox[2] - m_bbox[0]);
      yscale = (w-1) / (m_bbox[3] - m_bbox[1]);
      
      xtrans = -xscale * m_bbox[0];
      ytrans = -yscale * m_bbox[1];
    }
  }

  paint->scaleTranslate2(xscale, yscale, xtrans, ytrans);
}

/** Map a point from pixel coordinates to world coordinates. */

void MJ_GraphicsArea::unmapPoint(int x, int y, float *xf, float *yf)
{
  float xrel = (float) x / (float) width();
  float yrel = (float) y / (float) height();

  *xf = xrel * (m_bbox[2] - m_bbox[0]) + m_bbox[0];
  *yf = yrel * (m_bbox[3] - m_bbox[1]) + m_bbox[1];
}
