/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_gui_int.h>
#include <mj_gui_utils.h>
#include <mj_line_edit.h>
#include <mj_error.h>

#include <qcombobox.h> 
#include <qlayout.h> 
#include <qlabel.h> 
#include <qscrollview.h> 
#include <qslider.h>

MJ_GuiInt::MJ_GuiInt()
{}

MJ_GuiInt::~MJ_GuiInt()
{}

void MJ_GuiInt::setRange(int, int)
{}

/** Emits a signal to that tells that the current value is final. */
void MJ_GuiInt::done()
{
  if(!m_id)
    emit emitDone();
  else
    emit emitDone2(m_id);
}


/** Used by child classes to emit new value. Returns the updated
    value.*/
int MJ_GuiInt::emitInternal(int d)
{
  int tmp = 0;
  if(!m_id) {
    emit emitValue(d);
    emit emitGetValue(&tmp);
  }
  else {
    emit emitValue2(m_id, d);
    emit emitGetValue2(m_id, &tmp);
  }
  return tmp;
}

/** Emits a signal to get the current parameter value. */
void MJ_GuiInt::emitGet(int *d)
{
  if(!m_id) 
    emit emitGetValue(d);
  else
    emit emitGetValue2(m_id, d);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_GuiIntCombo::MJ_GuiIntCombo()
  : m_label(0),
    m_combo(0)
{
  m_vals.setAutoDelete(true);
}

MJ_GuiIntCombo::MJ_GuiIntCombo(QObject *parent) 
  : MJ_GuiInt(parent),
    m_label(0),
    m_combo(0)
{
  m_vals.setAutoDelete(true);
}

MJ_GuiIntCombo::~MJ_GuiIntCombo()
{}

void MJ_GuiIntCombo::setValue(int d)
{
  checkReady();

  int index = 0;

  bool b = m_combo->signalsBlocked();
  m_combo->blockSignals(true);
  
  for(uint i=0; i < m_vals.count(); i++) {
    if(*m_vals.at(i) == d) {
      index = i;
      i = m_vals.count();
    }
  }
  m_combo->setCurrentItem(index);
  m_combo->blockSignals(b);
}


void MJ_GuiIntCombo::setEnabled(bool ena)
{
  if(m_combo) m_combo->setEnabled(ena);
}

/** Creates objects relative to a parent. Also connects the widgets'
    signals to this object. Returns false if operations fail (may
    happe when out of memory etc.). */
bool MJ_GuiIntCombo::createObjs(QWidget *parent)
{
  m_label = new QLabel(parent);
  m_combo = new QComboBox(parent);

  if(!m_label || !m_combo) return false;

  for(uint i=0; i < m_vals.count(); i++) {
    QString s;
    s.setNum(*m_vals.at(i));
    m_combo->insertItem(s);
  }

  connect(m_combo, SIGNAL(activated(int)),
	  this, SLOT(newInd(int)));

  return true;
}


void MJ_GuiIntCombo::addChildren(QScrollView *s)
{
  if(m_label) s->addChild(m_label);
  if(m_combo) s->addChild(m_combo);
}

/** Limits the minimum sizes of the widgets to the minimum size
    implied by their "sizeHint". */
void MJ_GuiIntCombo::limMinSizes()
{
  checkReady();
  MJ_limMinSize(m_combo);
  MJ_limMinSize(m_label);
}

/** Deletes objects. */
void MJ_GuiIntCombo::deleteObjs()
{
  if(m_label) delete m_label; m_label = 0;
  if(m_combo) delete m_combo; m_combo = 0;
}

/** Puts the objects to a horizontal grid. First label, then the
    combo. If you need more detailed control over widgets' location,
    you must do it by hand.*/
void MJ_GuiIntCombo::toGridHorizontal(QGridLayout *g, int row, int firstCol)
{
  checkReady();
  g->addWidget(m_label, row, firstCol);
  g->addWidget(m_combo, row, firstCol + 1, AlignLeft);
}

/** Receives signals from the combo. */
void MJ_GuiIntCombo::newInd(int i)
{
  emitInternal(* m_vals.at(i));
  done();
}


void MJ_GuiIntCombo::show()
{
  if(m_label) m_label->show();
  if(m_combo) m_combo->show();
}

/** Check that the necessary child objects do exist. If the objects do
    not exist, fatal error is generated.*/

void MJ_GuiIntCombo::checkReady()
{
  if(!m_label || !m_combo) 
    qDebug("MJ_GuiIntCombo::checkReady NULL pointer(s)");
}


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_GuiIntSlider::MJ_GuiIntSlider()
  : m_label(0), m_slider(0), m_edit(0)
{}

MJ_GuiIntSlider::MJ_GuiIntSlider(QObject * parent)
  : MJ_GuiInt(parent), m_label(0), m_slider(0), m_edit(0)
{}

MJ_GuiIntSlider::~MJ_GuiIntSlider()
{}

MJ_GuiIntSlider &MJ_GuiIntSlider::operator = (const MJ_GuiIntSlider &)
{
  qDebug("MJ_GuiIntSlider::operator = DO NOT CALL");
  return *this;
}

void MJ_GuiIntSlider::setValue(int d)
{
  // qDebug("MJ_GuiIntSlider::setValue %f", d);
  
  m_slider->setValue(d);
  if(m_edit) m_edit->setValueInt(d);
}

void MJ_GuiIntSlider::setEnabled(bool ena)
{
  if(m_slider) m_slider->setEnabled(ena);
  if(m_edit) m_edit->setEnabled(ena);
}

/** Creates and connects child widgets. */
bool MJ_GuiIntSlider::createObjs(QWidget *parent, bool withLineEdit)
{
  m_label  = new QLabel(parent);
  m_slider = new QSlider(QSlider::Horizontal, parent);
  if(withLineEdit) m_edit   = new MJ_LineEdit(parent);

  if(!m_label || !m_slider || (!m_edit && withLineEdit)) return false;
  
  if(m_edit) m_edit->useDouble(false);

  connectObjs();
  return true;
}


void MJ_GuiIntSlider::addChildren(QScrollView *s)
{
  if(m_label) s->addChild(m_label);
  if(m_slider) s->addChild(m_slider);
  if(m_edit) s->addChild(m_edit);
}


/** Connects child widgets. */
void MJ_GuiIntSlider::connectObjs()
{
  checkReady();

  connect(m_slider, SIGNAL(valueChanged(int)),
	  this, SLOT(transmitValue(int)));


  connect(m_slider, SIGNAL(sliderReleased()),
	  this, SLOT(done()));

  if(m_edit) {
  connect(m_slider, SIGNAL(valueChanged(int)), 
	  m_edit, SLOT(setValueInt(int)));

  connect(m_edit, SIGNAL(emitValueInt(int)), 
	  m_slider, SLOT(setValue(int)));
  
  connect(m_edit, SIGNAL(emitValueInt(int)), 
	  this, SLOT(transmitValue(int)));
  
  connect(m_edit, SIGNAL(emitGetValueInt(int *)), 
	  this, SLOT(emitGet(int *)));
  
  connect(m_edit, SIGNAL(emitDone()),
	  this, SLOT(done()));
  }
}


/** Limits minimum size of the child objects to that implied by
    "sizeHint".*/
void MJ_GuiIntSlider::limMinSizes()
{
  checkReady();
  MJ_limMinSize(m_label);
  MJ_limMinSize(m_slider);
  if(m_edit) MJ_limMinSize(m_edit);
}

/** Deletes child objects. */
void MJ_GuiIntSlider::deleteObjs()
{
  if(m_label)  delete m_label;  m_label = 0;
  if(m_slider) delete m_slider; m_slider = 0;
  if(m_edit)   delete m_edit;   m_edit = 0;
}


/** Forgets child objects. The objects are not deleted. */
void MJ_GuiIntSlider::forgetObjs()
{
  m_label = 0;
  m_slider = 0;
  m_edit = 0;
}

/** Sets the range of the slider.  */
void MJ_GuiIntSlider::setRange(int minimum, int maximum)
{
  checkReady();
  m_slider->setRange(minimum, maximum);
}

/** Puts objects to grid. First label, then slider and then the text
    box.*/
void MJ_GuiIntSlider::toGridHorizontal
(QGridLayout *g, int row, int firstCol)
{
  checkReady();
  g->addWidget(m_label,  row, firstCol);

  if(m_edit) {
    g->addWidget(m_edit,   row, firstCol+2);
    g->addWidget(m_slider, row, firstCol+1);
  }
  else
    g->addMultiCellWidget(m_slider, row, row, firstCol+1, firstCol+2);
}

/** Shown all child objects. */
void MJ_GuiIntSlider::show()
{
  if(m_label) m_label->show();
  m_slider->show();
  if(m_edit) m_edit->show();
}

void MJ_GuiIntSlider::transmitValue(int d)
{
  emitInternal(d);
}

/** Checks all child objects exist. If they do not exist, fatal error
    is generated. */
void MJ_GuiIntSlider::checkReady()
{
  if(!m_label || !m_slider) 
    qDebug("MJ_GuiIntSlider::checkReady NULL pointer(s)");
}
