/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_GUI_INT_H
#define MJ_GUI_INT_H

#include <mj_gui_numeric.h>

class MJ_LineEdit;

class QComboBox;
class QGridLayout;
class QLabel;
class QSlider;
class QWidget;



/** Base class for multi-widget int valued GUI management. */
class MJ_GuiInt : public QObject, public MJ_HasId
{
  Q_OBJECT
public:
  MJ_GuiInt();
  MJ_GuiInt(QObject *parent) : QObject(parent) {}
  virtual ~MJ_GuiInt();

  virtual void setRange(int, int);

public slots:

  /** Sets the value displayed by the widgets. */
  virtual void setValue(int) = 0;

  virtual void done();

  virtual void setEnabled(bool) = 0;

  void emitGet(int *);

signals:

  /** Emitted when the value has changed. Used when no id has been set. */
  void emitValue(int);
  /** Emitted when the value has changed. Used when the first (but not
      the second) id has been set. */
  void emitValue2(const char *,int);

  /** Emitted to qeury value of the parameter. The object receiveing
      the values should be able to tell current value. It may well be
      that the value is not the value emitted byt this object (it may
      axceed some range or be badly formatted. ) and therefore the
      client needs to supply it. */
  void emitGetValue(int *);
  /** Emitted as one id has been set. */
  void emitGetValue2(const char *,int *);

  /** Signals that the change is now "final". If (for example) a
      slider sends many values while it is being dragged around the
      client may wish to know when it has reached the final
      value. This signal is emitted after the final value has been
      sent. */
  void emitDone();
  /** Emitted if one id has been set. */
  void emitDone2(const char *);

protected:

  int emitInternal(int);

};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


/** A Int GUI with label and combo-box for value selection. */
class MJ_GuiIntCombo : public MJ_GuiInt
{
  Q_OBJECT
public:
  MJ_GuiIntCombo();
  MJ_GuiIntCombo(QObject *parent);
  ~MJ_GuiIntCombo();

  void setValue(int);
  virtual void setEnabled(bool);

  bool createObjs(QWidget *parent);
  
  /** Adds the member widgets to scrollview with the respective
      "addChildren" method.*/

  void addChildren(QScrollView *);

  void limMinSizes();
  void deleteObjs();

  void toGridHorizontal(QGridLayout *, int row, int firstCol);

  QLabel    *label() { return m_label; }
  QComboBox *combo() { return m_combo; }

  void addValue(int d) { m_vals.append(new int(d)); }

  /** Show member widgets. */

  void show();

public slots:

  void newInd(int);

protected:
  
  void checkReady();

  QList<int> m_vals;

  QLabel    *m_label;
  QComboBox *m_combo;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/// Int GUI which packs a label, a slider and text-edit together.
class MJ_GuiIntSlider : public MJ_GuiInt
{
  Q_OBJECT
public:
  MJ_GuiIntSlider();
  MJ_GuiIntSlider(QObject * parent);
  ~MJ_GuiIntSlider();

  void setValue(int);
  virtual void setEnabled(bool);

  bool createObjs(QWidget *parent, bool withLineEdit = true);
  /** Adds the member widgets to scrollview with the respective
      "addChildren" method.*/

  void addChildren(QScrollView *);

  void connectObjs();
  void limMinSizes();
  void deleteObjs();
  void forgetObjs();

  void setRange(int minimum, int maximum);

  /** Set values to the child widgets. */
  void set(int val, int min, int max)
  { setRange(min, max); setValue(val); }

  void toGridHorizontal(QGridLayout *, int row, int firstCol);

  /// Returns the label widget.
  QLabel      *label()  { return m_label; }
  /// Returns the slider widget.
  QSlider     *slider() { return m_slider; }
  /// Returns the text-edit widget.
  MJ_LineEdit *edit()   { return m_edit; }

  void show();

  /// DO NOT CALL
  MJ_GuiIntSlider &operator = (const MJ_GuiIntSlider &);

public slots:

  void transmitValue(int);

protected:
  void checkReady();


  QLabel      *m_label;
  QSlider     *m_slider;
  MJ_LineEdit *m_edit;
};

#endif
