/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_GUI_NUMERIC_H
#define MJ_GUI_NUMERIC_H

#include <mj_has_id.h>

#include <qobject.h>
#include <qptrlist.h>

class MJ_Dial;
class MJ_Slider;
class MJ_LineEdit;

class QComboBox;
class QGridLayout;
class QLabel;
class QScrollView;
class QWidget;


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** Base class for multi-widget double valued GUI management. */
class MJ_GuiDouble : public QObject, public MJ_HasId
{
  Q_OBJECT
public:
  MJ_GuiDouble();
  MJ_GuiDouble(QObject *parent) : QObject(parent) {}
  virtual ~MJ_GuiDouble();

  virtual void setRange(double, double);
  virtual bool createObjs(QWidget *parent) = 0;
  virtual void limMinSizes() = 0;
  virtual void deleteObjs() = 0;
  virtual void toGridHorizontal(QGridLayout *, int row, int firstCol) = 0;

  virtual QLabel * label() = 0;
  virtual void show() = 0;

public slots:

  /** Sets the value displayed by the widgets. */
  virtual void setValue(double) = 0;

  virtual void done();

  void emitGet(double *);
  virtual void setEnabled(bool) = 0;

signals:

  /** Emitted when the value has changed. Used when no id has been set. */
  void emitValue(double);
  /** Emitted when the value has changed. Used when the first (but not
      the second) id has been set. */
  void emitValue2(const char *,double);

  /** Emitted to qeury value of the parameter. The object receiveing
      the values should be able to tell current value. It may well be
      that the value is not the value emitted byt this object (it may
      axceed some range or be badly formatted. ) and therefore the
      client needs to supply it. */
  void emitGetValue(double *);
  /** Emitted as one id has been set. */
  void emitGetValue2(const char *,double *);

  /** Signals that the change is now "final". If (for example) a
      slider sends many values while it is being dragged around the
      client may wish to know when it has reached the final
      value. This signal is emitted after the final value has been
      sent. */
  void emitDone();
  /** Emitted if one id has been set. */
  void emitDone2(const char *);

protected:

  double emitInternal(double);

};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


/** A Double GUI with label and combo-box for value selection. */
class MJ_GuiDoubleCombo : public MJ_GuiDouble
{
  Q_OBJECT
public:
  MJ_GuiDoubleCombo();
  MJ_GuiDoubleCombo(QObject *parent);
  virtual ~MJ_GuiDoubleCombo();

  void setValue(double);
  virtual void setEnabled(bool);

  bool createObjs(QWidget *parent);
  
  /** Adds the member widgets to scrollview with the respective
      "addChildren" method.*/

  void addChildren(QScrollView *);

  virtual void limMinSizes();
  virtual void deleteObjs();

  void toGridHorizontal(QGridLayout *, int row, int firstCol);

  QLabel    *label() { return m_label; }
  QComboBox *combo() { return m_combo; }

  void addValue(double d) { m_vals.append(new double(d)); }

  /** Show member widgets. */

  void show();

public slots:

  void newInd(int);

protected:
  
  void checkReady();

  QList<double> m_vals;

  QLabel    *m_label;
  QComboBox *m_combo;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/// Double GUI which packs a label, a slider and text-edit together.
class MJ_GuiDoubleSlider : public MJ_GuiDouble
{
  Q_OBJECT
public:
  MJ_GuiDoubleSlider();
  MJ_GuiDoubleSlider(QObject *parent);
  virtual ~MJ_GuiDoubleSlider();

  void setValue(double);
  virtual void setEnabled(bool);
  
  virtual bool createObjs(QWidget *parent);
  bool createObjs(QWidget *parent, bool withLineEdit);

  /** Adds the member widgets to scrollview with the respective
      "addChildren" method.*/

  void addChildren(QScrollView *);

  void connectObjs();
  virtual void limMinSizes();
  virtual void deleteObjs();

  void setRange(double minimum, double maximum);

  /** Set values to the child widgets. */
  void set(double val, double min, double max)
  { setRange(min, max); setValue(val); }

  void toGridHorizontal(QGridLayout *, int row, int firstCol);

  /// Returns the label widget.
  QLabel      *label()  { return m_label; }
  /// Returns the slider widget.
  MJ_Slider   *slider() { return m_slider; }
  /// Returns the text-edit widget.
  MJ_LineEdit *edit()   { return m_edit; }

  void show();

  /// DO NOT CALL
  MJ_GuiDoubleSlider &operator = (const MJ_GuiDoubleSlider &);

public slots:

  void transmitValue(double);

protected:
  void checkReady();


  QLabel      *m_label;
  MJ_Slider   *m_slider;
  MJ_LineEdit *m_edit;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/// Double GUI which packs a label, a slider and text-edit together.
class MJ_GuiDoubleDial : public MJ_GuiDouble
{
  Q_OBJECT
public:
  MJ_GuiDoubleDial();
  virtual ~MJ_GuiDoubleDial();

  void setValue(double);
  virtual void setEnabled(bool);
  
  virtual void limMinSizes();
  virtual void deleteObjs();

  virtual void setRange(double, double);
  virtual bool createObjs(QWidget *parent);
  bool createObjs(QWidget *parent, bool withDisplay);

  virtual void toGridHorizontal(QGridLayout *, int row, int firstCol);

  QLabel      *label()  { return m_label; }
  MJ_Dial     *dial() {return m_dial;}
  QLabel      *display() {return m_display;}

  virtual void show();

public slots:

  void transmitValue(double);

protected:

  void updateLabeltext();
  
  QLabel      *m_label;
  MJ_Dial     *m_dial;
  QLabel      *m_display;

};

#endif
