/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_gui_string.h>

#include <qcombobox.h> 
#include <qlabel.h> 
#include <qlayout.h> 
#include <qlineedit.h> 
#include <qpushbutton.h> 
#include <qscrollview.h> 

MJ_GuiString::MJ_GuiString()
  : m_label(0),
    m_lineEdit(0),
    m_button(0),
    m_fdialog(0)
{}

MJ_GuiString::MJ_GuiString(QObject * parent)
  : QObject(parent),
    m_label(0),
    m_lineEdit(0),
    m_button(0),
    m_fdialog(0)
{}

/** Destructor. Deletes all objects.*/

MJ_GuiString::~MJ_GuiString()
{
  deleteObjs();
}
   
/** Delete all objects. */
void MJ_GuiString::deleteObjs()
{
  if(m_label) {
    delete m_label;
    m_label = 0;
  }
  if(m_lineEdit) {
    delete m_lineEdit;
    m_lineEdit = 0;
  }
  if(m_button) {
    delete m_button;
    m_button = 0;
  }
  if(m_fdialog) {
    delete m_fdialog;
    m_fdialog = 0;
  }
}

/** Creates child objects relative to given parent. */
bool MJ_GuiString::createObjs(
			      /// Parent widget for the new widgets
			      QWidget *parent, 
			      /// Label text
			      const QString &label, 
			      /// Text in the entry box
			      const QString &text)
{
  deleteObjs();

  m_label = new QLabel(label, parent);
  m_lineEdit = new QLineEdit(text, parent);  

  if(!m_label || !m_lineEdit)
    return false;

  connect(m_lineEdit, SIGNAL(returnPressed()),
	  this, SLOT(activate()));
  
  return true;
}

/** Adds children to QScrollView */

void MJ_GuiString::addChildren(QScrollView *s)
{
  if(m_label) s->addChild(m_label);
  if(m_lineEdit) s->addChild(m_lineEdit);
  if(m_button) s->addChild(m_button);
}

/** Creates a browse-button. This method must be called before the
    objects are put to grid. */
bool MJ_GuiString::createBrowseButton(QWidget *parent,
				      const QString &browseText,
				      const QString &filter,
				      QFileDialog::Mode fmode)
{
  m_filter = filter;
  m_fmode  = fmode;

  m_button = new QPushButton(browseText, parent);
  
  if(!m_button) return false;
  
  connect(m_button, SIGNAL(pressed()), this, SLOT(browse()));
  return true;
}

/** Puts all child objects to a grid. */
void MJ_GuiString::toGridHorizontal(QGridLayout *g, 
				    int row, 
				    int firstCol, 
				    int entryWidth)
{
  g->addWidget(m_label, row, firstCol);
  
  g->addMultiCellWidget(m_lineEdit, row, row, firstCol+1, 
			firstCol + entryWidth);
  if(m_button) {
    g->addWidget(m_button, row, firstCol + entryWidth + 1, AlignLeft);
  }
}

/** Show child objects. If the child objects do not exist a seg fault
    is likely to take place. */
void MJ_GuiString::show()
{
  m_label->show();
  m_lineEdit->show();
  if(m_button) m_button->show();
}


void MJ_GuiString::activate()
{
  QString s = m_lineEdit->text();
  sendString(s);
}
 
/** Show a file browsing widget and get a filename. */
void MJ_GuiString::browse()
{
  if(m_fdialog) {
    m_fdialog->raise();
    m_fdialog->show();
    return;
  }

  m_fdialog = new QFileDialog(".", m_filter);
  m_fdialog->setMode(m_fmode);

  // Go to the last directory.
  QFileInfo fi(m_label->text());
  if(fi.exists())
    m_fdialog->setDir(fi.dirPath());

  connect(m_fdialog, SIGNAL(fileSelected(const QString &)),
	  this, SLOT(sendString(const QString &)));
  connect(m_fdialog, SIGNAL(dirEntered(const QString &)),
	  this, SLOT(sendDir(const QString &)));
  m_fdialog->show();
}

/** Emit the string via Qt signal. */
void MJ_GuiString::sendString(const QString &s)
{
  qDebug("MJ_GuiString::sendString %s", (const char *) s);

  QString s2 = m_lineEdit->text();

  if(s2 != s) m_lineEdit->setText(s);

  if(m_id.isEmpty()) emit textEmit(s);
  else emit textEmit(m_id, s);  
}

/** Emit directory name. If we are running in directory mode, the
    string is emitted. Else nothing is done. */
void MJ_GuiString::sendDir(const QString &)
{
  if(m_fmode == QFileDialog::Directory)
    sendString(m_fdialog->selectedFile());
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


MJ_GuiStringSelector::MJ_GuiStringSelector()
  : m_label(0),
    m_box(0)
{}

MJ_GuiStringSelector::~MJ_GuiStringSelector()
{

}
   
void MJ_GuiStringSelector::deleteObjs()
{
  if(m_label) {
    delete m_label;
    m_label = 0;
  }
  if(m_box) {
    delete m_box;
    m_box = 0;
  }
}
 
bool MJ_GuiStringSelector::createObjs(QWidget *parent, 
				      const QString &label)
{
  deleteObjs();
  
  m_label = new QLabel(label, parent);
  m_box = new QComboBox(parent);

  bool ok = m_label && m_box;

  if(ok) {
    connect(m_box, SIGNAL(activated(const QString &)), 
	    this, SLOT(sendString(const QString &)));
  }

  return ok;
}

void MJ_GuiStringSelector::addChildren(QScrollView *sv)
{
  if(m_label) sv->addChild(m_label);
  if(m_box) sv->addChild(m_box);
}

void MJ_GuiStringSelector::toGridHorizontal(QGridLayout *g, 
					    int row, 
					    int firstCol)
{
  g->addWidget(m_label, row, firstCol);
  g->addWidget(m_box, row, firstCol+1, AlignLeft);
}

void MJ_GuiStringSelector::show()
{
  if(m_label) m_label->show();
  if(m_box) m_box->show();
}
   

void MJ_GuiStringSelector::sendString(const QString &s)
{
  if(m_id.isEmpty()) 
    emit textEmit(s);
  else 
    emit textEmit(m_id, s);
}
