/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */
#include <mj_knob.h>
#include <mj_math.h>
#include <mj_misc.h>

#include <qpainter.h>

MJ_Knob::MJ_Knob(QWidget      *parent, 
		 const char   *name)
  : QWidget(parent, name),
    m_val(0.5),
    m_stepSize(0.01),
    m_angle(0.0),
    m_oldAngle(0.0),
    m_min(0.0),
    m_max(1.0),
    m_minAngle(-2.5),
    m_maxAngle(2.5)
{
  setMinimumSize(20,20);
}

MJ_Knob::~MJ_Knob()
{}

void MJ_Knob::incrementStep()
{
  setValue(m_val + m_stepSize);
}

void MJ_Knob::decrementStep()
{
  setValue(m_val - m_stepSize);
}

void MJ_Knob::setValue(double v)
{
  if(v > m_max) v = m_max;
  else if(v < m_min) v = m_min;

  if(v == m_val) return;

  m_val = v;
  m_oldAngle = m_angle;
  m_angle = (v - m_min) / (m_max - m_min) * 
    (m_maxAngle - m_minAngle) + m_minAngle;
  
  if(isVisible()) redrawFast();

  // turnTowards();
  emit emitValue(m_val);
}

void MJ_Knob::mousePressEvent(QMouseEvent *e)
{
  const ButtonState state = e->stateAfter();
  
  // qDebug("MJ_Knob::mousePressEvent");

  if(state & LeftButton) {
    int dx = e->x() - m_centerX;
    int dy = - e->y() + m_centerY;
    
    m_oldAngle = m_angle;
    m_angle = atan2((double) dx, (double) dy);

    if(m_angle == m_oldAngle) return;

    angleToVal();

    if(isVisible()) redrawFast();
    emit emitValue(m_val);
  }

  else if(state & MidButton)
    decrementStep();
  else if(state & RightButton)
    incrementStep();
}

void MJ_Knob::mouseMoveEvent(QMouseEvent *e)
{
  const ButtonState state = e->stateAfter();
  
  // qDebug("MJ_Knob::mouseMoveEvent");

  if(state & LeftButton) {
    mousePressEvent(e);
  }  
}

void MJ_Knob::paintEvent(QPaintEvent *e)
{
  QWidget::paintEvent(e);
  turnTowards();
}

QSize MJ_Knob::sizeHint() const
{ 
  return QSize(40, 40);
}

QSize MJ_Knob::minimumSizeHint() const
{
  return QSize(15, 15);
}

void MJ_Knob::turnTowards()
{  
  if(!isVisible()) return;

  // qDebug("MJ_Knob::turnTowards");

  QPainter painter(this);

  int w = width();
  int h = height();

  painter.eraseRect(0,0,w,h);

  const int centerX = w / 2, centerY = h / 2;
  const int lowerDim = MJ_min(w,h);
  const int diameter = (int) ((double) lowerDim * 0.45);

  m_centerX  = centerX;
  m_centerY  = centerY;
  m_diameter = diameter;

  // const QColorGroup & colors = colorGroup();
  
  redrawFast();
}

void MJ_Knob::redrawFast()
{
  QPainter painter(this);

  const int centerX  = m_centerX;
  const int centerY  = m_centerY;
  const int diameter = m_diameter;
  double diamD = (double) diameter;

  painter.setPen(backgroundColor());
  
  /* The trigonometrics may look a bit odd. The cause: Angle "zero" is
     directly upwards, not to the right. */

  // Remove the old line:
  int targetX = (int) (diamD * sin(m_oldAngle)) + centerX;
  int targetY = (int) (diamD * -cos(m_oldAngle)) + centerY;
  
  painter.drawLine(centerX, centerY, targetX, targetY);

  painter.setPen(foregroundColor());

  painter.drawEllipse(centerX - diameter, 
		      centerY - diameter, 
		      diameter * 2, 
		      diameter * 2);

  /* The trigonometrics may look a bit odd. The cause: Angle "zero" is
     directly upwards, not to the right. */
  targetX = (int) (diamD * sin(m_angle)) + centerX;
  targetY = (int) (diamD * -cos(m_angle)) + centerY;
  
  painter.drawLine(centerX, centerY, targetX, targetY);
}

void MJ_Knob::angleToVal()
{
  if(m_angle > m_maxAngle) m_angle = m_maxAngle;
  else if(m_angle < m_minAngle) m_angle = m_minAngle;

  m_val = (m_angle - m_minAngle) / (m_maxAngle - m_minAngle) * 
    (m_max - m_min);

  // qDebug("MJ_Knob::angleToVal angle: %f val: %f\n", m_angle, m_val);
}
