/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */
/* 
   Copyright (C) 1997--1999 Tommi Ilmonen
   This file is part of the Roaring Frog.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. */

#include <qpainter.h>

// These three are only needed by the test function
#include <qapplication.h>
#include <qlayout.h> 
#include <qtimer.h> 

#include <mj_led.h>

#define LED_WARNINGS 1



MJ_Led::MJ_Led(QWidget      *xpParent, 
	       const char   *cpName)
  : QWidget(xpParent, cpName),
    M_iColor(-1),
    M_xpColor(0),
    M_xShape(SHAPE_ELLIPSE)
{
  M_xColors.setAutoDelete(true);
  QSize xSize(5,5);
  setLedSize(xSize);
}

MJ_Led::~MJ_Led()
{}

void MJ_Led::clearColors()
{
  M_xColors.clear();
  M_iColor  = -1;
  M_xpColor = 0;
}

int MJ_Led::addColor(const QColor &xrC)
{
  QColor *xpTmp = new QColor(xrC);
  if(!xpTmp) return -1;

  M_xColors.append(xpTmp);
  return M_xColors.count() - 1;
}

int MJ_Led::colors()
{
  return M_xColors.count();
}

void MJ_Led::setLedSize(const QSize &xS)
{
  if(xS.height() < 1 || xS.width() < 1) {

#ifdef LED_WARNINGS
    qDebug("MJ_Led::setLedSize, size too small");
#endif
    return;
  }

  M_xSize = xS;
  setMinimumSize(xS);
}

void MJ_Led::setLedSize(int i)
{
  QSize xS(i,i);
  setLedSize(xS);
}

void MJ_Led::blackAndGreen()
{
  colorPair(black, green);
}

void MJ_Led::colorPair(const QColor &c1, const QColor &c2)
{
  clearColors();
  addColor(c1);
  addColor(c2);
}

void MJ_Led::test(QApplication *xpQA, int iArgc, char **cppArgv)
{
  (void) iArgc;
  (void) cppArgv;

  QWidget xW;

  xpQA->setMainWidget(&xW);

  QBoxLayout * xpBox = new QHBoxLayout(&xW, 1, 5);

  for(int i=0; i < 10; i++) {
    // New led:
    MJ_Led *xpLed = new MJ_Led(&xW);
    xpLed->setLedSize(QSize(10, 20));
    xpBox->addWidget(xpLed);

    /// Color space: black, green and red.
    xpLed->blackAndGreen();
    xpLed->addColor(red);

    /// Choose initial color:
    int iColor = 0;
    if(i < 6) iColor = 1;
    xpLed->chooseColor(iColor);

    /// Choose shape:
    int iRect = i % 2;
    xpLed->chooseShape(iRect);

    // Make the led blink:
    QTimer *xpTimer = new QTimer(&xW);
    connect(xpTimer, SIGNAL(timeout()),
	    xpLed, SLOT(incrementColor()));
    xpTimer->start(200 + i * 10);
    
  }
  xpBox->activate();
  xW.show();
  xpQA->exec();
}

void MJ_Led::chooseColor(int iColor)
{
  if(iColor >= (int) M_xColors.count() || iColor < 0) {
#ifdef LED_WARNINGS
    qDebug("MJ_Led::chooseColor, index exceeds colors");
#endif
    return;
  }

  if(iColor == M_iColor) return;

  M_iColor  = iColor;
  M_xpColor = M_xColors.at(M_iColor);

  draw();
}

void MJ_Led::chooseShape(int iS)
{
  shape xS = (shape) iS;
  if(xS == M_xShape) return;
  if((uint) xS > (uint) SHAPE_SIZEOF) {
#ifdef LED_WARNINGS
    qDebug("MJ_Led::chooseShape, index exceeds shapes");
#endif
    return;
  }

  M_xShape = xS;
  draw();
}

void MJ_Led::incrementColor()
{
  if(!M_xColors.count()) return;

  int iNewColor = (M_iColor + 1) % (int) M_xColors.count();
  chooseColor(iNewColor);
}

void MJ_Led::decrementColor()
{
  if(!M_xColors.count()) return;
  
  int iNewColor = M_iColor - 1;
  if(iNewColor == -1) 
    iNewColor = (int) M_xColors.count() - 1;
  
  chooseColor(iNewColor);
}

void MJ_Led::paintEvent(QPaintEvent *xpE)
{
  QWidget::paintEvent(xpE);

  draw();
}

void MJ_Led::draw()
{
  if(!isVisible()) return;

  if(!M_xpColor) return;

  QPainter paint(this);
  paint.eraseRect(0,0,width(), height());

  paint.setBrush(*M_xpColor); // Fill color
  // paint.setPen(*M_xpColor);   // Edge color

  int x = width() / 2 - M_xSize.width() / 2;
  int y = height() / 2 - M_xSize.height() / 2;

  if(M_xShape == SHAPE_ELLIPSE)
    paint.drawEllipse(x, y, M_xSize.width(), M_xSize.height());
  else if(M_xShape == SHAPE_RECTANGLE)
    paint.drawRect(x, y, M_xSize.width(), M_xSize.height());
}
