/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_level_display.h>
#include <mj_math.h>
#include <mj_slider.h>

#include <qapplication.h>
#include <qpainter.h>

#include <math.h>
#include <stdio.h>

template class QArray<MJ_LevelUnit_QG>;
template class QArray<MJ_LevelBar>;

void MJ_LevelBar::init()
{
  m_level = 0.0;
  m_peak = -1.0;
  m_levelPixels = 0;
  m_peakPixels = 0;
  m_pieceIndex = 0;
}

void MJ_LevelBar::setLevels(int index, 
			    QPainter *paint, 
			    MJ_LevelDisplay *dpy,
			    float level, 
			    float peak,
			    float fadeToBackground)
{
  // qDebug("MJ_LevelBar::setLevels ENTER");
  int left, right, height;

  m_level = level;
  m_peak  = peak;
  
  dpy->getGeom(index, &left, &right, &height);
  
  // Level:
  int levelPixels = (int) (height * level);
  if(levelPixels > height) levelPixels = height;
  else if(levelPixels < 0) levelPixels = 0;

  int peakPixels = (int) (height * peak);
  if(peakPixels > height) peakPixels = height;
  
  int endLevel = levelPixels;

  if(levelPixels > m_levelPixels)
    endLevel = fillTo(left, right, levelPixels, peakPixels, paint, dpy, 
		      fadeToBackground);
  else if(levelPixels < m_levelPixels)
    endLevel = blankTo(left, right, levelPixels, peakPixels, paint, dpy, 
		       fadeToBackground);
  else if(m_peakPixels != peakPixels) {
    if(m_peakPixels >= 0)
      undrawPeak(left, right, m_peakPixels, paint, dpy, fadeToBackground);
    if(peakPixels >= 0)
      drawPeak(left, right, peakPixels, paint, dpy);
  }

  m_levelPixels = endLevel;
  m_peakPixels = peakPixels;

  // qDebug("MJ_LevelBar::setLevels EXIT");
}

void MJ_LevelBar::redraw(int index, 
			 QPainter *paint, 
			 MJ_LevelDisplay *dpy,
			 float fadeToBackground)
{
  m_levelPixels = 0;
  m_peakPixels = 0;

  m_pieceIndex = 0;

  float l = m_level, p = m_peak;

  if(fadeToBackground < 1.0) {
    int left, right, height;
    
    dpy->getGeom(index, &left, &right, &height);

    QColor bgc = dpy->backgroundColor();
    QArray<MJ_LevelUnit_QG> & pieces = dpy->pieces();

    for(uint i=0; i < pieces.size(); i++) {      
      MJ_LevelUnit_QG &unit = pieces[(int)i]; 
      
      paint->setBrush(unit.fadeColor(bgc, fadeToBackground));
      paint->setPen(unit.fadeColor(bgc, fadeToBackground));
      dpy->fillRect(paint, left, unit.m_minPixels, right, unit.m_maxPixels);
      
    }
  }
  // float fade = (m_flags & FadeToBackground) ? 0.5 : 1.0;
  
  setLevels(index, paint, dpy, l, p, fadeToBackground);
}

int MJ_LevelBar::fillTo(int left, int right,
			 int pixels,
			 int peakPixels,
			 QPainter *paint, 
			 MJ_LevelDisplay *dpy,
			 float fadeToBackground)
{
  QArray<MJ_LevelUnit_QG> & pieces = dpy->pieces();

  bool loop = true;
  int nofPieces = pieces.size();
  
  int levelPixels = m_levelPixels;
  
  if(pixels == peakPixels) pixels--;

  while(loop) {
    MJ_LevelUnit_QG &piece = pieces[m_pieceIndex];
    
    paint->setBrush(piece.m_color);
    paint->setPen(piece.m_color);
    
    int topPixels;
    
    if(pixels > piece.m_maxPixels)
      topPixels = piece.m_maxPixels;
    else {
      topPixels = pixels;
      loop = false;
    }
    
    if(levelPixels != topPixels)
      dpy->fillRect(paint, 
		    left, 
		    levelPixels,
		    right, 
		    topPixels);
    else 
      dpy->drawLine(paint, 
		    left, 
		    levelPixels,
		    right, 
		    levelPixels);

    levelPixels = topPixels;

    if(loop) m_pieceIndex++;

    if(m_pieceIndex >= nofPieces) {
      m_pieceIndex--;
      loop = false;
    }
  }

  if(m_peakPixels >= 0 && m_peakPixels >= pixels && m_peakPixels != peakPixels)
    undrawPeak(left, right, m_peakPixels, paint, dpy, fadeToBackground);
    
  if(peakPixels >= 0 && peakPixels >= pixels)
    drawPeak(left, right, peakPixels, paint, dpy);

  return pixels;
}

int MJ_LevelBar::blankTo(int left, int right,
			 int pixels,
			 int peakPixels,
			 QPainter *paint, 
			 MJ_LevelDisplay *dpy,
			 float fadeToBackground)
{
  if(pixels == peakPixels) pixels--;
  
  QColor bgc = dpy->backgroundColor();

  bool loop = fadeToBackground < 1.0;

  QArray<MJ_LevelUnit_QG> & pieces = dpy->pieces();

  if(!loop) {

    paint->setBrush(bgc);
    paint->setPen(QPen(bgc, 1));
    
    dpy->fillRect(paint, left, pixels, right, m_levelPixels);
    
    for(; m_pieceIndex; m_pieceIndex--) {
      
      if(pieces[m_pieceIndex - 1].m_maxPixels <= pixels) {
	// return pixels;
	break;
      }
    }
  }
  else {

    int levelPixels = m_levelPixels;

    while(loop) {
      
      QColor currColor = pieces[m_pieceIndex].fadeColor(bgc, fadeToBackground);
      
      paint->setBrush(currColor);
      paint->setPen(currColor);
      
      int minPixels = pieces[m_pieceIndex].m_minPixels;
      
      if(minPixels > pixels) {
	dpy->fillRect(paint, left, minPixels, right, levelPixels);
	levelPixels = minPixels - 1;
	m_pieceIndex--;
      }
      else {
	dpy->fillRect(paint, left, pixels, right, levelPixels);
	levelPixels = pixels - 1;
	loop = false;
      }
      if(m_pieceIndex <= 0) {
	m_pieceIndex = 0;
	loop = false;
      }
    }
  }

  if(m_peakPixels >= 0 && 
     m_peakPixels >= pixels && 
     m_peakPixels != peakPixels)
    undrawPeak(left, right, m_peakPixels, paint, dpy, fadeToBackground);
    
  if(peakPixels >= 0 && peakPixels >= pixels)
    drawPeak(left, right, peakPixels, paint, dpy);

  return pixels;
}
  
void MJ_LevelBar::drawPeak(int left, int right,
			   int pixels,
			   QPainter *paint, 
			   MJ_LevelDisplay *dpy)
{
  paint->setPen(dpy->m_peakColor);
  dpy->drawLine(paint, left, pixels, right, pixels);
}

void MJ_LevelBar::undrawPeak(int left, int right,
			     int pixels,
			     QPainter *paint, 
			     MJ_LevelDisplay *dpy,
			     float fade)
{
  QColor bgc = dpy->backgroundColor();

  if(fade >= 1.0) {
    paint->setPen(bgc);
    dpy->drawLine(paint, left, pixels, right, pixels);
  }
  else {
    
    QArray<MJ_LevelUnit_QG> & pieces = dpy->pieces();

    for(uint i=0; i < pieces.size(); i++) {
      
      MJ_LevelUnit_QG &unit = pieces[(int)i]; 

      if(unit.m_minPixels <= pixels && unit.m_maxPixels >= pixels) {
	
	paint->setPen(unit.fadeColor(bgc, fade));
	dpy->drawLine(paint, left, pixels, right, pixels);
	
	return;
      }
    }
  }
}
  

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_LevelDisplay::MJ_LevelDisplay(QWidget      *xpParent, 
				 const char   *cpName,
				 BarOrient orient,
				 int flags )
  : QWidget(xpParent, cpName),
    m_border(5),
    m_distance(5),
    m_peakColor(QColor(255, 136, 0)),
    m_orient(orient),
    m_flags(flags)
{
  setNofBars(1);

  defaultColors();
  setBackgroundColor(black);
  setMinimumSize(5,5);
  erase();

  if(m_flags & Clickable)
    setCursor(pointingHandCursor);
}

MJ_LevelDisplay::~MJ_LevelDisplay()
{}

void MJ_LevelDisplay::defaultColors(int colors, 
				    const QColor & firstColor, 
				    const QColor & lastColor)
{
  int i;
  int cs = colors - 1;
  if(!cs) cs = 1;

  m_pieces.resize(colors);

  float fCs[3];
  float lCs[3];

  fCs[0] = firstColor.red();
  fCs[1] = firstColor.green();
  fCs[2] = firstColor.blue();

  lCs[0] = lastColor.red();
  lCs[1] = lastColor.green();
  lCs[2] = lastColor.blue();

  for(i=0; i < colors; i++) {
    float ratio1 = (float) i / (float) (cs);
    float ratio2 = 1.0 - ratio1;
    int r = (int) (ratio2 * fCs[0] + ratio1 * lCs[0]);
    int g = (int) (ratio2 * fCs[1] + ratio1 * lCs[1]);
    int b = (int) (ratio2 * fCs[2] + ratio1 * lCs[2]);
    
    m_pieces[i].m_color.setRgb(r, g, b);
    m_pieces[i].m_min = (float) i / (float) colors;
    m_pieces[i].m_max = (float) (i+1) / (float) colors;

    if(i)
      m_pieces[i].m_min = m_pieces[i-1].m_max;
  }
  determineUnitAreas();
}

void MJ_LevelDisplay::getGeom(int index, int *xleft, int *xright, int *h)
{
  int w;

  if(m_orient == BottomUp || m_orient == UpBottom) {
    w = width();
    *h = height() - m_border * 2;
  }
  else {
    *h = width() - m_border * 2;
    w = height();

    // qDebug("*h = %d width() = %d", *h, width());
  }
  
  int wactive = w - m_border * 2;
  int bordersTake = m_distance * (m_bars.size() - 1);

  float wunit = (float) (wactive - bordersTake) / (float) m_bars.size();
  float unitTotal = wunit + (float) m_distance;

  *xleft = (int) ((float) index * unitTotal);
  *xright =  *xleft + (int) wunit;

  /* qDebug("MJ_LevelDisplay::getGeom # %d %d %d %f", 
     index, *xleft, w, wunit); */
}

void MJ_LevelDisplay::setBorder(int b)
{
  m_border = b;
  paintEvent(0);
}

void MJ_LevelDisplay::setNofBars(int n)
{
  // qDebug("MJ_LevelDisplay::setNofBars %d", n);
  
  int old = 0;
  m_bars.resize(n);

  for(int i = old; i < n; i++) {
    m_bars[i].init();
  }
  paintEvent(0);
}

/** Set all bars' levels and peak displays. 
    
    The input arrays must have exactly correct number of elements or
    memory corruption may occur.  */

void MJ_LevelDisplay::set(float *levels, float *peaks)
{
  int s = m_bars.size();

  float peak;

  QPainter paint(this);

  float fade = fadeToBackGround();

  for(int i =0; i < s; i++) {

    if(peaks) peak = peaks[i];
    else peak = -1.0;
    
    m_bars[i].setLevels(i, & paint, this, levels[i], peak, fade);
  }
}


void MJ_LevelDisplay::setLevel(double l)
{
  // qDebug("MJ_LevelDisplay::setLevel %f", l);
  float fl = l;

  set(& fl, 0);
}

void MJ_LevelDisplay::mousePressEvent(QMouseEvent *e)
{
  if(e->stateAfter() != NoButton && (m_flags & Clickable))
    emit clicked(); 
}

void MJ_LevelDisplay::determineUnitAreas()
{
  uint s = m_pieces.size();

  if(!s) return;

  int useH = barHeight();
  
  float useHD = (float) useH;

  for(uint i = 0; i < s; i++) {
    MJ_LevelUnit_QG &piece = m_pieces[(int)i];

    piece.m_minPixels = (int) (piece.m_min * useHD + 0.5);
    piece.m_maxPixels = (int) (piece.m_max * useHD + 0.5);
  }

  m_pieces[(int)s - 1].m_maxPixels = useH;
}

void MJ_LevelDisplay::test(QApplication *xpQA)
{
  // MJ_LevelDisplay *ld = new MJ_LevelDisplay(0, "display", BottomUp);
  // MJ_LevelDisplay *ld = new MJ_LevelDisplay(0, "display", LeftRight);
  MJ_LevelDisplay *ld = new MJ_LevelDisplay(0, "display", RightLeft);
  xpQA->setMainWidget(ld);

  MJ_Slider *s = new MJ_Slider(0.0, 1.1, 0.0);
    
  connect(s, SIGNAL(valueChangedDouble(double)),
	  ld, SLOT(setLevel(double)));

  float l = 0.0, peak = 1.0;

  ld->set(&l, &peak);
  ld->show();
  xpQA->exec();
}

void MJ_LevelDisplay::paintEvent(QPaintEvent *p)
{
  // QWidget::paintEvent(p);
  (void) p;

  determineUnitAreas();

  QPainter paint(this);
  paint.setBrush(backgroundColor());
  paint.setPen(backgroundColor());
  paint.drawRect(0, 0, width(), height());

  float fade = fadeToBackGround();

  for(uint i=0; i < m_bars.size(); i++) {
    m_bars[(int)i].redraw(i, &paint, this, fade);
  }
}

void MJ_LevelDisplay::drawLine
(QPainter *paint, int x1, int y1, int x2, int y2)
{
  if(y1 < 0) return;

  x1 += m_border;
  x2 += m_border;
  y1 += m_border;
  y2 += m_border;

  if(m_orient == BottomUp) {
    int h = height();
    paint->drawLine(x1, h - y1, x2, h - y2);
  }
  else if(m_orient == UpBottom) {
    paint->drawLine(x1, y1, x2, y2);
  }
  else if(m_orient == LeftRight) {
    paint->drawLine(y1, x1, y2, x2);
  }
  else {
    int w = width();
    paint->drawLine(w - y1, x1, w - y2, x2);
  }
}

void MJ_LevelDisplay::fillRect(QPainter *paint, int x1, int y1, int x2, int y2)
{
  int w = x2 - x1 + 1;
  int h = y2 - y1 + 1;

  if(m_orient == BottomUp)
    paint->drawRect(x1 + m_border, height() - (y1 + m_border + h) + 1, w, h);
  else if(m_orient == UpBottom)
    paint->drawRect(x1 + m_border, y1 + m_border, w, h);
  else if(m_orient == LeftRight)
    paint->drawRect(y1 + m_border, x1 + m_border, h, w);
  else {
    int tmpx = width() - (y1 + m_border + h);
    paint->drawRect(tmpx, x1 + m_border, h, w);
  }
}
