/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */


#ifndef MJ_LEVEL_DISPLAY_H
#define MJ_LEVEL_DISPLAY_H

#include <qwidget.h>
#include <mj_color.h>

class MJ_LevelDisplay;

/** A helper class for MJ_LevelDisplay. */
class MJ_LevelUnit_QG
{
public:
  friend class MJ_LevelDisplay;
  friend class MJ_LevelBar;

  MJ_LevelUnit_QG() {}
  ~MJ_LevelUnit_QG() {}

  QColor fadeColor(QColor background, float bgWeight)
  { return (m_color * (1.0 - bgWeight)) + (background * bgWeight); }

protected:  
  
  float m_min;
  float m_max;

  int    m_minPixels, m_maxPixels;

  QColor m_color;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

class MJ_LevelBar
{
public:
  MJ_LevelBar() {}
  ~MJ_LevelBar() {}

  void init();

  void setLevels(int index, 
		 QPainter *paint, 
		 MJ_LevelDisplay *dpy,
		 float level, 
		 float peak,
		 float fadeToBackground);
  
  void redraw(int index, 
	      QPainter *paint, 
	      MJ_LevelDisplay *dpy,
	      float fadeToBackground);

protected:

  int fillTo(int left, int right,
	     int pixels,
	     int peakPixels,
	     QPainter *paint, 
	     MJ_LevelDisplay *dpy,
	     float fadeToBackground);

  int blankTo(int left, int right,
	       int pixels,
	       int peakPixels,
	       QPainter *paint, 
	       MJ_LevelDisplay *dpy,
	      float fadeToBackground);

  void drawPeak(int left, int right,
		int pixels,
		QPainter *paint, 
		MJ_LevelDisplay *dpy);

  void undrawPeak(int left, int right,
		  int pixels,
		  QPainter *paint, 
		  MJ_LevelDisplay *dpy,
		  float fade);
  
  float m_level, m_peak;
  int   m_levelPixels, m_peakPixels;
  int   m_pieceIndex;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** A level display widget. */
class MJ_LevelDisplay : public QWidget
{
  Q_OBJECT
public:
  friend class MJ_LevelBar;

  enum {
    Clickable = 0x1,
    FadeToBackground = 0x2
  };

  enum BarOrient {
    BottomUp,
    UpBottom,
    LeftRight,
    RightLeft
  };
  
  MJ_LevelDisplay(QWidget      *xpParent = 0, 
		  const char   *cpName   = 0,
		  BarOrient orient = BottomUp,
		  int flags = 0);

  virtual ~MJ_LevelDisplay();

  /// Create a color gradient.
  void defaultColors(int colors = 60, 
		     const QColor & firstColor = Qt::white, 
		     const QColor & lastColor = Qt::blue);

  /** Set border width. Default is 3 pixels. */
  void setBorder(int);

  int border() const { return m_border; }

  void setPeak(double);

  void setNofBars(int);

  int nofBars() { return m_bars.size(); }

  void set(float *levels, float *peaks = 0);

  int barHeight() 
  { return (m_orient == BottomUp) ? height() - m_border * 2 
      : width() - m_border * 2 ; }

  QArray<MJ_LevelUnit_QG> & pieces() { return m_pieces; }

  void setFlags(int flags) 
  { m_flags = flags; paintEvent(0); }

  float fadeToBackGround() 
  { return (m_flags & FadeToBackground) ? 0.5 : 1.0; }

  /// Self test.
  static void test(QApplication *xpQA);

public slots:
  /// Sets the fill level, range [0,1].
  void setLevel(double);

signals:

  void clicked();  
protected:
  virtual void mousePressEvent ( QMouseEvent * ) ;

  void determineUnitAreas();

  void paintEvent(QPaintEvent *);

  void setBar(int index, QPainter *paint, float level, float peak);

  void drawPeak(QPainter *paint);

  int barWidth() 
  { return (m_orient == BottomUp) ? width() : height(); }

  void getGeom(int index, int *xleft, int *xright, int *height);

  void drawLine(QPainter *paint, int x1, int y1, int x2, int y2);
  void fillRect(QPainter *paint, int x1, int y1, int x2, int y2);


  int    m_border;
  int    m_distance;

  int    m_pieceIndex;
  QArray<MJ_LevelUnit_QG> m_pieces;

  QColor m_peakColor;

  QArray<MJ_LevelBar> m_bars;

  BarOrient m_orient;

  bool  m_clickable;
  int   m_flags;
};

#endif
