/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <mj_slider.h>
#include <mj_error.h>
#include <mj_mapper.h>
#include <mj_math.h>
#include <mj_misc.h>

MJ_Slider::MJ_Slider(double dMinValue,
		     double dMaxValue,
		     double dValue,
		     QWidget      *xpParent, 
		     const char   *cpName,   
		     Orientation  xOrientation, 
		     int iMinValue, 
		     int iMaxValue, 
		     int iPageStep)
  : QSlider(iMinValue, 
	    iMaxValue, 
	    iPageStep, 
	    iMinValue, 
	    xOrientation, 
	    xpParent, 
	    cpName),
  M_dRealMin(dMinValue),
  M_dRealMax(dMaxValue),
  M_dValue(dValue),
  m_mapper(0)
{
  connect(this, SIGNAL(valueChanged(int)), this, SLOT(setValueInt(int)));
  setValueDouble(dValue);
  if(!xpParent) show();
}

MJ_Slider::~MJ_Slider()
{
  delete m_mapper;
}

void MJ_Slider::setRealRange(double dMin, double dMax)
{ 
  if(dMin > dMax) {
    qDebug("MJ_Slider::setRealRange %f > %f", dMin, dMax);
    return;
  }

  M_dRealMin = dMin; 
  M_dRealMax = dMax; 
  setValueDouble(M_dValue);
}

void MJ_Slider::setSliderRange(int iMin, int iMax)
{ 
  if(iMin >= iMax) {
    qDebug("MJ_Slider::setSliderRange %d >= %d", iMin, iMax);
    return;
  }
  setRange(iMin, iMax);
}

/** Sets the current mapper. Mappers are used to create sliders with
    non-linear feel. 
    
    Passing NULL as argument will result in no mapper being used (the
    default case).  */

void MJ_Slider::setMapper(MJ_Mapper  *m)
{
  delete m_mapper;
  m_mapper = m;
  setValueDouble(M_dValue);
}

void MJ_Slider::setValueDouble(double val)
{
  double localMin(M_dRealMin);
  double localMax(M_dRealMax);
  
  if(val < localMin) val = localMin;
  if(val > localMax) val = localMax;

  M_dValue = val;

  if(m_mapper) {
    localMin = m_mapper->map(localMin);
    localMax = m_mapper->map(localMax);
    val = m_mapper->map(val);
  }

  int iVal;
  
  if(MJ_abs(localMax - localMin) < 1e-10)
    iVal = (minValue() + maxValue ()) / 2;
  else
    iVal = (int) ((val - localMin) / (localMax - localMin) *
		  (double) (maxValue() - minValue())) + minValue();

  bool bDoEmit = signalsBlocked();
  blockSignals(true);
  setValue(iVal);   // Do not emit new value
  blockSignals(bDoEmit);
}

void MJ_Slider::setValueInt(int iVal)
{
  double val = (iVal - minValue()) / 
    (double) (maxValue() - minValue());
  
  if(m_mapper) {
    double localMin = m_mapper->map(M_dRealMin);
    double localMax = m_mapper->map(M_dRealMax);

    val = m_mapper->unmap(val * (localMax - localMin) + localMin);
  }
  else {
    val = val * (M_dRealMax - M_dRealMin) + M_dRealMin;
  }

  M_dValue = val;

  emit valueChangedDouble(M_dValue);
}

void MJ_Slider::getValueDouble(double *dpVal)
{
  /* int iVal = value();
     dVal = (iVal - minValue()) / 
     (double) (maxValue() - minValue()) * 
     (M_dRealMax - M_dRealMin) + M_dRealMin; */

  *dpVal = M_dValue;
}

void MJ_Slider::mouseReleaseEvent(QMouseEvent *e)
{
  QSlider::mouseReleaseEvent(e);
  emit emitDone();
}
