/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_auctrl.h>

#include <mj_app_action_auctrl.h>
#include <mj_global.h>
#include <mj_dsp_plugin.h>
#include <mj_error.h>
#include <mj_kernel.h>
#include <mj_message_qio.h>
#include <mj_module_vbap2.h>
#include <mj_message_string.h>
#include <mj_messages.h>
#include <mj_message_multi2.h>
#include <mj_parameters.h>
#include <mj_realtime_scheduling.h>
#include <mj_strs.h>
#include <mj_version.h>

#include <di_time.h>
#include <di_sleep.h>

#include <stdio.h>

MJ_AcBufferBank::MJ_AcBufferBank()
  : m_inUse(0) 
{ 
  m_list.setAutoDelete(true); 
  m_list.append(new AC_Buffer);
}

AC_Buffer * MJ_AcBufferBank::getUseBuffer()
{
  /* MJ_trace2("MJ_AcBufferBank::getUseBuffer # %p %d %d", 
     this, m_inUse, m_list.count()); */

  if((++m_inUse) > m_list.count()) {
    /* MJ_trace2("MJ_AcBufferBank::getUseBuffer # Adding one %p %d", 
       this, m_list.count()); */
    m_list.append(new AC_Buffer);
    return m_list.last();
  }
  
  return m_list.at(m_inUse - 1);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

bool MJ_ModuleAuCtrl::m_debug = false;

MJ_ModuleAuCtrl::MJ_ModuleAuCtrl()
  : m_currentCmd(0),
    m_vbap(0),
    m_vbap2(0),
    m_synth(0),
    m_appAction(new MJ_AppActionAuCtrl(this)),
    m_dspBuffer(0), m_threadBuffer(2),
    m_buffer(0),
    m_socket(0),
    m_synthPluginId("Tommi.Ilmonen_at_hut.fi_synth1"),
    m_threadActive(false)
{
  MJ_Kernel::actions()->addAction(m_appAction);
}

MJ_ModuleAuCtrl::~MJ_ModuleAuCtrl()
{
  m_appAction->done();

  stopThread();
  delete m_socket;
}

bool MJ_ModuleAuCtrl::dspPrepare()
{
  const char *fname = "MJ_ModuleAuCtrl::dspPrepare";
  
  char buf[32];

  trace3(m_debug, fname);

  stopThread(); // Just to be sure...

  m_requestStop = false;
  msgDisconnectOutput();
  
  /// Children of the main module.
  QList<MJ_ModuleDsp> children;

  MJ_ModuleDsp *host = parentModule();

  if(host) host->getChildren(children, true);
  else {
    return false;
  }

  QListIterator<MJ_ModuleDsp> iter(children);
  MJ_ModuleDsp *m;

  trace3(m_debug, "%s # Found %d children", fname, children.count());

  m_vbap  = 0;
  m_synth = 0;

  while((m = iter.current())) {

    MJ_ModuleVBAP  *vbap  = dynamic_cast<MJ_ModuleVBAP *>  (m);
    // MJ_ModuleSynth *synth = dynamic_cast<MJ_ModuleSynth *> (m);
    MJ_ModuleDsp *synth = 0; // dynamic_cast<MJ_ModuleSynth *> (m);
    if(m->dbe())
      if(m->dbe()->identifier() == m_synthPluginId)
	synth = m;

    if(vbap) {
      if(m_vbap) {
	error(MJ_ERR_UNKNOWN, fname, "Already one VBAP");
	continue;
      }

      trace3(m_debug, "%s # Found a vbap module", fname);
      
      uint i, chans = vbap->dspInChannels();

      for(i=0; i < chans; i++) {
	describeParam(buf, MJ_Strs::azi, i);
	vbap->msgConnectInput(id(), buf, buf);
	describeParam(buf, MJ_Strs::ele, i);
	vbap->msgConnectInput(id(), buf, buf);
      }
      m_vbap = vbap;

      MJ_ModuleVBAP2  *vbap2  = dynamic_cast<MJ_ModuleVBAP2 *>  (vbap);
      
      m_vbap2 = vbap2;

      if(vbap2) {
	for(i=0; i < chans; i++) {
	  describeParam(buf, MJ_Strs::dis, i);
	  vbap->msgConnectInput(id(), buf, buf);
	  describeParam(buf, MJ_Strs::gain, i);
	  vbap->msgConnectInput(id(), buf, buf);
	}
      }
    }
    
    if(synth) {
      if(m_synth) {
	error(MJ_ERR_UNKNOWN, fname, "Already one synth");
	continue;
      }

      synth->msgConnectInput(id(), MJ_Strs::noteon, MJ_Strs::noteon);
      synth->msgConnectInput(id(), MJ_Strs::noteoff, MJ_Strs::noteoff);
            
      m_synth = synth;
    }

    ++iter;
  }

  if(!m_socket) {    
    error(MJ_ERR_UNKNOWN, fname, 
	  "Please use remote socket with AC_Control to start processing");
    return false;
  }

  msgCompileOutputs();

  for(int i=0; i < AUCTRL_BUFFERS;i++)
    m_bufferBanks[i].clearUseCount();

  // Now span the thread:
  m_threadActive = true;
  m_endThread = false;
  m_endRequested = false;
  m_collectingContiguous = false;
  run(true);

  return true;
}


DI_Time tt1, tt2;

void MJ_ModuleAuCtrl::dspProcessPeriod()
{
  DI_Time t3;

  t3.getTimeOfDay();

  if(m_requestStop) {
    if(!m_endRequested && MJ_kernel) MJ_AppActions::stopDsp();
    m_endRequested = true;
    return;
  }

  tt1.getTimeOfDay();

  m_mutex.lock();
  
  if((m_dspBuffer + 1) == m_threadBuffer) {
    m_mutex.unlock();
    return;
  }

  MJ_AcBufferBank *threadBufBank = 
    & m_bufferBanks[(++m_dspBuffer) % AUCTRL_BUFFERS];

  // trace2("MJ_ModuleAuCtrl::dspProcessPeriod # Processing %d", m_dspBuffer);

  m_mutex.unlock();
  
  /* trace2("MJ_ModuleAuCtrl::dspProcessPeriod # %g seconds %g", 
     tt1.secsSince(), tt2.secsSince());*/
  
  tt2.getTimeOfDay();

  for(uint i = 0; i < threadBufBank->inUse(); i++) {

    /*trace2("MJ_ModuleAuCtrl::dspProcessPeriod # %d of %d", 
      i, threadBufBank->inUse());*/

    m_currentCmd = threadBufBank->get(i);

    if(!m_currentCmd) {
      return;
    }
    
    // trace2("MJ_ModuleAuCtrl::dspProcessPeriod # %p", m_currentCmd);

    m_currentCmd->readHeader(&m_header);
  
    switch(m_header.type()) {
    case AC_MT_PLAY_SAMPLE_PITCHED:
    case AC_MT_PLAY_SAMPLE:
      playSample();
      break;
      
    case AC_MT_PLAY_INPUT:
      playInput();
      break;
      
    case AC_MT_NOTE_OFF:
      noteOff();
      break;
      
    case AC_MT_CHANGE_SOURCE_ORIENTATION:
      changeSourceOrientation();
      break;
      
    case AC_MT_CHANGE_SOURCE_LOCATION:
      changeSourceLocation();
      break;
      
    case AC_MT_NOTE_GAIN:
      setNoteGain();
      break;
      
    case AC_MT_NOTE_RELATIVE_PITCH:
      setNoteRelativePitch();
      break;

    case AC_MT_MJ_MSG:
      processMjMessage();
      break;
      
    default:
      error(MJ_ERR_ARGUMENT, 
	    "MJ_ModuleAuCtrl::dspProcessPeriod # Bad type",
	    m_header.type());
      break;
    }
    
  }

  // Tell the thread that the buffer is consumed.
  m_currentCmd = 0;
  // m_condition.signal2(m_mutex);

  // trace2("MJ_ModuleAuCtrl::dspProcessPeriod # End %g seconds", t3.secsSince());
  
}


bool MJ_ModuleAuCtrl::dspStop()
{
  stopThread();
  
  m_appAction->dspStop();

  delete m_socket;
  m_socket = 0;

  m_vbap = 0;
  m_synth = 0;

  return true;
}

bool MJ_ModuleAuCtrl::msgProcess(const char * id, MJ_Message *m)
{
  const char *fname = "MJ_ModuleAuCtrl::msgProcess";

  trace3(m_debug, fname);

  if(!strcmp(id, MJ_Strs::address)) {
    MJ_MessageString *ms(m->toMessageString());
    if (ms) {
      m_address = ms->string();
      m_appAction->newAddress(m_address);
      return true;
    }
  }
  if(!strcmp(id, "synthtype")) {
    MJ_MessageString *ms(m->toMessageString());
    if (ms) {
      m_synthPluginId = ms->string();
      return true;
    }
  }
 
  return false;
}

MJ_Parameter * MJ_ModuleAuCtrl::msgGetInput(const char * id)
{
  trace3(m_debug, "MJ_ModuleAuCtrl::msgGetInput");
  
  if(!strcmp(id, MJ_Strs::address)) {
    return new 
      MJ_ParameterString(m_address, "Socket address", MJ_Strs::address);
  }
  if(!strcmp(id, "synthtype")) {
    return new 
      MJ_ParameterString(m_synthPluginId, "Synth type id", id);
  }
  
  return 0;
}

QCString MJ_ModuleAuCtrl::msgGetInputName(int i)
{
  QCString s;

  if(i == 0) s = MJ_Strs::address;
  if(i == 1) s = "synthtype";
  
  return s;
}

void MJ_ModuleAuCtrl::setSocket(DI_Socket *s)
{ 
  if(dspIsRealTime()) {
    MJ_fatal(MJ_ERR_INTERNAL,
	     "MJ_ModuleAuCtrl::setSocket # Must not call in real-time !");
  }

  delete m_socket; 
  m_socket = s; 
}

void MJ_ModuleAuCtrl::describeParam(char *buf, const char * name, int i)
{
  sprintf(buf, "%s_%d", name, i);
}

void MJ_ModuleAuCtrl::playSample()
{
  char buf[256];

  AC_Buffer   *cmd = m_currentCmd;
  
  int   loop = 0;

  cmd->readString(buf);
  float gain = cmd->readFloat32();
  float pitch;
  
  if(m_header.type() == AC_MT_PLAY_SAMPLE_PITCHED) {
    pitch = cmd->readFloat32();
  }
  else {
    pitch = -1.0;
  }

  float azi    = cmd->readFloat32();
  float ele    = cmd->readFloat32();
  float dist   = cmd->readFloat32();
  float dgain  = cmd->readFloat32();

  loop = cmd->readInt32();
  float skipSeconds = cmd->readFloat32();
  float interleave  = cmd->readFloat32();

  int sourceId = m_header.sourceId();

  trace3(m_debug, 
	 "MJ_ModuleAuCtrl::playSample # \"%s\" azi = %f, ele = %f, chan = %d,"
	 " loop = %d, pitch = %f",
	 buf, azi, ele, sourceId, loop, pitch);

  MJ_MessageMulti2 mm;

  mm.addChild(MJ_Strs::loop, new MJ_MessageBool(loop));
  mm.addChild(MJ_Strs::gain, new MJ_MessageDouble(gain));
  mm.addChild(MJ_Strs::filename, new MJ_MessageString(buf));
  mm.addChild(MJ_Strs::channel, new MJ_MessageInt(sourceId));
  mm.addChild("skipseconds", new MJ_MessageDouble(skipSeconds));

  // Trigger the sample:
  if(pitch > 0.0) {
    mm.addChild("voicetype", new MJ_MessageString("pitchsample"));
    mm.addChild("relpitch", new MJ_MessageDouble(pitch));
  }
  else {
    mm.addChild("voicetype", new MJ_MessageString("sample"));
    // MJ_MessagePlaySample mps(-1, buf, gain, sourceId, loop);
    // msgPushOutput(MJ_Strs::noteon, &mps);
  }

  if(interleave != 0.0) {
    mm.addChild("interleave", new MJ_MessageDouble(interleave));
  }

  msgPushOutput(MJ_Strs::noteon, &mm);

  // Pan the source:
  {
    // Azimuth:
    describeParam(buf, MJ_Strs::azi, sourceId);
    MJ_MessageDouble md(azi, -2);
    msgPushOutput(buf, &md);

    // Elevation:
    describeParam(buf, MJ_Strs::ele, sourceId);
    md.setValue(ele);
    msgPushOutput(buf, &md);
    
    // Distance
    describeParam(buf, MJ_Strs::dis, sourceId);
    md.setValue(MJ_metersToSeconds(dist));
    msgPushOutput(buf, &md);
    
    // Distance gain
    describeParam(buf, MJ_Strs::gain, sourceId);
    md.setValue(dgain);
    msgPushOutput(buf, &md);
  }
}

void MJ_ModuleAuCtrl::playInput()
{
  char buf[32];

  AC_Buffer   *cmd = m_currentCmd;
  
  int channel  = cmd->readInt32();
  float gain   = cmd->readFloat32();
  float azi    = cmd->readFloat32();
  float ele    = cmd->readFloat32();
  float dist   = cmd->readFloat32();
  float dgain  = cmd->readFloat32();

  int sourceId = m_header.sourceId();

  trace3(m_debug, "MJ_ModuleAuCtrl::playInput # %d -> %d, gain = %f",
	 channel, sourceId, gain);

  { 
    MJ_MessageMulti2 mm;

    mm.addChild(MJ_Strs::gain, new MJ_MessageDouble(gain));
    mm.addChild(MJ_Strs::channel, new MJ_MessageInt(sourceId));
    mm.addChild("source", new MJ_MessageInt(channel));
    mm.addChild("voicetype", new MJ_MessageString("input"));
    
    // MJ_MessagePlayInput mpi(-1, channel, sourceId, gain);
    msgPushOutput(MJ_Strs::noteon, &mm);
  }

  // Pan the source:
  {
    // Azimuth:
    describeParam(buf, MJ_Strs::azi, sourceId);
    MJ_MessageDouble md(azi, -2);
    msgPushOutput(buf, &md);

    // Elevation:
    describeParam(buf, MJ_Strs::ele, sourceId);
    md.setValue(ele);
    msgPushOutput(buf, &md);

    // Distance
    describeParam(buf, MJ_Strs::dis, sourceId);
    md.setValue(dist);
    msgPushOutput(buf, &md);
    
    // Distance gain
    describeParam(buf, MJ_Strs::gain, sourceId);
    md.setValue(MJ_metersToSeconds(dgain));
    msgPushOutput(buf, &md);
  }
}

void MJ_ModuleAuCtrl::noteOff()
{
  trace3(m_debug, "MJ_ModuleAuCtrl::noteOff # %d", m_header.sourceId());

  MJ_MessageInt mi(m_header.sourceId());

  msgPushOutput(MJ_Strs::noteoff, &mi);
}

void MJ_ModuleAuCtrl::changeSourceOrientation()
{
  char buf[32];

  AC_Buffer   *cmd = m_currentCmd;
  
  float interpTime = cmd->readFloat32();
  float azi  = cmd->readFloat32();
  float ele  = cmd->readFloat32();
  int channel = m_header.sourceId();
  
  long ts = -1;

  if(m_currentTime) {
    ts = (long) 
      (m_currentTime->stamp() + (long) (interpTime * dspSamplingRate()));
  }

  describeParam(buf, MJ_Strs::azi, channel);
  MJ_MessageDouble md(azi, ts);
  msgPushOutput(buf, &md);

  describeParam(buf, MJ_Strs::ele, channel);
  md.setValue(ele);
  msgPushOutput(buf, &md);
}

void MJ_ModuleAuCtrl::changeSourceLocation()
{
  char buf[32];

  AC_Buffer   *cmd = m_currentCmd;
  
  float interpTime = cmd->readFloat32();
  float azi   = cmd->readFloat32();
  float ele   = cmd->readFloat32();
  float dist  = cmd->readFloat32();
  float dgain  = cmd->readFloat32();
  int channel = m_header.sourceId();
  
  long ts = -1;

  if(m_currentTime) {
    ts = (long) 
      (m_currentTime->stamp() + (long) (interpTime * dspSamplingRate()));
  }

  describeParam(buf, MJ_Strs::azi, channel);
  MJ_MessageDouble md(azi, ts);
  msgPushOutput(buf, &md);

  describeParam(buf, MJ_Strs::ele, channel);
  md.setValue(ele);
  msgPushOutput(buf, &md);

  describeParam(buf, MJ_Strs::dis, channel);
  md.setValue(MJ_metersToSeconds(dist));
  msgPushOutput(buf, &md);

  describeParam(buf, MJ_Strs::gain, channel);
  md.setValue(dgain);
  msgPushOutput(buf, &md);
}

void MJ_ModuleAuCtrl::setNoteGain()
{
  if(!m_synth) return;

  long noteId = (long) m_currentCmd->readInt64();
  float gain = m_currentCmd->readFloat32();

  char buf[32];
  sprintf(buf, "note/%ld/gain", noteId);
  MJ_MessageDouble md(gain);

  m_synth->msgProcess(buf, &md);
}

void MJ_ModuleAuCtrl::setNoteRelativePitch()
{
  if(!m_synth) return;

  long noteId = (long) m_currentCmd->readInt64();
  float value = m_currentCmd->readFloat32();

  char buf[32];
  sprintf(buf, "note/%ld/relpitch", noteId);
  MJ_MessageDouble md(value);

  m_synth->msgProcess(buf, &md);
}

/** Gets an ordinary Mustajuuri message from the buffer and sends it
    to the proper recipient. */

void MJ_ModuleAuCtrl::processMjMessage()
{
  std::string tmp;

  m_currentCmd->readString(&tmp);
  MJ_ModuleDsp *m = 0;

  const char * address = & * tmp.begin();

  if(strncmp("$(SYNTH)/", address, 9) == 0) {
    m = m_synth;
    address += 9;
  }
  else if(strncmp("$(VBAP)/", address, 8) == 0) {
    m = m_vbap;
    address += 8;
  }
  
  if(!m) {
    trace3(m_debug, 
	   "MJ_ModuleAuCtrl::processMjMessage # Premature return \"%s\"",
	   address);
    
    return;
  }

  MJ_Message *msg = MJ_Message::readFromBuffer(m_currentCmd);

  if(msg) {

    if(m_debug)
      MJ_MessageQio::writeCout(msg);

    m->msgProcess(address, msg);
    delete msg;
  }
}

void MJ_ModuleAuCtrl::stopThread()
{
  trace3(m_debug, "MJ_ModuleAuCtrl::stopThread");

  if(m_threadActive) {

    m_endThread = true;

    m_currentCmd = 0;
    m_condition.signal2(m_mutex);

    waitEnd();
  }

  m_currentCmd = 0;
  m_endThread = false;
  m_threadActive = false;
}

void MJ_ModuleAuCtrl::childLoop()
{
  const char *fname = "MJ_ModuleAuCtrl::childLoop";
  
  trace3(m_debug, fname);

  if(MJ_realTimeScheduling(MJ_RTP_RT_IO)) 
  //if(MJ_realTimeScheduling(98)) 
    trace3(m_debug, "%s # Got RT-scheduling", fname);

  MJ_AcBufferBank *threadBufBank = 
    &m_bufferBanks[m_threadBuffer % AUCTRL_BUFFERS];

  /*DI_Time t1;
    t1.getTimeOfDay();*/
  

  while(true) {
    
    if(!m_collectingContiguous) {
      m_mutex.lock();
      
      if((m_threadBuffer - m_dspBuffer) == 1) {
	m_threadBuffer++;
	threadBufBank = & m_bufferBanks[m_threadBuffer % AUCTRL_BUFFERS];
	threadBufBank->clearUseCount();
      }
      m_mutex.unlock();
    }

    /* trace2("%s # X1 %g seconds", fname, t1.secsSince());
       t1.getTimeOfDay();*/

    // trace2("%s # Loop1", fname);

    // See if we should exit:
    if(m_endThread) {
      trace3(m_debug, "%s # EXIT", fname);
      return;
    }

    // Check that the socket is still ok:
    if(!m_socket->isOK() || m_socket->isEOF()) {
      if(!m_requestStop)
	error2(MJ_ERR_UNKNOWN, "%s # Socket connection lost", fname);
      m_requestStop = true;
      m_appAction->dspStop();
      m_endThread = true;
    }

    // trace2("%s # Loop2", fname);

    /* trace2("%s # X2 %g seconds", fname, t1.secsSince());
       t1.getTimeOfDay(); */
    
    if(m_requestStop) {
      DI_Sleep::sleepMs(100);
    }
    else if(m_socket->isPendingInput(500000)) {
      // Read data from the socket:
      
      if(!m_socket->isOK() || m_socket->isEOF())
	continue;

      m_buffer = threadBufBank->getUseBuffer();

      int bytes; 
      if((bytes = DI_readFromSocket(m_socket, m_buffer)) <= 0) 
	continue;

      trace3(m_debug, "%s # New event %d = 0x%x %d %p", 
	     fname, bytes, bytes, m_threadBuffer, m_buffer);

      m_buffer->readHeader(&m_header);

      if(threadHandleMsg())
	threadBufBank->unuseLastBuffer();

      /*trace2("%s # X3 %g seconds", fname, t1.secsSince());
	t1.getTimeOfDay(); */

    }

    /* trace2("%s # X4 %g seconds", fname, t1.secsSince());
       t1.getTimeOfDay(); */
  }
}

bool MJ_ModuleAuCtrl::threadHandleMsg()
{
  const char * fname = "MJ_ModuleAuCtrl::threadHandleMsg";
  
  // trace2("MJ_ModuleAuCtrl::threadHandleMsg # %d", (int) m_header.type());

  // int tmp;

  switch(m_header.type()) {
  case AC_MT_START:
    m_buffer->writeHeader(0, 0, AC_MT_STARTED, 0);
    DI_writeToSocket(m_socket, m_buffer);
    trace3(m_debug, "%s # Done start reply", fname);
    return true;

  case AC_MT_STOP:
    trace3(m_debug, "%s # Received STOP message", fname);
    m_buffer->writeHeader(0, 0, AC_MT_STOPPED, 0);
    DI_writeToSocket(m_socket, m_buffer);
    m_requestStop = true;
    return true;
    
  case AC_MT_IS_OK:
    trace3(m_debug, "%s # Received IS_OK message", fname);
    m_buffer->writeHeader(0, 0, AC_MT_IS_OK, 0);
    m_buffer->writeInt32(1);
    DI_writeToSocket(m_socket, m_buffer);
    return true;

  case AC_MT_QUERY_PANNING:
    trace3(m_debug, "%s # Received QUERY_PANNING message", fname);
    m_buffer->writeHeader(0, 0, AC_MT_PANNING, 0);
    m_buffer->writeInt32(m_vbap ? AC_PAN_VBAP : AC_PAN_UNKNOWN);
    DI_writeToSocket(m_socket, m_buffer);
    return true;

  case AC_MT_BEGIN_CONTIGUOUS:
    trace3(m_debug, "%s # Received BEGIN_CONTIGUOUS message", fname);
    m_collectingContiguous = true;
    return true;

  case AC_MT_END_CONTIGUOUS:
    trace3(m_debug, "%s # Received END_CONTIGUOUS message", fname);
    m_collectingContiguous = false;
    return true;
  }

  return false;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_ModuleDsp * MJ_createNew_ModuleModuleAuCtrl2() 
{
  return new MJ_ModuleAuCtrl;
}

MJ_PluginDbeDsp *MJ_createPlugin(int i)
{
  if(i) return 0;

  MJ_PluginDbeDsp *dbe = new MJ_PluginDbeDsp;
  dbe->setAuthors("Ilmonen Tommi");
  dbe->setMenuPosition("_Audio_Auralize_Control2");
  dbe->setCreator(MJ_createNew_ModuleModuleAuCtrl2);
  dbe->setIdentifier("Tommi.Ilmonen_at_hut.fi_aural_ctrl2");
  dbe->pluginVersion()->set(0,0,1);
  dbe->options()->setChans(0, 0, 0, 0);
  dbe->setPluginDebug(MJ_ModuleAuCtrl::debugPtr());
  dbe->setParameter(MJ_PP_COMPILE_TIME, DATETIME);
  
  return dbe;  
}

