/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

/* This file contains the necessary code to automatically generate
   audio configuration information on SGI and Linux/ALSA.
   
   Not exactly very beautifull, but does the job...  */

#include <mj_array.h>
#include <mj_audio_conf.h>
#include <mj_midi_conf.h>
#include <mj_error.h>

#include <di_audio_port.h> 

#include <qfileinfo.h> 
#include <qstring.h> 

#include <stdlib.h>
#include <stdio.h>

#define MJ_DEBUG_DIVA 1

/* Store the environment variable string to this variable so the
   string will exist as long as the plugin is loaded. */
static QCString MJ_divaAudioConfig;
static QCString MJ_divaMidiConfig;

static void MJDI_buildAudio();
static void MJDI_buildMidi();


void MJ_DivaBuildEnv()
{
  MJDI_buildAudio();
  MJDI_buildMidi();
}

void MJDI_buildMidi()
{
  if(getenv(MJ_MidiEnvVarName)) return;
  
#if defined(DI_AUDIO_IS_SGI)
  const char *devs = "=Input=Software Synth|Output=Software Synth";
#else
  QFileInfo fi("/dev/midi");
  const char *devs;

  if(fi.exists())
    devs = "=Input=/dev/midi|Output=/dev/midi";
  else 
    devs = "=Input=/dev/midi0|Output=/dev/midi0";
    
#endif

  MJ_divaMidiConfig = MJ_MidiEnvVarName;
  MJ_divaMidiConfig += devs;
  
  putenv(MJ_divaMidiConfig.data());
  MJ_trace("Default MIDI devices", MJ_divaMidiConfig.data());
}

// #if defined(DI_AUDIO_IS_ALSA) || defined(DI_AUDIO_IS_SGI)
#if defined(DI_AUDIO_IS_ALSA) || defined(DI_AUDIO_IS_OSS)

void MJDI_buildAudio()
{
  if(getenv(MJ_AudioEnvVarName)) return;
  
  MJ_divaAudioConfig = MJ_AudioEnvVarName;
  
#ifdef DI_AUDIO_IS_ALSA
#ifdef ALSA_IS_0_9
  const char *devs = "=Input=2=hw:0,0|Output=2=hw:0,0";
#else
  const char *devs = "=Input=2=0 0|Output=2=0 0";
#endif  // ALSA_IS_0_9
#else   // DI_AUDIO_IS_ALSA
  // For OSS audio drivers
  const char *devs = "=Input=2=/dev/dsp|Output=2=/dev/dsp";
#endif  // DI_AUDIO_IS_ALSA
  MJ_divaAudioConfig += devs;
  putenv(MJ_divaAudioConfig.data());
  MJ_trace("Default audio devices", MJ_divaAudioConfig.data());
}

#elif defined(DI_AUDIO_IS_SGI)

// Use this macro to get debugging information.

// #define MJ_DEBUG_DIVA

void MJDI_buildAudio()
{
  const char * fname = "MJ_DivaBuildEnv";

#ifdef MJ_DEBUG_DIVA
  MJ_trace(fname);
#endif

#if defined(DI_AUDIO_IS_ALSA) 
  
  /* Alsa support does not really work too well :-|
     
     The current code breaks the ALSA driver on my machine...

     Just initialize one stereo input and output. Hope this is enough
     for the user. Now nothing should crash. */
  
  if(!getenv(MJ_AudioEnvVarName)) {
    MJ_divaAudioConfig = MJ_AudioEnvVarName;
    MJ_divaAudioConfig += "=";
    MJ_divaAudioConfig  += "Stereo_input=input 1|Stereo_output=output 1";
    
    putenv(MJ_divaAudioConfig.data());
  }

  MJ_trace(fname, getenv(MJ_AudioEnvVarName));
  return;

#endif // DI_AUDIO_IS_ALSA


  // Do not override the value given by the user.
  const char *ae = getenv(MJ_AudioEnvVarName);
  if(ae) {
    if(strlen(ae))
      return;
  }

  ulong i;
  MJ_Array<ulong> ids;

  char buf[DI_AUDIO_MAX_NAME_LENGTH];
  char buf2[DI_AUDIO_MAX_NAME_LENGTH + 128];

  QCString config;

  // Input devices.
  ulong indevs = DI_AudioDevice::getNumberOfInputDevices();
  ids.resize(indevs);
  DI_AudioDevice::getInputDevices(ids.data());

  for(i=0; i < indevs; i++) {
    ulong id = ids.get(i);

    DI_AudioDevice::getInputDeviceName(id, buf);
    
#ifdef MJ_DEBUG_DIVA
    MJ_trace(fname, buf);
#endif    

    // See how large number of channels we can get:
    for(int channels = 64; channels > 0; channels--) {
      DI_AudioDevice d(id, true, channels);
      
      if(d.isOK() && d.getChannels() == channels) {
	sprintf(buf2, "|%s=%d=%s", MJ_AudioDeviceTypeKeys[MJ_ADT_INPUT], 
		channels,
		buf);
	config += buf2;
	channels = 1;
      }
    }
  }

#ifdef MJ_DEBUG_DIVA
  MJ_trace(fname,  config);
#endif    

  // Output devices.
  ulong outdevs = DI_AudioDevice::getNumberOfOutputDevices();
  ids.resize(outdevs);
  DI_AudioDevice::getOutputDevices(ids.data());
  
  for(i=0; i < outdevs; i++) {
    ulong id = ids.get(i);
    
    DI_AudioDevice::getOutputDeviceName(id, buf);
    
#ifdef MJ_DEBUG_DIVA
    MJ_trace(fname, buf);
#endif    

    // See how large number of channels we can get:
    for(int channels = 64; channels > 0; channels--) {
      DI_AudioDevice d(id, false, channels);
      
      if(d.isOK() && d.getChannels() == channels) {
	sprintf(buf2, "|%s=%d=%s", MJ_AudioDeviceTypeKeys[MJ_ADT_OUTPUT], 
		channels,
		buf);
	config += buf2;
	channels = 1;	
      }
    }
  }
  MJ_divaAudioConfig.resize(config.length() + 64);
  
  MJ_divaAudioConfig.sprintf("%s=%s", MJ_AudioEnvVarName, config.data());
  
  putenv(MJ_divaAudioConfig.data());
#ifdef MJ_DEBUG_DIVA
  MJ_trace(fname, MJ_divaAudioConfig);
#endif
}

#endif // defined(DI_AUDIO_IS_ALSA) || defined(DI_AUDIO_IS_SGI)

