/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_diva_af.h>

#include <mj_gui_string.h>
#include <mj_message_string.h>
#include <mj_module_dsp_row_gui.h>
#include <mj_module_level.h>
#include <mj_satellite.h>

#include <mj_error.h>
#include <mj_global.h>
#include <mj_iostream.h>
#include <mj_kernel.h>
#include <mj_messages.h>
#include <mj_parameters.h>
#include <mj_realtime_scheduling.h>
#include <mj_strs.h>

#include <qapplication.h>
#include <qcombobox.h> 
#include <qfiledialog.h> 

#include <string.h>

bool MJ_ModuleDivaAudioFile::m_debug = false;

MJ_ModuleDivaAudioFile::MJ_ModuleDivaAudioFile()
  : m_fileName("sound.wav")
{
  if(m_debug) trace("MJ_ModuleDivaAudioFile::MJ_ModuleDivaAudioFile");
  dspSetOutChannels(1);
}

MJ_ModuleDivaAudioFile::~MJ_ModuleDivaAudioFile()
{
  if(m_debug) trace("MJ_ModuleDivaAudioFile::~MJ_ModuleDivaAudioFile");
}

bool MJ_ModuleDivaAudioFile::dspStop()
{
  const char * fname = "MJ_ModuleDivaAudioFile::dspStop";

  if(m_debug) trace(fname);
  m_file.requestEnd();
  if(m_debug) trace(fname, "HALF");
  m_file.waitEnd();
  if(m_debug) trace(fname, "EXIT");
  return true;
}

bool MJ_ModuleDivaAudioFile::msgProcess(const char * i,MJ_Message *m)
{
  if(!strcmp(i, MJ_Strs::filename)) {
    MJ_MessageString *ms = dynamic_cast<MJ_MessageString *> (m);

    if(!ms) return false;
    if(ms->stringConst().isEmpty()) return false;
    
    m_fileName = ms->stringConst().copy();
    return true;
  }

  return false;
}

MJ_Parameter *MJ_ModuleDivaAudioFile::msgGetInput(const char * i)
{
  // trace("MJ_ModuleDivaAudio::msgGetInput", i);

  if(!strcmp(i, MJ_Strs::filename))
    return new MJ_ParameterFilename(m_fileName, 
				    "Audio file", 
				    MJ_Strs::filename,
				    "*.wav *.aiff *.aif *.aifc *.voc",
				    QFileDialog::ExistingFile);
  return 0;
}
						    
QCString MJ_ModuleDivaAudioFile::msgGetInputName(int i)
{
  return MJ_copyIfZero(MJ_Strs::filename, i);  
}


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

const char * MJ_ModuleDivaAudioFileIn::m_eofEndsStr = "eofEnds";

MJ_ModuleDivaAudioFileIn::MJ_ModuleDivaAudioFileIn()
  : m_eofEnds(true),
    m_loop(false)
{}

   MJ_ModuleDivaAudioFileIn::~MJ_ModuleDivaAudioFileIn()
{}

bool MJ_ModuleDivaAudioFileIn::dspAcceptIoChans(int * in, int * out)
{
  (void) in;

  const char * fname = "MJ_ModuleDivaAudioFileIn::dspAcceptIoChans";
  
  if(m_debug) trace(fname);

  QFileInfo  fi(m_fileName);
  
  if(!fi.exists()) 
    return false;

  DI_AudioFile af;
  if(!af.open(m_fileName, DI_IO_INPUT, DI_ASF_SAMPLE, 128))
    return false;
  
  bool change = false;

  /* if(*in) {
    *in = 0;
    change = true;
    }*/
  if(*out != (int) af.inChannels()) {
    *out = (int) af.inChannels();
    change = true;
    MJ_trace(fname, *out);
  }

  return change;
}

/** Open the file. */

bool MJ_ModuleDivaAudioFileIn::dspPrepare()
{
  const char * fname = "MJ_ModuleDivaAudioFileIn::dspPrepare";

  QFileInfo  fi(m_fileName);
  
  if(!fi.exists()) 
    return false;

  if(!m_signalOutput.size()) {
    error(MJ_ERR_INTERNAL, fname, "NO channels");
    return false;
  }
  
  const uint sr = dspSamplingRate();
  long startFrame = (long) m_currentTime->stamp();

  m_file.startRead(m_fileName, DI_ASF_FLOAT32, sr * 2, startFrame, 2000000000,
		   m_loop, MJ_RTP_FILE_IO);

  trace2("%s # Began reading file \"%s\" at frame %ld", 
	 fname, m_fileName.ascii(), startFrame);
  
  if(!m_file.userPrepare()) {
    error(MJ_ERR_INTERNAL, fname, "Open failed");
    dspStop();
    return false;
  }
  
  int chans = (int) m_file.file().inChannels();
  m_endFrame = m_file.file().frameCount();

  if((uint) chans != m_signalOutput.size() || !chans) {
    error2(MJ_ERR_ARGUMENT, " %s # bad number of channels, %d != %d",
	   fname, chans, m_signalOutput.size());
    dspStop();
    return false;
  }

  m_buffer.resize(chans * m_signalOutput[0].samples().size());


  return true;
}

void MJ_ModuleDivaAudioFileIn::dspProcessPeriod()
{
  // trace("MJ_ModuleDivaAudioFileIn::dspProcessPeriod");

  long frames = (long) m_signalOutput[0].samples().size();
  
  m_file.readFrames(m_buffer.data(), frames);

  for(uint i=0; i < m_signalOutput.size(); i++) {
    // Fill the signal portion
    m_signalOutput[i].samples().copySkipping
      (m_buffer.data(), 
       i,
       frames,
       m_signalOutput.size());
  }

  if(!m_file.loop() && 
     m_eofEnds && 
     (m_endFrame < (long) m_currentTime->stamp())) {
    stopProcessing();
  }
}

void MJ_ModuleDivaAudioFileIn::stopProcessing()
{
  trace2("MJ_ModuleDivaAudioFileIn::stopProcessing %d %d", 
	 (int) m_endFrame, (int) m_currentTime->stamp());
  if(MJ_kernel) MJ_kernel->stopPlaying(true);
}

bool MJ_ModuleDivaAudioFileIn::msgProcess(const char * i,MJ_Message *m)
{
  const char * fname = "MJ_ModuleDivaAudioFileIn::msgProcess";

  if(!strcmp(i, MJ_Strs::filename)) {
    return MJ_ModuleDivaAudioFile::msgProcess(i, m);
  }
  else if(!strcmp(i, MJ_Strs::loop)) {
    bool ok = true;
    bool loop = m->toBool(&ok);
    if(!ok) {
      error(MJ_ERR_ARGUMENT, fname, "Loop not a bool message");
      return false;
    }
    m_loop = loop;

    trace2("%s # Loop = %d", fname, (int) m_loop);

    return true;
  }
  else if(!strcmp(i, m_eofEndsStr)) {
    bool ok = true;
    bool eofEnds = m->toBool(&ok);
    if(!ok) {
      error(MJ_ERR_ARGUMENT, fname, "EofEnds not a bool message");
      return false;
    }
    m_eofEnds = eofEnds;

    trace2("%s # m_eofEnds = %d", fname, (int) m_eofEnds);

    return true;
    
  }

  return false;
}

MJ_Parameter *MJ_ModuleDivaAudioFileIn::msgGetInput(const char * i)
{
  // trace("MJ_ModuleDivaAudio::msgGetInput", i);

  if(!strcmp(i, MJ_Strs::filename)) {
    return MJ_ModuleDivaAudioFile::msgGetInput(i);
  }
  else if(!strcmp(i, MJ_Strs::loop)) {
    return new MJ_ParameterBool(m_loop, MJ_Strs::Loop, MJ_Strs::loop);
  }
  else if(!strcmp(i, m_eofEndsStr)) {
    return new MJ_ParameterBool
      (m_eofEnds, "Stop at the end of the file", m_eofEndsStr);
  }

  return 0;
}
						    
QCString MJ_ModuleDivaAudioFileIn::msgGetInputName(int i)
{
  if(!i) return MJ_ModuleDivaAudioFile::msgGetInputName(i);

  QCString str;

  if(i==1)
    str = MJ_Strs::loop;
  else if(i==2) 
    str = m_eofEndsStr;

  return str;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_ModuleDivaAudioFileOut::MJ_ModuleDivaAudioFileOut()
  : m_sformat(DI_ASF_INT16),
    m_format(DI_AFF_WAVE),
    m_overWriteFiles(true)
{
  if(m_debug) trace("MJ_ModuleDivaAudioFileOut::MJ_ModuleDivaAudioFileOut");
}

MJ_ModuleDivaAudioFileOut::~MJ_ModuleDivaAudioFileOut()
{
  if(m_debug) trace("MJ_ModuleDivaAudioFileOut::~MJ_ModuleDivaAudioFileOut");
}

bool MJ_ModuleDivaAudioFileOut::dspAcceptIoChans(int *, int *)
{
  return false;
}

bool MJ_ModuleDivaAudioFileOut::dspPrepare()
{
  const char * fname = "MJ_ModuleDivaAudioFileOut::dspPrepare";

  if(m_debug) trace(fname);

  /* #ifdef HAVE_PCAPS
  char * capstr = cap_to_text(cap_get_proc(), 0);
  trace(fname, capstr);
  #endif */
  bool mod;
  QString fileName(m_overWriteFiles 
		   ? m_fileName : newWriteFile(m_fileName, & mod));

  QFileInfo  fi(fileName);
  
  uint chans = m_signalInput.size();

  if(!chans) {
    error(MJ_ERR_INTERNAL, fname, "NO channels");
    return false;
  }
  
  uint sr = dspSamplingRate();

  m_file.startWrite(fileName, 
		    DI_ASF_FLOAT32, 
		    sr,
		    chans, 
		    sr * 2, 
		    MJ_RTP_FILE_IO,
		    m_format,
		    m_sformat);
  
  if(!m_file.userPrepare()) {
    error2(MJ_ERR_INTERNAL, "%s # Open \"%s\"failed", fname, fileName.ascii());
    dspStop();
    return false;
  }

  // MJ_trace(fname, "Sampling rate = ", sr);
  // MJ_trace(fname, "frame size = ", samples);
  
  if(chans != m_file.file().outChannels() || !chans) {
    error(MJ_ERR_ARGUMENT, fname, "bad number of channels");
    dspStop();
    return false;
  }
  
  m_buffer.resize(chans * m_signalInput[0].samples().size());

  return true;
}

void MJ_ModuleDivaAudioFileOut::dspProcessPeriod()
{
  long frames = (long) m_signalInput[0].samples().size();
  
  for(uint i=0; i < m_signalInput.size(); i++) {
    // Fill the signal portion
    m_signalInput[i].samplesConst().exportSkipping
      (m_buffer.data(), 
       i,
       frames,
       m_signalInput.size());
  }

  m_file.writeFrames(m_buffer.data(), frames);
}
  

bool MJ_ModuleDivaAudioFileOut::msgProcess(const char *id, MJ_Message *m)
{
  const char * fname = "MJ_ModuleDivaAudioFileOut::msgProcess";

  MJ_cpf(m, fname);

  if(m_debug) trace2("%s # %s %p", fname, id, m);

  if(!strcmp(id, MJ_Strs::filename)) {
    return MJ_ModuleDivaAudioFile::msgProcess(id, m);
  }
  else if(!strcmp(id, MJ_Strs::bits)) {

    bool ok = true;
    int bits = m->toInt(&ok);

    if(!ok)
      return false;

    switch(bits) {
    case 16:
      m_sformat = DI_ASF_INT16;
      break;
    case 24:
      m_sformat = DI_ASF_INT24;
      break;
    case 32:
      m_sformat = DI_ASF_FLOAT32;
      break;
    }

    if(m_debug) trace2("%s # format = %d", fname, (int) m_sformat);
    
    return true;
  }
  else if(!strcmp(id, MJ_Strs::format)) {
    MJ_MessageString *ms = dynamic_cast<MJ_MessageString *> (m);

    if(!ms) {
      error(MJ_ERR_ARGUMENT, fname, "Not a string message");
      return false;
    }

    // trace(fname, (const char *) ms->string());
    
    int format = DI_AudioFile::determineFormat(ms->string());
    
    // trace2("%s file format = %d", fname, format);

    if(format >= 0) {
      m_format = (DI_AudioFileFormat) format;
      return true;
    }
    error2(MJ_ERR_ARGUMENT, "%s # Unknown file format \"%s\"",
	   fname, (const char *) ms->string());
  }
  else if(!strcmp(id, MJ_Strs::overwrite)) {
    bool ok = true;
    bool overWrite = m->toBool(&ok);
    
    if(!ok) {
      error(MJ_ERR_ARGUMENT, fname, "Not a bool message in overwrite");
      return false;
    }

    m_overWriteFiles = overWrite;

    return true;
  }

  return MJ_ModuleDivaAudioFileIn::msgProcess(id, m);
}

MJ_Parameter *MJ_ModuleDivaAudioFileOut::msgGetInput(const char *id)
{
  if(!strcmp(id, MJ_Strs::filename)) {
    return MJ_ModuleDivaAudioFile::msgGetInput(id);
  }
  else if(!strcmp(id, MJ_Strs::bits)) {

    int bits;

    switch(m_sformat) {
    case DI_ASF_INT16:
      bits = 16;
      break;
    case DI_ASF_INT24:
      bits = 24;
      break;
    case DI_ASF_FLOAT32:
      bits = 32;
      break;
    }

    MJ_ParameterInt *pi = new MJ_ParameterInt
      ("Sample bits", id, bits, 16, 32);
    
    pi->addAllowed(16);
    pi->addAllowed(24);
    pi->addAllowed(32);
    return pi;
  }
  else if(!strcmp(id, MJ_Strs::format)) {
    MJ_ParameterString *ms = 
      new MJ_ParameterString(DI_AudioFile::formatStr(m_format),
			     "File format",
			     id);
    ms->addAllowedStr("wav");
    ms->addAllowedStr("aiff");
    ms->addAllowedStr("aifc");
    ms->addAllowedStr("voc");
    return ms;
  }
  else if(!strcmp(id, MJ_Strs::overwrite)) {
    return new MJ_ParameterBool(m_overWriteFiles, 
				"Overwrite old files",
				id);
  }
  
  return MJ_ModuleDivaAudioFileIn::msgGetInput(id);
}

QCString MJ_ModuleDivaAudioFileOut::msgGetInputName(int i)
{
  QCString r;

  switch(i) {
  case 0:
    r = MJ_Strs::filename;
    break;
  case 1:
    r = MJ_Strs::bits;
    break;
  case 2:
    r = MJ_Strs::format;
    break;
  case 3:
    r = MJ_Strs::overwrite;
    break;
  }

  return r;
}

/** Generate a new filename if it seems we might overwrite an
    important file. */

QString MJ_ModuleDivaAudioFileOut::newWriteFile(const QString &name,
						bool *modified)
{
  QFileInfo  fi(name);

  if(!fi.exists()) {
    *modified = false;
    return name;
  }

  *modified = true;

  QString base = name;

  QString ext = fi.extension();

  if(!ext.isEmpty()) {
    base.truncate(base.length() - ext.length() - 1);
  }

  QString is;

  for(int i=1; fi.exists() && i < 200; i++) {
    is.sprintf("_%d.", i);
    QString tmp(base + is + ext);
    fi.setFile(tmp);
  }
  
  if(m_debug) trace("MJ_ModuleDivaAudioFileOut::newWriteFile", fi.filePath());

  return fi.filePath();
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


MJ_ModuleDivaAudioFileOutGui::MJ_ModuleDivaAudioFileOutGui
(MJ_RtControl *rtparent,
 MJ_ModuleDsp *module,
 QWidget      *parent,
 const char   *name)
  : MJ_ModuleDspGuiPostCard(rtparent, module, true,
			    100,
			    parent == 0  ? MJ_GUI_OWN_WINDOW : MJ_GUI_POSTCARD,
			    parent, name)
{}

MJ_ModuleDivaAudioFileOutGui::~MJ_ModuleDivaAudioFileOutGui() 
{}

MJ_ModuleDspGui *MJ_ModuleDivaAudioFileOutGui::createNew
(MJ_RtControl *rtparent,
 MJ_ModuleDsp *module,
 QWidget *parent, 
 const char *name)
{
  return new MJ_ModuleDivaAudioFileOutGui(rtparent, module, parent, name);
}

void MJ_ModuleDivaAudioFileOutGui::moduleModified(const char *id)
{
  const char * fname = "MJ_ModuleDivaAudioFileOutGui::moduleModified";

  trace(fname, id);

  if(!strcmp(id, MJ_Strs::format)) {

    lock();
    MJ_ParameterString *ps = (MJ_ParameterString *) 
      m_module->msgGetInput(MJ_Strs::format);
    unlock();

    int i = m_parameters.findParameter(id);

    if(i < 0) {
      error(MJ_ERR_ARGUMENT, 
	    "%s # Could not locate format parameter", fname);
      return;
    }
    else if(i >= (int) m_maxDisplayParameters)
      return;

    MJ_GuiStringSelector *ss = 
      dynamic_cast<MJ_GuiStringSelector *> (m_children.at(i));
    
    if(ss) {
      int index = ps->allowedStrs()->findIndex(ps->string());
      ss->box()->setCurrentItem(index);
    }
    delete ps;
  }
  else if(!strcmp(id, MJ_Strs::filename)) {
    
    // If the filename is modified, then update the file format as well

    lock();  
    MJ_Parameter *p = m_module->msgGetInput(MJ_Strs::filename);
    MJ_ParameterString *ps = dynamic_cast<MJ_ParameterString *> (p);
    unlock();
    
    MJ_cpf(ps, fname);
    
    const QString &str = ps->string();
    
    int dotPlace = str.findRev('.');
    int slashPlace = str.findRev('/');
    
    QString suffix;
    
    if(dotPlace > slashPlace && dotPlace > 0) {
      suffix = str.right(str.length() - 1 - dotPlace).lower();
    }

    trace2("%s suffix is %s", fname, (const char *) suffix);
    
    MJ_MessageString ms(suffix);

    m_module->msgProcess(MJ_Strs::format, &ms);
    m_module->satellite()->modified(MJ_Strs::format);

    delete p;
  }
}
