/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_diva_af_rec_play.h>

#include <mj_message_string.h>
#include <mj_parameters.h>
#include <mj_strs.h>

const char * MJ_ModuleDivaAudioFileRecPlay::m_modeIds[MODES] = {
  "bypass",
  "record",
  "play"
};

const char * MJ_ModuleDivaAudioFileRecPlay::m_modeLabels[MODES] = {
  "Bypass",
  "Record",
  "Play"
};

MJ_ModuleDivaAudioFileRecPlay::MJ_ModuleDivaAudioFileRecPlay()
  : m_mode(BYPASS)
{}

MJ_ModuleDivaAudioFileRecPlay::~MJ_ModuleDivaAudioFileRecPlay()
{}

bool MJ_ModuleDivaAudioFileRecPlay::dspAcceptIoChans(int *in, int *out)
{
  bool change1 = true;

  if(m_mode == RECORD)
    change1 = MJ_ModuleDivaAudioFileOut::dspAcceptIoChans(in, out);
  if(m_mode == PLAY)
    return MJ_ModuleDivaAudioFileIn::dspAcceptIoChans(in, out);
  
  bool change;
  
  if(*in != *out) {
    change = true;
    *out = *in;
  }
  else 
    change = false;

  return change && change1;
}

bool MJ_ModuleDivaAudioFileRecPlay::dspPrepare()
{
  if(m_mode == BYPASS) 
    return m_signalInput.size() == m_signalOutput.size();
  else if (m_mode == RECORD)
    return MJ_ModuleDivaAudioFileOut::dspPrepare();
  else if(m_mode == PLAY)
    return MJ_ModuleDivaAudioFileIn::dspPrepare();

  error(MJ_ERR_UNKNOWN, 
	"MJ_ModuleDivaAudioFileRecPlay::dspPrepare # Bad mode = %d", 
	(int) m_mode);

  return false;
}

void MJ_ModuleDivaAudioFileRecPlay::dspProcessPeriod()
{
  switch(m_mode) {
  case BYPASS:
    bypass();
    break;
  case RECORD:
    MJ_ModuleDivaAudioFileOut::dspProcessPeriod();
    bypass();
    break;
  case PLAY:
    MJ_ModuleDivaAudioFileIn::dspProcessPeriod();
    break;
  case MODES:
  default:
    break;
  }
}

bool MJ_ModuleDivaAudioFileRecPlay::dspStop()
{
  if(m_mode == RECORD)
    return MJ_ModuleDivaAudioFileOut::dspStop();
  else if(m_mode == PLAY)
    return MJ_ModuleDivaAudioFileIn::dspStop();

  return true;
}
  
bool MJ_ModuleDivaAudioFileRecPlay::msgProcess(const char *id, MJ_Message *m)
{
  const char * fname = "MJ_ModuleDivaAudioFileRecPlay::msgProcess";

  trace2("%s # \"%s\"", fname, id);

  if(strcmp(id, MJ_Strs::mode) == 0) {

    trace2("%s # Attempting to change mode", fname);

    MJ_MessageString *ms = dynamic_cast<MJ_MessageString *> (m);

    if(!ms) {
      return false;
    }

    int mode = determineMode(ms->string());
    
    if(mode < 0) {
      error2(MJ_ERR_ARGUMENT, 
	     "%s # Unsupported mode %s", fname, (const char *) ms->string());
      return false;
    }

    m_mode = (Mode) mode;

    trace2("MJ_ModuleDivaAudioFileRecPlay::msgProcess # Mode set to %d",
	   mode);

    return true;
  }

  return MJ_ModuleDivaAudioFileOut::msgProcess(id, m);
}

MJ_Parameter *MJ_ModuleDivaAudioFileRecPlay::msgGetInput(const char *id)
{
  if(strcmp(id, MJ_Strs::mode) == 0) {
    MJ_ParameterString *ms = 
      new MJ_ParameterString(m_modeIds[m_mode],
			     "Mode",
			     id);
    
    for(uint i=0; i < MODES; i++)
      ms->addAllowedStr(m_modeIds[i]);
    
    return ms;    
  }

  return MJ_ModuleDivaAudioFileOut::msgGetInput(id);
}

QCString MJ_ModuleDivaAudioFileRecPlay::msgGetInputName(int i)
{
  if(!i) return MJ_ModuleDivaAudioFile::msgGetInputName(i);

  QCString r;

  switch(i) {
  case 0:
    r = MJ_Strs::filename;
    break;
  case 1:
    r = MJ_Strs::mode;
    break;
  case 2:
    r = MJ_Strs::bits;
    break;
  case 3:
    r = MJ_Strs::format;
    break;
  case 4:
    r = MJ_Strs::overwrite;
    break;
  case 5:
    r = MJ_Strs::loop;
    break;
  case 6:
    r = m_eofEndsStr;
    break;
  }

  return r;
}

int MJ_ModuleDivaAudioFileRecPlay::determineMode(const char *str)
{
  return MJ_findStr(str, m_modeIds, MODES);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////



MJ_ModuleDivaAudioFileRecPlayGui::MJ_ModuleDivaAudioFileRecPlayGui
(MJ_RtControl *rtparent,
 MJ_ModuleDsp *module,
 QWidget      *parent,
 const char   *name)
  : MJ_ModuleDivaAudioFileOutGui(rtparent, module, parent, name)
{
  QPalette p = palette();

  m_palettes[MJ_ModuleDivaAudioFileRecPlay::BYPASS] = p;

  p.setColor(QColorGroup::Background, QColor(200, 50, 50));
  m_palettes[MJ_ModuleDivaAudioFileRecPlay::RECORD] = p;

  p.setColor(QColorGroup::Background, QColor(50, 170, 50));
  m_palettes[MJ_ModuleDivaAudioFileRecPlay::PLAY] = p;

  moduleModified(MJ_Strs::mode);
}
  
MJ_ModuleDivaAudioFileRecPlayGui::~MJ_ModuleDivaAudioFileRecPlayGui()
{}

MJ_ModuleDspGui *MJ_ModuleDivaAudioFileRecPlayGui::createNew
(MJ_RtControl *rtparent,
 MJ_ModuleDsp *module,
 QWidget *parent, 
 const char *name)
{
  return new MJ_ModuleDivaAudioFileRecPlayGui(rtparent, module, parent, name);
}

void MJ_ModuleDivaAudioFileRecPlayGui::moduleModified(const char *id)
{
  const char * fname = "MJ_ModuleDivaAudioFileRecPlayGui::moduleModified";

  trace2("%s # \"%s\"", fname, id);

  if(strcmp(id, MJ_Strs::mode) == 0) {

    trace2("%s # New mode");
    
    lock();  
    MJ_Parameter *p = m_module->msgGetInput(MJ_Strs::mode);
    MJ_ParameterString *ps = dynamic_cast<MJ_ParameterString *> (p);
    unlock();

    int mode = MJ_ModuleDivaAudioFileRecPlay::determineMode(ps->string());

    if(mode < 0) return;
    
    setPalette(m_palettes[mode]);

    trace2("%s # Palette set to %d", fname, mode);

    return;
  }

  MJ_ModuleDivaAudioFileOutGui::moduleModified(id);
}
