/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

// This needs to be here for SGI :-(
#include <di_types.h>

#include <mj_module_diva_audio.h>

#include <mj_module_dsp_row.h>

#include <mj_audio_conf.h> 
#include <mj_error.h>
#include <mj_math.h>
#include <mj_misc.h>
#include <mj_message_string.h>
#include <mj_parameters.h>
#include <mj_sleep.h>
#include <mj_strs.h>

#ifdef __linux__
#include <mj_rtc_hz.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

bool MJ_ModuleDivaAudioIn::m_debug = false;

MJ_ModuleDivaAudioIn::MJ_ModuleDivaAudioIn(DI_IoMode mode)
  : m_audioIo(0),
    m_mode(mode),
    m_totalMode(mode),
    m_totalChans(0),
    m_isOwner(true),
    m_friend(0),
    m_owner(0)
{
  const char * fname = "MJ_ModuleDivaAudioIn::MJ_ModuleDivaAudioIn";
  
  if(m_debug) trace(fname, mode);
  
  if(mode == DI_IO_INPUT) {
    dspSetOutChannels(1);
  } 
  else if (mode == DI_IO_OUTPUT) {
    dspSetInChannels(1);
  }
  else MJ_fatal(MJ_ERR_ARGUMENT, fname, "Bad mode");
}

MJ_ModuleDivaAudioIn::~MJ_ModuleDivaAudioIn()
{}

int MJ_ModuleDivaAudioIn::bufferLength()
{
#ifdef ALSA_IS_0_9
  // if(m_outWritten)
  
  if(m_mode & DI_IO_OUTPUT) {
    m_outAvail= m_audioIo->outEmptyFrames();

    if(m_outWritten && (m_totalMode & DI_IO_INPUT)) {
      m_inAvail = m_audioIo->inFilledFrames();
    }
  }
  else
    m_inAvail = m_audioIo->inFilledFrames();

  if(m_friend) {
    m_friend->m_inAvail = m_inAvail;
    m_friend->m_outAvail = m_outAvail;
  }
  if(m_owner) {
    m_owner->m_inAvail = m_inAvail;
    m_owner->m_outAvail = m_outAvail;
  }
#endif // ALSA_IS_0_9

  if(m_mode & DI_IO_OUTPUT) return m_audioIo->outFilledFrames();
  else if(m_mode & DI_IO_INPUT) return m_audioIo->inFilledFrames();
  return 0;
}

bool MJ_ModuleDivaAudioIn::conflicts(MJ_ModuleDsp *m)
{
  const char *fname = "MJ_ModuleDivaAudioIn::conflicts";

  if(m_debug) trace(fname);
  
  MJ_ModuleDivaAudioIn * mm = dynamic_cast<MJ_ModuleDivaAudioIn *> (m);
  
  if(!mm) return false;
  
  if(mm == this)
    MJ_fatal(MJ_ERR_INTERNAL, fname, "mm = this");

  if(mm->m_portName != m_portName) return false;

  if(m_debug) trace(fname, "1-WAY");
  
  // The resource is the same. Now check that we are not in conflict:
  
  /* uint sr1 = dspSamplingRate();
  uint sr2 = mm->dspSamplingRate();

  if((sr1 != sr2) && sr1 && sr2) {
    error2(MJ_ERR_INTERNAL, "%s # Unequal sampling rates: %d != %d" ,
	   fname, sr1, sr2);
    return true;
    }*/


  if(m_debug) trace(fname, "2-WAY");

  if(m_mode == mm->m_mode) {
    if(m_moduleDspDebug) 
      error(MJ_ERR_INTERNAL, fname, "Bad modes");
    return true;
  }

  if(m_debug) trace(fname, "HALF-WAY");

#ifdef DI_AUDIO_IS_SGI
  return false;
#else

  // if(m_friend || mm->m_friend) return true;

  // It is certain we are doing full duplex:
  m_totalMode = DI_IO_INPUT_OUTPUT;

  m_isOwner = true;
  m_friend = mm;
  m_owner = this;
  m_totalChans = MJ_max(chans(), mm->chans());

  mm->m_owner = this;
  mm->m_isOwner = false;

  if(m_debug) trace(fname, "EXIT OK");

  return false;
#endif // DI_AUDIO_IS_SGI
}

bool MJ_ModuleDivaAudioIn::dspPrepare()
{
  const char *fname = "MJ_ModuleDivaAudioIn::dspPrepare";

  const uint channels = chans();
  bool ok = true;

  if(!channels) MJ_fatal(MJ_ERR_INTERNAL, fname, "No channels");

  if(m_mode == DI_IO_INPUT)
    ok = m_buffer.resize(m_signalOutput.getFast(0).samples().size() * channels,
			 false);
  else
    ok = m_buffer.resize(m_signalInput.getFast(0).samples().size() * channels,
			 false);

  if(!ok) return false;
  
  m_buffer.setAll(0.0);

#ifdef ALSA_IS_0_9
  m_inAvail = 0;
  m_outAvail = 0;
  m_outWritten = 0;
  m_inRead = 0;
#endif

  return open();
}

bool MJ_ModuleDivaAudioIn::dspPrepareFast()
{
  bool ok(true);

  if(m_isOwner && m_audioIo)
    ok = m_audioIo->start();

  return ok;
}

void MJ_ModuleDivaAudioIn::dspProcessPeriod()
{
  const char *fname = "MJ_ModuleDivaAudioIn::dspProcessPeriod";

  const uint channels = m_signalOutput.size();

  // trace(fname, (int) m_mode);
  // trace(fname, channels);

  if(!channels || !m_audioIo) 
    MJ_fatal(MJ_ERR_INTERNAL, fname, "Audio In");
  
  uint frames = m_signalOutput[0].samples().size();
  
#ifdef ALSA_IS_0_9
      
  // bufferLength();
  
  // if(m_inRead + 3 < m_outWritten && m_outWritten) {
  if(m_inAvail < (int) frames || 
     (m_outWritten < 2 && m_mode & DI_IO_OUTPUT)) {
    trace2("%s # premature return m_inAvail = %d m_outWritten = %d", 
	     fname, (int) m_inAvail, (int) m_outWritten);
    return;
  }
  // }

  m_inRead++;

  /* while(m_inAvail < (int) framesx) {
     bufferLength();
     }*/
#else
  int avail = m_audioIo->inFilledFrames();
  if(avail < frames) {
    // MJ_trace("READFRA", avail);
    return;
  }
#endif // ALSA_IS_0_9

  // MJ_trace("READING", avail);
  
  m_audioIo->readFrames(m_buffer.data(), frames);

  // MJ_trace("READ", frames);
    
  for(uint i=0; i < m_signalOutput.size(); i++) {
    MJ_Sample *out = m_signalOutput[i].samples().data();
    
    MJ_Sample *buf = m_buffer.data() + i;

    for(uint j=0; j < frames; j++) {
      *out++ = buf[j * channels];
    }
  }
}

bool MJ_ModuleDivaAudioIn::dspStop()
{
  return close();
}

bool MJ_ModuleDivaAudioIn::msgProcess(const char * i, MJ_Message *m)
{
  const char *fname = "MJ_ModuleDivaAudioIn::msgProcess";

  if(m_debug) trace(fname, i);

  if(!strcmp(i, MJ_Strs::device)) {
    
    MJ_MessageString *ms = dynamic_cast<MJ_MessageString *> (m);

    if(ms) {
      m_portName = ms->string();
      if(m_debug) trace(fname, ms->string());
      return true;
    }
    else {
      error(MJ_ERR_ARGUMENT, fname, "Not a string message");
      return false;
    }
  }
  else if(m_debug)
    error(MJ_ERR_ARGUMENT, fname, "Bad id");
    
  return false;
}

MJ_Parameter *MJ_ModuleDivaAudioIn::msgGetInput(const char * i)
{
  const char *fname = "MJ_ModuleDivaAudioIn::msgGetInput";

  if(!i) MJ_fatal(MJ_ERR_NULL_ARGUMENT, fname);

  // trace(fname, i);

  if(!strcmp(i, MJ_Strs::device)) 
    return new MJ_ParameterString(m_portName, "Audio device",
				  MJ_Strs::device);
  return 0;
}

QCString MJ_ModuleDivaAudioIn::msgGetInputName(int i)
{
  return MJ_copyIfZero(MJ_Strs::device, i);
}

bool MJ_ModuleDivaAudioIn::open()
{
  const char *fname = "MJ_ModuleDivaAudioIn::open";

  if(m_debug) trace(fname, (int) m_isOwner);
  
  if(!m_isOwner) return true;
  
  m_audioIo = new DI_AudioPort;
  
  uint channels;
  DI_IoMode mode;  
  
  if(m_friend) {
    channels = m_totalChans;
    mode = m_totalMode;
  }
  else {
    channels = chans();
    mode = m_mode;
  }

  
  QCString name;

  if(!m_portName.isEmpty()) name = m_portName;
  else {
    QCString d;
    
    if(m_mode == DI_IO_INPUT)
      d = MJ_AudioConf::audioDevice(MJ_ADT_INPUT, 0);
    else if(m_mode == DI_IO_OUTPUT)
      d = MJ_AudioConf::audioDevice(MJ_ADT_OUTPUT, 0);

    if(m_debug) trace(fname, d);

    if(d.isEmpty())
      return false;

    name = d;
  }
  
  if(m_debug) {
    trace2("%s device = \"%s\" mode = %d sr = %d", 
	   fname, (const char *) name, mode, dspSamplingRate());
  }

  int sr = dspSamplingRate();

  int minBlock = dspFrameSize();

  if(minBlock < m_minimumBuffer / 2) {

    minBlock = m_minimumBuffer / 2;

    if(m_debug) trace2("%s # minBlock set to %d", fname, (int) minBlock);
  }

  /* First try with the real sampling rate. */
  if(m_audioIo->open(name,
		     mode, 
		     DI_ASF_SAMPLE,
		     minBlock * 5, 
		     sr, 
		     channels,
		     minBlock)) {

#ifdef ALSA_IS_0_9
    m_minimumBuffer = m_audioIo->minimumBlock();
#endif

    distributeObject();    
    return true;
  }

  /* Now a dirty hack to see if we can start it with any sampling rate. */
  if(!m_audioIo->open(name, 
		      mode, 
		      DI_ASF_SAMPLE,
		      minBlock * 4, 
		      0, 
		      channels,
		      minBlock)) {
    error2(MJ_ERR_INTERNAL, "%s # %s chans = %d # errno = \"%s\"", 
	   fname, (const char *) name, channels, strerror(errno));
    errno = 0;

    close();
    
    return false;
  }

  distributeObject();
  return true;
}

bool MJ_ModuleDivaAudioIn::close()
{
  const char *fname = "MJ_ModuleDivaAudioIn::close";
  
  if(m_debug) trace(fname);

  if(!m_isOwner || !m_audioIo) return true;

  

  delete m_audioIo;
  m_audioIo = 0;
  m_totalMode = DI_IO_NONE;
  m_totalChans = 0;

  unlinkObject();

  if(m_debug) trace(fname, "CLOSED");
  return true;
}

void MJ_ModuleDivaAudioIn::distributeObject()
{
  // const char *fname = "MJ_ModuleDivaAudioIn::distributeObject";
  
  if(m_friend) {
    m_friend->m_audioIo = m_audioIo;
    m_friend->m_totalMode = m_totalMode;

#ifdef ALSA_IS_0_9
    m_friend->m_minimumBuffer = m_minimumBuffer;
#endif

  }
}

void MJ_ModuleDivaAudioIn::unlinkObject()
{
  if(m_friend) {
    m_friend->m_audioIo = 0;
    m_friend->m_isOwner = true;
    m_friend->m_owner = 0;
    m_friend->m_totalMode = DI_IO_NONE;
    // Forget the friend:
    m_friend = 0;
  }
  m_totalMode = DI_IO_NONE;
}


uint MJ_ModuleDivaAudioIn::chans() const
{
  if(m_mode == DI_IO_INPUT) return m_signalOutput.size();
  return m_signalInput.size();
}


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


MJ_ModuleDivaAudioOut::~MJ_ModuleDivaAudioOut()
{}

void MJ_ModuleDivaAudioOut::dspProcessPeriod()
{
  const char *fname = "MJ_ModuleDivaAudioIn::dspProcessPeriod";

  const uint channels = m_signalInput.size();

  if(!channels || !m_audioIo)
    MJ_fatal(MJ_ERR_INTERNAL, fname, "Audio Out");
    
  // MJ_Sample sum = 0.0;

  uint samples = m_signalInput[0].samples().size();

  // trace2("%s # nsamples = %d", fname, samples);

  for(uint i=0; i < m_signalInput.size(); i++) {
    MJ_Sample *in = m_signalInput[i].samples().data();

    MJ_Sample *buf = m_buffer.data() + i;

    for(uint j=0; j < samples; j++) {
      buf[j * channels] = *in++;
      // sum += MJ_abs(m_buffer.get(j * channels + i));
    }
  }
  // printf(" %f ", sum);

#ifdef ALSA_IS_0_9

  /* char buf[64];
    sprintf(buf, "m_outAvail = %d",
	    m_outAvail);
	    trace(fname, buf); */
  
  while(!m_audioIo->outEmptyFrames()) {
    MJ_Sleep sleep;
#ifdef __linux__
    if(MJ_RtcHz::rtc())
      MJ_RtcHz::rtc()->waitTicks(2);
    else
      sleep.sleepMs(10);
#else
    sleep.sleepMs(2);
#endif
  }
    
  /* if(!m_outAvail)
       return; */
    
    /* trace(fname, "writeFrames");
       trace(fname, m_outAvail);*/
  m_outWritten++;
  if(m_friend) m_friend->m_outWritten = m_outWritten;
  else if(m_owner) m_owner->m_outWritten = m_outWritten;
#endif // ALSA_IS_0_9

  m_audioIo->writeFrames(m_buffer.data(), samples);

#ifdef ALSA_IS_0_9
  if(!m_audioIo->check()) {
    toCout();
    MJ_fatal(MJ_ERR_SYSTEM, fname);
  }
#endif // ALSA_IS_0_9
}
