/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_diva_midi.h>
#include <mj_module_dsp_row.h>

#include <mj_error.h>
#include <mj_global.h>
#include <mj_message_string.h>
#include <mj_messages.h>
#include <mj_parameters.h>
#include <mj_strs.h>

#include <errno.h>

static const char *MJ_MDM_inKeys[2] = {
  MJ_Strs::device,
  MJ_Strs::MIDI
};

bool MJ_ModuleDivaMidi::m_debug = false;

MJ_ModuleDivaMidi::MJ_ModuleDivaMidi(DI_IoMode mode)
  : m_midiIo(0),
    m_mode(mode),
    m_totalMode(mode),
    m_isOwner(true),
    m_owner(0)
{
  const char * fname = "MJ_ModuleDivaMidi::MJ_ModuleDivaMidi";
  
  if((mode != DI_IO_INPUT) && (mode != DI_IO_OUTPUT))
    MJ_fatal(MJ_ERR_ARGUMENT, fname, "Bad mode");
  
  m_portName = DI_defaultMidiDev();
}

MJ_ModuleDivaMidi::~MJ_ModuleDivaMidi()
{}

bool MJ_ModuleDivaMidi::conflicts(MJ_ModuleDsp *m)
{
  const char *fname = "MJ_ModuleDivaMidi::conflicts";

  trace(fname);

  MJ_ModuleDivaMidi * mm = dynamic_cast<MJ_ModuleDivaMidi *> (m);
  
  if(!mm) return false;
  
  if(mm->m_portName != m_portName) return false;
  
  /// If this module has an owner pointer distribute it to the other object.
  if(m_owner) {
    if(!mm->m_owner) {
      if(!mm->linkTo(m_owner))
	return true;
    }
    else if (mm->m_owner != m_owner) {
      error(MJ_ERR_INTERNAL, fname, "Objects differ");
      return true;
    }    
    return false;
  }

  /// If this module does not have an owner, get one for it:
  MJ_ModuleDivaMidi *owner = 0;
  if(mm->m_owner) owner = mm->m_owner;
  else {
    // Make the given object an owner object. 
    mm->makeOwner();
    owner = mm;
  }
  
  return !linkTo(owner);
}

bool MJ_ModuleDivaMidi::dspPrepare()
{
  if(!m_currentTime) return false;
  return open();
}

bool MJ_ModuleDivaMidi::dspPrepareFast()
{
  if(m_isOwner) m_midiIo->resetTiming();

  return true;
}

void MJ_ModuleDivaMidi::dspProcessPeriod()
{
  if(m_mode == DI_IO_INPUT) {
    while(m_midiIo->eventPending()) {
      m_midiIo->readEvent(& m_event.midiMsg());
      m_event.setTimeStamp((long) m_currentTime->stamp());

      if(m_debug) MJ_trace("MJ_ModuleDivaMidi::dspProcessPeriod # event");

      msgPushOutput(MJ_Strs::MIDI, & m_event);
    }
  }
  /* The output events get pushed out immediately, there is no need to
     do anything here. */
}

bool MJ_ModuleDivaMidi::dspStop()
{
  return close();
}

bool MJ_ModuleDivaMidi::msgProcess(const char * id, MJ_Message *m)
{
  const char *fname = "MJ_ModuleDivaMidi::msgProcess";

  int i = MJ_findStr(id, MJ_MDM_inKeys, 2);

  if(i == 1) {

    // Probably not real-time:
    if(!m_midiIo) return true;

    MJ_MessageMidiEvent *me = dynamic_cast<MJ_MessageMidiEvent *> (m);
    if(me) {
      m_midiIo->writeEvent(&me->midiMsg());
      return true;
    }
    else {
      error(MJ_ERR_ARGUMENT, fname, "Not a MIDI message");
      return false;
    }
  }
  else if(i == 0) {
    MJ_MessageString *ms = dynamic_cast<MJ_MessageString *> (m);
    if(ms) {
      m_portName = ms->string();
      return true;
    }
    else {
      error(MJ_ERR_ARGUMENT, fname, "Not a string message");
      return false;
    }
  }
  return false;
}

MJ_Parameter *MJ_ModuleDivaMidi::msgGetOutput(const char * i)
{
  if(m_mode == DI_IO_OUTPUT) 
    return 0;

  if(!strcmp(MJ_Strs::MIDI, i)) {
    return new MJ_ParameterMidi("MIDI output", i);
  }
  return 0;
}

QCString MJ_ModuleDivaMidi::msgGetOutputName(int i)
{
  QCString s;
  if(DI_IO_INPUT) 
    s = MJ_copyIfZero(MJ_Strs::MIDI, i);
  return s;
}

MJ_Parameter *MJ_ModuleDivaMidi::msgGetInput(const char * id)
{
  // trace("MJ_ModuleDivaMidi::msgGetInput", i);

  int i = MJ_findStr(id, MJ_MDM_inKeys, 2);
  // int tmp;

  /* if(m_mode == DI_IO_INPUT)
    tmp = i + 1;
  else
  tmp = i; */

  if(i == 1) {
    MJ_ParameterMidi *pm = 
      new MJ_ParameterMidi("MIDI events", MJ_Strs::MIDI);
    return pm;
  }
  else if(i == 0) {
    MJ_ParameterString *ps = new MJ_ParameterString();
    ps->setString(m_portName);
    ps->setName("MIDI device");
    ps->setId(id);
    return ps;
  }

  return 0;
}

QCString MJ_ModuleDivaMidi::msgGetInputName(int i)
{
  QCString s;

  if(i == 0) s = MJ_Strs::device;
  if(i == 1) s = MJ_Strs::MIDI;

  return s;
}

bool MJ_ModuleDivaMidi::open()
{
  const char *fname = "MJ_ModuleDivaMidi::open";

  trace(fname, m_portName);
  
  if(!m_isOwner) return true;
  
  m_midiIo = new DI_MidiIo;

  if(!m_midiIo->open(m_portName, m_totalMode)) {
    error2(MJ_ERR_INTERNAL, "%s # device = \"%s\" error = \"%s\"", 
	   fname, (const char *) m_portName, strerror(errno));
    return false;
  }

  distributeObject();
  
  return true;
}

bool MJ_ModuleDivaMidi::close()
{
  const char *fname = "MJ_ModuleDivaMidi::close";
  
  trace(fname);

  if(!m_isOwner) return true;

  delete m_midiIo;
  m_midiIo = 0;

  unlinkObject();
  return true;
}

void MJ_ModuleDivaMidi::distributeObject()
{
  const char *fname = "MJ_ModuleDivaMidi::distributeObject";

  for(uint i=0; i < m_ptrs.count(); i++) {
    MJ_ModuleDivaMidi *mm = (MJ_ModuleDivaMidi *) m_ptrs.at(i);

    if(mm->m_owner != this || mm->m_isOwner) {
      MJ_fatal(MJ_ERR_INTERNAL, fname);
    }

    mm->m_midiIo = m_midiIo;
  }
}

void MJ_ModuleDivaMidi::unlinkObject()
{
  while(m_ptrs.count()) {
    MJ_ModuleDivaMidi *mm = (MJ_ModuleDivaMidi *) m_ptrs.first();

    mm->m_midiIo = 0;
    mm->m_isOwner = true;
    mm->m_owner = 0;

    m_ptrs.removeFirst();
  }
}

void MJ_ModuleDivaMidi::makeOwner()
{
  m_owner = this;
  m_isOwner = true;
  m_totalMode = m_mode;
}

bool MJ_ModuleDivaMidi::linkTo(MJ_ModuleDivaMidi *m)
{
  const char *fname = "MJ_ModuleDivaMidi::linkTo";
  
  /* Two modules cannot read from the same source. Evil things would
     happen. */
  if((m->m_totalMode & DI_IO_INPUT) &&
     (m_mode & DI_IO_INPUT)) {
    error(MJ_ERR_ARGUMENT, fname, "Conflicting modes");
    return false;
  }

  m->m_totalMode = (DI_IoMode) ((int) m->m_totalMode | (int) m_mode);
  m_isOwner = false;
  m->m_ptrs.append(this);
  m_owner = m;
  
  return true;
}
