/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_compressor2.h>

#include <mj_error.h> 
#include <mj_mapper.h> 
#include <mj_math.h>
#include <mj_parameters.h> 
#include <mj_strs.h>

#include <mj_arraytmpl.C>

static bool MJ_compressorDebug = false;

const char * MJ_ParamsCompressor2Keys[MJ_PARCOMP2_SIZEOF] = {
  MJ_Strs::threshold,
  MJ_Strs::ratio,
  MJ_Strs::attack,
  "hold",
  "release",
  "sync",
  "extended"
};


MJ_Sample MJ_ChannelCompressor::newGain(MJ_Sample energy,
					MJ_Sample thresholdDb, 
					MJ_Sample ratio,
					uint attackTime,
					uint holdTime,
					uint decayTime)
{
  // if(MJ_compressorDebug) MJ_trace("MJ_ChannelCompressor::newGain");

  MJ_Sample rms = m_rmsFilter.putGetSample(energy);
  MJ_Sample rmsDb = MJ_squareToDb(rms);
  
  if(rmsDb > m_currentPeak && rmsDb > thresholdDb) {

    // Kick in attack:
    
    m_state = MJ_COMPRESSOR_ATTACK;
    MJ_Sample compressDb = (thresholdDb - rmsDb) * (1.0 - 1.0 / ratio);
    MJ_Sample gain = MJ_dbToAbs(compressDb);
    m_ramp.setTarget(attackTime, gain, MJ_RAMP_EXPONENTIAL);
    m_secondaryPeak = m_currentPeak;
    m_currentPeak = rmsDb;
  }
  else if(m_state==MJ_COMPRESSOR_HOLD || m_state==MJ_COMPRESSOR_ATTACK){
    // Keep track of the secondary peaks:
    
    if(rmsDb > m_secondaryPeak) {
      m_secondaryPeak = rmsDb;
    }
  }
  else if(m_state == MJ_COMPRESSOR_DECAY) {
    if(rmsDb > m_secondaryPeak) {
      m_secondaryPeak = rmsDb;

      MJ_Sample compressDb = (thresholdDb - rmsDb) * (1.0 - 1.0 / ratio);
      MJ_Sample gain = MJ_dbToAbs(compressDb);
      m_ramp.setTarget(decayTime, gain, MJ_RAMP_EXPONENTIAL);
    }
  }

  // Calculate output:
  m_ramp.advanceExponential();
  MJ_Sample result = m_ramp.get();

  // Now change state, if necessary:
  if(!m_ramp.isActive()) {

    // MJ_trace2("state was %d, gain = %f", (int) m_state, result);

    if(m_state == MJ_COMPRESSOR_IDLE) {
      m_currentPeak = thresholdDb;
      m_secondaryPeak = thresholdDb;
    }
    else if(m_state == MJ_COMPRESSOR_ATTACK) {
      m_ramp.setTarget(holdTime, m_ramp.get(), MJ_RAMP_EXPONENTIAL);
      m_state = MJ_COMPRESSOR_HOLD;
    }
    else if(m_state == MJ_COMPRESSOR_HOLD) {
      m_state = MJ_COMPRESSOR_DECAY;
      
      if(m_secondaryPeak < thresholdDb) m_secondaryPeak = thresholdDb;

      MJ_Sample compressDb = (thresholdDb - m_secondaryPeak) * 
	(1.0 - 1.0 / ratio);
      MJ_Sample gain = MJ_dbToAbs(compressDb);
      m_ramp.setTarget(decayTime, gain, MJ_RAMP_EXPONENTIAL);
      m_secondaryPeak = thresholdDb;
    }
    else {
      // m_state is known to be MJ_COMPRESSOR_DECAY

      if(m_ramp.target() == 1.0) {

	m_state = MJ_COMPRESSOR_IDLE;
	m_currentPeak = thresholdDb;
	m_secondaryPeak = thresholdDb;
      }
      else {
	
	if(m_secondaryPeak < thresholdDb) m_secondaryPeak = thresholdDb;

	MJ_Sample compressDb = (thresholdDb - m_secondaryPeak) * 
	  (1.0 - 1.0 / ratio);
	MJ_Sample gain = MJ_dbToAbs(compressDb);
	m_ramp.setTarget(decayTime, gain, MJ_RAMP_EXPONENTIAL);
	m_secondaryPeak = thresholdDb;
      }
    }

    // MJ_trace2("state now %d target = %f", (int) m_state, m_ramp.target());
  }
  return result;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


MJ_ModuleCompressor2::MJ_ModuleCompressor2()
  : m_threshold(-10.0),
    m_ratio(2.0),
    m_attackMs(5),
    m_holdMs(10),
    m_releaseMs(30),
    m_syncChannels(false),
    m_extendedRanges(false)
{
  dspSetInChannels(1);
  dspSetOutChannels(1);
}

MJ_ModuleCompressor2::~MJ_ModuleCompressor2()
{}
  
bool MJ_ModuleCompressor2::dspPrepare()
{
  uint inchans = dspInChannels();

  if(!inchans || inchans != (uint) dspOutChannels()) {
    error(MJ_ERR_INTERNAL, "MJ_ModuleCompressor2::dspPrepare");
    return false;
  }

  uint samples = m_signalInput.get(0).samples().size();

  MJ_Allocator *allocator = mjDspAllocator();

  m_compressVect.customResize(allocator, samples);
  m_compressVect.setAll(0.0);

  m_compressGains.resize(inchans);
  m_channelCompressors.resize(inchans);

  for(uint i=0; i < inchans; i++) {
    m_channelCompressors[i].prepare(m_threshold);
  }

  redesignFilters();

  calculateTimes();

  return true;
}
 
void MJ_ModuleCompressor2::dspProcessPeriod()
{
  if(m_syncChannels) processSync();
  else processNonSync();
}

bool MJ_ModuleCompressor2::dspStop()
{  
  MJ_Allocator *allocator = mjDspAllocator();

  m_compressVect.customResize(allocator, 0);
  return true;
}

bool MJ_ModuleCompressor2::msgProcess(const char *id, MJ_Message *m)
{
  trace3(MJ_compressorDebug, "MJ_ModuleCompressor2::msgProcess # %s %p",
	 id, m);

  MJ_ParamsCompressor2 pc = (MJ_ParamsCompressor2)
    MJ_findStr(id, MJ_ParamsCompressor2Keys, MJ_PARCOMP2_SIZEOF);
  
  if((int) pc == -1) {
    error(MJ_ERR_ARGUMENT, "MJ_ModuleCompressor2::msgProcess");
    return false;
  }

  bool ok(true);


  if(pc == MJ_PARCOMP2_SYNC) {
    bool sync = m->toBool(&ok);

    if(!ok) return false;

    m_syncChannels = sync;

    return true;
  }
 
  if(pc == MJ_PARCOMP2_EXTENDED) {
    bool extend = m->toBool(&ok);

    if(!ok) return false;

    m_extendedRanges = extend;

    return true;
  }

  MJ_Sample val = m->toDouble(&ok);

  if(!ok) {
    return false;
  }

  switch (pc) {
  case MJ_PARCOMP2_THRESHOLD:
    m_threshold = MJ_clampS(val, -80, 10);
    break;
  case MJ_PARCOMP2_RATIO:
    m_ratio = MJ_clampS(val, 0.1, 1000);
    break;
  case MJ_PARCOMP2_ATTACK:
    m_attackMs = MJ_clampS(val, 0.01, 1000);
    redesignFilters();
    break;
  case MJ_PARCOMP2_HOLD:
    m_holdMs = MJ_clampS(val, 0.01, 1000);
    break;
  case MJ_PARCOMP2_RELEASE:
    m_releaseMs = MJ_clampS(val, 0.01, 1000);
    break;
  case MJ_PARCOMP2_SYNC:
  case MJ_PARCOMP2_EXTENDED:
  case MJ_PARCOMP2_SIZEOF:
  default:
    break;
  }

  calculateTimes();
  
  return true;
}

MJ_Parameter * MJ_ModuleCompressor2::msgGetInput(const char *id)
{
  const double maxTime = m_extendedRanges ? 10000.0 : 300.0;
    
  MJ_ParamsCompressor2 pc = (MJ_ParamsCompressor2)
    MJ_findStr(id, MJ_ParamsCompressor2Keys, MJ_PARCOMP2_SIZEOF);

  if(pc == MJ_PARCOMP2_THRESHOLD) {
    return new MJ_ParameterDouble
      (MJ_Strs::Threshold, MJ_Strs::threshold, m_threshold, -80.0, 10, 
       MJ_UNIT_DECIBEL);
  }
  else if(pc == MJ_PARCOMP2_RATIO) {
    MJ_ParameterDouble *pd = new MJ_ParameterDouble
      (MJ_Strs::Ratio, MJ_Strs::ratio, m_ratio, 1.0, 50);
    pd->setMapper(new MJ_MapperLog);
    return pd;
  }
  else if(pc == MJ_PARCOMP2_ATTACK) {
    MJ_ParameterDouble *pd = new MJ_ParameterDouble
      ("Attack time", id, m_attackMs, 0.1, maxTime, 
       MJ_UNIT_MILLISECONDS);
    pd->setMapper(new MJ_MapperLog);
    return pd;
  }
  else if(pc == MJ_PARCOMP2_HOLD) {
    MJ_ParameterDouble *pd = new MJ_ParameterDouble
      ("Hold time", id, m_holdMs, 0.01, maxTime, 
       MJ_UNIT_MILLISECONDS);
    pd->setMapper(new MJ_MapperLog);
    return pd;
  }
  else if(pc == MJ_PARCOMP2_RELEASE) {
    MJ_ParameterDouble *pd = new MJ_ParameterDouble
      ("Release time", id, m_releaseMs, 0.1, maxTime, 
       MJ_UNIT_MILLISECONDS);
    pd->setMapper(new MJ_MapperLog);
    return pd;
  }
  else if(pc == MJ_PARCOMP2_SYNC) {
    return new MJ_ParameterBool
      (m_syncChannels, "Sync channels", id);
  }
  else if(pc == MJ_PARCOMP2_EXTENDED) {
    MJ_ParameterBool * pb = 
      new MJ_ParameterBool(m_extendedRanges, "Extended ranges", id);
    
    pb->setFlags(MJ_PFLAG_AFFECTS);

    return pb;
  }
  return 0;
}

QCString MJ_ModuleCompressor2::msgGetInputName(int i)
{
  QCString s;
  if(i < MJ_PARCOMP2_SIZEOF) s = MJ_ParamsCompressor2Keys[i];
  return s;
}

void MJ_ModuleCompressor2::processSync()
{
  uint i,j;
  const uint chans   = m_signalInput.size();
  const uint samples = m_signalInput.get(0).samples().size();

  if(m_compressVect.size() < samples) {
    MJ_Allocator *allocator = mjDspAllocator();
    m_compressVect.customResize(allocator, samples);
  }

  // Get the peaks of all input channels:
  m_compressVect.setAll(0.0);
  MJ_Sample *tmpBuf = m_compressVect.data();

  for(i=0; i < chans; i++) {

    const MJ_Sample *inSamples = m_signalInput.getFast(i).samples().data();

    for(j=0; j < samples; j++) {
      MJ_Sample tmp = *inSamples++;
      tmp *= tmp;
      if(tmpBuf[j] < tmp) tmpBuf[j] = tmp;
    }
  }

  // Now get compression gains for each sample:  
  MJ_ChannelCompressor *cc = & m_channelCompressors.get(0);

  for(j=0; j < samples; j++) {
    tmpBuf[j] = cc->newGain(tmpBuf[j], 
			    m_threshold, 
			    m_ratio,
			    m_attackSamples, 
			    m_holdSamples, 
			    m_releaseSamples);
  }

  m_compressGains.setAll(tmpBuf[samples - 1]);

  // Now create the output:  
  for(i=0; i < chans; i++) {
    m_signalInput.getFast(i).samples().mul
      (m_compressVect, &m_signalOutput.getFast(i).samples());
  }
}

void MJ_ModuleCompressor2::processNonSync()
{
  const uint chans   = m_signalInput.size();
  const uint samples = m_signalInput.get(0).samples().size();

  for(uint i=0; i < chans; i++) {

  
    const MJ_Sample *inSamples = m_signalInput.getFast(i).samples().data();
    MJ_Sample *outSamples = m_signalOutput.getFast(i).samples().data();
        
    MJ_ChannelCompressor *cc = & m_channelCompressors.get(i);

    MJ_Sample gain = 1.0;

    for(uint j=0; j < samples; j++) {
      /* trace3(MJ_compressorDebug, 
	     "MJ_ModuleCompressor2::processNonSync # %d %d",
	     i, j); */
      
      MJ_Sample tmp = inSamples[j];
      gain = cc->newGain(tmp * tmp, 
			 m_threshold, 
			 m_ratio,
			 m_attackSamples, 
			 m_holdSamples, 
			 m_releaseSamples);
      outSamples[j] = tmp * gain;
    }
    m_compressGains.getFast(i) = gain;
  }

  /* trace3(MJ_compressorDebug, 
     "MJ_ModuleCompressor2::processNonSync # EXIT"); */
}

void MJ_ModuleCompressor2::redesignFilters()
{
  MJ_Sample sr = dspSamplingRate();

  MJ_Sample rate = (2000.0 / m_attackMs) / sr; // 1000 * 2 to scale between 0,1

  for(uint i =0; i < m_channelCompressors.size(); i++) {
    m_channelCompressors.getFast(i).rmsFilter().designLowPass(rate);
  }
}

void MJ_ModuleCompressor2::calculateTimes()
{
  MJ_Sample sr = (MJ_Sample) dspSamplingRate();
  
  m_attackSamples  = MJ_max((uint) (sr * m_attackMs * 0.001 + 0.5), (uint) 1);
  m_holdSamples    = MJ_max((uint) (sr * m_holdMs * 0.001 + 0.5), (uint) 1);
  m_releaseSamples = MJ_max((uint) (sr * m_releaseMs * 0.001 + 0.5), (uint) 1);
}

template class MJ_Array<MJ_ChannelCompressor>;
