/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_compressor3.h>

#include <mj_error.h> 
#include <mj_mapper.h> 
#include <mj_math.h>
#include <mj_parameters.h> 
#include <mj_strs.h>

#include <mj_arraytmpl.C>

const char * MJ_ParamsCompressor3Keys[MJ_PARCOMP3_SIZEOF] = {
  MJ_Strs::threshold,
  MJ_Strs::ratio,
  MJ_Strs::attack,
  "release",
  "sync"
};


MJ_Sample MJ_ChannelCompressor3::newGain(MJ_Sample energy,
					 MJ_Sample thresholdSquare, 
					 MJ_Sample thresholdDb, 
					 MJ_Sample ratio)
{
  if(energy < thresholdSquare) energy = thresholdSquare;

  double a = m_attackFilter.putGetSample(energy);
  double levelSquare = m_releaseFilter.putGetSample(energy);

  // MJ_trace("MJ_ChannelCompressor3::newGain 2");

  if(a > levelSquare) {
    m_releaseFilter.resetMemory(a);
    levelSquare = a;
  }

  MJ_Sample levelDb = MJ_squareToDb(levelSquare);

  MJ_Sample compressDb = (thresholdDb - levelDb) * (1.0 - 1.0 / ratio);

  // MJ_Sample compressDb = (thresholdDb - levelDb) * (1.0 - 1.0 / ratio);

  MJ_Sample tmp = MJ_dbToAbs(compressDb);
  
  // MJ_trace2("MJ_ChannelCompressor3::newGain 4 %f", tmp);

  return tmp;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

bool MJ_ModuleCompressor3::m_debug = false;

MJ_ModuleCompressor3::MJ_ModuleCompressor3()
  : m_thresholdDb(-10.0),
    m_ratio(2.0),
    m_attackMs(5),
    m_releaseMs(30),
    m_syncChannels(false)
{
  MJ_Sample tmp = MJ_dbToAbs(m_thresholdDb);
  m_thresholdSquare = tmp * tmp;
  dspSetInChannels(1);
  dspSetOutChannels(1);
}

MJ_ModuleCompressor3::~MJ_ModuleCompressor3()
{}
  
bool MJ_ModuleCompressor3::dspPrepare()
{
  uint inchans = dspInChannels();

  if(!inchans || inchans != (uint) dspOutChannels()) {
    error(MJ_ERR_INTERNAL, "MJ_ModuleCompressor3::dspPrepare");
    return false;
  }

  uint samples = m_signalInput.get(0).samples().size();

  MJ_Allocator *allocator = mjDspAllocator();

  m_compressVect.customResize(allocator, samples);
  m_compressVect.setAll(0.0);

  m_compressGains.resize(inchans);
  m_channelCompressors.resize(inchans);

  redesignFilters();

  MJ_Sample tmp = MJ_dbToSquare(m_thresholdDb);
  m_thresholdSquare = tmp * tmp;

  for(uint i=0; i < inchans; i++) {
    m_channelCompressors[i].prepare(m_thresholdSquare);
  }

  return true;
}
 
void MJ_ModuleCompressor3::dspProcessPeriod()
{
  if(m_syncChannels) processSync();
  else processNonSync();
}

bool MJ_ModuleCompressor3::dspStop()
{  
  MJ_Allocator *allocator = mjDspAllocator();

  m_compressVect.customResize(allocator, 0);
  return true;
}

bool MJ_ModuleCompressor3::msgProcess(const char *id, MJ_Message *m)
{
  trace3(m_debug, "MJ_ModuleCompressor3::msgProcess # %s %p",
	 id, m);

  MJ_ParamsCompressor3 pc = (MJ_ParamsCompressor3)
    MJ_findStr(id, MJ_ParamsCompressor3Keys, MJ_PARCOMP3_SIZEOF);
  
  if((int) pc == -1) {
    error(MJ_ERR_ARGUMENT, "MJ_ModuleCompressor3::msgProcess");
    return false;
  }

  bool ok(true);


  if(pc == MJ_PARCOMP3_SYNC) {
    bool sync = m->toBool(&ok);

    if(!ok) return false;

    m_syncChannels = sync;

    return true;
  }

  MJ_Sample val = m->toDouble(&ok);

  if(!ok) {
    return false;
  }

  MJ_Sample tmp;

  switch (pc) {
  case MJ_PARCOMP3_THRESHOLD:
    m_thresholdDb = MJ_clampS(val, -80, 10);
    tmp = MJ_dbToAbs(m_thresholdDb);
    m_thresholdSquare = tmp * tmp;
    break;
  case MJ_PARCOMP3_RATIO:
    m_ratio = MJ_clampS(val, 0.1, 1000);
    break;
  case MJ_PARCOMP3_ATTACK:
    m_attackMs = MJ_clampS(val, 0.01, 1000);
    redesignFilters();
    break;
  case MJ_PARCOMP3_RELEASE:
    m_releaseMs = MJ_clampS(val, 0.01, 1000);
    redesignFilters();
    break;

    // The next two are there just to stop g++ from complaining:
  case MJ_PARCOMP3_SYNC: 
  case MJ_PARCOMP3_SIZEOF:
  default:
    break;
  }
  return true;
}

MJ_Parameter * MJ_ModuleCompressor3::msgGetInput(const char *id)
{
  MJ_ParamsCompressor3 pc = (MJ_ParamsCompressor3)
    MJ_findStr(id, MJ_ParamsCompressor3Keys, MJ_PARCOMP3_SIZEOF);

  if(pc == MJ_PARCOMP3_THRESHOLD) {
    return new MJ_ParameterDouble
      (MJ_Strs::Threshold, MJ_Strs::threshold, m_thresholdDb, -80.0, 10, 
       MJ_UNIT_DECIBEL);
  }
  else if(pc == MJ_PARCOMP3_RATIO) {
    MJ_ParameterDouble *pd = new MJ_ParameterDouble
      (MJ_Strs::Ratio, MJ_Strs::ratio, m_ratio, 1.0, 50);
    pd->setMapper(new MJ_MapperLog);
    return pd;
  }
  else if(pc == MJ_PARCOMP3_ATTACK) {
    MJ_ParameterDouble *pd = new MJ_ParameterDouble
      ("Attack time", id, m_attackMs, 0.1, 300, 
       MJ_UNIT_MILLISECONDS);
    pd->setMapper(new MJ_MapperLog);
    return pd;
  }
  else if(pc == MJ_PARCOMP3_RELEASE) {
    MJ_ParameterDouble *pd = new MJ_ParameterDouble
      ("Release time", id, m_releaseMs, 0.1, 300, 
       MJ_UNIT_MILLISECONDS);
    pd->setMapper(new MJ_MapperLog);
    return pd;
  }
  else if(pc == MJ_PARCOMP3_SYNC) {
    return new MJ_ParameterBool
      (m_syncChannels, "Sync channels", id);
  }
  return 0;
}

QCString MJ_ModuleCompressor3::msgGetInputName(int i)
{
  QCString s;
  if(i < MJ_PARCOMP3_SIZEOF) s = MJ_ParamsCompressor3Keys[i];
  return s;
}

void MJ_ModuleCompressor3::processSync()
{
  uint i,j;
  const uint chans   = m_signalInput.size();
  const uint samples = m_signalInput.get(0).samples().size();

  if(m_compressVect.size() < samples) {
    MJ_Allocator *allocator = mjDspAllocator();
    m_compressVect.customResize(allocator, samples);
  }

  // Get the peaks of all input channels:
  m_compressVect.setAll(0.0);
  MJ_Sample *tmpBuf = m_compressVect.data();

  for(i=0; i < chans; i++) {

    const MJ_Sample *inSamples = m_signalInput.getFast(i).samples().data();

    for(j=0; j < samples; j++) {
      MJ_Sample tmp = *inSamples++;
      tmp *= tmp;
      if(tmpBuf[j] < tmp) tmpBuf[j] = tmp;
    }
  }

  // Now get compression gains for each sample:  
  MJ_ChannelCompressor3 *cc = & m_channelCompressors[0];

  for(j=0; j < samples; j++) {
    tmpBuf[j] = cc->newGain(tmpBuf[j], 
			    m_thresholdSquare, 
			    m_thresholdDb, 
			    m_ratio);
  }
  
  m_compressGains.setAll(tmpBuf[samples - 1]);

  // Now create the output:  
  for(i=0; i < chans; i++) {
    m_signalInput.getFast(i).samples().mul
      (m_compressVect, &m_signalOutput.getFast(i).samples());
  }
}

void MJ_ModuleCompressor3::processNonSync()
{
  // trace3(m_debug, "MJ_ModuleCompressor3::processNonSync # ENTER");

  const uint chans   = m_signalInput.size();
  const uint samples = m_signalInput.get(0).samples().size();

  for(uint i=0; i < chans; i++) {
  
    const MJ_Sample *inSamples = m_signalInput.getFast(i).samples().data();
    MJ_Sample *outSamples = m_signalOutput.getFast(i).samples().data();
        
    MJ_ChannelCompressor3 *cc = & m_channelCompressors[i];

    MJ_Sample gain = 1.0;

    for(uint j = 0; j < samples; j++) {
      
      MJ_Sample tmp = inSamples[j];

      /* trace3(m_debug, 
	     "MJ_ModuleCompressor3::processNonSync # %d %d %f",
	     i, j, tmp);*/

      gain = cc->newGain(tmp * tmp, 
			 m_thresholdSquare, 
			 m_thresholdDb, 
			 m_ratio);

      outSamples[j] = tmp * gain;

    }
    m_compressGains.getFast(i) = gain;
  }

  // trace3(m_debug, "MJ_ModuleCompressor3::processNonSync # EXIT");
}

void MJ_ModuleCompressor3::redesignFilters()
{
  MJ_Sample sr = dspSamplingRate();

  MJ_Sample attackFreq = 
    (2000.0 / m_attackMs) / sr; // 1000 * 2 to scale between 0,1
  MJ_Sample releaseFreq = 
    (2000.0 / m_releaseMs) / sr; // 1000 * 2 to scale between 0,1

  for(uint i =0; i < m_channelCompressors.size(); i++) {
    m_channelCompressors[i].attackFilter().designLowPass(attackFreq);
    m_channelCompressors[i].releaseFilter().designLowPass(releaseFreq);
  }
}

// template class MJ_Array<MJ_ChannelCompressor3>;
