/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_limiter.h>

#include <mj_error.h> 
#include <mj_mapper.h> 
#include <mj_math.h>
#include <mj_parameters.h> 
#include <mj_strs.h>

#include <mj_arraytmpl.C>

const char * MJ_ParamsLimiterKeys[MJ_PARLIM_SIZEOF] = {
  MJ_Strs::threshold,
  MJ_Strs::attack,
  "release"
};

MJ_Sample MJ_ChannelLimiter::putGet(MJ_Sample insample,
				    MJ_Sample thresholdLog,
				    uint attackTime,
				    uint releaseTime)
{
  // Store pure data to delay buffer:
  m_buffer.put(insample);

  // Store decibels to delay buffer:
  insample = MJ_abs(insample);

  if(insample < 1e-10) insample = 1e-10;

  MJ_Sample insampleLog = MJ_max(MJ_log(insample), thresholdLog);

  m_logBuffer.put(insampleLog);

  int design = 0;

  // See if the new input value indicated new peak:
  if(m_untilPeak) 
  {
    MJ_Sample requiredGain = thresholdLog - insampleLog;
    MJ_Sample ats = (MJ_Sample) attackTime - 1;
    MJ_Sample tmp2 = m_step * ats + m_gain;

    if(tmp2 > requiredGain) {
      // MJ_Sample oldStep = m_step;
      m_step = (requiredGain - m_gain) / ats;
      m_untilPeak = attackTime;

      /* MJ_trace2("Good redesign %f %f -> %f %f # %f", m_gain, oldStep, 
	 m_step, insampleLog, requiredGain);  */
      design = 1;
    }
  }

  if(m_untilPeak) m_untilPeak--;
  
  // Seek new attack path 
  // if(true) {
  if(!design) {
  // if(!m_untilPeak) {

    m_step = 0.0;
    
    for(uint i=1; i <= attackTime; i++) {
      MJ_Sample tmp = m_logBuffer.getNewest(attackTime - i);
      MJ_Sample requiredGain = thresholdLog - tmp;
      MJ_Sample is = (MJ_Sample) i;
      MJ_Sample tmp2 = m_step * is + m_gain;
      
      if(tmp2 > requiredGain) {
	m_step = (requiredGain - m_gain) / is;
	m_untilPeak = i - 1;
	design = 2;
	/* MJ_trace2("Poor redesign %f %d # %f %f", 
	   m_gain + m_step * (m_untilPeak + 1), 
	   m_untilPeak, tmp, requiredGain); */
	
      }
    }
  }

  MJ_Sample delayedSample = m_buffer.getNewest(attackTime-1);
  m_level.put(m_logBuffer.getNewest(attackTime-1), thresholdLog, releaseTime);

  if(!design) {
    MJ_Sample tmp = m_level.peak();
    MJ_Sample requiredGain = thresholdLog - tmp;
    m_step = (requiredGain - m_gain) / (MJ_Sample) releaseTime;
    design = 3;
    // m_mul = MJ_pow(requiredGain / m_gain, 1.0 / (MJ_Sample) releaseTime);
  }

  m_gain += m_step;

  //      printf(" [inLog = %f %f %d]\n", insampleLog, m_gain, design);//  fflush(0);

  // MJ_fatal(MJ_ERR_UNKNOWN, "FOO"); // Halt here
  MJ_Sample gainLinear = MJ_exp(m_gain);
  MJ_Sample rval = delayedSample * gainLinear;
  
  MJ_Sample outLog = MJ_log(MJ_abs(rval));

  if(outLog > (thresholdLog+.001) || !finite(rval)){
    MJ_Sample gainDiff = thresholdLog - outLog;
    
    MJ_trace2("END vals %f %f %f %f < %f %f %f # %d %d %d", m_gain,
	      m_step, gainDiff, thresholdLog,  
	      outLog, m_logBuffer.getNewest(attackTime-1), insampleLog, 
	      m_untilPeak, attackTime, design);
    MJ_fatal(MJ_ERR_INTERNAL, "MJ_ChannelLimiter::putGet");


  }

  return rval;
}

/////////////////////////////////////////////////////////////////////////////  
/////////////////////////////////////////////////////////////////////////////  


MJ_ModuleLimiter::MJ_ModuleLimiter()
  : m_threshold(1.0),
    m_attackMs(0.1),
    m_releaseMs(30)
{
  dspSetInChannels(1);
  dspSetOutChannels(1);
}

MJ_ModuleLimiter::~MJ_ModuleLimiter()
{}
  
bool MJ_ModuleLimiter::dspPrepare()
{
  uint inchans = dspInChannels();

  if(!inchans || inchans != (uint) dspOutChannels()) {
    error(MJ_ERR_INTERNAL, "MJ_ModuleLimiter::dspPrepare");
    return false;
  }

  m_compressGains.resize(inchans);
  m_channelLimiters.resize(inchans);

  for(uint i=0; i < inchans; i++) {
    m_channelLimiters[i].prepare(m_threshold, (uint) (dspSamplingRate() * 0.1));
  }

  redesignFilters();

  calculateTimes();

  return true;
}
 
void MJ_ModuleLimiter::dspProcessPeriod()
{
  const uint chans   = m_signalInput.size();
  const uint samples = m_signalInput.get(0).samples().size();

  MJ_Sample thresholdLog = MJ_log(m_threshold);

  for(uint i=0; i < chans; i++) {

  
    const MJ_Sample *inSamples = m_signalInput.getFast(i).samples().data();
    MJ_Sample *outSamples = m_signalOutput.getFast(i).samples().data();
        
    MJ_ChannelLimiter *cl = & m_channelLimiters.get(i);

    for(uint j=0; j < samples; j++) {
      /* trace3(MJ_limiterDebug, 
	     "MJ_ModuleLimiter::processNonSync # %d %d",
	     i, j); */
      
      outSamples[j] = cl->putGet(inSamples[j],
				 thresholdLog, 
				 m_attackSamples, 
				 m_releaseSamples);
    }
    m_compressGains.getFast(i) = MJ_exp(cl->gain());
  }

  /* trace3(MJ_limiterDebug, 
     "MJ_ModuleLimiter::processNonSync # EXIT"); */
}

bool MJ_ModuleLimiter::msgProcess(const char *id, MJ_Message *m)
{
  /* trace3(MJ_limiterDebug, "MJ_ModuleLimiter::msgProcess # %s %p",
     id, m); */

  MJ_ParamsLimiter pc = (MJ_ParamsLimiter)
    MJ_findStr(id, MJ_ParamsLimiterKeys, MJ_PARLIM_SIZEOF);
  
  if((int) pc == -1) {
    error(MJ_ERR_ARGUMENT, "MJ_ModuleLimiter::msgProcess", id);
    return false;
  }

  bool ok(true);

  MJ_Sample val = m->toDouble(&ok);

  if(!ok) {
    return false;
  }

  switch (pc) {
  case MJ_PARLIM_THRESHOLD:
    m_threshold = MJ_dbToAbs(MJ_clampS(val, -80, 30));
    /* trace2("MJ_ModuleLimiter::msgProcess # threshold = %f (%f dB)", 
       m_threshold, val); */
    break;
  case MJ_PARLIM_ATTACK:
    m_attackMs = MJ_clampS(val, 0.01, 1.5);
    redesignFilters();
    break;
  case MJ_PARLIM_RELEASE:
    m_releaseMs = MJ_clampS(val, 0.01, 1000);
    break;
  case MJ_PARLIM_SIZEOF:
  default:
    break;
  }

  calculateTimes();
  
  return true;
}

MJ_Parameter * MJ_ModuleLimiter::msgGetInput(const char *id)
{
  MJ_ParamsLimiter pc = (MJ_ParamsLimiter)
    MJ_findStr(id, MJ_ParamsLimiterKeys, MJ_PARLIM_SIZEOF);

  if(pc == MJ_PARLIM_THRESHOLD) {
    return new MJ_ParameterDouble
      (MJ_Strs::Threshold, id, MJ_absToDb(m_threshold), -80.0, 10, 
       MJ_UNIT_DECIBEL);
  }
  else if(pc == MJ_PARLIM_ATTACK) {
    MJ_ParameterDouble *pd = new MJ_ParameterDouble
      ("Attack time", id, m_attackMs, 0.01, 1.0, 
       MJ_UNIT_MILLISECONDS);
    pd->setMapper(new MJ_MapperLog);
    return pd;
  }
  else if(pc == MJ_PARLIM_RELEASE) {
    MJ_ParameterDouble *pd = new MJ_ParameterDouble
      ("Release time", id, m_releaseMs, 0.1, 300, 
       MJ_UNIT_MILLISECONDS);
    pd->setMapper(new MJ_MapperLog);
    return pd;
  }

  return 0;
}

QCString MJ_ModuleLimiter::msgGetInputName(int i)
{
  QCString s;
  if(i < MJ_PARLIM_SIZEOF) s = MJ_ParamsLimiterKeys[i];
  return s;
}

void MJ_ModuleLimiter::redesignFilters()
{
  /*   MJ_Sample sr = dspSamplingRate();

  MJ_Sample rate = (2000.0 / m_attackMs) / sr; // 1000 * 2 to scale between 0,1

  for(uint i =0; i < m_channelLimiters.size(); i++) {
    ; // m_channelLimiters.getFast(i).absFilter().designLowPass(rate);
    }*/
}

void MJ_ModuleLimiter::calculateTimes()
{
  MJ_Sample sr = (MJ_Sample) dspSamplingRate();
  
  m_attackSamples  = MJ_max((uint) (sr * m_attackMs * 0.001 + 0.5), (uint) 2);
  m_releaseSamples = MJ_max((uint) (sr * m_releaseMs * 0.001 + 0.5), (uint) 2);
}


/////////////////////////////////////////////////////////////////////////////  
/////////////////////////////////////////////////////////////////////////////  

template class MJ_Array<MJ_ChannelLimiter>;
