/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_limiter2.h>

#include <mj_error.h> 
#include <mj_mapper.h> 
#include <mj_parameters.h> 
#include <mj_strs.h>

#include <mj_arraytmpl.C>

const char * MJ_ParamsLimiter2Keys[MJ_PARLIM2_SIZEOF] = {
  MJ_Strs::threshold,
  "release"
};


MJ_ModuleLimiter2::MJ_ModuleLimiter2()
  : m_threshold(1.0),
    m_releaseMs(30)
{
  dspSetInChannels(1);
  dspSetOutChannels(1);
}

MJ_ModuleLimiter2::~MJ_ModuleLimiter2()
{}
  
bool MJ_ModuleLimiter2::dspPrepare()
{
  uint inchans = dspInChannels();

  if(!inchans || inchans != (uint) dspOutChannels()) {
    error(MJ_ERR_INTERNAL, "MJ_ModuleLimiter2::dspPrepare");
    return false;
  }

  m_compressGains.resize(inchans);
  m_channelLimiters.resize(inchans);

  for(uint i=0; i < inchans; i++) {
    m_channelLimiters[i].prepare(m_threshold);
  }

  redesignFilters();

  return true;
}
 
void MJ_ModuleLimiter2::dspProcessPeriod()
{
  const uint chans   = m_signalInput.size();
  const uint samples = m_signalInput.get(0).samples().size();

  MJ_Sample gain = 1.0;

  for(uint i=0; i < chans; i++) {

  
    const MJ_Sample *inSamples = m_signalInput.getFast(i).samples().data();
    MJ_Sample *outSamples = m_signalOutput.getFast(i).samples().data();
        
    MJ_ChannelLimiter2 *cl = & m_channelLimiters.get(i);


    for(uint j=0; j < samples; j++) {      

      MJ_Sample input = inSamples[j];

      gain = cl->newGain(input,
			 m_threshold, 
			 m_releaseSamples);
      outSamples[j] = gain * input;
    }
    m_compressGains.getFast(i) = gain;
  }
}

bool MJ_ModuleLimiter2::msgProcess(const char *id, MJ_Message *m)
{
  /* trace3(MJ_limiter2Debug, "MJ_ModuleLimiter2::msgProcess # %s %p",
     id, m); */

  MJ_ParamsLimiter2 pc = (MJ_ParamsLimiter2)
    MJ_findStr(id, MJ_ParamsLimiter2Keys, MJ_PARLIM2_SIZEOF);
  
  if((int) pc == -1) {
    error(MJ_ERR_ARGUMENT, "MJ_ModuleLimiter2::msgProcess", id);
    return false;
  }

  bool ok(true);

  MJ_Sample val = m->toDouble(&ok);

  if(!ok) {
    return false;
  }

  switch (pc) {
  case MJ_PARLIM2_THRESHOLD:
    m_threshold = MJ_dbToAbs(MJ_clampS(val, -80, 30));
    /* trace2("MJ_ModuleLimiter2::msgProcess # threshold = %f (%f dB)", 
       m_threshold, val); */
    break;
  case MJ_PARLIM2_RELEASE:
    m_releaseMs = MJ_clampS(val, 0.01, 1000);
    redesignFilters();
    break;
  case MJ_PARLIM2_SIZEOF:
  default:
    break;
  }


  return true;
}

MJ_Parameter * MJ_ModuleLimiter2::msgGetInput(const char *id)
{
  MJ_ParamsLimiter2 pc = (MJ_ParamsLimiter2)
    MJ_findStr(id, MJ_ParamsLimiter2Keys, MJ_PARLIM2_SIZEOF);

  if(pc == MJ_PARLIM2_THRESHOLD) {
    return new MJ_ParameterDouble
      (MJ_Strs::Threshold, id, MJ_absToDb(m_threshold), -80.0, 10, 
       MJ_UNIT_DECIBEL);
  }
  else if(pc == MJ_PARLIM2_RELEASE) {
    MJ_ParameterDouble *pd = new MJ_ParameterDouble
      ("Release time", id, m_releaseMs, 1, 500, 
       MJ_UNIT_MILLISECONDS);
    pd->setMapper(new MJ_MapperLog);
    return pd;
  }

  return 0;
}

QCString MJ_ModuleLimiter2::msgGetInputName(int i)
{
  QCString s;
  if(i < MJ_PARLIM2_SIZEOF) s = MJ_ParamsLimiter2Keys[i];
  return s;
}

void MJ_ModuleLimiter2::redesignFilters()
{
  MJ_Sample sr = dspSamplingRate();

  MJ_Sample rate = (2000.0 / m_releaseMs) / sr; // 1000 * 2 to scale between 0,1

  for(uint i =0; i < m_channelLimiters.size(); i++) {
    m_channelLimiters.getFast(i).absFilter().designLowPass(rate);
  }

  m_releaseSamples = 
    MJ_max((uint) (sr * m_releaseMs * 0.001 + 0.5), (uint) 2);
}


/////////////////////////////////////////////////////////////////////////////  
/////////////////////////////////////////////////////////////////////////////  

template class MJ_Array<MJ_ChannelLimiter2>;
