/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_eq1.h>

#include <mj_global.h>
#include <mj_mapper.h>
#include <mj_parameters.h>
#include <mj_strs.h>

const char * MJ_MEQ1_params[3] = {
  MJ_Strs::gain,
  MJ_Strs::frequency,
  MJ_Strs::bandwidth
};

bool MJ_ModuleEq1::m_debug = false;

MJ_ModuleEq1::MJ_ModuleEq1()
  : m_gainDb(0.0),
    m_centerHz(100.0),
    m_bandWidthOctaves(0.33),
    m_targetGainDb(0.0),
    m_targetCenterHz(100.0),
    m_targetBandWidthOctaves(0.33),
    m_samplingRate(44100.0),
    m_interpSamples(0)
{}

MJ_ModuleEq1::~MJ_ModuleEq1()
{}

bool MJ_ModuleEq1::dspPrepare()
{ 
  m_interpSamples = 0;

  m_gainDb = m_targetGainDb;
  m_centerHz = m_targetCenterHz;
  m_bandWidthOctaves = m_targetBandWidthOctaves;
 
  if(!updateFilterCoeffs()) return false;

  if(m_signalInput[0].samples().size() != m_signalOutput[0].samples().size())
    return false;

  return true;
}

void MJ_ModuleEq1::dspProcessPeriod()
{
  const uint chans = m_signalInput.size();

  uint samples = m_signalInput.getFast(0).samples().size();
  uint i, j;

  uint startSample = MJ_min(m_interpSamples, (ulong) samples);
  uint calcSamples = samples - startSample;

  MJ_SignalSource *sinput = m_signalInput.data();
  MJ_SignalSink *soutput = m_signalOutput.data();

  // Do the interpolation
  if(startSample) {

    // cout << "." << m_interpSamples << flush;

    double scale = 1.0 / (double) m_interpSamples;

    m_interpSamples -= (ulong) startSample;

    double gainRate = (m_targetGainDb - m_gainDb) * scale;
    double centerRate = (m_targetCenterHz - m_centerHz) * scale;
    double bwRate = (m_targetBandWidthOctaves - m_bandWidthOctaves) * scale;

    MJ_FilterIIR2<MJ_Sample> *iir = & m_filters[0];

    for(j = 0; j < startSample; j++) {
      
      iir->designParametricB2((m_centerHz += centerRate),
			      m_samplingRate,
			      (m_gainDb += gainRate),
			      (m_bandWidthOctaves += bwRate));
      
      for(i = 0; i < chans; i++) {
	MJ_Sample tmp = 
	  iir->putGetSampleOther(&m_filters[i], 
				 sinput[i].samples().get(j));
	soutput[i].samples().get(j) = tmp;
      }
    }
    for(i = 1; i < chans; i++) {
      m_filters[i].copyCoefficients(*iir);
    }
  }

  // Do normal filtering
  if(calcSamples) {
    for(i = 0; i < chans; i++) {
      m_filters[i].putSamples(& soutput[i].samples().getFast(startSample), 
			      & sinput[i].samples().getFastConst(startSample), 
			      calcSamples);
    }
  }
}

bool MJ_ModuleEq1::msgProcess(const char * i, MJ_Message *m)
{
  const char *fname = "MJ_ModuleEq1::msgProcess";
  bool ok = true;

  /* Optionally trace program execution. */
  if(m_debug) trace(fname, i);

  int id = MJ_findStr(i, MJ_MEQ1_params, 3);
  
  if(id == -1) {
    if(m_debug) error(MJ_ERR_ARGUMENT, fname, id);
    return false;
  }
  double tmp = m->toDouble(&ok);
  
  double g = m_targetGainDb, hz = m_targetCenterHz, 
    bw = m_targetBandWidthOctaves;

  if(!ok) {
    if(m_debug) 
      error(MJ_ERR_ARGUMENT, fname, "Failure 1");
    return false; 
  }

  switch(id) {
  case 0:
    g = tmp;
    break;
  case 1:
    hz = tmp;
    break;
  case 2:
    bw = tmp;
    break;
  default:
    ok = false;
    break;
  }

  if(!ok) {
    if(m_debug) 
      error(MJ_ERR_ARGUMENT, fname, "Failure 2");
    return false; 
  }

  if(m_debug) trace(fname, "Success 1");
    
  if(dspIsRealTime()) {
    m_targetGainDb = g;
    m_targetCenterHz = hz;
    m_targetBandWidthOctaves = bw;
	
    if((long) m->timeStamp() < (long) m_currentTime->stamp()) {
      // The number of interpolation samples, slow interpolation
      m_interpSamples = (ulong) (m_samplingRate / 20.0) + 1;
    }
    else {
      m_interpSamples = m->timeStamp() - m_currentTime->stamp();
    }
    
    if(m_debug) {
      char buf[128];
      sprintf(buf, "m_interpSamples 0= %d", (int) m_interpSamples);
      trace(fname, buf);
    }
  }      
  else {
    m_gainDb = g;
    m_centerHz = hz;
    m_bandWidthOctaves = bw;
    
    m_targetGainDb = g;
    m_targetCenterHz = hz;
    m_targetBandWidthOctaves = bw;
    
    updateFilterCoeffs();
  }

  return ok;
}

MJ_Parameter * MJ_ModuleEq1::msgGetInput(const char * i)
{
  if(m_debug) trace("MJ_ModuleEq1::msgGetInput", i);
  
  if(!strcmp(i, MJ_Strs::gain)) 
    return new MJ_ParameterDouble(MJ_Strs::Gain, i, m_targetGainDb, 
				  -15, 15, MJ_UNIT_DECIBEL);
  else if(!strcmp(i, MJ_Strs::frequency)) {
    MJ_ParameterDouble *pd = 
      new MJ_ParameterDouble(MJ_Strs::Frequency, i, m_targetCenterHz, 
			     20, 20000, MJ_UNIT_HERTZ);
    pd->setMapper(new MJ_MapperLog);
    return pd;
  }
  else if(!strcmp(i, MJ_Strs::bandwidth))
    return new MJ_ParameterDouble(MJ_Strs::Bandwidth, i, 
				  m_targetBandWidthOctaves, 
				  0.05, 2.0, MJ_UNIT_BLANK);

  return 0;
}

QCString MJ_ModuleEq1::msgGetInputName(int i)
{
  QCString s;
  
  switch(i) {
  case 0:
    s = MJ_Strs::gain;
    break;

  case 1:
    s = MJ_Strs::frequency;
    break;

  case 2:
    s = MJ_Strs::bandwidth;
    break;

  default:
    break;
  }
  
  return s;
}
  
bool MJ_ModuleEq1::updateFilterCoeffs()
{
  const char * fname = "MJ_ModuleEq1::updateFilterCoeffs";

  uint i;
  const uint chans = m_signalInput.size();

  if(!chans) return false;
  
  MJ_SampleRateDef &sr = m_signalInput[0].sink()->sampleRate();

  uint oldChans = m_filters.size();
  m_filters.resize(chans);

  for(i=oldChans; i < chans; i++)
    m_filters.getFast(i).resetMemory();

  MJ_FilterIIR2<MJ_Sample> *iir = & m_filters.getFast(0);

  m_samplingRate = sr.rate();

  iir->designParametricB2(m_centerHz,
			  m_samplingRate, 
			  m_gainDb, 
			  m_bandWidthOctaves);
  
  if(m_debug) {
    char buf[256];
    iir->writeTo(buf);
    trace(fname, buf);
  }

  iir->resetMemory();

  for(i=1; i < chans; i++) {
    m_filters.getFast(i) = *iir;
    
  }

  return true;
}
