/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_eq2.h>

#include <mj_module_eq1.h>
#include <mj_error.h>

#include <mj_parameters.h>
#include <mj_mapper.h>
#include <mj_strs.h>

#include <mj_arraytmpl.C>
#include <mj_array2dtmpl.C>

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_Eq2FilterParams::MJ_Eq2FilterParams()
{
  reset();
}

void MJ_Eq2FilterParams::reset()
{
  m_gainDb = 0.0;
  m_centerHz = 100.0;
  m_bandWidthOctaves = 0.33;
  m_targetGainDb = 0.0;
  m_targetCenterHz = 100.0;
  m_targetBandWidthOctaves = 0.33;
  m_interpSamples = 0;  
}

void MJ_Eq2FilterParams::addNewParam(int p, 
				     MJ_Sample target, 
				     ulong interpSamples)
{
  /* MJ_trace2("MJ_Eq2FilterParams::addNewParam %d %f %d", 
     p, target, (int) interpSamples); */

  switch(p) {
  case 0:
    m_targetGainDb = target;
    break;
  case 1:
    if(target < 10) target = 10.0;
    m_targetCenterHz = target;
    break;
  case 2:
    if(target < 0.01) target = 0.01;
    m_targetBandWidthOctaves = target;
    break;
  default: 
    MJ_error(MJ_ERR_ARGUMENT, "MJ_Eq2FilterParams::addNewParam");
  }
  m_interpSamples = interpSamples;
}

void MJ_Eq2FilterParams::describe(char *buf)
{
  sprintf(buf, "TARGET: gain = %f, Hz = %f, bw = %f", 
	  m_targetGainDb, m_targetCenterHz, m_targetBandWidthOctaves);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


MJ_ModuleEq2::MJ_ModuleEq2(uint peqBands)
  : m_inLadspa(false)
{
  dspSetInChannels(1);
  dspSetOutChannels(1);
  resize(peqBands);
}

MJ_ModuleEq2::~MJ_ModuleEq2()
{
  m_filters.customResize(mjDspAllocator(), 0, 0);
}

bool MJ_ModuleEq2::dspPrepare()
{
  int in = dspInChannels();

  if(in < 1 || in != dspOutChannels())
    return false;

  m_filters.resize(stages(), in);  
  
  // Custom memory allocation for cache-hit performance
  m_filters.customResize(mjDspAllocator(), stages(), in);

  updateFilterCoeffs();

  clearFilterMemories();

  return true;
}

/** A huge bloated function. */
void MJ_ModuleEq2::dspProcessPeriod()
{
  const char * fname = "MJ_ModuleEq2::dspProcessPeriod";
  // Generic parameters:
  uint i, j, k;

  MJ_SignalSource *sinput = m_signalInput.data();
  const uint samples = sinput[0].samples().size();

  MJ_SignalSink *soutput = m_signalOutput.data();
  const uint chans = m_signalOutput.size();

  double sr = dspSamplingRate();

  // Low-shelve:
  
  MJ_Eq2FilterParams *fp = & m_params.get(0);

  ulong interpSamples = fp->interpSamples();
  uint startSample = MJ_min(interpSamples, (ulong) samples);
  uint calcSamples = samples - startSample;
  
  MJ_FilterIIR2<MJ_Sample> *filters = m_filters.getRow(0);

  if(m_filters.cols() != chans) MJ_fatal(MJ_ERR_INTERNAL, fname);

  // trace2("%s # rows = %d, cols = %d", fname, m_filters.rows(), m_filters.cols());
  
  if(startSample) {
    double scale = 1.0 / (double) fp->interpSamples();

    fp->interpSamples() -= (ulong) startSample;

    double gainRate = fp->gainRate(scale);
    double centerRate = fp->centerHzRate(scale);
    double bwRate = fp->bwRate(scale);

    double centerHz = fp->centerHz();
    double gainDb = fp->gainDb();
    double bandWidthOctaves = fp->bandWidthOctaves();

    MJ_FilterIIR2<MJ_Sample> *iir = filters;
    
    for(j = 0; j < startSample; j++) {
      
      iir-> designLowShelve2((centerHz += centerRate),
			     sr,
			     (gainDb += gainRate),
			     (bandWidthOctaves += bwRate));
      
      for(i = 0; i < chans; i++) {
	MJ_Sample tmp = 
	  iir->putGetSampleOther(&filters[i], 
				 sinput[i].samples().getFast(j));
	soutput[i].samples().getFast(j) = tmp;

	/* if(!finite(tmp))
	   MJ_fatal(MJ_ERR_FPE, fname, "non-finite"); */
      }
    }
    fp->centerHz() =  centerHz;
    fp->gainDb() = gainDb;
    fp->bandWidthOctaves() = bandWidthOctaves;
    
    for(i = 1; i < chans; i++) {
      filters[i].copyCoefficients(*iir);
    }
  }

  if(calcSamples) {
    for(i = 0; i < chans; i++) {
      filters[i].putSamples(& soutput[i].samples().getFast(startSample), 
			    & sinput[i].samples().getFastConst(startSample), 
			    calcSamples);
      /* printf(" # %d %f -> %f #", 
	 startSample,
	 soutput[i].samples().getFast(startSample), 
	 sinput[i].samples().getFastConst(startSample)); */
    }
  }

  // return;

  /* if(!soutput[0].samples().allElemsFinite())
     MJ_fatal(MJ_ERR_FPE, fname, "nonfinite low"); */


  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  // Parametric EQs:

  for(k=1; k < m_params.size() - 1; k++) {

    filters = m_filters.getRow(k);
    
    fp = & m_params.get(k);
  
    interpSamples = fp->interpSamples();
    startSample = MJ_min(interpSamples, (ulong) samples);
    calcSamples = samples - startSample;
    
    if(startSample) {
      // MJ_fatal(MJ_ERR_INTERNAL, fname);

      double scale = 1.0 / (double) fp->interpSamples();
      
      fp->interpSamples() -= (ulong) startSample;
      
      double gainRate = fp->gainRate(scale);
      double centerRate = fp->centerHzRate(scale);
      double bwRate = fp->bwRate(scale);
      
      double centerHz = fp->centerHz();
      double gainDb = fp->gainDb();
      double bandWidthOctaves = fp->bandWidthOctaves();

      MJ_FilterIIR2<MJ_Sample> *iir = filters;
      
      for(j = 0; j < startSample; j++) {
	
	iir->designParametricB2((centerHz += centerRate),
				sr,
				(gainDb += gainRate),
				(bandWidthOctaves += bwRate));
	
	for(i = 0; i < chans; i++) {
	  MJ_Sample tmp = 
	    iir->putGetSampleOther(&filters[i], 
				   soutput[i].samples().getFast(j));
	  soutput[i].samples().getFast(j) = tmp;
	}
      }

      fp->centerHz() =  centerHz;
      fp->gainDb() = gainDb;
      fp->bandWidthOctaves() = bandWidthOctaves;
    
      for(i = 1; i < chans; i++) {
	filters[i].copyCoefficients(*iir);
      }
    }

    // return ;

    if(calcSamples) {
      for(i = 0; i < chans; i++) {
	filters[i]. putGetSamplesInplace
	  (& soutput[i].samples().get(startSample), 
	   calcSamples);
      }
    }
  }
  

  /* if(!soutput[0].samples().allElemsFinite())
     MJ_fatal(MJ_ERR_FPE, fname, "nonfinite middle"); */


  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  // High-shelve:
  
  fp = & m_params.last();
  
  interpSamples = fp->interpSamples();
  startSample = MJ_min(interpSamples, (ulong) samples);
  calcSamples = samples - startSample;
  
  filters = m_filters.getLastRow();
  
  if(startSample) {
    double scale = 1.0 / (double) fp->interpSamples();

    fp->interpSamples() -= (ulong) startSample;

    double gainRate = fp->gainRate(scale);
    double centerRate = fp->centerHzRate(scale);
    double bwRate = fp->bwRate(scale);

    double centerHz = fp->centerHz();
    double gainDb = fp->gainDb();
    double bandWidthOctaves = fp->bandWidthOctaves();

    MJ_FilterIIR2<MJ_Sample> *iir = filters;
    
    for(j = 0; j < startSample; j++) {
      
      iir-> designHighShelve2((centerHz += centerRate),
			      sr,
			      (gainDb += gainRate),
			      (bandWidthOctaves += bwRate));
      
      for(i = 0; i < chans; i++) {
	MJ_Sample tmp = 
	  iir->putGetSampleOther(&filters[i], 
				 soutput[i].samples().get(j));
	soutput[i].samples().get(j) = tmp;
      }
    }

    fp->centerHz() =  centerHz;
    fp->gainDb() = gainDb;
    fp->bandWidthOctaves() = bandWidthOctaves;
    
    for(i = 1; i < chans; i++) {
      filters[i].copyCoefficients(*iir);
    }
  }

  /* if(!soutput[0].samples().allElemsFinite())
     MJ_fatal(MJ_ERR_FPE, fname, "nonfinite semifinal"); */
  
  if(calcSamples) {
    for(i = 0; i < chans; i++) {
      filters[i]. putGetSamplesInplace
	(& soutput[i].samples().getFast(startSample), 
	 calcSamples);
    }
  }

  /* trace2("%s # in1 = %p, out1 = %p", fname, 
     m_signalInput[0].samples().data(),
     m_signalOutput[0].samples().data()); */

  /* if(!soutput[0].samples().allElemsFinite())
     MJ_fatal(MJ_ERR_FPE, fname, "nonfinite final");*/
  
  /* if(soutput[0].samples().max() > 3.0 || soutput[0].samples().min() < -3.0 )
     MJ_fatal(MJ_ERR_FPE, fname, "large"); */
}

bool MJ_ModuleEq2::dspStop()
{
  m_filters.customResize(mjDspAllocator(), 0, 0);
  return true;
}

void MJ_ModuleEq2::msgGetInputs(MJ_Parameters *p)
{
  char buf[64];

  p->clear();

  if(!m_inLadspa)
    p->append(msgGetInput(MJ_Strs::bands));
  
  for(uint i=0; i < m_params.size(); i++) {
    for(uint j=0; j < 3; j++) {
      sprintf(buf, "%s_%d", MJ_MEQ1_params[j], i);
      p->append(msgGetInput(buf));
    }
  }
}

bool MJ_ModuleEq2::msgProcess(const char *id, MJ_Message *m)
{
  const char * fname = "MJ_ModuleEq2::msgProcess";
  // trace(fname, id);

  bool ok = true;

  if(strcmp(id, "inladspa") == 0) {
    m_inLadspa = true;
    return true;
  }

  // Handle resize
  if(strcmp(id, MJ_Strs::bands) == 0) {

    if(m_inLadspa) return false;

    int bands = m->toInt(&ok);

    if(!ok || dspIsRealTime() || bands < 3 || bands > 200) {
      error(MJ_ERR_ARGUMENT, fname, id);
      return false;
    }
    
    resize(bands);

    return true;
  }

  // Handle parameter change:
  
  int param = MJ_findStr(MJ_fieldC(id, 0, "_"), MJ_MEQ1_params, 3);
  
  uint stage = MJ_fieldC(id, 1, "_").toInt(&ok);
  
  if(stage >= m_params.size()) {
    error(MJ_ERR_ARGUMENT, fname, id);
    return false;
  }
  
  MJ_Eq2FilterParams *fp = & m_params.get(stage);
  MJ_Sample target = m->toDouble(&ok);

  if(!ok) return false;
  
  ulong interpSamples;

  if(dspIsRealTime()) {
    if((long) m->timeStamp() < (long) m_currentTime->stamp()) {
      // The number of interpolation samples, slow interpolation
      interpSamples = (ulong) (dspSamplingRate() * 0.1) + 1;
    }
    else {
      interpSamples = m->timeStamp() - m_currentTime->stamp();
    }
  }
  else {
    // trace(fname, "not rt");
    interpSamples = 0;
  }
  
  fp->addNewParam(param, target, interpSamples);

  // trace(fname, "EXIT OK");

  return true;
}

MJ_Parameter * MJ_ModuleEq2::msgGetInput(const char *id)
{
  const char * fname = "MJ_ModuleEq2::msgGetInput";

  // trace(fname, id);
  
  if(strcmp(id, MJ_Strs::bands) == 0) {
    return new MJ_ParameterInt(MJ_Strs::Bands, id, stages(),
			       3, 100);
  }

  int param, stage;

  determineParams(id, &param, & stage);
    
  if((uint) stage >= m_params.size() || param < 0) {
    error(MJ_ERR_ARGUMENT, fname, id);
    return 0;
  }
  
  MJ_Eq2FilterParams *fp = & m_params.get(stage);

  if(param == 0) {
    return new MJ_ParameterDouble(MJ_Strs::Gain, id, fp->targetGainDb(), 
				  -15, 15, MJ_UNIT_DECIBEL);
  }
  else if(param == 1) {
    MJ_ParameterDouble *pd = 
      new MJ_ParameterDouble(MJ_Strs::Frequency, id, fp->targetCenterHz(), 
			     20, 16500, MJ_UNIT_HERTZ);
    pd->setMapper(new MJ_MapperLog);
    return pd;
  }
  else if(param == 2) {
    const char *name(0);

    if(!stage || (uint) stage == (m_params.size() - 1)) 
      name = MJ_Strs::Slope;
    else
      name = MJ_Strs::Bandwidth;
    
    return new MJ_ParameterDouble(name, id, 
				  fp->targetBandWidthOctaves(), 
				  0.05, 2.0, MJ_UNIT_BLANK);
  }

  return 0;
}

uint MJ_ModuleEq2::closestBand(MJ_Sample freqHz)
{
  MJ_Sample freqMapped = MJ_log(freqHz);

  MJ_Sample bestFit = 1e+10;
  uint bestInd = 0;

  uint params = m_params.size();

  for(uint i=0; i < params; i++) {
    MJ_Sample absDiff = MJ_abs(MJ_log(m_params.get(i).targetCenterHz()) -
			       freqMapped);
    if(absDiff < bestFit) {
      bestFit = absDiff;
      bestInd = i;
    }
  }

  return bestInd;
}

/** Find the EQ band that is closest to the target. We are assuming
    that the center frequencies are ascending. This should always be
    true for graphical EQs. */

uint MJ_ModuleEq2::closestGraphicBand(MJ_Sample freqHz)
{

  uint params = m_params.size();

  if(freqHz < m_params.get(0).targetCenterHz())
    return 0;
  
  if(freqHz > m_params.get(params - 2).targetCenterHz())
    return params - 1;
  
  return closestBand(freqHz);
}

void MJ_ModuleEq2::formMsgId(char *buf, int type, int index)
{
  sprintf(buf, "%s_%d", MJ_MEQ1_params[type], index);
}

/** Determine the parameter this id string refers to. 

    This method complements formMsgId.
 */
void MJ_ModuleEq2::determineParams(const char *id, int *type, int *index)
{
  bool ok = true;

  *type = MJ_findStr(MJ_fieldC(id, 0, "_"), MJ_MEQ1_params, 3);
  *index = MJ_fieldC(id, 1, "_").toInt(&ok);
  
  if(!ok) *index = -1;
}

/** Default parameters for one band. */
MJ_Eq2FilterParams MJ_ModuleEq2::defaultBandParam(uint band, uint bands)
{
  MJ_Eq2FilterParams p;
  
  p.reset();

  MJ_Sample lowFreqHz;
  MJ_Sample highFreqHz;

  if(bands < 3 || band >= bands) {
    MJ_error(MJ_ERR_ARGUMENT, "MJ_ModuleEq2::defaultBandParam");
    return p;
  }

  // Determine edge frequencies:
  if(bands <= 3) {

    lowFreqHz = 150.0;
    highFreqHz = 5000.0;
  }
  else if(bands <= 5) {
    lowFreqHz = 100.0;
    highFreqHz = 10000.0;
  }
  else if(bands <= 7) {
    lowFreqHz = 60.0;
    highFreqHz = 13000.0;
  }
  else if(bands <= 15) {
    lowFreqHz = 40.0;
    highFreqHz = 16000.0;
  }
  else {
    lowFreqHz = 25.0;
    highFreqHz = 16500.0;
  }
  
  if(!band || band == (bands - 1)) {

    if(!band) p.m_targetCenterHz = lowFreqHz * 1.5;
    else p.m_targetCenterHz = highFreqHz * 0.75;

    p.m_targetBandWidthOctaves = 1.0; // Slope !!
  }
  else {

    MJ_Sample lowMapped = MJ_log(lowFreqHz);
    MJ_Sample highMapped = MJ_log(highFreqHz);
    
    MJ_Sample centerMapped = (highMapped - lowMapped) * 
      ((MJ_Sample) band / (MJ_Sample) (bands - 1)) + lowMapped;

    MJ_Sample centerHz = MJ_exp(centerMapped);
    p.m_targetCenterHz = centerHz;

    MJ_Sample freqRatio = highFreqHz / lowFreqHz;
    MJ_Sample octaves = MJ_log(freqRatio) / MJ_ln2;
    MJ_Sample bw = octaves * 1.2 / (MJ_Sample) (bands - 2);
    
    p.m_targetBandWidthOctaves = bw;

    /* char buf[256];
       p.describe(buf);
       MJ_trace("MJ_ModuleEq2::defaultBandParam", buf); */
  }

  p.copyTargetValues();

  return p;
}


void MJ_ModuleEq2::resize(uint stages)
{
  // trace("MJ_ModuleEq2::resize", stages);

  MJ_Eq2FilterParams fp;
  uint oldSize = m_params.size();

  if(oldSize && stages) {
    fp = m_params.last();
  }

  m_params.resize(stages);
  
  if(oldSize < stages) {
    for(uint i=oldSize; i < stages; i++) {
      m_params.get(i) = defaultBandParam(i, stages);
    }    
  }

  if(oldSize && stages) {
    m_params.last() = fp;
    if(oldSize < stages)
      m_params.get(oldSize - 1).reset();
  }
    
}

/* Update all filter parameters (abruptly without future
   interpolation) */

bool MJ_ModuleEq2::updateFilterCoeffs()
{
  uint i, j;

  MJ_Sample sr = (MJ_Sample) dspSamplingRate();

  const uint stages = m_filters.rows();
  uint chans = m_filters.cols();

  for(i=0; i < stages; i++) {
    MJ_FilterIIR2<MJ_Sample> *iir = & m_filters.get(i, 0);
    MJ_Eq2FilterParams *fp = & m_params.getFast(i);
    fp->copyTargetValues();
  
    // Design the filter
    if(!i) {
      iir->designLowShelve2(fp->targetCenterHz(),
			    sr,
			    fp->targetGainDb(),
			    fp->targetBandWidthOctaves());
    }
    else if(i == (stages - 1)) {
      iir->designHighShelve2(fp->targetCenterHz(),
			     sr,
			     fp->targetGainDb(),
			     fp->targetBandWidthOctaves());
    }
    else {
      iir->designParametricB2(fp->targetCenterHz(),
			      sr,
			      fp->targetGainDb(),
			      fp->targetBandWidthOctaves());
    }
    // Copy values to other similar filters:
    MJ_FilterIIR2<MJ_Sample> *iirTarget = iir;
    for(j=1; j < chans; j++) {
      iirTarget[j].copyCoefficients(*iir);
    }

    /* char buf[128], buf2[128];
    iir->writeTo(buf);
    fp->describe(buf2);
    trace2("MJ_ModuleEq2::updateFilterCoeffs %d %s # %s", i, buf, buf2); */
  }
  return true;
}

void MJ_ModuleEq2::clearFilterMemories()
{
  uint cols = m_filters.cols();

  for(uint i=0; i < m_filters.rows(); i++) {
    MJ_FilterIIR2<MJ_Sample> *frow = m_filters.getRow(i);
    for(uint j=0; j < cols; j++)
      frow[j].resetMemory();
  }
}


template class MJ_Array<MJ_Eq2FilterParams> ;
  
template class MJ_Array2D<MJ_FilterIIR2<MJ_Sample> > ;
  
