/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_eq2.h>
#include <mj_module_eq2_gui.h>

#include <mj_cstr_remap.h> 
#include <mj_error.h> 
#include <mj_freq_resp_gui.h> 
#include <mj_graphic_polyline.h> 
#include <mj_graphics_context_simple.h> 
#include <mj_gui_numeric.h> 
#include <mj_gui_utils.h>
#include <mj_int_remap.h> 
#include <mj_mapper.h> 
#include <mj_messages.h> 
#include <mj_module_menubar.h> 
#include <mj_parameters.h> 
#include <mj_satellite.h> 
#include <mj_slider.h> 
#include <mj_strs.h>
#include <mj_undo_info.h> 
#include <mj_version.h>

#include <qlabel.h> 
#include <qlayout.h> 

bool MJ_ModuleEq2_QG::m_debug = false;

MJ_ModuleEq2_QG::MJ_ModuleEq2_QG(MJ_RtControl *rtparent,
				 MJ_ModuleDsp *module,
				 bool          drawSliders,
				 QWidget      *parent,
				 const char   *name)
  : MJ_ModuleDspGuiPostCard(rtparent, module, true,
			    0,
			    MJ_GUI_POSTCARD,
			    parent, name),
  m_freqResp(0),
  m_sliderBase(0),
  m_sliderBaseLayout(0),
  m_sliderWindow(0),
  m_menuBar(0),
  m_activeBand(0),
  m_blockDraw(false)
{
  resize(600, 400);

  m_objects.setAutoDelete(true);
  
  drawEq(drawSliders);
  realTimeChange(isRealTime());
}

MJ_ModuleEq2_QG::~MJ_ModuleEq2_QG()
{}

void MJ_ModuleEq2_QG::moduleModified(const char *id)
{
  const char * fname = "MJ_ModuleEq2_QG::moduleModified";

  // Handle band number change
  if(strcmp(id, MJ_Strs::bands) == 0) {

    lock();
    uint stages = ((MJ_ModuleEq2 *) m_module)->stages();
    unlock();

    setModuleInternal();

    resizeFreqResp(stages);
    m_activeBand = 1;
    setActiveBand(1);

    if(m_sliderBase)
      rebuildSliders();

    updateEq();
    
    return;
  }

  MJ_ModuleDspGuiPostCard::moduleModified(id);

  // Update sliders:
  int param, stage;
  
  MJ_ModuleEq2::determineParams(id, & param, & stage);

  if((uint) param > 3)
    MJ_fatal(MJ_ERR_INTERNAL, fname);

  if((uint) stage != m_activeBand)
    setActiveBand(stage);

  if(!m_blockDraw)
    updateBand(stage);
  
  // m_freqResp->redraw(); 

  uint guiIndex = stage * 3 + param;
  
  if(guiIndex >= m_guiComponents.size())
    return;
  
  void *qo = m_guiComponents.get(guiIndex);

  lock();
  MJ_Parameter * p = m_module->msgGetInput(id);
  unlock();

  if(!p) {
    error(MJ_ERR_INTERNAL, fname, id);
    return;
  }
  
  double val = ((MJ_ParameterDouble *) p)->value();

  delete p;

  if(param == 1) 
    ((MJ_GuiDoubleSlider *) qo)->setValue(val);
  else
    ((MJ_Slider *) qo)->setValueDouble(val);
}

MJ_ModuleDspGui *MJ_ModuleEq2_QG::createNew(MJ_RtControl *rtparent,
					    MJ_ModuleDsp *module,
					    QWidget *parent, 
					    const char *name)
{ 
  return new MJ_ModuleEq2_QG(rtparent, module, true, parent, name);
}

int MJ_ModuleEq2_QG::menuChoiceToBands(int i)
{
  switch(i) {
  case 0:
    return 3;
  case 1:
    return 5;
  case 2:
    return 7;
  case 3:
    return 9;
  case 4:
    return 15;
  case 5:
    return 31;
  }
  return 0;
}

void MJ_ModuleEq2_QG::grMousePress(QMouseEvent *e, float x, float y)
{
  // qDebug("MJ_ModuleEq2_QG::grMousePress # %f %f", x, y);
  
  if(!e->state()) {
    seekActiveBand(x, y);
  }

  if(e->stateAfter())
    setPeak(x, y);
}

void MJ_ModuleEq2_QG::grMouseRelease(QMouseEvent *, float, float)
{
  char bufGain[128], bufHz[128];

  MJ_ModuleEq2::formMsgId(bufGain, 0, m_activeBand);

  doneValue(bufGain);

  if(m_sliderBase) {
    MJ_ModuleEq2::formMsgId(bufHz, 1, m_activeBand);
    doneValue(bufHz);
  }
}

/** Flat frequency response. This method uses the multi-message
    undo/redo capability of MJ_UndoInfo. */

void MJ_ModuleEq2_QG::flatten()
{
  if(m_debug) trace("MJ_ModuleEq2_QG::flatten");

  char buf[128];

  MJ_UndoInfo *ui = new MJ_UndoInfo;

  MJ_ModuleEq2 *eq = (MJ_ModuleEq2 *) m_module;

  lock();
  uint stages = eq->stages();
  unlock();


  for(uint i=0; i <  stages; i++) {
    
    MJ_MessageDouble *md = new MJ_MessageDouble(0.0);
    MJ_MessageCarrier *mi = new MJ_MessageCarrier(buf, md);

    MJ_ModuleEq2::formMsgId(buf, 0, i);

    ui->addMsg(mi, false);

    lock();
    MJ_Sample gainOld = eq->parameters().get(i).targetGainDb();
    eq->msgProcess(buf, md);
    unlock();

    md = new MJ_MessageDouble(gainOld);
    mi = new MJ_MessageCarrier(buf, md);
    
    ui->addCounterMsg(mi, false);

    moduleModified(buf); // Notify GUI of the change.
  }
  
  ui->setParamName(tr("Flat response"));
  
  eq->satellite()->addUndoInfo(ui);
}

void MJ_ModuleEq2_QG::setBands(int i)
{
  if(m_debug) trace("MJ_ModuleEq2_QG::setBands", i);

  MJ_MessageInt msg(i);
  
  if(!m_sliderBase)
    ((MJ_ModuleEq2 *) m_module)->resize(0);

  m_module->msgProcess(MJ_Strs::bands, &msg);
  m_module->satellite()->modified(MJ_Strs::bands);

  // trace("MJ_ModuleEq2_QG::setBands # exit");
}

void MJ_ModuleEq2_QG::realTimeChange (bool nowRt) 
{
  if(m_menuBar) 
    m_menuBar->edit()->setItemEnabled(m_bandItem, !nowRt);
}

void MJ_ModuleEq2_QG::drawEq(bool drawSliders)
{
  if(m_debug) trace("MJ_ModuleEq2_QG::draw");

  uint i;
  int j;

  int rows = m_layout->numRows();
  int cols = m_layout->numCols();

  m_layout->expand(rows + 3, cols);
  
  if(!parent()) {
    MJ_ModuleMenubar *mbar = new MJ_ModuleMenubar(this);
    mbar->edit()->insertSeparator();

    QPopupMenu *bandNum = new QPopupMenu(this);

    QString str;

    for(i = 0; (j = menuChoiceToBands(i)); i++) {
      str.setNum(j);
      MJ_IntRemap *map = new MJ_IntRemap(j);

      bandNum->insertItem(str, map, SLOT(call()));
      
      connect(map, SIGNAL(emitCall(int)), 
	      this, SLOT(setBands(int)));
      m_children.append(map);
    }

    m_bandItem = mbar->edit()->insertItem(tr("Bands"), bandNum);

    mbar->edit()->insertItem(tr("Flat response"), 
			     this, SLOT(flatten()));
    
    m_layout->setMenuBar(mbar);
    m_menuBar = mbar;
  }

  // Help label for the graph

  QLabel *label = 
    new QLabel(tr("You can adjust the EQ by clicking on the graphs"), this);
  m_layout->addMultiCellWidget(label, rows, rows, 0, cols - 1);
  rows++;

  lock();
  MJ_ModuleEq2 *eq = (MJ_ModuleEq2 *) m_module;
  uint stages = eq->stages();
  unlock();
  
  // Frequency response GUI
  m_freqResp = new MJ_FreqRespGui(stages, this, "eqgraph");

  resizeFreqResp(stages);
  setActiveBand(1);

  m_freqResp->setBackgroundColor(Qt::white);
  m_freqResp->setMinimumHeight(100);

  connect(m_freqResp, SIGNAL(mousePress(QMouseEvent *, float, float)), 
	  this, SLOT(grMousePress(QMouseEvent *, float, float)));

  connect(m_freqResp, SIGNAL(mouseMove(QMouseEvent *, float, float)), 
	  this, SLOT(grMousePress(QMouseEvent *, float, float)));

  connect(m_freqResp, SIGNAL(mouseRelease(QMouseEvent *, float, float)), 
	  this, SLOT(grMouseRelease(QMouseEvent *, float, float)));

  m_layout->addMultiCellWidget(m_freqResp, rows, rows, 0, cols - 1);
  m_layout->setRowStretch(rows, 90);

  updateEq();

  m_freqResp->show();

  if(drawSliders) {
    m_sliderBase = new QWidget(this);
    
    rows++;
    m_layout->addMultiCellWidget(m_sliderBase, rows , rows, 0, cols-1);
    m_layout->addRowSpacing(rows, 100);
    
    // Create sliders etc:
    rebuildSliders();  
    m_sliderBase->show();
  }
}

void MJ_ModuleEq2_QG::updateEq()
{
  const char * fname = "MJ_ModuleEq2_QG::updateEq";
  
  if(m_debug) trace(fname);

  uint i;

  MJ_Array<MJ_FilterIIR2<MJ_Sample> > iirs;

  MJ_ModuleEq2 *eq = (MJ_ModuleEq2 *) m_module;

  lock();

  uint stages = eq->stages();

  unlock();

  iirs.resize(stages);

  // Get all EQ curves:

  lock();

  double sr = eq->dspSamplingRate();

  for(i=0; i < stages; i++) {
    
    MJ_Eq2FilterParams *fp = & eq->parameters().get(i);

    MJ_FilterIIR2<MJ_Sample> *iir = & iirs.get(i);

    if(!i) {
      iir->designLowShelve2(fp->targetCenterHz(),
			    sr,
			    fp->targetGainDb(),
			    fp->targetBandWidthOctaves());
    }
    else if(i == (stages - 1)) {
      iir->designHighShelve2(fp->targetCenterHz(),
			     sr,
			     fp->targetGainDb(),
			     fp->targetBandWidthOctaves());
    }
    else {
      iir->designParametricB2(fp->targetCenterHz(),
			      sr,
			      fp->targetGainDb(),
			      fp->targetBandWidthOctaves());
    }
    unlock();
    lock(); // Give DSP thread chance to execute
  }
  unlock();
  
  for(i=0; i < stages; i++) {
    MJ_FilterIIR2<MJ_Sample> *iir = & iirs.get(i);
    
    m_freqResp->setNthFilter(i,
			     iir->bptr(), 2,
			     iir->aptr(), 2);
  }
  m_freqResp->calculateSum();

  if(m_debug) trace(fname, "EXIT");
}

void MJ_ModuleEq2_QG::updateBand(uint band)
{
  if(m_debug) trace("MJ_ModuleEq2_QG::updateBand");

  MJ_FilterIIR2<MJ_Sample> iir;
  
  MJ_ModuleEq2 *eq = (MJ_ModuleEq2 *) m_module;
  
  uint stages = m_freqResp->curves() - 1;

  lock();  

  double sr = eq->dspSamplingRate();
  MJ_Eq2FilterParams *fp = & eq->parameters().get(band);

  if(!band) {
    iir.designLowShelve2(fp->targetCenterHz(),
			  sr,
			  fp->targetGainDb(),
			  fp->targetBandWidthOctaves());
  }
  else if(band == (stages - 1)) {
    iir.designHighShelve2(fp->targetCenterHz(),
			   sr,
			   fp->targetGainDb(),
			   fp->targetBandWidthOctaves());
  }
  else {
    iir.designParametricB2(fp->targetCenterHz(),
			   sr,
			   fp->targetGainDb(),
			   fp->targetBandWidthOctaves());
  }
  unlock();

  MJ_GraphicPolyLine *curves = m_freqResp->curve();

  m_freqResp->undrawGraphic(&curves[band]);
  m_freqResp->undrawGraphic(&curves[stages]);

  m_freqResp->setNthFilter(band,
			   iir.bptr(), 2,
			   iir.aptr(), 2);

  m_freqResp->calculateSum();

  m_freqResp->drawAll();

  if(m_debug) trace("MJ_ModuleEq2_QG::updateBand # exit");
}

void MJ_ModuleEq2_QG::seekActiveBand(float x, float y)
{
  (void) y;

  float freqHz = posToFreq(x);
  
  MJ_ModuleEq2 *eq = (MJ_ModuleEq2 *) m_module;

  lock();
  m_activeBand = eq->closestBand(freqHz);
  unlock();
}

void MJ_ModuleEq2_QG::setPeak(float x, float y)
{
  float freqHz = posToFreq(x);
  // qDebug("freqHz = %f, freqBase = %f", freqHz, logFreq);
  
  setBand(freqHz, y);
  
  // updateEq();
  updateBand(m_activeBand);
  // m_freqResp->redraw();
  
  // trace(" MJ_ModuleEq2_QG::setPeak # EXIT");
}

float MJ_ModuleEq2_QG::posToFreq(float x)
{
  MJ_ModuleEq2 *eq = (MJ_ModuleEq2 *) m_module;
  
  lock();
  float nyquist = eq->dspSamplingRate() * 0.5;
  unlock();

  float logMax = MJ_log(nyquist);
  float logMin = MJ_log(nyquist * 0.001);
  
  float freqBase = (x - 0.001) / (0.999);

  float logFreq = freqBase * (logMax - logMin) + logMin;

  return expf(logFreq);  
}

void MJ_ModuleEq2_QG::setBand(double freqHz, double gainDb)
{
  char bufGain[128], bufHz[128];

  MJ_ModuleEq2::formMsgId(bufGain, 0, m_activeBand);
  MJ_ModuleEq2::formMsgId(bufHz, 1, m_activeBand);

  m_blockDraw = true;
  setValue(bufHz, freqHz);
  setValue(bufGain, gainDb);
  m_blockDraw = false;

  update2(bufGain);
  update2(bufHz);
}

void MJ_ModuleEq2_QG::rebuildSliders()
{
  const char * fname = "MJ_ModuleEq2_QG::rebuildSliders";
  
  if(m_debug) trace(fname);

  m_objects.clear();

  if(m_sliderWindow) {
    delete m_sliderWindow;
    m_sliderWindow = 0;
  }
  
  MJ_ModuleEq2 *eq = (MJ_ModuleEq2 *) m_module;

  lock();
  uint stages = eq->stages();
  unlock();

  m_sliderWindow = new QWidget(m_sliderBase);

  if(m_sliderBaseLayout) delete m_sliderBaseLayout;

  m_sliderBaseLayout = new QGridLayout(m_sliderBase, 1, 1);

  m_sliderBaseLayout->addWidget(m_sliderWindow, 0, 0);

  QGridLayout *layout = new QGridLayout(m_sliderWindow, stages + 1, 7);
  MJ_equalStretch(layout, 10, 10);
  layout->setColStretch(1, 50);
  layout->setColStretch(4, 30);
  
  QString str1, str2;

  char buf[128];

  m_guiComponents.resize(stages * 3);

  for(uint i=0; i < stages; i++) {

    lock();
    MJ_Eq2FilterParams *fp = & eq->parameters().get(i);
    MJ_Sample gainDb = fp->targetGainDb();
    MJ_Sample centerHz = fp->targetCenterHz();
    MJ_Sample bandWidthOctaves = fp->targetBandWidthOctaves();
    unlock();

    // Center frequency:

    // Label is used to identify the type
    if(!i) str1 = tr("Bass");
    else if(i == stages - 1) str1 = tr("Treble");
    else {
      str1 = tr(MJ_Strs::Band);
      str2.setNum(i);
      str1 += " " + str2;
    }
    
    MJ_GuiDoubleSlider *gds = new MJ_GuiDoubleSlider;
    gds->createObjs(m_sliderWindow);
    gds->slider()->setMapper(new MJ_MapperLog);
    gds->set(centerHz, 20.0, 16500.0);
    gds->label()->setText(str1);
    gds->toGridHorizontal(layout, i+1, 0);

    MJ_ModuleEq2::formMsgId(buf, 1, i);
    gds->setId(buf);
 
    connect(gds, SIGNAL(emitValue2(const char *,double)),
	    this, SLOT(setValue(const char *,double)));
    connect(gds, SIGNAL(emitGetValue2(const char *,double *)),
	    this, SLOT(getValue(const char *,double *)));
    connect(gds, SIGNAL(emitDone2(const char *)),
	    this, SLOT(doneValue(const char *)));
    m_objects.append(gds);

    gds->show();
    m_guiComponents.get(i*3 + 1) = gds;

    // Gain:
    MJ_ModuleEq2::formMsgId(buf, 0, i);
    MJ_Slider *s = new MJ_Slider(-15, 15.0, gainDb, 
				 m_sliderWindow, buf);
    MJ_CStrRemap *map = new MJ_CStrRemap(buf);

    connect(s, SIGNAL(valueChangedDouble(double)), 
	    map, SLOT(callDouble(double)));
    connect(s, SIGNAL(emitDone()), 
	    map, SLOT(call()));

    connect(map, SIGNAL(emitCallDouble(const char *,double)), 
	    this, SLOT(setValue(const char *,double)));
    connect(map, SIGNAL(emitCall(const char *)), 
	    this, SLOT(doneValue(const char *)));
    layout->addWidget(s, i+1, 4);
    m_objects.append(map);

    m_guiComponents.get(i*3 + 0) = s;
    s->show();

    // Bandwidth:

    MJ_ModuleEq2::formMsgId(buf, 2, i);
    s = new MJ_Slider(.03, 2.0, bandWidthOctaves, m_sliderWindow, buf);
    s->setMapper(new MJ_MapperLog);

    map = new MJ_CStrRemap(buf);

    connect(s, SIGNAL(valueChangedDouble(double)), 
	    map, SLOT(callDouble(double)));
    connect(s, SIGNAL(emitDone()), 
	    map, SLOT(call()));

    connect(map, SIGNAL(emitCallDouble(const char *,double)), 
	    this, SLOT(setValue(const char *,double)));
    connect(map, SIGNAL(emitCall(const char *)), 
	    this, SLOT(doneValue(const char *)));
    layout->addWidget(s, i+1, 6);
    m_objects.append(map);

    m_guiComponents.get(i*3 + 2) = s;
    s->show();

  }

  if(m_debug) trace(fname, "final");

  m_sliderWindow->show();
}

void MJ_ModuleEq2_QG::resizeFreqResp(uint stages)
{
  if(m_debug) trace("MJ_ModuleEq2_QG::resizeFreqResp", stages);

  m_freqResp->resizeFilters(stages);
  
  for(uint i=0; i < stages; i++)
    m_freqResp->setNthCurveGC(i, new MJ_GraphicsContextPen(Qt::gray));
   
   m_freqResp->setNthCurveGC
     (stages, new MJ_GraphicsContextPen(QPen(Qt::blue, 2)));
}

void MJ_ModuleEq2_QG::setActiveBand(uint band)
{
  if(m_debug) trace("MJ_ModuleEq2_QG::setActiveBand", band);

  if(band >= m_freqResp->curves()) {
    MJ_fatal(MJ_ERR_INTERNAL, "MJ_ModuleEq2_QG::setActiveBand");
  }

  m_freqResp->setNthCurveGC(m_activeBand, 
			    new MJ_GraphicsContextPen(Qt::gray));  
  m_freqResp->setNthCurveGC(band, 
			    new MJ_GraphicsContextPen(Qt::red));
  m_freqResp->sumToTop();
  m_freqResp->bringToTop(band);
  m_activeBand = band;
}
