/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include "mj_module_high_pass.h"

#include <mj_parameters.h>
#include <mj_strs.h>

MJ_ModuleHighPass::MJ_ModuleHighPass()
  : m_cutOffHz(20.0),
    m_q(0.5)
{}

MJ_ModuleHighPass::~MJ_ModuleHighPass()
{}

bool MJ_ModuleHighPass::dspPrepare()
{  
  int in = dspInChannels();

  if(in < 1 || in != dspOutChannels())
    return false;

  m_filters.resize(in);  

  updateFilterCoeffs();

  clearFilterMemories();

  return true;
}

void MJ_ModuleHighPass::dspProcessPeriod()
{
  MJ_SignalSource *sinput = m_signalInput.data();
  const uint samples = sinput[0].samples().size();

  MJ_SignalSink *soutput = m_signalOutput.data();
  const uint chans = m_signalOutput.size();

  for(uint c=0; c < chans; c++) {

    Filter filt = m_filters[c];
    const MJ_Sample * in = sinput[c].samples().data();
    const MJ_Sample * sentinel = in + samples;
    MJ_Sample * out = soutput[c].samples().data();

    while(in < sentinel)
      *out++ = filt.putGetSample(*in++);
  }
  
}

bool MJ_ModuleHighPass::msgProcess(const char *id, MJ_Message *m)
{
  bool ok = true;
  
  double v = m->toDouble(&ok);

  if(!ok) return false;
  
  if(strcmp(id, MJ_Strs::frequency) == 0)
    m_cutOffHz = MJ_clamp(v, 5.0, 16500.0);
  else if(strcmp(id, "q") == 0)
    m_q = MJ_clamp(v, 0.1, 4.0);
  else 
    return false;
  
  updateFilterCoeffs();

  return true;
}

MJ_Parameter * MJ_ModuleHighPass::msgGetInput(const char *id)
{
  if(strcmp(id, MJ_Strs::frequency) == 0)
    return new MJ_ParameterDouble(MJ_Strs::Frequency, id, m_cutOffHz, 
			       5, 16500, MJ_UNIT_HERTZ);
  else if(strcmp(id, "q") == 0)
    return new MJ_ParameterDouble("Q-value", id, m_q, 
			       0.1, 4.0);

  return 0;
}

QCString MJ_ModuleHighPass::msgGetInputName(int i)
{
  QCString r;

  if(i == 0)
    r = MJ_Strs::frequency;
  else if(i == 1)
    r = "q";

  return r;
}

void MJ_ModuleHighPass::updateFilterCoeffs()
{
  MJ_Sample sr = dspSamplingRate();

  for(uint i=0; i < m_filters.size(); i++)    
    m_filters[i].designLowPass2(m_cutOffHz, sr, m_q);
}

void MJ_ModuleHighPass::clearFilterMemories()
{
  for(uint i=0; i < m_filters.size(); i++)
    m_filters[i].resetMemory();
}
