/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

/* Include the header of this file. */ 
#include <mj_module_example1.h>

/* Include header of parameter settings. */
#include <mj_parameters.h>

/* Include plugin data base declarations. */
#include <mj_dsp_plugin.h>

/* Include globally accessible character strings. */
#include <mj_strs.h>

/* Needed to control the version number of the plugin. */
#include <mj_version.h>

bool MJ_ModuleLowPass1::m_debug = true;

MJ_ModuleLowPass1::MJ_ModuleLowPass1()
  : m_cutOffHz(1000.0) // Default to 1000 Hertz.
{
  /* Make one input and one output channels by default. */
  m_signalSourceDefs.resize(1);
  m_signalOutput.resize(1);
}

/* The descturctor does not really need to do anything. All member
   variables destroy themselves. */
MJ_ModuleLowPass1::~MJ_ModuleLowPass1()
{}

bool MJ_ModuleLowPass1::dspPrepare()
{
  /* Clear filer memory. */
  m_filter.resetMemory();
  
  /* The sampling rates may have changed since we last did any
     processing. Thus we update filter coefficients to make sure
     nothing is lost */
  updateFilterCoeffs();

  /* Check that the number of input and output channels equals 1. */
  if((m_signalInput.size() != 1) || (m_signalOutput.size() != 1)) 
    return false;

  /* Check the number of samples in input and output buffers is the same. */
  if(m_signalInput[0].samples().size() != m_signalOutput[0].samples().size())
    return false;

  return true;
}

void MJ_ModuleLowPass1::dspProcessPeriod()
{
  // Get input data source
  const MJ_Vector<MJ_Sample> *in = & m_signalInput[0].samples();
  // Get destination of processed data
  MJ_Vector<MJ_Sample> *out = & m_signalOutput[0].samples();
  /* Process the data. The filter class has a method for optimized
     vector operation. Thus there is no need to loop over each sample
     in this code. */
  m_filter.putMany(out->data(), in->dataConst(), in->size());
}

bool MJ_ModuleLowPass1::msgProcess(const char * id, MJ_Message *m)
{
  const char *fname = "MJ_ModuleLowPass1::msgProcess";
  bool ok = true;

  /* Optionally trace program execution. */
  if(m_debug) trace(fname);
  
  if(!strcmp(id, MJ_Strs::frequency)) {
    /* Transform the message to double regardless of its type.  If the
       cast fails (eg. the message just cannot be turned to numeric
       value) the boolean argument is modified to false. */
    double tmp = m->toDouble(&ok);
    
    /* Optionally output a warning. */
    if(m_debug) {
      trace2("%s # Value = %f", fname, tmp);
    }
    
    if(ok) {
      if(m_debug) trace(fname, "Success 1");
      m_cutOffHz = tmp;
      updateFilterCoeffs();
    }
    else if(m_debug) 
      trace(fname, "Failure 1");
  }
  else ok = false;

  return ok;
}

MJ_Parameter * MJ_ModuleLowPass1::msgGetInput(const char * id)
{
  if(m_debug) trace("MJ_ModuleLowPass1::msgGetInput");
  
  if(!strcmp(id, MJ_Strs::frequency)) {
    /* This module hasd only one parameter (gain) that can be edited
       via the message interface. The following two commands declare
       the parameter and give its limiting values (minimum/maximum),
       plaintext name (for GUI) and units (of general interest, also
       for GUI). */
    MJ_ParameterDouble *pd = new MJ_ParameterDouble;
    pd->set(
	    /* Name */
	    "Cut-off frequency", 
	    MJ_Strs::frequency,
	    /* Current value of the parameter. */
	    m_cutOffHz, 
	    /* Minimum value */
	    0.001, 
	    /* Maximum value (20 kHz). */
	    20000.0, 
	    /* Unit */
	    MJ_UNIT_HERTZ);
    return pd;
  }
  return 0;
}

QCString MJ_ModuleLowPass1::msgGetInputName(int i)
{
  QCString s;
  if(i == 0) s = MJ_Strs::frequency;
  return s;
}

void MJ_ModuleLowPass1::updateFilterCoeffs()
{
  /* Check the input signal exists. If the signal does not exist, we
     do not know the sampling rate and there is no point in
     calculating filter coefficients. */
  if(!m_signalInput.size()) return;
  
  /* Get sample ratedefinition. */
  MJ_SampleRateDef &sr = m_signalInput[0].sink()->sampleRate();

  /* Local copy of cut-off frequency. This value may be modufied few
     lines later.*/
  double cutOffHz = m_cutOffHz;

  /* Calculate Nyquist frequency, the filter cut-off frequency must
     not exceed this value. */
  double nyquist = sr.rate() / 2.0;
  if(cutOffHz > nyquist) cutOffHz = nyquist;
  
  /* Normalize the cutoff-frequency to range 0-1. Needed for filter design. */
  double cutOffVal =  cutOffHz / nyquist;
  
  /* Update the filter coefficients. */
  m_filter.designLowPass(cutOffVal);
}

// End of signal processing-related code.

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


/* The next section contains code that is used to declare this as a
   plugin. Without this code the plugin cannot be loaded from the
   shared library.*/

/* Factorty function that is used to create new plugin instances. This
   needs to be normal function so we can get a function pointer to it.
 */
MJ_ModuleDsp * MJ_createNew_ModuleExample1() 
{
  return new MJ_ModuleLowPass1;
}


/* This function returns a new plugin data base entry. Upon startup,
   Mustajuuri opens this shared library and tries to find a function
   with this name. It should return new object of type
   MJ_PluginDbeDsp.
*/
MJ_PluginDbeDsp *MJ_createPlugin(int i)
{
  if(i != 0) return 0; // This function has laready been called once.
  // Create the entry.
  MJ_PluginDbeDsp *dbe = new MJ_PluginDbeDsp;

  // Set author of of this plugin.
  dbe->setAuthors("Ilmonen Tommi S.");

  /* Set menu position. This name may be translated one field at a
     time. Note that we use "_" as the field separator, not "/" (which
     would adhere better to de-facto standards). The character "_" is
     chosen since it is more seldom used in natural text while "/" is
     fairly common (audio I/O etc.).
  */
  dbe->setMenuPosition("_Audio_Equalizers_Monophonic low pass 1");

  /* Factory (or creator) function. Unless this is defined the
     database entry cannot fabricate new modules.  */
  dbe->setCreator(MJ_createNew_ModuleExample1);

  /* Unique identifier. This string is used to define module type when
     saving module definitions to a file. It must not contain
     spaces. It is probably a good idea to use name that can be read
     and has some meaning. */
  dbe->setIdentifier("_equalizers_monophonic_low_pass_1");

  /* Version number of this plugin (not compulsory). */
  dbe->pluginVersion()->set(0,0,1);

  /* Define how many input and outptu channels this plugin type can
     handle. One can be much more specific about the IO configuration,
     but this is the bare miniumum. */
  dbe->options()->defaultChans(1);
  
  /* This is the base name of the translation file. 
     
     Now people can make translation files like mj_example1_fi.po or
     mj_example1_de.po that are automatically loaded by the system
     upon startup */
  dbe->setParameter(MJ_PP_TRANSLATION_FILE, "mj_example1");
  
  /* We return the data-base entry. Eventually the hosting system will
     delete it. */
  
  return dbe;  
}
