/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

/* Include the header for this class. */ 
#include <mj_module_example2.h>

/* Error reporting. */
#include <mj_error.h>
#include <mj_messages.h>
#include <mj_satellite.h>

/* For non-linear sliders. */
#include <mj_mapper.h> 

/* Include header of parameter settings. */
#include <mj_parameters.h>

/* Convenient strings. */
#include <mj_strs.h>

/* Miscellanous mathematics functions. */
#include <mj_math.h>

/* Translation engine. */
#include <mj_tr.h>

static const char * MJ_MMT1_strs[4] = {
  MJ_Strs::tempo,
  MJ_Strs::gain,
  MJ_Strs::wetratio,
  MJ_Strs::tap
};

static const char * MJ_MMT1_params[3] = {
  MJ_Strs::time,
  MJ_Strs::gain,
  MJ_Strs::pan 
};

bool MJ_ModuleMultiTap1::m_debug = false;

MJ_ModuleMultiTap1::MJ_ModuleMultiTap1()
  : m_tempo(60.0),
    m_masterGain(1.0),
    m_wetRatio(.5),
    m_sr(0),
    m_relativeDelays(MJ_EIGHT),
    m_gainDefs(MJ_EIGHT),
    m_targetGainDefs(MJ_EIGHT),
    m_interpolationTimes(MJ_EIGHT),
    m_delays(MJ_EIGHT),
    m_gains(MJ_EIGHT),
    m_targetGains(MJ_EIGHT)
{
  uint i;

  /* Make one input and two output channels (default). */
  m_signalSourceDefs.resize(1);
  m_signalOutput.resize(2);


  for(i=0; i < MJ_EIGHT; i++) {
    m_relativeDelays[i] = (i+1) * 0.125;
  }
  MJ_DualSample tmp;
  tmp.m_v[0] = 0.5;
  tmp.m_v[1] = 0.0;

  m_targetGainDefs.setAll(tmp);
  m_gainDefs.setAll(tmp);
  m_delays.setAll(0);

  calculateGains();
}

MJ_ModuleMultiTap1::~MJ_ModuleMultiTap1()
{}

bool MJ_ModuleMultiTap1::dspPrepare()
{
  /* Check that we do have input signal and two outputs. */ 
  if((!m_signalInput.size()) || (m_signalOutput.size() != 2)) 
    return false;
  /* Check the number of samples in input and output buffers is the same. */
  if(m_signalInput[0].samples().size() != m_signalOutput[0].samples().size())
    return false;

  m_sr = (uint) m_signalOutput[0].sampleRate().rate();

  /* Allocate memory for the delay line. We must allocate enough
     memory to handle the longest possible delay times with grace
     (tempo = 40 bpm and maximum delay = 4 quarters).*/
  m_delayBuffer.resize((long) (m_sr * 60.0 / 40.0 * 4.0));

  m_delayBuffer.setAll(0.0);

  m_gainDefs = m_targetGainDefs;
  
  calculateDelays();
  calculateGains();
  m_interpolationTimes.setAll(0);
  return true;
}

void MJ_ModuleMultiTap1::dspProcessPeriod()
{
  const ulong *delays = m_delays.dataConst();
  MJ_DualSample *gains = m_gains.data();
  const MJ_DualSample *targetGains = m_targetGains.dataConst();
  
  MJ_Vector<MJ_Sample> *out1 = & m_signalOutput[MJ_STEREO_LEFT].samples();
  MJ_Vector<MJ_Sample> *out2 = & m_signalOutput[MJ_STEREO_RIGHT].samples();

  const MJ_Vector<MJ_Sample> *in1 = & m_signalInput[0].samples();
  const MJ_Vector<MJ_Sample> *in2;
  
  // Direct sound to output:

  const MJ_Sample directGain = (1.0 - m_wetRatio) * m_masterGain;
  
  MJ_Vector<MJ_Sample>::vmlBegin(); // Use VML acceleration if possible

  in1->mulVml(directGain, out1);

  if(m_signalInput.size() > 1) {
    in2 = & m_signalInput[1].samples();
    in2->mulVml(directGain, out2);
  }
  else {
    in2 = 0;
    in1->mulVml(directGain, out2);
  }

  MJ_Vector<MJ_Sample>::vmlEnd();
  
  /** Load variables to local variables. This effectively loads the
      stuff to registers, making for fewer (unnecessary) load/store
      events in the processing. 

      It does make code uglier, but also faster. Luckily this is
      the only place where such optimization is really needed. */
  
  const uint len = in1->size();
  uint i, j;

  MJ_Sample *rinfBuf = m_delayBuffer.data();
  const ulong mask = m_delayBuffer.mask();
  const ulong pos = m_delayBuffer.getSampleCount();

  // Fill the buffer with current audio:
  for(i=0; i < len; i++) {

    // Put the newest values to the buffer.
    MJ_Sample curr = in1->getFastConst(i);
    if(in2) curr += in2->getFastConst(i);
    rinfBuf[pos + i & mask] = curr;
  }

  // Do the delays. 

  // Loop over taps.
  for(j=0; j < MJ_EIGHT; j++) {

    const ulong delay = delays[j];
    ulong counter = pos - delay;
    
    MJ_Sample gainLeft  = gains[j].m_v[MJ_STEREO_LEFT];
    MJ_Sample gainRight = gains[j].m_v[MJ_STEREO_RIGHT];

    // Interpolate this tap if necessary.
    ulong interpSamplesTotal = m_interpolationTimes[j];
    
    ulong interpSamples = MJ_min(interpSamplesTotal, (ulong) len);

    if(interpSamplesTotal) {
      
      MJ_trace3(m_debug, "MJ_ModuleMultiTap1::dspProcessPeriod # %f  %f ", 
		gainRight, gainLeft);

      m_interpolationTimes[j] -= interpSamples;

      const MJ_Sample targetGainLeft  = targetGains[j].m_v[MJ_STEREO_LEFT];
      const MJ_Sample targetGainRight = targetGains[j].m_v[MJ_STEREO_RIGHT];

      MJ_Sample interpSamplesS = (MJ_Sample) interpSamplesTotal;

      const MJ_Sample leftGainRate = 
	(targetGainLeft - gainLeft) / interpSamplesS;
      const MJ_Sample rightGainRate = 
	(targetGainRight - gainRight) / interpSamplesS;
	
      for(i = 0; i < interpSamples; i++) {
	const MJ_Sample &s = rinfBuf[counter++ & mask];

	gainLeft += leftGainRate;
	gainRight += rightGainRate;

	out1->getFast(i) += s * gainLeft * m_wetRatio;
	out2->getFast(i) += s * gainRight * m_wetRatio;
      }

      gains[j].m_v[MJ_STEREO_LEFT] = gainLeft;
      gains[j].m_v[MJ_STEREO_RIGHT] = gainRight;
    }

    // Loop over samples.
    for(i=interpSamples; i < len; i++) {
      const MJ_Sample &s = rinfBuf[counter++ & mask];
      out1->getFast(i) += s * gainLeft * m_wetRatio;
      out2->getFast(i) += s * gainRight * m_wetRatio;
    }
  }
  
  // Advance the buffer.
  m_delayBuffer.setSampleCount(pos + (ulong) len);
}

bool MJ_ModuleMultiTap1::dspStop()
{
  m_delayBuffer.resize(0); // Free memory.
  return true;
}

bool MJ_ModuleMultiTap1::msgProcess(const char * id, MJ_Message *m)
{
  const char *fname = "MJ_ModuleMultiTap1::msgProcess";
  bool ok = true;

  /* Optionally trace program execution. */
  if(m_debug) trace(fname, id);

  double tmp = m->toDouble(&ok);
  if(!ok) return false;

  bool realTime = false;

  if(m_currentTime)
    realTime = m_currentTime->isRealTime();

  ulong interpolationTime(0);

  if(realTime) {
    if((long)m->timeStamp() < (long)m_currentTime->stamp())
      interpolationTime = (ulong) (m_sr / 8 + 1); // 1/8 second interpolation
    else 
      interpolationTime = m->timeStamp() - m_currentTime->stamp();
  }

  if(m_debug) trace(fname, (int) interpolationTime);

  int mainType;
  int tapIndex;
  int tapParam;

  if(!determineType(id, &mainType, &tapIndex, &tapParam))
    return false;


  if(mainType == MJ_MT_TEMPO) {
     m_tempo = tmp;
     calculateDelays();
  }
  else if(mainType == MJ_MT_MASTER_GAIN) {
    m_masterGain = MJ_dbToAbs(tmp);
    calculateGains();
  }
  else if(mainType == MJ_MT_WET_RATIO) {
    trace3(m_debug, "%s # Got new wet-ratio: %f", fname, tmp);
    m_wetRatio = tmp;
  }
  else if(tapParam == MJ_MTP_TIME) {
    // Delay times.
    m_relativeDelays[tapIndex] = tmp;
    calculateDelay(tapIndex);
  }
  else if(tapParam == MJ_MTP_GAIN) {
    // Gain.
    m_targetGainDefs[tapIndex].m_v[0] = tmp;

    if(realTime) {
      m_interpolationTimes[tapIndex] = interpolationTime;
    }
    else {
      m_gainDefs[tapIndex].m_v[0] = tmp;
    }
    calculateGain(tapIndex);
  }
  else if(tapParam == MJ_MTP_PAN) {
    // Panning.

    m_targetGainDefs[tapIndex].m_v[1] = tmp;

    if(realTime) {
      m_interpolationTimes[tapIndex] = interpolationTime;
    }
    else {
      m_gainDefs[tapIndex].m_v[1] = tmp;
    }
    calculateGain(tapIndex);
  }
  else ok = false;

  return ok;
}

MJ_Parameter * MJ_ModuleMultiTap1::msgGetInput(const char *id)
{
  if(m_debug) trace("MJ_ModuleMultiTap1::msgGetInput", id);

  QString tmp, tmp2;

  int mainType;
  int tapIndex;
  int tapParam;

  if(!determineType(id, &mainType, &tapIndex, &tapParam))
    return 0;

  if(mainType == MJ_MT_TEMPO) { 
    /* This module hasd only one parameter (gain) that can be edited
       via the message interface. The following two commands declare
       the parameter and give its limiting values (minimum/maximum),
       plaintext name (for GUI) and units (of general interest, also
       for GUI). */
    MJ_ParameterDouble *pd = new MJ_ParameterDouble;
    pd->set(
	    /* Name */
	    MJ_Strs::Tempo, 
	    id,
	    /* Current value of the parameter. */
	    m_tempo, 
	    /* Minimum value */
	    30.0, 
	    /* Maximum value. */
	    350.0, 
	    /* Unit */
	    MJ_UNIT_BPM);
    // Logarithmic scale
    pd->setMapper(new MJ_MapperLog);
    return pd;
  }
  else if(mainType == MJ_MT_MASTER_GAIN) {
    MJ_ParameterDouble *pd = new MJ_ParameterDouble;

    pd->set(
	    /* Name */
	    "Master gain", 
	    MJ_Strs::gain,
	    /* Current value of the parameter. */
	    MJ_absToDb(m_masterGain), 
	    /* Minimum value */
	    -20.0, 
	    /* Maximum value. */
	    6.0, 
	    /* Unit */
	    MJ_UNIT_DECIBEL);
    return pd;
    
  }
  else if(mainType == MJ_MT_WET_RATIO) {
    MJ_ParameterDouble *pd = new MJ_ParameterDouble;
    pd->set(
	    /* Name */
	    "Wet ratio", 
	    id,
	    /* Current value of the parameter. */
	    m_wetRatio, 
	    /* Minimum value */
	    0.0, 
	    /* Maximum value. */
	    1.0, 
	    /* Unit */
	    MJ_UNIT_BLANK);
    return pd;
    
  }
  else if(tapParam == MJ_MTP_TIME) {
    // Delay times.

    tmp = MJ_Tr::tr("Delay time");
    tmp2.sprintf(" %d", (int) tapIndex + 1);
    tmp += tmp2;

    MJ_ParameterDouble *pd = new MJ_ParameterDouble;
    
    pd->set(
	    /* Name */
	    tmp, 
	    id,
	    /* Current value of the parameter. */
	    m_relativeDelays[tapIndex], 
	    /* Minimum value */
	    0.0, 
	    /* Maximum value  */
	    4.0, 
	    /* Unit */
	    MJ_UNIT_QUARTERS);
    return pd;
  }
  else if(tapParam == MJ_MTP_GAIN) {
    // Gain.

    MJ_ParameterDouble *pd = new MJ_ParameterDouble;
    
    tmp = MJ_Tr::tr("Gain");
    tmp2.sprintf(" %d", (int) tapIndex + 1);
    tmp += tmp2;

    pd->set(
	    /* Name */
	    tmp, 
	    id,
	    /* Current value of the parameter. */
	    m_targetGainDefs[tapIndex].m_v[0], 
	    /* Minimum value */
	    0.0, 
	    /* Maximum value  */
	    1.0, 
	    /* Unit */
	    MJ_UNIT_BLANK);
    return pd;
  }
  else if(tapParam == MJ_MTP_PAN) {
    // Panning.
    
    MJ_ParameterDouble *pd = new MJ_ParameterDouble;
    
    tmp = MJ_Tr::tr("Pan");
    tmp2.sprintf(" %d", (int) tapIndex + 1);
    tmp += tmp2;

    pd->set(
	    /* Name */
	    tmp, 
	    id,
	    /* Current value of the parameter. */
	    m_targetGainDefs[tapIndex].m_v[1], 
	    /* Minimum value */
	    -1.0, 
	    /* Maximum value  */
	    1.0, 
	    /* Unit */
	    MJ_UNIT_BLANK);
    return pd;
  }

  return 0;
}


QCString MJ_ModuleMultiTap1::msgGetInputName(int i)
{
  QCString r;
  char buf[64];

  uint ui = i;

  uint taps = m_gains.size();

  if(ui < MJ_MT_SIZEOF)
    r = MJ_MMT1_strs[i];
  else {
    ui -= MJ_MT_SIZEOF;

    uint tap = ui / 3;
    uint param = ui - tap * 3;

    if(tap < taps) {
      describeTap(buf, tap, param);
      r = buf;
    }
  }
  return r;
}

uint MJ_ModuleMultiTap1::closestTap(float timeQuarters)
{
  uint taps = m_relativeDelays.size();
  float minDiff = 1.0e+10;
  uint bestInd = 0;

  for(uint i=0; i < taps; i++) {
    float tmp = (float) MJ_abs(timeQuarters - m_relativeDelays.getConst(i));
    if(tmp < minDiff) {
      minDiff = tmp;
      bestInd = i;
    }
  }

  return bestInd;
}

void MJ_ModuleMultiTap1::getTapParams
(uint tap, float *quarters, float *leftGain, float *rightGain) const
{
  *quarters = m_relativeDelays.getConst(tap);

  const MJ_DualSample * gains = & m_targetGains.getConst(tap);

  *leftGain = gains->m_v[0];
  *rightGain = gains->m_v[1];
}  

void MJ_ModuleMultiTap1::describeTap(char * buf, int index, int param)
{
  sprintf(buf, "tap_%d_%s", index, MJ_MMT1_params[param]);
}

bool MJ_ModuleMultiTap1::determineType(const char *address, 
				       int *mainType, 
				       int *index,
				       int *tapParam)
{
  const char * fname = "MJ_ModuleMultiTap1::determineType";

  int idMain = MJ_findStr2(address, MJ_MMT1_strs, 4);

  if(idMain == -1) {
    if(m_debug) MJ_error(MJ_ERR_ARGUMENT, fname, 1);
    return false;
  }

  bool ok = true;

  *mainType = idMain;

  if(idMain < MJ_MT_SIZEOF)
    return true;

  QCString str(address);
  
  *index = MJ_fieldC(str, 1, "_").toInt(&ok);

  QCString str2(MJ_fieldC(str, 2, "_"));

  *tapParam = MJ_findStr(str2, MJ_MMT1_params, MJ_MTP_SIZEOF);

  if(*tapParam == -1 || !ok) {
    if(m_debug) MJ_error(MJ_ERR_ARGUMENT, fname, 2);
    return false;
  }

  MJ_trace3(m_debug, "%s # %s # %d  %d  %d", 
	    fname, address, *mainType, *index, *tapParam);

  return ok;
}


void MJ_ModuleMultiTap1::calculateDelays()
{
  uint i, lim = m_relativeDelays.size();
  for(i=0; i < lim; i++) calculateDelay(i);
}  

void MJ_ModuleMultiTap1::calculateGain(uint tap)
{ 
  MJ_DualSample &g1 = m_gainDefs.getFast(tap); 

  m_gains.getFast(tap).m_v[MJ_STEREO_LEFT]  = 
    g1.m_v[0] * (1.0 + g1.m_v[1]) * 0.5 * m_masterGain;

  m_gains.getFast(tap).m_v[MJ_STEREO_RIGHT] = 
    g1.m_v[0] * (1.0 - g1.m_v[1]) * 0.5 * m_masterGain;


  MJ_DualSample &g2 = m_targetGainDefs.getFast(tap); 

  m_targetGains.getFast(tap).m_v[MJ_STEREO_LEFT]  = 
    g2.m_v[0] * (1.0 - g2.m_v[1]) * 0.5 * m_masterGain;

  m_targetGains.getFast(tap).m_v[MJ_STEREO_RIGHT] = 
    g2.m_v[0] * (g2.m_v[1] + 1.0) * 0.5 * m_masterGain;

  MJ_trace3(m_debug, "MJ_ModuleMultiTap1::calculateGain # %d # %f  %f", 
	    tap, m_targetGains.getFast(tap).m_v[MJ_STEREO_LEFT], 
	    m_targetGains.getFast(tap).m_v[MJ_STEREO_RIGHT]);
}  

void MJ_ModuleMultiTap1::calculateGains()
{
  uint i, lim = m_gainDefs.size();
  for(i=0; i < lim; i++) calculateGain(i);
}


