/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#ifndef MJ_MODULE_EXAMPLE2_H
#define MJ_MODULE_EXAMPLE2_H

/* Include base definitions. */
#include <mj_module_dsp.h>

/* Multisample data types. Used to get two-sample (stereo) data type. */
#include <mj_multi_sample.h>
/* Ring buffer for the delay. */
#include <mj_ringbuffer.h>


#define MJ_EIGHT 8

/** Eight-tap delay plugin. 

   In future the number of taps may be dynamic.

   The user can specify the tempo and the delays in relation to
   tempo. Therefore it is easy to create rhythmic patterns that can be
   scaled to different tempos with ease. */
class MJ_ModuleMultiTap1 : public MJ_ModuleDsp
{
public:

  /** The base parameters (not the taps). */
  enum MJ_MT_Param {
    MJ_MT_TEMPO,
    MJ_MT_MASTER_GAIN,
    MJ_MT_WET_RATIO,
    MJ_MT_SIZEOF
  };
  
  /** Parameters associated to each tap. */
  enum MJ_MTP_TapParam {
    MJ_MTP_TIME,
    MJ_MTP_GAIN,
    MJ_MTP_PAN,
    MJ_MTP_SIZEOF
  };

  MJ_ModuleMultiTap1();
  virtual ~MJ_ModuleMultiTap1();

  /** This method is called by the host module before any signal
     processing takes place. Initializations can be performed here. */
  virtual bool dspPrepare();

  /** This method does the actual processing. The parent module calls this
     method repeatedly. Any practical plugin should redefine this method. */
  virtual void dspProcessPeriod();

  /** This method is called as reawl-time processing dspStops.  It frees
     working memory. This is needed because the module might use
     siginificant amounts of memory. If multiple setups are loaded to
     system memory the totawl memory consumption may be a problem. */
  virtual bool dspStop();

  /** This method is used to receive input messages (or events).  The
     various child classes of MJ_Message make it possible to pass
     any information via this interface. */
  virtual bool msgProcess(const char *, MJ_Message *);

  /** The types of input messages are declared here. It is hoped all
     modules would tell all of their input message types with this
     method.  */
  virtual MJ_Parameter * msgGetInput(const char *);
  
  /** Returns the name of the parameter (assuming the plugin has linear
     parameter space). */
  virtual QCString msgGetInputName(int);

  /** Returns tap parameters. */

  void getTapParams
  (uint tap, float *quarters, float *leftGain, float *rightGain) const;
  
  /** Returns the tap that is closest to given time. */
  uint closestTap(float timeQuarters);

  /* Returns the parameter number of a certain parameter. */
  static int paramIndex(int type, int index)
  { 
    if(!type) return index;
    return (type -1) * MJ_EIGHT + MJ_MT_SIZEOF + index ; 
  }

  /* Returns pointer to the debugging variable. Now the host system
     can turn on the debugging of this class. */
  static bool * debugPtr() { return & m_debug; }

  /** A utility method that is used to transform identifier string to
      integer type description of the parameter. */

  static bool determineType(const char *address, 
			    int *mainType, 
			    int *index,
			    int *tapParam);
  
  static void describeTap(char * buf, int index, int param);

protected:
  
  /** Calculate delay time in samples. Calculate the real delay time
      based on the tempo, relative delay time and sampling rate. */
  void calculateDelay(uint tap)
  { m_delays[tap] = (ulong) 
      (m_relativeDelays[tap] * (60.0 / m_tempo) * (double) m_sr); }

  /// Calculate all delay line lengths.
  void calculateDelays();

  /** Calculates the stereo gain to be used by the DSP. The
      calculation is based on tap gain and pan and master gain. */
  void calculateGain(uint tap);
  
  /// Calculate all gains.
  void calculateGains();


  /** Tempo, bpm.*/
  double            m_tempo;
  double            m_masterGain;
  /** Wet to dry ratio. Range 0-1.*/
  MJ_Sample         m_wetRatio;
  /** Sampling rate. */
  uint              m_sr;
  /** Relative delay times. */
  MJ_Array<double>  m_relativeDelays;

  /** Gain parameters. First value is the gain [0,1] and the second is
      panning [0,1]. Pan zero = full left, one = full right. */
  MJ_Array<MJ_DualSample>   m_gainDefs;

  MJ_Array<MJ_DualSample>   m_targetGainDefs;
  
  /** How long does it take to interpolate everything. */
  MJ_Array<ulong>           m_interpolationTimes;
  
  /** Actual delay times in samples. */
  MJ_Array<ulong>           m_delays;

  /**  Stereo gains for the delay taps. */
  MJ_Array<MJ_DualSample>   m_gains; 

  /**  Stereo target gains for the delay taps. */
  MJ_Array<MJ_DualSample>   m_targetGains; 

  /** Delay buffer for audio. */
  MJ_RingBuffer2<MJ_Sample> m_delayBuffer;

  static bool m_debug;
};

#endif
