/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_example2_gui.h>

#include <mj_error.h>
#include <mj_parameters.h>

/* Miscellanous GUI helps. */
#include <mj_gui_utils.h>
#include <mj_slider.h>

/* For non-linear sliders. */
#include <mj_mapper.h> 

/* Get the menu bar declarations. */
#include <mj_module_menubar.h> 

/* Needed to control the version number of the plugin. */
#include <mj_version.h>

/* Include plugin data base declarations. */
#include <mj_dsp_plugin.h>

/* Translation engine. */
#include <mj_tr.h>

/* Graphics widget to visualize the delay setup. */
#include <mj_graphics_area.h>
/* Graphics axes object. */
#include <mj_graphic_axes.h>
/* Polyline object. */
#include <mj_graphic_polyline.h>
/* Graphics context so we can change visual properties. */
#include <mj_graphics_context_simple.h>

#include <qlayout.h> 
#include <qlabel.h> 

#include <strings.h>

MJ_ModuleMultiTap1_QG::MJ_ModuleMultiTap1_QG(MJ_RtControl *rtparent,
					     MJ_ModuleDsp *module,
					     QWidget      *parent,
					     const char   *name)
  : MJ_ModuleDspGuiPostCard(rtparent, module, false,
			    (uint) -1, 
			    MJ_GUI_OWN_WINDOW,
			    parent, name),
  m_multiTap(0),
  m_graphs(0),
  m_activeTap(-1)
{
  m_remaps.setAutoDelete(true);

  /* The dynamic cast makes shure we have the right kind of module. 
    
     It should not be necessary (the system guarantees the module is
     of the right kind), but extra safety checks seldom harm anybody. */
  m_multiTap = dynamic_cast<MJ_ModuleMultiTap1 *> (module);

  // Now draw the contents:
  draw();
}


MJ_ModuleMultiTap1_QG::~MJ_ModuleMultiTap1_QG()
{
  /* Since Qt deletes all child widgets there is little we need to do. */
}

void MJ_ModuleMultiTap1_QG::moduleModified(const char *id)
{
  MJ_ModuleDspGuiPostCard::moduleModified(id);
  
  int type, tap, param;

  MJ_ModuleMultiTap1::determineType(id, &type, &tap, &param);
  
  if(type >= MJ_ModuleMultiTap1::MJ_MT_SIZEOF) {
    updateGraph(tap);
  }
}

int MJ_ModuleMultiTap1_QG::weight()
{
  return 60;
}

MJ_ModuleDspGui *MJ_ModuleMultiTap1_QG::createNew(MJ_RtControl *rtparent,
						  MJ_ModuleDsp *module,
						  QWidget *parent, 
						  const char *name)
{ 
  return new MJ_ModuleMultiTap1_QG(rtparent, module, parent, name);
}

void MJ_ModuleMultiTap1_QG::grMousePress(QMouseEvent *e, float x, float y)
{
  (void) y;

  if(x < 0.0) x = 0.0;
  else if(x > 4.0) x = 4.0;

  if(!e->state()) {
    lock();
    m_activeTap = m_multiTap->closestTap(x);
    unlock();
  }

  char id[64];

  MJ_ModuleMultiTap1::describeTap
    (id, m_activeTap, MJ_ModuleMultiTap1::MJ_MTP_TIME);

  m_delays[m_activeTap].setValue(x);
  // updateGraph(m_activeTap);
  setValue(id, x);
}

void MJ_ModuleMultiTap1_QG::update2(const char *id)
{
  double val;
  // const char *fname = "MJ_ModuleMultiTap1_QG::update2";

  // Gets current value of the variable.
  getValue(id, & val);

  // qDebug("%s %s %f", fname, id, val);

  /* Put the edit parameter here so it will not cause further satellite
     calls. */
  const QCString tmp(m_editParameter);
  m_editParameter = id;
  
  int type, tap, param;

  MJ_ModuleMultiTap1::determineType(id, &type, &tap, &param);
  
  if(type < MJ_ModuleMultiTap1::MJ_MT_SIZEOF) {
    m_generic[type].setValue(val);
  }
  else {

    if(param == MJ_ModuleMultiTap1::MJ_MTP_TIME)
      m_delays[tap].setValue(val);
    else if(param == MJ_ModuleMultiTap1::MJ_MTP_GAIN)
      m_gains[tap]->setValueDouble(val);
    else if(param == MJ_ModuleMultiTap1::MJ_MTP_PAN)
      m_pans[tap]->setValueDouble(val);
    
    updateGraph(tap);    
  }

  m_editParameter = tmp;
}


void MJ_ModuleMultiTap1_QG::draw()
{
  uint i;
  QString str;

  const char * fname = "MJ_ModuleMultiTap1_QG::draw";

  /* First get the current state of the module. We need to lock the
     object to guarantee the data is always valid. */
  lock();
  m_multiTap->msgGetInputs(&m_parameters);
  unlock();

  /* Create a grid layout to which the GUI objects will be put. */
  QVBoxLayout *layout = new QVBoxLayout(this, 5);

  /* If there is no parent (= this is a stand-alone window) make a
     menu bar. After these two lines of code the window has a menubar
     with normal "File" and "Edit" menus. The variable will be deleted
     automatically by Qt. */
  if(!parent())
    layout->setMenuBar(new MJ_ModuleMenubar(this));

  // Generic parameters.
  // QWidget *generics = new QWidget(this, "multitap generics");

  QGridLayout *grid = 
    new QGridLayout((QWidget *) 0, MJ_ModuleMultiTap1::MJ_MT_SIZEOF, 3, 5, 5);
  
  layout->addLayout(grid);

  for(i = 0; i < MJ_ModuleMultiTap1::MJ_MT_SIZEOF; i++) {

    /* Get value of the parameter. The static cast is kinda dangerous,
       but we know the message type from code above
       (MJ_ModuleMultiTap1::msgGetInput). */
    MJ_ParameterDouble *p = (MJ_ParameterDouble *) m_parameters.at(i);

    /* Each MJ_GuiDoubleSlider contains a few widgets. It is not
       widget class, but a widget holder class. */
    MJ_GuiDoubleSlider &ds = m_generic[i];
    
    if(!ds.createObjs(this)) 
      MJ_fatal(MJ_ERR_MEMORY, fname);

    ds.label()->setText(p->name());
    ds.set(p->value(), p->min(), p->max());
    ds.setId(p->id());

    if(p->mapper()) {
      ds.slider()->setMapper(p->mapper()->duplicate());
    }
    
    ds.toGridHorizontal(grid, i, 0);

    connect(&ds, SIGNAL(emitValue2(const char *,double)),
	    this, SLOT(setValue(const char *,double)));
    connect(&ds, SIGNAL(emitGetValue2(const char *,double *)),
	    this, SLOT(getValue(const char *,double *)));
    connect(&ds, SIGNAL(emitDone2(const char *)),
	    this, SLOT(doneValue(const char *)));
  }

  // Now the delays:
  
  m_delays.resize(MJ_EIGHT);

  /* First two columns for the delay time. The next column is a space
     and the next is gain and after that the panning. */
  grid = new QGridLayout((QWidget *) 0, MJ_EIGHT +  2, 7, 5, 5);
  layout->addLayout(grid);

  const int gainCol = 4;
  const int panCol  = 6;

  grid->setRowStretch(0, 50); // The graphs

  grid->setColStretch(0, 5);  // Index unmber
  grid->setColStretch(1, 70); // Delay time slider.
  grid->setColStretch(2, 5);  // Delay time entry.
  grid->addColSpacing(3, 20); // Blank
  grid->setColStretch(gainCol, 30);      // Gain
  grid->addColSpacing(gainCol + 1, 20);  // Blank
  grid->setColStretch(panCol, 30);       // Pan

  // Put the graph on top of the grid

  m_graphs = new MJ_GraphicsArea(this, "delay-graph");
  m_graphs->setBoundingBox(0.0, -1.0, 4.0, 1.0);
  m_graphs->setBackgroundColor(Qt::white);  
  m_graphs->setMinimumHeight(50);
  grid->addMultiCellWidget(m_graphs, 0, 0, 0, 6);

  connect(m_graphs, SIGNAL(mousePress(QMouseEvent *, float, float)), 
	  this, SLOT(grMousePress(QMouseEvent *, float, float)));

  
  connect(m_graphs, SIGNAL(mouseMove(QMouseEvent *, float, float)), 
	  this, SLOT(grMousePress(QMouseEvent *, float, float)));

  /* Axes */
  MJ_GraphicAxes *ax = new MJ_GraphicAxes();
  ax->setGraphicsContext
    (new MJ_GraphicsContextPen(QPen(QColor(100, 100, 255), 1)));
  ax->setAxes(5);
  m_graphs->appendGraphic(ax);

  /* Graph baseLine*/

  QArray<float> points(4);
  points[0] = 0.0;
  points[1] = 0.0;
  points[2] = 4.0;
  points[3] = 0.0;
  
  MJ_GraphicPolyLine *gpl = new MJ_GraphicPolyLine;
  gpl->setGraph(points.copy());

  m_graphs->appendGraphic(gpl);
  

  /* Graph taps. */

  m_tapGraphs.resize(MJ_EIGHT);
  bzero(&*m_tapGraphs.begin(), sizeof(MJ_Graphic *) * m_tapGraphs.size());
  

  for(i=0; i < MJ_EIGHT; i++) {
    gpl = new MJ_GraphicPolyLine;
    
    /* gpl->setGraphicsContext
       (new MJ_GraphicsContextPen(QPen(Qt::darkBlue, 1))); */

    lock();
    float quarters, leftGain, rightGain;
    m_multiTap->getTapParams(i, & quarters, & leftGain, & rightGain);
    unlock();
    
    points[0] =  quarters;
    points[1] =  leftGain;
    points[2] =  quarters;
    points[3] = -rightGain;
    
    gpl->setGraph(points.copy());

    gpl->setGraphicsContext
      (new MJ_GraphicsContextPen(QPen(Qt::black, 2)));
        
    m_graphs->appendGraphic(gpl);

    m_tapGraphs[i] = gpl;
    updateGraph(i);
  }

  /* updateGraph(0);
     updateGraph(7); */
  
  // Add labels on top of the grid

  // Compose a translated version of "Delay time (quarters)":
  str = MJ_appendUnit(MJ_Tr::tr("Delay time"), MJ_UNIT_QUARTERS);

  int labelRow = 1;

  QLabel *label = 
    new QLabel(str, this, "tempo");
  MJ_limMinSize(label);
  grid->addMultiCellWidget(label, labelRow, labelRow, 1, 2);

  label = new QLabel(MJ_Tr::tr("Gain"), this, "gain");
  MJ_limMinSize(label);
  grid->addWidget(label, labelRow, gainCol);

  label = new QLabel(MJ_Tr::tr("Pan"), this, "pan");
  MJ_limMinSize(label);
  grid->addWidget(label, labelRow, panCol);

  // Put the delay times to their places. 
  for(i = 0; i < MJ_EIGHT; i++) {

    int gridInd = i + 2;

    // Parameter index number. 
    uint pnum = i * 3 + MJ_ModuleMultiTap1::MJ_MTP_TIME + 
      MJ_ModuleMultiTap1::MJ_MT_SIZEOF;

    MJ_ParameterDouble *p = (MJ_ParameterDouble *) m_parameters.at(pnum);

    MJ_GuiDoubleSlider & ds = m_delays.getFast(i);
    
    if(!ds.createObjs(this)) 
      MJ_fatal(MJ_ERR_MEMORY, fname);
    str.setNum(gridInd);
    ds.label()->setText(str);

    ds.slider()->setMinimumWidth(100);
    
    ds.set(p->value(), p->min(), p->max());
    ds.setId(p->id());
    ds.toGridHorizontal(grid, gridInd, 0);

    connect(&ds, SIGNAL(emitValue2(const char *,double)),
	    this, SLOT(setValue(const char *,double)));    
    connect(&ds, SIGNAL(emitGetValue2(const char *,double *)),
	    this, SLOT(getValue(const char *,double *)));
    connect(&ds, SIGNAL(emitDone2(const char *)),
	    this, SLOT(doneValue(const char *)));
  }

  // Gains:
  m_gains.resize(MJ_EIGHT);
  for(i = 0; i < MJ_EIGHT; i++) {

    uint pnum = i * 3 + MJ_ModuleMultiTap1::MJ_MTP_GAIN + 
      MJ_ModuleMultiTap1::MJ_MT_SIZEOF;
    
    MJ_ParameterDouble *p = (MJ_ParameterDouble *) m_parameters.at(pnum);

    MJ_Slider *slider = new MJ_Slider(p->min(), p->max(), p->value(),
				      this);

    MJ_CStrRemap *remap = new MJ_CStrRemap(p->id());
    
    connect(slider, SIGNAL(valueChangedDouble(double)),
	    remap, SLOT(callDouble(double)));
    connect(slider, SIGNAL(emitDone()),
	    remap, SLOT(call()));

    connect(remap, SIGNAL(emitCallDouble(const char *,double)),
	    this, SLOT(setValue(const char *,double)));
    connect(remap, SIGNAL(emitCall(const char *)),
	    this, SLOT(doneValue(const char *)));

    m_remaps.append(remap);

    grid->addWidget(slider, i + 2, gainCol);

    m_gains[i] = slider;
  }

  // Pans:
  m_pans.resize(MJ_EIGHT);
  for(i = 0; i < MJ_EIGHT; i++) {

    uint pnum = i * 3 + MJ_ModuleMultiTap1::MJ_MTP_PAN + 
      MJ_ModuleMultiTap1::MJ_MT_SIZEOF;

    MJ_ParameterDouble *p = (MJ_ParameterDouble *) m_parameters.at(pnum);

    MJ_Slider *slider = new MJ_Slider(p->min(), p->max(), p->value(),
				      this);

    MJ_CStrRemap *remap = new MJ_CStrRemap(p->id());
    
    connect(slider, SIGNAL(valueChangedDouble(double)),
	    remap, SLOT(callDouble(double)));
    connect(slider, SIGNAL(emitDone()),
	    remap, SLOT(call()));

    connect(remap, SIGNAL(emitCallDouble(const char *,double)),
	    this, SLOT(setValue(const char *,double)));
    connect(remap, SIGNAL(emitCall(const char *)),
	    this, SLOT(doneValue(const char *)));

    m_remaps.append(remap);

    grid->addWidget(slider, i + 2, panCol);

    m_pans[i] = slider;
  }
}


void MJ_ModuleMultiTap1_QG::updateGraph(uint tap)
{
  MJ_GraphicPolyLine *gpl = (MJ_GraphicPolyLine *) m_tapGraphs[tap];

  lock();
  float quarters, leftGain, rightGain;
  m_multiTap->getTapParams(tap, & quarters, & leftGain, & rightGain);
  unlock();
  
  /* debugging crap {
    char buf[128];
    sprintf(buf, 
	    "MJ_ModuleMultiTap1_QG::updateGraph # tap = %d q = %f, l = %f, r =  %f", 
	    (int) tap, quarters, leftGain, rightGain);
    trace(buf);
    }*/

    
  m_graphs->removeGraphic(gpl->id(), false);
  
  float *points = gpl->points().points();
  
  points[0] =  quarters;
  points[1] =  leftGain;
  points[2] =  quarters;
  points[3] = -rightGain;
  
  m_graphs->appendGraphic(gpl);
  m_graphs->drawAll();
  // m_graphs->drawGraphic(gpl);
  // m_graphs->redraw();
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_ModuleDsp * MJ_createNew_ModuleExample2() 
{
  return new MJ_ModuleMultiTap1;
}

/* This function returns a new plugin data base entry. Upon startup,
   Mustajuuri opens this shared library and tries to find a function
   with this name. It should return a new object of type
   MJ_PluginDbeDsp.

*/

MJ_PluginDbeDsp *MJ_createPlugin(int i)
{
  if(i != 0) return 0; // This function has laready been called once.

  // Create the entry.
  MJ_PluginDbeDsp *dbe = new MJ_PluginDbeDsp;

  // Set author of of this plugin.
  dbe->setAuthors("Ilmonen Tommi S.");

  /* Set menu position. This name may be translated one field at a time. */
  dbe->setMenuPosition("_Audio_Delay_Eight tap delay");
  
  /* General information for tooltips etc. */
  dbe->setAbout("Tempo-controlled eight tap delay");

  /* Factory (or creator) function. Unless this is defined the
     database entry cannot fabricate new modules.  */
  dbe->setCreator(MJ_createNew_ModuleExample2);

  /* Unique identifier. This string is used to define module type when
     saving module definitions to a file. It must not contain
     spaces. It is probably a good idea to use name that can be read
     and has some meaning. */
  dbe->setIdentifier("/delay/multi_tap_delay1");

  /* Version number of this plugin. */
  dbe->pluginVersion()->set(0,0,1);

  /* Define that both input and output should have two channels. */
  dbe->options()->defaultChans(2);

  /* Further define that the minimum number of input channels is one. 

     In fact this overrides part of the effect of the previous
     command.  */
  dbe->options()->setValue(MJ_DSP_MIN_CHANNELS_IN, 1);

  dbe->setParameter(MJ_PP_TRANSLATION_FILE, "mj_example2");

  /* We only set this for a poscard GUI factory. In the absence of
     window-sized GUI factory the system will use the same factory
     object. */
  dbe->setGuiFactory(MJ_GUI_POSTCARD, MJ_ModuleMultiTap1_QG::createNew);
  
  dbe->setPluginDebug(MJ_ModuleMultiTap1::debugPtr());
  
  return dbe;  
}
