/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#ifndef MJ_MODULE_EXAMPLE2_GUI_H
#define MJ_MODULE_EXAMPLE2_GUI_H

/* Includes for the GUI: */
/* Include GUI base definitions. */
#include <mj_module_dsp_gui_default.h>
#include <mj_gui_numeric.h>
#include <mj_graphic.h>
#include <mj_cstr_remap.h>
#include <mj_parameter_numeric.h>

#include <mj_module_example2.h>

#include <vector>

class MJ_GraphicsArea;

/** A specialized GUI for MJ_ModuleMultiTap1. 
    
    <B>BACKGROUND AND MOTIVATION:</B> Mustajuuri can generate a GUI
    automatically, but that is only suitable for omst trivial plugin
    types. If the plugin has more internal features it is useful to
    create a specialized GUI for the module type. That way you can
    make a more user-friendly and pragmatic GUI.

    In this case all parameters are displayed as usual, but the
    widgets and layout are differ from what the default system wuold
    create. 

    Furthermore, one can create more than one GUI to match different
    usages. A small stamp-size GUI fits easily to window that shows
    many modules. Bit biffer GUI (postcard-sized) fits to vindow that
    shows a few modules. Finally the plugin can be opened in a
    separate window and it can use all available screen space (and
    even span new windows).

    <B>CODE DOCUMENTATION:</B> All module GUI classes must inherit the
    MJ_ModuleDspGui class. The base class provides a few convenient
    methods that may make life a bit easier. Since we have a threaded
    real-time system there are a feq tricks that need to be
    done. First of all - access to the module is restricted with a
    mutex. The parent class provides convenience methods "lock()" and
    "unlock()" to guarantee no other thread is modifying the module at
    the same time. Secondly the GUI must not lock the module for a
    long period of time (say longer than 500 us). Such locking may
    result is degraded latency behaviour. Usually it is best to 1)
    lock the mutex 2) get or set a parameter 3) do something with the
    parameter. REMEMBER: Nearly all graphics (=widget related)
    operations may take unspecified amount of time get
    accomplished. NEVER do then while the module is being locked by
    the GUI.
    
    Typically the system creates a new gui like this:

    <OL>

    <LI> Object A needs a new GUI object and asks it from the dsp module.

    <LI> DSP-module (or its data base entry) creates a new GUI object
    and returns it.

    <LI> Object A sets the real-time parent (used for mutex passing) and
    passes the relevant mutex to it.

    <LI> Gives the module to the GUI. Now the GUI can construct itself
    properly.

    </OL>

    This module (MJ_ModuleMultiTap1) only has one extra GUI defined
    and it is used when when the system requires postcard- or
    window-sized GUI.

*/
class MJ_ModuleMultiTap1_QG : public MJ_ModuleDspGuiPostCard
{
  Q_OBJECT
public:
  /** Constructs the GUI. The module is not given to the GUI at this
      stage. */
  MJ_ModuleMultiTap1_QG(MJ_RtControl *rtparent,
			MJ_ModuleDsp *module,
			QWidget      *parent = 0,
			const char   *name = 0);

  virtual ~MJ_ModuleMultiTap1_QG();

  /** Receives information that the module has been modified. */
  virtual void moduleModified(const char *);

  /** This widget can use more space than some others. */
  int weight();

  /// Static method that can be used as a factory function.
  static MJ_ModuleDspGui *createNew(MJ_RtControl *rtparent,
				    MJ_ModuleDsp *module,
				    QWidget *parent, 
				    const char *name);

public slots:

  void grMousePress(QMouseEvent *, float, float);

protected:
  void update2(const char *);

  /// Creates all child widgets. 
  void draw();
  
  /// Update nth graph
  void updateGraph(uint tap);

  MJ_ModuleMultiTap1      *m_multiTap;

  /** GUI objects for the generic parameters (tempo etc). */

  MJ_GuiDoubleSlider       m_generic[MJ_ModuleMultiTap1::MJ_MT_SIZEOF]; 
  
  MJ_Array<MJ_GuiDoubleSlider> m_delays;
  MJ_Array<MJ_Slider *>        m_gains;
  MJ_Array<MJ_Slider *>        m_pans;
  
  MJ_GraphicsArea             *m_graphs;
  std::vector<MJ_Graphic *>    m_tapGraphs;

  int                          m_activeTap;

  QList<QObject>           m_remaps;

};

#endif
