/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_graph.h>

#include <mj_strs.h>
#include <mj_math.h> 
#include <mj_messages.h> 
#include <mj_message_string.h> 
#include <mj_parameters.h> 
#include <mj_stream_utils.h> 

#include <qfiledialog.h> 

bool MJ_ModuleGraph::m_debug = true;

MJ_ModuleGraph::MJ_ModuleGraph()
  : m_fileName("processors.graph"),
    m_framesPerCycle(64),
    m_framesPerCycleNow(64),
    m_inChans(0),
    m_outChans(2)
{
  m_transfer.setId("transfer");
  m_transfer.setParentModule(this);
  // m_transferOut.setId(MJ_Strs::output);
  MJ_ModuleDspArray::dspSetTimer(&m_graphTimer);
}

MJ_ModuleGraph::~MJ_ModuleGraph()
{}
  
MJ_ModuleDsp *MJ_ModuleGraph::findModule(const char * id)
{
  // trace3(m_debug, "MJ_ModuleGraph::findModule # \"%s\"", id);
  trace2("MJ_ModuleGraph::findModule # \"%s\"", id);
  
  if(strcmp(id, MJ_Strs::input) == 0 || 
     strcmp(id, MJ_Strs::output) == 0)
    return & m_transfer;

  return MJ_ModuleDspArray::findModule(id);
}

bool MJ_ModuleGraph::dspSetTimer(MJ_TimeStamp *t)
{
  trace3(m_debug, "MJ_ModuleGraph::dspSetTimer # %p (%u)", 
	 t, m_framesPerCycle);

  MJ_ModuleDsp::dspSetTimer(t);

  if(t) {
    m_graphTimer = *t;
    m_graphTimer.setFrameSize(m_framesPerCycle);
  }

  m_transfer.dspSetTimer(&m_graphTimer);

  trace2("MJ_ModuleGraph::dspSetTimer # %u", m_transfer.dspFrameSize());

  return MJ_ModuleDspArray::dspSetTimer(&m_graphTimer);
}

bool MJ_ModuleGraph::dspAcceptIoChans(int *, int *out)
{
  bool change = false;

  if(*out != (int) m_outChans) {
    change = true;
    *out = m_outChans;
  }

  return change;
}

bool MJ_ModuleGraph::dspPrepare()
{
  const char * fname = "MJ_ModuleGraph::dspPrepare";

  if(m_outChans != m_signalOutput.size()) {
    error2(MJ_ERR_UNKNOWN, "%s # The number of output channels should be %u",
	   fname, m_outChans);

    return false;
  }

  m_transfer.dspSetInChannels(m_outChans);

  // appendModule(&m_transfer);

  bool ok = loadFile(m_fileName);
  
  if(!ok) return false;

  ok = MJ_ModuleDspArray::dspPrepare();
  
  if(!ok) return false;

  // Set up the transfer module:
  m_transfer.dspClearOutputs();
  ok = m_transfer.dspCompileInputs();
  
  if(!ok) {
    error2(MJ_ERR_UNKNOWN, "%s # Failed to configure transfer module", fname);
    return false;
  }
  
  // Prepare internal variables:
  m_framesPerCycleNow = m_framesPerCycle;

  m_doneSamples = 0;
  m_givenSamples = 0;
  m_externSamples = 0;
  m_externalFramesPerCycle = m_currentTime->frameSize();
    
  trace3(m_debug, "%s # %d %d", fname,
	 (int) m_framesPerCycleNow, (int) m_externalFramesPerCycle);

  return true;
}

void MJ_ModuleGraph::dspProcessPeriod()
{
  // trace("MJ_ModuleGraph::dspProcessPeriod");

  uint i, j;

  if(m_doneSamples > m_externSamples) {
    
    uint offset = m_givenSamples % m_framesPerCycleNow;
    
    uint transferFrames = MJ_min((uint) m_framesPerCycleNow - offset,
				 (uint) m_externalFramesPerCycle);
    
    // Read data from transfer buffers to output.
    for(i=0; i < m_signalOutput.size(); i++) {
      
      MJ_Sample * extBuf= m_signalOutput.getFast(i).samples().data();
      const MJ_Sample *inBuf = 
	m_transfer.dspInSignal(i).samples().data() + offset;
      
      for(j=0; j < transferFrames; j++) {
	*extBuf++ = *inBuf++;
      }
    }

    /* trace2("MJ_ModuleGraph::dspProcessPeriod # 1 # %u %u", 
       transferFrames, offset);*/

    m_givenSamples += transferFrames;
  }

  m_externSamples += m_externalFramesPerCycle;

  while(m_doneSamples < m_externSamples) {

    MJ_ModuleDspArray::dspProcessPeriod();

    // Read data from input to transfer buffers.
    /* for(i=0; i < m_signalInput.size(); i++) {

      const MJ_Sample * extBuf= m_signalInput.getFast(i).samples().data();
      MJ_Sample *inBuf = m_tarnsfer.m_signalInput.get(i).samples().data();
      
      for(uint j=0; j < transferFrames; j++) {
	*inBuf++ = *extBuf++;
      }
      }*/

    // Read data from transfer buffers to output:

    uint offset = m_givenSamples % m_externalFramesPerCycle;

    uint transferFrames = MJ_min((uint) m_framesPerCycleNow, 
				 (uint) m_externalFramesPerCycle - offset);    

    for(i=0; i < m_signalOutput.size(); i++) {

      MJ_Sample * extBuf = m_signalOutput.getFast(i).samples().data() + offset;
      const MJ_Sample *inBuf = m_transfer.dspInSignal(i).samples().data();
      
      for(j = 0; j < transferFrames; j++) {
	*extBuf++ = *inBuf++;
      }
    }

    /*trace2("MJ_ModuleGraph::dspProcessPeriod # 2 # %u %u %u %u", 
	   transferFrames, offset, m_transfer.dspInSignal(0).samples().size(),
	   m_transfer.dspFrameSize());*/
    
    m_givenSamples += transferFrames;
    m_doneSamples  += m_framesPerCycleNow;
  }

  m_graphTimer.incrementStamp(m_graphTimer.frameSize());
}

bool MJ_ModuleGraph::dspStop()
{
  removeModule(&m_transfer);

  clearArray(true);

  return true;
}

bool MJ_ModuleGraph::msgProcess(const char *id, MJ_Message *m)
{
  if(!strcmp(id, MJ_Strs::filename)) {
    MJ_MessageString *ms = dynamic_cast<MJ_MessageString *> (m);

    if(!ms) return false;
    if(ms->stringConst().isEmpty()) return false;
    
    m_fileName = ms->stringConst().copy();
    return true;
  }
  else if(!strcmp(id, MJ_Strs::framesPerCycle)) {
    bool ok = true;

    int fpc = m->toInt(&ok);

    fpc = MJ_clamp(fpc, 1, 256);

    if(!ok) return false;
    
    m_framesPerCycle = fpc;

    return true;
  }
  else if(!strcmp(id, MJ_Strs::outchannells)) {
    bool ok = true;

    int fpc = m->toInt(&ok);

    fpc = MJ_clamp(fpc, 1, 256);

    if(!ok) return false;
    
    m_outChans = fpc;

    return true;
  }

  return MJ_ModuleDspArray::msgProcess(id, m);
}

MJ_Parameter * MJ_ModuleGraph::msgGetInput(const char *id)
{
  if(!strcmp(id, MJ_Strs::filename))
    return new MJ_ParameterFilename(m_fileName, 
				    "Graph file", 
				    MJ_Strs::filename,
				    "*.graph",
				    QFileDialog::ExistingFile);
  if(!strcmp(id, MJ_Strs::framesPerCycle))
    return new MJ_ParameterInt("Frames per cycle", id,
			       m_framesPerCycle, 1, 256);
  if(!strcmp(id, MJ_Strs::outchannells))
    return new MJ_ParameterInt("Output channels", id,
			       m_outChans, 1, 256);
  return 0;
}

QCString MJ_ModuleGraph::msgGetInputName(int i)
{
  QCString r;

  switch(i) {
  case 0:
    r = MJ_Strs::filename;
    break;
  case 1:
    r = MJ_Strs::framesPerCycle;
    break;
  case 2:
    r = MJ_Strs::outchannells;
    break;
  }

  return r;
}

bool MJ_ModuleGraph::loadFile(const QString &fileName)
{
  const char * fname = "MJ_ModuleGraph::loadFile";
  
  trace3(m_debug, "%s # \"%s\"", fname, fileName.ascii());

  QFile file(fileName);

  bool ok = file.open(IO_ReadOnly);

  if(!ok) {
    error2(MJ_ERR_ARGUMENT, "%s # Could not open file \"%s\"", fname, 
	   fileName.ascii());
    return false;
  }
  
  QTextStream s(&file);

  bool loop = true;

  QString tmp;

  uint count = 0;

  while(loop && ! file.atEnd()) {

    tmp = s.readLine().stripWhiteSpace();

    if(tmp == MJ_Strs::end) {
      trace3(m_debug, "%s # Got %u modules from the file", fname, count);
      return true;
    }

    if(tmp == "output {") {
      ok = m_transfer.readIODefs(&s);
      
      if(!ok) {
	return false;
      }
    }
    else {
      if(tmp != "module {") {
	error2(MJ_ERR_ARGUMENT, "%s # \"module {\" expected near line %d",
	       fname, MJ_currentLine(&file));
	return false;
      }
      
      MJ_ModuleDsp * m = MJ_ModuleDsp::readFromStream(&s);
      
      if(m) {
	appendModule(m);
	count++;
      }
      else {
	error2(MJ_ERR_UNKNOWN, "%s # Failed to read module from file", fname);
	return false;
      }
    }

    tmp = s.readLine().stripWhiteSpace();

    if(tmp != "}") {
      error2(MJ_ERR_ARGUMENT, "%s # Closing brace \"}\" or \"end\""
	     "expected near line %d",
	     fname, MJ_currentLine(&file));
      return false;
    }
  }
  
  error2(MJ_ERR_ARGUMENT, "%s # File ended abruptly", fname);

  return false;
}
