/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_ks1.h>

#include <mj_messages.h>
#include <mj_parameters.h>
#include <mj_random.h>
#include <mj_strs.h>

#include <mj_dsp_plugin.h>

static MJ_Random ___MJ_KS_rand;

MJ_ModuleKS1::MJ_ModuleKS1()
{}

MJ_ModuleKS1::~MJ_ModuleKS1()
{}
  
bool MJ_ModuleKS1::dspPrepare()
{
  m_buffer.resize(dspSamplingRate() / 10);

  m_filter.reset();

  m_feedback = 0.9;
  m_delayTime = 100.0;

  m_note = -1;

  return true;
}

void MJ_ModuleKS1::dspProcessPeriod()
{
  MJ_Vector<MJ_Sample> *out = & m_signalOutput[0].samples();

  MJ_Sample *buf = out->data();
  uint samples = out->size();

  for(uint i=0; i < samples; i++) {
    float tmp = m_buffer.getNewest((uint) m_delayTime);
    
    tmp = m_filter.put(tmp) * m_feedback;

    m_buffer.put(tmp);
    
    *buf++ = tmp;
  }
}

bool MJ_ModuleKS1::msgProcess(const char *id, MJ_Message *m)
{
  if(!strcmp(id, MJ_Strs::MIDI)) {
    
    MJ_MessageMidiEvent * mm = dynamic_cast<MJ_MessageMidiEvent *> (m);
    
    if(mm) {
      const DI_MidiEvent & e = mm->midiMsg();
      
      if(e.isNoteOn()) {

	/* Scale gainDb to range [-30, 0] dB. */
	float gainDb = ((float) (e.noteVelocity() - 127)) * (30.0f / 127.0f);

	float gain = MJ_dbToAbs(gainDb);
	
	m_note = e.keyNumber();
	float pitch = DI_MidiEvent::keyNumberToHz(m_note);

	m_delayTime = dspSamplingRate() / pitch  - 0.5;
	
	// Fill the delay buffer with white noise:
	for(uint i=0; i < m_buffer.size(); i++) {
	  m_buffer.getIndex(i) = ___MJ_KS_rand.getFloat1(gain);
	}
	
	m_filter.reset();
	m_feedback = 0.999;

	trace2("MJ_ModuleKS1::msgProcess # GainDb = %f, pitch = %f",
	       gainDb, pitch);
      }
      else if(e.isNoteOff()) {
	
	if(e.keyNumber() != m_note)
	  return true;

	m_feedback = 0.95;
      }
      
    }
    
    return true;
  }
  
  return false;
}

MJ_Parameter * MJ_ModuleKS1::msgGetInput(const char *id)
{
  if(!strcmp(id, MJ_Strs::MIDI)) {
    MJ_ParameterMidi *pm = new MJ_ParameterMidi();
    pm->setName("MIDI input");
    pm->setId(id);
    return pm;
  }
  
  return 0;
}

QCString MJ_ModuleKS1::msgGetInputName(int i)
{
  return i ? QCString() : QCString(MJ_Strs::MIDI);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_ModuleDsp * MJ_createNew_ModuleKS1() 
{
  return new MJ_ModuleKS1;
}

MJ_PluginDbeDsp *MJ_createPlugin(int i)
{
  if(i) return 0;

  MJ_PluginDbeDsp *dbe = new MJ_PluginDbeDsp;

  dbe->setAuthors("Ilmonen Tommi S.");
  dbe->setMenuPosition("_Audio_Synthesis_Karplus-Strong");
  dbe->setCreator(MJ_createNew_ModuleKS1);
  dbe->setIdentifier("MUSTAJUURI_KS1");
  dbe->options()->defaultChans2(0, 1);
  
  return dbe;  

}
