/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_ladspa.h>

#include <mj_ladspa_driver.h>

#include <mj_error.h> 
#include <mj_mapper.h> 
#include <mj_parameters.h> 

#include <stdlib.h>

MJ_ModuleLadspa::MJ_ModuleLadspa(const LADSPA_Descriptor *handle)
  : m_handle(handle),
    m_plugin(0)
{
  m_data.resize(MJ_LadspaDriver::nofControlIns(handle));
  m_data.setAll(0);

  getDefaultValues();
  // trace("MJ_ModuleLadspa::MJ_ModuleLadspa", m_data.size());
}

MJ_ModuleLadspa::~MJ_ModuleLadspa()
{
  if(m_plugin && m_handle->cleanup)
    (*m_handle->cleanup)(m_plugin);
}

bool MJ_ModuleLadspa::dspPrepare()
{
  const char * fname = "MJ_ModuleLadspa::dspPrepare";

  uint i;
  const LADSPA_Descriptor *handle = m_handle;

  m_plugin = handle->instantiate(handle, dspSamplingRate());

  MJ_cpf((void *)handle->connect_port, fname);
  
  const uint ins = MJ_LadspaDriver::nofInputs(handle);
  const uint outs = MJ_LadspaDriver::nofOutputs(handle);
  const uint controls = MJ_LadspaDriver::nofControlIns(handle);

  if(m_signalInput.size() != ins || 
     m_signalOutput.size() != outs ||
     controls != m_data.size()) {
    
    error(MJ_ERR_INTERNAL, fname, "I/O mismatch");
    return false;
  }

  // Connect inputs
  for(i=0; i < ins; i++) {
    ulong port = (ulong) MJ_LadspaDriver::indexToPort
      (handle, LADSPA_PORT_INPUT | LADSPA_PORT_AUDIO, i);
    handle->connect_port(m_plugin, 
			 port, 
			 m_signalInput.get(i).samples().data());
  }

  // Connect outputs
  for(i=0; i < outs; i++) {
    ulong port = (ulong) MJ_LadspaDriver::indexToPort
      (handle, LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO, i);
    handle->connect_port(m_plugin, 
			 port, 
			 m_signalOutput.get(i).samples().data());
  }
  
  // Connect controls
  for(i=0; i < controls; i++) {
    ulong port = (ulong) MJ_LadspaDriver::indexToPort
      (handle, LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL, i);
    handle->connect_port(m_plugin, port, & m_data.get(i));
  }

  if(handle->activate)
    handle->activate(m_plugin);

  return true;
}

void MJ_ModuleLadspa::dspProcessPeriod()
{
  m_handle->run(m_plugin, m_currentTime->frameSize());
}

bool MJ_ModuleLadspa::dspStop()
{
  if(m_handle->deactivate)
    m_handle->deactivate(m_plugin);

  if(m_handle->cleanup)
    m_handle->cleanup(m_plugin);

  m_plugin = 0;

 return true;
}

bool MJ_ModuleLadspa::msgProcess(const char *id, MJ_Message *m)
{
  const char * fname = "MJ_ModuleLadspa::msgProcess";

  // trace(fname, id);

  bool ok = true;

  double v = m->toDouble(&ok);

  if(!ok || !id) {
    error(MJ_ERR_ARGUMENT, fname, id);
    return false;
  }

  int index = atoi(id);

  if((uint) index >= m_data.size()) {
    error2(MJ_ERR_ARGUMENT, 
	   "%s # parameter exceeds index range %d >= %d" ,
	   fname, index, (int) m_data.size());
    return false;
  }
  
  m_data.getFast(index) = v;
  return true;
}

MJ_Parameter * MJ_ModuleLadspa::msgGetInput(const char *id)
{
  const char * fname = "MJ_ModuleLadspa::msgGetInput";
  
  MJ_cpf((void *) id, fname);

  // trace(fname, id);
  
  int index = atoi(id);

  const int sr = dspSamplingRate();
  
  int port = MJ_LadspaDriver::indexToPort
    (m_handle, LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL, index);

  if(port < 0) {
    error(MJ_ERR_ARGUMENT, fname, "No such control port");
    return 0;
  }

  const LADSPA_PortRangeHint *hint = & m_handle->PortRangeHints[port];

  LADSPA_Data lowerBound = hint->LowerBound;
  LADSPA_Data upperBound = hint->UpperBound;

  if(lowerBound == upperBound || 
     !LADSPA_IS_HINT_BOUNDED_BELOW(hint->HintDescriptor) ||
     !LADSPA_IS_HINT_BOUNDED_ABOVE(hint->HintDescriptor)) {
    QString tmp;
    tmp.setNum(m_data.get(index));
    return new MJ_ParameterString(tmp, m_handle->PortNames[port], id);
    
  }
  
  if(LADSPA_IS_HINT_SAMPLE_RATE(hint->HintDescriptor)) {
    lowerBound *= sr;
    upperBound *= sr;
  }
  
  if(hint->HintDescriptor & LADSPA_HINT_TOGGLED) {
    // A boolean parameter.
    MJ_ParameterBool *pb = new MJ_ParameterBool(m_data.get(index) > 0.0,
						m_handle->PortNames[port],
						id);
    return pb;
  }
  else if(hint->HintDescriptor & LADSPA_HINT_INTEGER) {
    MJ_ParameterInt *pi = new 
      MJ_ParameterInt(m_handle->PortNames[port],
		      id,
		      (int) m_data.get(index),
		      (int) lowerBound,
		      (int) upperBound);
    return pi;
  }
  else {
    MJ_ParameterDouble *pd = 
      new MJ_ParameterDouble(m_handle->PortNames[port],
			     id,
			     m_data.get(index),
			     lowerBound,
			     upperBound);
    
    if(LADSPA_IS_HINT_LOGARITHMIC(hint->HintDescriptor) && 
       lowerBound > 1e-5)
      pd->setMapper(new MJ_MapperLog);
    
    return pd;
  }
}

QCString MJ_ModuleLadspa::msgGetInputName(int i)
{
  QCString ret;

  if((uint) i >= m_data.size())
    return ret;

  ret.setNum(i);
  return ret;
}

/** Guess some default parameter values. The guesses are based on the
    port range hints. */

void MJ_ModuleLadspa::getDefaultValues()
{
  LADSPA_Data value;

  for(uint i = 0; i < m_data.size(); i++) {
    
    int port = MJ_LadspaDriver::indexToPort
      (m_handle, LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL, i);
    
    const LADSPA_PortRangeHint *hint = & m_handle->PortRangeHints[port];

    LADSPA_Data lowerBound = hint->LowerBound;
    LADSPA_Data upperBound = hint->UpperBound;
    
    if(LADSPA_IS_HINT_BOUNDED_BELOW(hint->HintDescriptor)) {
      if(LADSPA_IS_HINT_BOUNDED_ABOVE(hint->HintDescriptor))
	value = (lowerBound + upperBound) * 0.5;
      else 
	value = lowerBound;
    }
    else if(LADSPA_IS_HINT_BOUNDED_ABOVE(hint->HintDescriptor))
      value = upperBound;
    else 
      value = 0.0;

    m_data.getFast(i) = value;
  }
}
