/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_latencytest.h>

#include <mj_dsp_plugin.h>
#include <mj_error.h>
#include <mj_math.h>
#include <mj_messages.h>
#include <mj_parameters.h>
#include <mj_tr.h> 

#include <qlabel.h> 
#include <qlayout.h> 

enum MJ_LatTestParams {
  MJ_LTP_THRESHOLD,
  MJ_LTP_INTERVAL,
  MJ_LTP_SIZEOF
};

// The number of these parameters = MJ_ModuleUiTest::m_params.
static const char *MJ_LatencyTestKeys[] = {
  "threshold",
  "interval"
};

MJ_ModuleLatencyTest::MJ_ModuleLatencyTest()
  : m_threshold(0.1),
    m_testIntervalMs(1000.0),
    m_timeSamples(-1)
{}

MJ_ModuleLatencyTest::~MJ_ModuleLatencyTest()
{}
  
bool MJ_ModuleLatencyTest::dspPrepare()
{
  m_listen = false;
  m_sendTime = 0;
  m_counter = 0;
  m_testIntervalSamples = (ulong) 
    (dspSamplingRate() * m_testIntervalMs / 1000.0);
  m_burstSamples = (ulong) (dspSamplingRate() / 1000.0);

  return m_signalInput.size() == 1 && m_signalOutput.size() == 1;
}

void MJ_ModuleLatencyTest::dspProcessPeriod()
{
  MJ_Vector<MJ_Sample> *v = &m_signalInput.get(0).samples();
  MJ_Sample *data =v->data();
  const uint l = v->size();
  uint i;

  // Scan input for signal peaks
  for(i=0; i < l; i++) {
    if(MJ_abs(data[i]) > m_threshold && m_listen) {
      m_timeSamples = (long) ((m_counter + i) - m_sendTime);
      m_listen = false;
    }
  }
  
  // Send noise to the output:
  
  ulong start = m_counter;
  
  data = m_signalOutput.get(0).samples().data();

  for(i=0; i < l; i++) {
    
    ulong tmp = (start + i) % m_testIntervalSamples;
    
    if(!tmp) {
      m_listen = true;
      m_sendTime = start + i;
    }

    if(tmp < m_burstSamples)
      data[i] = m_noiseSource.getFloat1();
    else 
      data[i] = 0.0;
  }
  
  m_counter += l;
}

  
bool MJ_ModuleLatencyTest::msgProcess(const char *id, MJ_Message *m)
{
  const char * fname = "MJ_ModuleLatencyTest::msgProcess";

  int ind = MJ_findStr(id, MJ_LatencyTestKeys, MJ_LTP_SIZEOF);
  
  if(ind < 0) {
    error(MJ_ERR_ARGUMENT, fname, "Bad id");
    return false;
  }

  bool ok = true;

  double tmp = m->toDouble(&ok);

  if(!ok) return false;
  
  switch(ind) {
  case MJ_LTP_THRESHOLD:
    m_threshold = tmp;
    break;

  case MJ_LTP_INTERVAL:
    m_testIntervalMs = tmp;
    dspPrepare();
    break;

  default:
    ok = false;
  }
  return ok;
}

MJ_Parameter * MJ_ModuleLatencyTest::msgGetInput(const char *id)
{
  int ind = MJ_findStr(id, MJ_LatencyTestKeys, MJ_LTP_SIZEOF);

  if(ind == MJ_LTP_THRESHOLD) {
    return new MJ_ParameterDouble("Threshold", id, m_threshold, 0, 1.0);
  }
  else if(ind == MJ_LTP_INTERVAL) {
    return new MJ_ParameterDouble("Interval", id, m_testIntervalMs, 
				  10.0, 1000.0, MJ_UNIT_MILLISECONDS);
  }

  error(MJ_ERR_ARGUMENT, "MJ_ModuleLatencyTest::msgGetInput");
  return 0;
}


QCString MJ_ModuleLatencyTest::msgGetInputName(int i)
{
  QCString r;

  if(i <  MJ_LTP_SIZEOF) r = MJ_LatencyTestKeys[i];

  return r;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_ModuleLatencyTest_QG::MJ_ModuleLatencyTest_QG(MJ_RtControl *rtparent,
						 MJ_ModuleDsp *module,
						 QWidget      *parent,
						 const char   *name)
  : MJ_ModuleDspGuiPostCard(rtparent, module, true,
			    MJ_LTP_SIZEOF, 
			    MJ_GUI_OWN_WINDOW,
			    parent, name),
  m_display(new QLabel(childParent()))
{
  const char * fname = "MJ_ModuleLatencyTest_QG::MJ_ModuleLatencyTest_QG";

  if(!dynamic_cast<MJ_ModuleLatencyTest *> (module)) 
    MJ_fatal(MJ_ERR_ARGUMENT, fname);

  QLabel *l = new QLabel(tr("Round-trip time (ms)"), childParent());

  MJ_cpf(l, m_layout, fname);

  int rows = m_layout->numRows();
  m_layout->expand(rows + 1, m_layout->numCols());

  m_layout->addWidget(l, MJ_LTP_SIZEOF, 0);
  m_layout->addWidget(m_display, MJ_LTP_SIZEOF, 1);
  l->show();
  m_display->show();
  
  connect(&m_timer, SIGNAL(timeout()), this, SLOT(update()));
  if(isRealTime()) {
    realTimeChange(true);
  }
}
  
void MJ_ModuleLatencyTest_QG::update()
{
  lock();
  long timeSamples = ((MJ_ModuleLatencyTest *) m_module)->timeSamples();
  int sr = m_module->dspSamplingRate();
  unlock();
  
  QString str;

  if(timeSamples > 0) {
    str.setNum((double) timeSamples / (double) sr * 1000.0);
    if(str != m_display->text())
      m_display->setText(str);
  }
}

MJ_ModuleDspGui *MJ_ModuleLatencyTest_QG::createNew(MJ_RtControl *rtparent,
						    MJ_ModuleDsp *module,
						    QWidget *parent, 
						    const char *name)
{
  return new MJ_ModuleLatencyTest_QG(rtparent, module, parent, name);
}

void MJ_ModuleLatencyTest_QG::realTimeChange (bool nowRt)
{
  if(nowRt) m_timer.start(300);
  else m_timer.stop();
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_ModuleDsp * MJ_createNew_ModuleLatencyTest() 
{
  return new MJ_ModuleLatencyTest;
}

MJ_PluginDbeDsp *MJ_createPlugin(int i)
{
  if(i != 0) return 0;

  MJ_PluginDbeDsp *dbe = new MJ_PluginDbeDsp;
  dbe->setAuthors("Ilmonen Tommi S.");
  dbe->setMenuPosition("_Audio_Tests_Latency test");
  dbe->setCreator(MJ_createNew_ModuleLatencyTest);
  dbe->setIdentifier("tcm.hut.fi_latencytest1");
  dbe->setParameter(MJ_PP_TRANSLATION_FILE, "mj_latencytest");
  dbe->options()->defaultChans(1);
  dbe->setGuiFactory(MJ_GUI_POSTCARD, MJ_ModuleLatencyTest_QG::createNew);
  dbe->setParameter(MJ_PP_COMPILE_TIME, DATETIME);

  return dbe;  
}
