/* 
 * COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Authors: Tommi Ilmonen and Benoit Gauduin, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 * benoit.gauduin@freesbee.fr
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */


/* Include the header of this file. */ 
#include "mjdi_module_filter_iir_multichanl.h"

/* Include header of parameter settings. */
#include "mj_parameters.h"
#include "mj_messages.h"
#include "mj_message_string.h"

/* Include plugin data base declarations. */
#include "mj_dsp_plugin.h"

/* Include globally accessible character strings. */
#include "mj_strs.h"

/* Needed to control the version number of the plugin. */
#include "mj_version.h"
#include "mj_arraytmpl.C"

/* Include header for file I/O operations */
#include <qfiledialog.h>
#include <qfile.h>
#include <qtextstream.h>

bool MJDI_ModuleFilterIIR_multichanl::m_debug = false;

MJDI_ModuleFilterIIR_multichanl::MJDI_ModuleFilterIIR_multichanl()
  : m_bypass(false)
{}


MJDI_ModuleFilterIIR_multichanl::~MJDI_ModuleFilterIIR_multichanl()
{}


bool MJDI_ModuleFilterIIR_multichanl::dspPrepare()
{
  /* Read the coefficients from a file. */

  uint channels = m_signalOutput.size();    // Check the number of channels
  m_filters.resize(channels);

  QFile f(m_fileName);                      // open the file
  if (!f.open(IO_ReadOnly)) {
    trace("\nIIR Filter ERROR : unable to open coeficients'file\n");
    return false;
  }

  QTextStream fich( &f );

  uint order,i,j;
  MJ_FilterCoeffsIIR<MJDI_FilterCoeff> coefficients;
  MJ_Sample temp=0;
  bool endfile=0;
  if (m_debug) trace2("\nIIR Filter:\n");

  /* Loop for each channel */
  for(j=0; j<channels; j++)
  {
    order=getOrder(fich, endfile);
    if (!endfile)
        {
    	coefficients.setOrder(order);
        if (m_debug) trace2("- channel %d order = %d\n",j,order);

    	for (i=0;i<order+1;i++) {     // Read coefs b[i]
	  fich >> temp;
	  coefficients.setNumerator(i,temp);
	  if (m_debug) trace2("- channel %d coef b%d = %f\n",
			      j, i, coefficients.getNumerator(i));  
	  }
    	for (i=0;i<order+1;i++) {     // Read coefs a[i]
	  fich >> temp;
	  coefficients.setDenominator(i,temp);
	  if (m_debug) trace2("- channel %d coef a%d = %f\n",
			      j, i, coefficients.getDenominator(i));
      	  }
        }
    m_filters[j].update(coefficients); // update coefficients for the j channel
    if (m_debug) 
      if (endfile) 
	trace2("-> same filter for channel %d ok.\n",j);
      else trace2("`-> channel %d ok.\n",j);
  }
  f.close();                         // close the file

  for(j=0; j<channels; j++)
    m_filters[j].resetMemory();

  /* Check that the number of input channels equals the number of output*/
  if( m_signalInput.size() != m_signalOutput.size()) 
    return false;

  /* Check the number of samples in input and output buffers is the same. */
  if(m_signalInput[0].samples().size() != m_signalOutput[0].samples().size())
    return false;

  return true;
}

/* Read the notes until the order in the file */
uint MJDI_ModuleFilterIIR_multichanl::getOrder(QTextStream &fich, 
					       bool &endfile)
{
  bool ok;
  QString blabla;
  uint order;
  endfile=0;
  do
  {
    blabla=fich.readLine();
    if (fich.atEnd()) endfile=1;
    order=blabla.toUInt(&ok);
    if ((m_debug) && ((blabla!=NULL) && (!ok))) 
      trace2("Notes : %s\n",(const char *)blabla);
  }while ((!ok) && (!endfile));
  return order;
}

void MJDI_ModuleFilterIIR_multichanl::dspProcessPeriod()
{

  MJ_Sample *input;
  MJ_Sample *output;
  MJ_Sample *sentinel;
  uint samples;
  
  MJ_FilterDFII<MJDI_FilterCoeff> *filter;
    
  for (uint chnl=0;chnl<m_signalOutput.size();chnl++) {
    // Get input data source
    input = m_signalInput.getFast(chnl).samples().data();
    // Get destination of processed data
    output = m_signalOutput.getFast(chnl).samples().data();
    // Get length of the frame
    samples =  m_signalInput.getFast(chnl).samples().size();
    
    if(!m_bypass) {
      
      filter = & m_filters[chnl];
    
      sentinel = input + samples;

      while(input < sentinel) {
	filter->put(*input++);
	*output++ = filter->get();
      }
    }
    else
      memcpy(output, input, sizeof(MJ_Sample) * samples);

    if(!m_signalOutput.getFast(chnl).samples().allElemsFinite())
      error(MJ_ERR_FPE, "MJDI_ModuleFilterIIR_multichanl::dspProcessPeriod");
  }
}

bool MJDI_ModuleFilterIIR_multichanl::dspStop()
{
  if(m_debug) printFilters();
  
  return true;
} 
 
bool MJDI_ModuleFilterIIR_multichanl::msgProcess(const char * i,MJ_Message *m)
{
  const char * fname = "MJDI_ModuleFilterIIR_multichanl::msgProcess";


  if(!strcmp(i, MJ_Strs::filename)) {
    MJ_MessageString *ms = dynamic_cast<MJ_MessageString *> (m);

    if(!ms) return false;
    if(ms->stringConst().isEmpty()) return false;
    
    m_fileName = ms->stringConst().copy();
    return true;
  }

  if(!strcmp(i, MJ_Strs::bypass)) {
    bool ok = true;
    bool bypass = m->toBool(&ok);
    if(!ok) {
      error(MJ_ERR_ARGUMENT, fname, "\"Bypass\" not a bool message");
      return false;
    }
    m_bypass = bypass;

    trace3(m_debug, "%s # m_bypass = %d", fname, (int) m_bypass);

    return true;
  }

  return false;
}

MJ_Parameter *MJDI_ModuleFilterIIR_multichanl::msgGetInput(const char * i)
{
  // trace("MJ_ModuleDivaAudio::msgGetInput", i);

  if(!strcmp(i, MJ_Strs::filename))
    return new MJ_ParameterFilename(m_fileName, 
				    "Coefficient file", 
				    MJ_Strs::filename,
				    "*",
				    QFileDialog::ExistingFile);
  else if(!strcmp(i, MJ_Strs::bypass)) {
    return new MJ_ParameterBool(m_bypass, MJ_Strs::Bypass, MJ_Strs::bypass);
  }
  return 0;
}
						    
QCString MJDI_ModuleFilterIIR_multichanl::msgGetInputName(int i)
{
  QCString r;

  if(i == 0)
    r = MJ_Strs::filename;
  else if(i == 1)
    r = MJ_Strs::bypass;

  return r;
}

void MJDI_ModuleFilterIIR_multichanl::printFilters()
{
  MJ_FilterDFII<MJDI_FilterCoeff> *filter;

  for(uint i=0; i< m_filters.size(); i++) {
    
    filter = & m_filters[i];
    uint order = filter->getOrder();

    trace2("MJDI_ModuleFilterIIR_multichanl::dspStop # %d order = %d %p %p",
	   i, order, filter->coeffs().a(), filter->coeffs().b());
    
    for(uint j=0; j <= order; j++) {
      trace2("%d a = %f   b = %f\n", i, 
	     filter->coeffs().a()[j],  filter->coeffs().b()[j]);
    }
  }
}

// End of signal processing-related code.

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

// Instantiate templates:

template class MJ_Array<MJ_FilterDFII<MJDI_FilterCoeff> >;

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


/* The next section contains code that is used to declare this as a
   plugin..*/

/* Factorty function that is used to create new plugin instances.*/
MJ_ModuleDsp * MJ_createNew_ModuleFiltreIIR_mutlichanl() 
{
  return new MJDI_ModuleFilterIIR_multichanl;
}

MJ_PluginDbeDsp *MJ_createPlugin(int i)
{
  if(i != 0) return 0;

  MJ_PluginDbeDsp *dbe = new MJ_PluginDbeDsp;
  dbe->setAuthors("Benoit Gauduin and Tommi Ilmonen");
  dbe->setMenuPosition("_Audio_Equalizers_IIR Filter (multi channels)");
  dbe->setCreator(MJ_createNew_ModuleFiltreIIR_mutlichanl);
  dbe->setIdentifier("_equalizers_iir_filter_diva_multi_channels");
  dbe->pluginVersion()->set(0,0,2);
  dbe->setPluginDebug(MJDI_ModuleFilterIIR_multichanl::debugPtr());

  dbe->options()->setValue(MJ_DSP_EQUAL_CHANNEL_IO, 1);
  dbe->options()->setValue(MJ_DSP_MIN_CHANNELS_IN, 1);
  dbe->options()->setValue(MJ_DSP_MIN_CHANNELS_OUT, 1);
  dbe->setParameter(MJ_PP_COMPILE_TIME, DATETIME);

  return dbe;  
}
