/* -*- C++ -*- */

/* 
 * COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Authors: Tommi Ilmonen and Benoit Gauduin, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 * benoit.gauduin@freesbee.fr
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#ifndef MJDI_MODULE_FILTER_IIR_MULTICHAN1_H
#define MJDI_MODULE_FILTER_IIR_MULTICHAN1_H

/* Include base  definitions. */
#include "mj_module_dsp.h"

/* Include Direct Form II digital IIR filter implementation */
#include "mj_filter_iir.h"

// Use double or float as the filter coefficients.
typedef float MJDI_FilterCoeff;
   
/** This plugin performs multi-channels IIR filtering at n order.

    The filter is a "Direct Form II Transposed"
    implementation of the standard difference equation:

\verbatim
 
    a(0)*y(n) = b(0)*x(n) + b(1)*x(n-1) + ... + b(norder)*x(n-norder)
                          - a(1)*y(n-1) - ... - a(norder)*y(n-norder)


   The filter reads the coefficients for each channel from a file.

   File format:
   ------------
   1/ Notes
   2/ Filter's order for channel 0
   3/ b Coeficients
   4/ a Coeficients
   5/ Notes
   6/ Filter's order for channel 1
   7/ b Coeficients
   8/ a Coeficients
   9/ ...
   
   Example:

   # Example for 2 channels
   # Filter for channel 0 : order 2
   2
   0.0675
   0.1349
   0.0675
   1
   -1.143
   0.4128
   # Filter for channel 1 : order 4
   4
   ...

\endverbatim
   
  if there are more channels than filters in the file,
   the last coefficents are applied to the next channels.
   So, to apply the same filter to X channels,
   you need to put only one filter in the file.


*/

class MJDI_ModuleFilterIIR_multichanl: public MJ_ModuleDsp
{
public:
  MJDI_ModuleFilterIIR_multichanl();

  virtual ~MJDI_ModuleFilterIIR_multichanl();

  virtual bool dspPrepare();

  virtual void dspProcessPeriod();

  virtual bool dspStop();
  
  virtual bool msgProcess(const char *,MJ_Message *);
  virtual MJ_Parameter *msgGetInput(const char *);
  virtual QCString msgGetInputName(int);

  static bool *debugPtr() {return &m_debug;}

protected:

  void printFilters();
  
  /* Filter for each channel */
  MJ_Array<MJ_FilterDFII<MJDI_FilterCoeff> > m_filters;

  /* File's name containing the order and the coefficients */
  QString m_fileName;

  bool    m_bypass;
  
  /* Static variable that can be used to switch on debugging mode.*/
  static bool m_debug;

  /* Function to load the filter's order from the file */
  uint getOrder(QTextStream &, bool &);
};

#endif
