/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_1ton.h>

#include <mj_error.h> 
#include <mj_math.h> 
#include <mj_parameters.h> 
#include <mj_strs.h>

#include <strings.h>

MJ_Module1ToN::MJ_Module1ToN()
  : m_outChannels(1),
    m_chosenChannel(0)
{}

MJ_Module1ToN::~MJ_Module1ToN()
{}
  
bool MJ_Module1ToN::dspAcceptIoChans(int *in, int *out)
{
  bool change = false;

  if(*in != 1) {
    change = true;
    *in = 1;
  }

  if(*out != (int) m_outChannels) {
    change = true;
    *out = m_outChannels;
  }

  return change;
}

bool MJ_Module1ToN::dspPrepare()
{
  const char * fname = "MJ_Module1ToN::dspPrepare";

  if(m_signalInput.size() != 1) {
    error2(MJ_ERR_INTERNAL, "%s # There should be one input channel, not %d",
	   fname, m_signalInput.size());
    return false;
  }

  if(m_signalOutput.size() != m_outChannels) {
    error2(MJ_ERR_INTERNAL, "%s # There should be one input channel, not %d",
	   fname, m_signalOutput.size());
    return false;
  }

  m_interpolators.resize(m_outChannels);

  for(MJ_Ramp<MJ_Sample> *r = &* m_interpolators.begin(); 
      r < &* m_interpolators.end(); r++) {
    r->clear();
  }

  m_interpolators[m_chosenChannel].reset(1.0);

  return true;
}

void MJ_Module1ToN::dspProcessPeriod()
{
  const MJ_Vector<MJ_Sample> * invect = & m_signalInput.getFast(0).samples();
  const MJ_Sample * in = invect->data();
  uint samples = invect->size();

  for(uint i=0; i < m_outChannels; i++) {

    MJ_Ramp<MJ_Sample> *r = & m_interpolators[i];
    MJ_Sample * out = m_signalOutput.getFast(i).samples().data();
    
    if(r->isActive() || (r->get() != 0.0)) {
      r->advanceLinearMul(out, in, samples);
    }
    else {
      bzero(out, sizeof(MJ_Sample) * samples);
    }
  }
}

bool MJ_Module1ToN::msgProcess(const char *id, MJ_Message *m)
{
  const char * fname = "MJ_Module1ToN::msgProcess";

  bool ok = true;

  int tmp = m->toInt(&ok);

  if(!ok) {
    error(MJ_ERR_ARGUMENT,fname, "not an integer message");
    return false;
  }

  if(MJ_Strs::equals(id, MJ_Strs::channels)) {
    m_outChannels = MJ_clamp(tmp, 1, m_maxChans);
  }
  else if(MJ_Strs::equals(id, MJ_Strs::channel)) {
    
    int newVal = MJ_clamp(tmp, 1, m_outChannels);

    // trace3(m_debug, "%s # %d -> %d", tmp newVal);

    if(m_currentTime && m_interpolators.size()) {
      
      newVal = MJ_min((uint) newVal, m_interpolators.size());

      long now = (long) m_currentTime->stamp();
      long msgTime = m->timeStamp();

      uint interpTime = msgTime > now ? (msgTime - now) : 
	(dspSamplingRate() / 50 + 10);
      
      m_interpolators[m_chosenChannel].setTarget
	(interpTime, 0.0, MJ_RAMP_LINEAR);

      m_chosenChannel = newVal - 1;
      
      m_interpolators[m_chosenChannel].setTarget
	(interpTime, 1.0, MJ_RAMP_LINEAR);      
    }
    else
      m_chosenChannel = newVal - 1;
  }

  return true;
}

MJ_Parameter * MJ_Module1ToN::msgGetInput(const char *id)
{
  if(MJ_Strs::equals(id, MJ_Strs::channels)) {
    MJ_ParameterInt * pi = 
      new MJ_ParameterInt("Output channels", id, m_outChannels, 1, m_maxChans);
    
    pi->setFlags(MJ_PFLAG_AFFECTS);
    
    return pi;
  }
  if(MJ_Strs::equals(id, MJ_Strs::channel)) {
    return new MJ_ParameterInt("Target channel", id, m_chosenChannel + 1, 
			       1, m_outChannels);
  }

  return 0;
}

QCString MJ_Module1ToN::msgGetInputName(int i)
{
  QCString r;
  
  if(i == 0) r = MJ_Strs::channels;
  else if(i == 1) r = MJ_Strs::channel;

  return r;
}
