/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_overecho.h>

#include <mj_arraytmpl.C>
#include <mj_iostream.h>
#include <mj_mapper.h>
#include <mj_math.h>
#include <mj_messages.h>
#include <mj_parameters.h>
#include <mj_random.h>

#include <qptrstack.h> 

const char *MJ_OverechoIds[MJ_OEP_SIZEOF] = {
  "wet-dry",

  "predelay-time",
  "predelay-cut-off",
  "predelay-gain",
  "predelay-pan",

  "predel-scale-time",
  "predel-max-time",

  "max-feedback-time",

  "all-pass-time-start",
  "all-pass-time-end",
  "all-pass-gain",

  "feedback-time",
  "feedback-cut-off",
  "feedback-peq-cf",
  "feedback-peq-gain",
  "feedback-peq-bw",
  "feedback-gain",

  "output-time",
  "output-gain",
  "use-primes"
};

// The number of these parameters = MJ_ModuleOverecho::m_params.
static const char *MJ_OverechoLabels[MJ_OEP_SIZEOF] = {
  "Wet/dry ratio",

  "Predelay time",
  "Cut-off frequency",
  "Gain",
  "Pan",

  "Predelay scale time",
  "Predelay max time",

  "Maximum feedback time",

  "All-pass first time",
  "All-pass second time",
  "All-pass coefficient",

  "Feedback time",
  "Feedback cut-off frequency",
  "Feedback parameteric EQ center frequency",
  "Feedback parameteric EQ gain",
  "Feedback parameteric EQ bandwidth",
  "Feedback gain",

  "Output time",
  "Output gain",
  "Use prime numbers in all-passes"
};

bool MJ_ModuleOverecho::m_debug = false;

MJ_ModuleOverecho::MJ_ModuleOverecho()
  : m_sr(44100)
{  
  trace("MJ_ModuleOverecho::MJ_ModuleOverecho");
  defaultParams();
}

MJ_ModuleOverecho::~MJ_ModuleOverecho()
{
  trace("MJ_ModuleOverecho::~MJ_ModuleOverecho");
}
  
bool MJ_ModuleOverecho::dspPrepare()
{
  if((m_signalInput.size() != 1 && 
      m_signalInput.size() != 2) || 
     m_signalOutput.size() != 2)
    return false;

  /* Resize all buffers. */
  const MJ_SampleRateDef & srd = m_signalOutput.getFast(0).sampleRate();
  
  double rate = srd.rate();
  
  uint preDelayLen = (uint) (rate * m_predelayMaxTime);

  m_predelayBuffer[0].resize(preDelayLen);
  m_predelayBuffer[0].setAll(0.0);

  m_predelayBuffer[1].resize(preDelayLen);
  m_predelayBuffer[1].setAll(0.0);

  uint apLen = (uint) (m_allpassMaxTime * rate);

  m_allpassBuffer.resize(apLen);
  m_allpassBuffer.setAll(0.0);

  /* Build run-time values */

  bool ok = true;

  ok = ok & updatePredelays();
  ok = ok & updateAllpasses(false);
  if(m_usePrimes)
    ok = ok & updateAllpasses(true);
  ok = ok & updateOutputs();
  
  updateFeedback();

  m_feedbackFilter1.resetMemory();

  return ok;
}

void MJ_ModuleOverecho::dspProcessPeriod()
{
  // Pointers to input
  const MJ_Vector<MJ_Sample> *inLeft = & m_signalInput.getFast(0).samples();

  const MJ_Vector<MJ_Sample> *inRight;
  if(m_signalInput.size() == 1)
    inRight = inLeft;
  else
    inRight =  & m_signalInput.getFast(1).samples();
  
  uint count = inRight->size();
  
  // Pointers to output
  MJ_Vector<MJ_Sample> *outLeft = & m_signalOutput.getFast(0).samples();
  MJ_Vector<MJ_Sample> *outRight = & m_signalOutput.getFast(1).samples();

  // Fill output with direct sound.

  MJ_Sample dry = 1.0 - m_wetDry;

  MJ_Vector<MJ_Sample>::vmlBegin();
  inLeft->mulVml(dry, outLeft);
  inRight->mulVml(dry, outRight);
  MJ_Vector<MJ_Sample>::vmlEnd();

  uint i,k;
  // uint sentinel;

  /* Go over the algorithm sample by sample. This may be slightly
     inefficient, but I assume that it is not such a problem ... at
     least not while testing. */
  for(i=0; i < count; i++) {

    // Advance main all-pass buffer.
    m_allpassBuffer.advance();
    m_allpassBuffer.getNewest() = 0.0;

    float leftSample = inLeft->getConst(i);
    float rightSample = inRight->getConst(i);

    float outLeftSample = 0.0;
    float outRightSample = 0.0;
    float predelaySum = 0.0;

    m_predelayBuffer[0].advance();
    m_predelayBuffer[0].getNewest() = leftSample;

    m_predelayBuffer[1].advance();
    m_predelayBuffer[1].getNewest() = rightSample;
    
    // float sum = 0.0;

    /* Sum predelays to output and to main delay buffer 

       There is no separate gain, the all-pole filter is designed to
       take care of that.  */

    for(k=0; k < m_predelayIirs[0].size(); k++) {

      MJ_FilterAP1<float> *iir = & m_predelayIirs[0].getFast(k);
      
      MJ_IntFloat *ifp1 = & m_predelayUse[0].getFast(k);

      iir->put(m_predelayBuffer[0].getNewest(ifp1->m_i));
      float tmp = iir->get();

      iir = & m_predelayIirs[1].getFast(k);

      MJ_IntFloat *ifp2 = & m_predelayUse[1].getFast(k);

      iir->put(m_predelayBuffer[1].getNewest(ifp1->m_i));
      float tmp2 = iir->get();

      outLeftSample += tmp * (1.0 - ifp1->m_f) + tmp2 * ifp1->m_f;
      outRightSample += tmp * (1.0 - ifp2->m_f) + tmp2 * ifp2->m_f;
      
      predelaySum += tmp + tmp2;
    }

    /* Perform all-pass filtering. */

    ulong mask = m_allpassBuffer.mask();
    ulong bufferIndex = m_allpassBuffer.getSampleCount();
    float * data = m_allpassBuffer.data();

    for(k=0; k < m_allPassUse.size(); k++) {
      MJ_IntIntFloat * iif = & m_allPassUse.get(k);
      float gain = iif->m_f;
      ulong indFirst = (bufferIndex - iif->m_i1) & mask;
      ulong indSecond = (bufferIndex - iif->m_i2) & mask;
      
      
      data[indFirst] -= data[indSecond] * gain;
      data[indSecond] += data[indFirst] * gain;
      /* float tmp1 = data[indFirst] * gain;
      float tmp2 = data[indSecond] * gain;

      data[indFirst] -= tmp2;
      data[indSecond] += tmp1; */
      
    }
    
    /* Throw results to output. */
    
    for(k=0; k < m_outputUse[0].size(); k++) {
      MJ_IntFloat *pp = & m_outputUse[0].getFast(k); // Parameter pointer
      outLeftSample += data[(bufferIndex - pp->m_i) & mask] * pp->m_f;

      pp = & m_outputUse[1].getFast(k);
      outRightSample += data[(bufferIndex - pp->m_i) & mask] * pp->m_f;
    }
    
    outLeft->get(i) += outLeftSample * m_wetDry;
    outRight->get(i) += outRightSample * m_wetDry;

    /* Feedback */
    m_allpassBuffer.getNewest() += m_feedbackFilter1.putGetSample
      (m_allpassBuffer.getNewest(m_feedbackSamples) * m_feedbackParams[5] +
       predelaySum);
  }
}

bool MJ_ModuleOverecho::dspStop()
{
  m_predelayBuffer[0].resize(0);
  m_predelayBuffer[1].resize(0);
  m_allpassBuffer.resize(1);

  return true;
}
  
bool MJ_ModuleOverecho::msgProcess(const char *id, MJ_Message *m)
{
  const char * fname = "MJ_ModuleOverecho::msgProcess";

  if(m_debug) trace(fname, id);
  
  QCString idCStr(id);
  QCString idCStr2(idCStr);
  
  int first = idCStr.find('_');

  if(first > 0)
    idCStr.truncate(first);
  
  int param = MJ_findStr(idCStr, MJ_OverechoIds, MJ_OEP_SIZEOF);
  
  if(param < 0) {
    error(MJ_ERR_ARGUMENT, fname, id);
    return false;
  }
  
  /* Get integers if they exist. */
  bool hasFirst = true, hasSecond = true;

  int firstNum = MJ_fieldC(idCStr2, 1, "_").toInt(&hasFirst);
  int secondNum = MJ_fieldC(idCStr2, 2, "_").toInt(&hasSecond);
  
  /* Get double if it exists. */
  bool hasDouble = true;
  bool hasBool = true;
  double val = m->toDouble(&hasDouble);
  double bval = m->toBool(&hasBool);

  if(!hasDouble && param != MJ_OEP_USE_PRIMES) return false;

  bool update = dspIsRealTime();
  
  switch(param) {
  case MJ_OEP_WET_DRY:
    m_wetDry = val;
    break;

  case MJ_OEP_PREDELAY_TIME:
  case MJ_OEP_PREDELAY_CUT_OFF:
  case MJ_OEP_PREDELAY_GAIN:
  case MJ_OEP_PREDELAY_PAN:

    if((uint) firstNum >= 2 || !hasFirst ||
       (uint) secondNum >= m_predelayParams[0].size() || !hasSecond)
      return false;
    
    m_predelayParams[firstNum].get(secondNum).m_v[param - MJ_OEP_PREDELAY_TIME]
      = val;

    if(update)
      updatePredelay(firstNum, secondNum);

    break;

  case MJ_OEP_DEL_BUF_MAX_TIME:

    m_allpassMaxTime = val;

    break;

  case MJ_OEP_PREDEL_BUF_MAX_TIME:

    m_predelayMaxTime = val;

    break;

  case MJ_OEP_PREDEL_BUF_SCALE_TIME:
    m_predelayScaleTime = MJ_min((float)val, m_predelayMaxTime);
    if(update)
      updatePredelays();
    break;

  case MJ_OEP_ALL_PASS_TIME_START:
  case MJ_OEP_ALL_PASS_TIME_END:
  case MJ_OEP_ALL_PASS_GAIN:

    if((uint) firstNum >= m_allpassParams.size() || !hasFirst)
      return false;

    m_allpassParams.get(firstNum).m_v[param - MJ_OEP_ALL_PASS_TIME_START]
      = val;

    if(update)
      updateAllpass(firstNum, m_usePrimes);
    break;

  case MJ_OEP_FEEDBACK_TIME:
  case MJ_OEP_FEEDBACK_CUT_OFF:
  case MJ_OEP_FEEDBACK_PEQ_CF:
  case MJ_OEP_FEEDBACK_PEQ_GAIN:
  case MJ_OEP_FEEDBACK_PEQ_BW:
  case MJ_OEP_FEEDBACK_GAIN:
    m_feedbackParams[param - MJ_OEP_FEEDBACK_TIME] = val;

    if(update) {
      updateFeedback();

      if(param == MJ_OEP_FEEDBACK_TIME) {
	updateAllpasses(m_usePrimes);
	updateOutputs();
      }
    }

    break;
    
  case MJ_OEP_OUTPUT_TIME:
  case MJ_OEP_OUTPUT_GAIN:
    
    if(((uint) firstNum) >= 2 || !hasFirst ||
       ((uint) secondNum) >= m_outputParams[0].size() || !hasSecond)
      return false;
    
    m_outputParams[firstNum].get(secondNum).m_v[param - MJ_OEP_OUTPUT_TIME]
      = val;

    if(update)
      updateOutput(firstNum, secondNum);
    break;
    
  case MJ_OEP_USE_PRIMES:
    if(!bval) return false;
    m_usePrimes = bval;
    break;
  }

  return true;
}

void MJ_ModuleOverecho::msgGetInputs(MJ_Parameters *p)
{
  char buf[64];
  int i, j;

  // Direct gains:
  for(i = 0; i < 2; i++) {
    p->append(msgGetInput(MJ_OverechoIds[MJ_OEP_WET_DRY]));
  }
  
  // Predelays:
  int predelays = (int) m_predelayParams[0].size();
  for(i = 0; i < 2; i++) {
    for(j = 0; j < predelays; j++) {
    
      // Time
      sprintf(buf, "%s_%d_%d", MJ_OverechoIds[MJ_OEP_PREDELAY_TIME], i, j);
      p->append(msgGetInput(buf));
      
      // Cut-off freq
      sprintf(buf, "%s_%d_%d", MJ_OverechoIds[MJ_OEP_PREDELAY_CUT_OFF], i, j);
      p->append(msgGetInput(buf));
      
      // Gain
      sprintf(buf, "%s_%d_%d", MJ_OverechoIds[MJ_OEP_PREDELAY_GAIN], i, j);
      p->append(msgGetInput(buf));

      // Pan
      sprintf(buf, "%s_%d_%d", MJ_OverechoIds[MJ_OEP_PREDELAY_PAN], i, j);
      p->append(msgGetInput(buf));
    }
  }

  p->append(msgGetInput(MJ_OverechoIds[MJ_OEP_PREDEL_BUF_SCALE_TIME]));
  p->append(msgGetInput(MJ_OverechoIds[MJ_OEP_PREDEL_BUF_MAX_TIME]));
  p->append(msgGetInput(MJ_OverechoIds[MJ_OEP_DEL_BUF_MAX_TIME]));
  
  // All-passes
  int allpasses = (int) m_allpassParams.size();
  for(i = 0; i < allpasses; i++) {

    // Start time
    sprintf(buf, "%s_%d", MJ_OverechoIds[MJ_OEP_ALL_PASS_TIME_START], i);
    p->append(msgGetInput(buf));

    // End time
    sprintf(buf, "%s_%d", MJ_OverechoIds[MJ_OEP_ALL_PASS_TIME_END], i);
    p->append(msgGetInput(buf));
    
    // Gain
    sprintf(buf, "%s_%d", MJ_OverechoIds[MJ_OEP_ALL_PASS_GAIN], i);
    p->append(msgGetInput(buf));
  }

  // Feedback stuff:

  // Time
  for(i = MJ_OEP_FEEDBACK_TIME; i <= MJ_OEP_FEEDBACK_GAIN; i++) 
    p->append(msgGetInput(MJ_OverechoIds[i]));
  
  int outputs = (int) m_outputParams[0].size();

  for(i = 0; i < 2; i++) {
    for(j = 0; j < outputs; j++) {
      
      // Time
      sprintf(buf, "%s_%d_%d", MJ_OverechoIds[MJ_OEP_OUTPUT_TIME], i, j);
      p->append(msgGetInput(buf));
      
      // Gain
      sprintf(buf, "%s_%d_%d", MJ_OverechoIds[MJ_OEP_OUTPUT_GAIN], i, j);
      p->append(msgGetInput(buf));
    }
  }
  
  p->append(msgGetInput(MJ_OverechoIds[MJ_OEP_USE_PRIMES]));
}

MJ_Parameter * MJ_ModuleOverecho::msgGetInput(const char *id)
{
  const char * fname = "MJ_ModuleOverecho::msgGetInput";

  if(m_debug) trace(fname, id);
  
  QCString idCStr(id);
  QCString idCStr2(idCStr);
  
  int first = idCStr.find('_');

  if(first > 0)
    idCStr.truncate(first);
  
  int param = MJ_findStr(idCStr, MJ_OverechoIds, MJ_OEP_SIZEOF);
  
  if(param < 0) {
    error(MJ_ERR_ARGUMENT, fname, id);
    return 0;
  }
  
  /* Get integers if they exist. */
  bool hasFirst = true, hasSecond = true;

  int i = MJ_fieldC(idCStr2, 1, "_").toInt(&hasFirst);
  int j = MJ_fieldC(idCStr2, 2, "_").toInt(&hasSecond);

  if(m_debug) {
    trace2("%s # %d %d %d", fname, (int) param, i, j);
  }

  MJ_ParameterDouble *pd;
  
  switch(param) {
  case MJ_OEP_WET_DRY:
    return new MJ_ParameterDouble(MJ_OverechoLabels[MJ_OEP_WET_DRY],
				  id, m_wetDry, 0.0, 2.0);
  case MJ_OEP_PREDELAY_TIME:
    return new MJ_ParameterDouble(MJ_OverechoLabels[MJ_OEP_PREDELAY_TIME],
				  id, m_predelayParams[i].get(j).m_v[0], 
				  0.001, 1.0, 
				  MJ_UNIT_SECONDS);
  case MJ_OEP_PREDELAY_CUT_OFF:
    pd =  new MJ_ParameterDouble(MJ_OverechoLabels[MJ_OEP_PREDELAY_CUT_OFF],
				 id, m_predelayParams[i].get(j).m_v[1], 
				 10.0, 20000.0, MJ_UNIT_HERTZ);
    pd->setMapper(new MJ_MapperLog);
    return pd;

  case MJ_OEP_PREDELAY_GAIN:
    return new MJ_ParameterDouble(MJ_OverechoLabels[MJ_OEP_PREDELAY_GAIN],
				  id, m_predelayParams[i].get(j).m_v[2], 
				  0.0, 1.0);
  case MJ_OEP_PREDELAY_PAN:
    return new MJ_ParameterDouble(MJ_OverechoLabels[MJ_OEP_PREDELAY_PAN],
				  id, m_predelayParams[i].get(j).m_v[3],
				  0.0, 1.0);

  case MJ_OEP_PREDEL_BUF_SCALE_TIME:
    pd = 
      new MJ_ParameterDouble(MJ_OverechoLabels[MJ_OEP_PREDEL_BUF_SCALE_TIME],
			     id, m_predelayScaleTime, 0.02, m_predelayMaxTime,
			     MJ_UNIT_SECONDS);
    pd->setMapper(new MJ_MapperLog);
    return pd;
    
  case MJ_OEP_PREDEL_BUF_MAX_TIME:
    return new 
      MJ_ParameterDouble(MJ_OverechoLabels[MJ_OEP_PREDEL_BUF_MAX_TIME],
			 id, m_predelayMaxTime, 0.001, MJ_MAX_PREDEL_BUFFER,
			 MJ_UNIT_SECONDS);

  case MJ_OEP_DEL_BUF_MAX_TIME:
    return new MJ_ParameterDouble(MJ_OverechoLabels[MJ_OEP_DEL_BUF_MAX_TIME],
				  id, m_allpassMaxTime, 0.5, MJ_MAX_DEL_BUFFER,
				  MJ_UNIT_SECONDS);

  case MJ_OEP_ALL_PASS_TIME_START:
    return 
      new MJ_ParameterDouble(MJ_OverechoLabels[MJ_OEP_ALL_PASS_TIME_START],
			     id, m_allpassParams.get(i).m_v[0], 
			     0.001, 1.0);

  case MJ_OEP_ALL_PASS_TIME_END:
    return new MJ_ParameterDouble(MJ_OverechoLabels[MJ_OEP_ALL_PASS_TIME_END],
				  id, m_allpassParams.get(i).m_v[1], 
				  0.001, 1.0);
  case MJ_OEP_ALL_PASS_GAIN:
    return new MJ_ParameterDouble(MJ_OverechoLabels[MJ_OEP_ALL_PASS_GAIN],
				  id, m_allpassParams.get(i).m_v[2], 
				  0.0, 0.95);
  case MJ_OEP_FEEDBACK_TIME:
    pd = 
      new MJ_ParameterDouble(MJ_OverechoLabels[MJ_OEP_FEEDBACK_TIME],
				  MJ_OverechoIds[MJ_OEP_FEEDBACK_TIME], 
				  m_feedbackParams[0], 0.1, m_allpassMaxTime,
				  MJ_UNIT_SECONDS);
    pd->setMapper(new MJ_MapperLog);
    return pd;

  case MJ_OEP_FEEDBACK_CUT_OFF:
    return new MJ_ParameterDouble(MJ_OverechoLabels[MJ_OEP_FEEDBACK_CUT_OFF],
				  MJ_OverechoIds[MJ_OEP_FEEDBACK_CUT_OFF], 
				  m_feedbackParams[1], 10, 20000.0,
				  MJ_UNIT_HERTZ);
  case MJ_OEP_FEEDBACK_PEQ_CF:
    return new MJ_ParameterDouble(MJ_OverechoLabels[MJ_OEP_FEEDBACK_PEQ_CF],
				  MJ_OverechoIds[MJ_OEP_FEEDBACK_PEQ_CF], 
				  m_feedbackParams[2], 10, 20000.0,
				  MJ_UNIT_HERTZ);
  case MJ_OEP_FEEDBACK_PEQ_GAIN:
    return new MJ_ParameterDouble(MJ_OverechoLabels[MJ_OEP_FEEDBACK_PEQ_GAIN],
				  MJ_OverechoIds[MJ_OEP_FEEDBACK_PEQ_GAIN], 
				  m_feedbackParams[3], -15.0, 0.0,
				  MJ_UNIT_DECIBEL);
  case MJ_OEP_FEEDBACK_PEQ_BW:
    return new MJ_ParameterDouble(MJ_OverechoLabels[MJ_OEP_FEEDBACK_PEQ_BW],
				  MJ_OverechoIds[MJ_OEP_FEEDBACK_PEQ_BW], 
				  m_feedbackParams[4], 0.3, 2.0,
				  MJ_UNIT_BLANK); // Should be octave...
  case MJ_OEP_FEEDBACK_GAIN:
    return new MJ_ParameterDouble(MJ_OverechoLabels[MJ_OEP_FEEDBACK_GAIN],
				  MJ_OverechoIds[MJ_OEP_FEEDBACK_GAIN], 
				  m_feedbackParams[5], 0.0, 0.95);
  case MJ_OEP_OUTPUT_TIME:
    return new MJ_ParameterDouble(MJ_OverechoLabels[MJ_OEP_OUTPUT_TIME],
				  id, m_outputParams[i].get(j).m_v[0], 
				  0.001, 1.0);
  case MJ_OEP_OUTPUT_GAIN:
    return new MJ_ParameterDouble(MJ_OverechoLabels[MJ_OEP_OUTPUT_GAIN],
				  id, m_outputParams[i].get(j).m_v[1], 
				  0.0, 2.0);
  case MJ_OEP_USE_PRIMES:
    return new MJ_ParameterBool(m_usePrimes,
				MJ_OverechoLabels[MJ_OEP_USE_PRIMES],
				id);

  }
  return 0;
}

void MJ_ModuleOverecho::defaultParams()
{
  const char * fname = "MJ_ModuleOverecho::defaultParams";

  if(m_debug) trace(fname, "ENTER");

  m_wetDry = 0.5;

  m_allpassMaxTime = 15.0;
  m_predelayMaxTime = 7.0;
  m_predelayScaleTime = 5.0;

  m_feedbackParams[0] = 3.0;    // Delay time
  m_feedbackParams[1] = 2000.0; // Cut-off
  m_feedbackParams[2] = 200.0;
  m_feedbackParams[3] = -5.0;
  m_feedbackParams[4] = 1.0;
  m_feedbackParams[5] = 0.7;
  
  m_usePrimes = true;

  setPredelays(0);
  setPredelays(8);

  if(m_debug) trace(fname, "GOT PREDELAYS");

  setAllpasses(0);
  setAllpasses(8);

  setoutputs(0);
  setoutputs(5);

  updateFeedback();

  if(m_debug) trace(fname, "EXIT");
}


int MJ_ModuleOverecho::allPassRecursion(int tap)
{
  int maxRec = 0;

  MJ_TriSample *pp = & m_allpassParams.get(tap);

  for(uint i=0; i < m_allpassParams.size(); i++) {
    MJ_TriSample *child = & m_allpassParams.getFast(i);

    MJ_Sample plow = MJ_min(pp->m_v[0], pp->m_v[1]);
    MJ_Sample phigh = MJ_max(pp->m_v[0], pp->m_v[1]);

    MJ_Sample clow = MJ_min(child->m_v[0], child->m_v[1]);
    MJ_Sample chigh = MJ_max(child->m_v[0], child->m_v[1]);
    
    if((clow > plow) && (chigh < phigh)) {
      
      int tmpRec = allPassRecursion(i) + 1;

      if(maxRec < tmpRec) maxRec = tmpRec;
    }
  }
  
  return maxRec;
}

int MJ_ModuleOverecho::maximumAllPassRecursion()
{
  int maxRec = 0;

  for(uint i=0; i < m_allpassParams.size(); i++) {
    int tmpRec = allPassRecursion(i);
    if(maxRec < tmpRec) maxRec = tmpRec;
  }
  
  return maxRec;
}

void MJ_ModuleOverecho::setPredelays(uint n)
{
  uint oldSize = m_predelayParams[0].size();

  m_predelayParams[0].resize(n);
  m_predelayParams[1].resize(n);

  for(uint i = oldSize; i < n; i++) {
    defaultPredelay(& m_predelayParams[0].get(i), 0, i);
    defaultPredelay(& m_predelayParams[1].get(i), 1, i);
  }
}

void MJ_ModuleOverecho::setAllpasses(uint n)
{
  uint oldSize = m_allpassParams.size();

  m_allpassParams.resize(n);

  
  for(uint i = oldSize; i < n; i++) {
    defaultAllpass(& m_allpassParams.get(i), i);
  }

  validateAllpasses();
}

void MJ_ModuleOverecho::setoutputs(uint n)
{
  uint oldSize = m_outputParams[0].size();

  m_outputParams[0].resize(n);
  m_outputParams[1].resize(n);

  for(uint i = oldSize; i < n; i++) {
    defaultOutput(& m_outputParams[0].get(i), 0, i);
    defaultOutput(& m_outputParams[1].get(i), 1, i);
  }
}

void MJ_ModuleOverecho::defaultPredelay(MJ_QuadSample *p, uint chan, uint n)
{

  uint predelays = MJ_max((uint) 1, m_predelayParams[0].size());
  
  float time = tanh((double) (n * 2 + chan + 1) / (double) predelays);

  while (time > 1.0) time -= 1.0;

  p->m_v[0] = time;
  p->m_v[1] = (1.2 - time) * 3000.0;
  p->m_v[2] = 1.0 / (1.0 + time);

  // Repeatable pseudo-random number:

  MJ_Random r(n * 2 + chan + 1);

  for(uint i=0; i < 10; i++)
    r.get();

  p->m_v[3] = r.getFloat0();

  trace2("MJ_ModuleOverecho::defaultPredelay pan = %g", p->m_v[3] );
}

void MJ_ModuleOverecho::defaultAllpass(MJ_TriSample *p, uint n)
{
  float duration  = 0.8;
  float startTime = duration * 0.1;

  for(uint i=0; i < n; i++) {
    duration *= 0.58117562;
    startTime += duration * 0.2;
  }
  
  p->m_v[0] = startTime;
  p->m_v[1] = startTime + duration;
  p->m_v[2] = (1.0 - MJ_sqrt(duration)) * 0.3
;
}

void MJ_ModuleOverecho::defaultOutput(MJ_DualSample *p, uint chan, uint n)
{
  float rel = tanh((float) (2 * n + ((chan + n) & 1)) / 6.0);
  p->m_v[0] = rel;
  p->m_v[1] = 1.0 - rel * 0.5;
}

void MJ_ModuleOverecho::validateAllpasses()
{
  const char * fname = "MJ_ModuleOverecho::validateAllpasses";
  char buf[128];

  int currTap = -1;
  float currTime = 0.0;
  float nextTime = 0.0;
  QPtrStack<int> tapStack;

  tapStack.setAutoDelete(true);

  while((currTap = findNextAllPass(currTime, & nextTime)) >= 0) {
    MJ_TriSample &ts = m_allpassParams.getFast(currTap);
    
    // The loop closes, check that it is ok.
    if(ts.m_v[1] == nextTime) {
      int *i = tapStack.pop();

      if(*i != currTap) {
	sprintf(buf, "%s # %d != %d, ", fname, *i, currTap);
	error(MJ_ERR_INTERNAL, buf);
      }
      delete i;
    }
    else {
      // The loop opens, push tap to stack.
      int *i = new int(currTap);
      tapStack.push(i);
    }
    currTime = nextTime;
  }
}

int MJ_ModuleOverecho::findNextAllPass(float prevTime, float *newTime)
{
  uint sentinel = m_allpassParams.size();
  
  int best = -1;

  float lowest = 1e+10;

  for(uint i=0; i < sentinel; i++) {
    MJ_TriSample &ts = m_allpassParams.getFast(i);

    float tmp;

    if(ts.m_v[0] > prevTime && (tmp = (ts.m_v[0] - prevTime)) < lowest) {
      *newTime = ts.m_v[0];
      lowest = tmp;
      best = i;
    }
    if(ts.m_v[1] > prevTime && (tmp = (ts.m_v[1] - prevTime)) < lowest) {
      *newTime = ts.m_v[1];
      lowest = tmp;
      best = i;
    }
  }
  return best;
}


int MJ_ModuleOverecho::findPreviousAllPassPoint(uint start)
{
  uint r = 0;

  for(uint i=0; i < m_allPassUse.size(); i++) {

    MJ_IntIntFloat *aptap = & m_allPassUse[i];
    
    if(aptap->m_i1 > r && aptap->m_i1 < start)
      r = aptap->m_i1;

    if(aptap->m_i2 > r && aptap->m_i2 < start)
      r = aptap->m_i2;
  }

  return r;
}

bool MJ_ModuleOverecho::updatePredelays()
{
  const char * fname = "MJ_ModuleOverecho::updatePredelays";

  uint predelays = m_predelayParams[0].size();
  
  if(!m_predelayUse[0].resize(predelays, false) ||
     !m_predelayUse[1].resize(predelays, false) ||
     !m_predelayIirs[0].resize(predelays, false) ||
     !m_predelayIirs[1].resize(predelays, false)) {
    error(MJ_ERR_MEMORY, fname);
    return false;
  }

  bool ok = true;

  for(uint i = 0; i < predelays; i++) {
    ok = ok & updatePredelay(0, i);
    ok = ok & updatePredelay(1, i);
  }
  return ok;
}

bool MJ_ModuleOverecho::updateAllpasses(bool usePrimes)
{
  uint allpasses = m_allpassParams.size();
  
  if(!m_allPassUse.resize(allpasses, false))
    return false;

  bool ok = true;

  for(uint i = 0; i < allpasses; i++) {
    ok = ok & updateAllpass(i, usePrimes);
  }

  return ok;
}

bool MJ_ModuleOverecho::updateOutputs()
{
  uint outputs = m_outputParams[0].size();

  if(!m_outputUse[0].resize(outputs, false) ||
     !m_outputUse[1].resize(outputs, false))
    return false;
  
  bool ok = true;

  for(uint i = 0; i < outputs; i++) {
    ok = ok & updateOutput(0, i);
    ok = ok & updateOutput(1, i);
  }

  return ok;
}


bool MJ_ModuleOverecho::updatePredelay(uint chan, uint i)
{
  MJ_QuadSample *pdp = & m_predelayParams[chan].get(i);
  
  MJ_FilterAP1<float> *iir = & m_predelayIirs[chan].get(i);
  
  iir->designLowPass(pdp->m_v[1] / m_sr * 2.0);
  iir->mulGain(pdp->m_v[2]);
  
  MJ_IntFloat *ifp = & m_predelayUse[chan].get(i);

  ifp->m_i = (uint) (pdp->m_v[0] * m_sr * m_predelayScaleTime);

  if(m_usePrimes && ifp->m_i > 30)
    ifp->m_i = MJ_previousPrime(ifp->m_i );

  ifp->m_f = pdp->m_v[3];
  
  return true;
}

bool MJ_ModuleOverecho::updateAllpass(uint i, bool usePrimes)
{
  MJ_TriSample *apparams = & m_allpassParams.get(i);
  MJ_IntIntFloat *apuse = & m_allPassUse.get(i);

  apuse->m_i1 = (uint) (apparams->m_v[0] * m_sr * feedbackDelayTime());
  apuse->m_i2 = (uint) (apparams->m_v[1] * m_sr * feedbackDelayTime());
  apuse->m_f  = apparams->m_v[2];

  if(usePrimes) {
    uint i2 = apuse->m_i2;

    uint limit = findPreviousAllPassPoint(i2) + 2;
    uint distance = i2 - apuse->m_i1;
    uint startDistance = distance;
    uint sentinel = i2 - limit;

    while(!MJ_isPrime(distance) && distance > sentinel) {
      distance--;
    }
    if(MJ_isPrime(distance))
      apuse->m_i2 = apuse->m_i1 + distance;

    trace3(m_debug, 
	   "MJ_ModuleOverecho::updateAllpass # tried primes: %d -> %d %d %d %d",
	   startDistance, distance, sentinel, 
	   limit, (int) MJ_findDivider(distance));
    
  }

  return (apuse->m_i2 > apuse->m_i1) && 
    (MJ_abs(apuse->m_f) < 1.0);
}

bool MJ_ModuleOverecho::updateOutput(uint chan, uint i)
{
  MJ_DualSample *oparams = & m_outputParams[chan].get(i);
  MJ_IntFloat *ouse = & m_outputUse[chan].get(i);

  ouse->m_i = (uint) (oparams->m_v[0] * m_sr * feedbackDelayTime());

  if(m_usePrimes && ouse->m_i > 30)
    ouse->m_i = MJ_previousPrime(ouse->m_i); 

  ouse->m_f = oparams->m_v[1];
  
  return true;
}

void MJ_ModuleOverecho::updateFeedback()
{
  float nyquist = dspSamplingRate() * 0.5;

  m_feedbackFilter1.designLowPass(m_feedbackParams[1] / nyquist);

  m_feedbackSamples = (uint) (m_feedbackParams[0] * dspSamplingRate());
}

template class MJ_Array<MJ_IntFloat>;
template class MJ_Array<MJ_IntIntFloat>;

template class  QStack<int>;

