/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#ifndef MJ_MODULE_OVERECHO_H
#define MJ_MODULE_OVERECHO_H

#include <mj_module_dsp.h>

#include <mj_filter_ap1.h>
#include <mj_multi_sample.h>
#include <mj_ringbuffer.h>

#define MJ_MAX_DEL_BUFFER 40.0
#define MJ_MAX_PREDEL_BUFFER 15.0

struct MJ_IntIntFloat {
  uint m_i1;
  uint m_i2;
  float m_f;
};

struct MJ_IntFloat {
  uint m_i;
  float m_f;
};

/** Parameters for MJ_ModuleOverecho. */

enum MJ_OverEchoParams {
  /// Direct gains
  MJ_OEP_WET_DRY,

  /// Predelay times, 2*n
  MJ_OEP_PREDELAY_TIME,
  /// Predelay cut-off frequency, 2*n
  MJ_OEP_PREDELAY_CUT_OFF,
  /// Predelay gains, 2*n
  MJ_OEP_PREDELAY_GAIN,
  /// Predelay gains, 2*n
  MJ_OEP_PREDELAY_PAN,

  /// Predelay buffer size
  MJ_OEP_PREDEL_BUF_SCALE_TIME,

  /// Predelay buffer size
  MJ_OEP_PREDEL_BUF_MAX_TIME,

  /// Maximum all-pass buffer length
  MJ_OEP_DEL_BUF_MAX_TIME,
  
  /// All-pass start times, m
  MJ_OEP_ALL_PASS_TIME_START,
  /// All-pass end times, m
  MJ_OEP_ALL_PASS_TIME_END,
  /// All-pass gain, m
  MJ_OEP_ALL_PASS_GAIN,

  /// Feedback cut-off frequency
  MJ_OEP_FEEDBACK_TIME,
  /// Feedback cut-off frequency
  MJ_OEP_FEEDBACK_CUT_OFF,
  /// Feedback parametric EQ center frequency
  MJ_OEP_FEEDBACK_PEQ_CF,
  /// Feedback parametric EQ band gain (must be negative)
  MJ_OEP_FEEDBACK_PEQ_GAIN,
  /// Feedback parametric EQ band band width
  MJ_OEP_FEEDBACK_PEQ_BW,
  /// Feedback gain
  MJ_OEP_FEEDBACK_GAIN,

  /// Output time, k
  MJ_OEP_OUTPUT_TIME,
  /// Output gain, k
  MJ_OEP_OUTPUT_GAIN,

  /// Use prime numbers when posible in the all-pass delay times
  MJ_OEP_USE_PRIMES,

  /// Number of parameter types.
  MJ_OEP_SIZEOF
};

extern const char * MJ_OverechoIds[MJ_OEP_SIZEOF];

/** Rather an extreme over-reverb. Continues the saga of my old NEX
    application.

    This is a largish Mustajuuri plugin. It is totally dynamic. The
    number of parameters may well grow to hundreds. Unfortunately
    there is no easy way automate their value creation, but we must
    depend on the user.

@see MJ_OverEchoParams */
class MJ_ModuleOverecho : public MJ_ModuleDsp
{
public:
  MJ_ModuleOverecho();

  virtual ~MJ_ModuleOverecho();
  

  virtual bool dspPrepare();
  virtual void dspProcessPeriod();
  
  /** Frees memory. Since this plugin uses a lot of memory it is best
      to free it when it is no longer needed. */
  virtual bool dspStop();

  /** Pretty-prints the message to the screen. */
  
  virtual bool msgProcess(const char *, MJ_Message *);

  virtual void msgGetInputs(MJ_Parameters *);
  virtual MJ_Parameter * msgGetInput(const char *);

  MJ_Array<MJ_QuadSample> & predelayParams(int chan) 
    { return m_predelayParams[chan]; }

  static float predelayMaxTime() { return 1.0; }
  
  MJ_Array<MJ_TriSample> & allpassParams() { return m_allpassParams; }

  float allpassMaxTime() const { return 1.0; }
  
  float *feedbackParams() { return m_feedbackParams; }

  MJ_Array<MJ_DualSample> & outputParams(uint i) 
    { return m_outputParams[i]; }

  /** Tells how many levels of recursion are beneath this all-pass
      structure. */
  int allPassRecursion(int tap);
  int maximumAllPassRecursion();

  void setPredelays(uint);
  void setAllpasses(uint);
  void setoutputs(uint);

  static bool * debugPtr() {return & m_debug;}
protected:

  void defaultParams();
  
  void defaultPredelay(MJ_QuadSample *p, uint chan, uint n);
  void defaultAllpass(MJ_TriSample *p, uint n);
  void defaultOutput(MJ_DualSample *p, uint chan, uint n);

  void validateAllpasses();

  bool allpassesValid();

  int findNextAllPass(float prevTime, float *newTime);

  int findPreviousAllPassPoint(uint start);

  bool updatePredelays();
  bool updateAllpasses(bool);
  bool updateOutputs();

  bool updatePredelay(uint chan, uint i);
  bool updateAllpass(uint, bool);
  bool updateOutput(uint chan, uint i);

  void updateFeedback();

  float feedbackDelayTime() { return m_feedbackParams[0]; }

  MJ_Sample m_wetDry;
  
  MJ_RingBuffer2<float>  m_predelayBuffer[2];

  /** Predelay parameters. Contains four values for each tap: 1) time,
      2) cut-off frequency (Hz), 3) gain (linear) and pan (0-1). */
  MJ_Array<MJ_QuadSample> m_predelayParams[2];

  MJ_Array<MJ_FilterAP1<float> > m_predelayIirs[2];
  /* Time and pan*/
  MJ_Array<MJ_IntFloat>                m_predelayUse[2];
  
  /// Maximum all delay time (seconds)
  float                  m_allpassMaxTime;

  /// Predelay scaling factor. 
  float                  m_predelayScaleTime;
  /** Maximum preedlay time (seconds). */
  float                  m_predelayMaxTime;

  MJ_RingBuffer2<float>  m_allpassBuffer;

  /** All-pass parameters. Contains three values for each loop: first
      time, second time and coefficient. The times are seconds. */
  MJ_Array<MJ_TriSample> m_allpassParams;

  /** Parameters that are used by the real DSP-code. */
  
  MJ_Array<MJ_IntIntFloat> m_allPassUse;

  /** Feedback parameters. Time, cutf, peq-cf, peq-gain, pew-bw,
      gain*/
  float                  m_feedbackParams[6];
  uint                   m_feedbackSamples;

  MJ_FilterAP1<float>    m_feedbackFilter1;

  /** Output parameters. Each tap has two values: time and gain. */
  
  MJ_Array<MJ_DualSample> m_outputParams[2];
  MJ_Array<MJ_IntFloat>   m_outputUse[2];
  
  float                   m_sr;
  bool                    m_usePrimes;

  static bool m_debug;
};

#endif
