/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_overecho_gui.h>

#include <mj_cstr_remap.h> 
#include <mj_dsp_plugin.h>
#include <mj_error.h>
#include <mj_graphic_ellipse.h> 
#include <mj_graphic_polygon.h> 
#include <mj_graphics_context_simple.h> 
#include <mj_gui_numeric.h> 
#include <mj_gui_utils.h> 
#include <mj_mapper.h>
#include <mj_messages.h>
#include <mj_module_menubar.h>
#include <mj_module_overecho.h>
#include <mj_parameters.h> 
#include <mj_slider.h> 
#include <mj_strs.h> 
#include <mj_tr.h> 


#include <qcheckbox.h> 
#include <qcombobox.h> 
#include <qfontinfo.h> 
#include <qlabel.h> 
#include <qlayout.h> 
#include <qpushbutton.h> 
#include <qtabwidget.h> 

static int MJ_OverEchoGrow(int i)
{
  return (i + 1) * 2;
}

MJ_OverEchoResize::MJ_OverEchoResize(MJ_ModuleOverecho_QG *echoGui,
				     MJ_RtControl *rtparent,
				     QWidget      *parent,
				     const char   *name)
  : QGroupBox(MJ_Tr::tr("Overecho size"), parent, name),
    MJ_RtControl(rtparent),
    m_echoGui(echoGui)
{
  MJ_ModuleOverecho *echo = echoGui->echo();

  lock();
  m_sizes[0] = echo->predelayParams(0).size();
  m_sizes[1] = echo->allpassParams().size();
  m_sizes[2] = echo->outputParams(0).size();
  unlock();
  
  const char * ids[] = {
    "Pre-delays",
    "Allpasses",
    "Outputs"
  };

  const char * slotStrs[] = {
    SLOT(setPredelays(int)),
    SLOT(setAllpasses(int)),
    SLOT(setOutputs(int)),
  };

  QFontMetrics fm(fontMetrics());

  QGridLayout *l = new QGridLayout(this, 
				   4, 3, fm.height(), 4);
  int i, j, k;

  QString tmp;

  for(i=0; i < 3; i++) {
    l->addWidget(new QLabel(MJ_Tr::tr(ids[i]), this), i, 0);
    QComboBox *cb = new QComboBox(this, ids[i]);

    for(j = 0; j < 6; j++) {
      k = MJ_OverEchoGrow(j);
      tmp.setNum(k);
      cb->insertItem(tmp);

      if(k == m_sizes[i])
	cb->setCurrentItem(j);
    }
    
    l->addWidget(cb, i, 1);
    connect(cb, SIGNAL(activated(int)),
	    this, slotStrs[i]);
    m_combos[i] = cb;
  }

  QHBoxLayout *hb = new QHBoxLayout();

  QPushButton *applyButton = new QPushButton(MJ_Tr::tr(MJ_Strs::Apply), this);
  connect(applyButton, SIGNAL(pressed()), this, SLOT(apply()));
  hb->addWidget(applyButton);
  m_applyButton = applyButton;
  l->addMultiCellLayout(hb, i, i, 0, 1);

  realTimeChange(isRealTime());
}

MJ_OverEchoResize::~MJ_OverEchoResize()
{
  m_echoGui->forgetResize();
}

void MJ_OverEchoResize::apply()
{
  m_echoGui->resizeEcho(m_sizes[0], m_sizes[1], m_sizes[2]);
  delete this;
}

void MJ_OverEchoResize::setPredelays(int p)
{
  m_sizes[0] = MJ_OverEchoGrow(p);
}

void MJ_OverEchoResize::setAllpasses(int p)
{
  m_sizes[1] = MJ_OverEchoGrow(p);
}

void MJ_OverEchoResize::setOutputs(int p)
{
  m_sizes[2] = MJ_OverEchoGrow(p);
}

void MJ_OverEchoResize::realTimeChange(bool nowRt)
{
  m_applyButton->setEnabled(! nowRt);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

bool MJ_ModuleOverecho_QG::m_debug = false;

MJ_ModuleOverecho_QG::MJ_ModuleOverecho_QG(MJ_RtControl *rtparent,
					   MJ_ModuleDsp *module,
					   QWidget      *parent,
					   const char   *name)
  : MJ_ModuleDspGuiPostCard(rtparent, module, true, 0, 
			    parent ? MJ_GUI_POSTCARD : MJ_GUI_OWN_WINDOW, 
			    parent, name),
  m_echo((MJ_ModuleOverecho *) module),
  m_graphics(this, "echographs"),
  m_resize(0),
  m_tw(0)
{
  for(int i=0; i< 5; i ++)
    m_twChildren[i] = 0;

  m_children.setAutoDelete(true);

  draw();
}

MJ_ModuleOverecho_QG::~MJ_ModuleOverecho_QG()
{
  delete m_resize;
}

void MJ_ModuleOverecho_QG::moduleModified(const char *)
{}

MJ_ModuleDspGui *MJ_ModuleOverecho_QG::createNew(MJ_RtControl *rtparent,
						 MJ_ModuleDsp *module,
						 QWidget *parent, 
						 const char *name)
{
  return new MJ_ModuleOverecho_QG(rtparent, module, parent, name);
}

void MJ_ModuleOverecho_QG::handleParam(const char *id, double v)
{
  const char * fname = "MJ_ModuleOverecho_QG::handleParam";

  if(m_debug) trace(fname, id);

  setValue(id, v);

  redrawGraphics();
  m_graphics.redraw(true);
  if(m_debug) trace(fname, "EXIT");
}

void MJ_ModuleOverecho_QG::resizeEcho
(int predelays, int allpasses, int outputs)
{
  const char * fname = "MJ_ModuleOverecho_QG::resizeEcho";

  if(isRealTime()) 
    MJ_fatal(MJ_ERR_INTERNAL, fname);

  if(m_debug) trace(fname, this);

  m_echo->setPredelays(predelays);
  m_echo->setAllpasses(allpasses);
  m_echo->setoutputs(outputs);
  
  // Get the new parameters and rebuild undo-stack
  setModuleInternal();

  rebuildTw();
  m_tw->show();

  if(m_debug) trace(fname, "EXIT");
}

void MJ_ModuleOverecho_QG::resizer()
{
  if(m_resize) {
    m_resize->show();
    m_resize->raise();
  }
  else {
    m_resize = new MJ_OverEchoResize(this, this, 0, 0);
    m_resize->show();
  }
}

void MJ_ModuleOverecho_QG::draw()
{
  const char * fname = "MJ_ModuleOverecho_QG::draw";

  if(m_debug) trace(fname, this);

  lock();
  m_parameters.clear();
  m_echo->msgGetInputs( & m_parameters);
  unlock();

  QVBoxLayout *mainBox = dynamic_cast<QVBoxLayout *> (child("mainlayout"));
  
  MJ_cpf(mainBox, fname);

  m_graphics.setBackgroundColor(Qt::white);

  redrawGraphics();

  m_graphics.resize(500, 300);
  m_graphics.setMinimumSize(300, 200);
  mainBox->addWidget(&m_graphics, 50);
  
  MJ_ModuleMenubar *mbar = 
    dynamic_cast<MJ_ModuleMenubar *> (child("mj_menubar"));

  if(mbar) {
    mbar->edit()->insertSeparator();
    mbar->edit()->insertItem(MJ_Tr::tr("Resize"), this, SLOT(resizer()));
  }

  rebuildTw();
}

void  MJ_ModuleOverecho_QG::redrawGraphics()
{
  const char * fname = "MJ_ModuleOverecho_QG::redrawGraphics";
  // Constants to guide the drawing:
  const float prebufLeft = 1.75; // Left border of predelay buffer
  const float prebufRight = 4.0; // Left border of predelay buffer
  const float prebufTop = 2.0;
  const float prebufBottom = 1.75;
  
  const float lpfLeft = prebufRight; // Left border of filter images
  const float lpfRight = 4.5;        // Right border of filter images
  const float lpTop = 1.5;
  const float lpBottom = 0.5;

  const float preGainLeft = 5.0;    // Left border of predelay gains
  const float preGainRight = 5.25;  // Right border of predelay gains

  const float preSumLeft = 5.5;     // Left border of predelay sum
  const float preSumRight = 5.75;   // Right border of predelay sum

  const float allPassLeft = preSumLeft;
  const float allPassRight = 10.0;
  const float allPassWidth = allPassRight - allPassLeft;
  
  const float allPassHeightp2 = (prebufTop - prebufBottom) * 0.5;

  const float preToAllLeft = (preSumLeft + preGainRight) * 0.5;
  const float preToAllY = allPassHeightp2 * 0.5;

  const float outputLeft = allPassRight + 0.5;
  const float outputTop     = prebufTop;
  const float outputBottom  = prebufBottom;
  
  MJ_GraphicsContextPen defaultGC;
  defaultGC.pen() = QPen(QColor(200, 40, 40));

  MJ_GraphicsContextPenBrush ellipseGC;
  ellipseGC.brush().setStyle(SolidPattern);
  ellipseGC.pen() = QPen(QColor(200, 40, 40));
  QColor gainColor(70, 0, 0);
  
  // defaultGC.clone();

  uint i, j;

  if(m_debug) trace(fname, this);

  MJ_PointArray2D points;

  m_graphics.clear();
  // Matches the xfig image (more or less)
  m_graphics.setBoundingBox(1.0, 2.5, 12.0, -2.5);
  
  // Predelay buffers:

  points.resize(4);
  points.setPoint(0, prebufLeft, prebufBottom);
  points.setPoint(1, prebufRight, prebufBottom);
  points.setPoint(2, prebufRight, prebufTop);
  points.setPoint(3, prebufLeft, prebufTop);

  addPolygon(&points, &defaultGC);

  points.mirrorY();
  addPolygon(&points, &defaultGC);

  // Filter boxes:

  lock();
  
  uint predelays = m_echo->predelayParams(0).size();
  float predelayMaxTime = m_echo->predelayMaxTime();

  unlock();

  const float lpHeight = (lpTop - lpBottom) / (float) ((predelays * 2) - 1);
  const float lpHeightp2 = lpHeight * 0.5;
  
  for(i=0; i < predelays; i++) {
    const float top = lpTop - i * 2 * lpHeight;
    const float bottom = lpTop - (i * 2 + 1) * lpHeight;

    points.setPoint(0, lpfLeft, bottom);
    points.setPoint(1, lpfRight, bottom);
    points.setPoint(2, lpfRight, top);
    points.setPoint(3, lpfLeft, top);
    
    addPolygon(&points, &defaultGC);

    points.mirrorY();
    addPolygon(&points, &defaultGC);

  }

  // Predelay time wires:

  MJ_QuadSample qs1, qs2;
  MJ_GraphicEllipse  *ellipse;

  const float preWireBottom = lpTop - ((predelays - 1) + 0.5) * lpHeight;

  for(i=0; i < predelays; i++) {
    // Predelay buffer to LP-filters:
    lock();
    qs1 = m_echo->predelayParams(0).get(i);
    qs2 = m_echo->predelayParams(1).get(i);
    unlock();
    
    float tapTime = qs1.m_v[0];
    float relative = tapTime / predelayMaxTime;

    float left = prebufLeft + (prebufRight - prebufLeft) * relative;
    float y = lpTop - (i * 2 + 0.5) * lpHeight;
    
    points.resize(3);
    points.setPoint(0, left, prebufBottom);
    points.setPoint(1, left, y);
    points.setPoint(2, lpfLeft, y);

    addPolyline(&points, & defaultGC);

    tapTime = qs2.m_v[0];
    relative = tapTime / predelayMaxTime;
    left = prebufLeft + (prebufRight - prebufLeft) * relative;

    points.setPoint(0, left, prebufBottom);
    points.setPoint(1, left, y);

    points.mirrorY();
    addPolyline(&points, & defaultGC);

    // LP-filters to gains:
    points.resize(2);
    points.setPoint(0, lpfRight, y);
    points.setPoint(1, preGainLeft, y);
    
    addPolyline(&points, & defaultGC);

    points.mirrorY();
    addPolyline(&points, & defaultGC);

    // Gains:
    ellipse = new MJ_GraphicEllipse(preGainLeft, y - lpHeightp2,
				    preGainRight, y + lpHeightp2);
    ellipseGC.brush().setColor(gainColor.light(100 + (int)(qs1.m_v[2] * 400)));
    ellipse->setGraphicsContext(ellipseGC.clone());
    m_graphics.appendGraphic(ellipse);

    ellipse = new MJ_GraphicEllipse(preGainLeft, - y - lpHeightp2,
				    preGainRight, - y + lpHeightp2);
    ellipseGC.brush().setColor(gainColor.light(100 + (int)(qs2.m_v[2] * 300)));
    ellipse->setGraphicsContext(ellipseGC.clone());
    m_graphics.appendGraphic(ellipse);

    // The wires to gain sums:
    points.setPoint(0, preGainRight, y);
    points.setPoint(1, preSumLeft, preWireBottom);
    
    addPolyline(&points, & defaultGC);

    points.mirrorY();
    addPolyline(&points, & defaultGC);
  }

  // Predelay sums:
  
  ellipse = new MJ_GraphicEllipse(preSumLeft, preWireBottom - lpHeightp2,
				  preSumRight, preWireBottom + lpHeightp2);
  ellipse->setGraphicsContext(defaultGC.clone());
  m_graphics.appendGraphic(ellipse);

  ellipse = new MJ_GraphicEllipse(preSumLeft, - preWireBottom - lpHeightp2,
				  preSumRight, - preWireBottom + lpHeightp2);
  ellipse->setGraphicsContext(defaultGC.clone());
  m_graphics.appendGraphic(ellipse);

  // Predelay to all-pass:

  points.resize(3);
  points.setPoint(0, (preSumLeft + preSumRight) * 0.5, 
		  preWireBottom - lpHeightp2);
  points.setPoint(1, preToAllLeft, preToAllY);
  points.setPoint(2, allPassLeft, preToAllY);

  addPolyline(&points, & defaultGC);

  points.mirrorY();
  addPolyline(&points, & defaultGC);

  // All-pass delay buffer:
  points.resize(4);
  points.setPoint(0, allPassLeft, - allPassHeightp2);
  points.setPoint(1, allPassRight, - allPassHeightp2);
  points.setPoint(2, allPassRight, allPassHeightp2);
  points.setPoint(3, allPassLeft, allPassHeightp2);

  addPolygon(&points, &defaultGC);

  // All-passes:

  MJ_GraphicsContextPen allPassGC; // (QPen(gray));
  QColor apcolor(200, 200, 250);

  lock();
  uint nofAllPasses = m_echo->allpassParams().size();
  int maxRec = m_echo->maximumAllPassRecursion();
  const float allpassMaxTime = m_echo->allpassMaxTime();
  unlock();

  if(!maxRec) maxRec = 1;

  const float allPassTop = (preWireBottom + allPassHeightp2) * 0.5;
  const float allPassBottom = (allPassTop - allPassHeightp2) / (float) maxRec;
  const float allPassRecursionAmplitude = allPassTop - allPassBottom;

  /* printf("allPassRecursionAmplitude = %f\n", allPassRecursionAmplitude);
     fflush(0); */
  MJ_TriSample ts;

  for(i=0; i < nofAllPasses; i++) {
    lock();
    int rec = m_echo->allPassRecursion(i);
    ts = m_echo->allpassParams().get(i);
    unlock();


    const float recRatio = (float) rec / (float) maxRec;
    const float leftRatio = ts.m_v[0] / allpassMaxTime;
    const float rightRatio = ts.m_v[1] / allpassMaxTime;

    float left  = allPassLeft + leftRatio * allPassWidth;
    float right = allPassLeft + rightRatio * allPassWidth;
    float top   = 
      allPassHeightp2 + allPassBottom + recRatio * allPassRecursionAmplitude;

    // qDebug("recursion: i = %d rec = %d, top = %f", i, rec, top);
    allPassGC.pen() = QPen(apcolor.dark(100 +(int)( ts.m_v[2] * 200.0)), 
			   (int) (ts.m_v[2] * 3.0));
    
    points.setPoint(0, left, allPassHeightp2);
    points.setPoint(1, left, top);
    points.setPoint(2, right, top);
    points.setPoint(3, right, allPassHeightp2);
    addPolyline(&points, &allPassGC);
  }

  // Outputs:

  const float outputWireY = (outputTop + outputBottom) * 0.5;
  
  lock();
  MJ_Array<MJ_DualSample> *outputParams = &m_echo->outputParams(0);
  uint nofOutputParams = outputParams->size();
  unlock();
  
  points.resize(3);

  for(i=0; i < nofOutputParams; i++) {
    for(j=0; j < 2; j++) {
      lock();
      MJ_DualSample out = outputParams[j].get(i);
      unlock();
      
      const float outTimeRatio = out.m_v[0] / allpassMaxTime;
      
      const float left = allPassLeft + outTimeRatio * allPassWidth;
      const float dent = left * 0.8 + outputLeft * 0.2;

      points.setPoint(0, left, allPassHeightp2);
      points.setPoint(1, dent, outputWireY);
      points.setPoint(2, outputLeft, outputWireY);

      allPassGC.pen() = QPen(apcolor);

      if(j) points.mirrorY();
      addPolyline(&points, &allPassGC);
    }
  }

  if(m_debug) trace(fname, "EXIT");
}


MJ_GraphicPolyLine *MJ_ModuleOverecho_QG::addPolyline(MJ_PointArray2D *points, 
						      MJ_GraphicsContext *gc)
{
  MJ_GraphicPolyLine *pl = new MJ_GraphicPolyLine(points->qarray().copy());

  if(gc) pl->setGraphicsContext(gc->clone());
  
  m_graphics.appendGraphic(pl);

  return pl;
}

MJ_GraphicPolygon *MJ_ModuleOverecho_QG::addPolygon(MJ_PointArray2D *points, 
						      MJ_GraphicsContext *gc)
{
  MJ_GraphicPolygon *pl = new MJ_GraphicPolygon(points->qarray().copy());

  if(gc) pl->setGraphicsContext(gc->clone());
  
  m_graphics.appendGraphic(pl);

  return pl;
}

void MJ_ModuleOverecho_QG::rebuildTw()
{
  const char * fname = "MJ_ModuleOverecho_QG::rebuildTw";

  if(m_debug) trace(fname, this);

  QVBoxLayout *mainBox = dynamic_cast<QVBoxLayout *> (child("mainlayout"));

  m_children.clear();

  for(int i=0; i < 5; i++) {
    if(m_twChildren[i]) delete m_twChildren[i];
    m_twChildren[i] = 0;
  }

  if(m_tw) delete m_tw;

  QTabWidget *tw = new QTabWidget(this);
  int ind = 0;

  m_twChildren[ind] = predelaySliders();
  tw->insertTab(m_twChildren[ind++], MJ_Tr::tr("Pre-delays"));

  m_twChildren[ind] = allpassSliders();
  tw->insertTab(m_twChildren[ind++], MJ_Tr::tr("All-passes"));
  
  m_twChildren[ind] = feedbackParams();
  tw->insertTab(m_twChildren[ind++], MJ_Tr::tr("AP Feedback"));

  m_twChildren[ind] = outputSliders();
  tw->insertTab(m_twChildren[ind++], MJ_Tr::tr("Outputs"));
  
  m_twChildren[ind] = miscellanous();
  tw->insertTab(m_twChildren[ind++], MJ_Tr::tr("Miscellanous"));

  mainBox->addWidget(tw, 20);

  m_tw = tw;

  if(m_debug) trace(fname, "EXIT");
}


QWidget *MJ_ModuleOverecho_QG::predelaySliders()
{  
  const char * fname = "MJ_ModuleOverecho_QG::predelaySliders";

  lock();
  uint predelays = m_echo->predelayParams(0).size();
  unlock();

  MJ_QuadSample qs1, qs2;

  const uint params = 4;

  QWidget * w = new QWidget(this, "predelay-sliders");
  QGridLayout *l = new QGridLayout(w, predelays * 2 + 2, 
				   params + 1, 5, 4);
  MJ_equalStretch(l, 20, 20);
  l->setColStretch(0, 1);
  l->setColStretch(1, 60);
  l->setRowStretch(predelays * 2 + 1, 90);

  l->addWidget(new QLabel(MJ_Tr::tr("Predelay time"), w), 0, 1);
  l->addWidget(new QLabel(MJ_Tr::tr(MJ_Strs::Gain), w), 0, 2);
  l->addWidget(new QLabel(MJ_Tr::tr(MJ_Strs::Pan), w), 0, 3);
  l->addWidget(new QLabel(MJ_Tr::tr(MJ_Strs::Frequency), w), 0, 4);

  for(uint i=0; i < predelays; i++) {
    
    lock();
    qs1 = m_echo->predelayParams(0).get(i);
    qs2 = m_echo->predelayParams(1).get(i);
    unlock();

    buildPredelay(l, w, & qs1, i + 1, i, 0);
    buildPredelay(l, w, & qs2, i + predelays + 1, i, 1);
  }    

  // The final row = scale time:

  int indParam = m_parameters.findParameter
    (MJ_OverechoIds[MJ_OEP_PREDEL_BUF_SCALE_TIME]);

  MJ_ParameterDouble *pd = 
    (MJ_ParameterDouble *) (m_parameters.at(indParam));
  MJ_cpf(pd, fname);
  
  QWidget *w2 = new QWidget(w);
  QGridLayout *newGrid = new QGridLayout(w2, 1, 3, 4, 5);
  int finalRow =  predelays * 2 + 1;
  l->addMultiCellWidget(w2, finalRow, finalRow, 0, params + 1);

  MJ_GuiDoubleSlider *ds = new MJ_GuiDoubleSlider;
  ds->setId(pd->id());
  ds->createObjs(w2);
  ds->label()->setText(pd->name());
  ds->set(pd->value(), pd->min(), pd->max());
  ds->toGridHorizontal(newGrid, 0, 0);
  ds->slider()->setMapper(new MJ_MapperLog);
  
  connect(ds, SIGNAL(emitValue2(const char *,double)),
	  this, SLOT(handleParam(const char *,double)));
  connect(ds, SIGNAL(emitGetValue2(const char *,double *)),
	  this, SLOT(getValue(const char *,double *)));
  connect(ds, SIGNAL(emitDone2(const char *)),
	  this, SLOT(doneValue(const char *)));

  m_children.append(ds);
  
  return w;
}

void MJ_ModuleOverecho_QG::buildPredelay(QGridLayout *l,
					 QWidget *w,
					 MJ_QuadSample *qs, 
					 int row, 
					 int predelay, 
					 int chan)
{
  QCString str;

  int used = 0;
  
  {
    QString s;
    s.setNum(predelay + 1);
    l->addWidget(new QLabel(s, w), row, used++);
  }
  // Predelay time:
  MJ_Slider *s = new MJ_Slider(0, MJ_ModuleOverecho::predelayMaxTime(),
			       qs->m_v[0], w);
  l->addWidget(s, row, used++);

  str.sprintf("%s_%d_%d", MJ_OverechoIds[MJ_OEP_PREDELAY_TIME], 
	      chan, predelay);
  sliderConnect(s, str);
  
  // Predelay gain:
  
  s = new MJ_Slider(0, 1.0, qs->m_v[2], w);
  l->addWidget(s, row, used++);

  str.sprintf("%s_%d_%d", MJ_OverechoIds[MJ_OEP_PREDELAY_GAIN], 
	      chan, predelay);
  sliderConnect(s, str);

  // Predelay pan:
  
  s = new MJ_Slider(0, 1.0, qs->m_v[3], w);
  l->addWidget(s, row, used++);
  str.sprintf("%s_%d_%d", MJ_OverechoIds[MJ_OEP_PREDELAY_PAN], 
	      chan, predelay);  
  sliderConnect(s, str);

  // Predelay filter cut-off frequency:

  s = new MJ_Slider(50.0, 20000, qs->m_v[1], w);
  s->setMapper(new MJ_MapperLog);
  l->addWidget(s, row, used++);
  str.sprintf("%s_%d_%d", MJ_OverechoIds[MJ_OEP_PREDELAY_CUT_OFF], 
	      chan, predelay);  
  sliderConnect(s, str);
}

QWidget *MJ_ModuleOverecho_QG::allpassSliders()
{
  lock();
  uint nofAllPasses = m_echo->allpassParams().size();
  float maxTime = m_echo->allpassMaxTime();
  unlock();

  QWidget * w = new QWidget(this, "allpass-sliders");
  QGridLayout *l = new QGridLayout(w, nofAllPasses * 2 + 2, 
				   3 , 5, 4);
  MJ_equalStretch(l, 20, 20);
  l->setColStretch(0, 1);
  l->setColStretch(1, 60);
  l->setRowStretch(nofAllPasses * 2 + 1, 90);

  l->addWidget(new QLabel(MJ_Tr::tr(MJ_Strs::Time), w), 0, 1);
  l->addWidget(new QLabel(MJ_Tr::tr(MJ_Strs::Gain), w), 0, 2);

  MJ_TriSample ts;

  for(uint i=0; i < nofAllPasses; i++) {
    lock();
    ts = m_echo->allpassParams().get(i);
    unlock();

    buildAllpass(l, w, &ts, i * 2 + 1, i, maxTime);
  }
  return w;
}

void MJ_ModuleOverecho_QG::buildAllpass(QGridLayout *l,
					QWidget *w,
					MJ_TriSample *ts, 
					int row, 
					int index,
					float maxTime)
{
  QCString str;
  
  {
    QString s1(MJ_Tr::tr("1:st time")), s2;
    s2.setNum(index + 1);
    s1 += " " + s2;
    l->addWidget(new QLabel(s1, w), row, 0);
    
    s1 = MJ_Tr::tr("2:dn time");
    s2.setNum(index + 1);
    s1 += " " + s2;
    l->addWidget(new QLabel(s1, w), row + 1, 0);
    
  }
  
  MJ_Slider *s = new MJ_Slider(0, maxTime, ts->m_v[0], w);
  l->addWidget(s, row, 1);
  str.sprintf("%s_%d", MJ_OverechoIds[MJ_OEP_ALL_PASS_TIME_START], index);
  sliderConnect(s, str);
  
  s = new MJ_Slider(0, maxTime, ts->m_v[1], w);
  l->addWidget(s, row + 1, 1);
  str.sprintf("%s_%d", MJ_OverechoIds[MJ_OEP_ALL_PASS_TIME_END], index);
  sliderConnect(s, str);

  s = new MJ_Slider(0, 0.95, ts->m_v[2], w);
  l->addMultiCellWidget(s, row, row + 1, 2, 2);
  str.sprintf("%s_%d", MJ_OverechoIds[MJ_OEP_ALL_PASS_GAIN], index);
  sliderConnect(s, str);
}

QWidget *MJ_ModuleOverecho_QG::outputSliders()
{
  lock();
  uint nofOutputs = m_echo->outputParams(0).size();
  float maxTime = m_echo->allpassMaxTime();
  unlock();

  QWidget * w = new QWidget(this, "allpass-sliders");
  QGridLayout *l = new QGridLayout(w, nofOutputs * 2 + 2, 
				   3 , 5, 4);
  
  MJ_equalStretch(l, 20, 20);
  l->setColStretch(0, 1);
  l->setColStretch(1, 60);
  l->setRowStretch(nofOutputs * 2 + 1, 60);

  l->addWidget(new QLabel(MJ_Tr::tr(MJ_Strs::Time), w), 0, 1);
  l->addWidget(new QLabel(MJ_Tr::tr(MJ_Strs::Gain), w), 0, 2);

  MJ_DualSample ds;

  for(uint j=0; j < 2; j++) {
    for(uint i=0; i < nofOutputs; i++) {
      lock();
      ds = m_echo->outputParams(j).get(i);
      unlock();
      
      buildOutput(l, w, &ds, i + j * nofOutputs + 1, i, j, maxTime);
    }
  }
  return w;
}

void MJ_ModuleOverecho_QG::buildOutput(QGridLayout *l,
				       QWidget *w,
				       MJ_DualSample *ds, 
				       int row, 
				       int index,
				       int channel,
				       float maxTime)
{
  QCString str;
  
  {
    QString s1(MJ_Tr::tr(!channel ? "Left output" : "Right output")), s2;
    s2.setNum(index + 1);
    s2 += ". " + s1;
    l->addWidget(new QLabel(s2, w), row, 0);
  }
  
  MJ_Slider *s = new MJ_Slider(0, maxTime, ds->m_v[0], w);
  l->addWidget(s, row, 1);
  str.sprintf("%s_%d_%d", MJ_OverechoIds[MJ_OEP_OUTPUT_TIME], channel, index);
  sliderConnect(s, str);
  
  s = new MJ_Slider(0, 1.0, ds->m_v[1], w);
  l->addWidget(s, row, 2);
  str.sprintf("%s_%d_%d", MJ_OverechoIds[MJ_OEP_OUTPUT_GAIN], channel, index);
  sliderConnect(s, str);
}

QWidget *MJ_ModuleOverecho_QG::miscellanous()
{
  QWidget *w = new QWidget(this);

  const uint miscParams = 2;

  const char * paramIds[miscParams]= 
    { MJ_OverechoIds[MJ_OEP_WET_DRY], MJ_OverechoIds[MJ_OEP_USE_PRIMES] };
  
  buildGui(w, paramIds, miscParams);

  return w;
}

QWidget *MJ_ModuleOverecho_QG::feedbackParams()
{
  // Feedback parameters:
  QWidget *w = new QWidget(this);
  buildGui(w, & MJ_OverechoIds[MJ_OEP_FEEDBACK_TIME], 6);
  return w;
}

void MJ_ModuleOverecho_QG::sliderConnect(QObject *s, const char * str)
{
  MJ_CStrRemap *remap = new MJ_CStrRemap(str);
  
  connect(s, SIGNAL(valueChangedDouble(double)), 
	  remap, SLOT(callDouble(double)));
  connect(remap, SIGNAL(emitCallDouble(const char *,double)),
	  this, SLOT(handleParam(const char *,double)));
  
  m_children.append(remap);
}

void MJ_ModuleOverecho_QG::buildGui(QWidget *w, 
				    const char **paramIds, 
				    uint nofIds)
{
  const char* fname = "MJ_ModuleOverecho_QG::buildGui";
  
  QGridLayout *l = new QGridLayout(w, nofIds + 1, 3, 5, 4);
  
  MJ_equalStretch(l, 20, 20);
  l->setColStretch(0, 1);
  l->setColStretch(1, 60);
  l->setRowStretch(nofIds, 60);
  
  for(uint i=0; i < nofIds; i++) {
    int indParam = m_parameters.findParameter
      (paramIds[i]);
    
    if(indParam < 0) MJ_fatal(MJ_ERR_INTERNAL, fname);

    MJ_ParameterDouble *pd = 
      dynamic_cast<MJ_ParameterDouble *> (m_parameters.at(indParam));
    MJ_ParameterBool *pb = 
      dynamic_cast<MJ_ParameterBool *> (m_parameters.at(indParam));

    if(pd) {

      MJ_GuiDoubleSlider *ds = new MJ_GuiDoubleSlider;
      ds->setId(pd->id());
      ds->createObjs(w);
      ds->label()->setText(pd->name());
      ds->set(pd->value(), pd->min(), pd->max());
      ds->toGridHorizontal(l, i, 0);
      
      if(pd->mapper())
	ds->slider()->setMapper(pd->mapper()->duplicate());
      
      connect(ds, SIGNAL(emitValue2(const char *,double)),
	      this, SLOT(handleParam(const char *,double)));
      connect(ds, SIGNAL(emitGetValue2(const char *,double *)),
	      this, SLOT(getValue(const char *,double *)));
      connect(ds, SIGNAL(emitDone2(const char *)),
	      this, SLOT(doneValue(const char *)));
      
      m_children.append(ds);
    }
    else if(pb) {
      QCheckBox *cb = new QCheckBox(pb->name(), w, pb->id());
      
      cb->setChecked(pb->value());
      
      MJ_CStrRemap *remap = new MJ_CStrRemap(pb->id());
      m_children.append(remap);
      
      connect(cb, SIGNAL(toggled(bool)), remap, SLOT(callBool(bool)));
      connect(remap, SIGNAL(emitCallBool(const char *,bool)),
	      this, SLOT(setValue(const char *,bool)));
      
      l->addMultiCellWidget(cb, i, i, 0, 2, AlignLeft);
    }
  }
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_ModuleDsp * MJ_createNew_ModuleExample1() 
{
  return new MJ_ModuleOverecho;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_PluginDbeDsp *MJ_createPlugin(int i)
{
  if(i != 0) return 0;

  MJ_PluginDbeDsp *dbe = new MJ_PluginDbeDsp;
  dbe->setAuthors("Ilmonen Tommi S.");
  dbe->setMenuPosition("_Audio_Tests_Overecho UNSTABLE !!");
  dbe->setCreator(MJ_createNew_ModuleExample1);
  dbe->setIdentifier("tommi.ilmonen_tml.hut.fi_overecho1");
  dbe->setParameter(MJ_PP_TRANSLATION_FILE, "mj_overecho");
  dbe->options()->setChans(1,2,2,2);
  dbe->setGuiFactory(MJ_GUI_POSTCARD, MJ_ModuleOverecho_QG::createNew);
  dbe->setPluginDebug(MJ_ModuleOverecho::debugPtr());
  dbe->setParameter(MJ_PP_COMPILE_TIME, DATETIME);

  return dbe;  
}
