/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_pitch1.h>

#include <mj_display_type.h> 
#include <mj_error.h>
#include <mj_parameters.h>
#include <mj_math.h>
#include <mj_strs.h>
#include <mj_tr.h>

#include <qlabel.h> 
#include <qlayout.h> 


MJ_ModulePitchAc::MJ_ModulePitchAc()
  : m_updateInterval(100),
    m_counter(20),
    m_correlationPeriod(100),
    m_threshold(0.0)
{}

MJ_ModulePitchAc::~MJ_ModulePitchAc()
{}

bool MJ_ModulePitchAc::dspPrepare()
{
  int sr = dspSamplingRate();
  
  m_correlationPeriod = sr / 20;

  m_delayBuffer.resize(sr);
  m_delayBuffer.setAll(0.0);

  uint testPitches = 12 * 4;

  m_delays.resize(testPitches);

  double lowFreq = 440.0 / 8;
  double mul = MJ_centToMul(100.0);
  double freq = lowFreq;
  double dsr = (double) sr;

  for(uint i=0; i < testPitches; i++) {
    double ddelay = dsr / freq;
    uint delay = uint (ddelay + 0.5);
    
    m_delays.get(i) = delay;
    
    freq *= mul;
  }

  m_correlationPeriod = m_delays.get(0) * 3;

  m_histogramBuffer.resize(m_delays.size());
  m_histogramBuffer.setAll(0.0);
  
  m_counter = m_updateInterval;
      
  return true;
}

void MJ_ModulePitchAc::dspProcessPeriod()
{
  const MJ_Vector<MJ_Sample> *in = & m_signalInput[0].samples();
  
  uint samples = in->size();

  for(uint i = 0; i < samples; i++) {

    m_delayBuffer.advance();
    m_delayBuffer.getNewest() = in->getFastConst(i);

    if(! --m_counter) {
      m_counter = m_updateInterval;
      estimatePitch();
    }
  }
}

bool MJ_ModulePitchAc::dspStop()
{
  // Free memory.
  m_delayBuffer.resize(0);
  m_histogramBuffer.resize(0);

  return true;
}

bool MJ_ModulePitchAc::msgProcess(const char *, MJ_Message *)
{
  return true;
}

MJ_Parameter * MJ_ModulePitchAc::msgGetInput(const char *)
{
  return 0;
}

QCString MJ_ModulePitchAc::msgGetInputName(int)
{
  return "";
}

void MJ_ModulePitchAc::estimatePitch()
{
  for(uint i=0; i < m_delays.size(); i++) {
    uint delay = m_delays.get(i);

    uint skip = delay / 8 + 1;
    // uint skip = 1;

    float scale = 1.0; // + (delay * 0.15);

    m_histogramBuffer.get(i) = m_delayBuffer.autoCorrelation2
      (delay, m_correlationPeriod, skip) * scale * (float) (skip);
  }

  uint location;
  float peak = m_histogramBuffer.max2(&location);

  if(peak < m_threshold) return;

  uint delay = m_delays.get(location);

  float sr = dspSamplingRate();

  float basicFreq = sr / (float) delay;

  m_freq = basicFreq;

  /*   trace2("MJ_ModulePitchAc::estimatePitch # delay = %d, m_freq = %f",
       delay, m_freq); */
}


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

enum MJ_Pitch1Param {
  MJ_P1P_FREQ_HIGH,
  MJ_P1P_FREQ_LOW,
  MJ_P1P_THRESHOLD,
  MJ_P1P_RELATIVE_LIMIT,
  MJ_P1P_SIZEOF
};

static const char * MJ_Pitch1ParamStrs [MJ_P1P_SIZEOF] = {
  "freq-high",
  "freq-low",
  MJ_Strs::threshold,
  "relative"
};

bool MJ_ModulePitch1::m_debug = false;

MJ_ModulePitch1::MJ_ModulePitch1()
  : m_cutOffHighHz(500.0),
    m_cutOffLowHz(40.0),
    m_threshold(.01),
    m_relativeLimit(0.6),
    m_holdSamples(1000)
{
  /* Make one input channel by default. */
  m_signalSourceDefs.resize(1);

  m_timeBuffer.resize(KEEP_TIMES);
}

/* The descturctor does not really need to do anything. All member
   variables destroy themselves. */
MJ_ModulePitch1::~MJ_ModulePitch1()
{}

bool MJ_ModulePitch1::dspPrepare()
{
  m_filter.resetMemory();
  updateFilterCoeffs();

  /* Check that the number of input and output channels equals 1. */
  if(m_signalInput.size() != 1) 
    return false;

  m_timeBuffer.setAll(0);
  m_times = 0;
  m_count = 0;
  m_freq = 0.0;
  m_sr = dspSamplingRate();

  m_level.reset();

  return true;
}

void MJ_ModulePitch1::dspProcessPeriod()
{
  const MJ_Vector<MJ_Sample> *in = & m_signalInput[0].samples();

  MJ_Sample oldS = m_filter.getSample();

  uint samples = in->size();

  for(uint i = 0; i < samples; i++) {
    MJ_Sample newS = m_filter.putGetSample(in->getConst(i));

    if(newS > 0.0 && oldS <= 0.0 && m_armed) {
      
      MJ_Sample diff = newS - oldS;
      float fraction = newS / diff;

      getFreq(m_count + i, fraction);
      m_armed = false;
    }

    m_level.put(newS, m_holdSamples);

    MJ_Sample threshold2 = m_relativeLimit * m_level.peak();
    
    if(newS > m_threshold && newS > threshold2) 
      m_armed = true;

    oldS = newS;
  }
  m_count += samples;

  if((m_timeBuffer.getNewest() + m_sr * 2) < m_count) {
    m_times = 0;
    m_freq  = 0.0;
  }

}

bool MJ_ModulePitch1::msgProcess(const char * id, MJ_Message *m)
{
  const char *fname = "MJ_ModulePitch1::msgProcess";
  bool ok = true;

  double tmp = m->toDouble(&ok);
    
  if(m_debug) {
    trace2("%s # %s = %f", fname, id, tmp);
  }
  
  if(!ok) {
    if(m_debug) 
      trace(fname, "Failure 1");
  }
  else 
    if(m_debug) trace(fname, "Success 1");

  MJ_Pitch1Param p = (MJ_Pitch1Param) 
    MJ_findStr(id, MJ_Pitch1ParamStrs, MJ_P1P_SIZEOF);
  
  if(p == MJ_P1P_FREQ_HIGH) {
    m_cutOffHighHz = tmp;
    updateFilterCoeffs();
  }
  else if(p == MJ_P1P_FREQ_LOW) {
    m_cutOffLowHz = tmp;
    updateFilterCoeffs();
  }
  else if(p == MJ_P1P_THRESHOLD) {
    m_threshold = tmp;
  }
  else if(p == MJ_P1P_RELATIVE_LIMIT) {
    m_relativeLimit = tmp;
  }
  else ok = false;

  if(m_debug) trace(fname, "EXIT");

  return ok;
}

MJ_Parameter * MJ_ModulePitch1::msgGetInput(const char * id)
{
  if(m_debug) trace("MJ_ModulePitch1::msgGetInput", id);
  
  MJ_Pitch1Param p = (MJ_Pitch1Param) 
    MJ_findStr(id, MJ_Pitch1ParamStrs, MJ_P1P_SIZEOF);
  

  if(p == MJ_P1P_FREQ_HIGH) {
    return new MJ_ParameterDouble("Cut-off frequency (high)", id,
				  m_cutOffHighHz, 0.40, 5000.0, 
				  MJ_UNIT_HERTZ);
  }

  if (p == MJ_P1P_FREQ_LOW) {
    return new MJ_ParameterDouble("Cut-off frequency (low)", id,
				  m_cutOffLowHz, 0.40, 1000.0, 
				  MJ_UNIT_HERTZ);
  }
  
  if(p == MJ_P1P_THRESHOLD) {
    return
      new MJ_ParameterDouble(MJ_Strs::Threshold, MJ_Strs::threshold,
			     m_threshold , 0.0, 1.0);
  }

  if(p == MJ_P1P_RELATIVE_LIMIT) {
    return new MJ_ParameterDouble("Relative limit", id, 
				  m_relativeLimit, 0.0, 0.9);
  }

  return 0;
}

QCString MJ_ModulePitch1::msgGetInputName(int i)
{
  QCString s;

  if(i < MJ_P1P_SIZEOF) 
    s = MJ_Pitch1ParamStrs[i];
      
  return s;
}

void MJ_ModulePitch1::updateFilterCoeffs()
{
  double sr = (double) dspSamplingRate();

  m_filter.designLowPass2(m_cutOffHighHz, sr, 1.0);
  m_holdSamples = (uint) (sr / m_cutOffLowHz);
}


void MJ_ModulePitch1::getFreq(ulong sampleInd, float fractionalSamples)
{
  sampleInd = (sampleInd << 8) + (ulong) (fractionalSamples * 256.0);

  m_timeBuffer.advance();

  m_timeBuffer.getNewest() = sampleInd;

  m_times++;

  if(m_times < 3) return;

  // Now calculate the pitch based on the cross-over times:
  
  uint crossOvers = 1;
  const uint maxCrossOvers = MJ_min((ulong) KEEP_TIMES, m_times) - 2;
  const ulong maxTime = (ulong) (m_sr / 3) << 8;
  
  ulong nextTime = sampleInd;

  ulong prevTime = nextTime;

  for(; crossOvers < maxCrossOvers && (sampleInd - prevTime) < maxTime;
      crossOvers++) {
    prevTime = m_timeBuffer.getNewest(crossOvers);
  }
  
  if(sampleInd == prevTime) {
    m_freq = 0.0;
    return;
  }

  /* char buf[512];
     sprintf(buf, "%d %d", (int) crossOvers, (int) (sampleInd - prevTime));
     MJ_trace("MJ_ModulePitch1::getFreq", buf); */

  const float timeS =  (float) (sampleInd - prevTime) / (float) m_sr / 256.0;
  m_freq = (float) (crossOvers - 1) / timeS;

  newFreq(sampleInd, m_freq);
}

void MJ_ModulePitch1::newFreq(ulong, float)
{}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_ModulePitch1_QG::MJ_ModulePitch1_QG(MJ_RtControl *rtparent,
				       MJ_ModuleDsp *module,
				       QWidget      *parent,
				       const char   *name)
  : MJ_ModuleDspGuiPostCard(rtparent, module, true,
			    MJ_P1P_SIZEOF,
			    parent == 0  ? MJ_GUI_OWN_WINDOW : MJ_GUI_POSTCARD,
			    parent, name),
  m_display(new QLabel(childParent()))
{
  const char * fname = "MJ_ModulePitch1_QG::MJ_ModulePitch1_QG";

  if(!dynamic_cast<MJ_ModulePitchBase *> (module)) 
    MJ_fatal(MJ_ERR_ARGUMENT, fname);

  QLabel *l = new QLabel(MJ_Tr::tr(MJ_Strs::Frequency), childParent());

  MJ_cpf(l, m_layout, fname);

  QFont qf = m_display->font();
  qf.setPointSize((int) ((float) qf.pointSize() * 1.5));
  qf.setBold(true);
  m_display->setFont(qf);

  int rows = m_layout->numRows();
  m_layout->expand(rows + 1, m_layout->numCols());

  m_layout->addWidget(l, MJ_P1P_SIZEOF, 0);
  m_layout->addWidget(m_display, MJ_P1P_SIZEOF, 1);
  l->show();
  m_display->show();
  
  connect(&m_timer, SIGNAL(timeout()), this, SLOT(update()));
  if(isRealTime()) {
    realTimeChange(true);
  }
  update();
}
  
void MJ_ModulePitch1_QG::update()
{
  lock();
  float freq = (dynamic_cast<MJ_ModulePitchBase *> (m_module))->freq();
  unlock();
  
  QString str;

  str.sprintf("%3.3f Hz",freq);

  if(str != m_display->text())
    m_display->setText(str);
}

MJ_ModuleDspGui *MJ_ModulePitch1_QG::createNew(MJ_RtControl *rtparent,
						    MJ_ModuleDsp *module,
						    QWidget *parent, 
						    const char *name)
{
  return new MJ_ModulePitch1_QG(rtparent, module, parent, name);
}

void MJ_ModulePitch1_QG::realTimeChange (bool nowRt)
{
  if(nowRt) m_timer.start(MJ_DisplayType::localDisplay() ? 20 : 200);
  else m_timer.stop();
}

