/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#ifndef MJ_MODULE_PITCH1_H
#define MJ_MODULE_PITCH1_H

#include <mj_filter_iir2.h>
#include <mj_level_meter.h> 
#include <mj_module_dsp.h>
#include <mj_ringbuffer.h>

#include <mj_module_dsp_gui_default.h>
#include <qtimer.h>  

/// Base class for pitch detectors.
class MJ_ModulePitchBase 
{
public:
  MJ_ModulePitchBase() : m_freq(0.0) {}
  virtual ~MJ_ModulePitchBase() {}
  
  inline float freq() const { return m_freq; }

protected:
  float       m_freq;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** Pitch detector that is based on autocorrelation and zero crossings. */
class MJ_ModulePitchAc : public MJ_ModuleDsp, public MJ_ModulePitchBase
{
public:
  MJ_ModulePitchAc();
  virtual ~MJ_ModulePitchAc();

  virtual bool dspPrepare();
  virtual void dspProcessPeriod();
  virtual bool dspStop();

  virtual bool msgProcess(const char *, MJ_Message *);
  virtual MJ_Parameter * msgGetInput(const char *);
  virtual QCString msgGetInputName(int);

protected:

  void estimatePitch();

  MJ_RingBuffer2<float> m_delayBuffer;
  MJ_Array<uint>        m_delays;
  MJ_Vector<float>      m_histogramBuffer;

  uint m_updateInterval;
  uint m_counter;
  uint m_correlationPeriod;

  float m_threshold;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


/** A simple pitch detector. 

    This class uses the zero-crossing times as the pitch estimation
    starting point.
*/
class MJ_ModulePitch1 : public MJ_ModuleDsp, public MJ_ModulePitchBase
{
public:

  static const int KEEP_TIMES = 256;

  MJ_ModulePitch1();

  virtual ~MJ_ModulePitch1();
  
  virtual bool dspPrepare();
  virtual void dspProcessPeriod();

  virtual bool msgProcess(const char *, MJ_Message *);
  virtual MJ_Parameter * msgGetInput(const char *);
  virtual QCString msgGetInputName(int);

protected:
  
  void updateFilterCoeffs();

  void getFreq(ulong sampleInd, float fractionalSamples);

  virtual void newFreq(ulong sampleInd, float freq);

  /** The past cross-over times in a strange format. The numbers are
      fixed-point floats with first 24 integer bits and then 8
      fractional bits. */

  MJ_RingBuffer2<ulong> m_timeBuffer;
  ulong                 m_times;
  ulong                 m_count;

  bool m_armed;

  double      m_cutOffHighHz;
  double      m_cutOffLowHz;
  MJ_Sample   m_threshold;
  MJ_Sample   m_relativeLimit;
  int         m_sr;
  uint        m_holdSamples;

  MJ_FilterIIR2<MJ_Sample> m_filter;
  MJ_LevelMeter<MJ_Sample> m_level;
  
  static bool m_debug;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

class QLabel;

/// GUI for MJ_ModulePitch1
class MJ_ModulePitch1_QG : public MJ_ModuleDspGuiPostCard
{
  Q_OBJECT
public:
  MJ_ModulePitch1_QG(MJ_RtControl *rtparent,
		     MJ_ModuleDsp *module,
		     QWidget      *parent = 0,
		     const char   *name = 0);
  
  static MJ_ModuleDspGui *createNew(MJ_RtControl *rtparent,
				    MJ_ModuleDsp *module,
				    QWidget *parent, 
				    const char *name);

public slots:
  virtual void update();

protected:

  virtual void realTimeChange (bool nowRt);

  QLabel *m_display;
  QTimer  m_timer;
};

#endif
