/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_reverb.h>

#include <mj_error.h> 
#include <mj_random.h> 
#include <mj_mapper.h>
#include <mj_math.h>
#include <mj_parameters.h> 
#include <mj_strs.h>


const char * MJ_ParamsReverb1Keys[MJ_PARREV_SIZEOF] = {
  "predelay-time",
  "predelay-spread-time",
  "predelay-spread",
  "predelay-color",
  "tail-time",
  "tail-spread",
  "tail-color",
  "tail-feedback",
  "dry-wet"
};

const char * MJ_ParamsReverb1Labels[MJ_PARREV_SIZEOF] = {
  "Predelay time",
  "Predelay spread time",
  "Predelay stereo spread",
  "Predelay color (dark/bright)",
  "Tail time",
  "Tail stereo spread",
  "Tail color (dark/bright)",
  "Tail feedback",
  "Dry/wet -ratio"
};

MJ_ModuleReverb1::MJ_ModuleReverb1()
  : m_predelayMs(50.0),
    m_predelaySpreadTimeMs(50.0),
    m_predelaySpread(0.5),
    m_predelayColor(0.5),
    m_tailSpreadTimeMs(250.0),
    m_tailSpread(0.5),
    m_tailColor(0.5),
    m_tailFeedback(0.5),
    m_wetDryRatio(0.5)
{
}

MJ_ModuleReverb1::~MJ_ModuleReverb1()
{}
  
bool MJ_ModuleReverb1::dspPrepare()
{
  m_predelayBuffer.resize(dspSamplingRate());

  allocateTails(4);
  calculateTails();
  
  m_feedbackFilterHighPass.designHighPass(0.01);
  m_feedbackFilterHighPass.resetMemory();

  return true;
}
 
void MJ_ModuleReverb1::dspProcessPeriod()
{
  uint j;
  uint chans = m_signalInput.size();
  uint samples = m_signalInput[0].samples().size();

  float useFeedBack = m_tailFeedback; //  / (float) m_tailNodes.size();

  for(uint i=0; i < samples; i++) {

    m_predelayBuffer.advance();

    // Sum all inputs to a shared global predelay buffer
    float *currentPtr = & m_predelayBuffer.getNewest();

    (*currentPtr) = 0.0;

    for(j=0; j < chans; j++)
      (*currentPtr) += m_signalInput.getFast(j).samples().getFast(i);
    
    // float predelayed = m_predelayBuffer.getNewest(128);
    // m_feedbackFilterHighPass.putGet(m_feedbackFilter.get()) * useFeedBack;
    
    float sum = 0.0;

    uint lastIndex = m_tailNodes.size() - 1;

    float firstFeedback = m_tailNodes[0].get();

    for(j=0; j < lastIndex; j++) {
      uint sourceNode = (j + 1);
      sum += m_tailNodes[j].putGet(&m_predelayBuffer, 
				   m_tailNodes[sourceNode].get(),
				   useFeedBack);
    }
    
    sum += m_tailNodes[lastIndex].putGet(&m_predelayBuffer, 
					 firstFeedback,
					 useFeedBack);

    // m_feedbackFilter.put(sum);

    for(j=0; j < m_signalOutput.size(); j++) {
      m_signalOutput.getFast(j).samples().getFast(i) = 
	m_tailSpread * m_tailNodes[j].get() + (1.0 - m_tailSpread) * sum;
    }
  }
}

bool MJ_ModuleReverb1::msgProcess(const char *id, MJ_Message *m)
{
  /* trace3(MJ_reverb1Debug, "MJ_ModuleReverb1::msgProcess # %s %p",
     id, m); */

  MJ_ParamsReverb1 pc = (MJ_ParamsReverb1)
    MJ_findStr(id, MJ_ParamsReverb1Keys, MJ_PARREV_SIZEOF);
  
  if((int) pc == -1) {
    error(MJ_ERR_ARGUMENT, "MJ_ModuleReverb1::msgProcess", id);
    return false;
  }

  bool ok(true);

  float val = m->toDouble(&ok);

  if(!ok) return false;

  // Dangerous:

  switch(pc) {
  case MJ_PARREV_PREDELAY_TIME:
    m_predelayMs = val;
    calculateTails();
    break;
  case MJ_PARREV_PREDELAY_SPREAD_TIME:
    m_predelaySpreadTimeMs = val;
    calculateTails();
    break;
  case MJ_PARREV_PREDELAY_SPREAD:
    m_predelaySpread = MJ_clampF(val, 0.0, 1.0);
    break;
  case MJ_PARREV_PREDELAY_COLOR:
    m_predelayColor = MJ_clampF(val, 0.05, 1.0);
    break;

  case MJ_PARREV_TAIL_TIME:
    m_tailSpreadTimeMs = val;
    calculateTails();
    break;
  case MJ_PARREV_TAIL_SPREAD:
    m_tailSpread = val;
    break;
  case MJ_PARREV_TAIL_COLOR:
    m_tailColor = MJ_clampF(val, 0.05, 1.0);
    calculateTails();
    break;
  case MJ_PARREV_TAIL_FEEDBACK:
    m_tailFeedback = MJ_clampF(val, 0.0, 1.0);
    break;

  case MJ_PARREV_DRY_WET:
    m_wetDryRatio = MJ_clampF(val, 0.0, 1.0);
    break;
  }
  
  return true;
}

MJ_Parameter * MJ_ModuleReverb1::msgGetInput(const char *id)
{
  MJ_ParamsReverb1 pr = (MJ_ParamsReverb1)
    MJ_findStr(id, MJ_ParamsReverb1Keys, MJ_PARREV_SIZEOF);

  MJ_ParameterDouble *pd;

  if((int) pr < 0) return 0;
  switch(pr){
  case MJ_PARREV_PREDELAY_TIME:
    pd = new MJ_ParameterDouble(MJ_ParamsReverb1Labels[pr],
				id, m_predelayMs, 1.0, 500.0);
    pd->setMapper(new MJ_MapperLog);
    return pd;
  case MJ_PARREV_PREDELAY_SPREAD_TIME:
    pd = new MJ_ParameterDouble(MJ_ParamsReverb1Labels[pr],
				id, m_predelaySpreadTimeMs, 1.0, 500.0);
    pd->setMapper(new MJ_MapperLog);
    return pd;
  case MJ_PARREV_PREDELAY_SPREAD:
    return new MJ_ParameterDouble(MJ_ParamsReverb1Labels[pr],
				  id, m_predelaySpread, 0.0, 1.0);
  case MJ_PARREV_PREDELAY_COLOR:
    return new MJ_ParameterDouble(MJ_ParamsReverb1Labels[pr],
				  id, m_predelayColor, 0.05, 1.0);

  case MJ_PARREV_TAIL_TIME:
    pd = new MJ_ParameterDouble(MJ_ParamsReverb1Labels[pr],
				  id, m_tailSpreadTimeMs, 10.0, 1000.0);
    pd->setMapper(new MJ_MapperLog);
    return pd;
  case MJ_PARREV_TAIL_SPREAD:
    return new MJ_ParameterDouble(MJ_ParamsReverb1Labels[pr],
				  id, m_tailSpread, 0.0, 1.0);
  case MJ_PARREV_TAIL_COLOR:
    return new MJ_ParameterDouble(MJ_ParamsReverb1Labels[pr],
				  id, m_tailColor, 0.05, 1.0);
  case MJ_PARREV_TAIL_FEEDBACK:
    return new MJ_ParameterDouble(MJ_ParamsReverb1Labels[pr],
				  id, m_tailFeedback, 0.0, 0.98);

  case MJ_PARREV_DRY_WET:
    return new MJ_ParameterDouble(MJ_ParamsReverb1Labels[pr],
				  id, m_wetDryRatio, 0.0, 1.0);
  }

  return 0;
}

QCString MJ_ModuleReverb1::msgGetInputName(int i)
{
  QCString s;
  if(i < MJ_PARREV_SIZEOF) s = MJ_ParamsReverb1Keys[i];
  return s;
}

void MJ_ModuleReverb1::allocateTails(uint n)
{
  m_tailNodes.resize(n);
  
  uint maxDelay = 
    20 + (uint) (/*m_tailSpreadTimeMs * 0.001 */ (float) dspSamplingRate());
  
  for(uint i=0; i < n; i++) {
    m_tailNodes[i].m_delay.resize(maxDelay);
    m_tailNodes[i].m_delay.setAll(0.0);
  }
}

void MJ_ModuleReverb1::calculateTails()
{
  float sr = dspSamplingRate();
  
  MJ_Random r(123456);

  float tailSpreadSamples = m_tailSpreadTimeMs * (float) sr * 0.001 * 0.5;

  for(uint i=0; i < m_tailNodes.size(); i++) {
    MJ_ReverbTailNode &node = m_tailNodes[i];

    uint predelayTime = (uint) 
      ((m_predelayMs + r.getFloat0() * m_predelaySpreadTimeMs) * 
       (float) sr * 0.001);

    uint t1 = MJ_previousPrime((uint) (tailSpreadSamples + 
				       r.getFloat0() * tailSpreadSamples));
    uint t2 = MJ_nextPrime
      ((uint) sqrt((double) t1) + (uint) (r.getFloat0() * 50));
    uint t3 = (uint) (r.getFloat0() * 25.0) + 1;

    // Make sure t3 is not even if i is not even (ad-hoc logic)
    if(!(t3 & 1) && (i & 1)) t3++;

    if(t2 >= t1) t2 = t1 - 47;
    if(t2 <= t3) t2 = t3 + 2;

    node.m_operators[0].set(0.35, 1, 1 + t1);
    node.m_operators[1].set(0.35, 2, 2 + t2);
    node.m_operators[2].set(0.45, 3, 3 + t3);

    /* node.m_operators[0].set(0, 20, 20 + t1);
    node.m_operators[1].set(0, 21, 21 + t2);
    node.m_operators[2].set(0, 22, 22 + t3);*/
    
    uint delayTime = MJ_previousPrime
      ((uint) (r.getFloat0() * tailSpreadSamples));

    while(delayTime < t1 + 3)
      delayTime += 177;
    
    node.m_predelayTime = predelayTime;
    node.m_delayTime = delayTime;

    trace2("MJ_ModuleReverb1::calculateTails %u [%u %u %u] %u", 
	   i, t1, t2, t3, delayTime);

    node.m_feedbackFilter.set(-(1.0 - m_tailColor), m_tailColor);
    node.m_feedbackFilter.resetMemory();
    node.m_val = 0.0;

    if(delayTime >= node.m_delay.size())
      error2(MJ_ERR_INTERNAL, "MJ_ModuleReverb1::calculateTails # %u >= %u",
	     delayTime, node.m_delay.size());

  }
}
