/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_scope1.h>

#include <mj_display_type.h> 
#include <mj_error.h>
#include <mj_graphic_axes.h>
#include <mj_graphic_polyline.h> 
#include <mj_graphics_context_simple.h>
#include <mj_parameters.h>
#include <mj_math.h>
#include <mj_strs.h>

#include <mj_dsp_plugin.h>
#include <mj_version.h>

#include <qlayout.h> 

bool MJ_ModuleScope1::m_debug = false;

MJ_ModuleScope1::MJ_ModuleScope1()
  : m_cutOffHz(1000.0)
{
  /* Make one input channel by default. */
  m_signalSourceDefs.resize(1);
}

MJ_ModuleScope1::~MJ_ModuleScope1()
{}

bool MJ_ModuleScope1::dspPrepare()
{
  /* Check that the number of input channels equals 1. */
  if(m_signalInput.size() < 1) 
    return false;

  updateFilterCoeffs();

  m_news = 0;

  return true;
}

void MJ_ModuleScope1::dspProcessPeriod()
{
  const MJ_Vector<MJ_Sample> *in = & m_signalInput[0].samples();
  
  m_news += m_scope.putSamples(in->dataConst(), in->size());
}

bool MJ_ModuleScope1::msgProcess(const char * id, MJ_Message *m)
{
  const char *fname = "MJ_ModuleScope1::msgProcess";
  bool ok = true;

  double tmp = m->toDouble(&ok);
    
  if(m_debug) {
    char buf[128];
    sprintf(buf, "%s # Value = %f", fname, tmp);
    trace(buf);
  }
  
  if(!ok) {
    if(m_debug) 
      trace(fname, "Failure 1");
    return false;
  }
  else 
    if(m_debug) trace(fname, "Success 1");
  
  if(!strcmp(id, MJ_Strs::frequency)) {
    m_cutOffHz = tmp;
    updateFilterCoeffs();
  }
  else if(!strcmp(id, MJ_Strs::threshold)) {
    m_scope.setThreshold(tmp);
  }
  else ok = false;

  return ok;
}

MJ_Parameter * MJ_ModuleScope1::msgGetInput(const char * id)
{
  if(m_debug) trace("MJ_ModuleScope1::msgGetInput");
  
  if(!strcmp(id, MJ_Strs::frequency)) {
    MJ_ParameterDouble *pd = new MJ_ParameterDouble;
    pd->set(
	    "Cut-off frequency", 
	    MJ_Strs::frequency,
	    m_cutOffHz, 
	    0.40, 
	    5000.0, 
	    MJ_UNIT_HERTZ);
    return pd;
  }
  if(!strcmp(id, MJ_Strs::threshold)) {
    return
      new MJ_ParameterDouble(MJ_Strs::Threshold, MJ_Strs::threshold,
			     m_scope.threshold(), 0.0, 1.0);
  }
  return 0;
}

QCString MJ_ModuleScope1::msgGetInputName(int i)
{
  QCString s;
  
  switch(i) {
  case 0:
    s = MJ_Strs::frequency;
    break;
    
  case 1:
    s = MJ_Strs::threshold;
    break;
  }    
    
  return s;
}

/** Gets the wave data and returns the length of the wave (in
    samples). */

ulong MJ_ModuleScope1::getWave(MJ_Array<float> *a)
{
  float *waveData = a->data();

  ulong waveStart = m_scope.waveStartSamples();
  ulong waveEnd = m_scope.waveEndSamples();
  
  ulong waveSamples = waveEnd - waveStart + 1;
  ulong waveBegin = m_scope.buffer().getSampleCount() - waveEnd;
  
  if(a->size() < waveSamples)
    return 0;

  ulong waveSentinel = waveBegin + waveSamples;

  for(ulong i=waveBegin; i < waveSentinel; i++) {
    *waveData++ = m_scope.buffer().getNewest(i);
  }
  
  return waveSamples;
}

void MJ_ModuleScope1::updateFilterCoeffs()
{
  if(!m_signalInput.size()) return;
  
  MJ_SampleRateDef &sr = m_signalInput[0].sink()->sampleRate();

  double cutOffHz = m_cutOffHz;
  double nyquist = sr.rate() / 2.0;
  if(cutOffHz > nyquist) cutOffHz = nyquist;
  double cutOffVal =  cutOffHz / nyquist * MJ_pi;
  
  m_scope.designFilter(cutOffVal);
  m_scope.buffer().resize((ulong) (sr.rate() / 10.0));
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


MJ_ModuleScope1_QG::MJ_ModuleScope1_QG(MJ_RtControl *rtparent,
				       MJ_ModuleDsp *module,
				       QWidget      *parent,
				       const char   *name)
  : MJ_ModuleDspGuiPostCard(rtparent, module, true,
			    2,
			    parent == 0  ? MJ_GUI_OWN_WINDOW : MJ_GUI_POSTCARD,
			    parent, name),
  m_graphics(this)
  
{  
  m_graphics.setMinimumSize(100, 100);

  int rows = m_layout->numRows();
  m_layout->expand(rows + 1, m_layout->numCols());

  // The axes to the graphics.
  MJ_GraphicAxes *axes = new MJ_GraphicAxes;
  
  MJ_GraphicsContextPen *gcpb = 
    new MJ_GraphicsContextPen(QPen(Qt::gray));
  axes->setGraphicsContext(gcpb);
  m_graphics.appendGraphic(axes);
  
  // The waveform display
  m_waveForm = new MJ_GraphicPolyLine;

  MJ_PointArray2D *points = & m_waveForm->points();

  const uint nofPoints = 300;

  points->resize(nofPoints);

  for(uint i=0; i < nofPoints; i++) {
    points->setPoint(i, (double) i / (double) (nofPoints - 1), 0.0);
  }
  gcpb = new MJ_GraphicsContextPen(QPen(Qt::green));
  m_waveForm->setGraphicsContext(gcpb);
  
  m_layout->addMultiCellWidget(&m_graphics, rows, rows, 0, 2);
  m_layout->setRowStretch(rows, 80);
  m_graphics.show();
  m_graphics.setBoundingBox(0.0, 1.5, 1.0, -1.5);

  m_graphics.appendGraphic(m_waveForm);
  m_graphics.setBackgroundColor(Qt::black);

  connect(&m_timer, SIGNAL(timeout()), this, SLOT(update()));

  if(isRealTime()) {
    realTimeChange(true);
  }
  m_waveData.resize(10000);
}

MJ_ModuleDspGui *MJ_ModuleScope1_QG::createNew(MJ_RtControl *rtparent,
					       MJ_ModuleDsp *module,
					       QWidget *parent, 
					       const char *name)
{
  return new MJ_ModuleScope1_QG(rtparent, module, parent, name);
}

void MJ_ModuleScope1_QG::update()
{
  MJ_ModuleScope1 *scope = (MJ_ModuleScope1 *) m_module;
  
  ulong news = scope->news();

  if(news == m_news) return;

  m_news = news;

  ulong waveSamples = scope->getWave(&m_waveData);

  if(!waveSamples) return;
  
  MJ_PointArray2D *points = & m_waveForm->points();
  
  // Get point positions:
  uint nofPoints = points->size();
  double scale = 1.0 / (nofPoints - 1);
  
  for(uint i=0; i < nofPoints; i++) {
    double index = i * scale * (double) (waveSamples - 1);
    points->setPointY(i, m_waveData.getInterpolated1(index));
  }
  
  m_graphics.redraw(MJ_DisplayType::localDisplay());
}

void MJ_ModuleScope1_QG::realTimeChange (bool nowRt)
{
  if(nowRt) {
    m_timer.start(MJ_DisplayType::localDisplay() ? 50 : 500);
    m_news = 0;
  }
  else m_timer.stop();
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_ModuleDsp * MJ_createNew_ModuleScope1() 
{
  return new MJ_ModuleScope1;
}

MJ_PluginDbeDsp * libmjscope1_CREATE_DSPMODULE()
{
  MJ_PluginDbeDsp *dbe = new MJ_PluginDbeDsp;

  dbe->setAuthors("Ilmonen Tommi S.");
  dbe->setMenuPosition("_Audio_Analysis_Scope");
  dbe->setCreator(MJ_createNew_ModuleScope1);
  dbe->setIdentifier("Tommi.Ilmonen@hut.fi_analysis_scope1");
  dbe->pluginVersion()->set(0,0,1);
  dbe->options()->defaultChans2(1, 0);
  // dbe->setParameter(MJ_PP_TRANSLATION_FILE, <mj_scope1");
  dbe->setGuiFactory(MJ_GUI_POSTCARD, MJ_ModuleScope1_QG::createNew);
  dbe->setParameter(MJ_PP_COMPILE_TIME, DATETIME);
  
  return dbe;  
}

MJ_PluginDbeDsp *MJ_createPlugin(int i)
{
  switch(i) {
  case 0:
    return libmjscope1_CREATE_DSPMODULE();
  }
  return 0;
}
