/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_spectrometer.h>

#include <mj_display_type.h> 
#include <mj_array2dtmpl.C> 
#include <mj_error.h>
#include <mj_level_display.h> 
#include <mj_math.h>
#include <mj_parameters.h>
#include <mj_random.h>
#include <mj_strs.h>

#include <mj_dsp_plugin.h>
#include <mj_version.h>

#include <qlayout.h> 

bool MJ_ModuleSpectrometer::m_debug = false;

MJ_ModuleSpectrometer::MJ_ModuleSpectrometer()
  : m_peakTime(4000), m_peakHoldTime(100000), 
    m_bands(15)
{}

MJ_ModuleSpectrometer::~MJ_ModuleSpectrometer()
{}

bool MJ_ModuleSpectrometer::dspPrepare()
{
  /* Check that the number of input channels exceeds zero. */
  if(m_signalInput.size() < 1) 
    return false;

  updateFilterCoeffs();

  return true;
}

void MJ_ModuleSpectrometer::dspProcessPeriod()
{
  uint bands = m_levels.cols();

  for(uint i=0; i < m_signalInput.size(); i++) {

    MJ_Vector<MJ_Sample> *inv =  & m_signalInput.getFast(i).samples();

    const MJ_Sample *in = inv->data();
    uint samples = inv->size();

    for(uint k=0; k < bands; k++) {

      MJ_FilterAndLevels *tmp = & m_levels.get(i, k);

      for(uint j=0; j < samples; j++) {

	tmp->filter().put(in[j]);
	MJ_Sample tmps = tmp->filter().get();
	
	tmps = tmp->absFilter().putGetSample(tmps * tmps);
	// tmps = tmp->absFilter().putGetSample(MJ_abs(tmps));
	tmp->level().put(tmps, m_peakHoldTime);
      }
    }
  }
}

bool MJ_ModuleSpectrometer::msgProcess(const char * id, MJ_Message *m)
{
  // const char *fname = "MJ_ModuleSpectrometer::msgProcess";
  bool ok = true;
  
  if(!strcmp(id, MJ_Strs::bands)) {
    uint bands = m->toInt(&ok);

    if(ok && bands >= 3 && bands <= 100)
      m_bands = bands;
    else
      return false;
  }

  return ok;
}

MJ_Parameter * MJ_ModuleSpectrometer::msgGetInput(const char * id)
{
  if(m_debug) trace("MJ_ModuleSpectrometer::msgGetInput", id);

  if(!strcmp(id, MJ_Strs::bands)) {
    return new MJ_ParameterInt(MJ_Strs::Bands, id, m_bands, 3, 100);
  }

  return 0;
}

QCString MJ_ModuleSpectrometer::msgGetInputName(int i)
{
  QCString s;
  if(!i) s = MJ_Strs::bands;
  return s;
}

void MJ_ModuleSpectrometer::updateFilterCoeffs()
{
  if(m_debug) trace("MJ_ModuleSpectrometer::updateFilterCoeffs");

  uint chans = dspInChannels();

  m_levels.resize(chans, m_bands);
  
  double sr = dspSamplingRate();
  double lowestLog = log(MJ_min(20.0, sr / 100));
  double highestLog = log(sr / 3.0);
  double nyquist = sr / 2.0;
  double step = (highestLog - lowestLog) / (double) (m_bands - 1);

  if(!chans) return;

  for(uint j=0; j < m_bands; j++) {

    double freq = exp(j * step + lowestLog);
    MJ_FilterAndLevels *levels = & m_levels.get(0, j);
    
    // double bandWidth;
    
    if(j == 0) {
      levels->filter().designLowPass2(freq, sr, 1.0);
      // bandWidth = freq;
    }
    else if(j == (m_bands - 1)) {
      
      // bandWidth = nyquist -  freq;
      levels->filter().designHighPass2(freq, sr, 1.0);
    }
    else {
      double prevFreq = exp((j-1) * step + lowestLog);
      double octaves = MJ_mulToCent(freq / prevFreq) / 600.0;
      // bandWidth = 2.0 * (freq - prevFreq);
      levels->filter().designBandPassB2(freq, sr, octaves);
      // levels->filter().designBandPassB2(freq, sr, 1.0);
    }
    
    MJ_Sample filterGain = whiteNoiseGain(&levels->filter(), 100000);

    if(m_debug) 
      trace2("MJ_ModuleSpectrometer::updateFilterCoeffs filter %d gain = %e",
	     j, filterGain);
    
    levels->filter().mulGain(1.0 / filterGain);
    
    levels->absFilter().designLowPass(MJ_min(1.0, freq * 0.005) / nyquist);
    
    levels->resetMemory();
    
    // Copy the filter to other channels.
    for(uint i=1; i < chans; i++) 
      m_levels.get(i, j) = *levels;
  }
}

MJ_Sample MJ_ModuleSpectrometer::whiteNoiseGain
(MJ_FilterIIR2<MJ_Sample> *filter, 
 uint samples)
{
  MJ_Random rand;
  MJ_FilterAP1<double> lowPass;
  lowPass.designLowPass(20.0 / samples);
  lowPass.resetMemory();
  filter->resetMemory();

  for(uint i=0; i < samples; i++) {
    MJ_Sample tmp = filter->putGetSample((MJ_Sample) rand.getFloat1(2.0));
    lowPass.put(tmp * tmp);
    // lowPass.put(MJ_abs(tmp));
  }
  return MJ_sqrt(lowPass.get());
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


MJ_ModuleSpectrometer_QG::MJ_ModuleSpectrometer_QG(MJ_RtControl *rtparent,
						   MJ_ModuleDsp *module,
						   QWidget      *parent,
						   const char   *name)
  : MJ_ModuleDspGuiPostCard(rtparent, module, true,
			    1,
			    parent == 0  ? MJ_GUI_OWN_WINDOW : MJ_GUI_POSTCARD,
			    parent, name),
  m_table(0),
  m_vertical(true)
{
  connect(&m_timer, SIGNAL(timeout()), this, SLOT(update()));

  if(isRealTime()) {
    realTimeChange(isRealTime());
  }
  else if(module->dspInChannels()) {
    draw();
  }
}

MJ_ModuleSpectrometer_QG::~MJ_ModuleSpectrometer_QG()
{}

MJ_ModuleDspGui *MJ_ModuleSpectrometer_QG::createNew(MJ_RtControl *rtparent,
						     MJ_ModuleDsp *module,
						     QWidget *parent, 
						     const char *name)
{
  return new MJ_ModuleSpectrometer_QG(rtparent, module, parent, name);
}

void MJ_ModuleSpectrometer_QG::update()
{
  const char * fname = "MJ_ModuleSpectrometer_QG::update";

  // trace(fname);

  MJ_ModuleSpectrometer *spectro = (MJ_ModuleSpectrometer *) m_module;
  
  lock();
  uint chans = spectro->dspInChannels();
  uint bands = spectro->bands();
  unlock();

  if(chans != m_graphs.count() || bands != (m_tmpBuffer.size() / 2)) {
    error2(MJ_ERR_INTERNAL, "%s # (%d != %d || %d != %d)", 
	   fname, chans, m_graphs.count(), bands, (m_tmpBuffer.size() / 2));
    return;
  }

  for(uint i=0; i < chans; i++) {

    lock();
    for(uint j=0; j < bands; j++) {
      MJ_FilterAndLevels *fal = & spectro->level(i, j);
      
      m_tmpBuffer.get(j) = 
	MJ_absToDb(MJ_sqrt(fal->absFilter().get()), -60) / 30 + 1.0;

      m_tmpBuffer.get(j + bands) = 
	MJ_absToDb(MJ_sqrt(fal->level().peak()), -60) / 30 + 1.0;
    }
    unlock();

    MJ_LevelDisplay *ld = (MJ_LevelDisplay *) m_graphs.at(i);
    ld->set(m_tmpBuffer.data(), m_tmpBuffer.data() + bands);
  }
  // trace(fname, "EXIT");
}

void MJ_ModuleSpectrometer_QG::realTimeChange (bool nowRt)
{
  if(nowRt) {
    m_timer.start(MJ_DisplayType::localDisplay() ? 31 : 102);

    draw();
  }
  else m_timer.stop();
}

void MJ_ModuleSpectrometer_QG::draw()
{
  const char * fname = "MJ_ModuleSpectrometer_QG::draw";

  MJ_ModuleSpectrometer *spectro = (MJ_ModuleSpectrometer *) m_module;

  lock();
  uint chans = spectro->dspInChannels();
  uint bands = spectro->upcomingBands();
  unlock();

  trace(fname, bands);

  if(!chans || !bands) {
    error(MJ_ERR_INTERNAL, fname, "Empty spectrometer");
    return;
  }

  // bool drawn;

  m_graphs.clear();

  if(m_table) {
    // drawn = true;
  
    if(m_graphLayout) delete  m_graphLayout;
  }
  else {
    int rows = m_layout->numRows();
    int cols = m_layout->numCols();
    
    m_layout->expand(rows + 1, cols);

    m_table = new QWidget(childParent(), "meter board");  
    m_layout->addMultiCellWidget(m_table, rows, rows, 0, cols - 1);
    m_layout->setRowStretch(rows, 100);
    // drawn = false;
  }
  
  int space = bands ? 10 : 0;
  
  if(m_vertical)
    m_graphLayout = new QVBoxLayout(m_table, 0, space);
  else
    m_graphLayout = new QHBoxLayout(m_table, 0, space);

  m_tmpBuffer.resize(bands * 2);
  
  for(uint i=0; i < chans; i++) {
    MJ_LevelDisplay *ld = new MJ_LevelDisplay(m_table);
    ld->setNofBars(bands);
    ld->setMinimumSize(100, 100);
    m_graphLayout->addWidget(ld);
    m_graphs.append(ld);
    ld->show();
  }
  
  m_table->show();

}

template class MJ_Array2D<MJ_FilterAndLevels>;

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


MJ_ModuleDsp * MJ_createNew_ModuleSpectrometer() 
{
  return new MJ_ModuleSpectrometer;
}

MJ_PluginDbeDsp *MJ_createPlugin(int i)
{
  if(i != 0) return 0;

  MJ_PluginDbeDsp *dbe = new MJ_PluginDbeDsp;

  dbe->setAuthors("Ilmonen Tommi S.");
  dbe->setMenuPosition("_Audio_Analysis_Spectrometer");
  dbe->setCreator(MJ_createNew_ModuleSpectrometer);
  dbe->setIdentifier("Tommi.Ilmonen@hut.fi_analysis_spectrometer");
  dbe->pluginVersion()->set(0,0,1);
  dbe->options()->setChans(1, 128, 0, 0);
  // dbe->setParameter(MJ_PP_TRANSLATION_FILE, <mj_spectrometer");
  dbe->setGuiFactory(MJ_GUI_POSTCARD, MJ_ModuleSpectrometer_QG::createNew);
  dbe->setParameter(MJ_PP_COMPILE_TIME, DATETIME);
  
  return dbe;  
}

