/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#ifndef MJ_ADDITIVE_VOICE_H
#define MJ_ADDITIVE_VOICE_H

#include <mj_synth_voice.h>

#include <mj_array.h>
#include <mj_ramp.h>

class MJ_AdditiveVoice;

enum MJ_SinusoidParam {
  MJ_SINPAR_FREQUENCY,
  MJ_SINPAR_AMPLITUDE,
  MJ_SINPAR_SPAN_TIME,
  MJ_SINPAR_SIZEOF
};

class MJ_SinusoidParams
{
public:

  void setParam(float v, MJ_SinusoidParam p) { m_params[p] = v; }
  float param(MJ_SinusoidParam p) const { return m_params[p]; }

protected:
  float m_params[MJ_SINPAR_SIZEOF];
};

/** A single sinusoid for MJ_AditiveSynth. */

class MJ_Sinusoid
{
public:
  MJ_Sinusoid();
  ~MJ_Sinusoid() {}
  
  void synthesize(MJ_Sample *buf, uint samples, MJ_AdditiveVoice *voice);
  
  void clear();

  /** Calculate the sinusoid of the sample.
      
      @arg v Phase angle in range 0 - m_tableElems.
  */
  float fastCos(float v)
  {
    unsigned index1 = (unsigned) v;
    float diff = v - index1;
    index1 = index1 & (m_tableMask);
    unsigned index2 = (index1 + 1) & (m_tableMask);
    return (1.0 - diff) * m_table[index1] + diff * m_table[index2];
  }

  static void tableInit();

protected:

  void updateParams(MJ_AdditiveVoice *voice);

  double          m_phase;
  uint            m_segmentEnter;

  MJ_Ramp<double> m_frequency;
  MJ_Ramp<double> m_amplitude;

  // For fast sinusoid table lookup:
  static const uint m_tableMask  = 0x3ff;
  static const uint m_tableElems = m_tableMask + 1;
  static double cycle() { return (double) m_tableElems; }
  /// Look-up table for cosine (speed).
  static float m_table[m_tableElems];
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** Voice that uses the additive synthesis approach.
*/

class MJ_AdditiveVoice : public MJ_SynthVoice
{
public:
  MJ_AdditiveVoice();
  virtual ~MJ_AdditiveVoice();

  virtual void synthesize(MJ_Sample *buf, uint samples);
  virtual MJ_SynthVoiceStatus status();

  virtual void calculateParams(const MJ_Sinusoid &, MJ_SinusoidParams *) {}
  
protected:
  MJ_Array<MJ_Sinusoid> m_sinusoids;
  uint                  m_age;
};

#endif
