/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_audio_sample.h>

#include <mj_error.h>
#include <mj_math.h>

#include <di_audio_file.h> 

#include <qfileinfo.h> 

#include <vector>

bool MJ_AudioSample::m_debug = true;

MJ_AudioSample::MJ_AudioSample()
  : m_sampleRate(0.0)
{}

MJ_AudioSample::~MJ_AudioSample()
{}

/** Load sample data from a file. This class uses caching to avoid
    reloading the file many times over. Usually the file will only be
    read if it has really changed.
    

    
*/

bool MJ_AudioSample::loadFile(const QString &fileName, 
			      bool forceRead)
{
  const char *fname = "MJ_AudioSample::loadFile";

  if(m_debug) MJ_trace(fname, (const char *) fileName);

  QFileInfo fi(fileName);
  
  if(!forceRead && 
     !m_fileLoadInfo.needsLoading(fi)) {

    if(m_debug) MJ_trace2("%s # Already loaded \"%s\"", 
			  fname, (const char *) fileName);
    return true;
  }

  DI_AudioFile af;

  const uint baxBlock = 4096;

  bool ok = af.open((const char *) fileName, 
		    DI_IO_INPUT, 
		    DI_ASF_FLOAT32, 
		    baxBlock);

  if(!ok) {
    MJ_error2(MJ_ERR_ARGUMENT, "%s # Could not load file \"%s\"", 
	      fname, (const char *) fileName);
    return false;
  }

  m_sampleRate = af.getSampleRate();
  
  int chans = (int) af.inChannels();

  if(chans != 1) {
    MJ_warning2(MJ_ERR_ARGUMENT, 
		"%s # Sound file \"%s\" has %d channels -> merging to mono", 
		fname, fileName.ascii(), chans);
  }

  uint frames = af.frameCount();

  ok = m_wave.resize(frames, false);

  if(!ok) {
    MJ_error2(MJ_ERR_MEMORY, "%s # Could not allocate memory (%u frames)", 
	      fname, frames);
    return false;
  }
  
  m_wave.setAll(0);

  uint location = 0;

  if(chans == 1) {

    while(location < frames) {
      
      uint leftFrames = frames - location;
      
    uint readNow = MJ_min(baxBlock, leftFrames);
    
    af.readFrames(& m_wave[location], readNow);
    location += readNow;
    }
  }
  else {
    std::vector<float> tmpBuf;
    tmpBuf.resize(baxBlock * chans);

    while(location < frames) {
      
      uint leftFrames = frames - location;
      
      uint readNow = MJ_min(baxBlock, leftFrames);
    
      readNow = af.readFrames(& tmpBuf[0], readNow);

      const float * source = & tmpBuf[0];

      for(uint i = 0; i < readNow; i++) {

	float sample = 0.0;
	
	for(uint j=0; j < (uint) chans; j++)
	  sample += *source++;
	
	m_wave[location + i] = sample / (MJ_Sample) chans;
      }

      location += readNow;
    }
  }

  m_fileLoadInfo.nowLoaded(fileName);

  if(m_debug) MJ_trace2("%s # loaded %d frames", fname, frames);

  return true;
}

  
