/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_granular_voice.h>

#include <mj_audio_sample.h>
#include <mj_error.h>
#include <mj_message_multi2.h>
#include <mj_math.h>
#include <mj_module_synth.h>
#include <mj_random.h>
#include <mj_strs.h>

static MJ_Random MJ_granuRand;

MJ_GranularVoice::MJ_GranularVoice()
  : m_sample(0),
    m_gain(1.0),
    m_pitch(1.0),
    m_pitchDistribution(0.1),
    m_rate(200.0),
    m_interval(1000.0),
    // m_sinceLast(0.0),
    m_untilNext(0),
    m_status(MJ_SVS_INACTIVE)
{
  m_grains.resize(128);
}

MJ_GranularVoice::~MJ_GranularVoice()
{} 

void MJ_GranularVoice::synthesize(MJ_Sample *buf, uint samples)
{
  for(Grain *grain = & * m_grains.begin(); 
      grain != & * m_grains.end(); 
      grain++) {

    if(grain->isActive()) {
      grain->synthesize(buf, samples, m_sample);

      if(!grain->isActive())
	m_activeCount--;
    }
  }

  // m_sinceLast += samples;
  m_untilNext -= samples;

  if(!m_activeCount && m_status == MJ_SVS_AT_END) {
    m_status = MJ_SVS_INACTIVE;
  }
  else if(m_status != MJ_SVS_AT_END) {
    
    // while(m_sinceLast > m_interval && m_activeCount < m_grains.size()) {
    while(m_untilNext < 0 && m_activeCount < m_grains.size()) {

      Grain *grain = findFreeGrain();
      
      grain->init(m_gain, 
		  m_pitch + MJ_granuRand.getFloat1(m_pitchDistribution));

      m_activeCount++;

      // Randomize the time until next grain.
      // m_sinceLast -= MJ_granuRand.getFloat0(m_interval * 2.0f);
      m_untilNext += (int) (MJ_granuRand.getFloat0() * m_interval * 2.0f);
      
      /* Now we synthesize a bit of sound, starting from random
         position. This helps us avoid the effect of all grains
         getting active at the cycle rate (audible artifact that would
         be). */

      // uint beginSample = MJ_granuRand.get() % samples;
      uint beginSample = -m_untilNext;
      beginSample = MJ_min(beginSample, samples);

      grain->synthesize(buf + beginSample, samples - beginSample, m_sample);
      
      if(!grain->isActive())
	m_activeCount--;
    }
  }
}

bool MJ_GranularVoice::msgProcess
(const char *id, MJ_Message *msg, MJ_ModuleSynth *synth)
{
  const char *fname = "MJ_GranularVoice::msgProcess";
  
  if(MJ_synthDebug)
    MJ_trace2("%s # %s", fname, id);

  if(!strcmp(id, MJ_Strs::noteon)) {
    m_activeCount = 0;
        
    MJ_MessageMulti2 *mm = dynamic_cast<MJ_MessageMulti2 *> (msg);
    
    QString sample;

    if(mm) {
      m_gain  = mm->getDouble(MJ_Strs::gain, 1.0);
      m_pitch = mm->getDouble(MJ_Strs::pitch, 1.0);
      m_pitchDistribution = mm->getDouble("pitchdistribution", m_pitch * 0.1);
      m_rate  = mm->getDouble(MJ_Strs::rate, 200.0);
      sample  = mm->getQString(MJ_Strs::sample, "grain.wav");
    }
    else
      sample = "grain.wav";

    m_interval = synth->dspSamplingRate() / m_rate;
    
    int sampleIndex = synth->findSample(sample);

    if(sampleIndex < 0) {
      MJ_error2(MJ_ERR_ARGUMENT, 
		"%s # No sample \"%s\"", fname, sample.ascii());
      
      return false;
    }

    m_sample = & synth->getSample(sampleIndex);
    
    clearGrains();

    m_untilNext = 0;
    m_status = MJ_SVS_ACTIVE;

    return true;
  }
  else if(!strcmp(id, MJ_Strs::noteoff)) {
    m_status = MJ_SVS_AT_END;
  }
  else if(!strcmp(id, MJ_Strs::noteoffnow)) {
    m_status = MJ_SVS_INACTIVE;
  }
  else { 
    bool ok = true;
    double value = msg->toDouble(&ok);
    
    if(!ok) return false;

    if(!strcmp(id, MJ_Strs::gain)) {
      m_gain = value;
      return true;
    }
    else if(!strcmp(id, MJ_Strs::pitch)) {
      m_pitch = value;
      return true;
    }
    else if(!strcmp(id, MJ_Strs::rate)) {
      m_rate = value;
      m_interval = synth->dspSamplingRate() / m_rate;
      return true;
    }
  }
  return false;
}

MJ_SynthVoiceStatus MJ_GranularVoice::status()
{
  return m_status;
}

MJ_GranularVoice::Grain * MJ_GranularVoice::findFreeGrain()
{
  for(Grain *grain = &*m_grains.begin(); grain != &*m_grains.end(); grain++)
    if(!grain->isActive())
      return grain;

  double maxCount = -1.0;

  Grain *best = &*m_grains.begin();

  for(Grain *grain = &*m_grains.begin(); grain != &*m_grains.end(); grain++) {
    if(grain->m_sampleCount > maxCount) {
      maxCount = grain->m_sampleCount;
      best = grain;
    }
  }

  return best;
}

void MJ_GranularVoice::clearGrains()
{
  Grain tmp;

  for(Grain *grain = &*m_grains.begin(); grain != &*m_grains.end(); grain++)
    *grain = tmp;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

void MJ_GranularVoice::Grain::synthesize
(MJ_Sample *buf, uint samples, const MJ_AudioSample *sample)
{
  if(!m_active) return;

  MJ_Sample *sentinel = buf + samples;

  double maxCount = (double) sample->wave().size();

  while(buf < sentinel) {

    if(m_sampleCount >= maxCount) {
      m_active = false;
      return;
    }
    
    *buf++ = sample->wave().getInterpolated1(m_sampleCount) * m_gain;
    m_sampleCount += m_pitchRelative;
  }
}


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_GranularVoiceFactory * MJ_GranularVoiceFactory::m_instance;

const char * MJ_GranularVoiceFactory::type()
{
  return "granular";
}

MJ_SynthVoice *MJ_GranularVoiceFactory::create()
{
  return new MJ_GranularVoice;
}
