/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#ifdef MJ_WITH_KARPLUS_STRONG

#include <mj_ks_voice.h>

#include <mj_error.h>
#include <mj_message_multi2.h>
#include <mj_message_string.h>
#include <mj_messages.h>
#include <mj_module_synth.h>
#include <mj_random.h>
#include <mj_strs.h>

static MJ_Random __MJ_KS_rand;

MJ_KS1Voice::MJ_KS1Voice(uint maxBuf)
  : m_buffer(maxBuf),
    m_feedback(0.99),
    m_delayTime(99),
    m_status(MJ_SVS_INACTIVE)
{}

MJ_KS1Voice::~MJ_KS1Voice()
{}

void MJ_KS1Voice::synthesize(MJ_Sample *buf, uint samples)
{
  if(m_status != MJ_SVS_ACTIVE)
    return;

  for(uint i=0; i < samples; i++) {
    float tmp = m_buffer.getNewest((uint) m_delayTime);
    
    tmp = m_filter.put(tmp) * m_feedback;

    m_buffer.put(tmp);
    
    *buf++ = tmp;
  }

  if(samples > m_untilEnd) {
    m_untilEnd = 0;
    m_status   = MJ_SVS_INACTIVE;
  }
  else
    m_untilEnd -= samples;
}

bool MJ_KS1Voice::msgProcess
(const char *id, MJ_Message *msg, MJ_ModuleSynth *synth)
{
  const char *fname = "MJ_KS1Voice::msgProcess";

  if(!strcmp(id, MJ_Strs::noteon)) {
    
    MJ_MessageMulti *mm = dynamic_cast<MJ_MessageMulti *> (msg);

    if(!mm) {
      MJ_error(MJ_ERR_ARGUMENT, fname, "Not a multi-noteon message");
      return false;
    }
    
    float gain  = mm->getDouble(MJ_Strs::gain, 1.0);
    float pitch = mm->getDouble(MJ_Strs::pitch, 220.0);
    m_feedback  = mm->getDouble(MJ_Strs::feedback, 0.999);

    m_delayTime = synth->dspSamplingRate() / pitch  - 0.5;

    // Fill the delay buffer with white noise:
    for(uint i=0; i < m_buffer.size(); i++) {
      m_buffer.getIndex(i) = __MJ_KS_rand.getFloat1(gain);
    }

    m_filter.reset();

    m_status = MJ_SVS_ACTIVE;

    m_untilEnd = MJ_MAX_UINT;
    
    return true;
  }
  else if(!strcmp(id, MJ_Strs::noteoff)) {

    bool ok = true;
    double tmp = msg->toDouble(&ok);

    if(ok)
      m_feedback = MJ_min((1.0 - tmp), 0.9);
    else
      m_feedback = 0.7;
    
    const double dyingThreshold = 0.001;

    double roundsBeforeDying = log(tmp) / log(dyingThreshold);
    
    double samplesBeforeDying = m_delayTime * roundsBeforeDying;

    m_untilEnd = (uint) samplesBeforeDying;

    const uint maxFadeOut = 20000;

    if(m_untilEnd > maxFadeOut) m_untilEnd = maxFadeOut;

    return true;
  }
  else if(!strcmp(id, MJ_Strs::noteoffnow)) {
    m_status = MJ_SVS_INACTIVE;
    m_untilEnd = 0;
    return true;    
  }

  return false;
}

MJ_SynthVoiceStatus MJ_KS1Voice::status()
{
  return m_status;
}
  
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

class MJ_KS1MidiConverter : public MJ_SynthMidiConverter
{
public:
  MJ_KS1MidiConverter() : m_targetChannel(0) {}
  virtual ~MJ_KS1MidiConverter() {}
  
  virtual bool convert(const DI_MidiEvent *e, MJ_ModuleSynth *synth);

  virtual uint maxChannel() { return m_targetChannel; }

protected:
  uint  m_targetChannel;
};


bool MJ_KS1MidiConverter::convert
(const DI_MidiEvent *e, MJ_ModuleSynth *synth)
{
  if(e->isNoteOn()) {
    float gain = MJ_dbToAbs (e->noteVelocity() * (40.0 / 127.0) - 40.0);

    MJ_MessageMulti2 mm;
    
    mm.addChild("voicetype", new MJ_MessageString("ks1"));
    mm.addChild(MJ_Strs::gain, new MJ_MessageDouble(gain));
    mm.addChild(MJ_Strs::pitch, new MJ_MessageDouble
		(DI_MidiEvent::keyNumberToHz(e->keyNumber())));
    mm.addChild(MJ_Strs::channel, new MJ_MessageInt(m_targetChannel));

    MJ_SynthVoiceHolder *svp = synth->noteOn(&mm);

    MJ_trace2("MJ_KS1MidiConverter::convert # Note-On %p", svp);
    
    if(svp) {
      svp->setMidi(e->keyNumber(), e->channel());
      return true;
    }
    return false;
  }

  return MJ_SynthMidiConverter::convert(e, synth);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_KS1VoiceFactory * MJ_KS1VoiceFactory::m_instance;

const char * MJ_KS1VoiceFactory::type()
{
  return "ks1";
}

MJ_SynthVoice *MJ_KS1VoiceFactory::create()
{
  return new MJ_KS1Voice;
}


MJ_SynthMidiConverter * MJ_KS1VoiceFactory::midiConverter()
{
  return new MJ_KS1MidiConverter;
}

#endif // MJ_WITH_KARPLUS_STRONG
