/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#ifndef MJ_KS_VOICE_H
#define MJ_KS_VOICE_H

#ifdef MJ_WITH_KARPLUS_STRONG

#include <mj_synth_voice.h>

#include <mj_math.h>
#include <mj_ringbuffer.h>

/** Simple Karplus-Strong voice model. 

    Block diagram:

<PRE>
     ----------------------------
   +-|          Z ^ -L          |--+---> output
   ^ ----------------------------  |
   |                               |
   |         y ------------ x      |
   +-----------| Modifier |--------+
               ------------
             /              \
          /                   \
       /                        \
    /                             \

   x ------------       --------
  ->-|G-feedback|--+----|Z ^ -1|-+
     ------------  |    -------- |
                   G0            G1
                   |             |   y
                   +-------------+----->
</PRE>

    Necessary formulas:

<PRE>
           Fs
   F0 = --------
         L + 1/2

         Fs     1
   L  = ---- - ---
         F0     2

   F0 = Synthesis frequency
   Fs = Sampling rate
   L  = delay line length

</PRE>

*/

class MJ_KS1Voice : public MJ_SynthVoice
{
public:
  MJ_KS1Voice(uint maxBuf = 4000);
  virtual ~MJ_KS1Voice();

  virtual void synthesize(MJ_Sample *buf, uint samples);
  virtual bool msgProcess(const char *id, MJ_Message *, MJ_ModuleSynth *);
  virtual MJ_SynthVoiceStatus status();
  
protected:

  class Fir2
  {
  public:
    Fir2() { reset(); }

    void reset()
    { m_g0 = 0.5f; m_g1 = 0.5f; m_m = 0.0f; }
    
    float put(float v) { 

#ifdef __OS_LINUX__
      if(MJ_abs(v) < 1e-6f) v = 0.0f;
#endif

      float out = v * m_g0 + m_m * m_g1;
      m_m = v;
      return out;
    }

    float m_g0, m_g1;
    float m_m;
  };


  MJ_RingBufferDelay<float> m_buffer;
  Fir2                      m_filter;
  float                     m_feedback;
  float                     m_delayTime;

  uint                      m_untilEnd;

  MJ_SynthVoiceStatus       m_status;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/// KS1 voice factory
class MJ_KS1VoiceFactory : public MJ_SynthVoiceFactory
{
public:
  MJ_KS1VoiceFactory() {}
  virtual ~MJ_KS1VoiceFactory() {}

  virtual const char * type();
  virtual MJ_SynthVoice *create();
  virtual MJ_SynthMidiConverter * midiConverter();

  static MJ_KS1VoiceFactory * instance() { return m_instance; }

  static void init()     { m_instance = new MJ_KS1VoiceFactory;}
  static void shutdown() { delete m_instance; m_instance = 0; }

protected:
  static MJ_KS1VoiceFactory * m_instance;
};

#endif // MJ_WITH_KARPLUS_STRONG

#endif
