/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_synth.h>

#include <mj_arraytmpl.C>
#include <mj_error.h>
#include <mj_input_voice.h>
#include <mj_message_string.h>
#include <mj_message_multi2.h>
#include <mj_messages.h>
#include <mj_misc.h>
#include <mj_strs.h>
#include <mj_parameters.h>

/// Voice types:
#include <mj_granular_voice.h>
#include <mj_noise_voice.h>
#include <mj_sample_voice.h>
#include <mj_stk_voice.h>
#include <mj_synth_voice.h>

#ifdef MJ_WITH_KARPLUS_STRONG
#include <mj_ks_voice.h>
#endif

#include <qfile.h> 
#include <qfiledialog.h> 
#include <qregexp.h>
#include <qtextstream.h> 

#include <typeinfo>

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** Clears a voice. This method also causes "note-off" to take
    place. */

void MJ_SynthVoiceHolder::clear(MJ_ModuleSynth *synth)
{
  if(MJ_synthDebug) MJ_trace2("MJ_SynthVoiceHolder::clear %p %p", this, m_voice);

  MJ_Message m;

  if(m_voice) m_voice->msgProcess(MJ_Strs::noteoffnow, &m, synth);

  m_voice = 0;
  m_targetChannel = 0;
  m_midiNote = -1;
  m_midiChannel = -1;
  m_id = -1;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

static const char * MJMS_sampleDirStr = "sampledir";
static const char * MJMS_forceReloads = "forceRelods";
static const char * MJMS_inChansStr = "inchans";

MJ_ModuleSynth::MJ_ModuleSynth()
  : m_idCount(0),
    m_requireInChannels(0),
    m_requireOutChannels(1),
    m_forceFileReloads(false)
{
  setPolyphony(128);
  setAudioSamples(128);

  // Add the voice factories:
  addFactory( MJ_SampleVoiceFactory::instance());
  addFactory( MJ_SampleVoicePitchedFactory::instance());
  addFactory( MJ_InputVoiceFactory::instance());
  addFactory( MJ_GranularVoiceFactory::instance());
  addFactory( MJ_NoiseVoiceFactory::instance());

#ifdef MJ_WITH_KARPLUS_STRONG
  addFactory( MJ_KS1VoiceFactory::instance());
#endif // MJ_WITH_KARPLUS_STRONG

  for(int i=0; true; i++) {
    MJ_StkVoiceFactory * factory = MJ_StkVoiceFactory::getInstance(i);
    if(factory)
      addFactory(factory);
    else
      break;
  }

  if(MJ_synthDebug) {
    for(voiceBankIterator it = m_voiceBanks.begin(); 
	it != m_voiceBanks.end(); it++)
      MJ_trace2("MJ_ModuleSynth::MJ_ModuleSynth # Has voice type \"%s\"",
		(*it).first.c_str());
  }

}

MJ_ModuleSynth::~MJ_ModuleSynth()
{}

bool MJ_ModuleSynth::dspAcceptIoChans(int *in, int *out)
{
  bool change = false;

  int needChans = needOutputChannels();
  
  if(*out != needChans) {
    *out = needChans;
    change = true;
  }

  if((uint) *in != m_requireInChannels) {
    *in = (int) m_requireInChannels;
    change = true;
  }

  return change;
}

bool MJ_ModuleSynth::dspPrepare()
{
  const char * fname = "MJ_ModuleSynth::dspPrepare";

  if(m_forceFileReloads) {
    // MJ_trace3(MJ_synthDebug, "%s # Reloading all sound files");
    
    // Delete all loaded samples.
    m_sampleMap.clear();
    uint s = m_samples.size();
    m_samples.resize(0);
    m_samples.resize(s);
  }

  bool ok = loadSetupFile(m_fileName);

  allNotesOff(true);
  
  m_idCount = 0;
  
  std::map<QString, int>::iterator it;

  for(it = m_sampleMap.begin(); it != m_sampleMap.end(); ++it ) {
    MJ_trace3(MJ_synthDebug, "%s # %s -> %d", fname, 
	      (const char *) (*it).first, (*it).second);
  }

  for(voiceBankIterator vb = m_voiceBanks.begin(); 
      vb != m_voiceBanks.end(); vb++) {

    bool vok = (*vb).second.factory()->dspPrepare(this);
    
    MJ_trace3(MJ_synthDebug, "%s # Voice type \"%s\"", 
	      fname, (*vb).first.c_str());

    ok = ok && vok;
  }


  return ok;
}

void MJ_ModuleSynth::dspProcessPeriod()
{
  uint frames = (uint) m_signalOutput[0].samples().size();

  // Zero all output buffers.
  dspClearOutputs();

  // Add synth voices to the output:
  for(uint i=0; i < m_activeVoices.size(); i++) {
    MJ_SynthVoiceHolder *vp = & m_activeVoices.getFast(i);

    MJ_SynthVoice *voice = vp->voice();
    
    if(voice) {
      MJ_SynthVoiceStatus status = voice->status();

      if(status == MJ_SVS_ACTIVE) {
	// printf("SYN ");
	uint targetChan = vp->targetChannel();
	voice->synthesize(m_signalOutput[targetChan].samples().data(),
			  frames);
      }
      else {
	if(MJ_synthDebug)
	  trace2("MJ_ModuleSynth::dspProcessPeriod # Cleared %u of %u", 
		 i, playingNotes());
	vp->clear(this);
      }
    }
  }
}

bool MJ_ModuleSynth::dspStop()
{
  trace3(MJ_synthDebug, "MJ_ModuleSynth::dspStop # %ld notes", m_idCount);

  allNotesOff(true);
  // clearSampleVoices();
  // clearInputVoices();

  return true;
}

bool MJ_ModuleSynth::msgProcess(const char * i, MJ_Message *m)
{
  const char * fname = "MJ_ModuleSynth::msgProcess";

  trace3(MJ_synthDebug, "%s # \"%s\"", fname, i);

  if(!strcmp(i, MJ_Strs::MIDI)) {
    MJ_MessageMidiEvent *md = dynamic_cast<MJ_MessageMidiEvent *> (m);
    
    if(!md) return false;

    return processMidiEvent(md->midiMsg());
  }

  if(!strcmp(i, MJ_Strs::noteon)) {

    MJ_MessageMulti2 *mm = dynamic_cast<MJ_MessageMulti2 *> (m);
    
    if(!mm) {
      error2(MJ_ERR_ARGUMENT, "%s # noteon # Bad message type \"%s\"", 
	     fname, typeid(*m).name());
      return false;
    }
    return noteOn(mm);
  }

  if(!strcmp(i, MJ_Strs::noteoff)) {
    bool ok = true;
    int id = m->toInt(&ok);

    if(!ok) return false;
    
    noteOff(id, 0.5);
    return true;
  }

  if(!strncmp(i, "note/", 5)) {

    if(strlen(i) < 6)
      return false;
    
    // Move pointer past end of "note/"
    i += 5;

    long noteId = atol(i);

    int noteIndex = findNote(noteId);

    if(noteIndex < 0) {
      if(MJ_synthDebug)
	MJ_trace2("%s # No note with id %ld", fname, noteId);
      return false;
    }

    const char * id2 = MJ_Strs::findNext(i, '/');
    
    if(!id2) {
      if(MJ_synthDebug)
	MJ_trace2("%s # Could not locate note parameter id", fname);
      return false;
    }

    bool ok = m_activeVoices.get(noteIndex).voice()->msgProcess
      (id2 + 1, m, this);
    
    return ok;
  }

  if(!strcmp(i, MJ_Strs::channels)) {
    bool ok = true;
    int v = m->toInt(&ok);

    if(!ok) return false;

    m_requireOutChannels = MJ_clamp(v, 1, MJ_SYNTH_MAX_CHANNELS);
    
    /* trace3(MJ_synthDebug, "%s # channels changed to %d", fname,
       m_requireChannels); */

    return true;
  }

  if(!strcmp(i, MJMS_inChansStr)) {
    bool ok = true;
    int v = m->toInt(&ok);

    if(!ok) return false;

    m_requireInChannels = MJ_clamp(v, 0, MJ_SYNTH_MAX_CHANNELS);

    /* trace3(MJ_synthDebug, "%s # channels changed to %d", fname,
       m_requireChannels); */

    return true;
  }

  else if(!strcmp(i, MJ_Strs::filename)) {
    MJ_MessageString *ms = dynamic_cast<MJ_MessageString *> (m);

    if(!ms) return false;
    if(ms->stringConst().isEmpty()) return false;
    
    m_fileName = ms->stringConst().copy();
    return true;
  }
  else if(!strcmp(i, MJMS_sampleDirStr)) {
    MJ_MessageString *ms = dynamic_cast<MJ_MessageString *> (m);

    if(!ms) return false;
    if(ms->stringConst().isEmpty()) return false;
    
    m_sampleDir = ms->stringConst().copy();
    return true;
  }
  else if(!strcmp(i, MJMS_forceReloads)) {

    bool v, ok = true;
    v = m->toBool(&ok);

    if(ok) m_forceFileReloads = v;
    else error(MJ_ERR_ARGUMENT, fname, "Not a boolean message");

    
    return ok;
  }
  if(!strcmp(i, "allnotesoff")) {
    bool ok = true;
    int now = m->toInt(&ok);
    allNotesOff(ok ? now : true);
    return true;
  }

  return false;
}

MJ_Parameter * MJ_ModuleSynth::msgGetInput(const char *i)
{

  if(!strcmp(i, MJ_Strs::channels)) {
    MJ_ParameterInt * pi = 
      new MJ_ParameterInt("Output channels", i, m_requireOutChannels,
			  1, MJ_SYNTH_MAX_CHANNELS);
    pi->setFlags(0);
    return pi;
  }
  if(!strcmp(i, MJMS_inChansStr))
    return new MJ_ParameterInt("Input channels", i, m_requireInChannels,
			       0, MJ_SYNTH_MAX_CHANNELS);
  
  if(!strcmp(i, MJ_Strs::filename))
    return new MJ_ParameterFilename(m_fileName, 
				    "Setup file", 
				    MJ_Strs::filename,
				    "*",
				    QFileDialog::ExistingFile);
  if(!strcmp(i, MJMS_sampleDirStr))
    return new MJ_ParameterFilename(m_sampleDir, 
				    "Sound file directory", 
				    MJMS_sampleDirStr,
				    "*",
				    QFileDialog::Directory);
  if(!strcmp(i, MJMS_forceReloads))
    return new MJ_ParameterBool(m_forceFileReloads, 
				"Force file reloads",
				MJMS_forceReloads);
  if(!strcmp(i, MJ_Strs::MIDI)) {
    MJ_ParameterMidi *pm = new MJ_ParameterMidi();
    pm->setName("MIDI input");
    pm->setId(i);
    return pm;
  }  

  return 0;
}

QCString MJ_ModuleSynth::msgGetInputName(int i)
{
  QCString r;

  if(i == 0) 
    r = MJ_Strs::channels;
  else if(i == 1) 
    r = MJMS_inChansStr;
  else if(i == 2) 
    r = MJ_Strs::filename;
  else if(i == 3) 
    r = MJMS_sampleDirStr;
  else if(i == 4) 
    r = MJMS_forceReloads;
  else if(i == 5) 
    r = MJ_Strs::MIDI;

  return r;
}

/** Process a MIDI event. */
bool MJ_ModuleSynth::processMidiEvent(const DI_MidiEvent &e)
{
  return m_midiSetup.convert(&e, this);
}

MJ_SynthVoiceHolder * MJ_ModuleSynth::noteOn(MJ_MessageMulti2 *mm)
{
  const char * fname = "MJ_ModuleSynth::noteOn";

  trace3(MJ_synthDebug, fname);

  QString type(mm->getQString("voicetype"));

  /* Update the id counter even if the voice cannot be
     initialized. This is done to guarantee that clients can guess the
     note ids properly. */
  long noteId = m_idCount++;
  
  if(type.isEmpty()) {
    error2(MJ_ERR_ARGUMENT, "%s # voicetype undefined", fname);
    return 0;
  }

  voiceBankIterator it = m_voiceBanks.find(type.ascii());

  if(it == m_voiceBanks.end()) {
    error2(MJ_ERR_ARGUMENT, "%s # No voices of type \"%s\" (available: %s)", 
	   fname, type.ascii(), (const char *) availableVoiceTypes());
    return 0;
  }

  VoiceBank *vb = & (*it).second;

  MJ_SynthVoice *sv = vb->findFreeVoice();
  int avi = allocateActiveVoice();

  if(!sv || avi < 0) {
    if(MJ_synthDebug)
      error2(MJ_ERR_INTERNAL, "%s # sv = %p avi = %d", fname, sv, avi);
    return 0;
  }

  bool ok = sv->msgProcess(MJ_Strs::noteon, mm, this);

  if(!ok) {
    
    error2(MJ_ERR_ARGUMENT, "%s # %s->msgProcess failed", 
	   fname, typeid(*sv).name());

    return 0;
  }
  
  MJ_SynthVoiceHolder *svp = m_activeVoices.begin() + avi;

  svp->setVoice(sv, noteId, this);

  uint chan = mm->getInt(MJ_Strs::channel, 0);
  if(chan  >= m_signalOutput.size()) {
    error2(MJ_ERR_ARGUMENT, "%s # Requested channel %u (>=%u)", 
	   fname, chan, m_signalOutput.size());
    chan = chan % m_signalOutput.size();
  }

  svp->setTargetChannel(chan);
  
  if(MJ_synthDebug)
    trace2("%s # \"%s\" EXIT SUCCESS", fname, typeid(*sv).name());

  return svp;
}

/** Generic note-off. 

    @return This method returns true if the note was found and
    note-off command given. Note that it is often ok to return false
    -- the note the user was referring to may have been finished all
    by itself.
*/

bool MJ_ModuleSynth::noteOff(long id, float velocity)
{
  int note = findNote(id);

  trace3(MJ_synthDebug, "MJ_ModuleSynth::noteOf # %d -> %d", 
	 (int) id, note);

  if(note < 0) return false;
  
  MJ_MessageDouble md(velocity);

  m_activeVoices.getFast(note).voice()->msgProcess
    (MJ_Strs::noteoff, &md, this);
  
  return true;
}

/** Silence all notes. */

void MJ_ModuleSynth::allNotesOff(bool forceNow)
{
  MJ_Message m;

  for(uint i=0; i < m_activeVoices.size(); i++) {
    MJ_SynthVoiceHolder *svp = & m_activeVoices.getFast(i);

    if(forceNow) svp->clear(this);
    else svp->voice()->msgProcess(MJ_Strs::noteoffnow, &m, this);
  }

  MJ_SynthVoiceHolder empty;
  m_activeVoices.setAll(empty);
}

/** Returns the number of output channels that are needed by this
    system. */
uint MJ_ModuleSynth::needOutputChannels()
{
  uint maxChan = m_midiSetup.maximumOutputChannel();

  if(maxChan >= MJ_SYNTH_MAX_CHANNELS) 
    trace("MJ_ModuleSynth::needOutputChannels # Many output channels");

  maxChan = MJ_max(maxChan + 1, m_requireOutChannels);

  return maxChan;
}

/** Returns the number of active notes. */

uint MJ_ModuleSynth::activeNotes()
{
  uint notes = 0;

  for(uint i=0; i < m_activeVoices.size(); i++)
    if(m_activeVoices.getFast(i).voice()) 
      notes++;

  return notes;
}

/** Load a setup file for the sample data etc. The file contains
    descriptions of audio files to be read, maximum polyphony etc.

    The file format is pretty simple. One line contains one packet of
    information. Each line starts with a keyword and the keyword is
    followed by corresponding data.
    
    Example:
    
\verbatim

# This is just a comment.

# Remove all STK voices from the setup. This is often needed if the
# user does not have all the STK-related subsystems configured
# properly (rawwave directories etc). This command should be before
# the "polyphony" command since otherwise the polyphony command may
# try to create voices that cannot be created.

unusevoices *-stk

# Set the maximum polyphony:
polyphony 64

# Set the maximum number of audio samples:
samples 128

# Load these audio files:
sample foo.wav
sample potato.aiff
sample end-of-the-world.aiff
sample linus-is-cool.wav

\endverbatim
    
    
*/

bool MJ_ModuleSynth::loadSetupFile(const QString &fileName)
{
  const char * fname = "MJ_ModuleSynth::loadSetupFile";

  if(MJ_synthDebug) trace(fname, fileName);

  static const char *fileCommandStrs[MJMS_FILECOM_SIZEOF] = {
    "polyphony",
    "samples",
    "sample",
    "midiconfig",
    "usevoices",
    "unusevoices"
  };

  bool ok;
  bool rval = ok = true;

  QFile file(fileName);

  if(!file.open(IO_ReadOnly)) {
    if(MJ_synthDebug) error2(MJ_ERR_ARGUMENT, fname, "Open failed");
    return false;
  }

  QTextStream stream(&file);

  while(!stream.atEnd()) {

    QString str = stream.readLine();

    QString field0(MJ_field(str, 0));
    QString field1(MJ_field(str, 1));
    QString field2(MJ_field(str, 2));

    int fields = MJ_fields(str);

    const char * cfield0 = field0;
    cfield0 = cfield0 ? cfield0 : "#";

    if(field0.isEmpty() || cfield0[0] == '#') {
      // Skip comments.
      ; 
    }
    else {
      MJMS_fileCommand cmd = (MJMS_fileCommand) 
	MJ_findStr(cfield0, fileCommandStrs, MJMS_FILECOM_SIZEOF);

      ok = true;

      if(cmd == MJMS_FILECOM_POLYPHONY) {
	uint voices = field1.toUInt(&ok);
	
	if(ok) setPolyphony(voices);
	else {
	  error(MJ_ERR_DATA_CORRUPTED, fname, "Could not get polyphony");
	}
	rval = rval && ok;
      }
      else if(cmd == MJMS_FILECOM_SAMPLES) {
	uint samples = field1.toUInt(&ok);

	if(ok) setAudioSamples(samples);
	else {
	  error(MJ_ERR_DATA_CORRUPTED, fname, "Could not get samples");
	}
	rval = rval && ok;
      }
      /* else if(cmd == MJMS_FILECOM_SAMPLES) {

      DUPLICATE ????
	uint samples = field1.toUInt(&ok);

	if(ok) setAudioSamples(samples);
	else {
	  error(MJ_ERR_DATA_CORRUPTED, fname, "Could not get samples");
	}
	rval = rval && ok;
	} */
      else if(cmd == MJMS_FILECOM_SAMPLE) {
	initSample(field1);
      }
      else if(cmd == MJMS_FILECOM_MIDICONF) {
	setMidiConf(field1, field2);
      }
      else if(cmd == MJMS_FILECOM_UNUSEVOICES) {
	for(int f=1; f < fields; f++) {
	  removeVoiceTypes(MJ_field(str, f));
	}
      }
    }
  }
  return rval;
}

bool MJ_ModuleSynth::initSample(const QString &fileName)
{
  const char * fname = "MJ_ModuleSynth::initSample";

  /* 
  MJ_trace3(MJ_synthDebug, "%s # \"%s\"", fname, (const char *) fileName);
  {
    std::map<QString, int>::iterator it;

    for(it = m_sampleMap.begin(); it != m_sampleMap.end(); ++it ) {
      MJ_trace3(MJ_synthDebug, "%s # %s -> %d", fname, 
		(const char *) (*it).first, (*it).second);
    }
    }*/

  QString tmp = m_sampleDir + fileName;

  int index = findSample(fileName);

  if(index < 0) {
    for(uint i=0; i < m_samples.size(); i++) {
      if(m_samples[i].fileName().isEmpty()) {
	index = (int) i;
	i = m_samples.size();
      }
    }
    if(index < 0) {
      MJ_error2(MJ_ERR_MEMORY, "%s # Out of polyphony", fname);
      return false;
    }
    MJ_trace3(MJ_synthDebug, "MJ_ModuleSynth::initSample # %d", index);
  }
  
  bool ok = m_samples.get(index).loadFile(tmp, false);
  
  if(ok) 
    m_sampleMap[fileName] = (int) index;
  
  return ok;
}

/** Sets the maximum polyphony to be used. The default is 128. This
    method should not be called during DSP.*/
bool MJ_ModuleSynth::setPolyphony(uint voices)
{
  if(MJ_synthDebug) trace("MJ_ModuleSynth::setPolyphony", voices);

  m_activeVoices.resize(voices);

  for(voiceBankIterator it = m_voiceBanks.begin();
      it != m_voiceBanks.end(); ++it)
    (*it).second.resize(voices);

  return true;
}

uint MJ_ModuleSynth::playingNotes()
{
  uint r = 0;

  for(uint i=0; i < m_activeVoices.size(); i++)
    if(m_activeVoices[i].voice())
      r++;

  return r;
}

/** Set the number of audio samples that can be loaded. The default is
    128. This method should not be called during DSP.*/
bool MJ_ModuleSynth::setAudioSamples(uint samples)
{
  if(dspIsRealTime())
    MJ_fatal(MJ_ERR_DO_NOT_CALL, "MJ_ModuleSynth::setAudioSamples");

  return m_samples.resize(samples, false);
}


/** Finds the sample that has the file loaded. 

    @arg file The file name of the sample.

    @return This method returns non-negative integer if the sample is
    found. If the sample is not found then -1 is returned.
*/
int MJ_ModuleSynth::findSample(const QString &file)
{
  std::map<QString, int>::iterator it = m_sampleMap.find(file);
  
  if(it == m_sampleMap.end()) return -1;
  
  return (*it).second;
}

/** Find a note based on MIDI parameters "note" and "channel". 

    @arg note MIDI key number
    @arg channel MIDI channel

    @return This method returns non-negative integer that is the index
    of the voice. If voice cannot be found, this method returns -1.
*/
int MJ_ModuleSynth::midiFindNote(short note, short channel)
{
  for(uint i=0; i < m_activeVoices.size(); i++) {
    MJ_SynthVoiceHolder *vp = & m_activeVoices.getFast(i);
    
    if(vp->voice() && vp->midiNote() == note && vp->midiChannel() == channel)
      return (int) i;
  }
  
  return -1;
}

int MJ_ModuleSynth::findNote(long id)
{
  if(id < 0) return -1;

  for(uint i=0; i < m_activeVoices.size(); i++)
    if(m_activeVoices.getFast(i).id() == id)
      return (int) i;
    
  return -1;
}

/** Find a free active voice and allocate it to the user. 

    @return This method returns non-negative integer that is the index
    of the voice. If all voices are active, this method returns -1.
*/
int MJ_ModuleSynth::allocateActiveVoice()
{
  for(uint i=0; i < m_activeVoices.size(); i++)
    if(!m_activeVoices.getFast(i).voice()) 
      return (int) i;
  
  return -1;
}

MJ_SynthVoice *MJ_ModuleSynth::allocateVoice(const std::string &type)
{
  voiceBankIterator it = m_voiceBanks.find(type);

  return (it == m_voiceBanks.end()) ? 0 : (*it).second.findFreeVoice() ;
}

void MJ_ModuleSynth::init()
{
  MJ_SampleVoiceFactory::init();
  MJ_SampleVoicePitchedFactory::init();
  MJ_InputVoiceFactory::init();
  MJ_GranularVoiceFactory::init();
  MJ_NoiseVoiceFactory::init();

#ifdef MJ_WITH_KARPLUS_STRONG
  MJ_KS1VoiceFactory::init();
#endif // MJ_WITH_KARPLUS_STRONG
}

void MJ_ModuleSynth::shutdown()
{
  MJ_SampleVoiceFactory::shutdown();
  MJ_SampleVoicePitchedFactory::shutdown();
  MJ_InputVoiceFactory::shutdown();
  MJ_GranularVoiceFactory::shutdown();
  MJ_NoiseVoiceFactory::shutdown();

#ifdef MJ_WITH_KARPLUS_STRONG
  MJ_KS1VoiceFactory::shutdown();
#endif // MJ_WITH_KARPLUS_STRONG
}

bool MJ_ModuleSynth::setMidiConf(const QString & channel, const QString & type)
{
  const char * fname = "MJ_ModuleSynth::setMidiConf";

  bool ok = true;
  uint chan = channel.toUInt(&ok) - 1;
  
  if(!ok || chan > DI_MIDI_CHANNELS) {
    error2(MJ_ERR_ARGUMENT, "%s # Invalid MIDI channel %s", 
	   fname, channel.ascii());
    return false;
  }

  voiceBankIterator it = m_voiceBanks.find(std::string(type.ascii()));

  if(it == m_voiceBanks.end()) {
    MJ_error2(MJ_ERR_ARGUMENT, "%s # No voice bank for type \"%s\"",
	      fname, type.ascii());
    return false;
  }

  MJ_SynthMidiConverter * converter = 
    (*it).second.m_factory->midiConverter();

  if(!converter) {
    MJ_error2(MJ_ERR_ARGUMENT, "%s # No MIDI converter for type \"%s\"",
	      fname, type.ascii());
    return false;
  }

  m_midiSetup.setConverter(chan, converter);

  MJ_trace2("%s # Successfully set the MIDI converter %u to \"%s\"",
	    fname, chan, type.ascii());
  
  return true;
}

/** Removes all voice types that match given regular expression. */

void MJ_ModuleSynth::removeVoiceTypes(const QString &str)
{
  QRegExp re(str, true, true);

  voiceBankIterator it = m_voiceBanks.begin();

  while(it != m_voiceBanks.end()) {
    
    QString type((*it).first.c_str());
    
    voiceBankIterator tmp(it);
    tmp++;

    if(type.find(re) >= 0) {
      m_voiceBanks.erase(it);
      MJ_trace2("MJ_ModuleSynth::removeVoiceTypes # Removing %s (%s)",
		type.ascii(), str.ascii());
    }
    it = tmp;
  }
}

QCString MJ_ModuleSynth::availableVoiceTypes() const
{
  QCString r;

  for(voiceBankIteratorC it = m_voiceBanks.begin(); 
      it !=m_voiceBanks.end(); it++) {
    r += (*it).first.c_str();
    r += " ";
  }

  return r;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_SynthVoice *MJ_ModuleSynth::VoiceBank::findFreeVoice()
{
  for(MJ_SynthVoice ** voice = &*m_voices.begin(); 
      voice != &*m_voices.end(); voice++)
    if((*voice)->status() == MJ_SVS_INACTIVE)
      return *voice;

  return 0;
}

void MJ_ModuleSynth::VoiceBank::resize(uint n)
{
  if(!m_factory) 
    MJ_fatal(MJ_ERR_INTERNAL, "MJ_ModuleSynth::VoiceBank::resize");

  if(n < m_voices.size()) {

    for(MJ_SynthVoice ** voice = &*m_voices.begin() + n; 
	voice != &*m_voices.end(); voice++)
      delete *voice;

    m_voices.resize(n);
  }
  else {
    while(n > m_voices.size()) {
      m_voices.push_back(m_factory->create());
    }
  }
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

// Instantiate the debugging variable:
bool MJ_synthDebug = false;

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

// Template instantations:

template class MJ_Array<MJ_SynthVoiceHolder>;
// template class MJ_Array<MJ_InputVoice>;
// template class MJ_Array<MJ_SampleVoice>;
template class MJ_Array<MJ_AudioSample>;

// template class std::pair<QString, int>;

