/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#ifndef MJ_MODULE_SYNTH_H
#define MJ_MODULE_SYNTH_H

#include <mj_audio_sample.h>
#include <mj_message.h>
#include <mj_module_dsp.h>
#include <mj_synth_midi.h>
#include <mj_synth_voice.h>

#include <map>
#include <string>
#include <vector>

#define MJ_SYNTH_MAX_CHANNELS 128

class MJ_MessageMulti2;

/** Helper for MJ_ModuleSynth. This class is used to hold pointer and
    information about a particular voice. 

    <B>ID:</B> Each note has a unique id. Whenever a new note-on takes
    place a new id is allocated. Thus it is possible for user to to
    know when a particular note is really being played.

    @author Tommi Ilmonen 

*/
class MJ_SynthVoiceHolder
{
public:
  MJ_SynthVoiceHolder() : m_voice(0) { clear(0); }
  ~MJ_SynthVoiceHolder() {}

  MJ_SynthVoice *voice() { return m_voice; }
  void setVoice(MJ_SynthVoice *v, long id, MJ_ModuleSynth *synth) 
  { clear(synth); m_voice = v; m_id = id; }

  uint targetChannel() const { return m_targetChannel; }
  void setTargetChannel(uint c) { m_targetChannel = c; }

  void clear(MJ_ModuleSynth *synth);

  void setMidi(short note, short channel)
  { m_midiNote = note; m_midiChannel = channel; }

  short midiNote() const { return m_midiNote; }
  short midiChannel() const { return m_midiChannel; }

  long id() const { return m_id; }

protected:
  MJ_SynthVoice *m_voice;
  uint           m_targetChannel;

  long           m_id;
  short          m_midiNote;
  short          m_midiChannel;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** Synthesizer module. This class uses "synth voices" to get any
    sound output to the user. The voices may be really simple (sample
    playback) or really complex (additive synth).

    This synth has <I>multiple control mechanisms</I>. It can be
    controlled via MIDI, but it can also be controlled via other
    methods. The besic philosophy is not particularly MIDI-oriented
    altough MIDI-compatibility is included.
    
    While this synth should be a decent platform for the most usual
    classical synthesis stuff I do not expect it is good for physical
    modeling (excect for monophonic sources).

    So far there is no GUI for this class. It is meant to be used with
    a configuration file system.

    A novel feature of this synth is that it can have a great number
    of output channels -> The resulting sound can be auralized
    efficiently.

    @author Tommi Ilmonen
*/
class MJ_ModuleSynth : public MJ_ModuleDsp
{
public:
  MJ_ModuleSynth();
  virtual ~MJ_ModuleSynth();
  
  virtual bool dspAcceptIoChans(int *, int *);

  virtual bool dspPrepare();
  virtual void dspProcessPeriod();
  virtual bool dspStop();

  virtual bool msgProcess(const char *, MJ_Message *);
  virtual MJ_Parameter * msgGetInput(const char *);
  virtual QCString msgGetInputName(int);

  /** @name Public MIDI methods */
  //@{
  bool processMidiEvent(const DI_MidiEvent &);
  /* MJ_SynthVoiceHolder * midiNoteOn(const MJ_MidiEvent &);
     bool midiNoteOff(const MJ_MidiEvent &);*/
  //@}

  /** @name Synthesis commands */
  //@{

  MJ_SynthVoiceHolder * noteOn(MJ_MessageMulti2 *);
  bool noteOff(long id, float velocity);
  void allNotesOff(bool forceNow);
  //@}

  uint needOutputChannels();

  uint activeNotes();

  int findSample(const QString &);
  MJ_AudioSample & getSample(uint i) { return m_samples.getFast(i); }
  int midiFindNote(short note, short channel);
  int findNote(long id);
  MJ_SynthVoiceHolder * getActiveVoice(int index)
  { return & m_activeVoices[index]; }

  void addFactory(MJ_SynthVoiceFactory *factory)
  { m_voiceBanks[factory->type()].m_factory = factory; }

  static void init();
  static void shutdown();

protected:

  enum MJMS_fileCommand {
    MJMS_FILECOM_POLYPHONY,
    MJMS_FILECOM_SAMPLES,
    MJMS_FILECOM_SAMPLE,
    MJMS_FILECOM_MIDICONF,
    MJMS_FILECOM_USEVOICES,
    MJMS_FILECOM_UNUSEVOICES,
    MJMS_FILECOM_SIZEOF
  };

  class VoiceBank {
  public:
    VoiceBank() : m_factory(0) {}

    MJ_SynthVoice *findFreeVoice();
    void resize(uint);

    MJ_SynthVoiceFactory *factory() { return m_factory; }

    std::vector<MJ_SynthVoice *> m_voices;
    MJ_SynthVoiceFactory *m_factory;

    bool m_use;
  };

  bool loadSetupFile(const QString &fileName);

  bool initSample(const QString &fileName);
  
  bool setPolyphony(uint);
  uint playingNotes();
  bool setAudioSamples(uint);
  
  int allocateActiveVoice();
  MJ_SynthVoice *allocateVoice(const std::string &type);
  bool setMidiConf(const QString & channel, const QString & type);

  void removeVoiceTypes(const QString &);

  QCString availableVoiceTypes() const;

  MJ_Array<MJ_SynthVoiceHolder>  m_activeVoices;

  std::map<std::string, VoiceBank> m_voiceBanks;

  typedef std::map<std::string, VoiceBank>::iterator voiceBankIterator;
  typedef std::map<std::string, VoiceBank>::const_iterator voiceBankIteratorC;

  MJ_Array<MJ_AudioSample>   m_samples;
  std::map<QString, int>     m_sampleMap;

  long                       m_idCount;

  QString                    m_fileName;

  QString                    m_sampleDir;

  uint                       m_requireInChannels;
  uint                       m_requireOutChannels;

  bool                       m_forceFileReloads;

  MJ_SynthMidiSetup          m_midiSetup;
};

#endif
