/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_noise_voice.h>

#include <mj_error.h>
#include <mj_message_multi.h>
#include <mj_module_synth.h>
#include <mj_strs.h>

#include <typeinfo>


MJ_NoiseVoice::MJ_NoiseVoice()
{
  m_location = 0;
  m_endLocation = 0;
  m_sr = 44100;
}

MJ_NoiseVoice::~MJ_NoiseVoice()
{}

void MJ_NoiseVoice::synthesize(MJ_Sample *buf, uint samples)
{
  if(m_location >= m_endLocation) return;

  uint lastSample = MJ_min(m_location + samples, m_endLocation);

  for(; m_location < lastSample; m_location++) {

    m_gain.advance();

    if(m_pitch.samplesToTarget() || m_octaves.samplesToTarget()) {
      m_pitch.advance();
      m_octaves.advance();
      redesignFilter();
    }

    MJ_Sample r = m_rand.getFloat1();
    m_filter.put(r);
    *buf = m_filter.get() * m_gain.get();

#ifdef __OS_LINUX__
    if(isinf(*buf) || isnan(*buf)) {
      char cbuf[128];
      m_filter.writeTo(cbuf);
      MJ_error2(MJ_ERR_UNKNOWN, 
		"MJ_NoiseVoice::synthesize # not finite # %f %f %f %f (%s)",
		(float) m_filter.get(), (float) m_gain.get(), 
		(float) m_pitch.get(),  (float) m_octaves.get(), cbuf);
    }
#endif // __OS_LINUX__
    
    buf++;
  }
}

bool MJ_NoiseVoice::msgProcess(const char *id, 
			       MJ_Message *msg, 
			       MJ_ModuleSynth *synth)
{
  // MJ_trace2("MJ_NoiseVoice::msgProces # %s %s", id, typeid(*msg).name());

  if(strcmp(id, MJ_Strs::noteon) == 0) {

    MJ_MessageMulti * mm = dynamic_cast<MJ_MessageMulti *> (msg);

    if(!mm)
      return false;
    
    float gain = mm->getDouble(MJ_Strs::gain, 1.0);
    float pitch = mm->getDouble(MJ_Strs::pitch, 1.0);
    float octaves = mm->getDouble("octaves", 1.0);

    m_pitch.reset(pitch);
    m_octaves.reset(octaves);
    m_gain.set(1000, 0.0, gain);

    m_location = 0;
    m_endLocation = (uint) -1;

    m_sr = synth->dspSamplingRate();

    redesignFilter();

    m_ending = false;
    m_filter.resetMemory();

    return true;
  }
  else if(strcmp(id, MJ_Strs::noteoff) == 0) {

    uint interpSamples = 1000;
    
    m_endLocation = MJ_min(m_endLocation, m_location + interpSamples);
    m_gain.setTarget(m_endLocation - m_location, 0.0);
    
    m_ending = true;

    return true;
  }
  else if(strcmp(id, MJ_Strs::gain) == 0) {
    
    uint interpSamples = 1000;

    bool ok = true;
    double v = msg->toDouble(&ok);

    /* If the voice is ending one cannot adjust the gain any more
       since it is being interpolated to zero. */
    if(ok && !m_ending) {
      m_gain.setTarget(interpSamples, v);
    }
    
    return ok;
  }
  else if(strcmp(id, MJ_Strs::pitch) == 0) {
    
    uint interpSamples = 1000;

    bool ok = true;
    double v = MJ_clamp(msg->toDouble(&ok), 20.0, 10000.0);

    if(ok) {
      m_pitch.setTarget(interpSamples, v, MJ_RAMP_EXPONENTIAL);
      /* MJ_trace2("MJ_NoiseVoice::msgProces # pitch %lf -> %lf", 
	 m_pitch.get(), v); */
    }
    
    return ok;
  }
  else if(strcmp(id, "octaves") == 0) {
    
    uint interpSamples = 1000;

    bool ok = true;
    double v = MJ_clamp(msg->toDouble(&ok), 0.05, 3.0);

    if(ok) {
      m_octaves.setTarget(interpSamples, v, MJ_RAMP_LINEAR);
    }
    
    return ok;
  }

  return false;
}

MJ_SynthVoiceStatus MJ_NoiseVoice::status()
{
  if(m_location >= m_endLocation)
    return MJ_SVS_INACTIVE;

  return MJ_SVS_ACTIVE;
}
  
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


MJ_NoiseVoiceFactory * MJ_NoiseVoiceFactory::m_instance = 0;

MJ_NoiseVoiceFactory::MJ_NoiseVoiceFactory()
{}

MJ_NoiseVoiceFactory::~MJ_NoiseVoiceFactory()
{}

const char * MJ_NoiseVoiceFactory::type()
{
  return "noise";
}

MJ_SynthVoice * MJ_NoiseVoiceFactory::create()
{
  return new MJ_NoiseVoice;
}
