/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_sample_voice.h>

#include <mj_audio_sample.h>
#include <mj_error.h>
#include <mj_math.h>
#include <mj_message_multi2.h>
#include <mj_messages.h>
#include <mj_message_string.h>
#include <mj_module_synth.h>
#include <mj_strs.h>
#include <mj_synth_midi.h>

#include <typeinfo>

MJ_SampleVoice::MJ_SampleVoice()
  : m_sample(0),
    m_startLocation(0),
    m_location(0),
    m_endLocation(0),
    m_loopInterleave(0),
    m_loop(false),
    m_ending(true),
    m_gain(0.5)
{}

MJ_SampleVoice::~MJ_SampleVoice()
{}

void MJ_SampleVoice::synthesize(MJ_Sample *buf, uint samples)
{
  if(m_location >= m_endLocation) return;

  MJ_Vector<float> *wave = & m_sample->wave();

  uint framesLeft = m_endLocation - m_location - m_loopInterleave;

  uint process = MJ_min(framesLeft, samples);

  // MJ_trace2("MJ_SampleVoice::synthesize # %d", process);

  float *waveData = wave->data() + m_location;
  float *sentinel = waveData + process;

  if(m_loopInterleave == 0) {

    while(waveData < sentinel) {
      m_ramp.advanceLinear();
      // printf(" %f ", *waveData * m_ramp.get());
      *buf++ += (*waveData++ * m_ramp.get());
    }
    
    m_location += process;

    /* MJ_trace2("MJ_SampleVoice::synthesize # %d %f", 
       (int) m_location, (float) m_ramp.get()); */
  }
  else {
    
    uint loc = m_location;
    uint endFade = m_startLocation + m_loopInterleave;
    uint distance = m_endLocation - m_startLocation - m_loopInterleave;
    MJ_Sample gainScale = (MJ_Sample) 1.0 / (MJ_Sample) m_loopInterleave;

    while(waveData < sentinel) {

      m_ramp.advanceLinear();

      if(loc < endFade) {
	// Fade beginning in and end out.

	MJ_Sample g1 = (loc - m_startLocation) * gainScale;

	if(g1 > 1.0)
	  g1 = 1.0;	

	MJ_Sample g2 = 1.0 - g1;
	MJ_Sample g3 = m_ramp.get();
	
	*buf += (*waveData * g1 * g3); // beginning 
	*buf += (* (waveData + distance) * g2 * g3); // end
	
      }
      else {
	*buf += (*waveData * m_ramp.get());
      }
      buf++;
      waveData++;
      loc++;
    }

    if(loc == m_endLocation - m_loopInterleave) {
      rewind();
      synthesize(buf, samples - process);
    }
    else {
      m_location = loc;
    }
  }
  
  if((m_location == m_endLocation) && m_loop) {
    rewind();
    synthesize(buf + process, samples - process);
  }
}

bool MJ_SampleVoice::msgProcess
(const char *id, MJ_Message *msg, MJ_ModuleSynth *synth)
{
  const char * fname = "MJ_SampleVoice::msgProcess";

  MJ_trace3(MJ_synthDebug, "%s # \"%s\" %p %p", fname, id, this, synth);

  // MJ_trace2("MJ_SampleVoice::command # %d %f", (int) cmd, value);

  if(strcmp(id, MJ_Strs::noteon) == 0) {

    MJ_MessageMulti2 * mm = dynamic_cast<MJ_MessageMulti2 *> (msg);

    if(!mm) {
      MJ_error(MJ_ERR_ARGUMENT, 
	       "%s # Wrong message type \"%s\"", fname, typeid(*msg).name());
      return false;
    }

    m_ramp.reset(mm->getDouble(MJ_Strs::gain, 1.0));
    m_loop = mm->getInt(MJ_Strs::loop, 0);
    m_ending = false;
    
    QString fileName(mm->getQString(MJ_Strs::filename, "sound.wav"));
    
    int sampleIndex = synth->findSample(fileName);

    if(sampleIndex < 0) {
      MJ_error2(MJ_ERR_ARGUMENT, 
		"%s # No sample \"%s\"", fname, fileName.ascii());
      
      return false;
    }

    m_sample = & synth->getSample(sampleIndex);
    m_endLocation = m_sample->wave().size();

    double startLoc = mm->getDouble("skipseconds", 0.0);
    m_startLocation = (uint) (startLoc * m_sample->sampleRate());

    double interleave = mm->getDouble("interleave", 0.0);
    m_loopInterleave = (uint) (interleave * m_sample->sampleRate());

    rewind();

    return true;
  }
  else if(strcmp(id, MJ_Strs::noteoff) == 0) {
    uint interpSamples = 1000;
    
    m_endLocation = MJ_min(m_endLocation, m_location + interpSamples);
    m_ramp.setTarget(m_endLocation - m_location, 0.0, MJ_RAMP_LINEAR);
    m_ending = true;
    return true;
  }
  else if(strcmp(id, MJ_Strs::noteoffnow) == 0) {
    m_sample = 0;
    m_location = 0;
    m_endLocation = 0;
    
    m_ending = true;
    return true;
  }
  else if(strcmp(id, MJ_Strs::gain) == 0) {
    
    bool ok = true;
    double value = msg->toDouble(&ok);

    if(!m_ending && ok) {
      m_gain = value;
      uint interpSamples = 1000;
      m_ramp.setTarget(interpSamples, value, MJ_RAMP_LINEAR);
    }
    return ok;
  }
  else if(strcmp(id, "interleave") == 0) {
    
    bool ok = true;
    double value = msg->toDouble(&ok);

    if(m_sample && ok) {
      m_loopInterleave = (uint) (value * m_sample->sampleRate());    
    }
    return ok;
  }
  return false;
}

MJ_SynthVoiceStatus MJ_SampleVoice::status()
{
  if(!m_sample) return MJ_SVS_INACTIVE;

  if(m_ending && m_ramp.get() == 0.0)
    return MJ_SVS_AT_END;
  
  return (m_location < m_endLocation) ? 
    MJ_SVS_ACTIVE : MJ_SVS_AT_END;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

class MJ_SampleMidiConverter : public MJ_SynthMidiConverter
{
public:
  MJ_SampleMidiConverter()  
    : m_sampleName("tom.aiff"),
      m_targetChannel(0),
      m_defaultGain(1.0) {}

  virtual ~MJ_SampleMidiConverter() {}

  virtual bool convert(const DI_MidiEvent *e, MJ_ModuleSynth *synth);

  virtual uint maxChannel() { return m_targetChannel; }
  
  
  QString m_sampleName;

  uint  m_targetChannel;
  float m_defaultGain;
};

bool MJ_SampleMidiConverter::convert
(const DI_MidiEvent *e, MJ_ModuleSynth *synth)
{
  if(e->isNoteOn()) {
    float gain = MJ_dbToAbs (e->noteVelocity() * (40.0 / 128.0) - 40.0);

    MJ_MessageMulti2 mm;
    
    mm.addChild("voicetype", new MJ_MessageString("sample"));
    mm.addChild(MJ_Strs::gain, new MJ_MessageDouble(gain));
    mm.addChild(MJ_Strs::channel, new MJ_MessageInt(m_targetChannel));
    mm.addChild(MJ_Strs::loop, new MJ_MessageInt(0));
    mm.addChild(MJ_Strs::filename, new MJ_MessageString(m_sampleName));

    MJ_SynthVoiceHolder *svp = synth->noteOn(&mm);
    
    if(svp) {
      svp->setMidi(e->keyNumber(), e->channel());
      return true;
    }
    return false;
  }
  
  return MJ_SynthMidiConverter::convert(e, synth);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_SampleVoiceFactory *MJ_SampleVoiceFactory::m_instance = 0;

const char * MJ_SampleVoiceFactory::type()
{
  return "sample";
}

MJ_SynthVoice *MJ_SampleVoiceFactory::create()
{
  return new MJ_SampleVoice;
}

MJ_SynthMidiConverter * MJ_SampleVoiceFactory::midiConverter()
{
  return new MJ_SampleMidiConverter;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_SampleVoicePitched::MJ_SampleVoicePitched()
  : m_relativePitch(1.0)
{
  m_resampler.setParameters(0.3, 2, 4);
}

MJ_SampleVoicePitched::~MJ_SampleVoicePitched()
{}

void MJ_SampleVoicePitched::synthesize(MJ_Sample *buf, uint samples)
{
  // MJ_trace("MJ_SampleVoicePitched::synthesize");

  if(m_location >= m_endLocation) return;

  MJ_Vector<float> *wave = & m_sample->wave();
  float *waveData = wave->begin() + m_location;
  float *sentinel = wave->end();
  float *sentinel2 = wave->begin() + m_endLocation;

  if(sentinel > sentinel2)
    sentinel = sentinel2;

  // printf("%p -> %p # ", waveData, sentinel);

  for(uint i=0; i < samples; i++) {
    
    while(m_resampler.needNewSample()) {

      if(waveData >= sentinel) {
	if(m_loop) {
	  m_location = m_startLocation;
	  waveData = wave->begin() + m_location;
	  // MJ_trace("MJ_SampleVoicePitched::synthesize # loop");
	}
	else {
	  m_endLocation = m_location;
	  m_ending = true;
	  m_ramp.reset(0.0);

	  // MJ_trace("MJ_SampleVoicePitched::synthesize # sound end reached");

	  return;
	}
      }

      m_resampler.put(*waveData++);
      m_location++;
      // m_resampler.put(0.0);
    }
    
    m_relativePitch.advanceExponential();

    m_ramp.advanceLinear();

    float newSample = m_resampler.getResampled(m_relativePitch.get());
    // float newSample = m_resampler.getResampled(1.0);
    *buf++ = newSample * m_ramp.get();
  }

  // m_location += samples;
}

bool MJ_SampleVoicePitched::msgProcess
(const char *id, MJ_Message *msg, MJ_ModuleSynth *synth)
{
  if(!strcmp(id, "relpitch")) {
    bool ok = true;

    double value = msg->toDouble(&ok);

    if(!ok) return false;

    value = MJ_clamp(value, 1e-3, 1e+3);
    
    uint interpsamples = 1000;
    m_relativePitch.setTarget(interpsamples, value, MJ_RAMP_EXPONENTIAL);

    return true;
  }
  else if(strcmp(id, MJ_Strs::noteon) == 0) {
    bool ok = MJ_SampleVoice::msgProcess(id, msg, synth);

    if(!ok) return false;

    MJ_MessageMulti2 * mm = dynamic_cast<MJ_MessageMulti2 *> (msg);

    if(mm) 
      m_relativePitch.reset(mm->getDouble("relpitch", 1.0));
    
    /* MJ_trace2("MJ_SampleVoicePitched::msgProcess # rp = %f", 
       (float) m_relativePitch.target()); */

    return true;
  }

  return MJ_SampleVoice::msgProcess(id, msg, synth);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

MJ_SampleVoicePitchedFactory * MJ_SampleVoicePitchedFactory::m_instance = 0;

MJ_SampleVoiceFactory::MJ_SampleVoiceFactory()
{
  MJ_trace2("MJ_SampleVoiceFactory::MJ_SampleVoiceFactory # %p", this);
}

MJ_SampleVoiceFactory::~MJ_SampleVoiceFactory()
{
  MJ_trace2("MJ_SampleVoiceFactory::~MJ_SampleVoiceFactory # %p", this);
}

const char * MJ_SampleVoicePitchedFactory::type()
{
  return "pitchsample";
}

MJ_SynthVoice *MJ_SampleVoicePitchedFactory::create()
{
  return new MJ_SampleVoicePitched;
}

