/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#ifndef MJ_STK_VOICE_TMPL_C
#define MJ_STK_VOICE_TMPL_C

#include <mj_stk_voice.h>

#include <mj_error.h>
#include <mj_message_multi.h>
#include <mj_module_synth.h>
#include <mj_strs.h>

#include <Stk.h>

template <class T>
MJ_StkVoice<T>::MJ_StkVoice()
  : m_status(MJ_SVS_INACTIVE),
    m_untilEnd(-1)
{}

template <class T>
MJ_StkVoice<T>::~MJ_StkVoice()
{}

template <class T>
void MJ_StkVoice<T>::synthesize(MJ_Sample *buf, uint samples)
{
  MJ_Sample *max = buf + samples;

  while(buf < max) {
    /* The "T::" removes virtualization even if the compiler would not
       otherwise realize that things should not be virtual ->
       faster. */
    *buf += m_instrument.T::tick();
    buf++;
  }

  if(m_untilEnd > 0) {
    if(m_untilEnd > (int) samples)
      m_untilEnd -= samples;
    else {
      m_untilEnd = 0;
      m_status = MJ_SVS_INACTIVE;
    }
  }
}

template <class T>
bool MJ_StkVoice<T>::msgProcess
(const char *id, MJ_Message *m, MJ_ModuleSynth *s)
{
  const char *fname = "MJ_StkVoice::msgProcess";

  MJ_trace2("%s # %s", fname, id);

  if(internalMsgProcess(id, m, s))
    return true;

  if(!strcmp(id, MJ_Strs::noteon)) {

    MJ_MessageMulti *mm = dynamic_cast<MJ_MessageMulti *> (m);

    if(mm) {
      if(noteOn(mm, s)) {
	m_status = MJ_SVS_ACTIVE;
	m_untilEnd = -1;
	return true;
      }
      return false;
    }
    
    return false;
  }
  
  else if(!strcmp(id, MJ_Strs::noteoff)) {

    MJ_trace2("%s # noteoff", fname);

    bool ok = true;
    
    double value = m->toDouble(&ok);

    if(ok)
      m_instrument.T::noteOff(value);

    // One second decay. Hopefully ok in all cases...
    m_untilEnd = s->dspSamplingRate();

    return ok;
  }
  else if(!strncmp(id, "controller_", 11)) {

    int parameter = atoi(id + 11);
    
    bool ok = true;
    
    double value = m->toDouble(&ok);
    
    MJ_trace2("%s # Setting %s to %lf (ok = %d)", fname, id, value, (int) ok);

    if(ok)
      m_instrument.T::controlChange(parameter, value);
    
    return ok;
  }
  else if(!strcmp(id, MJ_Strs::pitch)) {

    bool ok = true;
    
    double value = m->toDouble(&ok);

    if(ok)
      m_instrument.T::setFrequency(value);
    
    return ok;
  }
  else if(!strcmp(id, "noteoffnow")) {
    m_untilEnd = 0;
    m_status = MJ_SVS_INACTIVE;
  }
  else 
    MJ_error2(MJ_ERR_ARGUMENT, "%s # Bad message id = \"%s\"", fname, id);
  
  return false;
}


template <class T>
MJ_SynthVoiceStatus MJ_StkVoice<T>::status()
{
  return m_status;
}

/* template <class T>
MJ_SynthVoice * MJ_StkVoice<T>::clone() const
{
  return new MJ_StkVoice;
  }*/

/// Specialize to override
template <class T>
bool MJ_StkVoice<T>::noteOn(MJ_MessageMulti *mm, MJ_ModuleSynth *s)
{
  MY_FLOAT frequency = mm->getDouble(MJ_Strs::pitch, 220.0);
  MY_FLOAT amplitude = mm->getDouble(MJ_Strs::gain, 1.0);
    
  m_instrument.T::noteOn(frequency, amplitude);

  return true;
}

template <class T>
MJ_SynthVoice * MJ_StkVoice<T>::create()
{
  return new MJ_StkVoice;
}

/// Specialize this to a particular special parameter.
template <class T>
bool MJ_StkVoice<T>::internalMsgProcess
(const char *, MJ_Message *, MJ_ModuleSynth *)
{
  return false;
}

#endif
