/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */


#include <mj_module_vbap.h>

#include <mj_arraytmpl.C>
#include <mj_message_string.h>
#include <mj_messages.h>
#include <mj_parameters.h>
#include <mj_strs.h>

#include <qfiledialog.h> 
#include <qfileinfo.h> 

#include <stdlib.h>
#include <string.h>
#include <time.h>

bool MJ_debugVBAP = false;

MJ_ModuleVBAP::MJ_ModuleVBAP()
  : m_vbap(0)
{
  m_panners.resize(1);
}

MJ_ModuleVBAP::~MJ_ModuleVBAP()
{
  delete m_vbap;
}
  
bool MJ_ModuleVBAP::dspAcceptIoChans(int *, int * out)
{
  trace3(MJ_debugVBAP, "MJ_ModuleVBAP::dspAcceptIoChans # ENTRY %d", *out );

  loadSetup();

  if(!m_vbap) {
    return false;
  }

  trace3(MJ_debugVBAP, "MJ_ModuleVBAP::dspAcceptIoChans # %d", *out );

  if(*out != (int) m_vbap->nofLoudspeakers()) {
    *out = m_vbap->nofLoudspeakers();
    trace3(MJ_debugVBAP, "MJ_ModuleVBAP::dspAcceptIoChans # %d", *out); 
    return true;
  }
  return false;
}

bool MJ_ModuleVBAP::dspPrepare()
{
  const char * fname = "MJ_ModuleVBAP::dspPrepare";

  trace3(MJ_debugVBAP, fname);

  if(!m_vbap) {
    error(MJ_ERR_UNKNOWN, fname, "NULL VBAP-setup");
    return false;
  }

  if((uint) dspOutChannels() != (uint) m_vbap->nofLoudspeakers()) {
    error2(MJ_ERR_UNKNOWN, "%s # Bad number of channels: %d != %d",
	   fname, dspOutChannels(), m_vbap->nofLoudspeakers());
    return false;
  }
  uint chansIn = dspInChannels();

  if(!chansIn) {
    error(MJ_ERR_UNKNOWN, fname, "No input");
    return false;
  }
  
  if(m_panners.size() != chansIn) {
    trace3(MJ_debugVBAP, "%s # Currently %d channels", fname, chansIn);
    m_panners.resize(chansIn);
  }
  
  initPanners();

  if(MJ_debugVBAP) toCout();
  
  return true;
}

void MJ_ModuleVBAP::dspProcessPeriod()
{
  // dspPrepare();

  dspClearOutputs();

  uint samples = m_signalInput.getFast(0).samples().size();

  for(uint i=0; i < m_panners.size(); i++) {
    m_panners.getFast(i).process(m_signalInput.getFast(i).samples().data(),
				 samples);
  }
}

bool MJ_ModuleVBAP::dspStop()
{
  if(MJ_debugVBAP) toCout();

  return true;
}

bool MJ_ModuleVBAP::msgProcess(const char *id, MJ_Message *m)
{
  trace3(MJ_debugVBAP, "MJ_ModuleVBAP::msgProcess # \"%s\"", id);

  if(!strcmp(id, MJ_Strs::filename)) {
    MJ_MessageString *ms = dynamic_cast<MJ_MessageString *> (m);

    if(!ms) return false;
    if(ms->stringConst().isEmpty()) return false;
    
    m_fileName = ms->stringConst().copy();

    loadSetup();
    
    return true;
  }

  /* Now it is either azimuth or elevation command: 
     
  azi_1, ele_12 etc...
  */

  int slen = strlen(id);

  bool ok  = true;
  
  double v = m->toDouble(&ok);

  if(slen <= 4 || !ok) return false;
  if(id[3] != '_') return false;

  uint index = (uint) atoi(&id[4]);

  if(index >= m_panners.size()) {
    uint chans = m_signalInput.size();
    if(index < chans)
      m_panners.resize(chans);
    else 
      return false;
  }

  if(!strncmp(id, MJ_Strs::azi, 3)) {
    
    v = MJ_clamp(v, -180.0, 180.0);

    trace3(MJ_debugVBAP, "MJ_ModuleVBAP::msgProcess # azi[%d] = %f", index, v);

    m_panners.getFast(index).setAzi(v);
  }
  else if(!strncmp(id, MJ_Strs::ele, 3)) {

    v = MJ_clamp(v, -90.0, 90.0);    
    trace3(MJ_debugVBAP, "MJ_ModuleVBAP::msgProcess # ele[%d] = %f", index, v);
    m_panners.getFast(index).setEle(v);
  }
  else 
    return false;

  // Not in real-time yet.
  if(!m_currentTime) return true;

  long ts = m->timeStamp();
  ulong currTs = m_currentTime->stamp();
  uint interpolateSamples;
  
  if(ts == -2) // Move panners immediately to the target
    interpolateSamples = 1;
  else if(ts <= 0 || (ulong) ts <= currTs)
    interpolateSamples = dspSamplingRate() / 50;
  else {
    interpolateSamples = ts - currTs;
  }

  // Do not accept too long interpolations:
  if(interpolateSamples > 100000) interpolateSamples = 100000;

  updatePanner(index, interpolateSamples);

  /* trace2("MJ_ModuleVBAP::msgProcess # voices = %d, interp = %d ", 
     voicesInUse(), (int) interpolateSamples); */

  /* trace2("MJ_ModuleVBAP::msgProcess # channel %d, azi = %f, ele = %f ", 
	 index, m_panners.getFast(index).azi(), 
	 m_panners.getFast(index).ele());*/
  
  return true;
}

MJ_Parameter * MJ_ModuleVBAP::msgGetInput(const char *id)
{
  if(!strcmp(id, MJ_Strs::filename)) {
    return new MJ_ParameterFilename(m_fileName, 
				    "Setup file", 
				    MJ_Strs::filename,
				    "*.vbap",
				    QFileDialog::ExistingFile);
  }

  int slen = strlen(id);

  if(slen <= 4) return 0;
  if(id[3] != '_') return 0;

  uint index = (uint) atoi(&id[4]);

  if(index >= m_panners.size()) return 0;

  MJ_VBAPVoiceGroup *vvg = & m_panners.getFast(index);

  if(!strncmp(id, MJ_Strs::azi, 3)) {
    return new MJ_ParameterDouble
      ("Azimuth", id, vvg->azi(), -180, 180, MJ_UNIT_ANGLE_DEGREES);
  }
  if(!strncmp(id, MJ_Strs::ele, 3)) {
    return new MJ_ParameterDouble
      ("Elevation", id, vvg->ele(), -90, 90, MJ_UNIT_ANGLE_DEGREES);
  }

  return 0;
}

QCString MJ_ModuleVBAP::msgGetInputName(int i)
{
  QCString r;
  
  if(i == 0) r = MJ_Strs::filename;
  else {
    uint shifted = (uint) i-1;
    uint voice = shifted / 2;
    
    if(voice < m_panners.size()) {

      const char * paramName;

      if(shifted & 1)
	paramName = MJ_Strs::ele;
      else 
	paramName = MJ_Strs::azi;

      r.sprintf("%s_%d", paramName, voice);
      trace3(MJ_debugVBAP, "MJ_ModuleVBAP::msgGetInputName# %s", 
	     (const char *) r);
    }
  }
  return r;
}

void MJ_ModuleVBAP::initPanners()
{
  if(!m_vbap) return;

  double g[3];
  int ls[3];
  uint dim = m_vbap->dimension;

  MJ_Sample *targets[3];

  for(uint i = 0; i < m_panners.size(); i++) {
    MJ_VBAPVoiceGroup *vvg = & m_panners.getFast(i);

    float azi = vvg->azi();
    float ele = vvg->ele();
    
    VBAPGains(g, ls, azi, ele, dim);

    for(uint j=0; j < dim; j++) {
      targets[j] = m_signalOutput.get(ls[j]).samples().data();
    }
    
    vvg->start();
    vvg->newTarget(azi, ele, 1, dim, 
		   g,
		   targets);
  }
}

void MJ_ModuleVBAP::updatePanner(uint i, uint interpolateSamples)
{
  if(!m_vbap) return;

  double g[3];
  int ls[3];
  uint dim = m_vbap->dimension;

  MJ_Sample *targets[3];

  MJ_VBAPVoiceGroup *vvg = & m_panners.getFast(i);

  float azi = vvg->azi();
  float ele = vvg->ele();
  
  VBAPGains(g, ls, azi, ele, dim);
  
  for(uint j=0; j < dim; j++) {
    targets[j] = m_signalOutput.get(ls[j]).samples().data();
  }
  
  // vvg->start();
  vvg->newTarget(azi, ele, interpolateSamples, dim, g, targets);
}

void MJ_ModuleVBAP::angleToCart(float azi, float ele, double *res)
{
  const double atorad = MJ_2pi / 360.0;

  res[0] = (float) (cos((double) (azi * atorad)) * 
		    cos((double)  (ele * atorad)));
  res[1] = (float) (sin((double) (azi * atorad)) * 
		    cos((double) (ele * atorad)));
  res[2] = (float) (sin((double) (ele * atorad)));
}

void MJ_ModuleVBAP::VBAPGains
(double g[3], int ls[3], float azi, float ele, uint dim)
{

  // const char * fname = "MJ_ModuleVBAP::VBAPGains";

  /* calculates gain factors using loudspeaker setup and given direction */
  double cartdir[3];
  double power;
  uint i,j,k, triplets;
  double small_g;
  double big_sm_g, gtmp[3];
  // int winner_triplet;
  
  angleToCart(azi,ele, &cartdir[0]);  

  big_sm_g = -100000.0;

  triplets = m_vbap->nofTriplets();

  // trace3(MJ_debugVBAP, "%s # %d triplets", fname, triplets);

  for(i=0; i < triplets; i++){

    small_g = 10000000.0;

    for(j = 0; j < dim; j++){
      gtmp[j] = 0.0;
      for(k = 0; k < dim; k++)
 	gtmp[j] += cartdir[k] * m_vbap->lsm[i][k + j * dim]; 
      if(gtmp[j] < small_g)
	small_g = gtmp[j];
    }

    /* trace3(MJ_debugVBAP, "%s # triplet %d: g = %f (big_g = %f)",
       fname, i, small_g, big_sm_g); */


    if(small_g > big_sm_g){
      
      big_sm_g = small_g;
      // winner_triplet=i;
      g[0] = gtmp[0]; 
      g[1] = gtmp[1]; 
      ls[0] = m_vbap->lstripl[i][0]; 
      ls[1] = m_vbap->lstripl[i][1]; 
      if(dim == 3){
	g[2] = gtmp[2];
	ls[2] = m_vbap->lstripl[i][2];
      } else {
	g[2] = 0.0;
	ls[2] = 0;
      }
    }
  }

  ls[0]--; // Range between zero and n.
  ls[1]--;
  ls[2]--;

  power=sqrt(g[0]*g[0] + g[1]*g[1] + g[2]*g[2]);
  /*  power=g[0]+g[1];*/
  
  g[0] /= power; 
  g[1] /= power;
  g[2] /= power;

  trace3(MJ_debugVBAP,
	 "VBAP azi = %f, ele = %f, dim = %d, speakers = [%d %d %d] gains = [%f  %f  %f]\n", 
	 azi, ele, dim, ls[0], ls[1], ls[2], g[0], g[1], g[2]);
  
}

bool MJ_ModuleVBAP::loadSetup()
{
  const char * fname = "MJ_ModuleVBAP::loadSetup";

  trace3(MJ_debugVBAP, "%s # \"%s\"", fname, (const char *) m_fileName);
  
  QFileInfo fi(m_fileName);

  if(!fi.isReadable() || !fi.exists()) {
    error2(MJ_ERR_UNKNOWN,
	   "%s # Cannot read file \"%s\"", fname,
	   (const char *) m_fileName);
    return false;
  }

  /*if(m_vbap && !m_vbapLoadInfo.needsLoading(fi)) {
    // trace("MJ_ModuleVBAP::loadSetup # already loaded");
    return true;
    }*/
  delete m_vbap;

  bool ok = true;
  
  m_vbap = new VBAP_Setup((char *) (const char *) m_fileName, ok);

  if(!ok) {
    error2(MJ_ERR_UNKNOWN,
	   "%s # Could not interpret file \"%s\"", fname,
	   (const char *) m_fileName);

    delete m_vbap;
    m_vbap = 0;
    return false;
  }
  else {
    m_vbapLoadInfo.nowLoaded(m_fileName);
    return true;
  }
}

/** Returns the total number of voices in use. */

uint MJ_ModuleVBAP::voicesInUse()
{
  uint sum =0;

  for(uint i=0; i < m_panners.size(); i++) {
    sum += m_panners.getFast(i).activeVoices();
  }

  return sum;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

template class MJ_Array<MJ_VBAPVoiceGroup>;
