/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the General Public License (GPL),
 * versions 2. Extra clauses can be found from the file MUSTAJUURI_GPL
 * and further explanations from the file MUSTAJUURI_LICENSE. The
 * license text can be found in file "GPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *  
 */

#include <mj_module_vbap2.h>

#include <mj_arraytmpl.C>
#include <mj_parameters.h>
#include <mj_strs.h>

#include <stdlib.h>

static const double MAX_VBAP_DELAY = 0.15;

MJ_ModuleVBAP2::MJ_ModuleVBAP2()
{}

MJ_ModuleVBAP2::~MJ_ModuleVBAP2()
{}

bool MJ_ModuleVBAP2::dspPrepare()
{
  bool ok = MJ_ModuleVBAP::dspPrepare();

  if(!ok) return false;
  
  uint n = m_panners.size();

  m_preprocessors.resize(n);
  
  float sr = dspSamplingRate();

  for(uint i=0; i < n; i++) {
    m_preprocessors.getFast(i).prepare(MAX_VBAP_DELAY, sr);
  }

  uint samples = m_signalInput.getFast(0).samples().size();
  m_tmpBuffer.resize(samples);

  return true;
}

void MJ_ModuleVBAP2::dspProcessPeriod()
{
  // return;
  
  dspClearOutputs();

  uint samples = m_signalInput.getFast(0).samples().size();

  for(uint i=0; i < m_panners.size(); i++) {
    
    // m_tmpBuffer = m_signalInput.getFast(i).samples();
    
    m_preprocessors.getFast(i).line().process
      (m_tmpBuffer.data(), m_signalInput.getFast(i).samples().data(), samples);

    m_panners.getFast(i).process(m_tmpBuffer.data(),
				 samples);
  }
}

bool MJ_ModuleVBAP2::dspStop()
{
  return MJ_ModuleVBAP::dspStop();

  // return true;
}

bool MJ_ModuleVBAP2::msgProcess(const char *id, MJ_Message *m)
{
  trace3(MJ_debugVBAP, "MJ_ModuleVBAP2::msgProcess # \"%s\"", id);

  int slen = strlen(id);

  if(slen <= 4) return false;

  if(!strncmp(id, MJ_Strs::dis, 3)) {
    bool ok  = true;
  
    double v = m->toDouble(&ok);

    if(!ok) 
      return false;

    if(id[3] != '_') return false;
  
    uint index = (uint) atoi(&id[4]);

    if(index >= m_preprocessors.size()) {
      uint chans = m_signalInput.size();
      if(index < chans)
	m_preprocessors.resize(chans);
      else 
	return false;
    }
    v = MJ_clamp(v, 0.0, 1e+5);

    float sr = (float) dspSamplingRate();
      
    if(!m_currentTime) {
      m_preprocessors.getFast(index).setDelay(v, sr);
    }
    else {

      long ts = m->timeStamp();
      ulong currTs = m_currentTime->stamp();
      uint interpolateSamples;

      if(ts == -2) // Move panners immediately to the target
	interpolateSamples = 1;
      else if(ts <= 0 || (ulong) ts <= currTs) {
	interpolateSamples = (uint) (sr * 0.1);
      }
      else {
	interpolateSamples = ts - currTs;
      }
      
      // Do not accept too long interpolations:
      if(interpolateSamples > 100000) interpolateSamples = 100000;

      /* trace2("MJ_ModuleVBAP2::msgProcess # interpolateSamples = %d %f", 
	 interpolateSamples, v); */
      
      m_preprocessors.getFast(index).interpolateDelay
	(v, sr, interpolateSamples);
    }
  }
  else if(!strncmp(id, MJ_Strs::gain, 4)) {
    
    bool ok  = true;
  
    double v = m->toDouble(&ok);

    if(!ok) 
      return false;

    if(id[4] != '_') return false;
  
    uint index = (uint) atoi(&id[5]);

    if(index >= m_preprocessors.size()) {
      uint chans = m_signalInput.size();
      if(index < chans)
	m_preprocessors.resize(chans);
      else 
	return false;
    }
    v = MJ_clamp(v, 0.0, 1e+3);

    if(!m_currentTime) {
      m_preprocessors.getFast(index).setGain(v);
    }
    else {

      long ts = m->timeStamp();
      ulong currTs = m_currentTime->stamp();
      uint interpolateSamples;

      if(ts == -2) // Move panners immediately to the target
	interpolateSamples = 1;
      else if(ts <= 0 || (ulong) ts <= currTs) {
	interpolateSamples = (uint) ((float) dspSamplingRate() * 0.1);
      }
      else {
	interpolateSamples = ts - currTs;
      }
      
      // Do not accept too long interpolations:
      if(interpolateSamples > 100000) interpolateSamples = 100000;

      /* trace2("MJ_ModuleVBAP2::msgProcess # interpolateSamples = %d %f", 
	 interpolateSamples, v); */
      
      m_preprocessors.getFast(index).interpolateGain
	(v, interpolateSamples);
    }
  }
  else {
    return MJ_ModuleVBAP::msgProcess(id, m);
  }

  // trace3(MJ_debugVBAP, "MJ_ModuleVBAP2::msgProcess # EXIT");

  return true;
}

MJ_Parameter * MJ_ModuleVBAP2::msgGetInput(const char *id)
{
  trace3(MJ_debugVBAP, "MJ_ModuleVBAP2::msgGetInput # \"%s\"", id);

  int slen = strlen(id);

  if(slen <= 4) return 0;
  
  if(!strncmp(id, MJ_Strs::dis, 3)) {
    
    if(id[3] != '_') return 0;

    uint index = (uint) atoi(&id[4]);

    if(index >= m_preprocessors.size()) {
      if(index > m_panners.size())
	return 0;
      m_preprocessors.resize(m_panners.size());
    }

    return new MJ_ParameterDouble
      ("Delay time", id, m_preprocessors.getFast(index).delayTime(), 
       0.0, MAX_VBAP_DELAY, MJ_UNIT_SECONDS);
  }
  else if(!strncmp(id, MJ_Strs::gain, 4)) {
    
    if(id[4] != '_') return 0;

    uint index = (uint) atoi(&id[5]);

    if(index >= m_preprocessors.size()) return 0;

    return new MJ_ParameterDouble
      ("Gain", id, m_preprocessors.getFast(index).gainTarget(), 
       0.0, 1.0);
  }

  return MJ_ModuleVBAP::msgGetInput(id);
}

QCString MJ_ModuleVBAP2::msgGetInputName(int i)
{
  QCString r;
  
  if(i == 0) r = MJ_Strs::filename;
  else {
    uint shifted = (uint) i-1;
    uint voice = shifted / 4;
    
    if(voice < m_panners.size()) {

      const char * paramName;

      uint tmp = shifted % 4;

      if(tmp == 0)
	paramName = MJ_Strs::azi;
      else if(tmp == 1)
	paramName = MJ_Strs::ele;
      else if(tmp == 2)
	paramName = MJ_Strs::dis;
      else
	paramName = MJ_Strs::gain;

      r.sprintf("%s_%d", paramName, voice);
      trace3(MJ_debugVBAP, "MJ_ModuleVBAP::msgGetInputName# %s", 
	     (const char *) r);
    }
  }

  // trace2("MJ_ModuleVBAP2::msgGetInputName # %d \"%s\"", i, (const char *) r);

  return r;
}


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


template class MJ_Array<MJ_ModuleVBAP2::Preprocess>;
