/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef QAWL_COLOR_LIST_H
#define QAWL_COLOR_LIST_H

#include <qcolor.h>

#include <list>

class QAWL_ColorList
{
public:
  QAWL_ColorList() : m_next(m_colors.begin()) {}
  ~QAWL_ColorList() {}

  /// Possible rounding modes
  enum Rounding { Closest, Floor, Ceil };

  class Item
  {
  public:
    
    Item(const QColor &bg, const QColor &fg, 
	 long beginFrame, Rounding rounding = Closest)
      : m_bg(bg), m_fg(fg), m_beginFrame(beginFrame), m_rounding(rounding) {}

    long beginFrame() const { return m_beginFrame; }
    
    inline int toPixels(long visualBeginFrame, double framesPerPixel) const;

    QColor & background() { return m_bg; }
    QColor & foreground() { return m_fg; }

    QColor m_bg, m_fg;
    long   m_beginFrame;
    Rounding m_rounding;
  };

  void clear() { m_colors.clear(); rewind(); }

  void rewind() { m_next = m_colors.begin(); }
  void advance() { ++m_next; }

  bool atEnd() /* const*/ { return m_next == m_colors.end(); }
  
  Item & currentItem() { return *m_next; }

  void seekToItem(long frame)
  {
    rewind();
    while(!atEnd() && (*m_next).beginFrame() < frame) {
      std::list<Item>::iterator tmp(m_next);
      tmp++;
      if(tmp == m_colors.end() || (*tmp).beginFrame() >= frame)
	return;
      m_next++;
    }
  }

  void seekToItem(long visualBeginFrame, double framesPerPixel)
  {
    rewind();
    while(!atEnd()) {
      std::list<Item>::iterator tmp(m_next);
      tmp++;
      if(tmp == m_colors.end() || 
	 (*tmp).toPixels(visualBeginFrame, framesPerPixel) > 0)
	return;
      m_next++;
    }
  }

  int nextItemPixels(long visualBeginFrame, double framesPerPixel)
  {
    std::list<Item>::iterator tmp(m_next);
    tmp++;
    if(tmp == m_colors.end()) return 1000000;
    return (*tmp).toPixels(visualBeginFrame, framesPerPixel);
  }

  ulong nextItemFrames()
  {
    std::list<Item>::iterator tmp(m_next);
    tmp++;
    if(tmp == m_colors.end()) return (ulong) -1;
    return (*tmp).m_beginFrame;
  }

  void appendItem(const QColor &bg, const QColor &fg, 
		  long frame, Rounding rounding = Closest)
  {
    Item tmp(bg, fg, frame, rounding);
    m_colors.push_back(tmp);
  }

  uint size() { return m_colors.size(); }

  QAWL_ColorList & operator = (const QAWL_ColorList &that)
  { m_colors = that.m_colors; rewind(); return * this; }

protected:

  std::list<Item> m_colors;
  std::list<Item>::iterator m_next;
};

inline int QAWL_ColorList::Item::toPixels
(long visualBeginFrame, double framesPerPixel) const
{
  double tmp = (m_beginFrame - visualBeginFrame) / framesPerPixel;
  
  if(m_rounding == Closest)
    return (int) (tmp + 0.5);
  else if(m_rounding == Floor)
    return (int) tmp;
  else
    return (int) (tmp + 0.99999);
}

#endif
