/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <qawl_file_cache.h>

#include <qawl_preview_calculator.h>

#include <qfileinfo.h>

#define QAWL_BLOCK 3000

QAWL_FileCache::QAWL_FileCache()
  : m_firstFrame(0),
    m_lastFrame(0),
    m_chans(0),
    m_discardAfter(200)
{}

QAWL_FileCache::~QAWL_FileCache()
{
  qDebug("QAWL_FileCache::~QAWL_FileCache # %u %u", 
	 m_data.size(), m_tmpBuf.size());
}
  
bool QAWL_FileCache::setFile(const QString &f, uint scale)
{
  qDebug("QAWL_FileCache::setFile # \"%s\" %u", f.ascii(), scale);
  
  m_chans = 0;
  m_firstFrame = 0;
  m_lastFrame = 0;
  m_fileFrames = 0;

  m_fileName = QAWL_PreviewCalculator::previewFileName(f, scale);

  QFileInfo fi(m_fileName);

  if(!fi.exists()) {
    qDebug("QAWL_FileCache::setFile # Preview file does not exist");
    return false;
  }

  m_scale = scale;

  bool ok = m_infile.open(m_fileName.ascii(), 
			  DI_IO_INPUT, DI_ASF_INT32, QAWL_BLOCK);
  
  if(!ok) 
    return false;

  m_chans = m_infile.inChannels();
  m_fileFrames = m_infile.frameCount();
  
  m_tmpBuf.resize(QAWL_BLOCK * m_chans);

  m_data.resize(0);

  return true;
}

bool QAWL_FileCache::loadFrames(ulong first, ulong last)
{
  /* qDebug("QAWL_FileCache::loadFrames # [%lu %lu] -> [%lu %lu]", 
     m_firstFrame, m_lastFrame,  first, last); */

  last++;

  if(last > m_lastFrame && last < (m_lastFrame + m_discardAfter))
    last += m_discardAfter;

  if(last > m_fileFrames)
    last = m_fileFrames;

  if(first >= m_fileFrames)
    return false;

  if(first >= m_firstFrame && last <= m_lastFrame) {
    return true;
  }

  if(last < m_firstFrame || first >= m_lastFrame) {
    return fullReload(first, last);
  }
  
  bool loop = true, ok = true;

  // Scan frames from before the current cache.
  if(first < m_firstFrame) {

    if(first > m_discardAfter)
      first -= m_discardAfter;
    else
      first = 0;

    ulong newFrames = m_firstFrame - first;

    ulong newSamples = newFrames * m_chans;
    ulong dropSamples = (last + m_discardAfter) < m_lastFrame ? 
      (m_lastFrame - last) * m_chans : 0;

    m_data.resize(m_data.size() + newSamples - dropSamples);
    QAWL_MinMax *mm = & * m_data.begin();
    QAWL_MinMax *mmend = mm + newSamples;

    memmove(mmend, mm, 
	    (m_data.size() - newSamples) * sizeof(QAWL_MinMax));
    
    m_infile.seekFrame((long) first);

    while(loop) {
      
      ulong needFrames = m_firstFrame - m_infile.currentFrame();
      ulong readFrames = needFrames > QAWL_BLOCK ? QAWL_BLOCK : needFrames;
      
      if(readFrames == 0) {
	loop = false;
      }
      else {
	ulong tmp = m_infile.readFrames(& * m_tmpBuf.begin(), readFrames);

	if(tmp != readFrames) {
	  loop = ok = false;
	  qDebug("QAWL_FileCache::loadFrames # Read error");
	}

	tmp *= m_chans;
	
	for(uint i=0; i < tmp; i++) {

	  if(m_scale != 1)
	    mm->setInt32(m_tmpBuf[i]);
	  else
	    mm->setInt16(m_tmpBuf[i] >> 16);
	  // mm->setInt32(m_tmpBuf[i]);
	  mm++;
	}
      }
    }
    if(mm != mmend)
      qDebug("QAWL_FileCache::loadFrames # mm failure %p != %p (%d)",
	     mm, mmend, (int) (mmend - mm));
  }
  
  // Scan frames from after the current cache.
  if(last > m_lastFrame) {

    if(first > (m_firstFrame + m_discardAfter)) {
      // Drop samples from the beginning
      ulong dropSamples = (first - m_firstFrame) * m_chans;

      memmove( &* m_data.begin(),  &* m_data.begin() + dropSamples, 
	       (m_data.size() - dropSamples) * sizeof(QAWL_MinMax));

      m_data.resize(m_data.size() - dropSamples);

      // qDebug("Dropped %lu frames from the beginning", dropSamples / m_chans);
    }

    ulong newFrames = last - m_lastFrame /* - 1*/;

    // qDebug("Loading %lu new frames to the end", newFrames);

    ulong newSamples = newFrames * m_chans;
    ulong oldSize = m_data.size();
    
    m_data.resize(oldSize + newSamples);
    QAWL_MinMax *mm =  &* m_data.begin() + oldSize;
    
    // memmove(mm + newSamples, mm, newSamples * sizeof(QAWL_MinMax));
    
    m_infile.seekFrame((long) m_lastFrame);

    loop = true;

    while(loop) {
      
      ulong needFrames = last - m_infile.currentFrame();
      ulong readFrames = needFrames > QAWL_BLOCK ? QAWL_BLOCK : needFrames;
      
      if(readFrames == 0) {
	loop = false;
      }
      else {
	ulong tmp = m_infile.readFrames(& * m_tmpBuf.begin(), readFrames);

	if(tmp != readFrames) {
	  loop = ok = false;
	  qDebug("Read error 2");
	}

	tmp *= m_chans;
	
	for(uint i=0; i < tmp; i++) {

	if(m_scale != 1)
	  mm->setInt32(m_tmpBuf[i]);
	else
	  mm->setInt16(m_tmpBuf[i] >> 16);

	// mm->setInt32(m_tmpBuf[i]);

	  /*if(i < 2 || i+2 >= tmp)
	    qDebug("%u:nth loaded is [%hd %hd]", i, mm->min, mm->max); */

	  mm++;
	}
      }      
    }

    // Sanity check:
    if(mm != &*m_data.end())
      qDebug("We did not fill the buffer exactly %p != %p (%ld bytes)", 
	     mm, &*m_data.end(), (long) &*m_data.end() - (long) mm);
  }

  if(ok) {
    m_firstFrame = first;
    m_lastFrame = last;
  }

  return ok;
}

QAWL_MinMax QAWL_FileCache::getMinMax(ulong firstRealFrame, 
				      ulong lastRealFrame, 
				      uint channel)
{
  ulong scaledFirst =  firstRealFrame / m_scale;
  ulong scaledLast = lastRealFrame / m_scale;

  if(scaledFirst < m_firstFrame /* || scaledFirst > m_lastFrame*/ ) {

    qDebug("QAWL_FileCache::getMinMax # Bad argument");

    return QAWL_MinMax(0, 0);
  }

  if(scaledLast > m_fileFrames - 1) {
    scaledLast = m_fileFrames - 1;
  }

  if(scaledFirst >= m_fileFrames)
    return QAWL_MinMax(0, 0);

  if(scaledLast >= m_lastFrame && m_lastFrame < m_fileFrames) {
    
    // qDebug("QAWL_FileCache::getMinMax # Loading more frames");
    
    /// Load more frames from the file:
    loadFrames(m_firstFrame, scaledLast + m_discardAfter);

    if(scaledLast > m_lastFrame) scaledLast = m_lastFrame; 

  }

  scaledFirst -= m_firstFrame;
  scaledLast  -= m_firstFrame;

  QAWL_MinMax &mm = getSample(scaledFirst, channel);
  
  short mi = mm.min;
  short ma = mm.max;

  for(ulong i = scaledFirst + 1; i <= scaledLast; i++) {
    QAWL_MinMax mm2 = getSample(i, channel);
    
    if(mi > mm.min) mi = mm2.min;
    if(ma < mm.max) ma = mm2.max;

    if(i >= m_data.size())
      qDebug("QAWL_FileCache::getMinMax # Array exceeded %d >= %d (%d)",
	     (int) i, (int) m_data.size(), (int) m_lastFrame);
  }

  return QAWL_MinMax(mi, ma);
}

bool QAWL_FileCache::fullReload(ulong first, ulong last)
{
  // qDebug("QAWL_FileCache::fullReload");

  ulong frames = last - first;
  ulong samples = frames * m_chans;

  m_data.resize(samples);

  m_infile.seekFrame((long) first);
  
  QAWL_MinMax *mm =  &* m_data.begin();

  bool loop = true, ok = true;

  uint count = 0;

  while(loop) {
      
    ulong needFrames = last - m_infile.currentFrame();
    ulong readFrames = needFrames > QAWL_BLOCK ? QAWL_BLOCK : needFrames;
      
    if(readFrames == 0) {
      loop = false;
    }
    else {
      ulong tmp = m_infile.readFrames(& * m_tmpBuf.begin(), readFrames);

      if(tmp != readFrames) {
	loop = ok = false;
	qDebug("QAWL_FileCache::fullReload # read error %lu != %lu", 
	       tmp, readFrames);
      }
      
      tmp *= m_chans;

      for(uint i=0; i < tmp; i++) {

	if(m_scale != 1)
	  mm->setInt32(m_tmpBuf[i]);
	else
	  mm->setInt16(m_tmpBuf[i] >> 16);

	mm++;

	count++;
	if(mm > & * m_data.end())
	  qDebug("QAWL_FileCache::fullReload # memory error %d %ld", 
		 count, tmp);
      }
    }      
  }

  if(ok) {
  
    m_firstFrame = first;
    m_lastFrame = last;
  }

  return true;
}

