/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <qawl_gl_renderer.h>

#include <qawl_color_list.h>
#include <qawl_file_cache.h>

#include <GL/gl.h>

void QAWL_GlRenderer::render(QAWL_FileCache *cache, 
			     ulong firstRealFrame, ulong lastRealFrame,
			     uint chan,
			     float width, 
			     float xresolution,
			     float height,
			     float clipHeight,
			     QAWL_ColorList * colors,
			     /** Collector that is used to store
				 rendering information. If this
				 argument is non-zero, then the
				 rendering data is stored into this
				 variable and not rendered directly.*/
			     std::vector<QuadData> * collector)
{
  /* qDebug("QAWL_Renderer::render # p = %p, c = %p "
	 "%lu %lu, [%d, %d], [%d, %d]",
	 painter, cache, 
	 firstRealFrame, lastRealFrame, left, right, centerY, height);
  */
  uint count = 0;

  uint quads = (uint) (width / xresolution);
  double framesPerQuad = (double) (lastRealFrame - firstRealFrame) / quads;

  double heightScale = (double) height / 32768.0;

  ulong nextColorFrames = (ulong) -1;
  
  QColor color(255, 255, 255);

  if(colors) {
    
    colors->seekToItem((long) firstRealFrame,framesPerQuad);

    //qDebug("QAWL_Renderer::render # %d color segments", (int)colors->size());

    if(!colors->atEnd()) {
      nextColorFrames = colors->nextItemFrames();

      QAWL_ColorList::Item & item = colors->currentItem();

      color = item.foreground();

      if(!collector)
	glColor3ub(color.red(), color.green(), color.blue());

      // qDebug("QAWL_Renderer::render # %x, %d", 
      // (int) item.background().rgb(), operateAt);
    }
    else
      qDebug("QAWL_Renderer::render # No visible color segments");
  }

  if(!collector)
    glBegin(GL_QUAD_STRIP);
  else
    collector->resize(quads + 1);  

  for(uint i=0; i <= quads; i++) {

    ulong startFrame = firstRealFrame + 
      (ulong) (i * framesPerQuad);

    ulong endFrame = firstRealFrame + 
      (ulong) ((i+1) * framesPerQuad);

    QAWL_MinMax range = cache->getMinMax(startFrame, endFrame, chan);

    range.limitZeroPassing();
    
    double low = range.min * heightScale;
    if(low < -clipHeight) low = -clipHeight;

    double high = range.max * heightScale;
    if(high > clipHeight) high = clipHeight;

    float x = i * xresolution;

    if(!collector) {
      glVertex2f(x, low);
      glVertex2f(x, high);
    }
    else {
      QuadData & vd = (*collector)[i];
      vd.m_x = x;
      vd.m_lowY  = low;
      vd.m_highY = high;
      vd.m_rgb[0] = color.red();
      vd.m_rgb[1] = color.green();
      vd.m_rgb[2] = color.blue();
    }

    count++;

    if(endFrame >= nextColorFrames) {
      colors->advance();

      nextColorFrames = colors->nextItemFrames();

      QAWL_ColorList::Item & item = colors->currentItem();

      color = item.foreground();

      if(!collector)
	glColor3ub(color.red(), color.green(), color.blue());
    }

  }
  
  if(!collector)
    glEnd();

}

void QAWL_GlRenderer::renderCollector(std::vector<QuadData> *collector)
{
  glBegin(GL_QUAD_STRIP);

  QuadData * qd = & * collector->begin();
  QuadData * sentinel = & * collector->end();

  for(; qd != sentinel; qd++) {
    // printf("QD = %p", qd); fflush(0);
    QuadData qdr = *qd;
    glVertex2f(qdr.m_x, qdr.m_lowY);
    glVertex2f(qdr.m_x, qdr.m_highY);
    glColor3ubv(qdr.m_rgb);
  }

  glEnd();
}

