/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <qawl_preview_calculator.h>

#include <qawl_min_max.h>

#include <qfileinfo.h> 

#include <algorithm>

// #define QAWL_BLOCK 150000
// #define QAWL_BLOCK 3000
#define QAWL_BLOCK 30000

QAWL_FilePreviewCalculus::~QAWL_FilePreviewCalculus()
{
  clear();
}

bool QAWL_FilePreviewCalculus::open(const QAWL_PreviewRequest &req)
{
  m_file = req.fileName();
  m_previewFile = req.previewFileName();
  m_scale = req.scale();
  m_sourceScale = req.sourceScale();


  qDebug("QAWL_FilePreviewCalculus::open \"%s\" -> \"%s\"", 
	 m_file.latin1(), m_previewFile.latin1());
  
  clear();

  m_infile = new DI_AudioFile;

  QString fileTmp(m_sourceScale == 1 ? m_file : 
		  QAWL_PreviewCalculator::previewFileName
		  (m_file, m_sourceScale));

  bool ok = m_infile->open(fileTmp, 
			   DI_IO_INPUT, 
			   m_sourceScale == 1 ? DI_ASF_INT16 : DI_ASF_INT32, 
			   QAWL_BLOCK);

  if(!ok) {
    qDebug("QAWL: Failed to open input file \"%s\"", (const char *) m_file);
    clear();
    return false;
  }

  if(m_sourceScale == 1) {
    m_dataBuffer.resize(m_infile->inChannels() * QAWL_BLOCK);
    m_dataBufferInt.resize(0);
  }
  else {
    m_dataBuffer.resize(0);
    m_dataBufferInt.resize(m_infile->inChannels() * QAWL_BLOCK);
  }

  m_scaledBuffer.resize(m_infile->inChannels() * 
			(QAWL_BLOCK / (m_scale / m_sourceScale) + 1));
  
  m_outfile = new DI_AudioFile;

  m_outfile->setFileFormat(DI_AFF_WAVE);
  m_outfile->setSampleRate(44100);
  m_outfile->setSampleWidth(32);
  m_outfile->setSampleFormat(DI_AFSF_TWOSCOMP);
  m_outfile->setByteOrder(DI_AFBO_LITTLEENDIAN);
  m_outfile->setChannels(m_infile->inChannels());
  
  ok = m_outfile->open(m_previewFile, 
		       DI_IO_OUTPUT, 
		       DI_ASF_INT32, 
		       m_scaledBuffer.size());
  
  return ok;
}

bool QAWL_FilePreviewCalculus::processFrames(uint frames)
{
  if(!m_infile || !m_outfile) {
    qDebug("QAWL_FilePreviewCalculus::processFrames # use \"open()\" first");
    return false;
  }

  long fileFrames = m_infile->frameCount();
  long avail = fileFrames - m_readFrames;

  if((long) frames > avail) {
    m_done = true;
    frames = avail;
  }

  uint scale = m_scale / m_sourceScale;

  uint channels = m_infile->inChannels();
  
  uint scaledFrames = frames / scale;

  frames = scaledFrames * scale;

  uint samples = frames * channels;

  uint readFrames;
  
  qDebug("QAWL_FilePreviewCalculus::processFrames %s %u %d %u", 
	 m_previewFile.latin1(), scale,
	 (int) m_readFrames, frames);

  if(m_sourceScale == 1) {
    qDebug("Reading from original file");

    if(m_dataBuffer.size() < samples)
      m_dataBuffer.resize(samples);

    readFrames = (uint) m_infile->readFrames(& m_dataBuffer[0], frames);
  }
  else {
    qDebug("Reading from previous preview file");
    
    if(m_dataBufferInt.size() < samples)
      m_dataBufferInt.resize(samples);

    readFrames = (uint) m_infile->readFrames(& m_dataBufferInt[0], frames);
  }

  if(readFrames != frames) {
    qDebug("ERROR: Failed to read all frames");
    return false;
  }

  m_readFrames += frames;

  if(m_sourceScale == 1) {
    for(uint channel = 0; channel < channels; channel++) {

      short * ptr = & m_dataBuffer[channel];

      for(uint i=0; i < scaledFrames; i++) {

	QAWL_MinMax mm(*ptr, *ptr);
	
	for(uint j=1; j < scale; j++) {
	  mm.expand(*ptr);
	  ptr += channels;
	}

	m_scaledBuffer[i * channels + channel] = mm.getInt32();
      }
    }
  }
  else {
    
    // qDebug("Scanning data");
    
    for(uint channel = 0; channel < channels; channel++) {

      const int * ptr = & m_dataBufferInt[channel];

      for(uint i=0; i < scaledFrames; i++) {

	QAWL_MinMax mm;

	mm.setInt32(*ptr);
	
	ptr += channels;

	for(uint j=1; j < scale; j++) {
	  mm.expandInt32(*ptr);
	  ptr += channels;
	}

	/* qDebug("Writing %d (%d) = 0x%x", (int) i * channels + channel, 
	       m_scaledBuffer.size(),
	       mm.getInt32());*/

	m_scaledBuffer[i * channels + channel] = mm.getInt32();
      }
    }
  }

  
  // qDebug("Writing data");

  m_outfile->writeFrames(& m_scaledBuffer[0], scaledFrames);

  if(m_done) {
    finalizeProcessing();

    delete m_outfile;
    m_outfile = 0;

    delete m_infile;
    m_infile = 0;
  }

  qDebug("Processed frames %u / %u", frames, scaledFrames);

  return true;
}

/// Handles the final frames in the file.

void QAWL_FilePreviewCalculus::finalizeProcessing()
{
  long fileFrames = m_infile->frameCount();
  long avail = fileFrames - m_readFrames;

  if(!avail) return;

  uint scale = m_scale / m_sourceScale;

  if((ulong) avail > scale) {
    qDebug("QAWL_FilePreviewCalculus::finalizeProcessing # avail > scale??");
    return;
  }

  uint channels = m_infile->inChannels();

  uint readFrames;
  
  if(m_sourceScale == 1)
    readFrames = (uint) m_infile->readFrames(& m_dataBuffer[0], avail);
  else
    readFrames = (uint) m_infile->readFrames(& m_dataBufferInt[0], avail);

  if(readFrames != (uint) avail) {
    qDebug("QAWL_FilePreviewCalculus::finalizeProcessing # "
	   "readFrames != avail");
    return;
  }
  
  if(m_sourceScale == 1) {
    for(uint channel = 0; channel < channels; channel++) {

      short * ptr = & m_dataBuffer[channel];
      
      QAWL_MinMax mm(*ptr, *ptr);
      
      for(long j=1; j < avail; j++) {
	mm.expand(*ptr);
	ptr += channels;
      }
      
      m_scaledBuffer[channel] = mm.getInt32();
    }
  }
  else {
    for(uint channel = 0; channel < channels; channel++) {

      int * ptr = & m_dataBufferInt[channel];

      QAWL_MinMax mm;

      mm.setInt32(*ptr);
      
      for(long j=1; j < avail; j++) {
	mm.expandInt32(*ptr);
	ptr += channels;
      }

      m_scaledBuffer[channel] = mm.getInt32();
    }
  }

  m_outfile->writeFrames(& m_scaledBuffer[0], 1);
}

void QAWL_FilePreviewCalculus::clear()
{
  if(m_infile) {
    delete m_infile;
    m_infile = 0;
  }

  if(m_outfile) {
    delete m_outfile;
    m_outfile = 0;
  }

  m_dataBuffer.resize(0);
  m_dataBufferInt.resize(0);
  m_scaledBuffer.resize(0);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

QAWL_PreviewCalculator * QAWL_PreviewCalculator::m_calculator = 0;

QAWL_PreviewCalculator::QAWL_PreviewCalculator()
{
  if(!m_calculator) m_calculator = this;

  connect(&m_timer, SIGNAL(timeout()), this, SLOT(update()));
}

QAWL_PreviewCalculator::~QAWL_PreviewCalculator()
{
  if(m_calculator == this) m_calculator = 0;
}

bool QAWL_PreviewCalculator::lockFile(const QString &file)
{
  return m_calculator->lockFileInternal(file);
}

bool QAWL_PreviewCalculator::unlockFile(const QString &file)
{
  return m_calculator->unlockFileInternal(file);
}
  
QString QAWL_PreviewCalculator::previewFileName
(const QString &file, uint scale)
{
  if(scale == 1) return file;

  QString scaleStr;
  scaleStr.setNum(scale);
  
  return file + "-" + scaleStr + "-previewwav";
}

bool QAWL_PreviewCalculator::requestPreview
(const QString &file, uint scale, uint sourceScale )
{
  /* qDebug("QAWL_PreviewCalculator::requestPreview \"%s\"  %d",
     file.latin1(), scale);*/

  QString previewFile(previewFileName(file, scale));

  /*QString realFile = sourceScale == 1 ? 
    file : previewFileName(file, sourceScale);*/
  
  QFileInfo fiOrig(file);
  QFileInfo fiPreview(previewFile);

  if(!fiOrig.exists()) {
    qDebug("QAWL: File \"%s\" does not exist (!)", file.latin1());
    return false;
  }

  if(fiPreview.exists()) {
    if(fiPreview.lastModified() > fiOrig.lastModified()) {
      emit previewAvailable(file, scale);
      return true;
    }
  }

  QAWL_PreviewRequest req(file, previewFile, scale, sourceScale);

  std::list<QAWL_PreviewRequest>::iterator it = 
    std::find(m_requests.begin(), m_requests.end(), req);
  //std::find(((const QAWL_PreviewCalculator *) this)->m_requests.begin(), 
  // ((const QAWL_PreviewCalculator *) this)->m_requests.end(), req);
    
    // m_requests.find(req);
  
  if(it == m_requests.end())
    m_requests.push_back(req);

  m_timer.start(10);
  
  // qDebug("QAWL_PreviewCalculator::requestPreview done");
  
  return true;
}

bool QAWL_PreviewCalculator::requestNormalPreviews(const QString &file)
{
  for(uint i=0; i < nofNormalPreviews(); i++) {
    if(!requestPreview(file, normalPreviewScale(i), 
		       i ? normalPreviewScale(i - 1) : 1))
      return false;
  }
  return true;
}


bool QAWL_PreviewCalculator::isPreviewReady(const QString &file, uint scale)
{
  QString previewFile(previewFileName(file, scale));

  /*QString realFile = sourceScale == 1 ? 
    file : previewFileName(file, sourceScale);*/
  
  QFileInfo fiOrig(file);
  QFileInfo fiPreview(previewFile);

  if(!fiOrig.exists()) {
    qDebug("QAWL: File \"%s\" does not exist (!)", file.latin1());
    return false;
  }

  if(fiPreview.exists() && !isLocked(previewFile)) {
    if(fiPreview.lastModified() > fiOrig.lastModified()) {
      return true;
    }
  }
  
  return false;
}

uint QAWL_PreviewCalculator::nofNormalPreviews()
{
  return 5;
}

uint QAWL_PreviewCalculator::normalPreviewScale(uint preview)
{
  uint scale = 10;
  
  for(uint i=0; i < preview; i++)
    scale *= 10;

  return scale;
}

uint QAWL_PreviewCalculator::determinePreviewScale(double scale)
{
  uint tmp = normalPreviewScale(0);

  if(scale < (double) tmp)
    return 1;

  uint tests = nofNormalPreviews();

  for(uint i=1; i < tests; i++) {

    tmp = normalPreviewScale(i);

    if(scale < (double) tmp)
      return normalPreviewScale(i - 1);    
  }
  
  return tmp;
}


long QAWL_PreviewCalculator::fileFrames(const QString &file)
{
  QFileInfo fi(file);
  if(!fi.exists()) return -1;

  DI_AudioFile af;
  
  bool ok = af.open(file.ascii(), DI_IO_INPUT, DI_ASF_INT16, 100);
  
  if(!ok) return -1;

  return af.frameCount();
}

QAWL_PreviewCalculator & QAWL_PreviewCalculator::instance()
{
  if(!m_calculator)
    new QAWL_PreviewCalculator;

  return *m_calculator;
}

void QAWL_PreviewCalculator::deleteInstance()
{
  delete m_calculator;
  m_calculator = 0;
}

void QAWL_PreviewCalculator::update()
{
  // qDebug("QAWL_PreviewCalculator::update");

  m_mutex.lock();

  QTime entryTime(QTime::currentTime());

  while(entryTime.msecsTo(QTime::currentTime()) < 50) {

    std::list<QAWL_FilePreviewCalculus>::iterator it = findCalculusObject();

    if(it == m_activePreviews.end()) {

      if(m_requests.size()) {
	std::list<QAWL_PreviewRequest>::iterator prit = m_requests.begin();
	
	m_activePreviews.push_back(QAWL_FilePreviewCalculus());
	
	std::list<QAWL_FilePreviewCalculus>::iterator fpcit = 
	  m_activePreviews.end();
	fpcit--;
	
	if((*fpcit).open((*prit))) {
	  qDebug("QAWL_PreviewCalculator::update # opened file \"%s\"",
		 (const char *) (*prit).fileName());
	}
	m_requests.erase(prit);
      }
      else {
	m_mutex.unlock();
	return;
      }
    }
    else {
      // Process the existing previews:

      /* if((*it).processedFrames() < 10)
	 (*it).processFrames(100); */

      (*it).processFrames(QAWL_BLOCK);

      if((*it).isDone()) {
	emit previewAvailable((*it).fileName(), (*it).scale());
	m_activePreviews.erase(it);
      }
    }
  }

  m_mutex.unlock();
}

bool QAWL_PreviewCalculator::lockFileInternal(const QString &file)
{
  m_mutex.lock();

  std::list<QAWL_FileLock>::iterator it = m_calculator->findLocked(file);
  
  if(it != m_lockedFiles.end()) {
    (*it).addLock();
  }
  else 
    m_lockedFiles.push_back(QAWL_FileLock(file));
  
  m_mutex.unlock();
  
  return true;
}

bool QAWL_PreviewCalculator::unlockFileInternal(const QString &file)
{
  bool ok;

  m_mutex.lock();

  std::list<QAWL_FileLock>::iterator it = m_calculator->findLocked(file);
  
  if(it != m_lockedFiles.end()) {
    (*it).removeLock();
    ok = true;
  }
  else 
    ok = false;
  
  m_mutex.unlock();
  
  return ok;
}

std::list<QAWL_FileLock>::iterator QAWL_PreviewCalculator::findLocked
(const QString &file)
{
  std::list<QAWL_FileLock>::iterator it = m_lockedFiles.begin();

  while(it != m_lockedFiles.end()) 
    if((*it).fileName() == file)
      return it;
  
  return m_lockedFiles.end();
}

std::list<QAWL_FilePreviewCalculus>::iterator 
QAWL_PreviewCalculator::findCalculusObject()
{
  std::list<QAWL_FilePreviewCalculus>::iterator it = m_activePreviews.begin();
  
  while(it != m_activePreviews.end()) {
    if(findLocked((*it).fileName()) == m_lockedFiles.end())
      return it;
  }

  return it;
}
