/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */


#ifndef QAWL_PREVIEW_CALCULATOR_H
#define QAWL_PREVIEW_CALCULATOR_H

#include <di_audio_file.h> 
#include <di_mutex.h> 

#include <qobject.h>
#include <qtimer.h>

#include <list>
#include <vector>

class QAWL_FileLock
{
public:
  QAWL_FileLock() : m_locks(0) {}
  QAWL_FileLock(const QString &file) : m_file(file), m_locks(1) {}

  int addLock() {return ++m_locks;}
  int removeLock() {return --m_locks;}
  int locks() const {return m_locks;}

  const QString & fileName() const { return m_file; }

protected:

  QString m_file;
  int    m_locks;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


class QAWL_PreviewRequest
{
public:
  QAWL_PreviewRequest(const QString &file, 
		      const QString &previewFile, 
		      uint scale,
		      uint sourceScale = 1)

    : m_file(file), m_previewFile(previewFile), m_scale(scale),
      m_sourceScale(sourceScale) {}
  
  const QString & fileName() const { return m_file; }
  const QString & previewFileName() const { return m_previewFile; }
  uint    scale()       const { return m_scale; }
  uint    sourceScale() const { return m_sourceScale; }

  bool operator == (const QAWL_PreviewRequest &r) const
  { return m_file == r.m_file && m_previewFile == r.m_previewFile &&
      m_scale == r.m_scale && m_sourceScale == r.m_sourceScale; }
  
  bool operator != (const QAWL_PreviewRequest &r) const
  { return ! ((*this) == r); }

protected:
  QString m_file, m_previewFile;
  uint    m_scale;
  uint    m_sourceScale;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** Calculate data of a sound file to a preview file. 

    <B>About the preview file format:</B> The data is stored to the
    file in pairs of short. The file itself is written with 32-bit
    ints. The minimum and maximum values of one sample interval are
    packed to one 32-bit sample. The 32-bit sample breaks down to two
    16-bit integers. The 16 MSB are the maximum value and 16 LSB are
    the minimum value.
*/
class QAWL_FilePreviewCalculus
{
public:
  QAWL_FilePreviewCalculus() : m_infile(0), m_outfile(0), m_readFrames(0), 
			       m_done(false) {}
  
  ~QAWL_FilePreviewCalculus();

  bool open(const QAWL_PreviewRequest &);

  const QString & fileName() const { return m_file; }
  uint scale() const { return m_scale; }
  
  bool processFrames(uint frames);

  bool isDone() { return m_done; }

  long processedFrames() const { return m_readFrames; }

protected:

  void finalizeProcessing();

  void clear();

  QString            m_file, m_previewFile;
  uint               m_scale;
  uint               m_sourceScale;

  std::vector<short> m_dataBuffer;
  std::vector<int>   m_dataBufferInt;
  std::vector<int>   m_scaledBuffer;

  DI_AudioFile      *m_infile;
  DI_AudioFile      *m_outfile;

  long               m_readFrames;

  bool               m_done;
};


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


class QAWL_PreviewCalculator : public QObject
{
  Q_OBJECT
public:
  QAWL_PreviewCalculator();
  virtual ~QAWL_PreviewCalculator();

  static bool lockFile(const QString &);
  static bool unlockFile(const QString &);

  static QString previewFileName(const QString &file, uint scale);
  
  bool requestPreview(const QString &file, uint scale, uint sourceScale = 1);
  bool requestNormalPreviews(const QString &file);
  bool isPreviewReady(const QString &file, uint scale);

  static uint nofNormalPreviews();
  static uint normalPreviewScale(uint i);
  static uint determinePreviewScale(double);

  static long fileFrames(const QString &file);
  
  static QAWL_PreviewCalculator & instance();
  static void deleteInstance();

public slots:

  void update();

signals:
  void previewAvailable(const QString &, uint);

protected: 
  
  bool lockFileInternal(const QString &file);
  bool unlockFileInternal(const QString &file);
  bool isLocked(const QString &file)
  { return findLocked(file) != m_lockedFiles.end(); }
  
  std::list<QAWL_FileLock>::iterator findLocked(const QString &file);
  std::list<QAWL_FilePreviewCalculus>::iterator findCalculusObject();

  /* Objects that can be used from within other threads. */
  DI_MutexAuto             m_mutex;
  std::list<QAWL_FileLock> m_lockedFiles;

  std::list<QAWL_PreviewRequest>      m_requests;
  std::list<QAWL_FilePreviewCalculus> m_activePreviews;
  
  QTimer m_timer;
  
  static QAWL_PreviewCalculator * m_calculator;
};

#endif
