/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <qawl_renderer.h>

#include <qawl_color_list.h>
#include <qawl_file_cache.h>

#include <qpainter.h>

void QAWL_Renderer::render(QPainter *painter, 
			   QPointArray *points,
			   QAWL_FileCache *cache, 
			   ulong firstRealFrame, ulong lastRealFrame,
			   uint chan,
			   /// Pixels
			   int left, 
			   /// Pixels
			   int right, 
			   /// Pixels
			   int centerY,
			   /// Pixels
			   float height,
			   /// Clip at this number of pixels
			   int clipPixels,
			   QAWL_ColorList * colors)
{
  /* qDebug("QAWL_Renderer::render # p = %p, c = %p "
	 "%lu %lu, [%d, %d], [%d, %d]",
	 painter, cache, 
	 firstRealFrame, lastRealFrame, left, right, centerY, height);
  */
  const uint segmentsPerPass = 100;

  if(points->size() < segmentsPerPass * 2)
    points->resize(segmentsPerPass * 2);

  uint count = 0;

  double pixelScaler = 1.0 / (double) (right - left);
  double frameSpan = (double) (lastRealFrame - firstRealFrame);
  double framesPerPixel = frameSpan * pixelScaler;
  uint pixels = right - left;

  double heightScale = (double) height / 32768.0;
  double clipAt = (double) clipPixels;

  int operateAt = 1000000;

  if(colors) {
    
    colors->seekToItem((long) firstRealFrame, framesPerPixel);

    // qDebug("QAWL_Renderer::render # %d color segments", (int) colors->size());

    if(!colors->atEnd()) {
      operateAt = colors->nextItemPixels((long) firstRealFrame, 
					 framesPerPixel) - 1;

      QAWL_ColorList::Item & item = colors->currentItem();

      painter->setBackgroundColor(item.background());
      painter->setPen(item.foreground());

      painter->eraseRect(left, centerY - clipPixels, 
			 operateAt - left + 2, clipPixels * 2 + 1);

      /* qDebug("QAWL_Renderer::render # %x, %d", 
	 (int) item.background().rgb(), operateAt);*/
    }
    else
      qDebug("QAWL_Renderer::render # No visible color segments");
  }

  for(uint i=0; i <= pixels; i++) {

    ulong startFrame = firstRealFrame + 
      (ulong) (i * pixelScaler * frameSpan);

    ulong endFrame = firstRealFrame + 
      (ulong) ((i+1) * pixelScaler * frameSpan);

    QAWL_MinMax range = cache->getMinMax(startFrame, endFrame, chan);

    range.limitZeroPassing();
    
    double low = range.min * heightScale;
    if(low < -clipAt) low = -clipAt;

    double high = range.max * heightScale;
    if(high > clipAt) high = clipAt;
    
    points->setPoint(2 * count, left + i, centerY - (int) low);
    points->setPoint(2 * count + 1, left + i, centerY - (int) high);
    
    count++;
    
    if(count == segmentsPerPass || (i + left) > (uint) operateAt) {
      painter->drawLineSegments(*points, 0, count);
      count = 0;

      if(colors && (i + left) > (uint) operateAt) {

	colors->advance();

	operateAt = colors->nextItemPixels((long)firstRealFrame, framesPerPixel) - 1;
	
	painter->setBackgroundColor(colors->currentItem().background());
	painter->setPen(colors->currentItem().foreground());

	painter->eraseRect(i + left + 1, centerY - clipPixels, 
			   operateAt - (i + left) + 2, clipPixels * 2 + 1);
      }
    }
  }
  
  painter->drawLineSegments(*points, 0, count);

}
