  /* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 *
 * This file is licensed under the Lesser GNU Public License
 * (LGPL). Extra clauses can be found from the file MUSTAJUURI_LGPL and
 * further explanations from the file MUSTAJUURI_LICENSE. The license
 * text can be found in file "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#include <qawl_view.h>

#include <qawl_file_cache.h>
#include <qawl_keyboard_handler.h>
#include <qawl_preview_calculator.h>
#include <qawl_renderer.h>

#include <qcursor.h> 
#include <qfileinfo.h> 
#include <qlayout.h>
#include <qpainter.h> 
#include <qpixmap.h> 
#include <qpointarray.h> 
#include <qscrollbar.h> 

#include <math.h>

class QAWL_View::Internal
{
  friend class QAWL_View;

  QPixmap m_pixmap;
  QPixmap m_cursor;  

  QDateTime m_lastModified;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

QPointArray QAWL_View::m_pointArray;

QAWL_View::QAWL_View(QWidget * parent, const char * name)
  : QWidget(parent, name),
    m_firstFrame(0),
    m_fileFrames(-1),
    m_fileSamplingRate(44100.0),
    m_samplesPerPixel(1500.0),
    m_magnify(1.0),
    m_fontHeight(10),
    m_cursorAt(-1),
    m_cursorRenderedAt(-1),
    m_previewAvailable(false),
    m_cache(new QAWL_FileCache),
    m_xScale(TimeScale),
    m_render(true),
    m_keyboardHandler(0),
    m_internal(new Internal)
{
  m_previewScale = 
    QAWL_PreviewCalculator::determinePreviewScale(m_samplesPerPixel);

  setMinimumSize(100, 50);

  QVBoxLayout *layout = new QVBoxLayout(this);
  layout->addSpacing(50);
  layout->addStretch(90);
  
  /* m_display = new QAWL_ViewDisplay(this);
  m_display->setMinimumHeight(100);
  layout->addWidget(m_display, 90);*/

  m_scroll = new QScrollBar(Qt::Horizontal, this, "qawl-scroll");
  layout->addWidget(m_scroll);  

  connect(m_scroll, SIGNAL(valueChanged(int)), this, SLOT(scrollMove(int)));
  m_scroll->setRange(0, 2000);

  connect(&QAWL_PreviewCalculator::instance(), 
	  SIGNAL(previewAvailable(const QString &, uint)),
	  this, SLOT(fileAvailable(const QString &, uint)));
}

QAWL_View::~QAWL_View()
{
  delete m_cache;
  delete m_internal;
}

void QAWL_View::setFile(const QString &file)
{
  QFileInfo fi(file);
  if(!fi.exists()) return;

  if(m_file == file && fi.lastModified() == m_internal->m_lastModified) 
    return;

  m_internal->m_lastModified = fi.lastModified();

  m_file = file;

  DI_AudioFile af;
  
  bool ok = af.open(file.ascii(), DI_IO_INPUT, DI_ASF_INT16, 100);
  
  if(!ok) return;
  
  m_fileFrames = af.frameCount();
  m_fileSamplingRate = af.getSampleRate();

  // m_fileFrames = QAWL_PreviewCalculator::fileFrames(file);

  ok = QAWL_PreviewCalculator::instance().requestNormalPreviews(file);

  if(!ok) {
    qDebug("QAWL_View::setFile # Cannot generate previews");
    return;
  }

  m_scroll->setMaxValue(m_fileFrames);

  recalculateScrollSteps();
}

void QAWL_View::moveFirst(long firstFrame)
{
  // qDebug("QAWL_View::moveFirst # %ld", firstFrame);

  m_firstFrame = firstFrame;
  
  /*ulong lastFrame = 
    (ulong) (m_firstFrame + m_samplesPerPixel * m_display->width());*/
  
  // m_cache->loadRealFrames(m_firstFrame, lastFrame);

  if(m_render)
    repaint(false);
}

void QAWL_View::renderDisplay(QPaintEvent *)
{
  if(!m_previewAvailable) return;

  int w = width();
  int h = height() - m_scroll->height();

  int margin = 25;

  // Use a pixmap to get flicker-free double-buffered updates.
  // QPixmap pxm(size());
  m_internal->m_pixmap.resize(size());
  m_internal->m_cursor.resize(1, height());
  

  QPainter p;

  p.begin(&m_internal->m_pixmap, this);

  QColor scaleBg(0xd5d6f2);

  p.setBackgroundColor(scaleBg);
  p.eraseRect(0, 0, w, margin + 1);

  p.setBackgroundColor(backgroundColor());

  // qDebug("QAWL_View::renderDisplay # %d", (int) m_colors.size());

  if(!m_colors.size()) {
    p.eraseRect(0, margin + 1, w, h - margin);
  }

  h -= margin;

  QFont f(font());
  f.setPixelSize(m_fontHeight);
  p.setFont(f);

  if(m_xScale == TimeScale)
    renderXScaleTime(&p);
  else if(m_xScale == FramesScale)
    renderXScaleFrames(&p);

  QAWL_Renderer renderer;

  uint chans = m_cache->channels();

  int h2 = (int) (0.5 * (double) h / (double) chans );

  ulong lastFrame = (ulong) (m_firstFrame + m_samplesPerPixel * w);
  
  m_cache->loadRealFrames(m_firstFrame, lastFrame);

  for(uint channel=0; channel < chans; channel++) {
    
    QAWL_ColorList *colors = 0;

    if(m_colors.size())
      colors = & m_colors[channel % m_colors.size()];

    int centerY = channel * h2 * 2 + h2 + margin;
    
    renderer.render(&p, &m_pointArray, m_cache,
		    m_firstFrame, 
		    lastFrame,
		    channel,
		    0, w, centerY, h2 * m_magnify, h2, colors);
  }  
  p.end();

  p.begin(&m_internal->m_cursor);
    
  p.setPen(white);
  p.drawLine(0, 0, 0, height());
  
  p.end();

  int cursorPixel = frameToPixel(m_cursorAt);

  if(cursorPixel >= 0 && cursorPixel < width() && m_cursorAt >= 0) {

    bitBlt(&m_internal->m_pixmap, cursorPixel, 0, 
	   &m_internal->m_cursor, 0, 0, 1, height(), XorROP);
    m_cursorRenderedAt = cursorPixel;
  }
  else
    m_cursorRenderedAt = -1;

  bitBlt(this, 0, 0, &m_internal->m_pixmap);
}
  
long QAWL_View::pixelToFrame(int pixel) const
{
  return m_firstFrame + (long) (m_samplesPerPixel * pixel);
}

int  QAWL_View::frameToPixel(long frame) const
{
  double tmp = (double) (frame - m_firstFrame) / (double) m_samplesPerPixel;

  return tmp >= 0.0  ? (int) (tmp + 0.5) : -1;
}

void QAWL_View::jumpForward(double jumpFactor)
{
  long newFirstFrame = m_firstFrame + 
    (long) (jumpFactor * width() * m_samplesPerPixel);

  if(newFirstFrame >= m_fileFrames)
    newFirstFrame = m_fileFrames - 1;
  if(newFirstFrame < 0)
    newFirstFrame = 0;

  m_scroll->setValue(newFirstFrame);
}

void QAWL_View::setMagnification(double m, bool repaintNow)
{ 
  m_magnify = m;

  if(repaintNow && m_render)
    repaint(false);
}

void QAWL_View::setCursorFrame(long frame)
{
  int cursorPixel = frameToPixel(m_cursorAt);
  
  m_cursorAt = frame;

  int h = height();

  if(cursorPixel == m_cursorRenderedAt)
    return;
  
  if(m_cursorRenderedAt >= 0)
    bitBlt(this, m_cursorRenderedAt, 0, 
	   &m_internal->m_cursor, 0, 0, 1, h, XorROP);

  if(cursorPixel < 0 || cursorPixel >= width()) {    
    m_cursorRenderedAt = -1;
    return;
  }
  
  m_cursorRenderedAt = cursorPixel;
    
  bitBlt(this, cursorPixel, 0, &m_internal->m_cursor, 0, 0, 1, h, XorROP);
}

void QAWL_View::fileAvailable(const QString &file, uint scale)
{
  if(file != m_file || scale != m_previewScale)
    return;

  qDebug("QAWL_View::fileAvailable # %s %u", file.ascii(), scale);

  bool ok = m_cache->setFile(file, scale);

  m_previewAvailable = ok;

  if(!ok) qDebug("QAWL_View::fileAvailable # m_cache->setFile failed");

  if(m_render)
    repaint(false);
}


void QAWL_View::setArea(long first, double scale)
{
  if(first == m_firstFrame && scale == m_samplesPerPixel)
    return;

  qDebug("QAWL_View::setArea # %ld, %2.2lf", first, scale);

  bool tmp = m_render;

  m_render = false;
  setScale(scale);
  
  m_scroll->setValue(first);
  recalculateScrollSteps();

  m_render = tmp;

  if(m_render)
    repaint(false);

  // m_firstFrame = first;  
}

void QAWL_View::setScale(double s)
{
  m_samplesPerPixel = s;

  uint previewScale = 
    QAWL_PreviewCalculator::determinePreviewScale(m_samplesPerPixel);
  
  recalculateScrollSteps();

  if(m_previewScale == previewScale) {
    if(m_render)
      repaint(false);
    return;
  }

  m_previewScale = previewScale;
  
  if(previewScale == 1) {
    fileAvailable(m_file, 1);
  }
  else {
    m_previewAvailable = false;

    QAWL_PreviewCalculator::instance().requestNormalPreviews(m_file);
  }

  /* m_cache->setFile(m_file, previewScale);
     m_previewScale = previewScale;*/
}

void QAWL_View::scrollMove(int v)
{
  double relative = (double) v / (double) m_scroll->maxValue();

  long firstFrame = (long) (relative * m_fileFrames);

  moveFirst(firstFrame);
}

void QAWL_View::setScaleType(int t)
{
  m_xScale = (XScale) t;

  if(m_render)
    repaint(false);
}

void QAWL_View::resizeEvent ( QResizeEvent * e)
{
  QWidget::resizeEvent(e);

  recalculateScrollSteps();
}

void QAWL_View::paintEvent ( QPaintEvent * e)
{
  renderDisplay(e);
}

void QAWL_View::keyPressEvent ( QKeyEvent * e )
{
  if(m_keyboardHandler) m_keyboardHandler->viewKeyPressEvent(this, e);
}

void QAWL_View::keyReleaseEvent ( QKeyEvent * e )
{
  if(m_keyboardHandler) m_keyboardHandler->viewKeyReleaseEvent(this, e);
}

void QAWL_View::enterEvent ( QEvent * )
{
  setFocus(); 
}


void QAWL_View::recalculateScrollSteps()
{
  int w = width();
  int framesPerDisplay = (int) (w * m_samplesPerPixel);

  int maxStep = m_fileFrames / 2;  

  m_scroll->setPageStep
    (framesPerDisplay < maxStep ? framesPerDisplay : maxStep);

  int lineStep = (int) (20 * m_samplesPerPixel);

  maxStep /= 10;

  if(lineStep > maxStep) lineStep = maxStep;
  if(lineStep < 1) lineStep = 1;

  m_scroll->setLineStep(lineStep);

  /* qDebug("QAWL_View::recalculateScrollSteps # w = %d, ps = %d, ls = %d",
     w, framesPerDisplay, lineStep); */
}


void QAWL_View::renderXScaleTime(QPainter *p)
{
  QString str;
    
  const uint serCount = 6;
  const int series[serCount] = {1, 2, 5, 10, 15, 30};

  const int margin = 2;
  const int lineBeginsAt = m_fontHeight + 2 * margin;
  const int lineEndsAt = 25;
  const int wp2 = 50;

  int minDistMajor = 100;
  uint i;
 
  long fs = (long) m_fileSamplingRate;
  long finalFrame = (long) (m_samplesPerPixel * (width() + 1) + m_firstFrame);

  long samplesInMajor = (long) (minDistMajor * m_samplesPerPixel);

  long samplesPerMinute = fs * 60;
  // long samplesPerHour = samplesPerMinute * 60;

  long secondsInSpan = (samplesInMajor * 10) / fs;
  long minutesInSpan = secondsInSpan / 60;
  // long hoursInMajor = minutesInMajor / 60;

  if(minutesInSpan > 10) {
    // Render full minutes

    long minutesPerMajor = 1;

    for(i=0; i < serCount; i++) {
      minutesPerMajor = series[i];
      if(series[i] * samplesPerMinute > samplesInMajor)
	break;
    }

    // samplesInMajor / samplesPerMinute + 1;
    long framesPerMajor = minutesPerMajor * samplesPerMinute;
    
    long times = m_firstFrame / framesPerMajor;
    
    if(m_firstFrame % framesPerMajor) times++;
    
    long minute = times * minutesPerMajor;

    long firstTick = times * framesPerMajor;
    
    for(long tickNow = firstTick; 
	tickNow < finalFrame; 
	tickNow += framesPerMajor) {
      
      int x = frameToPixel(tickNow);

      if(minute >= 60) {
	str.sprintf("%d:%.2d:00", (int) (minute / 60), (int) (minute % 60));
      }
      else
	str.sprintf("%d:00", (int) minute);
      
      p->drawText(x - wp2, margin, wp2 * 2, 30, AlignHCenter | AlignTop, str);

      p->drawLine(x, lineBeginsAt, x, lineEndsAt);

      minute += minutesPerMajor;
    }
  }
  else if(secondsInSpan > 5) {

    // printf full seconds:

    long secondsPerMajor;

    for(i=0; i < serCount; i++) {
      secondsPerMajor = series[i];
      if(series[i] * fs > samplesInMajor)
	break;
    }
    
    // = samplesInMajor / fs + 1;

    /*long mod = 2;

    if(secondsInSpan > 60) {
      mod = 10;
    }
    else if(secondsInSpan > 40) {
      mod = 5;
    }

    long tmp = secondsPerMajor % mod;
    if(tmp)
      secondsPerMajor += (mod - tmp);
    */
    long framesPerMajor = secondsPerMajor * fs;
    
    long times = m_firstFrame / framesPerMajor;

    if(m_firstFrame % framesPerMajor) times++;
    
    long second = times * secondsPerMajor;

    long firstTick = times * framesPerMajor;
    
    for(long tickNow = firstTick; 
	tickNow < finalFrame; 
	tickNow += framesPerMajor) {
      
      int x = frameToPixel(tickNow);
      
      secondToStr(&str, second);
      
      p->drawText(x - wp2, margin, wp2 * 2, 30, AlignHCenter | AlignTop, str);

      p->drawLine(x, lineBeginsAt, x, lineEndsAt);
      
      second += secondsPerMajor;
    }
    
  }
  else {

    // Render fractional parts of the second

    const int divCount = 3;
    int dividors[divCount] = {2, 5, 10};

    int dividor = 1;
    int divStep = 1;
    int dividorBase = 1;

    bool found = false;

    for(i=0; i < 6 && !found; i++) {
      for(int j=0; j < divCount && !found; j++) {
	
	dividor = dividorBase * dividors[j];

	if(m_fileSamplingRate / dividor < (samplesInMajor * 2)) {
	  found = true;
	  divStep =  (10 / dividors[j]);
	}
      }

      dividorBase *= 10;
    }

    long second = m_firstFrame / fs;
    
    long tmpFrame = second * fs;
    long tmpFrame2 = m_firstFrame - tmpFrame;

    double framesPerDivision = m_fileSamplingRate / (double) dividor;

    double untilFirstVisibleFrame = tmpFrame2 / framesPerDivision;

    long times = (long) ceil(untilFirstVisibleFrame);

    /* long tmp = times % divStep;
       if(tmp) times += (divStep - tmp);*/

    long dividorCount = times * divStep;

    if(dividorCount == dividorBase) {
      dividorCount = 0;
      second++;
    }

    bool done = false;

    char format[32];

    sprintf(format, ".%%.%dld", (int) digits10(dividorBase) - 1);
    
    QString fractStr;

    /* qDebug("Trash: %d %d %d #  %ld %ld %ld", 
	   dividor, divStep, dividorBase, 
	   dividorCount, times, second);*/


    while(!done) {
      
      double samplesTmp = times * framesPerDivision;

      long frame = tmpFrame + (long) samplesTmp;
      
      int x = frameToPixel(frame);

      if(x >= width()) {
	done = true;
	break;
      }
      
      secondToStr(&str, second);
      fractStr.sprintf(format, dividorCount);

      str += fractStr;

      p->drawText(x - wp2, margin, wp2 * 2, 30, AlignHCenter | AlignTop, str);

      p->drawLine(x, lineBeginsAt, x, lineEndsAt);
      
      times++;

      dividorCount += divStep;
      
      if(dividorCount == dividorBase) {
	dividorCount = 0;
	second++;
      }
    }
  }

}

void QAWL_View::renderXScaleFrames(QPainter *p)
{
  // Pixels between successive text points
  int minDistMajor = 100;
  
  long framesPerMajor = (long) (m_samplesPerPixel * minDistMajor);
  long finalFrame = (long) (m_samplesPerPixel * (width() + 1) + m_firstFrame);

  long   firstTick;

  if(framesPerMajor < 10.0) {
    framesPerMajor = (1 + (int) framesPerMajor);
    firstTick = m_firstFrame;
  }
  else {
    // long tmp1 = digits10((long) finalFrame + m_firstFrame);
    long tmp2 = digits10((long) m_samplesPerPixel * 500);
    long ignoreDigits = tmp2 - 2;

    if(ignoreDigits < 0) ignoreDigits = 0;
    
    framesPerMajor = roundUp10(framesPerMajor, ignoreDigits);

    // firstTick = roundUp10(m_firstFrame, ignoreDigits);
  }

  long times = m_firstFrame / framesPerMajor;

  if(m_firstFrame % framesPerMajor) times++;
  
  firstTick = times * framesPerMajor;

  const int margin = 2;
  const int lineBeginsAt = m_fontHeight + 2 * margin;
  const int lineEndsAt = 25;
  const int wp2 = 50;

  QString str;

  for(long tickNow = firstTick; 
      tickNow < finalFrame; 
      tickNow += (long) framesPerMajor) {
    
    int x = frameToPixel(tickNow);

    str.setNum(tickNow);
    
    p->drawText(x - wp2, margin, wp2 * 2, 30, AlignHCenter | AlignTop, str);

    p->drawLine(x, lineBeginsAt, x, lineEndsAt);
  }

}

long QAWL_View::roundUp10(long initialVal, long ignoreDigits)
{
  long mod = 10;

  for(long i=0; i < ignoreDigits; i++) {
    long tmp = (initialVal % mod);

    if(tmp)
      initialVal += (mod - tmp);

    mod *= 10;
  }

  return initialVal;
}

long QAWL_View::digits10(long v)
{
  if(v < 0) v *= -1;

  long r = 1;
  long cmp = 10;
  long prev = 1;

  while(prev < cmp) {
    if(cmp > v)
      return r;

    prev = cmp;
    cmp *= 10;
    r++;
  }
  return r;
}

void QAWL_View::secondToStr(QString *str, long second)
{
  if(second >= 60) {
    
    long minute = second / 60;
    long second = second % 60;
    
    if(minute >= 60) {
      str->sprintf("%ld:%.2ld:%.2ld", (minute / 60), minute % 60, second);
    }
    else
      str->sprintf("%ld:%.2ld", minute, second);
  }
  else
    str->sprintf("%d", (int) second);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

QAWL_View2::QAWL_View2(QWidget * parent, const char * name)
  : QAWL_View(parent, name),
    m_selectionStart(-1),
    m_selectionEnd(-1),
    m_jumpMode(NoJump)
{
  qDebug("QAWL_View2::QAWL_View2 # %p", this);
  
  connect(&m_timer, SIGNAL(timeout()), this, SLOT(jump()));
}

QAWL_View2::~QAWL_View2()
{
  qDebug("QAWL_View2::~QAWL_View2 # %p", this);
}

void QAWL_View2::setFile(const QString &f)
{
  QAWL_View::setFile(f);
  unselect();
}

void QAWL_View2::unselect(bool repaintNow)
{
  m_selectionStart = -1;
  m_selectionEnd = -1;

  m_colors.resize(0);
  
  if(repaintNow && m_render)
    repaint(false);

  emit isSelected(false);
}

void QAWL_View2::selectAll(bool repaintNow)
{
  m_selectionStart = 0;
  m_selectionEnd = m_fileFrames;

  m_timer.stop();
  m_jumpMode = NoJump;
  
  buildColors(m_selectionStart, m_selectionEnd, 0x77777);

  if(repaintNow && m_render)
    repaint(false);

  emit isSelected(true);
}

void QAWL_View2::jump()
{
  // qDebug("QAWL_View2::jump # %d", (int) m_jumpMode);

  long selectionEnd = pixelToFrame(mapFromGlobal(QCursor::pos()).x());

  if(selectionEnd < 0) selectionEnd = 0;
  if(selectionEnd > m_fileFrames) selectionEnd = m_fileFrames;

  buildColors(m_selectionStart, selectionEnd, red);

  if(m_jumpMode == JumpForward)
    jumpForward(0.1);
  else if(m_jumpMode == JumpBackward)
    jumpForward(-0.1);
}

void QAWL_View2::mousePressEvent ( QMouseEvent * e )
{
  if(e->button() == LeftButton) {

    m_selectionStart = pixelToFrame(e->x());

    if(m_selectionStart > m_fileFrames) m_selectionStart = m_fileFrames;

    m_selectionEnd = -1;

    if(m_selectionStart < 0) return;

    buildColors(m_selectionStart, m_selectionStart, red);

    if(m_render)
      repaint(false);

    qDebug("QAWL_View2::mousePressEvent # Selection starts at %ld (%d)",
	   m_selectionStart, e->x());
  }
}

void QAWL_View2::mouseReleaseEvent ( QMouseEvent * e )
{
  if(e->button() == LeftButton && m_selectionStart >= 0) {

    if(m_jumpMode != NoJump) {
      m_timer.stop();
      m_jumpMode = NoJump;
    }

    m_selectionEnd = pixelToFrame(e->x());
    
    if(m_selectionEnd < 0) m_selectionEnd = 0;
    if(m_selectionEnd > m_fileFrames) m_selectionEnd = m_fileFrames;

    if(m_selectionStart == m_selectionEnd) {
      unselect();
      return;
    }

    if(m_selectionStart > m_selectionEnd) {
      long tmp = m_selectionStart;
      m_selectionStart = m_selectionEnd;
      m_selectionEnd = tmp;
    }

    // Add samplesPerPixel to get all frames within the pixel.
    m_selectionEnd += (long) samplesPerPixel();
    
    qDebug("QAWL_View2::mouseReleaseEvent # Selection ends at %ld (%d,%ld)",
	   m_selectionEnd, e->x(), m_selectionStart);

    buildColors(m_selectionStart, m_selectionEnd, 0x77777);

    if(m_render)
      repaint(false);

    emit selectedArea(m_selectionStart, m_selectionEnd);
    emit isSelected(true);
  }
}

void QAWL_View2::mouseMoveEvent ( QMouseEvent * e )
{
  if(m_selectionStart < 0 || m_selectionEnd > 0) return;

  long selectionEnd = pixelToFrame(e->x());

  if(selectionEnd < 0) selectionEnd = 0;
  if(selectionEnd > m_fileFrames) selectionEnd = m_fileFrames;

  buildColors(m_selectionStart, selectionEnd, red);

  if(e->x() >= width()) {
    m_jumpMode = JumpForward;
    jump();
    m_timer.start(50);
  }
  else if(e->x() < 0) {
    m_jumpMode = JumpBackward;
    jump();
    m_timer.start(50);
  }
  else {

    if(m_jumpMode != NoJump) {
      m_timer.stop();
      m_jumpMode = NoJump;
    }
    
    if(m_render)
      repaint(false);
  }
}

void QAWL_View2::buildColors
(long firstSelected, long lastSelected, QColor selectColor)
{
  if(firstSelected > lastSelected) {
    long tmp = firstSelected;
    firstSelected = lastSelected;
    lastSelected = tmp;
  }

  m_colors.resize(1);

  QAWL_ColorList * colors = &m_colors[0];

  colors->clear();

  QColor bg(backgroundColor());
  QColor fg(foregroundColor());

  if(firstSelected != 0)
    colors->appendItem(bg, fg, 0);

  colors->appendItem(white, selectColor, firstSelected, QAWL_ColorList::Floor);
  colors->appendItem(bg, fg, lastSelected, QAWL_ColorList::Ceil);

}

