/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef QAWL_VIEW_H
#define QAWL_VIEW_H

#include <qawl_color_list.h>

#include <qtimer.h>
#include <qwidget.h>

#include <vector>

class QAWL_FileCache;
class QAWL_KeyboardHandler;
class QAWL_View;

class QScrollBar;

/** An audio file viewer widget. */

class QAWL_View : public QWidget
{
  Q_OBJECT
public:

  enum XScale {
    TimeScale,
    FramesScale
  };

  QAWL_View(QWidget * parent, const char * name = 0);
  virtual ~QAWL_View();

  virtual void setFile(const QString &);

  virtual void moveFirst(long);

  virtual void renderDisplay(QPaintEvent *);

  long pixelToFrame(int pixel) const;
  int  frameToPixel(long frame) const;
  
  double samplesPerPixel() const { return m_samplesPerPixel; }

  void jumpForward(double jumpFactor = 0.1);

  void setMagnification(double m, bool repaintNow = true);

  const QString & fileName() const { return m_file; }

  long firstFrame() const { return m_firstFrame; }

  long fileFrames() const {  return  m_fileFrames; }
  
  void setKeyboardHandler(QAWL_KeyboardHandler *kbh) 
  { setFocusPolicy(QWidget::StrongFocus); m_keyboardHandler = kbh; }

  void setCursorFrame(long frame);

public slots:

  virtual void fileAvailable(const QString &, uint);

  virtual void setArea(long, double);

  virtual void setScale(double);

  virtual void scrollMove(int);

  virtual void setScaleType(int);

protected:

  virtual void resizeEvent ( QResizeEvent * );

  virtual void paintEvent ( QPaintEvent * );

  virtual void keyPressEvent ( QKeyEvent * e );
  virtual void keyReleaseEvent ( QKeyEvent * e );
  virtual void enterEvent ( QEvent * );

  void recalculateScrollSteps();

  void renderXScaleTime(QPainter *);
  void renderXScaleFrames(QPainter *);

  static long roundUp10(long initialVal, long ignoreDigits);
  static long digits10(long v);
  static void secondToStr(QString *, long);

  QString m_file;

  long   m_firstFrame;
  long   m_fileFrames;
  double m_fileSamplingRate;
  double m_samplesPerPixel;
  double m_magnify;
  uint   m_previewScale;
  int    m_fontHeight;
  long   m_cursorAt;
  int    m_cursorRenderedAt;

  bool   m_previewAvailable;

  QAWL_FileCache *m_cache;

  QScrollBar *m_scroll;

  std::vector<QAWL_ColorList> m_colors;

  XScale m_xScale;

  bool   m_render;
  
  QAWL_KeyboardHandler *m_keyboardHandler;

  class Internal;
  Internal *m_internal;

  static QPointArray m_pointArray;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** Audio file viewer with selection possibilities. */

class QAWL_View2 : public QAWL_View
{
  Q_OBJECT

public:

  QAWL_View2(QWidget * parent, const char * name = 0);
  virtual ~QAWL_View2();
  
  virtual void setFile(const QString &);

  void unselect(bool repaintNow = true);
  void selectAll(bool repaintNow = true);

  long selectionStart() const { return m_selectionStart; }
  long selectionEnd() const   { return m_selectionEnd; }

public slots:

  void jump();

signals:
  void selectedArea(long, long);
  void isSelected(bool);

protected:

  enum JumpMode { NoJump, JumpForward, JumpBackward };

  virtual void mousePressEvent ( QMouseEvent * e );
  virtual void mouseReleaseEvent ( QMouseEvent * e );
  virtual void mouseMoveEvent ( QMouseEvent * e );

  void buildColors(long firstSelected, long lastSelected, QColor selectColor);


  long m_selectionStart;
  long m_selectionEnd;

  JumpMode m_jumpMode;
  QTimer m_timer;
};

#endif
