/* COPYRIGHT
 *
 * This file is part of the DIVA IO library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000.
 * Jarmo.Hiipakka@hut.fi Tommi.Ilmonen@hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA IO library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */


#include <stdlib.h>

#include <math.h>
#include <stdio.h>
#include <unistd.h>

#include <di_config.h>
#include <di_error.h>
#include <di_io_defs.h>
#include <di_audio_file.h>

int main(int argc, char **argv)
{
  const int cdblock = 2352;

  const ulong blocksPerPass = 20;

  if(argc < 3) {
    printf("Usage:\n%s [-n -s -e(+arg)] infile outfile\n", argv[0]);
    fflush(0);
    return 0;
  }

  bool verbose = true;

  uint selectChan = 0;

  ulong i;
  
  for(i = 1; i < (ulong) argc && argv[i][0] == '-'; i++) {
    switch(argv[i][1]) {

    case 'c':
      selectChan = atoi(argv[++i]) - 1;
      break;

    case 's':
      verbose = false;
      break;
    }
  }

  int firstFile = i;

  const char *fin = argv[firstFile];
  const char *fout = argv[firstFile + 1];
  
  DI_AudioFile in, out;

  if(!in.open(fin, DI_IO_INPUT, DI_ASF_INT32, cdblock * blocksPerPass)) {
    printf("Failed to open input file \"%s\"", fin);
    fflush(0);
    return 0;
  }

  int chans = in.inChannels();

  if((int) selectChan >= chans) {
    fprintf(stderr, "Selected channel %d exceeds the number of channels (%d)\n",
	    (int) selectChan + 1, (int) chans);
    return -1;
  }

  if(verbose) {
    puts("Input opened");
  }

  out.setFileFormat(DI_AFF_WAVE);
  out.setSampleRate(in.getSampleRate());
  out.setSampleWidth(in.getSampleWidth());
  out.setSampleFormat(in.getSampleFormat());
  out.setByteOrder(in.getByteOrder());
  out.setChannels(1);

  const int blockSize = 10000;
  
  if(!out.open(fout, DI_IO_OUTPUT, DI_ASF_INT32, blockSize)) {
    printf("Failed to open output file \"%s\"", fout);
    fflush(0);
    return 0;
  }
  
  ulong frames = in.frameCount();

  int * buf = new int[blockSize * chans];

  if(verbose) {
    printf("Extracting track %d from \"%s\" to \"%s\"\n", 
	   selectChan, fin, fout);
    fflush(0);    
  }

  for(ulong frame = 0; frame < frames; ) {

    int rframes = in.readFrames(buf, blockSize);
    
    for(int j=0; j < rframes; j++)
      buf[j] = buf[chans * j + selectChan];

    out.writeFrames(buf, rframes);
    
    frame += rframes;
  }
  in.close();
  out.close();
  printf("DONE\n");
  return 0;
}


