/* COPYRIGHT
 *
 * This file is part of the DIVA IO library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000.
 * Jarmo.Hiipakka@hut.fi Tommi.Ilmonen@hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA IO library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */


#include <stdlib.h>

#include <math.h>
#include <stdio.h>
#include <unistd.h>

#include <di_config.h>
#include <di_error.h>
#include <di_io_defs.h>
#include <di_sleep.h>
#include <di_audio_file.h>


int main(int argc, char **argv)
{
  const int cdblock = 2352;

  const ulong blocksPerPass = 20;

  if(argc < 3) {
    printf("Usage:\n%s [-n -s -e(+arg)] infile outfile\n", argv[0]);
    fflush(0);
    return 0;
  }

  bool normalize = false;
  bool verbose = true;

  double clipFromEnd = 0.0;

  uint firstFile = 1;
  ulong i;
  
  for(i = 1; i < (ulong) argc && argv[i][0] == '-'; i++) {
    switch(argv[i][1]) {
    case 'e':
      clipFromEnd = atof(argv[++i]);
      break;
    case 'n':
      puts("Normalization enabled");
      normalize = true;
      break;      
    case 's':
      verbose = false;
      break;
    }
  }

  firstFile = i;

  const char *fin = argv[firstFile];
  const char *fout = argv[firstFile + 1];
  
  DI_AudioFile in, out;

  if(verbose) {
    printf("\"%s\" -> \"%s\"\n", fin, fout);
    fflush(0);
  }

  if(!in.open(fin, DI_IO_INPUT, DI_ASF_INT32, cdblock * blocksPerPass)) {
    printf("Failed to open input file \"%s\"", fin);
    fflush(0);
    return 0;
  }

  if(in.inChannels() != 2) {
    DI_error2(DI_ERR_ARGUMENT, 
	      "File should have 2 channels, not %d", (int) in.inChannels());
    return 0;
  }

  if(verbose) {
    puts("Input opened");
  }

  out.setFileFormat(DI_AFF_WAVE);
  out.setSampleRate(44100);
  out.setSampleWidth(16);
  out.setSampleFormat(DI_AFSF_TWOSCOMP);
  out.setByteOrder(DI_AFBO_LITTLEENDIAN);
  out.setChannels(2);
  
  if(!out.open(fout, DI_IO_OUTPUT, DI_ASF_INT32, cdblock * blocksPerPass)) {
    printf("Failed to open output file \"%s\"", fout);
    fflush(0);
    return 0;
  }
  
  ulong frames = in.frameCount() - (ulong) (clipFromEnd * 44100);

  ulong blocks = frames / cdblock;

  ulong informInterval = blocks / (10 * blocksPerPass);

  int buf[cdblock * 2 * blocksPerPass];

  double scale = 1.0;

  if(verbose) {
    double secs = (double) frames / 44100.0;
    int isecs = (int) secs;
    printf("File length: %d:%d:%d\n", 
	   isecs / 60, isecs % 60, (int) ((secs - (double) isecs) * 100.0));
  }

  if(normalize) {
    if(verbose) {
      printf("Looking for peak ");
      fflush(0);
    }
    double peak = DI_AudioFile::filePeak(fin, verbose);
    if(verbose)
      printf(" Peak amplitude = %f\n", peak);
    scale = 0.999 / peak;
  }

  if(verbose) {
    printf("truncating file ");
    fflush(0);
  }

  for(i=0; i < blocks; ) {

    ulong remain = blocks - i;

    ulong thisBlocks = remain < blocksPerPass ? remain : blocksPerPass;

    in.readFrames(buf, thisBlocks * cdblock);
    
    for(ulong j=0; j < thisBlocks * cdblock * 2; j++) {
      buf[j] = (int) (scale * (double) buf[j]);
      // buf[j] *= scale;
    }

    out.writeFrames(buf, thisBlocks * cdblock);

    if(((i / blocksPerPass + 1) % informInterval) == 0 && verbose) {
      printf(".");
      fflush(0);
    }
    i += thisBlocks;
  }
  in.close();
  out.close();
  printf("\n");
  return 0;
}


