/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the GNU Public License (GPL).
 *  
 */


#include <stdlib.h>

#include <math.h>
#include <stdio.h>
#include <unistd.h>

#include <di_config.h>
#include <di_error.h>
#include <di_io_defs.h>
#include <di_sleep.h>
#include <di_audio_file.h>

#include <mj_audio_sample.h>
#include <mj_error.h>
#include <mj_math.h>
#include <mj_ringbuffer.h>
#include <mj_ringbuffertmpl.C>

#include <vector>

int main(int argc, char **argv)
{
  ulong i;
  
  const char * file = "test.wav";
  float scale = 1.0;
  float amplitude = 1.0;
  uint undersample = 1;

  bool newMethod = true;

  for(i = 1; i < (ulong) argc && argv[i][0] == '-'; i++) {
    switch(argv[i][1]) {
    case 'a':
      amplitude = atof(argv[++i]);
      break;
    case 's':
      scale = atof(argv[++i]);
      break;
    case 'f':
      file = argv[++i];
      break;      
    case 'o':
      newMethod = false;
      break;      
    case 'u':
      undersample = atoi(argv[++i]);
      break;      
    }
  }

  MJ_trace2("Testing doppler with scale = %f and file = \"%s\"\n"
	    "undersample = %u, amplitude = %f",
	    scale, file, undersample, amplitude);

  MJ_AudioSample sample;

  bool ok = sample.loadFile(file, true);

  if(!ok)
    MJ_fatal(MJ_ERR_UNKNOWN, "Could not read file");
  
  double sr = 44100;
  uint maxDelay = (uint) (sr * 5.0);
  uint maxBlock = 100;

  MJ_RingBuffer2<uint> times;
  MJ_RingBuffer2<double> locations, audio;

  locations.resize(maxDelay);
  audio.resize(maxDelay);
  times.resize(maxDelay);

  times.setAll(0);
  locations.setAll(0);
  audio.setAll(0);

  DI_AudioFile af;
  af.setSampleWidth(16);
  af.setSampleRate((uint) sr);
  af.setChannels(1);
  af.setFileFormat(DI_AFF_WAVE);
  ok = af.open("doppler.wav", DI_IO_OUTPUT, DI_ASF_FLOAT32, maxBlock);

  if(!ok)
    MJ_fatal(MJ_ERR_UNKNOWN, "Cannot write file");
  
  std::vector<float> block;

  block.resize(maxBlock);

  uint locIndex = 0;

  double mps = 330.0; // Meters per second

  for(uint i=0; i < sample.wave().size(); i++) {
    audio.put(sample.wave().getFast(i));
    double distance = mps;
    if(i / sr > 0.5)
      distance -= (i / sr - 0.5) * mps * 0.2;

    distance = MJ_abs(distance * scale) + 20.0;

    if(i % undersample == 0) {
      times.put(i);
      locations.put(distance);
    }

    bool loop = true;

    double distanceTime;
    double time;
      
    while(loop) {
      
      distanceTime = locations.getIndex(locIndex) / mps;
      time = (i - times.getIndex(locIndex)) / (double) sr;
      
      if(time <= distanceTime) {
	loop = false;
      }
      else
	locIndex++;
    }

    // Calculate intersection of lines:
    double distanceTimeOld = locations.getIndex(locIndex - 1) / mps;
    double timeOld = (i - times.getIndex(locIndex - 1)) / (double) sr;

    double k = (distanceTime - distanceTimeOld) / (timeOld - time);
    double alpha = timeOld - distanceTimeOld;
    double x = alpha / (1.0 + k);
    double delay = timeOld - x;
    /* double y = k * alpha / (1.0 + k);
       double delay = timeOld - y;// * (time - timeOld); */

    // printf("d = %lf ", delay * sr);

    if(alpha < 0.0)
      puts("BAD ALPHA !!");
    // if(k <-0.001) printf("- %lf", k);
    // if(k > 0.001) printf("+ %lf", k);

    // if(distanceTime - distanceTimeOld < 0.001) printf("<");
    // if(distanceTime - distanceTimeOld > 0.001) printf("> %lf", distanceTime - distanceTimeOld );

    
    
    if(newMethod)
      block[i % maxBlock] = 
	audio.getInterpolatedNewest1D(delay * sr) * amplitude;
	// audio.getInterpolatedNewest1D(100.5) * amplitude;
	// audio.getNewest(delay * sr) * amplitude;
    else
      block[i % maxBlock] = 
	audio.getInterpolatedNewest1D(sr * distance / mps * amplitude);

    if(i % maxBlock == maxBlock - 1) {
      // printf(".%u", locIndex); fflush(0);
      af.writeFrames( & block[0], maxBlock);
    }
  }

  puts("DONE");

  return 0;
}

template class MJ_RingBuffer<unsigned>;

