/* 
   This is a launcher for Mustajuuri. It should be suitable for other
   similar audio-related apps as well.

   This app helps overcome the capability hassles of Linux.

   It is based on "sucap"-application by Finn Arne Gangstad.
   
   The source code of sucap (sucap.c) is included with Mustajuuri. 

   To compile and use this application you need:

   1) PATCHED LINUX KERNEL
   2) LIBCAP

   Further information about these can be found in:

   http://www.tml.hut.fi/~tilmonen/mustajuuri/doc/README.install.html#LINUX-LIBCAP

   Example compilation:

c++ -O2  -DPROGRAM_PATH="\"/usr/local/projects/tilmonen/mustajuuri-devel/bin/Linux-2.4.4/mustajuuri-bin\"" -Wall -fno-implicit-templates main.C -o mustajuuri-devel -lcap

c++ -O2  -DPROGRAM_PATH="\"/usr/local/projects/tilmonen/mustajuuri-current/bin/Linux-2.4.4/mustajuuri-bin\"" -Wall -fno-implicit-templates main.C -o mustajuuri-stable -lcap

   To use the binaries you need to make them setuid-root:
   
   mv mustajuuri-devel /usr/local/bin/
   chown root /usr/local/bin/mustajuuri-devel
   chmod 4755 /usr/local/bin/mustajuuri-devel

*/

#include <di_types.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>

#undef _POSIX_SOURCE
#include <sys/capability.h>

/* If the compilation fails on the preceding line, then you probably
   do not have the libcap installed. 
   
*/

// #include <pwd.h>

// #define __USE_BSD
#include <grp.h>
#include <unistd.h>
#include <sys/wait.h>
#include <errno.h>
#include <string.h>

/* This macro must be supplied by compiler args or similar path. It
   should contain a character string describing the full path to the
   binary you are going to use.
*/
const char *progPath = PROGRAM_PATH;

static void usage(void)
{
  fprintf(stderr,
 	  "usage: mjsucap [command-args...]\n\n"
	  "  This program is a wrapper that change UID but not privileges of a\n"
	  "  program to be executed.\n"
	  "  Note, this wrapper is intended to assist in overcoming a lack of support\n"
	  "  for filesystem capability attributes and should be used to launch other\n"
	  "  files.\n\n"
	  "[Copyright (c) 1998 Finn Arne Gangstad <finnag@guardian.no>]"
	  "[Copyright (c) 2000-2001 Tommi Ilmonen <Tommi.Ilmonen@hut.fi>]\n");

  exit(1);
}


static void
wait_on_fd(int fd)
{
  /* Wait until some data is available on a file descriptor, or until
   * end of file or an error is detected */
  char buf[1];
  while (read(fd, buf, sizeof(buf)) == -1 && errno == EINTR) {
    /* empty loop */
  }
}


int main(int argc, char **argv)
{
  const int bsize = 128;

  char lsBuf[bsize], b2[bsize], b3[bsize], buser[bsize], bgroup[bsize];

  cap_t old_caps;
  uid_t uid;
  pid_t pid, parent_pid, parent_pid2;
  gid_t gid;
  int pipe_fds[2];
  

  // Use /proc/ filesystem to get user and group names.

  parent_pid2 = getppid();

  sprintf(lsBuf, "/bin/ls -aln /proc/%d/stat", (int) parent_pid2);
  
  puts(lsBuf);

  FILE *ptr = popen(lsBuf, "r");

  if(ptr) {
    fscanf(ptr, "%s %s %s %s", b2, b3, buser, bgroup);
  }
  else {
    fprintf(stderr, "mjsucap: popen failed");
    exit(1);
  }

  uid = (uid_t) atol(buser);
  gid = (gid_t) atol(bgroup);
    
  /* set process group to current pid */
  if (setpgid(0, getpid())) {
    perror("mjsucap: Failed to set process group");
    exit(1);
  }
    
  if (pipe(pipe_fds)) {
    perror("mjsucap: pipe() failed");
    exit(1);
  }
    
  parent_pid = getpid();

  old_caps = cap_init();

  const unsigned nofCaps = 3;

  /* We need these capabilities:

     CAP_SYS_NICE      -> Real-time priority
     CAP_SYS_RESOURCE  -> RTC above 64 Hz
     CAP_IPC_LOCK      -> mlockall
  */

  cap_value_t capList[nofCaps] = 
  { CAP_SYS_NICE, CAP_SYS_RESOURCE, CAP_IPC_LOCK} ;

  cap_clear(old_caps);

  cap_set_flag(old_caps, CAP_EFFECTIVE, nofCaps, capList , CAP_SET);
  cap_set_flag(old_caps, CAP_INHERITABLE, nofCaps, capList , CAP_SET);
  cap_set_flag(old_caps, CAP_PERMITTED, nofCaps, capList , CAP_SET);
        
  {
    ssize_t x;
    printf("Caps: %s\n", cap_to_text(old_caps, &x));
  }


  /* fork off a child to do the hard work */
  fflush(NULL);
  pid = fork();
  if (pid == -1) {
    perror("mjsucap: fork failed");
    exit(1);
  }

  /* 1. mother process sets gid and uid
   * 2. child process sets capabilities of mother process
   * 3. mother process execs whatever is to be executed
   */

  if (pid) {
    /* Get rid of any supplemental groups */
    if (!getuid() && setgroups(0, 0)) {
      perror("mjsucap: setgroups failed");
      exit(1);
    }

    /* Set gid and uid (this probably clears capabilities) */
    setregid(gid, gid);
    setreuid(uid, uid);

    {
      ssize_t x;
      cap_t cap = cap_init();
      capgetp(0, cap);
      printf("Caps: %s\n", cap_to_text(cap, &x));
    }
	
    printf("[debug] uid:%d, real uid:%d\n", geteuid(), getuid());

    /* Signal child that we want our privileges updated */
    close(pipe_fds[1]); /* Child hangs in blocking read */

    /* Wait for child process to set our privileges */
    {
      int status = 0;
      if (wait(&status) == -1) {
	perror("mjsucap: wait failed");
	exit(1);
      }
      if (!WIFEXITED(status) || WEXITSTATUS(status)) {
	fprintf(stderr, "mjsucap: child did not exit cleanly.\n");
	exit(1);
      }
    }

    {
      ssize_t x;
      cap_t cap = cap_init();
      capgetp(0, cap);
      printf("Caps before execvp: %s\n", cap_to_text(cap, &x));
    }

    /*	printf("[debug] uid:%d, real uid:%d\n", geteuid(), getuid()); */
    /* exec the program indicated by args 2 ... */
    argv[0] = (char *) progPath;
    execvp(progPath, argv);
	
    /* if we fall through to here, our exec failed -- announce the fact */
    fprintf(stderr, "Unable to execute command: %s\n", strerror(errno));
    
    usage();
  } else {
    /* Child process */
    close(pipe_fds[1]);

    /* Wait for mother process to setuid */
    wait_on_fd(pipe_fds[0]);

    /* Set privileges on mother process */
    if (capsetp(parent_pid, old_caps)) {
      perror("mjsucaps: capsetp");
      _exit(1);
    }

    /* exit to signal mother process that we are ready */
    _exit(0);
  }
  return 0;
}
