/* 

   Compile commands:
   
   CC -n32 -fullwarn main.C mj_owns_pid.C mj_realtime_scheduling.C -o priogive

   g++ -Wall main.C mj_owns_pid.C mj_realtime_scheduling.C -o priogive

   or:

   make smain

   This program gives one user right to request real-time priority.
   Calls POSIX1.b real-time functionality.

   The program is very very big, because it has so many features.

   Caller must be in the database.
   
   Copyright info:

   Author : Tommi S. Ilmonen - Tommi.Ilmonen@hut.fi
   License: LPGL
   Date   : 15.11.1998
 */

#include <iostream.h>
#include <fstream.h>

#include <string.h>
#include <unistd.h>
#include <sched.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <pwd.h>

#include <mj_realtime_scheduling.h>
#include <mj_owns_pid.h>

/* Configuration file. SHOULD be located in /etc/ or some similar
   place. Normal users must not have access to the file. */
static const char *resourceFile = "priogiverc";
// static const char *resourceFile = "/etc/priogiverc";

// If this is zero the user is not checked from the database
#define MJ_PRIOGIVE_CHECK_USER 0

#if MJ_PRIOGIVE_CHECK_USER
bool MJ_priogiveAcceptedUser(const char *user)
{
  const int bsize = 1024;
  char  buf[bsize], buf2[bsize];

  ifstream in(resourceFile);

  if(!in.good()) {
    cerr << "No resource file" << endl;
    return false;
  }

  while(in.good() && ! in.eof()) {
    buf[0] = '\0';
    buf2[0] = '\0';
    
    in.getline(buf, bsize);

    sscanf(buf, "%s", buf2);
    
    if(strlen(buf2) && buf2[0] != '#') {
      if(strcmp(user, buf2) == 0) return true;
    }
  }
  return false;
}

#else   // MJ_PRIOGIVE_CHECK_USER

bool MJ_priogiveAcceptedUser(const char *)
{
  return true;
}

#endif

bool UserIsCaller(const char * cpCaller)
{
  // Get true user id (not effective)
  uid_t xUserID = getuid();
  struct passwd *xpPassWD = getpwnam(cpCaller);
  if(xpPassWD) {
    if(xpPassWD->pw_uid == xUserID)
      return true;
  }
  return false;
}

void Helper()
{
  cerr << "Options: \n" 
       << "\t -h      \tThis help\n"
       << "\t -i +arg \tPid                    - default: parent pid\n" 
       << "\t -p +arg \tPriority (1-99)        - default: 10\n"
       << "\t -r      \tRound robin scheduling - default: FIFO" << endl;
  exit(-1);
}

int main(int argc, char *argv[])
{
  /* Defaults when no command line arguments are present. */

  // The default pid is the pid of the calling process.
  int xPid = getppid();
  int iPri = 10;
  int iSchedPolicy = SCHED_FIFO;
  
  int c;
  while	((c = getopt(argc, argv, "hi:p:r")) != -1) {
    switch (c) {
    case 'h':
      Helper();      
    case 'i':
      xPid = atoi(optarg);
      cout << "Target process id = " << xPid << endl;
      break;
    case 'p':
      iPri = atoi(optarg);
      cout << "Requested priority = " << iPri << endl;
      break;
    case 'r':
      iSchedPolicy = SCHED_RR;
      cout << "Using Round Robin scheduling" << endl;
      break;
    default:
      Helper();
      return -1;
    }
  }
  
  if(!xPid) {
    cerr << "Undefined pid" << endl;
    return EINVAL;
  }  

  // Turn PID to corresponding passwd structure
  struct passwd *pw = MJ_pidOwner(xPid);
  if(!pw) {
    cerr << "No such process " << xPid << endl;
    return 0;
  }

  cout << "Name = " << pw->pw_name 
       << "\nUid  = " << pw->pw_uid << endl;

  // Check the user is in the database
  bool userOk = MJ_priogiveAcceptedUser(pw->pw_name);
  if(!userOk) {
    cerr << "user " << pw->pw_name << " not accepted" << endl;
    return 0;
  }
  
  // The content.
  if(!MJ_realTimeScheduling(iPri, xPid, iSchedPolicy)) {
    cerr << "Failed to get real-time scheduling\n" 
	 << strerror(errno) << endl;    
    return errno;
  }
  else
    cout << "Succeeded in changing priority" << endl;

  return 0;
}
