/* COPYRIGHT
 *
 * This file is part of Mustajuuri GPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2001.
 * Tommi.Ilmonen@hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
  
 * This file is licensed under the GNU Public License (GPL).

 *  
 */


#include <stdlib.h>

#include <math.h>
#include <stdio.h>
#include <unistd.h>

#include <di_config.h>
#include <di_error.h>
#include <di_io_defs.h>
#include <di_sleep.h>
#include <di_audio_file.h>

#include <mj_resampler.h>

#include <vector>

int main(int argc, char **argv)
{
  bool verbose = true;
  uint chan;

  const int cdblock = 2352;

  const ulong blocksPerPass = 20;

  if(argc < 4) {
    printf("Usage:\n%s scale infile outfile\n", argv[0]);
    fflush(0);
    return 0;
  }

  double scale = atof(argv[1]);

  const char *fin = argv[2];
  const char *fout = argv[3];
  
  DI_AudioFile in, out;

  if(verbose) {
    printf("\"%s\" -> \"%s\"\n", fin, fout);
    fflush(0);
  }

  if(!in.open(fin, DI_IO_INPUT, DI_ASF_FLOAT32, cdblock * blocksPerPass)) {
    printf("Failed to open input file \"%s\"", fin);
    fflush(0);
    return 0;
  }

  if(verbose) {
    puts("Input opened");
  }

  uint channels = in.inChannels();

  out.setFileFormat(DI_AFF_WAVE);
  out.setSampleRate(44100);
  out.setSampleWidth(16);
  out.setSampleFormat(DI_AFSF_TWOSCOMP);
  out.setByteOrder(DI_AFBO_LITTLEENDIAN);
  out.setChannels(channels);
  
  if(!out.open(fout, DI_IO_OUTPUT, DI_ASF_FLOAT32, cdblock * blocksPerPass)) {
    printf("Failed to open output file \"%s\"", fout);
    fflush(0);
    return 0;
  }
  
  uint blocksize = 1024;

  ulong frames = in.frameCount();

  float * inbuf = new float[blocksize * channels];
  float * outbuf = new float[blocksize * channels];

  double secs = (double) frames / 44100.0;
  int isecs = (int) secs;
  printf("File length: %d:%d:%d\n", 
	 isecs / 60, isecs % 60, (int) ((secs - (double) isecs) * 100.0));

  std::vector<MJ_Resampler<float> > samplers;
  samplers.resize(channels);

  float lim = 1.0/scale;

  if(lim > 1.0) lim = 1.0;

  for(chan = 0; chan < channels; chan++)
    samplers[chan].setParameters(lim);

  uint inframe = 0;

  uint oframes = 0;
  
  uint informInterval = (uint) (frames * scale * 0.1);

  while(inframe < frames) {

    uint oframe;
  
    for(oframe = 0; oframe < blocksize && inframe < frames; oframe++) {

      while(samplers[0].needNewSample() && inframe < frames) {

	int index = inframe % blocksize;

	if(index == 0)
	  in.readFrames(inbuf, blocksize);

	for(chan = 0; chan < channels; chan++) {
	  
	  float sample = inbuf[index * channels + chan];
	  samplers[chan].put(sample);
	}
	inframe++;
      }

      for(chan = 0; chan < channels; chan++) {
	outbuf[channels * oframe + chan] = samplers[chan].getResampled(scale);
      }
      
      oframes++;

      if(oframes % informInterval == 0) {
	printf(".");
	fflush(0);
      }
    }
    
    out.writeFrames(outbuf, oframe);

  }
  in.close();
  out.close();
  puts("DONE");
  return 0;
}


