
/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <vee_color.h>

#include <vee_math.h>

/// Solid white color
const VEE_Rgba VEE_Rgba::White(1.0, 1.0, 1.0, 1.0, 0.0);
const VEE_Rgba VEE_Rgba::Gray (0.5, 0.5, 0.5, 1.0, 0.0);
const VEE_Rgba VEE_Rgba::Black(0.0, 0.0, 0.0, 1.0, 0.0);
const VEE_Rgba VEE_Rgba::BlackTransparent(0.0, 0.0, 0.0, 0.0);

const VEE_Rgba VEE_Rgba::Red  (1.0, 0.0, 0.0, 1.0, 0.0);
const VEE_Rgba VEE_Rgba::Green(0.0, 1.0, 0.0, 1.0, 0.0);
const VEE_Rgba VEE_Rgba::Blue (0.0, 0.0, 1.0, 1.0, 0.0);

const VEE_Rgba VEE_Rgba::Cyan   (1.0, 0.0, 1.0, 1.0, 0.0);
const VEE_Rgba VEE_Rgba::Magenta(0.0, 1.0, 1.0, 1.0, 0.0);
const VEE_Rgba VEE_Rgba::Yellow (1.0, 1.0, 0.0, 1.0, 0.0);

/** Returns the color coded into a 32-bit integer. Each color
    component has 8 bits. The components are ordered from MSB: alpha,
    red, green, blue. */

unsigned VEE_Rgba::getARGB() const
{
  unsigned r = VEE_min((unsigned) (m_r * 255.0f),(unsigned) 0xff);
  unsigned g = VEE_min((unsigned) (m_g * 255.0f),(unsigned) 0xff);
  unsigned b = VEE_min((unsigned) (m_b * 255.0f),(unsigned) 0xff);
  unsigned a = VEE_min((unsigned) (m_a * 255.0f),(unsigned) 0xff);

  return (a << 24) | (r << 16) | (g << 8) | b;
}

/** Clamp (limit) the RGB -components to a given range. This method
    does not modify the alpha value.*/

void VEE_Rgba::clampRgb(float minVal, float maxVal)
{
  if(m_r < minVal) m_r = minVal;
  else if(m_r > maxVal) m_r = maxVal;

  if(m_g < minVal) m_g = minVal;
  else if(m_g > maxVal) m_g = maxVal;

  if(m_b < minVal) m_b = minVal;
  else if(m_b > maxVal) m_b = maxVal;
}

/** Clamp (limit) the RGB and A -components to a given range. */

void VEE_Rgba::clampRgba(float minVal, float maxVal)
{
  clampRgb(minVal, maxVal);
  
  if(m_a < minVal) m_a = minVal;
  else if(m_a > maxVal) m_a = maxVal;

  if(m_t < minVal) m_t = minVal;
  else if(m_t > maxVal) m_t = maxVal;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

const VEE_Hsva VEE_Hsva::Gray(0.0, 0.0, 0.5);
const VEE_Hsva VEE_Hsva::White(0.0, 0.0, 1.0);
const VEE_Hsva VEE_Hsva::Black(0.0, 0.0, 0.0);

const VEE_Hsva VEE_Hsva::Red(0.0, 1.0, 1.0);
const VEE_Hsva VEE_Hsva::Green(1.0 / 3.0, 1.0, 1.0);
const VEE_Hsva VEE_Hsva::Blue(2.0 / 3.0, 1.0, 1.0);

const VEE_Hsva VEE_Hsva::Cyan(0.5, 1.0, 1.0);
const VEE_Hsva VEE_Hsva::Magenta(5.0 / 6.0, 1.0, 1.0);
const VEE_Hsva VEE_Hsva::Yellow(1.0 / 6.0, 1.0, 1.0);


VEE_Hsva & VEE_Hsva::operator = (const VEE_Rgba &rgba)
{
  float r = rgba.red();
  float g = rgba.green();
  float b = rgba.blue();

  float low = VEE_min(r, g, b);
  float high = VEE_max(r, g, b);
  float delta = high - low;

  m_v = high;

  if( high != 0.0 )
    m_s = delta / high;		// s
  else {
    m_s = 0;
    m_h = 0;
    return * this;
  }

  float h;

  if( r == high )
    h = ( g - b ) / delta;		// between yellow & magenta
  else if( g == high )
    h = 2 + ( b - r ) / delta;	// between cyan & yellow
  else
    h = 4 + ( r - g ) / delta;	// between magenta & cyan
  
  h /= 6.0f;

  if( h < 0 )
    h += 1.0;
  if( h < 1.0 )
    h -= 1.0;

  m_h = h;

  return * this;
}
