/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <vee_emitter.h>

#include <vee_error.h>
#include <vee_math.h>
#include <vee_random.h>
#include <vee_scene.h>
#include <vee_strs.h>
#include <vee_values.h>
#include <vee_value_object.h>

#include <typeinfo>

VEE_Emitter::VEE_Emitter(VEE_Object *type, 
					   float creationRate)
  : m_interval(1.0f / creationRate),
    m_untilNext(0.0),
    m_rateMul(1.0f),
    m_afterCollisionInterval(1.0f / creationRate),
    m_inheritVelocity(false),
    m_accumulate(false),
    m_collisionNewRate(false),
    m_maxCount(-1),
    m_count(0),
    m_type(type)
{
  m_updateVelocity = false;
  m_lastPos = m_location;
  m_objectType = VEE_OBJECT_SOURCE;
}

VEE_Emitter::~VEE_Emitter()
{}

/** Receive and process a value. 

    Parameters:

    <UL>

    <LI> "rate", float, The child creation rate in Hertz, range: 0.0001 to inf

    <LI> "mulrate", float, rate multiplier, range: 1.0e-5 to 1.0e+10

    <LI> "type", VEE_ValueObject

    <LI> "type/", VEE_ValueObject, gives the message to the source object

    <LI> "childops/", VEE_ValueObject, message to child-operators

    <LI> "selfops/", VEE_ValueObject, message to self-operators

    </UL>
*/
bool VEE_Emitter::setValue(const char *id, VEE_Value *msg)
{
  bool ok = false;

  if(VEE_Strs::equals(id, VEE_Strs::rate)) {
    float v = -1.0;

    ok = msg->getFloat(&v);

    if(ok) {

      if(v < 1.0e-6) v = 1.0e-6;

      m_interval = 1.0f / v;

      if(m_untilNext > (m_interval / m_rateMul))
	m_untilNext = m_interval / m_rateMul;
    }
  }
  else if(VEE_Strs::equals(id, "collisionrate")) {
    float v = -1.0;

    ok = msg->getFloat(&v);

    if(ok) {

      if(v < 1.0e-6) v = 1.0e-6;

      m_afterCollisionInterval = 1.0f / v;
    }
  }
  else if(VEE_Strs::equals(id, "collisionnewrate")) {
    int tmp;

    ok = msg->getInt(&tmp);

    if(ok)
      m_collisionNewRate = tmp != 0;
  }
  else if(VEE_Strs::equals(id, "mulrate")) {
    float v;

    ok = msg->getFloat(&v);

    if(ok) {

      m_rateMul = VEE_clamp(v, 1.0e-5f, 1.0e+10f);
      
      if(m_untilNext > (m_interval / m_rateMul))
	m_untilNext = m_interval / m_rateMul;
    }
  }
  else if(VEE_Strs::equals(id, "maxcount")) {
    int v;

    bool ok = msg->getInt(&v);

    if(ok)
      m_maxCount = v;
    return ok;
  }
  else if(msg->getInt2(ok, id, "maxcount", & m_maxCount))
    return true;
  else if(VEE_Strs::equals(id, VEE_Strs::type)) {

    VEE_ValueObject * vo = dynamic_cast<VEE_ValueObject *> (msg);

    if(vo) {
      // delete m_type;
      m_type = vo->newObject();
      return true;
    }
    return false;
  }
  else if(strncmp(id, "type/", 5) == 0) {

    if(!m_type.ptr())
      return false;

    return m_type.ptr()->setValue(id + 5, msg);
  }
  else if(strncmp(id, "childops/", 9) == 0) {
    return m_plist.setValue(id + 9, msg);
  }
  else if(strcmp(id, "trigger") == 0) {
    create(0.0, 0.0);
    return true;
  }
  else  
    return VEE_Particle2::setValue(id, msg);

  return ok;
}

VEE_Variable VEE_Emitter::getVariable(const char *id) const
{  
  VEE_Variable v;

  if(VEE_Strs::equals(id, VEE_Strs::rate)) {

    v = 1.0f / m_interval;
  }
  else if(VEE_Strs::equals(id, "mulrate")) {
    v = m_rateMul;
  }
  else  
    v = VEE_Particle2::getVariable(id);
  
  return v;
}

VEE_Object *VEE_Emitter::clone() const
{
  return new VEE_Emitter(*this);
}

void VEE_Emitter::update(float dt)
{
  /* VEE_trace("VEE_Emitter::update # %p %f %f", 
     this, m_sinceLastCreated, m_interval); */

  if(!m_type.ptr()) 
    VEE_fatal("VEE_Emitter::update # NULL factory object");

  if(m_age == 0.0)
    m_lastPos = m_location;

  VEE_Particle2::update(dt);

  /* VEE_trace("VEE_Emitter::update # %p %f %f", 
     this, m_untilNext, m_interval); */
  
  /* VEE_trace("VEE_Emitter::update # %p %u self-operators", 
     this, m_slist.size()); */

  float realInterval = m_interval / m_rateMul;

  float realIntP2 = realInterval * 0.5f;

  if(m_accumulate)
    m_justCreated.clear();

  for(m_untilNext -= dt; 
      m_untilNext < 0.0 && (m_count < m_maxCount || m_maxCount < 0); 
      m_untilNext += VEE_rand.rand0X(realIntP2) + realIntP2) {
    
    VEE_Object * obj = create(-m_untilNext, dt);

    if(m_accumulate)
      m_justCreated.push_back(obj);

    m_count++;
    // VEE_trace("VEE_Emitter::update # New %s", typeid(*obj).name());
  }
  m_lastPos = m_location;
}

void VEE_Emitter::collision(const VEE_Vector3 &)
{
  if(m_collisionNewRate)
    m_interval = m_afterCollisionInterval;
}

bool VEE_Emitter::removeOperator(const char *name, bool alsoDelete)
{
  return m_plist.removeOperator(name, alsoDelete);
}

VEE_Object * VEE_Emitter::create( float dt, float dtfull)
{
  if(dt < 1e-5) dt = 1e-5;
  if(dtfull < 1e-5) dtfull = 1e-5;

  VEE_Object * obj = m_type.clone();

  VEE_ForceObject * force = dynamic_cast<VEE_ForceObject *> (obj);
  
  if(force)
    m_sceneBox->addNewForce(force);
  else
    m_sceneBox->addObject(obj);
  
  float newWeight = dt / dtfull;

  VEE_Vector3 loc = m_location * newWeight + m_lastPos * (1.0f - newWeight);

  // VEE_trace("VEE_Emitter::create # %p %s", force, typeid(*obj).name());

  obj->setValueVector3(VEE_Strs::location, loc);

  // VEE_trace("VEE_Emitter::create # DONE");
  
  if(m_inheritVelocity) {
    obj->setValueVector3(VEE_Strs::velocity, m_velocity);
  }
  
  m_plist.operate(obj, dt);

  obj->updateVelocity(dt);
  obj->update(dt);
  
  return obj;
}

