/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <vee_particle.h>

#include <vee_math.h>
#include <vee_renderer.h> 
#include <vee_scene.h>
#include <vee_strs.h>
#include <vee_values.h> 

VEE_Particle::VEE_Particle(const VEE_Particle &p)
  : VEE_Object(p),
    m_looks(p.m_looks ? p.m_looks->clone() : 0),
  m_visible(p.m_visible)
{
  if(m_looks) m_looks->setHost(this);
}

VEE_Particle::VEE_Particle(VEE_Vector3 location, VEE_Vector3 velocity,
			   float ageLimit, VEE_Looks *l)
  : VEE_Object(location, velocity, ageLimit),
    m_looks(l),
  m_visible(true)
{
  if(l) l->setHost(this);
  m_objectType = VEE_OBJECT_PARTICLE;
}

VEE_Particle::~VEE_Particle()
{
  delete m_looks;
}

/** Receive and process a value. 

    Parameters:

    <UL>
    
    <LI> "looks", VEE_Looks (VEE_ValueLooks)

    <LI> "visible", int, range 0 to 1

    </UL>

*/
bool VEE_Particle::setValue(const char *id, VEE_Value *msg)
{
  if(strcmp(id, VEE_Strs::looks) == 0) {

    VEE_ValueLooks * vl = dynamic_cast<VEE_ValueLooks *> (msg);
    
    if(!vl) return false;

    setLooks(vl->looks() ? vl->looks()->clone() : 0);

    return true;
  }
  else if(strcmp(id, VEE_Strs::visible) == 0) {
    int tmp;

    bool ok = msg->getInt(&tmp);
    
    if(ok)
      m_visible = tmp;

    return ok;
  }

  return VEE_Object::setValue(id, msg);
}

VEE_Variable VEE_Particle::getVariable(const char *id) const
{
  VEE_Variable v;

  if(strcmp(id, VEE_Strs::visible) == 0) {
    v = (int) m_visible;
  }
  else
    v = VEE_Object::getVariable(id);

  return v;
}

VEE_Object * VEE_Particle::clone() const
{
  VEE_Particle *p = new VEE_Particle(*this);
  return p;
}

void VEE_Particle::update(float dt)
{
  m_location += m_velocity * dt;
  m_age += dt;
  m_rotationRadians = VEE_checkRadianRange(m_rotationRadians + m_omega * dt);

  if(m_age > m_ageLimit)
    m_done = true;

  if(m_looks) 
    m_color = m_looks->color();
}

void VEE_Particle::render(VEE_Renderer *r)
{
  if(m_visible)
    r->renderParticle(this);
}

void VEE_Particle::setLooks(VEE_Looks *l)
{ 
  delete m_looks; 
  m_looks = l; 

  if(l) {
    l->setHost(this);
    m_color = m_looks->color();
  }
}

VEE_Object * VEE_Particle::factory()
{
  return new VEE_Particle();
}
