/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <vee_volume_grid.h>

#include <vee_bounding_box.h>
#include <vee_error.h>
#include <vee_math.h>
#include <vee_meta_balls.h>
#include <vee_random.h>
#include <vee_value.h>

#include <mj_array3d_tmpl.C>

template class MJ_Array3D<float>;

inline int VEE_floorDiv(int x, int y)
{
  if(x >= 0)
    return x / y;

  int r = x / y;
  if(r * y == x)
    return r;
  
  return r - 1;
}

VEE_VolumeGrid::VEE_VolumeGrid()
  : m_metaBalls(0)
{
  m_rotation.identity();
  m_rotationInv = m_rotation;
  m_scale.make(1.0);
}

VEE_VolumeGrid::~VEE_VolumeGrid()
{}

float VEE_VolumeGrid::inside(const VEE_Vector3 &v)
{
  if(!m_grid.elements())
    return 0.0f;
  // VEE_fatal("VEE_VolumeGrid::inside # Grid has no size");
  
  VEE_Vector3 loc(v);

  loc.descale(m_scale);

  loc = m_rotationInv * loc;

  loc = loc + 0.5;

  loc[0] *= m_grid.cols();
  loc[1] *= m_grid.rows();
  loc[2] *= m_grid.layers();

  if(loc[0] < 0.0 || loc[1] < 0.0 || loc[2] < 0.0)
    return 0.0f;

  if(loc[0] >= m_grid.cols() || 
     loc[1] >= m_grid.rows() || 
     loc[2] >= m_grid.layers())
    return 0.0f;

  uint col   = (uint) loc[0];
  uint row   = (uint) loc[1];
  uint layer = (uint) loc[2];
  
  return m_grid.get(layer, row, col);
}

void VEE_VolumeGrid::getBoundingBox(VEE_BoundingBox *bb)
{
  VEE_Vector3 corner;

  bb->clear();

  for(int i=0; i < 2; i++) {
    for(int j=0; j < 2; j++) {
      corner = m_scale * 0.5;
      if(i) corner[1] = - corner[1];
      if(j) corner[2] = - corner[2];

      VEE_Vector3 worldCorner = m_rotation * corner;
      bb->expand(worldCorner);
      worldCorner = - worldCorner;
      bb->expand(worldCorner);      
    }
  }
}

VEE_Vector3 VEE_VolumeGrid::newRandomPoint()
{
  VEE_Vector3 loc;

  uint i, n = 1000;

  if(!m_grid.elements())
    return VEE_Vector3ZERO;

  bool loop = true;

  for(i=0; i < n && loop; i++) {

    loc[0] = VEE_rand.rand0X(m_grid.cols() - 0.001);
    loc[1] = VEE_rand.rand0X(m_grid.rows() - 0.001);
    loc[2] = VEE_rand.rand0X(m_grid.layers() - 0.001);

    float ratio = VEE_rand.rand01();

    float tmp = m_grid.get((uint) loc[2], (uint) loc[1], (uint) loc[0]);
    
    if(tmp > ratio)
      loop = false;
  }
  
  loc[0] /= m_grid.cols();
  loc[1] /= m_grid.rows();
  loc[2] /= m_grid.layers();
  
  loc = loc - 0.5;

  loc.scale(m_scale);

  VEE_trace("VEE_VolumeGrid::newRandomPoint # %f %f %f after %u", 
	    loc.x, loc.y, loc.z, i);

  return m_rotation * loc;
}

bool VEE_VolumeGrid::setValue(const char *id, VEE_Value *v)
{
  bool ok = false;

  if(strcmp(id, "rotation") == 0) {
    VEE_Matrix3 tmp;
    
    ok = v->getMatrix3(&tmp);
    
    if(ok) {
      m_rotation = tmp;
      m_rotationInv = tmp;
      m_rotationInv.transpose();
    }
  }

  return ok;
}

VEE_Volume * VEE_VolumeGrid::clone() const
{
  return new VEE_VolumeGrid(*this);
}

VEE_Vector3 VEE_VolumeGrid::createFromMetaBalls(VEE_MetaBalls *balls, 
						uint density)
{
  const char * fname = "VEE_VolumeGrid::createFromMetaBalls";

  VEE_trace("%s # %p %p %u", fname, this, balls, density);

  uint i, j, k;

  VEE_BoundingBox bbox;

  balls->getBoundingBox(&bbox);
  
  VEE_Vector3 high = bbox.high();
  VEE_Vector3 low  = bbox.low();
  VEE_Vector3 center = bbox.center();

  high -= center;
  low  -= center;

  m_scale = high * 2.0;

  float maxDim = high[0];
  if(maxDim < high[1]) maxDim = high[1];
  if(maxDim < high[2]) maxDim = high[2];

  float division = maxDim / (float) density;

  uint rows   = VEE_max((uint) 2, (uint) (high[0] / division));
  uint cols   = VEE_max((uint) 2, (uint) (high[1] / division));
  uint layers = VEE_max((uint) 2, (uint) (high[2] / division));

  m_grid.resize(layers, rows, cols, true, false);

  m_grid.setAll(0.0);
  
  VEE_Vector3 loc;

  VEE_trace("%s # Preparing to calculate", fname);
  
  for(i=0; i < m_grid.layers(); i++) {

    loc[2] = (float) i / (float) (m_grid.layers() - 1);

    loc[2] -= 0.5;
    loc[2] *= m_scale[2];

    for(j=0; j < m_grid.rows(); j++) {

      loc[1] = (float) j / (float) (m_grid.rows() - 1);

      loc[1] -= 0.5;
      loc[1] *= m_scale[1];
      
      for(k=0; k < m_grid.cols(); k++) {

	loc[0] = (float) k / (float) (m_grid.cols() - 1);

	loc[0] -= 0.5;
	loc[0] *= m_scale[0];

	float presence = balls->presence(loc + center, 0.7);

	if(presence > 1.0)
	  presence = 1.0;

	m_grid.get(i, j, k) = presence;
      }
    }
  }

  VEE_trace("%s # Calculation done", fname);

  m_rotation.identity();  
  m_rotationInv = m_rotation;
  m_metaBalls = balls;

  updateMarchingCubes(0.8);

  VEE_trace("%s # EXIT", fname);

  return center;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

VEE_VolumeGrid2::VEE_VolumeGrid2()
  : m_scale(1.0),
    m_voxelSize(1.0),
    m_firstLevelDivision(4),
    m_absoluteCoords(false)
{
  m_bbox.clear();
  m_rotation.identity();
  m_rotationInv.identity();
}

VEE_VolumeGrid2::VEE_VolumeGrid2(const VEE_VolumeGrid2 &that)
{
  *this = that;
}

VEE_VolumeGrid2::~VEE_VolumeGrid2()
{}

float VEE_VolumeGrid2::inside(const VEE_Vector3 &loc)
{
  VEE_Vector3 l = m_rotation * loc;
  
  int i, index[3], index2[3], lim = m_firstLevelDivision - 1;

  for(i=0; i < 3; i++) {
    float tmp = l[i] * m_scale / m_firstLevelDivision;
    index[i] = VEE_floor(tmp);
    index2[i] = VEE_min((int) ((tmp - index[i]) * m_firstLevelDivision), lim);
  }

  grid_array3::iterator iz = m_grids.find(index[2]);
  if(iz == m_grids.end())
    return 0.0;
  
  grid_array2 & grid2 = (*iz).second;

  grid_array2::iterator iy = grid2.find(index[1]);
  if(iy == grid2.end())
    return 0.0;

  grid_array & grid = (*iy).second;

  grid_array::iterator ix = grid.find(index[0]);
  if(ix == grid.end())
    return 0.0;
  
  return (*ix).second.m_grid.get(index2[2], index2[1], index2[0]);
}

void VEE_VolumeGrid2::getBoundingBox(VEE_BoundingBox *bb)
{
  *bb = m_bbox;
}

VEE_Vector3 VEE_VolumeGrid2::newRandomPoint()
{
  if(!m_gridLinear.size())
    return VEE_Vector3ZERO;

  for(uint i = 1000; i ; i--) {

    float n = VEE_rand.rand0X(m_gridLinear.size() - 0.01);
    const GridGrid * g = m_gridLinear[(uint) n];
    float t = VEE_rand.rand01();
    
    float div = m_firstLevelDivision - 0.001;
    
    float fx = VEE_rand.rand0X(div);
    float fy = VEE_rand.rand0X(div);
    float fz = VEE_rand.rand0X(div);

    float gridVal = g->m_grid.get((int) fz, (int) fy, (int) fx);

    if(gridVal > t) {
      VEE_Vector3 res(fx * m_voxelSize, fy * m_voxelSize, fz * m_voxelSize);
      res /= (float) m_firstLevelDivision;
      res += g->m_bbox.low();

      if(!m_absoluteCoords)
	res -= m_bbox.center();

      res = m_rotationInv * res;

      return res;
    }
  }

  return VEE_Vector3ZERO;
}

bool VEE_VolumeGrid2::setValue(const char *id, VEE_Value *v)
{
  bool ok = false;

  // VEE_trace("VEE_VolumeGrid2::setValue # \"%s\"", id);

  if(strcmp(id, "rotation") == 0) {
    VEE_Matrix3 tmp;
    
    ok = v->getMatrix3(&tmp);
    
    if(ok) {
      m_rotation = tmp;
      m_rotationInv = tmp;
      m_rotationInv.transpose();
    }
    // VEE_trace("VEE_VolumeGrid2::setValue # Got rotation %d", (int) ok);
  }
  else if(strcmp(id, "scale") == 0) {
    float tmp;
    
    ok = v->getFloat(&tmp);
    
    if(ok) {
      m_scale = tmp;
      m_voxelSize = 1.0 / tmp;
    }
  }
  else if(strcmp(id, "division") == 0) {
    int tmp;
    
    ok = v->getInt(&tmp);
    
    if(ok)
      m_firstLevelDivision = VEE_max(1, tmp);

  }

  return ok;
}

VEE_Volume * VEE_VolumeGrid2::clone() const
{
  return new VEE_VolumeGrid2(*this);
}

void VEE_VolumeGrid2::addBall(const VEE_Vector3 &loc, float radius)
{
  VEE_MetaBalls::Ball ball;
  ball.m_center = loc;
  ball.m_radius = radius;

  m_balls.addBall(loc, radius);

  if(m_balls.nofBalls() == 1)
    m_bbox.set(loc, radius);
  else {
    VEE_BoundingBox tmpbox;
    tmpbox.set(loc, radius);
    m_bbox.expand(tmpbox);
  }    

  VEE_BoundingBox bbox;
  bbox.set(loc, radius);

  int xlow  = VEE_floor(bbox.low()[0] * m_scale);
  int xhigh = VEE_ceil(bbox.high()[0] * m_scale);

  int ylow  = VEE_floor(bbox.low()[1] * m_scale);
  int yhigh = VEE_ceil(bbox.high()[1] * m_scale);

  int zlow  = VEE_floor(bbox.low()[2] * m_scale);
  int zhigh = VEE_ceil(bbox.high()[2] * m_scale);

  GridGrid * gptr;

  for(int x = xlow; x <= xhigh; x++) {
    for(int y = ylow; y <= yhigh; y++) {
      for(int z = zlow; z <= zhigh; z++) {

	VEE_Vector3 point(x * m_voxelSize, y * m_voxelSize, z * m_voxelSize);
	// point += bbox.low();
	float val = ball.presence(point, 0.7);

	if(val > 0.0) {
	  float & vox = getVoxel(x, y, z, &gptr);
	  
	  if(vox < val) {
	    // Modify the voxel:
	    vox = val;
	    gptr->m_dirty = true;
	  }
	}

      }
    }
  }
}

void VEE_VolumeGrid2::clear()
{
  m_balls.clear();
  m_gridLinear.clear();
  m_grids.clear();
  m_bbox.clear();
}

/** Triangulates the grid with the marching cubes. */

void VEE_VolumeGrid2::triangulate()
{
  int s = m_firstLevelDivision;

  for(uint i=0; i < m_gridLinear.size(); i++) {
    GridGrid * g = m_gridLinear[i];

    if(g->m_dirty || true) {
      g->m_triangles.clear();

      g->m_triangles.
	triangulateSubGrid(this, g->m_ind[0], g->m_ind[1], g->m_ind[2], 
			   g->m_ind[0] + s, g->m_ind[1] + s, g->m_ind[2] + s);
      
      g->m_dirty = false;
    }
  }
}

uint VEE_VolumeGrid2::triangleCount() const
{
  uint n = 0;

  for(uint i=0; i < m_gridLinear.size(); i++)
    n += m_gridLinear[i]->m_triangles.triangleCount();

  return n;
}

uint VEE_VolumeGrid2::byteCount() const
{
  uint n = sizeof(*this);

  for(uint i=0; i < m_gridLinear.size(); i++)
    n += m_gridLinear[i]->byteCount();

  return n;
}

/** Get a voxel from the super-grid. The original grid is stored in
    the gptr argument if the argument is non-null. */

float & VEE_VolumeGrid2::getVoxel(int x, int y, int z, GridGrid ** gptr)
{
  int x2 = VEE_floorDiv(x, m_firstLevelDivision);
  int y2 = VEE_floorDiv(y, m_firstLevelDivision);
  int z2 = VEE_floorDiv(z, m_firstLevelDivision);

  GridGrid * g = & m_grids[z2][y2][x2];

  if(gptr)
    *gptr = g;

  if(!g->m_touched) {
    g->m_touched = true;
    m_gridLinear.push_back(g);

    // VEE_trace("VEE_VolumeGrid2::getVoxel # New grid %p", g);
    
    float voxelDim = m_firstLevelDivision * m_voxelSize;

    VEE_Vector3 secondVoxelSize(voxelDim, voxelDim, voxelDim);
    VEE_Vector3 low(x2 * voxelDim, y2 * voxelDim, z2 * voxelDim);
    
    g->m_bbox.set(low, low + secondVoxelSize);

    g->m_grid.resize(m_firstLevelDivision, 
		     m_firstLevelDivision, 
		     m_firstLevelDivision);
    g->m_grid.setAll(0.0);

    /* if(m_gridLinear.size() == 1)
      m_bbox = g->m_bbox;
    else
    m_bbox.expand(g->m_bbox);      */

    g->m_ind[0] = x2 * m_firstLevelDivision;
    g->m_ind[1] = y2 * m_firstLevelDivision;
    g->m_ind[2] = z2 * m_firstLevelDivision;
  }

  int x3 = x - x2 * m_firstLevelDivision;
  int y3 = y - y2 * m_firstLevelDivision;
  int z3 = z - z2 * m_firstLevelDivision;

  float & vox = g->m_grid.get(z3, y3, x3);

  return vox;
}

float VEE_VolumeGrid2::getVoxel2(int x, int y, int z)
{
  int x2 = VEE_floorDiv(x, m_firstLevelDivision);
  int y2 = VEE_floorDiv(y, m_firstLevelDivision);
  int z2 = VEE_floorDiv(z, m_firstLevelDivision);

  int x3 = x - x2 * m_firstLevelDivision;
  int y3 = y - y2 * m_firstLevelDivision;
  int z3 = z - z2 * m_firstLevelDivision;

  grid_array3::iterator iz = m_grids.find(z2);
  if(iz == m_grids.end())
    return 0.0;
  
  grid_array2 & grid2 = (*iz).second;

  grid_array2::iterator iy = grid2.find(y2);
  if(iy == grid2.end())
    return 0.0;

  grid_array & grid = (*iy).second;

  grid_array::iterator ix = grid.find(x2);
  if(ix == grid.end())
    return 0.0;
  
  return (*ix).second.m_grid.get(z3, y3, x3);
}

VEE_VolumeGrid2 & VEE_VolumeGrid2::operator = (const VEE_VolumeGrid2 & that)
{
  if(this == & that)
    return *this;

  // First copy all trivial values:
  m_grids = that.m_grids;
  m_scale = that.m_scale;
  m_voxelSize = that.m_voxelSize;
  m_firstLevelDivision = that.m_firstLevelDivision;
  m_bbox        = that.m_bbox;
  m_rotation    = that.m_rotation;
  m_rotationInv = that.m_rotationInv;
  m_balls       = that.m_balls;
  m_absoluteCoords = that.m_absoluteCoords;

  m_gridLinear.clear();

  // Now rebuild the difficult stuff:
  for(grid_array3::iterator it3 = m_grids.begin(); 
      it3 != m_grids.end(); it3++) {
    for(grid_array2::iterator it2 = (*it3).second.begin(); 
	it2 !=(*it3).second.end(); it2++) {
      for(grid_array::iterator it = (*it2).second.begin(); 
	  it !=(*it2).second.end(); it++) {

	if((*it).second.m_touched)
	  m_gridLinear.push_back(& (*it).second);
	  
      }
    }
  }

  return *this;
}
