
/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */


#include <vee_vortex.h>

#include <vee_math.h>

#include <vee_error.h>
#include <vee_strs.h>
#include <vee_particle.h>

#include <vee_values.h>

#include <math.h>

VEE_Vortex::VEE_Vortex()
  : m_axes(0.0f,0.0f,1.0f), 
    m_magnitude(1.0),
    m_epsilon(0.2),
    m_range(1.0),
    m_height(1.0),
    m_fadeRange(0.7),
    m_maxForce(1e+10),
    m_strength(1.0),
    m_radialForce(1.0),
    m_minRadius(0.0),
    m_vortexed(0.0),
    m_useHeight(false)
{
  m_forceType = VEE_FORCE_DRAG;
  // VEE_trace("VEE_Vortex::VEE_Vortex # %p", this);
}

VEE_Vortex::~VEE_Vortex()
{
  // VEE_trace("VEE_Vortex::~VEE_Vortex # %p", this);
}


/** Receive and process a value. 

    Parameters:

    <UL>
    
    <LI> "axes", Vector3

    <LI> "epsilon", float, range: 0 to 1

    <LI> "magnitude", float, range: 0 to inf

    <LI> "range", float, range: 0 to inf

    <LI> "strength", float, range: 0 to 1

    <LI> "maxforce", float, range: 0 to inf

    <LI> "gravitate", float, range: 0 to inf

    </UL>

*/
bool VEE_Vortex::setValue(const char *id, VEE_Value *msg)
{
  bool ok = false;

  if(VEE_Strs::equals(id, VEE_Strs::axes)) {
    VEE_Vector3 v;

    ok = msg->getVector3(&v);

    if(ok)
      setAxes(v);
  }
  if(VEE_Strs::equals(id, VEE_Strs::epsilon)) {
    float v;

    ok = msg->getFloat(&v);

    if(ok)
      m_epsilon = v;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::magnitude)) {
    float v;

    ok = msg->getFloat(&v);

    if(ok)
      m_magnitude = v;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::range)) {
    float v;

    ok = msg->getFloat(&v);

    if(ok) {
      m_range = v;
      if(!m_useHeight)
	m_height = v;
    }
  }
  else if(VEE_Strs::equals(id, "height")) {
    float v;

    ok = msg->getFloat(&v);

    if(ok) {
      m_height = v;
      m_useHeight = true;
    }
  }
  else if(VEE_Strs::equals(id, "strength")) {
    float v;

    ok = msg->getFloat(&v);

    if(ok)
      m_strength = VEE_max(v, 1.0f);
  }
  else if(VEE_Strs::equals(id, "faderange")) {
    float v;

    ok = msg->getFloat(&v);

    if(ok)
      m_fadeRange = v;
  }
  else if(VEE_Strs::equals(id, "maxforce")) {
    float v;

    ok = msg->getFloat(&v);

    if(ok)
      m_maxForce = v;
  }
  else if(VEE_Strs::equals(id, "gravitate")) {
    float v;

    ok = msg->getFloat(&v);

    if(ok)
      m_radialForce = v;
  }
  else if(VEE_Strs::equals(id, "minradius")) {
    float v;

    ok = msg->getFloat(&v);

    if(ok)
      m_minRadius = v;
  }
  else 
    return VEE_ForceObject::setValue(id, msg);
  
  return ok;
}

VEE_Variable VEE_Vortex::getVariable(const char *id) const
{
  VEE_Variable v;

  if(VEE_Strs::equals(id, VEE_Strs::axes)) {
    v = m_axes;
  }
  if(VEE_Strs::equals(id, VEE_Strs::epsilon)) {
    v = m_epsilon;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::magnitude)) {
    v = m_magnitude;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::range)) {
    v = m_range;
  }
  else if(VEE_Strs::equals(id, "height")) {
    v = m_height;
  }
  else if(VEE_Strs::equals(id, "strength")) {
    v = m_strength;
  }
  else if(VEE_Strs::equals(id, "faderange")) {
    v = m_fadeRange;
  }
  else if(VEE_Strs::equals(id, "maxforce")) {
    v = m_maxForce;
  }
  else if(VEE_Strs::equals(id, "gravitate")) {
    v = m_radialForce;
  }
  else if(VEE_Strs::equals(id, "minradius")) {
    v = m_minRadius;
  }
  else 
    v = VEE_ForceObject::getVariable(id);
  
  return v;

}

VEE_Object * VEE_Vortex::clone() const
{
  return new VEE_Vortex(*this);
}

void VEE_Vortex::setAxes(float a1, float a2, float a3)
{
  m_axes[0] = a1;
  m_axes[1] = a2;
  m_axes[2] = a3;
  m_axes.normalize();
}

void VEE_Vortex::update(float dt)
{
  
  VEE_ForceObject::update(dt);

  m_vortexed += dt * m_magnitude;
}

void VEE_Vortex::getForce(VEE_Vector3 *force,
			 const VEE_Vector3 &velocity,
			 const VEE_Vector3 &location)
{
  VEE_Vector3 vel;

  if(m_updateVelocity)
    vel = velocity - m_velocity;
  else
    vel = velocity;

  VEE_Vector3 p = location - m_location, p2;
  float d, d2, v, v2, a, t, idealV;

  VEE_Vector3 pl = dot(p, m_axes) * m_axes;
  d2 = pl.length(); // height

  VEE_Vector3 tmp = cross(m_axes, p);
  d = tmp.length(); // Distance to axis

  p = p - pl;

  if((d>m_range) || (d2>m_height) || (location==m_location)) {
    *force = VEE_Vector3ZERO;
    return;
  }

  p.normalize();
  p2 = cross(m_axes, p);
  v = dot(p2, vel);

  float fadeRange = m_fadeRange * m_range;

  if(d < fadeRange) {
    idealV = d * m_magnitude;
    a = 3 * (idealV - v);
  }
  else {
    float tmp = (m_range - d) / (m_range - fadeRange);
    t = sqrt(tmp);
    idealV = t * d * m_magnitude;
    a = 3 * (idealV - v);
  }

  /* if(d2 > 0.5 * m_range) {
    t = sqrt((m_range - d2) / 0.5);
    a *= t;
    }*/

  v2 = dot(p, vel);
  p2.normalize();
  p2 = p2 * (v + a);

  if(idealV * 1.5 > v)
    idealV = v;

  if(d < m_epsilon) d = m_epsilon;

  p = -p * (idealV * idealV/d + v2);
    
  /* p is tangetial force, p2 is radial. */

  if(d < m_minRadius) {
    p.normalize();
    p *= 3.0f;
  }
  *force = (p2 + (p * m_radialForce)) * m_strength;

  float f = force->length();

  if(f > m_maxForce)
    (*force) *= (m_maxForce / f);
}

void VEE_Vortex::boundingBoxOfForce(VEE_BoundingBox *box)
{
  float tmp = sqrt(m_range * m_range + m_height * m_height);

  VEE_Vector3 r(tmp, tmp, tmp);

  box->low() = m_location - r;
  box->high() = m_location + r;
}

VEE_Object * VEE_Vortex::factory()
{
  return new VEE_Vortex;
}
