/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */


#include <Opcode.h>

#include <vee_grid_force.h>

#include <vee_error.h>
#include <vee_math.h>
#include <vee_scene.h>
#include <vee_strs.h>
#include <vee_value.h>

#include <gee_group.h>
#include <gee_io.h>
#include <gee_polygon_mesh.h>

#include <Solar/BoundingBox.hpp>
#include <Solar/GraphicTriangleMesh.hpp>
#include <Solar/GraphicTriangleStrips.hpp>
#include <Solar/Loaders.hpp>
#include <Solar/Model.hpp>

#include <Grid.hpp>

#include <di_endian.h>

#include <mj_array3d_tmpl.C>


// #include <Stdafx.h>


#include <OPC_RayCollider.h>

#include <errno.h>

template class MJ_Array3D<VEE_VectorGridCell>;

class __VEE_GTri {
public:
  VEE_Vector3 m_vertices[3];
  void clear() 
  { 
    m_vertices[0].clear(); 
    m_vertices[1].clear(); 
    m_vertices[2].clear(); 
  }
};

/** This is a class that hides the stuff coming from "Grid.hpp". This
    hiding is done to minimize the visibility of third-party headers
    that may be removed/renamed in fiuture. */

class VEE_VectorGrid::JanneGrid3
{
public:
  JanneGrid3() : grid(0) {}
  ~JanneGrid3() { delete grid; }

  Grid3 * grid;
  VEE_BoundingBox bbox;

  std::vector<__VEE_GTri> triangles;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

static void VEE_addTriangle(Grid3 *g, 
			    const VEE_Vector3 &v1, 
			    const VEE_Vector3 &v2, 
			    const VEE_Vector3 &v3,
			    int id = 0)
{
  Vector3 lv1 = * ((Vector3 *) & v1);
  Vector3 lv2 = * ((Vector3 *) & v2);
  Vector3 lv3 = * ((Vector3 *) & v3);

  g->rasterizeTriangle(lv1, lv2, lv3, id);
}


static void VEE_normalize(VEE_VectorGrid::grid3d * grid)
{
  VEE_VectorGridCell *vp, *vsentinel;

  vp = grid->begin();
  vsentinel = grid->end();

  while(vp < vsentinel) {
    float l = vp->m_gradient.length();
    if(l > 1.0e-5) 
      vp->m_gradient /= l;
    // vp->m_gradient *= scale;
    vp++;
  }
}

static void VEE_blurFilter(MJ_Array3D<float> * filter, uint radius)
{
  uint division = radius * 2 + 1;

  filter->resize(division,division, division);

  uint i,j ,k;
  
  float center = radius;

  float maxDist = radius + 0.99;

  for(i=0; i < division; i++) {
    for(j=0; j < division; j++) {
      for(k=0; k < division; k++) {
	
	float dx = k - center;
	float dy = j - center;
	float dz = i - center;

	float dist = sqrtf(dx * dx + dy * dy + dz * dz);
	
	float w = (maxDist - dist) / maxDist;
	if(w < 0.0) w = 0.0;

	filter->get(i, j, k) = w;
      }
    }
  }

  float sum = 0.0;

  float *v = filter->begin();
  float *sentinel = filter->end();

  while(v < sentinel)
    sum += *v++;

  float scale = 1.0 / sum;

  v = filter->begin();
  
  while(v < sentinel)
    *v++ *= scale;
}

/** Transforms triangles to the form understood by Grid3. */

static void VEE_meshToTriangles(std::vector<__VEE_GTri> &triangles,
				GEE_Group * m)
{
  // const char * fname = "VEE_meshToTriangles";

  triangles.resize(0);

  for(GEE_Group::iterator it = m->begin(); it != m->end(); it++) {

    VEE_RefPtr<GEE_Geometry>* data2 = m->getGeometry(it);

    if(!data2)
      continue;

    GEE_Geometry * data = data2->ptr();

    GEE_TriangleMesh * tris = dynamic_cast<GEE_TriangleMesh *> (data);
    GEE_PolygonMesh * polys = dynamic_cast<GEE_PolygonMesh *> (data);

    if(polys)
      tris = polys->triangleMesh();

    if(!tris)
      continue;

    for(GEE_TriangleMesh::triangleIterator tit = tris->beginTriangle(); 
	tit != tris->endTriangle(); tit++) {
      
      __VEE_GTri tmp;

      const GEE_TriangleIndices & tri = GEE_TriangleMesh::getTriangle(tit);

      for(uint i = 0; i < 3; i++)
	tmp.m_vertices[i] = tris->getVertex(tri.m_indices[i])->m_location;
      
      triangles.push_back(tmp);
    }

    /* If the triangle mesh was manufactured from polygons, then
       delete the mesh: */
    if(polys)
      delete tris;

  }
}
/** Transforms triangles to the form understood by Grid3. */

static void VEE_meshToTriangles(std::vector<__VEE_GTri> &triangles,
				Solar::Model * m)
{
  const char * fname = "VEE_meshToTriangles";

  uint i, j, k, n = m->subObjectCount();
  
  /* Cell *vp, *vp2;
     Cell *vsentinel; */

  triangles.resize(0);

  m->calculateBoundingBox();
  
  for(i=0; i < n; i++) {
    Solar::Graphic * graphic = m->getSubObject(i);

    if(graphic->material() == "inside") {
      VEE_trace("%s # Skipping an inside surface", fname);
      continue;
    }
    
    Solar::GraphicTriangleMesh  * trimesh =
      dynamic_cast<Solar::GraphicTriangleMesh*> (graphic);

    Solar::GraphicTriangleStrips  * tristrips =
      dynamic_cast<Solar::GraphicTriangleStrips *> (graphic);
    
    if(trimesh) {

      /* if(debug) {
	trimesh->recalculateBoundingBox();

	Solar::BoundingBox bbox;

	trimesh->getBoundingBox(&bbox);

	VEE_trace("%s # Reading a triangle mesh (material: %s) [%f %f %f] to [%f %f %f]", 
		  fname, graphic->material().c_str(),
		  v1[0], v1[1], v1[2], v2[0], v2[1], v2[2]);
		  }*/

      uint tris = trimesh->nofTriangles();

      for(j=0; j < tris; j++) {
	Solar::GraphicTriangleMesh::Triangle tri = trimesh->triangle(j);

	__VEE_GTri tmp;
	  
	for(k=0; k < 3; k++) {
	  /* VEE_trace("Reading vertex %d of %d", (int) tri.m_vertices[k], 
	     (int) tris); */
	  tmp.m_vertices[k] = trimesh->vertex(tri.m_vertices[k]).location();
	}

	/*
	if(debug)
	  VEE_trace
	    ("%s # Adding triangle [%f %f %f] [%f %f %f] [%f %f %f]", fname, 
	     tmp.m_vertices[0][0], tmp.m_vertices[0][1], tmp.m_vertices[0][2],
	     tmp.m_vertices[1][0], tmp.m_vertices[1][1], tmp.m_vertices[1][2],
	     tmp.m_vertices[2][0], tmp.m_vertices[2][1], tmp.m_vertices[2][2]);
	*/   

	triangles.push_back(tmp);
      }
    }
    else if(tristrips) {

      
      for(uint s = 2; s < tristrips->nofStrips(); s++) {
	
	Solar::GraphicTriangleStrips::Strip strip = tristrips->strip(s);

	uint nofVertices = strip.m_length;

	for(uint j = 2; j < nofVertices; j++) {
	  
	  __VEE_GTri tmp;
	  
	  for(k=0; k < 3; k++) {
	    tmp.m_vertices[k] = 
	      tristrips->vertex(j - k + strip.m_begin).m_location;
	  }
	  
	  triangles.push_back(tmp);
	}
      }
    }

  }
}

/* Fills a grid based triangles. */

/* static void VEE_trianglesToGrid(Grid3 *g,
				VEE_BoundingBox * bounds,
				std::vector<__VEE_GTri> &triangles)
{

  
  *bounds = bbox;
}	*/			

static void VEE_fillGrid(Grid3 **grid, uint division,
			 VEE_Vector3 *low, VEE_Vector3 *high,
			 std::vector<__VEE_GTri> &triangles,
			 bool debug)
{
  const char * fname = "VEE_fillGrid";

  // VEE_BoundingBox bbox(low, high);

  // VEE_Vector3 span = bbox.high() - bbox.low();

  // span = span * 0.1;

  uint i;

  VEE_BoundingBox bbox;

  if(triangles.size())
    bbox.set(triangles[0].m_vertices[0]);
  else
    bbox.clear();

  for(i = 0; i < triangles.size(); i++) {

    const __VEE_GTri & tmp = triangles[i];

    for(uint j=0; j < 3; j++)
      bbox.expand(tmp.m_vertices[j]);
  }


  VEE_Vector3 v1, v2, v3, v4;

  v3 = bbox.span();

  float maxEdge = VEE_max(VEE_abs(v3[0]), 
			  VEE_max(VEE_abs(v3[1]), VEE_abs(v3[2])));
						  
  v3.make(maxEdge * 0.6);

  v4 = bbox.center();

  v1 = v4 - v3;
  v2 = v4 + v3;

  *low = v1;
  *high = v2;

  if(debug)
    VEE_trace("%s # Bounds are [%f %f %f] to [%f %f %f]", fname,
	      v1[0], v1[1], v1[2], v2[0], v2[1], v2[2]);

  Grid3 *g = new Grid3(division, * ((Vector3 *) & v1), * ((Vector3 *) & v2));

  *grid = g;

  for(i = 0; i < triangles.size(); i++) {

    const __VEE_GTri & tmp = triangles[i];

    VEE_addTriangle(g, 
		    tmp.m_vertices[0], 
		    tmp.m_vertices[1], 
		    tmp.m_vertices[2], 
		    i++);
  }
  
  // int count = 0;

  /* Cell *vp, *vp2;
     Cell *vsentinel; */

  
}
			    

static void VEE_setDistances(VEE_VectorGrid::grid3d * target,
			     VEE_VectorGrid::grid3d * source,
			     MJ_Array3D<float>      * source2,
			     int maxDist)
{
  VEE_trace("VEE_setDistances # %d", maxDist);

  if(!source && !source2)
    VEE_fatal("VEE_setDistances # Either source or source2 must be set");

  int layers = (int) target->layers();
  int rows   = (int) target->rows();
  int cols   = (int) target->cols();

  float maxDim = VEE_max(layers, rows, cols);

  for(int layer = 0; layer < layers; layer++) {
    for(int row = 0; row < rows; row++) {
      for(int col = 0; col < cols; col++) {

	VEE_VectorGridCell & targetCell = target->get(layer, row, col);

	if(targetCell.m_gradient.length() <= 0.0) {
	  targetCell.m_distance = VEE_GRID_SENTINEL;
	  continue;
	}
	
	float minDist = 1.0e+10;
	bool anything = false;

	int layer2 = VEE_max(layer - maxDist, 0);

	int layer2e = VEE_min(layer + maxDist + 1, layers);
	int row2e   = VEE_min(row + maxDist + 1, rows);
	int col2e   = VEE_min(col + maxDist + 1, cols);
	
	// bool foundOne = false;

	for(; layer2 < layer2e; layer2++) {

	  int row2   = VEE_max(row - maxDist, 0);
	  for(; row2 < row2e; row2++) {

	    int col2   = VEE_max(col - maxDist, 0);
	    for(; col2 < col2e; col2++) {

	      bool onEdge = false;

	      if(source)
		onEdge = source->get(layer2, row2, col2).m_distance == 0.0;
	      else if(source2)
		onEdge = source2->get(layer2, row2, col2) != 0.0;

	      if(onEdge) {

		int dx = col - col2;
		int dy = row - row2;
		int dz = layer - layer2;
		
		float dist = 
		  sqrtf(dz * dz + dy * dy + dx * dx);
		
		if(minDist > dist) {
		  // printf("### %.3f > %.3f ", minDist, dist);
		  minDist = dist;
		  anything = true;
		}
	      }
	      
	    }
	  }
	} // One voxel

	if(anything)
	  targetCell.m_distance = minDist;
	else
	  targetCell.m_distance = VEE_GRID_SENTINEL;
	// VEE_trace("alpha set to %.3f", minDist);
	
      }
    }
  }
  
  VEE_VectorGridCell *voxel = target->begin();
  VEE_VectorGridCell *sentinel = target->end();

  float fmaxDist = maxDist + 1;

  for(; voxel < sentinel; voxel++) {
    float l = voxel->m_distance;
    if(l > fmaxDist)
      voxel->m_distance = VEE_GRID_SENTINEL;
    else if(l != VEE_GRID_SENTINEL)
      voxel->m_distance = l / maxDim;
  }
}


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

bool VEE_VectorGrid::m_debug = true;

VEE_VectorGrid::VEE_VectorGrid()
{
  m_scale = 1.0;
}

VEE_VectorGrid::~VEE_VectorGrid()
{}

/** Gets the gradient and alpha in location loc01. The location is
    expected in the original object coordinates range. */

VEE_VectorGridCell VEE_VectorGrid::getGradient
(const VEE_Vector3 &location) const
{
  VEE_VectorGridCell res;

  VEE_Vector3 loc = location;

  loc /= m_scale;
  loc = loc + 0.5;
  loc.scale(VEE_Vector3((float) m_field.cols(), 
			(float) m_field.rows(), 
			(float) m_field.layers()));

  
  if(loc[0] < 0.0 || loc[1] < 0.0 || loc[2] < 0.0 ||
     loc[0] >= m_field.cols() || 
     loc[1] >= m_field.rows() || 
     loc[2] >= m_field.layers()) {
    res.clear();
    return res;
  }

  res = m_field.get((uint) loc[2], (uint) loc[1], (uint) loc[0]);

  res.m_distance *= m_scale;

  return res;
}

VEE_VectorGridCell VEE_VectorGrid::interpolateGradient
(const VEE_Vector3 &location) const
{
  VEE_VectorGridCell res;

  VEE_Vector3 loc = location;

  loc /= m_scale;
  loc = loc + 0.5;
  loc.scale(VEE_Vector3((float) m_field.cols(), 
			(float) m_field.rows(), 
			(float) m_field.layers()));

  
  if(loc[0] < 0.0 || loc[1] < 0.0 || loc[2] < 0.0 ||
     loc[0] >= m_field.cols() - 1 || 
     loc[1] >= m_field.rows() - 1|| 
     loc[2] >= m_field.layers() - 1) {
    res.clear();
    return res;
  }
  
  uint x = (uint) loc[0];
  uint y = (uint) loc[1];
  uint z = (uint) loc[2];

  float dx = loc[0]- x;
  float dy = loc[1]- y;
  float dz = loc[2]- z;

  VEE_VectorGridCell c;
  c.clear();

  float sum = 0.0;

  float zw = 1.0 - dz;

  for(uint i=0; i < 2; i++) {

    float yw = 1.0 - dy;

    for(uint j=0; j < 2; j++) {
      
      float xw = 1.0 - dx;

      for(uint k=0; k < 2; k++) {

	float w = zw * yw * xw;

	VEE_VectorGridCell tmp = m_field.get(i + z, j + y, k + x);
	tmp.scale(w);
	c.add(tmp);

	sum += w;

	xw = dx;
      }

      yw = dy;
    }
    zw = dz;
  }
  
  c.scale(1.0 / sum);

  return c;
}

float VEE_VectorGrid::meanDistance()
{
  double sum = 0.0;

  const VEE_VectorGridCell * v = m_field.begin();
  const VEE_VectorGridCell * sentinel = m_field.end();

  while(v < sentinel) {
    sum += v->m_distance;
    v++;
  }

  return sum / m_field.elements();
}

float VEE_VectorGrid::meanGradientLength()
{

  double sum = 0.0;

  const VEE_VectorGridCell * v = m_field.begin();
  const VEE_VectorGridCell * sentinel = m_field.end();

  while(v < sentinel) {
    sum += v->m_gradient.length();
    v++;
  }

  return sum / m_field.elements();
}

  
/** Creates a grid from a model. */

bool VEE_VectorGrid::createFromModel(Solar::Model *m, 
				     uint division,
				     int blurRepeats,
				     int blurRadius, 
				     int pad)
{
  VEE_Vector3 low, high;

  JanneGrid3 grid;

  // 1) Get the stuff to the grid:

  VEE_meshToTriangles(grid.triangles, m);
  // VEE_trianglesToGrid(grid.grid, grid.triangles);

  VEE_fillGrid(&grid.grid, division, &low, &high, grid.triangles, m_debug);
  
  grid.bbox.set(low, high);

  bool ok = createFromGrid(& grid, low, high, 
			   division, blurRepeats, blurRadius, pad);

  setInsides(& grid);

  return ok;
}

bool VEE_VectorGrid::createFromModel(GEE_Group *m, 
				     uint division,
				     int blurRepeats,
				     int blurRadius,
				     int pad)
{
  VEE_Vector3 low, high;

  JanneGrid3 grid;

  // 1) Get the stuff to the grid:

  VEE_meshToTriangles(grid.triangles, m);
  // VEE_trianglesToGrid(grid.grid, grid.triangles);

  VEE_fillGrid(&grid.grid, division, &low, &high, grid.triangles, m_debug);
  
  grid.bbox.set(low, high);

  bool ok = createFromGrid(& grid, low, high, 
			   division, blurRepeats, blurRadius, pad);

  setInsides(& grid);

  return ok;
}

 
bool VEE_VectorGrid::createFromModel2(Solar::Model *m, 
				     uint division,
				     int blurRepeats,
				     int blurRadius)
{
  // const char * fname = "VEE_VectorGrid::createFromModel2";

  VEE_Vector3 low, high;

  JanneGrid3 grid;

  // 1) Get the stuff to the grid:

  VEE_meshToTriangles(grid.triangles, m);
  // VEE_trianglesToGrid(grid.grid, grid.triangles);

  VEE_fillGrid(&grid.grid, division, &low, &high, grid.triangles, m_debug);

  VEE_fillGrid(&grid.grid, division, &low, &high, grid.triangles, m_debug);

  return createFromGrid2(& grid, low, high, 
			 division, blurRepeats, blurRadius);
}

static inline void VEE_writeIntLE(FILE * f, const int *v)
{
  int r;
#ifdef WORDS_BIGENDIAN
  r = DI_byteSwap4(*v);
#else
  r = *v;
#endif
  fwrite(&r, 4, 1, f);
}

static inline int VEE_readIntLE(FILE * f)
{
  int r;

  fread(&r, 4, 1, f);
  
#ifdef WORDS_BIGENDIAN
  r = DI_byteSwap4(r);
#endif

  return r;
}

static inline void VEE_writeIntLE(FILE * f, int v)
{
  VEE_writeIntLE(f, & v);
}

static inline void VEE_writeFloatLE(FILE * f, float v)
{
  VEE_writeIntLE(f, (int *) & v);
}

static inline void VEE_writeVector3LE(FILE * f, const VEE_Vector3 & v)
{
  VEE_writeFloatLE(f, v[0]);
  VEE_writeFloatLE(f, v[1]);
  VEE_writeFloatLE(f, v[2]);
}

static inline float VEE_readFloatLE(FILE * f)
{
  int tmp = VEE_readIntLE(f);
  
  return * ((float *) & tmp);
}

static VEE_Vector3 VEE_readVector3LE(FILE * f)
{
  VEE_Vector3 v;

  v[0] = VEE_readFloatLE(f);
  v[1] = VEE_readFloatLE(f);
  v[2] = VEE_readFloatLE(f);

  return v;
}

bool VEE_VectorGrid::writeGridToFile(const char *filename)
{
  FILE * f = fopen(filename, "w");

  if(!f)
    return false;

  VEE_writeIntLE(f, m_field.layers());
  VEE_writeIntLE(f, m_field.rows());
  VEE_writeIntLE(f, m_field.cols());

  VEE_writeFloatLE(f, m_scale);

  const VEE_VectorGridCell * v = m_field.begin();
  const VEE_VectorGridCell * sentinel = m_field.end();

  while(v < sentinel) {
    VEE_writeVector3LE(f, v->m_gradient);
    VEE_writeFloatLE(f, v->m_distance);
    v++;
  }

  fclose(f);

  return true;
}

bool VEE_VectorGrid::readGridFromFile(const char *filename)
{
  FILE * f = fopen(filename, "r");

  if(!f)
    return false;

  int layers = VEE_readIntLE(f);
  int rows = VEE_readIntLE(f);
  int cols = VEE_readIntLE(f);

  m_scale = VEE_readFloatLE(f);
  
  uint elements = layers * rows * cols;

  uint bytes = elements * sizeof(VEE_Vector4);

  VEE_trace("VEE_VectorGrid::readGridFromFile # Size %d %d %d -> %u kb",
	    layers, rows, cols, bytes >> 10);
  
  m_field.resize(layers, rows, cols);

  VEE_VectorGridCell * v = m_field.begin();
  VEE_VectorGridCell * sentinel = m_field.end();

  while(v < sentinel) {
    v->m_gradient = VEE_readVector3LE(f);
    v->m_distance = VEE_readFloatLE(f);
    v++;
  }

  fclose(f);
  
  VEE_trace("VEE_VectorGrid::readGridFromFile # Dist ave = %f, grad ave = %f",
	    meanDistance(), meanGradientLength());

  return true;
}

/** Stores the field into an ASCII file. Use Matlab or something to
    load the file and plot the contents to the screen.*/

bool VEE_VectorGrid::writeAsciiFile
(const char *filename, int selectLayer) const
{
  FILE * f = fopen(filename, "w");
  
  if(!f) {
    VEE_error("VEE_VectorGrid::writeMatlabFile # "
	      "Could not open file \"%s\" because of %s",
	      filename, strerror(errno));
    errno = 0;
    return false;
  }

  fprintf(f, "SIZE %u %u %u\n", 
	  m_field.layers(), m_field.rows(), m_field.cols());

  for(uint layer = 0; layer < m_field.layers(); layer++) {

    if(selectLayer >= 0)
      layer = selectLayer;

    for(uint row = 0; row < m_field.rows(); row++) {
      for(uint col = 0; col < m_field.cols(); col++) {
	const VEE_VectorGridCell & c = m_field.get(layer, row, col);
	fprintf(f, "[ %e %e %e %e ] ", 
		c.m_gradient.x, c.m_gradient.y, c.m_gradient.z, c.m_distance);
      }
      // fprintf(f, "\n");
    }

    if(selectLayer >= 0)
      layer = m_field.layers();
  }

  return true;
}

bool VEE_VectorGrid::fromModelToFile(const char * fieldFile,
				     const char * solarModel, 
				     int division,
				     int blurRepeats,
				     int blurRadius,
				     int method)
{
  const char * fname = "VEE_VectorGrid::fromModelToFile";

  if(m_debug)
    VEE_trace("%s # Loading file \"%s\"", fname, solarModel);

  Solar::Model m;

  bool ok = Solar::Loaders::load(&m , solarModel);

  if(!ok) {
    VEE_error("%s # Could not load file \"%s\"", fname, solarModel);
    return false;
  }

  if(m_debug)
    VEE_trace("%s # Creating a grid from model", fname);

  VEE_VectorGrid grid;

  if(method == 0)
    ok = grid.createFromModel(&m, division, blurRepeats, blurRadius);
  else
    ok = grid.createFromModel2(&m, division, blurRepeats, blurRadius);

  if(!ok) {
    VEE_error("%s # Failed to turn the model \"%s\" into a grid", 
	      fname, solarModel);
    return false;
  }

  if(m_debug)
    VEE_trace("%s # Writing the grid to file \"%s\"", fname, fieldFile);

  ok = grid.writeGridToFile(fieldFile);

  if(!ok) {
    VEE_error("%s # Failed to save the field grid into \"%s\"", 
	      fname, fieldFile);
    return false;
  }

  VEE_trace("%s # Converted model from file \"%s\" to grid file \"%s\"",
	    fname, solarModel, fieldFile);
  
  return true;
}

bool VEE_VectorGrid::createFromGrid(JanneGrid3 *jgrid,
				    const VEE_Vector3 & low, 
				    const VEE_Vector3 & high,
				    uint division,
				    int blurRepeats,
				    int blurRadius,
				    int pad)
{
  const char * fname = "VEE_VectorGrid::createFromGrid";
  
  Grid3 * grid = jgrid->grid;
  std::vector<__VEE_GTri> & triangles = jgrid->triangles;

  grid3d edges;
  VEE_VectorGridCell *vp, *vp2;
  VEE_VectorGridCell *vsentinel;

  uint i, j, k;
  
  // 2) Now transform the stuff from Janne's grid to mine:

  // Surround the object with free space
  int offset = pad >= 0 ? pad : blurRepeats * blurRadius ;
  int offset2 = offset * 2;

  edges.resize(division + offset2, division + offset2, division + offset2);

  // Get the scale to be used when returning to original coordinates.
  m_scale = high.x - low.x;
  float padding = (float) offset2 / (float) division;
  m_scale *= (1.0 + padding);
  
  VEE_Vector3 v1;

  v1.clear();

  VEE_VectorGridCell c;
  c.m_gradient.clear();
  c.m_distance = 0.0;

  edges.setAll(c);

  for(i=0; i < division; i++) {
    for(j=0; j < division; j++) {
      for(k=0; k < division; k++) {
	
	Vector3i cellId;
	
	cellId.make(k, j, i);
	
	Grid3::Cell * c = & grid->getCell(cellId);
	
	VEE_Vector3 normal;
	normal.clear();
	float l = VEE_GRID_SENTINEL;

	for(int t = 0; t < c->getTriangleCount(); t++) {
	  __VEE_GTri tri = triangles[c->getTriangle(t)];
	  
	  normal += cross(tri.m_vertices[1] - tri.m_vertices[0],
			  tri.m_vertices[2] - tri.m_vertices[0]);
	  l = 0.0;
	}

	/* VEE_Vector3 normal = tri.m_vertices[0] + tri.m_vertices[1] + 
	   tri.m_vertices[2]; */

	normal.normalize();
	VEE_VectorGridCell & cell = 
	  edges.get(i + offset, j + offset, k + offset);

	cell.m_gradient = normal;
	cell.m_distance = l;
      }
    }
  }
  
  // 3) Do the filtering to blur the object edges.

  m_field = edges;

  /* uint blurRepeats = 0;
     int blurRadius  = 4; */
  
  grid3d res;
  res = m_field;

  // int nmax = (int) res.layers();

  VEE_Vector3 cellSize ;
  cellSize.make(m_scale);
  cellSize[0] /= m_field.cols();
  cellSize[1] /= m_field.rows();
  cellSize[2] /= m_field.layers();

  MJ_Array3D<float> filter;
    
  VEE_blurFilter(& filter, blurRadius);

  for(i=0; i < (uint) blurRepeats; i++) {

    VEE_trace("%s # Blurring %u of %u", fname, i + 1, blurRepeats);

    res.setAll(c);

    for(int z = 0; z < (int) m_field.layers(); z++) {

      printf("."); fflush(0);

      for(int y = 0; y < (int) m_field.rows(); y++) {
	for(int x = 0; x < (int) m_field.cols(); x++) {
	  
	  VEE_VectorGridCell & cell = res.get(z, y, x);

	  VEE_Vector3 & target = cell.m_gradient;

	  target.clear();

	  VEE_Vector3 location((float) x, (float) y, (float) z);
	  location.scale(cellSize);
      
	  // Filter one voxel:
	  for(int zz = -blurRadius; zz <= blurRadius; zz++) {
	    for(int yy = -blurRadius; yy <= blurRadius; yy++) {
	      for(int xx = -blurRadius; xx <= blurRadius; xx++) {
		
		int ix = x + xx;
		int iy = y + yy;
		int iz = z + zz;

		if(ix < 0 || ix >= (int) m_field.cols() ||
		   iy < 0 || iy >= (int) m_field.rows() ||
		   iz < 0 || iz >= (int) m_field.layers())
		  continue;

		int ii = xx + blurRadius;
		int jj = yy + blurRadius;
		int kk = zz + blurRadius;

		float w = filter.get(kk, jj, ii);
		  
		VEE_VectorGridCell & cell2 = m_field.get(iz, iy, ix);
		
		VEE_Vector3 location2((float) ix, (float) iy, (float) iz);
		location2.scale(cellSize);
		
		target += (cell2.m_gradient * w);
	      }
	    }
	  } // One voxel
	  
	}
      }
    } // One grid

    m_field = res;
    
    /* Add the original edge data to make sure the edges are and
       remain somewhat sharp. */
    // if(i + 1 < (uint) blurRepeats) {
    {
      vp = m_field.begin();
      vsentinel = m_field.end();
      
      vp2 = edges.begin();
      
      while(vp < vsentinel) {
	if(vp2->m_distance == 0.0) {
	  vp->m_gradient = vp2->m_gradient;
	  vp->m_distance = vp2->m_distance;
	}
	vp++;
	vp2++;
      }
      
    }
    
  } // One blur round


  // 4) Normalize the force field

  VEE_normalize(& m_field);

  VEE_setDistances(& m_field, &edges, 0, (blurRepeats * blurRadius) + 1);

  // delete grid;

  return true;
}

bool VEE_VectorGrid::createFromGrid2(JanneGrid3 *jgrid,
				     const VEE_Vector3 & low, 
				     const VEE_Vector3 & high,
				     uint division,
				     int blurRepeats,
				     int blurRadius)
{
  const char * fname = "VEE_VectorGrid::createFromGrid2";

  uint i, j, k;

  Grid3 * grid = jgrid->grid;

  int offset = blurRepeats * blurRadius;
  int offset2 = offset * 2;

  MJ_Array3D<float> edges, potential2, potential3;

  edges.resize(division + offset2, division + offset2, division + offset2);
  
  m_scale = high.x - low.x;
  float padding = (float) offset2 / (float) division;
  m_scale *= (1.0 + padding);

  // 2) Now transform the stuff from Janne's grid to potential field:
  
  edges.setAll(0);

  potential2 = edges;

  for(i=0; i < division; i++) {
    for(j=0; j < division; j++) {
      for(k=0; k < division; k++) {
	
	Vector3i cellId;
	
	cellId.make(k, j, i);
	
	Grid3::Cell * c = & grid->getCell(cellId);
	
	if(c->getTriangleCount())
	  edges.get(i + offset, j + offset, k + offset) = 1.0;
      }
    }
  }
  
  MJ_Array3D<float> filter;
    
  VEE_blurFilter(& filter, blurRadius);

  // 3) Blur the potential field
  
  int nmax = (int) potential2.layers();

  potential3 = edges;

  for(i=0; i < (uint) blurRepeats; i++) {

    potential2.setAll(0);

    VEE_trace("%s # Blurring %u of %u", fname, i + 1, blurRepeats);

    for(int z = 0; z < (int) potential3.layers(); z++) {

      printf("."); fflush(0);

      for(int y = 0; y < (int) potential3.rows(); y++) {
	for(int x = 0; x < (int) potential3.cols(); x++) {
	  
	  
	  for(int zz = -blurRadius; zz <= blurRadius; zz++) {
	    for(int yy = -blurRadius; yy <= blurRadius; yy++) {
	      for(int xx = -blurRadius; xx <= blurRadius; xx++) {
		
		int ix = x + xx;
		int iy = y + yy;
		int iz = z + zz;
		
		
		int ii = xx + blurRadius;
		int jj = yy + blurRadius;
		int kk = zz + blurRadius;

		float w = filter.get(kk, jj, ii);

		if(ix >= 0 && ix < nmax && 
		   iy >= 0 && iy < nmax && 
		   iz >= 0 && iz < nmax) 
		  potential2.get(z, y, x) += w * potential3.get(iz, iy, ix);
		

	      }
	    }
	  } // One voxel
	  
	  
	}
      }
    } // One grid

    potential3 = potential2;

    // Add the the old stuff:
    float * fp = potential3.begin();
    float * sentinel = potential3.end();

    float * fp2 = edges.begin();

    while(fp < sentinel) {
      if(*fp2 > 0.0)
	*fp = 1.0;
      fp++;
      fp2++;
    }
    
  } // One blur round

  // 4) Convert the potential data to a gradient grid;
  
  VEE_VectorGridCell c;
  c.m_gradient.clear();
  c.m_distance = VEE_GRID_SENTINEL;

  m_field.resize(potential2.layers(), potential2.rows(), potential2.cols());

  m_field.setAll(c);
  
  for(int z = 1; z + 1< (int) potential2.layers(); z++) {
    for(int y = 1; y + 1 < (int) potential2.rows(); y++) {
      for(int x = 1; x + 1< (int) potential2.cols(); x++) {

	float dx = potential2.get(z, y, x+1) - potential2.get(z, y, x-1);
	float dy = potential2.get(z, y+1, x) - potential2.get(z, y-1, x);
	float dz = potential2.get(z+1, y, x) - potential2.get(z-1, y, x);
	
	c.m_gradient.make(-dx, -dy, -dz);
	
	m_field.get(z, y, x) = c;

      }
    }
  }
  
  // 5) Normalize the results
    
  VEE_normalize(&m_field);
  VEE_setDistances(& m_field, 0, &edges, (blurRepeats * blurRadius) + 1);

  return true;
}

/** Sets the inside / outside information of each cell. */

void VEE_VectorGrid::setInsides(JanneGrid3 *jg)
{
  VEE_trace("VEE_VectorGrid::setInsides # Enter");

  // First construct the relevant intersection test setup with OPCODE

  std::vector<__VEE_GTri> & triangles = jg->triangles;
  std::vector<uint> indices;
  indices.resize(triangles.size() * 3);

  for(uint i=0; i < indices.size(); i++)
    indices[i] = i;
  
  
  Opcode::OPCODECREATE   opcCreate;
  Opcode::MeshInterface  meshinterface;
  Opcode::CollisionFaces collisionFaces;
  Opcode::RayCollider    rayCollider;
  Opcode::Model          opcModel;
  
  meshinterface.SetPointers
    ((ICEMATHS_API IndexedTriangle *) & indices[0],
     (ICEMATHS_API Point *) & triangles[0]);

  meshinterface.SetNbTriangles(triangles.size());
  meshinterface.SetNbVertices(indices.size());

  opcCreate.mIMesh = & meshinterface;

  opcModel.Build(opcCreate);

  rayCollider.SetCulling(false);
  rayCollider.SetClosestHit(true);
  rayCollider.SetDestination(& collisionFaces);
  
  // Now check for each voxel if it is inside the object or not:

  ICEMATHS_API Ray ray;

  VEE_Vector3 low = jg->bbox.low();
  VEE_Vector3 scale = jg->bbox.span();
  
  scale[0] /= m_field.cols();
  scale[1] /= m_field.rows();
  scale[2] /= m_field.layers();

  uint insides = 0, outsides = 0;

  for(uint layer = 0; layer < m_field.layers(); layer++) {
    for(uint row = 0; row < m_field.rows(); row++) {
      for(uint col = 0; col < m_field.cols(); col++) {
	
	VEE_VectorGridCell & cell = m_field.get(layer, row, col);

	if(cell.m_distance <= VEE_GRID_SENTINEL * 0.1)
	  continue;

	VEE_Vector3 centerLoc(col, row, layer);
	centerLoc.scale(scale);
	centerLoc += low;

	ray.mOrig = * ( ICEMATHS_API Point *) & centerLoc;
	ray.mDir = * ( ICEMATHS_API Point *) & cell.m_gradient;
	
	rayCollider.Collide(ray, opcModel);
	
	bool contact = rayCollider.GetContactStatus();

	if(!contact) {
	  outsides++;
	  continue;
	}

	int face = collisionFaces.GetFaces()->mFaceID;
	const __VEE_GTri & tri = triangles[face];
	VEE_Vector3 faceNormal = 
	  cross(tri.m_vertices[1] - tri.m_vertices[0],
		tri.m_vertices[2] - tri.m_vertices[0]);

	float dir = dot(faceNormal, cell.m_gradient);
	if(dir > 0.0) {
	  // inside (since normal and gradient to same direction)
	  insides++;
	  cell.m_distance *= -1.0;
	}
	else
	  outsides++;
      }
    }
  }
  
  VEE_trace("VEE_VectorGrid::setInsides # %u voxels inside, %u outside", 
	    insides, outsides);
}
  
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

VEE_VectorGridTree::Branch::Branch()
  : m_leafs(0)
{}

VEE_VectorGridTree::Branch::~Branch()
{
  delete m_leafs;
}
  
/** Tries to collapse this branch (remove all children in favor of
    one average value. */
  
bool VEE_VectorGridTree::Branch::tryCollapse(float maximumVectorSpread)
{
  //const char * fname = "VEE_VectorGridTree::Branch::tryCollapse";

  if(!m_leafs)
    return true;

  // First we calculate average direction
  uint i, n = m_leafs->elements();

  VEE_Vector3 average(0, 0, 0);

  bool anything = false;

  for(i=0; i < n; i++) {
    const VEE_Vector4 & plane = m_leafs->begin()[i].m_plane;

    if(plane.w != VEE_GRID_SENTINEL) {
      average += plane.vector3();
      anything = true;
    }
  }

  if(!anything) {
    delete m_leafs;
    m_leafs = 0;
    return true;
  }

  average.normalize();

  /* Now check if there are cells that are too far from the average
     value. Also seek the most average voxel to represent the whole
     supervoxel. */

  float bestMatch = VEE_GRID_SENTINEL;
  int bestIndex = -1;

  for(i=0; i < n; i++) {
    VEE_Vector4 &v = m_leafs->begin()[i].m_plane;

    if(v.w == VEE_GRID_SENTINEL)
      continue;

    float variation = (v.vector3() - average).length();

    if(variation > maximumVectorSpread) {
      /* VEE_trace("%s # Cannot collapse (%f > %f - %f %f %f %f)", 
		fname, variation, maximumVectorSpread, 
		v.x, v.y, v.z, v.w); */
      return false;
    }

    if(variation < bestMatch) {
      bestMatch = variation;
      bestIndex = i;
    }
  }

  // VEE_trace("%s # Collapsed %p", fname, this);

  // We got through -> Can be collapsed
  
  delete m_leafs;
  m_leafs = 0;
  
  if(bestIndex != -1)
    m_topLeaf = m_leafs->begin()[bestIndex];
  else
    m_topLeaf.clear();
  
  return true;
}

uint VEE_VectorGridTree::Branch::bytesUsed() const
{
  uint bytes = sizeof(*this);

  /* uint i, n = m_children.elements();

  for(i=0; i < n; i++)
    bytes += m_children.begin()[i].bytesUsed();
  */

  if(m_leafs)
    bytes += m_leafs->elements() * sizeof(Leaf) + sizeof(*m_leafs);
  
  return bytes;
}

bool VEE_VectorGridTree::Branch::write(GEE_Io *io)
{
  /* io->writeInt32(m_children.rows());
     io->writeInt32(m_children.cols());
     io->writeInt32(m_children.layers());

  uint i, n = m_children.elements();

  for(i=0; i < n; i++)
    m_children.begin()[i].write(io);
  */

  if(m_leafs) {

    io->writeInt32(m_leafs->rows());
    io->writeInt32(m_leafs->cols());
    io->writeInt32(m_leafs->layers());
    
    uint n = m_leafs->elements();
    
    if(n)
      io->writeFloat32v(& m_leafs->begin()->m_plane[0], n * 4);
  }
  else
    io->writeInt32(0);
    
  io->writeVector4(m_topLeaf.m_plane);
  
  return true;
}

bool VEE_VectorGridTree::Branch::read(GEE_Io *io)
{
  /* uint rows = io->readInt32();
  uint cols = io->readInt32();
  uint layers = io->readInt32();

  m_children.resize(rows, cols, layers);

  uint i, n = m_children.elements();

  for(i=0; i < n; i++)
    m_children.begin()[i].read(io);
  */
  uint rows = io->readInt32();

  if(rows) {

    uint cols = io->readInt32();
    uint layers = io->readInt32();
    
    resize(rows, cols, layers);
    
    uint n = m_leafs->elements();
    
    if(n)
      io->readFloat32v(& m_leafs->begin()->m_plane[0], n * 4);
    
    m_topLeaf.m_plane = io->readVector4();
  }

  return true;
}

VEE_VectorGridTree::VEE_VectorGridTree()
{}

VEE_VectorGridTree::~VEE_VectorGridTree()
{}

/// Returns the number of bytes this object takes.
uint VEE_VectorGridTree::bytesUsed() const
{
  uint bytes = sizeof(*this);
  
  uint i, n = m_tree.elements();

  for(i=0; i < n; i++)
    bytes += m_tree.begin()[i].bytesUsed();

  return bytes;
}

uint VEE_VectorGridTree::lowestVoxels() const
{
  uint count = 0, n = m_tree.elements();

  for(uint i=0; i < n; i++) {
    const Branch & br = m_tree.begin()[i];
    if(br.m_leafs)
      count += br.m_leafs->elements();
  }

  return count;
}

/** Builds a plane tree out of homogenously sampled vector grid.

*/

void VEE_VectorGridTree::build(VEE_VectorGrid * source, 
			       uint lowestDivision, 
			       float maximumVectorSpread)
{
  MJ_Array3D<Branch> tmp;
  
  uint division = source->rows();
  
  uint secondDivision = lowestDivision;
  uint firstDivision = division / secondDivision;

  if(firstDivision * secondDivision < division)
    firstDivision++;

  m_totalVoxelDim = firstDivision * secondDivision;
  m_voxelSize = source->scale() / division;
  m_bottomDivision = secondDivision;

  // Clear grid, (Also deletes all old data)
  m_tree.resize(0, 0, 0); 
  m_tree.resize(firstDivision, firstDivision, firstDivision);

  uint i, n = m_tree.elements();

  for(i=0; i < n; i++)
    m_tree.begin()[i].resize(secondDivision, secondDivision, secondDivision);

  // uint x, y, z, x2, y2, z2;

  // First we fill the tree grid:

  for(uint z = 0; z < division; z++) {
    for(uint y = 0; y < division; y++) {
      for(uint x = 0; x < division; x++) {

	uint x2 = x / secondDivision;
	uint y2 = y / secondDivision;
	uint z2 = z / secondDivision;

	uint x3 = x % secondDivision;
	uint y3 = y % secondDivision;
	uint z3 = z % secondDivision;
	
	const VEE_VectorGridCell & cell = source->grid().getSafe(z, y, x);
	VEE_Vector3 normal = cell.m_gradient;

	// printf("cell = %f %f %f %f\n", normal.x, normal.y, normal.z, cell.m_distance);

	if(cell.m_distance == VEE_GRID_SENTINEL)
	  continue;

	// float centerDistance = voxelNorma
	
	VEE_Vector3 voxelLocation(x, y, z);
	voxelLocation -= (0.5f * division);
	voxelLocation *= source->scale();
	float w = -dot(normal, voxelLocation);
	
	w -= cell.m_distance * source->scale();
	
	m_tree.getSafe(z2, y2, x2).m_leafs->getSafe(z3, y3, x3).m_plane.make
	  (normal.x, normal.y, normal.z, w);
      }
    }
  }
  
  // Now try to collapse all unrelevant areas:
  for(i=0; i < n; i++)
    m_tree.begin()[i].tryCollapse(maximumVectorSpread);
}

bool VEE_VectorGridTree::write(const char * filename)
{
  GEE_File file;
  if(!file.open(filename, DI_IO_OUTPUT))
    return false;

  return write(&file);
}

bool VEE_VectorGridTree::read(const char * filename)
{
  GEE_File file;
  if(!file.open(filename, DI_IO_INPUT))
    return false;

  return read(&file);
}

bool VEE_VectorGridTree::write(GEE_Io *io)
{  
  io->writeInt32(m_tree.rows());
  io->writeInt32(m_tree.cols());
  io->writeInt32(m_tree.layers());

  uint i, n = m_tree.elements();

  for(i=0; i < n; i++)
    m_tree.begin()[i].write(io);

  return true;
}

bool VEE_VectorGridTree::read(GEE_Io *io)
{
  uint rows = io->readInt32();
  uint cols = io->readInt32();
  uint layers = io->readInt32();

  m_tree.resize(rows, cols, layers);

  uint i, n = m_tree.elements();

  for(i=0; i < n; i++)
    m_tree.begin()[i].read(io);
  
  return true;
}

const VEE_Vector4 & VEE_VectorGridTree::getPlane(const VEE_Vector3 &loc)
{
  VEE_Vector3 voxLoc = loc / m_voxelSize + 0.5f * (float) m_totalVoxelDim;

  float voxLim = m_totalVoxelDim;

  if(voxLoc.x >= voxLim || voxLoc.x < 0.0f ||
     voxLoc.y >= voxLim || voxLoc.y < 0.0f ||
     voxLoc.z >= voxLim || voxLoc.z < 0.0f)
    return m_zeroLeaf.m_plane;

  uint x = (uint) voxLoc.x;
  uint y = (uint) voxLoc.y;
  uint z = (uint) voxLoc.z;

  uint x2 = x / m_bottomDivision;
  uint y2 = y / m_bottomDivision;
  uint z2 = z / m_bottomDivision;
  
  const Branch & branch = m_tree.getSafe(z2, y2, x2);

  if(branch.m_leafs == 0)
    return branch.m_topLeaf.m_plane;

  uint x3 = x % m_bottomDivision;
  uint y3 = y % m_bottomDivision;
  uint z3 = z % m_bottomDivision;

  return branch.m_leafs->getSafe(z3, y3, x3).m_plane;
}

template class MJ_Array3D<VEE_VectorGridTree::Branch>;
template class MJ_Array3D<VEE_VectorGridTree::Leaf>;

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

VEE_GridForceBase::VEE_GridForceBase()
  : m_magnitude(1.0),
    m_tangentForce(0.0),
    m_tangentGrab(0.0),
    m_bounce(0.0),
    m_drag(0.0),
    m_dragPower(2.0),
    m_bounceDistance(1.0),
    m_flowDistance(1.0),
    m_dragDistance(0.3),
    m_frameTime(0.5),
    m_impactMethod(true),
    m_onlyImpact(false)
{
  m_scale = 1.0;
  // m_offset = 0.0;
  m_rotation.identity();
  m_rotationInv = m_rotation;
}

VEE_GridForceBase::~VEE_GridForceBase()
{}

void VEE_GridForceBase::getForce(VEE_Vector3 *,
				 const VEE_Vector3 &,
				 const VEE_Vector3 &)
{
  VEE_error("VEE_GridForceBase::getForce # DO NOT CALL");
}

void VEE_GridForceBase::boundingBoxOfForce(VEE_BoundingBox *)
{
  VEE_error("VEE_GridForceBase::boundingBoxOfForce # DO NOT CALL");
}

bool VEE_GridForceBase::setValue(const char *id, VEE_Value *v)
{
  bool ok = VEE_ForceObject::setValue(id, v);
  
  if(ok) return true;

  if(strcmp(id, "rotation") == 0) {
    VEE_Matrix3 tmp;
    
    ok = v->getMatrix3(&tmp);
    
    if(ok)
      setRotation(tmp);
  }
  else if(strcmp(id, VEE_Strs::magnitude) == 0) {
    float tmp;
    
    ok = v->getFloat(&tmp);
    
    if(ok)
      m_magnitude = tmp;
  }
  else if(strcmp(id, "scalemul") == 0) {
    float tmp;
    
    ok = v->getFloat(&tmp);
    
    if(ok)
      m_scale *= tmp;
  }
  else if(strcmp(id, "scale") == 0) {
    float tmp;
    
    ok = v->getFloat(&tmp);
    
    if(ok)
      m_scale = tmp;
  }
  else if(strcmp(id, "bounce") == 0) {
    float tmp;
    
    ok = v->getFloat(&tmp);
    
    if(ok)
      m_bounce = tmp;
  }
  else if(strcmp(id, "drag") == 0) {
    float tmp;
    
    ok = v->getFloat(&tmp);
    
    if(ok)
      m_drag = tmp;
  }
  else if(strcmp(id, "dragpower") == 0) {
    float tmp;
    
    ok = v->getFloat(&tmp);
    
    if(ok)
      m_dragPower = tmp;
  }
  else if(strcmp(id, "tangentforce") == 0) {
    float tmp;
    
    ok = v->getFloat(&tmp);
    
    if(ok)
      m_tangentForce = tmp;
  }
  else if(strcmp(id, "tangentgrab") == 0) {
    float tmp;
    
    ok = v->getFloat(&tmp);
    
    if(ok)
      m_tangentGrab = tmp;
  }
  else if(strcmp(id, "bouncedistance") == 0) {
    float tmp;
    
    ok = v->getFloat(&tmp);
    
    if(ok)
      m_bounceDistance = tmp;
  }
  else if(strcmp(id, "flowdistance") == 0) {
    float tmp;
    
    ok = v->getFloat(&tmp);
    
    if(ok)
      m_flowDistance = tmp;
  }
  else if(strcmp(id, "dragdistance") == 0) {
    float tmp;
    
    ok = v->getFloat(&tmp);
    
    if(ok)
      m_dragDistance = tmp;
  }
  else if(strcmp(id, "impact") == 0) {
    int tmp;
    
    ok = v->getInt(&tmp);
    
    if(ok)
      m_impactMethod = tmp != 0;
  }
  
  return ok;
}

VEE_Variable VEE_GridForceBase::getVariable(const char *id) const
{
  VEE_Variable v = VEE_ForceObject::getVariable(id);

  if(v.type() != VEE_VAR_INVALID)
    return v;

  if(strcmp(id, "rotation") == 0)
    v = m_rotation;
  else if(strcmp(id, VEE_Strs::magnitude) == 0)
    v = m_magnitude;
  else if(strcmp(id, "bounce") == 0)
    v = m_bounce;
  else if(strcmp(id, "drag") == 0)
    v = m_drag;
  else if(strcmp(id, "scale") == 0)
    v = m_scale;
  else if(strcmp(id, "dragpower") == 0)
    v = m_dragPower;  
  else if(strcmp(id, "tangentforce") == 0)
    v = m_tangentForce;
  else if(strcmp(id, "tangentgrab") == 0)
    v = m_tangentGrab;
  else if(strcmp(id, "bouncedistance") == 0)
    v = m_bounceDistance;
  else if(strcmp(id, "flowdistance") == 0)
    v = m_flowDistance;
  else if(strcmp(id, "dragdistance") == 0)
    v = m_dragDistance;
  else if(strcmp(id, "impact") == 0)
    v = (int) m_impactMethod;

  return v;
}


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

VEE_GridForce::VEE_GridForce()
{}

VEE_GridForce::~VEE_GridForce()
{}

void VEE_GridForce::getForce(VEE_Vector3 *force,
			     const VEE_Vector3 &velocity,
			     const VEE_Vector3 &location)
{
  VEE_Vector3 loc = location - m_location;
  loc = m_rotationInv * loc;

  loc /= m_scale;
  
  VEE_VectorGridCell cell = (*m_grid).getGradient(loc);
  
  if(cell.m_distance <= 0.0f) {
    force->clear();
    return;
  }
  
  VEE_Vector3 vel = m_rotationInv * velocity;

  float hitRatio = dot(cell.m_gradient, vel);

  // float minMag = m_magnitude * 0.1;

  VEE_Vector3 localForce;

  if(!m_impactMethod && cell.m_distance < m_bounceDistance) {

    float magnitude;
    float balpha = 1.0f - cell.m_distance / m_bounceDistance;

    if(hitRatio < 0.0) {
      // Coming in
      magnitude = m_magnitude * balpha;
    }
    else {
      float balpha = 1.0 - cell.m_distance / m_bounceDistance;

      // Going out:
      magnitude = m_magnitude * balpha * m_bounce;

    }
    localForce = cell.m_gradient * magnitude;
  }
  else if(cell.m_distance < m_bounceDistance * 0.1)
    localForce = cell.m_gradient * m_magnitude * 0.05;
  else
    localForce.clear();

  if((m_tangentForce != 0.0 || m_tangentGrab != 0) && 
     cell.m_distance < m_flowDistance) {

    // float talpha = 1.0f - cell.m_distance / m_flowDistance;

    VEE_Vector3 velocityNormal = 
      cross(cross(cell.m_gradient, vel), vel);

    velocityNormal.normalize();
    
    float tangentForce;

    if(hitRatio < 0.0)
      tangentForce = hitRatio * m_tangentForce;
    else
      tangentForce = hitRatio * m_tangentGrab;

    localForce += velocityNormal * tangentForce/* * talpha*/;
  }

  // Add the drag component if necessary:
  if(m_drag > 0.0 && cell.m_distance < m_dragDistance) {
    
    float dalpha = 1.0f - cell.m_distance / m_dragDistance;

    // VEE_Vector3 tangent = 
    // cross(cross(velocity, cell.m_gradient), cell.m_gradient);
    
    VEE_Vector3 tangent = velocity - hitRatio * cell.m_gradient;

    tangent.normalize();
    
    float tangentVelocity = dot(tangent, velocity);
    
    float drag = powf(VEE_abs(tangentVelocity), m_dragPower) * m_drag;

    localForce -= (tangent * drag * dalpha);
  }

  *force = m_rotation * localForce;
}

void VEE_GridForce::boundingBoxOfForce(VEE_BoundingBox *bb)
{
  bb->clear();

  VEE_Vector3 tmp;
  tmp.make(m_scale * 0.5);
  VEE_calculateCubeBounds(bb, m_rotation, tmp);

  bb->move(m_location);
}

void VEE_GridForce::update(float dt)
{
  m_frameTime = dt;
  VEE_ForceObject::update(dt);
  
  VEE_BoundingBox bb;

  boundingBoxOfForce(&bb);

  VEE_SceneBox *scene = m_sceneBox;

  while(!scene->boundingBox().contains(bb) &&
	scene->parent()) {
    scene = scene->parent();
  }
  recursiveImpact(scene, &bb);
}

VEE_Object * VEE_GridForce::clone() const
{
  // VEE_fatal("VEE_GridForce::clone # %p", this);
  return new VEE_GridForce(*this);
}

void VEE_GridForce::recursiveImpact(VEE_SceneBox *scene, VEE_BoundingBox *bb)
{
  for(VEE_SceneBox::iterator it = scene->begin(); it != scene->end(); it++) {

    VEE_Object * o = VEE_SceneBox::getObject(it);

    VEE_Vector3 loc = o->location() - m_location;
    loc = loc * m_rotation;
    
    loc /= m_scale;
    
    VEE_VectorGridCell cell = (*m_grid).getGradient(loc);
        
    if(cell.m_distance <= VEE_GRID_SENTINEL * 0.001)
      continue;

    /* if(cell.m_distance < 0.0)
       VEE_trace("VEE_GridForce::recursiveImpact # %f", cell.m_distance); */

    cell.m_distance *= m_scale;

    /* VEE_trace("VEE_GridForce::recursiveImpact # %f %f %f %f", 
       loc.x, loc.y, loc.z, cell.m_distance); */

    VEE_Vector3 vel = /* m_rotationInv **/ o->velocity();
    cell.m_gradient = m_rotation * cell.m_gradient;
    float gamma = dot(cell.m_gradient, vel);

    if(gamma < 0.0) {

      if(gamma * m_frameTime + cell.m_distance < m_bounceDistance) {
	o->collision(cell.m_gradient);
	vel = vel - (gamma * (1.0f + m_bounce)) * cell.m_gradient;
	o->setVelocity(/* m_rotation * */ vel);
	// VEE_trace("Reflected %p", o);
      }
    }
  }

  for(uint i=0; i < VEE_SB_CHILDREN && scene->children(); i++) {
    VEE_SceneBox *tmp = scene->child(i);
    if(bb->intersects(tmp->boundingBox()))
      recursiveImpact(tmp, bb);
  }
}


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

VEE_GridForce2::VEE_GridForce2()
{}

VEE_GridForce2::~VEE_GridForce2()
{}

void VEE_GridForce2::getForce(VEE_Vector3 *force,
			     const VEE_Vector3 &velocity,
			     const VEE_Vector3 &location)
{
  VEE_Vector3 loc = location - m_location;
  loc = m_rotationInv * loc;

  loc /= m_scale;

  VEE_Vector4 plane = (*m_grid).getPlane(loc);

  if(plane.w == VEE_GRID_SENTINEL) {
    force->clear();
    return;
  }
  
  VEE_Vector3 vel = m_rotationInv * velocity;

  float hitRatio = dot3(plane, vel);
  float distance = dot4(plane, location);

  // float minMag = m_magnitude * 0.1;

  VEE_Vector3 localForce;

  if(!m_impactMethod && distance < m_bounceDistance) {

    float magnitude;
    float balpha = 1.0f - distance / m_bounceDistance;

    if(hitRatio < 0.0) {
      // Coming in
      magnitude = m_magnitude * balpha;
    }
    else {
      float balpha = 1.0 - distance / m_bounceDistance;

      // Going out:
      magnitude = m_magnitude * balpha * m_bounce;

    }
    localForce = plane.vector3() * magnitude;
  }
  else if(distance < m_bounceDistance * 0.1)
    localForce = plane.vector3() * m_magnitude * 0.05;
  else
    localForce.clear();

  if((m_tangentForce != 0.0 || m_tangentGrab != 0) && 
     distance < m_flowDistance) {

    // float talpha = 1.0f - cell.m_distance / m_flowDistance;

    VEE_Vector3 velocityNormal = 
      cross(cross(plane.vector3(), vel), vel);

    velocityNormal.normalize();
    
    float tangentForce;

    if(hitRatio < 0.0)
      tangentForce = hitRatio * m_tangentForce;
    else
      tangentForce = hitRatio * m_tangentGrab;

    localForce += velocityNormal * tangentForce/* * talpha*/;
  }

  // Add the drag component if necessary:
  if(m_drag > 0.0 && distance < m_dragDistance) {
    
    float dalpha = 1.0f - distance / m_dragDistance;

    // VEE_Vector3 tangent = 
    // cross(cross(velocity, cell.m_gradient), cell.m_gradient);
    
    VEE_Vector3 tangent = velocity - hitRatio * plane.vector3();

    tangent.normalize();
    
    float tangentVelocity = dot(tangent, velocity);
    
    float drag = powf(VEE_abs(tangentVelocity), m_dragPower) * m_drag;

    localForce -= (tangent * drag * dalpha);
  }

  *force = m_rotation * localForce;
}

void VEE_GridForce2::boundingBoxOfForce(VEE_BoundingBox *bb)
{
  bb->clear();

  VEE_Vector3 tmp;
  tmp.make(m_scale * 0.5);
  VEE_calculateCubeBounds(bb, m_rotation, tmp);

  bb->move(m_location);
}

void VEE_GridForce2::update(float dt)
{
  m_frameTime = dt;
  VEE_ForceObject::update(dt);
  
  VEE_BoundingBox bb;

  boundingBoxOfForce(&bb);

  VEE_SceneBox *scene = m_sceneBox;

  while(!scene->boundingBox().contains(bb) &&
	scene->parent()) {
    scene = scene->parent();
  }
  recursiveImpact(scene, &bb);
}

VEE_Object * VEE_GridForce2::clone() const
{
  // VEE_fatal("VEE_GridForce2::clone # %p", this);
  return new VEE_GridForce2(*this);
}

void VEE_GridForce2::recursiveImpact(VEE_SceneBox *scene, VEE_BoundingBox *bb)
{
  for(VEE_SceneBox::iterator it = scene->begin(); it != scene->end(); it++) {

    VEE_Object * o = VEE_SceneBox::getObject(it);

    VEE_Vector3 loc = o->location() - m_location;
    loc = loc * m_rotation;
    
    loc /= m_scale;
    
    VEE_Vector4 plane = (*m_grid).getPlane(loc);

    if(plane.w == VEE_GRID_SENTINEL)
      continue;

    float distance = dot4(plane, loc);
    
    /* if(cell.m_distance < 0.0)
       VEE_trace("VEE_GridForce2::recursiveImpact # %f", cell.m_distance); */

    distance *= m_scale;

    /* VEE_trace("VEE_GridForce2::recursiveImpact # %f %f %f %f", 
       loc.x, loc.y, loc.z, distance); */

    VEE_Vector3 vel = /* m_rotationInv **/ o->velocity();
    float gamma = dot3(plane, vel);

    if(gamma < 0.0) {

      if(gamma * m_frameTime + distance < m_bounceDistance) {
	o->collision(plane.vector3());
	vel = vel - (gamma * (1.0f + m_bounce)) * plane.vector3();
	o->setVelocity(/* m_rotation * */ vel);
	// VEE_trace("Reflected %p", o);
      }
    }
  }

  for(uint i=0; i < VEE_SB_CHILDREN && scene->children(); i++) {
    VEE_SceneBox *tmp = scene->child(i);
    if(bb->intersects(tmp->boundingBox()))
      recursiveImpact(tmp, bb);
  }
}

