/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <gee_move_around_path.h>

#include <gee_io.h>
#include <gee_macros.h>
#include <gee_world.h>

#include <vee_error.h>

GEE_MoveAroundPath::GEE_MoveAroundPath(GEE_id_t pathId, 
				       GEE_id_t geoId,
				       bool rotate,
				       bool useNapra,
				       float radiusNorm,
				       float exaggerateRadius)
  : GEE_PathToX(pathId),
    m_radius(2.0),
    m_exaggerateRadius(exaggerateRadius),
    m_radiusNorm(radiusNorm),
    m_pressure(1.0),
    m_rotate(rotate),
    m_useNapra(useNapra)
{
  if(geoId >= 0)
    m_geoIds.insert(geoId);
}

GEE_MoveAroundPath::~GEE_MoveAroundPath()
{}
  
GEE_IdContainer::id_container * GEE_MoveAroundPath::getExtraIds()
{
  return & m_geoIds;
}

bool GEE_MoveAroundPath::act(GEE_World *w, const GEE_TimeStamp &time)
{
  // const char * fname = "GEE_MoveAroundPath::act";

  m_moves.clear();

  if(!GEE_PathToX::act(w, time))
    return false;

  if(m_moves.size())
    doMove(w);

  return true;
}

static const ushort GEE_GeoId      = GEE_CONST2('G', 'I');
static const ushort GEE_RadiusId   = GEE_CONST2('R', 'A');
static const ushort GEE_ExRadiusId = GEE_CONST2('E', 'R');
static const ushort GEE_RadiusNormId = GEE_CONST2('R', 'N');
static const ushort GEE_RotateId   = GEE_CONST2('R', 'O');
static const ushort GEE_PressureId = GEE_CONST2('P', 'R');
static const ushort GEE_UseNapraId = GEE_CONST2('U', 'N');
static const ushort GEE_EndId      = GEE_CONST2('E', 'N');


bool GEE_MoveAroundPath::read(GEE_Io * io)
{
  std::string str;
  io->readNullTerminatedStdString(str);
  if(str != type())
    return false;
  
  readBasics(io);
  readPathId(io);

  int v = io->readInt32();

  if(v != 1)
    return false;

  m_geoIds.clear();

  ushort id = io->readInt16();

  while(id != GEE_EndId) {
    if(id == GEE_GeoId)
      m_geoIds.insert(io->readInt64());
    else if(id == GEE_PressureId)
      m_pressure = io->readFloat32();
    else if(id == GEE_RadiusId)
      m_radius = io->readFloat32();
    else if(id == GEE_ExRadiusId)
      m_exaggerateRadius = io->readFloat32();
    else if(id == GEE_RadiusNormId)
      m_radiusNorm = io->readFloat32();
    else if(id == GEE_RotateId)
      m_rotate = io->readByte();
    else if(id == GEE_UseNapraId)
      m_useNapra = io->readByte();
    else {
      int pos = (int) io->position();
      VEE_error("GEE_MoveAroundPath::read # Bad id at %d = %x (%c%c)", 
		pos, pos, (uchar) (id >> 8), (uchar) id);
      return false;
    }
    id = io->readInt16();
  }
  
  // Backwards compatibility:
  if(m_geoIds.size() == 1) {
    if((*m_geoIds.begin()) == -1)
      m_geoIds.clear();
  }

  return true;
}

bool GEE_MoveAroundPath::write(GEE_Io * io)
{
  io->writeNullTerminatedString(type());

  writeBasics(io);
  writePathId(io);
  
  io->writeInt32(1); // version

  for(id_container::iterator it = m_geoIds.begin(); 
      it != m_geoIds.end(); it++) {
    io->writeInt16(GEE_GeoId);
    io->writeInt64(*it);
  }

  io->writeInt16(GEE_RadiusId);
  io->writeFloat32(m_radius);

  io->writeInt16(GEE_ExRadiusId);
  io->writeFloat32(m_exaggerateRadius);

  io->writeInt16(GEE_RadiusNormId);
  io->writeFloat32(m_radiusNorm);

  io->writeInt16(GEE_RotateId);
  io->writeByte(m_rotate);

  io->writeInt16(GEE_PressureId);
  io->writeFloat32(m_pressure);

  io->writeInt16(GEE_UseNapraId);
  io->writeByte(m_useNapra);

  io->writeInt16(GEE_EndId);

  return true;
}

const char * GEE_MoveAroundPath::type() 
{ 
  return "GEE_MoveAroundPath"; 
}

GEE_Operator * GEE_MoveAroundPath::create()
{
  return new GEE_MoveAroundPath();
}

void GEE_MoveAroundPath::getBoundingBox(VEE_BoundingBox *bb)
{
  *bb = m_bbox;
}

bool GEE_MoveAroundPath::move(VEE_Vector3 *v)
{
  const Move * it = & * m_moves.begin();
  const Move * end = & * m_moves.end();

  bool hit = false;

  VEE_Vector3 v2 = *v;

  float radius = m_radius;

  for(; it != end; it++) {
    
    if(m_useNapra)
      radius = it->m_radius;

    float d = (v2 - it->m_origin).length();
    
    float w = (radius - d) / radius * m_pressure;
    
    if(w <= 0.0)
      continue;

    if(w > 1.0f)
      w = 1.0f;

    hit = true;
    
    v2 += it->m_move * w;
  }

  if(hit) {
    /* VEE_trace("GEE_MoveAroundPath::move # %f %f %f -> %f %f %f",
       v->x, v->y, v->z, v2.x, v2.y, v2.z); */
    *v = v2;
  }

  return hit;
}

void GEE_MoveAroundPath::moved(GEE_Geometry *g)
{
  addEditedGeometry(g);
}

void GEE_MoveAroundPath::doMove(GEE_World *w)
{
  GEE_FNAME = "GEE_MoveAroundPath::doMove";

  /* VEE_trace("GEE_MoveAroundPath::doMove # %d %d", 
     (int) m_layer, (int) m_geoId); */

  GEE_Geometry * g;

  // Update bounding box information

  float radius = m_useNapra ? m_moves[0].m_radius : m_radius;

  m_bbox.set(m_moves[0].m_origin, m_radius);
  for(uint i = 1; i < m_moves.size(); i++) {
    if(m_useNapra)
      radius = m_moves[i].m_radius;
    m_bbox.expand(m_moves[i].m_origin, radius);
  }

  if(!m_geoIds.size()) {
    
    for(GEE_LayerGeometryIterator it(w, m_layer); (g = *it); it.nextVisible()){
      // VEE_trace("GEE_MoveAroundPath::doMove %p", g);
      g->moveData(this);
    }
  }
  else {
    for(id_container::iterator itg = m_geoIds.begin(); 
      itg != m_geoIds.end(); itg++) {
      
      GEE_id_t geoId = *itg;

      g = w->geometry(geoId);
    
      if(!g) {
	VEE_error("%s # No geometry %d", fname, (int) geoId);
	continue;
      }
      
      const Move * it  = & * m_moves.begin();
      const Move * end = & * m_moves.end();

      VEE_Matrix4d tmp = g->transformation();
      
      for(; it != end; it++) {
	
	// std::cout << "first: " << tmp << std::endl;
	
	if(m_rotate) {
	  VEE_Vector3d origin = it->m_origin;
	  
	  VEE_Matrix4d m1, m2, m3;
	  // Move object to the origin:
	  m1.identity();
	  m1.addToColumn(3, - origin);
	  
	  // Rotate the object:
	  m2.identity();
	  m2.setRotation(it->m_rotate);

	  // Move back to the original location:
	  m3.identity();
	  m3.addToColumn(3, origin);
	  
	  // Apply the rotation:
	  tmp = m3 * m2 * m1 * tmp;
	  
	  // std::cout << "m1: " << m1 << "\n";
	  // std::cout << "m2: " << m2 << "\n";
	  // std::cout << "m3: " << m3 << "\n";
	  
	}

	// std::cout << "move: " << it->m_move << "\n";
	
	// Move the object:
	tmp.addToColumn(3, it->m_move);
	
	// std::cout << "loop: " << tmp << "\n";
      }
      
      // std::cout << "last: " << tmp << std::endl;
      
      g->setTransformation(tmp);
    }
  }
}

bool GEE_MoveAroundPath::handleNewVertex(GEE_Path3D::iterator &it)
{
  const GEE_PathVertex3D & pv = GEE_Path3D::getVertex(it);
    
  VEE_Vector3 loc = pv.m_location;

  VEE_Matrix3d r(pv.m_rotation.data());

  // The following lines normalize the rotation matrix:
  if(m_rotate) {
    VEE_Vector3d axis;
    double angle;
    r.getRotateAroundAxis(axis, angle);

    // std::cout << "axis & angle: " << axis << "  " << angle << "\n";

    r.rotateAroundAxis(axis, angle);
  }

  if(processedPathVertices()) {

    Move m;
    m.m_origin = m_curLoc;
    m.m_move = loc - m.m_origin;
    
    if(m_rotate) {
      VEE_Matrix3d tmp = m_curRot;
      tmp.transpose(); // Inverse
      m.m_rotate = r * tmp;
    }
    else
      m.m_rotate.identity();
    
    if(m_useNapra) {
      m.m_radius = 
	pv.calculateToolRadius(m_radius, m_radiusNorm, m_exaggerateRadius);
      m.m_origin = pv.calculateToolCenter();
    }

    m_moves.push_back(m);
  }

  m_curLoc = loc;
  m_curRot = r;

  return true;
}

