/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <gee_operator.h>

#include <gee_copy_geometry.h>
#include <gee_color_around_path.h>
#include <gee_debug.h>
#include <gee_erase_around_path.h>
#include <gee_group_around_path.h>
#include <gee_group_geos.h>
#include <gee_id_select.h>
#include <gee_insert_particle.h>
#include <gee_io.h>
#include <gee_macros.h>
#include <gee_move_around_path.h>
#include <gee_particle_rotate.h>
#include <gee_path_insert.h>
#include <gee_path_multiply.h>
#include <gee_path_to_lines.h>
#include <gee_path_to_mesh.h>
#include <gee_path_to_particles.h>
#include <gee_scale_around_path.h>
#include <gee_texture_around_path.h>
#include <gee_turn_around_path.h>
#include <gee_ungroup_geos.h>

#include <vee_error.h>
#include <vee_math.h>

#include <typeinfo>

typedef std::map<std::string, GEE_Operator *(*)()> factoryContainer;
static factoryContainer __factories;

GEE_Operator::GEE_Operator()
  : m_group(-1),
    m_keepAlive(false),
    m_manager(0)
{
  addIdMember(&m_group);
}

GEE_Operator::~GEE_Operator()
{}

bool GEE_Operator::read(GEE_Io *)
{
  VEE_error("GEE_Operator::read # unimplemented (in %s)", 
	    typeid(*this).name());
  return false;
}

bool GEE_Operator::write(GEE_Io *)
{
  VEE_error("GEE_Operator::write # unimplemented (in %s)", 
	    typeid(*this).name());
  return false;
}

const char * GEE_Operator::type()
{
  VEE_error("GEE_Operator::type # unimplemented (in %s)", 
	    typeid(*this).name());
  return "UNKNOWN_TYPE";
}

bool GEE_Operator::readBasics(GEE_Io *io)
{
  int v = io->readInt32();

  m_id    = io->readInt64();
  m_layer = io->readInt64();
  m_group = io->readInt64();
  bool ok = m_interval.read(io);

  return (v == 1) && ok;
}

bool GEE_Operator::writeBasics(GEE_Io *io)
{
  io->writeInt32(1); // version;
  
  io->writeInt64(m_id);
  io->writeInt64(m_layer);
  io->writeInt64(m_group);
  m_interval.write(io);

  return true;
}

void GEE_Operator::copyBasics(const GEE_Operator &source)
{
  m_id       = source.m_id;
  m_layer    = source.m_layer;
  m_interval = source.m_interval;
  m_group    = source.m_group;
  m_keepAlive= source.m_keepAlive;
}

/** A virtual function that starts to do the work. This method is
    called once before "act" is called.

    The default implementation returns true. */

bool GEE_Operator::start(GEE_World *, const GEE_TimeStamp &)
{
  return true;
}

/** A virtual function that does something. This method is called at
    least once between "start" and "stop". 

    The default implementation returns true.*/

bool GEE_Operator::act(GEE_World *, const GEE_TimeStamp &)
{
  return true;
}

/** A virtual function that ends/finalizes the work. This method is
    called once after "act" has been called.

    The default implementation returns true. */

bool GEE_Operator::stop(GEE_World *, const GEE_TimeStamp &)
{
  return true;
}

GEE_Operator * GEE_Operator::create(const char * type)
{
  static bool x = false;

  if(!type)
    return 0;

  if(!x) {
    init();
    x = true;
  }

  factoryContainer::iterator it = __factories.find(type);

  if(it == __factories.end())
    return 0;

  return (*(*it).second)();
}

void GEE_Operator::init()
{
  __factories["GEE_CopyGeometry"]      = GEE_CopyGeometry::create;
  __factories["GEE_ColorAroundPath"]   = GEE_ColorAroundPath::create;
  __factories["GEE_IdSelect"]          = GEE_IdSelect::create;
  __factories["GEE_InsertParticle"]    = GEE_InsertParticle::create;
  __factories["GEE_EraseAroundPath"]   = GEE_EraseAroundPath::create;
  __factories["GEE_GroupAroundPath"]   = GEE_GroupAroundPath::create;
  __factories["GEE_GroupGeos"]         = GEE_GroupGeos::create;
  __factories["GEE_MoveAroundPath"]    = GEE_MoveAroundPath::create;
  __factories["GEE_ParticleRotate"]    = GEE_ParticleRotate::create;
  __factories["GEE_PathInsert"]    = GEE_PathInsert::create;
  __factories["GEE_PathMultiply"]  = GEE_PathMultiply::create;
  __factories["GEE_PathToLines"]   = GEE_PathToLines::create;
  __factories["GEE_PathToMesh"]    = GEE_PathToMesh::create;
  __factories["GEE_PathToParticles"]   = GEE_PathToParticles::create;
  __factories["GEE_ScaleAroundPaths"]  = GEE_ScaleAroundPaths::create;
  __factories["GEE_TextureAroundPath"] = GEE_TextureAroundPath::create;
  __factories["GEE_TurnAroundPath"]    = GEE_TurnAroundPath::create;
  __factories["GEE_UnGroupGeos"]       = GEE_UnGroupGeos::create;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

GEE_OperatorTimer::GEE_OperatorTimer(const GEE_Operator *a)
  : m_time(a->interval().begin()),
    m_id(a->id())
{}

GEE_OperatorTimer::GEE_OperatorTimer(const GEE_OperatorGroup *g)
  : m_time(g->offset()),
    m_id(g->id())
{}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

bool GEE_OperatorGroup::m_debug = false;

GEE_OperatorGroup::GEE_OperatorGroup()
  : m_scale(1.0),
    m_group(-1),
    m_keepAlive(false),
    m_running(false)

{
  addIdMember(&m_group);
  m_nextOperator = m_operators.begin();

  m_debug = GEE_Debug::debug("GEE_OperatorGroup");
}

GEE_OperatorGroup::~GEE_OperatorGroup()
{}
  
bool GEE_OperatorGroup::start
(GEE_OperatorManager *om, GEE_World *, const GEE_TimeStamp &)
{
  if(!m_operators.size())
    return true;

  int fps = om->fps();

  GEE_TimeStamp begin = 
    (*m_operators.begin()).second.ptr()->interval().begin();

  m_frame = (begin.stamp() / fps);
  m_nextOperator = m_operators.begin();
  m_currentOperators.clear();
  m_running = true;
  return true;
}

bool GEE_OperatorGroup::act
(GEE_OperatorManager *om, GEE_World *w, const GEE_TimeStamp &ts)
{
  GEE_FNAME = "GEE_OperatorGroup::act";

  if(m_debug)
    VEE_trace("%s # ENTER", fname);

  GEE_TimeStamp local = ts - m_offset;
  GEE_TimeStamp ts2, ts3;

  uint fps = om->fps();
    
  if(m_scale > 1.0e-5) 
    local /= m_scale;    
  else
    local = m_endTime;

  if(m_keepAlive) {
    /* This is special case, we are running in interactive mode. */
    ts2.setSeconds(100000.0);
    local += ts2;
    nextFrame(w, local);
  }
  else {
    while(true) {

      ts2.setSeconds(m_frame / fps);
      ts3.setSeconds((m_frame % fps) / (double) fps);

      m_time = ts2 + ts3;
      nextFrame(w, m_time);

      if(m_time < local)
	m_frame++;
      else {
	if(m_debug)
	  VEE_trace("%s # EXIT - TIME", fname);
	return true;
      }
    }
  }

  if(m_debug)
    VEE_trace("%s # EXIT - NORM", fname);

  return true;
}

bool GEE_OperatorGroup::stop(GEE_OperatorManager *, 
GEE_World *w, const GEE_TimeStamp &ts)
{
  for(listIter lit = m_currentOperators.begin();
      lit != m_currentOperators.end(); lit++)
    (*lit).ptr()->stop(w, ts);
  
  m_currentOperators.clear();

  m_nextOperator = m_operators.end();
  m_keepAlive = false;
  m_running = false;
  return true;
}

void GEE_OperatorGroup::addOperator(VEE_RefPtr<GEE_Operator> & o)
{
  GEE_Operator * op = o.ptr();

  if(op->keepAlive())
    setKeepAlive(true);

  iterator it = 
    m_operators.insert(m_operators.begin(),
		       container::value_type(GEE_OperatorTimer(op), o));
  
  m_endTime = VEE_max(op->interval().end(), m_endTime);

  if(m_running) {
    it ++;
    if(it == m_nextOperator) {
      m_nextOperator--;
      if(m_debug)
	VEE_trace("GEE_OperatorGroup::addOperator # Rewinding one step");
    }
  }
}

void GEE_OperatorGroup::clearOperators()
{
  m_operators.clear();
  m_nextOperator = m_operators.begin();
  m_currentOperators.clear();
  m_endTime = 0;
}


static const ushort GEE_IdId         = GEE_CONST2('I', 'D');
static const ushort GEE_TsId         = GEE_CONST2('T', 'S');
static const ushort GEE_ScaleId      = GEE_CONST2('S', 'C');
static const ushort GEE_EndId        = GEE_CONST2('E', 'N');

bool GEE_OperatorGroup::read(GEE_Io *io)
{
  std::string str;
  io->readNullTerminatedStdString(str);
  if(str != "GEE_OperatorGroup")
    return false;
  
  int v = io->readInt32();

  if(v != 1)
    return false;

  ushort id = io->readInt16();

  while(id != GEE_EndId) {
    if(id == GEE_IdId)
      m_id = io->readInt64();
    else if(id == GEE_TsId)
      m_offset = io->readInt64();
    else if(id == GEE_ScaleId)
      m_scale = io->readFloat64();
    else {
      VEE_error("GEE_OperatorGroup::read # Bad id at %d (%c%c)", 
		(int) io->position(), (uchar) (id >> 8), (uchar) id);
      return false;
    }
    id = io->readInt16();
  }

  return true;
}

bool GEE_OperatorGroup::write(GEE_Io *io)
{
  io->writeNullTerminatedString("GEE_OperatorGroup");
  
  io->writeInt32(1); // version

  io->writeInt16(GEE_IdId);
  io->writeInt64(m_id);
  
  io->writeInt16(GEE_TsId);
  io->writeInt64(m_offset.stamp());

  io->writeInt16(GEE_ScaleId);
  io->writeFloat64(m_scale);

  io->writeInt16(GEE_EndId);
  
  return true;
}

void GEE_OperatorGroup::calculateEndTime()
{
  m_endTime = 0;
  
  for(iterator it = m_operators.begin(); it != m_operators.end(); it++) {
    GEE_TimeStamp ts = (*it).second.ptr()->interval().end();
    if(m_endTime < ts)
      m_endTime = ts;
  }
}

  
/** Calculates one step forward in local time coordinates. */

void GEE_OperatorGroup::nextFrame(GEE_World *w, const GEE_TimeStamp &ts)
{
  GEE_FNAME = "GEE_OperatorGroup::nextFrame";

  if(m_debug)
    VEE_trace("%s # ENTER %lf", fname, ts.secondsD());
  
  for(listIter lit = m_currentOperators.begin();
      lit != m_currentOperators.end(); ) {
    
    GEE_Operator * o = (*lit).ptr();
    
    if(!o->keepAlive() && o->interval().end() <= ts) {
      o->stop(w, ts);

      if(m_debug)
	VEE_trace("%s # Stopped %s", fname, typeid(*o).name());
      
      listIter tmp = lit;
      tmp ++;
      m_currentOperators.erase(lit);
      lit = tmp;
    }
    else {
      o->act(w, ts);
      lit++;
      if(m_debug)
	VEE_trace("%s # Acted %s", fname, typeid(*o).name());
    }
  }

  while((*m_nextOperator).first.time() <= ts && 
	m_nextOperator != m_operators.end()) {
    
    VEE_RefPtr<GEE_Operator> & ra = (*m_nextOperator).second;
    
    GEE_Operator * a = ra.ptr();
    
    bool ok = a->start(w, ts);

    if(!ok) {
      VEE_error("%s # start failed for %s", fname, typeid(*a).name());
      m_nextOperator++;
      continue;
    }

    if(m_debug)
      VEE_trace("%s # started %s", fname, typeid(*a).name());


    ok = a->act(w, ts);

    if(!ok) {
      VEE_error("%s # act failed for %s (new)", fname, typeid(*a).name());
      m_nextOperator++;
      continue;
    }

    if(!a->keepAlive() && a->interval().end() <= ts)
      a->stop(w, ts);
    else {
      if(m_debug)
	VEE_trace("%s # Added operator of type %s", fname, typeid(*a).name());
      m_currentOperators.push_back(ra);
    }
    
    m_nextOperator++;
  }

  if(m_debug)
    VEE_trace("%s # EXIT", fname);
  
  return;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

bool GEE_OperatorManager::m_debug = true;

GEE_OperatorManager::GEE_OperatorManager()
  : m_dirtyPlan(true),
    m_running(false),
    m_fps(25),
    m_frame(0),
    m_world(0)
{
  m_nextGroup = m_plan.begin();
  m_undoIter = m_undoList.end();

  m_debug = GEE_Debug::debug("GEE_OperatorManager");
}

GEE_OperatorManager::~GEE_OperatorManager()
{}

bool GEE_OperatorManager::addOperator(GEE_Operator * a)
{
  if(!a)
    return false;

  if(a->id() < 0)
    a->autoId();

  VEE_RefPtr<GEE_Operator> a2(a);
  addOperatorInternal(a2);

  return true;
}

bool GEE_OperatorManager::addOperatorGroup(const GEE_OperatorGroup &g)
{
  VEE_RefPtr<GEE_OperatorGroup> tmp(new GEE_OperatorGroup(g));

  return addOperatorGroup2(tmp);
}

bool GEE_OperatorManager::addOperatorGroup2(VEE_RefPtr<GEE_OperatorGroup> & g)
{
  GEE_OperatorGroup * gr = g.ptr();

  if(!gr)
    VEE_fatal("GEE_OperatorManager::addOperatorGroup2 # gr = NULL");

  if(gr->id() < 0)
    gr->autoId();

  m_groups[gr->id()] = g;

  if(m_running) {
    iterator it = 
      m_plan.insert(m_plan.begin(), 
		    container::value_type(GEE_OperatorTimer(gr), g));
    it++;
    if(it == m_nextGroup)
      m_nextGroup--;
  }

  return true;
}

/** Remaps some time stamp found in the operator to real-time. */

GEE_TimeStamp GEE_OperatorManager::operatorTimeToRealTime
(GEE_TimeStamp ts, GEE_id_t groupId)
{
  while(groupId >= 0) {
    GEE_OperatorGroup *g = getOperatorGroup(groupId);

    if(!g) {
      VEE_error("GEE_OperatorManager::operatorTimeToRealTime # No group %d", 
		(int) groupId);
      return 0;
    }
    
    ts *= g->scale();
    ts += g->offset();

    groupId = g->group();
  }

  return ts;
}

GEE_TimeStamp GEE_OperatorManager::realTimeToOperatorTime
(GEE_TimeStamp ts, GEE_id_t groupId)
{
  while(groupId >= 0) {
    GEE_OperatorGroup *g = getOperatorGroup(groupId);

    if(!g) {
      VEE_error("GEE_OperatorManager::realTimeToOperatorTime # No group %d", 
		(int) groupId);
      return 0;
    }
    
    ts -= g->offset();
    double scale = g->scale();
    if(scale > 1.0e-6)
      ts /= g->scale();
    else
      return GEE_TimeStamp().setSeconds(1000000.0);

    groupId = g->group();
  }

  return ts;
}

void GEE_OperatorManager::undo(uint n)
{
  GEE_FNAME = "GEE_OperatorManager::undo";

  if(m_debug)
    VEE_trace(fname);

  if(m_running)
    VEE_error("%s Cannot undo properly while running (!)", fname);

  for(uint i = 0; (i < n) && (m_undoIter != m_undoList.begin()); i++) {

    m_undoIter--;
    
    /* if(m_undoIter == m_undoList.begin()) {
      VEE_trace("%s # Nothing more undo (%d elems)", fname, 
		(int) m_undoList.size());
		}*/
  }
}

void GEE_OperatorManager::redo(uint n)
{
  const char *fname = "GEE_OperatorManager::redo";

  if(m_debug)
    VEE_trace(fname);

  if(m_running)
    VEE_error("%s Cannot redo properly while running (!)", fname);

  for(uint i = 0; (i < n) && (m_undoIter != m_undoList.end()); i++) {
    m_undoIter++;
  }
}

static const int nextId = GEE_CONST('N', 'E', 'X', 'T');
static const int endId  = GEE_CONST('E', 'N', 'D', 'D');
static const int undoId = GEE_CONST('U', 'N', 'D', 'O');

bool GEE_OperatorManager::read(GEE_Io *io, const char * )
{
  GEE_FNAME = "GEE_OperatorManager::read";

  if(m_debug)
    VEE_trace(fname);

  clear();

  std::string str;

  io->readNullTerminatedStdString(str);
  if(str != "GEE_OPERATOR_MANAGER")
    return false;
  
  int v = io->readInt32();

  if(v != 1)
    return false;

  io->readNullTerminatedStdString(str);
  if(str != "OPERATORS")
    return false;
  
  uint undos = 0;

  if(m_debug)
    VEE_trace("%s # 1", fname);

  std::list<GEE_Operator *> operators;

  while(true) {
    int command = io->readInt32();

    if(command == endId)
      break;

    bool isUndone = (command == undoId);
    if(isUndone) {
      command = io->readInt32();
      undos = 1;
    }

    if(command != nextId)
      return false;

    long pos = io->position();

    io->readNullTerminatedStdString(str);
    GEE_Operator * a = GEE_Operator::create(str.c_str());
    
    io->seekPosition(pos);

    if(!a)
      VEE_error("%s # Failed to create \"%s\"", fname, str.c_str());
    else {
      if(m_debug)
	VEE_trace("%s # Reading %s", fname, typeid(*a).name());

      if(a->read(io)) {
	VEE_trace("%s # read %p %s", fname, a, typeid(*a).name());
	operators.push_back(a);
	 // addOperator(a);
      }
      else
	 VEE_trace("%s # Failed to read %s", 
		   fname, typeid(*a).name());
    }

    if(undos && !isUndone) {
      undos++;
    }
  }

  // Read all groups:

  io->readNullTerminatedStdString(str);
  if(str != "GROUPS")
    return false;
  
  while(true) {
    int command = io->readInt32();

    if(command == endId)
      break;
    
    GEE_OperatorGroup og;
    if(og.read(io))
      addOperatorGroup(og);
  }

  for(std::list<GEE_Operator *>::iterator oit = operators.begin(); 
      oit != operators.end(); oit++)
    addOperator(*oit);
  
  // Do undos:
  for(uint i = 0; i < undos; i++)
    undo();

  rebuildPlan();

  if(m_debug)
    VEE_trace("%s # EXIT", fname);

  return true;
}

/** Writes current state of the manager. 

    <OL>

    <LI> Store all operators in the order that they happen to be in
    the undo list.

    <LI> Store all operator groups.

    </OL>
    
 */

bool GEE_OperatorManager::write(GEE_Io *io, const char *)
{
  GEE_FNAME = "GEE_OperatorManager::write";

  io->writeNullTerminatedString("GEE_OPERATOR_MANAGER");
  io->writeInt32(1); // version

  io->writeNullTerminatedString("OPERATORS");

  VEE_trace("GEE_OperatorManager::write # %d", (int) m_undoList.size());

  for(operatorListIter li = m_undoList.begin(); li != m_undoIter; li++) {
    
    io->writeInt32(nextId);

    GEE_Operator * a = (*li).ptr();

    if(a->write(io))
      VEE_trace("%s # wrote %s", 
		fname, typeid(*a).name());
    else
      VEE_error("%s # Could not write %s", 
		fname, typeid(*a).name());
  }
  
  io->writeInt32(endId);

  io->writeNullTerminatedString("GROUPS");

  for(groupIterator gi = m_groups.begin(); gi != m_groups.end(); gi++) {
    GEE_OperatorGroup * gr = (*gi).second.ptr();
    
    if(gr->operatorCount()) {
      
      io->writeInt32(nextId);
      gr->write(io);
    }
  }

  io->writeInt32(endId);

  return true;
}


void GEE_OperatorManager::collectIds(GEE_IdContainer::id_remap &idMap)
{
  // Collect all ids:
  for(operatorListIter lit = m_undoList.begin(); 
      lit != m_undoList.end(); lit++)
    (*lit).ptr()->addIdsToRemap(idMap);

  for(groupIterator gi = m_groups.begin(); gi != m_groups.end(); gi++)
    (*gi).second.ptr()->addIdsToRemap(idMap);
}
  
void GEE_OperatorManager::remapIds(GEE_IdContainer::id_remap &idMap)
{
  // Apply remaps;
  operatorListIter lit;
  for(lit = m_undoList.begin(); lit != m_undoList.end(); lit++)
    (*lit).ptr()->remapIdMembers(idMap);

  // Now update the indexing information in local map:

  groupContainer gtmp;
  for(groupIterator gi = m_groups.begin(); gi != m_groups.end(); gi++) {
    VEE_RefPtr<GEE_OperatorGroup> & gref = (*gi).second;
    GEE_OperatorGroup * og = gref.ptr();
    og->remapIdMembers(idMap);
    gtmp[og->id()] = gref;
  }
  
  m_groups.swap(gtmp);
  
  rebuildPlan();
}
  
void GEE_OperatorManager::update(GEE_TimeStamp ts)
{
  // GEE_FNAME = "GEE_OperatorManager::update";

  if(!m_running) {
    VEE_trace("GEE_OperatorManager::update # Not yet running -> rewinding");
    rewind(0);
  }

  GEE_TimeStamp ts2, ts3;

  while(true) {
    ts2.setSeconds(m_frame / m_fps);
    ts3.setSeconds((m_frame % m_fps) / (double) m_fps);

    nextFrame(ts2 + ts3);
    if(m_time < ts)
      m_frame++;
    else
      return;
  }
}

void GEE_OperatorManager::rewind(GEE_TimeStamp ts)
{
  if(m_dirtyPlan)
    rebuildPlan();

  m_nextGroup = m_plan.begin();

  m_frame = 0;
  m_time = 0;

  m_running = true;

  update(ts);
}


void GEE_OperatorManager::stop()
{
  for(groupListIter it = m_currentGroups.begin(); 
      it != m_currentGroups.end(); it++)
    (*it).ptr()->stop(this, m_world, m_time);
  
  m_currentGroups.clear();

  m_frame = 0;
  m_time = 0;

  m_running = false;
}

void GEE_OperatorManager::clear()
{
  m_plan.clear();
  m_nextGroup = m_plan.begin();
  m_currentGroups.clear();
  m_undoList.clear();
  m_undoIter = m_undoList.end();
  m_time = 0;
  m_groups.clear();
}

void GEE_OperatorManager::merge(GEE_OperatorManager &that)
{
  for(groupIterator git = that.m_groups.begin(); 
      git != that.m_groups.end(); git++)
    m_groups[(*git).second.ptr()->id()] = (*git).second;

  for(operatorListIter it = that.m_undoList.begin(); 
      it != that.m_undoIter; it++)
    addOperatorInternal((*it));
}

void GEE_OperatorManager::rebuildPlan()
{
  GEE_FNAME = "GEE_OperatorManager::rebuildPlan";

  m_plan.clear();
  m_currentGroups.clear();

  // Make all groups forget their operators:

  groupIterator git;

  for(git = m_groups.begin(); git != m_groups.end(); git++) {
    (*git).second.ptr()->clearOperators();
  }
  
  // Default group for all lone operators
  m_groups[-1] = new GEE_OperatorGroup();

  // Put all operators back to the groups:
  
  for(operatorListIter it = m_undoList.begin(); it != m_undoIter; it++) {

    VEE_RefPtr<GEE_Operator> & op = (*it);
    GEE_id_t gid = op.ptr()->group();

    GEE_OperatorGroup * gr = getOperatorGroup(gid);
    if(!gr)
      VEE_error("%s # No group %d for %s", 
		fname, (int) gid, typeid(*op.ptr()).name());
    else
      gr->addOperator(op);
  }
  
  // Now build the master plan:
  
  for(git = m_groups.begin(); git != m_groups.end(); git++) {
    VEE_RefPtr<GEE_OperatorGroup> & gr = (*git).second;
    m_plan[GEE_OperatorTimer(gr.ptr())] = gr;
  }  
  
  if(m_debug)
    VEE_trace("%s # Plan has %u elements", fname, (uint) m_plan.size());

  m_nextGroup = m_plan.begin();
}

void GEE_OperatorManager::addOperatorInternal(VEE_RefPtr<GEE_Operator> & a)
{
  GEE_FNAME = "GEE_OperatorManager::addOperatorInternal";

  GEE_Operator * op = a.ptr();

  if(m_debug || true) {
    VEE_trace("%s # %s %d [%lf - %lf]", fname, typeid(*op).name(), 
	      (int) op->id(),
	      op->interval().begin().secondsD(),
	      op->interval().end().secondsD());
  }

  op->setManager(this);

  if(m_running) {
    GEE_OperatorGroup * gr = getOperatorGroup(op->group());
    if(gr)
      gr->addOperator(a);
  }

  VEE_trace("%s # About to clear old actions", fname);

  if(m_undoIter != m_undoList.end()) {
    operatorListIter it = m_undoList.end();
    m_undoList.erase(m_undoIter, it);
  }

  VEE_trace("%s # Nearly done", fname);

  m_undoList.push_back(a);

  m_undoIter = m_undoList.end();

  VEE_trace("%s # EXIT", fname);
}


void GEE_OperatorManager::nextFrame(GEE_TimeStamp ts)
{
  GEE_FNAME = "GEE_OperatorManager::nextFrame";
  
  if(m_debug)
    VEE_trace("%s # %lf %u %u", fname, ts.secondsD(), (uint) m_plan.size(),
	      (uint) m_currentGroups.size());

  m_time = ts;

  for(groupListIter lit = m_currentGroups.begin();
      lit != m_currentGroups.end(); ) {

    GEE_OperatorGroup * g = (*lit).ptr();

    GEE_TimeStamp gend = g->endTime();

    if(!g->keepAlive() && gend <= ts) {
      g->stop(this, m_world, ts);
      groupListIter tmp = lit;
      tmp++;
      m_currentGroups.erase(lit);
      lit = tmp;
      
      if(m_debug)
	VEE_trace("%s # Stopped a group", fname);
    }
    else {

      if(m_debug)
	VEE_trace("%s # Group %d continues %d %lf", fname, (int) g->id(),
		  (int) g->keepAlive(), gend.secondsD());
      
      
      if(!g->act(this, m_world, ts))
	VEE_error("%s # act failed for group %d (in current)", 
		  fname, (int) g->id());
      lit++;
    }
  }
  
  // VEE_trace("Next time = %lf", (*m_nextOperator).first.secondsD());

  while((*m_nextGroup).first.time() <= ts && 
	m_nextGroup != m_plan.end()) {

    VEE_RefPtr<GEE_OperatorGroup> & rg = (*m_nextGroup).second;
    
    GEE_OperatorGroup * g = rg.ptr();
    
    // VEE_trace("New operator of type %s", typeid(*a).name());

    int gid = (int) g->id();

    bool ok = g->start(this, m_world, ts);

    if(m_debug)
      VEE_trace("%s # Hit group %d (%d)", fname, (int) gid, (int) ok);

    if(!ok) {
      VEE_error("%s # start failed for %d", fname, gid);
      m_nextGroup++;
      continue;
    }

    ok = g->act(this, m_world, ts);
    if(!ok) {
      VEE_error("%s # act failed for %d (new)", fname, gid);
      m_nextGroup++;
      continue;
    }
    
    GEE_TimeStamp gend = g->endTime();
    
    if(!g->keepAlive() && gend <= ts)
      g->stop(this, m_world, ts);
    else {
      VEE_trace("%s # Added group %d", fname, gid);
      m_currentGroups.push_back(rg);
    }

    m_nextGroup++;
  }

  if(m_debug)
    VEE_trace("%s # EXIT", fname);
}
