/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <gee_path_to_particles.h>

#include <gee_io.h>
#include <gee_macros.h>
#include <gee_world.h>

#include <vee_error.h>
#include <vee_particle.h>
#include <vee_time.h>

GEE_PathToParticles::GEE_PathToParticles(GEE_id_t pathId, 
					 GEE_id_t particleId)
  : GEE_PathToX(pathId),
    m_particleId(particleId),
    m_prevVertex(0,0,0),
    m_rate(10.0),
    m_radius(0.0),
    m_exaggerateRadius(0.0),
    m_radiusNorm(1.0),
    m_rotationRange(0.0f),
    m_useNapra(false)
{
  addIdMember(&m_particleId);  
  m_particle.m_particle = new VEE_Particle();

  // Fabricate a random seed:
  static uint counter = 0;
  VEE_Time tmp;
  tmp.getTimeOfDay();
  m_randomSeed = tmp.usecs() + counter++;

  m_writer.addMemberFloat(GEE_CONST2('H', 'Z'), & m_rate);
  m_writer.addMemberFloat(GEE_CONST2('E', 'R'), & m_exaggerateRadius);
  m_writer.addMemberFloat(GEE_CONST2('R', 'N'), & m_radiusNorm);
  m_writer.addMemberId   (GEE_CONST2('P', 'I'), & m_particleId);
  m_writer.addMemberFloat(GEE_CONST2('R', 'A'), & m_radius);
  m_writer.addMemberUInt (GEE_CONST2('R', 'S'), & m_randomSeed);
  m_writer.addMemberBool (GEE_CONST2('U', 'N'), & m_useNapra);
  m_writer.addMemberFloat(GEE_CONST2('R', 'R'), & m_rotationRange);
}

GEE_PathToParticles::~GEE_PathToParticles()
{}

bool GEE_PathToParticles::start(GEE_World *w, const GEE_TimeStamp &ts)
{
  // GEE_FNAME = "GEE_PathToParticles::start";

  if(!GEE_PathToX::start(w, ts))
    return false;  

  GEE_ParticleObject * particles = new GEE_ParticleObject();
  m_particles = particles;
  particles->setId(m_particleId);
  particles->setLayer(m_layer);

  m_particleId = w->addGeometry2(m_particles);
  m_remaining = 0.0;
  m_prevTime  = 0.0;
  m_prevVertex.clear();
  particles->attachEditor();


  /* m_rand.setSeed(m_randomSeed);

  VEE_trace("%s # %p %f", fname, m_particle.m_particle.ptr(), m_rand.rand11());
  */
  m_rand.setSeed(m_randomSeed);

  return true;
}


bool GEE_PathToParticles::stop(GEE_World *w, const GEE_TimeStamp &ts)
{
  // VEE_trace("GEE_PathToParticles::stop");

  GEE_PathToX::stop(w, ts);

  m_particles.ptr()->detachEditor();

  /*
  GEE_ParticleObject * po = 
    dynamic_cast<GEE_ParticleObject *> (m_particles.ptr());

  for(GEE_ParticleObject::iterator it = po->begin();
      it != po->end(); it++)
    VEE_trace("GEE_PathToParticles::stop # %p %p", po, (*it).m_particle.ptr());
  */

  return true;
}

bool GEE_PathToParticles::read(GEE_Io * io)
{
  std::string str;
  io->readNullTerminatedStdString(str);
  if(str != type())
    return false;
  
  readBasics(io);
  readPathId(io);

  int v = io->readInt32();

  if(v != 1)
    return false;

  m_writer.readParameters(io);

  /* ushort id = io->readInt16();

  while(id != GEE_EndId) {
    if(id == GEE_RateId)
      m_rate = io->readFloat32();
    else if(id == GEE_ParticleId)
      m_particleId = io->readInt64();
    else if(id == GEE_RadiusId)
      m_radius = io->readFloat32();
    else if(id == GEE_ExRadiusId)
      m_exaggerateRadius = io->readFloat32();
    else if(id == GEE_RadiusNormId)
      m_radiusNorm = io->readFloat32();
    else if(id == GEE_SeedId)
      m_randomSeed = io->readInt32();
    else if(id == GEE_UseNapraId)
      m_useNapra = io->readByte();
    else {
      VEE_error("GEE_PathToParticles::read # Bad id at %d (%c%c)", 
		(int) io->position(), (uchar) (id >> 8), (uchar) id);
      return false;
    }
    id = io->readInt16();
    }*/
  
  return m_particle.read(io);
}

bool GEE_PathToParticles::write(GEE_Io * io)
{
  io->writeNullTerminatedString(type());

  writeBasics(io);
  writePathId(io);

  io->writeInt32(1); // version

  // io->writeInt16(GEE_RateId);
  // io->writeFloat32(m_rate);

  // io->writeInt16(GEE_ExRadiusId);
  // io->writeFloat32(m_exaggerateRadius);

  // io->writeInt16(GEE_RadiusNormId);
  // io->writeFloat32(m_radiusNorm);

  // io->writeInt16(GEE_ParticleId);
  // io->writeInt64(m_particleId);

  // io->writeInt16(GEE_RadiusId);
  // io->writeFloat32(m_radius);

  // io->writeInt16(GEE_SeedId);
  // io->writeInt32(m_randomSeed);

  // io->writeInt16(GEE_UseNapraId);
  // io->writeByte(m_useNapra);

  m_writer.writeParameters(io);

  // io->writeInt16(GEE_EndId);

  return m_particle.write(io);
}

GEE_Operator * GEE_PathToParticles::create()
{
  return new GEE_PathToParticles;
}

void GEE_PathToParticles::setTextureName(const std::string &name)
{
  m_particle.setTexture(name, GEE_World::textureId(name));
}

bool GEE_PathToParticles::handleNewVertex(GEE_Path3D::iterator &it)
{
  // VEE_trace("GEE_PathToParticles::handleNewVertex");

  GEE_ParticleObject * particles = 
    dynamic_cast<GEE_ParticleObject *> (m_particles.ptr());

  if(!particles) {
    VEE_error("GEE_PathToParticles::handleNewVertex # particles = NULL");
    return false;
  }
  
  const GEE_PathVertex3D &pv = GEE_Path3D::getVertex(it);
  float time = GEE_Path3D::getTime(it);

  float radius;
  VEE_Vector3 newLoc;

  if(m_useNapra) {
    radius = 
      pv.calculateToolRadius(m_radius, m_radiusNorm, m_exaggerateRadius);
    newLoc = pv.calculateToolCenter();
  }
  else {
    newLoc = pv.m_location;
    radius = m_radius;
  }

  if(!processedPathVertices()) {
    m_prevTime = time;
    m_prevVertex = newLoc;
    return true;
  }

  /* VEE_trace("GEE_PathToParticles::handleNewVertex # %f %f %f",
     newLoc.x, newLoc.y, newLoc.z); */

  // Now calculate where the next particle will be located.
  float interval = 1.0 / m_rate;
  float dt = time - m_prevTime;

  VEE_trace("GEE_PathToParticles::handleNewVertex # time: %f %f", 
	    dt, m_remaining);

  while(m_remaining < dt) {
    
    float rel = m_remaining / dt;
    
    VEE_Vector3 loc = m_prevVertex * (1.0f - rel) + newLoc * rel;
    VEE_Vector3 randv = m_rand.randomVectorInSphere() * radius;
    loc += randv;
    
    m_particle.setLocation(loc);
    m_particle.setVelocity(m_rand.randomVectorInSphere());
    
    if(m_rotationRange != 0.0f) {
      m_particle.setRotation(m_rand.rand11() * m_rotationRange);
    }

    particles->addParticle(m_particle);
    
    m_remaining += interval;
    
    VEE_trace("GEE_PathToParticles::handleNewVertex # %f %f %f (%f %f %f) "
	      "[%f %f %f]",
	      loc.x, loc.y, loc.z, newLoc.x, newLoc.y, newLoc.z,
	      randv.x, randv.y, randv.z);
  }
  m_remaining -= dt;

  m_prevTime = time;
  m_prevVertex = newLoc;
  
  return true;
}
