/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <gee_scale_around_path.h>

#include <gee_io.h>
#include <gee_macros.h>
#include <gee_world.h>

#include <vee_error.h>
#include <vee_math.h>

GEE_ScaleAroundPaths::GEE_ScaleAroundPaths(GEE_id_t pathId1, 
					   GEE_id_t pathId2, 
					   GEE_id_t geoId, 
					   bool uniform)
  : GEE_PathToX(pathId1, pathId2),
    m_uniform(uniform)
{
  if(geoId >= 0)
    m_geoIds.insert(geoId);
}

GEE_ScaleAroundPaths::~GEE_ScaleAroundPaths()
{}
  
GEE_IdContainer::id_container * GEE_ScaleAroundPaths::getExtraIds()
{
  return & m_geoIds;
}

bool GEE_ScaleAroundPaths::act(GEE_World *w, const GEE_TimeStamp &time)
{
  // const char * fname = "GEE_ScaleAroundPaths::act";

  m_scales.clear();

  if(!GEE_PathToX::act(w, time))
    return false;

  if(m_scales.size())
    doScale(w);

  return true;
}

static const ushort GEE_GeoId      = GEE_CONST2('G', 'I');
static const ushort GEE_UniformId  = GEE_CONST2('U', 'N');
static const ushort GEE_EndId      = GEE_CONST2('E', 'N');


bool GEE_ScaleAroundPaths::read(GEE_Io * io)
{
  std::string str;
  io->readNullTerminatedStdString(str);
  if(str != type())
    return false;
  
  readBasics(io);
  readPathId(io);

  int v = io->readInt32();

  if(v != 1)
    return false;

  ushort id = io->readInt16();

  m_geoIds.clear();

  while(id != GEE_EndId) {
    if(id == GEE_GeoId)
      m_geoIds.insert(io->readInt64());
    else if(id == GEE_UniformId)
      m_uniform = io->readByte();
    else {
      VEE_error("GEE_ScaleAroundPaths::read # Bad id at %d (%c%c)", 
		(int) io->position(), (uchar) (id >> 8), (uchar) id);
      return false;
    }
    id = io->readInt16();
  }

  return true;
}

bool GEE_ScaleAroundPaths::write(GEE_Io * io)
{
  io->writeNullTerminatedString(type());

  writeBasics(io);
  writePathId(io);
  
  io->writeInt32(1); // version

  for(id_container::iterator it = m_geoIds.begin(); 
      it != m_geoIds.end(); it++) {
    io->writeInt16(GEE_GeoId);
    io->writeInt64(*it);
  }

  io->writeInt16(GEE_UniformId);
  io->writeByte(m_uniform);

  io->writeInt16(GEE_EndId);

  return true;
}

const char * GEE_ScaleAroundPaths::type() 
{ 
  return "GEE_ScaleAroundPaths"; 
}

GEE_Operator * GEE_ScaleAroundPaths::create()
{
  return new GEE_ScaleAroundPaths();
}

void GEE_ScaleAroundPaths::doScale(GEE_World *w)
{
  GEE_FNAME = "GEE_ScaleAroundPaths::doScale";

  for(id_container::iterator itg = m_geoIds.begin(); 
      itg != m_geoIds.end(); itg++) {

    GEE_id_t geoId = *itg;

    GEE_Geometry * g = w->geometry(geoId);
    
    if(!g) {
      VEE_error("%s # No geometry %d", fname, (int) geoId);
      return;
    }
    
    const Scale * it = & * m_scales.begin();
    const Scale * end = & * m_scales.end();
  
    VEE_Matrix4d tmp = g->transformation();
    
    VEE_trace(fname);

    for(; it != end; it++) {
    
      const Scale & s = *it;

      VEE_Matrix4d scale, toOrigin, fromOrigin;

      scale.clear();
      float scaling = s.m_size.length() / s.m_prevSize.length();

      VEE_trace("%s # scale = %f", fname, scaling);

      if(m_uniform) {
	scale.setDiagonal(VEE_Vector3(scaling));
      }
      else {
	VEE_Vector3 gd = s.m_size;
	double azimuth = atan2(gd.z, gd.x);
	double elevation = -atan2(gd.y, sqrt(gd.x * gd.x + gd.z * gd.z));
	
	VEE_Matrix3d align, tmp, alignInv, scaler;
	align.rotateY(azimuth);
	tmp.rotateZ(elevation);
	align = tmp * align;
	alignInv = align.inverse();
	
	scaler.identity();
	scaler[0][0] = scaling;
	
	/* std::cout << "align:    " << align    << "\n";
	   std::cout << "scaler:   " << scaler   << "\n";
	   std::cout << "alignInv: " << alignInv << "\n";*/
	
	scale.identity();
	scale.setRotation(alignInv * scaler * align);
      }
      
      toOrigin.identity();
      toOrigin.setColumn3(3, -s.m_center);
      
      fromOrigin.identity();
      fromOrigin.setColumn3(3, s.m_center);
      
      tmp = fromOrigin * scale * toOrigin * tmp;
      tmp.addToColumn(3, s.m_move);
      
      // std::cout << "last: " << tmp << std::endl;
    }
    
    g->setTransformation(tmp);
  }
}

bool GEE_ScaleAroundPaths::handleNewVertices(GEE_Path3D::iterator & it1, 
					     GEE_Path3D::iterator & it2)
{
  VEE_trace("GEE_ScaleAroundPaths::handleNewVertices # %d",
	    (int) processedPathVertices());

  const GEE_PathVertex3D & pv1 = GEE_Path3D::getVertex(it1);
  const GEE_PathVertex3D & pv2 = GEE_Path3D::getVertex(it2);

  VEE_Vector3 center = (pv1.m_location + pv2.m_location) * 0.5f;
  VEE_Vector3 size = pv1.m_location - pv2.m_location;

  if(processedPathVertices()) {

    Scale s;

    s.m_move   = center - m_curLoc;
    s.m_center = center;
    s.m_size   = size;
    s.m_prevSize = m_curSize;

    m_scales.push_back(s);
  }
  else {
    m_grabDirection = size;
  }

  m_curLoc = center;
  m_curSize = size;
  

  return true;
}

