/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <gee_world.h>

#include <gee_debug.h>
#include <gee_io.h>
#include <gee_macros.h>

#include <vee_bounding_box.h>
#include <vee_error.h>
#include <vee_render_class.h>

#include <ConfigReader.hpp>

static bool m_debug;

GEE_World::GEE_World(GEE_Engine * host)
  : m_deleted(0),
    m_bgColor(0.2, 0.2, 0.2, 1.0, 0.0),
    m_host(host)
{
  autoId();
  m_debug = GEE_Debug::debug("GEE_World");

  VEE_RenderClass rclass;

  rclass.defaults();

  rclass.setAlign(VEE_ALIGN_BILLBOARD);
  addRenderClass("billboard", & rclass);

  rclass.setAlign(VEE_ALIGN_MOTION_TOTAL);
  addRenderClass("motion", & rclass);

  rclass.setAlign(VEE_ALIGN_FIXED);
  addRenderClass("fixed", & rclass);
}

GEE_World::~GEE_World()
{}

GEE_id_t GEE_World::addGeometry(GEE_Geometry *geometry)
{
  if(!geometry) {
    VEE_error("GEE_World::addGeometry # NULL geometry");
    return -1;
  }

  if(m_deleted)
    geometry->setGarbageBin(m_deleted);

  GEE_id_t id = geometry->id();

  if(id < 0)
    id = GEE_IdContainer::allocateId();

  m_geometry[id] = geometry;
  geometry->setWorld(this);
  geometry->setId(id);

  /* if(isVisual)
     m_visuals[id] = m_geometry[id]; */

  return id;
}

GEE_id_t GEE_World::addGeometry2(VEE_RefPtr<GEE_Geometry> & geometry)
{
  if(!geometry.ptr()) {
    VEE_error("GEE_World::addGeometry2 # NULL geometry");
    return -1;
  }

  GEE_Geometry * d = geometry.ptr();
  
  if(m_deleted)
    d->setGarbageBin(m_deleted);

  GEE_id_t id = d->id();

  if(id < 0)
    id = GEE_IdContainer::allocateId();

  m_geometry[id] = geometry;
  d->setWorld(this);
  d->setId(id);

  /* if(isVisual)
     m_visuals[id] = geometry; */

  return id;  
}

bool GEE_World::eraseGeometry(GEE_id_t id)
{
  iterator it = m_geometry.find(id);
  if(it == m_geometry.end())
    return false;

  getGeometry(it).ptr()->erase();

  m_geometry.erase(it);

  return true;
}

void GEE_World::render(GEE_Renderer *r)
{
  // VEE_trace("GEE_World::render # %u", (uint) m_visuals.size());

  iterator sentinel = m_geometry.end();

  for(iterator it = m_geometry.begin(); it != sentinel; it++) {
    GEE_Geometry * d = (*it).second.ptr();
    if(d->visible())
      d->render(r);
  }
}

void GEE_World::clear()
{
  m_geometry.clear();
}

static const ushort GEE_BgColorId  = GEE_CONST2('B', 'G');
static const ushort GEE_GeoId      = GEE_CONST2('G', 'E');
static const ushort GEE_IdId       = GEE_CONST2('I', 'D');
static const ushort GEE_EndId      = GEE_CONST2('E', 'N');

bool GEE_World::read(GEE_Io *io)
{
  if(m_debug)
    VEE_trace("GEE_World::read");

  std::string str;
  io->readNullTerminatedStdString(str);
  if(str != "GEE_World")
    return false;
  
  int v = io->readInt32();

  if(v != 1)
    return false;

  ushort id = io->readInt16();

  while(id != GEE_EndId) {

    if(id == GEE_BgColorId)
      m_bgColor = io->readRgba();
    else if(id == GEE_IdId)
      m_id = io->readInt64();
    else if(GEE_GeoId) {
      GEE_Geometry * g = GEE_Geometry::readGeometry(io);
      if(g) addGeometry(g);
    }
    else {
      VEE_error("GEE_World::read # Bad id at %d (%c%c)", 
		(int) io->position(), (uchar) (id >> 8), (uchar) id);
      return false;
    }
    
    id = io->readInt16();
  }

  for(iterator it = m_geometry.begin(); it != m_geometry.end(); it++)
    (*it).second.ptr()->finalizeRead();

  if(m_debug)
    VEE_trace("GEE_World::read # EXIT (%d items)", (int) m_geometry.size());

  return true;
}

bool GEE_World::write(GEE_Io *io)
{
  if(m_debug)
    VEE_trace("GEE_World::write");

  io->writeNullTerminatedString("GEE_World");
  
  io->writeInt32(1); // version

  io->writeInt16(GEE_IdId);
  io->writeInt64(m_id);

  io->writeInt16(GEE_BgColorId);
  io->writeRgba(m_bgColor);

  for(iterator it = m_geometry.begin(); it != m_geometry.end(); it++) {
    io->writeInt16(GEE_GeoId);
    (*it).second.ptr()->write(io);
  }

  io->writeInt16(GEE_EndId);

  return true;
}

void GEE_World::collectIds(GEE_IdContainer::id_remap &idMap)
{
  addIdsToRemap(idMap);
  
  for(iterator it = m_geometry.begin(); it != m_geometry.end(); it++)
    (*it).second.ptr()->addIdsToRemap(idMap);
}

void GEE_World::remapIds(GEE_IdContainer::id_remap &idMap)
{
  remapIdMembers(idMap);

  container tmp;
  
  for(iterator it = m_geometry.begin(); it != m_geometry.end(); it++) {
    VEE_RefPtr<GEE_Geometry> & gptr = (*it).second;
    gptr.ptr()->remapIdMembers(idMap);
    tmp[gptr.ptr()->id()] = gptr;
  }
  
  m_geometry.swap(tmp);
}
  
void GEE_World::setAllGenerations(int g)
{
  for(iterator it = m_geometry.begin(); it != m_geometry.end(); it++)
    (*it).second.ptr()->setGeneration(g);
  
}

/** Sets the "nearly selected" bit of all geometries to zero. Useful
    for clearing away old selections. */

void GEE_World::deNearlySelectAll()
{
  for(iterator it = m_geometry.begin(); it != m_geometry.end(); it++)
    (*it).second.ptr()->setNearlySelected(false);
}

void GEE_World::selectAll(uint selectionState)
{
  for(iterator it = m_geometry.begin(); it != m_geometry.end(); it++) {
    GEE_Geometry * g = (*it).second.ptr();
    if(g->visible())
      g->setSelectionState(selectionState); 
  }
}


void GEE_World::dump()
{
  puts("GEE_World::dump {");
  for(iterator it = m_geometry.begin(); it != m_geometry.end(); it++) {
    printf("  %d \t \"%s\"\n", 
	   (int) (*it).first, typeid(*(*it).second.ptr()).name());
  }
  puts("}");
}

static std::map<std::string, int> __texIds;
static std::map<int, std::string> __texIds2;

void GEE_World::addTextureId(const std::string & name, int id)
{
  __texIds[name] = id;
  __texIds2[id] = name;
}

int GEE_World::textureId(const std::string & name)
{
  if(!__texIds.size()) {
    VEE_error("GEE_World::textureId # No textures");
    return -1;
  }

  std::map<std::string, int>::iterator it = __texIds.find(name);
  return it == __texIds.end() ? -1 : (*it).second;
}

std::string GEE_World::textureName(int id)
{
  std::map<int, std::string>::iterator it = __texIds2.find(id);
  return it == __texIds2.end() ? std::string("") : (*it).second;
}

void GEE_World::loadTextureDefinitions(const char * filename)
{
  ConfigReader::Config conf;

  if(filename)
    ConfigReader::readConfig(&conf, filename);
  
  ConfigReader::Chunk chunk = conf.get("TextureParams");

  uint rows = chunk.get("division").getInt(4);
  uint elems = rows * rows;
  
  char buf[64];

  for(uint i = 0; i < elems; i++) {
    
    sprintf(buf, "Texture-%u", i);
    
    chunk = conf.get(buf);

    std::string name = chunk.get("name");
    if(name.size())
      addTextureId(name, i);
  }

  VEE_trace("GEE_World::loadTextureDefinitions # EXIT with %d textures",
	    (int) __texIds.size());
}

typedef std::map<std::string, VEE_RenderClass> rcontainer; 

static rcontainer __renderClasses;

void GEE_World::addRenderClass(const std::string & id, 
			       const VEE_RenderClass * rc)
{
  __renderClasses[id] = * rc;
}

VEE_RenderClass * GEE_World::renderClass(const std::string & id)
{
  rcontainer::iterator it = __renderClasses.find(id);

  if(it == __renderClasses.end())
    return 0;

  return & (*it).second;
}

void getRenderClassNames(std::list<std::string> * classes)
{
  classes->clear();

  for(rcontainer::iterator it = __renderClasses.begin();
      it != __renderClasses.end(); it++)
    classes->push_back((*it).first);
}

void GEE_World::merge(GEE_World &w)
{
  for(iterator it = w.m_geometry.begin(); it != w.m_geometry.end(); it++)
    addGeometry2((*it).second);
}

void GEE_World::allGeometriesToLayer(GEE_id_t layer)
{
  for(iterator it = m_geometry.begin(); it != m_geometry.end(); it++)
    (*it).second.ptr()->setLayer(layer);
}

void GEE_World::getBoundingBox(VEE_BoundingBox *bb)
{
  bb->clear();

  if(!m_geometry.size()) {
    return;
  }

  bool hasOne = false;

  VEE_BoundingBox box;

  for(iterator it = m_geometry.begin(); it != m_geometry.end(); it++) {

    GEE_Geometry * g = (*it).second.ptr();
    
    if(!g->visible())
      continue;
    
    bool hit = g->getBoundingBox(&box);

    if(hit) {

      VEE_Matrix4 tmp(g->transformation().data());

      box = tmp * box;

      if(!hasOne) {
	*bb = box;
	hasOne = true;
      }
      else
	bb->expand(box);
    }
  }
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

void GEE_LayerGeometryIterator::eraseAndNextVisible()
{
  GEE_World::iterator it = m_it;
 
  GEE_id_t current = (*it).first;

  GEE_Geometry * geo = GEE_World::getGeometry(it).ptr();
  
  GEE_id_t topLevel = geo->masterGroup();

  m_world->eraseGeometry(topLevel);

  m_it = m_world->m_geometry.lower_bound(current);
  
  if(m_it == m_end)
    return;
  
  if(m_layer == -1) {
    while((m_it != m_end) && !(*m_it).second.ptr()->visible())
      m_it++;
  }
  else {
    while((m_it != m_end) &&
	  (((*m_it).second.ptr()->layer() != m_layer) ||
	   !(*m_it).second.ptr()->visible()))
      m_it++;
  }
}
