/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */




#include <vee_cg_gl_renderer.h>

#include <vee_error.h>

#ifndef GL_TEXTURE0_ARB
#error "Need GL multitexture ARB"
#endif

const char * VEE_CgParamNames[] = {
  "ModelViewProj",
  "campos",
  "age",
  "color",
  "additive",
  "location",
  "velocity",
  "corner_uv",
  "ScreenScale"
};

VEE_CgGlDriver::VEE_CgGlDriver()
  : m_vertexProgram(0),
    m_fragmentProgram(0)
{}

VEE_CgGlDriver::~VEE_CgGlDriver()
{}

bool VEE_CgGlDriver::createProgram(const char *program,
				   CGcontext context, 
				   CGprofile vp, 
				   CGprofile fp)
{
  const char * fname = "VEE_CgGlDriver::createProgram";

  if(!program) {
    m_programName = "(null)";
    return false;
  }
  
  m_programName = program;

  CGerror err = cgGetError();

  if(err != CG_NO_ERROR)
    VEE_error("%s # When entering (%s)", fname, cgGetErrorString(err));

  std::string prog;


  // Load fragment program:

  prog = program;
  prog += "-fragment.cg";
  m_fragmentProgram = cgCreateProgramFromFile(context, CG_SOURCE, prog.c_str(),
					      fp, 0, 0);
  if(m_fragmentProgram)
    cgGLLoadProgram(m_fragmentProgram);

  err = cgGetError();

  if(err != CG_NO_ERROR) {
    VEE_error
      ("%s # While loading fragment program %s: (%s) %s", 
       fname, prog.c_str(), cgGetErrorString(err), cgGetLastListing(context));
    cgDestroyProgram(m_fragmentProgram);
    m_fragmentProgram = 0;
  }
  else 
    VEE_trace("%s # Loaded fragment program %s", fname, prog.c_str());

  // Load vertex program:
  prog = program;
  prog += "-vertex.cg";

  m_vertexProgram = cgCreateProgramFromFile(context, CG_SOURCE, prog.c_str(),
					    vp, 0, 0);
  if(m_vertexProgram)
    cgGLLoadProgram(m_vertexProgram);
  
  err = cgGetError();

  if(err != CG_NO_ERROR) {
    VEE_error
      ("%s # While loading vertex program %s: (%s) %s", 
       fname, prog.c_str(), cgGetErrorString(err), cgGetLastListing(context));
    cgDestroyProgram(m_vertexProgram);
    m_vertexProgram = 0;
  }
  else 
    VEE_trace("%s # Loaded vertex program %s", fname, prog.c_str());

  if(!m_fragmentProgram && !m_vertexProgram)
    return false;

  // Get parameters:

  for(uint i=0; i < VEE_CGP_SIZEOF; i++) {

    CGparameter tmp = 
      cgGetNamedParameter(m_vertexProgram, VEE_CgParamNames[i]);

    if(tmp) {
      m_params[i] = tmp;
      continue;
    }

    tmp = cgGetNamedParameter(m_fragmentProgram, VEE_CgParamNames[i]);

    m_params[i] = tmp;

  }

  /* if(VEE_CgGlRenderer::debug()) 
     VEE_trace("%s # OK", fname); */

  cgGetError();

  return true;
}

bool VEE_CgGlDriver::bindProgram(VEE_CgGlRenderer *r)
{
  if(m_vertexProgram) {
    cgGLEnableProfile(r->cgVertexProfile());
    cgGLBindProgram(m_vertexProgram);
  }
  else
    cgGLDisableProfile(r->cgVertexProfile());
    
  if(m_fragmentProgram) {
    cgGLEnableProfile(r->cgFragmentProfile());
    cgGLBindProgram(m_fragmentProgram);
  }
  else
    cgGLDisableProfile(r->cgFragmentProfile());

  if(m_params[VEE_CGP_MODELVIEW_PROJECTION])
    cgGLSetStateMatrixParameter(m_params[VEE_CGP_MODELVIEW_PROJECTION], 
				CG_GL_MODELVIEW_PROJECTION_MATRIX, 
				CG_GL_MATRIX_IDENTITY);

  CGerror err = cgGetError();

  if(err != CG_NO_ERROR) {
    VEE_error
      ("VEE_CgGlDriver::bindProgram # While binding %s: (%s)", 
       m_programName.c_str(), cgGetErrorString(err));
    return false;
  }

  if(m_params[VEE_CGP_CAMPOS])
    cgGLSetParameter3fv(m_params[VEE_CGP_CAMPOS], r->camPos().data());

  if(m_params[VEE_CGP_SCREEN_SCALE]) {
    
    float texw = r->depthMap().width();
    float texh = r->depthMap().height();

    float xs = 1.0f / texw, ys = 1.0f / texh;

    cgGLSetParameter2f(m_params[VEE_CGP_SCREEN_SCALE], xs, ys);

    VEE_trace("VEE_CgGlDriver::bindProgram # Screen scale set %f %f",
	      xs, ys);
  }

  err = cgGetError();

  if(err != CG_NO_ERROR) {
    VEE_error
      ("VEE_CgGlDriver::bindProgram # After parameters %s: (%s)", 
       m_programName.c_str(), cgGetErrorString(err));
    return false;
  }

  glActiveTexture(GL_TEXTURE1); 
  r->depthMap().use();
  glActiveTexture(GL_TEXTURE0); 
  
  /* if(VEE_CgGlRenderer::debug()) 
     VEE_trace("VEE_CgGlRenderer::bindProgram # OK"); */
  
  return true;
}

/** Issues the commands that will render a particle using quads. All
    relevant texture modes, rendering modes etc. are assumed to be
    already on. 
    
    You can override this method in your own classes to handle the
    rendering in some other way.
*/

void VEE_CgGlDriver::renderParticle(VEE_Particle *p,
				    VEE_CgGlRenderer *r)
{
  if(!m_vertexProgram || !m_params[VEE_CGP_CAMPOS] || true) {
    r->VEE_GlRenderer::renderParticle(p);
    return;
  }

  VEE_Vector3 pos = p->location();
  const VEE_Vector3 relPos = r->camPos() - pos;
  float dist = relPos.length();
  
  // Skip objects that are not in the range.
  if(dist > r->farDistance() || dist <= r->nearDistance())
    return;
  
  // Set color:
  { 
    VEE_Rgba color = r->transformColor(p->color());
    
    if(dist < r->fadeDistance()) {
      float dfade = (dist - r->nearDistance()) / r->fadeRange();
      color.setAlpha(color.alpha() * dfade);
    }
    
    glColor4fv(color.data());

    if(m_params[VEE_CGP_ADDITIVE])
      cgGLSetParameter1f(m_params[VEE_CGP_ADDITIVE], color.theta());
  }

  // Set other stuff:
  if(m_params[VEE_CGP_AGE])
    cgGLSetParameter1f(m_params[VEE_CGP_AGE], p->age());

  if(m_params[VEE_CGP_VELOCITY])
    cgGLSetParameter3fv(m_params[VEE_CGP_VELOCITY], p->velocity().data());
  
  cgGLSetParameter4f(m_params[VEE_CGP_CORNER_UV], 0.0, 0.0, 0.0, 0.0);
  glVertex3fv(pos.data());

  cgGLSetParameter4f(m_params[VEE_CGP_CORNER_UV], 1.0, 0.0, 1.0, 0.0);
  glVertex3fv(pos.data());

  cgGLSetParameter4f(m_params[VEE_CGP_CORNER_UV], 1.0, 1.0, 1.0, 1.0);
  glVertex3fv(pos.data());

  cgGLSetParameter4f(m_params[VEE_CGP_CORNER_UV], 0.0, 1.0, 0.0, 1.0);
  glVertex3fv(pos.data());
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


bool VEE_CgGlRenderer::m_debug = true;

VEE_CgGlRenderer::VEE_CgGlRenderer(VEE_SceneBox *box)
  : VEE_GlRenderer(box),
    m_cgContext(0),
    m_currentDriver(0)
{
  m_useDepthMap = true;
  // m_farDistance = 10.0f;
}

VEE_CgGlRenderer::~VEE_CgGlRenderer()
{}

void VEE_CgGlRenderer::initializeGL()
{
  if(m_debug) VEE_trace("VEE_CgGlRenderer::initializeGL");

  VEE_GlRenderer::initializeGL();

  m_cgContext = cgCreateContext();
  
  m_cgVertexProfile   = cgGLGetLatestProfile(CG_GL_VERTEX);
  m_cgFragmentProfile = cgGLGetLatestProfile(CG_GL_FRAGMENT);
  
  if (cgGLIsProfileSupported(CG_PROFILE_VP30))
    m_cgVertexProfile = CG_PROFILE_VP30;

  if (cgGLIsProfileSupported(CG_PROFILE_FP30))
    m_cgFragmentProfile = CG_PROFILE_FP30;

  // cgGLSetOptimalOptions(m_cgVertexProfile);
  // cgGLSetOptimalOptions(m_cgFragmentProfile);

  loadShaders();
  m_currentShader = 0;
}

void VEE_CgGlRenderer::endGlMode()
{
  VEE_GlRenderer::endGlMode();

  disableCg();
}

void VEE_CgGlRenderer::renderParticle(VEE_Particle *p)
{
  VEE_RenderClass * rc = p->renderClass();

  int shader = rc ? rc->shader() : 0;
  
  if(!shader) {
    if(m_currentShader)
      disableCg();
    
    VEE_GlRenderer::renderParticleVisual(p);
    return;
  }
  else if(!m_currentShader)
    enableCg();

  // Change shaders if necessary:

  if(shader != m_currentShader) {
    driver_iter dit = m_drivers.find(shader);
    if(dit == m_drivers.end()) {
      VEE_error("VEE_CgGlRenderer::renderParticle # No shader %d", 
		shader);
      return;
    }

    newMode(RenderNothing);

    m_currentShader = shader;
    (*dit).second.ptr()->bindProgram(this);
    m_currentDriver = (*dit).second.ptr();
  }

  // Now do the rendering:
  
  newMode(RenderQuads);
  
  m_currentDriver->renderParticle(p, this);
}

void VEE_CgGlRenderer::disableCg()
{
  cgGLDisableProfile(m_cgVertexProfile);
  cgGLDisableProfile(m_cgFragmentProfile);

  glActiveTexture(GL_TEXTURE1); 
  glDisable(GL_TEXTURE_2D);
  glActiveTexture(GL_TEXTURE0); 

  m_currentShader = 0;
}

void VEE_CgGlRenderer::enableCg()
{
  glActiveTexture(GL_TEXTURE1); 
  m_depthMap.use();
  glActiveTexture(GL_TEXTURE0); 

  /* cgGLEnableProfile(m_cgVertexProfile);
  cgGLEnableProfile(m_cgFragmentProfile);  

  CGerror err = cgGetError();

  if(err != CG_NO_ERROR)
    VEE_error("VEE_CgGlRenderer::enableCg # %s", cgGetErrorString(err));
  */
}


/** This method loads all requested shaders. If the compilation of a
    particular shader fails, then that shader is removed from the
    shader table. */

void VEE_CgGlRenderer::loadShaders()
{  
  for(std::map<std::string, int>::iterator it = m_shaderNames.begin();
      it != m_shaderNames.end(); ) {
    
    std::string sh = (*it).first;

    VEE_CgGlDriver * driver = new VEE_CgGlDriver();

    if(!driver->createProgram(sh.c_str(),
			      m_cgContext,
			      m_cgVertexProfile,
			      m_cgFragmentProfile)) {
      delete driver;
      std::map<std::string, int>::iterator tmp = it;
      tmp++;
      m_shaderNames.erase(it);
      it = tmp;
    }
    else {
      int id = (*it).second;
      
      m_drivers[id] = driver;
      
      VEE_trace("VEE_CgGlRenderer::loadShaders # Added shader %s (%d)", 
		sh.c_str(), id);
      it++;
    }
  }

  if(m_drivers.size())
    m_applyDepthMap = true;
}
