/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <vee_gl_renderer.h>

#include <vee_blow.h>
#include <vee_collector.h>
#include <vee_curve_particle.h>
#include <vee_error.h>
#include <vee_explosion.h>
#include <vee_gl_texture.h>
#include <vee_gravity.h>
#include <vee_gravitate.h>
#include <vee_image.h>
#include <vee_math.h>
#include <vee_matrix2.h>
#include <vee_mgc.h>
#include <vee_orderer.h>
#include <vee_particle.h>
#include <vee_polygon.h>
#include <vee_random.h>
#include <vee_rigid_particle.h>
#include <vee_scene.h>
#include <vee_strs.h>
#include <vee_tail_particle.h>
#include <vee_tcb_spline3.h>
#include <vee_tube.h>
#include <vee_value.h>
#include <vee_value_str.h>
#include <vee_values.h>
#include <vee_vector2.h>
#include <vee_volume_grid.h>
#include <vee_volume_sphere.h>
#include <vee_vortex.h>
#include <vee_wind.h>

#include <vee_grid_force.h>

#include <ConfigReader.hpp>


static VEE_GlTexture2D VEE_blobTexture;
// static VEE_GlTexture2D VEE_triTexture;
static VEE_GlTexture2D VEE_polygonTexture;

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

void VEE_CameraPosition::describe(char * buf)
{
  sprintf(buf, "%.3f, %.3f, %.3f, %.3f, %.3f",
	  m_camPos[0], m_camPos[1], m_camPos[2], m_azi, m_ele);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

static const VEE_Vector3 quadcorners[4] = {
  VEE_Vector3(-0.5, -0.5, 0.0),
  VEE_Vector3(-0.5,  0.5, 0.0),
  VEE_Vector3( 0.5,  0.5, 0.0),
  VEE_Vector3( 0.5, -0.5, 0.0)
};

static const VEE_Vector2 quadtexcoords[4] = {
  VEE_Vector2(0.05, 0.95),
  VEE_Vector2(0.05, 0.05),
  VEE_Vector2(0.95, 0.05),
  VEE_Vector2(0.95, 0.95)
};

const uint explosionVertexCount = 18;
static VEE_Vector3 explosionVertices[explosionVertexCount * 3];

class VEE_GlRendererInit
{
public:

  VEE_GlRendererInit()
  {
    VEE_Vector3 scale(0.06, 0.06, 1.0);

    for(uint i=0; i < explosionVertexCount; i++) {

      for(uint j=0; j < 3; j++) {

	VEE_Vector3 loc(VEE_rand.rand11(), 
			VEE_rand.rand11(), 
			VEE_rand.rand11());
	loc.normalize();
	loc *= scale[j];
	explosionVertices[i * 3 + j] = loc;
      }
    }
  }
};

static VEE_GlRendererInit _veeGlInit;

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

bool VEE_GlRenderer::m_privateTextures = true;


void VEE_GlRenderer::newMode(Mode m)
{
  if(m == m_mode) return;

  m_modeChanges++;

  if(m_mode != RenderNothing)
    glEnd();

  GLenum e = glGetError();
  if(e) 
    VEE_error("VEE_GlRenderer::newMode %p # GL ERROR %s", 
	      this, gluErrorString(e));

  if(m == RenderQuads || m == RenderTriangles) {

    if(m_useTextures && !m_renderOnlyOutline)
      glEnable(GL_TEXTURE_2D);
    else
      glDisable(GL_TEXTURE_2D);
    
    if(m == RenderTriangles)
      glBegin(GL_TRIANGLES);
    else
      glBegin(GL_QUADS);
  }
  else if(m == RenderPoints) {
    glPointSize(m_particleSize * 15.0);
    glDisable(GL_TEXTURE_2D);
    glBegin(GL_POINTS);
  }
  else if(m == RenderLines) {
    glDisable(GL_TEXTURE_2D);

    int lineWidth = (int) (m_particleSize * 20);
    if(lineWidth < 1) lineWidth = 1;

    glLineWidth(lineWidth);

    glBegin(GL_LINES);
  }
  else 
    glDisable(GL_TEXTURE_2D);

  m_mode = m;

}

void VEE_GlRenderer::beginTextured(Mode m, GLuint tex)
{
  if((m_currentTexture == tex) && (m_mode == m))
    return;
  
  if(m_mode != RenderNothing)
    glEnd();
 
  // VEE_trace("VEE_GlRenderer::beginTextured %d %d", (int) m, tex);
 
  if(tex != (GLuint) -1 && m_useTextures) {
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, tex);
    // VEE_trace("VEE_GlRenderer::beginTextured # Enabling textures");
  }
  else
    glDisable(GL_TEXTURE_2D);
  
  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  
  GLenum e = glGetError();
  if(e) 
    VEE_error("VEE_GlRenderer::beginTextured # GL ERROR \"%s\"", 
	      gluErrorString(e));

  else if(m == RenderQuads) {
    glBegin(GL_QUADS);
  }
  
  m_mode = m;
  m_currentTexture = tex;
}

void VEE_GlRenderer::endTextured()
{  
  if(m_mode != RenderNothing)
    glEnd();
 
  GLenum e = glGetError();
  if(e) 
    VEE_error("VEE_GlRenderer::endTextured # GL ERROR 1 %s", 
	      gluErrorString(e));
 
  glDisable(GL_TEXTURE_2D);

  if(m_texId != (GLuint) -1)
    glBindTexture(GL_TEXTURE_2D, m_texId);

  m_currentTexture = (GLuint) -1;
  m_currentMaterial = -1;
  m_mode = RenderNothing;
  
  e = glGetError();
  if(e) 
    VEE_error("VEE_GlRenderer::endTextured # GL ERROR 2 %s", 
	      gluErrorString(e));
}

void VEE_GlRenderer::renderParticleVisual(VEE_ParticleVisualInfo *p)
{
  if(m_particleMode == RenderNothing)
    return;

  VEE_Vector3 pos = p->location();
  VEE_Vector3 velocity = p->velocity();

  if(m_useMatrix) {
    VEE_Matrix4 m = m_matrix;
    VEE_Vector3 projPos = m * pos;
    float speed = velocity.length();
    VEE_Vector3 projPos2 = m * (pos + velocity);
    velocity = projPos2 - projPos;
    velocity.normalize(speed);
    pos = projPos;

    /*VEE_Matrix4 m = m_matrix;
    VEE_trace("[\n  %f %f %f %f\n  %f %f %f %f\n  %f %f %f %f\n  %f %f %f %f]",
	      m[0][0], m[0][1], m[0][2], m[0][3], 
	      m[1][0], m[1][1], m[1][2], m[1][3],
	      m[2][0], m[2][1], m[2][2], m[2][3],
	      m[3][0], m[3][1], m[3][2], m[3][3]);*/
  }

  VEE_RenderClass *rc = p->renderClass();
  if(!rc) rc = & m_defaultRenderClass;

  VEE_MotionAlign align = rc->align();
  
  const VEE_Vector3 relPos = m_camPos - pos;
  float dist = relPos.length();
  
  // Skip objects that are not in the range.
  if(dist > m_farDistance || dist <= m_nearDistance)
    return;
  
  if(!m_renderOnlyOutline) {
    VEE_Rgba color = transformColor(p->color());
    
    if(dist < m_fadeDistance) {
      float dfade = (dist - m_nearDistance) / m_fadeRange;
      color.setAlpha(color.alpha() * dfade);
    }
    
    glColor4fv(color.data());
  }

  /* VEE_trace(
     "VEE_GlRenderer::renderParticle # %p RGBA = [%.3f %.3f %.3f %.3f]",
     p, color.red(), color.green(), color.blue(), color.alpha());*/
  
  newMode(m_particleMode);

  if(m_particleMode == RenderPoints) {
    glVertex3f(pos.x, pos.y, pos.z);
    return;
  }

  if(m_particleMode == RenderLines) {
    glVertex3fv(pos.data());
    glVertex3fv((pos - (velocity * (rc->lengthScale() * 0.3))).data());
    return;
  }

  VEE_Vector3 u = cross(velocity, relPos);
  
  uint  textureRow;
  uint  textureCol;

  int ptex = p->texture();

  m_currentMaterial = ptex;

  if(ptex < 0) {
    textureRow = m_textureRow;
    textureCol = m_textureCol;
  }
  else {
    textureRow = (ptex / m_textureDivision) % m_textureDivision;
    textureCol = ptex % m_textureDivision;
  }

  float s = p->size();

  if(s == 0.0) 
    s = m_particleSize;
  else
    s *= m_particleSize;
    

  VEE_Matrix2 rot;
  rot.rotate(p->rotation());

  if(align == VEE_ALIGN_MOTION_TOTAL) {

    uint i;

    if(u.length() < 0.001)
      return;

    u.normalize();
    
    float vel = velocity.length();

    float stretch = VEE_min(vel * rc->lengthScale(), rc->maxLengthScale());
    if(stretch < 1.0f)
      velocity /= vel;
    else
      velocity *= rc->lengthScale();
    
    for(i = 0; i < 4; i++) {
      
      float tx = quadtexcoords[i][0];
      float ty = quadtexcoords[i][1];
      tx = tx * m_textureStep + m_textureStep * textureCol;
      ty = ty * m_textureStep + m_textureStep * textureRow;
      
      glTexCoord2f(tx, ty);
      
      VEE_Vector3 tmp = quadcorners[i] * s;
      
      tmp = pos + (tmp[0] * u + tmp[1] * velocity);
      
      glVertex3fv(tmp.data());
    }
  }
  else if(align == VEE_ALIGN_BILLBOARD || u.length() < 0.001) {
    uint i;
    float horizontal = sqrtf(relPos.x * relPos.x + relPos.z * relPos.z);
      
    float elevation = atan2f(relPos.y, horizontal);
    float azimuth   = atan2f(relPos.x, relPos.z);
      
    // Kludke:
    /* if(relPos.z < 0.0)
	 elevation = - elevation; */
      
    VEE_Matrix3 rotation, rotation2;
      
    rotation.rotateX(-elevation);
    rotation2.rotateY(azimuth);
    rotation = rotation2 * rotation;
    // rotation.identity();

    float turn = p->rotation();

    if(turn != 0.0f) {
      rotation2.rotateZ(turn);
      rotation = rotation * rotation2;
    }
      
    for(i = 0; i < 4; i++) {
      
      float tx = quadtexcoords[i][0];
      float ty = quadtexcoords[i][1];
      tx = tx * m_textureStep + m_textureStep * textureCol;
      ty = ty * m_textureStep + m_textureStep * textureRow;
      
      // glTexCoord2f(quadtexcoords[i][0], quadtexcoords[i][1]);
      glTexCoord2f(tx, ty);
      glVertex3fv((pos + (rotation * (quadcorners[i] * s))).data());
    }
  }
  else if(align == VEE_ALIGN_MOTION_TEXTURE) {
    // Stretched particles:
    VEE_Vector3 v = cross(u, relPos);
      
    u.normalize();
    v.normalize();

    float stretch = VEE_clamp(VEE_abs(dot(v, velocity)) * rc->lengthScale(),
			      1.0f, rc->maxLengthScale());
      
    v *= stretch;

    uint i;
      
    for(i = 0; i < 4; i++) {
      
      float tx = quadtexcoords[i][1];
      float ty = quadtexcoords[i][0];
      tx = tx * m_textureStep + m_textureStep * textureCol;
      ty = ty * m_textureStep + m_textureStep * textureRow;
      
      glTexCoord2f(tx, ty);
      
      const VEE_Vector3 & tmp3 = quadcorners[i];

      VEE_Vector2 tmp(tmp3.x, tmp3.y);
      tmp *= s;
      tmp = rot * tmp;

      VEE_Vector3 tmp2 = pos + (tmp[1] * u + tmp[0] * v);
      glVertex3fv(tmp2.data());
    }
  }
  else if(align == VEE_ALIGN_MOTION_SHEAR) {
    // Stretched particles:
    float xvel = VEE_abs(dot(velocity, m_vpRight)) * rc->lengthScale();
    float yvel = VEE_abs(dot(velocity, m_vpUp)) * rc->lengthScale();

    VEE_Vector2 dir(xvel, yvel);
    float l = dir.length();
    float maxScale = rc->maxLengthScale();

    if(l > maxScale)
      dir.normalize(maxScale);
    else if(l < 1.0f)
      dir.normalize();

    VEE_Vector2 dir2(dir);
    VEE_Vector2 normal2(yvel, -xvel);
    dir2.normalize();
    normal2.normalize();
    
    for(int i = 0; i < 4; i++) {
      
      float tx = quadtexcoords[i][0];
      float ty = quadtexcoords[i][1];
      tx = tx * m_textureStep + m_textureStep * textureCol;
      ty = ty * m_textureStep + m_textureStep * textureRow;
      
      glTexCoord2f(tx, ty);
      
      VEE_Vector2 tmp = quadcorners[i].data();
      tmp *= s;
      tmp = rot * tmp;

      tmp = Fluid::dot(tmp, dir2) * dir + 
	Fluid::dot(tmp, normal2) * normal2;
      VEE_Vector3 tmp2 = tmp.x * m_vpRight + tmp.y * m_vpUp;

      glVertex3fv((pos + tmp2).data());
    }

  }
  else if(align == VEE_ALIGN_FIXED) {
    VEE_Matrix3 rot;
    rot.rotateAroundAxis(velocity, p->rotation());

    for(int i = 0; i < 4; i++) {
      
      float tx = quadtexcoords[i][0];
      float ty = quadtexcoords[i][1];
      tx = tx * m_textureStep + m_textureStep * textureCol;
      ty = ty * m_textureStep + m_textureStep * textureRow;
      
      glTexCoord2f(tx, ty);
      
      VEE_Vector2 tmp = quadcorners[i].data();
      tmp *= s;

      glVertex3fv((pos + rot * tmp).data());
    }
  }
}

void VEE_GlRenderer::renderTailParticle(VEE_TailParticle *p)
{
  newMode(RenderNothing);

  /* VEE_trace("VEE_GlRenderer::renderTailParticle # %p %d", 
     p, (int) p->pointsToRender()); */

  uint n = p->pointsToRender();

  float s = p->size();
  if(s <= 0.0)
    s = m_particleSize;
    
  if(m_particleMode == RenderPoints) {
    newMode(RenderPoints);

    for(uint i=0; i < n; i++) {
      const VEE_TailParticle::Point *po = & p->getDelayedPoint(i);
      VEE_Rgba color = transformColor(po->color());
      // color.setAlpha(color.alpha() * m_alphaMul);
      glColor4fv(color.data());

      glVertex3fv(po->m_location.data());
    }
  }
  else if(m_particleMode == RenderLines) {

    int lineWidth = (int) (s * 20);
    if(lineWidth < 1) lineWidth = 1;
    glLineWidth(lineWidth);

    float alphaMul = m_alphaMul * p->color().alpha();

    glBegin(GL_LINE_STRIP);
    
    for(uint i=0; i < n; i++) {
      const VEE_TailParticle::Point *po = & p->getDelayedPoint(i);

      VEE_Rgba color = po->color();
      color.setAlpha(color.alpha() * alphaMul);
      glColor4fv(color.data());

      glVertex3fv(po->m_location.data());
    }

    glEnd();
  }
  else if(m_continuousTails) {

    uint  textureRow;
    uint  textureCol;

    int ptex = p->texture();

    if(ptex < 0) {
      textureRow = m_textureRow;
      textureCol = m_textureCol;
    }
    else { 
      textureRow = (ptex / m_textureDivision) % m_textureDivision;
      textureCol = ptex % m_textureDivision;
    }
    
    if(m_useTextures)
      glEnable(GL_TEXTURE_2D);

    glBegin(GL_QUAD_STRIP);

    VEE_Vector3 prevPos;

    for(uint i=0; i < n; i++) {

      const VEE_TailParticle::Point *po = & p->getDelayedPoint(i);
      /* Previous and next points */
      const VEE_TailParticle::Point *ppo = i ? & p->getDelayedPoint(i-1) : 0;
      const VEE_TailParticle::Point *npo = 
	(i + 1) < n ? & p->getDelayedPoint(i + 1) : 0;

      VEE_Vector3 pos = po->m_location;

      if(((pos - prevPos).length() < s * 0.5) && (i != n-1) && i)
	continue;

      {
	VEE_Rgba color = po->color();
	color.setAlpha(color.alpha() * m_alphaMul);
	glColor4fv(color.data());
      }

      VEE_Vector3 velocity;

      if(!ppo && npo) {
	velocity = pos - npo->location();
      }
      else if(!npo && ppo) {
	velocity = ppo->location() - pos;
      }
      else if(npo && ppo) {
	VEE_Vector3 vel1 = ppo->location() - pos;
	VEE_Vector3 vel2 = pos - npo->location();
	vel1.normalize();
	vel2.normalize();
	velocity = vel1 + vel2;
      }
      else 
	velocity = po->velocity();

      const VEE_Vector3 relPos = m_camPos - pos;

      VEE_Vector3 u = cross(velocity, relPos);      
      // VEE_Vector3 v = cross(u, relPos);
      
      u.normalize(s);
      // v.normalize();

      float oldW = (float) i / (float) (n - 1);
      float newW = 1.0f - oldW;
      
      float tx = quadtexcoords[0][0] * newW + quadtexcoords[2][0] * oldW;
      float ty = quadtexcoords[0][1];
      tx = tx * m_textureStep + m_textureStep * textureCol;
      ty = ty * m_textureStep + m_textureStep * textureRow;

      VEE_Vector3 v = pos + u; // Vertex Point

      glTexCoord2f(tx, ty);
      glVertex3f(v[0], v[1], v[2]);

      v = pos - u; // Vertex Point
      ty = quadtexcoords[2][1];
      ty = ty * m_textureStep + m_textureStep * textureRow;
      glTexCoord2f(tx, ty);
      glVertex3f(v[0], v[1], v[2]);
      
      prevPos = pos;
    }

    glEnd();

    if(m_useTextures)
      glDisable(GL_TEXTURE_2D);
  }
  else {
    // Render tail-particles with a large number of particles
    
    float dist = s * m_tailDispersion;
    
    VEE_ParticleVisualInfo visual;

    for(float index = 0.0; index < (float) n; index += dist) {
    
      uint lowInd = (uint) index;
      uint highInd = lowInd + 1;
      float highW = index - lowInd;
      float lowW = 1.0 - highW;
      
      const VEE_TailParticle::Point *lowP  = & p->getDelayedPoint(lowInd);
      const VEE_TailParticle::Point *highP = & p->getDelayedPoint(highInd);
      
      visual.setLocation(lowP->location() * lowW + highP->location() * highW);
      visual.setVelocity(lowP->velocity() * lowW + highP->velocity() * highW);
      
      visual.setColor(lowP->color().mulRgba(lowW).addRgba
	(highP->color().mulRgba(highW)));
      
      renderParticleVisual(&visual);

    }
    
  }
}

void VEE_GlRenderer::renderCurveParticle(VEE_CurveParticle *p)
{
  newMode(RenderNothing);

  uint i, n = VEE_max(p->pointsToRender(), 4u);

  float s = p->size();
  if(s <= 0.0)
    s = m_particleSize;
  
  // Get the control points for spline:
  for(i=0; i < n; i++) {
    m_tcbSpline->setPoint(i, p->location(i));
  }
  m_tcbSpline->setKnots(n);
  m_tcbSpline->recalculate();

  VEE_Vector3 pos = p->location(0);
  
  const VEE_Vector3 relPos = m_camPos - pos;
  float dist = relPos.length();
  
  VEE_Rgba color = p->color();
  
  color.setAlpha(color.alpha() * m_alphaMul);

  if(dist < m_fadeDistance) {
    float dfade = 1.0f - (dist - m_nearDistance) / m_fadeRange;
    color.setAlpha(color.alpha() * dfade);
  }

  glColor4fv(color.data());

  // Render the particle as line strip:

  float step = 0.1;

  if(m_particleMode == RenderLines) {
    glBegin(GL_LINE_STRIP);

    for(float t = 0; t < (float) n - 1; t += step) {
      WmlVector3 v = m_tcbSpline->GetPosition(t);
      glVertex3fv(toFluid(&v)->data());
    }
  /*
    for(i=0; i < n; i++) {
    glVertex3fv(p->location(i).data());
    }*/
    
    glEnd();
  }
  else {

    // Render with one quad strip

    uint  textureRow;
    uint  textureCol;

    int ptex = p->texture();

    if(ptex < 0) {
      textureRow = m_textureRow;
      textureCol = m_textureCol;
    }
    else { 
      textureRow = (ptex / m_textureDivision) % m_textureDivision;
      textureCol = ptex % m_textureDivision;
    }
    
    if(m_useTextures)
      glEnable(GL_TEXTURE_2D);

    glBegin(GL_QUAD_STRIP);
    
    float end = (float) n - 1;
    
    for(float t = 0; t < end; t += step) {
      WmlVector3 ml = m_tcbSpline->GetPosition(t);
      WmlVector3 mv = m_tcbSpline->GetFirstDerivative(t);

      VEE_Vector3 loc = * toFluid(&ml);
      VEE_Vector3 vel = * toFluid(&mv);

      const VEE_Vector3 relPos = m_camPos - loc;
      VEE_Vector3 u = cross(vel, relPos); 
      // VEE_Vector3 v = cross(u, relPos);
      
      u.normalize(s);
      
      float oldW = t / end;
      float newW = 1.0f - oldW;
      
      float tx = quadtexcoords[0][0] * newW + quadtexcoords[2][0] * oldW;
      float ty = quadtexcoords[0][1];
      tx = tx * m_textureStep + m_textureStep * textureCol;
      ty = ty * m_textureStep + m_textureStep * textureRow;
      
      VEE_Vector3 v = loc + u; // Vertex Point

      glTexCoord2f(tx, ty);
      glVertex3f(v[0], v[1], v[2]);


      v = loc - u; // Vertex Point
      ty = quadtexcoords[2][1];
      ty = ty * m_textureStep + m_textureStep * textureRow;
      glTexCoord2f(tx, ty);
      glVertex3f(v[0], v[1], v[2]);
  
    }
    
    glEnd();
  }
}

void VEE_GlRenderer::renderRigidParticle(VEE_RigidParticle *p)
{
  // VEE_trace("VEE_GlRenderer::renderRigidParticle # %p", p);

  meshStore::iterator it = m_meshes.find(p->mesh());

  if(it == m_meshes.end())
    it = m_meshes.find(0);

  Solar::GraphicTriangleMesh * mesh = & (*it).second;

  newMode(RenderNothing);
  
  glPushAttrib(~0);

  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);

  // glColor4fv(p->color().data());
  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, p->color().data());

  glPushMatrix();

  VEE_Vector3 v = p->location();

  glTranslatef(v.x, v.y, v.z);
  
  v = p->rotationAxis();
  glRotatef(VEE_rad2deg(p->rotation()), v.x, v.y, v.z);
  
  float s = p->size();

  if(s <= 0.0) 
    s = m_particleSize;
  /* else
     s *= m_particleSize; */
    
  glScalef(s, s, s);

  int tex = p->texture();

  if(tex < 0)
    glDisable(GL_TEXTURE_2D);

  const uint stride = sizeof(Solar::GraphicTriangleMesh::VertexData);
  
  glVertexPointer(3, GL_FLOAT, stride, mesh->vertexPtr());
  glNormalPointer(GL_FLOAT, stride, & (mesh->vertexPtr()->normal()));
  glTexCoordPointer(2, GL_FLOAT, stride, & (mesh->vertexPtr()->uv()));

  const Solar::GraphicTriangleMesh::Triangle *tri = & mesh->triangle(0);
  const Solar::GraphicTriangleMesh::Triangle *sentinel = 
    tri + mesh->nofTriangles();

  // VEE_Rgba color = transformColor(p->color());

  glBegin(GL_TRIANGLES);
  
  while(tri != sentinel) {

    glArrayElement(tri->vertex(0));
    glArrayElement(tri->vertex(1));
    glArrayElement(tri->vertex(2));

    tri++;

    // VEE_trace("VEE_GlRenderer::renderRigidParticle # %p %p", p, tri);
  }

  glEnd();

  glPopMatrix();
  glPopAttrib();
}

VEE_GlRenderer::VEE_GlRenderer(VEE_SceneBox *box)
  : m_tcbSpline(new VEE_TCBSpline3(1000)),
    m_modeChanges(0),
    m_blobTexture(&VEE_blobTexture),
    m_polygonTexture(&VEE_polygonTexture),
    m_configFile("vee-gl.config")    
{
  m_cylinder = gluNewQuadric();
  m_blend = true;
  m_useTextures = true;
  m_mode = RenderNothing;

  m_farDistance = 500.0;
  m_fadeDistance = 0.5;
  m_nearDistance = 0.1;
  m_fadeRange = m_fadeDistance - m_nearDistance;
  m_renderOnlyOutline = false;
  
  m_texId = (GLuint) -1;
  m_currentTexture = (GLuint) -1;
  m_currentMaterial = -1;

  m_camPos.make(0.0, 1.5, 10.0);
  
  m_velocity.clear();

  m_azi = VEE_PI;
  // m_azi = 0.0;
  m_ele = 0.0;

  m_camUp.x = 0.0;
  m_camUp.y = 1.0;
  m_camUp.z = 0.0;

  m_direction.make(0.0, 0.0, 1.0);
  
  m_forceFrontColor = 0;
  m_forceBackColor = 0;

  m_mode = RenderNothing;
  m_particleMode = RenderQuads;
  m_forceMode = RenderPoints;
  // m_polygonMode = RenderLines;
  m_polygonMode = RenderQuads;
  m_drawForceBoxes = false;

  m_pointSize = 7.0;
  m_particleSize = 1.0;
  // m_lengthScale = .2;
  // rc->maxLengthScale() = 4.0;
  m_alphaMul = 1.0;
  m_fade = 0.0;
  m_fixedFade = 0.0;

  m_fov = 70.0;

  // m_stretchParticles = true;
  // m_alignParticlesWithVelocity = false;
  m_continuousTails = true;
  m_tailDispersion = 1.0;
  m_renderOrder = OrderBackToFront1;

  m_vps = box;

  m_textureDivision = 1;
  m_textureRow = 0;
  m_textureCol = 0;
  m_textureStep = 1.0f;  

  m_bgColor = VEE_Rgba::Black;
  m_blendType = VEE_BLEND_NEW;

  // Generate a default mesh = a single triangle

  Solar::GraphicTriangleMesh mesh;
  Solar::GraphicTriangleMesh::VertexData vertex;

  vertex.m_normal.make(0.0, 0.0, 1.0);
  vertex.m_location.make(-0.5, -0.5, 0.0);
  vertex.m_uv.make(0.0, 0.0);
  mesh.addVertex(vertex);

  vertex.m_location.make(+0.5, -0.5, 0.0);
  vertex.m_uv.make(1.0, 0.0);
  mesh.addVertex(vertex);

  vertex.m_location.make(+0.5, 1.0, 0.0);
  vertex.m_uv.make(0.5, 1.0);
  mesh.addVertex(vertex);

  mesh.addTriangle(0, 1, 2);

  m_meshes[0] = mesh;

  m_matrix.identity();
  m_useMatrix = false;

  m_useDepthMap = false;
  m_applyDepthMap = false;
  // m_blendType = VEE_BLEND_ADDITIVE;
}

void VEE_GlRenderer::initRenderState()
{
  prepareGlMode();
}

void VEE_GlRenderer::closeRenderState()
{
  endGlMode();  
}

void VEE_GlRenderer::renderParticle(VEE_Particle *p)
{
  renderParticleVisual(p);
}  

void VEE_GlRenderer::renderPolygon(VEE_Polygon *p)
{

  // VEE_Vector3 *pos = p->centralLocation();

  // glVertex3fv(*pos);
  
  uint tricount = p->triangleCount();

  if(m_polygonMode == RenderLines) {

    newMode(RenderLines);
  
    glColor4f(1.0,1.0,1.0,1.0);

    // Render the poor bastard with lines -- it is easier to follow what
    // is happening:
    
    for(uint i=0; i < tricount; i++) {
      const uint * tri = p->getTriangle(i);
      
      glVertex3fv(p->getVertex(tri[0]).data());
      glVertex3fv(p->getVertex(tri[1]).data());

      glVertex3fv(p->getVertex(tri[1]).data());
      glVertex3fv(p->getVertex(tri[2]).data());

      glVertex3fv(p->getVertex(tri[2]).data());
      glVertex3fv(p->getVertex(tri[0]).data());
    }

  }
  else if(m_polygonMode == RenderTriangles) {

    // p->printTextureUV();

    const VEE_GlMaterial & mat = m_materials.getMaterial(p->getMaterial());

    beginTextured(RenderTriangles, mat.textureId());

    VEE_Rgba color;

    if(p->looks()) color = p->looks()->color();
    else color = VEE_Rgba::White;
    
    color.setAlpha(color.alpha() * m_alphaMul);

    glColor4fv(color.data());
    
    for(uint i=0; i < tricount; i++) {

      const uint * tri = p->getTriangle(i);
      
      /*
      for(uint j = 0; j < 3; j++)
	glArrayElement(tri[j]);
      */

      uint v0 = tri[0];
      uint v1 = tri[1];
      uint v2 = tri[2];


      glTexCoord2fv(p->getTextureCoordinate(v0).data());
      glVertex3fv(p->getVertex(v0).data());

      glTexCoord2fv(p->getTextureCoordinate(v1).data());
      glVertex3fv(p->getVertex(v1).data());

      glTexCoord2fv(p->getTextureCoordinate(v2).data());
      glVertex3fv(p->getVertex(v2).data());

    }
    endTextured();
  }
}

static std::string vortexTex("vortex-force");
    
void VEE_GlRenderer::renderForceObject(VEE_ForceObject *f)
{
  if(m_forceMode == RenderNothing) 
    return;
  
  // static const char * fname = "VEE_GlRenderer::renderForceObject";

  VEE_Blow  *blow     = dynamic_cast<VEE_Blow *> (f);
  VEE_Vortex *vortex  = dynamic_cast<VEE_Vortex *> (f);
  VEE_Tube  *tube     = dynamic_cast<VEE_Tube *> (f);
  VEE_Wind  *wind     = dynamic_cast<VEE_Wind *> (f);
  VEE_GridForce *grid = dynamic_cast<VEE_GridForce *> (f);
  VEE_Explosion *explosion = dynamic_cast<VEE_Explosion *> (f);

  
  // glColor4f(0.5,1.0,0.5,1.0);

  newMode(RenderNothing);
  glLineWidth(2);

  if(blow) {

    VEE_Matrix3 rot;
    rot.rotateY(VEE_HALF_PI);
    
    pushTransform((*blow->rotation()) * rot,
		  blow->location());
    
    glPushAttrib(~0);

    if(m_forceFrontColor)
      glColor4fv(m_forceFrontColor->data());
    else
      glColor4f(1.0, 0.3, 0.0, 1.0);

    uint sides = 16;

    float range = blow->range();
    float offset = fmod(0.3 * blow->blown(), range * 0.5);

    glDisable(GL_TEXTURE_2D);
    glDisable(GL_BLEND);
    glDepthMask(GL_TRUE);

    glPushMatrix();
    glTranslatef(0.0, 0.0, offset);
    gluQuadricDrawStyle(m_cylinder, GLU_FILL);

    for(uint n=0; n < 2; n++) {
      gluDisk(m_cylinder, range * 0.22, range * 0.25, sides, 1);
      glTranslatef(0.0, 0.0, range * 0.5);
    }

    glPopMatrix();
    
    glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    glEnable(GL_BLEND);
    glDepthMask(GL_FALSE);
    glDisable(GL_CULL_FACE);
    glDisable(GL_LIGHTING);    

    const VEE_GlMaterial & mat = m_materials.getMaterial(BLOW);

    if(mat.hasTexture()) {
      gluQuadricDrawStyle(m_cylinder, GLU_FILL);
      gluQuadricTexture(m_cylinder, true);
      mat.useTexture();
    }
    else {
      gluQuadricDrawStyle(m_cylinder, GLU_LINE);
      gluQuadricTexture(m_cylinder, false);
    }
    
    gluCylinder(m_cylinder, range * 0.25, range * 0.25, range, sides, 2);

    glPopAttrib();

    popTransform();
  }
  else if(vortex) {

    VEE_Vector3 axes = vortex->getAxes();
    axes.normalize();

    VEE_Vector3 rotateAxis = VEE_Vector3(0.0, 0.0, 1.0) + axes;

    float range  = vortex->getRange();
    float vheight = vortex->height();

    int sides = 8;
    
    const VEE_GlMaterial & mat = m_materials.getMaterial(SWIRL);

    pushTransform(VEE_Matrix3IDENTITY,
		  vortex->location());

    glRotatef(180.0f, rotateAxis[0], rotateAxis[1], rotateAxis[2]);

    glTranslatef(0.0, 0.0, -vheight);
    float angle = vortex->vortexed() * 90.0f / VEE_TWO_PI;

    // VEE_trace("%s # vortex angle = %f", fname, angle);

    glRotatef(angle, 0.0f, 0.0f, 1.0f);

    glPushAttrib(~0);

    if(m_forceFrontColor)
      glColor4fv(m_forceFrontColor->data());
    else
      glColor4f(0.1, 0.3, 1.0, 1.0);

    // Render the fans

    glDisable(GL_TEXTURE_2D);
    glDisable(GL_BLEND);
    glDepthMask(GL_TRUE);

    float xlow = range * 0.8;

    for(uint n = 0; n < 4; n++) {

      glRotatef(90, 0.0f, 0.0f, 1.0f);

      glBegin(GL_QUADS);

      glVertex3f(xlow,  0, 0);
      glVertex3f(range, 0, 0);
      glVertex3f(range, 0, vheight * 2.0f);
      glVertex3f(xlow,  0, vheight * 2.0f);

      glEnd();    
    }

    // Render the tube
    glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    glEnable(GL_BLEND);
    glDepthMask(GL_FALSE);
    glDisable(GL_CULL_FACE);
    glDisable(GL_LIGHTING);

    if(mat.hasTexture()) {

      gluQuadricDrawStyle(m_cylinder, GLU_FILL);
      gluQuadricTexture(m_cylinder, true);

      mat.useTexture();
      sides = 32;
    }
    else {
      gluQuadricDrawStyle(m_cylinder, GLU_LINE);
      gluQuadricTexture(m_cylinder, false);
    }

    gluCylinder(m_cylinder, range, range, vortex->height() * 2.0f, sides, 2);

    glPopAttrib();

    glDisable(GL_TEXTURE_2D);

    // Render the center line
    glBegin(GL_LINES);

    glVertex3f(0, 0, 0);
    glVertex3f(0, 0, vortex->height() * 2.0f);

    glEnd();

    popTransform();
    
  }
  else if(tube) {
    
    pushTransform(tube->rotation(),
		  tube->location());

    glRotatef(-90.0, 1.0, 0.0, 0.0);
    
    glTranslatef(0.0, 0.0, -tube->height());
    
    glColor4f(0.0, 0.6, 1.0, 1.0);
    gluQuadricDrawStyle(m_cylinder, GLU_LINE);
    gluCylinder(m_cylinder, 
		tube->radius(), tube->radius(),  
		tube->height() * 2.0, 6, 2);
    popTransform();

  }
  else if(wind) {
    if(!wind->volume()) 
      return;

    VEE_VolumeGrid * vg = dynamic_cast<VEE_VolumeGrid *> (wind->volume());

    if(vg) {
      drawVolumeGrid(wind->location(), vg);
      return;
    }
    
    VEE_VolumeGrid2 * vg2 = dynamic_cast<VEE_VolumeGrid2 *> (wind->volume());

    if(vg2) {
      drawVolumeGrid2(wind->location(), vg2);
      return;
    }

    VEE_VolumeSphere * vs = dynamic_cast<VEE_VolumeSphere *> (wind->volume());

    if(!vs) return;

    pushTransform(VEE_Matrix3IDENTITY,
		  wind->location());

    gluQuadricDrawStyle(m_cylinder, GLU_LINE);
    gluSphere(m_cylinder, vs->outerRadius(), 6, 6);

    popTransform();
  }
  else if(grid) {

    pushTransform(grid->rotation(), grid->location());
    
    VEE_Vector3 scale;
    scale.make(grid->totalScale());

    newMode(RenderLines);
    drawBox(-scale, scale);
    newMode(RenderNothing);

    VEE_VectorGrid::grid3d *vg = & grid->grid().grid();
    
    VEE_Vector3 cellSize = scale;

    cellSize[0] /= vg->cols();
    cellSize[1] /= vg->rows();
    cellSize[2] /= vg->layers();

    newMode(RenderLines);
    
    // VEE_Vector3 low = grid->location() - scale * 0.5 + cellSize * 0.5;

    VEE_Vector3 low = scale * -0.5 + cellSize * 0.5;
 
    for(uint i=0; i < vg->layers(); i++) {
      for(uint j=0; j < vg->rows(); j++) {
	for(uint k=0; k < vg->cols(); k++) {
	  float d = vg->get(i,j,k).m_distance;
	  if(d < 0.1) {
	    VEE_Vector3 gradient = vg->get(i,j,k).m_gradient;
	    VEE_Vector3 kji;

	    kji.make(k, j, i);
	    
	    VEE_Vector3 startPoint = kji;
	    startPoint.scale(cellSize);
	    startPoint += low;

	    glColor3f(0.0, 0.0, 1.0);
	    glVertex3fv(startPoint.data());
	    glColor3f(1.0, 0.7, 0.0);
	    glVertex3fv((startPoint + gradient * 0.3).data());
	  }
	    
	}
      }
    }

    newMode(RenderNothing);
    popTransform();

  }
  else if (explosion) {
    newMode(RenderNothing);

    glPushAttrib(~0);

    const VEE_Vector3 & loc = explosion->location();
    float range = explosion->currentShockRange();

    glPushMatrix();
    glTranslatef(loc.x, loc.y, loc.z);
    glScalef(range, range, range);

    glDisable(GL_TEXTURE_2D);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    glBegin(GL_TRIANGLES);
    
    if(m_forceFrontColor)
      glColor4fv(m_forceFrontColor->data());
    else
      glColor4f(1.0, 0.05, 0.05, 0.5);

    for(uint i=0; i < explosionVertexCount * 3; i++) {
      glVertex3fv(explosionVertices[i].data());
    }

    glEnd();

    glPopMatrix();
    glPopAttrib();
  }
  else {
  
    newMode(RenderPoints);

    VEE_Vector3 *pos = f->centralLocation();
    
    glVertex3fv(pos->data());
  }

  if(m_drawForceBoxes) {
    
    newMode(RenderLines);

    VEE_Vector3 loc = f->location();
    
    VEE_ForceObject::boxStore * boxes = & f->boxes();
    
    // if(boxes->size() > 50) return;

    glColor4f(1.0,1.0,1.0,1.0);
    
    for(boxes->toFirst(); !boxes->isAtEnd(); boxes->advance()) {
      VEE_Vector3 center = boxes->current()->boundingBox().center();

      glVertex3fv(loc.data());
      glVertex3fv(center.data());
    }
  }
}

VEE_GlRenderer::~VEE_GlRenderer()
{
  VEE_trace("VEE_GlRenderer::~VEE_GlRenderer # %p", this);
  gluDeleteQuadric(m_cylinder);
  delete m_tcbSpline;
}

/** Receive and process a value. 

    Parameters:

    <UL>
  
    <LI> "size", float, range 1e-5 to 100
  
    <LI> "alphamul", float, range 1e-5 to 1

    <LI> "fov" (Field of view), float, range 10 to 175
  
    <LI> "fade", float, range 0 to 10
  
    <LI> "fixedfade", float, range 0 to 1
  
    <LI> "lengthscale" (Field of view), float, range 0 to 100
  
    <LI> "subtexture", int, range 0 to subTextureCount() - 1 
  
    <LI> "stretch" (Should we stretch textured particles), boolean (int)
  
    <LI> "renderorder", int, range 0 to OrderSizeof
  
    <LI> "blendtype", int, 0 to VEE_BLEND_USUAL

    <LI> "bgcolor" (background color) , RGBA or vector
  
    </UL>

*/

bool VEE_GlRenderer::setValue(const char *id, VEE_Value *v)
{
  // const char * fname = "VEE_GlRenderer::setValue";

  bool ok = false;

  if(VEE_Strs::equals(id, VEE_Strs::size)) {
    float tmp;
    ok = v->getFloat(&tmp);
    
    if(ok)
      m_particleSize = VEE_clamp(tmp, 1.0e-5f, 1000.0f);
  }
  else if(VEE_Strs::equals(id, "alphamul")) {
    float tmp;
    ok = v->getFloat(&tmp);
    
    if(ok)
      m_alphaMul = VEE_clamp(tmp, 1.0e-5f, 10.0f);
  }
  else if(VEE_Strs::equals(id, "fov")) {
    float tmp;
    ok = v->getFloat(&tmp);
    
    if(ok) {
      m_fov = VEE_clamp(tmp, 5.0f, 175.0f);
      resizeGL(m_width, m_height);
    }
  }
  else if(VEE_Strs::equals(id, "fade")) {
    float tmp;
    ok = v->getFloat(&tmp);
    
    if(ok) {
      m_fade = VEE_clamp(tmp, 0.0f, 10.0f);
    }
  }
  else if(VEE_Strs::equals(id, "fixedfade")) {
    float tmp;
    ok = v->getFloat(&tmp);
    
    if(ok) {
      m_fixedFade = VEE_clamp(tmp, 0.0f, 1.0f);
    }
  }
  else if(VEE_Strs::equals(id, "lengthscale")) {
    float tmp;
    ok = v->getFloat(&tmp);
    
    if(ok)
      m_defaultRenderClass.setLengthScale(VEE_clamp(tmp, 0.0f, 100.0f));
  }
  else if(VEE_Strs::equals(id, "maxlengthscale")) {
    float tmp;
    ok = v->getFloat(&tmp);
    
    if(ok)
      m_defaultRenderClass.setMaxLengthScale(VEE_max(tmp, 1.0f));
  }
  else if(VEE_Strs::equals(id, "subtexture")) {
    int tmp;
    ok = v->getInt(&tmp);
    
    if(ok) {
      m_textureRow = tmp / m_textureDivision;
      m_textureCol = tmp % m_textureDivision;

      /* VEE_trace("%s # New sub-texture %u %u (%d)", 
	 fname, m_textureRow, m_textureCol, tmp); */
    }
  }
  /* else if(VEE_Strs::equals(id, "stretch")) {
    int tmp;
    ok = v->getInt(&tmp);
    
    if(ok)
      m_stretchParticles = (tmp != 0);
      
      }*/
  else if(VEE_Strs::equals(id, "taildispersion")) {
    float tmp;
    ok = v->getFloat(&tmp);
    
    if(ok)
      m_tailDispersion = tmp;
  }
  else if(VEE_Strs::equals(id, "continuoustails")) {
    int tmp;
    ok = v->getInt(&tmp);
    
    if(ok)
      m_continuousTails = (tmp != 0);
  }
  else if(VEE_Strs::equals(id, "renderorder")) {
    int tmp;
    ok = v->getInt(&tmp);
    
    if(ok)
      m_renderOrder = (Order) VEE_clamp(tmp, 0, OrderSizeof - 1);
  }
  else if(VEE_Strs::equals(id, "blendtype")) {
    int tmp;
    ok = v->getInt(&tmp);
    
    if(ok)
      m_blendType = (VEE_Blending) tmp;
  }
  else if(VEE_Strs::equals(id, "config")) {
    VEE_ValueString * vs = dynamic_cast<VEE_ValueString *> (v);

    if(vs) {
      m_configFile = vs->string();
      ok = true;
    }
    else
      ok = false;
      
  }
  else if(VEE_Strs::equals(id, "bgcolor")) {

    VEE_ValueRgba * rgba = dynamic_cast<VEE_ValueRgba *> (v);
    
    if(rgba) {
      m_bgColor = rgba->getValue();
      return true;
    }

    VEE_Vector3 tmp;

    ok = v->getVector3(&tmp);
    
    if(ok) {
      m_bgColor.set(tmp[0], tmp[1], tmp[2], 1.0);
    }
  }

  return ok;
}

void VEE_GlRenderer::initializeGL()
{
  if(m_blobTexture && m_polygonTexture) {
    if(!m_blobTexture->isDefined()) {
      buildTextures(m_blobTexture, 0.45, m_configFile.c_str());
      // buildTextures(m_triTexture,  0.2887, m_configFile.c_str()); // tan(pi / 6) * 0.5
      buildPolygonTextures(m_polygonTexture);
    }
  }

  VEE_Image im, im2;
  VEE_GlTexture2D tex;
  unsigned char color1 = 0xFF;
  unsigned char color2 = 0x20;

  im.resize(32, 32, 1);
  im2.resize(32, 32, 1);
  im.skotchTexture(6, 32, & color1, & color2);
  im2.skotchTexture(2, 8, & color1, & color2);
  im.add(im2);
  tex.genTexture(&im, true);
  
  m_materials.addMaterial(CHECK, VEE_GlMaterial(tex.id()));

  GLenum e = glGetError();
  if(e) 
    VEE_error("VEE_GlRenderer::initializeGL # GL ERROR %s", 
	      gluErrorString(e));

  glClearColor(m_bgColor.red(), m_bgColor.green(), m_bgColor.blue(), 
	       m_bgColor.alpha());
  glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}

void VEE_GlRenderer::paintGL (bool drawBoxEdges, 
			      bool moveCamera, 
			      bool clear, 
			      float dt)
{
  const char * fname = "VEE_GlRenderer::paintGL";

  m_drawBoxEdges = drawBoxEdges;

  if(!m_privateTextures) 
    VEE_fatal("VEE_GlRenderer::paintGL # Must have private textures" );

  if(moveCamera) {
    glLoadIdentity ();

    if(m_fade > 0.005 || m_fixedFade > 0.005) {

      /* GLint drawBuf = 0;
	 // Just a test
      glGetIntegerv(GL_DRAW_BUFFER, &drawBuf);

      if(drawBuf == GL_BACK)
	VEE_trace("Back");
      else if(drawBuf == GL_FRONT)
	VEE_trace("Front");
      else
      VEE_trace("Unknown: %d", (int) drawBuf);*/
	
      float attenuationPerSecond = powf(0.05, 1.0f / m_fade);

      float alpha = 1.0f - powf(attenuationPerSecond, dt);

      if(m_fixedFade > 0.0) {
	alpha = 1.0f - m_fixedFade;
      }

      // This should not affect anything:
      /* glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
      
      // Set the matrices for bit transfers:
      glMatrixMode (GL_PROJECTION);
      glLoadIdentity ();
      gluOrtho2D(0, m_width, 0, m_height);
      glMatrixMode (GL_MODELVIEW);
      glLoadIdentity ();*/

      /* glReadBuffer(GL_FRONT);
	 glDrawBuffer(GL_BACK);*/
      
      if(alpha > 0.01) {

	// glRasterPos3f(0, 3, 0);
	// glCopyPixels(0, 0, m_width, m_height, GL_COLOR);
	
	glDisable(GL_CULL_FACE);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
  
	// glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	// glClear (GL_DEPTH_BUFFER_BIT);
	glBlendEquation(GL_FUNC_ADD);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	
	glColor4f(m_bgColor.red(), m_bgColor.green(), m_bgColor.blue(), 
		alpha);
	// glColor4f(0.0, 0.0, 1.0, 1.0);
	
	glBegin(GL_TRIANGLE_STRIP);
	
	glVertex3f(-50, -50, -2);
	glVertex3f(-50,  50, -2);
	glVertex3f( 50, -50, -2);
	glVertex3f( 50,  50, -2);
	
	glEnd();
	
      // Reset old settings:
      /* glMatrixMode (GL_PROJECTION);
	 glLoadIdentity ();
	 
	 gluPerspective (m_fov,  
	 (float) m_width / (float) m_height,
	 0.1,  
	 500.0); 
	 
	 glMatrixMode (GL_MODELVIEW);*/

	if(!m_blend)
	  glDisable(GL_BLEND);
      }
    }
    
    moveGlCamera();
  }
    
  if(clear && m_fade <= 0.005 && m_fixedFade <= 0.005) {
    clearGlScreen();
  }

  setTexId();

  prepareGlMode();
  
  glPushMatrix();

  if(clear) {
    newMode(RenderNothing);
    glColor4f(1.0,1.0,1.0,1.0);
  }

  if(m_renderOrder == OrderRandom1)
    drawRecursed(m_vps);
  else if(m_renderOrder == OrderRandom2) {
    VEE_Collector collector;
    collector.addScene(m_vps);
    drawRandom(&collector);
  }
  else if(m_renderOrder == OrderBackToFront1)
    drawBackToFront(m_vps);
  else if(m_renderOrder == OrderBackToFront2) {
    VEE_OrdererCollector collector;
    collector.setCameraPosition(m_camPos);
    collector.addScene(m_vps);
    drawBackToFront2(&collector);
  }
  else
    VEE_error("%s # Render Order %d is unknown", fname, (int) m_renderOrder);
  
  endGlMode();

  glPopMatrix();

  //   glFlush();
}


void VEE_GlRenderer::resizeGL (GLsizei w, GLsizei h)
{
  glViewport (0, 0, w, h);


  /* Aseta projektio */

  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();               /* Aloitetaan puhtaalta p|yd{lt{ */
  
  gluPerspective (m_fov,  /* N{k|kent{n leveys asteina */
                  (float) w / (float) h, 
                  m_nearDistance,   /* near plane */
                  m_farDistance);  /* far plane */

  m_width = w;
  m_height = h;

  glMatrixMode (GL_MODELVIEW);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

#ifdef GL_TEXTURE1
  if(m_useDepthMap) {
    glActiveTexture(GL_TEXTURE1); 
    m_depthMap.allocateArea(w, h, GL_DEPTH_COMPONENT);
    // m_depthMap.allocateArea(w, h, GL_RGB);
    glDisable(GL_TEXTURE_2D);
    glActiveTexture(GL_TEXTURE0);     
  }
#endif
  // glFlush();  /* Varmistetaan, ett{ kaikki menee perille X-serveriin. */
}

void VEE_GlRenderer::prepareGlMode()
{
  if(m_blend)
    glEnable(GL_BLEND);
  else
    glDisable(GL_BLEND);
  
  glDisable(GL_CULL_FACE);
  glDisable(GL_LIGHTING);
  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, 1);
  
  glEnable(GL_DEPTH_TEST);
  glDepthMask(GL_FALSE);
  
  if(m_blendType == VEE_BLEND_NEW) {
    glBlendEquation(GL_FUNC_ADD);
    glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
  }
  else if(m_blendType == VEE_BLEND_ADDITIVE) {
    glBlendEquation(GL_FUNC_ADD);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE);
  }
  else if(m_blendType == VEE_BLEND_SUBTRACTIVE) {
    glBlendEquation(GL_FUNC_REVERSE_SUBTRACT);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE);
  }
  else if(m_blendType == VEE_BLEND_USUAL) {
    glBlendEquation(GL_FUNC_ADD);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  }

  glClearColor(m_bgColor.red(), m_bgColor.green(), m_bgColor.blue(), 
	       m_bgColor.alpha());

  if(!m_renderOnlyOutline) {
    int lineWidth = VEE_max((int) (m_particleSize * 20), 1);
    glLineWidth(lineWidth);
  }

  if(m_texId != (GLuint) -1)
    glBindTexture(GL_TEXTURE_2D, m_texId);

  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

  /* glEnableClientState(GL_VERTEX_ARRAY);
     glEnableClientState(GL_NORMAL_ARRAY);
     glEnableClientState(GL_TEXTURE_COORD_ARRAY); */
  // glEnable(GL_COLOR_MATERIAL);  
  
  m_mode = RenderNothing;

  GLenum e = glGetError();
  if(e) 
    VEE_error("VEE_GlRenderer::prepareGlMode # GL ERROR %s", 
	      gluErrorString(e));

  m_currentTexture = (GLuint) -1;
  m_currentMaterial = -1;
  glDisable(GL_TEXTURE_2D);

  VEE_Vector3 vpRight = cross(m_direction, m_camUp);
  vpRight.normalize();
  VEE_Vector3 vpUp = cross(vpRight, m_direction);
  vpUp.normalize();

  m_vpUp    = vpUp;
  m_vpRight = vpRight;
}

void VEE_GlRenderer::endGlMode()
{
  if(m_mode != RenderNothing) {
    glEnd();
    m_mode = RenderNothing;
  }

  GLenum e = glGetError();
  if(e) 
    VEE_error("VEE_GlRenderer::endGlMode # GL ERROR 1 %s", 
	      gluErrorString(e));

  glDepthMask(GL_TRUE);
}

void VEE_GlRenderer::drawRecursed(VEE_SceneBox *scene)
{
  // const char * fname = "VEE_GlRenderer::drawRecursed";
  
  if(m_drawBoxEdges) {

    newMode(RenderNothing);
    glDisable(GL_TEXTURE_2D);
    
    VEE_Vector3 low = scene->boundingBox().low();
    VEE_Vector3 high = scene->boundingBox().high();

    
    glLineWidth(2);

    glBegin(GL_LINES);
    glColor4f(1.0,0.0,0.0,1.0);
    drawBox(low, high);
    glEnd();
  }

  if(!scene->isSplit() && !scene->numObjects())
    return;
  
  {

    VEE_SceneBox::iterator it = scene->begin();
    VEE_SceneBox::iterator sentinel = scene->end();
    
    while(it != sentinel) {
      
      VEE_Object * obj = *it;

      int t = obj->objectType();

      if(t == VEE_OBJECT_PARTICLE) {
	VEE_Particle *particle = (VEE_Particle *) obj;
	if(particle->visible())
	  renderParticle(particle);
      }
      else if(t == VEE_OBJECT_TAIL_PARTICLE)
	renderTailParticle((VEE_TailParticle *) obj);
      else if(t == VEE_OBJECT_CURVE_PARTICLE)
	renderCurveParticle((VEE_CurveParticle *) obj);
      else if(t == VEE_OBJECT_RIGID_PARTICLE)
	renderRigidParticle((VEE_RigidParticle *) obj);
      else
	obj->render(this);

      it++;
    }
  }

  if(scene->isSplit()) {

    uint i = 0;

    for(i = 0; i < 8; i++) 
      drawRecursed(scene->child(i));
  }
}

void VEE_GlRenderer::drawBackToFront(VEE_SceneBox *scene)
{
  VEE_Orderer collector;

  collector.setCameraPosition(m_camPos);
  collector.addScene(scene);

  VEE_Orderer::iterator it = collector.end();

  if(it == collector.begin())
    return;

  /* VEE_trace("VEE_GlRenderer::drawBackToFront # Rendering %u particles",
     (uint) collector.size()); */

  do {
    it--;
    VEE_Object * obj = VEE_Orderer::getObject(it);

    int t = obj->objectType();

    if(t == VEE_OBJECT_PARTICLE) {
      VEE_Particle * p = (VEE_Particle *) obj;
      if(p->visible())
	renderParticle(p);
    }
    else if(t == VEE_OBJECT_TAIL_PARTICLE)
      renderTailParticle((VEE_TailParticle *) obj);
    else if(t == VEE_OBJECT_CURVE_PARTICLE)
      renderCurveParticle((VEE_CurveParticle *) obj);
    else if(t == VEE_OBJECT_RIGID_PARTICLE)
      renderRigidParticle((VEE_RigidParticle *) obj);
    else
      obj->render(this);

  } while(it != collector.begin());
}

void VEE_GlRenderer::drawBackToFront2(VEE_OrdererCollector *collector)
{
  VEE_OrdererCollector::iterator it = collector->end();

  if(it == collector->begin())
    return;

  for(VEE_OrdererCollector::forceIterator fit = collector->forceBegin();
      fit != collector->forceEnd(); fit++) {
    renderForceObject(VEE_OrdererCollector::getForce(fit));
  }  

  /* VEE_trace("VEE_GlRenderer::drawBackToFront2 # Rendering %u particles",
     (uint) collector->size()); */

  do {
    it--;

    VEE_ParticleVisualInfo & p = VEE_OrdererCollector::getObject(it);

    renderParticleVisual(&p);

  } while(it != collector->begin());

}

void VEE_GlRenderer::drawRandom(VEE_Collector *collector)
{
  VEE_Collector::iterator it = collector->begin();

  uint i, n = collector->collectedObjects();

  // VEE_trace("VEE_GlRenderer::drawRandom # %u", n);

  for(i=0; i < n; i++) {
    renderParticleVisual(& VEE_Collector::getObject(it));
    it++;
  }

  for(VEE_Collector::forceIterator fit = collector->forceBegin();
      fit != collector->forceEnd(); fit++) {
    renderForceObject(VEE_OrdererCollector::getForce(fit));
  }  

}

/** Clears the OpenGL screen. */

void VEE_GlRenderer::clearGlScreen()
{
  glClearColor(m_bgColor.red(), m_bgColor.green(), m_bgColor.blue(), 
	       m_bgColor.alpha());
  glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}

/** Moves camera to accordingto camera parameters. */
void VEE_GlRenderer::moveGlCamera()
{
  glLoadIdentity ();

  VEE_Vector3 tmp(0.0f, 0.0f, 1.0f);
  VEE_Matrix3 r1, r2;
  
  r1.rotateX(-m_ele);
  r2.rotateY(m_azi);
  // rotation.rotateXYZ(m_ele, m_azi, 0.0f);
  
  m_direction = r2 * (r1 * tmp);
  tmp = m_camPos + m_direction;
  
  gluLookAt(m_camPos.x, m_camPos.y, m_camPos.z,
	    tmp.x, tmp.y, tmp.z,
	    m_camUp.x, m_camUp.y, m_camUp.z);
}

void VEE_GlRenderer::moveCamera(float dt)
{
  // Move foward:
  m_camPos += m_direction * (m_velocity[2] * dt);
  // Move sideways:
  VEE_Vector3 lateral = cross(m_direction, m_camUp);
  lateral.normalize();
  m_camPos += lateral * (m_velocity[0] * dt);

  /* VEE_trace("VEE_GlRenderer::moveCamera # %f %f", 
     m_velocity[2], m_velocity[0]); */
}

void VEE_GlRenderer::setRange(float minDist, float fadeDist, float maxDist)
{ 
  m_nearDistance = minDist; 
  m_fadeDistance = fadeDist; 
  m_fadeRange = m_fadeDistance - m_nearDistance; 
  m_farDistance = maxDist; 
}

/** Creates the blurry textures that are to be used in rendering. A
    great number of small textures are merged into a single larger
    texture that contains small "tiles". 

    This method also loads the shader names.
*/
void VEE_GlRenderer::buildTextures(VEE_GlTexture2D *tex, 
				   float radius,
				   const char * configFile)
{
  const char * fname = "VEE_GlRenderer::buildTextures";

  VEE_Image bitmap, mastertexture;

  ConfigReader::Config conf;

  if(configFile)
    ConfigReader::readConfig(&conf, configFile);

  // conf.dump(std::cout);

  ConfigReader::Chunk chunk = conf.get("TextureParams");

  uint bitmapSize = chunk.get("texturesize").getInt(32);
  uint rows = chunk.get("division").getInt(4);
  // uint tiles = rows * rows;
  uint masterMapSize = bitmapSize * rows;
  
  bitmap.resize(bitmapSize, bitmapSize, 4);
  bitmap.evenGray(0xff);

  mastertexture.resize(masterMapSize, masterMapSize, 4);

  uint index = 0;

  float optimal = 100.0;
  // float radius = 0.45;

  uint row;

  // A Series of cosine textures:
  for(row = 0; row < 2; row++) {
    for(uint col = 0; col < rows; col++) {

      float power = 3.0f * (float) index / (float) (rows * 2);
      index++;
  
      float tmp = VEE_abs(power - 1.0);
      
      if(optimal > tmp) {
	optimal = tmp;
	m_textureRow = row;
	m_textureCol = col;
      }
      
      /* bitmap.cosinePeak(VEE_Image::AlphaChannel, 1.0, 0.25);
	 VEE_triTexture.genTexture(&bitmap, false); */
  
      bitmap.cosinePeak(VEE_Image::AlphaChannel, power, radius);
      
      mastertexture.copyPart(bitmap, 0, bitmapSize - 1, 0, bitmapSize - 1,
			     col * bitmapSize, row * bitmapSize);
    }
  }

  // A series of power-based textures:
  index = 0;

  for(row = 2; row < 4; row++) {
    for(uint col = 0; col < rows; col++) {
      
      float power = 6.0f * (float) index / (float) (rows * 2);
      index++;
        
      bitmap.powerPeak(VEE_Image::AlphaChannel, power, radius);
      
      mastertexture.copyPart(bitmap, 0, bitmapSize - 1, 0, bitmapSize - 1,
			     col * bitmapSize, row * bitmapSize);
    }
  }

  mastertexture.evenGray(VEE_Image::RedChannel, 0xff);
  mastertexture.evenGray(VEE_Image::GreenChannel, 0xff);
  mastertexture.evenGray(VEE_Image::BlueChannel, 0xff);

  char buf[64];

  for(row = 0; row < rows; row++) {
    for(uint col=0; col < rows; col++) {

      uint index = row * rows + col;

      sprintf(buf, "-%u", index);

      std::string name("Texture");
      name += buf;

      chunk = conf.get(name);

      if(chunk.size()) {

	std::string name = chunk.get("name");
	bitmap.evenGray(0xff);

	std::string texName = chunk.get("bitmap");

	VEE_trace("%s # Loading texture %s (%u %u)", 
		  fname, texName.c_str(), row, col);

	if(texName.size()) {
	  if(!bitmap.load(texName.c_str(), bitmapSize, bitmapSize, 4)) {
	    VEE_error("%s # Failed to load \"%s\"", fname, texName.c_str());
	    continue;
	  }
	}

	if(name.size())
	  m_textureNames[name] = index;


	std::string noiseDef = chunk.get("noise");

	if(noiseDef.size()) {
	  float att = 0.5;
	  float scale = 1.0;
	  float z = 0.0;
	  float amplitude = 1.0;
	  // int seed = -1;
	  uint octaves = 4;
	  sscanf(noiseDef.c_str(), "%f %u %f %f %f", 
		 &att, &octaves, &scale, &z, &amplitude);
	  bitmap.evenGray(0xff);
	  bitmap.perlinNoise(VEE_Image::AlphaChannel, 
			     att, octaves, scale, z * scale, amplitude);
	}

	ConfigReader::Variant v = chunk.get("boost");

	if(!v.isEmpty()) {
	  int boost = (int) (v.getDouble(0.2) * 255.0);
	  bitmap.boostRgb(boost);
	}

	v = chunk.get("decaypower");

	if(chunk.get("luminance_alpha").getString("0") == "1") {
	  bitmap.alphaFromImageLuminance(bitmap);
	  VEE_trace("%s # alpha from luminance", fname);
	}
	else if(!v.isEmpty()) {
	  if(v.getString() == "null")
	    ;
	  else if(v.getString() == "flat")
	    bitmap.evenGray(VEE_Image::AlphaChannel, 0xff);
	  else {
	    float power = v.getDouble(1.0);	
	    bitmap.powerPeak(VEE_Image::AlphaChannel, power, radius);
	  }
	}
	
	v = chunk.get("brigthness");
	if(!v.isEmpty()) {
	  float brigthness = v.getDouble(1.0);
	  bitmap.multiply(brigthness);
	}
	
	if(m_blendType == VEE_BLEND_NEW) {

	  if(chunk.get("maintain_colors").getString("0") == "0") {
	    bitmap.channelMultiply
	      (VEE_Image::AlphaChannel, VEE_Image::AlphaChannel, 
	       VEE_Image::RedChannel);
	    
	    bitmap.channelCopy(VEE_Image::RedChannel, 
			       VEE_Image::AlphaChannel);
	    bitmap.channelCopy(VEE_Image::GreenChannel, 
			       VEE_Image::AlphaChannel);
	    bitmap.channelCopy(VEE_Image::BlueChannel, 
			       VEE_Image::AlphaChannel);
	  }
	  else {
	    bitmap.channelMultiply
	      (VEE_Image::RedChannel, VEE_Image::RedChannel,
	       VEE_Image::AlphaChannel);
	    bitmap.channelMultiply
	      (VEE_Image::GreenChannel, VEE_Image::GreenChannel,
	       VEE_Image::AlphaChannel);
	    bitmap.channelMultiply
	      (VEE_Image::BlueChannel, VEE_Image::BlueChannel,
	       VEE_Image::AlphaChannel);
	  }
	  
	}
	mastertexture.copyPart(bitmap, 0, bitmapSize - 1, 0, bitmapSize - 1,
			       col * bitmapSize, row * bitmapSize);
      }
      else
	VEE_error("%s # No bitmap defined for %s", fname, name.c_str());

    }
  }

  

  // Finalize the texture:
  m_textureDivision = rows;
  m_textureStep = 1.0f / (float) m_textureDivision;
  
  VEE_trace("VEE_GlRenderer::buildTextures # %d rows", m_textureDivision);

  tex->genTexture(&mastertexture, true);
  
  VEE_trace("VEE_GlRenderer::buildTextures # Created a texture (id = %d)", 
	    (int) tex->id());
  
  VEE_GlMaterial tmp(tex->id());
  
  VEE_GlMaterialBank::setDefaultMaterial(tmp);

  // Optionally store the texture to a file:
  std::string storeFile = conf.get("Store").get("texturefile").getString();
  if(storeFile.size()) {
    mastertexture.save(storeFile.c_str());
  }

  // Now load the vortex material:

  chunk = conf.get("Forces");
  
  static const char * chunknames[] = {
    "vortextex",
    "blowtex",
    0
  };

  static const char * defaultFiles[] = {
    "images/vortex.png",
    "images/blow.png",
    0
  };
   
  static Material materials[] = {
    SWIRL,
    BLOW
  };
 
  int i;

  for(i = 0; chunknames[i]; i++) {
    
    ConfigReader::Variant v = chunk.get(chunknames[i]);
    
    std::string fileName = v.getString(defaultFiles[i]);
    
    bool ok = mastertexture.load(fileName.c_str());
    
    if(ok) {
      VEE_GlTexture2D tex;
      tex.genTexture(&mastertexture, true);
      m_materials.addMaterial(materials[i], VEE_GlMaterial(tex.id()));
    }
  }
  
  // Load shaders:
  chunk = conf.get("Shaders");
  
  for(i = 0; i < 100; i++) {
    sprintf(buf, "shader-%d", i);
    std::string shader = chunk.get(buf);
    if(shader.size()) {
      int index = m_shaderNames.size() + 1;
      m_shaderNames[shader] = index;
    }
  }
}

void VEE_GlRenderer::buildPolygonTextures(VEE_GlTexture2D *tex)
{
  VEE_Image bitmap;
  float radius = 0.5;
  float power = 0.5;
  uint bitmapSize = 16;

  bitmap.resize(bitmapSize, bitmapSize, 4);
  bitmap.evenGray(0xff);
  
  bitmap.cosinePeak(VEE_Image::AlphaChannel, power, radius);

  bitmap.evenGray(VEE_Image::RedChannel, 200);
  bitmap.addRandom(VEE_Image::RedChannel, 60);
  bitmap.channelCopy(VEE_Image::GreenChannel, VEE_Image::RedChannel);
  bitmap.channelCopy(VEE_Image::BlueChannel, VEE_Image::RedChannel);
  
  tex->genTexture(&bitmap, true);
  VEE_GlMaterial tmp(tex->id());
  
  m_materials.addMaterial(POLYGON, tmp);
}

void VEE_GlRenderer::setTexId()
{
  if(m_blobTexture)
    m_texId = m_blobTexture->id();
}

void VEE_GlRenderer::drawBox(VEE_Vector3 low, VEE_Vector3 high)
{
  // The face with low y
  glVertex3f(low.x, low.y, low.z);
  glVertex3f(high.x, low.y, low.z);

  glVertex3f(low.x, low.y, low.z);
  glVertex3f(low.x, low.y, high.z);

  glVertex3f(high.x, low.y, high.z);
  glVertex3f(low.x, low.y, high.z);

  glVertex3f(high.x, low.y, high.z);
  glVertex3f(high.x, low.y, low.z);

  // The face with high y

  glVertex3f(low.x, high.y, low.z);
  glVertex3f(high.x, high.y, low.z);

  glVertex3f(low.x, high.y, low.z);
  glVertex3f(low.x, high.y, high.z);

  glVertex3f(high.x, high.y, high.z);
  glVertex3f(low.x, high.y, high.z);

  glVertex3f(high.x, high.y, high.z);
  glVertex3f(high.x, high.y, low.z);

  // The connecting lines:

  glVertex3f(low.x, low.y, low.z);
  glVertex3f(low.x, high.y, low.z);

  glVertex3f(high.x, low.y, low.z);
  glVertex3f(high.x, high.y, low.z);

  glVertex3f(low.x, low.y, high.z);
  glVertex3f(low.x, high.y, high.z);

  glVertex3f(high.x, low.y, high.z);
  glVertex3f(high.x, high.y, high.z);
}

int VEE_GlRenderer::textureNameToId(const char *name)
{
  std::map<std::string, int>::iterator it = m_textureNames.find(name);

  if(it != m_textureNames.end())
    return (*it).second;

  return -1;
}

int VEE_GlRenderer::shaderNameToId(const char *name)
{
  std::map<std::string, int>::iterator it = m_shaderNames.find(name);

  if(it != m_shaderNames.end())
    return (*it).second;

  return -1;
}


void VEE_GlRenderer::drawVolumeGrid(const VEE_Vector3 &loc, VEE_VolumeGrid *g)
{
  newMode(RenderNothing);

  const VEE_MarchingCubes * mc = & g->marchingCubes();

  uint n = mc->triangleCount();

  const VEE_GlMaterial & mat = m_materials.getMaterial(CHECK);
  mat.useTexture();

  glPushAttrib(~0);

  glEnable(GL_BLEND);
  glBlendEquation(GL_FUNC_ADD);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE);
  
  glEnable(GL_DEPTH_TEST);
  glDepthMask(GL_FALSE);

  VEE_Rgba frontColor(1.0, 1.0, 0.0, 1.0);
  VEE_Rgba backColor(0.2, 0.2, 0.0, 1.0);

  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, 1);

  glDisable(GL_COLOR_MATERIAL);
  glEnable(GL_LIGHTING);
  glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);

  glMaterialfv(GL_FRONT, GL_EMISSION, frontColor.data());
  glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, frontColor.data());

  glMaterialfv(GL_BACK,  GL_EMISSION, backColor.data());
  glMaterialfv(GL_BACK,  GL_AMBIENT_AND_DIFFUSE, backColor.data());

  pushTransform(g->rotation(), loc);

  const VEE_Vector3 & scale = g->scale();
  
  float maxScale = VEE_max(scale.x, scale.y, scale.z) * 6.0;

  glScalef(scale.x, scale.y, scale.z);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE);

  glBegin(GL_TRIANGLES);

  for(uint i=0; i < n; i++) {

    const VEE_MarchingCubes::Triangle & tri = mc->triangle(i);
    
    for(uint j=0; j < 3; j++) {

      VEE_Vector3 v = tri.m_vertices[j];

      glTexCoord2f(v.x * maxScale, v.z * maxScale);
      glVertex3fv(v.data());
    }
  }

  glEnd();

  glPopMatrix();

  glPopAttrib();

  // Just to be safe:
  glBindTexture(GL_TEXTURE_2D, m_texId);
}

void VEE_GlRenderer::drawVolumeGrid2(const VEE_Vector3 &loc, 
				     VEE_VolumeGrid2 *g,
				     bool blend)
{
  // VEE_trace("VEE_GlRenderer::drawVolumeGrid2");

  newMode(RenderNothing);

  VEE_VolumeGrid2::GridGrid ** grids = g->grids();
  uint count = g->gridCount();

  const VEE_GlMaterial & mat = m_materials.getMaterial(CHECK);
  mat.useTexture();

  glPushAttrib(~0);

  // glDisable(GL_TEXTURE_2D);

  if(blend) {
    glEnable(GL_BLEND);
    glDepthMask(GL_FALSE);
  }
  else {
    glDisable(GL_BLEND);
    glDepthMask(GL_TRUE);
  }

  glBlendEquation(GL_FUNC_ADD);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE);
  
  glEnable(GL_DEPTH_TEST);

  VEE_Rgba frontColor(1.0, 1.0, 0.0, 1.0);
  VEE_Rgba backColor(0.2, 0.2, 0.0, 1.0);

  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, 1);

  glDisable(GL_COLOR_MATERIAL);
  glEnable(GL_LIGHTING);
  glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);

  glMaterialfv(GL_FRONT, GL_EMISSION, frontColor.data());
  glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, frontColor.data());

  glMaterialfv(GL_BACK,  GL_EMISSION, backColor.data());
  glMaterialfv(GL_BACK,  GL_AMBIENT_AND_DIFFUSE, backColor.data());

  pushTransform(g->rotation(), loc);

  float scale = g->scale();
  
  glBlendFunc(GL_SRC_ALPHA, GL_ONE);

  glBegin(GL_TRIANGLES);

  for(uint grid = 0; grid < count; grid++) {
    
    const VEE_MarchingCubes * mc = & (*grids)->m_triangles;
    
    uint n = mc->triangleCount();
    
    for(uint i=0; i < n; i++) {
      
      const VEE_MarchingCubes::Triangle & tri = mc->triangle(i);
      
      for(uint j=0; j < 3; j++) {
	
	VEE_Vector3 v = tri.m_vertices[j];
	
	glTexCoord2f(v.x * scale, v.z * scale);
	glVertex3fv(v.data());
      }
    }
    grids++;
  }

  glEnd();

  glPopMatrix();

  glPopAttrib();

  // Just to be safe:
  glBindTexture(GL_TEXTURE_2D, m_texId);
}

void VEE_GlRenderer::setRenderOnlyOutline(bool outlines) 
{ 
  m_renderOnlyOutline = outlines; 

  if(outlines)
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
  else
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
}


/** Creates the depth map texture into the GL_TEXTURE1_ARB. */
void VEE_GlRenderer::createDepthMap()
{
  if(!m_useDepthMap)
    return;

  // VEE_trace("VEE_GlRenderer::createDepthMap 0");

  glGetError();

#ifdef GL_TEXTURE1
  // #if 1
  // VEE_trace("VEE_GlRenderer::createDepthMap");

  glActiveTexture(GL_TEXTURE1); 

  m_depthMap.use();
  // glBindTexture(GL_TEXTURE_2D, m_texId);
  
  glCopyTexSubImage2D(GL_TEXTURE_2D,
		      0,
		      0, 0,
		      0, 0,
		      m_width, m_height);

  GLenum e = glGetError();
  
  if(e)
    VEE_trace("VEE_GlRenderer::createDepthMap # glCopyTexImage2D failed %s",
	      gluErrorString(e));

  glDisable(GL_TEXTURE_2D);

  glActiveTexture(GL_TEXTURE0); 

  glEnable(GL_TEXTURE_2D);
  
  /*
  m_depthMap.use();

  glDisable(GL_LIGHTING);
  glColor3f(0.7, 0.7, 0.7);

  glBegin(GL_QUADS);

  glNormal3f(0, 0, 1);

  glTexCoord2f(0, 0);
  glVertex2f(0, 0);

  glTexCoord2f(1, 0);
  glVertex2f(3, 0);

  glTexCoord2f(1, 1);
  glVertex2f(3, 3);

  glTexCoord2f(0, 1);
  glVertex2f(0, 3);
  
  glEnd();

  glDisable(GL_TEXTURE_2D);
  */
#endif 
}


void VEE_GlRenderer::pushTransform(const VEE_Matrix3 &rot,
				   const VEE_Vector3 &loc)
{  
  glPushMatrix();

  // OpenGL transformation matrix
  float oglMatrix[16];

  oglMatrix[0] = rot[0][0];
  oglMatrix[1] = rot[1][0];
  oglMatrix[2] = rot[2][0];
  oglMatrix[3] = 0.0f;

  oglMatrix[4] = rot[0][1];
  oglMatrix[5] = rot[1][1];
  oglMatrix[6] = rot[2][1];
  oglMatrix[7] = 0.0f;

  oglMatrix[8]  = rot[0][2];
  oglMatrix[9]  = rot[1][2];
  oglMatrix[10] = rot[2][2];
  oglMatrix[11] = 0.0f;

  oglMatrix[12] = loc[0];
  oglMatrix[13] = loc[1];
  oglMatrix[14] = loc[2];
  oglMatrix[15] = 1.0f;

  glMultMatrixf(oglMatrix);
}
