/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <vee_image.h>

#include <vee_error.h>
#include <vee_math.h>
#include <vee_perlin_noise.h>

#include <iostream>

#include <math.h>
#include <sched.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <di_mutex.h>
#include <di_types.h>

#include <vector>

// #include <Magick++/Image.h>
// #include <magick/api.h>

extern "C" {
#include <jpeglib.h>
}

static std::vector<VEE_ImageIO *> __io_plugins;
static DI_MutexAuto __io_mutex;

VEE_Image::VEE_Image()
  : m_image(0),
    m_rowSize(0),
    m_bytesPerPixel(0),
    m_width(0),
    m_height(0)
{}

VEE_Image::~VEE_Image()
{
  delete [] m_image;
}

bool VEE_Image::loadJpeg(const char *fileName)
{
  const char *fname = "VEE_Image::loadJpeg";

  // cout << fname << std::endl;

  FILE * infile;
  jpeg_decompress_struct cinfo;

  struct jpeg_error_mgr jerr;
  cinfo.err = jpeg_std_error(&jerr);
  jpeg_create_decompress(&cinfo);  

  if (! (infile = fopen(fileName, "rb"))) {
    VEE_error("%s # can't open %s", fname, fileName);
    return false;
  }

  jpeg_stdio_src(&cinfo, infile);  
  jpeg_read_header(&cinfo, TRUE);

  resize(cinfo.image_width, cinfo.image_height, cinfo.num_components);

  /* cout << "Image size: " << m_width << " x " << m_height 
     << " bytes: " << m_bytesPerPixel << std::endl; */

  jpeg_start_decompress(&cinfo);
  // cout << fname << " 2 " << std::endl;
  uint r = 0;
  {
    JSAMPROW row_pointer[1];
    uint i, rowBytes = m_width * m_bytesPerPixel;
    
    for(i = 0; i < m_height; i++) {
      row_pointer[0] = & m_image[rowBytes * i];
      r += jpeg_read_scanlines(&cinfo, row_pointer, 1);
    }
  }
  // cout << fname << " 2.5 read scanlines: " << r << std::endl;

  jpeg_finish_decompress(&cinfo);  

  // cout << fname << " 3 " << std::endl;

  jpeg_destroy_decompress(&cinfo);

  fclose(infile);

  return true;
}

/** Loads an image using some plugin */

bool VEE_Image::load(const char *filename, int w, int h, int channels)
{
  // const char * fname = "VEE_Image::load";

  bool ok = false;

  __io_mutex.lock();

  for(uint i=0; i < __io_plugins.size() && !ok; i++) {
    ok = __io_plugins[i]->read(this, filename, w, h, channels);
  }

  /* if(!ok && w < 0 && h < 0)
     ok = loadJpeg(filename); */

  __io_mutex.unlock();

  return ok;
}

bool VEE_Image::save(const char *filename)
{
  // const char * fname = "VEE_Image::save";
  
  bool ok = false;

  __io_mutex.lock();

  for(uint i=0; i < __io_plugins.size() && !ok; i++) {
    ok = __io_plugins[i]->write(this, filename);
  }
  __io_mutex.unlock();

  return ok;
}

void VEE_Image::addAlpha()
{
  uint rowSize = m_width * 4;
  unsigned char *image = new unsigned char[m_height * rowSize];
  
  uint bytesPerPixel = m_bytesPerPixel;

  for(uint i=0; i < m_height; i++) {
    for(uint j=0; j < m_width; j++) {
      unsigned char * target = image + rowSize * i + j * 4;
      const unsigned char * source = pixelConst(j, i);

      for(uint k=0; k < bytesPerPixel; k++)
	target[k] = source[k];
    }
  }

  delete m_image;

  m_image = image;
  m_rowSize = rowSize;
  m_bytesPerPixel = 4;
}

void VEE_Image::alphaFromImageLuminance(const VEE_Image &img)
{
  
  for(uint i=0; i < m_height; i++) {
    for(uint j=0; j < m_width; j++) {
      pixel(j, i)[3] = img.pixelLuminance(j, i);
    }
  }
}

double VEE_Image::channelAverage(uint chan)
{
  double sum = 0;
  
  for(uint i=0; i < m_height; i++) {
    double tmpSum = 0.0;
    for(uint j=0; j < m_width; j++) {
      tmpSum += pixelConst(j, i)[chan];
    }
    sum += tmpSum;
  }
  return sum / (m_height * m_width);
}

void VEE_Image::resize(uint w, uint h, uint bytesPerPixel)
{
  uint nowBytes = m_rowSize * m_height;
  uint needBytes = w * h * bytesPerPixel;

  if(nowBytes == needBytes) {
    m_bytesPerPixel = bytesPerPixel;
    m_rowSize = bytesPerPixel * w;
    m_width = w;
    m_height = h;
    return;
  }

  delete [] m_image;
  
  m_image = 0;

  if(bytesPerPixel > 4)
    VEE_error("VEE_Image::resize # %u bytes per pixel", bytesPerPixel);

  /* std::cout << "VEE_Image::resize " << w << " x " 
     << h << " bytes: " << bytesPerPixel << std::endl; */

  uint s = w * h * bytesPerPixel;
  m_bytesPerPixel = bytesPerPixel;
  m_rowSize = bytesPerPixel * w;

  if(s) {
    m_image = new unsigned char[s];

    if(!m_image) {
      VEE_fatal("FATAL VEE_Image::resize");
      exit(0);
    }
    m_width = w;
    m_height = h;
  }
  else {
    m_image = 0;
    m_width = 0;
    m_height = 0;
  }
}

void VEE_Image::resizeImage(uint w, uint h)
{
  if(w > m_width || h > m_height) {
    VEE_error("VEE_Image::resizeImage BAD ARGS");
    return;
  }

  if(w == m_width && h == m_height)
    return;

  uint rowSize = w * m_bytesPerPixel;
  unsigned char *image = new unsigned char[h * rowSize];

  for(uint i=0; i < h; i++) {
    memcpy(image + i * rowSize, m_image + i * m_rowSize, rowSize);
  }

  delete m_image;

  m_image = image;
  m_rowSize = rowSize;
  m_width = w;
  m_height = h;
}
  
void VEE_Image::resizeImgToPowOfTwo()
{
  uint s = 0;
  for(uint i=1; i < 32; i++) {
    uint tmp = 1 << i;
    if(tmp > m_width || tmp > m_height) {
      s = 1 << (i-1);
      break;
    }
  }
  resizeImage(s, s);
}

void VEE_Image::mirrorXY()
{
  uint bpp = m_bytesPerPixel;

  for(uint i=0; i < m_height / 2; i++) {
    for(uint j=0; j < m_width; j++) {
      unsigned char *pixel1 = pixel(j, i);
      unsigned char *pixel2 = pixel(m_width - j - 1, m_height - i - 1);
      
      for(uint k = 0; k < bpp; k++) {
	unsigned char tmp = *pixel1;
	*pixel1++ = *pixel2;
	*pixel2++ = tmp;
      }
    }
  }
}

void VEE_Image::randomValues()
{
  uint bytes = m_width * m_height * m_bytesPerPixel;
  for(uint i=0; i < bytes; i++) 
    m_image[i] = (unsigned char) random();
}

void VEE_Image::perlinNoise(uint channel,
			    double att, int octaves,  double scale, 
			    double z, double amplitude)
{
  double dh = h() - 1;
  double dw = w() - 1;

  for(uint i=0; i < h(); i++) {

    double y = scale * ((double) i / dh - 0.5);

    for(uint j=0; j < w(); j++) {

      double x = scale * ((double) j / dw - 0.5);
      
      double r = VEE_PerlinNoise::noise3d(x, y, z, att, octaves) * amplitude;
      
      unsigned char *px = pixel(j, i);

      int v = (int) ((r + 0.5) * 256.0);

      v = VEE_clamp(v, 0, 255);
      px[channel] = v;
    }
  }
}

/** Generates a check-board texture. Enjoy. */

void VEE_Image::checkTexture(uint horizontalInterval,
			     uint verticalInterval,
			     const unsigned char * color1,
			     const unsigned char * color2)
{  
  unsigned char white[4] = {0xFF, 0xFF, 0xFF, 0xFF };
  unsigned char black[4] = {0, 0, 0, 0xFF };

  if(!color1)
    color1 = white;
  if(!color2)
    color2 = black;

  bool b1 = true, b2;

  uint bpp = m_bytesPerPixel;

  for(uint i=0; i < h(); i++) {

    if(i % verticalInterval == 0)
      b1 = !b1;

    //if(!(i & 0x1))
    b2 = true; // !(i & 0x1);

    for(uint j=0; j < w(); j++) {

      if(j % horizontalInterval == 0)
	b2 = !b2;

      const unsigned char * source;

      if((b1 && b2) || (!b1 && !b2))
	source = color1;
      else 
	source = color2;

      unsigned char *px = pixel(j, i);
      
      for(uint k = 0; k < bpp; k++)
	*px++ = *source++;

    }
  }
}

/** Generates a skotch texture. Enjoy. */

void VEE_Image::skotchTexture(uint stripeWidth,
			      uint stripeInterval,
			      const unsigned char * color1,
			      const unsigned char * color2)
{  
  unsigned char white[4] = {0xFF, 0xFF, 0xFF, 0xFF };
  unsigned char black[4] = {0, 0, 0, 0xFF };

  if(!color1)
    color1 = white;
  if(!color2)
    color2 = black;

  uint bpp = m_bytesPerPixel;

  for(uint i=0; i < h(); i++) {

    if(i % stripeInterval < stripeWidth) {

      for(uint j=0; j < w(); j++) {
	unsigned char *px = pixel(j, i);

	for(uint k = 0; k < bpp; k++)
	  *px++ = color1[k];
      }
    }
    else {

      for(uint j=0; j < w(); j++) {
	
	unsigned char *px = pixel(j, i);

	const unsigned char * source = (j % stripeInterval < stripeWidth) ?
	  color1 : color2;

	for(uint k = 0; k < bpp; k++)
	  *px++ = *source++;
      }
    }

  }
}

void VEE_Image::stripeTexture(uint stripeWidth,
			      uint stripeInterval,
			      const unsigned char * color1,
			      const unsigned char * color2)
{  
  unsigned char white[4] = {0xFF, 0xFF, 0xFF, 0xFF };
  unsigned char black[4] = {0, 0, 0, 0xFF };

  if(!color1)
    color1 = white;
  if(!color2)
    color2 = black;

  uint bpp = m_bytesPerPixel;

  for(uint i=0; i < h(); i++) {
    
    const unsigned char * source = i % stripeInterval < stripeWidth ?
      color1 : color2;
    
    for(uint j=0; j < w(); j++) {
      unsigned char *px = pixel(j, i);
      
      for(uint k = 0; k < bpp; k++)
	*px++ = source[k];
    }
  }
}

unsigned char *VEE_Image::pixel(uint x, uint y)
{ 
  if(x >= m_width || y >= m_height)
    VEE_fatal("VEE_Image::pixel");

  return & m_image[m_rowSize * y + x * m_bytesPerPixel]; 
}

const unsigned char *VEE_Image::pixelConst(uint x, uint y) const
{ 
  if(x >= m_width || y >= m_height)
    VEE_fatal("VEE_Image::pixelConst");

  return & m_image[m_rowSize * y + x * m_bytesPerPixel]; 
}


unsigned char VEE_Image::pixelLuminance(uint x, uint y) const
{
  float sum = 0.0;

  const unsigned char *p = pixelConst(x, y);

  uint bpp = m_bytesPerPixel;
  if(bpp > 3) bpp = 3;
  
  for(uint i=0; i < m_bytesPerPixel; i++) {
    sum += (float) p[i];
  }
  sum /= bpp;

  int v = (int) sum;
  if(v > 255) v = 255;
  return (unsigned char) v;
}

void VEE_Image::copyPart(const VEE_Image &img, 
			 uint startRow, uint endRow,
			 uint startCol, uint endCol,
			 uint offsetX,  uint offsetY)
{
  if(startRow >= img.h() || endRow >= img.h() || 
     startCol >= img.w() || endCol >= img.w()) {
    VEE_fatal("VEE_Image::copyPart # %u %u %u %u", 
	      startRow, endRow, startCol, endCol);
  }

  uint w = endCol - startCol + 1;
  uint h = endRow - startRow + 1;
  uint bytes = img.m_bytesPerPixel;

  resize(VEE_max(w, m_width), VEE_max(h, m_height), bytes);
  
  for(uint i = startRow; i <= endRow; i++) {
    for(uint j = startCol; j <= endCol; j++) {
      const unsigned char *fromPixel = img.pixelConst(j, i);
      unsigned char *toPixel = 
	pixel(j - startCol + offsetX, i - startRow + offsetY);
      for(uint k = 0; k < bytes; k++)
	toPixel[k] = fromPixel[k];
    }
  }
}

void VEE_Image::copyPartEnlarge(const VEE_Image &img, 
			       uint ratio,
			       uint startRow, uint endRow,
			       uint startCol, uint endCol)
{
  const char *fname = "VEE_Image::copyPartEnlarge";
  if(startRow >= img.h() || endRow >= img.h() || 
     startCol >= img.w() || endCol >= img.w()) {
    VEE_fatal(fname);
  }

  const uint w = (endCol - startCol + 1) * ratio;
  const uint h = (endRow - startRow + 1) * ratio;
  const uint bytes = img.m_bytesPerPixel;
  
  resize(w, h, bytes);
  
  double dratio = ratio;

  for(uint i = 0; i < h; i++) {
    double y = (double) i / dratio + (double) startRow;
    for(uint j = 0; j < w; j++) {
      unsigned char *toPixel = pixel(j, i);
      
      double x = (double) j / dratio + (double) startCol;
      img.getSubpixel(toPixel, x, y);
    }
  }
}

void VEE_Image::copyPartSmaller(const VEE_Image &img, 
			       uint ratio,
			       uint startRow, uint endRow,
			       uint startCol, uint endCol)
{
  const char *fname = "VEE_Image::copyPartSmaller";

  if(startRow >= img.h() || endRow >= img.h() || 
     startCol >= img.w() || endCol >= img.w()) {
    VEE_fatal(fname);
  }
  
  const uint w = (endCol - startCol + 1) / ratio;
  const uint h = (endRow - startRow + 1) / ratio;
  const uint bytes = img.m_bytesPerPixel;
  
  if((img.w() % ratio) != 0 || (img.h() % ratio) != 0) {
    printf("w = %d h = %d ratio = %d\n", w, h, ratio);
    VEE_fatal(fname, "2");
  }

  resize(w, h, bytes);

  for(uint i = 0; i < h; i++) {
    for(uint j = 0; j < w; j++) {
      unsigned char *toPixel = pixel(j, i);
      img.getMultipixel(toPixel, 
			j * ratio + startCol, 
			i * ratio + startRow,
			ratio);
    }
  }
}

void VEE_Image::evenGray(unsigned char v)
{
  uint bytes = m_width * m_height * m_bytesPerPixel;
  for(uint i=0; i < bytes; i++) 
    m_image[i] = v;
}

void VEE_Image::evenGray(uint channel, unsigned char v)
{
  unsigned char * target = m_image + channel;
  
  uint pixels = m_width * m_height;

  for(uint i=0; i < pixels; i++) {
    *target = v;
    target += m_bytesPerPixel;
  }
}

void VEE_Image::addRandom(uint channel, unsigned char amplitude)
{
  unsigned char * target = m_image + channel;
  
  uint pixels = m_width * m_height;

  for(uint i=0; i < pixels; i++) {
    int val = *target;
    char r = (char) (random() % (long) amplitude);
    val += r;

    if(val < 0) val = 0;
    else if(val > 0xff) val = 0xff;
    
    *target = val;

    target += m_bytesPerPixel;
  }
}

void VEE_Image::cosinePeak(uint channel, float power, float radius)
{
  float centerY = m_height * 0.5;
  float centerX = m_width  * 0.5;
  
  float xscale = VEE_PI / (radius * 2.0f * (float) m_width);
  float yscale = VEE_PI / (radius * 2.0f * (float) m_height);

  for(uint x=0; x < m_width; x++) {
    for(uint y=0; y < m_height; y++) {

      float xdiff = (centerX - (float) x) * xscale;
      float ydiff = (centerY - (float) y) * yscale;

      float distanceToCenter = sqrtf(xdiff * xdiff + ydiff * ydiff);
      
      float v;
      
      if(distanceToCenter >= VEE_HALF_PI) 
	v = 0.0;
      else {
	v = powf(cosf(distanceToCenter), power);
      }

      int iv = (int) (v * 256);

      if(iv < 0) iv = 0;
      else if(iv > 0xff) iv = 0xff;
      
      unsigned char *data = pixel(x, y);
      data[channel] = (unsigned char) iv;
    }
  }
}

void VEE_Image::channelCopy(uint targetChannel, uint sourceChannel)
{
  unsigned char * target = m_image + targetChannel;
  const unsigned char * source = m_image + sourceChannel;
  
  uint pixels = m_width * m_height;

  for(uint i=0; i < pixels; i++) {
    *target = *source;
    target += m_bytesPerPixel;
    source += m_bytesPerPixel;
  }
}

void VEE_Image::channelSwap(uint c1, uint c2)
{
  unsigned char * target = m_image + c1;
  unsigned char * source = m_image + c2;
  
  uint pixels = m_width * m_height;

  for(uint i=0; i < pixels; i++) {
    uchar tmp = *target;
    *target = *source;
    *source = tmp;

    target += m_bytesPerPixel;
    source += m_bytesPerPixel;
  }
}

void VEE_Image::channelMultiply(uint target, uint source1, uint source2)
{
  unsigned char * t  = m_image + target;
  const unsigned char * s1 = m_image + source1;
  const unsigned char * s2 = m_image + source2;
  
  uint pixels = m_width * m_height;

  for(uint i=0; i < pixels; i++) {
    
    uint tmp = ((uint) *s1 * (uint) *s2) / 255;
    
    *t = (uchar) tmp;

    t += m_bytesPerPixel;
    s1 += m_bytesPerPixel;
    s2 += m_bytesPerPixel;
  }
}

void VEE_Image::powerPeak(uint channel, float power, float radius)
{
  float centerY = m_height * 0.5;
  float centerX = m_width  * 0.5;
  
  for(uint x=0; x < m_width; x++) {
    for(uint y=0; y < m_height; y++) {

      float xdiff = (centerX - (float) x) / centerX;
      float ydiff = (centerY - (float) y) / centerY;

      float distanceToCenter = 
	0.5f * sqrtf(xdiff * xdiff + ydiff * ydiff) / radius;

      float v;

      if(distanceToCenter >= 1.0) 
	v = 0.0;
      else 
	v = powf(1.0f - distanceToCenter, power);

      unsigned char *data = pixel(x, y);
      float tmp = data[channel] * v;

      data[channel] = VEE_min((int) tmp, 255);
    }
  }  
}

void VEE_Image::boostRgb(int boost)
{
  uint cb = m_bytesPerPixel >= 3 ? 3 : 1;

  for(uint x=0; x < m_width; x++) {
    for(uint y=0; y < m_height; y++) {

      unsigned char *data = pixel(x, y);

      for(uint i=0; i < cb; i++) {
	int tmp = data[i];
	tmp += boost;
	data[i] = VEE_max(VEE_min(tmp, 255), 0);
      }

    }
  }  
}

void VEE_Image::multiply(float mul)
{
  unsigned char * ptr = m_image;
  unsigned char * sentinel = ptr + m_width * m_height * m_bytesPerPixel;
  
  for(; ptr < sentinel; ptr++) {
    *ptr = (unsigned char) VEE_min(*ptr * mul, 255.0f);
  }
}

bool VEE_Image::add(const VEE_Image &that)
{
  if(that.w() != w() || that.h() != h() || 
     that.m_bytesPerPixel != m_bytesPerPixel)
    return false;

  unsigned char * ptr = m_image;
  unsigned char * ptr2 = that.m_image;
  unsigned char * sentinel = ptr + m_width * m_height * m_bytesPerPixel;
  
  for(; ptr < sentinel; ptr++) {
    *ptr = (unsigned char) VEE_min((int) (*ptr + *ptr2++), 255);
  }

  return true;
}

void VEE_Image::test()
{
  VEE_Image i1;
  i1.loadJpeg("test.jpg");
}

void VEE_Image::getSubpixel(unsigned char *target, double x, double y) const
{
  if(x >= m_width || y >= m_height)
    VEE_fatal("VEE_Image::getSubpixel");

  uint xl = (uint) x;
  uint xh = xl + 1;

  uint yl = (uint) y;
  uint yh = yl + 1;

  // Upper left corner:
  const unsigned char *ul = pixelConst(xl, yl);
  // Upper right corner:
  const unsigned char *ur;
  if(xh < m_width) ur = & ul[m_bytesPerPixel];
  else ur = ul;

  // Lower left:
  const unsigned char *ll;
  if(yh < m_height) ll = pixelConst(xl, yl+1);
  else ll = ul;
  // Lower right:
  const unsigned char *lr;
  if(xh < m_width) lr = & ll[m_bytesPerPixel];
  else lr = ll;

  double dxl = x - xl;
  double dxln = 1.0 - dxl;
  double dyl = y - yl;
  double dyln = 1.0 - dyl;
  
  // Weights for all end points.
  double wul = dxln * dyln;
  double wur = dxl * dyln;
  double wll = dxln * dyl;
  double wlr = dxl * dyl;

  // double wsum = wul + wur + wll + wlr;

  for(uint i=0; i < m_bytesPerPixel; i++) {
    double s = wul * (double) ul[i];
    s += wur * (double) ur[i];
    s += wll * (double) ll[i];
    s += wlr * (double) lr[i];

    // s /= wsum;

    if(s <= 255.0) target[i] = (unsigned char) s;
    else target[i] = 255;
  }
}

void VEE_Image::getMultipixel
(unsigned char *target, uint x, uint y, uint size) const 
{

  if(x + size > m_width || y + size > m_height) {
    VEE_error(" x = %u, x + size = %u", x, x + size);
    VEE_fatal("VEE_Image::getMultipixel");
  }
  uint sums[4] = {0,0,0,0};
  uint bytesPerPixel = m_bytesPerPixel;

  for(uint i = 0; i < size; i++) {
    const unsigned char *rowPixels = pixelConst(x, y + i);
    for(uint j = 0; j < size; j++) {      
      for(uint k = 0; k < bytesPerPixel; k++) {
	sums[k] += (uint) *rowPixels++;
      }
    }
  }
  uint square = size * size;

  for(uint k = 0; k < bytesPerPixel; k++) {
    target[k] = (unsigned char) (sums[k] / square);
  }
}

VEE_Image & VEE_Image::operator = (const VEE_Image &i)
{
  resize(i.w(), i.h(), i.bytesPerPixel());  
  memcpy(m_image, i.m_image, m_rowSize * m_height);
  return *this;
}

void VEE_Image::addIO(VEE_ImageIO *io)
{
  __io_plugins.push_back(io);
}
