/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <vee_qt.h>

#include <gee_engine.h>
#include <vee_error.h>
#include <vee_gl_renderer.h>
#include <vee_gl_texture.h>
#include <vee_image_blender.h>
#include <vee_math.h>
#include <vee_scene.h>
#include <vee_value.h>
#include <vee_value_gui.h>

#ifdef HAVE_GLSL_ARB
#include <vee_gl_sl_renderer.h>
#endif

#ifdef HAVE_CGGL
#include <vee_cg_gl_renderer.h>
#endif

#include <gee_display_lists.h>
#include <gee_engine.h>
#include <gee_geometry.h>
#include <gee_gl_renderer.h>
#include <gee_gl_material.h>

#include <mj_str_remap.h>

#include <di_sleep.h>

#include <time.h>

#include <qcolordialog.h>
#include <qfiledialog.h> 
#include <qfileinfo.h>
#include <qimage.h>
#include <qkeycode.h>
#include <qlayout.h>
#include <qpopupmenu.h>
#include <qpushbutton.h>
#include <qtimer.h>


#include <stdio.h>
#include <strings.h>

class VEE_QtHidden::Internal
{
public:

  Internal() 
    : m_frameFileName("Vee-Frame-%.5d.png"), 
      m_fileFrameCount(0),
      m_toFileCount(0),
      m_toFileOverSample(6),
      m_simulationTime(0),
      m_updateType(0),
      m_displaySolids(true)
      // m_threadedCalculus(0)
  {}

  ~Internal() { /* delete m_threadedCalculus;*/ }

  QString frameFile()
  { 
    QString tmp;
    tmp.sprintf(m_frameFileName, (int) ++m_fileFrameCount);
    return tmp;
  }
  
  bool outputFrames() const 
  { return m_frameFileName.size() > 0 && m_toFileCount; }

  QImage * image() 
  { 
    return & m_image; 
  }
  
  VEE_GlTexture2DBank  m_textures;
  QCString             m_frameFileName;
  QImage               m_image;
  VEE_ImageBlender     m_accumulator;
  uint                 m_fileFrameCount;
  uint                 m_toFileCount;
  uint                 m_toFileOverSample;
  double               m_simulationTime;
  int                  m_updateType;

  bool                 m_displaySolids;
  VEE_GarbageBin       m_veeGarbage;
  QPoint               m_mousePos;
  
  // VEE_CalculusThreads *m_threadedCalculus;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


VEE_QtHidden::VEE_QtHidden(VEE_Qt * host, 
			   VEE_SceneBox *box, 
			   QWidget *parent, 
			   const char *name,
			   VEE_GlRendererType type)
  : QGLWidget(parent,name),
    m_lastFrameCheckFrames(0),
    m_fixedRate(0.0),
    m_playSpeed(1.0),
    m_frames(0),
    m_visualUpdate(true),
    m_calculusUpdate(true),
    m_showBoxes(false),
    m_pause(false),
    m_showNObj(false),
    m_splitAt(40),
    m_collapseAt(20),
    m_maximumRecursion(1),
    m_overSample(1),
    m_scene(box),
    m_host(host),
    m_internal(new Internal)
{
  m_veeGl = 0;

#ifdef HAVE_GLSL_ARB
  if(type == VEE_GL_SL)
    m_veeGl = new VEE_GlSlRenderer(box);
#endif

#ifdef HAVE_CGGL
  if(type == VEE_GL_CG)
    m_veeGl = new VEE_CgGlRenderer(box);
#endif

  if(!m_veeGl)
    m_veeGl = new VEE_GlRenderer(box);

  VEE_trace("VEE_QtHidden::VEE_QtHidden # VEE renderer = %s (%d)",
	    typeid(*m_veeGl).name(), (int) type);

  m_geeMaterials = new GEE_GlMaterialBank();
  m_geeGl        = new GEE_GlRenderer();
  m_geeLists     = new GEE_GlDisplayLists();
  m_geeEngine    = new GEE_Engine();

  m_geeGl->setMaterials(m_geeMaterials);
  m_geeGl->setDisplayLists(m_geeLists);
  m_geeGl->setVeeRenderer(m_veeGl);
  m_geeGl->loadLights("lights/lights-%d", 30);
  m_geeEngine->world().setGarbageBin(&m_internal->m_veeGarbage);

  GEE_World::loadTextureDefinitions(m_veeGl->configFile().c_str());

  // QGLFormat fmt;
  // fmt.

  // setContext(new VEE_GlContext(fmt, this), 0);

  setFocus(); 
}

VEE_QtHidden::~VEE_QtHidden()
{
  VEE_trace("VEE_QtHidden::~VEE_QtHidden # %d frames", m_frames);
  delete m_veeGl;
  delete m_geeGl;
  delete m_geeMaterials;
  delete m_geeLists;
  delete m_geeEngine;

  delete m_internal;
}


void VEE_QtHidden::mouseMoveEvent(QMouseEvent *m)
{
  float ratio = -0.006;

  float aziChange = m->x() - m_internal->m_mousePos.x();
  float eleChange = m->y() - m_internal->m_mousePos.y();

  m_veeGl->addCamAzi(aziChange * ratio);
  m_veeGl->addCamEle(eleChange * ratio);

  m_internal->m_mousePos = m->pos();
  
  // VEE_trace("VEE_QtHidden::mouseMoveEvent");
}

void VEE_QtHidden::mousePressEvent(QMouseEvent *m)
{
  m_internal->m_mousePos = m->pos();
  VEE_trace("VEE_QtHidden::mousePressEvent");
}

void VEE_QtHidden::enterEvent ( QEvent * )
{
  setFocus(); 
}

// GLenum depthTexture;
// const GLenum textype = GL_DEPTH_COMPONENT;

void VEE_QtHidden::initializeGL()
{
  m_veeGl->initializeGL();

  // Load textures:
  QString texdir("texturedir");

  VEE_Qt::loadTextureDir( & m_internal->m_textures, texdir);

  m_lastFrameTime.getTimeOfDay();

  if(m_geeEngine)
    m_geeEngine->rewind(0);
  m_internal->m_simulationTime = 0.0;
#if 0

  // Allocate a 256 x 256 texture
  glGenTextures(1, &depthTexture);
  glBindTexture(GL_TEXTURE_2D, depthTexture);
  glTexImage2D(GL_TEXTURE_2D, 0, textype, 
	       256, 256,
	       0, textype, GL_UNSIGNED_BYTE, NULL);
    
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

#endif

}

bool VEE_QtHidden::displaySolids() const
{
  return m_internal->m_displaySolids;
}

void VEE_QtHidden::reFresh()
{
  preRender();

  m_frames++;

  VEE_Time tmpTime;
  tmpTime.getTimeOfDay();
  
  float secs = (float) m_checkFpsTime.secsTo(tmpTime);
  
  if(secs > 2.0) {
    float fps = (m_frames - m_lastFrameCheckFrames) / secs;
    
    emit status(fps, m_veeGl->sceneBox()->totalChildren());
    
    m_checkFpsTime = tmpTime;
    m_lastFrameCheckFrames = m_frames;

  }
  
  if(m_visualUpdate)
    this->updateGL();
}

void VEE_QtHidden::paintGL (void)
{
  // VEE_trace("VEE_QtHidden::paintGL # frames = %d", m_frames);

  if(m_pause) {
    m_veeGl->moveCamera(m_lastFrameDuration);
    m_veeGl->paintGL(m_showBoxes, true, true);
    return;
  }

  if(m_geeGl)
    m_geeGl->resetCounters();

  if(m_visualUpdate)
    veeRender();
  
  // File output:
  if(m_internal->outputFrames()) {
    storeFrameToFile();
  }

#if 0
  int w = width();
  int h = height();

  // Test move-to-texture
  // glBindTexture(GL_TEXTURE_2D, depthTexture);
  m_veeGl->depthMap().use();
  
  /* glCopyTexSubImage2D(GL_TEXTURE_2D,
		      0,
		      0, 0,
		      0, 0,
		      w, h);*/

  GLenum e = glGetError();
  
  if(e)
    VEE_trace("VEE_QtHidden::paintGL # glCopyTexImage2D failed %s",
	      gluErrorString(e));
  
  glEnable(GL_TEXTURE_2D);
  // glDisable(GL_TEXTURE_2D);

  glDisable(GL_LIGHTING);
  glColor3f(0.7, 0.7, 0.7);

  glBegin(GL_QUADS);

  glNormal3f(0, 0, 1);

  glTexCoord2f(0, 0);
  glVertex2f(0, 0);

  glTexCoord2f(1, 0);
  glVertex2f(3, 0);

  glTexCoord2f(1, 1);
  glVertex2f(3, 3);

  glTexCoord2f(0, 1);
  glVertex2f(0, 3);
  
  glEnd();

#endif

  // VEE_trace("VEE_QtHidden::paintGL # %u frames EXIT", (uint) m_frames);
}

void VEE_QtHidden::preRender()
{
  // const char * fname = "VEE_QtHidden::preRender";

  VEE_Time tmpTime;
  tmpTime.getTimeOfDay();

  if(m_calculusUpdate) {

    float dt;

    if(m_fixedRate > 0.0) {
      dt = m_playSpeed / m_fixedRate;
      float passed = m_lastFrameTime.secsTo(tmpTime);
      if(dt > passed)
	DI_Sleep::sleepMs((uint) ((dt - passed) * 1000.0));
    }
    else if(m_visualUpdate) {
      dt = m_lastFrameTime.secsTo(tmpTime) * m_playSpeed;
    }
    else {
      dt = 0.1;
    }
    
    m_lastFrameDuration = dt;

    m_internal->m_simulationTime += dt;

    float tmp = dt / (float) m_overSample;

    VEE_SceneBox * scene = m_veeGl->sceneBox();

    for(uint i=0; i < m_overSample; i++) {

      // if(m_internal->m_updateType == UPDATE_USUAL)
      scene->update
	(m_splitAt, m_collapseAt, m_maximumRecursion, tmp);
    }

    GEE_TimeStamp ts;
    ts.setSeconds(m_internal->m_simulationTime);
    m_geeEngine->update(ts);
  }
  m_lastFrameTime = tmpTime;
}

void VEE_QtHidden::veeRender(bool clear)
{
  // VEE_trace("VEE_QtHidden::veeRender");
  
  m_veeGl->setBackgroundColor(m_geeEngine->world().backgroundColor());

  m_veeGl->moveCamera(m_lastFrameDuration);
  
  m_camPos = m_veeGl->cameraParams().m_camPos;

  m_veeGl->moveGlCamera();
  if(clear)
    m_veeGl->clearGlScreen();

  if(m_internal->m_displaySolids && m_geeGl) {

    // glPushAttrib(~0);

    m_geeLists->freeLists(& m_internal->m_veeGarbage);
    
    m_geeGl->initGlState();
    
    m_geeEngine->collector().setCameraLocation(m_camPos);
    m_geeEngine->collectObjects();

    m_geeGl->renderFrontToBack(& m_geeEngine->collector().solids());

    if(m_veeGl->applyDepthMap())
      m_veeGl->createDepthMap();

    m_geeGl->renderBackToFront(& m_geeEngine->collector().transparencies());

    m_geeGl->closeGlState();

    /* VEE_trace("VEE_QtHidden::veeRender # %u %u",
       m_geeEngine->collector().solids().size(),
       m_geeEngine->collector().transparencies().size()); */

    // glPopAttrib();
  }

  m_internal->m_veeGarbage.clear();

  glPushAttrib(~0);
  m_veeGl->paintGL(m_showBoxes, false, false, m_lastFrameDuration);
  glPopAttrib();
}

void VEE_QtHidden::storeFrameToFile()
{
  const char * fname = "VEE_QtHidden::storeFrameToFile";

  QImage * image = m_internal->image();

  grab(image);

  m_internal->m_toFileCount--;

  if((m_internal->m_toFileCount % m_internal->m_toFileOverSample) != 0) {
    m_internal->m_accumulator.add(image);
    return;
  }
  else if(m_internal->m_toFileOverSample > 1) {
    m_internal->m_accumulator.add(image);
    m_internal->m_accumulator.divide(m_internal->m_toFileOverSample);
    m_internal->m_accumulator.store(image);
    m_internal->m_accumulator.clear();
  }

  // Save to a file
  QString frameName = m_internal->frameFile();
  VEE_trace("%s # %s", fname, frameName.ascii());
    
  QFileInfo fi(frameName);
  QString suffix(fi.extension(false).upper());

  // Qt does not recognize "JPG" as JPEG:
  if(suffix == "JPG") suffix = "JPEG";

  bool saved = image->save(frameName, suffix, 80);

  if(!saved) {
    VEE_error("%s # Could not find writer for image format \"%s\"", 
	      fname, suffix.ascii());

    QStringList list = image->outputFormatList();
      
    QStringList::Iterator it = list.begin();
    while( it != list.end() ) {
      VEE_trace("%s # Has format \"%s\" ", fname, (*it).ascii());
      ++it;
    }
  }

  m_lastFrameTime.getTimeOfDay();

}

void VEE_QtHidden::grab(QImage *image)
{
  int w = width();
  int h = height();

  image->create(w, h, 32);

  uint pixels = w * h;

  // Grab the image:
  glReadPixels(0, 0, w, h, GL_RGBA, GL_UNSIGNED_BYTE, image->bits());

  *image = image->mirror(false, true);

  // Convert from OpenGL to Qt:
  uchar * bits = image->bits();

  for(uint i=0; i < pixels; i++) {

    uint b1 = bits[0];
    uint b2 = bits[1];
    uint b3 = bits[2];
    uint b4 = bits[3];

    uint argb = b4 << 24 | b1 << 16 | b2 << 8 | b3;
    
    * ((uint *) bits) = argb;

    bits += 4;
  }
}

void VEE_QtHidden::resizeGL (int w, int h )
{
  m_veeGl->resizeGL(w, h);
  m_lastFrameTime.getTimeOfDay();
  
#if 0
  glBindTexture(GL_TEXTURE_2D, depthTexture);

  int s = 2;

  while(s < w || s < h)
    s = s << 1;

  glTexImage2D(GL_TEXTURE_2D, 0, textype, 
	       s, s,
	       0, textype, GL_UNSIGNED_BYTE, NULL);
#endif

}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

class VEE_Qt::Internal
{
public:
  Internal() 
    : m_tools(0), m_params(0), m_acc(10.0)
  {}

  ~Internal() 
  { 
    delete m_params; 
  }

  QPopupMenu *m_tools;
  QPopupMenu *m_keyPointMenu;
  QWidget    *m_params;
  int         m_stretchId;
  int         m_alignId;
  int         m_showObjsId;
  float       m_acc;

  std::vector<VEE_CameraPosition> m_keyPoints;
};


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

VEE_Qt::VEE_Qt(VEE_SceneBox *box,
	       VEE_GlRendererType type,
	       QWidget *parent, 
	       const char *name,
	       bool createGlWidget)
  : QWidget(parent, name),
    m_glWidget(0),
    m_internal(new Internal),
    m_valueWindow(0)
{
  initImageIO();

  int itmp;

  m_menu = new QMenuBar( this, "menu" );
  
  resize(600, 400);

  QPopupMenu * view = new QPopupMenu(m_menu);
  CHECK_PTR( view );
  m_menu->insertItem( "&File", view );
  view->insertItem( "S&ave GEE stuff as...", this, SLOT(saveGeeAs()));
  view->insertItem( "Load GEE stuff...", this, SLOT(loadGee()));
  view->insertItem( "Merge GEE stuff...", this, SLOT(mergeGee()));
  view->insertItem( "&NumObjs", this, SLOT(toggleNObj()), ALT+Key_O );
  view->insertItem( "&Quit", qApp, SLOT(quit()), CTRL+Key_Q );

  QPopupMenu * edit = new QPopupMenu(m_menu);
  CHECK_PTR( edit );
  m_menu->insertItem( "&Edit", edit);

  edit->insertItem( "Undo", this, SLOT(undo()), CTRL + Key_Z);
  edit->insertItem( "redo", this, SLOT(redo()), CTRL + Key_R);
  edit->insertItem( "Rewind", this, SLOT(rewind()));


  QPopupMenu * tools = new QPopupMenu(m_menu);
  CHECK_PTR( tools );
  m_menu->insertItem( "&Tools", tools);
  tools->insertItem( "Params...", this, SLOT(paramsVisible()));
  tools->insertItem( "&Pause", this, SLOT(togglePause()), CTRL + Key_P);
  tools->insertItem( "&Visual updates", this, SLOT(toggleVisualUpdate()), 
		     CTRL + Key_V);
  tools->insertItem( "&Calculus updates", this, SLOT(toggleCalculusUpdate()), 
		     CTRL + Key_C);
  tools->insertItem( "&Blend", this, SLOT(toggleBlend()), CTRL + Key_B);
  tools->insertItem( "Te&xtures", this, SLOT(toggleTextureMode()), CTRL+Key_X);
  tools->insertItem( "B&oxes visible", this, SLOT(toggleShowBoxes()), 
		     CTRL + Key_O);
  tools->insertItem( "&Explosion", this, SLOT(holdTime()), CTRL + Key_E);
  tools->insertItem( "Poly&gon mode", this, SLOT(togglePolygonMode()), 
		     CTRL + Key_G);
  tools->insertItem( "Fo&rce mode", this, SLOT(toggleForceMode()));
  tools->insertItem( "Force box mode", this, SLOT(toggleForceBoxMode()));
  
  /*itmp = tools->insertItem( "Streth particles", this, SLOT(toggleStretch()));
    m_internal->m_stretchId = itmp; */

  itmp = tools->insertItem( "Align particles to motion", 
			    this, SLOT(toggleAlign()));
  // m_internal->m_alignId = itmp;

  tools->insertItem( "Continuous tails", this, SLOT(toggleTails()));

  tools->insertSeparator();
  // tools->insertItem( "&Blend", this, SLOT(holdTime()), Key_E);
  tools->insertItem( "Triangles", this, SLOT(useTriTextures()), CTRL + Key_T);
  tools->insertItem( "Quads", this, SLOT(useQuadTextures()), CTRL + Key_U);
  tools->insertItem( "Points", this, SLOT(usePoints()), CTRL + Key_P);
  tools->insertItem( "Lines", this, SLOT(useLines()), CTRL + Key_L);
  tools->insertItem( "Nothing", this, SLOT(useNothing()), CTRL + Key_N);
  m_internal->m_tools = tools;

  QPopupMenu * options = new QPopupMenu(m_menu);

  m_menu->insertItem( "&Options", options);

  m_internal->m_showObjsId = 
    options->insertItem( "Display &solids", this, SLOT(toggleSolids()));
  options->setItemChecked(m_internal->m_showObjsId, true);
  
  // Show objs:  

  options->insertSeparator();

  // Fade modes:
  options->insertItem( "&No fade", this, SLOT(nofade()));
  options->insertItem( "&Little fade", this, SLOT(fadeLittle()));
  options->insertItem( "&More fade", this, SLOT(fadeMore()));
  options->insertItem( "&Extreme fade", this, SLOT(fadeExtreme()));

  // Background colors:
  options->insertSeparator();
    
  options->insertItem( "&White", this, SLOT(white()));    
  options->insertItem( "&Black", this, SLOT(black()));
  options->insertItem( "&Gray",  this, SLOT(gray()));
  options->insertItem( "Bl&ue",  this, SLOT(blue()));
  options->insertItem( "Custom color...",  this, SLOT(bgColorDialog()));

  // Blending modes:
  options->insertSeparator();

  options->insertItem( "&New", this, SLOT(newBlend()));    
  options->insertItem( "&Additive", this, SLOT(additive()));    
  options->insertItem( "&Subtractive", this, SLOT(subtractive()));
  options->insertItem( "&Usual",  this, SLOT(usual()));

  options->insertSeparator();

  const char * orders[VEE_GlRenderer::OrderSizeof] = {
    "Random order",
    "Random order (via store)",
    "Back to front 1",
    "Back to front 2"
  };

  QString tmp, tmp2;
  uint i;

  for(i=0; i < VEE_GlRenderer::OrderSizeof; i++) {

    tmp2.sprintf("%u", i);
    MJ_StrRemap * remap = new MJ_StrRemap(this, tmp2);

    connect(remap, SIGNAL(emitCall(const QString &)), 
	    this, SLOT(setRenderOrder(const QString &)));

    options->insertItem(orders[i], remap, SLOT(call()));
    
  }

  options->insertSeparator();

  // Update types
  const char * updates[VEE_QtHidden::UPDATE_SIZEOF] = {
    "Normal updates",
    "Pseudo-thread updates (test)",
    "Thread updates (test)"
  };

  for(i=0; i < VEE_QtHidden::UPDATE_SIZEOF; i++) {

    tmp2.sprintf("%u", i);
    MJ_StrRemap * remap = new MJ_StrRemap(this, tmp2);

    connect(remap, SIGNAL(emitCall(const QString &)), 
	    this, SLOT(setUpdateType(const QString &)));

    options->insertItem(updates[i], remap, SLOT(call()));
    
  }


  // Allow explicit (and exact resizing) of the GL widget: 

  /* Size pairs, the first number is width, the second height. */
  int sizes[] = { 320, 256, 480, 320, 640, 480, 720, 576, 1024, 768, 0, 0 };

  options->insertSeparator();

  for(uint i=0; sizes[i*2]; i++) {
    int w = sizes[i * 2];
    int h = sizes[i * 2 + 1];

    tmp.sprintf("%d x %d", w, h);
    tmp2.sprintf("%d %d", w, h);

    MJ_StrRemap * remap = new MJ_StrRemap(this, tmp2);
    connect(remap, SIGNAL(emitCall(const QString &)), 
	    this, SLOT(resizeGlWidget(const QString &)));

    options->insertItem(tmp, remap, SLOT(call()));
  }

  // Create the keypoint menu:
  QPopupMenu * keyPoints = new QPopupMenu(m_menu);
  m_menu->insertItem( "&Navigation", keyPoints);
  keyPoints->insertItem("&Fit GEE", this, SLOT(fitGee()));
  
  // Acceleration speed menu:
  QPopupMenu * speeds = new QPopupMenu(keyPoints);

  const char * speedNames[] = {
    "Snail",  "0.3",
    "Slow",   "2.0",
    "Medium", "10.0",
    "Fast",   "50.0",
    "Rocket", "250.0",
    0
  };

  for(i = 0; speedNames[i * 2]; i++) {
    MJ_StrRemap * remap = new MJ_StrRemap(this, speedNames[i * 2 + 1]);
    connect(remap, SIGNAL(emitCall(const QString &)), 
	    this, SLOT(setAcceleration(const QString &)));
    speeds->insertItem(speedNames[i * 2], remap, SLOT(call()));
  }
  keyPoints->insertItem("Motion", speeds);

  keyPoints->insertItem("&Capture keypoint", this, SLOT(addKeyPoint()));

  m_internal->m_keyPointMenu = keyPoints;

  addKeyPoint2(0, 0, 10, VEE_PI, 0.2);
  addKeyPoint2(0, 10, 10, VEE_PI, -0.4);
  addKeyPoint2(10, 0, 0, -VEE_HALF_PI, 0.2);


  // Do the non-menu stuff:

  m_layout = new QVBoxLayout(this);
  QHBoxLayout * hlayout = new QHBoxLayout(m_layout);

  m_layout->setMenuBar(m_menu);


  QString title("VEE");
  setCaption(title);
  setIconText(title);

  QPushButton *pb = new QPushButton("Screenshot", this);
  connect(pb, SIGNAL(pressed()), this, SLOT(screenshot()));  
  pb->setFocusPolicy(NoFocus);
  hlayout->addWidget(pb, 1, Qt::AlignLeft);

  pb = new QPushButton("Largeshot", this);
  connect(pb, SIGNAL(pressed()), this, SLOT(screenshotLarge()));  
  pb->setFocusPolicy(NoFocus);
  hlayout->addWidget(pb, 1, Qt::AlignLeft);

  pb = new QPushButton("Begin movie", this);
  connect(pb, SIGNAL(pressed()), this, SLOT(movieBegin()));  
  pb->setFocusPolicy(NoFocus);
  hlayout->addWidget(pb, 1, Qt::AlignLeft);

  pb = new QPushButton("End movie", this);
  connect(pb, SIGNAL(pressed()), this, SLOT(movieEnd()));  
  pb->setFocusPolicy(NoFocus);
  hlayout->addWidget(pb, 1, Qt::AlignLeft);

  hlayout->addStretch(90);

  if(createGlWidget) {
    setGlWidget(new VEE_QtHidden(this, box, this, "vee-glw", type));
    /* tools->setItemChecked(itmp, renderer()->stretchParticles());
       tools->setItemChecked(m_internal->m_alignId, 
       renderer()->stretchParticles()); */
  }
}

VEE_Qt::~VEE_Qt()
{
  VEE_trace("VEE_Qt::~VEE_Qt # %p %p", this, m_valueWindow);
  delete m_valueWindow;
  delete m_internal;
}

bool VEE_Qt::setValue(const char *id, VEE_Value *v)
{
  if(!m_glWidget)
    VEE_fatal("VEE_Qt::setValue # m_glWidget = NULL");

  bool ok = false;

  if(strcmp(id, "splitat") == 0) {
    int tmp;
    ok = v->getInt(&tmp);
    if(ok) 
      m_glWidget->m_splitAt = VEE_clamp(tmp, 5, 10000);
  }
  else if(strcmp(id, "collapseat") == 0) {
    int tmp;
    ok = v->getInt(&tmp);
    if(ok) 
      m_glWidget->m_collapseAt = VEE_clamp(tmp, 5, 10000);
  }
  else if(strcmp(id, "maxrecursion") == 0) {
    int tmp;
    ok = v->getInt(&tmp);
    if(ok) 
      m_glWidget->m_maximumRecursion = VEE_clamp(tmp, 1, 15);
  }
  else if(strcmp(id, "rate") == 0) {
    float tmp;
    ok = v->getFloat(&tmp);
    if(ok) 
      m_glWidget->m_fixedRate = VEE_clamp(tmp, 0.0f, 180.0f);
  }
  else if(strcmp(id, "speed") == 0) {
    float tmp;
    ok = v->getFloat(&tmp);
    if(ok) 
      m_glWidget->m_playSpeed = VEE_clamp(tmp, 0.01f, 10.0f);
  }
  else if(strcmp(id, "fileoversample") == 0) {
    int tmp;
    ok = v->getInt(&tmp);
    if(ok) 
      m_glWidget->m_internal->m_toFileOverSample = VEE_clamp(tmp, 1, 1000);
  }
  else if(strncmp(id, "GL/", 3) == 0) {
    return m_glWidget->m_veeGl->setValue(id + 3, v);
  }
  return ok;
}

void VEE_Qt::setGlWidget(VEE_QtHidden *glw)
{
  delete m_glWidget;

  m_glWidget = glw;

  m_glWidget->m_checkFpsTime.getTimeOfDay();
  m_glWidget->m_lastFrameTime = m_glWidget->m_checkFpsTime;
  m_glWidget->m_lastFrameDuration = 0.1;

  connect(m_glWidget, SIGNAL(status(float,int)), 
	  this, SLOT(showStatus(float,int)));
  m_glWidget->m_veeGl->sceneBox()->refreshTimeLastUpdate();

  QTimer *t = new QTimer(this);

  connect(t, SIGNAL(timeout()),
          m_glWidget, SLOT(reFresh()) );
  t->start(0);

  m_layout->addWidget(m_glWidget, 100);
  m_glWidget->show();

  // Light selection:

  QPopupMenu * lights = new QPopupMenu(m_menu);
  m_menu->insertItem( "&Lights", lights);
  
  GEE_GlRenderer * geeGl = m_glWidget->m_geeGl;

  QString tmp;

  for(uint i = 0; i < geeGl->lightCount(); i++) {

    tmp.sprintf("%u", i);
    MJ_StrRemap * remap = new MJ_StrRemap(this, tmp);
    
    connect(remap, SIGNAL(emitCall(const QString &)), 
	    this, SLOT(selectLights(const QString &)));

    tmp.sprintf("Light set %u", i);

    lights->insertItem(tmp, remap, SLOT(call()));    
  }
}    


void VEE_Qt::setFrameFile(const char *name)
{
  m_glWidget->m_internal->m_frameFileName = name;
}

void VEE_Qt::setScreenShotCount(int n)
{
  m_glWidget->m_internal->m_toFileCount = n;
}

void VEE_Qt::setScreenShotOverSample(int i)
{
  m_glWidget->m_internal->m_toFileOverSample = i;
}

void VEE_Qt::initImageIO()
{
  static bool done = false;
  
  if(done) return;

  done = true;
  VEE_Image::addIO(new VEE_ImageIOQt);
}

int VEE_Qt::addKeyPoint(const VEE_Vector3 & loc, float azi, float ele)
{
  VEE_CameraPosition p(loc, azi, ele);
  char buf[256];
  p.describe(buf);
  uint n = m_internal->m_keyPointMenu->count() - 1;

  QString tmp;
  tmp.sprintf("%u", n);
  MJ_StrRemap * remap = new MJ_StrRemap(this, tmp);
  
  connect(remap, SIGNAL(emitCall(const QString &)), 
	  this, SLOT(selectKeyPoint(const QString &)));

  m_internal->m_keyPointMenu->insertItem(buf, remap, SLOT(call()));

  m_internal->m_keyPoints.push_back(p);

  VEE_trace("VEE_Qt::addKeypoint # Added keypoint %s", buf);

  return (int) n;
}

void VEE_Qt::selectKeyPoint(const QString &str)
{
  uint i = str.toUInt();
  selectKeyPoint2(i);
}

void VEE_Qt::selectKeyPoint2(int i)
{
  i = i % m_internal->m_keyPoints.size();
  VEE_CameraPosition p = m_internal->m_keyPoints[i];
  renderer()->setCameraParams(p);
}

void VEE_Qt::selectLights(const QString &str)
{
  uint i = str.toUInt();
  m_glWidget->m_geeGl->selectLights(i);
}

void VEE_Qt::rewind()
{
  m_glWidget->m_geeEngine->rewind(0);
  m_glWidget->m_internal->m_simulationTime = 0.0;
}

void VEE_Qt::undo()
{
  m_glWidget->m_geeEngine->undo();
}

void VEE_Qt::redo()
{
  m_glWidget->m_geeEngine->redo();
}

static QString filename("file.gee");

void VEE_Qt::saveGeeAs()
{
  filename = QFileDialog::getSaveFileName(filename, "*.gee");

  if(filename.isEmpty())
    return;

  m_glWidget->geeEngine()->write(filename.ascii());
}

void VEE_Qt::loadGee()
{
  filename = QFileDialog::getOpenFileName(filename, "*.gee");
  
  if(filename.isEmpty())
    return;

  m_glWidget->geeEngine()->read(filename.ascii());
  rewind();

}

void VEE_Qt::mergeGee()
{
  filename = QFileDialog::getOpenFileName(filename, "*.gee");
  
  if(filename.isEmpty())
    return;

  m_glWidget->geeEngine()->read(filename.ascii(), true);
  rewind();
}

/** Approximately fits all GEE stuff into the OpenGL window. */

void VEE_Qt::fitGee()
{
  VEE_BoundingBox bb;
  bb.clear();

  m_glWidget->geeEngine()->world().getBoundingBox(&bb);

  VEE_GlRenderer * gl = m_glWidget->m_veeGl;

  VEE_Vector3 span = bb.span();
  float fov = gl->fov();

  float frontDiagonal = span.vector2().length();
  float fovRatio = sinf(fov * (VEE_PI / 180.0f));
  float distance = 1.3 * frontDiagonal / fovRatio;
  
  VEE_Vector3 loc = bb.low();
  loc.x += span.x * 0.5f;
  loc.y += span.y * 0.5f;

  loc.z -= distance;

  VEE_trace("VEE_Qt::fitGee # %f %f %f", loc.x, loc.y, loc.z);

  VEE_CameraPosition p(loc, 0.0, 0.0);

  gl->setCameraParams(p);
}

void VEE_Qt::setAcceleration(const QString &str)
{
  bool ok = true;
  float acc = str.toFloat(&ok);
  if(ok)
    m_internal->m_acc = acc;
}

void VEE_Qt::toggleNObj()
{
  m_glWidget->m_showNObj = ! m_glWidget->m_showNObj;
}

void VEE_Qt::togglePause() 
{
  m_glWidget->m_pause = !m_glWidget->m_pause;

  if(!m_glWidget->m_pause) 
    m_glWidget->m_veeGl->sceneBox()->refreshTimeLastUpdate();
}
void VEE_Qt::toggleVisualUpdate()
{
  m_glWidget->m_visualUpdate = !m_glWidget->m_visualUpdate;

  if(m_glWidget->m_visualUpdate)
    m_glWidget->m_veeGl->sceneBox()->refreshTimeLastUpdate();
}

void VEE_Qt::toggleCalculusUpdate()
{
  m_glWidget->m_calculusUpdate = !m_glWidget->m_calculusUpdate;

  m_glWidget->m_veeGl->sceneBox()->refreshTimeLastUpdate();
  m_glWidget->m_lastFrameTime.getTimeOfDay();
}

void VEE_Qt::toggleShowBoxes()
{
  m_glWidget->m_showBoxes = !m_glWidget->m_showBoxes;
}

void VEE_Qt::toggleBlend()
{
  m_glWidget->m_veeGl->setBlend(!m_glWidget->m_veeGl->getBlend());
}

void VEE_Qt::togglePolygonMode()
{
  VEE_GlRenderer::Mode m = 
    m_glWidget->m_veeGl->getPolygonMode() == VEE_GlRenderer::RenderTriangles ? 
    VEE_GlRenderer::RenderLines : VEE_GlRenderer::RenderTriangles;
    
  m_glWidget->m_veeGl->setPolygonMode(m);
}

void VEE_Qt::toggleForceMode()
{
  VEE_GlRenderer::Mode m = 
    m_glWidget->m_veeGl->getForceMode() == VEE_GlRenderer::RenderPoints ? 
    VEE_GlRenderer::RenderNothing : VEE_GlRenderer::RenderPoints;
    
  m_glWidget->m_veeGl->setForceMode(m);

  VEE_trace("VEE_Qt::toggleForceMode # %d", (int) m);
}

void VEE_Qt::toggleForceBoxMode()
{
  m_glWidget->m_veeGl->setDrawForceBoxes
    (!m_glWidget->m_veeGl->getDrawForceBoxes());
}

void VEE_Qt::toggleAlign()
{
  renderer()->nextAlignMent();
  VEE_trace("VEE_Qt::toggleAlign # %d", (int) renderer()->align());
}

void VEE_Qt::toggleTails()
{
  renderer()->setContinuousTails(!renderer()->continuousTails());
}

void VEE_Qt::toggleSolids()
{
  if(!m_glWidget)
    return;

  bool &display = m_glWidget->m_internal->m_displaySolids;

  display = !display;

  m_menu->setItemChecked
    (m_internal->m_showObjsId, display);
}


void VEE_Qt::toggleTextureMode()
{
  m_glWidget->m_veeGl->setUseTextures(!m_glWidget->m_veeGl->getUseTextures());
  //VEE_trace("VEE_Qt::toggleTextureMode # %d", (int) m_glWidget->m_veeGl->getUseTextures());
}


void VEE_Qt::holdTime()
{
  m_glWidget->m_veeGl->sceneBox()->update
    (m_glWidget->m_splitAt, m_glWidget->m_collapseAt, m_glWidget->m_maximumRecursion, 1.0);
}


void VEE_Qt::useQuadTextures()
{
  m_glWidget->m_veeGl->setParticleMode(VEE_GlRenderer::RenderQuads);
}

void VEE_Qt::useTriTextures()
{
  // m_glWidget->m_veeGl->setParticleMode(VEE_GlRenderer::RenderTriangles);
}

void VEE_Qt::usePoints()
{
  m_glWidget->m_veeGl->setParticleMode(VEE_GlRenderer::RenderPoints);  
}

void VEE_Qt::useLines()
{
  m_glWidget->m_veeGl->setParticleMode(VEE_GlRenderer::RenderLines);  
}

void VEE_Qt::useNothing()
{
  m_glWidget->m_veeGl->setParticleMode(VEE_GlRenderer::RenderNothing);  
}

void VEE_Qt::showStatus(float fps, int objects)
{
  QString title;
  
  GEE_GlRenderer * gl = m_glWidget->m_geeGl;

  title.sprintf
    ("VEE %.1f Hz, %d objects (%.2lf secs, %ld gs, %ld vs, %ld ts, %ld ps)", 
     fps, objects, m_glWidget->m_internal->m_simulationTime,
     gl->geometryCount(), gl->vertexCount(), 
     gl->triangleCount(), gl->particleCount());
  setCaption(title);
  setIconText(title);
}

void VEE_Qt::nofade()
{
  m_glWidget->m_veeGl->setFade(0.0);
}

void VEE_Qt::fadeLittle()
{
  m_glWidget->m_veeGl->setFade(0.2);
}

void VEE_Qt::fadeMore()
{
  m_glWidget->m_veeGl->setFade(0.7);
}

void VEE_Qt::fadeExtreme()
{
  m_glWidget->m_veeGl->setFade(1.7);
}

void VEE_Qt::resizeGlWidget(const QString &size)
{
  int w = 640;
  int h = 480;
  
  int n = sscanf(size.ascii(), "%d %d", &w, &h);

  if(n != 2) {
    VEE_error("VEE_Qt::resizeGL(const QString &) # Bad Argument \"%s\"",
	      size.ascii());
    return;
  }

  resizeGlWidget(w, h);
}

void VEE_Qt::screenshot()
{
  m_glWidget->m_internal->m_toFileCount += 
    m_glWidget->m_internal->m_toFileOverSample;
}

void VEE_Qt::screenshotLarge()
{
  const char * fname = "VEE_Qt::screenshotLarge";

  int w = m_glWidget->width();
  int h = m_glWidget->height();

  QPixmap pm = m_glWidget->renderPixmap(w * 3, h * 3, true);

  QImage qi = pm.convertToImage();
  
  QString frameName = m_glWidget->m_internal->frameFile();
  VEE_trace("%s # %s", fname, frameName.ascii());
    
  QFileInfo fi(frameName);
  QString suffix(fi.extension(false).upper());

  // Qt does not recognize "JPG" as JPEG:
  if(suffix == "JPG") suffix = "JPEG";

  bool saved = qi.save(frameName, suffix, 80);

  if(!saved) {
    VEE_error("%s # Could not write image \"%s\"", 
	      fname, frameName.ascii());
  }

  // m_lastFrameTime.getTimeOfDay();
}

void VEE_Qt::movieBegin()
{
  m_glWidget->m_internal->m_toFileCount = 100000;
  m_glWidget->m_fixedRate = 25.0 * m_glWidget->m_internal->m_toFileOverSample;
}

void VEE_Qt::movieEnd()
{
  m_glWidget->m_internal->m_toFileCount = 0;
  m_glWidget->m_fixedRate = 0.0;
}

void VEE_Qt::paramsVisible()
{
  QWidget * params = m_internal->m_params;

  if(params) {
    params->raise();
    params->show();
    return;
  }

  VEE_GlRenderer * gl = m_glWidget->m_veeGl;

  VEE_ValueWindow * paramView = new VEE_ValueWindow();
  
  paramView->addFloatParameter("Particle size", "GL/size",
			       this, gl->particleSize(), 0.01, 30, 
			       VEE_MAP_LOGARITHMIC);
  
  paramView->addFloatParameter("Particle alpha", "GL/alphamul",
			       this, gl->alphaMul(), 0.01, 3.0, 
			       VEE_MAP_LINEAR);
    
  paramView->addFloatParameter("Lengthscale", "GL/lengthscale",
			       this, gl->lengthScale(), 0.05, 10.0, 
			       VEE_MAP_LOGARITHMIC);
  
  paramView->addFloatParameter("Max lengthscale", "GL/maxlengthscale",
			       this, gl->maxLengthScale(), 1.0, 50.0, 
			       VEE_MAP_LOGARITHMIC);
  
  paramView->addFloatParameter("Tail dispersion", "GL/taildispersion",
			       this, gl->tailDispersion(), 0.05, 10.0, 
			       VEE_MAP_LOGARITHMIC);
  
  paramView->addFloatParameter("Field of View", "GL/fov",
			       this, gl->fov(), 15, 140.0, 
			       VEE_MAP_LOGARITHMIC);
  
  paramView->addFloatParameter("Texture", "GL/subtexture",
			       this, renderer()->currentSubTexture(),
			       0, renderer()->subTextureCount() - 1,
			       VEE_MAP_LINEAR);

  paramView->addFloatParameter("Movie multisampling", "fileoversample",
			       this, 
			       m_glWidget->m_internal->m_toFileOverSample, 
			       1, 20, VEE_MAP_LINEAR);

  paramView->addFloatParameter("Split boxes at", "splitat",
			       this, m_glWidget->m_splitAt, 6.0, 2000.0, 
			       VEE_MAP_LOGARITHMIC);

  paramView->addFloatParameter("Collapse boxes at", "collapseat",
			       this, m_glWidget->m_collapseAt, 6.0, 2000.0, 
			       VEE_MAP_LOGARITHMIC);
  
  paramView->addFloatParameter("Max recursion", "maxrecursion", this, 
			       m_glWidget->m_maximumRecursion, 0.0, 15.0, 
			       VEE_MAP_LINEAR);

  QString tmp("VEE simulation parameters");

  paramView->setCaption(tmp);
  paramView->setIconText(tmp);
  
  paramView->show();
  
  m_internal->m_params = paramView;
}

void VEE_Qt::white()
{
  deliverBgColor(VEE_Rgba::White);
}

void VEE_Qt::black()
{
  deliverBgColor(VEE_Rgba::Black);
}

void VEE_Qt::gray()
{
  deliverBgColor(VEE_Rgba::Gray);
}

void VEE_Qt::blue()
{
  deliverBgColor(VEE_Rgba(0.15, 0.15, 0.3, 1.0, 0.0));
}

void VEE_Qt::newBlend()
{
  m_glWidget->m_veeGl->setBlendType(VEE_BLEND_NEW);
}


void VEE_Qt::additive()
{
  m_glWidget->m_veeGl->setBlendType(VEE_BLEND_ADDITIVE);
}

void VEE_Qt::subtractive()
{
  m_glWidget->m_veeGl->setBlendType(VEE_BLEND_SUBTRACTIVE);
}

void VEE_Qt::usual()
{
  m_glWidget->m_veeGl->setBlendType(VEE_BLEND_USUAL);
}

void VEE_Qt::setRenderOrder(const QString &str)
{
  int order = str.toInt();

  m_glWidget->m_veeGl->setValueInt("renderorder", order);
}

void VEE_Qt::setUpdateType(const QString &str)
{
  int type = str.toInt();
  
  m_glWidget->m_internal->m_updateType = type;
}

void VEE_Qt::bgColorDialog()
{
  VEE_Rgba rgba = m_glWidget->m_geeEngine->world().backgroundColor();

  QRgb qrgb = rgba.getARGB();

  // VEE_trace("VEE_Qt::bgColorDialog # %x", rgba.getARGB());

  qrgb = QColorDialog::getRgba(qrgb);

  float scale = 1.0f / 255.0f;

  rgba.set(qRed(qrgb) * scale, 
	   qGreen(qrgb) * scale, 
	   qBlue(qrgb) * scale, 
	   qAlpha(qrgb) * scale);
  
  deliverBgColor(rgba);

}


void VEE_Qt::addKeyPoint()
{
  VEE_CameraPosition p = renderer()->cameraParams();
  addKeyPoint(p.m_camPos, p.m_azi, p.m_ele);
}


void VEE_Qt::loadTextureDir(VEE_GlTexture2DBank *store,
			    const QString &dirname)
{
  const char * fname = "VEE_Qt::loadTextureDir";
  
  QDir dir(dirname);
  
  QStringList files = dir.entryList("*.png");
  
  VEE_trace("%s # About to load %d textures from \"%s\"", 
	    fname, (int) files.count(), dirname.ascii());
  
  for(uint i = 0; i < files.count(); i++) {
    QString fileBase(files[i]);
    QString file(dirname + "/" + fileBase);
    VEE_Image img;

    if(img.load(file.ascii())) {
      
      VEE_GlTexture2D tex;
      tex.genTexture(&img, true);
      
      const char * tmp1 = fileBase.ascii();
      const char * tmp2 = strchr(tmp1, '.');
      
      if(tmp2)
	fileBase.truncate(tmp2 - tmp1);
      
      store->addTexture(fileBase.ascii(), tex);
      VEE_trace("%s # Loaded texture %s (%s)", 
		fname, fileBase.ascii(), file.ascii());
    }
    else
      VEE_error("%s # Could not load texture %s", fname, file.ascii());
  }
}
  

void VEE_Qt::keyPressEvent(QKeyEvent *k)
{
  float acc = m_internal->m_acc;

  VEE_GlRenderer * gl = m_glWidget->m_veeGl;

  float dt = m_glWidget->m_lastFrameDuration;

  switch ( k->key() ) {
  case Key_Right:
    gl->addCamAzi(-0.05);
    break;
  case Key_Left:
    gl->addCamAzi(+0.05);
    break;
  case Key_Up:
    gl->addCamEle(+0.05);
    break;
  case Key_Down:
    gl->addCamEle(-0.05);
    break;
  case Key_F:
    m_glWidget->m_showFps = !m_glWidget->m_showFps;
    m_glWidget->m_checkFpsTime.getTimeOfDay();
    m_glWidget->m_lastFrameCheckFrames = m_glWidget->m_frames;
    break;

  case Key_A:
    gl->forwardAcceleration
      (acc * dt / m_glWidget->m_playSpeed);
    break;
  case Key_Z:
    gl->forwardAcceleration
      (-acc * dt / m_glWidget->m_playSpeed);
    break;
  case Key_X:
    gl->sidewaysAcceleration
      (acc * dt / m_glWidget->m_playSpeed);
    break;
  case Key_Less:
    gl->sidewaysAcceleration
      (-acc * dt / m_glWidget->m_playSpeed);
    break;

  case Key_Space:
    gl->stopMotion();
    break;

  case Key_PageUp:
    gl->moveCameraUp(acc * 0.01);
    break;
  case Key_PageDown:
    gl->moveCameraUp(acc *-0.01);
    break;

  case Key_Plus:
    m_glWidget->m_playSpeed *= 1.3;
    break;
  case Key_Minus:
    m_glWidget->m_playSpeed /= 1.3;
    break;

  case Key_Escape:
    QApplication::exit();
    break;
  }
}

void VEE_Qt::deliverBgColor(const VEE_Rgba & color)
{
  m_glWidget->m_geeEngine->world().setBackgroundColor(color);
}


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

bool VEE_ImageIOQt::read(VEE_Image *image, const char *filename, 
			 int w, int h, int chans)
{
  QImage qi;
  if(!qi.load(filename)) {
    VEE_error("VEE_ImageIOQt::read # Failed to load \"%s\"", filename);
    return false;
  }
  
  qi = qi.convertDepth(32);

  uint wi = qi.width();
  uint he = qi.height();

  if(w > 0 || h > 0) {
    w = w > 0 ? w : wi;
    h = h > 0 ? h : he;
    qi = qi.smoothScale(w, h);
  }
  else {
    w = wi;
    h = he;
  }

  if(chans <= 0)
    chans = 3;

  bool hasAlpha = qi.hasAlphaBuffer();

  if(hasAlpha)
    chans = 4;
    
  image->resize(w, h, chans);

  for(int y=0; y < h; y++) {
    for(int x=0; x < w; x++) {
      uint c = qi.pixel(x, y);
      unsigned char * px = image->pixel(x, y);

      if(chans == 1)
	*px = VEE_max(qRed(c), qGreen(c), qBlue(c));
      else if(chans >= 3) {
	px[0] = qRed(c);
	px[1] = qGreen(c);
	px[2] = qBlue(c);
	if(chans == 4) {
	  if(hasAlpha)
	    px[3] = qAlpha(c);
	  else
	    px[3] = 0xff;
	}
      }
    }
  }

  return true;
}

bool VEE_ImageIOQt::write(VEE_Image *image, const char *filename)
{
  QFileInfo fi(filename);
  QString suffix(fi.extension(false).upper());
  if(suffix == "JPG") suffix = "JPEG";

  QImage qi;

  uint w = image->w();
  uint h = image->h();
  uint chans = image->bytesPerPixel();

  if(!qi.create(w, h, 32))
    return false;

  if(chans == 4)
    qi.setAlphaBuffer(true);

  for(uint y=0; y < h; y++) {
    for(uint x=0; x < w; x++) {

      unsigned char * px = image->pixel(x, y);

      if(chans < 3)
	qi.setPixel(x, y, qRgb(*px, *px, *px));
      else if(chans == 3)
	qi.setPixel(x, y, qRgb(px[0], px[1], px[2]));
      else 
	qi.setPixel(x, y, qRgba(px[0], px[1], px[2], px[3]));
    }
  }

  bool ok = qi.save(filename, suffix, 80);

  return ok;
}
