/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <vee_rman_renderer.h>

#include <vee_error.h>
#include <vee_looks.h>
#include <vee_particle.h>
#include <vee_polygon.h>
#include <vee_scene.h>

#include <ri.h>

VEE_RManRenderer::VEE_RManRenderer(float frameDuration)
  : m_particleSize(0.25),
    m_frameDuration(frameDuration),
    m_particleMaterial("vee_particle"),
    m_polygonMaterial("vee_polygon")
{}

VEE_RManRenderer::~VEE_RManRenderer()
{}

void VEE_RManRenderer::initRenderState()
{
  RiSurface(&* m_particleMaterial.begin(), RI_NULL);
}

void VEE_RManRenderer::closeRenderState()
{}

void VEE_RManRenderer::renderParticle(VEE_Particle *p)
{
  VEE_Vector3 loc = p->location();
  VEE_Vector3 vel = p->velocity();

  VEE_Looks *looks = p->looks();
  
  RiTransformBegin();

  
  VEE_Rgba rgba;

  if(looks) {
    rgba = looks->color();
  }
  else 
    rgba = VEE_Rgba::Green;

  RtColor color = { rgba.red(), rgba.green(), rgba.blue() };
  RiColor(color);  
    
  float tmp = rgba.alpha();

  RtColor opacity = { tmp , tmp, tmp };
  RiOpacity(opacity);
  
  RiMotionBegin(2, 0.0, 1.0);
  {
    // The old location:
    VEE_Vector3 loc1 = loc - (vel * m_frameDuration);
    RiTranslate(loc1[0], loc1[1], loc1[2]);
    // The current location:
    RiTranslate(loc[0], loc[1], loc[2]);
  }
  RiMotionEnd();

  RiSphere(m_particleSize, -m_particleSize, m_particleSize, 360.0, RI_NULL);

  RiTransformEnd();
}

void VEE_RManRenderer::renderPolygon(VEE_Polygon *p)
{
  // VEE_trace("VEE_RManRenderer::renderPolygon");

  uint rows = p->getMeshRows() + 1;
  uint cols = p->getMeshColumns() + 1;

  if(rows < 3 || cols < 3) {
    VEE_error("VEE_RManRenderer::renderPolygon # Bad polygon");
    return;
  }

  // RiAttributeBegin();

  RiSurface(&* m_polygonMaterial.begin(), RI_NULL);
  
  VEE_Rgba rgba;

  VEE_Looks *looks = p->looks();

  if(looks) {
    rgba = looks->color();
  }
  else 
    rgba = VEE_Rgba::Green;

  RtColor color = { rgba.red(), rgba.green(), rgba.blue() };
  RiColor(color);  
    
  float tmp = 1.0; // rgba.alpha();

  RtColor opacity = { tmp , tmp, tmp };
  RiOpacity(opacity);

  RiBasis(RiCatmullRomBasis, RI_CATMULLROMSTEP, 
	  RiCatmullRomBasis, RI_CATMULLROMSTEP);

  RiPatchMesh(RI_BICUBIC, cols, RI_NONPERIODIC, rows, RI_NONPERIODIC,
	      RI_P, (RtPointer) & p->getVertex(0), RI_NULL);

  RiSurface(&*m_particleMaterial.begin(), RI_NULL);

  // RiAttributeEnd();
}

void VEE_RManRenderer::renderForceObject(VEE_ForceObject *)
{
  // float size = m_particleSize * 3.0;
  // RiCylinder(size, size, size, 360.0, 0);
}

void VEE_RManRenderer::paintRecursive(VEE_SceneBox *box)
{
  // VEE_trace("VEE_RManRenderer::paintRecursive # %p %p", this, box);

  // RiTransformBegin();
  // RiTransformEnd();

  {
    VEE_SceneBox::iterator it = box->begin();
    VEE_SceneBox::iterator sentinel = box->end();
    
    while(it != sentinel) {
      (*it)->render(this);
      it++;
    }
  }

  // Render the child boxes:
  if(box->isSplit()) {
    uint i = 0;
    for(i = 0; i < 8; i++) 
      paintRecursive(box->child(i));
  }

}

void VEE_RManRenderer::aimZ(const VEE_Vector3 &dir)
{
  const float PI = 3.14159265359;

  double xzlen, yzlen, yrot, xrot;

  if ( ( dir[0] == 0 )
       && ( dir[1] == 0 )
       && ( dir[2] == 0 ) )
    return;

  xzlen = sqrt(dir[0]*dir[0]+dir[2]*dir[2]);

  if (xzlen == 0)
    yrot = ( dir[1] < 0 ) ? 180 : 0;
  else
    yrot = 180*acos(dir[2]/xzlen)/PI;

  yzlen = sqrt(dir[1]*dir[1]+xzlen*xzlen);
  xrot = 180*acos(xzlen/yzlen)/PI;

  if ( dir[1] > 0 )
    RiRotate(xrot,1.0,0.0,0.0);
  else
    RiRotate(-xrot,1.0,0.0,0.0);

  if ( dir[0] > 0 )
    RiRotate(-yrot,0.0,1.0,0.0);
  else
    RiRotate(yrot,0.0,1.0,0.0);
}

void VEE_RManRenderer::placeCamera(const VEE_Vector3 &pos,
				   const VEE_Vector3 &dir,
				   float roll)
{
  RiIdentity();
  RiRotate(-roll,0.0,0.0,1.0);
  aimZ(dir);
  RiTranslate(-pos[0],-pos[1],-pos[2]);
}
