/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <vee_vrs_app.h>

#include <gee_group.h>
#include <gee_solar_conversion.h>
#include <gee_world.h>

#include <vee_error.h>

#include <Solar/Object.hpp>

#include <SDL/SDL.h>

#include <GL/gl.h>
#include <GL/glu.h>

static void runSDL(VEE_SdlApp & app)
{
  const char * fname = "VEE/SDL/run";

  VEE_trace(fname);

  if ( SDL_Init(SDL_INIT_VIDEO) < 0 ) {
    VEE_error("%s # Unable to initialize SDL: %s\n", fname, SDL_GetError());
    return;
  }

  int w = 640;
  int h = 480;

  /* Create a 640x480 OpenGL screen */
  if ( SDL_SetVideoMode(w, h, 0, SDL_OPENGL/* | SDL_NOFRAME*/ ) == NULL ) {
    VEE_error("%s # Unable to create OpenGL screen: %s\n", fname,
	    SDL_GetError());
    SDL_Quit();
    return;
  }

  /* Set the title bar in environments that support it */
  SDL_WM_SetCaption("VEE loader", NULL);

  VEE_SdlApp::GlData gld;

  app.initGl( & gld);
  
  bool done = false;

  while(!done) {
    
    SDL_Event event;

    while ( SDL_PollEvent(&event) ) {
      if ( event.type == SDL_QUIT )
	done = 1;
      if ( event.type == SDL_KEYDOWN ) {
	if ( event.key.keysym.sym == SDLK_ESCAPE )
	  done = 1;
      }
    }

    if(app.isDone())
      done = 1;
    
    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glMatrixMode(GL_PROJECTION);

    glLoadIdentity ();

    gluPerspective (70.0f,
		    (float) w / (float) h, 
		    0.05,   /* near plane */
		    10000.0f);  /* far plane */

    glMatrixMode(GL_MODELVIEW);

    glLoadIdentity ();

    gluLookAt(0, 0, 0,
	      0, 0, 1, 
	      0, 1, 0);
    app.update();
    app.draw( & gld);

    SDL_GL_SwapBuffers();
  }

  SDL_Quit();
}

int main(int argc, char **argv)
{
  int i;

  std::vector<const char *> filenames;
  const char * materials = "default.mtl";  
  const char * vrs = 0;

  VEE_SdlApp a;
  VEE_VrsApp va(&a);

  for(i = 1; i < argc; i++) {
    if(strcmp(argv[i], "--vrs") == 0 && i + 1 < argc)
      vrs = argv[++i];
    else if(strcmp(argv[i], "--obj") == 0 && i + 1 < argc)
      filenames.push_back(argv[++i]);
    else if(strcmp(argv[i], "--mtl") == 0 && i + 1 < argc)
      materials = argv[++i];
    else if(strcmp(argv[i], "--nolists") == 0)
      a.useLists(false);
    else if(strcmp(argv[i], "--maxframes") == 0)
      a.setMaxFrames(atoi(argv[++i]));
    else if(strcmp(argv[i], "--videowall") == 0)
      a.setAppType(VEE_SdlApp::TEXTURE_STREAM);
    else {
      VEE_error("%s # Undefined command line options \"%s\"", argv[0],
		argv[i]);
      return 0;
    }
  }

  puts("FOO");

  a.setMtlFile(materials);

  a.init();

  for(i = 0; i < (int) filenames.size(); i++) {

    GEE_World world;
    GEE_Group * g = new GEE_Group();
    
    world.addGeometry(g);
    
    if(!g->loadModel(filenames[i]))
      continue;

    g->convertPolygonsToTriangles();
    
    Solar::Model * model = new Solar::Model();

    GEE_ToSolar::convert(model, g);
    
    Solar::Object * solarobj = new Solar::Object();
      
    solarobj->setModel(model);

    model->normalize(20.0, 1);
    
    a.addModel(model);
    a.addObject(solarobj);
  }

  VEE_Qt::initImageIO();

  puts("FOO2");

  if(vrs) {
    if(va.loadVrConfig(vrs))
      va.exec(false);
  }
  else
    runSDL(a);

  return 0;
}



