/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <vee_sdl_app.h>

#include <vee_error.h>
#include <vee_image.h>
#include <vee_math.h>

#include <Solar/Object.hpp>

#include <GL/gl.h>


#define FRAMES 20
#define XS 320
#define YS 240
#define SIZE (XSa*YS)
#define STEP (2.0*M_PI/14.0)

unsigned char *frame[FRAMES];

unsigned char *loadImage(char *path,int *width,int *height)
{
  VEE_Image im;

  bool ok = im.load(path);

  *width  = im.w();
  *height = im.h();

  if(!ok) {
    VEE_error("loadImage # No file %s", path);
    return 0;
  }

  uint bytes = im.bytesPerPixel() * im.w() * im.h();
  uchar * buf = (uchar *) malloc(bytes);
  
  memcpy(buf, im.pixel(0, 0), bytes);

  return buf;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

void initGlLights()
{
  static GLfloat lightLocation[] = { 5.0, 5.0, 5.0, 0.0 };
  static GLfloat material[]  = { 1.0, 1.0, 1.0, 1.0 };
  static GLfloat shininess[] = { 127.0 };
  static GLfloat ambientLight[]  = { 0.3, 0.3, 0.3, 1.0 };
  static GLfloat diffuseLight[]  = { 0.7, 0.7, 0.7, 1.0 };
  static GLfloat specularLight[] = { 0.6, 0.6, 0.6, 1.0 };
  
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, material);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shininess);

  glLightfv (GL_LIGHT0, GL_AMBIENT, ambientLight);
  glLightfv (GL_LIGHT0, GL_DIFFUSE, diffuseLight);
  glLightfv (GL_LIGHT0, GL_SPECULAR, specularLight);
  glLightfv (GL_LIGHT0, GL_POSITION, lightLocation);

  glLightModelfv (GL_LIGHT_MODEL_AMBIENT, ambientLight);
}

void context_init(void)
{
  glPixelStorei(GL_UNPACK_ALIGNMENT,1);
  glBindTexture(GL_TEXTURE_2D,1);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);

  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);

  glTexImage2D(GL_TEXTURE_2D,0,GL_RGB,512,512,0,GL_RGB,GL_UNSIGNED_BYTE,NULL);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

void VEE_SdlApp::GlData::initGl(const std::string & mtlFile, bool useLists)
{
  VEE_TestQt::loadMaterials(mtlFile, & m_textures, & m_materials);

  if(useLists)
    m_renderer.setDisplayLists( & m_lists);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

VEE_SdlApp::VEE_SdlApp() 
  : m_withLists(true), 
    m_frame(0), 
    m_maxFrames(1000000), 
    m_maxTime(90.0f),
    m_type(OBJECT_DISPLAY)
{
  m_constructionTime.getTimeOfDay();
}

VEE_SdlApp::~VEE_SdlApp()
{
  DI_Time now;
  now.getTimeOfDay();

  float initTime = m_constructionTime.secsTo(m_startTime);
  float runTime  = m_startTime.secsTo(now);

  printf("INIT time = %.2f seconds\n"
	 "RUN time  = %.2f seconds\n"
	 "frames    = %u \n"
	 "fps       = %.2f Hz\n", 
	 initTime, runTime, m_frame, (m_frame - 2) / runTime);
  fflush(0);
}

void VEE_SdlApp::init()
{
  if(m_type == TEXTURE_STREAM) {
    int   n,dummy;
    char  name[]="00xx.jpg";
    
    for(n=0;n<FRAMES;n++) {
      name[2]=n/10+'0';
      name[3]=n%10+'0';
      frame[n]=loadImage(name,&dummy,&dummy);
      if(frame[n]==NULL)
	return;
    }
  }
}

void VEE_SdlApp::initGl(GlData * gld)
{
  if(m_type == OBJECT_DISPLAY)
    gld->initGl(m_mtlFile, m_withLists);
  else if(m_type == TEXTURE_STREAM)
    context_init();
}

void VEE_SdlApp::update()
{
  m_frame++;

  if(m_frame == 2)
    m_startTime.getTimeOfDay();
}

void VEE_SdlApp::draw(GlData * gld)
{
  int steps = 256;
  
  if(m_type == TEXTURE_STREAM) {
    steps = 5000;
  }

  float phase = 360.0f * (m_frame % steps) / (float) steps;
  glRotatef(phase, 0, 1, 0);    

  if(m_type == OBJECT_DISPLAY) {
    Solar::GlRenderer * r = & gld->m_renderer;
    
    initGlLights();
    
    
    Solar::Object ** obj = &* m_objects.begin();
    Solar::Object ** sentinel = &* m_objects.end();
    
    glEnable (GL_LIGHTING);
    glEnable (GL_LIGHT0);
    r->initGlState();
    
    while(obj != sentinel) {
      (*obj)->update(0.01);
      (*obj)->render(r);
      obj++;
    }
  }
  else if(m_type == TEXTURE_STREAM) {
    double y,x,px1,px2,pz1,pz2;
    
    glEnable(GL_DEPTH_TEST);

    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D,1);
    glTexSubImage2D(GL_TEXTURE_2D,0, 0,0, 320, 240, GL_RGB,GL_UNSIGNED_BYTE,
		    frame[(m_frame / 10) % FRAMES]);

    for(y=0;y<6;y++) {
      glPushMatrix();
      glTranslatef(0,y-3,0);
    
      glBegin(GL_QUADS);
      for(x=0;x<2*M_PI;x+=STEP) {
	px1=-sin(x)*3;
	px2=-sin(x+STEP)*3;
	pz1=-cos(x)*3-2;
	pz2=-cos(x+STEP)*3-2;

	glTexCoord2f(0,0);
	glVertex3f(px1,1,pz1);

	glTexCoord2f(XS/512.0,0);
	glVertex3f(px2,1,pz2);

	glTexCoord2f(XS/512.0,YS/512.0);
	glVertex3f(px2,0,pz2);

	glTexCoord2f(0,YS/512.0);
	glVertex3f(px1,0,pz1);
      }
      glEnd();
      glPopMatrix();
    }
  }
  else {
    printf("VEE_SdlApp::draw # Bad test type %d\n", (int) m_type);
    fflush(0);
    exit(0);
  }
}

bool VEE_SdlApp::isDone() 
{ 
  // puts("VEE_SdlApp::isDone");

  if(m_frame >= m_maxFrames)
    return true;

  if(m_frame < 10)
    return false;

  DI_Time now;
  now.getTimeOfDay();

  if(m_startTime.secsTo(now) > m_maxTime)
    return true;

  // puts("NOT DONE");

  return false;
}
