/* COPYRIGHT
 *
 * This file is part of the Simple Object Loader and Renderer - SOLAR
 *
 * Read the file "SOLAR-LICENSE" for the license.
 *
 * Authors & Copyright:
 *
 * Tommi Ilmonen, 2001-2003, except LWO-code by RJ, 2001-2003
 *
 * Tommi's email: firstname.lastname@hut.fi
 *
 */

#ifndef SOLAR_GRAPHIC_TRIANGLES_STRIPS_HPP
#define SOLAR_GRAPHIC_TRIANGLES_STRIPS_HPP

#include <Fluid/Vector2.hpp>

#include <Solar/BoundingBox.hpp>
#include <Solar/Graphic.hpp>

#include <vector>

namespace Solar {

  /** A collection of triangle strips. 
      
  A Usual way to create triangle strips is by:

  \code

  GraphicTriangleStrips * strips = new GraphicTriangleStrips();

  for(int s=0; s < myStripCount; s++) {

    for(int i=0; i < nmyvertices; i++)
      strips->addVertex(getMyVertex());

    strips->finalizeStrip();
  }

  \endcode

  */

class GraphicTriangleStrips : public Graphic
{
public:
  class Vertex {
  public:

    Fluid::Vector3 m_location;
    Fluid::Vector3 m_normal;
    Fluid::Vector2 m_uv;

    void clear() { m_location.clear(); m_normal.clear(); m_uv.clear(); }
  };

  class Strip {
  public:
    uint m_begin;
    /// The number of vertices
    uint m_length; 
  };


  GraphicTriangleStrips();
  virtual ~GraphicTriangleStrips();

  virtual Type type() const;
  /// The bounding box of the object in local coordinates.
  virtual void getBoundingBox(BoundingBox *bbox);
  virtual int nofPrimitives() const;
  
  virtual void render(Renderer *);
  virtual void scale(const Fluid::Vector3 &);
  virtual void move(const Fluid::Vector3 &);
  virtual void transform(const Fluid::Matrix3 &);
  
  virtual void write(std::ostream *);

  void addVertex(const Vertex &v);
  void addVertex(const Fluid::Vector3 & location,
		 const Fluid::Vector3 & normal,
		 const Fluid::Vector2 & uv)
  {
    Vertex v; 
    v.m_location = location; 
    v.m_normal = normal; 
    v.m_uv = uv; 
    addVertex(v);
  }

  unsigned nofVertices() const { return m_vertices.size(); }

  Vertex & vertex(unsigned n) { return m_vertices[n]; }
  const Vertex & vertex(unsigned n) const { return m_vertices[n]; }

  Vertex & vertex(unsigned strip, unsigned vertex) 
  { return m_vertices[m_strips[strip].m_begin + vertex]; }
  
  unsigned nofStrips() const { return m_strips.size(); }
  Strip & strip(unsigned n) { return m_strips[n]; }

  void finalizeStrip();

  void clear();

protected:

  std::vector<Vertex> m_vertices;
  std::vector<Strip>  m_strips;

  BoundingBox m_boundingBox;

};  

} // namespace Solar

#endif
