
/* COPYRIGHT
 *
 * This file is part of the Simple Object Loader and Renderer - SOLAR
 *
 * Read the file "SOLAR-LICENSE" for the license.
 *
 * Authors & Copyright:
 *
 * Tommi Ilmonen, 2001-2003, except LWO-code by RJ, 2001-2003
 *
 * Tommi's email: firstname.lastname@hut.fi
 *
 */

#include <Solar/GraphicTriangles.hpp>

#include <Solar/Renderer.hpp>
#include <Solar/Stream.hpp>

#include <iostream>

namespace Solar {

void GraphicTriangles::Triangle::clear()
{
  for(uint i=0; i < 3; i++) {
    m_vertices[i].m_vertex.clear();
    m_vertices[i].m_normal.clear();
    m_vertices[i].m_uv.clear();
  }
}

void GraphicTriangles::Triangle::write(std::ostream *s)
{
  uint i;

  (*s) << "TRIANGLE ";

  for(i=0; i < 3; i++) {
    (*s) << "p[" << i << "] = ";
    Write(s, m_vertices[i].m_vertex);
  }

  (*s) << "n = ";
  
  for(i=0; i < 3; i++) {
    (*s) << "n[" << i << "] = ";
    Write(s, m_vertices[i].m_normal);
  }

  (*s) << "u = ";

  for(i=0; i < 3; i++) {
    (*s) << "u[" << i << "] = ";
    Write(s, m_vertices[i].m_uv);
  }

  // << m_normals << m_uv << "\n";
}
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

GraphicTriangles::GraphicTriangles()
{
  clear();
}

GraphicTriangles::~GraphicTriangles()
{
  clear();
}

Graphic::Type GraphicTriangles::type() const
{ 
  return Graphic::Triangles;
}

void GraphicTriangles::getBoundingBox(BoundingBox *bbox)
{
  (*bbox) = m_boundingBox;
}

int GraphicTriangles::nofPrimitives() const
{
  return m_triangles.size();
}

void GraphicTriangles::render(Renderer *renderer)
{
  renderer->renderTriangles(this);
}


void GraphicTriangles::scale(const Fluid::Vector3 &v)
{
  for(uint i = m_triangles.size(); i--; )
    m_triangles[i].scale(v);

  m_boundingBox.scale(v);
}

void GraphicTriangles::transform(const Fluid::Matrix3 &m)
{
  for(uint i = m_triangles.size(); i--; )
    m_triangles[i].transform(m);

  recalculateBoundingBox();
}

void GraphicTriangles::move(const Fluid::Vector3 &v)
{
  for(uint i = m_triangles.size(); i--; )
    m_triangles[i].move(v);

  m_boundingBox.move(v);
}

void GraphicTriangles::write(std::ostream *s)
{
  (*s) << "TRIANGLES\n";

  Triangle *tri = &* m_triangles.begin();
  Triangle *sentinel = &* m_triangles.end();

  uint i=0;

  while(tri != sentinel) {
    (*s) << i++ << "   "; 
    tri->write(s);
    (*s) << "\n";
    tri++;
  }
}

void GraphicTriangles::clear()
{
  m_triangles.resize(0);

  m_boundingBox.clear();
}

void GraphicTriangles::recalculateBoundingBox()
{
  if(!m_triangles.size()) 
    m_boundingBox.clear();
  else {
    m_boundingBox.clear(m_triangles[0].m_vertices[0].m_vertex);
    
    for(uint i=0; i < m_triangles.size(); i++) {

      const Triangle &tri = m_triangles[i];

      for(uint j=0; j < 3; j++)
	m_boundingBox.expand(tri.m_vertices[i].m_vertex);
    }
  }
}

void GraphicTriangles::addTriangle(const Triangle &tri)
{
  if(!m_triangles.size()) {
    m_boundingBox.clear(tri.m_vertices[0].m_vertex);
    m_boundingBox.expand(tri.m_vertices[1].m_vertex);
    m_boundingBox.expand(tri.m_vertices[2].m_vertex);
  }
  else {
    for(uint i=0; i < 3; i++) 
      m_boundingBox.expand(tri.m_vertices[i].m_vertex);
  }

  m_triangles.push_back(tri);
}

} // namespace Solar
