/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Simple Object Loader and Renderer - SOLAR
 *
 * Read the file "SOLAR-LICENSE" for the license.
 *
 * Authors & Copyright:
 *
 * Tommi Ilmonen, 2001-2003, except LWO-code by RJ, 2001-2003
 *
 * Tommi's email: firstname.lastname@hut.fi
 *
 */

#ifndef SOLAR_MODEL_H
#define SOLAR_MODEL_H

#include <Solar/BoundingBox.hpp>
#include <Solar/Garbage.hpp>

#include <Fluid/Matrix3.hpp>

#include <vector>
#include <string>

namespace Solar {

  class Graphic;
  class Renderer;

  /** Visual (solid) model. A model is composed of several
      sub-objects. These sub-objects are derived from Solar::Graphic
      class. The sub-objects are of the kind that can be easily rendered
      on any simple, but sane graphics system (triangle strips, triangle
      arrays etc). 

      @author Tommi Ilmonen
  */

  class Model : public Garbage
  {
  public:
    Model();
    ~Model();

    virtual void setGarbageBin(GarbageBin *bin);

    void render(Renderer *);
    void scale(const Fluid::Vector3 &);
    void scale(float s) { scale(Fluid::Vector3(s, s, s)); }
    void move(const Fluid::Vector3 &);
    void transform(const Fluid::Matrix3 &);

    void addGraphic(Graphic *);

    void clear();

    void merge(Model *);
  
    /// Returns the number of sub-objects
    uint subObjectCount() const { return m_graphics.size(); }

    /// Gets one of the sub-objects
    Graphic *getSubObject(uint i) { return m_graphics[i]; }

    void calculateBoundingBox();

    const BoundingBox & boundingBox() const { return m_boundingBox; }

    void normalize(float maximumSpan, bool center);

    void centralize();

    int nofPrimitives() const;


    void createPyramid(float size, const std::string &baseMaterial, 
		       bool shareMaterials = false);
    void createCube(float size, const std::string &baseMaterial,
		    bool shareMaterials = false);

    void createFromHeightMap(const unsigned char *map, 
			     unsigned w, unsigned h,
			     const char * material = "ground");
  
  protected:
    std::vector<Graphic *> m_graphics;

    BoundingBox m_boundingBox;

  private:
    Model(const Model &) {}
    Model & operator = (const Model &) { return *this; }
  };

} // namespace Solar

#endif
