/* COPYRIGHT
 *
 * This file is part of the Simple Object Loader and Renderer - SOLAR
 *
 * Read the file "SOLAR-LICENSE" for the license.
 *
 * Authors & Copyright:
 *
 * Tommi Ilmonen, 2001-2003, except LWO-code by RJ, 2001-2003
 *
 * Tommi's email: firstname.lastname@hut.fi
 *
 */

#include <Solar/GlMaterial.hpp>

#include <Solar/Error.hpp>
#include <Solar/GlRenderer.hpp>

namespace Solar {

  GlMaterial::GlMaterial(float shine, 
			 const Rgba & color,
			 GLuint textureId,
			 int flags)
    : m_textureId(textureId),
      m_shine(shine),
      m_ambient(color), 
      m_diffuse(color), 
      m_specular(color),
      m_uvScale(1.0f),
      m_flags(flags)
  {}

  GlMaterial::GlMaterial()
    : m_textureId((GLuint)-1),
      m_shine(0.1),
      m_ambient(Rgba::Gray), 
      m_diffuse(Rgba::Gray), 
      m_specular(Rgba::Gray),
      m_uvScale(1.0f),
      m_flags(LIGHTING | SMOOTH)
  {}

  GlMaterial::~GlMaterial()
  {}

  void GlMaterial::applyGl(GlRenderer *renderer) const
  {
    renderer->applyGlLighting(m_flags & LIGHTING);
    renderer->applyGlBlend(m_flags & BLEND);
    renderer->applyGlSmooth(m_flags & SMOOTH);
    renderer->applyGlTwoSided(m_flags & TWO_SIDED);
    renderer->applyGlTexture2D(m_textureId);
  
    // glColor4fv(m_color.data());
    float shine = m_shine * 128.0f;
    if(shine > 128.0f) shine = 128.0f;
    glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, & shine);
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, m_specular.data());
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, m_ambient.data());
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, m_diffuse.data());
  }

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  /** The default material. Usually just fairly normal gray. */
  GlMaterial GlMaterialBank::m_default
  (70.0, Rgba(0.7, 0.7, 0.7, 1.0));

  GlMaterialBank::GlMaterialBank()
  {}

  GlMaterialBank::~GlMaterialBank()
  {}

  void GlMaterialBank::addMaterial(const std::string &name, 
				   const GlMaterial &material)
  {
    // trace("GlMaterialBank::addMaterial # \"%s\"", name.c_str());
    m_materials[name] = material;
  }

  /** Gets a material by name. If the material is not found, then the
      default material is returned. */

  const GlMaterial & GlMaterialBank::getMaterial(const std::string &s)
  {
    const std::string *s2 = m_forceMaterial.size() ? & m_forceMaterial : & s;

    std::map<std::string, GlMaterial>::iterator it = m_materials.find(*s2);

    if(it != m_materials.end()) {
      // trace("GlMaterialBank::getMaterial # \"%s\"", s2->c_str());
      return (*it).second;
    }

    return m_default;
  }

  void GlMaterialBank::addUsualColors()
  {
    addMaterial("black", GlMaterial(0.5, Rgba::Black));
    addMaterial("gray",  GlMaterial(0.5, Rgba::Gray));
    addMaterial("white", GlMaterial(0.5, Rgba::White));

    addMaterial("red",   GlMaterial(0.5, Rgba::Red));
    addMaterial("green", GlMaterial(0.5, Rgba::Green));
    addMaterial("blue",  GlMaterial(0.5, Rgba::Blue));
  }

} // namespace Solar
