/* COPYRIGHT
 *
 * This file is part of the Simple Object Loader and Renderer - SOLAR
 *
 * Read the file "SOLAR-LICENSE" for the license.
 *
 * Authors & Copyright:
 *
 * Tommi Ilmonen, 2001-2003, except LWO-code by RJ, 2001-2003
 *
 * Tommi's email: firstname.lastname@hut.fi
 *
 */

#ifndef SOLAR_GL_MATERIAL_HPP
#define SOLAR_GL_MATERIAL_HPP

#include <GL/gl.h>

#include <Solar/Color.hpp>

#include <map>
#include <string>

namespace Solar {

  class GlRenderer;

  /** Material definition for OpenGL. Includes color, texture and
      shine values. */
  class GlMaterial
  {
  public:

    enum {
      /// Use blending (use alpha channel)
      BLEND = 0x1,
      /// Use lighting
      LIGHTING = 0x2,
      /// Smooth shading
      SMOOTH = 0x4,
      /// Two sided object
      TWO_SIDED = 0x8,
      /// With mesh
      WITH_MESH = 0x10
    };

    GlMaterial(float shine, 
	       const Rgba & color,
	       GLuint textureId = (GLuint) -1,
	       int flags = LIGHTING | SMOOTH);
    GlMaterial();

    virtual ~GlMaterial();

    void set(float shine, const Rgba & color, GLuint textureId = (GLuint) -1)
    { m_textureId = textureId; m_shine = shine;  setColor(color); }

    void setShine(float s) { m_shine = s; }
    int  flags() const   { return m_flags; }
    void setFlags(int f) { m_flags = f; }
    void setTextureId(GLuint i) { m_textureId = i; }

    /// Sets all colors at once
    void setColor(const Rgba & c) { m_ambient = m_diffuse = m_specular = c; }
    void setAmbient(const Rgba & color)  { m_ambient = color; }
    void setDiffuse(const Rgba & color)  { m_diffuse = color; }
    void setSpecular(const Rgba & color) { m_specular = color; }

    void applyGl(GlRenderer *) const;

    GLuint   textureId()    const { return m_textureId; }

    const Rgba & ambient()  const { return m_ambient; }
    const Rgba & diffuse()  const { return m_diffuse; }
    const Rgba & specular() const { return m_specular; }

    float uvScale() const { return m_uvScale; }
    void  setUvScale(float uvScale) { m_uvScale = uvScale; }
  protected:

    GLuint   m_textureId;
    float    m_shine;
    Rgba     m_ambient, m_diffuse, m_specular;
    float    m_uvScale;

    int      m_flags;
  };

  /////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////

  /** Collection of materials.  */

  class GlMaterialBank
  {
  public:
    GlMaterialBank();
    virtual ~GlMaterialBank();

    void addMaterial(const std::string &, const GlMaterial &);

    const GlMaterial & getMaterial(const std::string &);
  
    void addUsualColors();

    void forceMaterial(const std::string & mat) { m_forceMaterial = mat; }
    void unforceMaterial() { m_forceMaterial.clear(); }

  protected:

    static GlMaterial m_default;

    std::map<std::string, GlMaterial> m_materials;

    std::string m_forceMaterial;
  };

} // namespace Solar

#endif
