/* COPYRIGHT
 *
 * This file is part of the Simple Object Loader and Renderer - SOLAR
 *
 * Read the file "SOLAR-LICENSE" for the license.
 *
 * Authors & Copyright:
 *
 * Tommi Ilmonen, 2001-2003, except LWO-code by RJ, 2001-2003
 *
 * Tommi's email: firstname.lastname@hut.fi
 *
 */

#include <Solar/GlRenderer.hpp>



#include <Solar/Error.hpp>
#include <Solar/GlDisplayLists.hpp>
#include <Solar/GraphicTriangles.hpp>
#include <Solar/GraphicTriangleMesh.hpp>
#include <Solar/GraphicTriangleStrips.hpp>
#include <Solar/GlMaterial.hpp>
#include <Solar/Model.hpp>

#include <GL/glu.h>

#define BUFFER_OFFSET(i) ((char *)NULL + (i))

namespace Solar {

  GlRenderer::GlRenderer(GlMaterialBank *materials)
    : m_materials(materials), 
      m_currentMaterial(0),
      m_displayLists(0),
      m_currentList(0),
      m_currentTexture((GLuint) -1),
      m_blend(false),
      m_lighting(true),
      m_smooth(true),
      m_twoSided(false),
      m_listRenderCount(0)
  {}

  GlRenderer::GlRenderer() 
    : m_materials(0), 
      m_currentMaterial(0),
      m_displayLists(0),
      m_currentList(0),
      m_currentTexture((GLuint) -1),
      m_blend(false),
      m_lighting(true),
      m_smooth(true),
      m_twoSided(false)
  {}

  void GlRenderer::begin(Model *)
  {
    if(!m_materials) 
      fatal("GlRenderer::begin #  m_materials undefined");
  }

  GlRenderer::~GlRenderer()
  {
    trace("GlRenderer::~GlRenderer # Rendered %ld lists", m_listRenderCount);
  }
  
  void GlRenderer::prepareToRender()
  {
    initGlState();
  }

  void GlRenderer::end()
  {}

  void GlRenderer::renderTriangles(GraphicTriangles *triangles)
  {
    preRender(triangles);

    bool twice = m_currentMaterial ? 
      m_currentMaterial->flags() & GlMaterial::WITH_MESH : false;

    ListMode m = twice ? NO_LIST : renderDisplayList(triangles);

    if(m == RENDERED_LIST)
      return;

    uint i, n = triangles->nofTriangles();

    /*
#ifdef HAVE_GL_BUFFERS_ARB

    // Test the use of vertex buffers. They do not seem to help at all.

    if(m == GENERATE_LIST) {
    
      puts("GlRenderer::renderTriangles # VBO");
    
      glEndList();
      
      m_currentList = & m_displayLists->getList(triangles);
      m_currentList->bufferId();
      
      glBindBufferARB(GL_ARRAY_BUFFER_ARB, m_currentList->bufferId());
      
      const void * data = (const void *) & triangles->triangle(0);
      uint bytes = triangles->bytes();
      
      glBufferDataARB(GL_ARRAY_BUFFER_ARB, bytes, data, GL_STATIC_DRAW_ARB);

      glNewList(m_currentList->listId(), GL_COMPILE_AND_EXECUTE);
      
      glBindBufferARB(GL_ARRAY_BUFFER_ARB, m_currentList->bufferId());

      glEnableClientState(GL_VERTEX_ARRAY);
      glEnableClientState(GL_NORMAL_ARRAY);
      glEnableClientState(GL_TEXTURE_COORD_ARRAY);
      
      glVertexPointer(3, GL_FLOAT, sizeof(GraphicTriangles::Vertex), 0);
      glNormalPointer(GL_FLOAT, sizeof(GraphicTriangles::Vertex), 
		      BUFFER_OFFSET(sizeof(VEE_Vector3)));
      glTexCoordPointer(2, GL_FLOAT, sizeof(GraphicTriangles::Vertex), 
			BUFFER_OFFSET(2 * sizeof(VEE_Vector3)));

      const uint maxArray = 2000;
      i = 0;
      uint left = n - i;

      while(left) {

	// Test how chunking affects the performance -> no effect.

	uint consume = left > maxArray ? maxArray : left;
	
	glDrawArrays(GL_TRIANGLES, i * 3, consume * 3);
	
	i += consume;

	left = n - i;
      }

      glDisableClientState(GL_VERTEX_ARRAY);
      glDisableClientState(GL_NORMAL_ARRAY);
      glDisableClientState(GL_TEXTURE_COORD_ARRAY);
      
      glEndList();
      
      return;
    }
#endif
    */

    glBegin(GL_TRIANGLES);

    for(i = 0; i < n; i++) {

      const GraphicTriangles::Triangle &t = triangles->triangle(i);
    
      for(uint j=0; j < 3; j++) {
	glTexCoord2fv(t.m_vertices[j].m_uv.data());
	glNormal3fv(t.m_vertices[j].m_normal.data());
	glVertex3fv(t.m_vertices[j].m_vertex.data());
      }
    }

    glEnd();

    if(twice) {

      // puts("OHO");

      // glDisable(GL_POLYGON_OFFSET_FILL);
      
      applyGlBlend(false);
      applyGlTexture2D(false); 
      
      glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

      VEE_Vector3 c;
      c.copy(m_currentMaterial->ambient().data());
      for(i=0; i < 3; i++)
	c[i] = c[i] > 0.5f ? 0.0f : 1.0f;
      
      glColor3fv(c.data());

      glLineWidth(2);

      glBegin(GL_TRIANGLES);

      for(i = 0; i < n; i++) {
	
	const GraphicTriangles::Triangle &t = triangles->triangle(i);
	
	for(uint j=0; j < 3; j++) {
	  // glNormal3fv(&t.m_normals[j][0]);
	  glVertex3fv( t.m_vertices[j].m_vertex.data());
	}
      }

      glEnd();

      glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    }

    if(m == GENERATE_LIST)
      glEndList();
  }

  void GlRenderer::renderTriangleMesh(GraphicTriangleMesh *mesh)
  {
    preRender(mesh);

    ListMode m = renderDisplayList(mesh);

    if(m == RENDERED_LIST)
      return;

    float scale = m_currentMaterial->uvScale();
    
    // trace("FOOFOFO %f", scale);
    
    const GraphicTriangleMesh::Triangle *tri = & mesh->triangle(0);
    const GraphicTriangleMesh::Triangle *sentinel = 
      tri + mesh->nofTriangles();
    
    glBegin(GL_TRIANGLES);
    
    while(tri != sentinel) {
      
      for(uint j=0; j < 3; j++) {
	const GraphicTriangleMesh::VertexData & v = 
	  mesh->vertex(tri->vertex(j));
	
	glNormal3fv(& v.normal()[0]);
	Fluid::Vector2 uv = v.uv() * scale;
	glTexCoord2f(uv.x, uv.y);
	glVertex3fv(& v.location()[0]);
      }
      
      tri++;
    }
    
    glEnd();

    if(m == GENERATE_LIST)
      glEndList();
  }

  void GlRenderer::renderTriangleStrips(GraphicTriangleStrips *strips)
  {
    preRender(strips);

    ListMode m = renderDisplayList(strips);

    if(m == RENDERED_LIST)
      return;

    uint nstrips = strips->nofStrips();

    for(uint s = 0; s < nstrips; s++) {

      glBegin(GL_TRIANGLE_STRIP);

      GraphicTriangleStrips::Strip stripData = strips->strip(s);
    
      for(uint i=0; i < stripData.m_length; i++) {

	const GraphicTriangleStrips::Vertex &v = 
	  strips->vertex(stripData.m_begin + i);
      
	glTexCoord2f(v.m_uv[0], v.m_uv[1]);
	glNormal3fv(v.m_normal.data());
	glVertex3fv(v.m_location.data());
      }
    
      glEnd();
    }

    if(m == GENERATE_LIST)
      glEndList();
  }

  /** Sets materials, textures, colors, transparencies etc. */

  void GlRenderer::preRender(Graphic *gr)
  {
    const GlMaterial & mat = m_materials->getMaterial(gr->material());

    GLenum e;

    e = glGetError();
    if(e) 
      error("GlRenderer::preRender # GL ERROR ALREADY \"%s\"", gluErrorString(e));

    /* trace("GlRenderer::preRender # \"%s\" [%.2f %.2f %.2f] %d", 
       gr->material().c_str(), 
       mat.color().red(), mat.color().green(), mat.color().blue(),
       (int) mat.textureId()); */

    m_currentMaterial = & mat;
    mat.applyGl(this);
  
    e = glGetError();
    if(e) 
      error("GlRenderer::preRender # GL ERROR \"%s\" (%s)", 
	    gluErrorString(e), gr->material().c_str());    
  }

  /** Checks if there is a diplay list for this graphic. If there is
      no list storage, then no list-related stuff is done and NO_LIST
      is returned. If the list exists, then it is rendered. If the the
      list does not exist, but there is a list storage, then its
      recording is began. */
  GlRenderer::ListMode GlRenderer::renderDisplayList(Graphic *g)
  {    
    m_currentList = 0;

    if(m_displayLists) {

      GlDisplayLists::List * l = & m_displayLists->getList(g);

      if(l->needsUpdate(g->generation())) {
	glNewList(l->listId(), GL_COMPILE_AND_EXECUTE);
	l->setGeneration(g->generation());
	trace("Create list for %p", g);
	m_currentList = l;
	return GENERATE_LIST;
      }
      else {
	glCallList(l->listId());
	m_listRenderCount++;
	return RENDERED_LIST;
      }
    }
    return NO_LIST;
  }

  void GlRenderer::pushTransform(const Fluid::Matrix3 &rot,
				 const Fluid::Vector3 &loc,
				 float scale)
  {  
    glPushMatrix();

    float oglMatrix[16];

    oglMatrix[0] = rot[0][0];
    oglMatrix[1] = rot[1][0];
    oglMatrix[2] = rot[2][0];
    oglMatrix[3] = 0.0f;

    oglMatrix[4] = rot[0][1];
    oglMatrix[5] = rot[1][1];
    oglMatrix[6] = rot[2][1];
    oglMatrix[7] = 0.0f;

    oglMatrix[8] = rot[0][2];
    oglMatrix[9] = rot[1][2];
    oglMatrix[10] = rot[2][2];
    oglMatrix[11] = 0.0f;

    oglMatrix[12] = loc[0];
    oglMatrix[13] = loc[1];
    oglMatrix[14] = loc[2];
    oglMatrix[15] = 1.0f;

    /*
      oglMatrix[0] = rot[0][0];
      oglMatrix[1] = rot[0][1];
      oglMatrix[2] = rot[0][2];
      oglMatrix[3] = 0.0f;

      oglMatrix[4] = rot[1][0];
      oglMatrix[5] = rot[1][1];
      oglMatrix[6] = rot[1][2];
      oglMatrix[7] = 0.0f;

      oglMatrix[8] = rot[2][0];
      oglMatrix[9] = rot[2][1];
      oglMatrix[10] = rot[2][2];
      oglMatrix[11] = 0.0f;

      oglMatrix[12] = loc[0];
      oglMatrix[13] = loc[1];
      oglMatrix[14] = loc[2];
      oglMatrix[15] = 1.0f;  
    */

    glMultMatrixf(oglMatrix);

    if(scale != 1.0f)
      glScalef(scale, scale, scale);

    // glTranslatef(loc[0], loc[1], loc[2]);  
  }

  void GlRenderer::popTransform()
  {
    glPopMatrix();
  }

  /** Initializes the OpenGL state machine for rendering the
      objects. This method needs to be called if the OpenGL state is
      wrong for the rendering. */
  void GlRenderer::initGlState()
  {
    GLenum e = glGetError();
    if(e) 
      error("GlRenderer::prepareGlMode # GL ERROR AT BEGIN %s", 
	    gluErrorString(e));

    glDisable(GL_COLOR_MATERIAL);
    glDisable(GL_BLEND);
    glDisable(GL_TEXTURE_2D);
    glEnable(GL_LIGHTING);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);
    glShadeModel(GL_SMOOTH);
    glDepthMask(GL_TRUE);
    glEnable(GL_NORMALIZE);

    m_currentTexture = (GLuint) -1;
    m_blend = false;
    m_lighting = true;
    m_smooth = true;
    m_twoSided = false;
  }

  /** This method switches OpenGL to a new 2D texture. If the texture is
      the same as the previous texture, then no operation is performed.

      @arg tex The OpenGL texture id. If the texture id equals -1, then
      this method disables textures.
  */
  void GlRenderer::applyGlTexture2D(GLuint tex)
  {
    if(tex == m_currentTexture) 
      return;

    // trace("GlRenderer::applyGlTexture2D # New texture = %d", (int) tex);
  
    if(tex == (GLuint) -1) {
      glDisable(GL_TEXTURE_2D);
    }
    else {
      if(m_currentTexture == (GLuint) -1)
	glEnable(GL_TEXTURE_2D);
    
      glBindTexture(GL_TEXTURE_2D, tex);
    }
    m_currentTexture = tex;
  }

  /** This method enables/disables blending of the OpenGL state
      machine. */
  void GlRenderer::applyGlBlend(bool blend)
  {
    if(blend == m_blend) return;

    if(blend) {
      glEnable(GL_BLEND);
      glDisable(GL_DEPTH_TEST);
      glDisable(GL_CULL_FACE);
    }
    else {
      glDisable(GL_BLEND);
      glEnable(GL_DEPTH_TEST);
      glEnable(GL_CULL_FACE);
    }

    m_blend = blend;
  }

  /** Switch OpenGL lights on/off. */

  void GlRenderer::applyGlLighting(bool lighting)
  {
    if(lighting == m_lighting) return;

    if(lighting)
      glEnable(GL_LIGHTING);
    else
      glDisable(GL_LIGHTING);
  
    m_lighting = lighting;
  }

  void GlRenderer::applyGlSmooth(bool smooth)
  {
    if(smooth == m_smooth) return;

    if(smooth)
      glShadeModel(GL_SMOOTH);
    else
      glShadeModel(GL_FLAT);

    m_smooth = smooth;
  }

  void GlRenderer::applyGlTwoSided(bool twoSided)
  {
    if(m_twoSided == twoSided) return;

    if(twoSided) 
      glDisable(GL_CULL_FACE);
    else
      glEnable(GL_CULL_FACE);

    m_twoSided = twoSided;
  }

} // namespace Solar
