/* COPYRIGHT
 *
 * This file is part of the Simple Object Loader and Renderer - SOLAR
 *
 * Read the file "SOLAR-LICENSE" for the license.
 *
 * Authors & Copyright:
 *
 * Tommi Ilmonen, 2001-2003, except LWO-code by RJ, 2001-2003
 *
 * Tommi's email: firstname.lastname@hut.fi
 *
 */

#include <Solar/PfConvert.hpp>

#include <Solar/Error.hpp>

#include <Solar/GraphicTriangles.hpp>
#include <Solar/GraphicTriangleStrips.hpp>
#include <Solar/Model.hpp>

#include <assert.h>

extern "C"
{
  /*#define PF_C_API 1
    #define PF_CPLUSPLUS_API 0 */
#include <Performer/pfdu.h>
#include <Performer/pr.h>
#include <Performer/prmath.h>
}

#include <iostream>

namespace Solar {

  bool PfConvert::convert(Solar::Model *model, pfGroup *p)
  {
    const char * fname = "PfConvert::convert";

    int children = pfGetNumChildren(p);

    trace("%s # Reading %d children", fname, children);
    
    for(int child = 0; child < children; child++) {
      // while(pfGetNumChildren(p)) {
      
      pfNode *node = children > 1 ? pfGetChild(p,child) : (pfNode *) p;

      if(node == 0) break;

      const char *name = pfGetNodeName(node);
      trace("%s # Node: \"%s\"", fname, name);
      
      if(pfIsDerivedFrom(pfGetType(node),pfGetGroupClassType()))
	convert(model,(pfGroup*)node);
      else if(pfIsDerivedFrom(pfGetType(node),pfGetGeodeClassType())) {

	pfGeode *g = (pfGeode*) node;
	for(int j = 0; j < pfGetNumGSets(g); j++) {

	  pfGeoSet *gset = pfGetGSet(g,j);
	  pfGeoState *gstate = pfGetGSetGState(gset);
	  pfRef(gstate); // to prevent Performer from deallocating the geostate

	  

	  int primType = pfGetGSetPrimType(gset);

	  pfVec3 *vertices = 0;
	  pfVec3 *normals = 0;
	  pfVec2 *texcoord2 = 0;
	  ushort *indices = 0;

	  pfGetGSetAttrLists(gset,PFGS_COORD3,(void**)&vertices, &indices);
	  pfGetGSetAttrLists(gset,PFGS_NORMAL3,(void**)&normals, &indices);
	  //assert(indices == 0);
	  
	  if(indices) {
	    error("%s # indexed attribute-lists are not supported", 
		  fname);
	    continue;
	  }
	  if(!normals) {
	    error("%s # Normals undefined", fname);
	  }

	  pfGetGSetAttrLists(gset,PFGS_TEXCOORD2, 
			     (void **)&texcoord2,&indices); 
	  assert(indices == 0);
	  
	  using namespace std;
	  switch(primType)
	    {
	    case PFGS_POINTS:
	      cout << "Points\n";
	      assert(0);
	      break;
	    case PFGS_LINES:
	      cout << "Lines\n";
	      assert(0);
	      break;
	    case PFGS_LINESTRIPS:
	      cout << "Linestrips\n";
	      assert(0);
	      break;
	    case PFGS_FLAT_LINESTRIPS:
	      cout << "Flat Linestrips\n";
	      assert(0);
	      break;
	    case PFGS_TRIS:
	      {
		int triCount = pfGetGSetNumPrims(gset);
		cout << "Triangles: " << triCount << std::endl;

		GraphicTriangles * tris = new GraphicTriangles();

		GraphicTriangles::Triangle tri;
		tri.clear();

		for(int i = 0; i < triCount; i++) {
		  for(int j = 0; j < 3; j++) {
		    tri.m_vertices[j].make(vertices[3*i+j][0],vertices[3*i+j][1],vertices[3*i+j][2]);
		    if(texcoord2)
		      tri.m_uv[j].make(texcoord2[3*i+j][0],texcoord2[3*i+j][1]);
		    if(normals)
		      tri.m_normals[j].make(normals[3*i+j][0],normals[3*i+j][1],normals[3*i+j][2]);
		  }
		  tris->addTriangle(tri);
		}
		
		model->addGraphic(tris);
	      }
	      break;
	    case PFGS_QUADS:
	      {
		error("%s # Quads not supported (yet)", fname);
		/* int quadCount = pfGetGSetNumPrims(gset);
		cout << "Quads: " << quadCount << std::endl;
		int baseIndex = mesh->getVertexCount();
		
		GraphicTriangles * tris = new GraphicTriangles();

		GraphicTriangles::Triangle tri;
		tri.clear();

		for(int i = 0; i < quadCount; i++) {
		  for(int j = 0; j < 4; j++) {
			mesh->addVertex(Vector3(vertices[4*i+j][0],vertices[4*i+j][1],vertices[4*i+j][2]));
			if(texcoord2)
			  mesh->addTexCoord2(Vector2(texcoord2[4*i+j][0],texcoord2[4*i+j][1]));
			else mesh->addTexCoord2(Vector2(0.0,0.0));
		      }
		    mesh->addTriangle(Vector3i(baseIndex+4*i,baseIndex+4*i+1,baseIndex+4*i+2),gstate);
		    mesh->addTriangle(Vector3i(baseIndex+4*i,baseIndex+4*i+2,baseIndex+4*i+3),gstate);
		    }*/
	      }
	      break;
	    case PFGS_FLAT_TRISTRIPS:
	    case PFGS_TRISTRIPS:
	      {
		int stripcount = pfGetGSetNumPrims(gset);
		cout << ((primType == PFGS_FLAT_TRISTRIPS) ? 
			 "Flat tristrips " : "Smooth tristrips ")
		     << stripcount << std::endl;
		int *lengths = pfGetGSetPrimLengths(gset);
		std::cout << "Queried vertices" << std::endl;
		int idx = 0;
		int idxn = 0;

		GraphicTriangleStrips::Vertex v;
		v.clear();

		GraphicTriangleStrips * tristrips = 
		  new GraphicTriangleStrips();
		  
		for(int i = 0; i < stripcount; i++) {

		  for(int j = 0; j < lengths[i]; j++) {

		    v.m_location.make(vertices[idx][0],
				      vertices[idx][1],
				      vertices[idx][2]);

		    if(texcoord2)
		      v.m_uv.make(texcoord2[idx][0],
				  texcoord2[idx][1]);

		    if(normals) {
		      v.m_normal.make(normals[idxn][0],
				      normals[idxn][1],
				      normals[idxn][2]);
		      if(primType == PFGS_TRISTRIPS || j >= 2)
			idxn++;
		    }
		    tristrips->addVertex(v);
		    
		    idx++;
		  }
		  tristrips->finalizeStrip();
		}
		model->addGraphic(tristrips);
	      }
	      break;
	    case PFGS_TRIFANS:
	      cout << "Trifans\n";
	      assert(0);
	      break;
	    case PFGS_FLAT_TRIFANS:
	      cout << "Flat trifans\n";
	      assert(0);
	      break;
	    case PFGS_POLYS:
	      cout << "Polygons\n";
	      assert(0);
	      break;
	    default:
	      std::cout << "Panic - unknown primitive" << std::endl;
	    }
	}
      }
      /* if(children > 1)
	 pfRemoveChild(p,node); */
    } 
    
    return true;
  }
  

  
  bool PfConvert::load(Solar::Model *m, const char *filename)
  {

    pfdInitConverter(filename);
    
    if(!pfIsConfiged())
      pfConfig();

    pfNode *objNode = pfdLoadFile(filename);

    if(!objNode)
      return false;

    pfFlatten(objNode,0);
    
    bool ok =  convert(m, (pfGroup*) objNode);

    pfFree(objNode);

    return ok;
  }
   
}
