/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <gee_debug.h>
#include <gee_engine.h>

#include <vee_blow.h>
#include <vee_color.h>
#include <vee_damping.h>
#include <vee_error.h>
#include <vee_force_plane.h>
#include <vee_gl_renderer.h>
#include <vee_gravitate.h>
#include <vee_gravity.h>
#include <vee_looks_color_table.h>
#include <vee_magic_rain.h>
#include <vee_math.h>
#include <vee_meta_balls.h>
#include <vee_emitter.h>
#include <vee_polygon.h>
// #include <vee_polygon_clouds.h>
#include <vee_polygon_fire.h>
#include <vee_scene.h>
#include <vee_volume_box.h>
#include <vee_volume_ellipsoid.h>
#include <vee_strs.h>
#include <vee_vortex.h>
// #include <vee_vortex_effect.h>
#include <vee_test_qt.h>
#include <vee_value_operator.h>
#include <vee_value_random.h>
#include <vee_wind.h>
#include <vee_values.h>
#include <vee_volume_grid.h>

#include <Solar/Loaders.hpp>
#include <Solar/Model.hpp>
#include <Solar/RotatingObject.hpp>

#include <qapplication.h>

#include <stdio.h>

static void addGravity(VEE_SceneBox *box, float force)
{
  VEE_Gravity *grav = new VEE_Gravity(VEE_Vector3(0.0f, -force, 0.0f));
  box->addNewForce(grav);
}

static void addRain(VEE_SceneBox *box, float rate, float y)
{
  VEE_MagicRain *rain = 
    new VEE_MagicRain(VEE_Vector3(0.0, y - 10.0, 0.0), 
		      VEE_Vector3(0.0, 0.0, 0.0), 
		      1e+10, 
		      5.0,
		      0.0,
		      rate);
  box->addNewObject(rain);
}

/* static void addVortex(VEE_SceneBox *box, float magnitude)
{
  VEE_Vortex *swrl = new VEE_Vortex();
  swrl->setRange(5.0);
  swrl->setAxes(1.0, 0.0, 0.0);
  box->addNewForce(swrl);
  }*/

/* static void addVortexeff(VEE_SceneBox *box, float magnitude)
{
  VEE_VortexEffect *swrleff =
    new VEE_VortexEffect(VEE_Vector3(0.0, 2.0, 4.0),
			VEE_Vector3(0.0, 0.0, 0.0),
			magnitude, 10000.0, 5.0, 1.0);
  swrleff->setAxes(0.0, 1.0, 0.0);
  box->addNewForce(swrleff);
  }*/


#if 0
static void addBlow(VEE_SceneBox *box)
{
  VEE_Matrix3 blm;

  VEE_Blow *bl = new VEE_Blow(VEE_Vector3(-3.0, -10.0, 0.0),
			      12.0, 9.0, 10000.0);
  blm.rotateZ(VEE_HALF_PI);
  bl->setRotation(blm);
  box->addNewObject(bl);

  /* VEE_Blow *bl = new VEE_Blow(VEE_Vector3(3.0, -10.0, 0.0),
			       12.0, 9.0, 10000.0);
  blm.FromEulerAnglesXYZ(0.0, 0.1 * VEE_PI, 0.2 * VEE_PI);
  bl->setRotation(blm);
  box->addNewObject(bl); */

  bl = new VEE_Blow(VEE_Vector3(0.0, -5.0, 0.0),
		    12.0, 9.0, 10000.0);
  bl->setDirection(VEE_Vector3(1.0, 1.0, 0.0));
  box->addNewObject(bl);
}
#endif

static void addPolygon(VEE_SceneBox *box, int div, float size)
{
  static VEE_RgbaTable colorTable;

  for(uint i=0; i < 20; i++) {
    colorTable.appendColor(VEE_Rgba(1.0, 0.5, 0.5, 0.6));
    colorTable.appendColor(VEE_Rgba(0.7, 0.7, 0.0, 0.6));
  }
  colorTable.appendColor(VEE_Rgba::Black);


  // VEE_Polygon *poly = VEE_Polygon::createNewMesh(div, div, size, size);
  
  VEE_Polygon *poly = VEE_Polygon::createNewQuadMesh(div, div, size, size);

  poly->setLooks(new VEE_LooksColorTable(&colorTable));
  poly->setStiffness(3.0);
  poly->setAgeLimit(220.0);

  printf("%d %d\n", poly->getMeshColumns(), poly->getMeshRows());
  box->addNewObject(poly);
}

static void addDamping(VEE_SceneBox *box, float magnitude)
{
  float radius = 100.0;

  VEE_VolumeEllipsoid *se = 
    new VEE_VolumeEllipsoid(VEE_Vector3(radius, radius, radius));

  VEE_Damping *d = new VEE_Damping(se, magnitude, 1e+10);
  
  box->addNewObject(d);
}

static void addPolyFire(VEE_SceneBox *box, float flameHz, float flameAge,
			float vortexHz, float gustHz)
{
  VEE_PolygonFire *pf = 
    new VEE_PolygonFire(flameHz, flameAge, vortexHz, gustHz);

  addGravity(box, -4.0);
  box->addNewObject(pf);
}

/* static void addPolyClouds(VEE_SceneBox *box, float interval, float magnitude)
{
  VEE_PolygonClouds *pc = 
    new VEE_PolygonClouds(VEE_Vector3(0.0, 5.0, 0.0),
			  VEE_Vector3ZERO,
			  5.0, 
			  300,
			  interval,
			  magnitude);

  box->addNewObject(pc);
  }*/

static void addWind(VEE_SceneBox *box, float velocity)
{
  VEE_Vector3 direction(2.0, 0.0, -1.0);

  direction.normalize();

  VEE_Wind * wind = new VEE_Wind(direction * velocity);
  box->addNewObject(wind);

  VEE_MetaBalls mb;
  mb.addBall(VEE_Vector3(-1.0, 0.0, 0.0), 1.0);
  mb.addBall(VEE_Vector3(1.5, -1.0, 0.0), 1.0);
  mb.addBall(VEE_Vector3(0.0,  0.0, 0.0), 1.3);
  mb.addBall(VEE_Vector3(1.5, -1.0, 1.0), 1.0);
  mb.addBall(VEE_Vector3(1.8, -1.0, 0.0), 1.0);

  VEE_VolumeGrid * vg = new VEE_VolumeGrid();

  vg->createFromMetaBalls(& mb, 40);

  VEE_ValueVolume vv(vg);

  wind->setValue("volume", &vv);
}

static void addCreator2(VEE_SceneBox *box)
{
  VEE_Particle *p = new VEE_Particle;
  p->setAgeLimit(3.0);

  VEE_Emitter *creator = new VEE_Emitter(p);
  
  creator->appendOperator
    (new VEE_ValueOperator(VEE_Strs::addvelocity, new VEE_ValueRandomVector3));

  creator->appendOperator
    (new VEE_ValueOperator(VEE_Strs::addlocation, new VEE_ValueRandomVector3));

  creator->appendOperator
    (new VEE_ValueOperator(VEE_Strs::mass, 
			   new VEE_ValueRandomFloat(0.5, 1.0)));

  box->addNewObject(creator);
}

static void addGravitates(VEE_SceneBox *box, float rate, float range, 
			  float magnitude)
{
  VEE_Gravitate *g = new VEE_Gravitate(VEE_Vector3ZERO,
				       VEE_Vector3ZERO,
				       3.0f,
				       0.5,
				       true);
  
  
  VEE_Emitter *creator = new VEE_Emitter(g);
  
  creator->setRate(rate);

  VEE_ValueRandomVector3 *rv = new VEE_ValueRandomVector3;
  
  rv->setRanges(VEE_Vector3(-5.0, 3.0, -5.0), VEE_Vector3(5.0, 3.5, 5.0));

  creator->appendOperator
    (new VEE_ValueOperator(VEE_Strs::addlocation, rv->clone()));

  rv->setRanges(VEE_Vector3(-3.0, 0.0, -3.0), VEE_Vector3(3.0, 0, 3.0));

  creator->appendOperator
    (new VEE_ValueOperator(VEE_Strs::addvelocity, rv));

  creator->appendOperator
    (new VEE_ValueOperator(VEE_Strs::range, range));

  creator->appendOperator
    (new VEE_ValueOperator(VEE_Strs::magnitude, magnitude));

  box->addNewObject(creator);
}

static void defaultSystem(VEE_SceneBox *box)
{
  addRain(box, 5.0, 15.0);
  addGravity(box, 3.0);
}

static bool isNewCommand(const char *str)
{
  if(str[0] == '-')
    if(str[1] == '-')
      return true;

  return false;
}

int main(int argc, char **argv)
{
  int i;
  bool usePyramid = true;
  bool useCube = false;
  bool scan = true;
  const char * renderManFile = 0;
  const char * mtlFile = 0;
  const char * geeFile = 0;
  const char * shotFiles = 0;
  uint renderManWidth = 320;
  uint renderManHeight = 256;
  float frameRate = 0.0; // ZERO = Unspecified

  std::vector<const char *> objFiles;

  clock_t t1 = clock();

  VEE_SceneBox *box = new VEE_SceneBox(0, VEE_Vector3(-100.0f,-100.0f,-100.0f)
                                        , VEE_Vector3(100.0f,100.0f,100.0f));

  if(argc == 1)
    defaultSystem(box);
  
  for(i=1; i < argc && scan; i++) {

    if(argv[i][0] != '-' || !strcmp("--done", argv[i])) {
      scan = false;
    }

    int j = i;

    const char * arg0 = argv[i];
    const char * arg1 = 0;
    const char * arg2 = 0;
    const char * arg3 = 0;
    const char * arg4 = 0;

    if((i + 1) < argc && scan) {
      if(!isNewCommand(argv[i+1]))
	arg1 = argv[++i];
    }
    if((i + 1) < argc && scan) {
      if(!isNewCommand(argv[i+1]))
	arg2 = argv[++i];
    }
    if((i + 1) < argc && scan) {
      if(!isNewCommand(argv[i+1]))
	arg3 = argv[++i];
    }
    if((i + 1) < argc && scan) {
      if(!isNewCommand(argv[i+1]))
	arg4 = argv[++i];
    }

    if(strcmp(arg0, "--damping") == 0) {
      float magnitude = 3.0;

      if(arg1) sscanf(arg1, "%f", & magnitude);

      addDamping(box, magnitude);
    }
    else if(strcmp(arg0, "--debug") == 0) {
      if(j + 1 >= argc || argv[j + 1][0] == '-') {
	GEE_Debug::setDebugAll(1);
	j++;
      }
      else {
	for(j++; 
	    j  < argc && argv[j][0] != '-'; 
	    j++) {
	  GEE_Debug::setDebug(argv[j]);
	}
      }	
      i = j;
    }
    else if(strcmp(arg0, "--default") == 0) {
      defaultSystem(box);
    }
    else if(strcmp(arg0, "--shotfile") == 0) {
      shotFiles = arg1;
    }
    else if(strcmp(arg0, "--gravity") == 0) {
      float force = 3.0;

      if(arg1) sscanf(arg1, "%f", & force);

      addGravity(box, force);
    }
    else if(strcmp(arg0, "--magicrain") == 0) {
      float rate = 10.0;
      float y = 15.0;

      if(arg1) sscanf(arg1, "%f", & rate);
      if(arg2) sscanf(arg2, "%f", & y);

      addRain(box, rate, y);
    }
	
    /* else if(strcmp(arg0, "--vortex") == 0) {
      float magnitude = 1.0;

      if(arg1) sscanf(arg1, "%f", & magnitude);

      addVortex(box, magnitude);
      }*/

    /* else if(strcmp(arg0, "--vortexeff") == 0) {
      float magnitude = 1.0;

      if(arg1) sscanf(arg1, "%f", & magnitude);

      addVortexeff(box, magnitude);
      }*/
    
    /* else if(strcmp(arg0, "--blow") == 0) {

      addBlow(box);
      }*/

    else if(strcmp(arg0, "--polygon") == 0) {
      int div = 6;
      float size = 2.0;

      if(arg1) sscanf(arg1, "%d", & div);
      if(arg2) sscanf(arg2, "%f", & size);

      addPolygon(box, div, size);
    }
    
    else if(strcmp(arg0, "--polygonfire") == 0) {
      float flameHz = 3.0;
      float flameAge = 5.0;
      float vortexRate = 3.0;
      float gustRate = 1.0;

      if(arg1) sscanf(arg1, "%f", & flameHz);
      if(arg2) sscanf(arg2, "%f", & flameAge);
      if(arg3) sscanf(arg3, "%f", & vortexRate);
      if(arg4) sscanf(arg4, "%f", & gustRate);

      addPolyFire(box, flameHz, flameAge, vortexRate, gustRate);
    }
    
    /* else if(strcmp(arg0, "--polygonclouds") == 0) {
      float cloudInterval = 3.0;
      float magnitude = 0.01;

      addPolyClouds(box, cloudInterval, magnitude);
      }*/
    
    else if(strcmp(arg0, "--nopyramid") == 0) {
      usePyramid = false;
    }
    
    else if(strcmp(arg0, "--wind") == 0) {
      float speed = 3.0;

      if(arg1) sscanf(arg1, "%f", & speed);

      addWind(box, speed);
    }
    else if(strcmp(arg0, "--framerate") == 0) {
      float rate = 25.0;

      if(arg1) sscanf(arg1, "%f", & rate);

      frameRate = rate;
    }
    else if(strcmp(arg0, "--renderman") == 0) {

      if(arg1) renderManFile = arg1;
      if(arg2) renderManWidth = atoi(arg2);
      if(arg3) renderManHeight = atoi(arg2);
    }
    else if(strcmp(arg0, "--done") == 0) {
      
      ; // Nothing to do
    }
    else if(strcmp(arg0, "--cube") == 0) {
      useCube = true;
    }
    else if(strcmp(arg0, "--creator2") == 0) {
      addCreator2(box);
    }
    else if(strcmp(arg0, "--gee") == 0) {
      geeFile = arg1;
      usePyramid = false;      
    }
    else if(strcmp(arg0, "--gravitates") == 0) {

      float rate = 2.0;
      float range = 2.0;
      float magnitude = arg3 ? atof(arg3) : 10.0;

      if(arg1) rate = atof(arg1);
      if(arg2) range = atof(arg2);

      addGravitates(box, rate, range, magnitude);
    }
    else if(strcmp(arg0, "--obj") == 0) {
      if(arg1) objFiles.push_back(arg1);
      if(arg2) objFiles.push_back(arg2);
      if(arg3) objFiles.push_back(arg3);
      if(arg4) objFiles.push_back(arg4);
    }
    else if(strcmp(arg0, "--mtl") == 0) {
      if(arg1) mtlFile = arg1;
    }
    else {
      fprintf(stderr, "\nUnrecognized option \"%s\"\n", arg0);
      fflush(0);
    }
    
  }

  QApplication * a = new QApplication( argc, argv );

  VEE_Qt *q = new VEE_Qt(box, VEE_GL_SL, 0, 0, false);

  VEE_TestQt * q2 = new 
    VEE_TestQt(q, box, renderManFile, renderManWidth, 
	       renderManHeight, VEE_GL_SL, q);

  q->setGlWidget(q2);

  q2->materials().addUsualColors();

  q->setFrameRate(frameRate);

  if(mtlFile)
    q2->setMtlFile(mtlFile);

  if(geeFile)
    q2->geeEngine()->read(geeFile);

  Solar::Model * model;
  Solar::RotatingObject *rotate;
  
  // Load a few objects -- just for fun !

  VEE_trace("Preparing to load %u 3D objects", (uint) objFiles.size());

  for(i = 0; i < (int) objFiles.size(); i++) {

    model = new Solar::Model;
    const char * filename = objFiles[i];

    VEE_trace("Trying to load file \"%s\"", filename);
    bool ok = Solar::Loaders::load(model, filename);

    if(!ok) {
      VEE_error("Failed to load file \"%s\"", filename);
      delete model;
      continue;
    }

    model->normalize(3.5, true);
    rotate = new Solar::RotatingObject;

    rotate->setOmega(0.7 + i * 0.6);
    rotate->setModel(model);
    rotate->getLocation()[1] = 1.0 + i * 2;

    q2->addModel(model);
    q2->addObject(rotate);
  }

  if(usePyramid) {
    model = new Solar::Model;

    model->createPyramid(2.5, "Pyramid");
  
    rotate = new Solar::RotatingObject;

    rotate->setOmega(-0.5);
    rotate->setModel(model);
    rotate->getLocation()[0] = 2.0;
    
    q2->addModel(model);
    q2->addObject(rotate);

    q2->materials().addMaterial("PyramidFace0", 
				Solar::GlMaterial(40, Solar::Rgba::Red));
  }

  if(useCube) {
    model = new Solar::Model;

    model->createCube(3.0, "Cube");
  
    rotate = new Solar::RotatingObject;

    rotate->setOmega(0.3);
    rotate->setModel(model);
    rotate->setLocation(VEE_Vector3(-4.4, -0.4, 0.0));
    // rotate->getLocation()[1] = -1.0;
    
    q2->addModel(model);
    q2->addObject(rotate);

    q2->materials().addMaterial("CubeFace0", 
			      Solar::GlMaterial(40, Solar::Rgba::Red));
    q2->materials().addMaterial("CubeFace1", 
			      Solar::GlMaterial(40, Solar::Rgba::Blue));
  }

  q->setGeometry( 100, 100, 640, 480 );

  a->setMainWidget( q );

  if(shotFiles) {
    VEE_trace("Storing images files as \"%s\"", shotFiles);
    q->setFrameFile(shotFiles);
  }

  q->show();

  int r = a->exec();

  ulong frames = q->frames();
  ulong modeChanges = q->renderer()->modeChanges();

  VEE_trace("%lu mode changes per frame", modeChanges / frames);

#ifndef VEE_FAST
  VEE_trace("%d particles in box", (int) box->totalChildren());
  VEE_trace("%d objects left", (int) VEE_Object::count());
  VEE_trace("%d scene boxes left", (int) VEE_SceneBox::count());

  {
    std::set<VEE_ForceObject *> forces;
    box->getForces(&forces);
    
    VEE_trace("%d forces left", (int) forces.size());
  }

  VEE_trace("Deleting all particles now");
  
  {
    VEE_SceneBox::ObjectStore objs;
    box->getObjects(&objs);
    
    for(VEE_SceneBox::iterator it = objs.begin(); it != objs.end(); it++)
      delete *it;
  }

  VEE_trace("Deleted objects");
  VEE_trace("%u boxes left", box->numChildBoxes());

  /* forces.clear();
  box->getForces(&forces);
  VEE_trace("%d forces left", (int) forces.size());*/

  delete box;
  box = 0;

  VEE_trace("%d objects left", (int) VEE_Object::count());
  VEE_trace("%d scene boxes left", (int) VEE_SceneBox::count());

#else
  
  box->write(& std::cout);

  std::cout << std::endl;

  VEE_trace("%u boxes left", box->numChildBoxes());
  delete box;
  box = 0;

#endif

  q2->deleteSolarObjectsAndModels();

  delete q;
  delete a;

  q = 0;
  a = 0;

  float t = (float) (clock()-t1) / (float) CLOCKS_PER_SEC;

  VEE_trace("\nCPU time = %.3f seconds", t);

  return r;
}



