/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <vee_test_qt.h>

#include <vee_gl_renderer.h>
#include <vee_gl_texture.h>
#include <vee_error.h>
#include <vee_image.h>

#ifdef VEE_WITH_RENDERMAN
#error "SUPPORT IS OBSOLOTE"
#include <vee_rman_renderer.h>
#include <ri.h>
#endif

#include <gee_geometry.h>
#include <gee_engine.h>
#include <gee_display_lists.h>
#include <gee_gl_material.h>
#include <gee_gl_renderer.h>

#include <Solar/GlRenderer.hpp>
#include <Solar/Model.hpp>
#include <Solar/MtlParser.hpp>
#include <Solar/Object.hpp>

#include <ConfigReader.hpp>

#include <stdio.h>

#ifdef VEE_WITH_RENDERMAN
static void VEE_RiLigths ()
{  
  RtPoint light0_pos={8.0,8.0,8.0};
  RtPoint light0_aim={0.0,0.0,0.0};
  RtFloat light0_intensity=2.0;
  RtColor light0_color={1.0,1.0,1.0};

  RtColor ambient_color={0.7,0.7,0.7};
  RtFloat ambient_intensity=0.6;

  RiAttributeBegin();

  RtLightHandle lh1 = RiLightSource(RI_AMBIENTLIGHT,RI_INTENSITY,
				    (RtPointer)&ambient_intensity,
				    RI_LIGHTCOLOR,ambient_color,RI_NULL);


  RiAttributeEnd();
  RiAttributeBegin();

  RtLightHandle lh2 = RiLightSource(RI_DISTANTLIGHT, 
				    RI_INTENSITY, (RtPointer)&light0_intensity,
				    RI_LIGHTCOLOR,light0_color,
				    RI_FROM,(RtPointer)&light0_pos,
				    RI_TO,(RtPointer)&light0_aim,
				    RI_NULL);

  RiAttributeEnd();

  RiAttributeBegin();
  RiTransformBegin();
  
  RtFloat tmpf = 0.0; // 1009500.0;
  RtColor colBall={1.0, 1.0, 1.0};  

  RtLightHandle lh3 = 
    RiAreaLightSource("arealight", 
		      "intensity", & tmpf, 
		      "lightcolor", & colBall, RI_NULL);
  
  RiTranslate(10.0, 0.0, 0.0);
  
  const float ballSize = 4.0;
  RiSphere(ballSize, -ballSize, ballSize, 360, RI_NULL);
  
  RiTransformEnd();
  RiAttributeEnd();

  RiIlluminate(lh1, 1);
  RiIlluminate(lh2, 1);
  RiIlluminate(lh3, 1);
}
#endif

VEE_TestQt::VEE_TestQt(VEE_Qt * host,
		       VEE_SceneBox *box, 
		       const char *rendermanFile,
		       uint width, 
		       uint height,
		       VEE_GlRendererType type,
		       QWidget *parent, 
		       const char *name)
  : VEE_QtHidden(host, box, parent, name, type),
    m_renderer(&m_materials)
{
  m_lastUpdate.getTimeOfDay();

  m_renderer.setDisplayLists(&m_lists);

  (void) width;
  (void) height;
  (void) rendermanFile;

  if(rendermanFile && strlen(rendermanFile)) {
#ifdef VEE_WITH_RENDERMAN
    m_rmanRenderer = new VEE_RManRenderer(1.0 / 25);
    RiBegin((char *) rendermanFile);
    // RiBegin(0);

    RiFormat(width, height, 0.0);
    // RiFormat(640, 496, 0.0);
    RiCropWindow(0.0, 1.0, 0.0, 1.0);
#else
    VEE_trace("VEE_TestQt::VEE_TestQt # WARNING # "
	      "No renderman support this time");
    m_rmanRenderer = 0;
#endif
  }
  else {
    m_rmanRenderer = 0;
  }
}

VEE_TestQt::~VEE_TestQt()
{
#ifdef VEE_WITH_RENDERMAN
  if(m_rmanRenderer) {
    delete m_rmanRenderer;
    RiEnd();
  }
#endif

  m_textures.clear();

  fflush(0);
}

void VEE_TestQt::deleteSolarObjectsAndModels()
{
  for(Solar::Model **m = &* m_models.begin(); m != &* m_models.end(); m++) 
    delete *m;
  for(Solar::Object **o = &* m_objects.begin(); o != &* m_objects.end(); o++) 
    delete *o;

  m_models.clear();
  m_objects.clear();
}

void VEE_TestQt::initializeGL()
{
  VEE_QtHidden::initializeGL();

  /* Set the light: */

  static GLfloat lightLocation[] = { 5.0, 5.0, 5.0, 0.0 };
  static GLfloat material[]  = { 1.0, 1.0, 1.0, 1.0 };
  static GLfloat shininess[] = { 127.0 };
  static GLfloat ambientLight[]  = { 0.3, 0.3, 0.3, 1.0 };
  static GLfloat diffuseLight[]  = { 0.7, 0.7, 0.7, 1.0 };
  static GLfloat specularLight[] = { 0.6, 0.6, 0.6, 1.0 };
  
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, material);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shininess);

  GLenum e = glGetError();
  if(e) 
    VEE_error("VEE_TestQt::initializeGL # GL ERROR 1 %s", gluErrorString(e));

  glLightfv (GL_LIGHT0, GL_AMBIENT, ambientLight);
  glLightfv (GL_LIGHT0, GL_DIFFUSE, diffuseLight);
  glLightfv (GL_LIGHT0, GL_SPECULAR, specularLight);
  glLightfv (GL_LIGHT0, GL_POSITION, lightLocation);

  e = glGetError();
  if(e) 
    VEE_error("VEE_TestQt::initializeGL # GL ERROR 2 %s", gluErrorString(e));

  glLightModelfv (GL_LIGHT_MODEL_AMBIENT, ambientLight);

  loadMaterials(m_mtlFile, & m_textures, & m_materials);

  
  e = glGetError();
  if(e) 
    VEE_error("VEE_TestQt::initializeGL # GL ERROR 3 %s", gluErrorString(e));
}


void VEE_TestQt::veeRender(bool clear)
{
#ifdef VEE_WITH_RENDERMAN
  if(m_rmanRenderer) {

    char framename[128];
    sprintf(framename, "frames/frame-%.4d.tif", m_frames);
    
    float fov = 80.0;

    // RiOrientation(RI_RH);

    RiProjection(RI_PERSPECTIVE, RI_FOV, & fov, RI_NULL);

    Fluid::Vector3 camPos(0.0, 0.0, 8.0);
    Fluid::Vector3 camDir(0.0, 0.3, -1.0);
    
    VEE_RManRenderer::placeCamera(camPos, camDir);
    // VEE_RManRenderer::placeCamera(m_veeGl->camPos(), m_veeGl->camDir());

    // RiIdentity();
    // RiRotate(180.0, 0.0, 1.0, 0.0);
    // RiTranslate(0.0, 6.0, 0.0);
    // RiTranslate(0.0, 0.0, 0.0);

    RiFrameBegin(m_frames);
    RiDisplay(framename,RI_FILE,RI_RGBA,RI_NULL);

    RiShutter(0.0, 1.0);
    RiWorldBegin();

    RiIdentity();

    // VEE_RiLigths();

    RtColor mcolor={.8,0.5,0.2};
    RiColor(mcolor);

    m_rmanRenderer->initRenderState();

    if(m_fixedRate > 0.0)
      m_rmanRenderer->setFrameDur(1.0 / m_fixedRate);

    m_rmanRenderer->paintRecursive(m_scene);
    
    RiWorldEnd();
    RiFrameEnd();  
  }
#endif

  // float distanceToOrigin = m_veeGl->camPos().length();

  GLenum e = glGetError();

  if(e) 
    VEE_error("VEE_TestQt::veeRender # GL ERROR %s", gluErrorString(e));
  
  float dt = m_lastFrameDuration;

  Solar::Object ** obj = &* m_objects.begin();
  Solar::Object ** sentinel = &* m_objects.end();
  
  glEnable (GL_LIGHT0);
  
  if(clear)
    m_veeGl->clearGlScreen();

  m_veeGl->moveCamera(m_lastFrameDuration);
  m_veeGl->moveGlCamera();

  m_renderer.initGlState();

  bool renderSolids = displaySolids();

  while(obj != sentinel && renderSolids) {
    (*obj)->update(dt);
    if(renderSolids)
      (*obj)->render(&m_renderer);
    obj++;
  }

  VEE_QtHidden::veeRender(false);
}

/** Load a few materials for the pyramid. */
void VEE_TestQt::loadMaterials(const std::string     &mtlFile, 
			       VEE_GlTexture2DBank   *textures,
			       Solar::GlMaterialBank *materials)
{
  const char *fname = "VEE_TestQt::loadMaterials";

  std::string path("images/");
  std::string base("PyramidFace");
  
  textures->loadTextures(path, base, ".jpg", 5);

  char buf[16];

  VEE_GlTexture2D *oldTex = 0;

  materials->addUsualColors();

  for(uint i=0; i < 5; i++) {

    sprintf(buf, "%s%d", base.c_str(), i);
    
    VEE_GlTexture2D *tex = textures->getTexture(buf);
    
    if(tex) {
	materials->addMaterial
	  (buf, Solar::GlMaterial(50, Solar::Rgba::White, tex->id()
				  /*,  Solar::GlMaterial::BLEND*/ ));
    }
    else if(oldTex) {
      materials->addMaterial
	(buf, Solar::GlMaterial(50, Solar::Rgba::White, oldTex->id()
				/*, Solar::GlMaterial::BLEND*/ ));
    }

    if(tex) oldTex = tex;
  }

  if(!mtlFile.size()) {
    VEE_trace("%s # No MTL file", fname);
    return;
  }

  ConfigReader::Config conf;

  bool ok = Solar::MtlParser::load(&conf, mtlFile.c_str());
  
  if(!ok) {
    VEE_error("%s # No materials", fname);
    return;
  }

  VEE_trace("%s # Loading %u materials", fname, conf.size());

  // conf.dump(std::cout);

  for(ConfigReader::Config::iterator it = conf.begin(); it != conf.end(); 
      it++) {
    
    Solar::GlMaterial mat; 

    ConfigReader::Chunk &chunk = ConfigReader::Config::getType(it);
    std::string name = ConfigReader::Config::getName(it);
    
    VEE_Rgba color = VEE_Rgba::White;

    // First we read color information:
    ConfigReader::Variant v = chunk.get("Ka");

    if(!v.isEmpty()) {
      ok = v.getFloats(color.data(), 3);
      if(!ok)
	VEE_error("%s # Could not get Ka", fname);
      else
	mat.setAmbient(color);
    }
    
    v = chunk.get("Kd");

    if(!v.isEmpty()) {
      ok = v.getFloats(color.data(), 3);
      if(!ok)
	VEE_error("%s # Could not get Kd", fname);
    }
    
    mat.setDiffuse(color);

    color = VEE_Rgba::Gray;

    v = chunk.get("Ks");

    if(!v.isEmpty()) {
      ok = v.getFloats(color.data(), 3);
      if(!ok)
	VEE_error("%s # Could not get Ks", fname);
    }
    
    mat.setSpecular(color);

    v = chunk.get("UVScale");

    if(!v.isEmpty()) {
      float scale = v.getDouble(1.0);
      mat.setUvScale(scale);
    }

    // VEE_trace("%s # Getting color [%s]", fname, v.getString().c_str());

    // Texture
    v = chunk.get("map_Ka");
    if(!v.isEmpty()) {
      textures->loadTexture(v, name);
    }
    
    float shine = chunk.get("Ns").getDouble(50.0) / 128.0;
    mat.setShine(shine);

    bool smooth   = chunk.get("flat").getInt(0) == 0;
    bool twoSided = chunk.get("twosided").getInt(0) == 1;
    
    VEE_GlTexture2D *tex = textures->getTexture(name);

    int flags = Solar::GlMaterial::LIGHTING;

    if(smooth)
      flags = flags | Solar::GlMaterial::SMOOTH;
    if(twoSided)
      flags = flags | Solar::GlMaterial::TWO_SIDED;

    VEE_trace("%s # Adding material \"%s\" %u %.2f %.2f %.2f %p %x", 
	      fname, name.c_str(), name.size(),
	      color.red(), color.green(), color.blue(), tex, flags);
    
    mat.setFlags(flags);
    mat.setTextureId(tex ? tex->id() : (GLuint) -1);
    materials->addMaterial(name, mat);
  }
}
