
/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <vee_calculus_thread.h>

#include <vee_error.h>
#include <vee_scene.h>

#define VEE_CALCULUS_SECOND -999

VEE_CalculusThread::VEE_CalculusThread()
  : m_mutex(false, false, false),
    m_scene(0),
    m_allowed(IDLE),
    m_current(IDLE),
    m_continue(true)
{}

VEE_CalculusThread::~VEE_CalculusThread()
{}

void VEE_CalculusThread::start(VEE_SceneBox * scene,
			       int calculusThreshold,
			       int beginBox,
			       int endBox,
			       float dt)
{
  m_scene = scene;
  m_calculusThreshold = calculusThreshold;
  m_beginBox = beginBox;
  m_endBox = endBox;
  m_dt = dt;

  run(true);
}
	     

void VEE_CalculusThread::doVelocityUpdate()
{
  m_allowed = VELOCITY;
  m_current = VELOCITY;
  m_cond.signal2(m_mutex);
}

void VEE_CalculusThread::waitVelocityUpdate()
{
  m_mutex.lock();
  while(m_current != VELOCITYDONE && m_continue)
    m_cond.wait(m_mutex);
  m_mutex.unlock();
}

void VEE_CalculusThread::doLocationUpdate()
{
  m_allowed = LOCATION;
  m_cond.signal2(m_mutex);
}

void VEE_CalculusThread::waitLocationUpdate()
{
  m_mutex.lock();
  while(m_current != LOCATIONDONE && m_continue)
    m_cond.wait(m_mutex);
  m_mutex.unlock();
}
  
void VEE_CalculusThread::stop()
{
  m_continue = false;
  m_cond.signal2(m_mutex);  
}

void VEE_CalculusThread::childLoop()
{
  const char *fname = "VEE_CalculusThread::childLoop";

  VEE_trace("%snthreads # %p ENTER", fname, this);
  
  while(m_continue) {
        
    // VEE_trace("%s # LOOP %p", fname, this);

    // Wait to get permission to execute
    m_mutex.lock();
    while(m_allowed != VELOCITY && m_continue)
      m_cond.wait(m_mutex);
    m_mutex.unlock();
    
    m_current = VELOCITY;
    m_cond.signal2(m_mutex);

    if(m_continue) {
      
      if(m_calculusThreshold != VEE_CALCULUS_SECOND)
	m_scene->updateVelocitiesPart
	  (m_calculusThreshold, m_beginBox, m_endBox, m_dt);
      else
	m_scene->updateVelocitiesPart2(m_beginBox, m_endBox, m_dt);
    }
    
    m_allowed = IDLE;
    m_current = VELOCITYDONE;
    m_cond.signal2(m_mutex);

    // Wait to get permission to execute
    /* m_mutex.lock();
    while(m_allowed != LOCATION && m_continue)
      m_cond.wait(m_mutex);
      m_mutex.unlock();
    
    m_current = LOCATION;
    m_cond.signal2(m_mutex);*/

    /* if(m_continue) {
      m_scene->updateLocationsPart
	(m_calculusThreshold, m_beginBox, m_endBox, m_dt);
	}*/
    
    /* m_current = LOCATIONDONE;
       m_cond.signal2(m_mutex); */
  }  

  VEE_trace("%s # EXIT", fname);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


VEE_CalculusThreads::VEE_CalculusThreads()
  : m_splitAt(40),
    m_collapseAt(20),
    m_maximumRecursion(6),
    m_method(0),
    m_dt(0.1)
{}

VEE_CalculusThreads::~VEE_CalculusThreads()
{
}

void VEE_CalculusThreads::addThread(int calculusThreshold,
				    int beginBox,
				    int endBox)
{
  if(!m_scene)
    VEE_fatal("VEE_CalculusThreads::addThread # Set scene before spanning threads !!");

  VEE_CalculusThread * calc = new VEE_CalculusThread;

  calc->start(m_scene, calculusThreshold, beginBox, endBox, 0.1);

  m_threads.push_back(calc);
}

void VEE_CalculusThreads::setDt(float dt)
{
  uint i;

  m_dt = dt;

  for(i=0; i < m_threads.size(); i++) {
    m_threads[i]->setDt(dt);
  }
}

void VEE_CalculusThreads::update()
{
  uint i;

  if(!m_scene)
    VEE_fatal("VEE_CalculusThreads::update # Set scene before this !!");

  for(i=0; i < m_threads.size(); i++) {
    m_threads[i]->doVelocityUpdate();
  }
  for(i=0; i < m_threads.size(); i++) {
    m_threads[i]->waitVelocityUpdate();
  }

  /* for(i=0; i < m_threads.size(); i++) {
    m_threads[i]->doLocationUpdate();
    }*/

  m_scene->updateR(m_dt);

  /* for(i=0; i < m_threads.size(); i++) {
    m_threads[i]->waitLocationUpdate();
    }*/

  m_scene->updateScene();
  m_scene->updateTree(m_splitAt, m_collapseAt, m_maximumRecursion);
  
  
}

void VEE_CalculusThreads::stop()
{
  uint i;

  for(i=0; i < m_threads.size(); i++) {
    m_threads[i]->stop();
  }
  
  for(i=0; i < m_threads.size(); i++) {
    m_threads[i]->waitEnd();
  }
  
}

/** Create 5 threads to do calculus. Great for SMP machines that can
    spare 5 CPUs for this :-) */

void VEE_CalculusThreads::init5(int calculusThreshold, 
				/** Use the linear splitting method
				    for distributing computational
				    load. */
				bool linearSplit)
{
  if(linearSplit) {

    int nthreads = 8;
    
    for(int i=0; i < nthreads; i++) {
      VEE_CalculusThread * calc = new VEE_CalculusThread;
      calc->start(m_scene, VEE_CALCULUS_SECOND, nthreads, i, 0.1);
      m_threads.push_back(calc);
    }

    m_method = 1;
    m_maximumRecursion = 0;
  }
  else {
    addThread(calculusThreshold, 0, 1);
    addThread(calculusThreshold, 2, 3);
    addThread(calculusThreshold, 4, 5);
    addThread(calculusThreshold, 6, 7);
    
    addThread(calculusThreshold, -1, -1);

    m_method = 0;
  }
}

